(window["webpackJsonp"]=window["webpackJsonp"]||[]).push([["chunk-vendors"],{"00ee":function(t,e,n){var i=n("b622"),r=i("toStringTag"),o={};o[r]="z",t.exports="[object z]"===String(o)},"00fd":function(t,e,n){var i=n("9e69"),r=Object.prototype,o=r.hasOwnProperty,a=r.toString,s=i?i.toStringTag:void 0;function l(t){var e=o.call(t,s),n=t[s];try{t[s]=void 0;var i=!0}catch(l){}var r=a.call(t);return i&&(e?t[s]=n:delete t[s]),r}t.exports=l},"010e":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("uz-latn",{months:"Yanvar_Fevral_Mart_Aprel_May_Iyun_Iyul_Avgust_Sentabr_Oktabr_Noyabr_Dekabr".split("_"),monthsShort:"Yan_Fev_Mar_Apr_May_Iyun_Iyul_Avg_Sen_Okt_Noy_Dek".split("_"),weekdays:"Yakshanba_Dushanba_Seshanba_Chorshanba_Payshanba_Juma_Shanba".split("_"),weekdaysShort:"Yak_Dush_Sesh_Chor_Pay_Jum_Shan".split("_"),weekdaysMin:"Ya_Du_Se_Cho_Pa_Ju_Sha".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"D MMMM YYYY, dddd HH:mm"},calendar:{sameDay:"[Bugun soat] LT [da]",nextDay:"[Ertaga] LT [da]",nextWeek:"dddd [kuni soat] LT [da]",lastDay:"[Kecha soat] LT [da]",lastWeek:"[O'tgan] dddd [kuni soat] LT [da]",sameElse:"L"},relativeTime:{future:"Yaqin %s ichida",past:"Bir necha %s oldin",s:"soniya",ss:"%d soniya",m:"bir daqiqa",mm:"%d daqiqa",h:"bir soat",hh:"%d soat",d:"bir kun",dd:"%d kun",M:"bir oy",MM:"%d oy",y:"bir yil",yy:"%d yil"},week:{dow:1,doy:7}});return e}))},"020f":function(t,e,n){var i=n("242e"),r=n("1304");function o(t,e){return t&&i(t,r(e))}t.exports=o},"02fb":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ml",{months:"ജനുവരി_ഫെബ്രുവരി_മാർച്ച്_ഏപ്രിൽ_മേയ്_ജൂൺ_ജൂലൈ_ഓഗസ്റ്റ്_സെപ്റ്റംബർ_ഒക്ടോബർ_നവംബർ_ഡിസംബർ".split("_"),monthsShort:"ജനു._ഫെബ്രു._മാർ._ഏപ്രി._മേയ്_ജൂൺ_ജൂലൈ._ഓഗ._സെപ്റ്റ._ഒക്ടോ._നവം._ഡിസം.".split("_"),monthsParseExact:!0,weekdays:"ഞായറാഴ്ച_തിങ്കളാഴ്ച_ചൊവ്വാഴ്ച_ബുധനാഴ്ച_വ്യാഴാഴ്ച_വെള്ളിയാഴ്ച_ശനിയാഴ്ച".split("_"),weekdaysShort:"ഞായർ_തിങ്കൾ_ചൊവ്വ_ബുധൻ_വ്യാഴം_വെള്ളി_ശനി".split("_"),weekdaysMin:"ഞാ_തി_ചൊ_ബു_വ്യാ_വെ_ശ".split("_"),longDateFormat:{LT:"A h:mm -നു",LTS:"A h:mm:ss -നു",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm -നു",LLLL:"dddd, D MMMM YYYY, A h:mm -നു"},calendar:{sameDay:"[ഇന്ന്] LT",nextDay:"[നാളെ] LT",nextWeek:"dddd, LT",lastDay:"[ഇന്നലെ] LT",lastWeek:"[കഴിഞ്ഞ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s കഴിഞ്ഞ്",past:"%s മുൻപ്",s:"അൽപ നിമിഷങ്ങൾ",ss:"%d സെക്കൻഡ്",m:"ഒരു മിനിറ്റ്",mm:"%d മിനിറ്റ്",h:"ഒരു മണിക്കൂർ",hh:"%d മണിക്കൂർ",d:"ഒരു ദിവസം",dd:"%d ദിവസം",M:"ഒരു മാസം",MM:"%d മാസം",y:"ഒരു വർഷം",yy:"%d വർഷം"},meridiemParse:/രാത്രി|രാവിലെ|ഉച്ച കഴിഞ്ഞ്|വൈകുന്നേരം|രാത്രി/i,meridiemHour:function(t,e){return 12===t&&(t=0),"രാത്രി"===e&&t>=4||"ഉച്ച കഴിഞ്ഞ്"===e||"വൈകുന്നേരം"===e?t+12:t},meridiem:function(t,e,n){return t<4?"രാത്രി":t<12?"രാവിലെ":t<17?"ഉച്ച കഴിഞ്ഞ്":t<20?"വൈകുന്നേരം":"രാത്രി"}});return e}))},"0366":function(t,e,n){var i=n("1c0b");t.exports=function(t,e,n){if(i(t),void 0===e)return t;switch(n){case 0:return function(){return t.call(e)};case 1:return function(n){return t.call(e,n)};case 2:return function(n,i){return t.call(e,n,i)};case 3:return function(n,i,r){return t.call(e,n,i,r)}}return function(){return t.apply(e,arguments)}}},"03dd":function(t,e,n){var i=n("eac5"),r=n("57a5"),o=Object.prototype,a=o.hasOwnProperty;function s(t){if(!i(t))return r(t);var e=[];for(var n in Object(t))a.call(t,n)&&"constructor"!=n&&e.push(n);return e}t.exports=s},"03ec":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("cv",{months:"кӑрлач_нарӑс_пуш_ака_май_ҫӗртме_утӑ_ҫурла_авӑн_юпа_чӳк_раштав".split("_"),monthsShort:"кӑр_нар_пуш_ака_май_ҫӗр_утӑ_ҫур_авн_юпа_чӳк_раш".split("_"),weekdays:"вырсарникун_тунтикун_ытларикун_юнкун_кӗҫнерникун_эрнекун_шӑматкун".split("_"),weekdaysShort:"выр_тун_ытл_юн_кӗҫ_эрн_шӑм".split("_"),weekdaysMin:"вр_тн_ыт_юн_кҫ_эр_шм".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD-MM-YYYY",LL:"YYYY [ҫулхи] MMMM [уйӑхӗн] D[-мӗшӗ]",LLL:"YYYY [ҫулхи] MMMM [уйӑхӗн] D[-мӗшӗ], HH:mm",LLLL:"dddd, YYYY [ҫулхи] MMMM [уйӑхӗн] D[-мӗшӗ], HH:mm"},calendar:{sameDay:"[Паян] LT [сехетре]",nextDay:"[Ыран] LT [сехетре]",lastDay:"[Ӗнер] LT [сехетре]",nextWeek:"[Ҫитес] dddd LT [сехетре]",lastWeek:"[Иртнӗ] dddd LT [сехетре]",sameElse:"L"},relativeTime:{future:function(t){var e=/сехет$/i.exec(t)?"рен":/ҫул$/i.exec(t)?"тан":"ран";return t+e},past:"%s каялла",s:"пӗр-ик ҫеккунт",ss:"%d ҫеккунт",m:"пӗр минут",mm:"%d минут",h:"пӗр сехет",hh:"%d сехет",d:"пӗр кун",dd:"%d кун",M:"пӗр уйӑх",MM:"%d уйӑх",y:"пӗр ҫул",yy:"%d ҫул"},dayOfMonthOrdinalParse:/\d{1,2}-мӗш/,ordinal:"%d-мӗш",week:{dow:1,doy:7}});return e}))},"0558":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t){return t%100===11||t%10!==1}function n(t,n,i,r){var o=t+" ";switch(i){case"s":return n||r?"nokkrar sekúndur":"nokkrum sekúndum";case"ss":return e(t)?o+(n||r?"sekúndur":"sekúndum"):o+"sekúnda";case"m":return n?"mínúta":"mínútu";case"mm":return e(t)?o+(n||r?"mínútur":"mínútum"):n?o+"mínúta":o+"mínútu";case"hh":return e(t)?o+(n||r?"klukkustundir":"klukkustundum"):o+"klukkustund";case"d":return n?"dagur":r?"dag":"degi";case"dd":return e(t)?n?o+"dagar":o+(r?"daga":"dögum"):n?o+"dagur":o+(r?"dag":"degi");case"M":return n?"mánuður":r?"mánuð":"mánuði";case"MM":return e(t)?n?o+"mánuðir":o+(r?"mánuði":"mánuðum"):n?o+"mánuður":o+(r?"mánuð":"mánuði");case"y":return n||r?"ár":"ári";case"yy":return e(t)?o+(n||r?"ár":"árum"):o+(n||r?"ár":"ári")}}var i=t.defineLocale("is",{months:"janúar_febrúar_mars_apríl_maí_júní_júlí_ágúst_september_október_nóvember_desember".split("_"),monthsShort:"jan_feb_mar_apr_maí_jún_júl_ágú_sep_okt_nóv_des".split("_"),weekdays:"sunnudagur_mánudagur_þriðjudagur_miðvikudagur_fimmtudagur_föstudagur_laugardagur".split("_"),weekdaysShort:"sun_mán_þri_mið_fim_fös_lau".split("_"),weekdaysMin:"Su_Má_Þr_Mi_Fi_Fö_La".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY [kl.] H:mm",LLLL:"dddd, D. MMMM YYYY [kl.] H:mm"},calendar:{sameDay:"[í dag kl.] LT",nextDay:"[á morgun kl.] LT",nextWeek:"dddd [kl.] LT",lastDay:"[í gær kl.] LT",lastWeek:"[síðasta] dddd [kl.] LT",sameElse:"L"},relativeTime:{future:"eftir %s",past:"fyrir %s síðan",s:n,ss:n,m:n,mm:n,h:"klukkustund",hh:n,d:n,dd:n,M:n,MM:n,y:n,yy:n},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return i}))},"057f":function(t,e,n){var i=n("fc6a"),r=n("241c").f,o={}.toString,a="object"==typeof window&&window&&Object.getOwnPropertyNames?Object.getOwnPropertyNames(window):[],s=function(t){try{return r(t)}catch(e){return a.slice()}};t.exports.f=function(t){return a&&"[object Window]"==o.call(t)?s(t):r(i(t))}},"0621":function(t,e,n){var i=n("9e69"),r=n("d370"),o=n("6747"),a=i?i.isConcatSpreadable:void 0;function s(t){return o(t)||r(t)||!!(a&&t&&t[a])}t.exports=s},"06c5":function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));n("a630"),n("fb6a"),n("b0c0"),n("d3b7"),n("25f0"),n("3ca3");var i=n("6b75");function r(t,e){if(t){if("string"===typeof t)return Object(i["a"])(t,e);var n=Object.prototype.toString.call(t).slice(8,-1);return"Object"===n&&t.constructor&&(n=t.constructor.name),"Map"===n||"Set"===n?Array.from(t):"Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)?Object(i["a"])(t,e):void 0}}},"06cf":function(t,e,n){var i=n("83ab"),r=n("d1e7"),o=n("5c6c"),a=n("fc6a"),s=n("c04e"),l=n("5135"),u=n("0cfb"),c=Object.getOwnPropertyDescriptor;e.f=i?c:function(t,e){if(t=a(t),e=s(e,!0),u)try{return c(t,e)}catch(n){}if(l(t,e))return o(!r.f.call(t,e),t[e])}},"0721":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fo",{months:"januar_februar_mars_apríl_mai_juni_juli_august_september_oktober_november_desember".split("_"),monthsShort:"jan_feb_mar_apr_mai_jun_jul_aug_sep_okt_nov_des".split("_"),weekdays:"sunnudagur_mánadagur_týsdagur_mikudagur_hósdagur_fríggjadagur_leygardagur".split("_"),weekdaysShort:"sun_mán_týs_mik_hós_frí_ley".split("_"),weekdaysMin:"su_má_tý_mi_hó_fr_le".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D. MMMM, YYYY HH:mm"},calendar:{sameDay:"[Í dag kl.] LT",nextDay:"[Í morgin kl.] LT",nextWeek:"dddd [kl.] LT",lastDay:"[Í gjár kl.] LT",lastWeek:"[síðstu] dddd [kl] LT",sameElse:"L"},relativeTime:{future:"um %s",past:"%s síðani",s:"fá sekund",ss:"%d sekundir",m:"ein minuttur",mm:"%d minuttir",h:"ein tími",hh:"%d tímar",d:"ein dagur",dd:"%d dagar",M:"ein mánaður",MM:"%d mánaðir",y:"eitt ár",yy:"%d ár"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},"079e":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ja",{eras:[{since:"2019-05-01",offset:1,name:"令和",narrow:"㋿",abbr:"R"},{since:"1989-01-08",until:"2019-04-30",offset:1,name:"平成",narrow:"㍻",abbr:"H"},{since:"1926-12-25",until:"1989-01-07",offset:1,name:"昭和",narrow:"㍼",abbr:"S"},{since:"1912-07-30",until:"1926-12-24",offset:1,name:"大正",narrow:"㍽",abbr:"T"},{since:"1873-01-01",until:"1912-07-29",offset:6,name:"明治",narrow:"㍾",abbr:"M"},{since:"0001-01-01",until:"1873-12-31",offset:1,name:"西暦",narrow:"AD",abbr:"AD"},{since:"0000-12-31",until:-1/0,offset:1,name:"紀元前",narrow:"BC",abbr:"BC"}],eraYearOrdinalRegex:/(元|\d+)年/,eraYearOrdinalParse:function(t,e){return"元"===e[1]?1:parseInt(e[1]||t,10)},months:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"日曜日_月曜日_火曜日_水曜日_木曜日_金曜日_土曜日".split("_"),weekdaysShort:"日_月_火_水_木_金_土".split("_"),weekdaysMin:"日_月_火_水_木_金_土".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日 dddd HH:mm",l:"YYYY/MM/DD",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日(ddd) HH:mm"},meridiemParse:/午前|午後/i,isPM:function(t){return"午後"===t},meridiem:function(t,e,n){return t<12?"午前":"午後"},calendar:{sameDay:"[今日] LT",nextDay:"[明日] LT",nextWeek:function(t){return t.week()!==this.week()?"[来週]dddd LT":"dddd LT"},lastDay:"[昨日] LT",lastWeek:function(t){return this.week()!==t.week()?"[先週]dddd LT":"dddd LT"},sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}日/,ordinal:function(t,e){switch(e){case"y":return 1===t?"元年":t+"年";case"d":case"D":case"DDD":return t+"日";default:return t}},relativeTime:{future:"%s後",past:"%s前",s:"数秒",ss:"%d秒",m:"1分",mm:"%d分",h:"1時間",hh:"%d時間",d:"1日",dd:"%d日",M:"1ヶ月",MM:"%dヶ月",y:"1年",yy:"%d年"}});return e}))},"07ac":function(t,e,n){var i=n("23e7"),r=n("6f53").values;i({target:"Object",stat:!0},{values:function(t){return r(t)}})},"07c7":function(t,e){function n(){return!1}t.exports=n},"087d":function(t,e){function n(t,e){var n=-1,i=e.length,r=t.length;while(++n<i)t[r+n]=e[n];return t}t.exports=n},"08cc":function(t,e,n){var i=n("1a8c");function r(t){return t===t&&!i(t)}t.exports=r},"0977":function(module,__webpack_exports__,__webpack_require__){"use strict";(function(global){var axios__WEBPACK_IMPORTED_MODULE_0__=__webpack_require__("bc3a"),axios__WEBPACK_IMPORTED_MODULE_0___default=__webpack_require__.n(axios__WEBPACK_IMPORTED_MODULE_0__),version="1.2.8";function _typeof(t){return _typeof="function"===typeof Symbol&&"symbol"===typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"===typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},_typeof(t)}function _classCallCheck(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")}function _defineProperties(t,e){for(var n=0;n<e.length;n++){var i=e[n];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,i.key,i)}}function _createClass(t,e,n){return e&&_defineProperties(t.prototype,e),n&&_defineProperties(t,n),t}function _defineProperty(t,e,n){return e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t}function _inherits(t,e){if("function"!==typeof e&&null!==e)throw new TypeError("Super expression must either be null or a function");t.prototype=Object.create(e&&e.prototype,{constructor:{value:t,writable:!0,configurable:!0}}),e&&_setPrototypeOf(t,e)}function _getPrototypeOf(t){return _getPrototypeOf=Object.setPrototypeOf?Object.getPrototypeOf:function(t){return t.__proto__||Object.getPrototypeOf(t)},_getPrototypeOf(t)}function _setPrototypeOf(t,e){return _setPrototypeOf=Object.setPrototypeOf||function(t,e){return t.__proto__=e,t},_setPrototypeOf(t,e)}function _assertThisInitialized(t){if(void 0===t)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return t}function _possibleConstructorReturn(t,e){return!e||"object"!==typeof e&&"function"!==typeof e?_assertThisInitialized(t):e}function _superPropBase(t,e){while(!Object.prototype.hasOwnProperty.call(t,e))if(t=_getPrototypeOf(t),null===t)break;return t}function _get(t,e,n){return _get="undefined"!==typeof Reflect&&Reflect.get?Reflect.get:function(t,e,n){var i=_superPropBase(t,e);if(i){var r=Object.getOwnPropertyDescriptor(i,e);return r.get?r.get.call(n):r.value}},_get(t,e,n||t)}function _slicedToArray(t,e){return _arrayWithHoles(t)||_iterableToArrayLimit(t,e)||_nonIterableRest()}function _toConsumableArray(t){return _arrayWithoutHoles(t)||_iterableToArray(t)||_nonIterableSpread()}function _arrayWithoutHoles(t){if(Array.isArray(t)){for(var e=0,n=new Array(t.length);e<t.length;e++)n[e]=t[e];return n}}function _arrayWithHoles(t){if(Array.isArray(t))return t}function _iterableToArray(t){if(Symbol.iterator in Object(t)||"[object Arguments]"===Object.prototype.toString.call(t))return Array.from(t)}function _iterableToArrayLimit(t,e){if(Symbol.iterator in Object(t)||"[object Arguments]"===Object.prototype.toString.call(t)){var n=[],i=!0,r=!1,o=void 0;try{for(var a,s=t[Symbol.iterator]();!(i=(a=s.next()).done);i=!0)if(n.push(a.value),e&&n.length===e)break}catch(l){r=!0,o=l}finally{try{i||null==s["return"]||s["return"]()}finally{if(r)throw o}}return n}}function _nonIterableSpread(){throw new TypeError("Invalid attempt to spread non-iterable instance")}function _nonIterableRest(){throw new TypeError("Invalid attempt to destructure non-iterable instance")}var commonjsGlobal="undefined"!==typeof globalThis?globalThis:"undefined"!==typeof window?window:"undefined"!==typeof global?global:"undefined"!==typeof self?self:{};function createCommonjsModule(t,e){return e={exports:{}},t(e,e.exports),e.exports}var check=function(t){return t&&t.Math==Math&&t},global_1=check("object"==typeof globalThis&&globalThis)||check("object"==typeof window&&window)||check("object"==typeof self&&self)||check("object"==typeof commonjsGlobal&&commonjsGlobal)||Function("return this")(),fails=function(t){try{return!!t()}catch(e){return!0}},descriptors=!fails((function(){return 7!=Object.defineProperty({},"a",{get:function(){return 7}}).a})),nativePropertyIsEnumerable={}.propertyIsEnumerable,getOwnPropertyDescriptor=Object.getOwnPropertyDescriptor,NASHORN_BUG=getOwnPropertyDescriptor&&!nativePropertyIsEnumerable.call({1:2},1),f=NASHORN_BUG?function(t){var e=getOwnPropertyDescriptor(this,t);return!!e&&e.enumerable}:nativePropertyIsEnumerable,objectPropertyIsEnumerable={f:f},createPropertyDescriptor=function(t,e){return{enumerable:!(1&t),configurable:!(2&t),writable:!(4&t),value:e}},toString={}.toString,classofRaw=function(t){return toString.call(t).slice(8,-1)},split="".split,indexedObject=fails((function(){return!Object("z").propertyIsEnumerable(0)}))?function(t){return"String"==classofRaw(t)?split.call(t,""):Object(t)}:Object,requireObjectCoercible=function(t){if(void 0==t)throw TypeError("Can't call method on "+t);return t},toIndexedObject=function(t){return indexedObject(requireObjectCoercible(t))},isObject=function(t){return"object"===typeof t?null!==t:"function"===typeof t},toPrimitive=function(t,e){if(!isObject(t))return t;var n,i;if(e&&"function"==typeof(n=t.toString)&&!isObject(i=n.call(t)))return i;if("function"==typeof(n=t.valueOf)&&!isObject(i=n.call(t)))return i;if(!e&&"function"==typeof(n=t.toString)&&!isObject(i=n.call(t)))return i;throw TypeError("Can't convert object to primitive value")},hasOwnProperty={}.hasOwnProperty,has=function(t,e){return hasOwnProperty.call(t,e)},document=global_1.document,EXISTS=isObject(document)&&isObject(document.createElement),documentCreateElement=function(t){return EXISTS?document.createElement(t):{}},ie8DomDefine=!descriptors&&!fails((function(){return 7!=Object.defineProperty(documentCreateElement("div"),"a",{get:function(){return 7}}).a})),nativeGetOwnPropertyDescriptor=Object.getOwnPropertyDescriptor,f$1=descriptors?nativeGetOwnPropertyDescriptor:function(t,e){if(t=toIndexedObject(t),e=toPrimitive(e,!0),ie8DomDefine)try{return nativeGetOwnPropertyDescriptor(t,e)}catch(n){}if(has(t,e))return createPropertyDescriptor(!objectPropertyIsEnumerable.f.call(t,e),t[e])},objectGetOwnPropertyDescriptor={f:f$1},anObject=function(t){if(!isObject(t))throw TypeError(String(t)+" is not an object");return t},nativeDefineProperty=Object.defineProperty,f$2=descriptors?nativeDefineProperty:function(t,e,n){if(anObject(t),e=toPrimitive(e,!0),anObject(n),ie8DomDefine)try{return nativeDefineProperty(t,e,n)}catch(i){}if("get"in n||"set"in n)throw TypeError("Accessors not supported");return"value"in n&&(t[e]=n.value),t},objectDefineProperty={f:f$2},createNonEnumerableProperty=descriptors?function(t,e,n){return objectDefineProperty.f(t,e,createPropertyDescriptor(1,n))}:function(t,e,n){return t[e]=n,t},setGlobal=function(t,e){try{createNonEnumerableProperty(global_1,t,e)}catch(n){global_1[t]=e}return e},SHARED="__core-js_shared__",store=global_1[SHARED]||setGlobal(SHARED,{}),sharedStore=store,shared=createCommonjsModule((function(t){(t.exports=function(t,e){return sharedStore[t]||(sharedStore[t]=void 0!==e?e:{})})("versions",[]).push({version:"3.4.4",mode:"global",copyright:"© 2019 Denis Pushkarev (zloirock.ru)"})})),functionToString=shared("native-function-to-string",Function.toString),WeakMap=global_1.WeakMap,nativeWeakMap="function"===typeof WeakMap&&/native code/.test(functionToString.call(WeakMap)),id=0,postfix=Math.random(),uid=function(t){return"Symbol("+String(void 0===t?"":t)+")_"+(++id+postfix).toString(36)},keys=shared("keys"),sharedKey=function(t){return keys[t]||(keys[t]=uid(t))},hiddenKeys={},WeakMap$1=global_1.WeakMap,set,get,has$1,enforce=function(t){return has$1(t)?get(t):set(t,{})},getterFor=function(t){return function(e){var n;if(!isObject(e)||(n=get(e)).type!==t)throw TypeError("Incompatible receiver, "+t+" required");return n}};if(nativeWeakMap){var store$1=new WeakMap$1,wmget=store$1.get,wmhas=store$1.has,wmset=store$1.set;set=function(t,e){return wmset.call(store$1,t,e),e},get=function(t){return wmget.call(store$1,t)||{}},has$1=function(t){return wmhas.call(store$1,t)}}else{var STATE=sharedKey("state");hiddenKeys[STATE]=!0,set=function(t,e){return createNonEnumerableProperty(t,STATE,e),e},get=function(t){return has(t,STATE)?t[STATE]:{}},has$1=function(t){return has(t,STATE)}}var internalState={set:set,get:get,has:has$1,enforce:enforce,getterFor:getterFor},redefine=createCommonjsModule((function(t){var e=internalState.get,n=internalState.enforce,i=String(functionToString).split("toString");shared("inspectSource",(function(t){return functionToString.call(t)})),(t.exports=function(t,e,r,o){var a=!!o&&!!o.unsafe,s=!!o&&!!o.enumerable,l=!!o&&!!o.noTargetGet;"function"==typeof r&&("string"!=typeof e||has(r,"name")||createNonEnumerableProperty(r,"name",e),n(r).source=i.join("string"==typeof e?e:"")),t!==global_1?(a?!l&&t[e]&&(s=!0):delete t[e],s?t[e]=r:createNonEnumerableProperty(t,e,r)):s?t[e]=r:setGlobal(e,r)})(Function.prototype,"toString",(function(){return"function"==typeof this&&e(this).source||functionToString.call(this)}))})),path=global_1,aFunction=function(t){return"function"==typeof t?t:void 0},getBuiltIn=function(t,e){return arguments.length<2?aFunction(path[t])||aFunction(global_1[t]):path[t]&&path[t][e]||global_1[t]&&global_1[t][e]},ceil=Math.ceil,floor=Math.floor,toInteger=function(t){return isNaN(t=+t)?0:(t>0?floor:ceil)(t)},min=Math.min,toLength=function(t){return t>0?min(toInteger(t),9007199254740991):0},max=Math.max,min$1=Math.min,toAbsoluteIndex=function(t,e){var n=toInteger(t);return n<0?max(n+e,0):min$1(n,e)},createMethod=function(t){return function(e,n,i){var r,o=toIndexedObject(e),a=toLength(o.length),s=toAbsoluteIndex(i,a);if(t&&n!=n){while(a>s)if(r=o[s++],r!=r)return!0}else for(;a>s;s++)if((t||s in o)&&o[s]===n)return t||s||0;return!t&&-1}},arrayIncludes={includes:createMethod(!0),indexOf:createMethod(!1)},indexOf=arrayIncludes.indexOf,objectKeysInternal=function(t,e){var n,i=toIndexedObject(t),r=0,o=[];for(n in i)!has(hiddenKeys,n)&&has(i,n)&&o.push(n);while(e.length>r)has(i,n=e[r++])&&(~indexOf(o,n)||o.push(n));return o},enumBugKeys=["constructor","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","toLocaleString","toString","valueOf"],hiddenKeys$1=enumBugKeys.concat("length","prototype"),f$3=Object.getOwnPropertyNames||function(t){return objectKeysInternal(t,hiddenKeys$1)},objectGetOwnPropertyNames={f:f$3},f$4=Object.getOwnPropertySymbols,objectGetOwnPropertySymbols={f:f$4},ownKeys=getBuiltIn("Reflect","ownKeys")||function(t){var e=objectGetOwnPropertyNames.f(anObject(t)),n=objectGetOwnPropertySymbols.f;return n?e.concat(n(t)):e},copyConstructorProperties=function(t,e){for(var n=ownKeys(e),i=objectDefineProperty.f,r=objectGetOwnPropertyDescriptor.f,o=0;o<n.length;o++){var a=n[o];has(t,a)||i(t,a,r(e,a))}},replacement=/#|\.prototype\./,isForced=function(t,e){var n=data[normalize(t)];return n==POLYFILL||n!=NATIVE&&("function"==typeof e?fails(e):!!e)},normalize=isForced.normalize=function(t){return String(t).replace(replacement,".").toLowerCase()},data=isForced.data={},NATIVE=isForced.NATIVE="N",POLYFILL=isForced.POLYFILL="P",isForced_1=isForced,getOwnPropertyDescriptor$1=objectGetOwnPropertyDescriptor.f,_export=function(t,e){var n,i,r,o,a,s,l=t.target,u=t.global,c=t.stat;if(i=u?global_1:c?global_1[l]||setGlobal(l,{}):(global_1[l]||{}).prototype,i)for(r in e){if(a=e[r],t.noTargetGet?(s=getOwnPropertyDescriptor$1(i,r),o=s&&s.value):o=i[r],n=isForced_1(u?r:l+(c?".":"#")+r,t.forced),!n&&void 0!==o){if(typeof a===typeof o)continue;copyConstructorProperties(a,o)}(t.sham||o&&o.sham)&&createNonEnumerableProperty(a,"sham",!0),redefine(i,r,a,t)}},aFunction$1=function(t){if("function"!=typeof t)throw TypeError(String(t)+" is not a function");return t},bindContext=function(t,e,n){if(aFunction$1(t),void 0===e)return t;switch(n){case 0:return function(){return t.call(e)};case 1:return function(n){return t.call(e,n)};case 2:return function(n,i){return t.call(e,n,i)};case 3:return function(n,i,r){return t.call(e,n,i,r)}}return function(){return t.apply(e,arguments)}},html=getBuiltIn("document","documentElement"),userAgent=getBuiltIn("navigator","userAgent")||"",isIos=/(iphone|ipod|ipad).*applewebkit/i.test(userAgent),location=global_1.location,set$1=global_1.setImmediate,clear=global_1.clearImmediate,process=global_1.process,MessageChannel=global_1.MessageChannel,Dispatch=global_1.Dispatch,counter=0,queue={},ONREADYSTATECHANGE="onreadystatechange",defer,channel,port,run=function(t){if(queue.hasOwnProperty(t)){var e=queue[t];delete queue[t],e()}},runner=function(t){return function(){run(t)}},listener=function(t){run(t.data)},post=function(t){global_1.postMessage(t+"",location.protocol+"//"+location.host)};set$1&&clear||(set$1=function(t){var e=[],n=1;while(arguments.length>n)e.push(arguments[n++]);return queue[++counter]=function(){("function"==typeof t?t:Function(t)).apply(void 0,e)},defer(counter),counter},clear=function(t){delete queue[t]},"process"==classofRaw(process)?defer=function(t){process.nextTick(runner(t))}:Dispatch&&Dispatch.now?defer=function(t){Dispatch.now(runner(t))}:MessageChannel&&!isIos?(channel=new MessageChannel,port=channel.port2,channel.port1.onmessage=listener,defer=bindContext(port.postMessage,port,1)):!global_1.addEventListener||"function"!=typeof postMessage||global_1.importScripts||fails(post)?defer=ONREADYSTATECHANGE in documentCreateElement("script")?function(t){html.appendChild(documentCreateElement("script"))[ONREADYSTATECHANGE]=function(){html.removeChild(this),run(t)}}:function(t){setTimeout(runner(t),0)}:(defer=post,global_1.addEventListener("message",listener,!1)));var task={set:set$1,clear:clear},FORCED=!global_1.setImmediate||!global_1.clearImmediate;_export({global:!0,bind:!0,enumerable:!0,forced:FORCED},{setImmediate:task.set,clearImmediate:task.clear});var setImmediate$1=path.setImmediate,eventemitter3=createCommonjsModule((function(t){var e=Object.prototype.hasOwnProperty,n="~";function i(){}function r(t,e,n){this.fn=t,this.context=e,this.once=n||!1}function o(t,e,i,o,a){if("function"!==typeof i)throw new TypeError("The listener must be a function");var s=new r(i,o||t,a),l=n?n+e:e;return t._events[l]?t._events[l].fn?t._events[l]=[t._events[l],s]:t._events[l].push(s):(t._events[l]=s,t._eventsCount++),t}function a(t,e){0===--t._eventsCount?t._events=new i:delete t._events[e]}function s(){this._events=new i,this._eventsCount=0}Object.create&&(i.prototype=Object.create(null),(new i).__proto__||(n=!1)),s.prototype.eventNames=function(){var t,i,r=[];if(0===this._eventsCount)return r;for(i in t=this._events)e.call(t,i)&&r.push(n?i.slice(1):i);return Object.getOwnPropertySymbols?r.concat(Object.getOwnPropertySymbols(t)):r},s.prototype.listeners=function(t){var e=n?n+t:t,i=this._events[e];if(!i)return[];if(i.fn)return[i.fn];for(var r=0,o=i.length,a=new Array(o);r<o;r++)a[r]=i[r].fn;return a},s.prototype.listenerCount=function(t){var e=n?n+t:t,i=this._events[e];return i?i.fn?1:i.length:0},s.prototype.emit=function(t,e,i,r,o,a){var s=n?n+t:t;if(!this._events[s])return!1;var l,u,c=this._events[s],d=arguments.length;if(c.fn){switch(c.once&&this.removeListener(t,c.fn,void 0,!0),d){case 1:return c.fn.call(c.context),!0;case 2:return c.fn.call(c.context,e),!0;case 3:return c.fn.call(c.context,e,i),!0;case 4:return c.fn.call(c.context,e,i,r),!0;case 5:return c.fn.call(c.context,e,i,r,o),!0;case 6:return c.fn.call(c.context,e,i,r,o,a),!0}for(u=1,l=new Array(d-1);u<d;u++)l[u-1]=arguments[u];c.fn.apply(c.context,l)}else{var h,f=c.length;for(u=0;u<f;u++)switch(c[u].once&&this.removeListener(t,c[u].fn,void 0,!0),d){case 1:c[u].fn.call(c[u].context);break;case 2:c[u].fn.call(c[u].context,e);break;case 3:c[u].fn.call(c[u].context,e,i);break;case 4:c[u].fn.call(c[u].context,e,i,r);break;default:if(!l)for(h=1,l=new Array(d-1);h<d;h++)l[h-1]=arguments[h];c[u].fn.apply(c[u].context,l)}}return!0},s.prototype.on=function(t,e,n){return o(this,t,e,n,!1)},s.prototype.once=function(t,e,n){return o(this,t,e,n,!0)},s.prototype.removeListener=function(t,e,i,r){var o=n?n+t:t;if(!this._events[o])return this;if(!e)return a(this,o),this;var s=this._events[o];if(s.fn)s.fn!==e||r&&!s.once||i&&s.context!==i||a(this,o);else{for(var l=0,u=[],c=s.length;l<c;l++)(s[l].fn!==e||r&&!s[l].once||i&&s[l].context!==i)&&u.push(s[l]);u.length?this._events[o]=1===u.length?u[0]:u:a(this,o)}return this},s.prototype.removeAllListeners=function(t){var e;return t?(e=n?n+t:t,this._events[e]&&a(this,e)):(this._events=new i,this._eventsCount=0),this},s.prototype.off=s.prototype.removeListener,s.prototype.addListener=s.prototype.on,s.prefixed=n,s.EventEmitter=s,t.exports=s})),IsEmptyObject=function(t){return t&&t.constructor===Object&&0===Object.keys(t).length},RandomStr=function(){for(var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:8,e="0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ".split(""),n="",i=0;i<t;i++)n+=e[62*Math.random()|0];return n};function _genList(t){for(var e=[],n=t.split("|"),i=0;i<n.length;i++){var r=n[i].trim();(0!==i||r)&&((i!==n.length-1||r)&&e.push(r))}return e}function _genItem(t,e){for(var n={},i=0;i<t.length;i++)n[t[i]]=e[i];return n}var ParseSettlementContent=function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:"";if(""===t)return t;for(var e=t.split("\n"),n="",i={positionClosed:{title:"平仓明细 Position Closed",colNames:[]},transactionRecords:{title:"成交记录 Transaction Record",colNames:[]}},r={account:{},positionClosed:[],transactionRecords:[]},o=0;o<e.length;o++){var a=e[o].trim();if(a.indexOf("资金状况")>-1)n="account",o++;else if(a.indexOf("平仓明细")>-1||a.indexOf("成交记录")>-1){n=a.indexOf("平仓明细")>-1?"positionClosed":"transactionRecords";while(o++){var s=e[o].trim();if(""===s.replace(/-/g,"")){if(0===i[n].colNames.length)continue;break}i[n].colNames=_genList(s)}}else if("account"===n){if(0===a.length||""===a.replace("-","")){n="";continue}for(var l=a.match(/([A-Z][a-zA-Z\.\/\(\)\s]+)[:：]+/g),u=a.match(/(-?[\d]+\.\d\d)/g),c=0;c<l.length;c++)r.account[l[c].split(/[:：]/)[0]]=u[c]}else if("positionClosed"===n||"transactionRecords"===n){if(0===a.length||""===a.replace(/-/g,"")){n="";continue}var d=i[n].colNames,h=_genList(a),f=_genItem(d,h);if(d.length!==h.length&&"transactionRecords"===n)for(var p=d.indexOf("Lots"),m=d.indexOf("Fee"),g=1;g<h.length;g++)if(/^\d+$/.test(h[g])){f.Lots=h[g],f.Fee=h[g+(m-p)];break}r[n].push(f)}}return r},TqWebsocket=function(_EventEmitter){function TqWebsocket(t){var e,n=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};return _classCallCheck(this,TqWebsocket),e=_possibleConstructorReturn(this,_getPrototypeOf(TqWebsocket).call(this)),e.urlList=t instanceof Array?t:[t],e.ws=null,e.queue=[],e.reconnect=!0,e.reconnectTask=null,e.reconnectInterval=n.reconnectInterval?n.reconnectInterval:3e3,e.reconnectMaxTimes=n.reconnectMaxTimes?n.reconnectMaxTimes:2,e.WebSocket=n.WebSocket?n.WebSocket:WebSocket,e.reconnectTimes=0,e.reconnectUrlIndex=0,e.STATUS={CONNECTING:0,OPEN:1,CLOSING:2,CLOSED:3},e.__init(!1),e}return _inherits(TqWebsocket,_EventEmitter),_createClass(TqWebsocket,[{key:"send",value:function(t){var e=JSON.stringify(t);this.isReady()?this.ws.send(e):this.queue.push(e)}},{key:"isReady",value:function(){return this.ws.readyState===this.WebSocket.OPEN}},{key:"__init",value:function __init(){var isReconnection=!(arguments.length>0&&void 0!==arguments[0])||arguments[0];this.ws=new this.WebSocket(this.urlList[this.reconnectUrlIndex]),isReconnection&&this.reconnectUrlIndex===this.urlList.length-1&&(this.reconnectTimes+=1);var _this=this;this.ws.onmessage=function(message){var data=eval("("+message.data+")");_this.emit("message",data),setImmediate((function(){_this.ws.send('{"aid":"peek_message"}')}))},this.ws.onclose=function(t){_this.emit("close"),_this.queue=[],_this.reconnect&&(_this.reconnectMaxTimes<=_this.reconnectTimes?(clearTimeout(_this.reconnectTask),_this.emit("death",{msg:"超过重连次数"+_this.reconnectMaxTimes})):_this.reconnectTask=setTimeout((function(){3===_this.ws.readyState&&(_this.reconnectUrlIndex=_this.reconnectUrlIndex+1<_this.urlList.length?_this.reconnectUrlIndex+1:0,_this.__init(!0),_this.emit("reconnect",{msg:"发起重连第 "+_this.reconnectTimes+" 次"}))}),_this.reconnectInterval))},this.ws.onerror=function(t){_this.emit("error",t),_this.ws.close()},this.ws.onopen=function(){_this.emit("open",{msg:"发起重连第 "+_this.reconnectTimes+" 次, 成功"}),this.reconnectTask&&clearTimeout(_this.reconnectTask);while(_this.queue.length>0){if(1!==_this.ws.readyState)break;_this.ws.send(_this.queue.shift())}}}},{key:"close",value:function(){this.ws.onclose=function(){},this.ws.close()}}]),TqWebsocket}(eventemitter3),TqTradeWebsocket=function(t){function e(t,n){var i,r=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};return _classCallCheck(this,e),i=_possibleConstructorReturn(this,_getPrototypeOf(e).call(this,t,r)),i.dm=n,i.req_login=null,i.init(),i}return _inherits(e,t),_createClass(e,[{key:"init",value:function(){var t=this;this.on("message",(function(e){if("rtn_data"===e.aid){for(var n=t._separateNotifies(e.data),i=0;i<n.length;i++)t.emit("notify",n[i]);t.dm.mergeData(e.data)}else if("rtn_brokers"===e.aid)t.emit("rtn_brokers",e.brokers);else if("qry_settlement_info"===e.aid){var r=ParseSettlementContent(e.settlement_info);t.dm.mergeData({trade:_defineProperty({},e.user_name,{his_settlements:_defineProperty({},e.trading_day,r)})}),Tqsdk.store&&Tqsdk.store.setContent(e.user_name,e.trading_day,e.settlement_info)}})),this.on("reconnect",(function(){t.req_login&&t.send(t.req_login)}))}},{key:"_separateNotifies",value:function(t){for(var e=[],n=0;n<t.length;n++)if(t[n].notify){var i=t.splice(n--,1)[0].notify;for(var r in i)e.push(i[r])}return e}},{key:"send",value:function(t){"req_login"===t.aid&&(this.req_login=t),_get(_getPrototypeOf(e.prototype),"send",this).call(this,t)}}]),e}(TqWebsocket),TqQuoteWebsocket=function(t){function e(t,n){var i,r=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};return _classCallCheck(this,e),i=_possibleConstructorReturn(this,_getPrototypeOf(e).call(this,t,r)),i.dm=n,i.subscribe_quote=null,i.charts={},i.init(),i}return _inherits(e,t),_createClass(e,[{key:"init",value:function(){var t=this;this.on("message",(function(e){"rtn_data"===e.aid&&t.dm.mergeData(e.data)})),this.on("reconnect",(function(e){for(var n in t.subscribe_quote&&t.send(t.subscribe_quote),t.charts)t.charts[n].view_width>0&&t.send(t.charts[n])}))}},{key:"send",value:function(t){"subscribe_quote"===t.aid?null!==this.subscribe_quote&&JSON.stringify(t.ins_list)===JSON.stringify(this.subscribe_quote.ins_list)||(this.subscribe_quote=t,_get(_getPrototypeOf(e.prototype),"send",this).call(this,t)):"set_chart"===t.aid&&(0===t.view_width?this.charts[t.chart_id]&&delete this.charts[t.chart_id]:this.charts[t.chart_id]=t,_get(_getPrototypeOf(e.prototype),"send",this).call(this,t))}}]),e}(TqWebsocket),TqRecvOnlyWebsocket=function(t){function e(t,n){var i,r=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};return _classCallCheck(this,e),i=_possibleConstructorReturn(this,_getPrototypeOf(e).call(this,t,r)),i.dm=n,i.init(),i}return _inherits(e,t),_createClass(e,[{key:"init",value:function(){var t=this;this.on("message",(function(e){"rtn_data"===e.aid&&t.dm.mergeData(e.data)})),this.on("reconnect",(function(t){}))}}]),e}(TqWebsocket),Quote=function(){function t(){_classCallCheck(this,t),this.instrument_id="",this.datetime="",this._last_price="-",this.ask_price1="-",this.ask_volume1="-",this.bid_price1="-",this.bid_volume1="-",this.highest="-",this.lowest="-",this.open="-",this.close="-",this.average="-",this.volume="-",this.amount="-",this.open_interest="-",this.lower_limit="-",this.upper_limit="-",this.settlement="-",this.change="-",this.change_percent="-",this.strike_price=NaN,this.pre_open_interest="-",this.pre_close="-",this.pre_volume="-",this._pre_settlement="-",this.margin="-",this.commission="-"}return _createClass(t,[{key:"setChange",value:function(){Number.isFinite(this._last_price)&&Number.isFinite(this._pre_settlement)&&0!==this._pre_settlement&&(this.change=this._last_price-this._pre_settlement,this.change_percent=this.change/this._pre_settlement*100)}},{key:"last_price",set:function(t){this._last_price=t,this.setChange()},get:function(){return this._last_price}},{key:"pre_settlement",set:function(t){this._pre_settlement=t,this.setChange()},get:function(){return this._pre_settlement}}]),t}(),Chart=function t(){var e=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{};_classCallCheck(this,t),this.left_id=-1,this.right_id=-1,this.more_data=!0,this.state=e},DataManager=function(t){function e(){var t,n=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{};return _classCallCheck(this,e),t=_possibleConstructorReturn(this,_getPrototypeOf(e).call(this)),t._epoch=0,t._data=n,t._diffs=[],t}return _inherits(e,t),_createClass(e,[{key:"mergeData",value:function(t){var e=!(arguments.length>1&&void 0!==arguments[1])||arguments[1],n=!(arguments.length>2&&void 0!==arguments[2])||arguments[2],i=Array.isArray(t)?t:[t];e&&(this._epoch+=1,this._diffs=i);var r=!0,o=!1,a=void 0;try{for(var s,l=i[Symbol.iterator]();!(r=(s=l.next()).done);r=!0){var u=s.value;null===u||IsEmptyObject(u)||this._mergeObject(this._data,u,this._epoch,n)}}catch(c){o=!0,a=c}finally{try{r||null==l["return"]||l["return"]()}finally{if(o)throw a}}e&&this._data._epoch===this._epoch&&this.emit("data",null)}},{key:"_mergeObject",value:function(t,e){var n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:0,i=!(arguments.length>3&&void 0!==arguments[3])||arguments[3];for(var r in e){var o=e[r],a=_typeof(o);if(["string","boolean","number"].includes(a))t[r]="NaN"===o?NaN:o;else if(null===o&&i)delete t[r];else if(Array.isArray(o))t[r]=o,o._epoch||Object.defineProperty(o,"_epoch",{configurable:!1,enumerable:!1,writable:!0}),o._epoch=n;else if("object"===a)if(t[r]=t[r]||("data"===r?[]:{}),"quotes"===r)for(var s in o){var l=o[s];null!==l?(t[r][s]||(t[r][s]=new Quote),this._mergeObject(t[r][s],l,n,i)):i&&s&&delete t[r][s]}else this._mergeObject(t[r],o,n,i)}t._root||Object.defineProperty(t,"_root",{value:this._data,configurable:!1,enumerable:!1,writable:!1}),t._epoch||Object.defineProperty(t,"_epoch",{configurable:!1,enumerable:!1,writable:!0}),t._epoch=n}},{key:"isChanging",value:function(t){if(Array.isArray(t)){for(var e=this._data,n=0;n<t.length;n++){if(e=e[t[n]],void 0===e)return!1;if(e._epoch&&e._epoch===this._epoch)return!0}return!1}return!(!t||!t._epoch)&&t._epoch===this._epoch}},{key:"setDefault",value:function(t){for(var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{},n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:this._data,i=n,r=0;r<t.length;r++){if("string"!==typeof t[r]&&"number"!==typeof t[r])break;var o=t[r];if(o in i||(i[o]=r===t.length-1?e:{}),r===t.length-1)return i[o];i=i[o]}return i}},{key:"getByPath",value:function(t){for(var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:this._data,n=e,i=0;i<t.length;i++)if(n=n[t[i]],void 0===n||null===n)return null;return n}}]),e}(eventemitter3),Tqsdk=function(t){function e(){var t,n=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{},i=n.symbolsServerUrl,r=void 0===i?"https://openmd.shinnytech.com/t/md/symbols/latest.json":i,o=n.wsQuoteUrl,a=void 0===o?"wss://openmd.shinnytech.com/t/md/front/mobile":o,s=n.clientSystemInfo,l=void 0===s?"":s,u=n.clientAppId,c=void 0===u?"":u,d=n.autoInit,h=void 0===d||d,f=n.data,p=void 0===f?{klines:{},quotes:{},charts:{},ticks:{},trade:{}}:f,m=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};_classCallCheck(this,e),t=_possibleConstructorReturn(this,_getPrototypeOf(e).call(this)),t._insUrl=r,t._mdUrl=a,t.clientSystemInfo=l,t.clientAppId=c,t.wsOption=m,t._prefix="TQJS_";var g=_assertThisInitialized(t);return t.dm=new DataManager(p),t.dm.on("data",(function(){g.emit("rtn_data",null)})),t.brokers_list=null,t.brokers=null,t.trade_accounts={},t.subscribeQuotesSet=new Set,t.quotesWs=null,t.quotesInfo={},h&&(t.initMdWebsocket(),t.initTdWebsocket()),t}return _inherits(e,t),_createClass(e,[{key:"initMdWebsocket",value:function(){if(!this.quotesWs){var t=this;axios__WEBPACK_IMPORTED_MODULE_0___default.a.get(this._insUrl,{headers:{Accept:"application/json; charset=utf-8"}}).then((function(e){t.quotesInfo=e.data,t.emit("ready"),t.emit("rtn_data",null)}))["catch"]((function(e){return t.emit("error",e),e})),this.quotesWs=new TqQuoteWebsocket(this._mdUrl,this.dm,this.wsOption)}}},{key:"initTdWebsocket",value:function(){if(!this.brokers){var t=this;axios__WEBPACK_IMPORTED_MODULE_0___default.a.get("https://files.shinnytech.com/broker-list.json",{headers:{Accept:"application/json; charset=utf-8"}}).then((function(e){t.brokers_list=e.data,t.brokers=Object.keys(e.data).filter((function(t){return!t.endsWith(" ")})).sort(),t.emit("rtn_brokers",t.brokers)}))["catch"]((function(e){return t.emit("error",e),e}))}}},{key:"addWebSocket",value:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:"";return t?new TqRecvOnlyWebsocket(t,this.dm,this.wsOption):null}},{key:"get",value:function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{},e=t.name,n=void 0===e?"users":e,i=t.bid,r=void 0===i?"":i,o=t.user_id,a=void 0===o?"":o,s=t.currency,l=void 0===s?"CNY":s,u=t.symbol,c=void 0===u?"":u,d=t.order_id,h=void 0===d?"":d,f=t.trade_id,p=void 0===f?"":f,m=t.trading_day,g=void 0===m?"":m,y=t.chart_id,v=void 0===y?"":y,_=t.input,b=void 0===_?"":_,x=t.duration,w=void 0===x?0:x;if("users"===n)return Object.keys(this.trade_accounts);if(a){var M=this._getAccountInfoByPaths({bid:r,user_id:a},[]);if(null===M&&(M=this.dm.getByPath("trade",a)),"user"===n)return M;if(["session","accounts","positions","orders","trades","his_settlements"].indexOf(n)>-1)return M&&M[n]?M[n]:null;if(M&&M[n+"s"]){var S="account"===n?l:"position"===n?c:"order"===n?h:"trade"===n?p:"his_settlement"===n?g:"";return M[n+"s"][S]}return null}return"quotes"===n?b?this.get_quotes_by_input(b):[]:"quote"===n?this.getQuote(c):"klines"===n?this.getKlines(c,w):"ticks"===n?this.getTicks(c):"charts"===n?this.dm.getByPath(["charts"]):"chart"===n?this.dm.getByPath(["charts",v]):void 0}},{key:"getByPath",value:function(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:this.dm;return e.getByPath(t)}},{key:"getQuotesByInput",value:function(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};if("string"!==typeof t)return[];var n={input:t.toLowerCase(),symbol:"boolean"!==typeof e.symbol||e.symbol,pinyin:"boolean"!==typeof e.pinyin||e.pinyin,include_expired:"boolean"===typeof e.include_expired&&e.include_expired,FUTURE:"boolean"!==typeof e.future||e.future,FUTURE_INDEX:"boolean"===typeof e.future_index&&e.future_index,FUTURE_CONT:"boolean"===typeof e.future_cont&&e.future_cont,OPTION:"boolean"===typeof e.option&&e.option,COMBINE:"boolean"===typeof e.combine&&e.combine},i=[];for(var r in this.quotesInfo)this._filterSymbol(n,this.quotesInfo[r])&&i.push(r);return i}},{key:"_filterSymbol",value:function(t,e){if(t[e["class"]]&&(t.include_expired||!t.include_expired&&!e.expired)){if(t.symbol)if(e.underlying_product){var n=e.underlying_product.toLowerCase().split("."),i=_slicedToArray(n,2),r=i[0],o=i[1];if(r===t.input||o===t.input)return!0}else{if(e.product_id&&e.product_id.toLowerCase()===t.input)return!0;if(t.input.length>2&&e.instrument_id.toLowerCase().indexOf(t.input)>-1)return!0}if(t.pinyin){var a=e.py.split(","),s=!0,l=!1,u=void 0;try{for(var c,d=a[Symbol.iterator]();!(s=(c=d.next()).done);s=!0){var h=c.value;if(h.indexOf(t.input)>-1)return!0}}catch(f){l=!0,u=f}finally{try{s||null==d["return"]||d["return"]()}finally{if(l)throw u}}}}return!1}},{key:"getQuote",value:function(t){if(""===t)return{};if(!this.quotesInfo[t])return{};var e=this.dm.setDefault(["quotes",t],new Quote);if(!e["class"]){var n=e.last_price?e.last_price:this.quotesInfo[t].last_price;Object.assign(e,this.quotesInfo[t],{last_price:n})}return this.subscribeQuote(t),e}},{key:"setChart",value:function(t){var e={};t.trading_day_start||t.trading_day_count?(e.trading_day_start=t.trading_day_start?t.trading_day_start:0,e.trading_day_count=t.trading_day_count?t.trading_day_count:864e11):(e.view_width=t.view_width?t.view_width:500,t.left_kline_id?e.left_kline_id=t.left_kline_id:t.focus_datetime&&(e.focus_datetime=t.focus_datetime,e.focus_position=t.focus_position?t.focus_position:0));var n=Object.assign({aid:"set_chart",chart_id:t.chart_id?t.chart_id:this._prefix+"kline_chart",ins_list:t.ins_list?t.ins_list.join(","):t.symbol,duration:t.duration},e);return this.quotesWs.send(n),this.dm.setDefault(["charts",n.chart_id],new Chart(n))}},{key:"getChart",value:function(t){return""===t?null:this.dm.getByPath(["charts",t])}},{key:"getKlines",value:function(t,e){if(""===t)return null;var n=this.dm.getByPath(["klines",t,e]);return n&&n.data&&-1!==n.last_id||(this.dm.mergeData({klines:_defineProperty({},t,_defineProperty({},e,{trading_day_end_id:-1,trading_day_start_id:-1,last_id:-1,data:{}}))},!1,!1),n=this.dm.getByPath(["klines",t,e])),n}},{key:"getTicks",value:function(t){if(""===t)return null;var e=this.dm.getByPath(["ticks",t]);return e&&e.data||this.dm.mergeData({ticks:_defineProperty({},t,{last_id:-1,data:{}})},!1,!1),this.dm.getByPath(["ticks",t])}},{key:"isChanging",value:function(t){return t&&t._epoch&&t._root&&t._root._epoch?t._epoch===t._root._epoch:!!Array.isArray(t)&&this.dm.isChanging(t)}},{key:"subscribeQuote",value:function(){var t,e=arguments.length>0&&void 0!==arguments[0]?arguments[0]:[],n=this.subscribeQuotesSet.size;for(var i in this.trade_accounts){var r=this.getPositions(this.trade_accounts[i]);if(r)for(var o in r)this.subscribeQuotesSet.add(o)}e="string"===typeof e?e.split(","):e;var a=!0,s=!1,l=void 0;try{for(var u,c=e[Symbol.iterator]();!(a=(u=c.next()).done);a=!0){var d=u.value;this.subscribeQuotesSet.add(d)}}catch(h){s=!0,l=h}finally{try{a||null==c["return"]||c["return"]()}finally{if(s)throw l}}n!==this.subscribeQuotesSet.size&&this.quotesWs.send({aid:"subscribe_quote",ins_list:(t=[]).concat.apply(t,_toConsumableArray(this.subscribeQuotesSet)).join(",")})}},{key:"addAccount",value:function(t){if(this.brokers){if(t.bid&&t.user_id&&t.password){if(-1===this.brokers.indexOf(t.bid))return null;var e=this._getAccountKey(t);if(e&&!this.trade_accounts[e]){var n=new DataManager({trade:_defineProperty({},t.user_id,{accounts:{CNY:{}},trades:{},positions:{},orders:{},his_settlements:{}})}),i=[this.brokers_list[t.bid].url+"?access_token=eyJhbGciOiJSUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICJobi1MZ3ZwbWlFTTJHZHAtRmlScjV5MUF5MnZrQmpLSFFyQVlnQ0UwR1JjIn0.eyJqdGkiOiJlZjgyYjBjNS02MDYwLTRjZTYtYjZhMS1kYjkxYmMxYjcyOTQiLCJleHAiOjE2NDIwNjc5NTYsIm5iZiI6MCwiaWF0IjoxNjEwNTMxOTU2LCJpc3MiOiJodHRwOi8vYXV0aC5zaGlubnl0ZWNoLmNvbS9hdXRoL3JlYWxtcy9zaGlubnl0ZWNoIiwic3ViIjoiNGJiMmEwZGMtMjZkMS00M2Y1LTlkNjctOTlkYzVkZGE4Y2ZlIiwidHlwIjoiQmVhcmVyIiwiYXpwIjoic2hpbm55X3dlYiIsImF1dGhfdGltZSI6MCwic2Vzc2lvbl9zdGF0ZSI6IjU0NmY5OWM4LWJmNjctNDhlYy1iNjFmLTY3MGUyZWZhYmQ4ZSIsImFjciI6IjEiLCJzY29wZSI6ImF0dHJpYnV0ZXMiLCJncmFudHMiOnsiZmVhdHVyZXMiOlsiIl0sImFjY291bnRzIjpbIioiXX19.B1pQIiu3iS6F7eusKSzjLZaZKtV084l4JYPNjkmBnNp2hBCEPfGwK4u-H4KMFXhbK5l4acvRY-QS7t3UMLEDbn-UQYEfcOeQVnWYlVDfSWUbSCGxLCIaQYHCct7VcWhcLgg0_xNmJMNdnnLvq44J8S1o4PFNbXJosScspenZUZmuZi75KTdFiA8KU_83X2-gJ7UjFuJ03YoaxFMBJhG7wpEXB5P_yjrgvjUGkD9oy-IRC2YH65ys0XpdzH5U4NzePi7gG01mh3mQM1DuyyS7YVkcxRrjvVUkx6lHChSjEMGtWWoC0aTg323dPTPQ8u4wPL650KOV3eGELBICUUVM8Q",this.brokers_list[t.bid].url],r=new TqTradeWebsocket(i,n,this.wsOption),o=this;n.on("data",(function(){o.emit("rtn_data",null)})),r.on("notify",(function(e){o.emit("notify",Object.assign(e,{bid:t.bid,user_id:t.user_id}))})),this.trade_accounts[e]={bid:t.bid,user_id:t.user_id,password:t.password,ws:r,dm:n}}return this.trade_accounts[e]}return null}}},{key:"removeAccount",value:function(t){var e=this._getAccountKey(t);this.trade_accounts[e]&&(this.trade_accounts[e].ws.close(),delete this.trade_accounts[e])}},{key:"login",value:function(t){var e=this._getAccountRef(t);IsEmptyObject(e)&&(e=this.addAccount(t));var n={aid:"req_login"};this.clientAppId&&(n.client_app_id=this.clientAppId,n.client_system_info=this.clientSystemInfo),e&&e.ws&&(n.bid=t.bid,n.user_name=t.user_id,n.password=t.password||e.password,e.ws.send(n))}},{key:"isLogined",value:function(t){var e=this._getAccountRef(t);if(e&&e.dm){var n=e.dm.getByPath(["trade",t.user_id,"session"]),i=e.dm.getByPath(["trade",e.user_id,"trade_more_data"]);if(n&&n.trading_day&&!1===i)return!0}return!1}},{key:"refreshAccount",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&(e.ws.send({aid:"qry_account_info"}),e.ws.send({aid:"qry_account_register"}))}},{key:"refreshAccounts",value:function(){for(var t in this.trade_accounts)this.trade_accounts[t].ws.send({aid:"qry_account_info"}),this.trade_accounts[t].ws.send({aid:"qry_account_register"})}},{key:"getAllAccounts",value:function(t){var e=[];return Object.values(this.trade_accounts).forEach((function(t){e.push({bid:t.bid,user_id:t.user_id,password:t.password})})),e}},{key:"getAccount",value:function(t){return this._getAccountInfoByPaths(t,["accounts","CNY"])}},{key:"_getAccountKey",value:function(t){if(t&&t.user_id){if(t.bid)return t.bid+","+t.user_id;var e="";for(var n in this.trade_accounts)if(t.user_id===this.trade_accounts[n].user_id){if(e)return"";e=this.trade_accounts[n].bid}return e+","+t.user_id}return""}},{key:"_getAccountRef",value:function(t){var e=this._getAccountKey(t);return this.trade_accounts[e]||{}}},{key:"insertOrder",value:function(t){if(!this.isLogined(t))return null;var e=this._getAccountRef(t),n=this._prefix+RandomStr(8),i={user_id:t.user_id,order_id:n,exchange_id:t.exchange_id,instrument_id:t.instrument_id,direction:t.direction,offset:t.offset,price_type:t.price_type?t.price_type:"LIMIT",limit_price:Number(t.limit_price),volume_condition:"ANY",time_condition:"ANY"===t.price_type?"IOC":"GFD"},r=Object.assign({aid:"insert_order",volume:t.volume},i);e.ws.send(r);var o=Object.assign({volume_orign:t.volume,status:"ALIVE",volume_left:t.volume},i);return e.dm.mergeData({trade:_defineProperty({},t.user_id,{orders:_defineProperty({},n,o)})},!1,!1),e.dm.getByPath(["trade",t.user_id,"orders",n])}},{key:"autoInsertOrder",value:function(t){if(!this.is_logined(t))return null;var e={bid:t.bid,user_id:t.user_id,exchange_id:t.exchange_id,instrument_id:t.instrument_id,direction:t.direction,price_type:t.price_type?t.price_type:"LIMIT",volume_condition:"ANY",time_condition:"ANY"===t.price_type?"IOC":"GFD",limit_price:Number(t.limit_price)};if("SHFE"!==t.exchange_id&&"INE"!==t.exchange_id||"CLOSE"!==t.offset)return[this.insert_order(Object.assign({offset:t.offset,volume:t.volume},e))];var n=this.getPosition({bid:t.bid,user_id:t.user_id,symbol:t.exchange_id+"."+t.instrument_id}),i=0;"BUY"===t.direction&&n.volume_short_today>0?i=Math.min(n.volume_short_today,t.volume):"SELL"===t.direction&&n.volume_long_today>0&&(i=Math.min(n.volume_long_today,t.volume));var r=[];if(i>0){var o=this.insert_order(Object.assign({offset:"CLOSETODAY",volume:i},e));r.push(o)}if(t.volume-i>0){var a=this.insert_order(Object.assign({offset:"CLOSE",volume:t.volume-i},e));r.push(a)}return r}},{key:"cancelOrder",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&e.ws.send({aid:"cancel_order",user_id:t.user_id,order_id:t.order_id})}},{key:"getPosition",value:function(t){return this._getAccountInfoByPaths(t,["positions",t.symbol])}},{key:"getPositions",value:function(t){return this._getAccountInfoByPaths(t,["positions"])}},{key:"getOrder",value:function(t){return this._getAccountInfoByPaths(t,["orders",t.order_id])}},{key:"getOrders",value:function(t){return this._getAccountInfoByPaths(t,["orders"])}},{key:"getOrdersBySymbol",value:function(t){var e=this._getAccountInfoByPaths(t,["orders"]),n=t.symbol.split("."),i=_slicedToArray(n,2),r=i[0],o=i[1],a={};for(var s in e)e[s].exchange_id===r&&e[s].instrument_id===o&&(a[s]=e[s]);return a}},{key:"getTrade",value:function(t){return this._getAccountInfoByPaths(t,["trades",t.trade_id])}},{key:"getTrades",value:function(t){return this._getAccountInfoByPaths(t,["trades"])}},{key:"getTradesByOrder",value:function(t){var e=this._getAccountInfoByPaths(t,["trades"]),n={};for(var i in e)e[i].order_id===t.order_id&&(n[i]=e[i]);return n}},{key:"getTradesBySymbol",value:function(t){var e=this._getAccountInfoByPaths(t,["trades"]),n=t.symbol.split("."),i=_slicedToArray(n,2),r=i[0],o=i[1],a={};for(var s in e)e[s].exchange_id===r&&e[s].instrument_id===o&&(a[s]=e[s]);return a}},{key:"getHisSettlements",value:function(t){return this._getAccountInfoByPaths(t,["his_settlements"])}},{key:"getHisSettlement",value:function(t){return this._getAccountInfoByPaths(t,["his_settlements",t.trading_day])}},{key:"_getAccountInfoByPaths",value:function(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:[],n=this._getAccountRef(t);return n&&n.dm?n.dm.getByPath(["trade",n.user_id].concat(e)):null}},{key:"confirmSettlement",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&!e._hasConfirmed&&(e._hasConfirmed=!0,e.ws.send({aid:"confirm_settlement"}))}},{key:"transfer",value:function(t){var e=this._getAccountRef(t);e&&e.ws&&e.ws.send({aid:"req_transfer",bank_id:t.bank_id,bank_password:t.bank_password,future_account:t.future_account,future_password:t.future_password,currency:t.currency||"CNY",amount:t.amount})}},{key:"hisSettlement",value:function(t){if(!e.store)return null;var n=this._getAccountRef(t);if(n&&n.dm){var i=n.dm.getByPath(["trade",t.user_id,"his_settlements",t.trading_day]);if(i)return;e.store.getContent(t.user_id,t.trading_day).then((function(e){if(null===e)n.ws.send({aid:"qry_settlement_info",trading_day:Number(t.trading_day)});else{var i=ParseSettlementContent(e);n.dm.mergeData({trade:_defineProperty({},t.user_id,{his_settlements:_defineProperty({},t.trading_day,i)})},!0,!1)}}))["catch"]((function(t){}))}}}]),e}(eventemitter3);Tqsdk.prototype.subscribe_quote=Tqsdk.prototype.subscribeQuote,Tqsdk.prototype.his_settlement=Tqsdk.prototype.hisSettlement,Tqsdk.prototype.confirm_settlement=Tqsdk.prototype.confirmSettlement,Tqsdk.prototype.add_account=Tqsdk.prototype.addAccount,Tqsdk.prototype.remove_account=Tqsdk.prototype.removeAccount,Tqsdk.prototype.get_by_path=Tqsdk.prototype.getByPath,Tqsdk.prototype.get_quotes_by_input=Tqsdk.prototype.getQuotesByInput,Tqsdk.prototype.get_quote=Tqsdk.prototype.getQuote,Tqsdk.prototype.set_chart=Tqsdk.prototype.setChart,Tqsdk.prototype.get_user=Tqsdk.prototype.getAccount,Tqsdk.prototype.is_logined=Tqsdk.prototype.isLogined,Tqsdk.prototype.is_changed=Tqsdk.prototype.isChanging,Tqsdk.prototype.insert_order=Tqsdk.prototype.insertOrder,Tqsdk.prototype.auto_insert_order=Tqsdk.prototype.autoInsertOrder,Tqsdk.prototype.cancel_order=Tqsdk.prototype.cancelOrder,Tqsdk.TqWebsocket=TqWebsocket,Tqsdk.DataManager=DataManager,Tqsdk.version=version,__webpack_exports__["a"]=Tqsdk}).call(this,__webpack_require__("c8ba"))},"099a":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("i",{class:t.classes,style:t.styles,on:{click:t.handleClick}})},r=[],o=(n("a9e3"),n("ade3")),a="ivu-icon",s={name:"Icon",props:{type:{type:String,default:""},size:[Number,String],color:String,custom:{type:String,default:""}},computed:{classes:function(){var t;return["".concat(a),(t={},Object(o["a"])(t,"".concat(a,"-").concat(this.type),""!==this.type),Object(o["a"])(t,"".concat(this.custom),""!==this.custom),t)]},styles:function(){var t={};return this.size&&(t["font-size"]="".concat(this.size,"px")),this.color&&(t.color=this.color),t}},methods:{handleClick:function(t){this.$emit("click",t)}}},l=s,u=n("2877"),c=Object(u["a"])(l,i,r,!1,null,null,null);e["a"]=c.exports},"0a06":function(t,e,n){"use strict";var i=n("c532"),r=n("30b5"),o=n("f6b49"),a=n("5270"),s=n("4a7b");function l(t){this.defaults=t,this.interceptors={request:new o,response:new o}}l.prototype.request=function(t){"string"===typeof t?(t=arguments[1]||{},t.url=arguments[0]):t=t||{},t=s(this.defaults,t),t.method?t.method=t.method.toLowerCase():this.defaults.method?t.method=this.defaults.method.toLowerCase():t.method="get";var e=[a,void 0],n=Promise.resolve(t);this.interceptors.request.forEach((function(t){e.unshift(t.fulfilled,t.rejected)})),this.interceptors.response.forEach((function(t){e.push(t.fulfilled,t.rejected)}));while(e.length)n=n.then(e.shift(),e.shift());return n},l.prototype.getUri=function(t){return t=s(this.defaults,t),r(t.url,t.params,t.paramsSerializer).replace(/^\?/,"")},i.forEach(["delete","get","head","options"],(function(t){l.prototype[t]=function(e,n){return this.request(s(n||{},{method:t,url:e,data:(n||{}).data}))}})),i.forEach(["post","put","patch"],(function(t){l.prototype[t]=function(e,n,i){return this.request(s(i||{},{method:t,url:e,data:n}))}})),t.exports=l},"0a3c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),i=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],r=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es-do",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,i){return t?/-MMM-/.test(i)?n[t.month()]:e[t.month()]:e},monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY h:mm A",LLLL:"dddd, D [de] MMMM [de] YYYY h:mm A"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return o}))},"0a84":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ar-ma",{months:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),monthsShort:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"احد_اثنين_ثلاثاء_اربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},week:{dow:1,doy:4}});return e}))},"0b07":function(t,e,n){var i=n("34ac"),r=n("3698");function o(t,e){var n=r(t,e);return i(n)?n:void 0}t.exports=o},"0b49":function(t,e,n){t.exports=n("f542")},"0caa":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r={s:["thoddea sekondamni","thodde sekond"],ss:[t+" sekondamni",t+" sekond"],m:["eka mintan","ek minut"],mm:[t+" mintamni",t+" mintam"],h:["eka voran","ek vor"],hh:[t+" voramni",t+" voram"],d:["eka disan","ek dis"],dd:[t+" disamni",t+" dis"],M:["eka mhoinean","ek mhoino"],MM:[t+" mhoineamni",t+" mhoine"],y:["eka vorsan","ek voros"],yy:[t+" vorsamni",t+" vorsam"]};return i?r[n][0]:r[n][1]}var n=t.defineLocale("gom-latn",{months:{standalone:"Janer_Febrer_Mars_Abril_Mai_Jun_Julai_Agost_Setembr_Otubr_Novembr_Dezembr".split("_"),format:"Janerachea_Febrerachea_Marsachea_Abrilachea_Maiachea_Junachea_Julaiachea_Agostachea_Setembrachea_Otubrachea_Novembrachea_Dezembrachea".split("_"),isFormat:/MMMM(\s)+D[oD]?/},monthsShort:"Jan._Feb._Mars_Abr._Mai_Jun_Jul._Ago._Set._Otu._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Aitar_Somar_Mongllar_Budhvar_Birestar_Sukrar_Son'var".split("_"),weekdaysShort:"Ait._Som._Mon._Bud._Bre._Suk._Son.".split("_"),weekdaysMin:"Ai_Sm_Mo_Bu_Br_Su_Sn".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"A h:mm [vazta]",LTS:"A h:mm:ss [vazta]",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY A h:mm [vazta]",LLLL:"dddd, MMMM Do, YYYY, A h:mm [vazta]",llll:"ddd, D MMM YYYY, A h:mm [vazta]"},calendar:{sameDay:"[Aiz] LT",nextDay:"[Faleam] LT",nextWeek:"[Fuddlo] dddd[,] LT",lastDay:"[Kal] LT",lastWeek:"[Fattlo] dddd[,] LT",sameElse:"L"},relativeTime:{future:"%s",past:"%s adim",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}(er)/,ordinal:function(t,e){switch(e){case"D":return t+"er";default:case"M":case"Q":case"DDD":case"d":case"w":case"W":return t}},week:{dow:0,doy:3},meridiemParse:/rati|sokallim|donparam|sanje/,meridiemHour:function(t,e){return 12===t&&(t=0),"rati"===e?t<4?t:t+12:"sokallim"===e?t:"donparam"===e?t>12?t:t+12:"sanje"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"rati":t<12?"sokallim":t<16?"donparam":t<20?"sanje":"rati"}});return n}))},"0cfb":function(t,e,n){var i=n("83ab"),r=n("d039"),o=n("cc12");t.exports=!i&&!r((function(){return 7!=Object.defineProperty(o("div"),"a",{get:function(){return 7}}).a}))},"0d24":function(t,e,n){(function(t){var i=n("2b3e"),r=n("07c7"),o=e&&!e.nodeType&&e,a=o&&"object"==typeof t&&t&&!t.nodeType&&t,s=a&&a.exports===o,l=s?i.Buffer:void 0,u=l?l.isBuffer:void 0,c=u||r;t.exports=c}).call(this,n("62e4")(t))},"0df6":function(t,e,n){"use strict";t.exports=function(t){return function(e){return t.apply(null,e)}}},"0e49":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fr-ch",{months:"janvier_février_mars_avril_mai_juin_juillet_août_septembre_octobre_novembre_décembre".split("_"),monthsShort:"janv._févr._mars_avr._mai_juin_juil._août_sept._oct._nov._déc.".split("_"),monthsParseExact:!0,weekdays:"dimanche_lundi_mardi_mercredi_jeudi_vendredi_samedi".split("_"),weekdaysShort:"dim._lun._mar._mer._jeu._ven._sam.".split("_"),weekdaysMin:"di_lu_ma_me_je_ve_sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Aujourd’hui à] LT",nextDay:"[Demain à] LT",nextWeek:"dddd [à] LT",lastDay:"[Hier à] LT",lastWeek:"dddd [dernier à] LT",sameElse:"L"},relativeTime:{future:"dans %s",past:"il y a %s",s:"quelques secondes",ss:"%d secondes",m:"une minute",mm:"%d minutes",h:"une heure",hh:"%d heures",d:"un jour",dd:"%d jours",M:"un mois",MM:"%d mois",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(er|e)/,ordinal:function(t,e){switch(e){default:case"M":case"Q":case"D":case"DDD":case"d":return t+(1===t?"er":"e");case"w":case"W":return t+(1===t?"re":"e")}},week:{dow:1,doy:4}});return e}))},"0e6b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-au",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:0,doy:4}});return e}))},"0e81":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"'inci",5:"'inci",8:"'inci",70:"'inci",80:"'inci",2:"'nci",7:"'nci",20:"'nci",50:"'nci",3:"'üncü",4:"'üncü",100:"'üncü",6:"'ncı",9:"'uncu",10:"'uncu",30:"'uncu",60:"'ıncı",90:"'ıncı"},n=t.defineLocale("tr",{months:"Ocak_Şubat_Mart_Nisan_Mayıs_Haziran_Temmuz_Ağustos_Eylül_Ekim_Kasım_Aralık".split("_"),monthsShort:"Oca_Şub_Mar_Nis_May_Haz_Tem_Ağu_Eyl_Eki_Kas_Ara".split("_"),weekdays:"Pazar_Pazartesi_Salı_Çarşamba_Perşembe_Cuma_Cumartesi".split("_"),weekdaysShort:"Paz_Pts_Sal_Çar_Per_Cum_Cts".split("_"),weekdaysMin:"Pz_Pt_Sa_Ça_Pe_Cu_Ct".split("_"),meridiem:function(t,e,n){return t<12?n?"öö":"ÖÖ":n?"ös":"ÖS"},meridiemParse:/öö|ÖÖ|ös|ÖS/,isPM:function(t){return"ös"===t||"ÖS"===t},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[bugün saat] LT",nextDay:"[yarın saat] LT",nextWeek:"[gelecek] dddd [saat] LT",lastDay:"[dün] LT",lastWeek:"[geçen] dddd [saat] LT",sameElse:"L"},relativeTime:{future:"%s sonra",past:"%s önce",s:"birkaç saniye",ss:"%d saniye",m:"bir dakika",mm:"%d dakika",h:"bir saat",hh:"%d saat",d:"bir gün",dd:"%d gün",w:"bir hafta",ww:"%d hafta",M:"bir ay",MM:"%d ay",y:"bir yıl",yy:"%d yıl"},ordinal:function(t,n){switch(n){case"d":case"D":case"Do":case"DD":return t;default:if(0===t)return t+"'ıncı";var i=t%10,r=t%100-i,o=t>=100?100:null;return t+(e[i]||e[r]||e[o])}},week:{dow:1,doy:7}});return n}))},"0f0f":function(t,e,n){var i=n("8eeb"),r=n("9934");function o(t,e){return t&&i(e,r(e),t)}t.exports=o},"0f14":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("da",{months:"januar_februar_marts_april_maj_juni_juli_august_september_oktober_november_december".split("_"),monthsShort:"jan_feb_mar_apr_maj_jun_jul_aug_sep_okt_nov_dec".split("_"),weekdays:"søndag_mandag_tirsdag_onsdag_torsdag_fredag_lørdag".split("_"),weekdaysShort:"søn_man_tir_ons_tor_fre_lør".split("_"),weekdaysMin:"sø_ma_ti_on_to_fr_lø".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd [d.] D. MMMM YYYY [kl.] HH:mm"},calendar:{sameDay:"[i dag kl.] LT",nextDay:"[i morgen kl.] LT",nextWeek:"på dddd [kl.] LT",lastDay:"[i går kl.] LT",lastWeek:"[i] dddd[s kl.] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"%s siden",s:"få sekunder",ss:"%d sekunder",m:"et minut",mm:"%d minutter",h:"en time",hh:"%d timer",d:"en dag",dd:"%d dage",M:"en måned",MM:"%d måneder",y:"et år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},"0f38":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tl-ph",{months:"Enero_Pebrero_Marso_Abril_Mayo_Hunyo_Hulyo_Agosto_Setyembre_Oktubre_Nobyembre_Disyembre".split("_"),monthsShort:"Ene_Peb_Mar_Abr_May_Hun_Hul_Ago_Set_Okt_Nob_Dis".split("_"),weekdays:"Linggo_Lunes_Martes_Miyerkules_Huwebes_Biyernes_Sabado".split("_"),weekdaysShort:"Lin_Lun_Mar_Miy_Huw_Biy_Sab".split("_"),weekdaysMin:"Li_Lu_Ma_Mi_Hu_Bi_Sab".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"MM/D/YYYY",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY HH:mm",LLLL:"dddd, MMMM DD, YYYY HH:mm"},calendar:{sameDay:"LT [ngayong araw]",nextDay:"[Bukas ng] LT",nextWeek:"LT [sa susunod na] dddd",lastDay:"LT [kahapon]",lastWeek:"LT [noong nakaraang] dddd",sameElse:"L"},relativeTime:{future:"sa loob ng %s",past:"%s ang nakalipas",s:"ilang segundo",ss:"%d segundo",m:"isang minuto",mm:"%d minuto",h:"isang oras",hh:"%d oras",d:"isang araw",dd:"%d araw",M:"isang buwan",MM:"%d buwan",y:"isang taon",yy:"%d taon"},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:function(t){return t},week:{dow:1,doy:4}});return e}))},"0f5c":function(t,e,n){var i=n("159a");function r(t,e,n){return null==t?t:i(t,e,n)}t.exports=r},"0ff2":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("eu",{months:"urtarrila_otsaila_martxoa_apirila_maiatza_ekaina_uztaila_abuztua_iraila_urria_azaroa_abendua".split("_"),monthsShort:"urt._ots._mar._api._mai._eka._uzt._abu._ira._urr._aza._abe.".split("_"),monthsParseExact:!0,weekdays:"igandea_astelehena_asteartea_asteazkena_osteguna_ostirala_larunbata".split("_"),weekdaysShort:"ig._al._ar._az._og._ol._lr.".split("_"),weekdaysMin:"ig_al_ar_az_og_ol_lr".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY[ko] MMMM[ren] D[a]",LLL:"YYYY[ko] MMMM[ren] D[a] HH:mm",LLLL:"dddd, YYYY[ko] MMMM[ren] D[a] HH:mm",l:"YYYY-M-D",ll:"YYYY[ko] MMM D[a]",lll:"YYYY[ko] MMM D[a] HH:mm",llll:"ddd, YYYY[ko] MMM D[a] HH:mm"},calendar:{sameDay:"[gaur] LT[etan]",nextDay:"[bihar] LT[etan]",nextWeek:"dddd LT[etan]",lastDay:"[atzo] LT[etan]",lastWeek:"[aurreko] dddd LT[etan]",sameElse:"L"},relativeTime:{future:"%s barru",past:"duela %s",s:"segundo batzuk",ss:"%d segundo",m:"minutu bat",mm:"%d minutu",h:"ordu bat",hh:"%d ordu",d:"egun bat",dd:"%d egun",M:"hilabete bat",MM:"%d hilabete",y:"urte bat",yy:"%d urte"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return e}))},"100e":function(t,e,n){var i=n("cd9d"),r=n("2286"),o=n("c1c9");function a(t,e){return o(r(t,e,i),t+"")}t.exports=a},1041:function(t,e,n){var i=n("8eeb"),r=n("a029");function o(t,e){return i(t,r(t),e)}t.exports=o},"10e8":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("th",{months:"มกราคม_กุมภาพันธ์_มีนาคม_เมษายน_พฤษภาคม_มิถุนายน_กรกฎาคม_สิงหาคม_กันยายน_ตุลาคม_พฤศจิกายน_ธันวาคม".split("_"),monthsShort:"ม.ค._ก.พ._มี.ค._เม.ย._พ.ค._มิ.ย._ก.ค._ส.ค._ก.ย._ต.ค._พ.ย._ธ.ค.".split("_"),monthsParseExact:!0,weekdays:"อาทิตย์_จันทร์_อังคาร_พุธ_พฤหัสบดี_ศุกร์_เสาร์".split("_"),weekdaysShort:"อาทิตย์_จันทร์_อังคาร_พุธ_พฤหัส_ศุกร์_เสาร์".split("_"),weekdaysMin:"อา._จ._อ._พ._พฤ._ศ._ส.".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY เวลา H:mm",LLLL:"วันddddที่ D MMMM YYYY เวลา H:mm"},meridiemParse:/ก่อนเที่ยง|หลังเที่ยง/,isPM:function(t){return"หลังเที่ยง"===t},meridiem:function(t,e,n){return t<12?"ก่อนเที่ยง":"หลังเที่ยง"},calendar:{sameDay:"[วันนี้ เวลา] LT",nextDay:"[พรุ่งนี้ เวลา] LT",nextWeek:"dddd[หน้า เวลา] LT",lastDay:"[เมื่อวานนี้ เวลา] LT",lastWeek:"[วัน]dddd[ที่แล้ว เวลา] LT",sameElse:"L"},relativeTime:{future:"อีก %s",past:"%sที่แล้ว",s:"ไม่กี่วินาที",ss:"%d วินาที",m:"1 นาที",mm:"%d นาที",h:"1 ชั่วโมง",hh:"%d ชั่วโมง",d:"1 วัน",dd:"%d วัน",w:"1 สัปดาห์",ww:"%d สัปดาห์",M:"1 เดือน",MM:"%d เดือน",y:"1 ปี",yy:"%d ปี"}});return e}))},1148:function(t,e,n){"use strict";var i=n("a691"),r=n("1d80");t.exports="".repeat||function(t){var e=String(r(this)),n="",o=i(t);if(o<0||o==1/0)throw RangeError("Wrong number of repetitions");for(;o>0;(o>>>=1)&&(e+=e))1&o&&(n+=e);return n}},1276:function(t,e,n){"use strict";var i=n("d784"),r=n("44e7"),o=n("825a"),a=n("1d80"),s=n("4840"),l=n("8aa5"),u=n("50c4"),c=n("14c3"),d=n("9263"),h=n("d039"),f=[].push,p=Math.min,m=4294967295,g=!h((function(){return!RegExp(m,"y")}));i("split",2,(function(t,e,n){var i;return i="c"=="abbc".split(/(b)*/)[1]||4!="test".split(/(?:)/,-1).length||2!="ab".split(/(?:ab)*/).length||4!=".".split(/(.?)(.?)/).length||".".split(/()()/).length>1||"".split(/.?/).length?function(t,n){var i=String(a(this)),o=void 0===n?m:n>>>0;if(0===o)return[];if(void 0===t)return[i];if(!r(t))return e.call(i,t,o);var s,l,u,c=[],h=(t.ignoreCase?"i":"")+(t.multiline?"m":"")+(t.unicode?"u":"")+(t.sticky?"y":""),p=0,g=new RegExp(t.source,h+"g");while(s=d.call(g,i)){if(l=g.lastIndex,l>p&&(c.push(i.slice(p,s.index)),s.length>1&&s.index<i.length&&f.apply(c,s.slice(1)),u=s[0].length,p=l,c.length>=o))break;g.lastIndex===s.index&&g.lastIndex++}return p===i.length?!u&&g.test("")||c.push(""):c.push(i.slice(p)),c.length>o?c.slice(0,o):c}:"0".split(void 0,0).length?function(t,n){return void 0===t&&0===n?[]:e.call(this,t,n)}:e,[function(e,n){var r=a(this),o=void 0==e?void 0:e[t];return void 0!==o?o.call(e,r,n):i.call(String(r),e,n)},function(t,r){var a=n(i,t,this,r,i!==e);if(a.done)return a.value;var d=o(t),h=String(this),f=s(d,RegExp),y=d.unicode,v=(d.ignoreCase?"i":"")+(d.multiline?"m":"")+(d.unicode?"u":"")+(g?"y":"g"),_=new f(g?d:"^(?:"+d.source+")",v),b=void 0===r?m:r>>>0;if(0===b)return[];if(0===h.length)return null===c(_,h)?[h]:[];var x=0,w=0,M=[];while(w<h.length){_.lastIndex=g?w:0;var S,L=c(_,g?h:h.slice(w));if(null===L||(S=p(u(_.lastIndex+(g?0:w)),h.length))===x)w=l(h,w,y);else{if(M.push(h.slice(x,w)),M.length===b)return M;for(var k=1;k<=L.length-1;k++)if(M.push(L[k]),M.length===b)return M;w=x=S}}return M.push(h.slice(x)),M}]}),!g)},1290:function(t,e){function n(t){var e=typeof t;return"string"==e||"number"==e||"symbol"==e||"boolean"==e?"__proto__"!==t:null===t}t.exports=n},1304:function(t,e,n){var i=n("cd9d");function r(t){return"function"==typeof t?t:i}t.exports=r},"130f":function(t,e,n){var i=n("23e7"),r=n("da84"),o=n("2cf4"),a=!r.setImmediate||!r.clearImmediate;i({global:!0,bind:!0,enumerable:!0,forced:a},{setImmediate:o.set,clearImmediate:o.clear})},1310:function(t,e){function n(t){return null!=t&&"object"==typeof t}t.exports=n},1368:function(t,e,n){var i=n("da03"),r=function(){var t=/[^.]+$/.exec(i&&i.keys&&i.keys.IE_PROTO||"");return t?"Symbol(src)_1."+t:""}();function o(t){return!!r&&r in t}t.exports=o},"13e9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={words:{ss:["секунда","секунде","секунди"],m:["један минут","једне минуте"],mm:["минут","минуте","минута"],h:["један сат","једног сата"],hh:["сат","сата","сати"],dd:["дан","дана","дана"],MM:["месец","месеца","месеци"],yy:["година","године","година"]},correctGrammaticalCase:function(t,e){return 1===t?e[0]:t>=2&&t<=4?e[1]:e[2]},translate:function(t,n,i){var r=e.words[i];return 1===i.length?n?r[0]:r[1]:t+" "+e.correctGrammaticalCase(t,r)}},n=t.defineLocale("sr-cyrl",{months:"јануар_фебруар_март_април_мај_јун_јул_август_септембар_октобар_новембар_децембар".split("_"),monthsShort:"јан._феб._мар._апр._мај_јун_јул_авг._сеп._окт._нов._дец.".split("_"),monthsParseExact:!0,weekdays:"недеља_понедељак_уторак_среда_четвртак_петак_субота".split("_"),weekdaysShort:"нед._пон._уто._сре._чет._пет._суб.".split("_"),weekdaysMin:"не_по_ут_ср_че_пе_су".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D. M. YYYY.",LL:"D. MMMM YYYY.",LLL:"D. MMMM YYYY. H:mm",LLLL:"dddd, D. MMMM YYYY. H:mm"},calendar:{sameDay:"[данас у] LT",nextDay:"[сутра у] LT",nextWeek:function(){switch(this.day()){case 0:return"[у] [недељу] [у] LT";case 3:return"[у] [среду] [у] LT";case 6:return"[у] [суботу] [у] LT";case 1:case 2:case 4:case 5:return"[у] dddd [у] LT"}},lastDay:"[јуче у] LT",lastWeek:function(){var t=["[прошле] [недеље] [у] LT","[прошлог] [понедељка] [у] LT","[прошлог] [уторка] [у] LT","[прошле] [среде] [у] LT","[прошлог] [четвртка] [у] LT","[прошлог] [петка] [у] LT","[прошле] [суботе] [у] LT"];return t[this.day()]},sameElse:"L"},relativeTime:{future:"за %s",past:"пре %s",s:"неколико секунди",ss:e.translate,m:e.translate,mm:e.translate,h:e.translate,hh:e.translate,d:"дан",dd:e.translate,M:"месец",MM:e.translate,y:"годину",yy:e.translate},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},"14c3":function(t,e,n){var i=n("c6b6"),r=n("9263");t.exports=function(t,e){var n=t.exec;if("function"===typeof n){var o=n.call(t,e);if("object"!==typeof o)throw TypeError("RegExp exec method returned something other than an Object or null");return o}if("RegExp"!==i(t))throw TypeError("RegExp#exec called on incompatible receiver");return r.call(t,e)}},"159a":function(t,e,n){var i=n("32b3"),r=n("e2e4"),o=n("c098"),a=n("1a8c"),s=n("f4d6");function l(t,e,n,l){if(!a(t))return t;e=r(e,t);var u=-1,c=e.length,d=c-1,h=t;while(null!=h&&++u<c){var f=s(e[u]),p=n;if("__proto__"===f||"constructor"===f||"prototype"===f)return t;if(u!=d){var m=h[f];p=l?l(m,f,h):void 0,void 0===p&&(p=a(m)?m:o(e[u+1])?[]:{})}i(h,f,p),h=h[f]}return t}t.exports=l},"159b":function(t,e,n){var i=n("da84"),r=n("fdbc"),o=n("17c2"),a=n("9112");for(var s in r){var l=i[s],u=l&&l.prototype;if(u&&u.forEach!==o)try{a(u,"forEach",o)}catch(c){u.forEach=o}}},"15f3":function(t,e,n){var i=n("89d9"),r=n("8604");function o(t,e){return i(t,e,(function(e,n){return r(t,n)}))}t.exports=o},"167b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("oc-lnc",{months:{standalone:"genièr_febrièr_març_abril_mai_junh_julhet_agost_setembre_octòbre_novembre_decembre".split("_"),format:"de genièr_de febrièr_de març_d'abril_de mai_de junh_de julhet_d'agost_de setembre_d'octòbre_de novembre_de decembre".split("_"),isFormat:/D[oD]?(\s)+MMMM/},monthsShort:"gen._febr._març_abr._mai_junh_julh._ago._set._oct._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"dimenge_diluns_dimars_dimècres_dijòus_divendres_dissabte".split("_"),weekdaysShort:"dg._dl._dm._dc._dj._dv._ds.".split("_"),weekdaysMin:"dg_dl_dm_dc_dj_dv_ds".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM [de] YYYY",ll:"D MMM YYYY",LLL:"D MMMM [de] YYYY [a] H:mm",lll:"D MMM YYYY, H:mm",LLLL:"dddd D MMMM [de] YYYY [a] H:mm",llll:"ddd D MMM YYYY, H:mm"},calendar:{sameDay:"[uèi a] LT",nextDay:"[deman a] LT",nextWeek:"dddd [a] LT",lastDay:"[ièr a] LT",lastWeek:"dddd [passat a] LT",sameElse:"L"},relativeTime:{future:"d'aquí %s",past:"fa %s",s:"unas segondas",ss:"%d segondas",m:"una minuta",mm:"%d minutas",h:"una ora",hh:"%d oras",d:"un jorn",dd:"%d jorns",M:"un mes",MM:"%d meses",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(r|n|t|è|a)/,ordinal:function(t,e){var n=1===t?"r":2===t?"n":3===t?"r":4===t?"t":"è";return"w"!==e&&"W"!==e||(n="a"),t+n},week:{dow:1,doy:4}});return e}))},1693:function(t,e,n){var i=n("afb9"),r=n("9934");function o(t){return null==t?[]:i(t,r(t))}t.exports=o},1766:function(t,e,n){"use strict";e["a"]={name:"RenderCell",functional:!0,props:{render:Function},render:(t,e)=>e.props.render(t)}},"17c2":function(t,e,n){"use strict";var i=n("b727").forEach,r=n("a640"),o=n("ae40"),a=r("forEach"),s=o("forEach");t.exports=a&&s?[].forEach:function(t){return i(this,t,arguments.length>1?arguments[1]:void 0)}},1838:function(t,e,n){var i=n("c05f"),r=n("9b02"),o=n("8604"),a=n("f608"),s=n("08cc"),l=n("20ec"),u=n("f4d6"),c=1,d=2;function h(t,e){return a(t)&&s(e)?l(u(t),e):function(n){var a=r(n,t);return void 0===a&&a===e?o(n,t):i(e,a,c|d)}}t.exports=h},"18d2":function(t,e,n){"use strict";var i=n("18e9");t.exports=function(t){t=t||{};var e=t.reporter,n=t.batchProcessor,r=t.stateHandler.getState;if(!e)throw new Error("Missing required dependency: reporter.");function o(t,e){function n(){e(t)}if(i.isIE(8))r(t).object={proxy:n},t.attachEvent("onresize",n);else{var o=l(t);if(!o)throw new Error("Element is not detectable by this strategy.");o.contentDocument.defaultView.addEventListener("resize",n)}}function a(e){var n=t.important?" !important; ":"; ";return(e.join(n)+n).trim()}function s(t,o,s){s||(s=o,o=t,t=null),t=t||{};t.debug;function l(o,s){var l=a(["display: block","position: absolute","top: 0","left: 0","width: 100%","height: 100%","border: none","padding: 0","margin: 0","opacity: 0","z-index: -1000","pointer-events: none"]),u=!1,c=window.getComputedStyle(o),d=o.offsetWidth,h=o.offsetHeight;function f(){function n(){if("static"===c.position){o.style.setProperty("position","relative",t.important?"important":"");var n=function(e,n,i,r){function o(t){return t.replace(/[^-\d\.]/g,"")}var a=i[r];"auto"!==a&&"0"!==o(a)&&(e.warn("An element that is positioned static has style."+r+"="+a+" which is ignored due to the static positioning. The element will need to be positioned relative, so the style."+r+" will be set to 0. Element: ",n),n.style.setProperty(r,"0",t.important?"important":""))};n(e,o,c,"top"),n(e,o,c,"right"),n(e,o,c,"bottom"),n(e,o,c,"left")}}function a(){function t(e,n){if(!e.contentDocument){var i=r(e);return i.checkForObjectDocumentTimeoutId&&window.clearTimeout(i.checkForObjectDocumentTimeoutId),void(i.checkForObjectDocumentTimeoutId=setTimeout((function(){i.checkForObjectDocumentTimeoutId=0,t(e,n)}),100))}n(e.contentDocument)}u||n();var e=this;t(e,(function(t){s(o)}))}""!==c.position&&(n(c),u=!0);var d=document.createElement("object");d.style.cssText=l,d.tabIndex=-1,d.type="text/html",d.setAttribute("aria-hidden","true"),d.onload=a,i.isIE()||(d.data="about:blank"),r(o)&&(o.appendChild(d),r(o).object=d,i.isIE()&&(d.data="about:blank"))}r(o).startSize={width:d,height:h},n?n.add(f):f()}i.isIE(8)?s(o):l(o,s)}function l(t){return r(t).object}function u(t){if(r(t)){var e=l(t);e&&(i.isIE(8)?t.detachEvent("onresize",e.proxy):t.removeChild(e),r(t).checkForObjectDocumentTimeoutId&&window.clearTimeout(r(t).checkForObjectDocumentTimeoutId),delete r(t).object)}}return{makeDetectable:s,addListener:o,uninstall:u}}},"18d8":function(t,e,n){var i=n("234d"),r=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,o=/\\(\\)?/g,a=i((function(t){var e=[];return 46===t.charCodeAt(0)&&e.push(""),t.replace(r,(function(t,n,i,r){e.push(i?r.replace(o,"$1"):n||t)})),e}));t.exports=a},"18e9":function(t,e,n){"use strict";var i=t.exports={};i.isIE=function(t){function e(){var t=navigator.userAgent.toLowerCase();return-1!==t.indexOf("msie")||-1!==t.indexOf("trident")||-1!==t.indexOf(" edge/")}if(!e())return!1;if(!t)return!0;var n=function(){var t,e=3,n=document.createElement("div"),i=n.getElementsByTagName("i");do{n.innerHTML="\x3c!--[if gt IE "+ ++e+"]><i></i><![endif]--\x3e"}while(i[0]);return e>4?e:t}();return t===n},i.isLegacyOpera=function(){return!!window.opera}},"18f6":function(t,e,n){var i=n("30e0"),r=n("1304"),o=n("9934");function a(t,e){return null==t?t:i(t,r(e),o)}t.exports=a},"193b":function(t,e,n){var i=n("159a");function r(t,e,n,r){return r="function"==typeof r?r:void 0,null==t?t:i(t,e,n,r)}t.exports=r},"19aa":function(t,e){t.exports=function(t,e,n){if(!(t instanceof e))throw TypeError("Incorrect "+(n?n+" ":"")+"invocation");return t}},"1a2d":function(t,e,n){var i=n("42a2"),r=n("1310"),o="[object Map]";function a(t){return r(t)&&i(t)==o}t.exports=a},"1a8c":function(t,e){function n(t){var e=typeof t;return null!=t&&("object"==e||"function"==e)}t.exports=n},"1b45":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("mt",{months:"Jannar_Frar_Marzu_April_Mejju_Ġunju_Lulju_Awwissu_Settembru_Ottubru_Novembru_Diċembru".split("_"),monthsShort:"Jan_Fra_Mar_Apr_Mej_Ġun_Lul_Aww_Set_Ott_Nov_Diċ".split("_"),weekdays:"Il-Ħadd_It-Tnejn_It-Tlieta_L-Erbgħa_Il-Ħamis_Il-Ġimgħa_Is-Sibt".split("_"),weekdaysShort:"Ħad_Tne_Tli_Erb_Ħam_Ġim_Sib".split("_"),weekdaysMin:"Ħa_Tn_Tl_Er_Ħa_Ġi_Si".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Illum fil-]LT",nextDay:"[Għada fil-]LT",nextWeek:"dddd [fil-]LT",lastDay:"[Il-bieraħ fil-]LT",lastWeek:"dddd [li għadda] [fil-]LT",sameElse:"L"},relativeTime:{future:"f’ %s",past:"%s ilu",s:"ftit sekondi",ss:"%d sekondi",m:"minuta",mm:"%d minuti",h:"siegħa",hh:"%d siegħat",d:"ġurnata",dd:"%d ġranet",M:"xahar",MM:"%d xhur",y:"sena",yy:"%d sni"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},"1bac":function(t,e,n){var i=n("7d1f"),r=n("a029"),o=n("9934");function a(t){return i(t,o,r)}t.exports=a},"1be4":function(t,e,n){var i=n("d066");t.exports=i("document","documentElement")},"1c0b":function(t,e){t.exports=function(t){if("function"!=typeof t)throw TypeError(String(t)+" is not a function");return t}},"1c3c":function(t,e,n){var i=n("9e69"),r=n("2474"),o=n("9638"),a=n("a2be"),s=n("edfa"),l=n("ac41"),u=1,c=2,d="[object Boolean]",h="[object Date]",f="[object Error]",p="[object Map]",m="[object Number]",g="[object RegExp]",y="[object Set]",v="[object String]",_="[object Symbol]",b="[object ArrayBuffer]",x="[object DataView]",w=i?i.prototype:void 0,M=w?w.valueOf:void 0;function S(t,e,n,i,w,S,L){switch(n){case x:if(t.byteLength!=e.byteLength||t.byteOffset!=e.byteOffset)return!1;t=t.buffer,e=e.buffer;case b:return!(t.byteLength!=e.byteLength||!S(new r(t),new r(e)));case d:case h:case m:return o(+t,+e);case f:return t.name==e.name&&t.message==e.message;case g:case v:return t==e+"";case p:var k=s;case y:var T=i&u;if(k||(k=l),t.size!=e.size&&!T)return!1;var D=L.get(t);if(D)return D==e;i|=c,L.set(t,e);var C=a(k(t),k(e),i,w,S,L);return L["delete"](t),C;case _:if(M)return M.call(t)==M.call(e)}return!1}t.exports=S},"1c7e":function(t,e,n){var i=n("b622"),r=i("iterator"),o=!1;try{var a=0,s={next:function(){return{done:!!a++}},return:function(){o=!0}};s[r]=function(){return this},Array.from(s,(function(){throw 2}))}catch(l){}t.exports=function(t,e){if(!e&&!o)return!1;var n=!1;try{var i={};i[r]=function(){return{next:function(){return{done:n=!0}}}},t(i)}catch(l){}return n}},"1cdc":function(t,e,n){var i=n("342f");t.exports=/(iphone|ipod|ipad).*applewebkit/i.test(i)},"1cec":function(t,e,n){var i=n("0b07"),r=n("2b3e"),o=i(r,"Promise");t.exports=o},"1cfd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"1",2:"2",3:"3",4:"4",5:"5",6:"6",7:"7",8:"8",9:"9",0:"0"},n=function(t){return 0===t?0:1===t?1:2===t?2:t%100>=3&&t%100<=10?3:t%100>=11?4:5},i={s:["أقل من ثانية","ثانية واحدة",["ثانيتان","ثانيتين"],"%d ثوان","%d ثانية","%d ثانية"],m:["أقل من دقيقة","دقيقة واحدة",["دقيقتان","دقيقتين"],"%d دقائق","%d دقيقة","%d دقيقة"],h:["أقل من ساعة","ساعة واحدة",["ساعتان","ساعتين"],"%d ساعات","%d ساعة","%d ساعة"],d:["أقل من يوم","يوم واحد",["يومان","يومين"],"%d أيام","%d يومًا","%d يوم"],M:["أقل من شهر","شهر واحد",["شهران","شهرين"],"%d أشهر","%d شهرا","%d شهر"],y:["أقل من عام","عام واحد",["عامان","عامين"],"%d أعوام","%d عامًا","%d عام"]},r=function(t){return function(e,r,o,a){var s=n(e),l=i[t][n(e)];return 2===s&&(l=l[r?0:1]),l.replace(/%d/i,e)}},o=["يناير","فبراير","مارس","أبريل","مايو","يونيو","يوليو","أغسطس","سبتمبر","أكتوبر","نوفمبر","ديسمبر"],a=t.defineLocale("ar-ly",{months:o,monthsShort:o,weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/‏M/‏YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم عند الساعة] LT",nextDay:"[غدًا عند الساعة] LT",nextWeek:"dddd [عند الساعة] LT",lastDay:"[أمس عند الساعة] LT",lastWeek:"dddd [عند الساعة] LT",sameElse:"L"},relativeTime:{future:"بعد %s",past:"منذ %s",s:r("s"),ss:r("s"),m:r("m"),mm:r("m"),h:r("h"),hh:r("h"),d:r("d"),dd:r("d"),M:r("M"),MM:r("M"),y:r("y"),yy:r("y")},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:6,doy:12}});return a}))},"1d2b":function(t,e,n){"use strict";t.exports=function(t,e){return function(){for(var n=new Array(arguments.length),i=0;i<n.length;i++)n[i]=arguments[i];return t.apply(e,n)}}},"1d80":function(t,e){t.exports=function(t){if(void 0==t)throw TypeError("Can't call method on "+t);return t}},"1dde":function(t,e,n){var i=n("d039"),r=n("b622"),o=n("2d00"),a=r("species");t.exports=function(t){return o>=51||!i((function(){var e=[],n=e.constructor={};return n[a]=function(){return{foo:1}},1!==e[t](Boolean).foo}))}},"1efc":function(t,e){function n(t){var e=this.has(t)&&delete this.__data__[t];return this.size-=e?1:0,e}t.exports=n},"1fc1":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e){var n=t.split("_");return e%10===1&&e%100!==11?n[0]:e%10>=2&&e%10<=4&&(e%100<10||e%100>=20)?n[1]:n[2]}function n(t,n,i){var r={ss:n?"секунда_секунды_секунд":"секунду_секунды_секунд",mm:n?"хвіліна_хвіліны_хвілін":"хвіліну_хвіліны_хвілін",hh:n?"гадзіна_гадзіны_гадзін":"гадзіну_гадзіны_гадзін",dd:"дзень_дні_дзён",MM:"месяц_месяцы_месяцаў",yy:"год_гады_гадоў"};return"m"===i?n?"хвіліна":"хвіліну":"h"===i?n?"гадзіна":"гадзіну":t+" "+e(r[i],+t)}var i=t.defineLocale("be",{months:{format:"студзеня_лютага_сакавіка_красавіка_траўня_чэрвеня_ліпеня_жніўня_верасня_кастрычніка_лістапада_снежня".split("_"),standalone:"студзень_люты_сакавік_красавік_травень_чэрвень_ліпень_жнівень_верасень_кастрычнік_лістапад_снежань".split("_")},monthsShort:"студ_лют_сак_крас_трав_чэрв_ліп_жнів_вер_каст_ліст_снеж".split("_"),weekdays:{format:"нядзелю_панядзелак_аўторак_сераду_чацвер_пятніцу_суботу".split("_"),standalone:"нядзеля_панядзелак_аўторак_серада_чацвер_пятніца_субота".split("_"),isFormat:/\[ ?[Ууў] ?(?:мінулую|наступную)? ?\] ?dddd/},weekdaysShort:"нд_пн_ат_ср_чц_пт_сб".split("_"),weekdaysMin:"нд_пн_ат_ср_чц_пт_сб".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY г.",LLL:"D MMMM YYYY г., HH:mm",LLLL:"dddd, D MMMM YYYY г., HH:mm"},calendar:{sameDay:"[Сёння ў] LT",nextDay:"[Заўтра ў] LT",lastDay:"[Учора ў] LT",nextWeek:function(){return"[У] dddd [ў] LT"},lastWeek:function(){switch(this.day()){case 0:case 3:case 5:case 6:return"[У мінулую] dddd [ў] LT";case 1:case 2:case 4:return"[У мінулы] dddd [ў] LT"}},sameElse:"L"},relativeTime:{future:"праз %s",past:"%s таму",s:"некалькі секунд",m:n,mm:n,h:n,hh:n,d:"дзень",dd:n,M:"месяц",MM:n,y:"год",yy:n},meridiemParse:/ночы|раніцы|дня|вечара/,isPM:function(t){return/^(дня|вечара)$/.test(t)},meridiem:function(t,e,n){return t<4?"ночы":t<12?"раніцы":t<17?"дня":"вечара"},dayOfMonthOrdinalParse:/\d{1,2}-(і|ы|га)/,ordinal:function(t,e){switch(e){case"M":case"d":case"DDD":case"w":case"W":return t%10!==2&&t%10!==3||t%100===12||t%100===13?t+"-ы":t+"-і";case"D":return t+"-га";default:return t}},week:{dow:1,doy:7}});return i}))},"1fc8":function(t,e,n){var i=n("4245");function r(t,e){var n=i(this,t),r=n.size;return n.set(t,e),this.size+=n.size==r?0:1,this}t.exports=r},"201b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ka",{months:"იანვარი_თებერვალი_მარტი_აპრილი_მაისი_ივნისი_ივლისი_აგვისტო_სექტემბერი_ოქტომბერი_ნოემბერი_დეკემბერი".split("_"),monthsShort:"იან_თებ_მარ_აპრ_მაი_ივნ_ივლ_აგვ_სექ_ოქტ_ნოე_დეკ".split("_"),weekdays:{standalone:"კვირა_ორშაბათი_სამშაბათი_ოთხშაბათი_ხუთშაბათი_პარასკევი_შაბათი".split("_"),format:"კვირას_ორშაბათს_სამშაბათს_ოთხშაბათს_ხუთშაბათს_პარასკევს_შაბათს".split("_"),isFormat:/(წინა|შემდეგ)/},weekdaysShort:"კვი_ორშ_სამ_ოთხ_ხუთ_პარ_შაბ".split("_"),weekdaysMin:"კვ_ორ_სა_ოთ_ხუ_პა_შა".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[დღეს] LT[-ზე]",nextDay:"[ხვალ] LT[-ზე]",lastDay:"[გუშინ] LT[-ზე]",nextWeek:"[შემდეგ] dddd LT[-ზე]",lastWeek:"[წინა] dddd LT-ზე",sameElse:"L"},relativeTime:{future:function(t){return t.replace(/(წამ|წუთ|საათ|წელ|დღ|თვ)(ი|ე)/,(function(t,e,n){return"ი"===n?e+"ში":e+n+"ში"}))},past:function(t){return/(წამი|წუთი|საათი|დღე|თვე)/.test(t)?t.replace(/(ი|ე)$/,"ის წინ"):/წელი/.test(t)?t.replace(/წელი$/,"წლის წინ"):t},s:"რამდენიმე წამი",ss:"%d წამი",m:"წუთი",mm:"%d წუთი",h:"საათი",hh:"%d საათი",d:"დღე",dd:"%d დღე",M:"თვე",MM:"%d თვე",y:"წელი",yy:"%d წელი"},dayOfMonthOrdinalParse:/0|1-ლი|მე-\d{1,2}|\d{1,2}-ე/,ordinal:function(t){return 0===t?t:1===t?t+"-ლი":t<20||t<=100&&t%20===0||t%100===0?"მე-"+t:t+"-ე"},week:{dow:1,doy:7}});return e}))},"20ec":function(t,e){function n(t,e){return function(n){return null!=n&&(n[t]===e&&(void 0!==e||t in Object(n)))}}t.exports=n},2266:function(t,e,n){var i=n("825a"),r=n("e95a"),o=n("50c4"),a=n("0366"),s=n("35a1"),l=n("2a62"),u=function(t,e){this.stopped=t,this.result=e};t.exports=function(t,e,n){var c,d,h,f,p,m,g,y=n&&n.that,v=!(!n||!n.AS_ENTRIES),_=!(!n||!n.IS_ITERATOR),b=!(!n||!n.INTERRUPTED),x=a(e,y,1+v+b),w=function(t){return c&&l(c),new u(!0,t)},M=function(t){return v?(i(t),b?x(t[0],t[1],w):x(t[0],t[1])):b?x(t,w):x(t)};if(_)c=t;else{if(d=s(t),"function"!=typeof d)throw TypeError("Target is not iterable");if(r(d)){for(h=0,f=o(t.length);f>h;h++)if(p=M(t[h]),p&&p instanceof u)return p;return new u(!1)}c=d.call(t)}m=c.next;while(!(g=m.call(c)).done){try{p=M(g.value)}catch(S){throw l(c),S}if("object"==typeof p&&p&&p instanceof u)return p}return new u(!1)}},2286:function(t,e,n){var i=n("85e3"),r=Math.max;function o(t,e,n){return e=r(void 0===e?t.length-1:e,0),function(){var o=arguments,a=-1,s=r(o.length-e,0),l=Array(s);while(++a<s)l[a]=o[e+a];a=-1;var u=Array(e+1);while(++a<e)u[a]=o[a];return u[e]=n(l),i(t,this,u)}}t.exports=o},"22f8":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ko",{months:"1월_2월_3월_4월_5월_6월_7월_8월_9월_10월_11월_12월".split("_"),monthsShort:"1월_2월_3월_4월_5월_6월_7월_8월_9월_10월_11월_12월".split("_"),weekdays:"일요일_월요일_화요일_수요일_목요일_금요일_토요일".split("_"),weekdaysShort:"일_월_화_수_목_금_토".split("_"),weekdaysMin:"일_월_화_수_목_금_토".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"YYYY.MM.DD.",LL:"YYYY년 MMMM D일",LLL:"YYYY년 MMMM D일 A h:mm",LLLL:"YYYY년 MMMM D일 dddd A h:mm",l:"YYYY.MM.DD.",ll:"YYYY년 MMMM D일",lll:"YYYY년 MMMM D일 A h:mm",llll:"YYYY년 MMMM D일 dddd A h:mm"},calendar:{sameDay:"오늘 LT",nextDay:"내일 LT",nextWeek:"dddd LT",lastDay:"어제 LT",lastWeek:"지난주 dddd LT",sameElse:"L"},relativeTime:{future:"%s 후",past:"%s 전",s:"몇 초",ss:"%d초",m:"1분",mm:"%d분",h:"한 시간",hh:"%d시간",d:"하루",dd:"%d일",M:"한 달",MM:"%d달",y:"일 년",yy:"%d년"},dayOfMonthOrdinalParse:/\d{1,2}(일|월|주)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"일";case"M":return t+"월";case"w":case"W":return t+"주";default:return t}},meridiemParse:/오전|오후/,isPM:function(t){return"오후"===t},meridiem:function(t,e,n){return t<12?"오전":"오후"}});return e}))},"22fe":function(t,e,n){},"234d":function(t,e,n){var i=n("e380"),r=500;function o(t){var e=i(t,(function(t){return n.size===r&&n.clear(),t})),n=e.cache;return e}t.exports=o},"23cb":function(t,e,n){var i=n("a691"),r=Math.max,o=Math.min;t.exports=function(t,e){var n=i(t);return n<0?r(n+e,0):o(n,e)}},"23e2":function(t,e,n){var i=n("242e");function r(t,e,n,r){return i(t,(function(t,i,o){e(r,n(t),i,o)})),r}t.exports=r},"23e7":function(t,e,n){var i=n("da84"),r=n("06cf").f,o=n("9112"),a=n("6eeb"),s=n("ce4e"),l=n("e893"),u=n("94ca");t.exports=function(t,e){var n,c,d,h,f,p,m=t.target,g=t.global,y=t.stat;if(c=g?i:y?i[m]||s(m,{}):(i[m]||{}).prototype,c)for(d in e){if(f=e[d],t.noTargetGet?(p=r(c,d),h=p&&p.value):h=c[d],n=u(g?d:m+(y?".":"#")+d,t.forced),!n&&void 0!==h){if(typeof f===typeof h)continue;l(f,h)}(t.sham||h&&h.sham)&&o(f,"sham",!0),a(c,d,f,t)}}},2411:function(t,e,n){var i=n("f909"),r=n("2ec1"),o=r((function(t,e,n,r){i(t,e,n,r)}));t.exports=o},"241c":function(t,e,n){var i=n("ca84"),r=n("7839"),o=r.concat("length","prototype");e.f=Object.getOwnPropertyNames||function(t){return i(t,o)}},2421:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"١",2:"٢",3:"٣",4:"٤",5:"٥",6:"٦",7:"٧",8:"٨",9:"٩",0:"٠"},n={"١":"1","٢":"2","٣":"3","٤":"4","٥":"5","٦":"6","٧":"7","٨":"8","٩":"9","٠":"0"},i=["کانونی دووەم","شوبات","ئازار","نیسان","ئایار","حوزەیران","تەمموز","ئاب","ئەیلوول","تشرینی یەكەم","تشرینی دووەم","كانونی یەکەم"],r=t.defineLocale("ku",{months:i,monthsShort:i,weekdays:"یه‌كشه‌ممه‌_دووشه‌ممه‌_سێشه‌ممه‌_چوارشه‌ممه‌_پێنجشه‌ممه‌_هه‌ینی_شه‌ممه‌".split("_"),weekdaysShort:"یه‌كشه‌م_دووشه‌م_سێشه‌م_چوارشه‌م_پێنجشه‌م_هه‌ینی_شه‌ممه‌".split("_"),weekdaysMin:"ی_د_س_چ_پ_ه_ش".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},meridiemParse:/ئێواره‌|به‌یانی/,isPM:function(t){return/ئێواره‌/.test(t)},meridiem:function(t,e,n){return t<12?"به‌یانی":"ئێواره‌"},calendar:{sameDay:"[ئه‌مرۆ كاتژمێر] LT",nextDay:"[به‌یانی كاتژمێر] LT",nextWeek:"dddd [كاتژمێر] LT",lastDay:"[دوێنێ كاتژمێر] LT",lastWeek:"dddd [كاتژمێر] LT",sameElse:"L"},relativeTime:{future:"له‌ %s",past:"%s",s:"چه‌ند چركه‌یه‌ك",ss:"چركه‌ %d",m:"یه‌ك خوله‌ك",mm:"%d خوله‌ك",h:"یه‌ك كاتژمێر",hh:"%d كاتژمێر",d:"یه‌ك ڕۆژ",dd:"%d ڕۆژ",M:"یه‌ك مانگ",MM:"%d مانگ",y:"یه‌ك ساڵ",yy:"%d ساڵ"},preparse:function(t){return t.replace(/[١٢٣٤٥٦٧٨٩٠]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:6,doy:12}});return r}))},"242e":function(t,e,n){var i=n("72af"),r=n("ec69");function o(t,e){return t&&i(t,e,r)}t.exports=o},2430:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes,attrs:{name:t.name}},[t._t("default")],2)},r=[],o=(n("99af"),n("4160"),n("a9e3"),n("159b"),n("ade3")),a=n("3a27"),s=n("2507"),l="ivu-radio-group",u=0,c=Date.now(),d=function(){return"ivuRadioGroup_".concat(c,"_").concat(u++)},h={name:"RadioGroup",mixins:[s["a"]],props:{value:{type:[String,Number],default:""},size:{validator:function(t){return Object(a["i"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},type:{validator:function(t){return Object(a["i"])(t,["button"])}},vertical:{type:Boolean,default:!1},name:{type:String,default:d}},data:function(){return{currentValue:this.value,childrens:[]}},computed:{classes:function(){var t;return["".concat(l),(t={},Object(o["a"])(t,"".concat(l,"-").concat(this.size),!!this.size),Object(o["a"])(t,"ivu-radio-".concat(this.size),!!this.size),Object(o["a"])(t,"".concat(l,"-").concat(this.type),!!this.type),Object(o["a"])(t,"".concat(l,"-vertical"),this.vertical),t)]}},mounted:function(){this.updateValue()},methods:{updateValue:function(){var t=this;this.childrens=Object(a["f"])(this,"Radio"),this.childrens&&this.childrens.forEach((function(e){e.currentValue=t.currentValue===e.label,e.group=!0}))},change:function(t){this.currentValue=t.value,this.updateValue(),this.$emit("input",t.value),this.$emit("on-change",t.value),this.dispatch("FormItem","on-form-change",t.value)}},watch:{value:function(){var t=this;this.currentValue!==this.value&&(this.currentValue=this.value,this.$nextTick((function(){t.updateValue()})))}}},f=h,p=n("2877"),m=Object(p["a"])(f,i,r,!1,null,null,null);e["a"]=m.exports},2444:function(t,e,n){"use strict";(function(e){var i=n("c532"),r=n("c8af"),o={"Content-Type":"application/x-www-form-urlencoded"};function a(t,e){!i.isUndefined(t)&&i.isUndefined(t["Content-Type"])&&(t["Content-Type"]=e)}function s(){var t;return("undefined"!==typeof XMLHttpRequest||"undefined"!==typeof e&&"[object process]"===Object.prototype.toString.call(e))&&(t=n("b50d")),t}var l={adapter:s(),transformRequest:[function(t,e){return r(e,"Accept"),r(e,"Content-Type"),i.isFormData(t)||i.isArrayBuffer(t)||i.isBuffer(t)||i.isStream(t)||i.isFile(t)||i.isBlob(t)?t:i.isArrayBufferView(t)?t.buffer:i.isURLSearchParams(t)?(a(e,"application/x-www-form-urlencoded;charset=utf-8"),t.toString()):i.isObject(t)?(a(e,"application/json;charset=utf-8"),JSON.stringify(t)):t}],transformResponse:[function(t){if("string"===typeof t)try{t=JSON.parse(t)}catch(e){}return t}],timeout:0,xsrfCookieName:"XSRF-TOKEN",xsrfHeaderName:"X-XSRF-TOKEN",maxContentLength:-1,maxBodyLength:-1,validateStatus:function(t){return t>=200&&t<300},headers:{common:{Accept:"application/json, text/plain, */*"}}};i.forEach(["delete","get","head"],(function(t){l.headers[t]={}})),i.forEach(["post","put","patch"],(function(t){l.headers[t]=i.merge(o)})),t.exports=l}).call(this,n("4362"))},"245a":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("li",{class:t.classes,on:{click:t.handleClick}},[t._t("default")],2)},r=[],o=(n("b0c0"),n("a9e3"),n("ade3")),a=n("3a27"),s="ivu-dropdown-item",l={name:"DropdownItem",props:{name:{type:[String,Number]},disabled:{type:Boolean,default:!1},selected:{type:Boolean,default:!1},divided:{type:Boolean,default:!1}},computed:{classes:function(){var t;return["".concat(s),(t={},Object(o["a"])(t,"".concat(s,"-disabled"),this.disabled),Object(o["a"])(t,"".concat(s,"-selected"),this.selected),Object(o["a"])(t,"".concat(s,"-divided"),this.divided),t)]}},methods:{handleClick:function(){if(!this.disabled){var t=Object(a["e"])(this,"Dropdown"),e=this.$parent&&"Dropdown"===this.$parent.$options.name;e?this.$parent.$emit("on-haschild-click"):t&&"Dropdown"===t.$options.name&&t.$emit("on-hover-click"),t.$emit("on-click",this.name)}}}},u=l,c=n("2877"),d=Object(c["a"])(u,i,r,!1,null,null,null);e["a"]=d.exports},2474:function(t,e,n){var i=n("2b3e"),r=i.Uint8Array;t.exports=r},2478:function(t,e,n){var i=n("4245");function r(t){return i(this,t).get(t)}t.exports=r},2507:function(t,e,n){"use strict";function i(t,e,n){this.$children.forEach(r=>{const o=r.$options.name;o===t?r.$emit.apply(r,[e].concat(n)):i.apply(r,[t,e].concat([n]))})}e["a"]={methods:{dispatch(t,e,n){let i=this.$parent||this.$root,r=i.$options.name;while(i&&(!r||r!==t))i=i.$parent,i&&(r=i.$options.name);i&&i.$emit.apply(i,[e].concat(n))},broadcast(t,e,n){i.call(this,t,e,n)}}}},2524:function(t,e,n){var i=n("6044"),r="__lodash_hash_undefined__";function o(t,e){var n=this.__data__;return this.size+=this.has(t)?0:1,n[t]=i&&void 0===e?r:e,this}t.exports=o},2532:function(t,e,n){"use strict";var i=n("23e7"),r=n("5a34"),o=n("1d80"),a=n("ab13");i({target:"String",proto:!0,forced:!a("includes")},{includes:function(t){return!!~String(o(this)).indexOf(r(t),arguments.length>1?arguments[1]:void 0)}})},"253c":function(t,e,n){var i=n("3729"),r=n("1310"),o="[object Arguments]";function a(t){return r(t)&&i(t)==o}t.exports=a},2554:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n){var i=t+" ";switch(n){case"ss":return i+=1===t?"sekunda":2===t||3===t||4===t?"sekunde":"sekundi",i;case"m":return e?"jedna minuta":"jedne minute";case"mm":return i+=1===t?"minuta":2===t||3===t||4===t?"minute":"minuta",i;case"h":return e?"jedan sat":"jednog sata";case"hh":return i+=1===t?"sat":2===t||3===t||4===t?"sata":"sati",i;case"dd":return i+=1===t?"dan":"dana",i;case"MM":return i+=1===t?"mjesec":2===t||3===t||4===t?"mjeseca":"mjeseci",i;case"yy":return i+=1===t?"godina":2===t||3===t||4===t?"godine":"godina",i}}var n=t.defineLocale("bs",{months:"januar_februar_mart_april_maj_juni_juli_august_septembar_oktobar_novembar_decembar".split("_"),monthsShort:"jan._feb._mar._apr._maj._jun._jul._aug._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedjelja_ponedjeljak_utorak_srijeda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sri._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedjelju] [u] LT";case 3:return"[u] [srijedu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[jučer u] LT",lastWeek:function(){switch(this.day()){case 0:case 3:return"[prošlu] dddd [u] LT";case 6:return"[prošle] [subote] [u] LT";case 1:case 2:case 4:case 5:return"[prošli] dddd [u] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"prije %s",s:"par sekundi",ss:e,m:e,mm:e,h:e,hh:e,d:"dan",dd:e,M:"mjesec",MM:e,y:"godinu",yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},2593:function(t,e,n){var i=n("15f3"),r=n("c6cf"),o=r((function(t,e){return null==t?{}:i(t,e)}));t.exports=o},"25eb":function(t,e,n){var i=n("23e7"),r=n("c20d");i({target:"Number",stat:!0,forced:Number.parseInt!=r},{parseInt:r})},"25f0":function(t,e,n){"use strict";var i=n("6eeb"),r=n("825a"),o=n("d039"),a=n("ad6d"),s="toString",l=RegExp.prototype,u=l[s],c=o((function(){return"/a/b"!=u.call({source:"a",flags:"b"})})),d=u.name!=s;(c||d)&&i(RegExp.prototype,s,(function(){var t=r(this),e=String(t.source),n=t.flags,i=String(void 0===n&&t instanceof RegExp&&!("flags"in l)?a.call(t):n);return"/"+e+"/"+i}),{unsafe:!0})},2626:function(t,e,n){"use strict";var i=n("d066"),r=n("9bf2"),o=n("b622"),a=n("83ab"),s=o("species");t.exports=function(t){var e=i(t),n=r.f;a&&e&&!e[s]&&n(e,s,{configurable:!0,get:function(){return this}})}},"266a":function(t,e,n){var i=n("7948");function r(t,e){return i(e,(function(e){return t[e]}))}t.exports=r},"26e8":function(t,e){function n(t,e){return null!=t&&e in Object(t)}t.exports=n},"26f9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={ss:"sekundė_sekundžių_sekundes",m:"minutė_minutės_minutę",mm:"minutės_minučių_minutes",h:"valanda_valandos_valandą",hh:"valandos_valandų_valandas",d:"diena_dienos_dieną",dd:"dienos_dienų_dienas",M:"mėnuo_mėnesio_mėnesį",MM:"mėnesiai_mėnesių_mėnesius",y:"metai_metų_metus",yy:"metai_metų_metus"};function n(t,e,n,i){return e?"kelios sekundės":i?"kelių sekundžių":"kelias sekundes"}function i(t,e,n,i){return e?o(n)[0]:i?o(n)[1]:o(n)[2]}function r(t){return t%10===0||t>10&&t<20}function o(t){return e[t].split("_")}function a(t,e,n,a){var s=t+" ";return 1===t?s+i(t,e,n[0],a):e?s+(r(t)?o(n)[1]:o(n)[0]):a?s+o(n)[1]:s+(r(t)?o(n)[1]:o(n)[2])}var s=t.defineLocale("lt",{months:{format:"sausio_vasario_kovo_balandžio_gegužės_birželio_liepos_rugpjūčio_rugsėjo_spalio_lapkričio_gruodžio".split("_"),standalone:"sausis_vasaris_kovas_balandis_gegužė_birželis_liepa_rugpjūtis_rugsėjis_spalis_lapkritis_gruodis".split("_"),isFormat:/D[oD]?(\[[^\[\]]*\]|\s)+MMMM?|MMMM?(\[[^\[\]]*\]|\s)+D[oD]?/},monthsShort:"sau_vas_kov_bal_geg_bir_lie_rgp_rgs_spa_lap_grd".split("_"),weekdays:{format:"sekmadienį_pirmadienį_antradienį_trečiadienį_ketvirtadienį_penktadienį_šeštadienį".split("_"),standalone:"sekmadienis_pirmadienis_antradienis_trečiadienis_ketvirtadienis_penktadienis_šeštadienis".split("_"),isFormat:/dddd HH:mm/},weekdaysShort:"Sek_Pir_Ant_Tre_Ket_Pen_Šeš".split("_"),weekdaysMin:"S_P_A_T_K_Pn_Š".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY [m.] MMMM D [d.]",LLL:"YYYY [m.] MMMM D [d.], HH:mm [val.]",LLLL:"YYYY [m.] MMMM D [d.], dddd, HH:mm [val.]",l:"YYYY-MM-DD",ll:"YYYY [m.] MMMM D [d.]",lll:"YYYY [m.] MMMM D [d.], HH:mm [val.]",llll:"YYYY [m.] MMMM D [d.], ddd, HH:mm [val.]"},calendar:{sameDay:"[Šiandien] LT",nextDay:"[Rytoj] LT",nextWeek:"dddd LT",lastDay:"[Vakar] LT",lastWeek:"[Praėjusį] dddd LT",sameElse:"L"},relativeTime:{future:"po %s",past:"prieš %s",s:n,ss:a,m:i,mm:a,h:i,hh:a,d:i,dd:a,M:i,MM:a,y:i,yy:a},dayOfMonthOrdinalParse:/\d{1,2}-oji/,ordinal:function(t){return t+"-oji"},week:{dow:1,doy:4}});return s}))},"27f3":function(t,e,n){var i=n("72f0"),r=n("43ad"),o=n("cd9d"),a=Object.prototype,s=a.toString,l=r((function(t,e,n){null!=e&&"function"!=typeof e.toString&&(e=s.call(e)),t[e]=n}),i(o));t.exports=l},2877:function(t,e,n){"use strict";function i(t,e,n,i,r,o,a,s){var l,u="function"===typeof t?t.options:t;if(e&&(u.render=e,u.staticRenderFns=n,u._compiled=!0),i&&(u.functional=!0),o&&(u._scopeId="data-v-"+o),a?(l=function(t){t=t||this.$vnode&&this.$vnode.ssrContext||this.parent&&this.parent.$vnode&&this.parent.$vnode.ssrContext,t||"undefined"===typeof __VUE_SSR_CONTEXT__||(t=__VUE_SSR_CONTEXT__),r&&r.call(this,t),t&&t._registeredComponents&&t._registeredComponents.add(a)},u._ssrRegister=l):r&&(l=s?function(){r.call(this,(u.functional?this.parent:this).$root.$options.shadowRoot)}:r),l)if(u.functional){u._injectStyles=l;var c=u.render;u.render=function(t,e){return l.call(e),c(t,e)}}else{var d=u.beforeCreate;u.beforeCreate=d?[].concat(d,l):[l]}return{exports:t,options:u}}n.d(e,"a",(function(){return i}))},"28c9":function(t,e){function n(){this.__data__=[],this.size=0}t.exports=n},2921:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("vi",{months:"tháng 1_tháng 2_tháng 3_tháng 4_tháng 5_tháng 6_tháng 7_tháng 8_tháng 9_tháng 10_tháng 11_tháng 12".split("_"),monthsShort:"Thg 01_Thg 02_Thg 03_Thg 04_Thg 05_Thg 06_Thg 07_Thg 08_Thg 09_Thg 10_Thg 11_Thg 12".split("_"),monthsParseExact:!0,weekdays:"chủ nhật_thứ hai_thứ ba_thứ tư_thứ năm_thứ sáu_thứ bảy".split("_"),weekdaysShort:"CN_T2_T3_T4_T5_T6_T7".split("_"),weekdaysMin:"CN_T2_T3_T4_T5_T6_T7".split("_"),weekdaysParseExact:!0,meridiemParse:/sa|ch/i,isPM:function(t){return/^ch$/i.test(t)},meridiem:function(t,e,n){return t<12?n?"sa":"SA":n?"ch":"CH"},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM [năm] YYYY",LLL:"D MMMM [năm] YYYY HH:mm",LLLL:"dddd, D MMMM [năm] YYYY HH:mm",l:"DD/M/YYYY",ll:"D MMM YYYY",lll:"D MMM YYYY HH:mm",llll:"ddd, D MMM YYYY HH:mm"},calendar:{sameDay:"[Hôm nay lúc] LT",nextDay:"[Ngày mai lúc] LT",nextWeek:"dddd [tuần tới lúc] LT",lastDay:"[Hôm qua lúc] LT",lastWeek:"dddd [tuần trước lúc] LT",sameElse:"L"},relativeTime:{future:"%s tới",past:"%s trước",s:"vài giây",ss:"%d giây",m:"một phút",mm:"%d phút",h:"một giờ",hh:"%d giờ",d:"một ngày",dd:"%d ngày",w:"một tuần",ww:"%d tuần",M:"một tháng",MM:"%d tháng",y:"một năm",yy:"%d năm"},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:function(t){return t},week:{dow:1,doy:4}});return e}))},"293c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={words:{ss:["sekund","sekunda","sekundi"],m:["jedan minut","jednog minuta"],mm:["minut","minuta","minuta"],h:["jedan sat","jednog sata"],hh:["sat","sata","sati"],dd:["dan","dana","dana"],MM:["mjesec","mjeseca","mjeseci"],yy:["godina","godine","godina"]},correctGrammaticalCase:function(t,e){return 1===t?e[0]:t>=2&&t<=4?e[1]:e[2]},translate:function(t,n,i){var r=e.words[i];return 1===i.length?n?r[0]:r[1]:t+" "+e.correctGrammaticalCase(t,r)}},n=t.defineLocale("me",{months:"januar_februar_mart_april_maj_jun_jul_avgust_septembar_oktobar_novembar_decembar".split("_"),monthsShort:"jan._feb._mar._apr._maj_jun_jul_avg._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedjelja_ponedjeljak_utorak_srijeda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sri._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sjutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedjelju] [u] LT";case 3:return"[u] [srijedu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[juče u] LT",lastWeek:function(){var t=["[prošle] [nedjelje] [u] LT","[prošlog] [ponedjeljka] [u] LT","[prošlog] [utorka] [u] LT","[prošle] [srijede] [u] LT","[prošlog] [četvrtka] [u] LT","[prošlog] [petka] [u] LT","[prošle] [subote] [u] LT"];return t[this.day()]},sameElse:"L"},relativeTime:{future:"za %s",past:"prije %s",s:"nekoliko sekundi",ss:e.translate,m:e.translate,mm:e.translate,h:e.translate,hh:e.translate,d:"dan",dd:e.translate,M:"mjesec",MM:e.translate,y:"godinu",yy:e.translate},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},"29f3":function(t,e){var n=Object.prototype,i=n.toString;function r(t){return i.call(t)}t.exports=r},"2a62":function(t,e,n){var i=n("825a");t.exports=function(t){var e=t["return"];if(void 0!==e)return i(e.call(t)).value}},"2b0e":function(t,e,n){"use strict";(function(t){
/*!
 * Vue.js v2.6.12
 * (c) 2014-2020 Evan You
 * Released under the MIT License.
 */
var n=Object.freeze({});function i(t){return void 0===t||null===t}function r(t){return void 0!==t&&null!==t}function o(t){return!0===t}function a(t){return!1===t}function s(t){return"string"===typeof t||"number"===typeof t||"symbol"===typeof t||"boolean"===typeof t}function l(t){return null!==t&&"object"===typeof t}var u=Object.prototype.toString;function c(t){return"[object Object]"===u.call(t)}function d(t){return"[object RegExp]"===u.call(t)}function h(t){var e=parseFloat(String(t));return e>=0&&Math.floor(e)===e&&isFinite(t)}function f(t){return r(t)&&"function"===typeof t.then&&"function"===typeof t.catch}function p(t){return null==t?"":Array.isArray(t)||c(t)&&t.toString===u?JSON.stringify(t,null,2):String(t)}function m(t){var e=parseFloat(t);return isNaN(e)?t:e}function g(t,e){for(var n=Object.create(null),i=t.split(","),r=0;r<i.length;r++)n[i[r]]=!0;return e?function(t){return n[t.toLowerCase()]}:function(t){return n[t]}}g("slot,component",!0);var y=g("key,ref,slot,slot-scope,is");function v(t,e){if(t.length){var n=t.indexOf(e);if(n>-1)return t.splice(n,1)}}var _=Object.prototype.hasOwnProperty;function b(t,e){return _.call(t,e)}function x(t){var e=Object.create(null);return function(n){var i=e[n];return i||(e[n]=t(n))}}var w=/-(\w)/g,M=x((function(t){return t.replace(w,(function(t,e){return e?e.toUpperCase():""}))})),S=x((function(t){return t.charAt(0).toUpperCase()+t.slice(1)})),L=/\B([A-Z])/g,k=x((function(t){return t.replace(L,"-$1").toLowerCase()}));function T(t,e){function n(n){var i=arguments.length;return i?i>1?t.apply(e,arguments):t.call(e,n):t.call(e)}return n._length=t.length,n}function D(t,e){return t.bind(e)}var C=Function.prototype.bind?D:T;function I(t,e){e=e||0;var n=t.length-e,i=new Array(n);while(n--)i[n]=t[n+e];return i}function A(t,e){for(var n in e)t[n]=e[n];return t}function Y(t){for(var e={},n=0;n<t.length;n++)t[n]&&A(e,t[n]);return e}function O(t,e,n){}var P=function(t,e,n){return!1},E=function(t){return t};function N(t,e){if(t===e)return!0;var n=l(t),i=l(e);if(!n||!i)return!n&&!i&&String(t)===String(e);try{var r=Array.isArray(t),o=Array.isArray(e);if(r&&o)return t.length===e.length&&t.every((function(t,n){return N(t,e[n])}));if(t instanceof Date&&e instanceof Date)return t.getTime()===e.getTime();if(r||o)return!1;var a=Object.keys(t),s=Object.keys(e);return a.length===s.length&&a.every((function(n){return N(t[n],e[n])}))}catch(u){return!1}}function R(t,e){for(var n=0;n<t.length;n++)if(N(t[n],e))return n;return-1}function H(t){var e=!1;return function(){e||(e=!0,t.apply(this,arguments))}}var j="data-server-rendered",z=["component","directive","filter"],W=["beforeCreate","created","beforeMount","mounted","beforeUpdate","updated","beforeDestroy","destroyed","activated","deactivated","errorCaptured","serverPrefetch"],B={optionMergeStrategies:Object.create(null),silent:!1,productionTip:!1,devtools:!1,performance:!1,errorHandler:null,warnHandler:null,ignoredElements:[],keyCodes:Object.create(null),isReservedTag:P,isReservedAttr:P,isUnknownElement:P,getTagNamespace:O,parsePlatformTagName:E,mustUseProp:P,async:!0,_lifecycleHooks:W},F=/a-zA-Z\u00B7\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u037D\u037F-\u1FFF\u200C-\u200D\u203F-\u2040\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD/;function V(t){var e=(t+"").charCodeAt(0);return 36===e||95===e}function G(t,e,n,i){Object.defineProperty(t,e,{value:n,enumerable:!!i,writable:!0,configurable:!0})}var $=new RegExp("[^"+F.source+".$_\\d]");function U(t){if(!$.test(t)){var e=t.split(".");return function(t){for(var n=0;n<e.length;n++){if(!t)return;t=t[e[n]]}return t}}}var q,X="__proto__"in{},Z="undefined"!==typeof window,J="undefined"!==typeof WXEnvironment&&!!WXEnvironment.platform,K=J&&WXEnvironment.platform.toLowerCase(),Q=Z&&window.navigator.userAgent.toLowerCase(),tt=Q&&/msie|trident/.test(Q),et=Q&&Q.indexOf("msie 9.0")>0,nt=Q&&Q.indexOf("edge/")>0,it=(Q&&Q.indexOf("android"),Q&&/iphone|ipad|ipod|ios/.test(Q)||"ios"===K),rt=(Q&&/chrome\/\d+/.test(Q),Q&&/phantomjs/.test(Q),Q&&Q.match(/firefox\/(\d+)/)),ot={}.watch,at=!1;if(Z)try{var st={};Object.defineProperty(st,"passive",{get:function(){at=!0}}),window.addEventListener("test-passive",null,st)}catch(Ma){}var lt=function(){return void 0===q&&(q=!Z&&!J&&"undefined"!==typeof t&&(t["process"]&&"server"===t["process"].env.VUE_ENV)),q},ut=Z&&window.__VUE_DEVTOOLS_GLOBAL_HOOK__;function ct(t){return"function"===typeof t&&/native code/.test(t.toString())}var dt,ht="undefined"!==typeof Symbol&&ct(Symbol)&&"undefined"!==typeof Reflect&&ct(Reflect.ownKeys);dt="undefined"!==typeof Set&&ct(Set)?Set:function(){function t(){this.set=Object.create(null)}return t.prototype.has=function(t){return!0===this.set[t]},t.prototype.add=function(t){this.set[t]=!0},t.prototype.clear=function(){this.set=Object.create(null)},t}();var ft=O,pt=0,mt=function(){this.id=pt++,this.subs=[]};mt.prototype.addSub=function(t){this.subs.push(t)},mt.prototype.removeSub=function(t){v(this.subs,t)},mt.prototype.depend=function(){mt.target&&mt.target.addDep(this)},mt.prototype.notify=function(){var t=this.subs.slice();for(var e=0,n=t.length;e<n;e++)t[e].update()},mt.target=null;var gt=[];function yt(t){gt.push(t),mt.target=t}function vt(){gt.pop(),mt.target=gt[gt.length-1]}var _t=function(t,e,n,i,r,o,a,s){this.tag=t,this.data=e,this.children=n,this.text=i,this.elm=r,this.ns=void 0,this.context=o,this.fnContext=void 0,this.fnOptions=void 0,this.fnScopeId=void 0,this.key=e&&e.key,this.componentOptions=a,this.componentInstance=void 0,this.parent=void 0,this.raw=!1,this.isStatic=!1,this.isRootInsert=!0,this.isComment=!1,this.isCloned=!1,this.isOnce=!1,this.asyncFactory=s,this.asyncMeta=void 0,this.isAsyncPlaceholder=!1},bt={child:{configurable:!0}};bt.child.get=function(){return this.componentInstance},Object.defineProperties(_t.prototype,bt);var xt=function(t){void 0===t&&(t="");var e=new _t;return e.text=t,e.isComment=!0,e};function wt(t){return new _t(void 0,void 0,void 0,String(t))}function Mt(t){var e=new _t(t.tag,t.data,t.children&&t.children.slice(),t.text,t.elm,t.context,t.componentOptions,t.asyncFactory);return e.ns=t.ns,e.isStatic=t.isStatic,e.key=t.key,e.isComment=t.isComment,e.fnContext=t.fnContext,e.fnOptions=t.fnOptions,e.fnScopeId=t.fnScopeId,e.asyncMeta=t.asyncMeta,e.isCloned=!0,e}var St=Array.prototype,Lt=Object.create(St),kt=["push","pop","shift","unshift","splice","sort","reverse"];kt.forEach((function(t){var e=St[t];G(Lt,t,(function(){var n=[],i=arguments.length;while(i--)n[i]=arguments[i];var r,o=e.apply(this,n),a=this.__ob__;switch(t){case"push":case"unshift":r=n;break;case"splice":r=n.slice(2);break}return r&&a.observeArray(r),a.dep.notify(),o}))}));var Tt=Object.getOwnPropertyNames(Lt),Dt=!0;function Ct(t){Dt=t}var It=function(t){this.value=t,this.dep=new mt,this.vmCount=0,G(t,"__ob__",this),Array.isArray(t)?(X?At(t,Lt):Yt(t,Lt,Tt),this.observeArray(t)):this.walk(t)};function At(t,e){t.__proto__=e}function Yt(t,e,n){for(var i=0,r=n.length;i<r;i++){var o=n[i];G(t,o,e[o])}}function Ot(t,e){var n;if(l(t)&&!(t instanceof _t))return b(t,"__ob__")&&t.__ob__ instanceof It?n=t.__ob__:Dt&&!lt()&&(Array.isArray(t)||c(t))&&Object.isExtensible(t)&&!t._isVue&&(n=new It(t)),e&&n&&n.vmCount++,n}function Pt(t,e,n,i,r){var o=new mt,a=Object.getOwnPropertyDescriptor(t,e);if(!a||!1!==a.configurable){var s=a&&a.get,l=a&&a.set;s&&!l||2!==arguments.length||(n=t[e]);var u=!r&&Ot(n);Object.defineProperty(t,e,{enumerable:!0,configurable:!0,get:function(){var e=s?s.call(t):n;return mt.target&&(o.depend(),u&&(u.dep.depend(),Array.isArray(e)&&Rt(e))),e},set:function(e){var i=s?s.call(t):n;e===i||e!==e&&i!==i||s&&!l||(l?l.call(t,e):n=e,u=!r&&Ot(e),o.notify())}})}}function Et(t,e,n){if(Array.isArray(t)&&h(e))return t.length=Math.max(t.length,e),t.splice(e,1,n),n;if(e in t&&!(e in Object.prototype))return t[e]=n,n;var i=t.__ob__;return t._isVue||i&&i.vmCount?n:i?(Pt(i.value,e,n),i.dep.notify(),n):(t[e]=n,n)}function Nt(t,e){if(Array.isArray(t)&&h(e))t.splice(e,1);else{var n=t.__ob__;t._isVue||n&&n.vmCount||b(t,e)&&(delete t[e],n&&n.dep.notify())}}function Rt(t){for(var e=void 0,n=0,i=t.length;n<i;n++)e=t[n],e&&e.__ob__&&e.__ob__.dep.depend(),Array.isArray(e)&&Rt(e)}It.prototype.walk=function(t){for(var e=Object.keys(t),n=0;n<e.length;n++)Pt(t,e[n])},It.prototype.observeArray=function(t){for(var e=0,n=t.length;e<n;e++)Ot(t[e])};var Ht=B.optionMergeStrategies;function jt(t,e){if(!e)return t;for(var n,i,r,o=ht?Reflect.ownKeys(e):Object.keys(e),a=0;a<o.length;a++)n=o[a],"__ob__"!==n&&(i=t[n],r=e[n],b(t,n)?i!==r&&c(i)&&c(r)&&jt(i,r):Et(t,n,r));return t}function zt(t,e,n){return n?function(){var i="function"===typeof e?e.call(n,n):e,r="function"===typeof t?t.call(n,n):t;return i?jt(i,r):r}:e?t?function(){return jt("function"===typeof e?e.call(this,this):e,"function"===typeof t?t.call(this,this):t)}:e:t}function Wt(t,e){var n=e?t?t.concat(e):Array.isArray(e)?e:[e]:t;return n?Bt(n):n}function Bt(t){for(var e=[],n=0;n<t.length;n++)-1===e.indexOf(t[n])&&e.push(t[n]);return e}function Ft(t,e,n,i){var r=Object.create(t||null);return e?A(r,e):r}Ht.data=function(t,e,n){return n?zt(t,e,n):e&&"function"!==typeof e?t:zt(t,e)},W.forEach((function(t){Ht[t]=Wt})),z.forEach((function(t){Ht[t+"s"]=Ft})),Ht.watch=function(t,e,n,i){if(t===ot&&(t=void 0),e===ot&&(e=void 0),!e)return Object.create(t||null);if(!t)return e;var r={};for(var o in A(r,t),e){var a=r[o],s=e[o];a&&!Array.isArray(a)&&(a=[a]),r[o]=a?a.concat(s):Array.isArray(s)?s:[s]}return r},Ht.props=Ht.methods=Ht.inject=Ht.computed=function(t,e,n,i){if(!t)return e;var r=Object.create(null);return A(r,t),e&&A(r,e),r},Ht.provide=zt;var Vt=function(t,e){return void 0===e?t:e};function Gt(t,e){var n=t.props;if(n){var i,r,o,a={};if(Array.isArray(n)){i=n.length;while(i--)r=n[i],"string"===typeof r&&(o=M(r),a[o]={type:null})}else if(c(n))for(var s in n)r=n[s],o=M(s),a[o]=c(r)?r:{type:r};else 0;t.props=a}}function $t(t,e){var n=t.inject;if(n){var i=t.inject={};if(Array.isArray(n))for(var r=0;r<n.length;r++)i[n[r]]={from:n[r]};else if(c(n))for(var o in n){var a=n[o];i[o]=c(a)?A({from:o},a):{from:a}}else 0}}function Ut(t){var e=t.directives;if(e)for(var n in e){var i=e[n];"function"===typeof i&&(e[n]={bind:i,update:i})}}function qt(t,e,n){if("function"===typeof e&&(e=e.options),Gt(e,n),$t(e,n),Ut(e),!e._base&&(e.extends&&(t=qt(t,e.extends,n)),e.mixins))for(var i=0,r=e.mixins.length;i<r;i++)t=qt(t,e.mixins[i],n);var o,a={};for(o in t)s(o);for(o in e)b(t,o)||s(o);function s(i){var r=Ht[i]||Vt;a[i]=r(t[i],e[i],n,i)}return a}function Xt(t,e,n,i){if("string"===typeof n){var r=t[e];if(b(r,n))return r[n];var o=M(n);if(b(r,o))return r[o];var a=S(o);if(b(r,a))return r[a];var s=r[n]||r[o]||r[a];return s}}function Zt(t,e,n,i){var r=e[t],o=!b(n,t),a=n[t],s=te(Boolean,r.type);if(s>-1)if(o&&!b(r,"default"))a=!1;else if(""===a||a===k(t)){var l=te(String,r.type);(l<0||s<l)&&(a=!0)}if(void 0===a){a=Jt(i,r,t);var u=Dt;Ct(!0),Ot(a),Ct(u)}return a}function Jt(t,e,n){if(b(e,"default")){var i=e.default;return t&&t.$options.propsData&&void 0===t.$options.propsData[n]&&void 0!==t._props[n]?t._props[n]:"function"===typeof i&&"Function"!==Kt(e.type)?i.call(t):i}}function Kt(t){var e=t&&t.toString().match(/^\s*function (\w+)/);return e?e[1]:""}function Qt(t,e){return Kt(t)===Kt(e)}function te(t,e){if(!Array.isArray(e))return Qt(e,t)?0:-1;for(var n=0,i=e.length;n<i;n++)if(Qt(e[n],t))return n;return-1}function ee(t,e,n){yt();try{if(e){var i=e;while(i=i.$parent){var r=i.$options.errorCaptured;if(r)for(var o=0;o<r.length;o++)try{var a=!1===r[o].call(i,t,e,n);if(a)return}catch(Ma){ie(Ma,i,"errorCaptured hook")}}}ie(t,e,n)}finally{vt()}}function ne(t,e,n,i,r){var o;try{o=n?t.apply(e,n):t.call(e),o&&!o._isVue&&f(o)&&!o._handled&&(o.catch((function(t){return ee(t,i,r+" (Promise/async)")})),o._handled=!0)}catch(Ma){ee(Ma,i,r)}return o}function ie(t,e,n){if(B.errorHandler)try{return B.errorHandler.call(null,t,e,n)}catch(Ma){Ma!==t&&re(Ma,null,"config.errorHandler")}re(t,e,n)}function re(t,e,n){if(!Z&&!J||"undefined"===typeof console)throw t}var oe,ae=!1,se=[],le=!1;function ue(){le=!1;var t=se.slice(0);se.length=0;for(var e=0;e<t.length;e++)t[e]()}if("undefined"!==typeof Promise&&ct(Promise)){var ce=Promise.resolve();oe=function(){ce.then(ue),it&&setTimeout(O)},ae=!0}else if(tt||"undefined"===typeof MutationObserver||!ct(MutationObserver)&&"[object MutationObserverConstructor]"!==MutationObserver.toString())oe="undefined"!==typeof setImmediate&&ct(setImmediate)?function(){setImmediate(ue)}:function(){setTimeout(ue,0)};else{var de=1,he=new MutationObserver(ue),fe=document.createTextNode(String(de));he.observe(fe,{characterData:!0}),oe=function(){de=(de+1)%2,fe.data=String(de)},ae=!0}function pe(t,e){var n;if(se.push((function(){if(t)try{t.call(e)}catch(Ma){ee(Ma,e,"nextTick")}else n&&n(e)})),le||(le=!0,oe()),!t&&"undefined"!==typeof Promise)return new Promise((function(t){n=t}))}var me=new dt;function ge(t){ye(t,me),me.clear()}function ye(t,e){var n,i,r=Array.isArray(t);if(!(!r&&!l(t)||Object.isFrozen(t)||t instanceof _t)){if(t.__ob__){var o=t.__ob__.dep.id;if(e.has(o))return;e.add(o)}if(r){n=t.length;while(n--)ye(t[n],e)}else{i=Object.keys(t),n=i.length;while(n--)ye(t[i[n]],e)}}}var ve=x((function(t){var e="&"===t.charAt(0);t=e?t.slice(1):t;var n="~"===t.charAt(0);t=n?t.slice(1):t;var i="!"===t.charAt(0);return t=i?t.slice(1):t,{name:t,once:n,capture:i,passive:e}}));function _e(t,e){function n(){var t=arguments,i=n.fns;if(!Array.isArray(i))return ne(i,null,arguments,e,"v-on handler");for(var r=i.slice(),o=0;o<r.length;o++)ne(r[o],null,t,e,"v-on handler")}return n.fns=t,n}function be(t,e,n,r,a,s){var l,u,c,d;for(l in t)u=t[l],c=e[l],d=ve(l),i(u)||(i(c)?(i(u.fns)&&(u=t[l]=_e(u,s)),o(d.once)&&(u=t[l]=a(d.name,u,d.capture)),n(d.name,u,d.capture,d.passive,d.params)):u!==c&&(c.fns=u,t[l]=c));for(l in e)i(t[l])&&(d=ve(l),r(d.name,e[l],d.capture))}function xe(t,e,n){var a;t instanceof _t&&(t=t.data.hook||(t.data.hook={}));var s=t[e];function l(){n.apply(this,arguments),v(a.fns,l)}i(s)?a=_e([l]):r(s.fns)&&o(s.merged)?(a=s,a.fns.push(l)):a=_e([s,l]),a.merged=!0,t[e]=a}function we(t,e,n){var o=e.options.props;if(!i(o)){var a={},s=t.attrs,l=t.props;if(r(s)||r(l))for(var u in o){var c=k(u);Me(a,l,u,c,!0)||Me(a,s,u,c,!1)}return a}}function Me(t,e,n,i,o){if(r(e)){if(b(e,n))return t[n]=e[n],o||delete e[n],!0;if(b(e,i))return t[n]=e[i],o||delete e[i],!0}return!1}function Se(t){for(var e=0;e<t.length;e++)if(Array.isArray(t[e]))return Array.prototype.concat.apply([],t);return t}function Le(t){return s(t)?[wt(t)]:Array.isArray(t)?Te(t):void 0}function ke(t){return r(t)&&r(t.text)&&a(t.isComment)}function Te(t,e){var n,a,l,u,c=[];for(n=0;n<t.length;n++)a=t[n],i(a)||"boolean"===typeof a||(l=c.length-1,u=c[l],Array.isArray(a)?a.length>0&&(a=Te(a,(e||"")+"_"+n),ke(a[0])&&ke(u)&&(c[l]=wt(u.text+a[0].text),a.shift()),c.push.apply(c,a)):s(a)?ke(u)?c[l]=wt(u.text+a):""!==a&&c.push(wt(a)):ke(a)&&ke(u)?c[l]=wt(u.text+a.text):(o(t._isVList)&&r(a.tag)&&i(a.key)&&r(e)&&(a.key="__vlist"+e+"_"+n+"__"),c.push(a)));return c}function De(t){var e=t.$options.provide;e&&(t._provided="function"===typeof e?e.call(t):e)}function Ce(t){var e=Ie(t.$options.inject,t);e&&(Ct(!1),Object.keys(e).forEach((function(n){Pt(t,n,e[n])})),Ct(!0))}function Ie(t,e){if(t){for(var n=Object.create(null),i=ht?Reflect.ownKeys(t):Object.keys(t),r=0;r<i.length;r++){var o=i[r];if("__ob__"!==o){var a=t[o].from,s=e;while(s){if(s._provided&&b(s._provided,a)){n[o]=s._provided[a];break}s=s.$parent}if(!s)if("default"in t[o]){var l=t[o].default;n[o]="function"===typeof l?l.call(e):l}else 0}}return n}}function Ae(t,e){if(!t||!t.length)return{};for(var n={},i=0,r=t.length;i<r;i++){var o=t[i],a=o.data;if(a&&a.attrs&&a.attrs.slot&&delete a.attrs.slot,o.context!==e&&o.fnContext!==e||!a||null==a.slot)(n.default||(n.default=[])).push(o);else{var s=a.slot,l=n[s]||(n[s]=[]);"template"===o.tag?l.push.apply(l,o.children||[]):l.push(o)}}for(var u in n)n[u].every(Ye)&&delete n[u];return n}function Ye(t){return t.isComment&&!t.asyncFactory||" "===t.text}function Oe(t,e,i){var r,o=Object.keys(e).length>0,a=t?!!t.$stable:!o,s=t&&t.$key;if(t){if(t._normalized)return t._normalized;if(a&&i&&i!==n&&s===i.$key&&!o&&!i.$hasNormal)return i;for(var l in r={},t)t[l]&&"$"!==l[0]&&(r[l]=Pe(e,l,t[l]))}else r={};for(var u in e)u in r||(r[u]=Ee(e,u));return t&&Object.isExtensible(t)&&(t._normalized=r),G(r,"$stable",a),G(r,"$key",s),G(r,"$hasNormal",o),r}function Pe(t,e,n){var i=function(){var t=arguments.length?n.apply(null,arguments):n({});return t=t&&"object"===typeof t&&!Array.isArray(t)?[t]:Le(t),t&&(0===t.length||1===t.length&&t[0].isComment)?void 0:t};return n.proxy&&Object.defineProperty(t,e,{get:i,enumerable:!0,configurable:!0}),i}function Ee(t,e){return function(){return t[e]}}function Ne(t,e){var n,i,o,a,s;if(Array.isArray(t)||"string"===typeof t)for(n=new Array(t.length),i=0,o=t.length;i<o;i++)n[i]=e(t[i],i);else if("number"===typeof t)for(n=new Array(t),i=0;i<t;i++)n[i]=e(i+1,i);else if(l(t))if(ht&&t[Symbol.iterator]){n=[];var u=t[Symbol.iterator](),c=u.next();while(!c.done)n.push(e(c.value,n.length)),c=u.next()}else for(a=Object.keys(t),n=new Array(a.length),i=0,o=a.length;i<o;i++)s=a[i],n[i]=e(t[s],s,i);return r(n)||(n=[]),n._isVList=!0,n}function Re(t,e,n,i){var r,o=this.$scopedSlots[t];o?(n=n||{},i&&(n=A(A({},i),n)),r=o(n)||e):r=this.$slots[t]||e;var a=n&&n.slot;return a?this.$createElement("template",{slot:a},r):r}function He(t){return Xt(this.$options,"filters",t,!0)||E}function je(t,e){return Array.isArray(t)?-1===t.indexOf(e):t!==e}function ze(t,e,n,i,r){var o=B.keyCodes[e]||n;return r&&i&&!B.keyCodes[e]?je(r,i):o?je(o,t):i?k(i)!==e:void 0}function We(t,e,n,i,r){if(n)if(l(n)){var o;Array.isArray(n)&&(n=Y(n));var a=function(a){if("class"===a||"style"===a||y(a))o=t;else{var s=t.attrs&&t.attrs.type;o=i||B.mustUseProp(e,s,a)?t.domProps||(t.domProps={}):t.attrs||(t.attrs={})}var l=M(a),u=k(a);if(!(l in o)&&!(u in o)&&(o[a]=n[a],r)){var c=t.on||(t.on={});c["update:"+a]=function(t){n[a]=t}}};for(var s in n)a(s)}else;return t}function Be(t,e){var n=this._staticTrees||(this._staticTrees=[]),i=n[t];return i&&!e||(i=n[t]=this.$options.staticRenderFns[t].call(this._renderProxy,null,this),Ve(i,"__static__"+t,!1)),i}function Fe(t,e,n){return Ve(t,"__once__"+e+(n?"_"+n:""),!0),t}function Ve(t,e,n){if(Array.isArray(t))for(var i=0;i<t.length;i++)t[i]&&"string"!==typeof t[i]&&Ge(t[i],e+"_"+i,n);else Ge(t,e,n)}function Ge(t,e,n){t.isStatic=!0,t.key=e,t.isOnce=n}function $e(t,e){if(e)if(c(e)){var n=t.on=t.on?A({},t.on):{};for(var i in e){var r=n[i],o=e[i];n[i]=r?[].concat(r,o):o}}else;return t}function Ue(t,e,n,i){e=e||{$stable:!n};for(var r=0;r<t.length;r++){var o=t[r];Array.isArray(o)?Ue(o,e,n):o&&(o.proxy&&(o.fn.proxy=!0),e[o.key]=o.fn)}return i&&(e.$key=i),e}function qe(t,e){for(var n=0;n<e.length;n+=2){var i=e[n];"string"===typeof i&&i&&(t[e[n]]=e[n+1])}return t}function Xe(t,e){return"string"===typeof t?e+t:t}function Ze(t){t._o=Fe,t._n=m,t._s=p,t._l=Ne,t._t=Re,t._q=N,t._i=R,t._m=Be,t._f=He,t._k=ze,t._b=We,t._v=wt,t._e=xt,t._u=Ue,t._g=$e,t._d=qe,t._p=Xe}function Je(t,e,i,r,a){var s,l=this,u=a.options;b(r,"_uid")?(s=Object.create(r),s._original=r):(s=r,r=r._original);var c=o(u._compiled),d=!c;this.data=t,this.props=e,this.children=i,this.parent=r,this.listeners=t.on||n,this.injections=Ie(u.inject,r),this.slots=function(){return l.$slots||Oe(t.scopedSlots,l.$slots=Ae(i,r)),l.$slots},Object.defineProperty(this,"scopedSlots",{enumerable:!0,get:function(){return Oe(t.scopedSlots,this.slots())}}),c&&(this.$options=u,this.$slots=this.slots(),this.$scopedSlots=Oe(t.scopedSlots,this.$slots)),u._scopeId?this._c=function(t,e,n,i){var o=dn(s,t,e,n,i,d);return o&&!Array.isArray(o)&&(o.fnScopeId=u._scopeId,o.fnContext=r),o}:this._c=function(t,e,n,i){return dn(s,t,e,n,i,d)}}function Ke(t,e,i,o,a){var s=t.options,l={},u=s.props;if(r(u))for(var c in u)l[c]=Zt(c,u,e||n);else r(i.attrs)&&tn(l,i.attrs),r(i.props)&&tn(l,i.props);var d=new Je(i,l,a,o,t),h=s.render.call(null,d._c,d);if(h instanceof _t)return Qe(h,i,d.parent,s,d);if(Array.isArray(h)){for(var f=Le(h)||[],p=new Array(f.length),m=0;m<f.length;m++)p[m]=Qe(f[m],i,d.parent,s,d);return p}}function Qe(t,e,n,i,r){var o=Mt(t);return o.fnContext=n,o.fnOptions=i,e.slot&&((o.data||(o.data={})).slot=e.slot),o}function tn(t,e){for(var n in e)t[M(n)]=e[n]}Ze(Je.prototype);var en={init:function(t,e){if(t.componentInstance&&!t.componentInstance._isDestroyed&&t.data.keepAlive){var n=t;en.prepatch(n,n)}else{var i=t.componentInstance=on(t,In);i.$mount(e?t.elm:void 0,e)}},prepatch:function(t,e){var n=e.componentOptions,i=e.componentInstance=t.componentInstance;En(i,n.propsData,n.listeners,e,n.children)},insert:function(t){var e=t.context,n=t.componentInstance;n._isMounted||(n._isMounted=!0,jn(n,"mounted")),t.data.keepAlive&&(e._isMounted?Kn(n):Rn(n,!0))},destroy:function(t){var e=t.componentInstance;e._isDestroyed||(t.data.keepAlive?Hn(e,!0):e.$destroy())}},nn=Object.keys(en);function rn(t,e,n,a,s){if(!i(t)){var u=n.$options._base;if(l(t)&&(t=u.extend(t)),"function"===typeof t){var c;if(i(t.cid)&&(c=t,t=xn(c,u),void 0===t))return bn(c,e,n,a,s);e=e||{},xi(t),r(e.model)&&ln(t.options,e);var d=we(e,t,s);if(o(t.options.functional))return Ke(t,d,e,n,a);var h=e.on;if(e.on=e.nativeOn,o(t.options.abstract)){var f=e.slot;e={},f&&(e.slot=f)}an(e);var p=t.options.name||s,m=new _t("vue-component-"+t.cid+(p?"-"+p:""),e,void 0,void 0,void 0,n,{Ctor:t,propsData:d,listeners:h,tag:s,children:a},c);return m}}}function on(t,e){var n={_isComponent:!0,_parentVnode:t,parent:e},i=t.data.inlineTemplate;return r(i)&&(n.render=i.render,n.staticRenderFns=i.staticRenderFns),new t.componentOptions.Ctor(n)}function an(t){for(var e=t.hook||(t.hook={}),n=0;n<nn.length;n++){var i=nn[n],r=e[i],o=en[i];r===o||r&&r._merged||(e[i]=r?sn(o,r):o)}}function sn(t,e){var n=function(n,i){t(n,i),e(n,i)};return n._merged=!0,n}function ln(t,e){var n=t.model&&t.model.prop||"value",i=t.model&&t.model.event||"input";(e.attrs||(e.attrs={}))[n]=e.model.value;var o=e.on||(e.on={}),a=o[i],s=e.model.callback;r(a)?(Array.isArray(a)?-1===a.indexOf(s):a!==s)&&(o[i]=[s].concat(a)):o[i]=s}var un=1,cn=2;function dn(t,e,n,i,r,a){return(Array.isArray(n)||s(n))&&(r=i,i=n,n=void 0),o(a)&&(r=cn),hn(t,e,n,i,r)}function hn(t,e,n,i,o){if(r(n)&&r(n.__ob__))return xt();if(r(n)&&r(n.is)&&(e=n.is),!e)return xt();var a,s,l;(Array.isArray(i)&&"function"===typeof i[0]&&(n=n||{},n.scopedSlots={default:i[0]},i.length=0),o===cn?i=Le(i):o===un&&(i=Se(i)),"string"===typeof e)?(s=t.$vnode&&t.$vnode.ns||B.getTagNamespace(e),a=B.isReservedTag(e)?new _t(B.parsePlatformTagName(e),n,i,void 0,void 0,t):n&&n.pre||!r(l=Xt(t.$options,"components",e))?new _t(e,n,i,void 0,void 0,t):rn(l,n,t,i,e)):a=rn(e,n,t,i);return Array.isArray(a)?a:r(a)?(r(s)&&fn(a,s),r(n)&&pn(n),a):xt()}function fn(t,e,n){if(t.ns=e,"foreignObject"===t.tag&&(e=void 0,n=!0),r(t.children))for(var a=0,s=t.children.length;a<s;a++){var l=t.children[a];r(l.tag)&&(i(l.ns)||o(n)&&"svg"!==l.tag)&&fn(l,e,n)}}function pn(t){l(t.style)&&ge(t.style),l(t.class)&&ge(t.class)}function mn(t){t._vnode=null,t._staticTrees=null;var e=t.$options,i=t.$vnode=e._parentVnode,r=i&&i.context;t.$slots=Ae(e._renderChildren,r),t.$scopedSlots=n,t._c=function(e,n,i,r){return dn(t,e,n,i,r,!1)},t.$createElement=function(e,n,i,r){return dn(t,e,n,i,r,!0)};var o=i&&i.data;Pt(t,"$attrs",o&&o.attrs||n,null,!0),Pt(t,"$listeners",e._parentListeners||n,null,!0)}var gn,yn=null;function vn(t){Ze(t.prototype),t.prototype.$nextTick=function(t){return pe(t,this)},t.prototype._render=function(){var t,e=this,n=e.$options,i=n.render,r=n._parentVnode;r&&(e.$scopedSlots=Oe(r.data.scopedSlots,e.$slots,e.$scopedSlots)),e.$vnode=r;try{yn=e,t=i.call(e._renderProxy,e.$createElement)}catch(Ma){ee(Ma,e,"render"),t=e._vnode}finally{yn=null}return Array.isArray(t)&&1===t.length&&(t=t[0]),t instanceof _t||(t=xt()),t.parent=r,t}}function _n(t,e){return(t.__esModule||ht&&"Module"===t[Symbol.toStringTag])&&(t=t.default),l(t)?e.extend(t):t}function bn(t,e,n,i,r){var o=xt();return o.asyncFactory=t,o.asyncMeta={data:e,context:n,children:i,tag:r},o}function xn(t,e){if(o(t.error)&&r(t.errorComp))return t.errorComp;if(r(t.resolved))return t.resolved;var n=yn;if(n&&r(t.owners)&&-1===t.owners.indexOf(n)&&t.owners.push(n),o(t.loading)&&r(t.loadingComp))return t.loadingComp;if(n&&!r(t.owners)){var a=t.owners=[n],s=!0,u=null,c=null;n.$on("hook:destroyed",(function(){return v(a,n)}));var d=function(t){for(var e=0,n=a.length;e<n;e++)a[e].$forceUpdate();t&&(a.length=0,null!==u&&(clearTimeout(u),u=null),null!==c&&(clearTimeout(c),c=null))},h=H((function(n){t.resolved=_n(n,e),s?a.length=0:d(!0)})),p=H((function(e){r(t.errorComp)&&(t.error=!0,d(!0))})),m=t(h,p);return l(m)&&(f(m)?i(t.resolved)&&m.then(h,p):f(m.component)&&(m.component.then(h,p),r(m.error)&&(t.errorComp=_n(m.error,e)),r(m.loading)&&(t.loadingComp=_n(m.loading,e),0===m.delay?t.loading=!0:u=setTimeout((function(){u=null,i(t.resolved)&&i(t.error)&&(t.loading=!0,d(!1))}),m.delay||200)),r(m.timeout)&&(c=setTimeout((function(){c=null,i(t.resolved)&&p(null)}),m.timeout)))),s=!1,t.loading?t.loadingComp:t.resolved}}function wn(t){return t.isComment&&t.asyncFactory}function Mn(t){if(Array.isArray(t))for(var e=0;e<t.length;e++){var n=t[e];if(r(n)&&(r(n.componentOptions)||wn(n)))return n}}function Sn(t){t._events=Object.create(null),t._hasHookEvent=!1;var e=t.$options._parentListeners;e&&Dn(t,e)}function Ln(t,e){gn.$on(t,e)}function kn(t,e){gn.$off(t,e)}function Tn(t,e){var n=gn;return function i(){var r=e.apply(null,arguments);null!==r&&n.$off(t,i)}}function Dn(t,e,n){gn=t,be(e,n||{},Ln,kn,Tn,t),gn=void 0}function Cn(t){var e=/^hook:/;t.prototype.$on=function(t,n){var i=this;if(Array.isArray(t))for(var r=0,o=t.length;r<o;r++)i.$on(t[r],n);else(i._events[t]||(i._events[t]=[])).push(n),e.test(t)&&(i._hasHookEvent=!0);return i},t.prototype.$once=function(t,e){var n=this;function i(){n.$off(t,i),e.apply(n,arguments)}return i.fn=e,n.$on(t,i),n},t.prototype.$off=function(t,e){var n=this;if(!arguments.length)return n._events=Object.create(null),n;if(Array.isArray(t)){for(var i=0,r=t.length;i<r;i++)n.$off(t[i],e);return n}var o,a=n._events[t];if(!a)return n;if(!e)return n._events[t]=null,n;var s=a.length;while(s--)if(o=a[s],o===e||o.fn===e){a.splice(s,1);break}return n},t.prototype.$emit=function(t){var e=this,n=e._events[t];if(n){n=n.length>1?I(n):n;for(var i=I(arguments,1),r='event handler for "'+t+'"',o=0,a=n.length;o<a;o++)ne(n[o],e,i,e,r)}return e}}var In=null;function An(t){var e=In;return In=t,function(){In=e}}function Yn(t){var e=t.$options,n=e.parent;if(n&&!e.abstract){while(n.$options.abstract&&n.$parent)n=n.$parent;n.$children.push(t)}t.$parent=n,t.$root=n?n.$root:t,t.$children=[],t.$refs={},t._watcher=null,t._inactive=null,t._directInactive=!1,t._isMounted=!1,t._isDestroyed=!1,t._isBeingDestroyed=!1}function On(t){t.prototype._update=function(t,e){var n=this,i=n.$el,r=n._vnode,o=An(n);n._vnode=t,n.$el=r?n.__patch__(r,t):n.__patch__(n.$el,t,e,!1),o(),i&&(i.__vue__=null),n.$el&&(n.$el.__vue__=n),n.$vnode&&n.$parent&&n.$vnode===n.$parent._vnode&&(n.$parent.$el=n.$el)},t.prototype.$forceUpdate=function(){var t=this;t._watcher&&t._watcher.update()},t.prototype.$destroy=function(){var t=this;if(!t._isBeingDestroyed){jn(t,"beforeDestroy"),t._isBeingDestroyed=!0;var e=t.$parent;!e||e._isBeingDestroyed||t.$options.abstract||v(e.$children,t),t._watcher&&t._watcher.teardown();var n=t._watchers.length;while(n--)t._watchers[n].teardown();t._data.__ob__&&t._data.__ob__.vmCount--,t._isDestroyed=!0,t.__patch__(t._vnode,null),jn(t,"destroyed"),t.$off(),t.$el&&(t.$el.__vue__=null),t.$vnode&&(t.$vnode.parent=null)}}}function Pn(t,e,n){var i;return t.$el=e,t.$options.render||(t.$options.render=xt),jn(t,"beforeMount"),i=function(){t._update(t._render(),n)},new ni(t,i,O,{before:function(){t._isMounted&&!t._isDestroyed&&jn(t,"beforeUpdate")}},!0),n=!1,null==t.$vnode&&(t._isMounted=!0,jn(t,"mounted")),t}function En(t,e,i,r,o){var a=r.data.scopedSlots,s=t.$scopedSlots,l=!!(a&&!a.$stable||s!==n&&!s.$stable||a&&t.$scopedSlots.$key!==a.$key),u=!!(o||t.$options._renderChildren||l);if(t.$options._parentVnode=r,t.$vnode=r,t._vnode&&(t._vnode.parent=r),t.$options._renderChildren=o,t.$attrs=r.data.attrs||n,t.$listeners=i||n,e&&t.$options.props){Ct(!1);for(var c=t._props,d=t.$options._propKeys||[],h=0;h<d.length;h++){var f=d[h],p=t.$options.props;c[f]=Zt(f,p,e,t)}Ct(!0),t.$options.propsData=e}i=i||n;var m=t.$options._parentListeners;t.$options._parentListeners=i,Dn(t,i,m),u&&(t.$slots=Ae(o,r.context),t.$forceUpdate())}function Nn(t){while(t&&(t=t.$parent))if(t._inactive)return!0;return!1}function Rn(t,e){if(e){if(t._directInactive=!1,Nn(t))return}else if(t._directInactive)return;if(t._inactive||null===t._inactive){t._inactive=!1;for(var n=0;n<t.$children.length;n++)Rn(t.$children[n]);jn(t,"activated")}}function Hn(t,e){if((!e||(t._directInactive=!0,!Nn(t)))&&!t._inactive){t._inactive=!0;for(var n=0;n<t.$children.length;n++)Hn(t.$children[n]);jn(t,"deactivated")}}function jn(t,e){yt();var n=t.$options[e],i=e+" hook";if(n)for(var r=0,o=n.length;r<o;r++)ne(n[r],t,null,t,i);t._hasHookEvent&&t.$emit("hook:"+e),vt()}var zn=[],Wn=[],Bn={},Fn=!1,Vn=!1,Gn=0;function $n(){Gn=zn.length=Wn.length=0,Bn={},Fn=Vn=!1}var Un=0,qn=Date.now;if(Z&&!tt){var Xn=window.performance;Xn&&"function"===typeof Xn.now&&qn()>document.createEvent("Event").timeStamp&&(qn=function(){return Xn.now()})}function Zn(){var t,e;for(Un=qn(),Vn=!0,zn.sort((function(t,e){return t.id-e.id})),Gn=0;Gn<zn.length;Gn++)t=zn[Gn],t.before&&t.before(),e=t.id,Bn[e]=null,t.run();var n=Wn.slice(),i=zn.slice();$n(),Qn(n),Jn(i),ut&&B.devtools&&ut.emit("flush")}function Jn(t){var e=t.length;while(e--){var n=t[e],i=n.vm;i._watcher===n&&i._isMounted&&!i._isDestroyed&&jn(i,"updated")}}function Kn(t){t._inactive=!1,Wn.push(t)}function Qn(t){for(var e=0;e<t.length;e++)t[e]._inactive=!0,Rn(t[e],!0)}function ti(t){var e=t.id;if(null==Bn[e]){if(Bn[e]=!0,Vn){var n=zn.length-1;while(n>Gn&&zn[n].id>t.id)n--;zn.splice(n+1,0,t)}else zn.push(t);Fn||(Fn=!0,pe(Zn))}}var ei=0,ni=function(t,e,n,i,r){this.vm=t,r&&(t._watcher=this),t._watchers.push(this),i?(this.deep=!!i.deep,this.user=!!i.user,this.lazy=!!i.lazy,this.sync=!!i.sync,this.before=i.before):this.deep=this.user=this.lazy=this.sync=!1,this.cb=n,this.id=++ei,this.active=!0,this.dirty=this.lazy,this.deps=[],this.newDeps=[],this.depIds=new dt,this.newDepIds=new dt,this.expression="","function"===typeof e?this.getter=e:(this.getter=U(e),this.getter||(this.getter=O)),this.value=this.lazy?void 0:this.get()};ni.prototype.get=function(){var t;yt(this);var e=this.vm;try{t=this.getter.call(e,e)}catch(Ma){if(!this.user)throw Ma;ee(Ma,e,'getter for watcher "'+this.expression+'"')}finally{this.deep&&ge(t),vt(),this.cleanupDeps()}return t},ni.prototype.addDep=function(t){var e=t.id;this.newDepIds.has(e)||(this.newDepIds.add(e),this.newDeps.push(t),this.depIds.has(e)||t.addSub(this))},ni.prototype.cleanupDeps=function(){var t=this.deps.length;while(t--){var e=this.deps[t];this.newDepIds.has(e.id)||e.removeSub(this)}var n=this.depIds;this.depIds=this.newDepIds,this.newDepIds=n,this.newDepIds.clear(),n=this.deps,this.deps=this.newDeps,this.newDeps=n,this.newDeps.length=0},ni.prototype.update=function(){this.lazy?this.dirty=!0:this.sync?this.run():ti(this)},ni.prototype.run=function(){if(this.active){var t=this.get();if(t!==this.value||l(t)||this.deep){var e=this.value;if(this.value=t,this.user)try{this.cb.call(this.vm,t,e)}catch(Ma){ee(Ma,this.vm,'callback for watcher "'+this.expression+'"')}else this.cb.call(this.vm,t,e)}}},ni.prototype.evaluate=function(){this.value=this.get(),this.dirty=!1},ni.prototype.depend=function(){var t=this.deps.length;while(t--)this.deps[t].depend()},ni.prototype.teardown=function(){if(this.active){this.vm._isBeingDestroyed||v(this.vm._watchers,this);var t=this.deps.length;while(t--)this.deps[t].removeSub(this);this.active=!1}};var ii={enumerable:!0,configurable:!0,get:O,set:O};function ri(t,e,n){ii.get=function(){return this[e][n]},ii.set=function(t){this[e][n]=t},Object.defineProperty(t,n,ii)}function oi(t){t._watchers=[];var e=t.$options;e.props&&ai(t,e.props),e.methods&&pi(t,e.methods),e.data?si(t):Ot(t._data={},!0),e.computed&&ci(t,e.computed),e.watch&&e.watch!==ot&&mi(t,e.watch)}function ai(t,e){var n=t.$options.propsData||{},i=t._props={},r=t.$options._propKeys=[],o=!t.$parent;o||Ct(!1);var a=function(o){r.push(o);var a=Zt(o,e,n,t);Pt(i,o,a),o in t||ri(t,"_props",o)};for(var s in e)a(s);Ct(!0)}function si(t){var e=t.$options.data;e=t._data="function"===typeof e?li(e,t):e||{},c(e)||(e={});var n=Object.keys(e),i=t.$options.props,r=(t.$options.methods,n.length);while(r--){var o=n[r];0,i&&b(i,o)||V(o)||ri(t,"_data",o)}Ot(e,!0)}function li(t,e){yt();try{return t.call(e,e)}catch(Ma){return ee(Ma,e,"data()"),{}}finally{vt()}}var ui={lazy:!0};function ci(t,e){var n=t._computedWatchers=Object.create(null),i=lt();for(var r in e){var o=e[r],a="function"===typeof o?o:o.get;0,i||(n[r]=new ni(t,a||O,O,ui)),r in t||di(t,r,o)}}function di(t,e,n){var i=!lt();"function"===typeof n?(ii.get=i?hi(e):fi(n),ii.set=O):(ii.get=n.get?i&&!1!==n.cache?hi(e):fi(n.get):O,ii.set=n.set||O),Object.defineProperty(t,e,ii)}function hi(t){return function(){var e=this._computedWatchers&&this._computedWatchers[t];if(e)return e.dirty&&e.evaluate(),mt.target&&e.depend(),e.value}}function fi(t){return function(){return t.call(this,this)}}function pi(t,e){t.$options.props;for(var n in e)t[n]="function"!==typeof e[n]?O:C(e[n],t)}function mi(t,e){for(var n in e){var i=e[n];if(Array.isArray(i))for(var r=0;r<i.length;r++)gi(t,n,i[r]);else gi(t,n,i)}}function gi(t,e,n,i){return c(n)&&(i=n,n=n.handler),"string"===typeof n&&(n=t[n]),t.$watch(e,n,i)}function yi(t){var e={get:function(){return this._data}},n={get:function(){return this._props}};Object.defineProperty(t.prototype,"$data",e),Object.defineProperty(t.prototype,"$props",n),t.prototype.$set=Et,t.prototype.$delete=Nt,t.prototype.$watch=function(t,e,n){var i=this;if(c(e))return gi(i,t,e,n);n=n||{},n.user=!0;var r=new ni(i,t,e,n);if(n.immediate)try{e.call(i,r.value)}catch(o){ee(o,i,'callback for immediate watcher "'+r.expression+'"')}return function(){r.teardown()}}}var vi=0;function _i(t){t.prototype._init=function(t){var e=this;e._uid=vi++,e._isVue=!0,t&&t._isComponent?bi(e,t):e.$options=qt(xi(e.constructor),t||{},e),e._renderProxy=e,e._self=e,Yn(e),Sn(e),mn(e),jn(e,"beforeCreate"),Ce(e),oi(e),De(e),jn(e,"created"),e.$options.el&&e.$mount(e.$options.el)}}function bi(t,e){var n=t.$options=Object.create(t.constructor.options),i=e._parentVnode;n.parent=e.parent,n._parentVnode=i;var r=i.componentOptions;n.propsData=r.propsData,n._parentListeners=r.listeners,n._renderChildren=r.children,n._componentTag=r.tag,e.render&&(n.render=e.render,n.staticRenderFns=e.staticRenderFns)}function xi(t){var e=t.options;if(t.super){var n=xi(t.super),i=t.superOptions;if(n!==i){t.superOptions=n;var r=wi(t);r&&A(t.extendOptions,r),e=t.options=qt(n,t.extendOptions),e.name&&(e.components[e.name]=t)}}return e}function wi(t){var e,n=t.options,i=t.sealedOptions;for(var r in n)n[r]!==i[r]&&(e||(e={}),e[r]=n[r]);return e}function Mi(t){this._init(t)}function Si(t){t.use=function(t){var e=this._installedPlugins||(this._installedPlugins=[]);if(e.indexOf(t)>-1)return this;var n=I(arguments,1);return n.unshift(this),"function"===typeof t.install?t.install.apply(t,n):"function"===typeof t&&t.apply(null,n),e.push(t),this}}function Li(t){t.mixin=function(t){return this.options=qt(this.options,t),this}}function ki(t){t.cid=0;var e=1;t.extend=function(t){t=t||{};var n=this,i=n.cid,r=t._Ctor||(t._Ctor={});if(r[i])return r[i];var o=t.name||n.options.name;var a=function(t){this._init(t)};return a.prototype=Object.create(n.prototype),a.prototype.constructor=a,a.cid=e++,a.options=qt(n.options,t),a["super"]=n,a.options.props&&Ti(a),a.options.computed&&Di(a),a.extend=n.extend,a.mixin=n.mixin,a.use=n.use,z.forEach((function(t){a[t]=n[t]})),o&&(a.options.components[o]=a),a.superOptions=n.options,a.extendOptions=t,a.sealedOptions=A({},a.options),r[i]=a,a}}function Ti(t){var e=t.options.props;for(var n in e)ri(t.prototype,"_props",n)}function Di(t){var e=t.options.computed;for(var n in e)di(t.prototype,n,e[n])}function Ci(t){z.forEach((function(e){t[e]=function(t,n){return n?("component"===e&&c(n)&&(n.name=n.name||t,n=this.options._base.extend(n)),"directive"===e&&"function"===typeof n&&(n={bind:n,update:n}),this.options[e+"s"][t]=n,n):this.options[e+"s"][t]}}))}function Ii(t){return t&&(t.Ctor.options.name||t.tag)}function Ai(t,e){return Array.isArray(t)?t.indexOf(e)>-1:"string"===typeof t?t.split(",").indexOf(e)>-1:!!d(t)&&t.test(e)}function Yi(t,e){var n=t.cache,i=t.keys,r=t._vnode;for(var o in n){var a=n[o];if(a){var s=Ii(a.componentOptions);s&&!e(s)&&Oi(n,o,i,r)}}}function Oi(t,e,n,i){var r=t[e];!r||i&&r.tag===i.tag||r.componentInstance.$destroy(),t[e]=null,v(n,e)}_i(Mi),yi(Mi),Cn(Mi),On(Mi),vn(Mi);var Pi=[String,RegExp,Array],Ei={name:"keep-alive",abstract:!0,props:{include:Pi,exclude:Pi,max:[String,Number]},created:function(){this.cache=Object.create(null),this.keys=[]},destroyed:function(){for(var t in this.cache)Oi(this.cache,t,this.keys)},mounted:function(){var t=this;this.$watch("include",(function(e){Yi(t,(function(t){return Ai(e,t)}))})),this.$watch("exclude",(function(e){Yi(t,(function(t){return!Ai(e,t)}))}))},render:function(){var t=this.$slots.default,e=Mn(t),n=e&&e.componentOptions;if(n){var i=Ii(n),r=this,o=r.include,a=r.exclude;if(o&&(!i||!Ai(o,i))||a&&i&&Ai(a,i))return e;var s=this,l=s.cache,u=s.keys,c=null==e.key?n.Ctor.cid+(n.tag?"::"+n.tag:""):e.key;l[c]?(e.componentInstance=l[c].componentInstance,v(u,c),u.push(c)):(l[c]=e,u.push(c),this.max&&u.length>parseInt(this.max)&&Oi(l,u[0],u,this._vnode)),e.data.keepAlive=!0}return e||t&&t[0]}},Ni={KeepAlive:Ei};function Ri(t){var e={get:function(){return B}};Object.defineProperty(t,"config",e),t.util={warn:ft,extend:A,mergeOptions:qt,defineReactive:Pt},t.set=Et,t.delete=Nt,t.nextTick=pe,t.observable=function(t){return Ot(t),t},t.options=Object.create(null),z.forEach((function(e){t.options[e+"s"]=Object.create(null)})),t.options._base=t,A(t.options.components,Ni),Si(t),Li(t),ki(t),Ci(t)}Ri(Mi),Object.defineProperty(Mi.prototype,"$isServer",{get:lt}),Object.defineProperty(Mi.prototype,"$ssrContext",{get:function(){return this.$vnode&&this.$vnode.ssrContext}}),Object.defineProperty(Mi,"FunctionalRenderContext",{value:Je}),Mi.version="2.6.12";var Hi=g("style,class"),ji=g("input,textarea,option,select,progress"),zi=function(t,e,n){return"value"===n&&ji(t)&&"button"!==e||"selected"===n&&"option"===t||"checked"===n&&"input"===t||"muted"===n&&"video"===t},Wi=g("contenteditable,draggable,spellcheck"),Bi=g("events,caret,typing,plaintext-only"),Fi=function(t,e){return qi(e)||"false"===e?"false":"contenteditable"===t&&Bi(e)?e:"true"},Vi=g("allowfullscreen,async,autofocus,autoplay,checked,compact,controls,declare,default,defaultchecked,defaultmuted,defaultselected,defer,disabled,enabled,formnovalidate,hidden,indeterminate,inert,ismap,itemscope,loop,multiple,muted,nohref,noresize,noshade,novalidate,nowrap,open,pauseonexit,readonly,required,reversed,scoped,seamless,selected,sortable,translate,truespeed,typemustmatch,visible"),Gi="http://www.w3.org/1999/xlink",$i=function(t){return":"===t.charAt(5)&&"xlink"===t.slice(0,5)},Ui=function(t){return $i(t)?t.slice(6,t.length):""},qi=function(t){return null==t||!1===t};function Xi(t){var e=t.data,n=t,i=t;while(r(i.componentInstance))i=i.componentInstance._vnode,i&&i.data&&(e=Zi(i.data,e));while(r(n=n.parent))n&&n.data&&(e=Zi(e,n.data));return Ji(e.staticClass,e.class)}function Zi(t,e){return{staticClass:Ki(t.staticClass,e.staticClass),class:r(t.class)?[t.class,e.class]:e.class}}function Ji(t,e){return r(t)||r(e)?Ki(t,Qi(e)):""}function Ki(t,e){return t?e?t+" "+e:t:e||""}function Qi(t){return Array.isArray(t)?tr(t):l(t)?er(t):"string"===typeof t?t:""}function tr(t){for(var e,n="",i=0,o=t.length;i<o;i++)r(e=Qi(t[i]))&&""!==e&&(n&&(n+=" "),n+=e);return n}function er(t){var e="";for(var n in t)t[n]&&(e&&(e+=" "),e+=n);return e}var nr={svg:"http://www.w3.org/2000/svg",math:"http://www.w3.org/1998/Math/MathML"},ir=g("html,body,base,head,link,meta,style,title,address,article,aside,footer,header,h1,h2,h3,h4,h5,h6,hgroup,nav,section,div,dd,dl,dt,figcaption,figure,picture,hr,img,li,main,ol,p,pre,ul,a,b,abbr,bdi,bdo,br,cite,code,data,dfn,em,i,kbd,mark,q,rp,rt,rtc,ruby,s,samp,small,span,strong,sub,sup,time,u,var,wbr,area,audio,map,track,video,embed,object,param,source,canvas,script,noscript,del,ins,caption,col,colgroup,table,thead,tbody,td,th,tr,button,datalist,fieldset,form,input,label,legend,meter,optgroup,option,output,progress,select,textarea,details,dialog,menu,menuitem,summary,content,element,shadow,template,blockquote,iframe,tfoot"),rr=g("svg,animate,circle,clippath,cursor,defs,desc,ellipse,filter,font-face,foreignObject,g,glyph,image,line,marker,mask,missing-glyph,path,pattern,polygon,polyline,rect,switch,symbol,text,textpath,tspan,use,view",!0),or=function(t){return ir(t)||rr(t)};function ar(t){return rr(t)?"svg":"math"===t?"math":void 0}var sr=Object.create(null);function lr(t){if(!Z)return!0;if(or(t))return!1;if(t=t.toLowerCase(),null!=sr[t])return sr[t];var e=document.createElement(t);return t.indexOf("-")>-1?sr[t]=e.constructor===window.HTMLUnknownElement||e.constructor===window.HTMLElement:sr[t]=/HTMLUnknownElement/.test(e.toString())}var ur=g("text,number,password,search,email,tel,url");function cr(t){if("string"===typeof t){var e=document.querySelector(t);return e||document.createElement("div")}return t}function dr(t,e){var n=document.createElement(t);return"select"!==t||e.data&&e.data.attrs&&void 0!==e.data.attrs.multiple&&n.setAttribute("multiple","multiple"),n}function hr(t,e){return document.createElementNS(nr[t],e)}function fr(t){return document.createTextNode(t)}function pr(t){return document.createComment(t)}function mr(t,e,n){t.insertBefore(e,n)}function gr(t,e){t.removeChild(e)}function yr(t,e){t.appendChild(e)}function vr(t){return t.parentNode}function _r(t){return t.nextSibling}function br(t){return t.tagName}function xr(t,e){t.textContent=e}function wr(t,e){t.setAttribute(e,"")}var Mr=Object.freeze({createElement:dr,createElementNS:hr,createTextNode:fr,createComment:pr,insertBefore:mr,removeChild:gr,appendChild:yr,parentNode:vr,nextSibling:_r,tagName:br,setTextContent:xr,setStyleScope:wr}),Sr={create:function(t,e){Lr(e)},update:function(t,e){t.data.ref!==e.data.ref&&(Lr(t,!0),Lr(e))},destroy:function(t){Lr(t,!0)}};function Lr(t,e){var n=t.data.ref;if(r(n)){var i=t.context,o=t.componentInstance||t.elm,a=i.$refs;e?Array.isArray(a[n])?v(a[n],o):a[n]===o&&(a[n]=void 0):t.data.refInFor?Array.isArray(a[n])?a[n].indexOf(o)<0&&a[n].push(o):a[n]=[o]:a[n]=o}}var kr=new _t("",{},[]),Tr=["create","activate","update","remove","destroy"];function Dr(t,e){return t.key===e.key&&(t.tag===e.tag&&t.isComment===e.isComment&&r(t.data)===r(e.data)&&Cr(t,e)||o(t.isAsyncPlaceholder)&&t.asyncFactory===e.asyncFactory&&i(e.asyncFactory.error))}function Cr(t,e){if("input"!==t.tag)return!0;var n,i=r(n=t.data)&&r(n=n.attrs)&&n.type,o=r(n=e.data)&&r(n=n.attrs)&&n.type;return i===o||ur(i)&&ur(o)}function Ir(t,e,n){var i,o,a={};for(i=e;i<=n;++i)o=t[i].key,r(o)&&(a[o]=i);return a}function Ar(t){var e,n,a={},l=t.modules,u=t.nodeOps;for(e=0;e<Tr.length;++e)for(a[Tr[e]]=[],n=0;n<l.length;++n)r(l[n][Tr[e]])&&a[Tr[e]].push(l[n][Tr[e]]);function c(t){return new _t(u.tagName(t).toLowerCase(),{},[],void 0,t)}function d(t,e){function n(){0===--n.listeners&&h(t)}return n.listeners=e,n}function h(t){var e=u.parentNode(t);r(e)&&u.removeChild(e,t)}function f(t,e,n,i,a,s,l){if(r(t.elm)&&r(s)&&(t=s[l]=Mt(t)),t.isRootInsert=!a,!p(t,e,n,i)){var c=t.data,d=t.children,h=t.tag;r(h)?(t.elm=t.ns?u.createElementNS(t.ns,h):u.createElement(h,t),w(t),_(t,d,e),r(c)&&x(t,e),v(n,t.elm,i)):o(t.isComment)?(t.elm=u.createComment(t.text),v(n,t.elm,i)):(t.elm=u.createTextNode(t.text),v(n,t.elm,i))}}function p(t,e,n,i){var a=t.data;if(r(a)){var s=r(t.componentInstance)&&a.keepAlive;if(r(a=a.hook)&&r(a=a.init)&&a(t,!1),r(t.componentInstance))return m(t,e),v(n,t.elm,i),o(s)&&y(t,e,n,i),!0}}function m(t,e){r(t.data.pendingInsert)&&(e.push.apply(e,t.data.pendingInsert),t.data.pendingInsert=null),t.elm=t.componentInstance.$el,b(t)?(x(t,e),w(t)):(Lr(t),e.push(t))}function y(t,e,n,i){var o,s=t;while(s.componentInstance)if(s=s.componentInstance._vnode,r(o=s.data)&&r(o=o.transition)){for(o=0;o<a.activate.length;++o)a.activate[o](kr,s);e.push(s);break}v(n,t.elm,i)}function v(t,e,n){r(t)&&(r(n)?u.parentNode(n)===t&&u.insertBefore(t,e,n):u.appendChild(t,e))}function _(t,e,n){if(Array.isArray(e)){0;for(var i=0;i<e.length;++i)f(e[i],n,t.elm,null,!0,e,i)}else s(t.text)&&u.appendChild(t.elm,u.createTextNode(String(t.text)))}function b(t){while(t.componentInstance)t=t.componentInstance._vnode;return r(t.tag)}function x(t,n){for(var i=0;i<a.create.length;++i)a.create[i](kr,t);e=t.data.hook,r(e)&&(r(e.create)&&e.create(kr,t),r(e.insert)&&n.push(t))}function w(t){var e;if(r(e=t.fnScopeId))u.setStyleScope(t.elm,e);else{var n=t;while(n)r(e=n.context)&&r(e=e.$options._scopeId)&&u.setStyleScope(t.elm,e),n=n.parent}r(e=In)&&e!==t.context&&e!==t.fnContext&&r(e=e.$options._scopeId)&&u.setStyleScope(t.elm,e)}function M(t,e,n,i,r,o){for(;i<=r;++i)f(n[i],o,t,e,!1,n,i)}function S(t){var e,n,i=t.data;if(r(i))for(r(e=i.hook)&&r(e=e.destroy)&&e(t),e=0;e<a.destroy.length;++e)a.destroy[e](t);if(r(e=t.children))for(n=0;n<t.children.length;++n)S(t.children[n])}function L(t,e,n){for(;e<=n;++e){var i=t[e];r(i)&&(r(i.tag)?(k(i),S(i)):h(i.elm))}}function k(t,e){if(r(e)||r(t.data)){var n,i=a.remove.length+1;for(r(e)?e.listeners+=i:e=d(t.elm,i),r(n=t.componentInstance)&&r(n=n._vnode)&&r(n.data)&&k(n,e),n=0;n<a.remove.length;++n)a.remove[n](t,e);r(n=t.data.hook)&&r(n=n.remove)?n(t,e):e()}else h(t.elm)}function T(t,e,n,o,a){var s,l,c,d,h=0,p=0,m=e.length-1,g=e[0],y=e[m],v=n.length-1,_=n[0],b=n[v],x=!a;while(h<=m&&p<=v)i(g)?g=e[++h]:i(y)?y=e[--m]:Dr(g,_)?(C(g,_,o,n,p),g=e[++h],_=n[++p]):Dr(y,b)?(C(y,b,o,n,v),y=e[--m],b=n[--v]):Dr(g,b)?(C(g,b,o,n,v),x&&u.insertBefore(t,g.elm,u.nextSibling(y.elm)),g=e[++h],b=n[--v]):Dr(y,_)?(C(y,_,o,n,p),x&&u.insertBefore(t,y.elm,g.elm),y=e[--m],_=n[++p]):(i(s)&&(s=Ir(e,h,m)),l=r(_.key)?s[_.key]:D(_,e,h,m),i(l)?f(_,o,t,g.elm,!1,n,p):(c=e[l],Dr(c,_)?(C(c,_,o,n,p),e[l]=void 0,x&&u.insertBefore(t,c.elm,g.elm)):f(_,o,t,g.elm,!1,n,p)),_=n[++p]);h>m?(d=i(n[v+1])?null:n[v+1].elm,M(t,d,n,p,v,o)):p>v&&L(e,h,m)}function D(t,e,n,i){for(var o=n;o<i;o++){var a=e[o];if(r(a)&&Dr(t,a))return o}}function C(t,e,n,s,l,c){if(t!==e){r(e.elm)&&r(s)&&(e=s[l]=Mt(e));var d=e.elm=t.elm;if(o(t.isAsyncPlaceholder))r(e.asyncFactory.resolved)?Y(t.elm,e,n):e.isAsyncPlaceholder=!0;else if(o(e.isStatic)&&o(t.isStatic)&&e.key===t.key&&(o(e.isCloned)||o(e.isOnce)))e.componentInstance=t.componentInstance;else{var h,f=e.data;r(f)&&r(h=f.hook)&&r(h=h.prepatch)&&h(t,e);var p=t.children,m=e.children;if(r(f)&&b(e)){for(h=0;h<a.update.length;++h)a.update[h](t,e);r(h=f.hook)&&r(h=h.update)&&h(t,e)}i(e.text)?r(p)&&r(m)?p!==m&&T(d,p,m,n,c):r(m)?(r(t.text)&&u.setTextContent(d,""),M(d,null,m,0,m.length-1,n)):r(p)?L(p,0,p.length-1):r(t.text)&&u.setTextContent(d,""):t.text!==e.text&&u.setTextContent(d,e.text),r(f)&&r(h=f.hook)&&r(h=h.postpatch)&&h(t,e)}}}function I(t,e,n){if(o(n)&&r(t.parent))t.parent.data.pendingInsert=e;else for(var i=0;i<e.length;++i)e[i].data.hook.insert(e[i])}var A=g("attrs,class,staticClass,staticStyle,key");function Y(t,e,n,i){var a,s=e.tag,l=e.data,u=e.children;if(i=i||l&&l.pre,e.elm=t,o(e.isComment)&&r(e.asyncFactory))return e.isAsyncPlaceholder=!0,!0;if(r(l)&&(r(a=l.hook)&&r(a=a.init)&&a(e,!0),r(a=e.componentInstance)))return m(e,n),!0;if(r(s)){if(r(u))if(t.hasChildNodes())if(r(a=l)&&r(a=a.domProps)&&r(a=a.innerHTML)){if(a!==t.innerHTML)return!1}else{for(var c=!0,d=t.firstChild,h=0;h<u.length;h++){if(!d||!Y(d,u[h],n,i)){c=!1;break}d=d.nextSibling}if(!c||d)return!1}else _(e,u,n);if(r(l)){var f=!1;for(var p in l)if(!A(p)){f=!0,x(e,n);break}!f&&l["class"]&&ge(l["class"])}}else t.data!==e.text&&(t.data=e.text);return!0}return function(t,e,n,s){if(!i(e)){var l=!1,d=[];if(i(t))l=!0,f(e,d);else{var h=r(t.nodeType);if(!h&&Dr(t,e))C(t,e,d,null,null,s);else{if(h){if(1===t.nodeType&&t.hasAttribute(j)&&(t.removeAttribute(j),n=!0),o(n)&&Y(t,e,d))return I(e,d,!0),t;t=c(t)}var p=t.elm,m=u.parentNode(p);if(f(e,d,p._leaveCb?null:m,u.nextSibling(p)),r(e.parent)){var g=e.parent,y=b(e);while(g){for(var v=0;v<a.destroy.length;++v)a.destroy[v](g);if(g.elm=e.elm,y){for(var _=0;_<a.create.length;++_)a.create[_](kr,g);var x=g.data.hook.insert;if(x.merged)for(var w=1;w<x.fns.length;w++)x.fns[w]()}else Lr(g);g=g.parent}}r(m)?L([t],0,0):r(t.tag)&&S(t)}}return I(e,d,l),e.elm}r(t)&&S(t)}}var Yr={create:Or,update:Or,destroy:function(t){Or(t,kr)}};function Or(t,e){(t.data.directives||e.data.directives)&&Pr(t,e)}function Pr(t,e){var n,i,r,o=t===kr,a=e===kr,s=Nr(t.data.directives,t.context),l=Nr(e.data.directives,e.context),u=[],c=[];for(n in l)i=s[n],r=l[n],i?(r.oldValue=i.value,r.oldArg=i.arg,Hr(r,"update",e,t),r.def&&r.def.componentUpdated&&c.push(r)):(Hr(r,"bind",e,t),r.def&&r.def.inserted&&u.push(r));if(u.length){var d=function(){for(var n=0;n<u.length;n++)Hr(u[n],"inserted",e,t)};o?xe(e,"insert",d):d()}if(c.length&&xe(e,"postpatch",(function(){for(var n=0;n<c.length;n++)Hr(c[n],"componentUpdated",e,t)})),!o)for(n in s)l[n]||Hr(s[n],"unbind",t,t,a)}var Er=Object.create(null);function Nr(t,e){var n,i,r=Object.create(null);if(!t)return r;for(n=0;n<t.length;n++)i=t[n],i.modifiers||(i.modifiers=Er),r[Rr(i)]=i,i.def=Xt(e.$options,"directives",i.name,!0);return r}function Rr(t){return t.rawName||t.name+"."+Object.keys(t.modifiers||{}).join(".")}function Hr(t,e,n,i,r){var o=t.def&&t.def[e];if(o)try{o(n.elm,t,n,i,r)}catch(Ma){ee(Ma,n.context,"directive "+t.name+" "+e+" hook")}}var jr=[Sr,Yr];function zr(t,e){var n=e.componentOptions;if((!r(n)||!1!==n.Ctor.options.inheritAttrs)&&(!i(t.data.attrs)||!i(e.data.attrs))){var o,a,s,l=e.elm,u=t.data.attrs||{},c=e.data.attrs||{};for(o in r(c.__ob__)&&(c=e.data.attrs=A({},c)),c)a=c[o],s=u[o],s!==a&&Wr(l,o,a);for(o in(tt||nt)&&c.value!==u.value&&Wr(l,"value",c.value),u)i(c[o])&&($i(o)?l.removeAttributeNS(Gi,Ui(o)):Wi(o)||l.removeAttribute(o))}}function Wr(t,e,n){t.tagName.indexOf("-")>-1?Br(t,e,n):Vi(e)?qi(n)?t.removeAttribute(e):(n="allowfullscreen"===e&&"EMBED"===t.tagName?"true":e,t.setAttribute(e,n)):Wi(e)?t.setAttribute(e,Fi(e,n)):$i(e)?qi(n)?t.removeAttributeNS(Gi,Ui(e)):t.setAttributeNS(Gi,e,n):Br(t,e,n)}function Br(t,e,n){if(qi(n))t.removeAttribute(e);else{if(tt&&!et&&"TEXTAREA"===t.tagName&&"placeholder"===e&&""!==n&&!t.__ieph){var i=function(e){e.stopImmediatePropagation(),t.removeEventListener("input",i)};t.addEventListener("input",i),t.__ieph=!0}t.setAttribute(e,n)}}var Fr={create:zr,update:zr};function Vr(t,e){var n=e.elm,o=e.data,a=t.data;if(!(i(o.staticClass)&&i(o.class)&&(i(a)||i(a.staticClass)&&i(a.class)))){var s=Xi(e),l=n._transitionClasses;r(l)&&(s=Ki(s,Qi(l))),s!==n._prevClass&&(n.setAttribute("class",s),n._prevClass=s)}}var Gr,$r={create:Vr,update:Vr},Ur="__r",qr="__c";function Xr(t){if(r(t[Ur])){var e=tt?"change":"input";t[e]=[].concat(t[Ur],t[e]||[]),delete t[Ur]}r(t[qr])&&(t.change=[].concat(t[qr],t.change||[]),delete t[qr])}function Zr(t,e,n){var i=Gr;return function r(){var o=e.apply(null,arguments);null!==o&&Qr(t,r,n,i)}}var Jr=ae&&!(rt&&Number(rt[1])<=53);function Kr(t,e,n,i){if(Jr){var r=Un,o=e;e=o._wrapper=function(t){if(t.target===t.currentTarget||t.timeStamp>=r||t.timeStamp<=0||t.target.ownerDocument!==document)return o.apply(this,arguments)}}Gr.addEventListener(t,e,at?{capture:n,passive:i}:n)}function Qr(t,e,n,i){(i||Gr).removeEventListener(t,e._wrapper||e,n)}function to(t,e){if(!i(t.data.on)||!i(e.data.on)){var n=e.data.on||{},r=t.data.on||{};Gr=e.elm,Xr(n),be(n,r,Kr,Qr,Zr,e.context),Gr=void 0}}var eo,no={create:to,update:to};function io(t,e){if(!i(t.data.domProps)||!i(e.data.domProps)){var n,o,a=e.elm,s=t.data.domProps||{},l=e.data.domProps||{};for(n in r(l.__ob__)&&(l=e.data.domProps=A({},l)),s)n in l||(a[n]="");for(n in l){if(o=l[n],"textContent"===n||"innerHTML"===n){if(e.children&&(e.children.length=0),o===s[n])continue;1===a.childNodes.length&&a.removeChild(a.childNodes[0])}if("value"===n&&"PROGRESS"!==a.tagName){a._value=o;var u=i(o)?"":String(o);ro(a,u)&&(a.value=u)}else if("innerHTML"===n&&rr(a.tagName)&&i(a.innerHTML)){eo=eo||document.createElement("div"),eo.innerHTML="<svg>"+o+"</svg>";var c=eo.firstChild;while(a.firstChild)a.removeChild(a.firstChild);while(c.firstChild)a.appendChild(c.firstChild)}else if(o!==s[n])try{a[n]=o}catch(Ma){}}}}function ro(t,e){return!t.composing&&("OPTION"===t.tagName||oo(t,e)||ao(t,e))}function oo(t,e){var n=!0;try{n=document.activeElement!==t}catch(Ma){}return n&&t.value!==e}function ao(t,e){var n=t.value,i=t._vModifiers;if(r(i)){if(i.number)return m(n)!==m(e);if(i.trim)return n.trim()!==e.trim()}return n!==e}var so={create:io,update:io},lo=x((function(t){var e={},n=/;(?![^(]*\))/g,i=/:(.+)/;return t.split(n).forEach((function(t){if(t){var n=t.split(i);n.length>1&&(e[n[0].trim()]=n[1].trim())}})),e}));function uo(t){var e=co(t.style);return t.staticStyle?A(t.staticStyle,e):e}function co(t){return Array.isArray(t)?Y(t):"string"===typeof t?lo(t):t}function ho(t,e){var n,i={};if(e){var r=t;while(r.componentInstance)r=r.componentInstance._vnode,r&&r.data&&(n=uo(r.data))&&A(i,n)}(n=uo(t.data))&&A(i,n);var o=t;while(o=o.parent)o.data&&(n=uo(o.data))&&A(i,n);return i}var fo,po=/^--/,mo=/\s*!important$/,go=function(t,e,n){if(po.test(e))t.style.setProperty(e,n);else if(mo.test(n))t.style.setProperty(k(e),n.replace(mo,""),"important");else{var i=vo(e);if(Array.isArray(n))for(var r=0,o=n.length;r<o;r++)t.style[i]=n[r];else t.style[i]=n}},yo=["Webkit","Moz","ms"],vo=x((function(t){if(fo=fo||document.createElement("div").style,t=M(t),"filter"!==t&&t in fo)return t;for(var e=t.charAt(0).toUpperCase()+t.slice(1),n=0;n<yo.length;n++){var i=yo[n]+e;if(i in fo)return i}}));function _o(t,e){var n=e.data,o=t.data;if(!(i(n.staticStyle)&&i(n.style)&&i(o.staticStyle)&&i(o.style))){var a,s,l=e.elm,u=o.staticStyle,c=o.normalizedStyle||o.style||{},d=u||c,h=co(e.data.style)||{};e.data.normalizedStyle=r(h.__ob__)?A({},h):h;var f=ho(e,!0);for(s in d)i(f[s])&&go(l,s,"");for(s in f)a=f[s],a!==d[s]&&go(l,s,null==a?"":a)}}var bo={create:_o,update:_o},xo=/\s+/;function wo(t,e){if(e&&(e=e.trim()))if(t.classList)e.indexOf(" ")>-1?e.split(xo).forEach((function(e){return t.classList.add(e)})):t.classList.add(e);else{var n=" "+(t.getAttribute("class")||"")+" ";n.indexOf(" "+e+" ")<0&&t.setAttribute("class",(n+e).trim())}}function Mo(t,e){if(e&&(e=e.trim()))if(t.classList)e.indexOf(" ")>-1?e.split(xo).forEach((function(e){return t.classList.remove(e)})):t.classList.remove(e),t.classList.length||t.removeAttribute("class");else{var n=" "+(t.getAttribute("class")||"")+" ",i=" "+e+" ";while(n.indexOf(i)>=0)n=n.replace(i," ");n=n.trim(),n?t.setAttribute("class",n):t.removeAttribute("class")}}function So(t){if(t){if("object"===typeof t){var e={};return!1!==t.css&&A(e,Lo(t.name||"v")),A(e,t),e}return"string"===typeof t?Lo(t):void 0}}var Lo=x((function(t){return{enterClass:t+"-enter",enterToClass:t+"-enter-to",enterActiveClass:t+"-enter-active",leaveClass:t+"-leave",leaveToClass:t+"-leave-to",leaveActiveClass:t+"-leave-active"}})),ko=Z&&!et,To="transition",Do="animation",Co="transition",Io="transitionend",Ao="animation",Yo="animationend";ko&&(void 0===window.ontransitionend&&void 0!==window.onwebkittransitionend&&(Co="WebkitTransition",Io="webkitTransitionEnd"),void 0===window.onanimationend&&void 0!==window.onwebkitanimationend&&(Ao="WebkitAnimation",Yo="webkitAnimationEnd"));var Oo=Z?window.requestAnimationFrame?window.requestAnimationFrame.bind(window):setTimeout:function(t){return t()};function Po(t){Oo((function(){Oo(t)}))}function Eo(t,e){var n=t._transitionClasses||(t._transitionClasses=[]);n.indexOf(e)<0&&(n.push(e),wo(t,e))}function No(t,e){t._transitionClasses&&v(t._transitionClasses,e),Mo(t,e)}function Ro(t,e,n){var i=jo(t,e),r=i.type,o=i.timeout,a=i.propCount;if(!r)return n();var s=r===To?Io:Yo,l=0,u=function(){t.removeEventListener(s,c),n()},c=function(e){e.target===t&&++l>=a&&u()};setTimeout((function(){l<a&&u()}),o+1),t.addEventListener(s,c)}var Ho=/\b(transform|all)(,|$)/;function jo(t,e){var n,i=window.getComputedStyle(t),r=(i[Co+"Delay"]||"").split(", "),o=(i[Co+"Duration"]||"").split(", "),a=zo(r,o),s=(i[Ao+"Delay"]||"").split(", "),l=(i[Ao+"Duration"]||"").split(", "),u=zo(s,l),c=0,d=0;e===To?a>0&&(n=To,c=a,d=o.length):e===Do?u>0&&(n=Do,c=u,d=l.length):(c=Math.max(a,u),n=c>0?a>u?To:Do:null,d=n?n===To?o.length:l.length:0);var h=n===To&&Ho.test(i[Co+"Property"]);return{type:n,timeout:c,propCount:d,hasTransform:h}}function zo(t,e){while(t.length<e.length)t=t.concat(t);return Math.max.apply(null,e.map((function(e,n){return Wo(e)+Wo(t[n])})))}function Wo(t){return 1e3*Number(t.slice(0,-1).replace(",","."))}function Bo(t,e){var n=t.elm;r(n._leaveCb)&&(n._leaveCb.cancelled=!0,n._leaveCb());var o=So(t.data.transition);if(!i(o)&&!r(n._enterCb)&&1===n.nodeType){var a=o.css,s=o.type,u=o.enterClass,c=o.enterToClass,d=o.enterActiveClass,h=o.appearClass,f=o.appearToClass,p=o.appearActiveClass,g=o.beforeEnter,y=o.enter,v=o.afterEnter,_=o.enterCancelled,b=o.beforeAppear,x=o.appear,w=o.afterAppear,M=o.appearCancelled,S=o.duration,L=In,k=In.$vnode;while(k&&k.parent)L=k.context,k=k.parent;var T=!L._isMounted||!t.isRootInsert;if(!T||x||""===x){var D=T&&h?h:u,C=T&&p?p:d,I=T&&f?f:c,A=T&&b||g,Y=T&&"function"===typeof x?x:y,O=T&&w||v,P=T&&M||_,E=m(l(S)?S.enter:S);0;var N=!1!==a&&!et,R=Go(Y),j=n._enterCb=H((function(){N&&(No(n,I),No(n,C)),j.cancelled?(N&&No(n,D),P&&P(n)):O&&O(n),n._enterCb=null}));t.data.show||xe(t,"insert",(function(){var e=n.parentNode,i=e&&e._pending&&e._pending[t.key];i&&i.tag===t.tag&&i.elm._leaveCb&&i.elm._leaveCb(),Y&&Y(n,j)})),A&&A(n),N&&(Eo(n,D),Eo(n,C),Po((function(){No(n,D),j.cancelled||(Eo(n,I),R||(Vo(E)?setTimeout(j,E):Ro(n,s,j)))}))),t.data.show&&(e&&e(),Y&&Y(n,j)),N||R||j()}}}function Fo(t,e){var n=t.elm;r(n._enterCb)&&(n._enterCb.cancelled=!0,n._enterCb());var o=So(t.data.transition);if(i(o)||1!==n.nodeType)return e();if(!r(n._leaveCb)){var a=o.css,s=o.type,u=o.leaveClass,c=o.leaveToClass,d=o.leaveActiveClass,h=o.beforeLeave,f=o.leave,p=o.afterLeave,g=o.leaveCancelled,y=o.delayLeave,v=o.duration,_=!1!==a&&!et,b=Go(f),x=m(l(v)?v.leave:v);0;var w=n._leaveCb=H((function(){n.parentNode&&n.parentNode._pending&&(n.parentNode._pending[t.key]=null),_&&(No(n,c),No(n,d)),w.cancelled?(_&&No(n,u),g&&g(n)):(e(),p&&p(n)),n._leaveCb=null}));y?y(M):M()}function M(){w.cancelled||(!t.data.show&&n.parentNode&&((n.parentNode._pending||(n.parentNode._pending={}))[t.key]=t),h&&h(n),_&&(Eo(n,u),Eo(n,d),Po((function(){No(n,u),w.cancelled||(Eo(n,c),b||(Vo(x)?setTimeout(w,x):Ro(n,s,w)))}))),f&&f(n,w),_||b||w())}}function Vo(t){return"number"===typeof t&&!isNaN(t)}function Go(t){if(i(t))return!1;var e=t.fns;return r(e)?Go(Array.isArray(e)?e[0]:e):(t._length||t.length)>1}function $o(t,e){!0!==e.data.show&&Bo(e)}var Uo=Z?{create:$o,activate:$o,remove:function(t,e){!0!==t.data.show?Fo(t,e):e()}}:{},qo=[Fr,$r,no,so,bo,Uo],Xo=qo.concat(jr),Zo=Ar({nodeOps:Mr,modules:Xo});et&&document.addEventListener("selectionchange",(function(){var t=document.activeElement;t&&t.vmodel&&ra(t,"input")}));var Jo={inserted:function(t,e,n,i){"select"===n.tag?(i.elm&&!i.elm._vOptions?xe(n,"postpatch",(function(){Jo.componentUpdated(t,e,n)})):Ko(t,e,n.context),t._vOptions=[].map.call(t.options,ea)):("textarea"===n.tag||ur(t.type))&&(t._vModifiers=e.modifiers,e.modifiers.lazy||(t.addEventListener("compositionstart",na),t.addEventListener("compositionend",ia),t.addEventListener("change",ia),et&&(t.vmodel=!0)))},componentUpdated:function(t,e,n){if("select"===n.tag){Ko(t,e,n.context);var i=t._vOptions,r=t._vOptions=[].map.call(t.options,ea);if(r.some((function(t,e){return!N(t,i[e])}))){var o=t.multiple?e.value.some((function(t){return ta(t,r)})):e.value!==e.oldValue&&ta(e.value,r);o&&ra(t,"change")}}}};function Ko(t,e,n){Qo(t,e,n),(tt||nt)&&setTimeout((function(){Qo(t,e,n)}),0)}function Qo(t,e,n){var i=e.value,r=t.multiple;if(!r||Array.isArray(i)){for(var o,a,s=0,l=t.options.length;s<l;s++)if(a=t.options[s],r)o=R(i,ea(a))>-1,a.selected!==o&&(a.selected=o);else if(N(ea(a),i))return void(t.selectedIndex!==s&&(t.selectedIndex=s));r||(t.selectedIndex=-1)}}function ta(t,e){return e.every((function(e){return!N(e,t)}))}function ea(t){return"_value"in t?t._value:t.value}function na(t){t.target.composing=!0}function ia(t){t.target.composing&&(t.target.composing=!1,ra(t.target,"input"))}function ra(t,e){var n=document.createEvent("HTMLEvents");n.initEvent(e,!0,!0),t.dispatchEvent(n)}function oa(t){return!t.componentInstance||t.data&&t.data.transition?t:oa(t.componentInstance._vnode)}var aa={bind:function(t,e,n){var i=e.value;n=oa(n);var r=n.data&&n.data.transition,o=t.__vOriginalDisplay="none"===t.style.display?"":t.style.display;i&&r?(n.data.show=!0,Bo(n,(function(){t.style.display=o}))):t.style.display=i?o:"none"},update:function(t,e,n){var i=e.value,r=e.oldValue;if(!i!==!r){n=oa(n);var o=n.data&&n.data.transition;o?(n.data.show=!0,i?Bo(n,(function(){t.style.display=t.__vOriginalDisplay})):Fo(n,(function(){t.style.display="none"}))):t.style.display=i?t.__vOriginalDisplay:"none"}},unbind:function(t,e,n,i,r){r||(t.style.display=t.__vOriginalDisplay)}},sa={model:Jo,show:aa},la={name:String,appear:Boolean,css:Boolean,mode:String,type:String,enterClass:String,leaveClass:String,enterToClass:String,leaveToClass:String,enterActiveClass:String,leaveActiveClass:String,appearClass:String,appearActiveClass:String,appearToClass:String,duration:[Number,String,Object]};function ua(t){var e=t&&t.componentOptions;return e&&e.Ctor.options.abstract?ua(Mn(e.children)):t}function ca(t){var e={},n=t.$options;for(var i in n.propsData)e[i]=t[i];var r=n._parentListeners;for(var o in r)e[M(o)]=r[o];return e}function da(t,e){if(/\d-keep-alive$/.test(e.tag))return t("keep-alive",{props:e.componentOptions.propsData})}function ha(t){while(t=t.parent)if(t.data.transition)return!0}function fa(t,e){return e.key===t.key&&e.tag===t.tag}var pa=function(t){return t.tag||wn(t)},ma=function(t){return"show"===t.name},ga={name:"transition",props:la,abstract:!0,render:function(t){var e=this,n=this.$slots.default;if(n&&(n=n.filter(pa),n.length)){0;var i=this.mode;0;var r=n[0];if(ha(this.$vnode))return r;var o=ua(r);if(!o)return r;if(this._leaving)return da(t,r);var a="__transition-"+this._uid+"-";o.key=null==o.key?o.isComment?a+"comment":a+o.tag:s(o.key)?0===String(o.key).indexOf(a)?o.key:a+o.key:o.key;var l=(o.data||(o.data={})).transition=ca(this),u=this._vnode,c=ua(u);if(o.data.directives&&o.data.directives.some(ma)&&(o.data.show=!0),c&&c.data&&!fa(o,c)&&!wn(c)&&(!c.componentInstance||!c.componentInstance._vnode.isComment)){var d=c.data.transition=A({},l);if("out-in"===i)return this._leaving=!0,xe(d,"afterLeave",(function(){e._leaving=!1,e.$forceUpdate()})),da(t,r);if("in-out"===i){if(wn(o))return u;var h,f=function(){h()};xe(l,"afterEnter",f),xe(l,"enterCancelled",f),xe(d,"delayLeave",(function(t){h=t}))}}return r}}},ya=A({tag:String,moveClass:String},la);delete ya.mode;var va={props:ya,beforeMount:function(){var t=this,e=this._update;this._update=function(n,i){var r=An(t);t.__patch__(t._vnode,t.kept,!1,!0),t._vnode=t.kept,r(),e.call(t,n,i)}},render:function(t){for(var e=this.tag||this.$vnode.data.tag||"span",n=Object.create(null),i=this.prevChildren=this.children,r=this.$slots.default||[],o=this.children=[],a=ca(this),s=0;s<r.length;s++){var l=r[s];if(l.tag)if(null!=l.key&&0!==String(l.key).indexOf("__vlist"))o.push(l),n[l.key]=l,(l.data||(l.data={})).transition=a;else;}if(i){for(var u=[],c=[],d=0;d<i.length;d++){var h=i[d];h.data.transition=a,h.data.pos=h.elm.getBoundingClientRect(),n[h.key]?u.push(h):c.push(h)}this.kept=t(e,null,u),this.removed=c}return t(e,null,o)},updated:function(){var t=this.prevChildren,e=this.moveClass||(this.name||"v")+"-move";t.length&&this.hasMove(t[0].elm,e)&&(t.forEach(_a),t.forEach(ba),t.forEach(xa),this._reflow=document.body.offsetHeight,t.forEach((function(t){if(t.data.moved){var n=t.elm,i=n.style;Eo(n,e),i.transform=i.WebkitTransform=i.transitionDuration="",n.addEventListener(Io,n._moveCb=function t(i){i&&i.target!==n||i&&!/transform$/.test(i.propertyName)||(n.removeEventListener(Io,t),n._moveCb=null,No(n,e))})}})))},methods:{hasMove:function(t,e){if(!ko)return!1;if(this._hasMove)return this._hasMove;var n=t.cloneNode();t._transitionClasses&&t._transitionClasses.forEach((function(t){Mo(n,t)})),wo(n,e),n.style.display="none",this.$el.appendChild(n);var i=jo(n);return this.$el.removeChild(n),this._hasMove=i.hasTransform}}};function _a(t){t.elm._moveCb&&t.elm._moveCb(),t.elm._enterCb&&t.elm._enterCb()}function ba(t){t.data.newPos=t.elm.getBoundingClientRect()}function xa(t){var e=t.data.pos,n=t.data.newPos,i=e.left-n.left,r=e.top-n.top;if(i||r){t.data.moved=!0;var o=t.elm.style;o.transform=o.WebkitTransform="translate("+i+"px,"+r+"px)",o.transitionDuration="0s"}}var wa={Transition:ga,TransitionGroup:va};Mi.config.mustUseProp=zi,Mi.config.isReservedTag=or,Mi.config.isReservedAttr=Hi,Mi.config.getTagNamespace=ar,Mi.config.isUnknownElement=lr,A(Mi.options.directives,sa),A(Mi.options.components,wa),Mi.prototype.__patch__=Z?Zo:O,Mi.prototype.$mount=function(t,e){return t=t&&Z?cr(t):void 0,Pn(this,t,e)},Z&&setTimeout((function(){B.devtools&&ut&&ut.emit("init",Mi)}),0),e["a"]=Mi}).call(this,n("c8ba"))},"2b10":function(t,e){function n(t,e,n){var i=-1,r=t.length;e<0&&(e=-e>r?0:r+e),n=n>r?r:n,n<0&&(n+=r),r=e>n?0:n-e>>>0,e>>>=0;var o=Array(r);while(++i<r)o[i]=t[i+e];return o}t.exports=n},"2b3e":function(t,e,n){var i=n("585a"),r="object"==typeof self&&self&&self.Object===Object&&self,o=i||r||Function("return this")();t.exports=o},"2bfb":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("af",{months:"Januarie_Februarie_Maart_April_Mei_Junie_Julie_Augustus_September_Oktober_November_Desember".split("_"),monthsShort:"Jan_Feb_Mrt_Apr_Mei_Jun_Jul_Aug_Sep_Okt_Nov_Des".split("_"),weekdays:"Sondag_Maandag_Dinsdag_Woensdag_Donderdag_Vrydag_Saterdag".split("_"),weekdaysShort:"Son_Maa_Din_Woe_Don_Vry_Sat".split("_"),weekdaysMin:"So_Ma_Di_Wo_Do_Vr_Sa".split("_"),meridiemParse:/vm|nm/i,isPM:function(t){return/^nm$/i.test(t)},meridiem:function(t,e,n){return t<12?n?"vm":"VM":n?"nm":"NM"},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Vandag om] LT",nextDay:"[Môre om] LT",nextWeek:"dddd [om] LT",lastDay:"[Gister om] LT",lastWeek:"[Laas] dddd [om] LT",sameElse:"L"},relativeTime:{future:"oor %s",past:"%s gelede",s:"'n paar sekondes",ss:"%d sekondes",m:"'n minuut",mm:"%d minute",h:"'n uur",hh:"%d ure",d:"'n dag",dd:"%d dae",M:"'n maand",MM:"%d maande",y:"'n jaar",yy:"%d jaar"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return e}))},"2cef":function(t,e,n){"use strict";t.exports=function(){var t=1;function e(){return t++}return{generate:e}}},"2cf4":function(t,e,n){var i,r,o,a=n("da84"),s=n("d039"),l=n("0366"),u=n("1be4"),c=n("cc12"),d=n("1cdc"),h=n("605d"),f=a.location,p=a.setImmediate,m=a.clearImmediate,g=a.process,y=a.MessageChannel,v=a.Dispatch,_=0,b={},x="onreadystatechange",w=function(t){if(b.hasOwnProperty(t)){var e=b[t];delete b[t],e()}},M=function(t){return function(){w(t)}},S=function(t){w(t.data)},L=function(t){a.postMessage(t+"",f.protocol+"//"+f.host)};p&&m||(p=function(t){var e=[],n=1;while(arguments.length>n)e.push(arguments[n++]);return b[++_]=function(){("function"==typeof t?t:Function(t)).apply(void 0,e)},i(_),_},m=function(t){delete b[t]},h?i=function(t){g.nextTick(M(t))}:v&&v.now?i=function(t){v.now(M(t))}:y&&!d?(r=new y,o=r.port2,r.port1.onmessage=S,i=l(o.postMessage,o,1)):a.addEventListener&&"function"==typeof postMessage&&!a.importScripts&&f&&"file:"!==f.protocol&&!s(L)?(i=L,a.addEventListener("message",S,!1)):i=x in c("script")?function(t){u.appendChild(c("script"))[x]=function(){u.removeChild(this),w(t)}}:function(t){setTimeout(M(t),0)}),t.exports={set:p,clear:m}},"2d00":function(t,e,n){var i,r,o=n("da84"),a=n("342f"),s=o.process,l=s&&s.versions,u=l&&l.v8;u?(i=u.split("."),r=i[0]+i[1]):a&&(i=a.match(/Edge\/(\d+)/),(!i||i[1]>=74)&&(i=a.match(/Chrome\/(\d+)/),i&&(r=i[1]))),t.exports=r&&+r},"2d48":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("label",{class:t.wrapClasses},[n("span",{class:t.radioClasses},[n("span",{class:t.innerClasses}),n("input",{class:t.inputClasses,attrs:{type:"radio",disabled:t.itemDisabled,name:t.groupName},domProps:{checked:t.currentValue},on:{change:t.change,focus:t.onFocus,blur:t.onBlur}})]),t._t("default",[t._v(t._s(t.label))])],2)},r=[],o=(n("b0c0"),n("a9e3"),n("ade3")),a=n("3a27"),s=n("2507"),l=n("9f81"),u="ivu-radio",c={name:"Radio",mixins:[s["a"],l["a"]],props:{value:{type:[String,Number,Boolean],default:!1},trueValue:{type:[String,Number,Boolean],default:!0},falseValue:{type:[String,Number,Boolean],default:!1},label:{type:[String,Number]},disabled:{type:Boolean,default:!1},size:{validator:function(t){return Object(a["i"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},name:{type:String},border:{type:Boolean,default:!1}},data:function(){return{currentValue:this.value,group:!1,groupName:this.name,parent:Object(a["e"])(this,"RadioGroup"),focusWrapper:!1,focusInner:!1}},computed:{wrapClasses:function(){var t;return["".concat(u,"-wrapper"),(t={},Object(o["a"])(t,"".concat(u,"-group-item"),this.group),Object(o["a"])(t,"".concat(u,"-wrapper-checked"),this.currentValue),Object(o["a"])(t,"".concat(u,"-wrapper-disabled"),this.itemDisabled),Object(o["a"])(t,"".concat(u,"-").concat(this.size),!!this.size),Object(o["a"])(t,"".concat(u,"-focus"),this.focusWrapper),Object(o["a"])(t,"".concat(u,"-border"),this.border),t)]},radioClasses:function(){var t;return["".concat(u),(t={},Object(o["a"])(t,"".concat(u,"-checked"),this.currentValue),Object(o["a"])(t,"".concat(u,"-disabled"),this.itemDisabled),t)]},innerClasses:function(){return["".concat(u,"-inner"),Object(o["a"])({},"".concat(u,"-focus"),this.focusInner)]},inputClasses:function(){return"".concat(u,"-input")}},mounted:function(){this.parent&&(this.group=!0,this.name&&this.name!==this.parent.name?console.warn:this.groupName=this.parent.name),this.group?this.parent.updateValue():this.updateValue()},methods:{change:function(t){if(this.itemDisabled)return!1;var e=t.target.checked;this.currentValue=e;var n=e?this.trueValue:this.falseValue;this.$emit("input",n),this.group?void 0!==this.label&&this.parent.change({value:this.label,checked:this.value}):(this.$emit("on-change",n),this.dispatch("FormItem","on-form-change",n))},updateValue:function(){this.currentValue=this.value===this.trueValue},onBlur:function(){this.focusWrapper=!1,this.focusInner=!1},onFocus:function(){this.group&&"button"===this.parent.type?this.focusWrapper=!0:this.focusInner=!0}},watch:{value:function(t){if(t!==this.trueValue&&t!==this.falseValue)throw"Value should be trueValue or falseValue.";this.updateValue()}}},d=c,h=n("2877"),f=Object(h["a"])(d,i,r,!1,null,null,null),p=f.exports,m=n("2430");p.Group=m["a"];e["a"]=p},"2d7c":function(t,e){function n(t,e){var n=-1,i=null==t?0:t.length,r=0,o=[];while(++n<i){var a=t[n];e(a,n,t)&&(o[r++]=a)}return o}t.exports=n},"2d83":function(t,e,n){"use strict";var i=n("387f");t.exports=function(t,e,n,r,o){var a=new Error(t);return i(a,e,n,r,o)}},"2dcb":function(t,e,n){var i=n("91e9"),r=i(Object.getPrototypeOf,Object);t.exports=r},"2e67":function(t,e,n){"use strict";t.exports=function(t){return!(!t||!t.__CANCEL__)}},"2e8c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("uz",{months:"январ_феврал_март_апрел_май_июн_июл_август_сентябр_октябр_ноябр_декабр".split("_"),monthsShort:"янв_фев_мар_апр_май_июн_июл_авг_сен_окт_ноя_дек".split("_"),weekdays:"Якшанба_Душанба_Сешанба_Чоршанба_Пайшанба_Жума_Шанба".split("_"),weekdaysShort:"Якш_Душ_Сеш_Чор_Пай_Жум_Шан".split("_"),weekdaysMin:"Як_Ду_Се_Чо_Па_Жу_Ша".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"D MMMM YYYY, dddd HH:mm"},calendar:{sameDay:"[Бугун соат] LT [да]",nextDay:"[Эртага] LT [да]",nextWeek:"dddd [куни соат] LT [да]",lastDay:"[Кеча соат] LT [да]",lastWeek:"[Утган] dddd [куни соат] LT [да]",sameElse:"L"},relativeTime:{future:"Якин %s ичида",past:"Бир неча %s олдин",s:"фурсат",ss:"%d фурсат",m:"бир дакика",mm:"%d дакика",h:"бир соат",hh:"%d соат",d:"бир кун",dd:"%d кун",M:"бир ой",MM:"%d ой",y:"бир йил",yy:"%d йил"},week:{dow:1,doy:7}});return e}))},"2ec1":function(t,e,n){var i=n("100e"),r=n("9aff");function o(t){return i((function(e,n){var i=-1,o=n.length,a=o>1?n[o-1]:void 0,s=o>2?n[2]:void 0;a=t.length>3&&"function"==typeof a?(o--,a):void 0,s&&r(n[0],n[1],s)&&(a=o<3?void 0:a,o=1),e=Object(e);while(++i<o){var l=n[i];l&&t(e,l,i,a)}return e}))}t.exports=o},"2f62":function(t,e,n){"use strict";(function(t){
/*!
 * vuex v3.6.0
 * (c) 2020 Evan You
 * @license MIT
 */
function n(t){var e=Number(t.version.split(".")[0]);if(e>=2)t.mixin({beforeCreate:i});else{var n=t.prototype._init;t.prototype._init=function(t){void 0===t&&(t={}),t.init=t.init?[i].concat(t.init):i,n.call(this,t)}}function i(){var t=this.$options;t.store?this.$store="function"===typeof t.store?t.store():t.store:t.parent&&t.parent.$store&&(this.$store=t.parent.$store)}}var i="undefined"!==typeof window?window:"undefined"!==typeof t?t:{},r=i.__VUE_DEVTOOLS_GLOBAL_HOOK__;function o(t){r&&(t._devtoolHook=r,r.emit("vuex:init",t),r.on("vuex:travel-to-state",(function(e){t.replaceState(e)})),t.subscribe((function(t,e){r.emit("vuex:mutation",t,e)}),{prepend:!0}),t.subscribeAction((function(t,e){r.emit("vuex:action",t,e)}),{prepend:!0}))}function a(t,e){return t.filter(e)[0]}function s(t,e){if(void 0===e&&(e=[]),null===t||"object"!==typeof t)return t;var n=a(e,(function(e){return e.original===t}));if(n)return n.copy;var i=Array.isArray(t)?[]:{};return e.push({original:t,copy:i}),Object.keys(t).forEach((function(n){i[n]=s(t[n],e)})),i}function l(t,e){Object.keys(t).forEach((function(n){return e(t[n],n)}))}function u(t){return null!==t&&"object"===typeof t}function c(t){return t&&"function"===typeof t.then}function d(t,e){return function(){return t(e)}}var h=function(t,e){this.runtime=e,this._children=Object.create(null),this._rawModule=t;var n=t.state;this.state=("function"===typeof n?n():n)||{}},f={namespaced:{configurable:!0}};f.namespaced.get=function(){return!!this._rawModule.namespaced},h.prototype.addChild=function(t,e){this._children[t]=e},h.prototype.removeChild=function(t){delete this._children[t]},h.prototype.getChild=function(t){return this._children[t]},h.prototype.hasChild=function(t){return t in this._children},h.prototype.update=function(t){this._rawModule.namespaced=t.namespaced,t.actions&&(this._rawModule.actions=t.actions),t.mutations&&(this._rawModule.mutations=t.mutations),t.getters&&(this._rawModule.getters=t.getters)},h.prototype.forEachChild=function(t){l(this._children,t)},h.prototype.forEachGetter=function(t){this._rawModule.getters&&l(this._rawModule.getters,t)},h.prototype.forEachAction=function(t){this._rawModule.actions&&l(this._rawModule.actions,t)},h.prototype.forEachMutation=function(t){this._rawModule.mutations&&l(this._rawModule.mutations,t)},Object.defineProperties(h.prototype,f);var p=function(t){this.register([],t,!1)};function m(t,e,n){if(e.update(n),n.modules)for(var i in n.modules){if(!e.getChild(i))return void 0;m(t.concat(i),e.getChild(i),n.modules[i])}}p.prototype.get=function(t){return t.reduce((function(t,e){return t.getChild(e)}),this.root)},p.prototype.getNamespace=function(t){var e=this.root;return t.reduce((function(t,n){return e=e.getChild(n),t+(e.namespaced?n+"/":"")}),"")},p.prototype.update=function(t){m([],this.root,t)},p.prototype.register=function(t,e,n){var i=this;void 0===n&&(n=!0);var r=new h(e,n);if(0===t.length)this.root=r;else{var o=this.get(t.slice(0,-1));o.addChild(t[t.length-1],r)}e.modules&&l(e.modules,(function(e,r){i.register(t.concat(r),e,n)}))},p.prototype.unregister=function(t){var e=this.get(t.slice(0,-1)),n=t[t.length-1],i=e.getChild(n);i&&i.runtime&&e.removeChild(n)},p.prototype.isRegistered=function(t){var e=this.get(t.slice(0,-1)),n=t[t.length-1];return!!e&&e.hasChild(n)};var g;var y=function(t){var e=this;void 0===t&&(t={}),!g&&"undefined"!==typeof window&&window.Vue&&A(window.Vue);var n=t.plugins;void 0===n&&(n=[]);var i=t.strict;void 0===i&&(i=!1),this._committing=!1,this._actions=Object.create(null),this._actionSubscribers=[],this._mutations=Object.create(null),this._wrappedGetters=Object.create(null),this._modules=new p(t),this._modulesNamespaceMap=Object.create(null),this._subscribers=[],this._watcherVM=new g,this._makeLocalGettersCache=Object.create(null);var r=this,a=this,s=a.dispatch,l=a.commit;this.dispatch=function(t,e){return s.call(r,t,e)},this.commit=function(t,e,n){return l.call(r,t,e,n)},this.strict=i;var u=this._modules.root.state;w(this,u,[],this._modules.root),x(this,u),n.forEach((function(t){return t(e)}));var c=void 0!==t.devtools?t.devtools:g.config.devtools;c&&o(this)},v={state:{configurable:!0}};function _(t,e,n){return e.indexOf(t)<0&&(n&&n.prepend?e.unshift(t):e.push(t)),function(){var n=e.indexOf(t);n>-1&&e.splice(n,1)}}function b(t,e){t._actions=Object.create(null),t._mutations=Object.create(null),t._wrappedGetters=Object.create(null),t._modulesNamespaceMap=Object.create(null);var n=t.state;w(t,n,[],t._modules.root,!0),x(t,n,e)}function x(t,e,n){var i=t._vm;t.getters={},t._makeLocalGettersCache=Object.create(null);var r=t._wrappedGetters,o={};l(r,(function(e,n){o[n]=d(e,t),Object.defineProperty(t.getters,n,{get:function(){return t._vm[n]},enumerable:!0})}));var a=g.config.silent;g.config.silent=!0,t._vm=new g({data:{$$state:e},computed:o}),g.config.silent=a,t.strict&&D(t),i&&(n&&t._withCommit((function(){i._data.$$state=null})),g.nextTick((function(){return i.$destroy()})))}function w(t,e,n,i,r){var o=!n.length,a=t._modules.getNamespace(n);if(i.namespaced&&(t._modulesNamespaceMap[a],t._modulesNamespaceMap[a]=i),!o&&!r){var s=C(e,n.slice(0,-1)),l=n[n.length-1];t._withCommit((function(){g.set(s,l,i.state)}))}var u=i.context=M(t,a,n);i.forEachMutation((function(e,n){var i=a+n;L(t,i,e,u)})),i.forEachAction((function(e,n){var i=e.root?n:a+n,r=e.handler||e;k(t,i,r,u)})),i.forEachGetter((function(e,n){var i=a+n;T(t,i,e,u)})),i.forEachChild((function(i,o){w(t,e,n.concat(o),i,r)}))}function M(t,e,n){var i=""===e,r={dispatch:i?t.dispatch:function(n,i,r){var o=I(n,i,r),a=o.payload,s=o.options,l=o.type;return s&&s.root||(l=e+l),t.dispatch(l,a)},commit:i?t.commit:function(n,i,r){var o=I(n,i,r),a=o.payload,s=o.options,l=o.type;s&&s.root||(l=e+l),t.commit(l,a,s)}};return Object.defineProperties(r,{getters:{get:i?function(){return t.getters}:function(){return S(t,e)}},state:{get:function(){return C(t.state,n)}}}),r}function S(t,e){if(!t._makeLocalGettersCache[e]){var n={},i=e.length;Object.keys(t.getters).forEach((function(r){if(r.slice(0,i)===e){var o=r.slice(i);Object.defineProperty(n,o,{get:function(){return t.getters[r]},enumerable:!0})}})),t._makeLocalGettersCache[e]=n}return t._makeLocalGettersCache[e]}function L(t,e,n,i){var r=t._mutations[e]||(t._mutations[e]=[]);r.push((function(e){n.call(t,i.state,e)}))}function k(t,e,n,i){var r=t._actions[e]||(t._actions[e]=[]);r.push((function(e){var r=n.call(t,{dispatch:i.dispatch,commit:i.commit,getters:i.getters,state:i.state,rootGetters:t.getters,rootState:t.state},e);return c(r)||(r=Promise.resolve(r)),t._devtoolHook?r.catch((function(e){throw t._devtoolHook.emit("vuex:error",e),e})):r}))}function T(t,e,n,i){t._wrappedGetters[e]||(t._wrappedGetters[e]=function(t){return n(i.state,i.getters,t.state,t.getters)})}function D(t){t._vm.$watch((function(){return this._data.$$state}),(function(){0}),{deep:!0,sync:!0})}function C(t,e){return e.reduce((function(t,e){return t[e]}),t)}function I(t,e,n){return u(t)&&t.type&&(n=e,e=t,t=t.type),{type:t,payload:e,options:n}}function A(t){g&&t===g||(g=t,n(g))}v.state.get=function(){return this._vm._data.$$state},v.state.set=function(t){0},y.prototype.commit=function(t,e,n){var i=this,r=I(t,e,n),o=r.type,a=r.payload,s=(r.options,{type:o,payload:a}),l=this._mutations[o];l&&(this._withCommit((function(){l.forEach((function(t){t(a)}))})),this._subscribers.slice().forEach((function(t){return t(s,i.state)})))},y.prototype.dispatch=function(t,e){var n=this,i=I(t,e),r=i.type,o=i.payload,a={type:r,payload:o},s=this._actions[r];if(s){try{this._actionSubscribers.slice().filter((function(t){return t.before})).forEach((function(t){return t.before(a,n.state)}))}catch(u){0}var l=s.length>1?Promise.all(s.map((function(t){return t(o)}))):s[0](o);return new Promise((function(t,e){l.then((function(e){try{n._actionSubscribers.filter((function(t){return t.after})).forEach((function(t){return t.after(a,n.state)}))}catch(u){0}t(e)}),(function(t){try{n._actionSubscribers.filter((function(t){return t.error})).forEach((function(e){return e.error(a,n.state,t)}))}catch(u){0}e(t)}))}))}},y.prototype.subscribe=function(t,e){return _(t,this._subscribers,e)},y.prototype.subscribeAction=function(t,e){var n="function"===typeof t?{before:t}:t;return _(n,this._actionSubscribers,e)},y.prototype.watch=function(t,e,n){var i=this;return this._watcherVM.$watch((function(){return t(i.state,i.getters)}),e,n)},y.prototype.replaceState=function(t){var e=this;this._withCommit((function(){e._vm._data.$$state=t}))},y.prototype.registerModule=function(t,e,n){void 0===n&&(n={}),"string"===typeof t&&(t=[t]),this._modules.register(t,e),w(this,this.state,t,this._modules.get(t),n.preserveState),x(this,this.state)},y.prototype.unregisterModule=function(t){var e=this;"string"===typeof t&&(t=[t]),this._modules.unregister(t),this._withCommit((function(){var n=C(e.state,t.slice(0,-1));g.delete(n,t[t.length-1])})),b(this)},y.prototype.hasModule=function(t){return"string"===typeof t&&(t=[t]),this._modules.isRegistered(t)},y.prototype.hotUpdate=function(t){this._modules.update(t),b(this,!0)},y.prototype._withCommit=function(t){var e=this._committing;this._committing=!0,t(),this._committing=e},Object.defineProperties(y.prototype,v);var Y=j((function(t,e){var n={};return R(e).forEach((function(e){var i=e.key,r=e.val;n[i]=function(){var e=this.$store.state,n=this.$store.getters;if(t){var i=z(this.$store,"mapState",t);if(!i)return;e=i.context.state,n=i.context.getters}return"function"===typeof r?r.call(this,e,n):e[r]},n[i].vuex=!0})),n})),O=j((function(t,e){var n={};return R(e).forEach((function(e){var i=e.key,r=e.val;n[i]=function(){var e=[],n=arguments.length;while(n--)e[n]=arguments[n];var i=this.$store.commit;if(t){var o=z(this.$store,"mapMutations",t);if(!o)return;i=o.context.commit}return"function"===typeof r?r.apply(this,[i].concat(e)):i.apply(this.$store,[r].concat(e))}})),n})),P=j((function(t,e){var n={};return R(e).forEach((function(e){var i=e.key,r=e.val;r=t+r,n[i]=function(){if(!t||z(this.$store,"mapGetters",t))return this.$store.getters[r]},n[i].vuex=!0})),n})),E=j((function(t,e){var n={};return R(e).forEach((function(e){var i=e.key,r=e.val;n[i]=function(){var e=[],n=arguments.length;while(n--)e[n]=arguments[n];var i=this.$store.dispatch;if(t){var o=z(this.$store,"mapActions",t);if(!o)return;i=o.context.dispatch}return"function"===typeof r?r.apply(this,[i].concat(e)):i.apply(this.$store,[r].concat(e))}})),n})),N=function(t){return{mapState:Y.bind(null,t),mapGetters:P.bind(null,t),mapMutations:O.bind(null,t),mapActions:E.bind(null,t)}};function R(t){return H(t)?Array.isArray(t)?t.map((function(t){return{key:t,val:t}})):Object.keys(t).map((function(e){return{key:e,val:t[e]}})):[]}function H(t){return Array.isArray(t)||u(t)}function j(t){return function(e,n){return"string"!==typeof e?(n=e,e=""):"/"!==e.charAt(e.length-1)&&(e+="/"),t(e,n)}}function z(t,e,n){var i=t._modulesNamespaceMap[n];return i}function W(t){void 0===t&&(t={});var e=t.collapsed;void 0===e&&(e=!0);var n=t.filter;void 0===n&&(n=function(t,e,n){return!0});var i=t.transformer;void 0===i&&(i=function(t){return t});var r=t.mutationTransformer;void 0===r&&(r=function(t){return t});var o=t.actionFilter;void 0===o&&(o=function(t,e){return!0});var a=t.actionTransformer;void 0===a&&(a=function(t){return t});var l=t.logMutations;void 0===l&&(l=!0);var u=t.logActions;void 0===u&&(u=!0);var c=t.logger;return void 0===c&&(c=console),function(t){var d=s(t.state);"undefined"!==typeof c&&(l&&t.subscribe((function(t,o){var a=s(o);if(n(t,d,a)){var l=V(),u=r(t),h="mutation "+t.type+l;B(c,h,e),c.log("%c prev state","color: #9E9E9E; font-weight: bold",i(d)),c.log("%c mutation","color: #03A9F4; font-weight: bold",u),c.log("%c next state","color: #4CAF50; font-weight: bold",i(a)),F(c)}d=a})),u&&t.subscribeAction((function(t,n){if(o(t,n)){var i=V(),r=a(t),s="action "+t.type+i;B(c,s,e),c.log("%c action","color: #03A9F4; font-weight: bold",r),F(c)}})))}}function B(t,e,n){var i=n?t.groupCollapsed:t.group;try{i.call(t,e)}catch(r){t.log(e)}}function F(t){try{t.groupEnd()}catch(e){t.log("—— log end ——")}}function V(){var t=new Date;return" @ "+$(t.getHours(),2)+":"+$(t.getMinutes(),2)+":"+$(t.getSeconds(),2)+"."+$(t.getMilliseconds(),3)}function G(t,e){return new Array(e+1).join(t)}function $(t,e){return G("0",e-t.toString().length)+t}var U={Store:y,install:A,version:"3.6.0",mapState:Y,mapMutations:O,mapGetters:P,mapActions:E,createNamespacedHelpers:N,createLogger:W};e["a"]=U}).call(this,n("c8ba"))},"2fcc":function(t,e){function n(t){var e=this.__data__,n=e["delete"](t);return this.size=e.size,n}t.exports=n},"30b5":function(t,e,n){"use strict";var i=n("c532");function r(t){return encodeURIComponent(t).replace(/%3A/gi,":").replace(/%24/g,"$").replace(/%2C/gi,",").replace(/%20/g,"+").replace(/%5B/gi,"[").replace(/%5D/gi,"]")}t.exports=function(t,e,n){if(!e)return t;var o;if(n)o=n(e);else if(i.isURLSearchParams(e))o=e.toString();else{var a=[];i.forEach(e,(function(t,e){null!==t&&"undefined"!==typeof t&&(i.isArray(t)?e+="[]":t=[t],i.forEach(t,(function(t){i.isDate(t)?t=t.toISOString():i.isObject(t)&&(t=JSON.stringify(t)),a.push(r(e)+"="+r(t))})))})),o=a.join("&")}if(o){var s=t.indexOf("#");-1!==s&&(t=t.slice(0,s)),t+=(-1===t.indexOf("?")?"?":"&")+o}return t}},"30c9":function(t,e,n){var i=n("9520"),r=n("b218");function o(t){return null!=t&&r(t.length)&&!i(t)}t.exports=o},"30e0":function(t,e,n){var i=n("99cd"),r=i(!0);t.exports=r},"313e":function(t,e,n){"use strict";n.d(e,"a",(function(){return tM}));var i={};n.r(i),n.d(i,"extendShape",(function(){return zd})),n.d(i,"extendPath",(function(){return Bd})),n.d(i,"registerShape",(function(){return Fd})),n.d(i,"getShapeClass",(function(){return Vd})),n.d(i,"makePath",(function(){return Gd})),n.d(i,"makeImage",(function(){return $d})),n.d(i,"mergePath",(function(){return qd})),n.d(i,"resizePath",(function(){return Xd})),n.d(i,"subPixelOptimizeLine",(function(){return Zd})),n.d(i,"subPixelOptimizeRect",(function(){return Jd})),n.d(i,"subPixelOptimize",(function(){return Kd})),n.d(i,"updateProps",(function(){return th})),n.d(i,"initProps",(function(){return eh})),n.d(i,"removeElement",(function(){return nh})),n.d(i,"removeElementWithFadeOut",(function(){return rh})),n.d(i,"isElementRemoved",(function(){return oh})),n.d(i,"getTransform",(function(){return ah})),n.d(i,"applyTransform",(function(){return sh})),n.d(i,"transformDirection",(function(){return lh})),n.d(i,"groupTransition",(function(){return dh})),n.d(i,"clipPointsByRect",(function(){return hh})),n.d(i,"clipRectByRect",(function(){return fh})),n.d(i,"createIcon",(function(){return ph})),n.d(i,"linePolygonIntersect",(function(){return mh})),n.d(i,"lineLineIntersect",(function(){return gh})),n.d(i,"setTooltipConfig",(function(){return _h})),n.d(i,"Group",(function(){return Mr})),n.d(i,"Image",(function(){return wl})),n.d(i,"Text",(function(){return Gl})),n.d(i,"Circle",(function(){return Dc})),n.d(i,"Ellipse",(function(){return Ac})),n.d(i,"Sector",(function(){return qc})),n.d(i,"Ring",(function(){return Jc})),n.d(i,"Polygon",(function(){return rd})),n.d(i,"Polyline",(function(){return sd})),n.d(i,"Rect",(function(){return Al})),n.d(i,"Line",(function(){return dd})),n.d(i,"BezierCurve",(function(){return gd})),n.d(i,"Arc",(function(){return _d})),n.d(i,"IncrementalDisplayable",(function(){return Nd})),n.d(i,"CompoundPath",(function(){return xd})),n.d(i,"LinearGradient",(function(){return Ld})),n.d(i,"RadialGradient",(function(){return Td})),n.d(i,"BoundingRect",(function(){return Ji})),n.d(i,"OrientedBoundingRect",(function(){return Od})),n.d(i,"Point",(function(){return Wi})),n.d(i,"Path",(function(){return pl}));
/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
var r=function(t,e){return r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)Object.prototype.hasOwnProperty.call(e,n)&&(t[n]=e[n])},r(t,e)};function o(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var a=function(){return a=Object.assign||function(t){for(var e,n=1,i=arguments.length;n<i;n++)for(var r in e=arguments[n],e)Object.prototype.hasOwnProperty.call(e,r)&&(t[r]=e[r]);return t},a.apply(this,arguments)};Object.create;function s(){for(var t=0,e=0,n=arguments.length;e<n;e++)t+=arguments[e].length;var i=Array(t),r=0;for(e=0;e<n;e++)for(var o=arguments[e],a=0,s=o.length;a<s;a++,r++)i[r]=o[a];return i}Object.create;var l=function(){function t(){this.firefox=!1,this.ie=!1,this.edge=!1,this.newEdge=!1,this.weChat=!1}return t}(),u=function(){function t(){this.browser=new l,this.node=!1,this.wxa=!1,this.worker=!1,this.canvasSupported=!1,this.svgSupported=!1,this.touchEventsSupported=!1,this.pointerEventsSupported=!1,this.domSupported=!1,this.transformSupported=!1,this.transform3dSupported=!1}return t}(),c=new u;function d(t,e){var n=e.browser,i=t.match(/Firefox\/([\d.]+)/),r=t.match(/MSIE\s([\d.]+)/)||t.match(/Trident\/.+?rv:(([\d.]+))/),o=t.match(/Edge?\/([\d.]+)/),a=/micromessenger/i.test(t);i&&(n.firefox=!0,n.version=i[1]),r&&(n.ie=!0,n.version=r[1]),o&&(n.edge=!0,n.version=o[1],n.newEdge=+o[1].split(".")[0]>18),a&&(n.weChat=!0),e.canvasSupported=!!document.createElement("canvas").getContext,e.svgSupported="undefined"!==typeof SVGRect,e.touchEventsSupported="ontouchstart"in window&&!n.ie&&!n.edge,e.pointerEventsSupported="onpointerdown"in window&&(n.edge||n.ie&&+n.version>=11),e.domSupported="undefined"!==typeof document;var s=document.documentElement.style;e.transform3dSupported=(n.ie&&"transition"in s||n.edge||"WebKitCSSMatrix"in window&&"m11"in new WebKitCSSMatrix||"MozPerspective"in s)&&!("OTransition"in s),e.transformSupported=e.transform3dSupported||n.ie&&+n.version>=9}"object"===typeof wx&&"function"===typeof wx.getSystemInfoSync?(c.wxa=!0,c.canvasSupported=!0,c.touchEventsSupported=!0):"undefined"===typeof document&&"undefined"!==typeof self?(c.worker=!0,c.canvasSupported=!0):"undefined"===typeof navigator?(c.node=!0,c.canvasSupported=!0,c.svgSupported=!0):d(navigator.userAgent,c);var h=c,f={"[object Function]":!0,"[object RegExp]":!0,"[object Date]":!0,"[object Error]":!0,"[object CanvasGradient]":!0,"[object CanvasPattern]":!0,"[object Image]":!0,"[object Canvas]":!0},p={"[object Int8Array]":!0,"[object Uint8Array]":!0,"[object Uint8ClampedArray]":!0,"[object Int16Array]":!0,"[object Uint16Array]":!0,"[object Int32Array]":!0,"[object Uint32Array]":!0,"[object Float32Array]":!0,"[object Float64Array]":!0},m=Object.prototype.toString,g=Array.prototype,y=g.forEach,v=g.filter,_=g.slice,b=g.map,x=function(){}.constructor,w=x?x.prototype:null,M={};var S=2311;function L(){return S++}function k(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e]}function T(t){if(null==t||"object"!==typeof t)return t;var e=t,n=m.call(t);if("[object Array]"===n){if(!pt(t)){e=[];for(var i=0,r=t.length;i<r;i++)e[i]=T(t[i])}}else if(p[n]){if(!pt(t)){var o=t.constructor;if(o.from)e=o.from(t);else{e=new o(t.length);for(i=0,r=t.length;i<r;i++)e[i]=T(t[i])}}}else if(!f[n]&&!pt(t)&&!tt(t))for(var a in e={},t)t.hasOwnProperty(a)&&(e[a]=T(t[a]));return e}function D(t,e,n){if(!J(e)||!J(t))return n?T(e):t;for(var i in e)if(e.hasOwnProperty(i)){var r=t[i],o=e[i];!J(o)||!J(r)||$(o)||$(r)||tt(o)||tt(r)||K(o)||K(r)||pt(o)||pt(r)?!n&&i in t||(t[i]=T(e[i])):D(r,o,n)}return t}function C(t,e){for(var n=t[0],i=1,r=t.length;i<r;i++)n=D(n,t[i],e);return n}function I(t,e){if(Object.assign)Object.assign(t,e);else for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n]);return t}function A(t,e,n){for(var i=B(e),r=0;r<i.length;r++){var o=i[r];(n?null!=e[o]:null==t[o])&&(t[o]=e[o])}return t}var Y=function(){return M.createCanvas()};function O(t,e){if(t){if(t.indexOf)return t.indexOf(e);for(var n=0,i=t.length;n<i;n++)if(t[n]===e)return n}return-1}function P(t,e){var n=t.prototype;function i(){}for(var r in i.prototype=e.prototype,t.prototype=new i,n)n.hasOwnProperty(r)&&(t.prototype[r]=n[r]);t.prototype.constructor=t,t.superClass=e}function E(t,e,n){if(t="prototype"in t?t.prototype:t,e="prototype"in e?e.prototype:e,Object.getOwnPropertyNames)for(var i=Object.getOwnPropertyNames(e),r=0;r<i.length;r++){var o=i[r];"constructor"!==o&&(n?null!=e[o]:null==t[o])&&(t[o]=e[o])}else A(t,e,n)}function N(t){return!!t&&("string"!==typeof t&&"number"===typeof t.length)}function R(t,e,n){if(t&&e)if(t.forEach&&t.forEach===y)t.forEach(e,n);else if(t.length===+t.length)for(var i=0,r=t.length;i<r;i++)e.call(n,t[i],i,t);else for(var o in t)t.hasOwnProperty(o)&&e.call(n,t[o],o,t)}function H(t,e,n){if(!t)return[];if(!e)return lt(t);if(t.map&&t.map===b)return t.map(e,n);for(var i=[],r=0,o=t.length;r<o;r++)i.push(e.call(n,t[r],r,t));return i}function j(t,e,n,i){if(t&&e){for(var r=0,o=t.length;r<o;r++)n=e.call(i,n,t[r],r,t);return n}}function z(t,e,n){if(!t)return[];if(!e)return lt(t);if(t.filter&&t.filter===v)return t.filter(e,n);for(var i=[],r=0,o=t.length;r<o;r++)e.call(n,t[r],r,t)&&i.push(t[r]);return i}function W(t,e,n){if(t&&e)for(var i=0,r=t.length;i<r;i++)if(e.call(n,t[i],i,t))return t[i]}function B(t){if(!t)return[];if(Object.keys)return Object.keys(t);var e=[];for(var n in t)t.hasOwnProperty(n)&&e.push(n);return e}function F(t,e){for(var n=[],i=2;i<arguments.length;i++)n[i-2]=arguments[i];return function(){return t.apply(e,n.concat(_.call(arguments)))}}M.createCanvas=function(){return document.createElement("canvas")};var V=w&&U(w.bind)?w.call.bind(w.bind):F;function G(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return function(){return t.apply(this,e.concat(_.call(arguments)))}}function $(t){return Array.isArray?Array.isArray(t):"[object Array]"===m.call(t)}function U(t){return"function"===typeof t}function q(t){return"string"===typeof t}function X(t){return"[object String]"===m.call(t)}function Z(t){return"number"===typeof t}function J(t){var e=typeof t;return"function"===e||!!t&&"object"===e}function K(t){return!!f[m.call(t)]}function Q(t){return!!p[m.call(t)]}function tt(t){return"object"===typeof t&&"number"===typeof t.nodeType&&"object"===typeof t.ownerDocument}function et(t){return null!=t.colorStops}function nt(t){return null!=t.image}function it(t){return"[object RegExp]"===m.call(t)}function rt(t){return t!==t}function ot(){for(var t=[],e=0;e<arguments.length;e++)t[e]=arguments[e];for(var n=0,i=t.length;n<i;n++)if(null!=t[n])return t[n]}function at(t,e){return null!=t?t:e}function st(t,e,n){return null!=t?t:null!=e?e:n}function lt(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];return _.apply(t,e)}function ut(t){if("number"===typeof t)return[t,t,t,t];var e=t.length;return 2===e?[t[0],t[1],t[0],t[1]]:3===e?[t[0],t[1],t[2],t[1]]:t}function ct(t,e){if(!t)throw new Error(e)}function dt(t){return null==t?null:"function"===typeof t.trim?t.trim():t.replace(/^[\s\uFEFF\xA0]+|[\s\uFEFF\xA0]+$/g,"")}var ht="__ec_primitive__";function ft(t){t[ht]=!0}function pt(t){return t[ht]}var mt=function(){function t(e){this.data={};var n=$(e);this.data={};var i=this;function r(t,e){n?i.set(t,e):i.set(e,t)}e instanceof t?e.each(r):e&&R(e,r)}return t.prototype.get=function(t){return this.data.hasOwnProperty(t)?this.data[t]:null},t.prototype.set=function(t,e){return this.data[t]=e},t.prototype.each=function(t,e){for(var n in this.data)this.data.hasOwnProperty(n)&&t.call(e,this.data[n],n)},t.prototype.keys=function(){return B(this.data)},t.prototype.removeKey=function(t){delete this.data[t]},t}();function gt(t){return new mt(t)}function yt(t,e){for(var n=new t.constructor(t.length+e.length),i=0;i<t.length;i++)n[i]=t[i];var r=t.length;for(i=0;i<e.length;i++)n[i+r]=e[i];return n}function vt(t,e){var n;if(Object.create)n=Object.create(t);else{var i=function(){};i.prototype=t,n=new i}return e&&I(n,e),n}function _t(t,e){return t.hasOwnProperty(e)}function bt(){}
/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */var xt=function(t,e){return xt=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)Object.prototype.hasOwnProperty.call(e,n)&&(t[n]=e[n])},xt(t,e)};function wt(t,e){function n(){this.constructor=t}xt(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}Object.create;Object.create;function Mt(t,e){return null==t&&(t=0),null==e&&(e=0),[t,e]}function St(t,e){return t[0]=e[0],t[1]=e[1],t}function Lt(t){return[t[0],t[1]]}function kt(t,e,n){return t[0]=e,t[1]=n,t}function Tt(t,e,n){return t[0]=e[0]+n[0],t[1]=e[1]+n[1],t}function Dt(t,e,n,i){return t[0]=e[0]+n[0]*i,t[1]=e[1]+n[1]*i,t}function Ct(t,e,n){return t[0]=e[0]-n[0],t[1]=e[1]-n[1],t}function It(t){return Math.sqrt(At(t))}function At(t){return t[0]*t[0]+t[1]*t[1]}function Yt(t,e,n){return t[0]=e[0]*n,t[1]=e[1]*n,t}function Ot(t,e){var n=It(e);return 0===n?(t[0]=0,t[1]=0):(t[0]=e[0]/n,t[1]=e[1]/n),t}function Pt(t,e){return Math.sqrt((t[0]-e[0])*(t[0]-e[0])+(t[1]-e[1])*(t[1]-e[1]))}var Et=Pt;function Nt(t,e){return(t[0]-e[0])*(t[0]-e[0])+(t[1]-e[1])*(t[1]-e[1])}var Rt=Nt;function Ht(t,e,n,i){return t[0]=e[0]+i*(n[0]-e[0]),t[1]=e[1]+i*(n[1]-e[1]),t}function jt(t,e,n){var i=e[0],r=e[1];return t[0]=n[0]*i+n[2]*r+n[4],t[1]=n[1]*i+n[3]*r+n[5],t}function zt(t,e,n){return t[0]=Math.min(e[0],n[0]),t[1]=Math.min(e[1],n[1]),t}function Wt(t,e,n){return t[0]=Math.max(e[0],n[0]),t[1]=Math.max(e[1],n[1]),t}var Bt=function(){function t(t,e){this.target=t,this.topTarget=e&&e.topTarget}return t}(),Ft=function(){function t(t){this.handler=t,t.on("mousedown",this._dragStart,this),t.on("mousemove",this._drag,this),t.on("mouseup",this._dragEnd,this)}return t.prototype._dragStart=function(t){var e=t.target;while(e&&!e.draggable)e=e.parent;e&&(this._draggingTarget=e,e.dragging=!0,this._x=t.offsetX,this._y=t.offsetY,this.handler.dispatchToElement(new Bt(e,t),"dragstart",t.event))},t.prototype._drag=function(t){var e=this._draggingTarget;if(e){var n=t.offsetX,i=t.offsetY,r=n-this._x,o=i-this._y;this._x=n,this._y=i,e.drift(r,o,t),this.handler.dispatchToElement(new Bt(e,t),"drag",t.event);var a=this.handler.findHover(n,i,e).target,s=this._dropTarget;this._dropTarget=a,e!==a&&(s&&a!==s&&this.handler.dispatchToElement(new Bt(s,t),"dragleave",t.event),a&&a!==s&&this.handler.dispatchToElement(new Bt(a,t),"dragenter",t.event))}},t.prototype._dragEnd=function(t){var e=this._draggingTarget;e&&(e.dragging=!1),this.handler.dispatchToElement(new Bt(e,t),"dragend",t.event),this._dropTarget&&this.handler.dispatchToElement(new Bt(this._dropTarget,t),"drop",t.event),this._draggingTarget=null,this._dropTarget=null},t}(),Vt=Ft,Gt=function(){function t(t){t&&(this._$eventProcessor=t)}return t.prototype.on=function(t,e,n,i){this._$handlers||(this._$handlers={});var r=this._$handlers;if("function"===typeof e&&(i=n,n=e,e=null),!n||!t)return this;var o=this._$eventProcessor;null!=e&&o&&o.normalizeQuery&&(e=o.normalizeQuery(e)),r[t]||(r[t]=[]);for(var a=0;a<r[t].length;a++)if(r[t][a].h===n)return this;var s={h:n,query:e,ctx:i||this,callAtLast:n.zrEventfulCallAtLast},l=r[t].length-1,u=r[t][l];return u&&u.callAtLast?r[t].splice(l,0,s):r[t].push(s),this},t.prototype.isSilent=function(t){var e=this._$handlers;return!e||!e[t]||!e[t].length},t.prototype.off=function(t,e){var n=this._$handlers;if(!n)return this;if(!t)return this._$handlers={},this;if(e){if(n[t]){for(var i=[],r=0,o=n[t].length;r<o;r++)n[t][r].h!==e&&i.push(n[t][r]);n[t]=i}n[t]&&0===n[t].length&&delete n[t]}else delete n[t];return this},t.prototype.trigger=function(t){for(var e=[],n=1;n<arguments.length;n++)e[n-1]=arguments[n];if(!this._$handlers)return this;var i=this._$handlers[t],r=this._$eventProcessor;if(i)for(var o=e.length,a=i.length,s=0;s<a;s++){var l=i[s];if(!r||!r.filter||null==l.query||r.filter(t,l.query))switch(o){case 0:l.h.call(l.ctx);break;case 1:l.h.call(l.ctx,e[0]);break;case 2:l.h.call(l.ctx,e[0],e[1]);break;default:l.h.apply(l.ctx,e);break}}return r&&r.afterTrigger&&r.afterTrigger(t),this},t.prototype.triggerWithContext=function(t){if(!this._$handlers)return this;var e=this._$handlers[t],n=this._$eventProcessor;if(e)for(var i=arguments,r=i.length,o=i[r-1],a=e.length,s=0;s<a;s++){var l=e[s];if(!n||!n.filter||null==l.query||n.filter(t,l.query))switch(r){case 0:l.h.call(o);break;case 1:l.h.call(o,i[0]);break;case 2:l.h.call(o,i[0],i[1]);break;default:l.h.apply(o,i.slice(1,r-1));break}}return n&&n.afterTrigger&&n.afterTrigger(t),this},t}(),$t=Gt,Ut=Math.log(2);function qt(t,e,n,i,r,o){var a=i+"-"+r,s=t.length;if(o.hasOwnProperty(a))return o[a];if(1===e){var l=Math.round(Math.log((1<<s)-1&~r)/Ut);return t[n][l]}var u=i|1<<n,c=n+1;while(i&1<<c)c++;for(var d=0,h=0,f=0;h<s;h++){var p=1<<h;p&r||(d+=(f%2?-1:1)*t[n][h]*qt(t,e-1,c,u,r|p,o),f++)}return o[a]=d,d}function Xt(t,e){var n=[[t[0],t[1],1,0,0,0,-e[0]*t[0],-e[0]*t[1]],[0,0,0,t[0],t[1],1,-e[1]*t[0],-e[1]*t[1]],[t[2],t[3],1,0,0,0,-e[2]*t[2],-e[2]*t[3]],[0,0,0,t[2],t[3],1,-e[3]*t[2],-e[3]*t[3]],[t[4],t[5],1,0,0,0,-e[4]*t[4],-e[4]*t[5]],[0,0,0,t[4],t[5],1,-e[5]*t[4],-e[5]*t[5]],[t[6],t[7],1,0,0,0,-e[6]*t[6],-e[6]*t[7]],[0,0,0,t[6],t[7],1,-e[7]*t[6],-e[7]*t[7]]],i={},r=qt(n,8,0,0,0,i);if(0!==r){for(var o=[],a=0;a<8;a++)for(var s=0;s<8;s++)null==o[s]&&(o[s]=0),o[s]+=((a+s)%2?-1:1)*qt(n,7,0===a?1:0,1<<a,1<<s,i)/r*e[a];return function(t,e,n){var i=e*o[6]+n*o[7]+1;t[0]=(e*o[0]+n*o[1]+o[2])/i,t[1]=(e*o[3]+n*o[4]+o[5])/i}}}var Zt="___zrEVENTSAVED",Jt=[];function Kt(t,e,n,i,r){return Qt(Jt,e,i,r,!0)&&Qt(t,n,Jt[0],Jt[1])}function Qt(t,e,n,i,r){if(e.getBoundingClientRect&&h.domSupported&&!ne(e)){var o=e[Zt]||(e[Zt]={}),a=te(e,o),s=ee(a,o,r);if(s)return s(t,n,i),!0}return!1}function te(t,e){var n=e.markers;if(n)return n;n=e.markers=[];for(var i=["left","right"],r=["top","bottom"],o=0;o<4;o++){var a=document.createElement("div"),s=a.style,l=o%2,u=(o>>1)%2;s.cssText=["position: absolute","visibility: hidden","padding: 0","margin: 0","border-width: 0","user-select: none","width:0","height:0",i[l]+":0",r[u]+":0",i[1-l]+":auto",r[1-u]+":auto",""].join("!important;"),t.appendChild(a),n.push(a)}return n}function ee(t,e,n){for(var i=n?"invTrans":"trans",r=e[i],o=e.srcCoords,a=[],s=[],l=!0,u=0;u<4;u++){var c=t[u].getBoundingClientRect(),d=2*u,h=c.left,f=c.top;a.push(h,f),l=l&&o&&h===o[d]&&f===o[d+1],s.push(t[u].offsetLeft,t[u].offsetTop)}return l&&r?r:(e.srcCoords=a,e[i]=n?Xt(s,a):Xt(a,s))}function ne(t){return"CANVAS"===t.nodeName.toUpperCase()}var ie="undefined"!==typeof window&&!!window.addEventListener,re=/^(?:mouse|pointer|contextmenu|drag|drop)|click/,oe=[];function ae(t,e,n,i){return n=n||{},i||!h.canvasSupported?se(t,e,n):h.browser.firefox&&null!=e.layerX&&e.layerX!==e.offsetX?(n.zrX=e.layerX,n.zrY=e.layerY):null!=e.offsetX?(n.zrX=e.offsetX,n.zrY=e.offsetY):se(t,e,n),n}function se(t,e,n){if(h.domSupported&&t.getBoundingClientRect){var i=e.clientX,r=e.clientY;if(ne(t)){var o=t.getBoundingClientRect();return n.zrX=i-o.left,void(n.zrY=r-o.top)}if(Qt(oe,t,i,r))return n.zrX=oe[0],void(n.zrY=oe[1])}n.zrX=n.zrY=0}function le(t){return t||window.event}function ue(t,e,n){if(e=le(e),null!=e.zrX)return e;var i=e.type,r=i&&i.indexOf("touch")>=0;if(r){var o="touchend"!==i?e.targetTouches[0]:e.changedTouches[0];o&&ae(t,o,e,n)}else{ae(t,e,e,n);var a=ce(e);e.zrDelta=a?a/120:-(e.detail||0)/3}var s=e.button;return null==e.which&&void 0!==s&&re.test(e.type)&&(e.which=1&s?1:2&s?3:4&s?2:0),e}function ce(t){var e=t.wheelDelta;if(e)return e;var n=t.deltaX,i=t.deltaY;if(null==n||null==i)return e;var r=0!==i?Math.abs(i):Math.abs(n),o=i>0?-1:i<0?1:n>0?-1:1;return 3*r*o}function de(t,e,n,i){ie?t.addEventListener(e,n,i):t.attachEvent("on"+e,n)}function he(t,e,n,i){ie?t.removeEventListener(e,n,i):t.detachEvent("on"+e,n)}var fe=ie?function(t){t.preventDefault(),t.stopPropagation(),t.cancelBubble=!0}:function(t){t.returnValue=!1,t.cancelBubble=!0};function pe(t){return 2===t.which||3===t.which}var me=function(){function t(){this._track=[]}return t.prototype.recognize=function(t,e,n){return this._doTrack(t,e,n),this._recognize(t)},t.prototype.clear=function(){return this._track.length=0,this},t.prototype._doTrack=function(t,e,n){var i=t.touches;if(i){for(var r={points:[],touches:[],target:e,event:t},o=0,a=i.length;o<a;o++){var s=i[o],l=ae(n,s,{});r.points.push([l.zrX,l.zrY]),r.touches.push(s)}this._track.push(r)}},t.prototype._recognize=function(t){for(var e in ve)if(ve.hasOwnProperty(e)){var n=ve[e](this._track,t);if(n)return n}},t}();function ge(t){var e=t[1][0]-t[0][0],n=t[1][1]-t[0][1];return Math.sqrt(e*e+n*n)}function ye(t){return[(t[0][0]+t[1][0])/2,(t[0][1]+t[1][1])/2]}var ve={pinch:function(t,e){var n=t.length;if(n){var i=(t[n-1]||{}).points,r=(t[n-2]||{}).points||i;if(r&&r.length>1&&i&&i.length>1){var o=ge(i)/ge(r);!isFinite(o)&&(o=1),e.pinchScale=o;var a=ye(i);return e.pinchX=a[0],e.pinchY=a[1],{type:"pinch",target:t[0].target,event:e}}}}},_e="silent";function be(t,e,n){return{type:t,event:n,target:e.target,topTarget:e.topTarget,cancelBubble:!1,offsetX:n.zrX,offsetY:n.zrY,gestureEvent:n.gestureEvent,pinchX:n.pinchX,pinchY:n.pinchY,pinchScale:n.pinchScale,wheelDelta:n.zrDelta,zrByTouch:n.zrByTouch,which:n.which,stop:xe}}function xe(){fe(this.event)}var we=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.handler=null,e}return wt(e,t),e.prototype.dispose=function(){},e.prototype.setCursor=function(){},e}($t),Me=function(){function t(t,e){this.x=t,this.y=e}return t}(),Se=["click","dblclick","mousewheel","mouseout","mouseup","mousedown","mousemove","contextmenu"],Le=function(t){function e(e,n,i,r){var o=t.call(this)||this;return o._hovered=new Me(0,0),o.storage=e,o.painter=n,o.painterRoot=r,i=i||new we,o.proxy=null,o.setHandlerProxy(i),o._draggingMgr=new Vt(o),o}return wt(e,t),e.prototype.setHandlerProxy=function(t){this.proxy&&this.proxy.dispose(),t&&(R(Se,(function(e){t.on&&t.on(e,this[e],this)}),this),t.handler=this),this.proxy=t},e.prototype.mousemove=function(t){var e=t.zrX,n=t.zrY,i=Te(this,e,n),r=this._hovered,o=r.target;o&&!o.__zr&&(r=this.findHover(r.x,r.y),o=r.target);var a=this._hovered=i?new Me(e,n):this.findHover(e,n),s=a.target,l=this.proxy;l.setCursor&&l.setCursor(s?s.cursor:"default"),o&&s!==o&&this.dispatchToElement(r,"mouseout",t),this.dispatchToElement(a,"mousemove",t),s&&s!==o&&this.dispatchToElement(a,"mouseover",t)},e.prototype.mouseout=function(t){var e=t.zrEventControl;"only_globalout"!==e&&this.dispatchToElement(this._hovered,"mouseout",t),"no_globalout"!==e&&this.trigger("globalout",{type:"globalout",event:t})},e.prototype.resize=function(){this._hovered=new Me(0,0)},e.prototype.dispatch=function(t,e){var n=this[t];n&&n.call(this,e)},e.prototype.dispose=function(){this.proxy.dispose(),this.storage=null,this.proxy=null,this.painter=null},e.prototype.setCursorStyle=function(t){var e=this.proxy;e.setCursor&&e.setCursor(t)},e.prototype.dispatchToElement=function(t,e,n){t=t||{};var i=t.target;if(!i||!i.silent){var r="on"+e,o=be(e,t,n);while(i)if(i[r]&&(o.cancelBubble=!!i[r].call(i,o)),i.trigger(e,o),i=i.__hostTarget?i.__hostTarget:i.parent,o.cancelBubble)break;o.cancelBubble||(this.trigger(e,o),this.painter&&this.painter.eachOtherLayer&&this.painter.eachOtherLayer((function(t){"function"===typeof t[r]&&t[r].call(t,o),t.trigger&&t.trigger(e,o)})))}},e.prototype.findHover=function(t,e,n){for(var i=this.storage.getDisplayList(),r=new Me(t,e),o=i.length-1;o>=0;o--){var a=void 0;if(i[o]!==n&&!i[o].ignore&&(a=ke(i[o],t,e))&&(!r.topTarget&&(r.topTarget=i[o]),a!==_e)){r.target=i[o];break}}return r},e.prototype.processGesture=function(t,e){this._gestureMgr||(this._gestureMgr=new me);var n=this._gestureMgr;"start"===e&&n.clear();var i=n.recognize(t,this.findHover(t.zrX,t.zrY,null).target,this.proxy.dom);if("end"===e&&n.clear(),i){var r=i.type;t.gestureEvent=r;var o=new Me;o.target=i.target,this.dispatchToElement(o,r,i.event)}},e}($t);function ke(t,e,n){if(t[t.rectHover?"rectContain":"contain"](e,n)){var i=t,r=void 0,o=!1;while(i){if(i.ignoreClip&&(o=!0),!o){var a=i.getClipPath();if(a&&!a.contain(e,n))return!1;i.silent&&(r=!0)}var s=i.__hostTarget;i=s||i.parent}return!r||_e}return!1}function Te(t,e,n){var i=t.painter;return e<0||e>i.getWidth()||n<0||n>i.getHeight()}R(["click","mousedown","mouseup","mousewheel","dblclick","contextmenu"],(function(t){Le.prototype[t]=function(e){var n,i,r=e.zrX,o=e.zrY,a=Te(this,r,o);if("mouseup"===t&&a||(n=this.findHover(r,o),i=n.target),"mousedown"===t)this._downEl=i,this._downPoint=[e.zrX,e.zrY],this._upEl=i;else if("mouseup"===t)this._upEl=i;else if("click"===t){if(this._downEl!==this._upEl||!this._downPoint||Et(this._downPoint,[e.zrX,e.zrY])>4)return;this._downPoint=null}this.dispatchToElement(n,t,e)}}));var De=Le,Ce=32,Ie=7;function Ae(t){var e=0;while(t>=Ce)e|=1&t,t>>=1;return t+e}function Ye(t,e,n,i){var r=e+1;if(r===n)return 1;if(i(t[r++],t[e])<0){while(r<n&&i(t[r],t[r-1])<0)r++;Oe(t,e,r)}else while(r<n&&i(t[r],t[r-1])>=0)r++;return r-e}function Oe(t,e,n){n--;while(e<n){var i=t[e];t[e++]=t[n],t[n--]=i}}function Pe(t,e,n,i,r){for(i===e&&i++;i<n;i++){var o,a=t[i],s=e,l=i;while(s<l)o=s+l>>>1,r(a,t[o])<0?l=o:s=o+1;var u=i-s;switch(u){case 3:t[s+3]=t[s+2];case 2:t[s+2]=t[s+1];case 1:t[s+1]=t[s];break;default:while(u>0)t[s+u]=t[s+u-1],u--}t[s]=a}}function Ee(t,e,n,i,r,o){var a=0,s=0,l=1;if(o(t,e[n+r])>0){s=i-r;while(l<s&&o(t,e[n+r+l])>0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s),a+=r,l+=r}else{s=r+1;while(l<s&&o(t,e[n+r-l])<=0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s);var u=a;a=r-l,l=r-u}a++;while(a<l){var c=a+(l-a>>>1);o(t,e[n+c])>0?a=c+1:l=c}return l}function Ne(t,e,n,i,r,o){var a=0,s=0,l=1;if(o(t,e[n+r])<0){s=r+1;while(l<s&&o(t,e[n+r-l])<0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s);var u=a;a=r-l,l=r-u}else{s=i-r;while(l<s&&o(t,e[n+r+l])>=0)a=l,l=1+(l<<1),l<=0&&(l=s);l>s&&(l=s),a+=r,l+=r}a++;while(a<l){var c=a+(l-a>>>1);o(t,e[n+c])<0?l=c:a=c+1}return l}function Re(t,e){var n,i,r=Ie,o=0,a=0;o=t.length;var s=[];function l(t,e){n[a]=t,i[a]=e,a+=1}function u(){while(a>1){var t=a-2;if(t>=1&&i[t-1]<=i[t]+i[t+1]||t>=2&&i[t-2]<=i[t]+i[t-1])i[t-1]<i[t+1]&&t--;else if(i[t]>i[t+1])break;d(t)}}function c(){while(a>1){var t=a-2;t>0&&i[t-1]<i[t+1]&&t--,d(t)}}function d(r){var o=n[r],s=i[r],l=n[r+1],u=i[r+1];i[r]=s+u,r===a-3&&(n[r+1]=n[r+2],i[r+1]=i[r+2]),a--;var c=Ne(t[l],t,o,s,0,e);o+=c,s-=c,0!==s&&(u=Ee(t[o+s-1],t,l,u,u-1,e),0!==u&&(s<=u?h(o,s,l,u):f(o,s,l,u)))}function h(n,i,o,a){var l=0;for(l=0;l<i;l++)s[l]=t[n+l];var u=0,c=o,d=n;if(t[d++]=t[c++],0!==--a)if(1!==i){var h,f,p,m=r;while(1){h=0,f=0,p=!1;do{if(e(t[c],s[u])<0){if(t[d++]=t[c++],f++,h=0,0===--a){p=!0;break}}else if(t[d++]=s[u++],h++,f=0,1===--i){p=!0;break}}while((h|f)<m);if(p)break;do{if(h=Ne(t[c],s,u,i,0,e),0!==h){for(l=0;l<h;l++)t[d+l]=s[u+l];if(d+=h,u+=h,i-=h,i<=1){p=!0;break}}if(t[d++]=t[c++],0===--a){p=!0;break}if(f=Ee(s[u],t,c,a,0,e),0!==f){for(l=0;l<f;l++)t[d+l]=t[c+l];if(d+=f,c+=f,a-=f,0===a){p=!0;break}}if(t[d++]=s[u++],1===--i){p=!0;break}m--}while(h>=Ie||f>=Ie);if(p)break;m<0&&(m=0),m+=2}if(r=m,r<1&&(r=1),1===i){for(l=0;l<a;l++)t[d+l]=t[c+l];t[d+a]=s[u]}else{if(0===i)throw new Error;for(l=0;l<i;l++)t[d+l]=s[u+l]}}else{for(l=0;l<a;l++)t[d+l]=t[c+l];t[d+a]=s[u]}else for(l=0;l<i;l++)t[d+l]=s[u+l]}function f(n,i,o,a){var l=0;for(l=0;l<a;l++)s[l]=t[o+l];var u=n+i-1,c=a-1,d=o+a-1,h=0,f=0;if(t[d--]=t[u--],0!==--i)if(1!==a){var p=r;while(1){var m=0,g=0,y=!1;do{if(e(s[c],t[u])<0){if(t[d--]=t[u--],m++,g=0,0===--i){y=!0;break}}else if(t[d--]=s[c--],g++,m=0,1===--a){y=!0;break}}while((m|g)<p);if(y)break;do{if(m=i-Ne(s[c],t,n,i,i-1,e),0!==m){for(d-=m,u-=m,i-=m,f=d+1,h=u+1,l=m-1;l>=0;l--)t[f+l]=t[h+l];if(0===i){y=!0;break}}if(t[d--]=s[c--],1===--a){y=!0;break}if(g=a-Ee(t[u],s,0,a,a-1,e),0!==g){for(d-=g,c-=g,a-=g,f=d+1,h=c+1,l=0;l<g;l++)t[f+l]=s[h+l];if(a<=1){y=!0;break}}if(t[d--]=t[u--],0===--i){y=!0;break}p--}while(m>=Ie||g>=Ie);if(y)break;p<0&&(p=0),p+=2}if(r=p,r<1&&(r=1),1===a){for(d-=i,u-=i,f=d+1,h=u+1,l=i-1;l>=0;l--)t[f+l]=t[h+l];t[d]=s[c]}else{if(0===a)throw new Error;for(h=d-(a-1),l=0;l<a;l++)t[h+l]=s[l]}}else{for(d-=i,u-=i,f=d+1,h=u+1,l=i-1;l>=0;l--)t[f+l]=t[h+l];t[d]=s[c]}else for(h=d-(a-1),l=0;l<a;l++)t[h+l]=s[l]}return o<120?5:o<1542?10:o<119151?19:40,n=[],i=[],{mergeRuns:u,forceMergeRuns:c,pushRun:l}}function He(t,e,n,i){n||(n=0),i||(i=t.length);var r=i-n;if(!(r<2)){var o=0;if(r<Ce)return o=Ye(t,n,i,e),void Pe(t,n,i,n+o,e);var a=Re(t,e),s=Ae(r);do{if(o=Ye(t,n,i,e),o<s){var l=r;l>s&&(l=s),Pe(t,n,n+l,n+o,e),o=l}a.pushRun(n,o),a.mergeRuns(),r-=o,n+=o}while(0!==r);a.forceMergeRuns()}}var je=1,ze=2,We=4,Be=!1;function Fe(){Be||(Be=!0)}function Ve(t,e){return t.zlevel===e.zlevel?t.z===e.z?t.z2-e.z2:t.z-e.z:t.zlevel-e.zlevel}var Ge,$e=function(){function t(){this._roots=[],this._displayList=[],this._displayListLen=0,this.displayableSortFunc=Ve}return t.prototype.traverse=function(t,e){for(var n=0;n<this._roots.length;n++)this._roots[n].traverse(t,e)},t.prototype.getDisplayList=function(t,e){e=e||!1;var n=this._displayList;return!t&&n.length||this.updateDisplayList(e),n},t.prototype.updateDisplayList=function(t){this._displayListLen=0;for(var e=this._roots,n=this._displayList,i=0,r=e.length;i<r;i++)this._updateAndAddDisplayable(e[i],null,t);n.length=this._displayListLen,h.canvasSupported&&He(n,Ve)},t.prototype._updateAndAddDisplayable=function(t,e,n){if(!t.ignore||n){t.beforeUpdate(),t.update(),t.afterUpdate();var i=t.getClipPath();if(t.ignoreClip)e=null;else if(i){e=e?e.slice():[];var r=i,o=t;while(r)r.parent=o,r.updateTransform(),e.push(r),o=r,r=r.getClipPath()}if(t.childrenRef){for(var a=t.childrenRef(),s=0;s<a.length;s++){var l=a[s];t.__dirty&&(l.__dirty|=je),this._updateAndAddDisplayable(l,e,n)}t.__dirty=0}else{var u=t;e&&e.length?u.__clipPaths=e:u.__clipPaths&&u.__clipPaths.length>0&&(u.__clipPaths=[]),isNaN(u.z)&&(Fe(),u.z=0),isNaN(u.z2)&&(Fe(),u.z2=0),isNaN(u.zlevel)&&(Fe(),u.zlevel=0),this._displayList[this._displayListLen++]=u}var c=t.getDecalElement&&t.getDecalElement();c&&this._updateAndAddDisplayable(c,e,n);var d=t.getTextGuideLine();d&&this._updateAndAddDisplayable(d,e,n);var h=t.getTextContent();h&&this._updateAndAddDisplayable(h,e,n)}},t.prototype.addRoot=function(t){t.__zr&&t.__zr.storage===this||this._roots.push(t)},t.prototype.delRoot=function(t){if(t instanceof Array)for(var e=0,n=t.length;e<n;e++)this.delRoot(t[e]);else{var i=O(this._roots,t);i>=0&&this._roots.splice(i,1)}},t.prototype.delAllRoots=function(){this._roots=[],this._displayList=[],this._displayListLen=0},t.prototype.getRoots=function(){return this._roots},t.prototype.dispose=function(){this._displayList=null,this._roots=null},t}(),Ue=$e;Ge="undefined"!==typeof window&&(window.requestAnimationFrame&&window.requestAnimationFrame.bind(window)||window.msRequestAnimationFrame&&window.msRequestAnimationFrame.bind(window)||window.mozRequestAnimationFrame||window.webkitRequestAnimationFrame)||function(t){return setTimeout(t,16)};var qe=Ge,Xe={linear:function(t){return t},quadraticIn:function(t){return t*t},quadraticOut:function(t){return t*(2-t)},quadraticInOut:function(t){return(t*=2)<1?.5*t*t:-.5*(--t*(t-2)-1)},cubicIn:function(t){return t*t*t},cubicOut:function(t){return--t*t*t+1},cubicInOut:function(t){return(t*=2)<1?.5*t*t*t:.5*((t-=2)*t*t+2)},quarticIn:function(t){return t*t*t*t},quarticOut:function(t){return 1- --t*t*t*t},quarticInOut:function(t){return(t*=2)<1?.5*t*t*t*t:-.5*((t-=2)*t*t*t-2)},quinticIn:function(t){return t*t*t*t*t},quinticOut:function(t){return--t*t*t*t*t+1},quinticInOut:function(t){return(t*=2)<1?.5*t*t*t*t*t:.5*((t-=2)*t*t*t*t+2)},sinusoidalIn:function(t){return 1-Math.cos(t*Math.PI/2)},sinusoidalOut:function(t){return Math.sin(t*Math.PI/2)},sinusoidalInOut:function(t){return.5*(1-Math.cos(Math.PI*t))},exponentialIn:function(t){return 0===t?0:Math.pow(1024,t-1)},exponentialOut:function(t){return 1===t?1:1-Math.pow(2,-10*t)},exponentialInOut:function(t){return 0===t?0:1===t?1:(t*=2)<1?.5*Math.pow(1024,t-1):.5*(2-Math.pow(2,-10*(t-1)))},circularIn:function(t){return 1-Math.sqrt(1-t*t)},circularOut:function(t){return Math.sqrt(1- --t*t)},circularInOut:function(t){return(t*=2)<1?-.5*(Math.sqrt(1-t*t)-1):.5*(Math.sqrt(1-(t-=2)*t)+1)},elasticIn:function(t){var e,n=.1,i=.4;return 0===t?0:1===t?1:(!n||n<1?(n=1,e=i/4):e=i*Math.asin(1/n)/(2*Math.PI),-n*Math.pow(2,10*(t-=1))*Math.sin((t-e)*(2*Math.PI)/i))},elasticOut:function(t){var e,n=.1,i=.4;return 0===t?0:1===t?1:(!n||n<1?(n=1,e=i/4):e=i*Math.asin(1/n)/(2*Math.PI),n*Math.pow(2,-10*t)*Math.sin((t-e)*(2*Math.PI)/i)+1)},elasticInOut:function(t){var e,n=.1,i=.4;return 0===t?0:1===t?1:(!n||n<1?(n=1,e=i/4):e=i*Math.asin(1/n)/(2*Math.PI),(t*=2)<1?n*Math.pow(2,10*(t-=1))*Math.sin((t-e)*(2*Math.PI)/i)*-.5:n*Math.pow(2,-10*(t-=1))*Math.sin((t-e)*(2*Math.PI)/i)*.5+1)},backIn:function(t){var e=1.70158;return t*t*((e+1)*t-e)},backOut:function(t){var e=1.70158;return--t*t*((e+1)*t+e)+1},backInOut:function(t){var e=2.5949095;return(t*=2)<1?t*t*((e+1)*t-e)*.5:.5*((t-=2)*t*((e+1)*t+e)+2)},bounceIn:function(t){return 1-Xe.bounceOut(1-t)},bounceOut:function(t){return t<1/2.75?7.5625*t*t:t<2/2.75?7.5625*(t-=1.5/2.75)*t+.75:t<2.5/2.75?7.5625*(t-=2.25/2.75)*t+.9375:7.5625*(t-=2.625/2.75)*t+.984375},bounceInOut:function(t){return t<.5?.5*Xe.bounceIn(2*t):.5*Xe.bounceOut(2*t-1)+.5}},Ze=Xe,Je=function(){function t(t){this._initialized=!1,this._startTime=0,this._pausedTime=0,this._paused=!1,this._life=t.life||1e3,this._delay=t.delay||0,this.loop=null!=t.loop&&t.loop,this.gap=t.gap||0,this.easing=t.easing||"linear",this.onframe=t.onframe,this.ondestroy=t.ondestroy,this.onrestart=t.onrestart}return t.prototype.step=function(t,e){if(this._initialized||(this._startTime=t+this._delay,this._initialized=!0),!this._paused){var n=(t-this._startTime-this._pausedTime)/this._life;n<0&&(n=0),n=Math.min(n,1);var i=this.easing,r="string"===typeof i?Ze[i]:i,o="function"===typeof r?r(n):n;if(this.onframe&&this.onframe(o),1===n){if(!this.loop)return!0;this._restart(t),this.onrestart&&this.onrestart()}return!1}this._pausedTime+=e},t.prototype._restart=function(t){var e=(t-this._startTime-this._pausedTime)%this._life;this._startTime=t-e+this.gap,this._pausedTime=0},t.prototype.pause=function(){this._paused=!0},t.prototype.resume=function(){this._paused=!1},t}(),Ke=Je,Qe=function(){function t(t){this.value=t}return t}(),tn=function(){function t(){this._len=0}return t.prototype.insert=function(t){var e=new Qe(t);return this.insertEntry(e),e},t.prototype.insertEntry=function(t){this.head?(this.tail.next=t,t.prev=this.tail,t.next=null,this.tail=t):this.head=this.tail=t,this._len++},t.prototype.remove=function(t){var e=t.prev,n=t.next;e?e.next=n:this.head=n,n?n.prev=e:this.tail=e,t.next=t.prev=null,this._len--},t.prototype.len=function(){return this._len},t.prototype.clear=function(){this.head=this.tail=null,this._len=0},t}(),en=function(){function t(t){this._list=new tn,this._maxSize=10,this._map={},this._maxSize=t}return t.prototype.put=function(t,e){var n=this._list,i=this._map,r=null;if(null==i[t]){var o=n.len(),a=this._lastRemovedEntry;if(o>=this._maxSize&&o>0){var s=n.head;n.remove(s),delete i[s.key],r=s.value,this._lastRemovedEntry=s}a?a.value=e:a=new Qe(e),a.key=t,n.insertEntry(a),i[t]=a}return r},t.prototype.get=function(t){var e=this._map[t],n=this._list;if(null!=e)return e!==n.tail&&(n.remove(e),n.insertEntry(e)),e.value},t.prototype.clear=function(){this._list.clear(),this._map={}},t.prototype.len=function(){return this._list.len()},t}(),nn=en,rn={transparent:[0,0,0,0],aliceblue:[240,248,255,1],antiquewhite:[250,235,215,1],aqua:[0,255,255,1],aquamarine:[127,255,212,1],azure:[240,255,255,1],beige:[245,245,220,1],bisque:[255,228,196,1],black:[0,0,0,1],blanchedalmond:[255,235,205,1],blue:[0,0,255,1],blueviolet:[138,43,226,1],brown:[165,42,42,1],burlywood:[222,184,135,1],cadetblue:[95,158,160,1],chartreuse:[127,255,0,1],chocolate:[210,105,30,1],coral:[255,127,80,1],cornflowerblue:[100,149,237,1],cornsilk:[255,248,220,1],crimson:[220,20,60,1],cyan:[0,255,255,1],darkblue:[0,0,139,1],darkcyan:[0,139,139,1],darkgoldenrod:[184,134,11,1],darkgray:[169,169,169,1],darkgreen:[0,100,0,1],darkgrey:[169,169,169,1],darkkhaki:[189,183,107,1],darkmagenta:[139,0,139,1],darkolivegreen:[85,107,47,1],darkorange:[255,140,0,1],darkorchid:[153,50,204,1],darkred:[139,0,0,1],darksalmon:[233,150,122,1],darkseagreen:[143,188,143,1],darkslateblue:[72,61,139,1],darkslategray:[47,79,79,1],darkslategrey:[47,79,79,1],darkturquoise:[0,206,209,1],darkviolet:[148,0,211,1],deeppink:[255,20,147,1],deepskyblue:[0,191,255,1],dimgray:[105,105,105,1],dimgrey:[105,105,105,1],dodgerblue:[30,144,255,1],firebrick:[178,34,34,1],floralwhite:[255,250,240,1],forestgreen:[34,139,34,1],fuchsia:[255,0,255,1],gainsboro:[220,220,220,1],ghostwhite:[248,248,255,1],gold:[255,215,0,1],goldenrod:[218,165,32,1],gray:[128,128,128,1],green:[0,128,0,1],greenyellow:[173,255,47,1],grey:[128,128,128,1],honeydew:[240,255,240,1],hotpink:[255,105,180,1],indianred:[205,92,92,1],indigo:[75,0,130,1],ivory:[255,255,240,1],khaki:[240,230,140,1],lavender:[230,230,250,1],lavenderblush:[255,240,245,1],lawngreen:[124,252,0,1],lemonchiffon:[255,250,205,1],lightblue:[173,216,230,1],lightcoral:[240,128,128,1],lightcyan:[224,255,255,1],lightgoldenrodyellow:[250,250,210,1],lightgray:[211,211,211,1],lightgreen:[144,238,144,1],lightgrey:[211,211,211,1],lightpink:[255,182,193,1],lightsalmon:[255,160,122,1],lightseagreen:[32,178,170,1],lightskyblue:[135,206,250,1],lightslategray:[119,136,153,1],lightslategrey:[119,136,153,1],lightsteelblue:[176,196,222,1],lightyellow:[255,255,224,1],lime:[0,255,0,1],limegreen:[50,205,50,1],linen:[250,240,230,1],magenta:[255,0,255,1],maroon:[128,0,0,1],mediumaquamarine:[102,205,170,1],mediumblue:[0,0,205,1],mediumorchid:[186,85,211,1],mediumpurple:[147,112,219,1],mediumseagreen:[60,179,113,1],mediumslateblue:[123,104,238,1],mediumspringgreen:[0,250,154,1],mediumturquoise:[72,209,204,1],mediumvioletred:[199,21,133,1],midnightblue:[25,25,112,1],mintcream:[245,255,250,1],mistyrose:[255,228,225,1],moccasin:[255,228,181,1],navajowhite:[255,222,173,1],navy:[0,0,128,1],oldlace:[253,245,230,1],olive:[128,128,0,1],olivedrab:[107,142,35,1],orange:[255,165,0,1],orangered:[255,69,0,1],orchid:[218,112,214,1],palegoldenrod:[238,232,170,1],palegreen:[152,251,152,1],paleturquoise:[175,238,238,1],palevioletred:[219,112,147,1],papayawhip:[255,239,213,1],peachpuff:[255,218,185,1],peru:[205,133,63,1],pink:[255,192,203,1],plum:[221,160,221,1],powderblue:[176,224,230,1],purple:[128,0,128,1],red:[255,0,0,1],rosybrown:[188,143,143,1],royalblue:[65,105,225,1],saddlebrown:[139,69,19,1],salmon:[250,128,114,1],sandybrown:[244,164,96,1],seagreen:[46,139,87,1],seashell:[255,245,238,1],sienna:[160,82,45,1],silver:[192,192,192,1],skyblue:[135,206,235,1],slateblue:[106,90,205,1],slategray:[112,128,144,1],slategrey:[112,128,144,1],snow:[255,250,250,1],springgreen:[0,255,127,1],steelblue:[70,130,180,1],tan:[210,180,140,1],teal:[0,128,128,1],thistle:[216,191,216,1],tomato:[255,99,71,1],turquoise:[64,224,208,1],violet:[238,130,238,1],wheat:[245,222,179,1],white:[255,255,255,1],whitesmoke:[245,245,245,1],yellow:[255,255,0,1],yellowgreen:[154,205,50,1]};function on(t){return t=Math.round(t),t<0?0:t>255?255:t}function an(t){return t=Math.round(t),t<0?0:t>360?360:t}function sn(t){return t<0?0:t>1?1:t}function ln(t){var e=t;return e.length&&"%"===e.charAt(e.length-1)?on(parseFloat(e)/100*255):on(parseInt(e,10))}function un(t){var e=t;return e.length&&"%"===e.charAt(e.length-1)?sn(parseFloat(e)/100):sn(parseFloat(e))}function cn(t,e,n){return n<0?n+=1:n>1&&(n-=1),6*n<1?t+(e-t)*n*6:2*n<1?e:3*n<2?t+(e-t)*(2/3-n)*6:t}function dn(t,e,n){return t+(e-t)*n}function hn(t,e,n,i,r){return t[0]=e,t[1]=n,t[2]=i,t[3]=r,t}function fn(t,e){return t[0]=e[0],t[1]=e[1],t[2]=e[2],t[3]=e[3],t}var pn=new nn(20),mn=null;function gn(t,e){mn&&fn(mn,e),mn=pn.put(t,mn||e.slice())}function yn(t,e){if(t){e=e||[];var n=pn.get(t);if(n)return fn(e,n);t+="";var i=t.replace(/ /g,"").toLowerCase();if(i in rn)return fn(e,rn[i]),gn(t,e),e;var r=i.length;if("#"!==i.charAt(0)){var o=i.indexOf("("),a=i.indexOf(")");if(-1!==o&&a+1===r){var s=i.substr(0,o),l=i.substr(o+1,a-(o+1)).split(","),u=1;switch(s){case"rgba":if(4!==l.length)return 3===l.length?hn(e,+l[0],+l[1],+l[2],1):hn(e,0,0,0,1);u=un(l.pop());case"rgb":return 3!==l.length?void hn(e,0,0,0,1):(hn(e,ln(l[0]),ln(l[1]),ln(l[2]),u),gn(t,e),e);case"hsla":return 4!==l.length?void hn(e,0,0,0,1):(l[3]=un(l[3]),vn(l,e),gn(t,e),e);case"hsl":return 3!==l.length?void hn(e,0,0,0,1):(vn(l,e),gn(t,e),e);default:return}}hn(e,0,0,0,1)}else{if(4===r||5===r){var c=parseInt(i.slice(1,4),16);return c>=0&&c<=4095?(hn(e,(3840&c)>>4|(3840&c)>>8,240&c|(240&c)>>4,15&c|(15&c)<<4,5===r?parseInt(i.slice(4),16)/15:1),gn(t,e),e):void hn(e,0,0,0,1)}if(7===r||9===r){c=parseInt(i.slice(1,7),16);return c>=0&&c<=16777215?(hn(e,(16711680&c)>>16,(65280&c)>>8,255&c,9===r?parseInt(i.slice(7),16)/255:1),gn(t,e),e):void hn(e,0,0,0,1)}}}}function vn(t,e){var n=(parseFloat(t[0])%360+360)%360/360,i=un(t[1]),r=un(t[2]),o=r<=.5?r*(i+1):r+i-r*i,a=2*r-o;return e=e||[],hn(e,on(255*cn(a,o,n+1/3)),on(255*cn(a,o,n)),on(255*cn(a,o,n-1/3)),1),4===t.length&&(e[3]=t[3]),e}function _n(t){if(t){var e,n,i=t[0]/255,r=t[1]/255,o=t[2]/255,a=Math.min(i,r,o),s=Math.max(i,r,o),l=s-a,u=(s+a)/2;if(0===l)e=0,n=0;else{n=u<.5?l/(s+a):l/(2-s-a);var c=((s-i)/6+l/2)/l,d=((s-r)/6+l/2)/l,h=((s-o)/6+l/2)/l;i===s?e=h-d:r===s?e=1/3+c-h:o===s&&(e=2/3+d-c),e<0&&(e+=1),e>1&&(e-=1)}var f=[360*e,n,u];return null!=t[3]&&f.push(t[3]),f}}function bn(t,e){var n=yn(t);if(n){for(var i=0;i<3;i++)n[i]=e<0?n[i]*(1-e)|0:(255-n[i])*e+n[i]|0,n[i]>255?n[i]=255:n[i]<0&&(n[i]=0);return Ln(n,4===n.length?"rgba":"rgb")}}function xn(t){var e=yn(t);if(e)return((1<<24)+(e[0]<<16)+(e[1]<<8)+ +e[2]).toString(16).slice(1)}function wn(t,e,n){if(e&&e.length&&t>=0&&t<=1){n=n||[];var i=t*(e.length-1),r=Math.floor(i),o=Math.ceil(i),a=e[r],s=e[o],l=i-r;return n[0]=on(dn(a[0],s[0],l)),n[1]=on(dn(a[1],s[1],l)),n[2]=on(dn(a[2],s[2],l)),n[3]=sn(dn(a[3],s[3],l)),n}}function Mn(t,e,n,i){var r=yn(t);if(t)return r=_n(r),null!=e&&(r[0]=an(e)),null!=n&&(r[1]=un(n)),null!=i&&(r[2]=un(i)),Ln(vn(r),"rgba")}function Sn(t,e){var n=yn(t);if(n&&null!=e)return n[3]=sn(e),Ln(n,"rgba")}function Ln(t,e){if(t&&t.length){var n=t[0]+","+t[1]+","+t[2];return"rgba"!==e&&"hsva"!==e&&"hsla"!==e||(n+=","+t[3]),e+"("+n+")"}}function kn(t,e){var n=yn(t);return n?(.299*n[0]+.587*n[1]+.114*n[2])*n[3]/255+(1-n[3])*e:0}var Tn=Array.prototype.slice;function Dn(t,e,n){return(e-t)*n+t}function Cn(t,e,n){return n>.5?e:t}function In(t,e,n,i){for(var r=e.length,o=0;o<r;o++)t[o]=Dn(e[o],n[o],i)}function An(t,e,n,i){for(var r=e.length,o=r&&e[0].length,a=0;a<r;a++){t[a]||(t[a]=[]);for(var s=0;s<o;s++)t[a][s]=Dn(e[a][s],n[a][s],i)}}function Yn(t,e,n,i){for(var r=e.length,o=0;o<r;o++)t[o]=e[o]+n[o]*i;return t}function On(t,e,n,i){for(var r=e.length,o=r&&e[0].length,a=0;a<r;a++){t[a]||(t[a]=[]);for(var s=0;s<o;s++)t[a][s]=e[a][s]+n[a][s]*i}return t}function Pn(t,e,n){var i=t,r=e;if(i.push&&r.push){var o=i.length,a=r.length;if(o!==a){var s=o>a;if(s)i.length=a;else for(var l=o;l<a;l++)i.push(1===n?r[l]:Tn.call(r[l]))}var u=i[0]&&i[0].length;for(l=0;l<i.length;l++)if(1===n)isNaN(i[l])&&(i[l]=r[l]);else for(var c=0;c<u;c++)isNaN(i[l][c])&&(i[l][c]=r[l][c])}}function En(t,e){var n=t.length;if(n!==e.length)return!1;for(var i=0;i<n;i++)if(t[i]!==e[i])return!1;return!0}function Nn(t,e,n,i,r,o,a){var s=.5*(n-t),l=.5*(i-e);return(2*(e-n)+s+l)*a+(-3*(e-n)-2*s-l)*o+s*r+e}function Rn(t,e,n,i,r,o,a,s){for(var l=e.length,u=0;u<l;u++)t[u]=Nn(e[u],n[u],i[u],r[u],o,a,s)}function Hn(t,e,n,i,r,o,a,s){for(var l=e.length,u=e[0].length,c=0;c<l;c++){t[c]||(t[1]=[]);for(var d=0;d<u;d++)t[c][d]=Nn(e[c][d],n[c][d],i[c][d],r[c][d],o,a,s)}}function jn(t){if(N(t)){var e=t.length;if(N(t[0])){for(var n=[],i=0;i<e;i++)n.push(Tn.call(t[i]));return n}return Tn.call(t)}return t}function zn(t){return t[0]=Math.floor(t[0]),t[1]=Math.floor(t[1]),t[2]=Math.floor(t[2]),"rgba("+t.join(",")+")"}function Wn(t){return N(t&&t[0])?2:1}var Bn=[0,0,0,0],Fn=function(){function t(t){this.keyframes=[],this.maxTime=0,this.arrDim=0,this.interpolable=!0,this._needsSort=!1,this._isAllValueEqual=!0,this._lastFrame=0,this._lastFramePercent=0,this.propName=t}return t.prototype.isFinished=function(){return this._finished},t.prototype.setFinished=function(){this._finished=!0,this._additiveTrack&&this._additiveTrack.setFinished()},t.prototype.needsAnimate=function(){return!this._isAllValueEqual&&this.keyframes.length>=2&&this.interpolable},t.prototype.getAdditiveTrack=function(){return this._additiveTrack},t.prototype.addKeyframe=function(t,e){t>=this.maxTime?this.maxTime=t:this._needsSort=!0;var n=this.keyframes,i=n.length;if(this.interpolable)if(N(e)){var r=Wn(e);if(i>0&&this.arrDim!==r)return void(this.interpolable=!1);if(1===r&&"number"!==typeof e[0]||2===r&&"number"!==typeof e[0][0])return void(this.interpolable=!1);if(i>0){var o=n[i-1];this._isAllValueEqual&&(1===r&&En(e,o.value)||(this._isAllValueEqual=!1))}this.arrDim=r}else{if(this.arrDim>0)return void(this.interpolable=!1);if("string"===typeof e){var a=yn(e);a?(e=a,this.isValueColor=!0):this.interpolable=!1}else if("number"!==typeof e||isNaN(e))return void(this.interpolable=!1);if(this._isAllValueEqual&&i>0){o=n[i-1];(this.isValueColor&&!En(o.value,e)||o.value!==e)&&(this._isAllValueEqual=!1)}}var s={time:t,value:e,percent:0};return this.keyframes.push(s),s},t.prototype.prepare=function(t){var e=this.keyframes;this._needsSort&&e.sort((function(t,e){return t.time-e.time}));for(var n=this.arrDim,i=e.length,r=e[i-1],o=0;o<i;o++)e[o].percent=e[o].time/this.maxTime,n>0&&o!==i-1&&Pn(e[o].value,r.value,n);if(t&&this.needsAnimate()&&t.needsAnimate()&&n===t.arrDim&&this.isValueColor===t.isValueColor&&!t._finished){this._additiveTrack=t;var a=e[0].value;for(o=0;o<i;o++)0===n?this.isValueColor?e[o].additiveValue=Yn([],e[o].value,a,-1):e[o].additiveValue=e[o].value-a:1===n?e[o].additiveValue=Yn([],e[o].value,a,-1):2===n&&(e[o].additiveValue=On([],e[o].value,a,-1))}},t.prototype.step=function(t,e){if(!this._finished){this._additiveTrack&&this._additiveTrack._finished&&(this._additiveTrack=null);var n,i=null!=this._additiveTrack,r=i?"additiveValue":"value",o=this.keyframes,a=this.keyframes.length,s=this.propName,l=this.arrDim,u=this.isValueColor;if(e<0)n=0;else if(e<this._lastFramePercent){var c=Math.min(this._lastFrame+1,a-1);for(n=c;n>=0;n--)if(o[n].percent<=e)break;n=Math.min(n,a-2)}else{for(n=this._lastFrame;n<a;n++)if(o[n].percent>e)break;n=Math.min(n-1,a-2)}var d=o[n+1],h=o[n];if(h&&d){this._lastFrame=n,this._lastFramePercent=e;var f=d.percent-h.percent;if(0!==f){var p=(e-h.percent)/f,m=i?this._additiveValue:u?Bn:t[s];if((l>0||u)&&!m&&(m=this._additiveValue=[]),this.useSpline){var g=o[n][r],y=o[0===n?n:n-1][r],v=o[n>a-2?a-1:n+1][r],_=o[n>a-3?a-1:n+2][r];if(l>0)1===l?Rn(m,y,g,v,_,p,p*p,p*p*p):Hn(m,y,g,v,_,p,p*p,p*p*p);else if(u)Rn(m,y,g,v,_,p,p*p,p*p*p),i||(t[s]=zn(m));else{var b=void 0;b=this.interpolable?Nn(y,g,v,_,p,p*p,p*p*p):v,i?this._additiveValue=b:t[s]=b}}else if(l>0)1===l?In(m,h[r],d[r],p):An(m,h[r],d[r],p);else if(u)In(m,h[r],d[r],p),i||(t[s]=zn(m));else{b=void 0;b=this.interpolable?Dn(h[r],d[r],p):Cn(h[r],d[r],p),i?this._additiveValue=b:t[s]=b}i&&this._addToTarget(t)}}}},t.prototype._addToTarget=function(t){var e=this.arrDim,n=this.propName,i=this._additiveValue;0===e?this.isValueColor?(yn(t[n],Bn),Yn(Bn,Bn,i,1),t[n]=zn(Bn)):t[n]=t[n]+i:1===e?Yn(t[n],t[n],i,1):2===e&&On(t[n],t[n],i,1)},t}(),Vn=function(){function t(t,e,n){this._tracks={},this._trackKeys=[],this._delay=0,this._maxTime=0,this._paused=!1,this._started=0,this._clip=null,this._target=t,this._loop=e,e&&n?k("Can' use additive animation on looped animation."):this._additiveAnimators=n}return t.prototype.getTarget=function(){return this._target},t.prototype.changeTarget=function(t){this._target=t},t.prototype.when=function(t,e){return this.whenWithKeys(t,e,B(e))},t.prototype.whenWithKeys=function(t,e,n){for(var i=this._tracks,r=0;r<n.length;r++){var o=n[r],a=i[o];if(!a){a=i[o]=new Fn(o);var s=void 0,l=this._getAdditiveTrack(o);if(l){var u=l.keyframes[l.keyframes.length-1];s=u&&u.value,l.isValueColor&&s&&(s=zn(s))}else s=this._target[o];if(null==s)continue;0!==t&&a.addKeyframe(0,jn(s)),this._trackKeys.push(o)}a.addKeyframe(t,jn(e[o]))}return this._maxTime=Math.max(this._maxTime,t),this},t.prototype.pause=function(){this._clip.pause(),this._paused=!0},t.prototype.resume=function(){this._clip.resume(),this._paused=!1},t.prototype.isPaused=function(){return!!this._paused},t.prototype._doneCallback=function(){this._setTracksFinished(),this._clip=null;var t=this._doneList;if(t)for(var e=t.length,n=0;n<e;n++)t[n].call(this)},t.prototype._abortedCallback=function(){this._setTracksFinished();var t=this.animation,e=this._abortedList;if(t&&t.removeClip(this._clip),this._clip=null,e)for(var n=0;n<e.length;n++)e[n].call(this)},t.prototype._setTracksFinished=function(){for(var t=this._tracks,e=this._trackKeys,n=0;n<e.length;n++)t[e[n]].setFinished()},t.prototype._getAdditiveTrack=function(t){var e,n=this._additiveAnimators;if(n)for(var i=0;i<n.length;i++){var r=n[i].getTrack(t);r&&(e=r)}return e},t.prototype.start=function(t,e){if(!(this._started>0)){this._started=1;for(var n=this,i=[],r=0;r<this._trackKeys.length;r++){var o=this._trackKeys[r],a=this._tracks[o],s=this._getAdditiveTrack(o),l=a.keyframes;if(a.prepare(s),a.needsAnimate())i.push(a);else if(!a.interpolable){var u=l[l.length-1];u&&(n._target[a.propName]=u.value)}}if(i.length||e){var c=new Ke({life:this._maxTime,loop:this._loop,delay:this._delay,onframe:function(t){n._started=2;var e=n._additiveAnimators;if(e){for(var r=!1,o=0;o<e.length;o++)if(e[o]._clip){r=!0;break}r||(n._additiveAnimators=null)}for(o=0;o<i.length;o++)i[o].step(n._target,t);var a=n._onframeList;if(a)for(o=0;o<a.length;o++)a[o](n._target,t)},ondestroy:function(){n._doneCallback()}});this._clip=c,this.animation&&this.animation.addClip(c),t&&"spline"!==t&&(c.easing=t)}else this._doneCallback();return this}},t.prototype.stop=function(t){if(this._clip){var e=this._clip;t&&e.onframe(1),this._abortedCallback()}},t.prototype.delay=function(t){return this._delay=t,this},t.prototype.during=function(t){return t&&(this._onframeList||(this._onframeList=[]),this._onframeList.push(t)),this},t.prototype.done=function(t){return t&&(this._doneList||(this._doneList=[]),this._doneList.push(t)),this},t.prototype.aborted=function(t){return t&&(this._abortedList||(this._abortedList=[]),this._abortedList.push(t)),this},t.prototype.getClip=function(){return this._clip},t.prototype.getTrack=function(t){return this._tracks[t]},t.prototype.stopTracks=function(t,e){if(!t.length||!this._clip)return!0;for(var n=this._tracks,i=this._trackKeys,r=0;r<t.length;r++){var o=n[t[r]];o&&(e?o.step(this._target,1):1===this._started&&o.step(this._target,0),o.setFinished())}var a=!0;for(r=0;r<i.length;r++)if(!n[i[r]].isFinished()){a=!1;break}return a&&this._abortedCallback(),a},t.prototype.saveFinalToTarget=function(t,e){if(t){e=e||this._trackKeys;for(var n=0;n<e.length;n++){var i=e[n],r=this._tracks[i];if(r&&!r.isFinished()){var o=r.keyframes,a=o[o.length-1];if(a){var s=jn(a.value);r.isValueColor&&(s=zn(s)),t[i]=s}}}}},t.prototype.__changeFinalValue=function(t,e){e=e||B(t);for(var n=0;n<e.length;n++){var i=e[n],r=this._tracks[i];if(r){var o=r.keyframes;if(o.length>1){var a=o.pop();r.addKeyframe(a.time,t[i]),r.prepare(r.getAdditiveTrack())}}}},t}(),Gn=Vn,$n=function(t){function e(e){var n=t.call(this)||this;return n._running=!1,n._time=0,n._pausedTime=0,n._pauseStart=0,n._paused=!1,e=e||{},n.stage=e.stage||{},n.onframe=e.onframe||function(){},n}return wt(e,t),e.prototype.addClip=function(t){t.animation&&this.removeClip(t),this._clipsHead?(this._clipsTail.next=t,t.prev=this._clipsTail,t.next=null,this._clipsTail=t):this._clipsHead=this._clipsTail=t,t.animation=this},e.prototype.addAnimator=function(t){t.animation=this;var e=t.getClip();e&&this.addClip(e)},e.prototype.removeClip=function(t){if(t.animation){var e=t.prev,n=t.next;e?e.next=n:this._clipsHead=n,n?n.prev=e:this._clipsTail=e,t.next=t.prev=t.animation=null}},e.prototype.removeAnimator=function(t){var e=t.getClip();e&&this.removeClip(e),t.animation=null},e.prototype.update=function(t){var e=(new Date).getTime()-this._pausedTime,n=e-this._time,i=this._clipsHead;while(i){var r=i.next,o=i.step(e,n);o?(i.ondestroy&&i.ondestroy(),this.removeClip(i),i=r):i=r}this._time=e,t||(this.onframe(n),this.trigger("frame",n),this.stage.update&&this.stage.update())},e.prototype._startLoop=function(){var t=this;function e(){t._running&&(qe(e),!t._paused&&t.update())}this._running=!0,qe(e)},e.prototype.start=function(){this._running||(this._time=(new Date).getTime(),this._pausedTime=0,this._startLoop())},e.prototype.stop=function(){this._running=!1},e.prototype.pause=function(){this._paused||(this._pauseStart=(new Date).getTime(),this._paused=!0)},e.prototype.resume=function(){this._paused&&(this._pausedTime+=(new Date).getTime()-this._pauseStart,this._paused=!1)},e.prototype.clear=function(){var t=this._clipsHead;while(t){var e=t.next;t.prev=t.next=t.animation=null,t=e}this._clipsHead=this._clipsTail=null},e.prototype.isFinished=function(){return null==this._clipsHead},e.prototype.animate=function(t,e){e=e||{},this.start();var n=new Gn(t,e.loop);return this.addAnimator(n),n},e}($t),Un=$n,qn=300,Xn=h.domSupported,Zn=function(){var t=["click","dblclick","mousewheel","wheel","mouseout","mouseup","mousedown","mousemove","contextmenu"],e=["touchstart","touchend","touchmove"],n={pointerdown:1,pointerup:1,pointermove:1,pointerout:1},i=H(t,(function(t){var e=t.replace("mouse","pointer");return n.hasOwnProperty(e)?e:t}));return{mouse:t,touch:e,pointer:i}}(),Jn={mouse:["mousemove","mouseup"],pointer:["pointermove","pointerup"]},Kn=!1;function Qn(t){var e=t.pointerType;return"pen"===e||"touch"===e}function ti(t){t.touching=!0,null!=t.touchTimer&&(clearTimeout(t.touchTimer),t.touchTimer=null),t.touchTimer=setTimeout((function(){t.touching=!1,t.touchTimer=null}),700)}function ei(t){t&&(t.zrByTouch=!0)}function ni(t,e){return ue(t.dom,new ri(t,e),!0)}function ii(t,e){var n=e,i=!1;while(n&&9!==n.nodeType&&!(i=n.domBelongToZr||n!==e&&n===t.painterRoot))n=n.parentNode;return i}var ri=function(){function t(t,e){this.stopPropagation=bt,this.stopImmediatePropagation=bt,this.preventDefault=bt,this.type=e.type,this.target=this.currentTarget=t.dom,this.pointerType=e.pointerType,this.clientX=e.clientX,this.clientY=e.clientY}return t}(),oi={mousedown:function(t){t=ue(this.dom,t),this.__mayPointerCapture=[t.zrX,t.zrY],this.trigger("mousedown",t)},mousemove:function(t){t=ue(this.dom,t);var e=this.__mayPointerCapture;!e||t.zrX===e[0]&&t.zrY===e[1]||this.__togglePointerCapture(!0),this.trigger("mousemove",t)},mouseup:function(t){t=ue(this.dom,t),this.__togglePointerCapture(!1),this.trigger("mouseup",t)},mouseout:function(t){t=ue(this.dom,t);var e=t.toElement||t.relatedTarget;ii(this,e)||(this.__pointerCapturing&&(t.zrEventControl="no_globalout"),this.trigger("mouseout",t))},wheel:function(t){Kn=!0,t=ue(this.dom,t),this.trigger("mousewheel",t)},mousewheel:function(t){Kn||(t=ue(this.dom,t),this.trigger("mousewheel",t))},touchstart:function(t){t=ue(this.dom,t),ei(t),this.__lastTouchMoment=new Date,this.handler.processGesture(t,"start"),oi.mousemove.call(this,t),oi.mousedown.call(this,t)},touchmove:function(t){t=ue(this.dom,t),ei(t),this.handler.processGesture(t,"change"),oi.mousemove.call(this,t)},touchend:function(t){t=ue(this.dom,t),ei(t),this.handler.processGesture(t,"end"),oi.mouseup.call(this,t),+new Date-+this.__lastTouchMoment<qn&&oi.click.call(this,t)},pointerdown:function(t){oi.mousedown.call(this,t)},pointermove:function(t){Qn(t)||oi.mousemove.call(this,t)},pointerup:function(t){oi.mouseup.call(this,t)},pointerout:function(t){Qn(t)||oi.mouseout.call(this,t)}};R(["click","dblclick","contextmenu"],(function(t){oi[t]=function(e){e=ue(this.dom,e),this.trigger(t,e)}}));var ai={pointermove:function(t){Qn(t)||ai.mousemove.call(this,t)},pointerup:function(t){ai.mouseup.call(this,t)},mousemove:function(t){this.trigger("mousemove",t)},mouseup:function(t){var e=this.__pointerCapturing;this.__togglePointerCapture(!1),this.trigger("mouseup",t),e&&(t.zrEventControl="only_globalout",this.trigger("mouseout",t))}};function si(t,e){var n=e.domHandlers;h.pointerEventsSupported?R(Zn.pointer,(function(i){ui(e,i,(function(e){n[i].call(t,e)}))})):(h.touchEventsSupported&&R(Zn.touch,(function(i){ui(e,i,(function(r){n[i].call(t,r),ti(e)}))})),R(Zn.mouse,(function(i){ui(e,i,(function(r){r=le(r),e.touching||n[i].call(t,r)}))})))}function li(t,e){function n(n){function i(i){i=le(i),ii(t,i.target)||(i=ni(t,i),e.domHandlers[n].call(t,i))}ui(e,n,i,{capture:!0})}h.pointerEventsSupported?R(Jn.pointer,n):h.touchEventsSupported||R(Jn.mouse,n)}function ui(t,e,n,i){t.mounted[e]=n,t.listenerOpts[e]=i,de(t.domTarget,e,n,i)}function ci(t){var e=t.mounted;for(var n in e)e.hasOwnProperty(n)&&he(t.domTarget,n,e[n],t.listenerOpts[n]);t.mounted={}}var di=function(){function t(t,e){this.mounted={},this.listenerOpts={},this.touching=!1,this.domTarget=t,this.domHandlers=e}return t}(),hi=function(t){function e(e,n){var i=t.call(this)||this;return i.__pointerCapturing=!1,i.dom=e,i.painterRoot=n,i._localHandlerScope=new di(e,oi),Xn&&(i._globalHandlerScope=new di(document,ai)),si(i,i._localHandlerScope),i}return wt(e,t),e.prototype.dispose=function(){ci(this._localHandlerScope),Xn&&ci(this._globalHandlerScope)},e.prototype.setCursor=function(t){this.dom.style&&(this.dom.style.cursor=t||"default")},e.prototype.__togglePointerCapture=function(t){if(this.__mayPointerCapture=null,Xn&&+this.__pointerCapturing^+t){this.__pointerCapturing=t;var e=this._globalHandlerScope;t?li(this,e):ci(e)}},e}($t),fi=hi,pi=1;"undefined"!==typeof window&&(pi=Math.max(window.devicePixelRatio||window.screen&&window.screen.deviceXDPI/window.screen.logicalXDPI||1,1));var mi=pi,gi=.4,yi="#333",vi="#ccc",_i="#eee";function bi(){return[1,0,0,1,0,0]}function xi(t){return t[0]=1,t[1]=0,t[2]=0,t[3]=1,t[4]=0,t[5]=0,t}function wi(t,e){return t[0]=e[0],t[1]=e[1],t[2]=e[2],t[3]=e[3],t[4]=e[4],t[5]=e[5],t}function Mi(t,e,n){var i=e[0]*n[0]+e[2]*n[1],r=e[1]*n[0]+e[3]*n[1],o=e[0]*n[2]+e[2]*n[3],a=e[1]*n[2]+e[3]*n[3],s=e[0]*n[4]+e[2]*n[5]+e[4],l=e[1]*n[4]+e[3]*n[5]+e[5];return t[0]=i,t[1]=r,t[2]=o,t[3]=a,t[4]=s,t[5]=l,t}function Si(t,e,n){return t[0]=e[0],t[1]=e[1],t[2]=e[2],t[3]=e[3],t[4]=e[4]+n[0],t[5]=e[5]+n[1],t}function Li(t,e,n){var i=e[0],r=e[2],o=e[4],a=e[1],s=e[3],l=e[5],u=Math.sin(n),c=Math.cos(n);return t[0]=i*c+a*u,t[1]=-i*u+a*c,t[2]=r*c+s*u,t[3]=-r*u+c*s,t[4]=c*o+u*l,t[5]=c*l-u*o,t}function ki(t,e,n){var i=n[0],r=n[1];return t[0]=e[0]*i,t[1]=e[1]*r,t[2]=e[2]*i,t[3]=e[3]*r,t[4]=e[4]*i,t[5]=e[5]*r,t}function Ti(t,e){var n=e[0],i=e[2],r=e[4],o=e[1],a=e[3],s=e[5],l=n*a-o*i;return l?(l=1/l,t[0]=a*l,t[1]=-o*l,t[2]=-i*l,t[3]=n*l,t[4]=(i*s-a*r)*l,t[5]=(o*r-n*s)*l,t):null}function Di(t){var e=bi();return wi(e,t),e}var Ci=xi,Ii=5e-5;function Ai(t){return t>Ii||t<-Ii}var Yi,Oi,Pi=[],Ei=[],Ni=bi(),Ri=Math.abs,Hi=function(){function t(){}return t.prototype.setPosition=function(t){this.x=t[0],this.y=t[1]},t.prototype.setScale=function(t){this.scaleX=t[0],this.scaleY=t[1]},t.prototype.setSkew=function(t){this.skewX=t[0],this.skewY=t[1]},t.prototype.setOrigin=function(t){this.originX=t[0],this.originY=t[1]},t.prototype.needLocalTransform=function(){return Ai(this.rotation)||Ai(this.x)||Ai(this.y)||Ai(this.scaleX-1)||Ai(this.scaleY-1)},t.prototype.updateTransform=function(){var t=this.parent,e=t&&t.transform,n=this.needLocalTransform(),i=this.transform;n||e?(i=i||bi(),n?this.getLocalTransform(i):Ci(i),e&&(n?Mi(i,t.transform,i):wi(i,t.transform)),this.transform=i,this._resolveGlobalScaleRatio(i)):i&&Ci(i)},t.prototype._resolveGlobalScaleRatio=function(t){var e=this.globalScaleRatio;if(null!=e&&1!==e){this.getGlobalScale(Pi);var n=Pi[0]<0?-1:1,i=Pi[1]<0?-1:1,r=((Pi[0]-n)*e+n)/Pi[0]||0,o=((Pi[1]-i)*e+i)/Pi[1]||0;t[0]*=r,t[1]*=r,t[2]*=o,t[3]*=o}this.invTransform=this.invTransform||bi(),Ti(this.invTransform,t)},t.prototype.getLocalTransform=function(e){return t.getLocalTransform(this,e)},t.prototype.getComputedTransform=function(){var t=this,e=[];while(t)e.push(t),t=t.parent;while(t=e.pop())t.updateTransform();return this.transform},t.prototype.setLocalTransform=function(t){if(t){var e=t[0]*t[0]+t[1]*t[1],n=t[2]*t[2]+t[3]*t[3],i=Math.atan2(t[1],t[0]),r=Math.PI/2+i-Math.atan2(t[3],t[2]);n=Math.sqrt(n)*Math.cos(r),e=Math.sqrt(e),this.skewX=r,this.skewY=0,this.rotation=-i,this.x=+t[4],this.y=+t[5],this.scaleX=e,this.scaleY=n,this.originX=0,this.originY=0}},t.prototype.decomposeTransform=function(){if(this.transform){var t=this.parent,e=this.transform;t&&t.transform&&(Mi(Ei,t.invTransform,e),e=Ei);var n=this.originX,i=this.originY;(n||i)&&(Ni[4]=n,Ni[5]=i,Mi(Ei,e,Ni),Ei[4]-=n,Ei[5]-=i,e=Ei),this.setLocalTransform(e)}},t.prototype.getGlobalScale=function(t){var e=this.transform;return t=t||[],e?(t[0]=Math.sqrt(e[0]*e[0]+e[1]*e[1]),t[1]=Math.sqrt(e[2]*e[2]+e[3]*e[3]),e[0]<0&&(t[0]=-t[0]),e[3]<0&&(t[1]=-t[1]),t):(t[0]=1,t[1]=1,t)},t.prototype.transformCoordToLocal=function(t,e){var n=[t,e],i=this.invTransform;return i&&jt(n,n,i),n},t.prototype.transformCoordToGlobal=function(t,e){var n=[t,e],i=this.transform;return i&&jt(n,n,i),n},t.prototype.getLineScale=function(){var t=this.transform;return t&&Ri(t[0]-1)>1e-10&&Ri(t[3]-1)>1e-10?Math.sqrt(Ri(t[0]*t[3]-t[2]*t[1])):1},t.getLocalTransform=function(t,e){e=e||[];var n=t.originX||0,i=t.originY||0,r=t.scaleX,o=t.scaleY,a=t.rotation||0,s=t.x,l=t.y,u=t.skewX?Math.tan(t.skewX):0,c=t.skewY?Math.tan(-t.skewY):0;return n||i?(e[4]=-n*r-u*i*o,e[5]=-i*o-c*n*r):e[4]=e[5]=0,e[0]=r,e[3]=o,e[1]=c*r,e[2]=u*o,a&&Li(e,e,a),e[4]+=n+s,e[5]+=i+l,e},t.initDefaultProps=function(){var e=t.prototype;e.x=0,e.y=0,e.scaleX=1,e.scaleY=1,e.originX=0,e.originY=0,e.skewX=0,e.skewY=0,e.rotation=0,e.globalScaleRatio=1}(),t}(),ji=Hi,zi=function(){function t(t,e){this.x=t||0,this.y=e||0}return t.prototype.copy=function(t){return this.x=t.x,this.y=t.y,this},t.prototype.clone=function(){return new t(this.x,this.y)},t.prototype.set=function(t,e){return this.x=t,this.y=e,this},t.prototype.equal=function(t){return t.x===this.x&&t.y===this.y},t.prototype.add=function(t){return this.x+=t.x,this.y+=t.y,this},t.prototype.scale=function(t){this.x*=t,this.y*=t},t.prototype.scaleAndAdd=function(t,e){this.x+=t.x*e,this.y+=t.y*e},t.prototype.sub=function(t){return this.x-=t.x,this.y-=t.y,this},t.prototype.dot=function(t){return this.x*t.x+this.y*t.y},t.prototype.len=function(){return Math.sqrt(this.x*this.x+this.y*this.y)},t.prototype.lenSquare=function(){return this.x*this.x+this.y*this.y},t.prototype.normalize=function(){var t=this.len();return this.x/=t,this.y/=t,this},t.prototype.distance=function(t){var e=this.x-t.x,n=this.y-t.y;return Math.sqrt(e*e+n*n)},t.prototype.distanceSquare=function(t){var e=this.x-t.x,n=this.y-t.y;return e*e+n*n},t.prototype.negate=function(){return this.x=-this.x,this.y=-this.y,this},t.prototype.transform=function(t){if(t){var e=this.x,n=this.y;return this.x=t[0]*e+t[2]*n+t[4],this.y=t[1]*e+t[3]*n+t[5],this}},t.prototype.toArray=function(t){return t[0]=this.x,t[1]=this.y,t},t.prototype.fromArray=function(t){this.x=t[0],this.y=t[1]},t.set=function(t,e,n){t.x=e,t.y=n},t.copy=function(t,e){t.x=e.x,t.y=e.y},t.len=function(t){return Math.sqrt(t.x*t.x+t.y*t.y)},t.lenSquare=function(t){return t.x*t.x+t.y*t.y},t.dot=function(t,e){return t.x*e.x+t.y*e.y},t.add=function(t,e,n){t.x=e.x+n.x,t.y=e.y+n.y},t.sub=function(t,e,n){t.x=e.x-n.x,t.y=e.y-n.y},t.scale=function(t,e,n){t.x=e.x*n,t.y=e.y*n},t.scaleAndAdd=function(t,e,n,i){t.x=e.x+n.x*i,t.y=e.y+n.y*i},t.lerp=function(t,e,n,i){var r=1-i;t.x=r*e.x+i*n.x,t.y=r*e.y+i*n.y},t}(),Wi=zi,Bi=Math.min,Fi=Math.max,Vi=new Wi,Gi=new Wi,$i=new Wi,Ui=new Wi,qi=new Wi,Xi=new Wi,Zi=function(){function t(t,e,n,i){n<0&&(t+=n,n=-n),i<0&&(e+=i,i=-i),this.x=t,this.y=e,this.width=n,this.height=i}return t.prototype.union=function(t){var e=Bi(t.x,this.x),n=Bi(t.y,this.y);isFinite(this.x)&&isFinite(this.width)?this.width=Fi(t.x+t.width,this.x+this.width)-e:this.width=t.width,isFinite(this.y)&&isFinite(this.height)?this.height=Fi(t.y+t.height,this.y+this.height)-n:this.height=t.height,this.x=e,this.y=n},t.prototype.applyTransform=function(e){t.applyTransform(this,this,e)},t.prototype.calculateTransform=function(t){var e=this,n=t.width/e.width,i=t.height/e.height,r=bi();return Si(r,r,[-e.x,-e.y]),ki(r,r,[n,i]),Si(r,r,[t.x,t.y]),r},t.prototype.intersect=function(e,n){if(!e)return!1;e instanceof t||(e=t.create(e));var i=this,r=i.x,o=i.x+i.width,a=i.y,s=i.y+i.height,l=e.x,u=e.x+e.width,c=e.y,d=e.y+e.height,h=!(o<l||u<r||s<c||d<a);if(n){var f=1/0,p=0,m=Math.abs(o-l),g=Math.abs(u-r),y=Math.abs(s-c),v=Math.abs(d-a),_=Math.min(m,g),b=Math.min(y,v);o<l||u<r?_>p&&(p=_,m<g?Wi.set(Xi,-m,0):Wi.set(Xi,g,0)):_<f&&(f=_,m<g?Wi.set(qi,m,0):Wi.set(qi,-g,0)),s<c||d<a?b>p&&(p=b,y<v?Wi.set(Xi,0,-y):Wi.set(Xi,0,v)):_<f&&(f=_,y<v?Wi.set(qi,0,y):Wi.set(qi,0,-v))}return n&&Wi.copy(n,h?qi:Xi),h},t.prototype.contain=function(t,e){var n=this;return t>=n.x&&t<=n.x+n.width&&e>=n.y&&e<=n.y+n.height},t.prototype.clone=function(){return new t(this.x,this.y,this.width,this.height)},t.prototype.copy=function(e){t.copy(this,e)},t.prototype.plain=function(){return{x:this.x,y:this.y,width:this.width,height:this.height}},t.prototype.isFinite=function(){return isFinite(this.x)&&isFinite(this.y)&&isFinite(this.width)&&isFinite(this.height)},t.prototype.isZero=function(){return 0===this.width||0===this.height},t.create=function(e){return new t(e.x,e.y,e.width,e.height)},t.copy=function(t,e){t.x=e.x,t.y=e.y,t.width=e.width,t.height=e.height},t.applyTransform=function(e,n,i){if(i){if(i[1]<1e-5&&i[1]>-1e-5&&i[2]<1e-5&&i[2]>-1e-5){var r=i[0],o=i[3],a=i[4],s=i[5];return e.x=n.x*r+a,e.y=n.y*o+s,e.width=n.width*r,e.height=n.height*o,e.width<0&&(e.x+=e.width,e.width=-e.width),void(e.height<0&&(e.y+=e.height,e.height=-e.height))}Vi.x=$i.x=n.x,Vi.y=Ui.y=n.y,Gi.x=Ui.x=n.x+n.width,Gi.y=$i.y=n.y+n.height,Vi.transform(i),Ui.transform(i),Gi.transform(i),$i.transform(i),e.x=Bi(Vi.x,Gi.x,$i.x,Ui.x),e.y=Bi(Vi.y,Gi.y,$i.y,Ui.y);var l=Fi(Vi.x,Gi.x,$i.x,Ui.x),u=Fi(Vi.y,Gi.y,$i.y,Ui.y);e.width=l-e.x,e.height=u-e.y}else e!==n&&t.copy(e,n)},t}(),Ji=Zi,Ki={},Qi="12px sans-serif";function tr(t,e){return Yi||(Yi=Y().getContext("2d")),Oi!==e&&(Oi=Yi.font=e||Qi),Yi.measureText(t)}var er={measureText:tr};function nr(t,e){e=e||Qi;var n=Ki[e];n||(n=Ki[e]=new nn(500));var i=n.get(t);return null==i&&(i=er.measureText(t,e).width,n.put(t,i)),i}function ir(t,e,n,i){var r=nr(t,e),o=sr(e),a=or(0,r,n),s=ar(0,o,i),l=new Ji(a,s,r,o);return l}function rr(t,e,n,i){var r=((t||"")+"").split("\n"),o=r.length;if(1===o)return ir(r[0],e,n,i);for(var a=new Ji(0,0,0,0),s=0;s<r.length;s++){var l=ir(r[s],e,n,i);0===s?a.copy(l):a.union(l)}return a}function or(t,e,n){return"right"===n?t-=e:"center"===n&&(t-=e/2),t}function ar(t,e,n){return"middle"===n?t-=e/2:"bottom"===n&&(t-=e),t}function sr(t){return nr("国",t)}function lr(t,e){return"string"===typeof t?t.lastIndexOf("%")>=0?parseFloat(t)/100*e:parseFloat(t):t}function ur(t,e,n){var i=e.position||"inside",r=null!=e.distance?e.distance:5,o=n.height,a=n.width,s=o/2,l=n.x,u=n.y,c="left",d="top";if(i instanceof Array)l+=lr(i[0],n.width),u+=lr(i[1],n.height),c=null,d=null;else switch(i){case"left":l-=r,u+=s,c="right",d="middle";break;case"right":l+=r+a,u+=s,d="middle";break;case"top":l+=a/2,u-=r,c="center",d="bottom";break;case"bottom":l+=a/2,u+=o+r,c="center";break;case"inside":l+=a/2,u+=s,c="center",d="middle";break;case"insideLeft":l+=r,u+=s,d="middle";break;case"insideRight":l+=a-r,u+=s,c="right",d="middle";break;case"insideTop":l+=a/2,u+=r,c="center";break;case"insideBottom":l+=a/2,u+=o-r,c="center",d="bottom";break;case"insideTopLeft":l+=r,u+=r;break;case"insideTopRight":l+=a-r,u+=r,c="right";break;case"insideBottomLeft":l+=r,u+=o-r,d="bottom";break;case"insideBottomRight":l+=a-r,u+=o-r,c="right",d="bottom";break}return t=t||{},t.x=l,t.y=u,t.align=c,t.verticalAlign=d,t}var cr="__zr_normal__",dr=["x","y","scaleX","scaleY","originX","originY","rotation","ignore"],hr={x:!0,y:!0,scaleX:!0,scaleY:!0,originX:!0,originY:!0,rotation:!0,ignore:!1},fr={},pr=new Ji(0,0,0,0),mr=function(){function t(t){this.id=L(),this.animators=[],this.currentStates=[],this.states={},this._init(t)}return t.prototype._init=function(t){this.attr(t)},t.prototype.drift=function(t,e,n){switch(this.draggable){case"horizontal":e=0;break;case"vertical":t=0;break}var i=this.transform;i||(i=this.transform=[1,0,0,1,0,0]),i[4]+=t,i[5]+=e,this.decomposeTransform(),this.markRedraw()},t.prototype.beforeUpdate=function(){},t.prototype.afterUpdate=function(){},t.prototype.update=function(){this.updateTransform(),this.__dirty&&this.updateInnerText()},t.prototype.updateInnerText=function(t){var e=this._textContent;if(e&&(!e.ignore||t)){this.textConfig||(this.textConfig={});var n=this.textConfig,i=n.local,r=e.attachedTransform,o=void 0,a=void 0,s=!1;r.parent=i?this:null;var l=!1;if(r.x=e.x,r.y=e.y,r.originX=e.originX,r.originY=e.originY,r.rotation=e.rotation,r.scaleX=e.scaleX,r.scaleY=e.scaleY,null!=n.position){var u=pr;n.layoutRect?u.copy(n.layoutRect):u.copy(this.getBoundingRect()),i||u.applyTransform(this.transform),this.calculateTextPosition?this.calculateTextPosition(fr,n,u):ur(fr,n,u),r.x=fr.x,r.y=fr.y,o=fr.align,a=fr.verticalAlign;var c=n.origin;if(c&&null!=n.rotation){var d=void 0,h=void 0;"center"===c?(d=.5*u.width,h=.5*u.height):(d=lr(c[0],u.width),h=lr(c[1],u.height)),l=!0,r.originX=-r.x+d+(i?0:u.x),r.originY=-r.y+h+(i?0:u.y)}}null!=n.rotation&&(r.rotation=n.rotation);var f=n.offset;f&&(r.x+=f[0],r.y+=f[1],l||(r.originX=-f[0],r.originY=-f[1]));var p=null==n.inside?"string"===typeof n.position&&n.position.indexOf("inside")>=0:n.inside,m=this._innerTextDefaultStyle||(this._innerTextDefaultStyle={}),g=void 0,y=void 0,v=void 0;p&&this.canBeInsideText()?(g=n.insideFill,y=n.insideStroke,null!=g&&"auto"!==g||(g=this.getInsideTextFill()),null!=y&&"auto"!==y||(y=this.getInsideTextStroke(g),v=!0)):(g=n.outsideFill,y=n.outsideStroke,null!=g&&"auto"!==g||(g=this.getOutsideFill()),null!=y&&"auto"!==y||(y=this.getOutsideStroke(g),v=!0)),g=g||"#000",g===m.fill&&y===m.stroke&&v===m.autoStroke&&o===m.align&&a===m.verticalAlign||(s=!0,m.fill=g,m.stroke=y,m.autoStroke=v,m.align=o,m.verticalAlign=a,e.setDefaultTextStyle(m)),e.__dirty|=je,s&&e.dirtyStyle(!0)}},t.prototype.canBeInsideText=function(){return!0},t.prototype.getInsideTextFill=function(){return"#fff"},t.prototype.getInsideTextStroke=function(t){return"#000"},t.prototype.getOutsideFill=function(){return this.__zr&&this.__zr.isDarkMode()?vi:yi},t.prototype.getOutsideStroke=function(t){var e=this.__zr&&this.__zr.getBackgroundColor(),n="string"===typeof e&&yn(e);n||(n=[255,255,255,1]);for(var i=n[3],r=this.__zr.isDarkMode(),o=0;o<3;o++)n[o]=n[o]*i+(r?0:255)*(1-i);return n[3]=1,Ln(n,"rgba")},t.prototype.traverse=function(t,e){},t.prototype.attrKV=function(t,e){"textConfig"===t?this.setTextConfig(e):"textContent"===t?this.setTextContent(e):"clipPath"===t?this.setClipPath(e):"extra"===t?(this.extra=this.extra||{},I(this.extra,e)):this[t]=e},t.prototype.hide=function(){this.ignore=!0,this.markRedraw()},t.prototype.show=function(){this.ignore=!1,this.markRedraw()},t.prototype.attr=function(t,e){if("string"===typeof t)this.attrKV(t,e);else if(J(t))for(var n=t,i=B(n),r=0;r<i.length;r++){var o=i[r];this.attrKV(o,t[o])}return this.markRedraw(),this},t.prototype.saveCurrentToNormalState=function(t){this._innerSaveToNormal(t);for(var e=this._normalState,n=0;n<this.animators.length;n++){var i=this.animators[n],r=i.__fromStateTransition;if(!r||r===cr){var o=i.targetName,a=o?e[o]:e;i.saveFinalToTarget(a)}}},t.prototype._innerSaveToNormal=function(t){var e=this._normalState;e||(e=this._normalState={}),t.textConfig&&!e.textConfig&&(e.textConfig=this.textConfig),this._savePrimaryToNormal(t,e,dr)},t.prototype._savePrimaryToNormal=function(t,e,n){for(var i=0;i<n.length;i++){var r=n[i];null==t[r]||r in e||(e[r]=this[r])}},t.prototype.hasState=function(){return this.currentStates.length>0},t.prototype.getState=function(t){return this.states[t]},t.prototype.ensureState=function(t){var e=this.states;return e[t]||(e[t]={}),e[t]},t.prototype.clearStates=function(t){this.useState(cr,!1,t)},t.prototype.useState=function(t,e,n,i){var r=t===cr,o=this.hasState();if(o||!r){var a=this.currentStates,s=this.stateTransition;if(!(O(a,t)>=0)||!e&&1!==a.length){var l;if(this.stateProxy&&!r&&(l=this.stateProxy(t)),l||(l=this.states&&this.states[t]),l||r){r||this.saveCurrentToNormalState(l);var u=!!(l&&l.hoverLayer||i);u&&this._toggleHoverLayerFlag(!0),this._applyStateObj(t,l,this._normalState,e,!n&&!this.__inHover&&s&&s.duration>0,s);var c=this._textContent,d=this._textGuide;return c&&c.useState(t,e,n,u),d&&d.useState(t,e,n,u),r?(this.currentStates=[],this._normalState={}):e?this.currentStates.push(t):this.currentStates=[t],this._updateAnimationTargets(),this.markRedraw(),!u&&this.__inHover&&(this._toggleHoverLayerFlag(!1),this.__dirty&=~je),l}k("State "+t+" not exists.")}}},t.prototype.useStates=function(t,e,n){if(t.length){var i=[],r=this.currentStates,o=t.length,a=o===r.length;if(a)for(var s=0;s<o;s++)if(t[s]!==r[s]){a=!1;break}if(a)return;for(s=0;s<o;s++){var l=t[s],u=void 0;this.stateProxy&&(u=this.stateProxy(l,t)),u||(u=this.states[l]),u&&i.push(u)}var c=i[o-1],d=!!(c&&c.hoverLayer||n);d&&this._toggleHoverLayerFlag(!0);var h=this._mergeStates(i),f=this.stateTransition;this.saveCurrentToNormalState(h),this._applyStateObj(t.join(","),h,this._normalState,!1,!e&&!this.__inHover&&f&&f.duration>0,f);var p=this._textContent,m=this._textGuide;p&&p.useStates(t,e,d),m&&m.useStates(t,e,d),this._updateAnimationTargets(),this.currentStates=t.slice(),this.markRedraw(),!d&&this.__inHover&&(this._toggleHoverLayerFlag(!1),this.__dirty&=~je)}else this.clearStates()},t.prototype._updateAnimationTargets=function(){for(var t=0;t<this.animators.length;t++){var e=this.animators[t];e.targetName&&e.changeTarget(this[e.targetName])}},t.prototype.removeState=function(t){var e=O(this.currentStates,t);if(e>=0){var n=this.currentStates.slice();n.splice(e,1),this.useStates(n)}},t.prototype.replaceState=function(t,e,n){var i=this.currentStates.slice(),r=O(i,t),o=O(i,e)>=0;r>=0?o?i.splice(r,1):i[r]=e:n&&!o&&i.push(e),this.useStates(i)},t.prototype.toggleState=function(t,e){e?this.useState(t,!0):this.removeState(t)},t.prototype._mergeStates=function(t){for(var e,n={},i=0;i<t.length;i++){var r=t[i];I(n,r),r.textConfig&&(e=e||{},I(e,r.textConfig))}return e&&(n.textConfig=e),n},t.prototype._applyStateObj=function(t,e,n,i,r,o){var a=!(e&&i);e&&e.textConfig?(this.textConfig=I({},i?this.textConfig:n.textConfig),I(this.textConfig,e.textConfig)):a&&n.textConfig&&(this.textConfig=n.textConfig);for(var s={},l=!1,u=0;u<dr.length;u++){var c=dr[u],d=r&&hr[c];e&&null!=e[c]?d?(l=!0,s[c]=e[c]):this[c]=e[c]:a&&null!=n[c]&&(d?(l=!0,s[c]=n[c]):this[c]=n[c])}if(!r)for(u=0;u<this.animators.length;u++){var h=this.animators[u],f=h.targetName;h.__changeFinalValue(f?(e||n)[f]:e||n)}l&&this._transitionState(t,s,o)},t.prototype._attachComponent=function(t){if(t.__zr&&!t.__hostTarget)throw new Error("Text element has been added to zrender.");if(t===this)throw new Error("Recursive component attachment.");var e=this.__zr;e&&t.addSelfToZr(e),t.__zr=e,t.__hostTarget=this},t.prototype._detachComponent=function(t){t.__zr&&t.removeSelfFromZr(t.__zr),t.__zr=null,t.__hostTarget=null},t.prototype.getClipPath=function(){return this._clipPath},t.prototype.setClipPath=function(t){this._clipPath&&this._clipPath!==t&&this.removeClipPath(),this._attachComponent(t),this._clipPath=t,this.markRedraw()},t.prototype.removeClipPath=function(){var t=this._clipPath;t&&(this._detachComponent(t),this._clipPath=null,this.markRedraw())},t.prototype.getTextContent=function(){return this._textContent},t.prototype.setTextContent=function(t){var e=this._textContent;if(e!==t){if(e&&e!==t&&this.removeTextContent(),t.__zr&&!t.__hostTarget)throw new Error("Text element has been added to zrender.");t.attachedTransform=new ji,this._attachComponent(t),this._textContent=t,this.markRedraw()}},t.prototype.setTextConfig=function(t){this.textConfig||(this.textConfig={}),I(this.textConfig,t),this.markRedraw()},t.prototype.removeTextConfig=function(){this.textConfig=null,this.markRedraw()},t.prototype.removeTextContent=function(){var t=this._textContent;t&&(t.attachedTransform=null,this._detachComponent(t),this._textContent=null,this._innerTextDefaultStyle=null,this.markRedraw())},t.prototype.getTextGuideLine=function(){return this._textGuide},t.prototype.setTextGuideLine=function(t){this._textGuide&&this._textGuide!==t&&this.removeTextGuideLine(),this._attachComponent(t),this._textGuide=t,this.markRedraw()},t.prototype.removeTextGuideLine=function(){var t=this._textGuide;t&&(this._detachComponent(t),this._textGuide=null,this.markRedraw())},t.prototype.markRedraw=function(){this.__dirty|=je;var t=this.__zr;t&&(this.__inHover?t.refreshHover():t.refresh()),this.__hostTarget&&this.__hostTarget.markRedraw()},t.prototype.dirty=function(){this.markRedraw()},t.prototype._toggleHoverLayerFlag=function(t){this.__inHover=t;var e=this._textContent,n=this._textGuide;e&&(e.__inHover=t),n&&(n.__inHover=t)},t.prototype.addSelfToZr=function(t){this.__zr=t;var e=this.animators;if(e)for(var n=0;n<e.length;n++)t.animation.addAnimator(e[n]);this._clipPath&&this._clipPath.addSelfToZr(t),this._textContent&&this._textContent.addSelfToZr(t),this._textGuide&&this._textGuide.addSelfToZr(t)},t.prototype.removeSelfFromZr=function(t){this.__zr=null;var e=this.animators;if(e)for(var n=0;n<e.length;n++)t.animation.removeAnimator(e[n]);this._clipPath&&this._clipPath.removeSelfFromZr(t),this._textContent&&this._textContent.removeSelfFromZr(t),this._textGuide&&this._textGuide.removeSelfFromZr(t)},t.prototype.animate=function(t,e){var n=t?this[t]:this;if(n){var i=new Gn(n,e);return this.addAnimator(i,t),i}k('Property "'+t+'" is not existed in element '+this.id)},t.prototype.addAnimator=function(t,e){var n=this.__zr,i=this;t.during((function(){i.updateDuringAnimation(e)})).done((function(){var e=i.animators,n=O(e,t);n>=0&&e.splice(n,1)})),this.animators.push(t),n&&n.animation.addAnimator(t),n&&n.wakeUp()},t.prototype.updateDuringAnimation=function(t){this.markRedraw()},t.prototype.stopAnimation=function(t,e){for(var n=this.animators,i=n.length,r=[],o=0;o<i;o++){var a=n[o];t&&t!==a.scope?r.push(a):a.stop(e)}return this.animators=r,this},t.prototype.animateTo=function(t,e,n){gr(this,t,e,n)},t.prototype.animateFrom=function(t,e,n){gr(this,t,e,n,!0)},t.prototype._transitionState=function(t,e,n,i){for(var r=gr(this,e,n,i),o=0;o<r.length;o++)r[o].__fromStateTransition=t},t.prototype.getBoundingRect=function(){return null},t.prototype.getPaintRect=function(){return null},t.initDefaultProps=function(){var e=t.prototype;e.type="element",e.name="",e.ignore=!1,e.silent=!1,e.isGroup=!1,e.draggable=!1,e.dragging=!1,e.ignoreClip=!1,e.__inHover=!1,e.__dirty=je;var n={};function i(t,e,i){n[t+e+i]||(n[t+e+i]=!0)}function r(t,n,r,o){function a(t,e){Object.defineProperty(e,0,{get:function(){return t[r]},set:function(e){t[r]=e}}),Object.defineProperty(e,1,{get:function(){return t[o]},set:function(e){t[o]=e}})}Object.defineProperty(e,t,{get:function(){if(i(t,r,o),!this[n]){var e=this[n]=[];a(this,e)}return this[n]},set:function(e){i(t,r,o),this[r]=e[0],this[o]=e[1],this[n]=e,a(this,e)}})}Object.defineProperty&&(!h.browser.ie||h.browser.version>8)&&(r("position","_legacyPos","x","y"),r("scale","_legacyScale","scaleX","scaleY"),r("origin","_legacyOrigin","originX","originY"))}(),t}();function gr(t,e,n,i,r){n=n||{};var o=[];br(t,"",t,e,n,i,o,r);var a=o.length,s=!1,l=n.done,u=n.aborted,c=function(){s=!0,a--,a<=0&&(s?l&&l():u&&u())},d=function(){a--,a<=0&&(s?l&&l():u&&u())};a||l&&l(),o.length>0&&n.during&&o[0].during((function(t,e){n.during(e)}));for(var h=0;h<o.length;h++){var f=o[h];c&&f.done(c),d&&f.aborted(d),f.start(n.easing,n.force)}return o}function yr(t,e,n){for(var i=0;i<n;i++)t[i]=e[i]}function vr(t){return N(t[0])}function _r(t,e,n){if(N(e[n]))if(N(t[n])||(t[n]=[]),Q(e[n])){var i=e[n].length;t[n].length!==i&&(t[n]=new e[n].constructor(i),yr(t[n],e[n],i))}else{var r=e[n],o=t[n],a=r.length;if(vr(r))for(var s=r[0].length,l=0;l<a;l++)o[l]?yr(o[l],r[l],s):o[l]=Array.prototype.slice.call(r[l]);else yr(o,r,a);o.length=r.length}else t[n]=e[n]}function br(t,e,n,i,r,o,a,s){for(var l=[],u=[],c=B(i),d=r.duration,h=r.delay,f=r.additive,p=r.setToFinal,m=!J(o),g=0;g<c.length;g++){var y=c[g];if(null!=n[y]&&null!=i[y]&&(m||o[y]))if(J(i[y])&&!N(i[y])){if(e){s||(n[y]=i[y],t.updateDuringAnimation(e));continue}br(t,y,n[y],i[y],r,o&&o[y],a,s)}else l.push(y),u.push(y);else s||(n[y]=i[y],t.updateDuringAnimation(e),u.push(y))}var v=l.length;if(v>0||r.force&&!a.length){for(var _=t.animators,b=[],x=0;x<_.length;x++)_[x].targetName===e&&b.push(_[x]);if(!f&&b.length)for(x=0;x<b.length;x++){var w=b[x].stopTracks(u);if(w){var M=O(_,b[x]);_.splice(M,1)}}var S=void 0,L=void 0,k=void 0;if(s){L={},p&&(S={});for(x=0;x<v;x++){y=l[x];L[y]=n[y],p?S[y]=i[y]:n[y]=i[y]}}else if(p){k={};for(x=0;x<v;x++){y=l[x];k[y]=jn(n[y]),_r(n,i,y)}}var T=new Gn(n,!1,f?b:null);T.targetName=e,r.scope&&(T.scope=r.scope),p&&S&&T.whenWithKeys(0,S,l),k&&T.whenWithKeys(0,k,l),T.whenWithKeys(null==d?500:d,s?L:i,l).delay(h||0),t.addAnimator(T,e),a.push(T)}}E(mr,$t),E(mr,ji);var xr=mr,wr=function(t){function e(e){var n=t.call(this)||this;return n.isGroup=!0,n._children=[],n.attr(e),n}return wt(e,t),e.prototype.childrenRef=function(){return this._children},e.prototype.children=function(){return this._children.slice()},e.prototype.childAt=function(t){return this._children[t]},e.prototype.childOfName=function(t){for(var e=this._children,n=0;n<e.length;n++)if(e[n].name===t)return e[n]},e.prototype.childCount=function(){return this._children.length},e.prototype.add=function(t){if(t&&(t!==this&&t.parent!==this&&(this._children.push(t),this._doAdd(t)),t.__hostTarget))throw"This elemenet has been used as an attachment";return this},e.prototype.addBefore=function(t,e){if(t&&t!==this&&t.parent!==this&&e&&e.parent===this){var n=this._children,i=n.indexOf(e);i>=0&&(n.splice(i,0,t),this._doAdd(t))}return this},e.prototype.replaceAt=function(t,e){var n=this._children,i=n[e];if(t&&t!==this&&t.parent!==this&&t!==i){n[e]=t,i.parent=null;var r=this.__zr;r&&i.removeSelfFromZr(r),this._doAdd(t)}return this},e.prototype._doAdd=function(t){t.parent&&t.parent.remove(t),t.parent=this;var e=this.__zr;e&&e!==t.__zr&&t.addSelfToZr(e),e&&e.refresh()},e.prototype.remove=function(t){var e=this.__zr,n=this._children,i=O(n,t);return i<0||(n.splice(i,1),t.parent=null,e&&t.removeSelfFromZr(e),e&&e.refresh()),this},e.prototype.removeAll=function(){for(var t=this._children,e=this.__zr,n=0;n<t.length;n++){var i=t[n];e&&i.removeSelfFromZr(e),i.parent=null}return t.length=0,this},e.prototype.eachChild=function(t,e){for(var n=this._children,i=0;i<n.length;i++){var r=n[i];t.call(e,r,i)}return this},e.prototype.traverse=function(t,e){for(var n=0;n<this._children.length;n++){var i=this._children[n],r=t.call(e,i);i.isGroup&&!r&&i.traverse(t,e)}return this},e.prototype.addSelfToZr=function(e){t.prototype.addSelfToZr.call(this,e);for(var n=0;n<this._children.length;n++){var i=this._children[n];i.addSelfToZr(e)}},e.prototype.removeSelfFromZr=function(e){t.prototype.removeSelfFromZr.call(this,e);for(var n=0;n<this._children.length;n++){var i=this._children[n];i.removeSelfFromZr(e)}},e.prototype.getBoundingRect=function(t){for(var e=new Ji(0,0,0,0),n=t||this._children,i=[],r=null,o=0;o<n.length;o++){var a=n[o];if(!a.ignore&&!a.invisible){var s=a.getBoundingRect(),l=a.getLocalTransform(i);l?(Ji.applyTransform(e,s,l),r=r||e.clone(),r.union(e)):(r=r||s.clone(),r.union(s))}}return r||e},e}(xr);wr.prototype.type="group";var Mr=wr,Sr=!h.canvasSupported,Lr={},kr={};
/*!
* ZRender, a high performance 2d drawing library.
*
* Copyright (c) 2013, Baidu Inc.
* All rights reserved.
*
* LICENSE
* https://github.com/ecomfe/zrender/blob/master/LICENSE.txt
*/function Tr(t){delete kr[t]}function Dr(t){if(!t)return!1;if("string"===typeof t)return kn(t,1)<gi;if(t.colorStops){for(var e=t.colorStops,n=0,i=e.length,r=0;r<i;r++)n+=kn(e[r].color,1);return n/=i,n<gi}return!1}var Cr=function(){function t(t,e,n){var i=this;this._sleepAfterStill=10,this._stillFrameAccum=0,this._needsRefresh=!0,this._needsRefreshHover=!0,this._darkMode=!1,n=n||{},this.dom=e,this.id=t;var r=new Ue,o=n.renderer||"canvas";if(Sr)throw new Error("IE8 support has been dropped since 5.0");if(Lr[o]||(o=B(Lr)[0]),!Lr[o])throw new Error("Renderer '"+o+"' is not imported. Please import it first.");n.useDirtyRect=null!=n.useDirtyRect&&n.useDirtyRect;var a=new Lr[o](e,r,n,t);this.storage=r,this.painter=a;var s=h.node||h.worker?null:new fi(a.getViewportRoot(),a.root);this.handler=new De(r,a,s,a.root),this.animation=new Un({stage:{update:function(){return i._flush(!0)}}}),this.animation.start()}return t.prototype.add=function(t){t&&(this.storage.addRoot(t),t.addSelfToZr(this),this.refresh())},t.prototype.remove=function(t){t&&(this.storage.delRoot(t),t.removeSelfFromZr(this),this.refresh())},t.prototype.configLayer=function(t,e){this.painter.configLayer&&this.painter.configLayer(t,e),this.refresh()},t.prototype.setBackgroundColor=function(t){this.painter.setBackgroundColor&&this.painter.setBackgroundColor(t),this.refresh(),this._backgroundColor=t,this._darkMode=Dr(t)},t.prototype.getBackgroundColor=function(){return this._backgroundColor},t.prototype.setDarkMode=function(t){this._darkMode=t},t.prototype.isDarkMode=function(){return this._darkMode},t.prototype.refreshImmediately=function(t){t||this.animation.update(!0),this._needsRefresh=!1,this.painter.refresh(),this._needsRefresh=!1},t.prototype.refresh=function(){this._needsRefresh=!0,this.animation.start()},t.prototype.flush=function(){this._flush(!1)},t.prototype._flush=function(t){var e,n=(new Date).getTime();this._needsRefresh&&(e=!0,this.refreshImmediately(t)),this._needsRefreshHover&&(e=!0,this.refreshHoverImmediately());var i=(new Date).getTime();e?(this._stillFrameAccum=0,this.trigger("rendered",{elapsedTime:i-n})):this._sleepAfterStill>0&&(this._stillFrameAccum++,this._stillFrameAccum>this._sleepAfterStill&&this.animation.stop())},t.prototype.setSleepAfterStill=function(t){this._sleepAfterStill=t},t.prototype.wakeUp=function(){this.animation.start(),this._stillFrameAccum=0},t.prototype.addHover=function(t){},t.prototype.removeHover=function(t){},t.prototype.clearHover=function(){},t.prototype.refreshHover=function(){this._needsRefreshHover=!0},t.prototype.refreshHoverImmediately=function(){this._needsRefreshHover=!1,this.painter.refreshHover&&"canvas"===this.painter.getType()&&this.painter.refreshHover()},t.prototype.resize=function(t){t=t||{},this.painter.resize(t.width,t.height),this.handler.resize()},t.prototype.clearAnimation=function(){this.animation.clear()},t.prototype.getWidth=function(){return this.painter.getWidth()},t.prototype.getHeight=function(){return this.painter.getHeight()},t.prototype.pathToImage=function(t,e){if(this.painter.pathToImage)return this.painter.pathToImage(t,e)},t.prototype.setCursorStyle=function(t){this.handler.setCursorStyle(t)},t.prototype.findHover=function(t,e){return this.handler.findHover(t,e)},t.prototype.on=function(t,e,n){return this.handler.on(t,e,n),this},t.prototype.off=function(t,e){this.handler.off(t,e)},t.prototype.trigger=function(t,e){this.handler.trigger(t,e)},t.prototype.clear=function(){for(var t=this.storage.getRoots(),e=0;e<t.length;e++)t[e]instanceof Mr&&t[e].removeSelfFromZr(this);this.storage.delAllRoots(),this.painter.clear()},t.prototype.dispose=function(){this.animation.stop(),this.clear(),this.storage.dispose(),this.painter.dispose(),this.handler.dispose(),this.animation=this.storage=this.painter=this.handler=null,Tr(this.id)},t}();function Ir(t,e){var n=new Cr(L(),t,e);return kr[n.id]=n,n}function Ar(t,e){Lr[t]=e}var Yr=1e-4,Or=20;function Pr(t){return t.replace(/^\s+|\s+$/g,"")}function Er(t,e,n,i){var r=e[0],o=e[1],a=n[0],s=n[1],l=o-r,u=s-a;if(0===l)return 0===u?a:(a+s)/2;if(i)if(l>0){if(t<=r)return a;if(t>=o)return s}else{if(t>=r)return a;if(t<=o)return s}else{if(t===r)return a;if(t===o)return s}return(t-r)/l*u+a}function Nr(t,e){switch(t){case"center":case"middle":t="50%";break;case"left":case"top":t="0%";break;case"right":case"bottom":t="100%";break}return"string"===typeof t?Pr(t).match(/%$/)?parseFloat(t)/100*e:parseFloat(t):null==t?NaN:+t}function Rr(t,e,n){return null==e&&(e=10),e=Math.min(Math.max(0,e),Or),t=(+t).toFixed(e),n?t:+t}function Hr(t){return t.sort((function(t,e){return t-e})),t}function jr(t){if(t=+t,isNaN(t))return 0;if(t>1e-14)for(var e=1,n=0;n<15;n++,e*=10)if(Math.round(t*e)/e===t)return n;return zr(t)}function zr(t){var e=t.toString().toLowerCase(),n=e.indexOf("e"),i=n>0?+e.slice(n+1):0,r=n>0?n:e.length,o=e.indexOf("."),a=o<0?0:r-1-o;return Math.max(0,a-i)}function Wr(t,e){var n=Math.log,i=Math.LN10,r=Math.floor(n(t[1]-t[0])/i),o=Math.round(n(Math.abs(e[1]-e[0]))/i),a=Math.min(Math.max(-r+o,0),20);return isFinite(a)?a:20}function Br(t,e,n){if(!t[e])return 0;var i=j(t,(function(t,e){return t+(isNaN(e)?0:e)}),0);if(0===i)return 0;var r=Math.pow(10,n),o=H(t,(function(t){return(isNaN(t)?0:t)/i*r*100})),a=100*r,s=H(o,(function(t){return Math.floor(t)})),l=j(s,(function(t,e){return t+e}),0),u=H(o,(function(t,e){return t-s[e]}));while(l<a){for(var c=Number.NEGATIVE_INFINITY,d=null,h=0,f=u.length;h<f;++h)u[h]>c&&(c=u[h],d=h);++s[d],u[d]=0,++l}return s[e]/r}function Fr(t,e){var n=Math.max(jr(t),jr(e)),i=t+e;return n>Or?i:Rr(i,n)}var Vr=9007199254740991;function Gr(t){var e=2*Math.PI;return(t%e+e)%e}function $r(t){return t>-Yr&&t<Yr}var Ur=/^(?:(\d{4})(?:[-\/](\d{1,2})(?:[-\/](\d{1,2})(?:[T ](\d{1,2})(?::(\d{1,2})(?::(\d{1,2})(?:[.,](\d+))?)?)?(Z|[\+\-]\d\d:?\d\d)?)?)?)?)?$/;function qr(t){if(t instanceof Date)return t;if("string"===typeof t){var e=Ur.exec(t);if(!e)return new Date(NaN);if(e[8]){var n=+e[4]||0;return"Z"!==e[8].toUpperCase()&&(n-=+e[8].slice(0,3)),new Date(Date.UTC(+e[1],+(e[2]||1)-1,+e[3]||1,n,+(e[5]||0),+e[6]||0,+e[7]||0))}return new Date(+e[1],+(e[2]||1)-1,+e[3]||1,+e[4]||0,+(e[5]||0),+e[6]||0,+e[7]||0)}return null==t?new Date(NaN):new Date(Math.round(t))}function Xr(t){return Math.pow(10,Zr(t))}function Zr(t){if(0===t)return 0;var e=Math.floor(Math.log(t)/Math.LN10);return t/Math.pow(10,e)>=10&&e++,e}function Jr(t,e){var n,i=Zr(t),r=Math.pow(10,i),o=t/r;return n=e?o<1.5?1:o<2.5?2:o<4?3:o<7?5:10:o<1?1:o<2?2:o<3?3:o<5?5:10,t=n*r,i>=-20?+t.toFixed(i<0?-i:0):t}function Kr(t,e){var n=(t.length-1)*e+1,i=Math.floor(n),r=+t[i-1],o=n-i;return o?r+o*(t[i]-r):r}function Qr(t){t.sort((function(t,e){return s(t,e,0)?-1:1}));for(var e=-1/0,n=1,i=0;i<t.length;){for(var r=t[i].interval,o=t[i].close,a=0;a<2;a++)r[a]<=e&&(r[a]=e,o[a]=a?1:1-n),e=r[a],n=o[a];r[0]===r[1]&&o[0]*o[1]!==1?t.splice(i,1):i++}return t;function s(t,e,n){return t.interval[n]<e.interval[n]||t.interval[n]===e.interval[n]&&(t.close[n]-e.close[n]===(n?-1:1)||!n&&s(t,e,1))}}function to(t){var e=parseFloat(t);return e==t&&(0!==e||"string"!==typeof t||t.indexOf("x")<=0)?e:NaN}function eo(t){return!isNaN(to(t))}function no(){return Math.round(9*Math.random())}function io(t,e){return 0===e?t:io(e,t%e)}function ro(t,e){return null==t?e:null==e?t:t*e/io(t,e)}var oo="series\0",ao="\0_ec_\0";function so(t){return t instanceof Array?t:null==t?[]:[t]}function lo(t,e,n){if(t){t[e]=t[e]||{},t.emphasis=t.emphasis||{},t.emphasis[e]=t.emphasis[e]||{};for(var i=0,r=n.length;i<r;i++){var o=n[i];!t.emphasis[e].hasOwnProperty(o)&&t[e].hasOwnProperty(o)&&(t.emphasis[e][o]=t[e][o])}}}var uo=["fontStyle","fontWeight","fontSize","fontFamily","rich","tag","color","textBorderColor","textBorderWidth","width","height","lineHeight","align","verticalAlign","baseline","shadowColor","shadowBlur","shadowOffsetX","shadowOffsetY","textShadowColor","textShadowBlur","textShadowOffsetX","textShadowOffsetY","backgroundColor","borderColor","borderWidth","borderRadius","padding"];function co(t){return!J(t)||$(t)||t instanceof Date?t:t.value}function ho(t){return J(t)&&!(t instanceof Array)}function fo(t,e,n){var i="normalMerge"===n,r="replaceMerge"===n,o="replaceAll"===n;t=t||[],e=(e||[]).slice();var a=gt();R(e,(function(t,n){J(t)||(e[n]=null)}));var s=po(t,a,n);return(i||r)&&mo(s,t,a,e),i&&go(s,e),i||r?yo(s,e,r):o&&vo(s,e),_o(s),s}function po(t,e,n){var i=[];if("replaceAll"===n)return i;for(var r=0;r<t.length;r++){var o=t[r];o&&null!=o.id&&e.set(o.id,r),i.push({existing:"replaceMerge"===n||So(o)?null:o,newOption:null,keyInfo:null,brandNew:null})}return i}function mo(t,e,n,i){R(i,(function(r,o){if(r&&null!=r.id){var a=xo(r.id),s=n.get(a);if(null!=s){var l=t[s];ct(!l.newOption,'Duplicated option on id "'+a+'".'),l.newOption=r,l.existing=e[s],i[o]=null}}}))}function go(t,e){R(e,(function(n,i){if(n&&null!=n.name)for(var r=0;r<t.length;r++){var o=t[r].existing;if(!t[r].newOption&&o&&(null==o.id||null==n.id)&&!So(n)&&!So(o)&&bo("name",o,n))return t[r].newOption=n,void(e[i]=null)}}))}function yo(t,e,n){R(e,(function(e){if(e){var i,r=0;while((i=t[r])&&(i.newOption||So(i.existing)||i.existing&&null!=e.id&&!bo("id",e,i.existing)))r++;i?(i.newOption=e,i.brandNew=n):t.push({newOption:e,brandNew:n,existing:null,keyInfo:null}),r++}}))}function vo(t,e){R(e,(function(e){t.push({newOption:e,brandNew:!0,existing:null,keyInfo:null})}))}function _o(t){var e=gt();R(t,(function(t){var n=t.existing;n&&e.set(n.id,t)})),R(t,(function(t){var n=t.newOption;ct(!n||null==n.id||!e.get(n.id)||e.get(n.id)===t,"id duplicates: "+(n&&n.id)),n&&null!=n.id&&e.set(n.id,t),!t.keyInfo&&(t.keyInfo={})})),R(t,(function(t,n){var i=t.existing,r=t.newOption,o=t.keyInfo;if(J(r)){if(o.name=null!=r.name?xo(r.name):i?i.name:oo+n,i)o.id=xo(i.id);else if(null!=r.id)o.id=xo(r.id);else{var a=0;do{o.id="\0"+o.name+"\0"+a++}while(e.get(o.id))}e.set(o.id,t)}}))}function bo(t,e,n){var i=wo(e[t],null),r=wo(n[t],null);return null!=i&&null!=r&&i===r}function xo(t){return wo(t,"")}function wo(t,e){if(null==t)return e;var n=typeof t;return"string"===n?t:"number"===n||X(t)?t+"":e}function Mo(t){var e=t.name;return!(!e||!e.indexOf(oo))}function So(t){return t&&null!=t.id&&0===xo(t.id).indexOf(ao)}function Lo(t){return ao+t}function ko(t,e,n){R(t,(function(t){var i=t.newOption;J(i)&&(t.keyInfo.mainType=e,t.keyInfo.subType=To(e,i,t.existing,n))}))}function To(t,e,n,i){var r=e.type?e.type:n?n.subType:i.determineSubType(t,e);return r}function Do(t,e){var n={},i={};return r(t||[],n),r(e||[],i,n),[o(n),o(i)];function r(t,e,n){for(var i=0,r=t.length;i<r;i++){var o=wo(t[i].seriesId,null);if(null==o)return;for(var a=so(t[i].dataIndex),s=n&&n[o],l=0,u=a.length;l<u;l++){var c=a[l];s&&s[c]?s[c]=null:(e[o]||(e[o]={}))[c]=1}}}function o(t,e){var n=[];for(var i in t)if(t.hasOwnProperty(i)&&null!=t[i])if(e)n.push(+i);else{var r=o(t[i],!0);r.length&&n.push({seriesId:i,dataIndex:r})}return n}}function Co(t,e){return null!=e.dataIndexInside?e.dataIndexInside:null!=e.dataIndex?$(e.dataIndex)?H(e.dataIndex,(function(e){return t.indexOfRawIndex(e)})):t.indexOfRawIndex(e.dataIndex):null!=e.name?$(e.name)?H(e.name,(function(e){return t.indexOfName(e)})):t.indexOfName(e.name):void 0}function Io(){var t="__ec_inner_"+Ao++;return function(e){return e[t]||(e[t]={})}}var Ao=no();function Yo(t,e,n){var i=Oo(e,n),r=i.mainTypeSpecified,o=i.queryOptionMap,a=i.others,s=a,l=n?n.defaultMainType:null;return!r&&l&&o.set(l,{}),o.each((function(e,i){var r=No(t,i,e,{useDefault:l===i,enableAll:!n||null==n.enableAll||n.enableAll,enableNone:!n||null==n.enableNone||n.enableNone});s[i+"Models"]=r.models,s[i+"Model"]=r.models[0]})),s}function Oo(t,e){var n;if(q(t)){var i={};i[t+"Index"]=0,n=i}else n=t;var r=gt(),o={},a=!1;return R(n,(function(t,n){if("dataIndex"!==n&&"dataIndexInside"!==n){var i=n.match(/^(\w+)(Index|Id|Name)$/)||[],s=i[1],l=(i[2]||"").toLowerCase();if(s&&l&&!(e&&e.includeMainTypes&&O(e.includeMainTypes,s)<0)){a=a||!!s;var u=r.get(s)||r.set(s,{});u[l]=t}}else o[n]=t})),{mainTypeSpecified:a,queryOptionMap:r,others:o}}var Po={useDefault:!0,enableAll:!1,enableNone:!1},Eo={useDefault:!1,enableAll:!0,enableNone:!0};function No(t,e,n,i){i=i||Po;var r=n.index,o=n.id,a=n.name,s={models:null,specified:null!=r||null!=o||null!=a};if(!s.specified){var l=void 0;return s.models=i.useDefault&&(l=t.getComponent(e))?[l]:[],s}return"none"===r||!1===r?(ct(i.enableNone,'`"none"` or `false` is not a valid value on index option.'),s.models=[],s):("all"===r&&(ct(i.enableAll,'`"all"` is not a valid value on index option.'),r=o=a=null),s.models=t.queryComponents({mainType:e,index:r,id:o,name:a}),s)}function Ro(t,e,n){t.setAttribute?t.setAttribute(e,n):t[e]=n}function Ho(t,e){return t.getAttribute?t.getAttribute(e):t[e]}function jo(t){return"auto"===t?h.domSupported?"html":"richText":t||"html"}function zo(t,e){var n=gt(),i=[];return R(t,(function(t){var r=e(t);(n.get(r)||(i.push(r),n.set(r,[]))).push(t)})),{keys:i,buckets:n}}function Wo(t,e,n,i,r){var o=null==e||"auto"===e;if(null==i)return i;if("number"===typeof i){var a=Dn(n||0,i,r);return Rr(a,o?Math.max(jr(n||0),jr(i)):e)}if("string"===typeof i)return r<1?n:i;for(var s=[],l=n,u=i,c=Math.max(l?l.length:0,u.length),d=0;d<c;++d){var h=t.getDimensionInfo(d);if("ordinal"===h.type)s[d]=(r<1&&l?l:u)[d];else{var f=l&&l[d]?l[d]:0,p=u[d];a=Dn(f,p,r);s[d]=Rr(a,o?Math.max(jr(f),jr(p)):e)}}return s}var Bo=".",Fo="___EC__COMPONENT__CONTAINER___",Vo="___EC__EXTENDED_CLASS___";function Go(t){var e={main:"",sub:""};if(t){var n=t.split(Bo);e.main=n[0]||"",e.sub=n[1]||""}return e}function $o(t){ct(/^[a-zA-Z0-9_]+([.][a-zA-Z0-9_]+)?$/.test(t),'componentType "'+t+'" illegal')}function Uo(t){return!(!t||!t[Vo])}function qo(t,e){t.$constructor=t,t.extend=function(t){var e=this;function n(){for(var i=[],r=0;r<arguments.length;r++)i[r]=arguments[r];if(t.$constructor)t.$constructor.apply(this,arguments);else{if(Xo(e)){var o=vt(n.prototype,new(e.bind.apply(e,s([void 0],i))));return o}e.apply(this,arguments)}}return n[Vo]=!0,I(n.prototype,t),n.extend=this.extend,n.superCall=Qo,n.superApply=ta,P(n,this),n.superClass=e,n}}function Xo(t){return"function"===typeof t&&/^class\s/.test(Function.prototype.toString.call(t))}function Zo(t,e){t.extend=e.extend}var Jo=Math.round(10*Math.random());function Ko(t){var e=["__\0is_clz",Jo++].join("_");t.prototype[e]=!0,t.isInstance=function(t){return!(!t||!t[e])}}function Qo(t,e){for(var n=[],i=2;i<arguments.length;i++)n[i-2]=arguments[i];return this.superClass.prototype[e].apply(t,n)}function ta(t,e,n){return this.superClass.prototype[e].apply(t,n)}function ea(t){var e={};function n(t){var n=e[t.main];return n&&n[Fo]||(n=e[t.main]={},n[Fo]=!0),n}t.registerClass=function(t){var i=t.type||t.prototype.type;if(i){$o(i),t.prototype.type=i;var r=Go(i);if(r.sub){if(r.sub!==Fo){var o=n(r);o[r.sub]=t}}else e[r.main]=t}return t},t.getClass=function(t,n,i){var r=e[t];if(r&&r[Fo]&&(r=n?r[n]:null),i&&!r)throw new Error(n?"Component "+t+"."+(n||"")+" is used but not imported.":t+".type should be specified.");return r},t.getClassesByMainType=function(t){var n=Go(t),i=[],r=e[n.main];return r&&r[Fo]?R(r,(function(t,e){e!==Fo&&i.push(t)})):i.push(r),i},t.hasClass=function(t){var n=Go(t);return!!e[n.main]},t.getAllClassMainTypes=function(){var t=[];return R(e,(function(e,n){t.push(n)})),t},t.hasSubTypes=function(t){var n=Go(t),i=e[n.main];return i&&i[Fo]}}function na(t,e){for(var n=0;n<t.length;n++)t[n][1]||(t[n][1]=t[n][0]);return e=e||!1,function(n,i,r){for(var o={},a=0;a<t.length;a++){var s=t[a][1];if(!(i&&O(i,s)>=0||r&&O(r,s)<0)){var l=n.getShallow(s,e);null!=l&&(o[t[a][0]]=l)}}return o}}var ia=[["fill","color"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["opacity"],["shadowColor"]],ra=na(ia),oa=function(){function t(){}return t.prototype.getAreaStyle=function(t,e){return ra(this,t,e)},t}(),aa=new nn(50);function sa(t){if("string"===typeof t){var e=aa.get(t);return e&&e.image}return t}function la(t,e,n,i,r){if(t){if("string"===typeof t){if(e&&e.__zrImageSrc===t||!n)return e;var o=aa.get(t),a={hostEl:n,cb:i,cbPayload:r};return o?(e=o.image,!ca(e)&&o.pending.push(a)):(e=new Image,e.onload=e.onerror=ua,aa.put(t,e.__cachedImgObj={image:e,pending:[a]}),e.src=e.__zrImageSrc=t),e}return t}return e}function ua(){var t=this.__cachedImgObj;this.onload=this.onerror=this.__cachedImgObj=null;for(var e=0;e<t.pending.length;e++){var n=t.pending[e],i=n.cb;i&&i(this,n.cbPayload),n.hostEl.dirty()}t.pending.length=0}function ca(t){return t&&t.width&&t.height}var da=/\{([a-zA-Z0-9_]+)\|([^}]*)\}/g;function ha(t,e,n,i,r){if(!e)return"";var o=(t+"").split("\n");r=fa(e,n,i,r);for(var a=0,s=o.length;a<s;a++)o[a]=pa(o[a],r);return o.join("\n")}function fa(t,e,n,i){i=i||{};var r=I({},i);r.font=e,n=at(n,"..."),r.maxIterations=at(i.maxIterations,2);var o=r.minChar=at(i.minChar,0);r.cnCharWidth=nr("国",e);var a=r.ascCharWidth=nr("a",e);r.placeholder=at(i.placeholder,"");for(var s=t=Math.max(0,t-1),l=0;l<o&&s>=a;l++)s-=a;var u=nr(n,e);return u>s&&(n="",u=0),s=t-u,r.ellipsis=n,r.ellipsisWidth=u,r.contentWidth=s,r.containerWidth=t,r}function pa(t,e){var n=e.containerWidth,i=e.font,r=e.contentWidth;if(!n)return"";var o=nr(t,i);if(o<=n)return t;for(var a=0;;a++){if(o<=r||a>=e.maxIterations){t+=e.ellipsis;break}var s=0===a?ma(t,r,e.ascCharWidth,e.cnCharWidth):o>0?Math.floor(t.length*r/o):0;t=t.substr(0,s),o=nr(t,i)}return""===t&&(t=e.placeholder),t}function ma(t,e,n,i){for(var r=0,o=0,a=t.length;o<a&&r<e;o++){var s=t.charCodeAt(o);r+=0<=s&&s<=127?n:i}return o}function ga(t,e){null!=t&&(t+="");var n,i=e.overflow,r=e.padding,o=e.font,a="truncate"===i,s=sr(o),l=at(e.lineHeight,s),u="truncate"===e.lineOverflow,c=e.width;n=null!=c&&"break"===i||"breakAll"===i?t?La(t,e.font,c,"breakAll"===i,0).lines:[]:t?t.split("\n"):[];var d=n.length*l,h=at(e.height,d);if(d>h&&u){var f=Math.floor(h/l);n=n.slice(0,f)}var p=h,m=c;if(r&&(p+=r[0]+r[2],null!=m&&(m+=r[1]+r[3])),t&&a&&null!=m)for(var g=fa(c,o,e.ellipsis,{minChar:e.truncateMinChar,placeholder:e.placeholder}),y=0;y<n.length;y++)n[y]=pa(n[y],g);if(null==c){var v=0;for(y=0;y<n.length;y++)v=Math.max(nr(n[y],o),v);c=v}return{lines:n,height:h,outerHeight:p,lineHeight:l,calculatedLineHeight:s,contentHeight:d,width:c}}var ya=function(){function t(){}return t}(),va=function(){function t(t){this.tokens=[],t&&(this.tokens=t)}return t}(),_a=function(){function t(){this.width=0,this.height=0,this.contentWidth=0,this.contentHeight=0,this.outerWidth=0,this.outerHeight=0,this.lines=[]}return t}();function ba(t,e){var n=new _a;if(null!=t&&(t+=""),!t)return n;var i,r=e.width,o=e.height,a=e.overflow,s="break"!==a&&"breakAll"!==a||null==r?null:{width:r,accumWidth:0,breakAll:"breakAll"===a},l=da.lastIndex=0;while(null!=(i=da.exec(t))){var u=i.index;u>l&&xa(n,t.substring(l,u),e,s),xa(n,i[2],e,s,i[1]),l=da.lastIndex}l<t.length&&xa(n,t.substring(l,t.length),e,s);var c=[],d=0,h=0,f=e.padding,p="truncate"===a,m="truncate"===e.lineOverflow;function g(t,e,n){t.width=e,t.lineHeight=n,d+=n,h=Math.max(h,e)}t:for(var y=0;y<n.lines.length;y++){for(var v=n.lines[y],_=0,b=0,x=0;x<v.tokens.length;x++){var w=v.tokens[x],M=w.styleName&&e.rich[w.styleName]||{},S=w.textPadding=M.padding,L=S?S[1]+S[3]:0,k=w.font=M.font||e.font;w.contentHeight=sr(k);var T=at(M.height,w.contentHeight);if(w.innerHeight=T,S&&(T+=S[0]+S[2]),w.height=T,w.lineHeight=st(M.lineHeight,e.lineHeight,T),w.align=M&&M.align||e.align,w.verticalAlign=M&&M.verticalAlign||"middle",m&&null!=o&&d+w.lineHeight>o){x>0?(v.tokens=v.tokens.slice(0,x),g(v,b,_),n.lines=n.lines.slice(0,y+1)):n.lines=n.lines.slice(0,y);break t}var D=M.width,C=null==D||"auto"===D;if("string"===typeof D&&"%"===D.charAt(D.length-1))w.percentWidth=D,c.push(w),w.contentWidth=nr(w.text,k);else{if(C){var I=M.backgroundColor,A=I&&I.image;A&&(A=sa(A),ca(A)&&(w.width=Math.max(w.width,A.width*T/A.height)))}var Y=p&&null!=r?r-b:null;null!=Y&&Y<w.width?!C||Y<L?(w.text="",w.width=w.contentWidth=0):(w.text=ha(w.text,Y-L,k,e.ellipsis,{minChar:e.truncateMinChar}),w.width=w.contentWidth=nr(w.text,k)):w.contentWidth=nr(w.text,k)}w.width+=L,b+=w.width,M&&(_=Math.max(_,w.lineHeight))}g(v,b,_)}n.outerWidth=n.width=at(r,h),n.outerHeight=n.height=at(o,d),n.contentHeight=d,n.contentWidth=h,f&&(n.outerWidth+=f[1]+f[3],n.outerHeight+=f[0]+f[2]);for(y=0;y<c.length;y++){w=c[y];var O=w.percentWidth;w.width=parseInt(O,10)/100*n.width}return n}function xa(t,e,n,i,r){var o,a,s=""===e,l=r&&n.rich[r]||{},u=t.lines,c=l.font||n.font,d=!1;if(i){var h=l.padding,f=h?h[1]+h[3]:0;if(null!=l.width&&"auto"!==l.width){var p=lr(l.width,i.width)+f;u.length>0&&p+i.accumWidth>i.width&&(o=e.split("\n"),d=!0),i.accumWidth=p}else{var m=La(e,c,i.width,i.breakAll,i.accumWidth);i.accumWidth=m.accumWidth+f,a=m.linesWidths,o=m.lines}}else o=e.split("\n");for(var g=0;g<o.length;g++){var y=o[g],v=new ya;if(v.styleName=r,v.text=y,v.isLineHolder=!y&&!s,"number"===typeof l.width?v.width=l.width:v.width=a?a[g]:nr(y,c),g||d)u.push(new va([v]));else{var _=(u[u.length-1]||(u[0]=new va)).tokens,b=_.length;1===b&&_[0].isLineHolder?_[0]=v:(y||!b||s)&&_.push(v)}}}function wa(t){var e=t.charCodeAt(0);return e>=33&&e<=255}var Ma=j(",&?/;] ".split(""),(function(t,e){return t[e]=!0,t}),{});function Sa(t){return!wa(t)||!!Ma[t]}function La(t,e,n,i,r){for(var o=[],a=[],s="",l="",u=0,c=0,d=0;d<t.length;d++){var h=t.charAt(d);if("\n"!==h){var f=nr(h,e),p=!i&&!Sa(h);(o.length?c+f>n:r+c+f>n)?c?(s||l)&&(p?(s||(s=l,l="",u=0,c=u),o.push(s),a.push(c-u),l+=h,u+=f,s="",c=u):(l&&(s+=l,c+=u,l="",u=0),o.push(s),a.push(c),s=h,c=f)):p?(o.push(l),a.push(u),l=h,u=f):(o.push(h),a.push(f)):(c+=f,p?(l+=h,u+=f):(l&&(s+=l,l="",u=0),s+=h))}else l&&(s+=l,c+=u),o.push(s),a.push(c),s="",l="",u=0,c=0}return o.length||s||(s=t,l="",u=0),l&&(s+=l),s&&(o.push(s),a.push(c)),1===o.length&&(c+=r),{accumWidth:c,lines:o,linesWidths:a}}var ka="__zr_style_"+Math.round(10*Math.random()),Ta={shadowBlur:0,shadowOffsetX:0,shadowOffsetY:0,shadowColor:"#000",opacity:1,blend:"source-over"},Da={style:{shadowBlur:!0,shadowOffsetX:!0,shadowOffsetY:!0,shadowColor:!0,opacity:!0}};Ta[ka]=!0;var Ca=["z","z2","invisible"],Ia=["invisible"],Aa=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype._init=function(e){for(var n=B(e),i=0;i<n.length;i++){var r=n[i];"style"===r?this.useStyle(e[r]):t.prototype.attrKV.call(this,r,e[r])}this.style||this.useStyle({})},e.prototype.beforeBrush=function(){},e.prototype.afterBrush=function(){},e.prototype.innerBeforeBrush=function(){},e.prototype.innerAfterBrush=function(){},e.prototype.shouldBePainted=function(t,e,n,i){var r=this.transform;if(this.ignore||this.invisible||0===this.style.opacity||this.culling&&Pa(this,t,e)||r&&!r[0]&&!r[3])return!1;if(n&&this.__clipPaths)for(var o=0;o<this.__clipPaths.length;++o)if(this.__clipPaths[o].isZeroArea())return!1;if(i&&this.parent){var a=this.parent;while(a){if(a.ignore)return!1;a=a.parent}}return!0},e.prototype.contain=function(t,e){return this.rectContain(t,e)},e.prototype.traverse=function(t,e){t.call(e,this)},e.prototype.rectContain=function(t,e){var n=this.transformCoordToLocal(t,e),i=this.getBoundingRect();return i.contain(n[0],n[1])},e.prototype.getPaintRect=function(){var t=this._paintRect;if(!this._paintRect||this.__dirty){var e=this.transform,n=this.getBoundingRect(),i=this.style,r=i.shadowBlur||0,o=i.shadowOffsetX||0,a=i.shadowOffsetY||0;t=this._paintRect||(this._paintRect=new Ji(0,0,0,0)),e?Ji.applyTransform(t,n,e):t.copy(n),(r||o||a)&&(t.width+=2*r+Math.abs(o),t.height+=2*r+Math.abs(a),t.x=Math.min(t.x,t.x+o-r),t.y=Math.min(t.y,t.y+a-r));var s=this.dirtyRectTolerance;t.isZero()||(t.x=Math.floor(t.x-s),t.y=Math.floor(t.y-s),t.width=Math.ceil(t.width+1+2*s),t.height=Math.ceil(t.height+1+2*s))}return t},e.prototype.setPrevPaintRect=function(t){t?(this._prevPaintRect=this._prevPaintRect||new Ji(0,0,0,0),this._prevPaintRect.copy(t)):this._prevPaintRect=null},e.prototype.getPrevPaintRect=function(){return this._prevPaintRect},e.prototype.animateStyle=function(t){return this.animate("style",t)},e.prototype.updateDuringAnimation=function(t){"style"===t?this.dirtyStyle():this.markRedraw()},e.prototype.attrKV=function(e,n){"style"!==e?t.prototype.attrKV.call(this,e,n):this.style?this.setStyle(n):this.useStyle(n)},e.prototype.setStyle=function(t,e){return"string"===typeof t?this.style[t]=e:I(this.style,t),this.dirtyStyle(),this},e.prototype.dirtyStyle=function(t){t||this.markRedraw(),this.__dirty|=ze,this._rect&&(this._rect=null)},e.prototype.dirty=function(){this.dirtyStyle()},e.prototype.styleChanged=function(){return!!(this.__dirty&ze)},e.prototype.styleUpdated=function(){this.__dirty&=~ze},e.prototype.createStyle=function(t){return vt(Ta,t)},e.prototype.useStyle=function(t){t[ka]||(t=this.createStyle(t)),this.__inHover?this.__hoverStyle=t:this.style=t,this.dirtyStyle()},e.prototype.isStyleObject=function(t){return t[ka]},e.prototype._innerSaveToNormal=function(e){t.prototype._innerSaveToNormal.call(this,e);var n=this._normalState;e.style&&!n.style&&(n.style=this._mergeStyle(this.createStyle(),this.style)),this._savePrimaryToNormal(e,n,Ca)},e.prototype._applyStateObj=function(e,n,i,r,o,a){t.prototype._applyStateObj.call(this,e,n,i,r,o,a);var s,l=!(n&&r);if(n&&n.style?o?r?s=n.style:(s=this._mergeStyle(this.createStyle(),i.style),this._mergeStyle(s,n.style)):(s=this._mergeStyle(this.createStyle(),r?this.style:i.style),this._mergeStyle(s,n.style)):l&&(s=i.style),s)if(o){var u=this.style;if(this.style=this.createStyle(l?{}:u),l)for(var c=B(u),d=0;d<c.length;d++){var h=c[d];h in s&&(s[h]=s[h],this.style[h]=u[h])}var f=B(s);for(d=0;d<f.length;d++){h=f[d];this.style[h]=this.style[h]}this._transitionState(e,{style:s},a,this.getAnimationStyleProps())}else this.useStyle(s);var p=this.__inHover?Ia:Ca;for(d=0;d<p.length;d++){h=p[d];n&&null!=n[h]?this[h]=n[h]:l&&null!=i[h]&&(this[h]=i[h])}},e.prototype._mergeStates=function(e){for(var n,i=t.prototype._mergeStates.call(this,e),r=0;r<e.length;r++){var o=e[r];o.style&&(n=n||{},this._mergeStyle(n,o.style))}return n&&(i.style=n),i},e.prototype._mergeStyle=function(t,e){return I(t,e),t},e.prototype.getAnimationStyleProps=function(){return Da},e.initDefaultProps=function(){var t=e.prototype;t.type="displayable",t.invisible=!1,t.z=0,t.z2=0,t.zlevel=0,t.culling=!1,t.cursor="pointer",t.rectHover=!1,t.incremental=!1,t._rect=null,t.dirtyRectTolerance=0,t.__dirty=je|ze}(),e}(xr),Ya=new Ji(0,0,0,0),Oa=new Ji(0,0,0,0);function Pa(t,e,n){return Ya.copy(t.getBoundingRect()),t.transform&&Ya.applyTransform(t.transform),Oa.width=e,Oa.height=n,!Ya.intersect(Oa)}var Ea=Aa,Na=Math.pow,Ra=Math.sqrt,Ha=1e-8,ja=1e-4,za=Ra(3),Wa=1/3,Ba=Mt(),Fa=Mt(),Va=Mt();function Ga(t){return t>-Ha&&t<Ha}function $a(t){return t>Ha||t<-Ha}function Ua(t,e,n,i,r){var o=1-r;return o*o*(o*t+3*r*e)+r*r*(r*i+3*o*n)}function qa(t,e,n,i,r){var o=1-r;return 3*(((e-t)*o+2*(n-e)*r)*o+(i-n)*r*r)}function Xa(t,e,n,i,r,o){var a=i+3*(e-n)-t,s=3*(n-2*e+t),l=3*(e-t),u=t-r,c=s*s-3*a*l,d=s*l-9*a*u,h=l*l-3*s*u,f=0;if(Ga(c)&&Ga(d))if(Ga(s))o[0]=0;else{var p=-l/s;p>=0&&p<=1&&(o[f++]=p)}else{var m=d*d-4*c*h;if(Ga(m)){var g=d/c,y=(p=-s/a+g,-g/2);p>=0&&p<=1&&(o[f++]=p),y>=0&&y<=1&&(o[f++]=y)}else if(m>0){var v=Ra(m),_=c*s+1.5*a*(-d+v),b=c*s+1.5*a*(-d-v);_=_<0?-Na(-_,Wa):Na(_,Wa),b=b<0?-Na(-b,Wa):Na(b,Wa);p=(-s-(_+b))/(3*a);p>=0&&p<=1&&(o[f++]=p)}else{var x=(2*c*s-3*a*d)/(2*Ra(c*c*c)),w=Math.acos(x)/3,M=Ra(c),S=Math.cos(w),L=(p=(-s-2*M*S)/(3*a),y=(-s+M*(S+za*Math.sin(w)))/(3*a),(-s+M*(S-za*Math.sin(w)))/(3*a));p>=0&&p<=1&&(o[f++]=p),y>=0&&y<=1&&(o[f++]=y),L>=0&&L<=1&&(o[f++]=L)}}return f}function Za(t,e,n,i,r){var o=6*n-12*e+6*t,a=9*e+3*i-3*t-9*n,s=3*e-3*t,l=0;if(Ga(a)){if($a(o)){var u=-s/o;u>=0&&u<=1&&(r[l++]=u)}}else{var c=o*o-4*a*s;if(Ga(c))r[0]=-o/(2*a);else if(c>0){var d=Ra(c),h=(u=(-o+d)/(2*a),(-o-d)/(2*a));u>=0&&u<=1&&(r[l++]=u),h>=0&&h<=1&&(r[l++]=h)}}return l}function Ja(t,e,n,i,r,o){var a=(e-t)*r+t,s=(n-e)*r+e,l=(i-n)*r+n,u=(s-a)*r+a,c=(l-s)*r+s,d=(c-u)*r+u;o[0]=t,o[1]=a,o[2]=u,o[3]=d,o[4]=d,o[5]=c,o[6]=l,o[7]=i}function Ka(t,e,n,i,r,o,a,s,l,u,c){var d,h,f,p,m,g=.005,y=1/0;Ba[0]=l,Ba[1]=u;for(var v=0;v<1;v+=.05)Fa[0]=Ua(t,n,r,a,v),Fa[1]=Ua(e,i,o,s,v),p=Rt(Ba,Fa),p<y&&(d=v,y=p);y=1/0;for(var _=0;_<32;_++){if(g<ja)break;h=d-g,f=d+g,Fa[0]=Ua(t,n,r,a,h),Fa[1]=Ua(e,i,o,s,h),p=Rt(Fa,Ba),h>=0&&p<y?(d=h,y=p):(Va[0]=Ua(t,n,r,a,f),Va[1]=Ua(e,i,o,s,f),m=Rt(Va,Ba),f<=1&&m<y?(d=f,y=m):g*=.5)}return c&&(c[0]=Ua(t,n,r,a,d),c[1]=Ua(e,i,o,s,d)),Ra(y)}function Qa(t,e,n,i,r,o,a,s,l){for(var u=t,c=e,d=0,h=1/l,f=1;f<=l;f++){var p=f*h,m=Ua(t,n,r,a,p),g=Ua(e,i,o,s,p),y=m-u,v=g-c;d+=Math.sqrt(y*y+v*v),u=m,c=g}return d}function ts(t,e,n,i){var r=1-i;return r*(r*t+2*i*e)+i*i*n}function es(t,e,n,i){return 2*((1-i)*(e-t)+i*(n-e))}function ns(t,e,n,i,r){var o=t-2*e+n,a=2*(e-t),s=t-i,l=0;if(Ga(o)){if($a(a)){var u=-s/a;u>=0&&u<=1&&(r[l++]=u)}}else{var c=a*a-4*o*s;if(Ga(c)){u=-a/(2*o);u>=0&&u<=1&&(r[l++]=u)}else if(c>0){var d=Ra(c),h=(u=(-a+d)/(2*o),(-a-d)/(2*o));u>=0&&u<=1&&(r[l++]=u),h>=0&&h<=1&&(r[l++]=h)}}return l}function is(t,e,n){var i=t+n-2*e;return 0===i?.5:(t-e)/i}function rs(t,e,n,i,r){var o=(e-t)*i+t,a=(n-e)*i+e,s=(a-o)*i+o;r[0]=t,r[1]=o,r[2]=s,r[3]=s,r[4]=a,r[5]=n}function os(t,e,n,i,r,o,a,s,l){var u,c=.005,d=1/0;Ba[0]=a,Ba[1]=s;for(var h=0;h<1;h+=.05){Fa[0]=ts(t,n,r,h),Fa[1]=ts(e,i,o,h);var f=Rt(Ba,Fa);f<d&&(u=h,d=f)}d=1/0;for(var p=0;p<32;p++){if(c<ja)break;var m=u-c,g=u+c;Fa[0]=ts(t,n,r,m),Fa[1]=ts(e,i,o,m);f=Rt(Fa,Ba);if(m>=0&&f<d)u=m,d=f;else{Va[0]=ts(t,n,r,g),Va[1]=ts(e,i,o,g);var y=Rt(Va,Ba);g<=1&&y<d?(u=g,d=y):c*=.5}}return l&&(l[0]=ts(t,n,r,u),l[1]=ts(e,i,o,u)),Ra(d)}function as(t,e,n,i,r,o,a){for(var s=t,l=e,u=0,c=1/a,d=1;d<=a;d++){var h=d*c,f=ts(t,n,r,h),p=ts(e,i,o,h),m=f-s,g=p-l;u+=Math.sqrt(m*m+g*g),s=f,l=p}return u}var ss=Math.min,ls=Math.max,us=Math.sin,cs=Math.cos,ds=2*Math.PI,hs=Mt(),fs=Mt(),ps=Mt();function ms(t,e,n){if(0!==t.length){for(var i=t[0],r=i[0],o=i[0],a=i[1],s=i[1],l=1;l<t.length;l++)i=t[l],r=ss(r,i[0]),o=ls(o,i[0]),a=ss(a,i[1]),s=ls(s,i[1]);e[0]=r,e[1]=a,n[0]=o,n[1]=s}}function gs(t,e,n,i,r,o){r[0]=ss(t,n),r[1]=ss(e,i),o[0]=ls(t,n),o[1]=ls(e,i)}var ys=[],vs=[];function _s(t,e,n,i,r,o,a,s,l,u){var c=Za,d=Ua,h=c(t,n,r,a,ys);l[0]=1/0,l[1]=1/0,u[0]=-1/0,u[1]=-1/0;for(var f=0;f<h;f++){var p=d(t,n,r,a,ys[f]);l[0]=ss(p,l[0]),u[0]=ls(p,u[0])}h=c(e,i,o,s,vs);for(f=0;f<h;f++){var m=d(e,i,o,s,vs[f]);l[1]=ss(m,l[1]),u[1]=ls(m,u[1])}l[0]=ss(t,l[0]),u[0]=ls(t,u[0]),l[0]=ss(a,l[0]),u[0]=ls(a,u[0]),l[1]=ss(e,l[1]),u[1]=ls(e,u[1]),l[1]=ss(s,l[1]),u[1]=ls(s,u[1])}function bs(t,e,n,i,r,o,a,s){var l=is,u=ts,c=ls(ss(l(t,n,r),1),0),d=ls(ss(l(e,i,o),1),0),h=u(t,n,r,c),f=u(e,i,o,d);a[0]=ss(t,r,h),a[1]=ss(e,o,f),s[0]=ls(t,r,h),s[1]=ls(e,o,f)}function xs(t,e,n,i,r,o,a,s,l){var u=zt,c=Wt,d=Math.abs(r-o);if(d%ds<1e-4&&d>1e-4)return s[0]=t-n,s[1]=e-i,l[0]=t+n,void(l[1]=e+i);if(hs[0]=cs(r)*n+t,hs[1]=us(r)*i+e,fs[0]=cs(o)*n+t,fs[1]=us(o)*i+e,u(s,hs,fs),c(l,hs,fs),r%=ds,r<0&&(r+=ds),o%=ds,o<0&&(o+=ds),r>o&&!a?o+=ds:r<o&&a&&(r+=ds),a){var h=o;o=r,r=h}for(var f=0;f<o;f+=Math.PI/2)f>r&&(ps[0]=cs(f)*n+t,ps[1]=us(f)*i+e,u(s,ps,s),c(l,ps,l))}var ws={M:1,L:2,C:3,Q:4,A:5,Z:6,R:7},Ms=[],Ss=[],Ls=[],ks=[],Ts=[],Ds=[],Cs=Math.min,Is=Math.max,As=Math.cos,Ys=Math.sin,Os=Math.sqrt,Ps=Math.abs,Es=Math.PI,Ns=2*Es,Rs="undefined"!==typeof Float32Array,Hs=[];function js(t){var e=Math.round(t/Es*1e8)/1e8;return e%2*Es}function zs(t,e){var n=js(t[0]);n<0&&(n+=Ns);var i=n-t[0],r=t[1];r+=i,!e&&r-n>=Ns?r=n+Ns:e&&n-r>=Ns?r=n-Ns:!e&&n>r?r=n+(Ns-js(n-r)):e&&n<r&&(r=n-(Ns-js(r-n))),t[0]=n,t[1]=r}var Ws=function(){function t(t){this.dpr=1,this._xi=0,this._yi=0,this._x0=0,this._y0=0,this._len=0,t&&(this._saveData=!1),this._saveData&&(this.data=[])}return t.prototype.increaseVersion=function(){this._version++},t.prototype.getVersion=function(){return this._version},t.prototype.setScale=function(t,e,n){n=n||0,n>0&&(this._ux=Ps(n/mi/t)||0,this._uy=Ps(n/mi/e)||0)},t.prototype.setDPR=function(t){this.dpr=t},t.prototype.setContext=function(t){this._ctx=t},t.prototype.getContext=function(){return this._ctx},t.prototype.beginPath=function(){return this._ctx&&this._ctx.beginPath(),this.reset(),this},t.prototype.reset=function(){this._saveData&&(this._len=0),this._lineDash&&(this._lineDash=null,this._dashOffset=0),this._pathSegLen&&(this._pathSegLen=null,this._pathLen=0),this._version++},t.prototype.moveTo=function(t,e){return this._drawPendingPt(),this.addData(ws.M,t,e),this._ctx&&this._ctx.moveTo(t,e),this._x0=t,this._y0=e,this._xi=t,this._yi=e,this},t.prototype.lineTo=function(t,e){var n=Ps(t-this._xi),i=Ps(e-this._yi),r=n>this._ux||i>this._uy;if(this.addData(ws.L,t,e),this._ctx&&r&&(this._needsDash?this._dashedLineTo(t,e):this._ctx.lineTo(t,e)),r)this._xi=t,this._yi=e,this._pendingPtDist=0;else{var o=n*n+i*i;o>this._pendingPtDist&&(this._pendingPtX=t,this._pendingPtY=e,this._pendingPtDist=o)}return this},t.prototype.bezierCurveTo=function(t,e,n,i,r,o){return this.addData(ws.C,t,e,n,i,r,o),this._ctx&&(this._needsDash?this._dashedBezierTo(t,e,n,i,r,o):this._ctx.bezierCurveTo(t,e,n,i,r,o)),this._xi=r,this._yi=o,this},t.prototype.quadraticCurveTo=function(t,e,n,i){return this.addData(ws.Q,t,e,n,i),this._ctx&&(this._needsDash?this._dashedQuadraticTo(t,e,n,i):this._ctx.quadraticCurveTo(t,e,n,i)),this._xi=n,this._yi=i,this},t.prototype.arc=function(t,e,n,i,r,o){Hs[0]=i,Hs[1]=r,zs(Hs,o),i=Hs[0],r=Hs[1];var a=r-i;return this.addData(ws.A,t,e,n,n,i,a,0,o?0:1),this._ctx&&this._ctx.arc(t,e,n,i,r,o),this._xi=As(r)*n+t,this._yi=Ys(r)*n+e,this},t.prototype.arcTo=function(t,e,n,i,r){return this._ctx&&this._ctx.arcTo(t,e,n,i,r),this},t.prototype.rect=function(t,e,n,i){return this._ctx&&this._ctx.rect(t,e,n,i),this.addData(ws.R,t,e,n,i),this},t.prototype.closePath=function(){this._drawPendingPt(),this.addData(ws.Z);var t=this._ctx,e=this._x0,n=this._y0;return t&&(this._needsDash&&this._dashedLineTo(e,n),t.closePath()),this._xi=e,this._yi=n,this},t.prototype.fill=function(t){t&&t.fill(),this.toStatic()},t.prototype.stroke=function(t){t&&t.stroke(),this.toStatic()},t.prototype.setLineDash=function(t){if(t instanceof Array){this._lineDash=t,this._dashIdx=0;for(var e=0,n=0;n<t.length;n++)e+=t[n];this._dashSum=e,this._needsDash=!0}else this._lineDash=null,this._needsDash=!1;return this},t.prototype.setLineDashOffset=function(t){return this._dashOffset=t,this},t.prototype.len=function(){return this._len},t.prototype.setData=function(t){var e=t.length;this.data&&this.data.length===e||!Rs||(this.data=new Float32Array(e));for(var n=0;n<e;n++)this.data[n]=t[n];this._len=e},t.prototype.appendPath=function(t){t instanceof Array||(t=[t]);for(var e=t.length,n=0,i=this._len,r=0;r<e;r++)n+=t[r].len();Rs&&this.data instanceof Float32Array&&(this.data=new Float32Array(i+n));for(r=0;r<e;r++)for(var o=t[r].data,a=0;a<o.length;a++)this.data[i++]=o[a];this._len=i},t.prototype.addData=function(t,e,n,i,r,o,a,s,l){if(this._saveData){var u=this.data;this._len+arguments.length>u.length&&(this._expandData(),u=this.data);for(var c=0;c<arguments.length;c++)u[this._len++]=arguments[c]}},t.prototype._drawPendingPt=function(){this._pendingPtDist>0&&(this._ctx&&this._ctx.lineTo(this._pendingPtX,this._pendingPtY),this._pendingPtDist=0)},t.prototype._expandData=function(){if(!(this.data instanceof Array)){for(var t=[],e=0;e<this._len;e++)t[e]=this.data[e];this.data=t}},t.prototype._dashedLineTo=function(t,e){var n,i,r=this._dashSum,o=this._lineDash,a=this._ctx,s=this._dashOffset,l=this._xi,u=this._yi,c=t-l,d=e-u,h=Os(c*c+d*d),f=l,p=u,m=o.length;c/=h,d/=h,s<0&&(s=r+s),s%=r,f-=s*c,p-=s*d;while(c>0&&f<=t||c<0&&f>=t||0===c&&(d>0&&p<=e||d<0&&p>=e))i=this._dashIdx,n=o[i],f+=c*n,p+=d*n,this._dashIdx=(i+1)%m,c>0&&f<l||c<0&&f>l||d>0&&p<u||d<0&&p>u||a[i%2?"moveTo":"lineTo"](c>=0?Cs(f,t):Is(f,t),d>=0?Cs(p,e):Is(p,e));c=f-t,d=p-e,this._dashOffset=-Os(c*c+d*d)},t.prototype._dashedBezierTo=function(t,e,n,i,r,o){var a,s,l,u,c,d=this._ctx,h=this._dashSum,f=this._dashOffset,p=this._lineDash,m=this._xi,g=this._yi,y=0,v=this._dashIdx,_=p.length,b=0;for(f<0&&(f=h+f),f%=h,a=0;a<1;a+=.1)s=Ua(m,t,n,r,a+.1)-Ua(m,t,n,r,a),l=Ua(g,e,i,o,a+.1)-Ua(g,e,i,o,a),y+=Os(s*s+l*l);for(;v<_;v++)if(b+=p[v],b>f)break;a=(b-f)/y;while(a<=1)u=Ua(m,t,n,r,a),c=Ua(g,e,i,o,a),v%2?d.moveTo(u,c):d.lineTo(u,c),a+=p[v]/y,v=(v+1)%_;v%2!==0&&d.lineTo(r,o),s=r-u,l=o-c,this._dashOffset=-Os(s*s+l*l)},t.prototype._dashedQuadraticTo=function(t,e,n,i){var r=n,o=i;n=(n+2*t)/3,i=(i+2*e)/3,t=(this._xi+2*t)/3,e=(this._yi+2*e)/3,this._dashedBezierTo(t,e,n,i,r,o)},t.prototype.toStatic=function(){if(this._saveData){this._drawPendingPt();var t=this.data;t instanceof Array&&(t.length=this._len,Rs&&this._len>11&&(this.data=new Float32Array(t)))}},t.prototype.getBoundingRect=function(){Ls[0]=Ls[1]=Ts[0]=Ts[1]=Number.MAX_VALUE,ks[0]=ks[1]=Ds[0]=Ds[1]=-Number.MAX_VALUE;var t,e=this.data,n=0,i=0,r=0,o=0;for(t=0;t<this._len;){var a=e[t++],s=1===t;switch(s&&(n=e[t],i=e[t+1],r=n,o=i),a){case ws.M:n=r=e[t++],i=o=e[t++],Ts[0]=r,Ts[1]=o,Ds[0]=r,Ds[1]=o;break;case ws.L:gs(n,i,e[t],e[t+1],Ts,Ds),n=e[t++],i=e[t++];break;case ws.C:_s(n,i,e[t++],e[t++],e[t++],e[t++],e[t],e[t+1],Ts,Ds),n=e[t++],i=e[t++];break;case ws.Q:bs(n,i,e[t++],e[t++],e[t],e[t+1],Ts,Ds),n=e[t++],i=e[t++];break;case ws.A:var l=e[t++],u=e[t++],c=e[t++],d=e[t++],h=e[t++],f=e[t++]+h;t+=1;var p=!e[t++];s&&(r=As(h)*c+l,o=Ys(h)*d+u),xs(l,u,c,d,h,f,p,Ts,Ds),n=As(f)*c+l,i=Ys(f)*d+u;break;case ws.R:r=n=e[t++],o=i=e[t++];var m=e[t++],g=e[t++];gs(r,o,r+m,o+g,Ts,Ds);break;case ws.Z:n=r,i=o;break}zt(Ls,Ls,Ts),Wt(ks,ks,Ds)}return 0===t&&(Ls[0]=Ls[1]=ks[0]=ks[1]=0),new Ji(Ls[0],Ls[1],ks[0]-Ls[0],ks[1]-Ls[1])},t.prototype._calculateLength=function(){var t=this.data,e=this._len,n=this._ux,i=this._uy,r=0,o=0,a=0,s=0;this._pathSegLen||(this._pathSegLen=[]);for(var l=this._pathSegLen,u=0,c=0,d=0;d<e;){var h=t[d++],f=1===d;f&&(r=t[d],o=t[d+1],a=r,s=o);var p=-1;switch(h){case ws.M:r=a=t[d++],o=s=t[d++];break;case ws.L:var m=t[d++],g=t[d++],y=m-r,v=g-o;(Ps(y)>n||Ps(v)>i||d===e-1)&&(p=Math.sqrt(y*y+v*v),r=m,o=g);break;case ws.C:var _=t[d++],b=t[d++],x=(m=t[d++],g=t[d++],t[d++]),w=t[d++];p=Qa(r,o,_,b,m,g,x,w,10),r=x,o=w;break;case ws.Q:_=t[d++],b=t[d++],m=t[d++],g=t[d++];p=as(r,o,_,b,m,g,10),r=m,o=g;break;case ws.A:var M=t[d++],S=t[d++],L=t[d++],k=t[d++],T=t[d++],D=t[d++],C=D+T;d+=1;t[d++];f&&(a=As(T)*L+M,s=Ys(T)*k+S),p=Is(L,k)*Cs(Ns,Math.abs(D)),r=As(C)*L+M,o=Ys(C)*k+S;break;case ws.R:a=r=t[d++],s=o=t[d++];var I=t[d++],A=t[d++];p=2*I+2*A;break;case ws.Z:y=a-r,v=s-o;p=Math.sqrt(y*y+v*v),r=a,o=s;break}p>=0&&(l[c++]=p,u+=p)}return this._pathLen=u,u},t.prototype.rebuildPath=function(t,e){var n,i,r,o,a,s,l,u,c,d,h,f=this.data,p=this._ux,m=this._uy,g=this._len,y=e<1,v=0,_=0,b=0;if(!y||(this._pathSegLen||this._calculateLength(),l=this._pathSegLen,u=this._pathLen,c=e*u,c))t:for(var x=0;x<g;){var w=f[x++],M=1===x;switch(M&&(r=f[x],o=f[x+1],n=r,i=o),w){case ws.M:b>0&&(t.lineTo(d,h),b=0),n=r=f[x++],i=o=f[x++],t.moveTo(r,o);break;case ws.L:a=f[x++],s=f[x++];var S=Ps(a-r),L=Ps(s-o);if(S>p||L>m){if(y){var k=l[_++];if(v+k>c){var T=(c-v)/k;t.lineTo(r*(1-T)+a*T,o*(1-T)+s*T);break t}v+=k}t.lineTo(a,s),r=a,o=s,b=0}else{var D=S*S+L*L;D>b&&(d=a,h=s,b=D)}break;case ws.C:var C=f[x++],I=f[x++],A=f[x++],Y=f[x++],O=f[x++],P=f[x++];if(y){k=l[_++];if(v+k>c){T=(c-v)/k;Ja(r,C,A,O,T,Ms),Ja(o,I,Y,P,T,Ss),t.bezierCurveTo(Ms[1],Ss[1],Ms[2],Ss[2],Ms[3],Ss[3]);break t}v+=k}t.bezierCurveTo(C,I,A,Y,O,P),r=O,o=P;break;case ws.Q:C=f[x++],I=f[x++],A=f[x++],Y=f[x++];if(y){k=l[_++];if(v+k>c){T=(c-v)/k;rs(r,C,A,T,Ms),rs(o,I,Y,T,Ss),t.quadraticCurveTo(Ms[1],Ss[1],Ms[2],Ss[2]);break t}v+=k}t.quadraticCurveTo(C,I,A,Y),r=A,o=Y;break;case ws.A:var E=f[x++],N=f[x++],R=f[x++],H=f[x++],j=f[x++],z=f[x++],W=f[x++],B=!f[x++],F=R>H?R:H,V=Ps(R-H)>.001,G=j+z,$=!1;if(y){k=l[_++];v+k>c&&(G=j+z*(c-v)/k,$=!0),v+=k}if(V&&t.ellipse?t.ellipse(E,N,R,H,W,j,G,B):t.arc(E,N,F,j,G,B),$)break t;M&&(n=As(j)*R+E,i=Ys(j)*H+N),r=As(G)*R+E,o=Ys(G)*H+N;break;case ws.R:n=r=f[x],i=o=f[x+1],a=f[x++],s=f[x++];var U=f[x++],q=f[x++];if(y){k=l[_++];if(v+k>c){var X=c-v;t.moveTo(a,s),t.lineTo(a+Cs(X,U),s),X-=U,X>0&&t.lineTo(a+U,s+Cs(X,q)),X-=q,X>0&&t.lineTo(a+Is(U-X,0),s+q),X-=U,X>0&&t.lineTo(a,s+Is(q-X,0));break t}v+=k}t.rect(a,s,U,q);break;case ws.Z:if(b>0&&(t.lineTo(d,h),b=0),y){k=l[_++];if(v+k>c){T=(c-v)/k;t.lineTo(r*(1-T)+n*T,o*(1-T)+i*T);break t}v+=k}t.closePath(),r=n,o=i}}},t.CMD=ws,t.initDefaultProps=function(){var e=t.prototype;e._saveData=!0,e._needsDash=!1,e._dashOffset=0,e._dashIdx=0,e._dashSum=0,e._ux=0,e._uy=0,e._pendingPtDist=0,e._version=0}(),t}(),Bs=Ws;function Fs(t,e,n,i,r,o,a){if(0===r)return!1;var s=r,l=0,u=t;if(a>e+s&&a>i+s||a<e-s&&a<i-s||o>t+s&&o>n+s||o<t-s&&o<n-s)return!1;if(t===n)return Math.abs(o-t)<=s/2;l=(e-i)/(t-n),u=(t*i-n*e)/(t-n);var c=l*o-a+u,d=c*c/(l*l+1);return d<=s/2*s/2}function Vs(t,e,n,i,r,o,a,s,l,u,c){if(0===l)return!1;var d=l;if(c>e+d&&c>i+d&&c>o+d&&c>s+d||c<e-d&&c<i-d&&c<o-d&&c<s-d||u>t+d&&u>n+d&&u>r+d&&u>a+d||u<t-d&&u<n-d&&u<r-d&&u<a-d)return!1;var h=Ka(t,e,n,i,r,o,a,s,u,c,null);return h<=d/2}function Gs(t,e,n,i,r,o,a,s,l){if(0===a)return!1;var u=a;if(l>e+u&&l>i+u&&l>o+u||l<e-u&&l<i-u&&l<o-u||s>t+u&&s>n+u&&s>r+u||s<t-u&&s<n-u&&s<r-u)return!1;var c=os(t,e,n,i,r,o,s,l,null);return c<=u/2}var $s=2*Math.PI;function Us(t){return t%=$s,t<0&&(t+=$s),t}var qs=2*Math.PI;function Xs(t,e,n,i,r,o,a,s,l){if(0===a)return!1;var u=a;s-=t,l-=e;var c=Math.sqrt(s*s+l*l);if(c-u>n||c+u<n)return!1;if(Math.abs(i-r)%qs<1e-4)return!0;if(o){var d=i;i=Us(r),r=Us(d)}else i=Us(i),r=Us(r);i>r&&(r+=qs);var h=Math.atan2(l,s);return h<0&&(h+=qs),h>=i&&h<=r||h+qs>=i&&h+qs<=r}function Zs(t,e,n,i,r,o){if(o>e&&o>i||o<e&&o<i)return 0;if(i===e)return 0;var a=(o-e)/(i-e),s=i<e?1:-1;1!==a&&0!==a||(s=i<e?.5:-.5);var l=a*(n-t)+t;return l===r?1/0:l>r?s:0}var Js=Bs.CMD,Ks=2*Math.PI,Qs=1e-4;function tl(t,e){return Math.abs(t-e)<Qs}var el=[-1,-1,-1],nl=[-1,-1];function il(){var t=nl[0];nl[0]=nl[1],nl[1]=t}function rl(t,e,n,i,r,o,a,s,l,u){if(u>e&&u>i&&u>o&&u>s||u<e&&u<i&&u<o&&u<s)return 0;var c=Xa(e,i,o,s,u,el);if(0===c)return 0;for(var d=0,h=-1,f=void 0,p=void 0,m=0;m<c;m++){var g=el[m],y=0===g||1===g?.5:1,v=Ua(t,n,r,a,g);v<l||(h<0&&(h=Za(e,i,o,s,nl),nl[1]<nl[0]&&h>1&&il(),f=Ua(e,i,o,s,nl[0]),h>1&&(p=Ua(e,i,o,s,nl[1]))),2===h?g<nl[0]?d+=f<e?y:-y:g<nl[1]?d+=p<f?y:-y:d+=s<p?y:-y:g<nl[0]?d+=f<e?y:-y:d+=s<f?y:-y)}return d}function ol(t,e,n,i,r,o,a,s){if(s>e&&s>i&&s>o||s<e&&s<i&&s<o)return 0;var l=ns(e,i,o,s,el);if(0===l)return 0;var u=is(e,i,o);if(u>=0&&u<=1){for(var c=0,d=ts(e,i,o,u),h=0;h<l;h++){var f=0===el[h]||1===el[h]?.5:1,p=ts(t,n,r,el[h]);p<a||(el[h]<u?c+=d<e?f:-f:c+=o<d?f:-f)}return c}f=0===el[0]||1===el[0]?.5:1,p=ts(t,n,r,el[0]);return p<a?0:o<e?f:-f}function al(t,e,n,i,r,o,a,s){if(s-=e,s>n||s<-n)return 0;var l=Math.sqrt(n*n-s*s);el[0]=-l,el[1]=l;var u=Math.abs(i-r);if(u<1e-4)return 0;if(u>=Ks-1e-4){i=0,r=Ks;var c=o?1:-1;return a>=el[0]+t&&a<=el[1]+t?c:0}if(i>r){var d=i;i=r,r=d}i<0&&(i+=Ks,r+=Ks);for(var h=0,f=0;f<2;f++){var p=el[f];if(p+t>a){var m=Math.atan2(s,p);c=o?1:-1;m<0&&(m=Ks+m),(m>=i&&m<=r||m+Ks>=i&&m+Ks<=r)&&(m>Math.PI/2&&m<1.5*Math.PI&&(c=-c),h+=c)}}return h}function sl(t,e,n,i,r){for(var o,a,s=t.data,l=t.len(),u=0,c=0,d=0,h=0,f=0,p=0;p<l;){var m=s[p++],g=1===p;switch(m===Js.M&&p>1&&(n||(u+=Zs(c,d,h,f,i,r))),g&&(c=s[p],d=s[p+1],h=c,f=d),m){case Js.M:h=s[p++],f=s[p++],c=h,d=f;break;case Js.L:if(n){if(Fs(c,d,s[p],s[p+1],e,i,r))return!0}else u+=Zs(c,d,s[p],s[p+1],i,r)||0;c=s[p++],d=s[p++];break;case Js.C:if(n){if(Vs(c,d,s[p++],s[p++],s[p++],s[p++],s[p],s[p+1],e,i,r))return!0}else u+=rl(c,d,s[p++],s[p++],s[p++],s[p++],s[p],s[p+1],i,r)||0;c=s[p++],d=s[p++];break;case Js.Q:if(n){if(Gs(c,d,s[p++],s[p++],s[p],s[p+1],e,i,r))return!0}else u+=ol(c,d,s[p++],s[p++],s[p],s[p+1],i,r)||0;c=s[p++],d=s[p++];break;case Js.A:var y=s[p++],v=s[p++],_=s[p++],b=s[p++],x=s[p++],w=s[p++];p+=1;var M=!!(1-s[p++]);o=Math.cos(x)*_+y,a=Math.sin(x)*b+v,g?(h=o,f=a):u+=Zs(c,d,o,a,i,r);var S=(i-y)*b/_+y;if(n){if(Xs(y,v,b,x,x+w,M,e,S,r))return!0}else u+=al(y,v,b,x,x+w,M,S,r);c=Math.cos(x+w)*_+y,d=Math.sin(x+w)*b+v;break;case Js.R:h=c=s[p++],f=d=s[p++];var L=s[p++],k=s[p++];if(o=h+L,a=f+k,n){if(Fs(h,f,o,f,e,i,r)||Fs(o,f,o,a,e,i,r)||Fs(o,a,h,a,e,i,r)||Fs(h,a,h,f,e,i,r))return!0}else u+=Zs(o,f,o,a,i,r),u+=Zs(h,a,h,f,i,r);break;case Js.Z:if(n){if(Fs(c,d,h,f,e,i,r))return!0}else u+=Zs(c,d,h,f,i,r);c=h,d=f;break}}return n||tl(d,f)||(u+=Zs(c,d,h,f,i,r)||0),0!==u}function ll(t,e,n){return sl(t,0,!1,e,n)}function ul(t,e,n,i){return sl(t,e,!0,n,i)}var cl=A({fill:"#000",stroke:null,strokePercent:1,fillOpacity:1,strokeOpacity:1,lineDashOffset:0,lineWidth:1,lineCap:"butt",miterLimit:10,strokeNoScale:!1,strokeFirst:!1},Ta),dl={style:A({fill:!0,stroke:!0,strokePercent:!0,fillOpacity:!0,strokeOpacity:!0,lineDashOffset:!0,lineWidth:!0,miterLimit:!0},Da.style)},hl=["x","y","rotation","scaleX","scaleY","originX","originY","invisible","culling","z","z2","zlevel","parent"],fl=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.update=function(){var n=this;t.prototype.update.call(this);var i=this.style;if(i.decal){var r=this._decalEl=this._decalEl||new e;r.buildPath===e.prototype.buildPath&&(r.buildPath=function(t){n.buildPath(t,n.shape)}),r.silent=!0;var o=r.style;for(var a in i)o[a]!==i[a]&&(o[a]=i[a]);o.fill=i.fill?i.decal:null,o.decal=null,o.shadowColor=null,i.strokeFirst&&(o.stroke=null);for(var s=0;s<hl.length;++s)r[hl[s]]=this[hl[s]];r.__dirty|=je}else this._decalEl&&(this._decalEl=null)},e.prototype.getDecalElement=function(){return this._decalEl},e.prototype._init=function(e){var n=B(e);this.shape=this.getDefaultShape();var i=this.getDefaultStyle();i&&this.useStyle(i);for(var r=0;r<n.length;r++){var o=n[r],a=e[o];"style"===o?this.style?I(this.style,a):this.useStyle(a):"shape"===o?I(this.shape,a):t.prototype.attrKV.call(this,o,a)}this.style||this.useStyle({})},e.prototype.getDefaultStyle=function(){return null},e.prototype.getDefaultShape=function(){return{}},e.prototype.canBeInsideText=function(){return this.hasFill()},e.prototype.getInsideTextFill=function(){var t=this.style.fill;if("none"!==t){if(q(t)){var e=kn(t,0);return e>.5?yi:e>.2?_i:vi}if(t)return vi}return yi},e.prototype.getInsideTextStroke=function(t){var e=this.style.fill;if(q(e)){var n=this.__zr,i=!(!n||!n.isDarkMode()),r=kn(t,0)<gi;if(i===r)return e}},e.prototype.buildPath=function(t,e,n){},e.prototype.pathUpdated=function(){this.__dirty&=~We},e.prototype.createPathProxy=function(){this.path=new Bs(!1)},e.prototype.hasStroke=function(){var t=this.style,e=t.stroke;return!(null==e||"none"===e||!(t.lineWidth>0))},e.prototype.hasFill=function(){var t=this.style,e=t.fill;return null!=e&&"none"!==e},e.prototype.getBoundingRect=function(){var t=this._rect,e=this.style,n=!t;if(n){var i=!1;this.path||(i=!0,this.createPathProxy());var r=this.path;(i||this.__dirty&We)&&(r.beginPath(),this.buildPath(r,this.shape,!1),this.pathUpdated()),t=r.getBoundingRect()}if(this._rect=t,this.hasStroke()&&this.path&&this.path.len()>0){var o=this._rectWithStroke||(this._rectWithStroke=t.clone());if(this.__dirty||n){o.copy(t);var a=e.strokeNoScale?this.getLineScale():1,s=e.lineWidth;if(!this.hasFill()){var l=this.strokeContainThreshold;s=Math.max(s,null==l?4:l)}a>1e-10&&(o.width+=s/a,o.height+=s/a,o.x-=s/a/2,o.y-=s/a/2)}return o}return t},e.prototype.contain=function(t,e){var n=this.transformCoordToLocal(t,e),i=this.getBoundingRect(),r=this.style;if(t=n[0],e=n[1],i.contain(t,e)){var o=this.path;if(this.hasStroke()){var a=r.lineWidth,s=r.strokeNoScale?this.getLineScale():1;if(s>1e-10&&(this.hasFill()||(a=Math.max(a,this.strokeContainThreshold)),ul(o,a/s,t,e)))return!0}if(this.hasFill())return ll(o,t,e)}return!1},e.prototype.dirtyShape=function(){this.__dirty|=We,this._rect&&(this._rect=null),this._decalEl&&this._decalEl.dirtyShape(),this.markRedraw()},e.prototype.dirty=function(){this.dirtyStyle(),this.dirtyShape()},e.prototype.animateShape=function(t){return this.animate("shape",t)},e.prototype.updateDuringAnimation=function(t){"style"===t?this.dirtyStyle():"shape"===t?this.dirtyShape():this.markRedraw()},e.prototype.attrKV=function(e,n){"shape"===e?this.setShape(n):t.prototype.attrKV.call(this,e,n)},e.prototype.setShape=function(t,e){var n=this.shape;return n||(n=this.shape={}),"string"===typeof t?n[t]=e:I(n,t),this.dirtyShape(),this},e.prototype.shapeChanged=function(){return!!(this.__dirty&We)},e.prototype.createStyle=function(t){return vt(cl,t)},e.prototype._innerSaveToNormal=function(e){t.prototype._innerSaveToNormal.call(this,e);var n=this._normalState;e.shape&&!n.shape&&(n.shape=I({},this.shape))},e.prototype._applyStateObj=function(e,n,i,r,o,a){t.prototype._applyStateObj.call(this,e,n,i,r,o,a);var s,l=!(n&&r);if(n&&n.shape?o?r?s=n.shape:(s=I({},i.shape),I(s,n.shape)):(s=I({},r?this.shape:i.shape),I(s,n.shape)):l&&(s=i.shape),s)if(o){this.shape=I({},this.shape);for(var u={},c=B(s),d=0;d<c.length;d++){var h=c[d];"object"===typeof s[h]?this.shape[h]=s[h]:u[h]=s[h]}this._transitionState(e,{shape:u},a)}else this.shape=s,this.dirtyShape()},e.prototype._mergeStates=function(e){for(var n,i=t.prototype._mergeStates.call(this,e),r=0;r<e.length;r++){var o=e[r];o.shape&&(n=n||{},this._mergeStyle(n,o.shape))}return n&&(i.shape=n),i},e.prototype.getAnimationStyleProps=function(){return dl},e.prototype.isZeroArea=function(){return!1},e.extend=function(t){var n=function(e){function n(n){var i=e.call(this,n)||this;return t.init&&t.init.call(i,n),i}return wt(n,e),n.prototype.getDefaultStyle=function(){return T(t.style)},n.prototype.getDefaultShape=function(){return T(t.shape)},n}(e);for(var i in t)"function"===typeof t[i]&&(n.prototype[i]=t[i]);return n},e.initDefaultProps=function(){var t=e.prototype;t.type="path",t.strokeContainThreshold=5,t.segmentIgnoreThreshold=0,t.subPixelOptimize=!1,t.autoBatch=!1,t.__dirty=je|ze|We}(),e}(Ea),pl=fl,ml=A({strokeFirst:!0,font:Qi,x:0,y:0,textAlign:"left",textBaseline:"top",miterLimit:2},cl),gl=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return wt(e,t),e.prototype.hasStroke=function(){var t=this.style,e=t.stroke;return null!=e&&"none"!==e&&t.lineWidth>0},e.prototype.hasFill=function(){var t=this.style,e=t.fill;return null!=e&&"none"!==e},e.prototype.createStyle=function(t){return vt(ml,t)},e.prototype.setBoundingRect=function(t){this._rect=t},e.prototype.getBoundingRect=function(){var t=this.style;if(!this._rect){var e=t.text;null!=e?e+="":e="";var n=rr(e,t.font,t.textAlign,t.textBaseline);if(n.x+=t.x||0,n.y+=t.y||0,this.hasStroke()){var i=t.lineWidth;n.x-=i/2,n.y-=i/2,n.width+=i,n.height+=i}this._rect=n}return this._rect},e.initDefaultProps=function(){var t=e.prototype;t.dirtyRectTolerance=10}(),e}(Ea);gl.prototype.type="tspan";var yl=gl,vl=A({x:0,y:0},Ta),_l={style:A({x:!0,y:!0,width:!0,height:!0,sx:!0,sy:!0,sWidth:!0,sHeight:!0},Da.style)};function bl(t){return!!(t&&"string"!==typeof t&&t.width&&t.height)}var xl=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return wt(e,t),e.prototype.createStyle=function(t){return vt(vl,t)},e.prototype._getSize=function(t){var e=this.style,n=e[t];if(null!=n)return n;var i=bl(e.image)?e.image:this.__image;if(!i)return 0;var r="width"===t?"height":"width",o=e[r];return null==o?i[t]:i[t]/i[r]*o},e.prototype.getWidth=function(){return this._getSize("width")},e.prototype.getHeight=function(){return this._getSize("height")},e.prototype.getAnimationStyleProps=function(){return _l},e.prototype.getBoundingRect=function(){var t=this.style;return this._rect||(this._rect=new Ji(t.x||0,t.y||0,this.getWidth(),this.getHeight())),this._rect},e}(Ea);xl.prototype.type="image";var wl=xl;function Ml(t,e){var n,i,r,o,a,s=e.x,l=e.y,u=e.width,c=e.height,d=e.r;u<0&&(s+=u,u=-u),c<0&&(l+=c,c=-c),"number"===typeof d?n=i=r=o=d:d instanceof Array?1===d.length?n=i=r=o=d[0]:2===d.length?(n=r=d[0],i=o=d[1]):3===d.length?(n=d[0],i=o=d[1],r=d[2]):(n=d[0],i=d[1],r=d[2],o=d[3]):n=i=r=o=0,n+i>u&&(a=n+i,n*=u/a,i*=u/a),r+o>u&&(a=r+o,r*=u/a,o*=u/a),i+r>c&&(a=i+r,i*=c/a,r*=c/a),n+o>c&&(a=n+o,n*=c/a,o*=c/a),t.moveTo(s+n,l),t.lineTo(s+u-i,l),0!==i&&t.arc(s+u-i,l+i,i,-Math.PI/2,0),t.lineTo(s+u,l+c-r),0!==r&&t.arc(s+u-r,l+c-r,r,0,Math.PI/2),t.lineTo(s+o,l+c),0!==o&&t.arc(s+o,l+c-o,o,Math.PI/2,Math.PI),t.lineTo(s,l+n),0!==n&&t.arc(s+n,l+n,n,Math.PI,1.5*Math.PI)}var Sl=Math.round;function Ll(t,e,n){if(e){var i=e.x1,r=e.x2,o=e.y1,a=e.y2;t.x1=i,t.x2=r,t.y1=o,t.y2=a;var s=n&&n.lineWidth;return s?(Sl(2*i)===Sl(2*r)&&(t.x1=t.x2=Tl(i,s,!0)),Sl(2*o)===Sl(2*a)&&(t.y1=t.y2=Tl(o,s,!0)),t):t}}function kl(t,e,n){if(e){var i=e.x,r=e.y,o=e.width,a=e.height;t.x=i,t.y=r,t.width=o,t.height=a;var s=n&&n.lineWidth;return s?(t.x=Tl(i,s,!0),t.y=Tl(r,s,!0),t.width=Math.max(Tl(i+o,s,!1)-t.x,0===o?0:1),t.height=Math.max(Tl(r+a,s,!1)-t.y,0===a?0:1),t):t}}function Tl(t,e,n){if(!e)return t;var i=Sl(2*t);return(i+Sl(e))%2===0?i/2:(i+(n?1:-1))/2}var Dl=function(){function t(){this.x=0,this.y=0,this.width=0,this.height=0}return t}(),Cl={},Il=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultShape=function(){return new Dl},e.prototype.buildPath=function(t,e){var n,i,r,o;if(this.subPixelOptimize){var a=kl(Cl,e,this.style);n=a.x,i=a.y,r=a.width,o=a.height,a.r=e.r,e=a}else n=e.x,i=e.y,r=e.width,o=e.height;e.r?Ml(t,e):t.rect(n,i,r,o)},e.prototype.isZeroArea=function(){return!this.shape.width||!this.shape.height},e}(pl);Il.prototype.type="rect";var Al=Il,Yl={fill:"#000"},Ol=2,Pl={style:A({fill:!0,stroke:!0,fillOpacity:!0,strokeOpacity:!0,lineWidth:!0,fontSize:!0,lineHeight:!0,width:!0,height:!0,textShadowColor:!0,textShadowBlur:!0,textShadowOffsetX:!0,textShadowOffsetY:!0,backgroundColor:!0,padding:!0,borderColor:!0,borderWidth:!0,borderRadius:!0},Da.style)},El=function(t){function e(e){var n=t.call(this)||this;return n.type="text",n._children=[],n._defaultStyle=Yl,n.attr(e),n}return wt(e,t),e.prototype.childrenRef=function(){return this._children},e.prototype.update=function(){this.styleChanged()&&this._updateSubTexts();for(var e=0;e<this._children.length;e++){var n=this._children[e];n.zlevel=this.zlevel,n.z=this.z,n.z2=this.z2,n.culling=this.culling,n.cursor=this.cursor,n.invisible=this.invisible}var i=this.attachedTransform;if(i){i.updateTransform();var r=i.transform;r?(this.transform=this.transform||[],wi(this.transform,r)):this.transform=null}else t.prototype.update.call(this)},e.prototype.getComputedTransform=function(){return this.__hostTarget&&(this.__hostTarget.getComputedTransform(),this.__hostTarget.updateInnerText(!0)),this.attachedTransform?this.attachedTransform.getComputedTransform():t.prototype.getComputedTransform.call(this)},e.prototype._updateSubTexts=function(){this._childCursor=0,Hl(this.style),this.style.rich?this._updateRichTexts():this._updatePlainTexts(),this._children.length=this._childCursor,this.styleUpdated()},e.prototype.addSelfToZr=function(e){t.prototype.addSelfToZr.call(this,e);for(var n=0;n<this._children.length;n++)this._children[n].__zr=e},e.prototype.removeSelfFromZr=function(e){t.prototype.removeSelfFromZr.call(this,e);for(var n=0;n<this._children.length;n++)this._children[n].__zr=null},e.prototype.getBoundingRect=function(){if(this.styleChanged()&&this._updateSubTexts(),!this._rect){for(var t=new Ji(0,0,0,0),e=this._children,n=[],i=null,r=0;r<e.length;r++){var o=e[r],a=o.getBoundingRect(),s=o.getLocalTransform(n);s?(t.copy(a),t.applyTransform(s),i=i||t.clone(),i.union(t)):(i=i||a.clone(),i.union(a))}this._rect=i||t}return this._rect},e.prototype.setDefaultTextStyle=function(t){this._defaultStyle=t||Yl},e.prototype.setTextContent=function(t){throw new Error("Can't attach text on another text")},e.prototype._mergeStyle=function(t,e){if(!e)return t;var n=e.rich,i=t.rich||n&&{};return I(t,e),n&&i?(this._mergeRich(i,n),t.rich=i):i&&(t.rich=i),t},e.prototype._mergeRich=function(t,e){for(var n=B(e),i=0;i<n.length;i++){var r=n[i];t[r]=t[r]||{},I(t[r],e[r])}},e.prototype.getAnimationStyleProps=function(){return Pl},e.prototype._getOrCreateChild=function(t){var e=this._children[this._childCursor];return e&&e instanceof t||(e=new t),this._children[this._childCursor++]=e,e.__zr=this.__zr,e.parent=this,e},e.prototype._updatePlainTexts=function(){var t=this.style,e=t.font||Qi,n=t.padding,i=Fl(t),r=ga(i,t),o=Vl(t),a=!!t.backgroundColor,s=r.outerHeight,l=r.lines,u=r.lineHeight,c=this._defaultStyle,d=t.x||0,h=t.y||0,f=t.align||c.align||"left",p=t.verticalAlign||c.verticalAlign||"top",m=d,g=ar(h,r.contentHeight,p);if(o||n){var y=r.width;n&&(y+=n[1]+n[3]);var v=or(d,y,f),_=ar(h,s,p);o&&this._renderBackground(t,t,v,_,y,s)}g+=u/2,n&&(m=Bl(d,f,n),"top"===p?g+=n[0]:"bottom"===p&&(g-=n[2]));for(var b=0,x=!1,w=(Wl("fill"in t?t.fill:(x=!0,c.fill))),M=(zl("stroke"in t?t.stroke:a||c.autoStroke&&!x?null:(b=Ol,c.stroke))),S=t.textShadowBlur>0,L=null!=t.width&&("truncate"===t.overflow||"break"===t.overflow||"breakAll"===t.overflow),k=r.calculatedLineHeight,T=0;T<l.length;T++){var D=this._getOrCreateChild(yl),C=D.createStyle();D.useStyle(C),C.text=l[T],C.x=m,C.y=g,f&&(C.textAlign=f),C.textBaseline="middle",C.opacity=t.opacity,C.strokeFirst=!0,S&&(C.shadowBlur=t.textShadowBlur||0,C.shadowColor=t.textShadowColor||"transparent",C.shadowOffsetX=t.textShadowOffsetX||0,C.shadowOffsetY=t.textShadowOffsetY||0),M&&(C.stroke=M,C.lineWidth=t.lineWidth||b,C.lineDash=t.lineDash,C.lineDashOffset=t.lineDashOffset||0),w&&(C.fill=w),C.font=e,g+=u,L&&D.setBoundingRect(new Ji(or(C.x,t.width,C.textAlign),ar(C.y,k,C.textBaseline),t.width,k))}},e.prototype._updateRichTexts=function(){var t=this.style,e=Fl(t),n=ba(e,t),i=n.width,r=n.outerWidth,o=n.outerHeight,a=t.padding,s=t.x||0,l=t.y||0,u=this._defaultStyle,c=t.align||u.align,d=t.verticalAlign||u.verticalAlign,h=or(s,r,c),f=ar(l,o,d),p=h,m=f;a&&(p+=a[3],m+=a[0]);var g=p+i;Vl(t)&&this._renderBackground(t,t,h,f,r,o);for(var y=!!t.backgroundColor,v=0;v<n.lines.length;v++){var _=n.lines[v],b=_.tokens,x=b.length,w=_.lineHeight,M=_.width,S=0,L=p,k=g,T=x-1,D=void 0;while(S<x&&(D=b[S],!D.align||"left"===D.align))this._placeToken(D,t,w,m,L,"left",y),M-=D.width,L+=D.width,S++;while(T>=0&&(D=b[T],"right"===D.align))this._placeToken(D,t,w,m,k,"right",y),M-=D.width,k-=D.width,T--;L+=(i-(L-p)-(g-k)-M)/2;while(S<=T)D=b[S],this._placeToken(D,t,w,m,L+D.width/2,"center",y),L+=D.width,S++;m+=w}},e.prototype._placeToken=function(t,e,n,i,r,o,a){var s=e.rich[t.styleName]||{};s.text=t.text;var l=t.verticalAlign,u=i+n/2;"top"===l?u=i+t.height/2:"bottom"===l&&(u=i+n-t.height/2);var c=!t.isLineHolder&&Vl(s);c&&this._renderBackground(s,e,"right"===o?r-t.width:"center"===o?r-t.width/2:r,u-t.height/2,t.width,t.height);var d=!!s.backgroundColor,h=t.textPadding;h&&(r=Bl(r,o,h),u-=t.height/2-h[0]-t.innerHeight/2);var f=this._getOrCreateChild(yl),p=f.createStyle();f.useStyle(p);var m=this._defaultStyle,g=!1,y=0,v=zl("fill"in s?s.fill:"fill"in e?e.fill:(g=!0,m.fill)),_=zl("stroke"in s?s.stroke:"stroke"in e?e.stroke:d||a||m.autoStroke&&!g?null:(y=Ol,m.stroke)),b=s.textShadowBlur>0||e.textShadowBlur>0;p.text=t.text,p.x=r,p.y=u,b&&(p.shadowBlur=s.textShadowBlur||e.textShadowBlur||0,p.shadowColor=s.textShadowColor||e.textShadowColor||"transparent",p.shadowOffsetX=s.textShadowOffsetX||e.textShadowOffsetX||0,p.shadowOffsetY=s.textShadowOffsetY||e.textShadowOffsetY||0),p.textAlign=o,p.textBaseline="middle",p.font=t.font||Qi,p.opacity=st(s.opacity,e.opacity,1),_&&(p.lineWidth=st(s.lineWidth,e.lineWidth,y),p.lineDash=at(s.lineDash,e.lineDash),p.lineDashOffset=e.lineDashOffset||0,p.stroke=_),v&&(p.fill=v);var x=t.contentWidth,w=t.contentHeight;f.setBoundingRect(new Ji(or(p.x,x,p.textAlign),ar(p.y,w,p.textBaseline),x,w))},e.prototype._renderBackground=function(t,e,n,i,r,o){var a,s,l=t.backgroundColor,u=t.borderWidth,c=t.borderColor,d=l&&l.image,h=l&&!d,f=t.borderRadius,p=this;if(h||u&&c){a=this._getOrCreateChild(Al),a.useStyle(a.createStyle()),a.style.fill=null;var m=a.shape;m.x=n,m.y=i,m.width=r,m.height=o,m.r=f,a.dirtyShape()}if(h){var g=a.style;g.fill=l||null,g.fillOpacity=at(t.fillOpacity,1)}else if(d){s=this._getOrCreateChild(wl),s.onload=function(){p.dirtyStyle()};var y=s.style;y.image=l.image,y.x=n,y.y=i,y.width=r,y.height=o}if(u&&c){g=a.style;g.lineWidth=u,g.stroke=c,g.strokeOpacity=at(t.strokeOpacity,1),g.lineDash=t.borderDash,g.lineDashOffset=t.borderDashOffset||0,a.strokeContainThreshold=0,a.hasFill()&&a.hasStroke()&&(g.strokeFirst=!0,g.lineWidth*=2)}var v=(a||s).style;v.shadowBlur=t.shadowBlur||0,v.shadowColor=t.shadowColor||"transparent",v.shadowOffsetX=t.shadowOffsetX||0,v.shadowOffsetY=t.shadowOffsetY||0,v.opacity=st(t.opacity,e.opacity,1)},e.makeFont=function(t){var e="";if(t.fontSize||t.fontFamily||t.fontWeight){var n="";n="string"!==typeof t.fontSize||-1===t.fontSize.indexOf("px")&&-1===t.fontSize.indexOf("rem")&&-1===t.fontSize.indexOf("em")?isNaN(+t.fontSize)?"12px":t.fontSize+"px":t.fontSize,e=[t.fontStyle,t.fontWeight,n,t.fontFamily||"sans-serif"].join(" ")}return e&&dt(e)||t.textFont||t.font},e}(Ea),Nl={left:!0,right:1,center:1},Rl={top:1,bottom:1,middle:1};function Hl(t){return jl(t),R(t.rich,jl),t}function jl(t){if(t){t.font=El.makeFont(t);var e=t.align;"middle"===e&&(e="center"),t.align=null==e||Nl[e]?e:"left";var n=t.verticalAlign;"center"===n&&(n="middle"),t.verticalAlign=null==n||Rl[n]?n:"top";var i=t.padding;i&&(t.padding=ut(t.padding))}}function zl(t,e){return null==t||e<=0||"transparent"===t||"none"===t?null:t.image||t.colorStops?"#000":t}function Wl(t){return null==t||"none"===t?null:t.image||t.colorStops?"#000":t}function Bl(t,e,n){return"right"===e?t-n[1]:"center"===e?t+n[3]/2-n[1]/2:t+n[3]}function Fl(t){var e=t.text;return null!=e&&(e+=""),e}function Vl(t){return!!(t.backgroundColor||t.borderWidth&&t.borderColor)}var Gl=El,$l=Io(),Ul=1,ql={},Xl=Io(),Zl=0,Jl=1,Kl=2,Ql=["emphasis","blur","select"],tu=["normal","emphasis","blur","select"],eu=10,nu=9,iu="highlight",ru="downplay",ou="select",au="unselect",su="toggleSelect";function lu(t){return null!=t&&"none"!==t}var uu=new nn(100);function cu(t){if("string"!==typeof t)return t;var e=uu.get(t);return e||(e=bn(t,-.1),uu.put(t,e)),e}function du(t,e,n){t.onHoverStateChange&&(t.hoverState||0)!==n&&t.onHoverStateChange(e),t.hoverState=n}function hu(t){du(t,"emphasis",Kl)}function fu(t){t.hoverState===Kl&&du(t,"normal",Zl)}function pu(t){du(t,"blur",Jl)}function mu(t){t.hoverState===Jl&&du(t,"normal",Zl)}function gu(t){t.selected=!0}function yu(t){t.selected=!1}function vu(t,e,n){e(t,n)}function _u(t,e,n){vu(t,e,n),t.isGroup&&t.traverse((function(t){vu(t,e,n)}))}function bu(t,e){switch(e){case"emphasis":t.hoverState=Kl;break;case"normal":t.hoverState=Zl;break;case"blur":t.hoverState=Jl;break;case"select":t.selected=!0}}function xu(t,e,n,i){for(var r=t.style,o={},a=0;a<e.length;a++){var s=e[a],l=r[s];o[s]=null==l?i&&i[s]:l}for(a=0;a<t.animators.length;a++){var u=t.animators[a];u.__fromStateTransition&&u.__fromStateTransition.indexOf(n)<0&&"style"===u.targetName&&u.saveFinalToTarget(o,e)}return o}function wu(t,e,n,i){var r=n&&O(n,"select")>=0,o=!1;if(t instanceof pl){var a=Xl(t),s=r&&a.selectFill||a.normalFill,l=r&&a.selectStroke||a.normalStroke;if(lu(s)||lu(l)){i=i||{};var u=i.style||{};!lu(u.fill)&&lu(s)?(o=!0,i=I({},i),u=I({},u),u.fill=cu(s)):!lu(u.stroke)&&lu(l)&&(o||(i=I({},i),u=I({},u)),u.stroke=cu(l)),i.style=u}}if(i&&null==i.z2){o||(i=I({},i));var c=t.z2EmphasisLift;i.z2=t.z2+(null!=c?c:eu)}return i}function Mu(t,e,n){if(n&&null==n.z2){n=I({},n);var i=t.z2SelectLift;n.z2=t.z2+(null!=i?i:nu)}return n}function Su(t,e,n){var i=O(t.currentStates,e)>=0,r=t.style.opacity,o=i?null:xu(t,["opacity"],e,{opacity:1});n=n||{};var a=n.style||{};return null==a.opacity&&(n=I({},n),a=I({opacity:i?r:.1*o.opacity},a),n.style=a),n}function Lu(t,e){var n=this.states[t];if(this.style){if("emphasis"===t)return wu(this,t,e,n);if("blur"===t)return Su(this,t,n);if("select"===t)return Mu(this,t,n)}return n}function ku(t){t.stateProxy=Lu;var e=t.getTextContent(),n=t.getTextGuideLine();e&&(e.stateProxy=Lu),n&&(n.stateProxy=Lu)}function Tu(t,e){!Eu(t,e)&&!t.__highByOuter&&_u(t,hu)}function Du(t,e){!Eu(t,e)&&!t.__highByOuter&&_u(t,fu)}function Cu(t,e){t.__highByOuter|=1<<(e||0),_u(t,hu)}function Iu(t,e){!(t.__highByOuter&=~(1<<(e||0)))&&_u(t,fu)}function Au(t){_u(t,pu)}function Yu(t){_u(t,mu)}function Ou(t){_u(t,gu)}function Pu(t){_u(t,yu)}function Eu(t,e){return t.__highDownSilentOnTouch&&e.zrByTouch}function Nu(t){var e=t.getModel();e.eachComponent((function(e,n){var i="series"===e?t.getViewOfSeriesModel(n):t.getViewOfComponentModel(n);i.group.traverse((function(t){mu(t)}))}))}function Ru(t,e,n,i){var r=i.getModel();function o(t,e){for(var n=0;n<e.length;n++){var i=t.getItemGraphicEl(e[n]);i&&Yu(i)}}if(n=n||"coordinateSystem",null!=t&&e&&"none"!==e){var a=r.getSeriesByIndex(t),s=a.coordinateSystem;s&&s.master&&(s=s.master);var l=[];r.eachSeries((function(t){var r=a===t,u=t.coordinateSystem;u&&u.master&&(u=u.master);var c=u&&s?u===s:r;if(!("series"===n&&!r||"coordinateSystem"===n&&!c||"series"===e&&r)){var d=i.getViewOfSeriesModel(t);if(d.group.traverse((function(t){pu(t)})),N(e))o(t.getData(),e);else if(J(e))for(var h=B(e),f=0;f<h.length;f++)o(t.getData(h[f]),e[h[f]]);l.push(t)}})),r.eachComponent((function(t,e){if("series"!==t){var n=i.getViewOfComponentModel(e);n&&n.blurSeries&&n.blurSeries(l,r)}}))}}function Hu(t,e,n){if(null!=t&&null!=e){var i=n.getModel().getComponent(t,e);if(i){var r=n.getViewOfComponentModel(i);r&&r.focusBlurEnabled&&r.group.traverse((function(t){pu(t)}))}}}function ju(t,e,n){var i=t.seriesIndex,r=t.getData(e.dataType),o=Co(r,e);o=($(o)?o[0]:o)||0;var a=r.getItemGraphicEl(o);if(!a){var s=r.count(),l=0;while(!a&&l<s)a=r.getItemGraphicEl(l++)}if(a){var u=$l(a);Ru(i,u.focus,u.blurScope,n)}else{var c=t.get(["emphasis","focus"]),d=t.get(["emphasis","blurScope"]);null!=c&&Ru(i,c,d,n)}}function zu(t,e,n,i){var r={focusSelf:!1,dispatchers:null};if(null==t||"series"===t||null==e||null==n)return r;var o=i.getModel().getComponent(t,e);if(!o)return r;var a=i.getViewOfComponentModel(o);if(!a||!a.findHighDownDispatchers)return r;for(var s,l=a.findHighDownDispatchers(n),u=0;u<l.length;u++)if("self"===$l(l[u]).focus){s=!0;break}return{focusSelf:s,dispatchers:l}}function Wu(t,e,n){var i=$l(t),r=zu(i.componentMainType,i.componentIndex,i.componentHighDownName,n),o=r.dispatchers,a=r.focusSelf;o?(a&&Hu(i.componentMainType,i.componentIndex,n),R(o,(function(t){return Tu(t,e)}))):(Ru(i.seriesIndex,i.focus,i.blurScope,n),"self"===i.focus&&Hu(i.componentMainType,i.componentIndex,n),Tu(t,e))}function Bu(t,e,n){Nu(n);var i=$l(t),r=zu(i.componentMainType,i.componentIndex,i.componentHighDownName,n).dispatchers;r?R(r,(function(t){return Du(t,e)})):Du(t,e)}function Fu(t,e,n){if(ec(e)){var i=e.dataType,r=t.getData(i),o=Co(r,e);$(o)||(o=[o]),t[e.type===su?"toggleSelect":e.type===ou?"select":"unselect"](o,i)}}function Vu(t){var e=t.getAllData();R(e,(function(e){var n=e.data,i=e.type;n.eachItemGraphicEl((function(e,n){t.isSelected(n,i)?Ou(e):Pu(e)}))}))}function Gu(t){var e=[];return t.eachSeries((function(t){var n=t.getAllData();R(n,(function(n){n.data;var i=n.type,r=t.getSelectedDataIndices();if(r.length>0){var o={dataIndex:r,seriesIndex:t.seriesIndex};null!=i&&(o.dataType=i),e.push(o)}}))})),e}function $u(t,e,n){Ju(t,!0),_u(t,ku),Uu(t,e,n)}function Uu(t,e,n){var i=$l(t);null!=e?(i.focus=e,i.blurScope=n):i.focus&&(i.focus=null)}var qu=["emphasis","blur","select"],Xu={itemStyle:"getItemStyle",lineStyle:"getLineStyle",areaStyle:"getAreaStyle"};function Zu(t,e,n,i){n=n||"itemStyle";for(var r=0;r<qu.length;r++){var o=qu[r],a=e.getModel([o,n]),s=t.ensureState(o);s.style=i?i(a):a[Xu[n]]()}}function Ju(t,e){var n=!1===e,i=t;t.highDownSilentOnTouch&&(i.__highDownSilentOnTouch=t.highDownSilentOnTouch),n&&!i.__highDownDispatcher||(i.__highByOuter=i.__highByOuter||0,i.__highDownDispatcher=!n)}function Ku(t){return!(!t||!t.__highDownDispatcher)}function Qu(t,e,n){var i=$l(t);i.componentMainType=e.mainType,i.componentIndex=e.componentIndex,i.componentHighDownName=n}function tc(t){var e=ql[t];return null==e&&Ul<=32&&(e=ql[t]=Ul++),e}function ec(t){var e=t.type;return e===ou||e===au||e===su}function nc(t){var e=t.type;return e===iu||e===ru}function ic(t){var e=Xl(t);e.normalFill=t.style.fill,e.normalStroke=t.style.stroke;var n=t.states.select||{};e.selectFill=n.style&&n.style.fill||null,e.selectStroke=n.style&&n.style.stroke||null}var rc=Bs.CMD,oc=[[],[],[]],ac=Math.sqrt,sc=Math.atan2;function lc(t,e){var n,i,r,o,a,s,l=t.data,u=t.len(),c=rc.M,d=rc.C,h=rc.L,f=rc.R,p=rc.A,m=rc.Q;for(r=0,o=0;r<u;){switch(n=l[r++],o=r,i=0,n){case c:i=1;break;case h:i=1;break;case d:i=3;break;case m:i=2;break;case p:var g=e[4],y=e[5],v=ac(e[0]*e[0]+e[1]*e[1]),_=ac(e[2]*e[2]+e[3]*e[3]),b=sc(-e[1]/_,e[0]/v);l[r]*=v,l[r++]+=g,l[r]*=_,l[r++]+=y,l[r++]*=v,l[r++]*=_,l[r++]+=b,l[r++]+=b,r+=2,o=r;break;case f:s[0]=l[r++],s[1]=l[r++],jt(s,s,e),l[o++]=s[0],l[o++]=s[1],s[0]+=l[r++],s[1]+=l[r++],jt(s,s,e),l[o++]=s[0],l[o++]=s[1]}for(a=0;a<i;a++){var x=oc[a];x[0]=l[r++],x[1]=l[r++],jt(x,x,e),l[o++]=x[0],l[o++]=x[1]}}t.increaseVersion()}var uc=Math.sqrt,cc=Math.sin,dc=Math.cos,hc=Math.PI;function fc(t){return Math.sqrt(t[0]*t[0]+t[1]*t[1])}function pc(t,e){return(t[0]*e[0]+t[1]*e[1])/(fc(t)*fc(e))}function mc(t,e){return(t[0]*e[1]<t[1]*e[0]?-1:1)*Math.acos(pc(t,e))}function gc(t,e,n,i,r,o,a,s,l,u,c){var d=l*(hc/180),h=dc(d)*(t-n)/2+cc(d)*(e-i)/2,f=-1*cc(d)*(t-n)/2+dc(d)*(e-i)/2,p=h*h/(a*a)+f*f/(s*s);p>1&&(a*=uc(p),s*=uc(p));var m=(r===o?-1:1)*uc((a*a*(s*s)-a*a*(f*f)-s*s*(h*h))/(a*a*(f*f)+s*s*(h*h)))||0,g=m*a*f/s,y=m*-s*h/a,v=(t+n)/2+dc(d)*g-cc(d)*y,_=(e+i)/2+cc(d)*g+dc(d)*y,b=mc([1,0],[(h-g)/a,(f-y)/s]),x=[(h-g)/a,(f-y)/s],w=[(-1*h-g)/a,(-1*f-y)/s],M=mc(x,w);if(pc(x,w)<=-1&&(M=hc),pc(x,w)>=1&&(M=0),M<0){var S=Math.round(M/hc*1e6)/1e6;M=2*hc+S%2*hc}c.addData(u,v,_,a,s,b,M,d,o)}var yc=/([mlvhzcqtsa])([^mlvhzcqtsa]*)/gi,vc=/-?([0-9]*\.)?[0-9]+([eE]-?[0-9]+)?/g;function _c(t){var e=new Bs;if(!t)return e;var n,i=0,r=0,o=i,a=r,s=Bs.CMD,l=t.match(yc);if(!l)return e;for(var u=0;u<l.length;u++){for(var c=l[u],d=c.charAt(0),h=void 0,f=c.match(vc)||[],p=f.length,m=0;m<p;m++)f[m]=parseFloat(f[m]);var g=0;while(g<p){var y=void 0,v=void 0,_=void 0,b=void 0,x=void 0,w=void 0,M=void 0,S=i,L=r,k=void 0,T=void 0;switch(d){case"l":i+=f[g++],r+=f[g++],h=s.L,e.addData(h,i,r);break;case"L":i=f[g++],r=f[g++],h=s.L,e.addData(h,i,r);break;case"m":i+=f[g++],r+=f[g++],h=s.M,e.addData(h,i,r),o=i,a=r,d="l";break;case"M":i=f[g++],r=f[g++],h=s.M,e.addData(h,i,r),o=i,a=r,d="L";break;case"h":i+=f[g++],h=s.L,e.addData(h,i,r);break;case"H":i=f[g++],h=s.L,e.addData(h,i,r);break;case"v":r+=f[g++],h=s.L,e.addData(h,i,r);break;case"V":r=f[g++],h=s.L,e.addData(h,i,r);break;case"C":h=s.C,e.addData(h,f[g++],f[g++],f[g++],f[g++],f[g++],f[g++]),i=f[g-2],r=f[g-1];break;case"c":h=s.C,e.addData(h,f[g++]+i,f[g++]+r,f[g++]+i,f[g++]+r,f[g++]+i,f[g++]+r),i+=f[g-2],r+=f[g-1];break;case"S":y=i,v=r,k=e.len(),T=e.data,n===s.C&&(y+=i-T[k-4],v+=r-T[k-3]),h=s.C,S=f[g++],L=f[g++],i=f[g++],r=f[g++],e.addData(h,y,v,S,L,i,r);break;case"s":y=i,v=r,k=e.len(),T=e.data,n===s.C&&(y+=i-T[k-4],v+=r-T[k-3]),h=s.C,S=i+f[g++],L=r+f[g++],i+=f[g++],r+=f[g++],e.addData(h,y,v,S,L,i,r);break;case"Q":S=f[g++],L=f[g++],i=f[g++],r=f[g++],h=s.Q,e.addData(h,S,L,i,r);break;case"q":S=f[g++]+i,L=f[g++]+r,i+=f[g++],r+=f[g++],h=s.Q,e.addData(h,S,L,i,r);break;case"T":y=i,v=r,k=e.len(),T=e.data,n===s.Q&&(y+=i-T[k-4],v+=r-T[k-3]),i=f[g++],r=f[g++],h=s.Q,e.addData(h,y,v,i,r);break;case"t":y=i,v=r,k=e.len(),T=e.data,n===s.Q&&(y+=i-T[k-4],v+=r-T[k-3]),i+=f[g++],r+=f[g++],h=s.Q,e.addData(h,y,v,i,r);break;case"A":_=f[g++],b=f[g++],x=f[g++],w=f[g++],M=f[g++],S=i,L=r,i=f[g++],r=f[g++],h=s.A,gc(S,L,i,r,w,M,_,b,x,h,e);break;case"a":_=f[g++],b=f[g++],x=f[g++],w=f[g++],M=f[g++],S=i,L=r,i+=f[g++],r+=f[g++],h=s.A,gc(S,L,i,r,w,M,_,b,x,h,e);break}}"z"!==d&&"Z"!==d||(h=s.Z,e.addData(h),i=o,r=a),n=h}return e.toStatic(),e}var bc=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return wt(e,t),e.prototype.applyTransform=function(t){},e}(pl);function xc(t){return null!=t.setData}function wc(t,e){var n=_c(t),i=I({},e);return i.buildPath=function(t){if(xc(t)){t.setData(n.data);var e=t.getContext();e&&t.rebuildPath(e,1)}else{e=t;n.rebuildPath(e,1)}},i.applyTransform=function(t){lc(n,t),this.dirtyShape()},i}function Mc(t,e){return new bc(wc(t,e))}function Sc(t,e){var n=wc(t,e),i=function(t){function e(e){var i=t.call(this,e)||this;return i.applyTransform=n.applyTransform,i.buildPath=n.buildPath,i}return wt(e,t),e}(bc);return i}function Lc(t,e){for(var n=[],i=t.length,r=0;r<i;r++){var o=t[r];o.path||o.createPathProxy(),o.shapeChanged()&&o.buildPath(o.path,o.shape,!0),n.push(o.path)}var a=new pl(e);return a.createPathProxy(),a.buildPath=function(t){if(xc(t)){t.appendPath(n);var e=t.getContext();e&&t.rebuildPath(e,1)}},a}var kc=function(){function t(){this.cx=0,this.cy=0,this.r=0}return t}(),Tc=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultShape=function(){return new kc},e.prototype.buildPath=function(t,e,n){n&&t.moveTo(e.cx+e.r,e.cy),t.arc(e.cx,e.cy,e.r,0,2*Math.PI)},e}(pl);Tc.prototype.type="circle";var Dc=Tc,Cc=function(){function t(){this.cx=0,this.cy=0,this.rx=0,this.ry=0}return t}(),Ic=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultShape=function(){return new Cc},e.prototype.buildPath=function(t,e){var n=.5522848,i=e.cx,r=e.cy,o=e.rx,a=e.ry,s=o*n,l=a*n;t.moveTo(i-o,r),t.bezierCurveTo(i-o,r-l,i-s,r-a,i,r-a),t.bezierCurveTo(i+s,r-a,i+o,r-l,i+o,r),t.bezierCurveTo(i+o,r+l,i+s,r+a,i,r+a),t.bezierCurveTo(i-s,r+a,i-o,r+l,i-o,r),t.closePath()},e}(pl);Ic.prototype.type="ellipse";var Ac=Ic,Yc=Math.PI,Oc=2*Yc,Pc=Math.sin,Ec=Math.cos,Nc=Math.acos,Rc=Math.atan2,Hc=Math.abs,jc=Math.sqrt,zc=Math.max,Wc=Math.min,Bc=1e-4;function Fc(t,e,n,i,r,o,a,s){var l=n-t,u=i-e,c=a-r,d=s-o,h=d*l-c*u;if(!(h*h<Bc))return h=(c*(e-o)-d*(t-r))/h,[t+h*l,e+h*u]}function Vc(t,e,n,i,r,o,a){var s=t-n,l=e-i,u=(a?o:-o)/jc(s*s+l*l),c=u*l,d=-u*s,h=t+c,f=e+d,p=n+c,m=i+d,g=(h+p)/2,y=(f+m)/2,v=p-h,_=m-f,b=v*v+_*_,x=r-o,w=h*m-p*f,M=(_<0?-1:1)*jc(zc(0,x*x*b-w*w)),S=(w*_-v*M)/b,L=(-w*v-_*M)/b,k=(w*_+v*M)/b,T=(-w*v+_*M)/b,D=S-g,C=L-y,I=k-g,A=T-y;return D*D+C*C>I*I+A*A&&(S=k,L=T),{cx:S,cy:L,x01:-c,y01:-d,x11:S*(r/x-1),y11:L*(r/x-1)}}function Gc(t,e){var n=zc(e.r,0),i=zc(e.r0||0,0),r=n>0,o=i>0;if(r||o){if(r||(n=i,i=0),i>n){var a=n;n=i,i=a}var s,l=!!e.clockwise,u=e.startAngle,c=e.endAngle;if(u===c)s=0;else{var d=[u,c];zs(d,!l),s=Hc(d[0]-d[1])}var h=e.cx,f=e.cy,p=e.cornerRadius||0,m=e.innerCornerRadius||0;if(n>Bc)if(s>Oc-Bc)t.moveTo(h+n*Ec(u),f+n*Pc(u)),t.arc(h,f,n,u,c,!l),i>Bc&&(t.moveTo(h+i*Ec(c),f+i*Pc(c)),t.arc(h,f,i,c,u,l));else{var g=Hc(n-i)/2,y=Wc(g,p),v=Wc(g,m),_=v,b=y,x=n*Ec(u),w=n*Pc(u),M=i*Ec(c),S=i*Pc(c),L=void 0,k=void 0,T=void 0,D=void 0;if((y>Bc||v>Bc)&&(L=n*Ec(c),k=n*Pc(c),T=i*Ec(u),D=i*Pc(u),s<Yc)){var C=Fc(x,w,T,D,L,k,M,S);if(C){var I=x-C[0],A=w-C[1],Y=L-C[0],O=k-C[1],P=1/Pc(Nc((I*Y+A*O)/(jc(I*I+A*A)*jc(Y*Y+O*O)))/2),E=jc(C[0]*C[0]+C[1]*C[1]);_=Wc(v,(i-E)/(P-1)),b=Wc(y,(n-E)/(P+1))}}if(s>Bc)if(b>Bc){var N=Vc(T,D,x,w,n,b,l),R=Vc(L,k,M,S,n,b,l);t.moveTo(h+N.cx+N.x01,f+N.cy+N.y01),b<y?t.arc(h+N.cx,f+N.cy,b,Rc(N.y01,N.x01),Rc(R.y01,R.x01),!l):(t.arc(h+N.cx,f+N.cy,b,Rc(N.y01,N.x01),Rc(N.y11,N.x11),!l),t.arc(h,f,n,Rc(N.cy+N.y11,N.cx+N.x11),Rc(R.cy+R.y11,R.cx+R.x11),!l),t.arc(h+R.cx,f+R.cy,b,Rc(R.y11,R.x11),Rc(R.y01,R.x01),!l))}else t.moveTo(h+x,f+w),t.arc(h,f,n,u,c,!l);else t.moveTo(h+x,f+w);if(i>Bc&&s>Bc)if(_>Bc){N=Vc(M,S,L,k,i,-_,l),R=Vc(x,w,T,D,i,-_,l);t.lineTo(h+N.cx+N.x01,f+N.cy+N.y01),_<v?t.arc(h+N.cx,f+N.cy,_,Rc(N.y01,N.x01),Rc(R.y01,R.x01),!l):(t.arc(h+N.cx,f+N.cy,_,Rc(N.y01,N.x01),Rc(N.y11,N.x11),!l),t.arc(h,f,i,Rc(N.cy+N.y11,N.cx+N.x11),Rc(R.cy+R.y11,R.cx+R.x11),l),t.arc(h+R.cx,f+R.cy,_,Rc(R.y11,R.x11),Rc(R.y01,R.x01),!l))}else t.lineTo(h+M,f+S),t.arc(h,f,i,c,u,l);else t.lineTo(h+M,f+S)}else t.moveTo(h,f);t.closePath()}}var $c=function(){function t(){this.cx=0,this.cy=0,this.r0=0,this.r=0,this.startAngle=0,this.endAngle=2*Math.PI,this.clockwise=!0,this.cornerRadius=0,this.innerCornerRadius=0}return t}(),Uc=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultShape=function(){return new $c},e.prototype.buildPath=function(t,e){Gc(t,e)},e.prototype.isZeroArea=function(){return this.shape.startAngle===this.shape.endAngle||this.shape.r===this.shape.r0},e}(pl);Uc.prototype.type="sector";var qc=Uc,Xc=function(){function t(){this.cx=0,this.cy=0,this.r=0,this.r0=0}return t}(),Zc=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultShape=function(){return new Xc},e.prototype.buildPath=function(t,e){var n=e.cx,i=e.cy,r=2*Math.PI;t.moveTo(n+e.r,i),t.arc(n,i,e.r,0,r,!1),t.moveTo(n+e.r0,i),t.arc(n,i,e.r0,0,r,!0)},e}(pl);Zc.prototype.type="ring";var Jc=Zc;function Kc(t,e,n,i,r,o,a){var s=.5*(n-t),l=.5*(i-e);return(2*(e-n)+s+l)*a+(-3*(e-n)-2*s-l)*o+s*r+e}function Qc(t,e){for(var n=t.length,i=[],r=0,o=1;o<n;o++)r+=Pt(t[o-1],t[o]);var a=r/2;a=a<n?n:a;for(o=0;o<a;o++){var s=o/(a-1)*(e?n:n-1),l=Math.floor(s),u=s-l,c=void 0,d=t[l%n],h=void 0,f=void 0;e?(c=t[(l-1+n)%n],h=t[(l+1)%n],f=t[(l+2)%n]):(c=t[0===l?l:l-1],h=t[l>n-2?n-1:l+1],f=t[l>n-3?n-1:l+2]);var p=u*u,m=u*p;i.push([Kc(c[0],d[0],h[0],f[0],u,p,m),Kc(c[1],d[1],h[1],f[1],u,p,m)])}return i}function td(t,e,n,i){var r,o,a,s,l=[],u=[],c=[],d=[];if(i){a=[1/0,1/0],s=[-1/0,-1/0];for(var h=0,f=t.length;h<f;h++)zt(a,a,t[h]),Wt(s,s,t[h]);zt(a,a,i[0]),Wt(s,s,i[1])}for(h=0,f=t.length;h<f;h++){var p=t[h];if(n)r=t[h?h-1:f-1],o=t[(h+1)%f];else{if(0===h||h===f-1){l.push(Lt(t[h]));continue}r=t[h-1],o=t[h+1]}Ct(u,o,r),Yt(u,u,e);var m=Pt(p,r),g=Pt(p,o),y=m+g;0!==y&&(m/=y,g/=y),Yt(c,u,-m),Yt(d,u,g);var v=Tt([],p,c),_=Tt([],p,d);i&&(Wt(v,v,a),zt(v,v,s),Wt(_,_,a),zt(_,_,s)),l.push(v),l.push(_)}return n&&l.push(l.shift()),l}function ed(t,e,n){var i=e.smooth,r=e.points;if(r&&r.length>=2){if(i&&"spline"!==i){var o=td(r,i,n,e.smoothConstraint);t.moveTo(r[0][0],r[0][1]);for(var a=r.length,s=0;s<(n?a:a-1);s++){var l=o[2*s],u=o[2*s+1],c=r[(s+1)%a];t.bezierCurveTo(l[0],l[1],u[0],u[1],c[0],c[1])}}else{"spline"===i&&(r=Qc(r,n)),t.moveTo(r[0][0],r[0][1]);s=1;for(var d=r.length;s<d;s++)t.lineTo(r[s][0],r[s][1])}n&&t.closePath()}}var nd=function(){function t(){this.points=null,this.smooth=0,this.smoothConstraint=null}return t}(),id=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultShape=function(){return new nd},e.prototype.buildPath=function(t,e){ed(t,e,!0)},e}(pl);id.prototype.type="polygon";var rd=id,od=function(){function t(){this.points=null,this.percent=1,this.smooth=0,this.smoothConstraint=null}return t}(),ad=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new od},e.prototype.buildPath=function(t,e){ed(t,e,!1)},e}(pl);ad.prototype.type="polyline";var sd=ad,ld={},ud=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.percent=1}return t}(),cd=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new ud},e.prototype.buildPath=function(t,e){var n,i,r,o;if(this.subPixelOptimize){var a=Ll(ld,e,this.style);n=a.x1,i=a.y1,r=a.x2,o=a.y2}else n=e.x1,i=e.y1,r=e.x2,o=e.y2;var s=e.percent;0!==s&&(t.moveTo(n,i),s<1&&(r=n*(1-s)+r*s,o=i*(1-s)+o*s),t.lineTo(r,o))},e.prototype.pointAt=function(t){var e=this.shape;return[e.x1*(1-t)+e.x2*t,e.y1*(1-t)+e.y2*t]},e}(pl);cd.prototype.type="line";var dd=cd,hd=[],fd=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.cpx1=0,this.cpy1=0,this.percent=1}return t}();function pd(t,e,n){var i=t.cpx2,r=t.cpy2;return null===i||null===r?[(n?qa:Ua)(t.x1,t.cpx1,t.cpx2,t.x2,e),(n?qa:Ua)(t.y1,t.cpy1,t.cpy2,t.y2,e)]:[(n?es:ts)(t.x1,t.cpx1,t.x2,e),(n?es:ts)(t.y1,t.cpy1,t.y2,e)]}var md=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new fd},e.prototype.buildPath=function(t,e){var n=e.x1,i=e.y1,r=e.x2,o=e.y2,a=e.cpx1,s=e.cpy1,l=e.cpx2,u=e.cpy2,c=e.percent;0!==c&&(t.moveTo(n,i),null==l||null==u?(c<1&&(rs(n,a,r,c,hd),a=hd[1],r=hd[2],rs(i,s,o,c,hd),s=hd[1],o=hd[2]),t.quadraticCurveTo(a,s,r,o)):(c<1&&(Ja(n,a,l,r,c,hd),a=hd[1],l=hd[2],r=hd[3],Ja(i,s,u,o,c,hd),s=hd[1],u=hd[2],o=hd[3]),t.bezierCurveTo(a,s,l,u,r,o)))},e.prototype.pointAt=function(t){return pd(this.shape,t,!1)},e.prototype.tangentAt=function(t){var e=pd(this.shape,t,!0);return Ot(e,e)},e}(pl);md.prototype.type="bezier-curve";var gd=md,yd=function(){function t(){this.cx=0,this.cy=0,this.r=0,this.startAngle=0,this.endAngle=2*Math.PI,this.clockwise=!0}return t}(),vd=function(t){function e(e){return t.call(this,e)||this}return wt(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new yd},e.prototype.buildPath=function(t,e){var n=e.cx,i=e.cy,r=Math.max(e.r,0),o=e.startAngle,a=e.endAngle,s=e.clockwise,l=Math.cos(o),u=Math.sin(o);t.moveTo(l*r+n,u*r+i),t.arc(n,i,r,o,a,!s)},e}(pl);vd.prototype.type="arc";var _d=vd,bd=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="compound",e}return wt(e,t),e.prototype._updatePathDirty=function(){for(var t=this.shape.paths,e=this.shapeChanged(),n=0;n<t.length;n++)e=e||t[n].shapeChanged();e&&this.dirtyShape()},e.prototype.beforeBrush=function(){this._updatePathDirty();for(var t=this.shape.paths||[],e=this.getGlobalScale(),n=0;n<t.length;n++)t[n].path||t[n].createPathProxy(),t[n].path.setScale(e[0],e[1],t[n].segmentIgnoreThreshold)},e.prototype.buildPath=function(t,e){for(var n=e.paths||[],i=0;i<n.length;i++)n[i].buildPath(t,n[i].shape,!0)},e.prototype.afterBrush=function(){for(var t=this.shape.paths||[],e=0;e<t.length;e++)t[e].pathUpdated()},e.prototype.getBoundingRect=function(){return this._updatePathDirty.call(this),pl.prototype.getBoundingRect.call(this)},e}(pl),xd=bd,wd=function(){function t(t){this.colorStops=t||[]}return t.prototype.addColorStop=function(t,e){this.colorStops.push({offset:t,color:e})},t}(),Md=wd,Sd=function(t){function e(e,n,i,r,o,a){var s=t.call(this,o)||this;return s.x=null==e?0:e,s.y=null==n?0:n,s.x2=null==i?1:i,s.y2=null==r?0:r,s.type="linear",s.global=a||!1,s}return wt(e,t),e}(Md),Ld=Sd,kd=function(t){function e(e,n,i,r,o){var a=t.call(this,r)||this;return a.x=null==e?.5:e,a.y=null==n?.5:n,a.r=null==i?.5:i,a.type="radial",a.global=o||!1,a}return wt(e,t),e}(Md),Td=kd,Dd=[0,0],Cd=[0,0],Id=new Wi,Ad=new Wi,Yd=function(){function t(t,e){this._corners=[],this._axes=[],this._origin=[0,0];for(var n=0;n<4;n++)this._corners[n]=new Wi;for(n=0;n<2;n++)this._axes[n]=new Wi;t&&this.fromBoundingRect(t,e)}return t.prototype.fromBoundingRect=function(t,e){var n=this._corners,i=this._axes,r=t.x,o=t.y,a=r+t.width,s=o+t.height;if(n[0].set(r,o),n[1].set(a,o),n[2].set(a,s),n[3].set(r,s),e)for(var l=0;l<4;l++)n[l].transform(e);Wi.sub(i[0],n[1],n[0]),Wi.sub(i[1],n[3],n[0]),i[0].normalize(),i[1].normalize();for(l=0;l<2;l++)this._origin[l]=i[l].dot(n[0])},t.prototype.intersect=function(t,e){var n=!0,i=!e;return Id.set(1/0,1/0),Ad.set(0,0),!this._intersectCheckOneSide(this,t,Id,Ad,i,1)&&(n=!1,i)||!this._intersectCheckOneSide(t,this,Id,Ad,i,-1)&&(n=!1,i)||i||Wi.copy(e,n?Id:Ad),n},t.prototype._intersectCheckOneSide=function(t,e,n,i,r,o){for(var a=!0,s=0;s<2;s++){var l=this._axes[s];if(this._getProjMinMaxOnAxis(s,t._corners,Dd),this._getProjMinMaxOnAxis(s,e._corners,Cd),Dd[1]<Cd[0]||Dd[0]>Cd[1]){if(a=!1,r)return a;var u=Math.abs(Cd[0]-Dd[1]),c=Math.abs(Dd[0]-Cd[1]);Math.min(u,c)>i.len()&&(u<c?Wi.scale(i,l,-u*o):Wi.scale(i,l,c*o))}else if(n){u=Math.abs(Cd[0]-Dd[1]),c=Math.abs(Dd[0]-Cd[1]);Math.min(u,c)<n.len()&&(u<c?Wi.scale(n,l,u*o):Wi.scale(n,l,-c*o))}}return a},t.prototype._getProjMinMaxOnAxis=function(t,e,n){for(var i=this._axes[t],r=this._origin,o=e[0].dot(i)+r[t],a=o,s=o,l=1;l<e.length;l++){var u=e[l].dot(i)+r[t];a=Math.min(u,a),s=Math.max(u,s)}n[0]=a,n[1]=s},t}(),Od=Yd,Pd=[],Ed=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.notClear=!0,e.incremental=!0,e._displayables=[],e._temporaryDisplayables=[],e._cursor=0,e}return wt(e,t),e.prototype.traverse=function(t,e){t.call(e,this)},e.prototype.useStyle=function(){this.style={}},e.prototype.getCursor=function(){return this._cursor},e.prototype.innerAfterBrush=function(){this._cursor=this._displayables.length},e.prototype.clearDisplaybles=function(){this._displayables=[],this._temporaryDisplayables=[],this._cursor=0,this.markRedraw(),this.notClear=!1},e.prototype.clearTemporalDisplayables=function(){this._temporaryDisplayables=[]},e.prototype.addDisplayable=function(t,e){e?this._temporaryDisplayables.push(t):this._displayables.push(t),this.markRedraw()},e.prototype.addDisplayables=function(t,e){e=e||!1;for(var n=0;n<t.length;n++)this.addDisplayable(t[n],e)},e.prototype.getDisplayables=function(){return this._displayables},e.prototype.getTemporalDisplayables=function(){return this._temporaryDisplayables},e.prototype.eachPendingDisplayable=function(t){for(var e=this._cursor;e<this._displayables.length;e++)t&&t(this._displayables[e]);for(e=0;e<this._temporaryDisplayables.length;e++)t&&t(this._temporaryDisplayables[e])},e.prototype.update=function(){this.updateTransform();for(var t=this._cursor;t<this._displayables.length;t++){var e=this._displayables[t];e.parent=this,e.update(),e.parent=null}for(t=0;t<this._temporaryDisplayables.length;t++){e=this._temporaryDisplayables[t];e.parent=this,e.update(),e.parent=null}},e.prototype.getBoundingRect=function(){if(!this._rect){for(var t=new Ji(1/0,1/0,-1/0,-1/0),e=0;e<this._displayables.length;e++){var n=this._displayables[e],i=n.getBoundingRect().clone();n.needLocalTransform()&&i.applyTransform(n.getLocalTransform(Pd)),t.union(i)}this._rect=t}return this._rect},e.prototype.contain=function(t,e){var n=this.transformCoordToLocal(t,e),i=this.getBoundingRect();if(i.contain(n[0],n[1]))for(var r=0;r<this._displayables.length;r++){var o=this._displayables[r];if(o.contain(t,e))return!0}return!1},e}(Ea),Nd=Ed,Rd=Math.max,Hd=Math.min,jd={};function zd(t){return pl.extend(t)}var Wd=Sc;function Bd(t,e){return Wd(t,e)}function Fd(t,e){jd[t]=e}function Vd(t){if(jd.hasOwnProperty(t))return jd[t]}function Gd(t,e,n,i){var r=Mc(t,e);return n&&("center"===i&&(n=Ud(n,r.getBoundingRect())),Xd(r,n)),r}function $d(t,e,n){var i=new wl({style:{image:t,x:e.x,y:e.y,width:e.width,height:e.height},onload:function(t){if("center"===n){var r={width:t.width,height:t.height};i.setStyle(Ud(e,r))}}});return i}function Ud(t,e){var n,i=e.width/e.height,r=t.height*i;r<=t.width?n=t.height:(r=t.width,n=r/i);var o=t.x+t.width/2,a=t.y+t.height/2;return{x:o-r/2,y:a-n/2,width:r,height:n}}var qd=Lc;function Xd(t,e){if(t.applyTransform){var n=t.getBoundingRect(),i=n.calculateTransform(e);t.applyTransform(i)}}function Zd(t){return Ll(t.shape,t.shape,t.style),t}function Jd(t){return kl(t.shape,t.shape,t.style),t}var Kd=Tl;function Qd(t,e,n,i,r,o,a){var s,l=!1;"function"===typeof r?(a=o,o=r,r=null):J(r)&&(o=r.cb,a=r.during,l=r.isFrom,s=r.removeOpt,r=r.dataIndex);var u,c="update"===t,d="remove"===t;if(i&&i.ecModel){var h=i.ecModel.getUpdatePayload();u=h&&h.animation}var f=i&&i.isAnimationEnabled();if(d||e.stopAnimation("remove"),f){var p=void 0,m=void 0,g=void 0;u?(p=u.duration||0,m=u.easing||"cubicOut",g=u.delay||0):d?(s=s||{},p=at(s.duration,200),m=at(s.easing,"cubicOut"),g=0):(p=i.getShallow(c?"animationDurationUpdate":"animationDuration"),m=i.getShallow(c?"animationEasingUpdate":"animationEasing"),g=i.getShallow(c?"animationDelayUpdate":"animationDelay")),"function"===typeof g&&(g=g(r,i.getAnimationDelayParams?i.getAnimationDelayParams(e,r):null)),"function"===typeof p&&(p=p(r)),p>0?l?e.animateFrom(n,{duration:p,delay:g||0,easing:m,done:o,force:!!o||!!a,scope:t,during:a}):e.animateTo(n,{duration:p,delay:g||0,easing:m,done:o,force:!!o||!!a,setToFinal:!0,scope:t,during:a}):(e.stopAnimation(),!l&&e.attr(n),o&&o())}else e.stopAnimation(),!l&&e.attr(n),a&&a(1),o&&o()}function th(t,e,n,i,r,o){Qd("update",t,e,n,i,r,o)}function eh(t,e,n,i,r,o){Qd("init",t,e,n,i,r,o)}function nh(t,e,n,i,r,o){oh(t)||Qd("remove",t,e,n,i,r,o)}function ih(t,e,n,i){t.removeTextContent(),t.removeTextGuideLine(),nh(t,{style:{opacity:0}},e,n,i)}function rh(t,e,n){function i(){t.parent&&t.parent.remove(t)}t.isGroup?t.traverse((function(t){t.isGroup||ih(t,e,n,i)})):ih(t,e,n,i)}function oh(t){if(!t.__zr)return!0;for(var e=0;e<t.animators.length;e++){var n=t.animators[e];if("remove"===n.scope)return!0}return!1}function ah(t,e){var n=xi([]);while(t&&t!==e)Mi(n,t.getLocalTransform(),n),t=t.parent;return n}function sh(t,e,n){return e&&!N(e)&&(e=ji.getLocalTransform(e)),n&&(e=Ti([],e)),jt([],t,e)}function lh(t,e,n){var i=0===e[4]||0===e[5]||0===e[0]?1:Math.abs(2*e[4]/e[0]),r=0===e[4]||0===e[5]||0===e[2]?1:Math.abs(2*e[4]/e[2]),o=["left"===t?-i:"right"===t?i:0,"top"===t?-r:"bottom"===t?r:0];return o=sh(o,e,n),Math.abs(o[0])>Math.abs(o[1])?o[0]>0?"right":"left":o[1]>0?"bottom":"top"}function uh(t){return!t.isGroup}function ch(t){return null!=t.shape}function dh(t,e,n){if(t&&e){var i=r(t);e.traverse((function(t){if(uh(t)&&t.anid){var e=i[t.anid];if(e){var r=o(t);t.attr(o(e)),th(t,r,n,$l(t).dataIndex)}}}))}function r(t){var e={};return t.traverse((function(t){uh(t)&&t.anid&&(e[t.anid]=t)})),e}function o(t){var e={x:t.x,y:t.y,rotation:t.rotation};return ch(t)&&(e.shape=I({},t.shape)),e}}function hh(t,e){return H(t,(function(t){var n=t[0];n=Rd(n,e.x),n=Hd(n,e.x+e.width);var i=t[1];return i=Rd(i,e.y),i=Hd(i,e.y+e.height),[n,i]}))}function fh(t,e){var n=Rd(t.x,e.x),i=Hd(t.x+t.width,e.x+e.width),r=Rd(t.y,e.y),o=Hd(t.y+t.height,e.y+e.height);if(i>=n&&o>=r)return{x:n,y:r,width:i-n,height:o-r}}function ph(t,e,n){var i=I({rectHover:!0},e),r=i.style={strokeNoScale:!0};if(n=n||{x:-1,y:-1,width:2,height:2},t)return 0===t.indexOf("image://")?(r.image=t.slice(8),A(r,n),new wl(i)):Gd(t.replace("path://",""),i,n,"center")}function mh(t,e,n,i,r){for(var o=0,a=r[r.length-1];o<r.length;o++){var s=r[o];if(gh(t,e,n,i,s[0],s[1],a[0],a[1]))return!0;a=s}}function gh(t,e,n,i,r,o,a,s){var l=n-t,u=i-e,c=a-r,d=s-o,h=yh(c,d,l,u);if(vh(h))return!1;var f=t-r,p=e-o,m=yh(f,p,l,u)/h;if(m<0||m>1)return!1;var g=yh(f,p,c,d)/h;return!(g<0||g>1)}function yh(t,e,n,i){return t*i-n*e}function vh(t){return t<=1e-6&&t>=-1e-6}function _h(t){var e=t.itemTooltipOption,n=t.componentModel,i=t.itemName,r=q(e)?{formatter:e}:e,o=n.mainType,a=n.componentIndex,s={componentType:o,name:i,$vars:["name"]};s[o+"Index"]=a;var l=t.formatterParamsExtra;l&&R(B(l),(function(t){_t(s,t)||(s[t]=l[t],s.$vars.push(t))}));var u=$l(t.el);u.componentMainType=o,u.componentIndex=a,u.tooltipConfig={name:i,option:A({content:i,formatterParams:s},r)}}Fd("circle",Dc),Fd("ellipse",Ac),Fd("sector",qc),Fd("ring",Jc),Fd("polygon",rd),Fd("polyline",sd),Fd("rect",Al),Fd("line",dd),Fd("bezierCurve",gd),Fd("arc",_d);var bh={};function xh(t,e){for(var n=0;n<Ql.length;n++){var i=Ql[n],r=e[i],o=t.ensureState(i);o.style=o.style||{},o.style.text=r}var a=t.currentStates.slice();t.clearStates(!0),t.setStyle({text:e.normal}),t.useStates(a,!0)}function wh(t,e,n){var i,r=t.labelFetcher,o=t.labelDataIndex,a=t.labelDimIndex,s=e.normal;r&&(i=r.getFormattedLabel(o,"normal",null,a,s&&s.get("formatter"),null!=n?{interpolatedValue:n}:null)),null==i&&(i=U(t.defaultText)?t.defaultText(o,t,n):t.defaultText);for(var l={normal:i},u=0;u<Ql.length;u++){var c=Ql[u],d=e[c];l[c]=at(r?r.getFormattedLabel(o,c,null,a,d&&d.get("formatter")):null,i)}return l}function Mh(t,e,n,i){n=n||bh;for(var r=t instanceof Gl,o=!1,a=0;a<tu.length;a++){var s=e[tu[a]];if(s&&s.getShallow("show")){o=!0;break}}var l=r?t:t.getTextContent();if(o){r||(l||(l=new Gl,t.setTextContent(l)),t.stateProxy&&(l.stateProxy=t.stateProxy));var u=wh(n,e),c=e.normal,d=!!c.getShallow("show"),h=Lh(c,i&&i.normal,n,!1,!r);h.text=u.normal,r||t.setTextConfig(kh(c,n,!1));for(a=0;a<Ql.length;a++){var f=Ql[a];s=e[f];if(s){var p=l.ensureState(f),m=!!at(s.getShallow("show"),d);if(m!==d&&(p.ignore=!m),p.style=Lh(s,i&&i[f],n,!0,!r),p.style.text=u[f],!r){var g=t.ensureState(f);g.textConfig=kh(s,n,!0)}}}l.silent=!!c.getShallow("silent"),null!=l.style.x&&(h.x=l.style.x),null!=l.style.y&&(h.y=l.style.y),l.ignore=!d,l.useStyle(h),l.dirty(),n.enableTextSetter&&(Ph(l).setLabelText=function(t){var i=wh(n,e,t);xh(l,i)})}else l&&(l.ignore=!0);t.dirty()}function Sh(t,e){e=e||"label";for(var n={normal:t.getModel(e)},i=0;i<Ql.length;i++){var r=Ql[i];n[r]=t.getModel([r,e])}return n}function Lh(t,e,n,i,r){var o={};return Th(o,t,n,i,r),e&&I(o,e),o}function kh(t,e,n){e=e||{};var i,r={},o=t.getShallow("rotate"),a=at(t.getShallow("distance"),n?null:5),s=t.getShallow("offset");return i=t.getShallow("position")||(n?null:"inside"),"outside"===i&&(i=e.defaultOutsidePosition||"top"),null!=i&&(r.position=i),null!=s&&(r.offset=s),null!=o&&(o*=Math.PI/180,r.rotation=o),null!=a&&(r.distance=a),r.outsideFill="inherit"===t.get("color")?e.inheritColor||null:"auto",r}function Th(t,e,n,i,r){n=n||bh;var o,a=e.ecModel,s=a&&a.option.textStyle,l=Dh(e);if(l)for(var u in o={},l)if(l.hasOwnProperty(u)){var c=e.getModel(["rich",u]);Yh(o[u]={},c,s,n,i,r,!1,!0)}o&&(t.rich=o);var d=e.get("overflow");d&&(t.overflow=d);var h=e.get("minMargin");null!=h&&(t.margin=h),Yh(t,e,s,n,i,r,!0,!1)}function Dh(t){var e;while(t&&t!==t.ecModel){var n=(t.option||bh).rich;if(n){e=e||{};for(var i=B(n),r=0;r<i.length;r++){var o=i[r];e[o]=1}}t=t.parentModel}return e}var Ch=["fontStyle","fontWeight","fontSize","fontFamily","textShadowColor","textShadowBlur","textShadowOffsetX","textShadowOffsetY"],Ih=["align","lineHeight","width","height","tag","verticalAlign"],Ah=["padding","borderWidth","borderRadius","borderDashOffset","backgroundColor","borderColor","shadowColor","shadowBlur","shadowOffsetX","shadowOffsetY"];function Yh(t,e,n,i,r,o,a,s){n=!r&&n||bh;var l=i&&i.inheritColor,u=e.getShallow("color"),c=e.getShallow("textBorderColor"),d=at(e.getShallow("opacity"),n.opacity);"inherit"!==u&&"auto"!==u||(u=l||null),"inherit"!==c&&"auto"!==c||(c=l||null),o||(u=u||n.color,c=c||n.textBorderColor),null!=u&&(t.fill=u),null!=c&&(t.stroke=c);var h=at(e.getShallow("textBorderWidth"),n.textBorderWidth);null!=h&&(t.lineWidth=h);var f=at(e.getShallow("textBorderType"),n.textBorderType);null!=f&&(t.lineDash=f);var p=at(e.getShallow("textBorderDashOffset"),n.textBorderDashOffset);null!=p&&(t.lineDashOffset=p),r||null!=d||s||(d=i&&i.defaultOpacity),null!=d&&(t.opacity=d),r||o||null==t.fill&&i.inheritColor&&(t.fill=i.inheritColor);for(var m=0;m<Ch.length;m++){var g=Ch[m],y=at(e.getShallow(g),n[g]);null!=y&&(t[g]=y)}for(m=0;m<Ih.length;m++){g=Ih[m],y=e.getShallow(g);null!=y&&(t[g]=y)}if(null==t.verticalAlign){var v=e.getShallow("baseline");null!=v&&(t.verticalAlign=v)}if(!a||!i.disableBox){for(m=0;m<Ah.length;m++){g=Ah[m],y=e.getShallow(g);null!=y&&(t[g]=y)}var _=e.getShallow("borderType");null!=_&&(t.borderDash=_),"auto"!==t.backgroundColor&&"inherit"!==t.backgroundColor||!l||(t.backgroundColor=l),"auto"!==t.borderColor&&"inherit"!==t.borderColor||!l||(t.borderColor=l)}}function Oh(t,e){var n=e&&e.getModel("textStyle");return dt([t.fontStyle||n&&n.getShallow("fontStyle")||"",t.fontWeight||n&&n.getShallow("fontWeight")||"",(t.fontSize||n&&n.getShallow("fontSize")||12)+"px",t.fontFamily||n&&n.getShallow("fontFamily")||"sans-serif"].join(" "))}var Ph=Io();function Eh(t,e,n,i){if(t){var r=Ph(t);r.prevValue=r.value,r.value=n;var o=e.normal;r.valueAnimation=o.get("valueAnimation"),r.valueAnimation&&(r.precision=o.get("precision"),r.defaultInterpolatedText=i,r.statesModels=e)}}function Nh(t,e,n,i,r){var o=Ph(t);if(o.valueAnimation){var a=o.defaultInterpolatedText,s=at(o.interpolatedValue,o.prevValue),l=o.value;(null==s?eh:th)(t,{},i,e,null,u)}function u(i){var u=Wo(n,o.precision,s,l,i);o.interpolatedValue=1===i?null:u;var c=wh({labelDataIndex:e,labelFetcher:r,defaultText:a?a(u):u+""},o.statesModels,u);xh(t,c)}}var Rh=["textStyle","color"],Hh=new Gl,jh=function(){function t(){}return t.prototype.getTextColor=function(t){var e=this.ecModel;return this.getShallow("color")||(!t&&e?e.get(Rh):null)},t.prototype.getFont=function(){return Oh({fontStyle:this.getShallow("fontStyle"),fontWeight:this.getShallow("fontWeight"),fontSize:this.getShallow("fontSize"),fontFamily:this.getShallow("fontFamily")},this.ecModel)},t.prototype.getTextRect=function(t){return Hh.useStyle({text:t,fontStyle:this.getShallow("fontStyle"),fontWeight:this.getShallow("fontWeight"),fontSize:this.getShallow("fontSize"),fontFamily:this.getShallow("fontFamily"),verticalAlign:this.getShallow("verticalAlign")||this.getShallow("baseline"),padding:this.getShallow("padding"),lineHeight:this.getShallow("lineHeight"),rich:this.getShallow("rich")}),Hh.update(),Hh.getBoundingRect()},t}(),zh=jh,Wh=[["lineWidth","width"],["stroke","color"],["opacity"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["shadowColor"],["lineDash","type"],["lineDashOffset","dashOffset"],["lineCap","cap"],["lineJoin","join"],["miterLimit"]],Bh=na(Wh),Fh=function(){function t(){}return t.prototype.getLineStyle=function(t){return Bh(this,t)},t}(),Vh=[["fill","color"],["stroke","borderColor"],["lineWidth","borderWidth"],["opacity"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["shadowColor"],["lineDash","borderType"],["lineDashOffset","borderDashOffset"],["lineCap","borderCap"],["lineJoin","borderJoin"],["miterLimit","borderMiterLimit"]],Gh=na(Vh),$h=function(){function t(){}return t.prototype.getItemStyle=function(t,e){return Gh(this,t,e)},t}(),Uh=function(){function t(t,e,n){this.parentModel=e,this.ecModel=n,this.option=t}return t.prototype.init=function(t,e,n){for(var i=[],r=3;r<arguments.length;r++)i[r-3]=arguments[r]},t.prototype.mergeOption=function(t,e){D(this.option,t,!0)},t.prototype.get=function(t,e){return null==t?this.option:this._doGet(this.parsePath(t),!e&&this.parentModel)},t.prototype.getShallow=function(t,e){var n=this.option,i=null==n?n:n[t];if(null==i&&!e){var r=this.parentModel;r&&(i=r.getShallow(t))}return i},t.prototype.getModel=function(e,n){var i=null!=e,r=i?this.parsePath(e):null,o=i?this._doGet(r):this.option;return n=n||this.parentModel&&this.parentModel.getModel(this.resolveParentPath(r)),new t(o,n,this.ecModel)},t.prototype.isEmpty=function(){return null==this.option},t.prototype.restoreData=function(){},t.prototype.clone=function(){var t=this.constructor;return new t(T(this.option))},t.prototype.parsePath=function(t){return"string"===typeof t?t.split("."):t},t.prototype.resolveParentPath=function(t){return t},t.prototype.isAnimationEnabled=function(){if(!h.node&&this.option){if(null!=this.option.animation)return!!this.option.animation;if(this.parentModel)return this.parentModel.isAnimationEnabled()}},t.prototype._doGet=function(t,e){var n=this.option;if(!t)return n;for(var i=0;i<t.length;i++)if(t[i]&&(n=n&&"object"===typeof n?n[t[i]]:null,null==n))break;return null==n&&e&&(n=e._doGet(this.resolveParentPath(t),e.parentModel)),n},t}();qo(Uh),Ko(Uh),E(Uh,Fh),E(Uh,$h),E(Uh,oa),E(Uh,zh);var qh=Uh,Xh=Math.round(10*Math.random());function Zh(t){return[t||"",Xh++].join("_")}function Jh(t){var e={};t.registerSubTypeDefaulter=function(t,n){var i=Go(t);e[i.main]=n},t.determineSubType=function(n,i){var r=i.type;if(!r){var o=Go(n).main;t.hasSubTypes(n)&&e[o]&&(r=e[o](i))}return r}}function Kh(t,e){function n(t){var n={},o=[];return R(t,(function(a){var s=i(n,a),l=s.originalDeps=e(a),u=r(l,t);s.entryCount=u.length,0===s.entryCount&&o.push(a),R(u,(function(t){O(s.predecessor,t)<0&&s.predecessor.push(t);var e=i(n,t);O(e.successor,t)<0&&e.successor.push(a)}))})),{graph:n,noEntryList:o}}function i(t,e){return t[e]||(t[e]={predecessor:[],successor:[]}),t[e]}function r(t,e){var n=[];return R(t,(function(t){O(e,t)>=0&&n.push(t)})),n}t.topologicalTravel=function(t,e,i,r){if(t.length){var o=n(e),a=o.graph,s=o.noEntryList,l={};R(t,(function(t){l[t]=!0}));while(s.length){var u=s.pop(),c=a[u],d=!!l[u];d&&(i.call(r,u,c.originalDeps.slice()),delete l[u]),R(c.successor,d?f:h)}R(l,(function(){var t="";throw new Error(t)}))}function h(t){a[t].entryCount--,0===a[t].entryCount&&s.push(t)}function f(t){l[t]=!0,h(t)}}}function Qh(t,e){return D(D({},t,!0),e,!0)}var tf={time:{month:["January","February","March","April","May","June","July","August","September","October","November","December"],monthAbbr:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],dayOfWeek:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"],dayOfWeekAbbr:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"]},legend:{selector:{all:"All",inverse:"Inv"}},toolbox:{brush:{title:{rect:"Box Select",polygon:"Lasso Select",lineX:"Horizontally Select",lineY:"Vertically Select",keep:"Keep Selections",clear:"Clear Selections"}},dataView:{title:"Data View",lang:["Data View","Close","Refresh"]},dataZoom:{title:{zoom:"Zoom",back:"Zoom Reset"}},magicType:{title:{line:"Switch to Line Chart",bar:"Switch to Bar Chart",stack:"Stack",tiled:"Tile"}},restore:{title:"Restore"},saveAsImage:{title:"Save as Image",lang:["Right Click to Save Image"]}},series:{typeNames:{pie:"Pie chart",bar:"Bar chart",line:"Line chart",scatter:"Scatter plot",effectScatter:"Ripple scatter plot",radar:"Radar chart",tree:"Tree",treemap:"Treemap",boxplot:"Boxplot",candlestick:"Candlestick",k:"K line chart",heatmap:"Heat map",map:"Map",parallel:"Parallel coordinate map",lines:"Line graph",graph:"Relationship graph",sankey:"Sankey diagram",funnel:"Funnel chart",gauge:"Gauge",pictorialBar:"Pictorial bar",themeRiver:"Theme River Map",sunburst:"Sunburst"}},aria:{general:{withTitle:'This is a chart about "{title}"',withoutTitle:"This is a chart"},series:{single:{prefix:"",withName:" with type {seriesType} named {seriesName}.",withoutName:" with type {seriesType}."},multiple:{prefix:". It consists of {seriesCount} series count.",withName:" The {seriesId} series is a {seriesType} representing {seriesName}.",withoutName:" The {seriesId} series is a {seriesType}.",separator:{middle:"",end:""}}},data:{allData:"The data is as follows: ",partialData:"The first {displayCnt} items are: ",withName:"the data for {name} is {value}",withoutName:"{value}",separator:{middle:", ",end:". "}}}},ef={time:{month:["一月","二月","三月","四月","五月","六月","七月","八月","九月","十月","十一月","十二月"],monthAbbr:["1月","2月","3月","4月","5月","6月","7月","8月","9月","10月","11月","12月"],dayOfWeek:["星期日","星期一","星期二","星期三","星期四","星期五","星期六"],dayOfWeekAbbr:["日","一","二","三","四","五","六"]},legend:{selector:{all:"全选",inverse:"反选"}},toolbox:{brush:{title:{rect:"矩形选择",polygon:"圈选",lineX:"横向选择",lineY:"纵向选择",keep:"保持选择",clear:"清除选择"}},dataView:{title:"数据视图",lang:["数据视图","关闭","刷新"]},dataZoom:{title:{zoom:"区域缩放",back:"区域缩放还原"}},magicType:{title:{line:"切换为折线图",bar:"切换为柱状图",stack:"切换为堆叠",tiled:"切换为平铺"}},restore:{title:"还原"},saveAsImage:{title:"保存为图片",lang:["右键另存为图片"]}},series:{typeNames:{pie:"饼图",bar:"柱状图",line:"折线图",scatter:"散点图",effectScatter:"涟漪散点图",radar:"雷达图",tree:"树图",treemap:"矩形树图",boxplot:"箱型图",candlestick:"K线图",k:"K线图",heatmap:"热力图",map:"地图",parallel:"平行坐标图",lines:"线图",graph:"关系图",sankey:"桑基图",funnel:"漏斗图",gauge:"仪表盘图",pictorialBar:"象形柱图",themeRiver:"主题河流图",sunburst:"旭日图"}},aria:{general:{withTitle:"这是一个关于“{title}”的图表。",withoutTitle:"这是一个图表，"},series:{single:{prefix:"",withName:"图表类型是{seriesType}，表示{seriesName}。",withoutName:"图表类型是{seriesType}。"},multiple:{prefix:"它由{seriesCount}个图表系列组成。",withName:"第{seriesId}个系列是一个表示{seriesName}的{seriesType}，",withoutName:"第{seriesId}个系列是一个{seriesType}，",separator:{middle:"；",end:"。"}}},data:{allData:"其数据是——",partialData:"其中，前{displayCnt}项是——",withName:"{name}的数据是{value}",withoutName:"{value}",separator:{middle:"，",end:""}}}},nf="ZH",rf="EN",of=rf,af={},sf={},lf=h.domSupported?function(){var t=(document.documentElement.lang||navigator.language||navigator.browserLanguage).toUpperCase();return t.indexOf(nf)>-1?nf:of}():of;function uf(t,e){t=t.toUpperCase(),sf[t]=new qh(e),af[t]=e}function cf(t){if(q(t)){var e=af[t.toUpperCase()]||{};return t===nf||t===rf?T(e):D(T(e),T(af[of]),!1)}return D(T(t),T(af[of]),!1)}function df(t){return sf[t]}function hf(){return sf[of]}uf(rf,tf),uf(nf,ef);var ff=1e3,pf=60*ff,mf=60*pf,gf=24*mf,yf=365*gf,vf={year:"{yyyy}",month:"{MMM}",day:"{d}",hour:"{HH}:{mm}",minute:"{HH}:{mm}",second:"{HH}:{mm}:{ss}",millisecond:"{hh}:{mm}:{ss} {SSS}",none:"{yyyy}-{MM}-{dd} {hh}:{mm}:{ss} {SSS}"},_f="{yyyy}-{MM}-{dd}",bf={year:"{yyyy}",month:"{yyyy}-{MM}",day:_f,hour:_f+" "+vf.hour,minute:_f+" "+vf.minute,second:_f+" "+vf.second,millisecond:vf.none},xf=["year","month","day","hour","minute","second","millisecond"],wf=["year","half-year","quarter","month","week","half-week","day","half-day","quarter-day","hour","minute","second","millisecond"];function Mf(t,e){return t+="","0000".substr(0,e-t.length)+t}function Sf(t){switch(t){case"half-year":case"quarter":return"month";case"week":case"half-week":return"day";case"half-day":case"quarter-day":return"hour";default:return t}}function Lf(t){return t===Sf(t)}function kf(t){switch(t){case"year":case"month":return"day";case"millisecond":return"millisecond";default:return"second"}}function Tf(t,e,n,i){var r=qr(t),o=r[Af(n)](),a=r[Yf(n)]()+1,s=Math.floor((a-1)/4)+1,l=r[Of(n)](),u=r["get"+(n?"UTC":"")+"Day"](),c=r[Pf(n)](),d=(c-1)%12+1,h=r[Ef(n)](),f=r[Nf(n)](),p=r[Rf(n)](),m=i instanceof qh?i:df(i||lf)||hf(),g=m.getModel("time"),y=g.get("month"),v=g.get("monthAbbr"),_=g.get("dayOfWeek"),b=g.get("dayOfWeekAbbr");return(e||"").replace(/{yyyy}/g,o+"").replace(/{yy}/g,o%100+"").replace(/{Q}/g,s+"").replace(/{MMMM}/g,y[a-1]).replace(/{MMM}/g,v[a-1]).replace(/{MM}/g,Mf(a,2)).replace(/{M}/g,a+"").replace(/{dd}/g,Mf(l,2)).replace(/{d}/g,l+"").replace(/{eeee}/g,_[u]).replace(/{ee}/g,b[u]).replace(/{e}/g,u+"").replace(/{HH}/g,Mf(c,2)).replace(/{H}/g,c+"").replace(/{hh}/g,Mf(d+"",2)).replace(/{h}/g,d+"").replace(/{mm}/g,Mf(h,2)).replace(/{m}/g,h+"").replace(/{ss}/g,Mf(f,2)).replace(/{s}/g,f+"").replace(/{SSS}/g,Mf(p,3)).replace(/{S}/g,p+"")}function Df(t,e,n,i,r){var o=null;if("string"===typeof n)o=n;else if("function"===typeof n)o=n(t.value,e,{level:t.level});else{var a=I({},vf);if(t.level>0)for(var s=0;s<xf.length;++s)a[xf[s]]="{primary|"+a[xf[s]]+"}";var l=n?!1===n.inherit?n:A(n,a):a,u=Cf(t.value,r);if(l[u])o=l[u];else if(l.inherit){var c=wf.indexOf(u);for(s=c-1;s>=0;--s)if(l[u]){o=l[u];break}o=o||a.none}if($(o)){var d=null==t.level?0:t.level>=0?t.level:o.length+t.level;d=Math.min(d,o.length-1),o=o[d]}}return Tf(new Date(t.value),o,r,i)}function Cf(t,e){var n=qr(t),i=n[Yf(e)]()+1,r=n[Of(e)](),o=n[Pf(e)](),a=n[Ef(e)](),s=n[Nf(e)](),l=n[Rf(e)](),u=0===l,c=u&&0===s,d=c&&0===a,h=d&&0===o,f=h&&1===r,p=f&&1===i;return p?"year":f?"month":h?"day":d?"hour":c?"minute":u?"second":"millisecond"}function If(t,e,n){var i="number"===typeof t?qr(t):t;switch(e=e||Cf(t,n),e){case"year":return i[Af(n)]();case"half-year":return i[Yf(n)]()>=6?1:0;case"quarter":return Math.floor((i[Yf(n)]()+1)/4);case"month":return i[Yf(n)]();case"day":return i[Of(n)]();case"half-day":return i[Pf(n)]()/24;case"hour":return i[Pf(n)]();case"minute":return i[Ef(n)]();case"second":return i[Nf(n)]();case"millisecond":return i[Rf(n)]()}}function Af(t){return t?"getUTCFullYear":"getFullYear"}function Yf(t){return t?"getUTCMonth":"getMonth"}function Of(t){return t?"getUTCDate":"getDate"}function Pf(t){return t?"getUTCHours":"getHours"}function Ef(t){return t?"getUTCMinutes":"getMinutes"}function Nf(t){return t?"getUTCSeconds":"getSeconds"}function Rf(t){return t?"getUTCSeconds":"getSeconds"}function Hf(t){return t?"setUTCFullYear":"setFullYear"}function jf(t){return t?"setUTCMonth":"setMonth"}function zf(t){return t?"setUTCDate":"setDate"}function Wf(t){return t?"setUTCHours":"setHours"}function Bf(t){return t?"setUTCMinutes":"setMinutes"}function Ff(t){return t?"setUTCSeconds":"setSeconds"}function Vf(t){return t?"setUTCSeconds":"setSeconds"}function Gf(t){if(!eo(t))return q(t)?t:"-";var e=(t+"").split(".");return e[0].replace(/(\d{1,3})(?=(?:\d{3})+(?!\d))/g,"$1,")+(e.length>1?"."+e[1]:"")}function $f(t,e){return t=(t||"").toLowerCase().replace(/-(.)/g,(function(t,e){return e.toUpperCase()})),e&&t&&(t=t.charAt(0).toUpperCase()+t.slice(1)),t}var Uf=ut,qf=/([&<>"'])/g,Xf={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"};function Zf(t){return null==t?"":(t+"").replace(qf,(function(t,e){return Xf[e]}))}function Jf(t,e,n){var i="{yyyy}-{MM}-{dd} {hh}:{mm}:{ss}";function r(t){return t&&dt(t)?t:"-"}function o(t){return!(null==t||isNaN(t)||!isFinite(t))}var a="time"===e,s=t instanceof Date;if(a||s){var l=a?qr(t):t;if(!isNaN(+l))return Tf(l,i,n);if(s)return"-"}if("ordinal"===e)return X(t)?r(t):Z(t)&&o(t)?t+"":"-";var u=to(t);return o(u)?Gf(u):X(t)?r(t):"-"}var Kf=["a","b","c","d","e","f","g"],Qf=function(t,e){return"{"+t+(null==e?"":e)+"}"};function tp(t,e,n){$(e)||(e=[e]);var i=e.length;if(!i)return"";for(var r=e[0].$vars||[],o=0;o<r.length;o++){var a=Kf[o];t=t.replace(Qf(a),Qf(a,0))}for(var s=0;s<i;s++)for(var l=0;l<r.length;l++){var u=e[s][r[l]];t=t.replace(Qf(Kf[l],s),n?Zf(u):u)}return t}function ep(t,e,n){return R(e,(function(e,i){t=t.replace("{"+i+"}",n?Zf(e):e)})),t}function np(t,e){var n=q(t)?{color:t,extraCssText:e}:t||{},i=n.color,r=n.type;e=n.extraCssText;var o=n.renderMode||"html";if(!i)return"";if("html"===o)return"subItem"===r?'<span style="display:inline-block;vertical-align:middle;margin-right:8px;margin-left:3px;border-radius:4px;width:4px;height:4px;background-color:'+Zf(i)+";"+(e||"")+'"></span>':'<span style="display:inline-block;margin-right:4px;border-radius:10px;width:10px;height:10px;background-color:'+Zf(i)+";"+(e||"")+'"></span>';var a=n.markerId||"markerX";return{renderMode:o,content:"{"+a+"|}  ",style:"subItem"===r?{width:4,height:4,borderRadius:2,backgroundColor:i}:{width:10,height:10,borderRadius:5,backgroundColor:i}}}function ip(t,e){return e=e||"transparent",q(t)?t:J(t)&&t.colorStops&&(t.colorStops[0]||{}).color||e}function rp(t,e){if("_blank"===e||"blank"===e){var n=window.open();n.opener=null,n.location.href=t}else window.open(t,e)}var op=R,ap=["left","right","top","bottom","width","height"],sp=[["width","left","right"],["height","top","bottom"]];function lp(t,e,n,i,r){var o=0,a=0;null==i&&(i=1/0),null==r&&(r=1/0);var s=0;e.eachChild((function(l,u){var c,d,h=l.getBoundingRect(),f=e.childAt(u+1),p=f&&f.getBoundingRect();if("horizontal"===t){var m=h.width+(p?-p.x+h.x:0);c=o+m,c>i||l.newline?(o=0,c=m,a+=s+n,s=h.height):s=Math.max(s,h.height)}else{var g=h.height+(p?-p.y+h.y:0);d=a+g,d>r||l.newline?(o+=s+n,a=0,d=g,s=h.width):s=Math.max(s,h.width)}l.newline||(l.x=o,l.y=a,l.markRedraw(),"horizontal"===t?o=c+n:a=d+n)}))}var up=lp;G(lp,"vertical"),G(lp,"horizontal");function cp(t,e,n){var i=e.width,r=e.height,o=Nr(t.left,i),a=Nr(t.top,r),s=Nr(t.right,i),l=Nr(t.bottom,r);return(isNaN(o)||isNaN(parseFloat(t.left)))&&(o=0),(isNaN(s)||isNaN(parseFloat(t.right)))&&(s=i),(isNaN(a)||isNaN(parseFloat(t.top)))&&(a=0),(isNaN(l)||isNaN(parseFloat(t.bottom)))&&(l=r),n=Uf(n||0),{width:Math.max(s-o-n[1]-n[3],0),height:Math.max(l-a-n[0]-n[2],0)}}function dp(t,e,n){n=Uf(n||0);var i=e.width,r=e.height,o=Nr(t.left,i),a=Nr(t.top,r),s=Nr(t.right,i),l=Nr(t.bottom,r),u=Nr(t.width,i),c=Nr(t.height,r),d=n[2]+n[0],h=n[1]+n[3],f=t.aspect;switch(isNaN(u)&&(u=i-s-h-o),isNaN(c)&&(c=r-l-d-a),null!=f&&(isNaN(u)&&isNaN(c)&&(f>i/r?u=.8*i:c=.8*r),isNaN(u)&&(u=f*c),isNaN(c)&&(c=u/f)),isNaN(o)&&(o=i-s-u-h),isNaN(a)&&(a=r-l-c-d),t.left||t.right){case"center":o=i/2-u/2-n[3];break;case"right":o=i-u-h;break}switch(t.top||t.bottom){case"middle":case"center":a=r/2-c/2-n[0];break;case"bottom":a=r-c-d;break}o=o||0,a=a||0,isNaN(u)&&(u=i-h-o-(s||0)),isNaN(c)&&(c=r-d-a-(l||0));var p=new Ji(o+n[3],a+n[0],u,c);return p.margin=n,p}function hp(t,e,n,i,r){var o=!r||!r.hv||r.hv[0],a=!r||!r.hv||r.hv[1],s=r&&r.boundingMode||"all";if(o||a){var l;if("raw"===s)l="group"===t.type?new Ji(0,0,+e.width||0,+e.height||0):t.getBoundingRect();else if(l=t.getBoundingRect(),t.needLocalTransform()){var u=t.getLocalTransform();l=l.clone(),l.applyTransform(u)}var c=dp(A({width:l.width,height:l.height},e),n,i),d=o?c.x-l.x:0,h=a?c.y-l.y:0;"raw"===s?(t.x=d,t.y=h):(t.x+=d,t.y+=h),t.markRedraw()}}function fp(t,e){return null!=t[sp[e][0]]||null!=t[sp[e][1]]&&null!=t[sp[e][2]]}function pp(t){var e=t.layoutMode||t.constructor.layoutMode;return J(e)?e:e?{type:e}:null}function mp(t,e,n){var i=n&&n.ignoreSize;!$(i)&&(i=[i,i]);var r=a(sp[0],0),o=a(sp[1],1);function a(n,r){var o={},a=0,u={},c=0,d=2;if(op(n,(function(e){u[e]=t[e]})),op(n,(function(t){s(e,t)&&(o[t]=u[t]=e[t]),l(o,t)&&a++,l(u,t)&&c++})),i[r])return l(e,n[1])?u[n[2]]=null:l(e,n[2])&&(u[n[1]]=null),u;if(c!==d&&a){if(a>=d)return o;for(var h=0;h<n.length;h++){var f=n[h];if(!s(o,f)&&s(t,f)){o[f]=t[f];break}}return o}return u}function s(t,e){return t.hasOwnProperty(e)}function l(t,e){return null!=t[e]&&"auto"!==t[e]}function u(t,e,n){op(t,(function(t){e[t]=n[t]}))}u(sp[0],t,r),u(sp[1],t,o)}function gp(t){return yp({},t)}function yp(t,e){return e&&t&&op(ap,(function(n){e.hasOwnProperty(n)&&(t[n]=e[n])})),t}var vp=Io(),_p=function(t){function e(e,n,i){var r=t.call(this,e,n,i)||this;return r.uid=Zh("ec_cpt_model"),r}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n)},e.prototype.mergeDefaultAndTheme=function(t,e){var n=pp(this),i=n?gp(t):{},r=e.getTheme();D(t,r.get(this.mainType)),D(t,this.getDefaultOption()),n&&mp(t,i,n)},e.prototype.mergeOption=function(t,e){D(this.option,t,!0);var n=pp(this);n&&mp(this.option,t,n)},e.prototype.optionUpdated=function(t,e){},e.prototype.getDefaultOption=function(){var t=this.constructor;if(!Uo(t))return t.defaultOption;var e=vp(this);if(!e.defaultOption){var n=[],i=t;while(i){var r=i.prototype.defaultOption;r&&n.push(r),i=i.superClass}for(var o={},a=n.length-1;a>=0;a--)o=D(o,n[a],!0);e.defaultOption=o}return e.defaultOption},e.prototype.getReferringComponents=function(t,e){var n=t+"Index",i=t+"Id";return No(this.ecModel,t,{index:this.get(n,!0),id:this.get(i,!0)},e)},e.prototype.getBoxLayoutParams=function(){var t=this;return{left:t.get("left"),top:t.get("top"),right:t.get("right"),bottom:t.get("bottom"),width:t.get("width"),height:t.get("height")}},e.protoInitialize=function(){var t=e.prototype;t.type="component",t.id="",t.name="",t.mainType="",t.subType="",t.componentIndex=0}(),e}(qh);function bp(t){var e=[];return R(_p.getClassesByMainType(t),(function(t){e=e.concat(t.dependencies||t.prototype.dependencies||[])})),e=H(e,(function(t){return Go(t).main})),"dataset"!==t&&O(e,"dataset")<=0&&e.unshift("dataset"),e}Zo(_p,qh),ea(_p),Jh(_p),Kh(_p,bp);var xp=_p,wp="";"undefined"!==typeof navigator&&(wp=navigator.platform||"");var Mp="rgba(0, 0, 0, 0.2)",Sp={darkMode:"auto",color:["#5470c6","#91cc75","#fac858","#ee6666","#73c0de","#3ba272","#fc8452","#9a60b4","#ea7ccc"],gradientColor:["#f6efa6","#d88273","#bf444c"],aria:{decal:{decals:[{color:Mp,dashArrayX:[1,0],dashArrayY:[2,5],symbolSize:1,rotation:Math.PI/6},{color:Mp,symbol:"circle",dashArrayX:[[8,8],[0,8,8,0]],dashArrayY:[6,0],symbolSize:.8},{color:Mp,dashArrayX:[1,0],dashArrayY:[4,3],rotation:-Math.PI/4},{color:Mp,dashArrayX:[[6,6],[0,6,6,0]],dashArrayY:[6,0]},{color:Mp,dashArrayX:[[1,0],[1,6]],dashArrayY:[1,0,6,0],rotation:Math.PI/4},{color:Mp,symbol:"triangle",dashArrayX:[[9,9],[0,9,9,0]],dashArrayY:[7,2],symbolSize:.75}]}},textStyle:{fontFamily:wp.match(/^Win/)?"Microsoft YaHei":"sans-serif",fontSize:12,fontStyle:"normal",fontWeight:"normal"},blendMode:null,stateAnimation:{duration:300,easing:"cubicOut"},animation:"auto",animationDuration:1e3,animationDurationUpdate:500,animationEasing:"cubicInOut",animationEasingUpdate:"cubicInOut",animationThreshold:2e3,progressiveThreshold:3e3,progressive:400,hoverLayerThreshold:3e3,useUTC:!1},Lp=gt(["tooltip","label","itemName","itemId","seriesName"]),kp="original",Tp="arrayRows",Dp="objectRows",Cp="keyedColumns",Ip="typedArray",Ap="unknown",Yp="column",Op="row",Pp={Must:1,Might:2,Not:3},Ep=Io();function Np(t){Ep(t).datasetMap=gt()}function Rp(t,e,n){var i={},r=jp(e);if(!r||!t)return i;var o,a,s=[],l=[],u=e.ecModel,c=Ep(u).datasetMap,d=r.uid+"_"+n.seriesLayoutBy;t=t.slice(),R(t,(function(e,n){var r=J(e)?e:t[n]={name:e};"ordinal"===r.type&&null==o&&(o=n,a=p(r)),i[r.name]=[]}));var h=c.get(d)||c.set(d,{categoryWayDim:a,valueWayDim:0});function f(t,e,n){for(var i=0;i<n;i++)t.push(e+i)}function p(t){var e=t.dimsDef;return e?e.length:1}return R(t,(function(t,e){var n=t.name,r=p(t);if(null==o){var a=h.valueWayDim;f(i[n],a,r),f(l,a,r),h.valueWayDim+=r}else if(o===e)f(i[n],0,r),f(s,0,r);else{a=h.categoryWayDim;f(i[n],a,r),f(l,a,r),h.categoryWayDim+=r}})),s.length&&(i.itemName=s),l.length&&(i.seriesName=l),i}function Hp(t,e,n){var i={},r=jp(t);if(!r)return i;var o,a=e.sourceFormat,s=e.dimensionsDefine;a!==Dp&&a!==Cp||R(s,(function(t,e){"name"===(J(t)?t.name:t)&&(o=e)}));var l=function(){for(var t={},i={},r=[],l=0,u=Math.min(5,n);l<u;l++){var c=Bp(e.data,a,e.seriesLayoutBy,s,e.startIndex,l);r.push(c);var d=c===Pp.Not;if(d&&null==t.v&&l!==o&&(t.v=l),(null==t.n||t.n===t.v||!d&&r[t.n]===Pp.Not)&&(t.n=l),h(t)&&r[t.n]!==Pp.Not)return t;d||(c===Pp.Might&&null==i.v&&l!==o&&(i.v=l),null!=i.n&&i.n!==i.v||(i.n=l))}function h(t){return null!=t.v&&null!=t.n}return h(t)?t:h(i)?i:null}();if(l){i.value=[l.v];var u=null!=o?o:l.n;i.itemName=[u],i.seriesName=[u]}return i}function jp(t){var e=t.get("data",!0);if(!e)return No(t.ecModel,"dataset",{index:t.get("datasetIndex",!0),id:t.get("datasetId",!0)},Po).models[0]}function zp(t){return t.get("transform",!0)||t.get("fromTransformResult",!0)?No(t.ecModel,"dataset",{index:t.get("fromDatasetIndex",!0),id:t.get("fromDatasetId",!0)},Po).models:[]}function Wp(t,e){return Bp(t.data,t.sourceFormat,t.seriesLayoutBy,t.dimensionsDefine,t.startIndex,e)}function Bp(t,e,n,i,r,o){var a,s,l,u=5;if(Q(t))return Pp.Not;if(i){var c=i[o];J(c)?(s=c.name,l=c.type):q(c)&&(s=c)}if(null!=l)return"ordinal"===l?Pp.Must:Pp.Not;if(e===Tp){var d=t;if(n===Op){for(var h=d[o],f=0;f<(h||[]).length&&f<u;f++)if(null!=(a=b(h[r+f])))return a}else for(f=0;f<d.length&&f<u;f++){var p=d[r+f];if(p&&null!=(a=b(p[o])))return a}}else if(e===Dp){var m=t;if(!s)return Pp.Not;for(f=0;f<m.length&&f<u;f++){var g=m[f];if(g&&null!=(a=b(g[s])))return a}}else if(e===Cp){var y=t;if(!s)return Pp.Not;h=y[s];if(!h||Q(h))return Pp.Not;for(f=0;f<h.length&&f<u;f++)if(null!=(a=b(h[f])))return a}else if(e===kp){var v=t;for(f=0;f<v.length&&f<u;f++){g=v[f];var _=co(g);if(!$(_))return Pp.Not;if(null!=(a=b(_[o])))return a}}function b(t){var e=q(t);return null!=t&&isFinite(t)&&""!==t?e?Pp.Might:Pp.Not:e&&"-"!==t?Pp.Must:void 0}return Pp.Not}var Fp=gt();function Vp(t,e){ct(null==Fp.get(t)&&e),Fp.set(t,e)}function Gp(t,e,n){var i=Fp.get(e);if(!i)return n;var r=i(t);return r?n.concat(r):n}var $p=Io(),Up=Io(),qp=function(){function t(){}return t.prototype.getColorFromPalette=function(t,e,n){var i=so(this.get("color",!0)),r=this.get("colorLayer",!0);return Jp(this,$p,i,r,t,e,n)},t.prototype.clearColorPalette=function(){Kp(this,$p)},t}();function Xp(t,e,n,i){var r=so(t.get(["aria","decal","decals"]));return Jp(t,Up,r,null,e,n,i)}function Zp(t,e){for(var n=t.length,i=0;i<n;i++)if(t[i].length>e)return t[i];return t[n-1]}function Jp(t,e,n,i,r,o,a){o=o||t;var s=e(o),l=s.paletteIdx||0,u=s.paletteNameMap=s.paletteNameMap||{};if(u.hasOwnProperty(r))return u[r];var c=null!=a&&i?Zp(i,a):n;if(c=c||n,c&&c.length){var d=c[l];return r&&(u[r]=d),s.paletteIdx=(l+1)%c.length,d}}function Kp(t,e){e(t).paletteIdx=0,e(t).paletteNameMap={}}var Qp,tm,em;"undefined"!==typeof console&&console.warn&&console.log;function nm(t){0}function im(t){throw new Error(t)}var rm="\0_ec_inner",om=1;var am=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.init=function(t,e,n,i,r,o){i=i||{},this.option=null,this._theme=new qh(i),this._locale=new qh(r),this._optionManager=o},e.prototype.setOption=function(t,e,n){var i=dm(e);this._optionManager.setOption(t,n,i),this._resetOption(null,i)},e.prototype.resetOption=function(t,e){return this._resetOption(t,dm(e))},e.prototype._resetOption=function(t,e){var n=!1,i=this._optionManager;if(!t||"recreate"===t){var r=i.mountOption("recreate"===t);0,this.option&&"recreate"!==t?(this.restoreData(),this._mergeOption(r,e)):em(this,r),n=!0}if("timeline"!==t&&"media"!==t||this.restoreData(),!t||"recreate"===t||"timeline"===t){var o=i.getTimelineOption(this);o&&(n=!0,this._mergeOption(o,e))}if(!t||"recreate"===t||"media"===t){var a=i.getMediaOption(this);a.length&&R(a,(function(t){n=!0,this._mergeOption(t,e)}),this)}return n},e.prototype.mergeOption=function(t){this._mergeOption(t,null)},e.prototype._mergeOption=function(t,e){var n=this.option,i=this._componentsMap,r=this._componentsCount,o=[],a=gt(),s=e&&e.replaceMergeMainTypeMap;function l(e){var o=Gp(this,e,so(t[e])),a=i.get(e),l=a?s&&s.get(e)?"replaceMerge":"normalMerge":"replaceAll",u=fo(a,o,l);ko(u,e,xp),n[e]=null,i.set(e,null),r.set(e,0);var c=[],d=[],h=0;R(u,(function(t,n){var i=t.existing,r=t.newOption;if(r){var o="series"===e,a=xp.getClass(e,t.keyInfo.subType,!o);if(!a)return;if(i&&i.constructor===a)i.name=t.keyInfo.name,i.mergeOption(r,this),i.optionUpdated(r,!1);else{var s=I({componentIndex:n},t.keyInfo);i=new a(r,this,this,s),I(i,s),t.brandNew&&(i.__requireNewView=!0),i.init(r,this,this),i.optionUpdated(null,!0)}}else i&&(i.mergeOption({},this),i.optionUpdated({},!1));i?(c.push(i.option),d.push(i),h++):(c.push(void 0),d.push(void 0))}),this),n[e]=c,i.set(e,d),r.set(e,h),"series"===e&&Qp(this)}Np(this),R(t,(function(t,e){null!=t&&(xp.hasClass(e)?e&&(o.push(e),a.set(e,!0)):n[e]=null==n[e]?T(t):D(n[e],t,!0))})),s&&s.each((function(t,e){xp.hasClass(e)&&!a.get(e)&&(o.push(e),a.set(e,!0))})),xp.topologicalTravel(o,xp.getAllClassMainTypes(),l,this),this._seriesIndices||Qp(this)},e.prototype.getOption=function(){var t=T(this.option);return R(t,(function(e,n){if(xp.hasClass(n)){for(var i=so(e),r=i.length,o=!1,a=r-1;a>=0;a--)i[a]&&!So(i[a])?o=!0:(i[a]=null,!o&&r--);i.length=r,t[n]=i}})),delete t[rm],t},e.prototype.getTheme=function(){return this._theme},e.prototype.getLocaleModel=function(){return this._locale},e.prototype.getLocale=function(t){var e=this.getLocaleModel();return e.get(t)},e.prototype.setUpdatePayload=function(t){this._payload=t},e.prototype.getUpdatePayload=function(){return this._payload},e.prototype.getComponent=function(t,e){var n=this._componentsMap.get(t);if(n){var i=n[e||0];if(i)return i;if(null==e)for(var r=0;r<n.length;r++)if(n[r])return n[r]}},e.prototype.queryComponents=function(t){var e=t.mainType;if(!e)return[];var n,i=t.index,r=t.id,o=t.name,a=this._componentsMap.get(e);return a&&a.length?(null!=i?(n=[],R(so(i),(function(t){a[t]&&n.push(a[t])}))):n=null!=r?um("id",r,a):null!=o?um("name",o,a):z(a,(function(t){return!!t})),cm(n,t)):[]},e.prototype.findComponents=function(t){var e=t.query,n=t.mainType,i=o(e),r=i?this.queryComponents(i):z(this._componentsMap.get(n),(function(t){return!!t}));return a(cm(r,t));function o(t){var e=n+"Index",i=n+"Id",r=n+"Name";return!t||null==t[e]&&null==t[i]&&null==t[r]?null:{mainType:n,index:t[e],id:t[i],name:t[r]}}function a(e){return t.filter?z(e,t.filter):e}},e.prototype.eachComponent=function(t,e,n){var i=this._componentsMap;if(U(t)){var r=e,o=t;i.each((function(t,e){for(var n=0;t&&n<t.length;n++){var i=t[n];i&&o.call(r,e,i,i.componentIndex)}}))}else for(var a=q(t)?i.get(t):J(t)?this.findComponents(t):null,s=0;a&&s<a.length;s++){var l=a[s];l&&e.call(n,l,l.componentIndex)}},e.prototype.getSeriesByName=function(t){var e=wo(t,null);return z(this._componentsMap.get("series"),(function(t){return!!t&&null!=e&&t.name===e}))},e.prototype.getSeriesByIndex=function(t){return this._componentsMap.get("series")[t]},e.prototype.getSeriesByType=function(t){return z(this._componentsMap.get("series"),(function(e){return!!e&&e.subType===t}))},e.prototype.getSeries=function(){return z(this._componentsMap.get("series").slice(),(function(t){return!!t}))},e.prototype.getSeriesCount=function(){return this._componentsCount.get("series")},e.prototype.eachSeries=function(t,e){tm(this),R(this._seriesIndices,(function(n){var i=this._componentsMap.get("series")[n];t.call(e,i,n)}),this)},e.prototype.eachRawSeries=function(t,e){R(this._componentsMap.get("series"),(function(n){n&&t.call(e,n,n.componentIndex)}))},e.prototype.eachSeriesByType=function(t,e,n){tm(this),R(this._seriesIndices,(function(i){var r=this._componentsMap.get("series")[i];r.subType===t&&e.call(n,r,i)}),this)},e.prototype.eachRawSeriesByType=function(t,e,n){return R(this.getSeriesByType(t),e,n)},e.prototype.isSeriesFiltered=function(t){return tm(this),null==this._seriesIndicesMap.get(t.componentIndex)},e.prototype.getCurrentSeriesIndices=function(){return(this._seriesIndices||[]).slice()},e.prototype.filterSeries=function(t,e){tm(this);var n=[];R(this._seriesIndices,(function(i){var r=this._componentsMap.get("series")[i];t.call(e,r,i)&&n.push(i)}),this),this._seriesIndices=n,this._seriesIndicesMap=gt(n)},e.prototype.restoreData=function(t){Qp(this);var e=this._componentsMap,n=[];e.each((function(t,e){xp.hasClass(e)&&n.push(e)})),xp.topologicalTravel(n,xp.getAllClassMainTypes(),(function(n){R(e.get(n),(function(e){!e||"series"===n&&sm(e,t)||e.restoreData()}))}))},e.internalField=function(){Qp=function(t){var e=t._seriesIndices=[];R(t._componentsMap.get("series"),(function(t){t&&e.push(t.componentIndex)})),t._seriesIndicesMap=gt(e)},tm=function(t){0},em=function(t,e){t.option={},t.option[rm]=om,t._componentsMap=gt({series:[]}),t._componentsCount=gt();var n=e.aria;J(n)&&null==n.enabled&&(n.enabled=!0),lm(e,t._theme.option),D(e,Sp,!1),t._mergeOption(e,null)}}(),e}(qh);function sm(t,e){if(e){var n=e.seriesIndex,i=e.seriesId,r=e.seriesName;return null!=n&&t.componentIndex!==n||null!=i&&t.id!==i||null!=r&&t.name!==r}}function lm(t,e){var n=t.color&&!t.colorLayer;R(e,(function(e,i){"colorLayer"===i&&n||xp.hasClass(i)||("object"===typeof e?t[i]=t[i]?D(t[i],e,!1):T(e):null==t[i]&&(t[i]=e))}))}function um(t,e,n){if($(e)){var i=gt();return R(e,(function(t){if(null!=t){var e=wo(t,null);null!=e&&i.set(t,!0)}})),z(n,(function(e){return e&&i.get(e[t])}))}var r=wo(e,null);return z(n,(function(e){return e&&null!=r&&e[t]===r}))}function cm(t,e){return e.hasOwnProperty("subType")?z(t,(function(t){return t&&t.subType===e.subType})):t}function dm(t){var e=gt();return t&&R(so(t.replaceMerge),(function(t){e.set(t,!0)})),{replaceMergeMainTypeMap:e}}E(am,qp);var hm=am,fm=["getDom","getZr","getWidth","getHeight","getDevicePixelRatio","dispatchAction","isDisposed","on","off","getDataURL","getConnectedDataURL","getOption","getId","updateLabelLayout"],pm=function(){function t(t){R(fm,(function(e){this[e]=V(t[e],t)}),this)}return t}(),mm=pm,gm={},ym=function(){function t(){this._coordinateSystems=[]}return t.prototype.create=function(t,e){var n=[];R(gm,(function(i,r){var o=i.create(t,e);n=n.concat(o||[])})),this._coordinateSystems=n},t.prototype.update=function(t,e){R(this._coordinateSystems,(function(n){n.update&&n.update(t,e)}))},t.prototype.getCoordinateSystems=function(){return this._coordinateSystems.slice()},t.register=function(t,e){gm[t]=e},t.get=function(t){return gm[t]},t}(),vm=ym,_m=/^(min|max)?(.+)$/,bm=function(){function t(t){this._timelineOptions=[],this._mediaList=[],this._currentMediaIndices=[],this._api=t}return t.prototype.setOption=function(t,e,n){t&&(R(so(t.series),(function(t){t&&t.data&&Q(t.data)&&ft(t.data)})),R(so(t.dataset),(function(t){t&&t.source&&Q(t.source)&&ft(t.source)}))),t=T(t);var i=this._optionBackup,r=xm(t,e,!i);this._newBaseOption=r.baseOption,i?(r.timelineOptions.length&&(i.timelineOptions=r.timelineOptions),r.mediaList.length&&(i.mediaList=r.mediaList),r.mediaDefault&&(i.mediaDefault=r.mediaDefault)):this._optionBackup=r},t.prototype.mountOption=function(t){var e=this._optionBackup;return this._timelineOptions=e.timelineOptions,this._mediaList=e.mediaList,this._mediaDefault=e.mediaDefault,this._currentMediaIndices=[],T(t?e.baseOption:this._newBaseOption)},t.prototype.getTimelineOption=function(t){var e,n=this._timelineOptions;if(n.length){var i=t.getComponent("timeline");i&&(e=T(n[i.getCurrentIndex()]))}return e},t.prototype.getMediaOption=function(t){var e=this._api.getWidth(),n=this._api.getHeight(),i=this._mediaList,r=this._mediaDefault,o=[],a=[];if(!i.length&&!r)return a;for(var s=0,l=i.length;s<l;s++)wm(i[s].query,e,n)&&o.push(s);return!o.length&&r&&(o=[-1]),o.length&&!Sm(o,this._currentMediaIndices)&&(a=H(o,(function(t){return T(-1===t?r.option:i[t].option)}))),this._currentMediaIndices=o,a},t}();function xm(t,e,n){var i,r,o=[],a=t.baseOption,s=t.timeline,l=t.options,u=t.media,c=!!t.media,d=!!(l||s||a&&a.timeline);function h(t){R(e,(function(e){e(t,n)}))}return a?(r=a,r.timeline||(r.timeline=s)):((d||c)&&(t.options=t.media=null),r=t),c&&$(u)&&R(u,(function(t){t&&t.option&&(t.query?o.push(t):i||(i=t))})),h(r),R(l,(function(t){return h(t)})),R(o,(function(t){return h(t.option)})),{baseOption:r,timelineOptions:l||[],mediaDefault:i,mediaList:o}}function wm(t,e,n){var i={width:e,height:n,aspectratio:e/n},r=!0;return R(t,(function(t,e){var n=e.match(_m);if(n&&n[1]&&n[2]){var o=n[1],a=n[2].toLowerCase();Mm(i[a],t,o)||(r=!1)}})),r}function Mm(t,e,n){return"min"===n?t>=e:"max"===n?t<=e:t===e}function Sm(t,e){return t.join(",")===e.join(",")}var Lm=bm,km=R,Tm=J,Dm=["areaStyle","lineStyle","nodeStyle","linkStyle","chordStyle","label","labelLine"];function Cm(t){var e=t&&t.itemStyle;if(e)for(var n=0,i=Dm.length;n<i;n++){var r=Dm[n],o=e.normal,a=e.emphasis;o&&o[r]&&(t[r]=t[r]||{},t[r].normal?D(t[r].normal,o[r]):t[r].normal=o[r],o[r]=null),a&&a[r]&&(t[r]=t[r]||{},t[r].emphasis?D(t[r].emphasis,a[r]):t[r].emphasis=a[r],a[r]=null)}}function Im(t,e,n){if(t&&t[e]&&(t[e].normal||t[e].emphasis)){var i=t[e].normal,r=t[e].emphasis;i&&(n?(t[e].normal=t[e].emphasis=null,A(t[e],i)):t[e]=i),r&&(t.emphasis=t.emphasis||{},t.emphasis[e]=r,r.focus&&(t.emphasis.focus=r.focus),r.blurScope&&(t.emphasis.blurScope=r.blurScope))}}function Am(t){Im(t,"itemStyle"),Im(t,"lineStyle"),Im(t,"areaStyle"),Im(t,"label"),Im(t,"labelLine"),Im(t,"upperLabel"),Im(t,"edgeLabel")}function Ym(t,e){var n=Tm(t)&&t[e],i=Tm(n)&&n.textStyle;if(i){0;for(var r=0,o=uo.length;r<o;r++){var a=uo[r];i.hasOwnProperty(a)&&(n[a]=i[a])}}}function Om(t){t&&(Am(t),Ym(t,"label"),t.emphasis&&Ym(t.emphasis,"label"))}function Pm(t){if(Tm(t)){Cm(t),Am(t),Ym(t,"label"),Ym(t,"upperLabel"),Ym(t,"edgeLabel"),t.emphasis&&(Ym(t.emphasis,"label"),Ym(t.emphasis,"upperLabel"),Ym(t.emphasis,"edgeLabel"));var e=t.markPoint;e&&(Cm(e),Om(e));var n=t.markLine;n&&(Cm(n),Om(n));var i=t.markArea;i&&Om(i);var r=t.data;if("graph"===t.type){r=r||t.nodes;var o=t.links||t.edges;if(o&&!Q(o))for(var a=0;a<o.length;a++)Om(o[a]);R(t.categories,(function(t){Am(t)}))}if(r&&!Q(r))for(a=0;a<r.length;a++)Om(r[a]);if(e=t.markPoint,e&&e.data){var s=e.data;for(a=0;a<s.length;a++)Om(s[a])}if(n=t.markLine,n&&n.data){var l=n.data;for(a=0;a<l.length;a++)$(l[a])?(Om(l[a][0]),Om(l[a][1])):Om(l[a])}"gauge"===t.type?(Ym(t,"axisLabel"),Ym(t,"title"),Ym(t,"detail")):"treemap"===t.type?(Im(t.breadcrumb,"itemStyle"),R(t.levels,(function(t){Am(t)}))):"tree"===t.type&&Am(t.leaves)}}function Em(t){return $(t)?t:t?[t]:[]}function Nm(t){return($(t)?t[0]:t)||{}}function Rm(t,e){km(Em(t.series),(function(t){Tm(t)&&Pm(t)}));var n=["xAxis","yAxis","radiusAxis","angleAxis","singleAxis","parallelAxis","radar"];e&&n.push("valueAxis","categoryAxis","logAxis","timeAxis"),km(n,(function(e){km(Em(t[e]),(function(t){t&&(Ym(t,"axisLabel"),Ym(t.axisPointer,"label"))}))})),km(Em(t.parallel),(function(t){var e=t&&t.parallelAxisDefault;Ym(e,"axisLabel"),Ym(e&&e.axisPointer,"label")})),km(Em(t.calendar),(function(t){Im(t,"itemStyle"),Ym(t,"dayLabel"),Ym(t,"monthLabel"),Ym(t,"yearLabel")})),km(Em(t.radar),(function(t){Ym(t,"name"),t.name&&null==t.axisName&&(t.axisName=t.name,delete t.name),null!=t.nameGap&&null==t.axisNameGap&&(t.axisNameGap=t.nameGap,delete t.nameGap)})),km(Em(t.geo),(function(t){Tm(t)&&(Om(t),km(Em(t.regions),(function(t){Om(t)})))})),km(Em(t.timeline),(function(t){Om(t),Im(t,"label"),Im(t,"itemStyle"),Im(t,"controlStyle",!0);var e=t.data;$(e)&&R(e,(function(t){J(t)&&(Im(t,"label"),Im(t,"itemStyle"))}))})),km(Em(t.toolbox),(function(t){Im(t,"iconStyle"),km(t.feature,(function(t){Im(t,"iconStyle")}))})),Ym(Nm(t.axisPointer),"label"),Ym(Nm(t.tooltip).axisPointer,"label")}function Hm(t,e){for(var n=e.split(","),i=t,r=0;r<n.length;r++)if(i=i&&i[n[r]],null==i)break;return i}function jm(t,e,n,i){for(var r,o=e.split(","),a=t,s=0;s<o.length-1;s++)r=o[s],null==a[r]&&(a[r]={}),a=a[r];(i||null==a[o[s]])&&(a[o[s]]=n)}function zm(t){t&&R(Wm,(function(e){e[0]in t&&!(e[1]in t)&&(t[e[1]]=t[e[0]])}))}var Wm=[["x","left"],["y","top"],["x2","right"],["y2","bottom"]],Bm=["grid","geo","parallel","legend","toolbox","title","visualMap","dataZoom","timeline"],Fm=[["borderRadius","barBorderRadius"],["borderColor","barBorderColor"],["borderWidth","barBorderWidth"]];function Vm(t){var e=t&&t.itemStyle;if(e)for(var n=0;n<Fm.length;n++){var i=Fm[n][1],r=Fm[n][0];null!=e[i]&&(e[r]=e[i])}}function Gm(t){t&&"edge"===t.alignTo&&null!=t.margin&&null==t.edgeDistance&&(t.edgeDistance=t.margin)}function $m(t){t&&t.downplay&&!t.blur&&(t.blur=t.downplay)}function Um(t){t&&null!=t.focusNodeAdjacency&&(t.emphasis=t.emphasis||{},null==t.emphasis.focus&&(t.emphasis.focus="adjacency"))}function qm(t,e){if(t)for(var n=0;n<t.length;n++)e(t[n]),t[n]&&qm(t[n].children,e)}function Xm(t,e){Rm(t,e),t.series=so(t.series),R(t.series,(function(t){if(J(t)){var e=t.type;if("line"===e)null!=t.clipOverflow&&(t.clip=t.clipOverflow);else if("pie"===e||"gauge"===e){null!=t.clockWise&&(t.clockwise=t.clockWise),Gm(t.label);var n=t.data;if(n&&!Q(n))for(var i=0;i<n.length;i++)Gm(n[i]);null!=t.hoverOffset&&(t.emphasis=t.emphasis||{},(t.emphasis.scaleSize=null)&&(t.emphasis.scaleSize=t.hoverOffset))}else if("gauge"===e){var r=Hm(t,"pointer.color");null!=r&&jm(t,"itemStyle.color",r)}else if("bar"===e){Vm(t),Vm(t.backgroundStyle),Vm(t.emphasis);n=t.data;if(n&&!Q(n))for(i=0;i<n.length;i++)"object"===typeof n[i]&&(Vm(n[i]),Vm(n[i]&&n[i].emphasis))}else if("sunburst"===e){var o=t.highlightPolicy;o&&(t.emphasis=t.emphasis||{},t.emphasis.focus||(t.emphasis.focus=o)),$m(t),qm(t.data,$m)}else"graph"===e||"sankey"===e?Um(t):"map"===e&&(t.mapType&&!t.map&&(t.map=t.mapType),t.mapLocation&&A(t,t.mapLocation));null!=t.hoverAnimation&&(t.emphasis=t.emphasis||{},t.emphasis&&null==t.emphasis.scale&&(t.emphasis.scale=t.hoverAnimation)),zm(t)}})),t.dataRange&&(t.visualMap=t.dataRange),R(Bm,(function(e){var n=t[e];n&&($(n)||(n=[n]),R(n,(function(t){zm(t)})))}))}function Zm(t){var e=gt();t.eachSeries((function(t){var n=t.get("stack");if(n){var i=e.get(n)||e.set(n,[]),r=t.getData(),o={stackResultDimension:r.getCalculationInfo("stackResultDimension"),stackedOverDimension:r.getCalculationInfo("stackedOverDimension"),stackedDimension:r.getCalculationInfo("stackedDimension"),stackedByDimension:r.getCalculationInfo("stackedByDimension"),isStackedByIndex:r.getCalculationInfo("isStackedByIndex"),data:r,seriesModel:t};if(!o.stackedDimension||!o.isStackedByIndex&&!o.stackedByDimension)return;i.length&&r.setCalculationInfo("stackedOnSeries",i[i.length-1].seriesModel),i.push(o)}})),e.each(Jm)}function Jm(t){R(t,(function(e,n){var i=[],r=[NaN,NaN],o=[e.stackResultDimension,e.stackedOverDimension],a=e.data,s=e.isStackedByIndex,l=a.map(o,(function(o,l,u){var c,d,h=a.get(e.stackedDimension,u);if(isNaN(h))return r;s?d=a.getRawIndex(u):c=a.get(e.stackedByDimension,u);for(var f=NaN,p=n-1;p>=0;p--){var m=t[p];if(s||(d=m.data.rawIndexOf(m.stackedByDimension,c)),d>=0){var g=m.data.getByRawIndex(m.stackResultDimension,d);if(h>=0&&g>0||h<=0&&g<0){h=Fr(h,g),f=g;break}}}return i[0]=h,i[1]=f,i}));a.hostModel.setData(l),e.data=l}))}var Km,Qm,tg,eg,ng,ig=function(){function t(t){this.data=t.data||(t.sourceFormat===Cp?{}:[]),this.sourceFormat=t.sourceFormat||Ap,this.seriesLayoutBy=t.seriesLayoutBy||Yp,this.startIndex=t.startIndex||0,this.dimensionsDefine=t.dimensionsDefine,this.dimensionsDetectedCount=t.dimensionsDetectedCount,this.encodeDefine=t.encodeDefine,this.metaRawOption=t.metaRawOption}return t}();function rg(t){return t instanceof ig}function og(t,e,n,i){n=n||ug(t);var r=e.seriesLayoutBy,o=cg(t,n,r,e.sourceHeader,e.dimensions),a=new ig({data:t,sourceFormat:n,seriesLayoutBy:r,dimensionsDefine:o.dimensionsDefine,startIndex:o.startIndex,dimensionsDetectedCount:o.dimensionsDetectedCount,encodeDefine:lg(i),metaRawOption:T(e)});return a}function ag(t){return new ig({data:t,sourceFormat:Q(t)?Ip:kp})}function sg(t){return new ig({data:t.data,sourceFormat:t.sourceFormat,seriesLayoutBy:t.seriesLayoutBy,dimensionsDefine:T(t.dimensionsDefine),startIndex:t.startIndex,dimensionsDetectedCount:t.dimensionsDetectedCount,encodeDefine:lg(t.encodeDefine)})}function lg(t){return t?gt(t):null}function ug(t){var e=Ap;if(Q(t))e=Ip;else if($(t)){0===t.length&&(e=Tp);for(var n=0,i=t.length;n<i;n++){var r=t[n];if(null!=r){if($(r)){e=Tp;break}if(J(r)){e=Dp;break}}}}else if(J(t))for(var o in t)if(_t(t,o)&&N(t[o])){e=Cp;break}return e}function cg(t,e,n,i,r){var o,a;if(!t)return{dimensionsDefine:hg(r),startIndex:a,dimensionsDetectedCount:o};if(e===Tp){var s=t;"auto"===i||null==i?fg((function(t){null!=t&&"-"!==t&&(q(t)?null==a&&(a=1):a=0)}),n,s,10):a=Z(i)?i:i?1:0,r||1!==a||(r=[],fg((function(t,e){r[e]=null!=t?t+"":""}),n,s,1/0)),o=r?r.length:n===Op?s.length:s[0]?s[0].length:null}else if(e===Dp)r||(r=dg(t));else if(e===Cp)r||(r=[],R(t,(function(t,e){r.push(e)})));else if(e===kp){var l=co(t[0]);o=$(l)&&l.length||1}return{startIndex:a,dimensionsDefine:hg(r),dimensionsDetectedCount:o}}function dg(t){var e,n=0;while(n<t.length&&!(e=t[n++]));if(e){var i=[];return R(e,(function(t,e){i.push(e)})),i}}function hg(t){if(t){var e=gt();return H(t,(function(t,n){t=J(t)?t:{name:t};var i={name:t.name,displayName:t.displayName,type:t.type};if(null==i.name)return i;i.name+="",null==i.displayName&&(i.displayName=i.name);var r=e.get(i.name);return r?i.name+="-"+r.count++:e.set(i.name,{count:1}),i}))}}function fg(t,e,n,i){if(e===Op)for(var r=0;r<n.length&&r<i;r++)t(n[r]?n[r][0]:null,r);else{var o=n[0]||[];for(r=0;r<o.length&&r<i;r++)t(o[r],r)}}var pg=function(){function t(t,e){var n=rg(t)?t:ag(t);this._source=n;var i=this._data=n.data;n.sourceFormat===Ip&&(this._offset=0,this._dimSize=e,this._data=i),ng(this,i,n)}return t.prototype.getSource=function(){return this._source},t.prototype.count=function(){return 0},t.prototype.getItem=function(t,e){},t.prototype.appendData=function(t){},t.prototype.clean=function(){},t.protoInitialize=function(){var e=t.prototype;e.pure=!1,e.persistent=!0}(),t.internalField=function(){var t;ng=function(t,r,o){var a=o.sourceFormat,s=o.seriesLayoutBy,l=o.startIndex,u=o.dimensionsDefine,c=eg[Sg(a,s)];if(I(t,c),a===Ip)t.getItem=e,t.count=i,t.fillStorage=n;else{var d=yg(a,s);t.getItem=V(d,null,r,l,u);var h=bg(a,s);t.count=V(h,null,r,l,u)}};var e=function(t,e){t-=this._offset,e=e||[];for(var n=this._data,i=this._dimSize,r=i*t,o=0;o<i;o++)e[o]=n[r+o];return e},n=function(t,e,n,i){for(var r=this._data,o=this._dimSize,a=0;a<o;a++){for(var s=i[a],l=null==s[0]?1/0:s[0],u=null==s[1]?-1/0:s[1],c=e-t,d=n[a],h=0;h<c;h++){var f=r[h*o+a];d[t+h]=f,f<l&&(l=f),f>u&&(u=f)}s[0]=l,s[1]=u}},i=function(){return this._data?this._data.length/this._dimSize:0};function r(t){for(var e=0;e<t.length;e++)this._data.push(t[e])}t={},t[Tp+"_"+Yp]={pure:!0,appendData:r},t[Tp+"_"+Op]={pure:!0,appendData:function(){throw new Error('Do not support appendData when set seriesLayoutBy: "row".')}},t[Dp]={pure:!0,appendData:r},t[Cp]={pure:!0,appendData:function(t){var e=this._data;R(t,(function(t,n){for(var i=e[n]||(e[n]=[]),r=0;r<(t||[]).length;r++)i.push(t[r])}))}},t[kp]={appendData:r},t[Ip]={persistent:!1,pure:!0,appendData:function(t){this._data=t},clean:function(){this._offset+=this.count(),this._data=null}},eg=t}(),t}(),mg=function(t,e,n,i){return t[i]},gg=(Km={},Km[Tp+"_"+Yp]=function(t,e,n,i){return t[i+e]},Km[Tp+"_"+Op]=function(t,e,n,i){i+=e;for(var r=[],o=t,a=0;a<o.length;a++){var s=o[a];r.push(s?s[i]:null)}return r},Km[Dp]=mg,Km[Cp]=function(t,e,n,i){for(var r=[],o=0;o<n.length;o++){var a=n[o].name;0;var s=t[a];r.push(s?s[i]:null)}return r},Km[kp]=mg,Km);function yg(t,e){var n=gg[Sg(t,e)];return n}var vg=function(t,e,n){return t.length},_g=(Qm={},Qm[Tp+"_"+Yp]=function(t,e,n){return Math.max(0,t.length-e)},Qm[Tp+"_"+Op]=function(t,e,n){var i=t[0];return i?Math.max(0,i.length-e):0},Qm[Dp]=vg,Qm[Cp]=function(t,e,n){var i=n[0].name;var r=t[i];return r?r.length:0},Qm[kp]=vg,Qm);function bg(t,e){var n=_g[Sg(t,e)];return n}var xg=function(t,e,n){return null!=e?t[e]:t},wg=(tg={},tg[Tp]=xg,tg[Dp]=function(t,e,n){return null!=e?t[n]:t},tg[Cp]=xg,tg[kp]=function(t,e,n){var i=co(t);return null!=e&&i instanceof Array?i[e]:i},tg[Ip]=xg,tg);function Mg(t){var e=wg[t];return e}function Sg(t,e){return t===Tp?t+"_"+e:t}function Lg(t,e,n){if(t){var i=t.getRawDataItem(e);if(null!=i){var r,o,a=t.getProvider().getSource().sourceFormat,s=t.getDimensionInfo(n);return s&&(r=s.name,o=s.index),Mg(a)(i,o,r)}}}var kg=/\{@(.+?)\}/g,Tg=function(){function t(){}return t.prototype.getDataParams=function(t,e){var n=this.getData(e),i=this.getRawValue(t,e),r=n.getRawIndex(t),o=n.getName(t),a=n.getRawDataItem(t),s=n.getItemVisual(t,"style"),l=s&&s[n.getItemVisual(t,"drawType")||"fill"],u=s&&s.stroke,c=this.mainType,d="series"===c,h=n.userOutput;return{componentType:c,componentSubType:this.subType,componentIndex:this.componentIndex,seriesType:d?this.subType:null,seriesIndex:this.seriesIndex,seriesId:d?this.id:null,seriesName:d?this.name:null,name:o,dataIndex:r,data:a,dataType:e,value:i,color:l,borderColor:u,dimensionNames:h?h.dimensionNames:null,encode:h?h.encode:null,$vars:["seriesName","name","value"]}},t.prototype.getFormattedLabel=function(t,e,n,i,r,o){e=e||"normal";var a=this.getData(n),s=this.getDataParams(t,n);if(o&&(s.value=o.interpolatedValue),null!=i&&$(s.value)&&(s.value=s.value[i]),!r){var l=a.getItemModel(t);r=l.get("normal"===e?["label","formatter"]:[e,"label","formatter"])}if("function"===typeof r)return s.status=e,s.dimensionIndex=i,r(s);if("string"===typeof r){var u=tp(r,s);return u.replace(kg,(function(e,n){var i=n.length,r="["===n.charAt(0)&&"]"===n.charAt(i-1)?+n.slice(1,i-1):n,s=Lg(a,t,r);if(o&&$(o.interpolatedValue)){var l=a.getDimensionInfo(r);l&&(s=o.interpolatedValue[l.index])}return null!=s?s+"":""}))}},t.prototype.getRawValue=function(t,e){return Lg(this.getData(e),t)},t.prototype.formatTooltip=function(t,e,n){},t}();function Dg(t){var e,n;return J(t)?t.type&&(n=t):e=t,{markupText:e,markupFragment:n}}function Cg(t){return new Ig(t)}var Ig=function(){function t(t){t=t||{},this._reset=t.reset,this._plan=t.plan,this._count=t.count,this._onDirty=t.onDirty,this._dirty=!0}return t.prototype.perform=function(t){var e,n=this._upstream,i=t&&t.skip;if(this._dirty&&n){var r=this.context;r.data=r.outputData=n.context.outputData}this.__pipeline&&(this.__pipeline.currentTask=this),this._plan&&!i&&(e=this._plan(this.context));var o,a=c(this._modBy),s=this._modDataCount||0,l=c(t&&t.modBy),u=t&&t.modDataCount||0;function c(t){return!(t>=1)&&(t=1),t}a===l&&s===u||(e="reset"),(this._dirty||"reset"===e)&&(this._dirty=!1,o=this._doReset(i)),this._modBy=l,this._modDataCount=u;var d=t&&t.step;if(this._dueEnd=n?n._outputDueEnd:this._count?this._count(this.context):1/0,this._progress){var h=this._dueIndex,f=Math.min(null!=d?this._dueIndex+d:1/0,this._dueEnd);if(!i&&(o||h<f)){var p=this._progress;if($(p))for(var m=0;m<p.length;m++)this._doProgress(p[m],h,f,l,u);else this._doProgress(p,h,f,l,u)}this._dueIndex=f;var g=null!=this._settedOutputEnd?this._settedOutputEnd:f;0,this._outputDueEnd=g}else this._dueIndex=this._outputDueEnd=null!=this._settedOutputEnd?this._settedOutputEnd:this._dueEnd;return this.unfinished()},t.prototype.dirty=function(){this._dirty=!0,this._onDirty&&this._onDirty(this.context)},t.prototype._doProgress=function(t,e,n,i,r){Ag.reset(e,n,i,r),this._callingProgress=t,this._callingProgress({start:e,end:n,count:n-e,next:Ag.next},this.context)},t.prototype._doReset=function(t){var e,n;this._dueIndex=this._outputDueEnd=this._dueEnd=0,this._settedOutputEnd=null,!t&&this._reset&&(e=this._reset(this.context),e&&e.progress&&(n=e.forceFirstProgress,e=e.progress),$(e)&&!e.length&&(e=null)),this._progress=e,this._modBy=this._modDataCount=null;var i=this._downstream;return i&&i.dirty(),n},t.prototype.unfinished=function(){return this._progress&&this._dueIndex<this._dueEnd},t.prototype.pipe=function(t){(this._downstream!==t||this._dirty)&&(this._downstream=t,t._upstream=this,t.dirty())},t.prototype.dispose=function(){this._disposed||(this._upstream&&(this._upstream._downstream=null),this._downstream&&(this._downstream._upstream=null),this._dirty=!1,this._disposed=!0)},t.prototype.getUpstream=function(){return this._upstream},t.prototype.getDownstream=function(){return this._downstream},t.prototype.setOutputEnd=function(t){this._outputDueEnd=this._settedOutputEnd=t},t}(),Ag=function(){var t,e,n,i,r,o={reset:function(l,u,c,d){e=l,t=u,n=c,i=d,r=Math.ceil(i/n),o.next=n>1&&i>0?s:a}};return o;function a(){return e<t?e++:null}function s(){var o=e%r*n+Math.ceil(e/r),a=e>=t?null:o<i?o:e;return e++,a}}();function Yg(t,e){var n=e&&e.type;if("ordinal"===n){var i=e&&e.ordinalMeta;return i?i.parseAndCollect(t):t}return"time"===n&&"number"!==typeof t&&null!=t&&"-"!==t&&(t=+qr(t)),null==t||""===t?NaN:+t}var Og=gt({number:function(t){return parseFloat(t)},time:function(t){return+qr(t)},trim:function(t){return"string"===typeof t?dt(t):t}});function Pg(t){return Og.get(t)}var Eg={lt:function(t,e){return t<e},lte:function(t,e){return t<=e},gt:function(t,e){return t>e},gte:function(t,e){return t>=e}},Ng=function(){function t(t,e){if("number"!==typeof e){var n="";0,im(n)}this._opFn=Eg[t],this._rvalFloat=to(e)}return t.prototype.evaluate=function(t){return"number"===typeof t?this._opFn(t,this._rvalFloat):this._opFn(to(t),this._rvalFloat)},t}(),Rg=function(){function t(t,e){var n="desc"===t;this._resultLT=n?1:-1,null==e&&(e=n?"min":"max"),this._incomparable="min"===e?-1/0:1/0}return t.prototype.evaluate=function(t,e){var n=typeof t,i=typeof e,r="number"===n?t:to(t),o="number"===i?e:to(e),a=isNaN(r),s=isNaN(o);if(a&&(r=this._incomparable),s&&(o=this._incomparable),a&&s){var l="string"===n,u="string"===i;l&&(r=u?t:0),u&&(o=l?e:0)}return r<o?this._resultLT:r>o?-this._resultLT:0},t}(),Hg=function(){function t(t,e){this._rval=e,this._isEQ=t,this._rvalTypeof=typeof e,this._rvalFloat=to(e)}return t.prototype.evaluate=function(t){var e=t===this._rval;if(!e){var n=typeof t;n===this._rvalTypeof||"number"!==n&&"number"!==this._rvalTypeof||(e=to(t)===this._rvalFloat)}return this._isEQ?e:!e},t}();function jg(t,e){return"eq"===t||"ne"===t?new Hg("eq"===t,e):_t(Eg,t)?new Ng(t,e):null}var zg=function(){function t(){}return t.prototype.getRawData=function(){throw new Error("not supported")},t.prototype.getRawDataItem=function(t){throw new Error("not supported")},t.prototype.cloneRawData=function(){},t.prototype.getDimensionInfo=function(t){},t.prototype.cloneAllDimensionInfo=function(){},t.prototype.count=function(){},t.prototype.retrieveValue=function(t,e){},t.prototype.retrieveValueFromItem=function(t,e){},t.prototype.convertValue=function(t,e){return Yg(t,e)},t}();function Wg(t,e){var n=new zg,i=t.data,r=n.sourceFormat=t.sourceFormat,o=t.startIndex,a="";t.seriesLayoutBy!==Yp&&im(a);var s=[],l={},u=t.dimensionsDefine;if(u)R(u,(function(t,e){var n=t.name,i={index:e,name:n,displayName:t.displayName};if(s.push(i),null!=n){var r="";_t(l,n)&&im(r),l[n]=i}}));else for(var c=0;c<t.dimensionsDetectedCount;c++)s.push({index:c});var d=yg(r,Yp);e.__isBuiltIn&&(n.getRawDataItem=function(t){return d(i,o,s,t)},n.getRawData=V(Bg,null,t)),n.cloneRawData=V(Fg,null,t);var h=bg(r,Yp);n.count=V(h,null,i,o,s);var f=Mg(r);n.retrieveValue=function(t,e){var n=d(i,o,s,t);return p(n,e)};var p=n.retrieveValueFromItem=function(t,e){if(null!=t){var n=s[e];return n?f(t,e,n.name):void 0}};return n.getDimensionInfo=V(Vg,null,s,l),n.cloneAllDimensionInfo=V(Gg,null,s),n}function Bg(t){var e=t.sourceFormat;if(!Zg(e)){var n="";0,im(n)}return t.data}function Fg(t){var e=t.sourceFormat,n=t.data;if(!Zg(e)){var i="";0,im(i)}if(e===Tp){for(var r=[],o=0,a=n.length;o<a;o++)r.push(n[o].slice());return r}if(e===Dp){for(r=[],o=0,a=n.length;o<a;o++)r.push(I({},n[o]));return r}}function Vg(t,e,n){if(null!=n)return"number"===typeof n||!isNaN(n)&&!_t(e,n)?t[n]:_t(e,n)?e[n]:void 0}function Gg(t){return T(t)}var $g=gt();function Ug(t){t=T(t);var e=t.type,n="";e||im(n);var i=e.split(":");2!==i.length&&im(n);var r=!1;"echarts"===i[0]&&(e=i[1],r=!0),t.__isBuiltIn=r,$g.set(e,t)}function qg(t,e,n){var i=so(t),r=i.length,o="";r||im(o);for(var a=0,s=r;a<s;a++){var l=i[a];e=Xg(l,e,n,1===r?null:a),a!==s-1&&(e.length=Math.max(e.length,1))}return e}function Xg(t,e,n,i){var r="";e.length||im(r),J(t)||im(r);var o=t.type,a=$g.get(o);a||im(r);var s=H(e,(function(t){return Wg(t,a)})),l=so(a.transform({upstream:s[0],upstreamList:s,config:T(t.config)}));return H(l,(function(t,n){var i="";J(t)||im(i),t.data||im(i);var r,o=ug(t.data);Zg(o)||im(i);var a=e[0];if(a&&0===n&&!t.dimensions){var s=a.startIndex;s&&(t.data=a.data.slice(0,s).concat(t.data)),r={seriesLayoutBy:Yp,sourceHeader:s,dimensions:a.metaRawOption.dimensions}}else r={seriesLayoutBy:Yp,sourceHeader:0,dimensions:t.dimensions};return og(t.data,r,null,null)}))}function Zg(t){return t===Tp||t===Dp}var Jg=function(){function t(t){this._sourceList=[],this._upstreamSignList=[],this._versionSignBase=0,this._sourceHost=t}return t.prototype.dirty=function(){this._setLocalSource([],[])},t.prototype._setLocalSource=function(t,e){this._sourceList=t,this._upstreamSignList=e,this._versionSignBase++,this._versionSignBase>9e10&&(this._versionSignBase=0)},t.prototype._getVersionSign=function(){return this._sourceHost.uid+"_"+this._versionSignBase},t.prototype.prepareSource=function(){this._isDirty()&&this._createSource()},t.prototype._createSource=function(){this._setLocalSource([],[]);var t,e,n=this._sourceHost,i=this._getUpstreamSourceManagers(),r=!!i.length;if(Qg(n)){var o=n,a=void 0,s=void 0,l=void 0;if(r){var u=i[0];u.prepareSource(),l=u.getSource(),a=l.data,s=l.sourceFormat,e=[u._getVersionSign()]}else a=o.get("data",!0),s=Q(a)?Ip:kp,e=[];var c=this._getSourceMetaRawOption(),d=l?l.metaRawOption:null,h=at(c.seriesLayoutBy,d?d.seriesLayoutBy:null),f=at(c.sourceHeader,d?d.sourceHeader:null),p=at(c.dimensions,d?d.dimensions:null);t=[og(a,{seriesLayoutBy:h,sourceHeader:f,dimensions:p},s,o.get("encode",!0))]}else{var m=n;if(r){var g=this._applyTransform(i);t=g.sourceList,e=g.upstreamSignList}else{var y=m.get("source",!0);t=[og(y,this._getSourceMetaRawOption(),null,null)],e=[]}}this._setLocalSource(t,e)},t.prototype._applyTransform=function(t){var e,n=this._sourceHost,i=n.get("transform",!0),r=n.get("fromTransformResult",!0);if(null!=r){var o="";1!==t.length&&ty(o)}var a=[],s=[];return R(t,(function(t){t.prepareSource();var e=t.getSource(r||0),n="";null==r||e||ty(n),a.push(e),s.push(t._getVersionSign())})),i?e=qg(i,a,{datasetIndex:n.componentIndex}):null!=r&&(e=[sg(a[0])]),{sourceList:e,upstreamSignList:s}},t.prototype._isDirty=function(){var t=this._sourceList;if(!t.length)return!0;for(var e=this._getUpstreamSourceManagers(),n=0;n<e.length;n++){var i=e[n];if(i._isDirty()||this._upstreamSignList[n]!==i._getVersionSign())return!0}},t.prototype.getSource=function(t){return this._sourceList[t||0]},t.prototype._getUpstreamSourceManagers=function(){var t=this._sourceHost;if(Qg(t)){var e=jp(t);return e?[e.getSourceManager()]:[]}return H(zp(t),(function(t){return t.getSourceManager()}))},t.prototype._getSourceMetaRawOption=function(){var t,e,n,i=this._sourceHost;if(Qg(i))t=i.get("seriesLayoutBy",!0),e=i.get("sourceHeader",!0),n=i.get("dimensions",!0);else if(!this._getUpstreamSourceManagers().length){var r=i;t=r.get("seriesLayoutBy",!0),e=r.get("sourceHeader",!0),n=r.get("dimensions",!0)}return{seriesLayoutBy:t,sourceHeader:e,dimensions:n}},t}();function Kg(t){var e=t.option.transform;e&&ft(t.option.transform)}function Qg(t){return"series"===t.mainType}function ty(t){throw new Error(t)}var ey="line-height:1";function ny(t,e){var n=t.color||"#6e7079",i=t.fontSize||12,r=t.fontWeight||"400",o=t.color||"#464646",a=t.fontSize||14,s=t.fontWeight||"900";return"html"===e?{nameStyle:"font-size:"+Zf(i+"")+"px;color:"+Zf(n)+";font-weight:"+Zf(r+""),valueStyle:"font-size:"+Zf(a+"")+"px;color:"+Zf(o)+";font-weight:"+Zf(s+"")}:{nameStyle:{fontSize:i,fill:n,fontWeight:r},valueStyle:{fontSize:a,fill:o,fontWeight:s}}}var iy=[0,10,20,30],ry=["","\n","\n\n","\n\n\n"];function oy(t,e){return e.type=t,e}function ay(t){return _t(sy,t.type)&&sy[t.type]}var sy={section:{planLayout:function(t){var e=t.blocks.length,n=e>1||e>0&&!t.noHeader,i=0;R(t.blocks,(function(t){ay(t).planLayout(t);var e=t.__gapLevelBetweenSubBlocks;e>=i&&(i=e+(!n||e&&("section"!==t.type||t.noHeader)?0:1))})),t.__gapLevelBetweenSubBlocks=i},build:function(t,e,n,i){var r=e.noHeader,o=cy(e),a=ly(t,e,r?n:o.html,i);if(r)return a;var s=Jf(e.header,"ordinal",t.useUTC),l=ny(i,t.renderMode).nameStyle;return"richText"===t.renderMode?py(t,s,l)+o.richText+a:dy('<div style="'+l+";"+ey+';">'+Zf(s)+"</div>"+a,n)}},nameValue:{planLayout:function(t){t.__gapLevelBetweenSubBlocks=0},build:function(t,e,n,i){var r=t.renderMode,o=e.noName,a=e.noValue,s=!e.markerType,l=e.name,u=e.value,c=t.useUTC;if(!o||!a){var d=s?"":t.markupStyleCreator.makeTooltipMarker(e.markerType,e.markerColor||"#333",r),h=o?"":Jf(l,"ordinal",c),f=e.valueType,p=a?[]:$(u)?H(u,(function(t,e){return Jf(t,$(f)?f[e]:f,c)})):[Jf(u,$(f)?f[0]:f,c)],m=!s||!o,g=!s&&o,y=ny(i,r),v=y.nameStyle,_=y.valueStyle;return"richText"===r?(s?"":d)+(o?"":py(t,h,v))+(a?"":my(t,p,m,g,_)):dy((s?"":d)+(o?"":hy(h,!s,v))+(a?"":fy(p,m,g,_)),n)}}}};function ly(t,e,n,i){var r=[],o=e.blocks||[];ct(!o||$(o)),o=o||[];var a=t.orderMode;if(e.sortBlocks&&a){o=o.slice();var s={valueAsc:"asc",valueDesc:"desc"};if(_t(s,a)){var l=new Rg(s[a],null);o.sort((function(t,e){return l.evaluate(t.sortParam,e.sortParam)}))}else"seriesDesc"===a&&o.reverse()}var u=cy(e);if(R(o,(function(e,n){var o=ay(e).build(t,e,n>0?u.html:0,i);null!=o&&r.push(o)})),r.length)return"richText"===t.renderMode?r.join(u.richText):dy(r.join(""),n)}function uy(t,e,n,i,r,o){if(t){var a=ay(t);a.planLayout(t);var s={useUTC:r,renderMode:n,orderMode:i,markupStyleCreator:e};return a.build(s,t,0,o)}}function cy(t){var e=t.__gapLevelBetweenSubBlocks;return{html:iy[e],richText:ry[e]}}function dy(t,e){var n='<div style="clear:both"></div>',i="margin: "+e+"px 0 0";return'<div style="'+i+";"+ey+';">'+t+n+"</div>"}function hy(t,e,n){var i=e?"margin-left:2px":"";return'<span style="'+n+";"+i+'">'+Zf(t)+"</span>"}function fy(t,e,n,i){var r=n?"10px":"20px",o=e?"float:right;margin-left:"+r:"";return'<span style="'+o+";"+i+'">'+H(t,(function(t){return Zf(t)})).join("&nbsp;&nbsp;")+"</span>"}function py(t,e,n){return t.markupStyleCreator.wrapRichTextStyle(e,n)}function my(t,e,n,i,r){var o=[r],a=i?10:20;return n&&o.push({padding:[0,0,0,a],align:"right"}),t.markupStyleCreator.wrapRichTextStyle(e.join("  "),o)}function gy(t,e){var n=t.getData().getItemVisual(e,"style"),i=n[t.visualDrawType];return ip(i)}function yy(t,e){var n=t.get("padding");return null!=n?n:"richText"===e?[8,10]:10}var vy=function(){function t(){this.richTextStyles={},this._nextStyleNameId=no()}return t.prototype._generateStyleName=function(){return"__EC_aUTo_"+this._nextStyleNameId++},t.prototype.makeTooltipMarker=function(t,e,n){var i="richText"===n?this._generateStyleName():null,r=np({color:e,type:t,renderMode:n,markerId:i});return q(r)?r:(this.richTextStyles[i]=r.style,r.content)},t.prototype.wrapRichTextStyle=function(t,e){var n={};$(e)?R(e,(function(t){return I(n,t)})):I(n,e);var i=this._generateStyleName();return this.richTextStyles[i]=n,"{"+i+"|"+t+"}"},t}();function _y(t){var e,n,i,r,o=t.series,a=t.dataIndex,s=t.multipleSeries,l=o.getData(),u=l.mapDimensionsAll("defaultedTooltip"),c=u.length,d=o.getRawValue(a),h=$(d),f=gy(o,a);if(c>1||h&&!c){var p=by(d,o,a,u,f);e=p.inlineValues,n=p.inlineValueTypes,i=p.blocks,r=p.inlineValues[0]}else if(c){var m=l.getDimensionInfo(u[0]);r=e=Lg(l,a,u[0]),n=m.type}else r=e=h?d[0]:d;var g=Mo(o),y=g&&o.name||"",v=l.getName(a),_=s?y:v;return oy("section",{header:y,noHeader:s||!g,sortParam:r,blocks:[oy("nameValue",{markerType:"item",markerColor:f,name:_,noName:!dt(_),value:e,valueType:n})].concat(i||[])})}function by(t,e,n,i,r){var o=e.getData(),a=j(t,(function(t,e,n){var i=o.getDimensionInfo(n);return t||i&&!1!==i.tooltip&&null!=i.displayName}),!1),s=[],l=[],u=[];function c(t,e){var n=o.getDimensionInfo(e);n&&!1!==n.otherDims.tooltip&&(a?u.push(oy("nameValue",{markerType:"subItem",markerColor:r,name:n.displayName,value:t,valueType:n.type})):(s.push(t),l.push(n.type)))}return i.length?R(i,(function(t){c(Lg(o,n,t),t)})):R(t,c),{inlineValues:s,inlineValueTypes:l,blocks:u}}var xy=Io();function wy(t,e){return t.getName(e)||t.getId(e)}var My=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._selectedDataIndicesMap={},e}return o(e,t),e.prototype.init=function(t,e,n){this.seriesIndex=this.componentIndex,this.dataTask=Cg({count:ky,reset:Ty}),this.dataTask.context={model:this},this.mergeDefaultAndTheme(t,n);var i=xy(this).sourceManager=new Jg(this);i.prepareSource();var r=this.getInitialData(t,n);Cy(r,this),this.dataTask.context.data=r,xy(this).dataBeforeProcessed=r,Sy(this),this._initSelectedMapFromData(r)},e.prototype.mergeDefaultAndTheme=function(t,e){var n=pp(this),i=n?gp(t):{},r=this.subType;xp.hasClass(r)&&(r+="Series"),D(t,e.getTheme().get(this.subType)),D(t,this.getDefaultOption()),lo(t,"label",["show"]),this.fillDataTextStyle(t.data),n&&mp(t,i,n)},e.prototype.mergeOption=function(t,e){t=D(this.option,t,!0),this.fillDataTextStyle(t.data);var n=pp(this);n&&mp(this.option,t,n);var i=xy(this).sourceManager;i.dirty(),i.prepareSource();var r=this.getInitialData(t,e);Cy(r,this),this.dataTask.dirty(),this.dataTask.context.data=r,xy(this).dataBeforeProcessed=r,Sy(this),this._initSelectedMapFromData(r)},e.prototype.fillDataTextStyle=function(t){if(t&&!Q(t))for(var e=["show"],n=0;n<t.length;n++)t[n]&&t[n].label&&lo(t[n],"label",e)},e.prototype.getInitialData=function(t,e){},e.prototype.appendData=function(t){var e=this.getRawData();e.appendData(t.data)},e.prototype.getData=function(t){var e=Ay(this);if(e){var n=e.context.data;return null==t?n:n.getLinkedData(t)}return xy(this).data},e.prototype.getAllData=function(){var t=this.getData();return t&&t.getLinkedDataAll?t.getLinkedDataAll():[{data:t}]},e.prototype.setData=function(t){var e=Ay(this);if(e){var n=e.context;n.outputData=t,e!==this.dataTask&&(n.data=t)}xy(this).data=t},e.prototype.getSource=function(){return xy(this).sourceManager.getSource()},e.prototype.getRawData=function(){return xy(this).dataBeforeProcessed},e.prototype.getBaseAxis=function(){var t=this.coordinateSystem;return t&&t.getBaseAxis&&t.getBaseAxis()},e.prototype.formatTooltip=function(t,e,n){return _y({series:this,dataIndex:t,multipleSeries:e})},e.prototype.isAnimationEnabled=function(){if(h.node)return!1;var t=this.getShallow("animation");return t&&this.getData().count()>this.getShallow("animationThreshold")&&(t=!1),!!t},e.prototype.restoreData=function(){this.dataTask.dirty()},e.prototype.getColorFromPalette=function(t,e,n){var i=this.ecModel,r=qp.prototype.getColorFromPalette.call(this,t,e,n);return r||(r=i.getColorFromPalette(t,e,n)),r},e.prototype.coordDimToDataDim=function(t){return this.getRawData().mapDimensionsAll(t)},e.prototype.getProgressive=function(){return this.get("progressive")},e.prototype.getProgressiveThreshold=function(){return this.get("progressiveThreshold")},e.prototype.select=function(t,e){this._innerSelect(this.getData(e),t)},e.prototype.unselect=function(t,e){var n=this.option.selectedMap;if(n)for(var i=this.getData(e),r=0;r<t.length;r++){var o=t[r],a=wy(i,o);n[a]=!1,this._selectedDataIndicesMap[a]=-1}},e.prototype.toggleSelect=function(t,e){for(var n=[],i=0;i<t.length;i++)n[0]=t[i],this.isSelected(t[i],e)?this.unselect(n,e):this.select(n,e)},e.prototype.getSelectedDataIndices=function(){for(var t=this._selectedDataIndicesMap,e=B(t),n=[],i=0;i<e.length;i++){var r=t[e[i]];r>=0&&n.push(r)}return n},e.prototype.isSelected=function(t,e){var n=this.option.selectedMap;if(!n)return!1;var i=this.getData(e),r=wy(i,t);return n[r]||!1},e.prototype._innerSelect=function(t,e){var n,i,r=this.option.selectedMode,o=e.length;if(r&&o)if("multiple"===r)for(var a=this.option.selectedMap||(this.option.selectedMap={}),s=0;s<o;s++){var l=e[s],u=wy(t,l);a[u]=!0,this._selectedDataIndicesMap[u]=t.getRawIndex(l)}else if("single"===r||!0===r){var c=e[o-1];u=wy(t,c);this.option.selectedMap=(n={},n[u]=!0,n),this._selectedDataIndicesMap=(i={},i[u]=t.getRawIndex(c),i)}},e.prototype._initSelectedMapFromData=function(t){if(!this.option.selectedMap){var e=[];t.hasItemOption&&t.each((function(n){var i=t.getRawDataItem(n);i&&i.selected&&e.push(n)})),e.length>0&&this._innerSelect(t,e)}},e.registerClass=function(t){return xp.registerClass(t)},e.protoInitialize=function(){var t=e.prototype;t.type="series.__base__",t.seriesIndex=0,t.useColorPaletteOnData=!1,t.ignoreStyleOnData=!1,t.hasSymbolVisual=!1,t.defaultSymbol="circle",t.visualStyleAccessPath="itemStyle",t.visualDrawType="fill"}(),e}(xp);function Sy(t){var e=t.name;Mo(t)||(t.name=Ly(t)||e)}function Ly(t){var e=t.getRawData(),n=e.mapDimensionsAll("seriesName"),i=[];return R(n,(function(t){var n=e.getDimensionInfo(t);n.displayName&&i.push(n.displayName)})),i.join(" ")}function ky(t){return t.model.getRawData().count()}function Ty(t){var e=t.model;return e.setData(e.getRawData().cloneShallow()),Dy}function Dy(t,e){e.outputData&&t.end>e.outputData.count()&&e.model.getRawData().cloneShallow(e.outputData)}function Cy(t,e){R(s(t.CHANGABLE_METHODS,t.DOWNSAMPLE_METHODS),(function(n){t.wrapMethod(n,G(Iy,e))}))}function Iy(t,e){var n=Ay(t);return n&&n.setOutputEnd((e||this).count()),e}function Ay(t){var e=(t.ecModel||{}).scheduler,n=e&&e.getPipeline(t.uid);if(n){var i=n.currentTask;if(i){var r=i.agentStubMap;r&&(i=r.get(t.uid))}return i}}E(My,Tg),E(My,qp),Zo(My,xp);var Yy=My,Oy=function(){function t(){this.group=new Mr,this.uid=Zh("viewComponent")}return t.prototype.init=function(t,e){},t.prototype.render=function(t,e,n,i){},t.prototype.dispose=function(t,e){},t.prototype.updateView=function(t,e,n,i){},t.prototype.updateLayout=function(t,e,n,i){},t.prototype.updateVisual=function(t,e,n,i){},t.prototype.blurSeries=function(t,e){},t}();qo(Oy),ea(Oy);var Py=Oy;function Ey(){var t=Io();return function(e){var n=t(e),i=e.pipelineContext,r=!!n.large,o=!!n.progressiveRender,a=n.large=!(!i||!i.large),s=n.progressiveRender=!(!i||!i.progressiveRender);return!(r===a&&o===s)&&"reset"}}var Ny=Io(),Ry=Ey(),Hy=function(){function t(){this.group=new Mr,this.uid=Zh("viewChart"),this.renderTask=Cg({plan:Wy,reset:By}),this.renderTask.context={view:this}}return t.prototype.init=function(t,e){},t.prototype.render=function(t,e,n,i){},t.prototype.highlight=function(t,e,n,i){zy(t.getData(),i,"emphasis")},t.prototype.downplay=function(t,e,n,i){zy(t.getData(),i,"normal")},t.prototype.remove=function(t,e){this.group.removeAll()},t.prototype.dispose=function(t,e){},t.prototype.updateView=function(t,e,n,i){this.render(t,e,n,i)},t.prototype.updateLayout=function(t,e,n,i){this.render(t,e,n,i)},t.prototype.updateVisual=function(t,e,n,i){this.render(t,e,n,i)},t.markUpdateMethod=function(t,e){Ny(t).updateMethod=e},t.protoInitialize=function(){var e=t.prototype;e.type="chart"}(),t}();function jy(t,e,n){t&&("emphasis"===e?Cu:Iu)(t,n)}function zy(t,e,n){var i=Co(t,e),r=e&&null!=e.highlightKey?tc(e.highlightKey):null;null!=i?R(so(i),(function(e){jy(t.getItemGraphicEl(e),n,r)})):t.eachItemGraphicEl((function(t){jy(t,n,r)}))}function Wy(t){return Ry(t.model)}function By(t){var e=t.model,n=t.ecModel,i=t.api,r=t.payload,o=e.pipelineContext.progressiveRender,a=t.view,s=r&&Ny(r).updateMethod,l=o?"incrementalPrepareRender":s&&a[s]?s:"render";return"render"!==l&&a[l](e,n,i,r),Fy[l]}qo(Hy,["dispose"]),ea(Hy);var Fy={incrementalPrepareRender:{progress:function(t,e){e.view.incrementalRender(t,e.model,e.ecModel,e.api,e.payload)}},render:{forceFirstProgress:!0,progress:function(t,e){e.view.render(e.model,e.ecModel,e.api,e.payload)}}},Vy=Hy,Gy="\0__throttleOriginMethod",$y="\0__throttleRate",Uy="\0__throttleType";function qy(t,e,n){var i,r,o,a,s,l=0,u=0,c=null;function d(){u=(new Date).getTime(),c=null,t.apply(o,a||[])}e=e||0;var h=function(){for(var t=[],h=0;h<arguments.length;h++)t[h]=arguments[h];i=(new Date).getTime(),o=this,a=t;var f=s||e,p=s||n;s=null,r=i-(p?l:u)-f,clearTimeout(c),p?c=setTimeout(d,f):r>=0?d():c=setTimeout(d,-r),l=i};return h.clear=function(){c&&(clearTimeout(c),c=null)},h.debounceNextCall=function(t){s=t},h}function Xy(t,e,n,i){var r=t[e];if(r){var o=r[Gy]||r,a=r[Uy],s=r[$y];if(s!==n||a!==i){if(null==n||!i)return t[e]=o;r=t[e]=qy(o,n,"debounce"===i),r[Gy]=o,r[Uy]=i,r[$y]=n}return r}}function Zy(t,e){var n=t[e];n&&n[Gy]&&(t[e]=n[Gy])}var Jy=Io(),Ky={itemStyle:na(Vh,!0),lineStyle:na(Wh,!0)},Qy={lineStyle:"stroke",itemStyle:"fill"};function tv(t,e){var n=t.visualStyleMapper||Ky[e];return n||Ky.itemStyle}function ev(t,e){var n=t.visualDrawType||Qy[e];return n||"fill"}var nv={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){var n=t.getData(),i=t.visualStyleAccessPath||"itemStyle",r=t.getModel(i),o=tv(t,i),a=o(r),s=r.getShallow("decal");s&&(n.setVisual("decal",s),s.dirty=!0);var l=ev(t,i),u=a[l],c=U(u)?u:null,d="auto"===a.fill||"auto"===a.stroke;if(!a[l]||c||d){var h=t.getColorFromPalette(t.name,null,e.getSeriesCount());a[l]||(a[l]=h,n.setVisual("colorFromPalette",!0)),a.fill="auto"===a.fill||"function"===typeof a.fill?h:a.fill,a.stroke="auto"===a.stroke||"function"===typeof a.stroke?h:a.stroke}if(n.setVisual("style",a),n.setVisual("drawType",l),!e.isSeriesFiltered(t)&&c)return n.setVisual("colorFromPalette",!1),{dataEach:function(e,n){var i=t.getDataParams(n),r=I({},a);r[l]=c(i),e.setItemVisual(n,"style",r)}}}},iv=new qh,rv={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){if(!t.ignoreStyleOnData&&!e.isSeriesFiltered(t)){var n=t.getData(),i=t.visualStyleAccessPath||"itemStyle",r=tv(t,i),o=n.getVisual("drawType");return{dataEach:n.hasItemOption?function(t,e){var n=t.getRawDataItem(e);if(n&&n[i]){iv.option=n[i];var a=r(iv),s=t.ensureUniqueItemVisual(e,"style");I(s,a),iv.option.decal&&(t.setItemVisual(e,"decal",iv.option.decal),iv.option.decal.dirty=!0),o in a&&t.setItemVisual(e,"colorFromPalette",!1)}}:null}}}},ov={performRawSeries:!0,overallReset:function(t){var e=gt();t.eachSeries((function(t){if(t.useColorPaletteOnData){var n=e.get(t.type);n||(n={},e.set(t.type,n)),Jy(t).scope=n}})),t.eachSeries((function(e){if(e.useColorPaletteOnData&&!t.isSeriesFiltered(e)){var n=e.getRawData(),i={},r=e.getData(),o=Jy(e).scope,a=e.visualStyleAccessPath||"itemStyle",s=ev(e,a);r.each((function(t){var e=r.getRawIndex(t);i[e]=t})),n.each((function(t){var a=i[t],l=r.getItemVisual(a,"colorFromPalette");if(l){var u=r.ensureUniqueItemVisual(a,"style"),c=n.getName(t)||t+"",d=n.count();u[s]=e.getColorFromPalette(c,o,d)}}))}}))}},av=Math.PI;function sv(t,e){e=e||{},A(e,{text:"loading",textColor:"#000",fontSize:12,fontWeight:"normal",fontStyle:"normal",fontFamily:"sans-serif",maskColor:"rgba(255, 255, 255, 0.8)",showSpinner:!0,color:"#5470c6",spinnerRadius:10,lineWidth:5,zlevel:0});var n=new Mr,i=new Al({style:{fill:e.maskColor},zlevel:e.zlevel,z:1e4});n.add(i);var r,o=new Gl({style:{text:e.text,fill:e.textColor,fontSize:e.fontSize,fontWeight:e.fontWeight,fontStyle:e.fontStyle,fontFamily:e.fontFamily},zlevel:e.zlevel,z:10001}),a=new Al({style:{fill:"none"},textContent:o,textConfig:{position:"right",distance:10},zlevel:e.zlevel,z:10001});return n.add(a),e.showSpinner&&(r=new _d({shape:{startAngle:-av/2,endAngle:-av/2+.1,r:e.spinnerRadius},style:{stroke:e.color,lineCap:"round",lineWidth:e.lineWidth},zlevel:e.zlevel,z:10001}),r.animateShape(!0).when(1e3,{endAngle:3*av/2}).start("circularInOut"),r.animateShape(!0).when(1e3,{startAngle:3*av/2}).delay(300).start("circularInOut"),n.add(r)),n.resize=function(){var n=o.getBoundingRect().width,s=e.showSpinner?e.spinnerRadius:0,l=(t.getWidth()-2*s-(e.showSpinner&&n?10:0)-n)/2-(e.showSpinner&&n?0:5+n/2)+(e.showSpinner?0:n/2)+(n?0:s),u=t.getHeight()/2;e.showSpinner&&r.setShape({cx:l,cy:u}),a.setShape({x:l-s,y:u-s,width:2*s,height:2*s}),i.setShape({x:0,y:0,width:t.getWidth(),height:t.getHeight()})},n.resize(),n}var lv=function(){function t(t,e,n,i){this._stageTaskMap=gt(),this.ecInstance=t,this.api=e,n=this._dataProcessorHandlers=n.slice(),i=this._visualHandlers=i.slice(),this._allHandlers=n.concat(i)}return t.prototype.restoreData=function(t,e){t.restoreData(e),this._stageTaskMap.each((function(t){var e=t.overallTask;e&&e.dirty()}))},t.prototype.getPerformArgs=function(t,e){if(t.__pipeline){var n=this._pipelineMap.get(t.__pipeline.id),i=n.context,r=!e&&n.progressiveEnabled&&(!i||i.progressiveRender)&&t.__idxInPipeline>n.blockIndex,o=r?n.step:null,a=i&&i.modDataCount,s=null!=a?Math.ceil(a/o):null;return{step:o,modBy:s,modDataCount:a}}},t.prototype.getPipeline=function(t){return this._pipelineMap.get(t)},t.prototype.updateStreamModes=function(t,e){var n=this._pipelineMap.get(t.uid),i=t.getData(),r=i.count(),o=n.progressiveEnabled&&e.incrementalPrepareRender&&r>=n.threshold,a=t.get("large")&&r>=t.get("largeThreshold"),s="mod"===t.get("progressiveChunkMode")?r:null;t.pipelineContext=n.context={progressiveRender:o,modDataCount:s,large:a}},t.prototype.restorePipelines=function(t){var e=this,n=e._pipelineMap=gt();t.eachSeries((function(t){var i=t.getProgressive(),r=t.uid;n.set(r,{id:r,head:null,tail:null,threshold:t.getProgressiveThreshold(),progressiveEnabled:i&&!(t.preventIncremental&&t.preventIncremental()),blockIndex:-1,step:Math.round(i||700),count:0}),e._pipe(t,t.dataTask)}))},t.prototype.prepareStageTasks=function(){var t=this._stageTaskMap,e=this.api.getModel(),n=this.api;R(this._allHandlers,(function(i){var r=t.get(i.uid)||t.set(i.uid,{}),o="";ct(!(i.reset&&i.overallReset),o),i.reset&&this._createSeriesStageTask(i,r,e,n),i.overallReset&&this._createOverallStageTask(i,r,e,n)}),this)},t.prototype.prepareView=function(t,e,n,i){var r=t.renderTask,o=r.context;o.model=e,o.ecModel=n,o.api=i,r.__block=!t.incrementalPrepareRender,this._pipe(e,r)},t.prototype.performDataProcessorTasks=function(t,e){this._performStageTasks(this._dataProcessorHandlers,t,e,{block:!0})},t.prototype.performVisualTasks=function(t,e,n){this._performStageTasks(this._visualHandlers,t,e,n)},t.prototype._performStageTasks=function(t,e,n,i){i=i||{};var r=!1,o=this;function a(t,e){return t.setDirty&&(!t.dirtyMap||t.dirtyMap.get(e.__pipeline.id))}R(t,(function(t,s){if(!i.visualType||i.visualType===t.visualType){var l=o._stageTaskMap.get(t.uid),u=l.seriesTaskMap,c=l.overallTask;if(c){var d,h=c.agentStubMap;h.each((function(t){a(i,t)&&(t.dirty(),d=!0)})),d&&c.dirty(),o.updatePayload(c,n);var f=o.getPerformArgs(c,i.block);h.each((function(t){t.perform(f)})),c.perform(f)&&(r=!0)}else u&&u.each((function(s,l){a(i,s)&&s.dirty();var u=o.getPerformArgs(s,i.block);u.skip=!t.performRawSeries&&e.isSeriesFiltered(s.context.model),o.updatePayload(s,n),s.perform(u)&&(r=!0)}))}})),this.unfinished=r||this.unfinished},t.prototype.performSeriesTasks=function(t){var e;t.eachSeries((function(t){e=t.dataTask.perform()||e})),this.unfinished=e||this.unfinished},t.prototype.plan=function(){this._pipelineMap.each((function(t){var e=t.tail;do{if(e.__block){t.blockIndex=e.__idxInPipeline;break}e=e.getUpstream()}while(e)}))},t.prototype.updatePayload=function(t,e){"remain"!==e&&(t.context.payload=e)},t.prototype._createSeriesStageTask=function(t,e,n,i){var r=this,o=e.seriesTaskMap,a=e.seriesTaskMap=gt(),s=t.seriesType,l=t.getTargetSeries;function u(e){var s=e.uid,l=a.set(s,o&&o.get(s)||Cg({plan:fv,reset:pv,count:yv}));l.context={model:e,ecModel:n,api:i,useClearVisual:t.isVisual&&!t.isLayout,plan:t.plan,reset:t.reset,scheduler:r},r._pipe(e,l)}t.createOnAllSeries?n.eachRawSeries(u):s?n.eachRawSeriesByType(s,u):l&&l(n,i).each(u)},t.prototype._createOverallStageTask=function(t,e,n,i){var r=this,o=e.overallTask=e.overallTask||Cg({reset:uv});o.context={ecModel:n,api:i,overallReset:t.overallReset,scheduler:r};var a=o.agentStubMap,s=o.agentStubMap=gt(),l=t.seriesType,u=t.getTargetSeries,c=!0,d=!1,h="";function f(t){var e=t.uid,n=s.set(e,a&&a.get(e)||(d=!0,Cg({reset:cv,onDirty:hv})));n.context={model:t,overallProgress:c},n.agent=o,n.__block=c,r._pipe(t,n)}ct(!t.createOnAllSeries,h),l?n.eachRawSeriesByType(l,f):u?u(n,i).each(f):(c=!1,R(n.getSeries(),f)),d&&o.dirty()},t.prototype._pipe=function(t,e){var n=t.uid,i=this._pipelineMap.get(n);!i.head&&(i.head=e),i.tail&&i.tail.pipe(e),i.tail=e,e.__idxInPipeline=i.count++,e.__pipeline=i},t.wrapStageHandler=function(t,e){return U(t)&&(t={overallReset:t,seriesType:vv(t)}),t.uid=Zh("stageHandler"),e&&(t.visualType=e),t},t}();function uv(t){t.overallReset(t.ecModel,t.api,t.payload)}function cv(t){return t.overallProgress&&dv}function dv(){this.agent.dirty(),this.getDownstream().dirty()}function hv(){this.agent&&this.agent.dirty()}function fv(t){return t.plan?t.plan(t.model,t.ecModel,t.api,t.payload):null}function pv(t){t.useClearVisual&&t.data.clearAllVisual();var e=t.resetDefines=so(t.reset(t.model,t.ecModel,t.api,t.payload));return e.length>1?H(e,(function(t,e){return gv(e)})):mv}var mv=gv(0);function gv(t){return function(e,n){var i=n.data,r=n.resetDefines[t];if(r&&r.dataEach)for(var o=e.start;o<e.end;o++)r.dataEach(i,o);else r&&r.progress&&r.progress(e,i)}}function yv(t){return t.data.count()}function vv(t){_v=null;try{t(bv,xv)}catch(e){}return _v}var _v,bv={},xv={};function wv(t,e){for(var n in e.prototype)t[n]=bt}wv(bv,hm),wv(xv,mm),bv.eachSeriesByType=bv.eachRawSeriesByType=function(t){_v=t},bv.eachComponent=function(t){"series"===t.mainType&&t.subType&&(_v=t.subType)};var Mv=lv,Sv=["#37A2DA","#32C5E9","#67E0E3","#9FE6B8","#FFDB5C","#ff9f7f","#fb7293","#E062AE","#E690D1","#e7bcf3","#9d96f5","#8378EA","#96BFFF"],Lv={color:Sv,colorLayer:[["#37A2DA","#ffd85c","#fd7b5f"],["#37A2DA","#67E0E3","#FFDB5C","#ff9f7f","#E062AE","#9d96f5"],["#37A2DA","#32C5E9","#9FE6B8","#FFDB5C","#ff9f7f","#fb7293","#e7bcf3","#8378EA","#96BFFF"],Sv]},kv="#B9B8CE",Tv="#100C2A",Dv=function(){return{axisLine:{lineStyle:{color:kv}},splitLine:{lineStyle:{color:"#484753"}},splitArea:{areaStyle:{color:["rgba(255,255,255,0.02)","rgba(255,255,255,0.05)"]}},minorSplitLine:{lineStyle:{color:"#20203B"}}}},Cv=["#4992ff","#7cffb2","#fddd60","#ff6e76","#58d9f9","#05c091","#ff8a45","#8d48e3","#dd79ff"],Iv={darkMode:!0,color:Cv,backgroundColor:Tv,axisPointer:{lineStyle:{color:"#817f91"},crossStyle:{color:"#817f91"},label:{color:"#fff"}},legend:{textStyle:{color:kv}},textStyle:{color:kv},title:{textStyle:{color:"#EEF1FA"},subtextStyle:{color:"#B9B8CE"}},toolbox:{iconStyle:{borderColor:kv}},dataZoom:{borderColor:"#71708A",textStyle:{color:kv},brushStyle:{color:"rgba(135,163,206,0.3)"},handleStyle:{color:"#353450",borderColor:"#C5CBE3"},moveHandleStyle:{color:"#B0B6C3",opacity:.3},fillerColor:"rgba(135,163,206,0.2)",emphasis:{handleStyle:{borderColor:"#91B7F2",color:"#4D587D"},moveHandleStyle:{color:"#636D9A",opacity:.7}},dataBackground:{lineStyle:{color:"#71708A",width:1},areaStyle:{color:"#71708A"}},selectedDataBackground:{lineStyle:{color:"#87A3CE"},areaStyle:{color:"#87A3CE"}}},visualMap:{textStyle:{color:kv}},timeline:{lineStyle:{color:kv},label:{color:kv},controlStyle:{color:kv,borderColor:kv}},calendar:{itemStyle:{color:Tv},dayLabel:{color:kv},monthLabel:{color:kv},yearLabel:{color:kv}},timeAxis:Dv(),logAxis:Dv(),valueAxis:Dv(),categoryAxis:Dv(),line:{symbol:"circle"},graph:{color:Cv},gauge:{title:{color:kv},axisLine:{lineStyle:{color:[[1,"rgba(207,212,219,0.2)"]]}},axisLabel:{color:kv},detail:{color:"#EEF1FA"}},candlestick:{itemStyle:{color:"#f64e56",color0:"#54ea92",borderColor:"#f64e56",borderColor0:"#54ea92"}}};Iv.categoryAxis.splitLine.show=!1;var Av=Iv,Yv=function(){function t(){}return t.prototype.normalizeQuery=function(t){var e={},n={},i={};if(q(t)){var r=Go(t);e.mainType=r.main||null,e.subType=r.sub||null}else{var o=["Index","Name","Id"],a={name:1,dataIndex:1,dataType:1};R(t,(function(t,r){for(var s=!1,l=0;l<o.length;l++){var u=o[l],c=r.lastIndexOf(u);if(c>0&&c===r.length-u.length){var d=r.slice(0,c);"data"!==d&&(e.mainType=d,e[u.toLowerCase()]=t,s=!0)}}a.hasOwnProperty(r)&&(n[r]=t,s=!0),s||(i[r]=t)}))}return{cptQuery:e,dataQuery:n,otherQuery:i}},t.prototype.filter=function(t,e){var n=this.eventInfo;if(!n)return!0;var i=n.targetEl,r=n.packedEvent,o=n.model,a=n.view;if(!o||!a)return!0;var s=e.cptQuery,l=e.dataQuery;return u(s,o,"mainType")&&u(s,o,"subType")&&u(s,o,"index","componentIndex")&&u(s,o,"name")&&u(s,o,"id")&&u(l,r,"name")&&u(l,r,"dataIndex")&&u(l,r,"dataType")&&(!a.filterForExposedEvent||a.filterForExposedEvent(t,e.otherQuery,i,r));function u(t,e,n,i){return null==t[n]||e[i||n]===t[n]}},t.prototype.afterTrigger=function(){this.eventInfo=null},t}(),Ov={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){var n=t.getData();if(t.legendIcon&&n.setVisual("legendIcon",t.legendIcon),t.hasSymbolVisual){var i=t.get("symbol"),r=t.get("symbolSize"),o=t.get("symbolKeepAspect"),a=t.get("symbolRotate"),s=t.get("symbolOffset"),l=U(i),u=U(r),c=U(a),d=U(s),h=l||u||c||d,f=!l&&i?i:t.defaultSymbol,p=u?null:r,m=c?null:a,g=d?null:s;if(n.setVisual({legendIcon:t.legendIcon||f,symbol:f,symbolSize:p,symbolKeepAspect:o,symbolRotate:m,symbolOffset:g}),!e.isSeriesFiltered(t))return{dataEach:h?y:null}}function y(e,n){var o=t.getRawValue(n),h=t.getDataParams(n);l&&e.setItemVisual(n,"symbol",i(o,h)),u&&e.setItemVisual(n,"symbolSize",r(o,h)),c&&e.setItemVisual(n,"symbolRotate",a(o,h)),d&&e.setItemVisual(n,"symbolOffset",s(o,h))}}},Pv={createOnAllSeries:!0,performRawSeries:!0,reset:function(t,e){if(t.hasSymbolVisual&&!e.isSeriesFiltered(t)){var n=t.getData();return{dataEach:n.hasItemOption?i:null}}function i(t,e){var n=t.getItemModel(e),i=n.getShallow("symbol",!0),r=n.getShallow("symbolSize",!0),o=n.getShallow("symbolRotate",!0),a=n.getShallow("symbolOffset",!0),s=n.getShallow("symbolKeepAspect",!0);null!=i&&t.setItemVisual(e,"symbol",i),null!=r&&t.setItemVisual(e,"symbolSize",r),null!=o&&t.setItemVisual(e,"symbolRotate",o),null!=a&&t.setItemVisual(e,"symbolOffset",a),null!=s&&t.setItemVisual(e,"symbolKeepAspect",s)}}};function Ev(t,e,n){switch(n){case"color":var i=t.getItemVisual(e,"style");return i[t.getVisual("drawType")];case"opacity":return t.getItemVisual(e,"style").opacity;case"symbol":case"symbolSize":case"liftZ":return t.getItemVisual(e,n);default:0}}function Nv(t,e){switch(e){case"color":var n=t.getVisual("style");return n[t.getVisual("drawType")];case"opacity":return t.getVisual("style").opacity;case"symbol":case"symbolSize":case"liftZ":return t.getVisual(e);default:0}}function Rv(t,e,n,i){switch(n){case"color":var r=t.ensureUniqueItemVisual(e,"style");r[t.getVisual("drawType")]=i,t.setItemVisual(e,"colorFromPalette",!1);break;case"opacity":t.ensureUniqueItemVisual(e,"style").opacity=i;break;case"symbol":case"symbolSize":case"liftZ":t.setItemVisual(e,n,i);break;default:0}}var Hv=2*Math.PI,jv=Bs.CMD,zv=["top","right","bottom","left"];function Wv(t,e,n,i,r){var o=n.width,a=n.height;switch(t){case"top":i.set(n.x+o/2,n.y-e),r.set(0,-1);break;case"bottom":i.set(n.x+o/2,n.y+a+e),r.set(0,1);break;case"left":i.set(n.x-e,n.y+a/2),r.set(-1,0);break;case"right":i.set(n.x+o+e,n.y+a/2),r.set(1,0);break}}function Bv(t,e,n,i,r,o,a,s,l){a-=t,s-=e;var u=Math.sqrt(a*a+s*s);a/=u,s/=u;var c=a*n+t,d=s*n+e;if(Math.abs(i-r)%Hv<1e-4)return l[0]=c,l[1]=d,u-n;if(o){var h=i;i=Us(r),r=Us(h)}else i=Us(i),r=Us(r);i>r&&(r+=Hv);var f=Math.atan2(s,a);if(f<0&&(f+=Hv),f>=i&&f<=r||f+Hv>=i&&f+Hv<=r)return l[0]=c,l[1]=d,u-n;var p=n*Math.cos(i)+t,m=n*Math.sin(i)+e,g=n*Math.cos(r)+t,y=n*Math.sin(r)+e,v=(p-a)*(p-a)+(m-s)*(m-s),_=(g-a)*(g-a)+(y-s)*(y-s);return v<_?(l[0]=p,l[1]=m,Math.sqrt(v)):(l[0]=g,l[1]=y,Math.sqrt(_))}function Fv(t,e,n,i,r,o,a,s){var l=r-t,u=o-e,c=n-t,d=i-e,h=Math.sqrt(c*c+d*d);c/=h,d/=h;var f=l*c+u*d,p=f/h;s&&(p=Math.min(Math.max(p,0),1)),p*=h;var m=a[0]=t+p*c,g=a[1]=e+p*d;return Math.sqrt((m-r)*(m-r)+(g-o)*(g-o))}function Vv(t,e,n,i,r,o,a){n<0&&(t+=n,n=-n),i<0&&(e+=i,i=-i);var s=t+n,l=e+i,u=a[0]=Math.min(Math.max(r,t),s),c=a[1]=Math.min(Math.max(o,e),l);return Math.sqrt((u-r)*(u-r)+(c-o)*(c-o))}var Gv=[];function $v(t,e,n){var i=Vv(e.x,e.y,e.width,e.height,t.x,t.y,Gv);return n.set(Gv[0],Gv[1]),i}function Uv(t,e,n){for(var i,r,o=0,a=0,s=0,l=0,u=1/0,c=e.data,d=t.x,h=t.y,f=0;f<c.length;){var p=c[f++];1===f&&(o=c[f],a=c[f+1],s=o,l=a);var m=u;switch(p){case jv.M:s=c[f++],l=c[f++],o=s,a=l;break;case jv.L:m=Fv(o,a,c[f],c[f+1],d,h,Gv,!0),o=c[f++],a=c[f++];break;case jv.C:m=Ka(o,a,c[f++],c[f++],c[f++],c[f++],c[f],c[f+1],d,h,Gv),o=c[f++],a=c[f++];break;case jv.Q:m=os(o,a,c[f++],c[f++],c[f],c[f+1],d,h,Gv),o=c[f++],a=c[f++];break;case jv.A:var g=c[f++],y=c[f++],v=c[f++],_=c[f++],b=c[f++],x=c[f++];f+=1;var w=!!(1-c[f++]);i=Math.cos(b)*v+g,r=Math.sin(b)*_+y,f<=1&&(s=i,l=r);var M=(d-g)*_/v+g;m=Bv(g,y,_,b,b+x,w,M,h,Gv),o=Math.cos(b+x)*v+g,a=Math.sin(b+x)*_+y;break;case jv.R:s=o=c[f++],l=a=c[f++];var S=c[f++],L=c[f++];m=Vv(s,l,S,L,d,h,Gv);break;case jv.Z:m=Fv(o,a,s,l,d,h,Gv,!0),o=s,a=l;break}m<u&&(u=m,n.set(Gv[0],Gv[1]))}return u}var qv=new Wi,Xv=new Wi,Zv=new Wi,Jv=new Wi,Kv=new Wi;function Qv(t,e){if(t){var n=t.getTextGuideLine(),i=t.getTextContent();if(i&&n){var r=t.textGuideLineConfig||{},o=[[0,0],[0,0],[0,0]],a=r.candidates||zv,s=i.getBoundingRect().clone();s.applyTransform(i.getComputedTransform());var l=1/0,u=r.anchor,c=t.getComputedTransform(),d=c&&Ti([],c),h=e.get("length2")||0;u&&Zv.copy(u);for(var f=0;f<a.length;f++){var p=a[f];Wv(p,0,s,qv,Jv),Wi.scaleAndAdd(Xv,qv,Jv,h),Xv.transform(d);var m=t.getBoundingRect(),g=u?u.distance(Xv):t instanceof pl?Uv(Xv,t.path,Zv):$v(Xv,m,Zv);g<l&&(l=g,Xv.transform(c),Zv.transform(c),Zv.toArray(o[0]),Xv.toArray(o[1]),qv.toArray(o[2]))}n_(o,e.get("minTurnAngle")),n.setShape({points:o})}}}var t_=[],e_=new Wi;function n_(t,e){if(e<=180&&e>0){e=e/180*Math.PI,qv.fromArray(t[0]),Xv.fromArray(t[1]),Zv.fromArray(t[2]),Wi.sub(Jv,qv,Xv),Wi.sub(Kv,Zv,Xv);var n=Jv.len(),i=Kv.len();if(!(n<.001||i<.001)){Jv.scale(1/n),Kv.scale(1/i);var r=Jv.dot(Kv),o=Math.cos(e);if(o<r){var a=Fv(Xv.x,Xv.y,Zv.x,Zv.y,qv.x,qv.y,t_,!1);e_.fromArray(t_),e_.scaleAndAdd(Kv,a/Math.tan(Math.PI-e));var s=Zv.x!==Xv.x?(e_.x-Xv.x)/(Zv.x-Xv.x):(e_.y-Xv.y)/(Zv.y-Xv.y);if(isNaN(s))return;s<0?Wi.copy(e_,Xv):s>1&&Wi.copy(e_,Zv),e_.toArray(t[1])}}}}function i_(t,e,n){if(n<=180&&n>0){n=n/180*Math.PI,qv.fromArray(t[0]),Xv.fromArray(t[1]),Zv.fromArray(t[2]),Wi.sub(Jv,Xv,qv),Wi.sub(Kv,Zv,Xv);var i=Jv.len(),r=Kv.len();if(!(i<.001||r<.001)){Jv.scale(1/i),Kv.scale(1/r);var o=Jv.dot(e),a=Math.cos(n);if(o<a){var s=Fv(Xv.x,Xv.y,Zv.x,Zv.y,qv.x,qv.y,t_,!1);e_.fromArray(t_);var l=Math.PI/2,u=Math.acos(Kv.dot(e)),c=l+u-n;if(c>=l)Wi.copy(e_,Zv);else{e_.scaleAndAdd(Kv,s/Math.tan(Math.PI/2-c));var d=Zv.x!==Xv.x?(e_.x-Xv.x)/(Zv.x-Xv.x):(e_.y-Xv.y)/(Zv.y-Xv.y);if(isNaN(d))return;d<0?Wi.copy(e_,Xv):d>1&&Wi.copy(e_,Zv)}e_.toArray(t[1])}}}}function r_(t,e,n,i){var r="normal"===n,o=r?t:t.ensureState(n);o.ignore=e;var a=i.get("smooth");a&&!0===a&&(a=.3),o.shape=o.shape||{},a>0&&(o.shape.smooth=a);var s=i.getModel("lineStyle").getLineStyle();r?t.useStyle(s):o.style=s}function o_(t,e){var n=e.smooth,i=e.points;if(i)if(t.moveTo(i[0][0],i[0][1]),n>0&&i.length>=3){var r=Et(i[0],i[1]),o=Et(i[1],i[2]);if(!r||!o)return t.lineTo(i[1][0],i[1][1]),void t.lineTo(i[2][0],i[2][1]);var a=Math.min(r,o)*n,s=Ht([],i[1],i[0],a/r),l=Ht([],i[1],i[2],a/o),u=Ht([],s,l,.5);t.bezierCurveTo(s[0],s[1],s[0],s[1],u[0],u[1]),t.bezierCurveTo(l[0],l[1],l[0],l[1],i[2][0],i[2][1])}else for(var c=1;c<i.length;c++)t.lineTo(i[c][0],i[c][1])}function a_(t,e,n){var i=t.getTextGuideLine(),r=t.getTextContent();if(r){for(var o=e.normal,a=o.get("show"),s=r.ignore,l=0;l<tu.length;l++){var u=tu[l],c=e[u],d="normal"===u;if(c){var h=c.get("show"),f=d?s:at(r.states[u]&&r.states[u].ignore,s);if(f||!at(h,a)){var p=d?i:i&&i.states.normal;p&&(p.ignore=!0);continue}i||(i=new sd,t.setTextGuideLine(i),d||!s&&a||r_(i,!0,"normal",e.normal),t.stateProxy&&(i.stateProxy=t.stateProxy)),r_(i,!1,u,c)}}if(i){A(i.style,n),i.style.fill=null;var m=o.get("showAbove"),g=t.textGuideLineConfig=t.textGuideLineConfig||{};g.showAbove=m||!1,i.buildPath=o_}}else i&&t.removeTextGuideLine()}function s_(t,e){e=e||"labelLine";for(var n={normal:t.getModel(e)},i=0;i<Ql.length;i++){var r=Ql[i];n[r]=t.getModel([r,e])}return n}function l_(t){for(var e=[],n=0;n<t.length;n++){var i=t[n];if(!i.defaultAttr.ignore){var r=i.label,o=r.getComputedTransform(),a=r.getBoundingRect(),s=!o||o[1]<1e-5&&o[2]<1e-5,l=r.style.margin||0,u=a.clone();u.applyTransform(o),u.x-=l/2,u.y-=l/2,u.width+=l,u.height+=l;var c=s?new Od(a,o):null;e.push({label:r,labelLine:i.labelLine,rect:u,localRect:a,obb:c,priority:i.priority,defaultAttr:i.defaultAttr,layoutOption:i.computedLayoutOption,axisAligned:s,transform:o})}}return e}function u_(t,e,n,i,r,o){var a=t.length;if(!(a<2)){t.sort((function(t,n){return t.rect[e]-n.rect[e]}));for(var s,l=0,u=!1,c=[],d=0,h=0;h<a;h++){var f=t[h],p=f.rect;s=p[e]-l,s<0&&(p[e]-=s,f.label[e]-=s,u=!0);var m=Math.max(-s,0);c.push(m),d+=m,l=p[e]+p[n]}d>0&&o&&w(-d/a,0,a);var g,y,v=t[0],_=t[a-1];return b(),g<0&&M(-g,.8),y<0&&M(y,.8),b(),x(g,y,1),x(y,g,-1),b(),g<0&&S(-g),y<0&&S(y),u}function b(){g=v.rect[e]-i,y=r-_.rect[e]-_.rect[n]}function x(t,e,n){if(t<0){var i=Math.min(e,-t);if(i>0){w(i*n,0,a);var r=i+t;r<0&&M(-r*n,1)}else M(-t*n,1)}}function w(n,i,r){0!==n&&(u=!0);for(var o=i;o<r;o++){var a=t[o],s=a.rect;s[e]+=n,a.label[e]+=n}}function M(i,r){for(var o=[],s=0,l=1;l<a;l++){var u=t[l-1].rect,c=Math.max(t[l].rect[e]-u[e]-u[n],0);o.push(c),s+=c}if(s){var d=Math.min(Math.abs(i)/s,r);if(i>0)for(l=0;l<a-1;l++){var h=o[l]*d;w(h,0,l+1)}else for(l=a-1;l>0;l--){h=o[l-1]*d;w(-h,l,a)}}}function S(t){var e=t<0?-1:1;t=Math.abs(t);for(var n=Math.ceil(t/(a-1)),i=0;i<a-1;i++)if(e>0?w(n,0,i+1):w(-n,a-i-1,a),t-=n,t<=0)return}}function c_(t,e,n,i){return u_(t,"x","width",e,n,i)}function d_(t,e,n,i){return u_(t,"y","height",e,n,i)}function h_(t){var e=[];t.sort((function(t,e){return e.priority-t.priority}));var n=new Ji(0,0,0,0);function i(t){if(!t.ignore){var e=t.ensureState("emphasis");null==e.ignore&&(e.ignore=!1)}t.ignore=!0}for(var r=0;r<t.length;r++){var o=t[r],a=o.axisAligned,s=o.localRect,l=o.transform,u=o.label,c=o.labelLine;n.copy(o.rect),n.width-=.1,n.height-=.1,n.x+=.05,n.y+=.05;for(var d=o.obb,h=!1,f=0;f<e.length;f++){var p=e[f];if(n.intersect(p.rect)){if(a&&p.axisAligned){h=!0;break}if(p.obb||(p.obb=new Od(p.localRect,p.transform)),d||(d=new Od(s,l)),d.intersect(p.obb)){h=!0;break}}}h?(i(u),c&&i(c)):(u.attr("ignore",o.defaultAttr.ignore),c&&c.attr("ignore",o.defaultAttr.labelGuideIgnore),e.push(o))}}function f_(t){if(t){for(var e=[],n=0;n<t.length;n++)e.push(t[n].slice());return e}}function p_(t,e){var n=t.label,i=e&&e.getTextGuideLine();return{dataIndex:t.dataIndex,dataType:t.dataType,seriesIndex:t.seriesModel.seriesIndex,text:t.label.style.text,rect:t.hostRect,labelRect:t.rect,align:n.style.align,verticalAlign:n.style.verticalAlign,labelLinePoints:f_(i&&i.shape.points)}}var m_=["align","verticalAlign","width","height","fontSize"],g_=new ji,y_=Io(),v_=Io();function __(t,e,n){for(var i=0;i<n.length;i++){var r=n[i];null!=e[r]&&(t[r]=e[r])}}var b_=["x","y","rotation"],x_=function(){function t(){this._labelList=[],this._chartViewList=[]}return t.prototype.clearLabels=function(){this._labelList=[],this._chartViewList=[]},t.prototype._addLabel=function(t,e,n,i,r){var o=i.style,a=i.__hostTarget,s=a.textConfig||{},l=i.getComputedTransform(),u=i.getBoundingRect().plain();Ji.applyTransform(u,u,l),l?g_.setLocalTransform(l):(g_.x=g_.y=g_.rotation=g_.originX=g_.originY=0,g_.scaleX=g_.scaleY=1);var c,d=i.__hostTarget;if(d){c=d.getBoundingRect().plain();var h=d.getComputedTransform();Ji.applyTransform(c,c,h)}var f=c&&d.getTextGuideLine();this._labelList.push({label:i,labelLine:f,seriesModel:n,dataIndex:t,dataType:e,layoutOption:r,computedLayoutOption:null,rect:u,hostRect:c,priority:c?c.width*c.height:0,defaultAttr:{ignore:i.ignore,labelGuideIgnore:f&&f.ignore,x:g_.x,y:g_.y,scaleX:g_.scaleX,scaleY:g_.scaleY,rotation:g_.rotation,style:{x:o.x,y:o.y,align:o.align,verticalAlign:o.verticalAlign,width:o.width,height:o.height,fontSize:o.fontSize},cursor:i.cursor,attachedPos:s.position,attachedRot:s.rotation}})},t.prototype.addLabelsOfSeries=function(t){var e=this;this._chartViewList.push(t);var n=t.__model,i=n.get("labelLayout");(U(i)||B(i).length)&&t.group.traverse((function(t){if(t.ignore)return!0;var r=t.getTextContent(),o=$l(t);r&&!r.disableLabelLayout&&e._addLabel(o.dataIndex,o.dataType,n,r,i)}))},t.prototype.updateLayoutConfig=function(t){var e=t.getWidth(),n=t.getHeight();function i(t,e){return function(){Qv(t,e)}}for(var r=0;r<this._labelList.length;r++){var o=this._labelList[r],a=o.label,s=a.__hostTarget,l=o.defaultAttr,u=void 0;u="function"===typeof o.layoutOption?o.layoutOption(p_(o,s)):o.layoutOption,u=u||{},o.computedLayoutOption=u;var c=Math.PI/180;s&&s.setTextConfig({local:!1,position:null!=u.x||null!=u.y?null:l.attachedPos,rotation:null!=u.rotate?u.rotate*c:l.attachedRot,offset:[u.dx||0,u.dy||0]});var d=!1;if(null!=u.x?(a.x=Nr(u.x,e),a.setStyle("x",0),d=!0):(a.x=l.x,a.setStyle("x",l.style.x)),null!=u.y?(a.y=Nr(u.y,n),a.setStyle("y",0),d=!0):(a.y=l.y,a.setStyle("y",l.style.y)),u.labelLinePoints){var h=s.getTextGuideLine();h&&(h.setShape({points:u.labelLinePoints}),d=!1)}var f=y_(a);f.needsUpdateLabelLine=d,a.rotation=null!=u.rotate?u.rotate*c:l.rotation,a.scaleX=l.scaleX,a.scaleY=l.scaleY;for(var p=0;p<m_.length;p++){var m=m_[p];a.setStyle(m,null!=u[m]?u[m]:l.style[m])}if(u.draggable){if(a.draggable=!0,a.cursor="move",s){var g=o.seriesModel;if(null!=o.dataIndex){var y=o.seriesModel.getData(o.dataType);g=y.getItemModel(o.dataIndex)}a.on("drag",i(s,g.getModel("labelLine")))}}else a.off("drag"),a.cursor=l.cursor}},t.prototype.layout=function(t){var e=t.getWidth(),n=t.getHeight(),i=l_(this._labelList),r=z(i,(function(t){return"shiftX"===t.layoutOption.moveOverlap})),o=z(i,(function(t){return"shiftY"===t.layoutOption.moveOverlap}));c_(r,0,e),d_(o,0,n);var a=z(i,(function(t){return t.layoutOption.hideOverlap}));h_(a)},t.prototype.processLabelsOverall=function(){var t=this;R(this._chartViewList,(function(e){var n=e.__model,i=e.ignoreLabelLineUpdate,r=n.isAnimationEnabled();e.group.traverse((function(e){if(e.ignore)return!0;var o=!i,a=e.getTextContent();!o&&a&&(o=y_(a).needsUpdateLabelLine),o&&t._updateLabelLine(e,n),r&&t._animateLabels(e,n)}))}))},t.prototype._updateLabelLine=function(t,e){var n=t.getTextContent(),i=$l(t),r=i.dataIndex;if(n&&null!=r){var o=e.getData(i.dataType),a=o.getItemModel(r),s={},l=o.getItemVisual(r,"style"),u=o.getVisual("drawType");s.stroke=l[u];var c=a.getModel("labelLine");a_(t,s_(a),s),Qv(t,c)}},t.prototype._animateLabels=function(t,e){var n=t.getTextContent(),i=t.getTextGuideLine();if(n&&!n.ignore&&!n.invisible&&!t.disableLabelAnimation&&!oh(t)){var r=y_(n),o=r.oldLayout,a=$l(t),s=a.dataIndex,l={x:n.x,y:n.y,rotation:n.rotation},u=e.getData(a.dataType);if(o){n.attr(o);var c=t.prevStates;c&&(O(c,"select")>=0&&n.attr(r.oldLayoutSelect),O(c,"emphasis")>=0&&n.attr(r.oldLayoutEmphasis)),th(n,l,e,s)}else if(n.attr(l),!Ph(n).valueAnimation){var d=at(n.style.opacity,1);n.style.opacity=0,eh(n,{style:{opacity:d}},e,s)}if(r.oldLayout=l,n.states.select){var h=r.oldLayoutSelect={};__(h,l,b_),__(h,n.states.select,b_)}if(n.states.emphasis){var f=r.oldLayoutEmphasis={};__(f,l,b_),__(f,n.states.emphasis,b_)}Nh(n,s,u,e,e)}if(i&&!i.ignore&&!i.invisible){r=v_(i),o=r.oldLayout;var p={points:i.shape.points};o?(i.attr({shape:o}),th(i,{shape:p},e)):(i.setShape(p),i.style.strokePercent=0,eh(i,{style:{strokePercent:1}},e)),r.oldLayout=p}},t}(),w_=x_;function M_(t,e){function n(e,n){var i=[];return e.eachComponent({mainType:"series",subType:t,query:n},(function(t){i.push(t.seriesIndex)})),i}R([[t+"ToggleSelect","toggleSelect"],[t+"Select","select"],[t+"UnSelect","unselect"]],(function(t){e(t[0],(function(e,i,r){e=I({},e),r.dispatchAction(I(e,{type:t[1],seriesIndex:n(i,e)}))}))}))}function S_(t,e,n,i,r){var o=t+e;n.isSilent(o)||i.eachComponent({mainType:"series",subType:"pie"},(function(t){for(var e=t.seriesIndex,i=r.selected,a=0;a<i.length;a++)if(i[a].seriesIndex===e){var s=t.getData(),l=Co(s,r.fromActionPayload);n.trigger(o,{type:o,seriesId:t.id,name:$(l)?s.getName(l[0]):s.getName(l),selected:I({},t.option.selectedMap)})}}))}function L_(t,e,n){t.on("selectchanged",(function(t){var i=n.getModel();t.isFromClick?(S_("map","selectchanged",e,i,t),S_("pie","selectchanged",e,i,t)):"select"===t.fromAction?(S_("map","selected",e,i,t),S_("pie","selected",e,i,t)):"unselect"===t.fromAction&&(S_("map","unselected",e,i,t),S_("pie","unselected",e,i,t))}))}function k_(t,e,n){var i;while(t){if(e(t)&&(i=t,n))break;t=t.__hostTarget||t.parent}return i}var T_=Math.round(9*Math.random()),D_=function(){function t(){this._id="__ec_inner_"+T_++}return t.prototype.get=function(t){return this._guard(t)[this._id]},t.prototype.set=function(t,e){var n=this._guard(t);return"function"===typeof Object.defineProperty?Object.defineProperty(n,this._id,{value:e,enumerable:!1,configurable:!0}):n[this._id]=e,this},t.prototype["delete"]=function(t){return!!this.has(t)&&(delete this._guard(t)[this._id],!0)},t.prototype.has=function(t){return!!this._guard(t)[this._id]},t.prototype._guard=function(t){if(t!==Object(t))throw TypeError("Value of WeakMap is not a non-null object.");return t},t}(),C_=D_,I_=pl.extend({type:"triangle",shape:{cx:0,cy:0,width:0,height:0},buildPath:function(t,e){var n=e.cx,i=e.cy,r=e.width/2,o=e.height/2;t.moveTo(n,i-o),t.lineTo(n+r,i+o),t.lineTo(n-r,i+o),t.closePath()}}),A_=pl.extend({type:"diamond",shape:{cx:0,cy:0,width:0,height:0},buildPath:function(t,e){var n=e.cx,i=e.cy,r=e.width/2,o=e.height/2;t.moveTo(n,i-o),t.lineTo(n+r,i),t.lineTo(n,i+o),t.lineTo(n-r,i),t.closePath()}}),Y_=pl.extend({type:"pin",shape:{x:0,y:0,width:0,height:0},buildPath:function(t,e){var n=e.x,i=e.y,r=e.width/5*3,o=Math.max(r,e.height),a=r/2,s=a*a/(o-a),l=i-o+a+s,u=Math.asin(s/a),c=Math.cos(u)*a,d=Math.sin(u),h=Math.cos(u),f=.6*a,p=.7*a;t.moveTo(n-c,l+s),t.arc(n,l,a,Math.PI-u,2*Math.PI+u),t.bezierCurveTo(n+c-d*f,l+s+h*f,n,i-p,n,i),t.bezierCurveTo(n,i-p,n-c+d*f,l+s+h*f,n-c,l+s),t.closePath()}}),O_=pl.extend({type:"arrow",shape:{x:0,y:0,width:0,height:0},buildPath:function(t,e){var n=e.height,i=e.width,r=e.x,o=e.y,a=i/3*2;t.moveTo(r,o),t.lineTo(r+a,o+n),t.lineTo(r,o+n/4*3),t.lineTo(r-a,o+n),t.lineTo(r,o),t.closePath()}}),P_={line:dd,rect:Al,roundRect:Al,square:Al,circle:Dc,diamond:A_,pin:Y_,arrow:O_,triangle:I_},E_={line:function(t,e,n,i,r){r.x1=t,r.y1=e+i/2,r.x2=t+n,r.y2=e+i/2},rect:function(t,e,n,i,r){r.x=t,r.y=e,r.width=n,r.height=i},roundRect:function(t,e,n,i,r){r.x=t,r.y=e,r.width=n,r.height=i,r.r=Math.min(n,i)/4},square:function(t,e,n,i,r){var o=Math.min(n,i);r.x=t,r.y=e,r.width=o,r.height=o},circle:function(t,e,n,i,r){r.cx=t+n/2,r.cy=e+i/2,r.r=Math.min(n,i)/2},diamond:function(t,e,n,i,r){r.cx=t+n/2,r.cy=e+i/2,r.width=n,r.height=i},pin:function(t,e,n,i,r){r.x=t+n/2,r.y=e+i/2,r.width=n,r.height=i},arrow:function(t,e,n,i,r){r.x=t+n/2,r.y=e+i/2,r.width=n,r.height=i},triangle:function(t,e,n,i,r){r.cx=t+n/2,r.cy=e+i/2,r.width=n,r.height=i}},N_={};R(P_,(function(t,e){N_[e]=new t}));var R_=pl.extend({type:"symbol",shape:{symbolType:"",x:0,y:0,width:0,height:0},calculateTextPosition:function(t,e,n){var i=ur(t,e,n),r=this.shape;return r&&"pin"===r.symbolType&&"inside"===e.position&&(i.y=n.y+.4*n.height),i},buildPath:function(t,e,n){var i=e.symbolType;if("none"!==i){var r=N_[i];r||(i="rect",r=N_[i]),E_[i](e.x,e.y,e.width,e.height,r.shape),r.buildPath(t,r.shape,n)}}});function H_(t,e){if("image"!==this.type){var n=this.style;this.__isEmptyBrush?(n.stroke=t,n.fill=e||"#fff",n.lineWidth=2):"line"===this.shape.symbolType?n.stroke=t:n.fill=t,this.markRedraw()}}function j_(t,e,n,i,r,o,a){var s,l=0===t.indexOf("empty");return l&&(t=t.substr(5,1).toLowerCase()+t.substr(6)),s=0===t.indexOf("image://")?$d(t.slice(8),new Ji(e,n,i,r),a?"center":"cover"):0===t.indexOf("path://")?Gd(t.slice(7),{},new Ji(e,n,i,r),a?"center":"cover"):new R_({shape:{symbolType:t,x:e,y:n,width:i,height:r}}),s.__isEmptyBrush=l,s.setColor=H_,o&&s.setColor(o),s}function z_(t,e,n){var i=null==e.x?0:e.x,r=null==e.x2?1:e.x2,o=null==e.y?0:e.y,a=null==e.y2?0:e.y2;e.global||(i=i*n.width+n.x,r=r*n.width+n.x,o=o*n.height+n.y,a=a*n.height+n.y),i=isNaN(i)?0:i,r=isNaN(r)?1:r,o=isNaN(o)?0:o,a=isNaN(a)?0:a;var s=t.createLinearGradient(i,o,r,a);return s}function W_(t,e,n){var i=n.width,r=n.height,o=Math.min(i,r),a=null==e.x?.5:e.x,s=null==e.y?.5:e.y,l=null==e.r?.5:e.r;e.global||(a=a*i+n.x,s=s*r+n.y,l*=o);var u=t.createRadialGradient(a,s,0,a,s,l);return u}function B_(t,e,n){for(var i="radial"===e.type?W_(t,e,n):z_(t,e,n),r=e.colorStops,o=0;o<r.length;o++)i.addColorStop(r[o].offset,r[o].color);return i}function F_(t,e){if(t===e||!t&&!e)return!1;if(!t||!e||t.length!==e.length)return!0;for(var n=0;n<t.length;n++)if(t[n]!==e[n])return!0;return!1}function V_(t,e){return t&&"solid"!==t&&e>0?(e=e||1,"dashed"===t?[4*e,2*e]:"dotted"===t?[e]:Z(t)?[t]:$(t)?t:null):null}var G_=new Bs(!0);function $_(t){var e=t.stroke;return!(null==e||"none"===e||!(t.lineWidth>0))}function U_(t){var e=t.fill;return null!=e&&"none"!==e}function q_(t,e){if(null!=e.fillOpacity&&1!==e.fillOpacity){var n=t.globalAlpha;t.globalAlpha=e.fillOpacity*e.opacity,t.fill(),t.globalAlpha=n}else t.fill()}function X_(t,e){if(null!=e.strokeOpacity&&1!==e.strokeOpacity){var n=t.globalAlpha;t.globalAlpha=e.strokeOpacity*e.opacity,t.stroke(),t.globalAlpha=n}else t.stroke()}function Z_(t,e,n){var i=la(e.image,e.__image,n);if(ca(i)){var r=t.createPattern(i,e.repeat||"repeat");if("function"===typeof DOMMatrix&&r.setTransform){var o=new DOMMatrix;o.rotateSelf(0,0,(e.rotation||0)/Math.PI*180),o.scaleSelf(e.scaleX||1,e.scaleY||1),o.translateSelf(e.x||0,e.y||0),r.setTransform(o)}return r}}function J_(t,e,n,i){var r=$_(n),o=U_(n),a=n.strokePercent,s=a<1,l=!e.path;e.silent&&!s||!l||e.createPathProxy();var u=e.path||G_;if(!i){var c=n.fill,d=n.stroke,h=o&&!!c.colorStops,f=r&&!!d.colorStops,p=o&&!!c.image,m=r&&!!d.image,g=void 0,y=void 0,v=void 0,_=void 0,b=void 0;(h||f)&&(b=e.getBoundingRect()),h&&(g=e.__dirty?B_(t,c,b):e.__canvasFillGradient,e.__canvasFillGradient=g),f&&(y=e.__dirty?B_(t,d,b):e.__canvasStrokeGradient,e.__canvasStrokeGradient=y),p&&(v=e.__dirty||!e.__canvasFillPattern?Z_(t,c,e):e.__canvasFillPattern,e.__canvasFillPattern=v),m&&(_=e.__dirty||!e.__canvasStrokePattern?Z_(t,d,e):e.__canvasStrokePattern,e.__canvasStrokePattern=v),h?t.fillStyle=g:p&&(v?t.fillStyle=v:o=!1),f?t.strokeStyle=y:m&&(_?t.strokeStyle=_:r=!1)}var x=n.lineDash&&n.lineWidth>0&&V_(n.lineDash,n.lineWidth),w=n.lineDashOffset,M=!!t.setLineDash,S=e.getGlobalScale();if(u.setScale(S[0],S[1],e.segmentIgnoreThreshold),x){var L=n.strokeNoScale&&e.getLineScale?e.getLineScale():1;L&&1!==L&&(x=H(x,(function(t){return t/L})),w/=L)}var k=!0;(l||e.__dirty&We||x&&!M&&r)&&(u.setDPR(t.dpr),s?u.setContext(null):(u.setContext(t),k=!1),u.reset(),x&&!M&&(u.setLineDash(x),u.setLineDashOffset(w)),e.buildPath(u,e.shape,i),u.toStatic(),e.pathUpdated()),k&&u.rebuildPath(t,s?a:1),x&&M&&(t.setLineDash(x),t.lineDashOffset=w),i||(n.strokeFirst?(r&&X_(t,n),o&&q_(t,n)):(o&&q_(t,n),r&&X_(t,n))),x&&M&&t.setLineDash([])}function K_(t,e,n){var i=e.__image=la(n.image,e.__image,e,e.onload);if(i&&ca(i)){var r=n.x||0,o=n.y||0,a=e.getWidth(),s=e.getHeight(),l=i.width/i.height;if(null==a&&null!=s?a=s*l:null==s&&null!=a?s=a/l:null==a&&null==s&&(a=i.width,s=i.height),n.sWidth&&n.sHeight){var u=n.sx||0,c=n.sy||0;t.drawImage(i,u,c,n.sWidth,n.sHeight,r,o,a,s)}else if(n.sx&&n.sy){u=n.sx,c=n.sy;var d=a-u,h=s-c;t.drawImage(i,u,c,d,h,r,o,a,s)}else t.drawImage(i,r,o,a,s)}}function Q_(t,e,n){var i=n.text;if(null!=i&&(i+=""),i){t.font=n.font||Qi,t.textAlign=n.textAlign,t.textBaseline=n.textBaseline;var r=void 0;if(t.setLineDash){var o=n.lineDash&&n.lineWidth>0&&V_(n.lineDash,n.lineWidth),a=n.lineDashOffset;if(o){var s=n.strokeNoScale&&e.getLineScale?e.getLineScale():1;s&&1!==s&&(o=H(o,(function(t){return t/s})),a/=s),t.setLineDash(o),t.lineDashOffset=a,r=!0}}n.strokeFirst?($_(n)&&t.strokeText(i,n.x,n.y),U_(n)&&t.fillText(i,n.x,n.y)):(U_(n)&&t.fillText(i,n.x,n.y),$_(n)&&t.strokeText(i,n.x,n.y)),r&&t.setLineDash([])}}var tb=["shadowBlur","shadowOffsetX","shadowOffsetY"],eb=[["lineCap","butt"],["lineJoin","miter"],["miterLimit",10]];function nb(t,e,n,i,r){var o=!1;if(!i&&(n=n||{},e===n))return!1;if(i||e.opacity!==n.opacity){o||(fb(t,r),o=!0);var a=Math.max(Math.min(e.opacity,1),0);t.globalAlpha=isNaN(a)?Ta.opacity:a}(i||e.blend!==n.blend)&&(o||(fb(t,r),o=!0),t.globalCompositeOperation=e.blend||Ta.blend);for(var s=0;s<tb.length;s++){var l=tb[s];(i||e[l]!==n[l])&&(o||(fb(t,r),o=!0),t[l]=t.dpr*(e[l]||0))}return(i||e.shadowColor!==n.shadowColor)&&(o||(fb(t,r),o=!0),t.shadowColor=e.shadowColor||Ta.shadowColor),o}function ib(t,e,n,i,r){var o=pb(e,r.inHover),a=i?null:n&&pb(n,r.inHover)||{};if(o===a)return!1;var s=nb(t,o,a,i,r);if((i||o.fill!==a.fill)&&(s||(fb(t,r),s=!0),t.fillStyle=o.fill),(i||o.stroke!==a.stroke)&&(s||(fb(t,r),s=!0),t.strokeStyle=o.stroke),(i||o.opacity!==a.opacity)&&(s||(fb(t,r),s=!0),t.globalAlpha=null==o.opacity?1:o.opacity),e.hasStroke()){var l=o.lineWidth,u=l/(o.strokeNoScale&&e&&e.getLineScale?e.getLineScale():1);t.lineWidth!==u&&(s||(fb(t,r),s=!0),t.lineWidth=u)}for(var c=0;c<eb.length;c++){var d=eb[c],h=d[0];(i||o[h]!==a[h])&&(s||(fb(t,r),s=!0),t[h]=o[h]||d[1])}return s}function rb(t,e,n,i,r){return nb(t,pb(e,r.inHover),n&&pb(n,r.inHover),i,r)}function ob(t,e){var n=e.transform,i=t.dpr||1;n?t.setTransform(i*n[0],i*n[1],i*n[2],i*n[3],i*n[4],i*n[5]):t.setTransform(i,0,0,i,0,0)}function ab(t,e,n){for(var i=!1,r=0;r<t.length;r++){var o=t[r];i=i||o.isZeroArea(),ob(e,o),e.beginPath(),o.buildPath(e,o.shape),e.clip()}n.allClipped=i}function sb(t,e){return t&&e?t[0]!==e[0]||t[1]!==e[1]||t[2]!==e[2]||t[3]!==e[3]||t[4]!==e[4]||t[5]!==e[5]:!(!t&&!e)}var lb=1,ub=2,cb=3,db=4;function hb(t){var e=U_(t),n=$_(t);return!(t.lineDash||!(+e^+n)||e&&"string"!==typeof t.fill||n&&"string"!==typeof t.stroke||t.strokePercent<1||t.strokeOpacity<1||t.fillOpacity<1)}function fb(t,e){e.batchFill&&t.fill(),e.batchStroke&&t.stroke(),e.batchFill="",e.batchStroke=""}function pb(t,e){return e&&t.__hoverStyle||t.style}function mb(t,e){gb(t,e,{inHover:!1,viewWidth:0,viewHeight:0},!0)}function gb(t,e,n,i){var r=e.transform;if(!e.shouldBePainted(n.viewWidth,n.viewHeight,!1,!1))return e.__dirty&=~je,void(e.__isRendered=!1);var o=e.__clipPaths,a=n.prevElClipPaths,s=!1,l=!1;if(a&&!F_(o,a)||(a&&a.length&&(fb(t,n),t.restore(),l=s=!0,n.prevElClipPaths=null,n.allClipped=!1,n.prevEl=null),o&&o.length&&(fb(t,n),t.save(),ab(o,t,n),s=!0),n.prevElClipPaths=o),n.allClipped)e.__isRendered=!1;else{e.beforeBrush&&e.beforeBrush(),e.innerBeforeBrush();var u=n.prevEl;u||(l=s=!0);var c=e instanceof pl&&e.autoBatch&&hb(e.style);s||sb(r,u.transform)?(fb(t,n),ob(t,e)):c||fb(t,n);var d=pb(e,n.inHover);e instanceof pl?(n.lastDrawType!==lb&&(l=!0,n.lastDrawType=lb),ib(t,e,u,l,n),c&&(n.batchFill||n.batchStroke)||t.beginPath(),J_(t,e,d,c),c&&(n.batchFill=d.fill||"",n.batchStroke=d.stroke||"")):e instanceof yl?(n.lastDrawType!==cb&&(l=!0,n.lastDrawType=cb),ib(t,e,u,l,n),Q_(t,e,d)):e instanceof wl?(n.lastDrawType!==ub&&(l=!0,n.lastDrawType=ub),rb(t,e,u,l,n),K_(t,e,d)):e instanceof Nd&&(n.lastDrawType!==db&&(l=!0,n.lastDrawType=db),yb(t,e,n)),c&&i&&fb(t,n),e.innerAfterBrush(),e.afterBrush&&e.afterBrush(),n.prevEl=e,e.__dirty=0,e.__isRendered=!0}}function yb(t,e,n){var i=e.getDisplayables(),r=e.getTemporalDisplayables();t.save();var o,a,s={prevElClipPaths:null,prevEl:null,allClipped:!1,viewWidth:n.viewWidth,viewHeight:n.viewHeight,inHover:n.inHover};for(o=e.getCursor(),a=i.length;o<a;o++){var l=i[o];l.beforeBrush&&l.beforeBrush(),l.innerBeforeBrush(),gb(t,l,s,o===a-1),l.innerAfterBrush(),l.afterBrush&&l.afterBrush(),s.prevEl=l}for(var u=0,c=r.length;u<c;u++){l=r[u];l.beforeBrush&&l.beforeBrush(),l.innerBeforeBrush(),gb(t,l,s,u===c-1),l.innerAfterBrush(),l.afterBrush&&l.afterBrush(),s.prevEl=l}e.clearTemporalDisplayables(),e.notClear=!0,t.restore()}var vb,_b=new C_,bb=new nn(100),xb=["symbol","symbolSize","symbolKeepAspect","color","backgroundColor","dashArrayX","dashArrayY","maxTileWidth","maxTileHeight"];function wb(t,e){if("none"===t)return null;var n=e.getDevicePixelRatio(),i=e.getZr(),r="svg"===i.painter.type;t.dirty&&_b["delete"](t);var o=_b.get(t);if(o)return o;var a=A(t,{symbol:"rect",symbolSize:1,symbolKeepAspect:!0,color:"rgba(0, 0, 0, 0.2)",backgroundColor:null,dashArrayX:5,dashArrayY:5,rotation:0,maxTileWidth:512,maxTileHeight:512});"none"===a.backgroundColor&&(a.backgroundColor=null);var s={repeat:"repeat"};return l(s),s.rotation=a.rotation,s.scaleX=s.scaleY=r?1:1/n,_b.set(t,s),t.dirty=!1,s;function l(t){for(var e,o=[n],s=!0,l=0;l<xb.length;++l){var u=a[xb[l]],c=typeof u;if(null!=u&&!$(u)&&"string"!==c&&"number"!==c&&"boolean"!==c){s=!1;break}o.push(u)}if(s){e=o.join(",")+(r?"-svg":"");var d=bb.get(e);d&&(r?t.svgElement=d:t.image=d)}var h,f=Sb(a.dashArrayX),p=Lb(a.dashArrayY),m=Mb(a.symbol),g=kb(f),y=Tb(p),v=!r&&Y(),_=r&&i.painter.createSVGElement("g"),b=x();function x(){for(var t=1,e=0,n=g.length;e<n;++e)t=ro(t,g[e]);var i=1;for(e=0,n=m.length;e<n;++e)i=ro(i,m[e].length);t*=i;var r=y*g.length*m.length;return{width:Math.max(1,Math.min(t,a.maxTileWidth)),height:Math.max(1,Math.min(r,a.maxTileHeight))}}function w(){h&&(h.clearRect(0,0,v.width,v.height),a.backgroundColor&&(h.fillStyle=a.backgroundColor,h.fillRect(0,0,v.width,v.height)));for(var t=0,e=0;e<p.length;++e)t+=p[e];if(!(t<=0)){var o=-y,s=0,l=0,u=0;while(o<b.height){if(s%2===0){var c=l/2%m.length,d=0,g=0,x=0;while(d<2*b.width){var w=0;for(e=0;e<f[u].length;++e)w+=f[u][e];if(w<=0)break;if(g%2===0){var M=.5*(1-a.symbolSize),S=d+f[u][g]*M,L=o+p[s]*M,k=f[u][g]*a.symbolSize,T=p[s]*a.symbolSize,D=x/2%m[c].length;C(S,L,k,T,m[c][D])}d+=f[u][g],++x,++g,g===f[u].length&&(g=0)}++u,u===f.length&&(u=0)}o+=p[s],++l,++s,s===p.length&&(s=0)}}function C(t,e,o,s,l){var u=r?1:n,c=j_(l,t*u,e*u,o*u,s*u,a.color,a.symbolKeepAspect);r?_.appendChild(i.painter.paintOne(c)):mb(h,c)}}v&&(v.width=b.width*n,v.height=b.height*n,h=v.getContext("2d")),w(),s&&bb.put(e,v||_),t.image=v,t.svgElement=_,t.svgWidth=b.width,t.svgHeight=b.height}}function Mb(t){if(!t||0===t.length)return[["rect"]];if("string"===typeof t)return[[t]];for(var e=!0,n=0;n<t.length;++n)if("string"!==typeof t[n]){e=!1;break}if(e)return Mb([t]);var i=[];for(n=0;n<t.length;++n)"string"===typeof t[n]?i.push([t[n]]):i.push(t[n]);return i}function Sb(t){if(!t||0===t.length)return[[0,0]];if("number"===typeof t){var e=Math.ceil(t);return[[e,e]]}for(var n=!0,i=0;i<t.length;++i)if("number"!==typeof t[i]){n=!1;break}if(n)return Sb([t]);var r=[];for(i=0;i<t.length;++i)if("number"===typeof t[i]){e=Math.ceil(t[i]);r.push([e,e])}else{e=H(t[i],(function(t){return Math.ceil(t)}));e.length%2===1?r.push(e.concat(e)):r.push(e)}return r}function Lb(t){if(!t||"object"===typeof t&&0===t.length)return[0,0];if("number"===typeof t){var e=Math.ceil(t);return[e,e]}var n=H(t,(function(t){return Math.ceil(t)}));return t.length%2?n.concat(n):n}function kb(t){return H(t,(function(t){return Tb(t)}))}function Tb(t){for(var e=0,n=0;n<t.length;++n)e+=t[n];return t.length%2===1?2*e:e}function Db(t,e){t.eachRawSeries((function(n){if(!t.isSeriesFiltered(n)){var i=n.getData();i.hasItemVisual()&&i.each((function(t){var n=i.getItemVisual(t,"decal");if(n){var r=i.ensureUniqueItemVisual(t,"style");r.decal=wb(n,e)}}));var r=i.getVisual("decal");if(r){var o=i.getVisual("style");o.decal=wb(r,e)}}}))}function Cb(t){if(q(t)){var e=new DOMParser;t=e.parseFromString(t,"text/xml")}var n=t;9===n.nodeType&&(n=n.firstChild);while("svg"!==n.nodeName.toLowerCase()||1!==n.nodeType)n=n.nextSibling;return n}var Ib={fill:"fill",stroke:"stroke","stroke-width":"lineWidth",opacity:"opacity","fill-opacity":"fillOpacity","stroke-opacity":"strokeOpacity","stroke-dasharray":"lineDash","stroke-dashoffset":"lineDashOffset","stroke-linecap":"lineCap","stroke-linejoin":"lineJoin","stroke-miterlimit":"miterLimit","font-family":"fontFamily","font-size":"fontSize","font-style":"fontStyle","font-weight":"fontWeight","text-anchor":"textAlign",visibility:"visibility",display:"display"},Ab=B(Ib),Yb={"alignment-baseline":"textBaseline","stop-color":"stopColor"},Ob=B(Yb),Pb=function(){function t(){this._defs={},this._root=null}return t.prototype.parse=function(t,e){e=e||{};var n=Cb(t);if(!n)throw new Error("Illegal svg");this._defsUsePending=[];var i=new Mr;this._root=i;var r=[],o=n.getAttribute("viewBox")||"",a=parseFloat(n.getAttribute("width")||e.width),s=parseFloat(n.getAttribute("height")||e.height);isNaN(a)&&(a=null),isNaN(s)&&(s=null),zb(n,i,null,!0,!1);var l,u,c=n.firstChild;while(c)this._parseNode(c,i,r,null,!1,!1),c=c.nextSibling;if(Vb(this._defs,this._defsUsePending),this._defsUsePending=[],o){var d=$b(o);d.length>=4&&(l={x:parseFloat(d[0]||0),y:parseFloat(d[1]||0),width:parseFloat(d[2]),height:parseFloat(d[3])})}if(l&&null!=a&&null!=s&&(u=Qb(l,{x:0,y:0,width:a,height:s}),!e.ignoreViewBox)){var h=i;i=new Mr,i.add(h),h.scaleX=h.scaleY=u.scale,h.x=u.x,h.y=u.y}return e.ignoreRootClip||null==a||null==s||i.setClipPath(new Al({shape:{x:0,y:0,width:a,height:s}})),{root:i,width:a,height:s,viewBoxRect:l,viewBoxTransform:u,named:r}},t.prototype._parseNode=function(t,e,n,i,r,o){var a,s=t.nodeName.toLowerCase(),l=i;if("defs"===s&&(r=!0),"text"===s&&(o=!0),"defs"===s||"switch"===s)a=e;else{if(!r){var u=vb[s];if(u&&_t(vb,s)){a=u.call(this,t,e);var c=t.getAttribute("name");if(c){var d={name:c,namedFrom:null,svgNodeTagLower:s,el:a};n.push(d),"g"===s&&(l=d)}else i&&n.push({name:i.name,namedFrom:i,svgNodeTagLower:s,el:a});e.add(a)}}var h=Eb[s];if(h&&_t(Eb,s)){var f=h.call(this,t),p=t.getAttribute("id");p&&(this._defs[p]=f)}}if(a&&a.isGroup){var m=t.firstChild;while(m)1===m.nodeType?this._parseNode(m,a,n,l,r,o):3===m.nodeType&&o&&this._parseText(m,a),m=m.nextSibling}},t.prototype._parseText=function(t,e){var n=new yl({style:{text:t.textContent},silent:!0,x:this._textX||0,y:this._textY||0});Hb(e,n),zb(t,n,this._defsUsePending,!1,!1),Wb(n,e);var i=n.style,r=i.fontSize;r&&r<9&&(i.fontSize=9,n.scaleX*=r/9,n.scaleY*=r/9);var o=(i.fontSize||i.fontFamily)&&[i.fontStyle,i.fontWeight,(i.fontSize||12)+"px",i.fontFamily||"sans-serif"].join(" ");i.font=o;var a=n.getBoundingRect();return this._textX+=a.width,e.add(n),n},t.internalField=function(){vb={g:function(t,e){var n=new Mr;return Hb(e,n),zb(t,n,this._defsUsePending,!1,!1),n},rect:function(t,e){var n=new Al;return Hb(e,n),zb(t,n,this._defsUsePending,!1,!1),n.setShape({x:parseFloat(t.getAttribute("x")||"0"),y:parseFloat(t.getAttribute("y")||"0"),width:parseFloat(t.getAttribute("width")||"0"),height:parseFloat(t.getAttribute("height")||"0")}),n.silent=!0,n},circle:function(t,e){var n=new Dc;return Hb(e,n),zb(t,n,this._defsUsePending,!1,!1),n.setShape({cx:parseFloat(t.getAttribute("cx")||"0"),cy:parseFloat(t.getAttribute("cy")||"0"),r:parseFloat(t.getAttribute("r")||"0")}),n.silent=!0,n},line:function(t,e){var n=new dd;return Hb(e,n),zb(t,n,this._defsUsePending,!1,!1),n.setShape({x1:parseFloat(t.getAttribute("x1")||"0"),y1:parseFloat(t.getAttribute("y1")||"0"),x2:parseFloat(t.getAttribute("x2")||"0"),y2:parseFloat(t.getAttribute("y2")||"0")}),n.silent=!0,n},ellipse:function(t,e){var n=new Ac;return Hb(e,n),zb(t,n,this._defsUsePending,!1,!1),n.setShape({cx:parseFloat(t.getAttribute("cx")||"0"),cy:parseFloat(t.getAttribute("cy")||"0"),rx:parseFloat(t.getAttribute("rx")||"0"),ry:parseFloat(t.getAttribute("ry")||"0")}),n.silent=!0,n},polygon:function(t,e){var n,i=t.getAttribute("points");i&&(n=jb(i));var r=new rd({shape:{points:n||[]},silent:!0});return Hb(e,r),zb(t,r,this._defsUsePending,!1,!1),r},polyline:function(t,e){var n,i=t.getAttribute("points");i&&(n=jb(i));var r=new sd({shape:{points:n||[]},silent:!0});return Hb(e,r),zb(t,r,this._defsUsePending,!1,!1),r},image:function(t,e){var n=new wl;return Hb(e,n),zb(t,n,this._defsUsePending,!1,!1),n.setStyle({image:t.getAttribute("xlink:href"),x:+t.getAttribute("x"),y:+t.getAttribute("y"),width:+t.getAttribute("width"),height:+t.getAttribute("height")}),n.silent=!0,n},text:function(t,e){var n=t.getAttribute("x")||"0",i=t.getAttribute("y")||"0",r=t.getAttribute("dx")||"0",o=t.getAttribute("dy")||"0";this._textX=parseFloat(n)+parseFloat(r),this._textY=parseFloat(i)+parseFloat(o);var a=new Mr;return Hb(e,a),zb(t,a,this._defsUsePending,!1,!0),a},tspan:function(t,e){var n=t.getAttribute("x"),i=t.getAttribute("y");null!=n&&(this._textX=parseFloat(n)),null!=i&&(this._textY=parseFloat(i));var r=t.getAttribute("dx")||"0",o=t.getAttribute("dy")||"0",a=new Mr;return Hb(e,a),zb(t,a,this._defsUsePending,!1,!0),this._textX+=parseFloat(r),this._textY+=parseFloat(o),a},path:function(t,e){var n=t.getAttribute("d")||"",i=Mc(n);return Hb(e,i),zb(t,i,this._defsUsePending,!1,!1),i.silent=!0,i}}}(),t}(),Eb={lineargradient:function(t){var e=parseInt(t.getAttribute("x1")||"0",10),n=parseInt(t.getAttribute("y1")||"0",10),i=parseInt(t.getAttribute("x2")||"10",10),r=parseInt(t.getAttribute("y2")||"0",10),o=new Ld(e,n,i,r);return Nb(t,o),Rb(t,o),o},radialgradient:function(t){var e=parseInt(t.getAttribute("cx")||"0",10),n=parseInt(t.getAttribute("cy")||"0",10),i=parseInt(t.getAttribute("r")||"0",10),r=new Td(e,n,i);return Nb(t,r),Rb(t,r),r}};function Nb(t,e){var n=t.getAttribute("gradientUnits");"userSpaceOnUse"===n&&(e.global=!0)}function Rb(t,e){var n=t.firstChild;while(n){if(1===n.nodeType&&"stop"===n.nodeName.toLocaleLowerCase()){var i=n.getAttribute("offset"),r=void 0;r=i&&i.indexOf("%")>0?parseInt(i,10)/100:i?parseFloat(i):0;var o={};Jb(n,o,o);var a=o.stopColor||n.getAttribute("stop-color")||"#000000";e.colorStops.push({offset:r,color:a})}n=n.nextSibling}}function Hb(t,e){t&&t.__inheritedStyle&&(e.__inheritedStyle||(e.__inheritedStyle={}),A(e.__inheritedStyle,t.__inheritedStyle))}function jb(t){for(var e=$b(t),n=[],i=0;i<e.length;i+=2){var r=parseFloat(e[i]),o=parseFloat(e[i+1]);n.push([r,o])}return n}function zb(t,e,n,i,r){var o=e,a=o.__inheritedStyle=o.__inheritedStyle||{},s={};1===t.nodeType&&(Xb(t,e),Jb(t,a,s),i||Kb(t,a,s)),o.style=o.style||{},null!=a.fill&&(o.style.fill=Fb(o,"fill",a.fill,n)),null!=a.stroke&&(o.style.stroke=Fb(o,"stroke",a.stroke,n)),R(["lineWidth","opacity","fillOpacity","strokeOpacity","miterLimit","fontSize"],(function(t){null!=a[t]&&(o.style[t]=parseFloat(a[t]))})),R(["lineDashOffset","lineCap","lineJoin","fontWeight","fontFamily","fontStyle","textAlign"],(function(t){null!=a[t]&&(o.style[t]=a[t])})),r&&(o.__selfStyle=s),a.lineDash&&(o.style.lineDash=H($b(a.lineDash),(function(t){return parseFloat(t)}))),"hidden"!==a.visibility&&"collapse"!==a.visibility||(o.invisible=!0),"none"===a.display&&(o.ignore=!0)}function Wb(t,e){var n=e.__selfStyle;if(n){var i=n.textBaseline,r=i;i&&"auto"!==i?"baseline"===i?r="alphabetic":"before-edge"===i||"text-before-edge"===i?r="top":"after-edge"===i||"text-after-edge"===i?r="bottom":"central"!==i&&"mathematical"!==i||(r="middle"):r="alphabetic",t.style.textBaseline=r}var o=e.__inheritedStyle;if(o){var a=o.textAlign,s=a;a&&("middle"===a&&(s="center"),t.style.textAlign=s)}}var Bb=/^url\(\s*#(.*?)\)/;function Fb(t,e,n,i){var r=n&&n.match(Bb);if(!r)return"none"===n&&(n=null),n;var o=dt(r[1]);i.push([t,e,o])}function Vb(t,e){for(var n=0;n<e.length;n++){var i=e[n];i[0].style[i[1]]=t[i[2]]}}var Gb=/-?([0-9]*\.)?[0-9]+([eE]-?[0-9]+)?/g;function $b(t){return t.match(Gb)||[]}var Ub=/(translate|scale|rotate|skewX|skewY|matrix)\(([\-\s0-9\.eE,]*)\)/g,qb=Math.PI/180;function Xb(t,e){var n=t.getAttribute("transform");if(n){n=n.replace(/,/g," ");var i=[],r=null;n.replace(Ub,(function(t,e,n){return i.push(e,n),""}));for(var o=i.length-1;o>0;o-=2){var a=i[o],s=i[o-1],l=$b(a);switch(r=r||bi(),s){case"translate":Si(r,r,[parseFloat(l[0]),parseFloat(l[1]||"0")]);break;case"scale":ki(r,r,[parseFloat(l[0]),parseFloat(l[1]||l[0])]);break;case"rotate":Li(r,r,-parseFloat(l[0])*qb);break;case"skewX":var u=Math.tan(parseFloat(l[0])*qb);Mi(r,[1,0,u,1,0,0],r);break;case"skewY":var c=Math.tan(parseFloat(l[0])*qb);Mi(r,[1,c,0,1,0,0],r);break;case"matrix":r[0]=parseFloat(l[0]),r[1]=parseFloat(l[1]),r[2]=parseFloat(l[2]),r[3]=parseFloat(l[3]),r[4]=parseFloat(l[4]),r[5]=parseFloat(l[5]);break}}e.setLocalTransform(r)}}var Zb=/([^\s:;]+)\s*:\s*([^:;]+)/g;function Jb(t,e,n){var i=t.getAttribute("style");if(i){var r;Zb.lastIndex=0;while(null!=(r=Zb.exec(i))){var o=r[1],a=_t(Ib,o)?Ib[o]:null;a&&(e[a]=r[2]);var s=_t(Yb,o)?Yb[o]:null;s&&(n[s]=r[2])}}}function Kb(t,e,n){for(var i=0;i<Ab.length;i++){var r=Ab[i],o=t.getAttribute(r);null!=o&&(e[Ib[r]]=o)}for(i=0;i<Ob.length;i++){r=Ob[i],o=t.getAttribute(r);null!=o&&(n[Yb[r]]=o)}}function Qb(t,e){var n=e.width/t.width,i=e.height/t.height,r=Math.min(n,i);return{scale:r,x:-(t.x+t.width/2)*r+(e.x+e.width/2),y:-(t.y+t.height/2)*r+(e.y+e.height/2)}}function tx(t,e){var n=new Pb;return n.parse(t,e)}var ex=1e-8;function nx(t,e){return Math.abs(t-e)<ex}function ix(t,e,n){var i=0,r=t[0];if(!r)return!1;for(var o=1;o<t.length;o++){var a=t[o];i+=Zs(r[0],r[1],a[0],a[1],e,n),r=a}var s=t[0];return nx(r[0],s[0])&&nx(r[1],s[1])||(i+=Zs(r[0],r[1],s[0],s[1],e,n)),0!==i}var rx=[],ox=function(){function t(t){this.name=t}return t.prototype.getCenter=function(){},t}(),ax=function(t){function e(e,n,i){var r=t.call(this,e)||this;if(r.type="geoJSON",r.geometries=n,i)i=[i[0],i[1]];else{var o=r.getBoundingRect();i=[o.x+o.width/2,o.y+o.height/2]}return r._center=i,r}return o(e,t),e.prototype.getBoundingRect=function(){var t=this._rect;if(t)return t;for(var e=Number.MAX_VALUE,n=[e,e],i=[-e,-e],r=[],o=[],a=this.geometries,s=0;s<a.length;s++)if("polygon"===a[s].type){var l=a[s].exterior;ms(l,r,o),zt(n,n,r),Wt(i,i,o)}return 0===s&&(n[0]=n[1]=i[0]=i[1]=0),this._rect=new Ji(n[0],n[1],i[0]-n[0],i[1]-n[1])},e.prototype.contain=function(t){var e=this.getBoundingRect(),n=this.geometries;if(!e.contain(t[0],t[1]))return!1;t:for(var i=0,r=n.length;i<r;i++)if("polygon"===n[i].type){var o=n[i].exterior,a=n[i].interiors;if(ix(o,t[0],t[1])){for(var s=0;s<(a?a.length:0);s++)if(ix(a[s],t[0],t[1]))continue t;return!0}}return!1},e.prototype.transformTo=function(t,e,n,i){var r=this.getBoundingRect(),o=r.width/r.height;n?i||(i=n/o):n=o*i;for(var a=new Ji(t,e,n,i),s=r.calculateTransform(a),l=this.geometries,u=0;u<l.length;u++)if("polygon"===l[u].type){for(var c=l[u].exterior,d=l[u].interiors,h=0;h<c.length;h++)jt(c[h],c[h],s);for(var f=0;f<(d?d.length:0);f++)for(h=0;h<d[f].length;h++)jt(d[f][h],d[f][h],s)}r=this._rect,r.copy(a),this._center=[r.x+r.width/2,r.y+r.height/2]},e.prototype.cloneShallow=function(t){null==t&&(t=this.name);var n=new e(t,this.geometries,this._center);return n._rect=this._rect,n.transformTo=null,n},e.prototype.getCenter=function(){return this._center},e.prototype.setCenter=function(t){this._center=t},e}(ox),sx=function(t){function e(e,n){var i=t.call(this,e)||this;return i.type="geoSVG",i._elOnlyForCalculate=n,i}return o(e,t),e.prototype.getCenter=function(){var t=this._center;return t||(t=this._center=this._calculateCenter()),t},e.prototype._calculateCenter=function(){var t=this._elOnlyForCalculate,e=t.getBoundingRect(),n=[e.x+e.width/2,e.y+e.height/2],i=xi(rx),r=t;while(r&&!r.isGeoSVGGraphicRoot)Mi(i,r.getLocalTransform(),i),r=r.parent;return Ti(i,i),jt(n,n,i),n},e}(ox),lx=gt(["rect","circle","line","ellipse","polygon","polyline","path","text","tspan","g"]),ux=function(){function t(t,e){this.type="geoSVG",this._usedGraphicMap=gt(),this._freedGraphics=[],this._mapName=t,this._parsedXML=Cb(e)}return t.prototype.load=function(){var t=this._firstGraphic;if(!t){t=this._firstGraphic=this._buildGraphic(this._parsedXML),this._freedGraphics.push(t),this._boundingRect=this._firstGraphic.boundingRect.clone();var e=dx(t.named),n=e.regions,i=e.regionsMap;this._regions=n,this._regionsMap=i}return{boundingRect:this._boundingRect,regions:this._regions,regionsMap:this._regionsMap}},t.prototype._buildGraphic=function(t){var e,n;try{e=t&&tx(t,{ignoreViewBox:!0,ignoreRootClip:!0})||{},n=e.root,ct(null!=n)}catch(m){throw new Error("Invalid svg format\n"+m.message)}var i=new Mr;i.add(n),i.isGeoSVGGraphicRoot=!0;var r=e.width,o=e.height,a=e.viewBoxRect,s=this._boundingRect;if(!s){var l=void 0,u=void 0,c=void 0,d=void 0;if(null!=r?(l=0,c=r):a&&(l=a.x,c=a.width),null!=o?(u=0,d=o):a&&(u=a.y,d=a.height),null==l||null==u){var h=n.getBoundingRect();null==l&&(l=h.x,c=h.width),null==u&&(u=h.y,d=h.height)}s=this._boundingRect=new Ji(l,u,c,d)}if(a){var f=Qb(a,s);n.scaleX=n.scaleY=f.scale,n.x=f.x,n.y=f.y}i.setClipPath(new Al({shape:s.plain()}));var p=[];return R(e.named,(function(t){null!=lx.get(t.svgNodeTagLower)&&(p.push(t),cx(t.el))})),{root:i,boundingRect:s,named:p}},t.prototype.useGraphic=function(t){var e=this._usedGraphicMap,n=e.get(t);return n||(n=this._freedGraphics.pop()||this._buildGraphic(this._parsedXML),e.set(t,n),n)},t.prototype.freeGraphic=function(t){var e=this._usedGraphicMap,n=e.get(t);n&&(e.removeKey(t),this._freedGraphics.push(n))},t}();function cx(t){t.silent=!1,t.isGroup&&t.traverse((function(t){t.silent=!1}))}function dx(t){var e=[],n=gt();return R(t,(function(t){if(null==t.namedFrom){var i=new sx(t.name,t.el);e.push(i),n.set(t.name,i)}})),{regions:e,regionsMap:n}}function hx(t){if(!t.UTF8Encoding)return t;var e=t,n=e.UTF8Scale;null==n&&(n=1024);for(var i=e.features,r=0;r<i.length;r++){var o=i[r],a=o.geometry;if("Polygon"===a.type)for(var s=a.coordinates,l=0;l<s.length;l++)s[l]=fx(s[l],a.encodeOffsets[l],n);else if("MultiPolygon"===a.type)for(s=a.coordinates,l=0;l<s.length;l++)for(var u=s[l],c=0;c<u.length;c++)u[c]=fx(u[c],a.encodeOffsets[l][c],n)}return e.UTF8Encoding=!1,e}function fx(t,e,n){for(var i=[],r=e[0],o=e[1],a=0;a<t.length;a+=2){var s=t.charCodeAt(a)-64,l=t.charCodeAt(a+1)-64;s=s>>1^-(1&s),l=l>>1^-(1&l),s+=r,l+=o,r=s,o=l,i.push([s/n,l/n])}return i}function px(t,e){return t=hx(t),H(z(t.features,(function(t){return t.geometry&&t.properties&&t.geometry.coordinates.length>0})),(function(t){var n=t.properties,i=t.geometry,r=[];if("Polygon"===i.type){var o=i.coordinates;r.push({type:"polygon",exterior:o[0],interiors:o.slice(1)})}if("MultiPolygon"===i.type){o=i.coordinates;R(o,(function(t){t[0]&&r.push({type:"polygon",exterior:t[0],interiors:t.slice(1)})}))}var a=new ax(n[e||"name"],r,n.cp);return a.properties=n,a}))}for(var mx=[126,25],gx="南海诸岛",yx=[[[0,3.5],[7,11.2],[15,11.9],[30,7],[42,.7],[52,.7],[56,7.7],[59,.7],[64,.7],[64,0],[5,0],[0,3.5]],[[13,16.1],[19,14.7],[16,21.7],[11,23.1],[13,16.1]],[[12,32.2],[14,38.5],[15,38.5],[13,32.2],[12,32.2]],[[16,47.6],[12,53.2],[13,53.2],[18,47.6],[16,47.6]],[[6,64.4],[8,70],[9,70],[8,64.4],[6,64.4]],[[23,82.6],[29,79.8],[30,79.8],[25,82.6],[23,82.6]],[[37,70.7],[43,62.3],[44,62.3],[39,70.7],[37,70.7]],[[48,51.1],[51,45.5],[53,45.5],[50,51.1],[48,51.1]],[[51,35],[51,28.7],[53,28.7],[53,35],[51,35]],[[52,22.4],[55,17.5],[56,17.5],[53,22.4],[52,22.4]],[[58,12.6],[62,7],[63,7],[60,12.6],[58,12.6]],[[0,3.5],[0,93.1],[64,93.1],[64,0],[63,0],[63,92.4],[1,92.4],[1,3.5],[0,3.5]]],vx=0;vx<yx.length;vx++)for(var _x=0;_x<yx[vx].length;_x++)yx[vx][_x][0]/=10.5,yx[vx][_x][1]/=-14,yx[vx][_x][0]+=mx[0],yx[vx][_x][1]+=mx[1];function bx(t,e){if("china"===t){for(var n=0;n<e.length;n++)if(e[n].name===gx)return;e.push(new ax(gx,H(yx,(function(t){return{type:"polygon",exterior:t}})),mx))}}var xx={"南海诸岛":[32,80],"广东":[0,-10],"香港":[10,5],"澳门":[-10,10],"天津":[5,5]};function Mx(t,e){if("china"===t){var n=xx[e.name];if(n){var i=e.getCenter();i[0]+=n[0]/10.5,i[1]+=-n[1]/14,e.setCenter(i)}}}var Sx={Russia:[100,60],"United States":[-99,38],"United States of America":[-99,38]};function Lx(t,e){if("world"===t){var n=Sx[e.name];if(n){var i=[n[0],n[1]];e.setCenter(i)}}}var kx=[[[123.45165252685547,25.73527164402261],[123.49731445312499,25.73527164402261],[123.49731445312499,25.750734064600884],[123.45165252685547,25.750734064600884],[123.45165252685547,25.73527164402261]]];function Tx(t,e){"china"===t&&"台湾"===e.name&&e.geometries.push({type:"polygon",exterior:kx[0]})}var Dx="name",Cx=function(){function t(t,e,n){this.type="geoJSON",this._parsedMap=gt(),this._mapName=t,this._specialAreas=n,this._geoJSON=Ax(e)}return t.prototype.load=function(t,e){e=e||Dx;var n=this._parsedMap.get(e);if(!n){var i=this._parseToRegions(e);n=this._parsedMap.set(e,{regions:i,boundingRect:Ix(i)})}var r=gt(),o=[];return R(n.regions,(function(e){var n=e.name;t&&t.hasOwnProperty(n)&&(e=e.cloneShallow(n=t[n])),o.push(e),r.set(n,e)})),{regions:o,boundingRect:n.boundingRect||new Ji(0,0,0,0),regionsMap:r}},t.prototype._parseToRegions=function(t){var e,n=this._mapName,i=this._geoJSON;try{e=i?px(i,t):[]}catch(r){throw new Error("Invalid geoJson format\n"+r.message)}return bx(n,e),R(e,(function(t){var e=t.name;Mx(n,t),Lx(n,t),Tx(n,t);var i=this._specialAreas&&this._specialAreas[e];i&&t.transformTo(i.left,i.top,i.width,i.height)}),this),e},t.prototype.getMapForUser=function(){return{geoJson:this._geoJSON,geoJSON:this._geoJSON,specialAreas:this._specialAreas}},t}();function Ix(t){for(var e,n=0;n<t.length;n++){var i=t[n].getBoundingRect();e=e||i.clone(),e.union(i)}return e}function Ax(t){return q(t)?"undefined"!==typeof JSON&&JSON.parse?JSON.parse(t):new Function("return ("+t+");")():t}var Yx=gt(),Ox={registerMap:function(t,e,n){if(e.svg){var i=new ux(t,e.svg);Yx.set(t,i)}else{var r=e.geoJson||e.geoJSON;r&&!e.features?n=e.specialAreas:r=e;i=new Cx(t,r,n);Yx.set(t,i)}},getGeoResource:function(t){return Yx.get(t)},getMapForUser:function(t){var e=Yx.get(t);return e&&"geoJSON"===e.type&&e.getMapForUser()},load:function(t,e,n){var i=Yx.get(t);if(i)return i.load(e,n)}},Px=ct,Ex=R,Nx=U,Rx=J,Hx=O,jx="undefined"!==typeof window,zx=1,Wx=800,Bx=900,Fx=1e3,Vx=2e3,Gx=5e3,$x=1e3,Ux=1100,qx=2e3,Xx=3e3,Zx=4e3,Jx=4500,Kx=4600,Qx=5e3,tw=6e3,ew=7e3,nw={PROCESSOR:{FILTER:Fx,SERIES_FILTER:Wx,STATISTIC:Gx},VISUAL:{LAYOUT:$x,PROGRESSIVE_LAYOUT:Ux,GLOBAL:qx,CHART:Xx,POST_CHART_LAYOUT:Kx,COMPONENT:Zx,BRUSH:Qx,CHART_ITEM:Jx,ARIA:tw,DECAL:ew}},iw="__flagInMainProcess",rw="__optionUpdated",ow="__needsUpdateStatus",aw=/^[a-zA-Z0-9_]+$/,sw="__connectUpdateStatus",lw=0,uw=1,cw=2;function dw(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];if(!this.isDisposed())return fw(this,t,e);zw(this.id)}}function hw(t){return function(){for(var e=[],n=0;n<arguments.length;n++)e[n]=arguments[n];return fw(this,t,e)}}function fw(t,e,n){return n[0]=n[0]&&n[0].toLowerCase(),$t.prototype[e].apply(t,n)}var pw,mw,gw,yw,vw,_w,bw,xw,ww,Mw,Sw,Lw,kw,Tw,Dw,Cw,Iw,Aw,Yw,Ow,Pw,Ew=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e}($t),Nw=Ew.prototype;Nw.on=hw("on"),Nw.off=hw("off");var Rw=function(t){function e(e,n,i){var r=t.call(this,new Yv)||this;r._chartsViews=[],r._chartsMap={},r._componentsViews=[],r._componentsMap={},r._pendingActions=[],i=i||{},"string"===typeof n&&(n=qw[n]),r._dom=e;var o="canvas",a=!1,s=r._zr=Ir(e,{renderer:i.renderer||o,devicePixelRatio:i.devicePixelRatio,width:i.width,height:i.height,useDirtyRect:null==i.useDirtyRect?a:i.useDirtyRect});r._throttledZrFlush=qy(V(s.flush,s),17),n=T(n),n&&Xm(n,!0),r._theme=n,r._locale=cf(i.locale||lf),r._coordSysMgr=new vm;var l=r._api=Iw(r);function u(t,e){return t.__prio-e.__prio}return He(Uw,u),He(Fw,u),r._scheduler=new Mv(r,l,Fw,Uw),r._messageCenter=new Ew,r._labelManager=new w_,r._initEvents(),r.resize=V(r.resize,r),s.animation.on("frame",r._onframe,r),Mw(s,r),Sw(s,r),ft(r),r}return o(e,t),e.prototype._onframe=function(){if(!this._disposed){Pw(this);var t=this._scheduler;if(this[rw]){var e=this[rw].silent;this[iw]=!0,pw(this),yw.update.call(this),this._zr.flush(),this[iw]=!1,this[rw]=!1,xw.call(this,e),ww.call(this,e)}else if(t.unfinished){var n=zx,i=this._model,r=this._api;t.unfinished=!1;do{var o=+new Date;t.performSeriesTasks(i),t.performDataProcessorTasks(i),_w(this,i),t.performVisualTasks(i),Dw(this,this._model,r,"remain"),n-=+new Date-o}while(n>0&&t.unfinished);t.unfinished||this._zr.flush()}}},e.prototype.getDom=function(){return this._dom},e.prototype.getId=function(){return this.id},e.prototype.getZr=function(){return this._zr},e.prototype.setOption=function(t,e,n){if(this._disposed)zw(this.id);else{var i,r,o;if(Rx(e)&&(n=e.lazyUpdate,i=e.silent,r=e.replaceMerge,o=e.transition,e=e.notMerge),this[iw]=!0,!this._model||e){var a=new Lm(this._api),s=this._theme,l=this._model=new hm;l.scheduler=this._scheduler,l.init(null,null,null,s,this._locale,a)}this._model.setOption(t,{replaceMerge:r},Vw),Yw(this,o),n?(this[rw]={silent:i},this[iw]=!1,this.getZr().wakeUp()):(pw(this),yw.update.call(this),this._zr.flush(),this[rw]=!1,this[iw]=!1,xw.call(this,i),ww.call(this,i))}},e.prototype.setTheme=function(){},e.prototype.getModel=function(){return this._model},e.prototype.getOption=function(){return this._model&&this._model.getOption()},e.prototype.getWidth=function(){return this._zr.getWidth()},e.prototype.getHeight=function(){return this._zr.getHeight()},e.prototype.getDevicePixelRatio=function(){return this._zr.painter.dpr||jx&&window.devicePixelRatio||1},e.prototype.getRenderedCanvas=function(t){if(h.canvasSupported)return t=t||{},this._zr.painter.getRenderedCanvas({backgroundColor:t.backgroundColor||this._model.get("backgroundColor"),pixelRatio:t.pixelRatio||this.getDevicePixelRatio()})},e.prototype.getSvgDataURL=function(){if(h.svgSupported){var t=this._zr,e=t.storage.getDisplayList();return R(e,(function(t){t.stopAnimation(null,!0)})),t.painter.toDataURL()}},e.prototype.getDataURL=function(t){if(!this._disposed){t=t||{};var e=t.excludeComponents,n=this._model,i=[],r=this;Ex(e,(function(t){n.eachComponent({mainType:t},(function(t){var e=r._componentsMap[t.__viewId];e.group.ignore||(i.push(e),e.group.ignore=!0)}))}));var o="svg"===this._zr.painter.getType()?this.getSvgDataURL():this.getRenderedCanvas(t).toDataURL("image/"+(t&&t.type||"png"));return Ex(i,(function(t){t.group.ignore=!1})),o}zw(this.id)},e.prototype.getConnectedDataURL=function(t){if(this._disposed)zw(this.id);else if(h.canvasSupported){var e="svg"===t.type,n=this.group,i=Math.min,r=Math.max,o=1/0;if(Jw[n]){var a=o,s=o,l=-o,u=-o,c=[],d=t&&t.pixelRatio||this.getDevicePixelRatio();R(Zw,(function(o,d){if(o.group===n){var h=e?o.getZr().painter.getSvgDom().innerHTML:o.getRenderedCanvas(T(t)),f=o.getDom().getBoundingClientRect();a=i(f.left,a),s=i(f.top,s),l=r(f.right,l),u=r(f.bottom,u),c.push({dom:h,left:f.left,top:f.top})}})),a*=d,s*=d,l*=d,u*=d;var f=l-a,p=u-s,m=Y(),g=Ir(m,{renderer:e?"svg":"canvas"});if(g.resize({width:f,height:p}),e){var y="";return Ex(c,(function(t){var e=t.left-a,n=t.top-s;y+='<g transform="translate('+e+","+n+')">'+t.dom+"</g>"})),g.painter.getSvgRoot().innerHTML=y,t.connectedBackgroundColor&&g.painter.setBackgroundColor(t.connectedBackgroundColor),g.refreshImmediately(),g.painter.toDataURL()}return t.connectedBackgroundColor&&g.add(new Al({shape:{x:0,y:0,width:f,height:p},style:{fill:t.connectedBackgroundColor}})),Ex(c,(function(t){var e=new wl({style:{x:t.left*d-a,y:t.top*d-s,image:t.dom}});g.add(e)})),g.refreshImmediately(),m.toDataURL("image/"+(t&&t.type||"png"))}return this.getDataURL(t)}},e.prototype.convertToPixel=function(t,e){return vw(this,"convertToPixel",t,e)},e.prototype.convertFromPixel=function(t,e){return vw(this,"convertFromPixel",t,e)},e.prototype.containPixel=function(t,e){if(!this._disposed){var n,i=this._model,r=Yo(i,t);return R(r,(function(t,i){i.indexOf("Models")>=0&&R(t,(function(t){var r=t.coordinateSystem;if(r&&r.containPoint)n=n||!!r.containPoint(e);else if("seriesModels"===i){var o=this._chartsMap[t.__viewId];o&&o.containPoint&&(n=n||o.containPoint(e,t))}else 0}),this)}),this),!!n}zw(this.id)},e.prototype.getVisual=function(t,e){var n=this._model,i=Yo(n,t,{defaultMainType:"series"}),r=i.seriesModel;var o=r.getData(),a=i.hasOwnProperty("dataIndexInside")?i.dataIndexInside:i.hasOwnProperty("dataIndex")?o.indexOfRawIndex(i.dataIndex):null;return null!=a?Ev(o,a,e):Nv(o,e)},e.prototype.getViewOfComponentModel=function(t){return this._componentsMap[t.__viewId]},e.prototype.getViewOfSeriesModel=function(t){return this._chartsMap[t.__viewId]},e.prototype._initEvents=function(){var t=this;Ex(jw,(function(e){var n=function(n){var i,r=t.getModel(),o=n.target,a="globalout"===e;if(a?i={}:o&&k_(o,(function(t){var e=$l(t);if(e&&null!=e.dataIndex){var n=e.dataModel||r.getSeriesByIndex(e.seriesIndex);return i=n&&n.getDataParams(e.dataIndex,e.dataType)||{},!0}if(e.eventData)return i=I({},e.eventData),!0}),!0),i){var s=i.componentType,l=i.componentIndex;"markLine"!==s&&"markPoint"!==s&&"markArea"!==s||(s="series",l=i.seriesIndex);var u=s&&null!=l&&r.getComponent(s,l),c=u&&t["series"===u.mainType?"_chartsMap":"_componentsMap"][u.__viewId];0,i.event=n,i.type=e,t._$eventProcessor.eventInfo={targetEl:o,packedEvent:i,model:u,view:c},t.trigger(e,i)}};n.zrEventfulCallAtLast=!0,t._zr.on(e,n,t)})),Ex(Bw,(function(e,n){t._messageCenter.on(n,(function(t){this.trigger(n,t)}),t)})),Ex(["selectchanged"],(function(e){t._messageCenter.on(e,(function(t){this.trigger(e,t)}),t)})),L_(this._messageCenter,this,this._api)},e.prototype.isDisposed=function(){return this._disposed},e.prototype.clear=function(){this._disposed?zw(this.id):this.setOption({series:[]},!0)},e.prototype.dispose=function(){if(this._disposed)zw(this.id);else{this._disposed=!0,Ro(this.getDom(),Qw,"");var t=this._api,e=this._model;Ex(this._componentsViews,(function(n){n.dispose(e,t)})),Ex(this._chartsViews,(function(n){n.dispose(e,t)})),this._zr.dispose(),delete Zw[this.id]}},e.prototype.resize=function(t){if(this._disposed)zw(this.id);else{this._zr.resize(t);var e=this._model;if(this._loadingFX&&this._loadingFX.resize(),e){var n=e.resetOption("media"),i=t&&t.silent;this[rw]&&(null==i&&(i=this[rw].silent),n=!0,this[rw]=!1),this[iw]=!0,n&&pw(this),yw.update.call(this,{type:"resize",animation:I({duration:0},t&&t.animation)}),this[iw]=!1,xw.call(this,i),ww.call(this,i)}}},e.prototype.showLoading=function(t,e){if(this._disposed)zw(this.id);else if(Rx(t)&&(e=t,t=""),t=t||"default",this.hideLoading(),Xw[t]){var n=Xw[t](this._api,e),i=this._zr;this._loadingFX=n,i.add(n)}},e.prototype.hideLoading=function(){this._disposed?zw(this.id):(this._loadingFX&&this._zr.remove(this._loadingFX),this._loadingFX=null)},e.prototype.makeActionFromEvent=function(t){var e=I({},t);return e.type=Bw[t.type],e},e.prototype.dispatchAction=function(t,e){if(this._disposed)zw(this.id);else if(Rx(e)||(e={silent:!!e}),Ww[t.type]&&this._model)if(this[iw])this._pendingActions.push(t);else{var n=e.silent;bw.call(this,t,n);var i=e.flush;i?this._zr.flush():!1!==i&&h.browser.weChat&&this._throttledZrFlush(),xw.call(this,n),ww.call(this,n)}},e.prototype.updateLabelLayout=function(){var t=this._labelManager;t.updateLayoutConfig(this._api),t.layout(this._api),t.processLabelsOverall()},e.prototype.appendData=function(t){if(this._disposed)zw(this.id);else{var e=t.seriesIndex,n=this.getModel(),i=n.getSeriesByIndex(e);0,i.appendData(t),this._scheduler.unfinished=!0,this.getZr().wakeUp()}},e.internalField=function(){function t(t){for(var e=[],n=t.currentStates,i=0;i<n.length;i++){var r=n[i];"emphasis"!==r&&"blur"!==r&&"select"!==r&&e.push(r)}t.selected&&t.states.select&&e.push("select"),t.hoverState===Kl&&t.states.emphasis?e.push("emphasis"):t.hoverState===Jl&&t.states.blur&&e.push("blur"),t.useStates(e)}function e(t,e){var n=t._zr,i=n.storage,r=0;i.traverse((function(t){t.isGroup||r++})),r>e.get("hoverLayerThreshold")&&!h.node&&!h.worker&&e.eachSeries((function(e){if(!e.preventUsingHoverLayer){var n=t._chartsMap[e.__viewId];n.__alive&&n.group.traverse((function(t){t.states.emphasis&&(t.states.emphasis.hoverLayer=!0)}))}}))}function n(t,e){var n=t.get("blendMode")||null;e.group.traverse((function(t){t.isGroup||(t.style.blend=n),t.eachPendingDisplayable&&t.eachPendingDisplayable((function(t){t.style.blend=n}))}))}function i(t,e){t.preventAutoZ||r(e.group,t.get("z")||0,t.get("zlevel")||0,-1/0)}function r(t,e,n,i){var o=t.getTextContent(),a=t.getTextGuideLine(),s=t.isGroup;if(s)for(var l=t.childrenRef(),u=0;u<l.length;u++)i=Math.max(r(l[u],e,n,i),i);else t.z=e,t.zlevel=n,i=Math.max(t.z2,i);if(o&&(o.z=e,o.zlevel=n,isFinite(i)&&(o.z2=i+2)),a){var c=t.textGuideLineConfig;a.z=e,a.zlevel=n,isFinite(i)&&(a.z2=i+(c&&c.showAbove?1:-1))}return i}function a(t,e){e.group.traverse((function(t){if(!oh(t)){var e=t.getTextContent(),n=t.getTextGuideLine();t.stateTransition&&(t.stateTransition=null),e&&e.stateTransition&&(e.stateTransition=null),n&&n.stateTransition&&(n.stateTransition=null),t.hasState()?(t.prevStates=t.currentStates,t.clearStates()):t.prevStates&&(t.prevStates=null)}}))}function s(e,n){var i=e.getModel("stateAnimation"),r=e.isAnimationEnabled(),o=i.get("duration"),a=o>0?{duration:o,delay:i.get("delay"),easing:i.get("easing")}:null;n.group.traverse((function(e){if(e.states&&e.states.emphasis){if(oh(e))return;if(e instanceof pl&&ic(e),e.__dirty){var n=e.prevStates;n&&e.useStates(n)}if(r){e.stateTransition=a;var i=e.getTextContent(),o=e.getTextGuideLine();i&&(i.stateTransition=a),o&&(o.stateTransition=a)}e.__dirty&&t(e)}}))}pw=function(t){var e=t._scheduler;e.restorePipelines(t._model),e.prepareStageTasks(),mw(t,!0),mw(t,!1),e.plan()},mw=function(t,e){for(var n=t._model,i=t._scheduler,r=e?t._componentsViews:t._chartsViews,o=e?t._componentsMap:t._chartsMap,a=t._zr,s=t._api,l=0;l<r.length;l++)r[l].__alive=!1;function u(t){var l=t.__requireNewView;t.__requireNewView=!1;var u="_ec_"+t.id+"_"+t.type,c=!l&&o[u];if(!c){var d=Go(t.type),h=e?Py.getClass(d.main,d.sub):Vy.getClass(d.sub);0,c=new h,c.init(n,s),o[u]=c,r.push(c),a.add(c.group)}t.__viewId=c.__id=u,c.__alive=!0,c.__model=t,c.group.__ecComponentInfo={mainType:t.mainType,index:t.componentIndex},!e&&i.prepareView(c,t,n,s)}e?n.eachComponent((function(t,e){"series"!==t&&u(e)})):n.eachSeries(u);for(l=0;l<r.length;){var c=r[l];c.__alive?l++:(!e&&c.renderTask.dispose(),a.remove(c.group),c.dispose(n,s),r.splice(l,1),o[c.__id]===c&&delete o[c.__id],c.__id=c.group.__ecComponentInfo=null)}},gw=function(t,e,n,i,r){var o=t._model;if(o.setUpdatePayload(n),i){var a={};a[i+"Id"]=n[i+"Id"],a[i+"Index"]=n[i+"Index"],a[i+"Name"]=n[i+"Name"];var s={mainType:i,query:a};r&&(s.subType=r);var l,u=n.excludeSeriesId;null!=u&&(l=gt(),Ex(so(u),(function(t){var e=wo(t,null);null!=e&&l.set(e,!0)}))),nc(n)&&Nu(t._api),o&&o.eachComponent(s,(function(e){if(!l||null==l.get(e.id)){if(nc(n))if(e instanceof Yy)n.type!==iu||n.notBlur||ju(e,n,t._api);else{var r=zu(e.mainType,e.componentIndex,n.name,t._api),o=r.focusSelf,a=r.dispatchers;n.type===iu&&o&&!n.notBlur&&Hu(e.mainType,e.componentIndex,t._api),a&&Ex(a,(function(t){n.type===iu?Cu(t):Iu(t)}))}else ec(n)&&e instanceof Yy&&(Fu(e,n,t._api),Vu(e),Ow(t));c(t["series"===i?"_chartsMap":"_componentsMap"][e.__viewId])}}),t)}else Ex([].concat(t._componentsViews).concat(t._chartsViews),c);function c(i){i&&i.__alive&&i[e]&&i[e](i.__model,o,t._api,n)}},yw={prepareAndUpdate:function(t){pw(this),yw.update.call(this,t)},update:function(t){var e=this._model,n=this._api,i=this._zr,r=this._coordSysMgr,o=this._scheduler;if(e){e.setUpdatePayload(t),o.restoreData(e,t),o.performSeriesTasks(e),r.create(e,n),o.performDataProcessorTasks(e,t),_w(this,e),r.update(e,n),Lw(e),o.performVisualTasks(e,t),kw(this,e,n,t);var a=e.get("backgroundColor")||"transparent",s=e.get("darkMode");if(h.canvasSupported)i.setBackgroundColor(a),null!=s&&"auto"!==s&&i.setDarkMode(s);else{var l=yn(a);a=Ln(l,"rgb"),0===l[3]&&(a="transparent")}Cw(e,n)}},updateTransform:function(t){var e=this,n=this._model,i=this._api;if(n){n.setUpdatePayload(t);var r=[];n.eachComponent((function(o,a){if("series"!==o){var s=e.getViewOfComponentModel(a);if(s&&s.__alive)if(s.updateTransform){var l=s.updateTransform(a,n,i,t);l&&l.update&&r.push(s)}else r.push(s)}}));var o=gt();n.eachSeries((function(r){var a=e._chartsMap[r.__viewId];if(a.updateTransform){var s=a.updateTransform(r,n,i,t);s&&s.update&&o.set(r.uid,1)}else o.set(r.uid,1)})),Lw(n),this._scheduler.performVisualTasks(n,t,{setDirty:!0,dirtyMap:o}),Dw(this,n,i,t,o),Cw(n,this._api)}},updateView:function(t){var e=this._model;e&&(e.setUpdatePayload(t),Vy.markUpdateMethod(t,"updateView"),Lw(e),this._scheduler.performVisualTasks(e,t,{setDirty:!0}),kw(this,this._model,this._api,t),Cw(e,this._api))},updateVisual:function(t){var e=this,n=this._model;n&&(n.setUpdatePayload(t),n.eachSeries((function(t){t.getData().clearAllVisual()})),Vy.markUpdateMethod(t,"updateVisual"),Lw(n),this._scheduler.performVisualTasks(n,t,{visualType:"visual",setDirty:!0}),n.eachComponent((function(i,r){if("series"!==i){var o=e.getViewOfComponentModel(r);o&&o.__alive&&o.updateVisual(r,n,e._api,t)}})),n.eachSeries((function(i){var r=e._chartsMap[i.__viewId];r.updateVisual(i,n,e._api,t)})),Cw(n,this._api))},updateLayout:function(t){yw.update.call(this,t)}},vw=function(t,e,n,i){if(t._disposed)zw(t.id);else{for(var r,o=t._model,a=t._coordSysMgr.getCoordinateSystems(),s=Yo(o,n),l=0;l<a.length;l++){var u=a[l];if(u[e]&&null!=(r=u[e](o,s,i)))return r}0}},_w=function(t,e){var n=t._chartsMap,i=t._scheduler;e.eachSeries((function(t){i.updateStreamModes(t,n[t.__viewId])}))},bw=function(t,e){var n=this,i=this.getModel(),r=t.type,o=t.escapeConnect,a=Ww[r],s=a.actionInfo,l=(s.update||"update").split(":"),u=l.pop(),c=null!=l[0]&&Go(l[0]);this[iw]=!0;var d=[t],h=!1;t.batch&&(h=!0,d=H(t.batch,(function(e){return e=A(I({},e),t),e.batch=null,e})));var f,p=[],m=ec(t),g=nc(t);if(Ex(d,(function(e){if(f=a.action(e,n._model,n._api),f=f||I({},e),f.type=s.event||f.type,p.push(f),g){var i=Oo(t),r=i.queryOptionMap,o=i.mainTypeSpecified,l=o?r.keys()[0]:"series";gw(n,u,e,l),Ow(n)}else m?(gw(n,u,e,"series"),Ow(n)):c&&gw(n,u,e,c.main,c.sub)})),"none"===u||g||m||c||(this[rw]?(pw(this),yw.update.call(this,t),this[rw]=!1):yw[u].call(this,t)),f=h?{type:s.event||r,escapeConnect:o,batch:p}:p[0],this[iw]=!1,!e){var y=this._messageCenter;if(y.trigger(f.type,f),m){var v={type:"selectchanged",escapeConnect:o,selected:Gu(i),isFromClick:t.isFromClick||!1,fromAction:t.type,fromActionPayload:t};y.trigger(v.type,v)}}},xw=function(t){var e=this._pendingActions;while(e.length){var n=e.shift();bw.call(this,n,t)}},ww=function(t){!t&&this.trigger("updated")},Mw=function(t,e){t.on("rendered",(function(n){e.trigger("rendered",n),!t.animation.isFinished()||e[rw]||e._scheduler.unfinished||e._pendingActions.length||e.trigger("finished")}))},Sw=function(t,e){t.on("mouseover",(function(t){var n=t.target,i=k_(n,Ku);i&&(Wu(i,t,e._api),Ow(e))})).on("mouseout",(function(t){var n=t.target,i=k_(n,Ku);i&&(Bu(i,t,e._api),Ow(e))})).on("click",(function(t){var n=t.target,i=k_(n,(function(t){return null!=$l(t).dataIndex}),!0);if(i){var r=i.selected?"unselect":"select",o=$l(i);e._api.dispatchAction({type:r,dataType:o.dataType,dataIndexInside:o.dataIndex,seriesIndex:o.seriesIndex,isFromClick:!0})}}))},Lw=function(t){t.clearColorPalette(),t.eachSeries((function(t){t.clearColorPalette()}))},kw=function(t,e,n,i){Tw(t,e,n,i),Ex(t._chartsViews,(function(t){t.__alive=!1})),Dw(t,e,n,i),Ex(t._chartsViews,(function(t){t.__alive||t.remove(e,n)}))},Tw=function(t,e,n,r,o){Ex(o||t._componentsViews,(function(t){var o=t.__model;a(o,t),t.render(o,e,n,r),i(o,t),s(o,t)}))},Dw=function(t,r,o,l,u){var c=t._scheduler,d=t._labelManager;d.clearLabels();var h=!1;r.eachSeries((function(e){var i=t._chartsMap[e.__viewId];i.__alive=!0;var r=i.renderTask;c.updatePayload(r,l),a(e,i),u&&u.get(e.uid)&&r.dirty(),r.perform(c.getPerformArgs(r))&&(h=!0),e.__transientTransitionOpt=null,i.group.silent=!!e.get("silent"),n(e,i),Vu(e),d.addLabelsOfSeries(i)})),c.unfinished=h||c.unfinished,d.updateLayoutConfig(o),d.layout(o),d.processLabelsOverall(),r.eachSeries((function(e){var n=t._chartsMap[e.__viewId];i(e,n),s(e,n)})),e(t,r)},Cw=function(t,e){Ex($w,(function(n){n(t,e)}))},Ow=function(t){t[ow]=!0,t.getZr().wakeUp()},Pw=function(e){e[ow]&&(e.getZr().storage.traverse((function(e){oh(e)||t(e)})),e[ow]=!1)},Iw=function(t){return new(function(e){function n(){return null!==e&&e.apply(this,arguments)||this}return o(n,e),n.prototype.getCoordinateSystems=function(){return t._coordSysMgr.getCoordinateSystems()},n.prototype.getComponentByElement=function(e){while(e){var n=e.__ecComponentInfo;if(null!=n)return t._model.getComponent(n.mainType,n.index);e=e.parent}},n.prototype.enterEmphasis=function(e,n){Cu(e,n),Ow(t)},n.prototype.leaveEmphasis=function(e,n){Iu(e,n),Ow(t)},n.prototype.enterBlur=function(e){Au(e),Ow(t)},n.prototype.leaveBlur=function(e){Yu(e),Ow(t)},n.prototype.enterSelect=function(e){Ou(e),Ow(t)},n.prototype.leaveSelect=function(e){Pu(e),Ow(t)},n.prototype.getModel=function(){return t.getModel()},n.prototype.getViewOfComponentModel=function(e){return t.getViewOfComponentModel(e)},n.prototype.getViewOfSeriesModel=function(e){return t.getViewOfSeriesModel(e)},n}(mm))(t)},Aw=function(t){function e(t,e){for(var n=0;n<t.length;n++){var i=t[n];i[sw]=e}}Ex(Bw,(function(n,i){t._messageCenter.on(i,(function(n){if(Jw[t.group]&&t[sw]!==lw){if(n&&n.escapeConnect)return;var i=t.makeActionFromEvent(n),r=[];Ex(Zw,(function(e){e!==t&&e.group===t.group&&r.push(e)})),e(r,lw),Ex(r,(function(t){t[sw]!==uw&&t.dispatchAction(i)})),e(r,cw)}}))}))},Yw=function(t,e){var n=t._model;R(so(e),(function(t){var e,i=t.from,r=t.to;null==r&&im(e);var o={includeMainTypes:["series"],enableAll:!1,enableNone:!1},a=i?Yo(n,i,o):null,s=Yo(n,r,o),l=s.seriesModel;null==l&&(e=""),a&&a.seriesModel!==l&&(e=""),null!=e&&im(e),l.__transientTransitionOpt={from:i?i.dimension:null,to:r.dimension,dividingMethod:t.dividingMethod}}))}}(),e}($t),Hw=Rw.prototype;Hw.on=dw("on"),Hw.off=dw("off"),Hw.one=function(t,e,n){var i=this;function r(){for(var n=[],o=0;o<arguments.length;o++)n[o]=arguments[o];e&&e.apply&&e.apply(this,n),i.off(t,r)}nm("ECharts#one is deprecated."),this.on.call(this,t,r,n)};var jw=["click","dblclick","mouseover","mouseout","mousemove","mousedown","mouseup","globalout","contextmenu"];function zw(t){0}var Ww={},Bw={},Fw=[],Vw=[],Gw=[],$w=[],Uw=[],qw={},Xw={},Zw={},Jw={},Kw=+new Date-0,Qw=(new Date,"_echarts_instance_");function tM(t,e,n){var i=eM(t);if(i)return i;var r=new Rw(t,e,n);return r.id="ec_"+Kw++,Zw[r.id]=r,Ro(t,Qw,r.id),Aw(r),Ex(Gw,(function(t){t(r)})),r}function eM(t){return Zw[Ho(t,Qw)]}function nM(t,e){qw[t]=e}function iM(t){Hx(Vw,t)<0&&Vw.push(t)}function rM(t,e){hM(Fw,t,e,Vx)}function oM(t){Hx(Gw,t)<0&&t&&Gw.push(t)}function aM(t){Hx($w,t)<0&&t&&$w.push(t)}function sM(t,e,n){"function"===typeof e&&(n=e,e="");var i=Rx(t)?t.type:[t,t={event:e}][0];t.event=(t.event||i).toLowerCase(),e=t.event,Bw[e]||(Px(aw.test(i)&&aw.test(e)),Ww[i]||(Ww[i]={action:n,actionInfo:t}),Bw[e]=i)}function lM(t,e){vm.register(t,e)}function uM(t,e){hM(Uw,t,e,$x,"layout")}function cM(t,e){hM(Uw,t,e,Xx,"visual")}var dM=[];function hM(t,e,n,i,r){if((Nx(e)||Rx(e))&&(n=e,e=i),!(Hx(dM,n)>=0)){dM.push(n);var o=Mv.wrapStageHandler(n,r);o.__prio=e,o.__raw=n,t.push(o)}}function fM(t,e){Xw[t]=e}function pM(t,e,n){Ox.registerMap(t,e,n)}var mM=Ug;cM(qx,nv),cM(Jx,rv),cM(Jx,ov),cM(qx,Ov),cM(Jx,Pv),cM(ew,Db),iM(Xm),rM(Bx,Zm),fM("default",sv),sM({type:iu,event:iu,update:iu},bt),sM({type:ru,event:ru,update:ru},bt),sM({type:ou,event:ou,update:ou},bt),sM({type:au,event:au,update:au},bt),sM({type:su,event:su,update:su},bt),nM("light",Lv),nM("dark",Av);var gM=[],yM={registerPreprocessor:iM,registerProcessor:rM,registerPostInit:oM,registerPostUpdate:aM,registerAction:sM,registerCoordinateSystem:lM,registerLayout:uM,registerVisual:cM,registerTransform:mM,registerLoading:fM,registerMap:pM,PRIORITY:nw,ComponentModel:xp,ComponentView:Py,SeriesModel:Yy,ChartView:Vy,registerComponentModel:function(t){xp.registerClass(t)},registerComponentView:function(t){Py.registerClass(t)},registerSeriesModel:function(t){Yy.registerClass(t)},registerChartView:function(t){Vy.registerClass(t)},registerSubTypeDefaulter:function(t,e){xp.registerSubTypeDefaulter(t,e)},registerPainter:function(t,e){Ar(t,e)}};function vM(t){$(t)?R(t,(function(t){vM(t)})):O(gM,t)>=0||(gM.push(t),U(t)&&(t={install:t}),t.install(yM))}function _M(){return!1}function bM(t,e,n){var i=Y(),r=e.getWidth(),o=e.getHeight(),a=i.style;return a&&(a.position="absolute",a.left="0",a.top="0",a.width=r+"px",a.height=o+"px",i.setAttribute("data-zr-dom-id",t)),i.width=r*n,i.height=o*n,i}var xM=function(t){function e(e,n,i){var r,o=t.call(this)||this;o.motionBlur=!1,o.lastFrameAlpha=.7,o.dpr=1,o.virtual=!1,o.config={},o.incremental=!1,o.zlevel=0,o.maxRepaintRectCount=5,o.__dirty=!0,o.__firstTimePaint=!0,o.__used=!1,o.__drawIndex=0,o.__startIndex=0,o.__endIndex=0,o.__prevStartIndex=null,o.__prevEndIndex=null,i=i||mi,"string"===typeof e?r=bM(e,n,i):J(e)&&(r=e,e=r.id),o.id=e,o.dom=r;var a=r.style;return a&&(r.onselectstart=_M,a.webkitUserSelect="none",a.userSelect="none",a.webkitTapHighlightColor="rgba(0,0,0,0)",a["-webkit-touch-callout"]="none",a.padding="0",a.margin="0",a.borderWidth="0"),o.domBack=null,o.ctxBack=null,o.painter=n,o.config=null,o.dpr=i,o}return wt(e,t),e.prototype.getElementCount=function(){return this.__endIndex-this.__startIndex},e.prototype.afterBrush=function(){this.__prevStartIndex=this.__startIndex,this.__prevEndIndex=this.__endIndex},e.prototype.initContext=function(){this.ctx=this.dom.getContext("2d"),this.ctx.dpr=this.dpr},e.prototype.setUnpainted=function(){this.__firstTimePaint=!0},e.prototype.createBackBuffer=function(){var t=this.dpr;this.domBack=bM("back-"+this.id,this.painter,t),this.ctxBack=this.domBack.getContext("2d"),1!==t&&this.ctxBack.scale(t,t)},e.prototype.createRepaintRects=function(t,e,n,i){if(this.__firstTimePaint)return this.__firstTimePaint=!1,null;var r,o=[],a=this.maxRepaintRectCount,s=!1,l=new Ji(0,0,0,0);function u(t){if(t.isFinite()&&!t.isZero())if(0===o.length){var e=new Ji(0,0,0,0);e.copy(t),o.push(e)}else{for(var n=!1,i=1/0,r=0,u=0;u<o.length;++u){var c=o[u];if(c.intersect(t)){var d=new Ji(0,0,0,0);d.copy(c),d.union(t),o[u]=d,n=!0;break}if(s){l.copy(t),l.union(c);var h=t.width*t.height,f=c.width*c.height,p=l.width*l.height,m=p-h-f;m<i&&(i=m,r=u)}}if(s&&(o[r].union(t),n=!0),!n){e=new Ji(0,0,0,0);e.copy(t),o.push(e)}s||(s=o.length>=a)}}for(var c=this.__startIndex;c<this.__endIndex;++c){var d=t[c];if(d){var h=d.shouldBePainted(n,i,!0,!0),f=d.__isRendered&&(d.__dirty&je||!h)?d.getPrevPaintRect():null;f&&u(f);var p=h&&(d.__dirty&je||!d.__isRendered)?d.getPaintRect():null;p&&u(p)}}for(c=this.__prevStartIndex;c<this.__prevEndIndex;++c){d=e[c],h=d.shouldBePainted(n,i,!0,!0);if(d&&(!h||!d.__zr)&&d.__isRendered){f=d.getPrevPaintRect();f&&u(f)}}do{r=!1;for(c=0;c<o.length;)if(o[c].isZero())o.splice(c,1);else{for(var m=c+1;m<o.length;)o[c].intersect(o[m])?(r=!0,o[c].union(o[m]),o.splice(m,1)):m++;c++}}while(r);return this._paintRects=o,o},e.prototype.debugGetPaintRects=function(){return(this._paintRects||[]).slice()},e.prototype.resize=function(t,e){var n=this.dpr,i=this.dom,r=i.style,o=this.domBack;r&&(r.width=t+"px",r.height=e+"px"),i.width=t*n,i.height=e*n,o&&(o.width=t*n,o.height=e*n,1!==n&&this.ctxBack.scale(n,n))},e.prototype.clear=function(t,e,n){var i=this.dom,r=this.ctx,o=i.width,a=i.height;e=e||this.clearColor;var s=this.motionBlur&&!t,l=this.lastFrameAlpha,u=this.dpr,c=this;s&&(this.domBack||this.createBackBuffer(),this.ctxBack.globalCompositeOperation="copy",this.ctxBack.drawImage(i,0,0,o/u,a/u));var d=this.domBack;function h(t,n,i,o){if(r.clearRect(t,n,i,o),e&&"transparent"!==e){var a=void 0;et(e)?(a=e.__canvasGradient||B_(r,e,{x:0,y:0,width:i,height:o}),e.__canvasGradient=a):nt(e)&&(a=Z_(r,e,{dirty:function(){c.setUnpainted(),c.__painter.refresh()}})),r.save(),r.fillStyle=a||e,r.fillRect(t,n,i,o),r.restore()}s&&(r.save(),r.globalAlpha=l,r.drawImage(d,t,n,i,o),r.restore())}!n||s?h(0,0,o,a):n.length&&R(n,(function(t){h(t.x*u,t.y*u,t.width*u,t.height*u)}))},e}($t),wM=xM,MM=1e5,SM=314159,LM=.01,kM=.001;function TM(t){return parseInt(t,10)}function DM(t){return!!t&&(!!t.__builtin__||"function"===typeof t.resize&&"function"===typeof t.refresh)}function CM(t,e){var n=document.createElement("div");return n.style.cssText=["position:relative","width:"+t+"px","height:"+e+"px","padding:0","margin:0","border-width:0"].join(";")+";",n}var IM=function(){function t(t,e,n,i){this.type="canvas",this._zlevelList=[],this._prevDisplayList=[],this._layers={},this._layerConfig={},this._needsManuallyCompositing=!1,this.type="canvas";var r=!t.nodeName||"CANVAS"===t.nodeName.toUpperCase();this._opts=n=I({},n||{}),this.dpr=n.devicePixelRatio||mi,this._singleCanvas=r,this.root=t;var o=t.style;o&&(o.webkitTapHighlightColor="transparent",o.webkitUserSelect="none",o.userSelect="none",o["-webkit-touch-callout"]="none",t.innerHTML=""),this.storage=e;var a=this._zlevelList;this._prevDisplayList=[];var s=this._layers;if(r){var l=t,u=l.width,c=l.height;null!=n.width&&(u=n.width),null!=n.height&&(c=n.height),this.dpr=n.devicePixelRatio||1,l.width=u*this.dpr,l.height=c*this.dpr,this._width=u,this._height=c;var d=new wM(l,this,this.dpr);d.__builtin__=!0,d.initContext(),s[SM]=d,d.zlevel=SM,a.push(SM),this._domRoot=t}else{this._width=this._getSize(0),this._height=this._getSize(1);var h=this._domRoot=CM(this._width,this._height);t.appendChild(h)}}return t.prototype.getType=function(){return"canvas"},t.prototype.isSingleCanvas=function(){return this._singleCanvas},t.prototype.getViewportRoot=function(){return this._domRoot},t.prototype.getViewportRootOffset=function(){var t=this.getViewportRoot();if(t)return{offsetLeft:t.offsetLeft||0,offsetTop:t.offsetTop||0}},t.prototype.refresh=function(t){var e=this.storage.getDisplayList(!0),n=this._prevDisplayList,i=this._zlevelList;this._redrawId=Math.random(),this._paintList(e,n,t,this._redrawId);for(var r=0;r<i.length;r++){var o=i[r],a=this._layers[o];if(!a.__builtin__&&a.refresh){var s=0===r?this._backgroundColor:null;a.refresh(s)}}return this._opts.useDirtyRect&&(this._prevDisplayList=e.slice()),this},t.prototype.refreshHover=function(){this._paintHoverList(this.storage.getDisplayList(!1))},t.prototype._paintHoverList=function(t){var e=t.length,n=this._hoverlayer;if(n&&n.clear(),e){for(var i,r={inHover:!0,viewWidth:this._width,viewHeight:this._height},o=0;o<e;o++){var a=t[o];a.__inHover&&(n||(n=this._hoverlayer=this.getLayer(MM)),i||(i=n.ctx,i.save()),gb(i,a,r,o===e-1))}i&&i.restore()}},t.prototype.getHoverLayer=function(){return this.getLayer(MM)},t.prototype.paintOne=function(t,e){mb(t,e)},t.prototype._paintList=function(t,e,n,i){if(this._redrawId===i){n=n||!1,this._updateLayerStatus(t);var r=this._doPaintList(t,e,n),o=r.finished,a=r.needsRefreshHover;if(this._needsManuallyCompositing&&this._compositeManually(),a&&this._paintHoverList(t),o)this.eachLayer((function(t){t.afterBrush&&t.afterBrush()}));else{var s=this;qe((function(){s._paintList(t,e,n,i)}))}}},t.prototype._compositeManually=function(){var t=this.getLayer(SM).ctx,e=this._domRoot.width,n=this._domRoot.height;t.clearRect(0,0,e,n),this.eachBuiltinLayer((function(i){i.virtual&&t.drawImage(i.dom,0,0,e,n)}))},t.prototype._doPaintList=function(t,e,n){for(var i=this,r=[],o=this._opts.useDirtyRect,a=0;a<this._zlevelList.length;a++){var s=this._zlevelList[a],l=this._layers[s];l.__builtin__&&l!==this._hoverlayer&&(l.__dirty||n)&&r.push(l)}for(var u=!0,c=!1,d=function(a){var s,l=r[a],d=l.ctx,h=o&&l.createRepaintRects(t,e,f._width,f._height),p=n?l.__startIndex:l.__drawIndex,m=!n&&l.incremental&&Date.now,g=m&&Date.now(),y=l.zlevel===f._zlevelList[0]?f._backgroundColor:null;if(l.__startIndex===l.__endIndex)l.clear(!1,y,h);else if(p===l.__startIndex){var v=t[p];v.incremental&&v.notClear&&!n||l.clear(!1,y,h)}-1===p&&(p=l.__startIndex);var _=function(e){var n={inHover:!1,allClipped:!1,prevEl:null,viewWidth:i._width,viewHeight:i._height};for(s=p;s<l.__endIndex;s++){var r=t[s];if(r.__inHover&&(c=!0),i._doPaintEl(r,l,o,e,n,s===l.__endIndex-1),m){var a=Date.now()-g;if(a>15)break}}n.prevElClipPaths&&d.restore()};if(h)if(0===h.length)s=l.__endIndex;else for(var b=f.dpr,x=0;x<h.length;++x){var w=h[x];d.save(),d.beginPath(),d.rect(w.x*b,w.y*b,w.width*b,w.height*b),d.clip(),_(w),d.restore()}else d.save(),_(),d.restore();l.__drawIndex=s,l.__drawIndex<l.__endIndex&&(u=!1)},f=this,p=0;p<r.length;p++)d(p);return h.wxa&&R(this._layers,(function(t){t&&t.ctx&&t.ctx.draw&&t.ctx.draw()})),{finished:u,needsRefreshHover:c}},t.prototype._doPaintEl=function(t,e,n,i,r,o){var a=e.ctx;if(n){var s=t.getPaintRect();(!i||s&&s.intersect(i))&&(gb(a,t,r,o),t.setPrevPaintRect(s))}else gb(a,t,r,o)},t.prototype.getLayer=function(t,e){this._singleCanvas&&!this._needsManuallyCompositing&&(t=SM);var n=this._layers[t];return n||(n=new wM("zr_"+t,this,this.dpr),n.zlevel=t,n.__builtin__=!0,this._layerConfig[t]?D(n,this._layerConfig[t],!0):this._layerConfig[t-LM]&&D(n,this._layerConfig[t-LM],!0),e&&(n.virtual=e),this.insertLayer(t,n),n.initContext()),n},t.prototype.insertLayer=function(t,e){var n=this._layers,i=this._zlevelList,r=i.length,o=this._domRoot,a=null,s=-1;if(n[t])k("ZLevel "+t+" has been used already");else if(DM(e)){if(r>0&&t>i[0]){for(s=0;s<r-1;s++)if(i[s]<t&&i[s+1]>t)break;a=n[i[s]]}if(i.splice(s+1,0,t),n[t]=e,!e.virtual)if(a){var l=a.dom;l.nextSibling?o.insertBefore(e.dom,l.nextSibling):o.appendChild(e.dom)}else o.firstChild?o.insertBefore(e.dom,o.firstChild):o.appendChild(e.dom);e.__painter=this}else k("Layer of zlevel "+t+" is not valid")},t.prototype.eachLayer=function(t,e){for(var n=this._zlevelList,i=0;i<n.length;i++){var r=n[i];t.call(e,this._layers[r],r)}},t.prototype.eachBuiltinLayer=function(t,e){for(var n=this._zlevelList,i=0;i<n.length;i++){var r=n[i],o=this._layers[r];o.__builtin__&&t.call(e,o,r)}},t.prototype.eachOtherLayer=function(t,e){for(var n=this._zlevelList,i=0;i<n.length;i++){var r=n[i],o=this._layers[r];o.__builtin__||t.call(e,o,r)}},t.prototype.getLayers=function(){return this._layers},t.prototype._updateLayerStatus=function(t){function e(t){a&&(a.__endIndex!==t&&(a.__dirty=!0),a.__endIndex=t)}if(this.eachBuiltinLayer((function(t,e){t.__dirty=t.__used=!1})),this._singleCanvas)for(var n=1;n<t.length;n++){var i=t[n];if(i.zlevel!==t[n-1].zlevel||i.incremental){this._needsManuallyCompositing=!0;break}}var r,o,a=null,s=0;for(o=0;o<t.length;o++){i=t[o];var l=i.zlevel,u=void 0;r!==l&&(r=l,s=0),i.incremental?(u=this.getLayer(l+kM,this._needsManuallyCompositing),u.incremental=!0,s=1):u=this.getLayer(l+(s>0?LM:0),this._needsManuallyCompositing),u.__builtin__||k("ZLevel "+l+" has been used by unkown layer "+u.id),u!==a&&(u.__used=!0,u.__startIndex!==o&&(u.__dirty=!0),u.__startIndex=o,u.incremental?u.__drawIndex=-1:u.__drawIndex=o,e(o),a=u),i.__dirty&je&&!i.__inHover&&(u.__dirty=!0,u.incremental&&u.__drawIndex<0&&(u.__drawIndex=o))}e(o),this.eachBuiltinLayer((function(t,e){!t.__used&&t.getElementCount()>0&&(t.__dirty=!0,t.__startIndex=t.__endIndex=t.__drawIndex=0),t.__dirty&&t.__drawIndex<0&&(t.__drawIndex=t.__startIndex)}))},t.prototype.clear=function(){return this.eachBuiltinLayer(this._clearLayer),this},t.prototype._clearLayer=function(t){t.clear()},t.prototype.setBackgroundColor=function(t){this._backgroundColor=t,R(this._layers,(function(t){t.setUnpainted()}))},t.prototype.configLayer=function(t,e){if(e){var n=this._layerConfig;n[t]?D(n[t],e,!0):n[t]=e;for(var i=0;i<this._zlevelList.length;i++){var r=this._zlevelList[i];if(r===t||r===t+LM){var o=this._layers[r];D(o,n[t],!0)}}}},t.prototype.delLayer=function(t){var e=this._layers,n=this._zlevelList,i=e[t];i&&(i.dom.parentNode.removeChild(i.dom),delete e[t],n.splice(O(n,t),1))},t.prototype.resize=function(t,e){if(this._domRoot.style){var n=this._domRoot;n.style.display="none";var i=this._opts;if(null!=t&&(i.width=t),null!=e&&(i.height=e),t=this._getSize(0),e=this._getSize(1),n.style.display="",this._width!==t||e!==this._height){for(var r in n.style.width=t+"px",n.style.height=e+"px",this._layers)this._layers.hasOwnProperty(r)&&this._layers[r].resize(t,e);this.refresh(!0)}this._width=t,this._height=e}else{if(null==t||null==e)return;this._width=t,this._height=e,this.getLayer(SM).resize(t,e)}return this},t.prototype.clearLayer=function(t){var e=this._layers[t];e&&e.clear()},t.prototype.dispose=function(){this.root.innerHTML="",this.root=this.storage=this._domRoot=this._layers=null},t.prototype.getRenderedCanvas=function(t){if(t=t||{},this._singleCanvas&&!this._compositeManually)return this._layers[SM].dom;var e=new wM("image",this,t.pixelRatio||this.dpr);e.initContext(),e.clear(!1,t.backgroundColor||this._backgroundColor);var n=e.ctx;if(t.pixelRatio<=this.dpr){this.refresh();var i=e.dom.width,r=e.dom.height;this.eachLayer((function(t){t.__builtin__?n.drawImage(t.dom,0,0,i,r):t.renderToCanvas&&(n.save(),t.renderToCanvas(n),n.restore())}))}else for(var o={inHover:!1,viewWidth:this._width,viewHeight:this._height},a=this.storage.getDisplayList(!0),s=0,l=a.length;s<l;s++){var u=a[s];gb(n,u,o,s===l-1)}return e.dom},t.prototype.getWidth=function(){return this._width},t.prototype.getHeight=function(){return this._height},t.prototype._getSize=function(t){var e=this._opts,n=["width","height"][t],i=["clientWidth","clientHeight"][t],r=["paddingLeft","paddingTop"][t],o=["paddingRight","paddingBottom"][t];if(null!=e[n]&&"auto"!==e[n])return parseFloat(e[n]);var a=this.root,s=document.defaultView.getComputedStyle(a);return(a[i]||TM(s[n])||TM(a.style[n]))-(TM(s[r])||0)-(TM(s[o])||0)|0},t.prototype.pathToImage=function(t,e){e=e||this.dpr;var n=document.createElement("canvas"),i=n.getContext("2d"),r=t.getBoundingRect(),o=t.style,a=o.shadowBlur*e,s=o.shadowOffsetX*e,l=o.shadowOffsetY*e,u=t.hasStroke()?o.lineWidth:0,c=Math.max(u/2,-s+a),d=Math.max(u/2,s+a),h=Math.max(u/2,-l+a),f=Math.max(u/2,l+a),p=r.width+c+d,m=r.height+h+f;n.width=p*e,n.height=m*e,i.scale(e,e),i.clearRect(0,0,p,m),i.dpr=e;var g={x:t.x,y:t.y,scaleX:t.scaleX,scaleY:t.scaleY,rotation:t.rotation,originX:t.originX,originY:t.originY};t.x=c-r.x,t.y=h-r.y,t.rotation=0,t.scaleX=1,t.scaleY=1,t.updateTransform(),t&&gb(i,t,{inHover:!1,viewWidth:this._width,viewHeight:this._height},!0);var y=new wl({style:{x:0,y:0,image:n}});return I(t,g),y},t}(),AM=IM;function YM(t){t.registerPainter("canvas",AM)}function OM(t){return document.createElementNS("http://www.w3.org/2000/svg",t)}function PM(t,e,n){n||(n=function(t,e){return t===e}),t=t.slice(),e=e.slice();var i=e.length,r=t.length,o=1,a=i+r,s=[{newPos:-1,components:[]}],l=EM(s[0],e,t,0,n);if(s[0].newPos+1>=i&&l+1>=r){for(var u=[],c=0;c<e.length;c++)u.push(c);return[{indices:u,count:e.length,added:!1,removed:!1}]}function d(){for(var a=-1*o;a<=o;a+=2){var l,u=s[a-1],c=s[a+1],d=(c?c.newPos:0)-a;u&&(s[a-1]=void 0);var h=u&&u.newPos+1<i,f=c&&0<=d&&d<r;if(h||f){if(!h||f&&u.newPos<c.newPos?(l=HM(c),NM(l.components,!1,!0)):(l=u,l.newPos++,NM(l.components,!0,!1)),d=EM(l,e,t,a,n),l.newPos+1>=i&&d+1>=r)return RM(l.components);s[a]=l}else s[a]=void 0}o++}while(o<=a){var h=d();if(h)return h}}function EM(t,e,n,i,r){var o=e.length,a=n.length,s=t.newPos,l=s-i,u=0;while(s+1<o&&l+1<a&&r(e[s+1],n[l+1]))s++,l++,u++;return u&&t.components.push({count:u,added:!1,removed:!1,indices:[]}),t.newPos=s,l}function NM(t,e,n){var i=t[t.length-1];i&&i.added===e&&i.removed===n?t[t.length-1]={count:i.count+1,added:e,removed:n,indices:[]}:t.push({count:1,added:e,removed:n,indices:[]})}function RM(t){for(var e=0,n=t.length,i=0,r=0;e<n;e++){var o=t[e];if(o.removed){for(s=r;s<r+o.count;s++)o.indices.push(s);r+=o.count}else{for(var a=[],s=i;s<i+o.count;s++)a.push(s);o.indices=a,i+=o.count,o.added||(r+=o.count)}}return t}function HM(t){return{newPos:t.newPos,components:t.components.slice(0)}}function jM(t,e,n){return PM(t,e,n)}var zM="none",WM=Math.round,BM=Math.sin,FM=Math.cos,VM=Math.PI,GM=2*Math.PI,$M=180/VM,UM=1e-4;function qM(t){return WM(1e3*t)/1e3}function XM(t){return WM(1e4*t)/1e4}function ZM(t){return t<UM&&t>-UM}function JM(t){var e=t.fill;return null!=e&&e!==zM}function KM(t){var e=t.stroke;return null!=e&&e!==zM}function QM(t,e){e&&tS(t,"transform","matrix("+qM(e[0])+","+qM(e[1])+","+qM(e[2])+","+qM(e[3])+","+XM(e[4])+","+XM(e[5])+")")}function tS(t,e,n){(!n||"linear"!==n.type&&"radial"!==n.type)&&t.setAttribute(e,n)}function eS(t,e,n){t.setAttributeNS("http://www.w3.org/1999/xlink",e,n)}function nS(t,e,n){t.setAttributeNS("http://www.w3.org/XML/1998/namespace",e,n)}function iS(t,e,n){var i=null==e.opacity?1:e.opacity;if(n instanceof wl)t.style.opacity=i+"";else{if(JM(e)){var r=e.fill;r="transparent"===r?zM:r,tS(t,"fill",r),tS(t,"fill-opacity",(null!=e.fillOpacity?e.fillOpacity*i:i)+"")}else tS(t,"fill",zM);if(KM(e)){var o=e.stroke;o="transparent"===o?zM:o,tS(t,"stroke",o);var a=e.lineWidth,s=e.strokeNoScale?n.getLineScale():1;tS(t,"stroke-width",(s?a/s:0)+""),tS(t,"paint-order",e.strokeFirst?"stroke":"fill"),tS(t,"stroke-opacity",(null!=e.strokeOpacity?e.strokeOpacity*i:i)+"");var l=e.lineDash&&a>0&&V_(e.lineDash,a);if(l){var u=e.lineDashOffset;s&&1!==s&&(l=H(l,(function(t){return t/s})),u&&(u/=s,u=WM(u))),tS(t,"stroke-dasharray",l.join(",")),tS(t,"stroke-dashoffset",(u||0)+"")}else tS(t,"stroke-dasharray","");e.lineCap&&tS(t,"stroke-linecap",e.lineCap),e.lineJoin&&tS(t,"stroke-linejoin",e.lineJoin),e.miterLimit&&tS(t,"stroke-miterlimit",e.miterLimit+"")}else tS(t,"stroke",zM)}}var rS=function(){function t(){}return t.prototype.reset=function(){this._d=[],this._str=""},t.prototype.moveTo=function(t,e){this._add("M",t,e)},t.prototype.lineTo=function(t,e){this._add("L",t,e)},t.prototype.bezierCurveTo=function(t,e,n,i,r,o){this._add("C",t,e,n,i,r,o)},t.prototype.quadraticCurveTo=function(t,e,n,i){this._add("Q",t,e,n,i)},t.prototype.arc=function(t,e,n,i,r,o){this.ellipse(t,e,n,n,0,i,r,o)},t.prototype.ellipse=function(t,e,n,i,r,o,a,s){var l=0===this._d.length,u=a-o,c=!s,d=Math.abs(u),h=ZM(d-GM)||(c?u>=GM:-u>=GM),f=u>0?u%GM:u%GM+GM,p=!1;p=!!h||!ZM(d)&&f>=VM===!!c;var m=XM(t+n*FM(o)),g=XM(e+i*BM(o));h&&(u=c?GM-1e-4:1e-4-GM,p=!0,l&&this._d.push("M",m,g));var y=XM(t+n*FM(o+u)),v=XM(e+i*BM(o+u));if(isNaN(m)||isNaN(g)||isNaN(n)||isNaN(i)||isNaN(r)||isNaN($M)||isNaN(y)||isNaN(v))return"";this._d.push("A",XM(n),XM(i),WM(r*$M),+p,+c,y,v)},t.prototype.rect=function(t,e,n,i){this._add("M",t,e),this._add("L",t+n,e),this._add("L",t+n,e+i),this._add("L",t,e+i),this._add("L",t,e)},t.prototype.closePath=function(){this._d.length>0&&this._add("Z")},t.prototype._add=function(t,e,n,i,r,o,a,s,l){this._d.push(t);for(var u=1;u<arguments.length;u++){var c=arguments[u];if(isNaN(c))return void(this._invalid=!0);this._d.push(XM(c))}},t.prototype.generateStr=function(){this._str=this._invalid?"":this._d.join(" "),this._d=[]},t.prototype.getStr=function(){return this._str},t}(),oS={brush:function(t){var e=t.style,n=t.__svgEl;n||(n=OM("path"),t.__svgEl=n),t.path||t.createPathProxy();var i=t.path;t.shapeChanged()&&(i.beginPath(),t.buildPath(i,t.shape),t.pathUpdated());var r=i.getVersion(),o=t,a=o.__svgPathBuilder;(o.__svgPathVersion!==r||!a||t.style.strokePercent<1)&&(a||(a=o.__svgPathBuilder=new rS),a.reset(),i.rebuildPath(a,t.style.strokePercent),a.generateStr(),o.__svgPathVersion=r),tS(n,"d",a.getStr()),iS(n,e,t),QM(n,t.transform)}},aS={brush:function(t){var e=t.style,n=e.image;if(n instanceof HTMLImageElement?n=n.src:n instanceof HTMLCanvasElement&&(n=n.toDataURL()),n){var i=e.x||0,r=e.y||0,o=e.width,a=e.height,s=t.__svgEl;s||(s=OM("image"),t.__svgEl=s),n!==t.__imageSrc&&(eS(s,"href",n),t.__imageSrc=n),tS(s,"width",o+""),tS(s,"height",a+""),tS(s,"x",i+""),tS(s,"y",r+""),iS(s,e,t),QM(s,t.transform)}}},sS={left:"start",right:"end",center:"middle",middle:"middle"};function lS(t,e,n){return"top"===n?t+=e/2:"bottom"===n&&(t-=e/2),t}var uS={brush:function(t){var e=t.style,n=e.text;if(null!=n&&(n+=""),n&&!isNaN(e.x)&&!isNaN(e.y)){var i=t.__svgEl;i||(i=OM("text"),nS(i,"xml:space","preserve"),t.__svgEl=i);var r=e.font||Qi,o=i.style;o.font=r,i.textContent=n,iS(i,e,t),QM(i,t.transform);var a=e.x||0,s=lS(e.y||0,sr(r),e.textBaseline),l=sS[e.textAlign]||e.textAlign;tS(i,"dominant-baseline","central"),tS(i,"text-anchor",l),tS(i,"x",a+""),tS(i,"y",s+"")}}},cS="0",dS="1",hS=function(){function t(t,e,n,i,r){this.nextId=0,this._domName="_dom",this.createElement=OM,this._zrId=t,this._svgRoot=e,this._tagNames="string"===typeof n?[n]:n,this._markLabel=i,r&&(this._domName=r)}return t.prototype.getDefs=function(t){var e=this._svgRoot,n=this._svgRoot.getElementsByTagName("defs");if(0===n.length){if(t){var i=e.insertBefore(this.createElement("defs"),e.firstChild);return i.contains||(i.contains=function(t){var e=i.children;if(!e)return!1;for(var n=e.length-1;n>=0;--n)if(e[n]===t)return!0;return!1}),i}return null}return n[0]},t.prototype.doUpdate=function(t,e){if(t){var n=this.getDefs(!1);if(t[this._domName]&&n.contains(t[this._domName]))"function"===typeof e&&e(t);else{var i=this.add(t);i&&(t[this._domName]=i)}}},t.prototype.add=function(t){return null},t.prototype.addDom=function(t){var e=this.getDefs(!0);t.parentNode!==e&&e.appendChild(t)},t.prototype.removeDom=function(t){var e=this.getDefs(!1);e&&t[this._domName]&&(e.removeChild(t[this._domName]),t[this._domName]=null)},t.prototype.getDoms=function(){var t=this.getDefs(!1);if(!t)return[];var e=[];return R(this._tagNames,(function(n){for(var i=t.getElementsByTagName(n),r=0;r<i.length;r++)e.push(i[r])})),e},t.prototype.markAllUnused=function(){var t=this.getDoms(),e=this;R(t,(function(t){t[e._markLabel]=cS}))},t.prototype.markDomUsed=function(t){t&&(t[this._markLabel]=dS)},t.prototype.markDomUnused=function(t){t&&(t[this._markLabel]=cS)},t.prototype.isDomUnused=function(t){return t&&t[this._markLabel]!==dS},t.prototype.removeUnused=function(){var t=this,e=this.getDefs(!1);if(e){var n=this.getDoms();R(n,(function(n){t.isDomUnused(n)&&e.removeChild(n)}))}},t.prototype.getSvgProxy=function(t){return t instanceof pl?oS:t instanceof wl?aS:t instanceof yl?uS:oS},t.prototype.getSvgElement=function(t){return t.__svgEl},t}(),fS=hS;function pS(t){return"linear"===t.type}function mS(t){return"radial"===t.type}function gS(t){return t&&("linear"===t.type||"radial"===t.type)}var yS=function(t){function e(e,n){return t.call(this,e,n,["linearGradient","radialGradient"],"__gradient_in_use__")||this}return wt(e,t),e.prototype.addWithoutUpdate=function(t,e){if(e&&e.style){var n=this;R(["fill","stroke"],(function(i){var r=e.style[i];if(gS(r)){var o=r,a=n.getDefs(!0),s=void 0;o.__dom?(s=o.__dom,a.contains(o.__dom)||n.addDom(s)):s=n.add(o),n.markUsed(e);var l=s.getAttribute("id");t.setAttribute(i,"url(#"+l+")")}}))}},e.prototype.add=function(t){var e;if(pS(t))e=this.createElement("linearGradient");else{if(!mS(t))return k("Illegal gradient type."),null;e=this.createElement("radialGradient")}return t.id=t.id||this.nextId++,e.setAttribute("id","zr"+this._zrId+"-gradient-"+t.id),this.updateDom(t,e),this.addDom(e),e},e.prototype.update=function(t){if(gS(t)){var e=this;this.doUpdate(t,(function(){var n=t.__dom;if(n){var i=n.tagName,r=t.type;"linear"===r&&"linearGradient"===i||"radial"===r&&"radialGradient"===i?e.updateDom(t,t.__dom):(e.removeDom(t),e.add(t))}}))}},e.prototype.updateDom=function(t,e){if(pS(t))e.setAttribute("x1",t.x+""),e.setAttribute("y1",t.y+""),e.setAttribute("x2",t.x2+""),e.setAttribute("y2",t.y2+"");else{if(!mS(t))return void k("Illegal gradient type.");e.setAttribute("cx",t.x+""),e.setAttribute("cy",t.y+""),e.setAttribute("r",t.r+"")}t.global?e.setAttribute("gradientUnits","userSpaceOnUse"):e.setAttribute("gradientUnits","objectBoundingBox"),e.innerHTML="";for(var n=t.colorStops,i=0,r=n.length;i<r;++i){var o=this.createElement("stop");o.setAttribute("offset",100*n[i].offset+"%");var a=n[i].color;if(a.indexOf("rgba")>-1){var s=yn(a)[3],l=xn(a);o.setAttribute("stop-color","#"+l),o.setAttribute("stop-opacity",s+"")}else o.setAttribute("stop-color",n[i].color);e.appendChild(o)}t.__dom=e},e.prototype.markUsed=function(e){if(e.style){var n=e.style.fill;n&&n.__dom&&t.prototype.markDomUsed.call(this,n.__dom),n=e.style.stroke,n&&n.__dom&&t.prototype.markDomUsed.call(this,n.__dom)}},e}(fS),vS=yS;function _S(t){return t&&(!!t.image||!!t.svgElement)}var bS=new C_,xS=function(t){function e(e,n){return t.call(this,e,n,["pattern"],"__pattern_in_use__")||this}return wt(e,t),e.prototype.addWithoutUpdate=function(t,e){if(e&&e.style){var n=this;R(["fill","stroke"],(function(i){var r=e.style[i];if(_S(r)){var o=n.getDefs(!0),a=bS.get(r);a?o.contains(a)||n.addDom(a):a=n.add(r),n.markUsed(e);var s=a.getAttribute("id");t.setAttribute(i,"url(#"+s+")")}}))}},e.prototype.add=function(t){if(_S(t)){var e=this.createElement("pattern");return t.id=null==t.id?this.nextId++:t.id,e.setAttribute("id","zr"+this._zrId+"-pattern-"+t.id),e.setAttribute("x","0"),e.setAttribute("y","0"),e.setAttribute("patternUnits","userSpaceOnUse"),this.updateDom(t,e),this.addDom(e),e}},e.prototype.update=function(t){if(_S(t)){var e=this;this.doUpdate(t,(function(){var n=bS.get(t);e.updateDom(t,n)}))}},e.prototype.updateDom=function(t,e){var n=t.svgElement;if(n instanceof SVGElement)n.parentNode!==e&&(e.innerHTML="",e.appendChild(n),e.setAttribute("width",t.svgWidth+""),e.setAttribute("height",t.svgHeight+""));else{var i=void 0,r=e.getElementsByTagName("image");if(r.length){if(!t.image)return void e.removeChild(r[0]);i=r[0]}else t.image&&(i=this.createElement("image"));if(i){var o=void 0,a=t.image;if("string"===typeof a?o=a:a instanceof HTMLImageElement?o=a.src:a instanceof HTMLCanvasElement&&(o=a.toDataURL()),o){i.setAttribute("href",o),i.setAttribute("x","0"),i.setAttribute("y","0");var s={dirty:function(){}},l=la(o,i,s,(function(t){e.setAttribute("width",t.width+""),e.setAttribute("height",t.height+"")}));l&&l.width&&l.height&&(e.setAttribute("width",l.width+""),e.setAttribute("height",l.height+"")),e.appendChild(i)}}}var u=t.x||0,c=t.y||0,d=(t.rotation||0)/Math.PI*180,h=t.scaleX||1,f=t.scaleY||1,p="translate("+u+", "+c+") rotate("+d+") scale("+h+", "+f+")";e.setAttribute("patternTransform",p),bS.set(t,e)},e.prototype.markUsed=function(e){e.style&&(_S(e.style.fill)&&t.prototype.markDomUsed.call(this,bS.get(e.style.fill)),_S(e.style.stroke)&&t.prototype.markDomUsed.call(this,bS.get(e.style.stroke)))},e}(fS),wS=xS;function MS(t){var e=[];if(t)for(var n=0;n<t.length;n++){var i=t[n];e.push(i.id)}return e.join(",")}function SS(t){var e=t.__clipPaths;return e&&e.length>0}var LS=function(t){function e(e,n){var i=t.call(this,e,n,"clipPath","__clippath_in_use__")||this;return i._refGroups={},i._keyDuplicateCount={},i}return wt(e,t),e.prototype.markAllUnused=function(){for(var e in t.prototype.markAllUnused.call(this),this._refGroups)this.markDomUnused(this._refGroups[e]);this._keyDuplicateCount={}},e.prototype._getClipPathGroup=function(t,e){if(SS(t)){var n=t.__clipPaths,i=this._keyDuplicateCount,r=MS(n);return F_(n,e&&e.__clipPaths)&&(i[r]=i[r]||0,i[r]&&(r+="-"+i[r]),i[r]++),this._refGroups[r]||(this._refGroups[r]=this.createElement("g"))}},e.prototype.update=function(t,e){var n=this._getClipPathGroup(t,e);return n&&(this.markDomUsed(n),this.updateDom(n,t.__clipPaths)),n},e.prototype.updateDom=function(t,e){if(e&&e.length>0){var n=this.getDefs(!0),i=e[0],r=void 0,o=void 0;i._dom?(o=i._dom.getAttribute("id"),r=i._dom,n.contains(r)||n.appendChild(r)):(o="zr"+this._zrId+"-clip-"+this.nextId,++this.nextId,r=this.createElement("clipPath"),r.setAttribute("id",o),n.appendChild(r),i._dom=r);var a=this.getSvgProxy(i);a.brush(i);var s=this.getSvgElement(i);r.innerHTML="",r.appendChild(s),t.setAttribute("clip-path","url(#"+o+")"),e.length>1&&this.updateDom(r,e.slice(1))}else t&&t.setAttribute("clip-path","none")},e.prototype.markUsed=function(e){var n=this;e.__clipPaths&&R(e.__clipPaths,(function(e){e._dom&&t.prototype.markDomUsed.call(n,e._dom)}))},e.prototype.removeUnused=function(){t.prototype.removeUnused.call(this);var e={};for(var n in this._refGroups){var i=this._refGroups[n];this.isDomUnused(i)?i.parentNode&&i.parentNode.removeChild(i):e[n]=i}this._refGroups=e},e}(fS),kS=LS,TS=function(t){function e(e,n){var i=t.call(this,e,n,["filter"],"__filter_in_use__","_shadowDom")||this;return i._shadowDomMap={},i._shadowDomPool=[],i}return wt(e,t),e.prototype._getFromPool=function(){var t=this._shadowDomPool.pop();if(!t){t=this.createElement("filter"),t.setAttribute("id","zr"+this._zrId+"-shadow-"+this.nextId++);var e=this.createElement("feDropShadow");t.appendChild(e),this.addDom(t)}return t},e.prototype.update=function(t,e){var n=e.style;if(CS(n)){var i=IS(e),r=e._shadowDom=this._shadowDomMap[i];r||(r=this._getFromPool(),this._shadowDomMap[i]=r),this.updateDom(t,e,r)}else this.remove(t,e)},e.prototype.remove=function(t,e){null!=e._shadowDom&&(e._shadowDom=null,t.style.filter="")},e.prototype.updateDom=function(t,e,n){var i=n.children[0],r=e.style,o=e.getGlobalScale(),a=o[0],s=o[1];if(a&&s){var l=r.shadowOffsetX||0,u=r.shadowOffsetY||0,c=r.shadowBlur,d=r.shadowColor;i.setAttribute("dx",l/a+""),i.setAttribute("dy",u/s+""),i.setAttribute("flood-color",d);var h=c/2/a,f=c/2/s,p=h+" "+f;i.setAttribute("stdDeviation",p),n.setAttribute("x","-100%"),n.setAttribute("y","-100%"),n.setAttribute("width","300%"),n.setAttribute("height","300%"),e._shadowDom=n;var m=n.getAttribute("id");t.style.filter="url(#"+m+")"}},e.prototype.removeUnused=function(){var t=this.getDefs(!1);if(t){var e=this._shadowDomPool;for(var n in this._shadowDomMap){var i=this._shadowDomMap[n];e.push(i)}this._shadowDomMap={}}},e}(fS),DS=TS;function CS(t){return t&&(t.shadowBlur||t.shadowOffsetX||t.shadowOffsetY)}function IS(t){var e=t.style,n=t.getGlobalScale();return[e.shadowColor,(e.shadowBlur||0).toFixed(2),(e.shadowOffsetX||0).toFixed(2),(e.shadowOffsetY||0).toFixed(2),n[0],n[1]].join(",")}function AS(t){return parseInt(t,10)}function YS(t){return t instanceof pl?oS:t instanceof wl?aS:t instanceof yl?uS:oS}function OS(t,e){return e&&t&&e.parentNode!==t}function PS(t,e,n){if(OS(t,e)&&n){var i=n.nextSibling;i?t.insertBefore(e,i):t.appendChild(e)}}function ES(t,e){if(OS(t,e)){var n=t.firstChild;n?t.insertBefore(e,n):t.appendChild(e)}}function NS(t,e){e&&t&&e.parentNode===t&&t.removeChild(e)}function RS(t){t&&t.parentNode&&t.parentNode.removeChild(t)}function HS(t){return t.__svgEl}var jS=function(){function t(t,e,n,i){this.type="svg",this.refreshHover=zS("refreshHover"),this.pathToImage=zS("pathToImage"),this.configLayer=zS("configLayer"),this.root=t,this.storage=e,this._opts=n=I({},n||{});var r=OM("svg");r.setAttributeNS("http://www.w3.org/2000/xmlns/","xmlns","http://www.w3.org/2000/svg"),r.setAttributeNS("http://www.w3.org/2000/xmlns/","xmlns:xlink","http://www.w3.org/1999/xlink"),r.setAttribute("version","1.1"),r.setAttribute("baseProfile","full"),r.style.cssText="user-select:none;position:absolute;left:0;top:0;";var o=OM("g");r.appendChild(o);var a=OM("g");r.appendChild(a),this._gradientManager=new vS(i,a),this._patternManager=new wS(i,a),this._clipPathManager=new kS(i,a),this._shadowManager=new DS(i,a);var s=document.createElement("div");s.style.cssText="overflow:hidden;position:relative",this._svgDom=r,this._svgRoot=a,this._backgroundRoot=o,this._viewport=s,t.appendChild(s),s.appendChild(r),this.resize(n.width,n.height),this._visibleList=[]}return t.prototype.getType=function(){return"svg"},t.prototype.getViewportRoot=function(){return this._viewport},t.prototype.getSvgDom=function(){return this._svgDom},t.prototype.getSvgRoot=function(){return this._svgRoot},t.prototype.getViewportRootOffset=function(){var t=this.getViewportRoot();if(t)return{offsetLeft:t.offsetLeft||0,offsetTop:t.offsetTop||0}},t.prototype.refresh=function(){var t=this.storage.getDisplayList(!0);this._paintList(t)},t.prototype.setBackgroundColor=function(t){this._backgroundRoot&&this._backgroundNode&&this._backgroundRoot.removeChild(this._backgroundNode);var e=OM("rect");e.setAttribute("width",this.getWidth()),e.setAttribute("height",this.getHeight()),e.setAttribute("x",0),e.setAttribute("y",0),e.setAttribute("id",0),e.style.fill=t,this._backgroundRoot.appendChild(e),this._backgroundNode=e},t.prototype.createSVGElement=function(t){return OM(t)},t.prototype.paintOne=function(t){var e=YS(t);return e&&e.brush(t),HS(t)},t.prototype._paintList=function(t){var e=this._gradientManager,n=this._patternManager,i=this._clipPathManager,r=this._shadowManager;e.markAllUnused(),n.markAllUnused(),i.markAllUnused(),r.markAllUnused();for(var o=this._svgRoot,a=this._visibleList,s=t.length,l=[],u=0;u<s;u++){var c=t[u],d=YS(c),h=HS(c);c.invisible||(!c.__dirty&&h||(d&&d.brush(c),h=HS(c),h&&c.style&&(e.update(c.style.fill),e.update(c.style.stroke),n.update(c.style.fill),n.update(c.style.stroke),r.update(h,c)),c.__dirty=0),h&&l.push(c))}var f,p,m,g,y=jM(a,l);for(u=0;u<y.length;u++){var v=y[u];if(v.removed)for(var _=0;_<v.count;_++){c=a[v.indices[_]],h=HS(c);SS(c)?RS(h):NS(o,h)}}for(u=0;u<y.length;u++){v=y[u];if(!v.removed)for(_=0;_<v.count;_++){c=l[v.indices[_]];var b=i.update(c,m);b!==g&&(f=p,b&&(f?PS(o,b,f):ES(o,b),p=b,f=null),g=b);h=HS(c);f?PS(g||o,h,f):ES(g||o,h),f=h||f,g||(p=f),e.markUsed(c),e.addWithoutUpdate(h,c),n.markUsed(c),n.addWithoutUpdate(h,c),i.markUsed(c),m=c}}e.removeUnused(),n.removeUnused(),i.removeUnused(),r.removeUnused(),this._visibleList=l},t.prototype.resize=function(t,e){var n=this._viewport;n.style.display="none";var i=this._opts;if(null!=t&&(i.width=t),null!=e&&(i.height=e),t=this._getSize(0),e=this._getSize(1),n.style.display="",this._width!==t||this._height!==e){this._width=t,this._height=e;var r=n.style;r.width=t+"px",r.height=e+"px";var o=this._svgDom;o.setAttribute("width",t+""),o.setAttribute("height",e+"")}this._backgroundNode&&(this._backgroundNode.setAttribute("width",t),this._backgroundNode.setAttribute("height",e))},t.prototype.getWidth=function(){return this._width},t.prototype.getHeight=function(){return this._height},t.prototype._getSize=function(t){var e=this._opts,n=["width","height"][t],i=["clientWidth","clientHeight"][t],r=["paddingLeft","paddingTop"][t],o=["paddingRight","paddingBottom"][t];if(null!=e[n]&&"auto"!==e[n])return parseFloat(e[n]);var a=this.root,s=document.defaultView.getComputedStyle(a);return(a[i]||AS(s[n])||AS(a.style[n]))-(AS(s[r])||0)-(AS(s[o])||0)|0},t.prototype.dispose=function(){this.root.innerHTML="",this._svgRoot=this._backgroundRoot=this._svgDom=this._backgroundNode=this._viewport=this.storage=null},t.prototype.clear=function(){var t=this._viewport;t&&t.parentNode&&t.parentNode.removeChild(t)},t.prototype.toDataURL=function(){this.refresh();var t=this._svgDom,e=t.outerHTML||(t.parentNode&&t.parentNode).innerHTML,n=encodeURIComponent(e.replace(/></g,">\n\r<"));return"data:image/svg+xml;charset=UTF-8,"+n},t}();function zS(t){return function(){k('In SVG mode painter not support method "'+t+'"')}}var WS=jS;function BS(t){t.registerPainter("svg",WS)}function FS(t){return null==t?0:t.length||1}function VS(t){return t}var GS=function(){function t(t,e,n,i,r,o){this._old=t,this._new=e,this._oldKeyGetter=n||VS,this._newKeyGetter=i||VS,this.context=r,this._diffModeMultiple="multiple"===o}return t.prototype.add=function(t){return this._add=t,this},t.prototype.update=function(t){return this._update=t,this},t.prototype.updateManyToOne=function(t){return this._updateManyToOne=t,this},t.prototype.updateOneToMany=function(t){return this._updateOneToMany=t,this},t.prototype.remove=function(t){return this._remove=t,this},t.prototype.execute=function(){this[this._diffModeMultiple?"_executeMultiple":"_executeOneToOne"]()},t.prototype._executeOneToOne=function(){var t=this._old,e=this._new,n={},i=new Array(t.length),r=new Array(e.length);this._initIndexMap(t,null,i,"_oldKeyGetter"),this._initIndexMap(e,n,r,"_newKeyGetter");for(var o=0;o<t.length;o++){var a=i[o],s=n[a],l=FS(s);if(l>1){var u=s.shift();1===s.length&&(n[a]=s[0]),this._update&&this._update(u,o)}else 1===l?(n[a]=null,this._update&&this._update(s,o)):this._remove&&this._remove(o)}this._performRestAdd(r,n)},t.prototype._executeMultiple=function(){var t=this._old,e=this._new,n={},i={},r=[],o=[];this._initIndexMap(t,n,r,"_oldKeyGetter"),this._initIndexMap(e,i,o,"_newKeyGetter");for(var a=0;a<r.length;a++){var s=r[a],l=n[s],u=i[s],c=FS(l),d=FS(u);if(c>1&&1===d)this._updateManyToOne&&this._updateManyToOne(u,l),i[s]=null;else if(1===c&&d>1)this._updateOneToMany&&this._updateOneToMany(u,l),i[s]=null;else if(1===c&&1===d)this._update&&this._update(u,l),i[s]=null;else if(c>1)for(var h=0;h<c;h++)this._remove&&this._remove(l[h]);else this._remove&&this._remove(l)}this._performRestAdd(o,i)},t.prototype._performRestAdd=function(t,e){for(var n=0;n<t.length;n++){var i=t[n],r=e[i],o=FS(r);if(o>1)for(var a=0;a<o;a++)this._add&&this._add(r[a]);else 1===o&&this._add&&this._add(r);e[i]=null}},t.prototype._initIndexMap=function(t,e,n,i){for(var r=this._diffModeMultiple,o=0;o<t.length;o++){var a="_ec_"+this[i](t[o],o);if(r||(n[o]=a),e){var s=e[a],l=FS(s);0===l?(e[a]=o,r&&n.push(a)):1===l?e[a]=[s,o]:s.push(o)}}},t}(),$S=GS;function US(t){var e={},n=e.encode={},i=gt(),r=[],o=[],a=e.userOutput={dimensionNames:t.dimensions.slice(),encode:{}};R(t.dimensions,(function(e){var s=t.getDimensionInfo(e),l=s.coordDim;if(l){0;var u=s.coordDimIndex;qS(n,l)[u]=e,s.isExtraCoord||(i.set(l,1),ZS(s.type)&&(r[0]=e),qS(a.encode,l)[u]=s.index),s.defaultTooltip&&o.push(e)}Lp.each((function(t,e){var i=qS(n,e),r=s.otherDims[e];null!=r&&!1!==r&&(i[r]=s.name)}))}));var s=[],l={};i.each((function(t,e){var i=n[e];l[e]=i[0],s=s.concat(i)})),e.dataDimsOnCoord=s,e.encodeFirstDimNotExtra=l;var u=n.label;u&&u.length&&(r=u.slice());var c=n.tooltip;return c&&c.length?o=c.slice():o.length||(o=r.slice()),n.defaultedLabel=r,n.defaultedTooltip=o,e}function qS(t,e){return t.hasOwnProperty(e)||(t[e]=[]),t[e]}function XS(t){return"category"===t?"ordinal":"time"===t?"time":"float"}function ZS(t){return!("ordinal"===t||"time"===t)}var JS,KS,QS,tL,eL,nL,iL,rL,oL,aL,sL,lL,uL,cL,dL=function(){function t(t){this.otherDims={},null!=t&&I(this,t)}return t}(),hL=dL,fL=Math.floor,pL=J,mL=H,gL="undefined",yL=-1,vL="e\0\0",_L={float:typeof Float64Array===gL?Array:Float64Array,int:typeof Int32Array===gL?Array:Int32Array,ordinal:Array,number:Array,time:Array},bL=typeof Uint32Array===gL?Array:Uint32Array,xL=typeof Int32Array===gL?Array:Int32Array,wL=typeof Uint16Array===gL?Array:Uint16Array,ML=["hasItemOption","_nameList","_idList","_invertedIndicesMap","_rawData","_dimValueGetter","_count","_rawCount","_nameDimIdx","_idDimIdx","_nameRepeatCount"],SL=["_extent","_approximateExtent","_rawExtent"],LL=function(){function t(t,e){this.type="list",this._count=0,this._rawCount=0,this._storage={},this._storageArr=[],this._nameList=[],this._idList=[],this._visual={},this._layout={},this._itemVisuals=[],this._itemLayouts=[],this._graphicEls=[],this._rawExtent={},this._extent={},this._approximateExtent={},this._calculationInfo={},this.hasItemOption=!0,this.TRANSFERABLE_METHODS=["cloneShallow","downSample","lttbDownSample","map"],this.CHANGABLE_METHODS=["filterSelf","selectRange"],this.DOWNSAMPLE_METHODS=["downSample","lttbDownSample"],this.getRawIndex=eL,t=t||["x","y"];for(var n={},i=[],r={},o=0;o<t.length;o++){var a=t[o],s=q(a)?new hL({name:a}):a instanceof hL?a:new hL(a),l=s.name;s.type=s.type||"float",s.coordDim||(s.coordDim=l,s.coordDimIndex=0);var u=s.otherDims=s.otherDims||{};i.push(l),n[l]=s,s.index=o,s.createInvertedIndices&&(r[l]=[]),0===u.itemName&&(this._nameDimIdx=o,this._nameOrdinalMeta=s.ordinalMeta),0===u.itemId&&(this._idDimIdx=o,this._idOrdinalMeta=s.ordinalMeta)}this.dimensions=i,this._dimensionInfos=n,this.hostModel=e,this._dimensionsSummary=US(this),this._invertedIndicesMap=r,this.userOutput=this._dimensionsSummary.userOutput}return t.prototype.getDimension=function(t){return"number"!==typeof t&&(isNaN(t)||this._dimensionInfos.hasOwnProperty(t))||(t=this.dimensions[t]),t},t.prototype.getDimensionInfo=function(t){return this._dimensionInfos[this.getDimension(t)]},t.prototype.getDimensionsOnCoord=function(){return this._dimensionsSummary.dataDimsOnCoord.slice()},t.prototype.mapDimension=function(t,e){var n=this._dimensionsSummary;if(null==e)return n.encodeFirstDimNotExtra[t];var i=n.encode[t];return i?i[e]:null},t.prototype.mapDimensionsAll=function(t){var e=this._dimensionsSummary,n=e.encode[t];return(n||[]).slice()},t.prototype.initData=function(t,e,n){var i=rg(t)||N(t),r=i?new pg(t,this.dimensions.length):t;this._rawData=r;var o=r.getSource().sourceFormat;this._storage={},this._indices=null,this._dontMakeIdFromName=null!=this._idDimIdx||o===Ip||!!r.fillStorage,this._nameList=(e||[]).slice(),this._idList=[],this._nameRepeatCount={},n||(this.hasItemOption=!1),this.defaultDimValueGetter=JS[o],this._dimValueGetter=n=n||this.defaultDimValueGetter,this._dimValueGetterArrayRows=JS.arrayRows,this._rawExtent={},this._initDataFromProvider(0,r.count()),r.pure&&(this.hasItemOption=!1)},t.prototype.getProvider=function(){return this._rawData},t.prototype.appendData=function(t){var e=this._rawData,n=this.count();e.appendData(t);var i=e.count();e.persistent||(i+=n),this._initDataFromProvider(n,i,!0)},t.prototype.appendValues=function(t,e){for(var n=this._storage,i=this.dimensions,r=i.length,o=this._rawExtent,a=this.count(),s=a+Math.max(t.length,e?e.length:0),l=0;l<r;l++){var u=i[l];o[u]||(o[u]=lL()),tL(n,this._dimensionInfos[u],s,!0)}for(var c=mL(i,(function(t){return o[t]})),d=this._storageArr=mL(i,(function(t){return n[t]})),h=[],f=a;f<s;f++){for(var p=f-a,m=0;m<r;m++){u=i[m];var g=this._dimValueGetterArrayRows(t[p]||h,u,p,m);d[m][f]=g;var y=c[m];g<y[0]&&(y[0]=g),g>y[1]&&(y[1]=g)}e&&(this._nameList[f]=e[p],this._dontMakeIdFromName||oL(this,f))}this._rawCount=this._count=s,this._extent={},KS(this)},t.prototype._initDataFromProvider=function(t,e,n){if(!(t>=e)){for(var i=this._rawData,r=this._storage,o=this.dimensions,a=o.length,s=this._dimensionInfos,l=this._nameList,u=this._idList,c=this._rawExtent,d=i.getSource().sourceFormat,h=d===kp,f=0;f<a;f++){var p=o[f];c[p]||(c[p]=lL()),tL(r,s[p],e,n)}var m=this._storageArr=mL(o,(function(t){return r[t]})),g=mL(o,(function(t){return c[t]}));if(i.fillStorage)i.fillStorage(t,e,m,g);else for(var y=[],v=t;v<e;v++){y=i.getItem(v,y);for(var _=0;_<a;_++){p=o[_];var b=m[_],x=this._dimValueGetter(y,p,v,_);b[v]=x;var w=g[_];x<w[0]&&(w[0]=x),x>w[1]&&(w[1]=x)}if(h&&!i.pure&&y){var M=y.name;null==l[v]&&null!=M&&(l[v]=wo(M,null));var S=y.id;null==u[v]&&null!=S&&(u[v]=wo(S,null))}this._dontMakeIdFromName||oL(this,v)}!i.persistent&&i.clean&&i.clean(),this._rawCount=this._count=e,this._extent={},KS(this)}},t.prototype.count=function(){return this._count},t.prototype.getIndices=function(){var t,e=this._indices;if(e){var n=e.constructor,i=this._count;if(n===Array){t=new n(i);for(var r=0;r<i;r++)t[r]=e[r]}else t=new n(e.buffer,0,i)}else{n=QS(this);t=new n(this.count());for(r=0;r<t.length;r++)t[r]=r}return t},t.prototype.getByDimIdx=function(t,e){if(!(e>=0&&e<this._count))return NaN;var n=this._storageArr[t];return n?n[this.getRawIndex(e)]:NaN},t.prototype.get=function(t,e){if(!(e>=0&&e<this._count))return NaN;var n=this._storage[t];return n?n[this.getRawIndex(e)]:NaN},t.prototype.getByRawIndex=function(t,e){if(!(e>=0&&e<this._rawCount))return NaN;var n=this._storage[t];return n?n[e]:NaN},t.prototype.getValues=function(t,e){var n=[];$(t)||(e=t,t=this.dimensions);for(var i=0,r=t.length;i<r;i++)n.push(this.get(t[i],e));return n},t.prototype.hasValue=function(t){for(var e=this._dimensionsSummary.dataDimsOnCoord,n=0,i=e.length;n<i;n++)if(isNaN(this.get(e[n],t)))return!1;return!0},t.prototype.getDataExtent=function(t){t=this.getDimension(t);var e=this._storage[t],n=lL();if(!e)return n;var i,r=this.count(),o=!this._indices;if(o)return this._rawExtent[t].slice();if(i=this._extent[t],i)return i.slice();i=n;for(var a=i[0],s=i[1],l=0;l<r;l++){var u=this.getRawIndex(l),c=e[u];c<a&&(a=c),c>s&&(s=c)}return i=[a,s],this._extent[t]=i,i},t.prototype.getApproximateExtent=function(t){return t=this.getDimension(t),this._approximateExtent[t]||this.getDataExtent(t)},t.prototype.setApproximateExtent=function(t,e){e=this.getDimension(e),this._approximateExtent[e]=t.slice()},t.prototype.getCalculationInfo=function(t){return this._calculationInfo[t]},t.prototype.setCalculationInfo=function(t,e){pL(t)?I(this._calculationInfo,t):this._calculationInfo[t]=e},t.prototype.getSum=function(t){var e=this._storage[t],n=0;if(e)for(var i=0,r=this.count();i<r;i++){var o=this.get(t,i);isNaN(o)||(n+=o)}return n},t.prototype.getMedian=function(t){var e=[];this.each(t,(function(t){isNaN(t)||e.push(t)}));var n=e.sort((function(t,e){return t-e})),i=this.count();return 0===i?0:i%2===1?n[(i-1)/2]:(n[i/2]+n[i/2-1])/2},t.prototype.rawIndexOf=function(t,e){var n=t&&this._invertedIndicesMap[t];var i=n[e];return null==i||isNaN(i)?yL:i},t.prototype.indexOfName=function(t){for(var e=0,n=this.count();e<n;e++)if(this.getName(e)===t)return e;return-1},t.prototype.indexOfRawIndex=function(t){if(t>=this._rawCount||t<0)return-1;if(!this._indices)return t;var e=this._indices,n=e[t];if(null!=n&&n<this._count&&n===t)return t;var i=0,r=this._count-1;while(i<=r){var o=(i+r)/2|0;if(e[o]<t)i=o+1;else{if(!(e[o]>t))return o;r=o-1}}return-1},t.prototype.indicesOfNearest=function(t,e,n){var i=this._storage,r=i[t],o=[];if(!r)return o;null==n&&(n=1/0);for(var a=1/0,s=-1,l=0,u=0,c=this.count();u<c;u++){var d=this.getRawIndex(u),h=e-r[d],f=Math.abs(h);f<=n&&((f<a||f===a&&h>=0&&s<0)&&(a=f,s=h,l=0),h===s&&(o[l++]=u))}return o.length=l,o},t.prototype.getRawDataItem=function(t){if(this._rawData.persistent)return this._rawData.getItem(this.getRawIndex(t));for(var e=[],n=0;n<this.dimensions.length;n++){var i=this.dimensions[n];e.push(this.get(i,t))}return e},t.prototype.getName=function(t){var e=this.getRawIndex(t),n=this._nameList[e];return null==n&&null!=this._nameDimIdx&&(n=rL(this,this._nameDimIdx,this._nameOrdinalMeta,e)),null==n&&(n=""),n},t.prototype.getId=function(t){return iL(this,this.getRawIndex(t))},t.prototype.each=function(t,e,n,i){var r=this;if(this._count){"function"===typeof t&&(i=n,n=e,e=t,t=[]);var o=n||i||this,a=mL(aL(t),this.getDimension,this);0;for(var s=a.length,l=mL(a,(function(t){return r._dimensionInfos[t].index})),u=this._storageArr,c=0,d=this.count();c<d;c++){var h=this.getRawIndex(c);switch(s){case 0:e.call(o,c);break;case 1:e.call(o,u[l[0]][h],c);break;case 2:e.call(o,u[l[0]][h],u[l[1]][h],c);break;default:for(var f=0,p=[];f<s;f++)p[f]=u[l[f]][h];p[f]=c,e.apply(o,p)}}}},t.prototype.filterSelf=function(t,e,n,i){var r=this;if(this._count){"function"===typeof t&&(i=n,n=e,e=t,t=[]);var o=n||i||this,a=mL(aL(t),this.getDimension,this);0;for(var s=this.count(),l=QS(this),u=new l(s),c=[],d=a.length,h=0,f=mL(a,(function(t){return r._dimensionInfos[t].index})),p=f[0],m=this._storageArr,g=0;g<s;g++){var y=void 0,v=this.getRawIndex(g);if(0===d)y=e.call(o,g);else if(1===d){var _=m[p][v];y=e.call(o,_,g)}else{for(var b=0;b<d;b++)c[b]=m[f[b]][v];c[b]=g,y=e.apply(o,c)}y&&(u[h++]=v)}return h<s&&(this._indices=u),this._count=h,this._extent={},this.getRawIndex=this._indices?nL:eL,this}},t.prototype.selectRange=function(t){var e=this,n=this._count;if(n){var i=[];for(var r in t)t.hasOwnProperty(r)&&i.push(r);0;var o=i.length;if(o){var a=this.count(),s=QS(this),l=new s(a),u=0,c=i[0],d=mL(i,(function(t){return e._dimensionInfos[t].index})),h=t[c][0],f=t[c][1],p=this._storageArr,m=!1;if(!this._indices){var g=0;if(1===o){for(var y=p[d[0]],v=0;v<n;v++){var _=y[v];(_>=h&&_<=f||isNaN(_))&&(l[u++]=g),g++}m=!0}else if(2===o){y=p[d[0]];var b=p[d[1]],x=t[i[1]][0],w=t[i[1]][1];for(v=0;v<n;v++){_=y[v];var M=b[v];(_>=h&&_<=f||isNaN(_))&&(M>=x&&M<=w||isNaN(M))&&(l[u++]=g),g++}m=!0}}if(!m)if(1===o)for(v=0;v<a;v++){var S=this.getRawIndex(v);_=p[d[0]][S];(_>=h&&_<=f||isNaN(_))&&(l[u++]=S)}else for(v=0;v<a;v++){for(var L=!0,k=(S=this.getRawIndex(v),0);k<o;k++){var T=i[k];_=p[d[k]][S];(_<t[T][0]||_>t[T][1])&&(L=!1)}L&&(l[u++]=this.getRawIndex(v))}return u<a&&(this._indices=l),this._count=u,this._extent={},this.getRawIndex=this._indices?nL:eL,this}}},t.prototype.mapArray=function(t,e,n,i){"function"===typeof t&&(i=n,n=e,e=t,t=[]),n=n||i||this;var r=[];return this.each(t,(function(){r.push(e&&e.apply(this,arguments))}),n),r},t.prototype.map=function(t,e,n,i){var r=n||i||this,o=mL(aL(t),this.getDimension,this);var a=sL(this,o),s=a._storage;a._indices=this._indices,a.getRawIndex=a._indices?nL:eL;for(var l=[],u=o.length,c=this.count(),d=[],h=a._rawExtent,f=0;f<c;f++){for(var p=0;p<u;p++)d[p]=this.get(o[p],f);d[u]=f;var m=e&&e.apply(r,d);if(null!=m){"object"!==typeof m&&(l[0]=m,m=l);for(var g=this.getRawIndex(f),y=0;y<m.length;y++){var v=o[y],_=m[y],b=h[v],x=s[v];x&&(x[g]=_),_<b[0]&&(b[0]=_),_>b[1]&&(b[1]=_)}}}return a},t.prototype.downSample=function(t,e,n,i){for(var r=sL(this,[t]),o=r._storage,a=[],s=fL(1/e),l=o[t],u=this.count(),c=r._rawExtent[t],d=new(QS(this))(u),h=0,f=0;f<u;f+=s){s>u-f&&(s=u-f,a.length=s);for(var p=0;p<s;p++){var m=this.getRawIndex(f+p);a[p]=l[m]}var g=n(a),y=this.getRawIndex(Math.min(f+i(a,g)||0,u-1));l[y]=g,g<c[0]&&(c[0]=g),g>c[1]&&(c[1]=g),d[h++]=y}return r._count=h,r._indices=d,r.getRawIndex=nL,r},t.prototype.lttbDownSample=function(t,e){var n,i,r,o=sL(this,[]),a=o._storage,s=a[t],l=this.count(),u=new(QS(this))(l),c=0,d=fL(1/e),h=this.getRawIndex(0);u[c++]=h;for(var f=1;f<l-1;f+=d){for(var p=Math.min(f+d,l-1),m=Math.min(f+2*d,l),g=(m+p)/2,y=0,v=p;v<m;v++){var _=this.getRawIndex(v),b=s[_];isNaN(b)||(y+=b)}y/=m-p;var x=f,w=Math.min(f+d,l),M=f-1,S=s[h];n=-1,r=x;for(v=x;v<w;v++){_=this.getRawIndex(v),b=s[_];isNaN(b)||(i=Math.abs((M-g)*(b-S)-(M-v)*(y-S)),i>n&&(n=i,r=_))}u[c++]=r,h=r}return u[c++]=this.getRawIndex(l-1),o._count=c,o._indices=u,o.getRawIndex=nL,o},t.prototype.getItemModel=function(t){var e=this.hostModel,n=this.getRawDataItem(t);return new qh(n,e,e&&e.ecModel)},t.prototype.diff=function(t){var e=this;return new $S(t?t.getIndices():[],this.getIndices(),(function(e){return iL(t,e)}),(function(t){return iL(e,t)}))},t.prototype.getVisual=function(t){var e=this._visual;return e&&e[t]},t.prototype.setVisual=function(t,e){this._visual=this._visual||{},pL(t)?I(this._visual,t):this._visual[t]=e},t.prototype.getItemVisual=function(t,e){var n=this._itemVisuals[t],i=n&&n[e];return null==i?this.getVisual(e):i},t.prototype.hasItemVisual=function(){return this._itemVisuals.length>0},t.prototype.ensureUniqueItemVisual=function(t,e){var n=this._itemVisuals,i=n[t];i||(i=n[t]={});var r=i[e];return null==r&&(r=this.getVisual(e),$(r)?r=r.slice():pL(r)&&(r=I({},r)),i[e]=r),r},t.prototype.setItemVisual=function(t,e,n){var i=this._itemVisuals[t]||{};this._itemVisuals[t]=i,pL(e)?I(i,e):i[e]=n},t.prototype.clearAllVisual=function(){this._visual={},this._itemVisuals=[]},t.prototype.setLayout=function(t,e){if(pL(t))for(var n in t)t.hasOwnProperty(n)&&this.setLayout(n,t[n]);else this._layout[t]=e},t.prototype.getLayout=function(t){return this._layout[t]},t.prototype.getItemLayout=function(t){return this._itemLayouts[t]},t.prototype.setItemLayout=function(t,e,n){this._itemLayouts[t]=n?I(this._itemLayouts[t]||{},e):e},t.prototype.clearItemLayouts=function(){this._itemLayouts.length=0},t.prototype.setItemGraphicEl=function(t,e){var n=this.hostModel;if(e){var i=$l(e);i.dataIndex=t,i.dataType=this.dataType,i.seriesIndex=n&&n.seriesIndex,"group"===e.type&&e.traverse(uL,e)}this._graphicEls[t]=e},t.prototype.getItemGraphicEl=function(t){return this._graphicEls[t]},t.prototype.eachItemGraphicEl=function(t,e){R(this._graphicEls,(function(n,i){n&&t&&t.call(e,n,i)}))},t.prototype.cloneShallow=function(e){if(!e){var n=mL(this.dimensions,this.getDimensionInfo,this);e=new t(n,this.hostModel)}if(e._storage=this._storage,e._storageArr=this._storageArr,cL(e,this),this._indices){var i=this._indices.constructor;if(i===Array){var r=this._indices.length;e._indices=new i(r);for(var o=0;o<r;o++)e._indices[o]=this._indices[o]}else e._indices=new i(this._indices)}else e._indices=null;return e.getRawIndex=e._indices?nL:eL,e},t.prototype.wrapMethod=function(t,e){var n=this[t];"function"===typeof n&&(this.__wrappedMethods=this.__wrappedMethods||[],this.__wrappedMethods.push(t),this[t]=function(){var t=n.apply(this,arguments);return e.apply(this,[t].concat(lt(arguments)))})},t.internalField=function(){function e(t,e,n,i){return Yg(t[i],this._dimensionInfos[e])}function n(t){var e=t.constructor;return e===Array?t.slice():new e(t)}JS={arrayRows:e,objectRows:function(t,e,n,i){return Yg(t[e],this._dimensionInfos[e])},keyedColumns:e,original:function(t,e,n,i){var r=t&&(null==t.value?t:t.value);return!this._rawData.pure&&ho(t)&&(this.hasItemOption=!0),Yg(r instanceof Array?r[i]:r,this._dimensionInfos[e])},typedArray:function(t,e,n,i){return t[i]}},KS=function(t){var e=t._invertedIndicesMap;R(e,(function(n,i){var r=t._dimensionInfos[i],o=r.ordinalMeta;if(o){n=e[i]=new xL(o.categories.length);for(var a=0;a<n.length;a++)n[a]=yL;for(a=0;a<t._count;a++)n[t.get(i,a)]=a}}))},rL=function(t,e,n,i){var r,o=t._storageArr[e];return o&&(r=o[i],n&&n.categories.length&&(r=n.categories[r])),wo(r,null)},QS=function(t){return t._rawCount>65535?bL:wL},tL=function(t,e,n,i){var r=_L[e.type],o=e.name;if(i){var a=t[o],s=a&&a.length;if(s!==n){for(var l=new r(n),u=0;u<s;u++)l[u]=a[u];t[o]=l}}else t[o]=new r(n)},eL=function(t){return t},nL=function(t){return t<this._count&&t>=0?this._indices[t]:-1},iL=function(t,e){var n=t._idList[e];return null==n&&null!=t._idDimIdx&&(n=rL(t,t._idDimIdx,t._idOrdinalMeta,e)),null==n&&(n=vL+e),n},aL=function(t){return $(t)||(t=null!=t?[t]:[]),t},function(t,e){for(var n=0;n<e.length;n++)t._dimensionInfos[e[n]]},sL=function(e,i){var r=e.dimensions,o=new t(mL(r,e.getDimensionInfo,e),e.hostModel);cL(o,e);for(var a=o._storage={},s=e._storage,l=o._storageArr=[],u=0;u<r.length;u++){var c=r[u];s[c]&&(O(i,c)>=0?(a[c]=n(s[c]),o._rawExtent[c]=lL(),o._extent[c]=null):a[c]=s[c],l.push(a[c]))}return o},lL=function(){return[1/0,-1/0]},uL=function(t){var e=$l(t),n=$l(this);e.seriesIndex=n.seriesIndex,e.dataIndex=n.dataIndex,e.dataType=n.dataType},cL=function(t,e){R(ML.concat(e.__wrappedMethods||[]),(function(n){e.hasOwnProperty(n)&&(t[n]=e[n])})),t.__wrappedMethods=e.__wrappedMethods,R(SL,(function(n){t[n]=T(e[n])})),t._calculationInfo=I({},e._calculationInfo)},oL=function(t,e){var n=t._nameList,i=t._idList,r=t._nameDimIdx,o=t._idDimIdx,a=n[e],s=i[e];if(null==a&&null!=r&&(n[e]=a=rL(t,r,t._nameOrdinalMeta,e)),null==s&&null!=o&&(i[e]=s=rL(t,o,t._idOrdinalMeta,e)),null==s&&null!=a){var l=t._nameRepeatCount,u=l[a]=(l[a]||0)+1;s=a,u>1&&(s+="__ec__"+u),i[e]=s}}}(),t}(),kL=LL;function TL(t,e,n){rg(e)||(e=ag(e)),n=n||{},t=(t||[]).slice();for(var i=(n.dimsDef||[]).slice(),r=gt(),o=gt(),a=[],s=DL(e,t,i,n.dimCount),l=0;l<s;l++){var u=i[l],c=i[l]=I({},J(u)?u:{name:u}),d=c.name,h=a[l]=new hL;null!=d&&null==r.get(d)&&(h.name=h.displayName=d,r.set(d,l)),null!=c.type&&(h.type=c.type),null!=c.displayName&&(h.displayName=c.displayName)}var f=n.encodeDef;!f&&n.encodeDefaulter&&(f=n.encodeDefaulter(e,s));var p=gt(f);p.each((function(t,e){var n=so(t).slice();if(1===n.length&&!q(n[0])&&n[0]<0)p.set(e,!1);else{var i=p.set(e,[]);R(n,(function(t,n){var o=q(t)?r.get(t):t;null!=o&&o<s&&(i[n]=o,g(a[o],e,n))}))}}));var m=0;function g(t,e,n){null!=Lp.get(e)?t.otherDims[e]=n:(t.coordDim=e,t.coordDimIndex=n,o.set(e,!0))}R(t,(function(t){var e,n,i,r;if(q(t))e=t,r={};else{r=t,e=r.name;var o=r.ordinalMeta;r.ordinalMeta=null,r=T(r),r.ordinalMeta=o,n=r.dimsDef,i=r.otherDims,r.name=r.coordDim=r.coordDimIndex=r.dimsDef=r.otherDims=null}var s=p.get(e);if(!1!==s){if(s=so(s),!s.length)for(var l=0;l<(n&&n.length||1);l++){while(m<a.length&&null!=a[m].coordDim)m++;m<a.length&&s.push(m++)}R(s,(function(t,o){var s=a[t];if(g(A(s,r),e,o),null==s.name&&n){var l=n[o];!J(l)&&(l={name:l}),s.name=s.displayName=l.name,s.defaultTooltip=l.defaultTooltip}i&&A(s.otherDims,i)}))}}));var y=n.generateCoord,v=n.generateCoordCount,_=null!=v;v=y?v||1:0;for(var b=y||"value",x=0;x<s;x++){h=a[x]=a[x]||new hL;var w=h.coordDim;null==w&&(h.coordDim=CL(b,o,_),h.coordDimIndex=0,(!y||v<=0)&&(h.isExtraCoord=!0),v--),null==h.name&&(h.name=CL(h.coordDim,r,!1)),null!=h.type||Wp(e,x)!==Pp.Must&&(!h.isExtraCoord||null==h.otherDims.itemName&&null==h.otherDims.seriesName)||(h.type="ordinal")}return a}function DL(t,e,n,i){var r=Math.max(t.dimensionsDetectedCount||1,e.length,n.length,i||0);return R(e,(function(t){var e;J(t)&&(e=t.dimsDef)&&(r=Math.max(r,e.length))})),r}function CL(t,e,n){if(n||null!=e.get(t)){var i=0;while(null!=e.get(t+i))i++;t+=i}return e.set(t,!0),t}var IL=TL;function AL(t,e){return e=e||{},IL(e.coordDimensions||[],t,{dimsDef:e.dimensionsDefine||t.dimensionsDefine,encodeDef:e.encodeDefine||t.encodeDefine,dimCount:e.dimensionsCount,encodeDefaulter:e.encodeDefaulter,generateCoord:e.generateCoord,generateCoordCount:e.generateCoordCount})}var YL=function(){function t(t){this.coordSysDims=[],this.axisMap=gt(),this.categoryAxisMap=gt(),this.coordSysName=t}return t}();function OL(t){var e=t.get("coordinateSystem"),n=new YL(e),i=PL[e];if(i)return i(t,n,n.axisMap,n.categoryAxisMap),n}var PL={cartesian2d:function(t,e,n,i){var r=t.getReferringComponents("xAxis",Po).models[0],o=t.getReferringComponents("yAxis",Po).models[0];e.coordSysDims=["x","y"],n.set("x",r),n.set("y",o),EL(r)&&(i.set("x",r),e.firstCategoryDimIndex=0),EL(o)&&(i.set("y",o),null==e.firstCategoryDimIndex&&(e.firstCategoryDimIndex=1))},singleAxis:function(t,e,n,i){var r=t.getReferringComponents("singleAxis",Po).models[0];e.coordSysDims=["single"],n.set("single",r),EL(r)&&(i.set("single",r),e.firstCategoryDimIndex=0)},polar:function(t,e,n,i){var r=t.getReferringComponents("polar",Po).models[0],o=r.findAxisModel("radiusAxis"),a=r.findAxisModel("angleAxis");e.coordSysDims=["radius","angle"],n.set("radius",o),n.set("angle",a),EL(o)&&(i.set("radius",o),e.firstCategoryDimIndex=0),EL(a)&&(i.set("angle",a),null==e.firstCategoryDimIndex&&(e.firstCategoryDimIndex=1))},geo:function(t,e,n,i){e.coordSysDims=["lng","lat"]},parallel:function(t,e,n,i){var r=t.ecModel,o=r.getComponent("parallel",t.get("parallelIndex")),a=e.coordSysDims=o.dimensions.slice();R(o.parallelAxisIndex,(function(t,o){var s=r.getComponent("parallelAxis",t),l=a[o];n.set(l,s),EL(s)&&(i.set(l,s),null==e.firstCategoryDimIndex&&(e.firstCategoryDimIndex=o))}))}};function EL(t){return"category"===t.get("type")}function NL(t,e,n){n=n||{};var i,r,o,a,s=n.byIndex,l=n.stackedCoordDimension,u=!(!t||!t.get("stack"));if(R(e,(function(t,n){q(t)&&(e[n]=t={name:t}),u&&!t.isExtraCoord&&(s||i||!t.ordinalMeta||(i=t),r||"ordinal"===t.type||"time"===t.type||l&&l!==t.coordDim||(r=t))})),!r||s||i||(s=!0),r){o="__\0ecstackresult",a="__\0ecstackedover",i&&(i.createInvertedIndices=!0);var c=r.coordDim,d=r.type,h=0;R(e,(function(t){t.coordDim===c&&h++})),e.push({name:o,coordDim:c,coordDimIndex:h,type:d,isExtraCoord:!0,isCalculationCoord:!0}),h++,e.push({name:a,coordDim:a,coordDimIndex:h,type:d,isExtraCoord:!0,isCalculationCoord:!0})}return{stackedDimension:r&&r.name,stackedByDimension:i&&i.name,isStackedByIndex:s,stackedOverDimension:a,stackResultDimension:o}}function RL(t,e){return!!e&&e===t.getCalculationInfo("stackedDimension")}function HL(t,e){return RL(t,e)?t.getCalculationInfo("stackResultDimension"):e}function jL(t,e,n){n=n||{},rg(t)||(t=ag(t));var i,r=e.get("coordinateSystem"),o=vm.get(r),a=OL(e);a&&a.coordSysDims&&(i=H(a.coordSysDims,(function(t){var e={name:t},n=a.axisMap.get(t);if(n){var i=n.get("type");e.type=XS(i)}return e}))),i||(i=o&&(o.getDimensionsInfo?o.getDimensionsInfo():o.dimensions.slice())||["x","y"]);var s,l,u=n.useEncodeDefaulter,c=AL(t,{coordDimensions:i,generateCoord:n.generateCoord,encodeDefaulter:U(u)?u:u?G(Rp,i,e):null});a&&R(c,(function(t,e){var i=t.coordDim,r=a.categoryAxisMap.get(i);r&&(null==s&&(s=e),t.ordinalMeta=r.getOrdinalMeta(),n.createInvertedIndices&&(t.createInvertedIndices=!0)),null!=t.otherDims.itemName&&(l=!0)})),l||null==s||(c[s].otherDims.itemName=0);var d=NL(e,c),h=new kL(c,e);h.setCalculationInfo(d);var f=null!=s&&zL(t)?function(t,e,n,i){return i===s?n:this.defaultDimValueGetter(t,e,n,i)}:null;return h.hasItemOption=!1,h.initData(t,null,f),h}function zL(t){if(t.sourceFormat===kp){var e=WL(t.data||[]);return null!=e&&!$(co(e))}}function WL(t){var e=0;while(e<t.length&&null==t[e])e++;return t[e]}var BL=jL,FL=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.getInitialData=function(t){return BL(this.getSource(),this,{useEncodeDefaulter:!0})},e.prototype.getLegendIcon=function(t){var e=new Mr,n=j_("line",0,t.itemHeight/2,t.itemWidth,0,t.lineStyle.stroke,!1);e.add(n),n.setStyle(t.lineStyle);var i=this.getData().getVisual("symbol"),r=this.getData().getVisual("symbolRotate"),o="none"===i?"circle":i,a=.8*t.itemHeight,s=j_(o,(t.itemWidth-a)/2,(t.itemHeight-a)/2,a,a,t.itemStyle.fill);e.add(s),s.setStyle(t.itemStyle);var l="inherit"===t.iconRotate?r:t.iconRotate||0;return s.rotation=l*Math.PI/180,s.setOrigin([t.itemWidth/2,t.itemHeight/2]),o.indexOf("empty")>-1&&(s.style.stroke=s.style.fill,s.style.fill="#fff",s.style.lineWidth=2),e},e.type="series.line",e.dependencies=["grid","polar"],e.defaultOption={zlevel:0,z:3,coordinateSystem:"cartesian2d",legendHoverLink:!0,clip:!0,label:{position:"top"},endLabel:{show:!1,valueAnimation:!0,distance:8},lineStyle:{width:2,type:"solid"},emphasis:{scale:!0,lineStyle:{width:"bolder"}},step:!1,smooth:!1,smoothMonotone:null,symbol:"emptyCircle",symbolSize:4,symbolRotate:null,showSymbol:!0,showAllSymbol:"auto",connectNulls:!1,sampling:"none",animationEasing:"linear",progressive:0,hoverLayerThreshold:1/0},e}(Yy),VL=FL;function GL(t,e){var n=t.mapDimensionsAll("defaultedLabel"),i=n.length;if(1===i){var r=Lg(t,e,n[0]);return null!=r?r+"":null}if(i){for(var o=[],a=0;a<n.length;a++)o.push(Lg(t,e,n[a]));return o.join(" ")}}function $L(t,e){var n=t.mapDimensionsAll("defaultedLabel");if(!$(e))return e+"";for(var i=[],r=0;r<n.length;r++){var o=t.getDimensionInfo(n[r]);o&&i.push(e[o.index])}return i.join(" ")}var UL=function(t){function e(e,n,i,r){var o=t.call(this)||this;return o.updateData(e,n,i,r),o}return o(e,t),e.prototype._createSymbol=function(t,e,n,i,r){this.removeAll();var o=j_(t,-1,-1,2,2,null,r);o.attr({z2:100,culling:!0,scaleX:i[0]/2,scaleY:i[1]/2}),o.drift=qL,this._symbolType=t,this.add(o)},e.prototype.stopSymbolAnimation=function(t){this.childAt(0).stopAnimation(null,t)},e.prototype.getSymbolPath=function(){return this.childAt(0)},e.prototype.highlight=function(){Cu(this.childAt(0))},e.prototype.downplay=function(){Iu(this.childAt(0))},e.prototype.setZ=function(t,e){var n=this.childAt(0);n.zlevel=t,n.z=e},e.prototype.setDraggable=function(t){var e=this.childAt(0);e.draggable=t,e.cursor=t?"move":e.cursor},e.prototype.updateData=function(t,n,i,r){this.silent=!1;var o=t.getItemVisual(n,"symbol")||"circle",a=t.hostModel,s=e.getSymbolSize(t,n),l=o!==this._symbolType,u=r&&r.disableAnimation;if(l){var c=t.getItemVisual(n,"symbolKeepAspect");this._createSymbol(o,t,n,s,c)}else{var d=this.childAt(0);d.silent=!1;var h={scaleX:s[0]/2,scaleY:s[1]/2};u?d.attr(h):th(d,h,a,n)}if(this._updateCommon(t,n,s,i,r),l){d=this.childAt(0);if(!u){h={scaleX:this._sizeX,scaleY:this._sizeY,style:{opacity:d.style.opacity}};d.scaleX=d.scaleY=0,d.style.opacity=0,eh(d,h,a,n)}}u&&this.childAt(0).stopAnimation("remove"),this._seriesModel=a},e.prototype._updateCommon=function(t,e,n,i,r){var o,a,s,l,u,c,d,h,f=this.childAt(0),p=t.hostModel;if(i&&(o=i.emphasisItemStyle,a=i.blurItemStyle,s=i.selectItemStyle,l=i.focus,u=i.blurScope,c=i.labelStatesModels,d=i.hoverScale,h=i.cursorStyle),!i||t.hasItemOption){var m=i&&i.itemModel?i.itemModel:t.getItemModel(e),g=m.getModel("emphasis");o=g.getModel("itemStyle").getItemStyle(),s=m.getModel(["select","itemStyle"]).getItemStyle(),a=m.getModel(["blur","itemStyle"]).getItemStyle(),l=g.get("focus"),u=g.get("blurScope"),c=Sh(m),d=g.getShallow("scale"),h=m.getShallow("cursor")}var y=t.getItemVisual(e,"symbolRotate");f.attr("rotation",(y||0)*Math.PI/180||0);var v=t.getItemVisual(e,"symbolOffset")||0;v&&($(v)||(v=[v,v]),f.x=Nr(v[0],n[0]),f.y=Nr(at(v[1],v[0])||0,n[1])),h&&f.attr("cursor",h);var _=t.getItemVisual(e,"style"),b=_.fill;if(f instanceof wl){var x=f.style;f.useStyle(I({image:x.image,x:x.x,y:x.y,width:x.width,height:x.height},_))}else f.__isEmptyBrush?f.useStyle(I({},_)):f.useStyle(_),f.style.decal=null,f.setColor(b,r&&r.symbolInnerColor),f.style.strokeNoScale=!0;var w=t.getItemVisual(e,"liftZ"),M=this._z2;null!=w?null==M&&(this._z2=f.z2,f.z2+=w):null!=M&&(f.z2=M,this._z2=null);var S=r&&r.useNameLabel;function L(e){return S?t.getName(e):GL(t,e)}Mh(f,c,{labelFetcher:p,labelDataIndex:e,defaultText:L,inheritColor:b,defaultOpacity:_.opacity}),this._sizeX=n[0]/2,this._sizeY=n[1]/2;var k=f.ensureState("emphasis");if(k.style=o,f.ensureState("select").style=s,f.ensureState("blur").style=a,d){var T=Math.max(1.1,3/this._sizeY);k.scaleX=this._sizeX*T,k.scaleY=this._sizeY*T}this.setSymbolScale(1),$u(this,l,u)},e.prototype.setSymbolScale=function(t){this.scaleX=this.scaleY=t},e.prototype.fadeOut=function(t,e){var n=this.childAt(0),i=this._seriesModel,r=$l(this).dataIndex,o=e&&e.animation;if(this.silent=n.silent=!0,e&&e.fadeLabel){var a=n.getTextContent();a&&nh(a,{style:{opacity:0}},i,{dataIndex:r,removeOpt:o,cb:function(){n.removeTextContent()}})}else n.removeTextContent();nh(n,{style:{opacity:0},scaleX:0,scaleY:0},i,{dataIndex:r,cb:t,removeOpt:o})},e.getSymbolSize=function(t,e){var n=t.getItemVisual(e,"symbolSize");return $(n)?n.slice():[+n,+n]},e}(Mr);function qL(t,e){this.parent.drift(t,e)}var XL=UL;function ZL(t,e,n,i){return e&&!isNaN(e[0])&&!isNaN(e[1])&&!(i.isIgnore&&i.isIgnore(n))&&!(i.clipShape&&!i.clipShape.contain(e[0],e[1]))&&"none"!==t.getItemVisual(n,"symbol")}function JL(t){return null==t||J(t)||(t={isIgnore:t}),t||{}}function KL(t){var e=t.hostModel,n=e.getModel("emphasis");return{emphasisItemStyle:n.getModel("itemStyle").getItemStyle(),blurItemStyle:e.getModel(["blur","itemStyle"]).getItemStyle(),selectItemStyle:e.getModel(["select","itemStyle"]).getItemStyle(),focus:n.get("focus"),blurScope:n.get("blurScope"),hoverScale:n.get("scale"),labelStatesModels:Sh(e),cursorStyle:e.get("cursor")}}var QL=function(){function t(t){this.group=new Mr,this._SymbolCtor=t||XL}return t.prototype.updateData=function(t,e){e=JL(e);var n=this.group,i=t.hostModel,r=this._data,o=this._SymbolCtor,a=e.disableAnimation,s=KL(t),l={disableAnimation:a},u=e.getSymbolPoint||function(e){return t.getItemLayout(e)};r||n.removeAll(),t.diff(r).add((function(i){var r=u(i);if(ZL(t,r,i,e)){var a=new o(t,i,s,l);a.setPosition(r),t.setItemGraphicEl(i,a),n.add(a)}})).update((function(c,d){var h=r.getItemGraphicEl(d),f=u(c);if(ZL(t,f,c,e)){if(h){h.updateData(t,c,s,l);var p={x:f[0],y:f[1]};a?h.attr(p):th(h,p,i)}else h=new o(t,c),h.setPosition(f);n.add(h),t.setItemGraphicEl(c,h)}else n.remove(h)})).remove((function(t){var e=r.getItemGraphicEl(t);e&&e.fadeOut((function(){n.remove(e)}))})).execute(),this._getSymbolPoint=u,this._data=t},t.prototype.isPersistent=function(){return!0},t.prototype.updateLayout=function(){var t=this,e=this._data;e&&e.eachItemGraphicEl((function(e,n){var i=t._getSymbolPoint(n);e.setPosition(i),e.markRedraw()}))},t.prototype.incrementalPrepareUpdate=function(t){this._seriesScope=KL(t),this._data=null,this.group.removeAll()},t.prototype.incrementalUpdate=function(t,e,n){function i(t){t.isGroup||(t.incremental=!0,t.ensureState("emphasis").hoverLayer=!0)}n=JL(n);for(var r=t.start;r<t.end;r++){var o=e.getItemLayout(r);if(ZL(e,o,r,n)){var a=new this._SymbolCtor(e,r,this._seriesScope);a.traverse(i),a.setPosition(o),this.group.add(a),e.setItemGraphicEl(r,a)}}},t.prototype.remove=function(t){var e=this.group,n=this._data;n&&t?n.eachItemGraphicEl((function(t){t.fadeOut((function(){e.remove(t)}))})):e.removeAll()},t}(),tk=QL;function ek(t,e,n){var i=t.getBaseAxis(),r=t.getOtherAxis(i),o=nk(r,n),a=i.dim,s=r.dim,l=e.mapDimension(s),u=e.mapDimension(a),c="x"===s||"radius"===s?1:0,d=H(t.dimensions,(function(t){return e.mapDimension(t)})),h=!1,f=e.getCalculationInfo("stackResultDimension");return RL(e,d[0])&&(h=!0,d[0]=f),RL(e,d[1])&&(h=!0,d[1]=f),{dataDimsForPoint:d,valueStart:o,valueAxisDim:s,baseAxisDim:a,stacked:!!h,valueDim:l,baseDim:u,baseDataOffset:c,stackedOverDimension:e.getCalculationInfo("stackedOverDimension")}}function nk(t,e){var n=0,i=t.scale.getExtent();return"start"===e?n=i[0]:"end"===e?n=i[1]:i[0]>0?n=i[0]:i[1]<0&&(n=i[1]),n}function ik(t,e,n,i){var r=NaN;t.stacked&&(r=n.get(n.getCalculationInfo("stackedOverDimension"),i)),isNaN(r)&&(r=t.valueStart);var o=t.baseDataOffset,a=[];return a[o]=n.get(t.baseDim,i),a[1-o]=r,e.dataToPoint(a)}var rk="undefined"!==typeof Float32Array,ok=rk?Float32Array:Array;function ak(t){return $(t)?rk?new Float32Array(t):t:new ok(t)}function sk(t,e){var n=[];return e.diff(t).add((function(t){n.push({cmd:"+",idx:t})})).update((function(t,e){n.push({cmd:"=",idx:e,idx1:t})})).remove((function(t){n.push({cmd:"-",idx:t})})).execute(),n}function lk(t,e,n,i,r,o,a,s){for(var l=sk(t,e),u=[],c=[],d=[],h=[],f=[],p=[],m=[],g=ek(r,e,a),y=t.getLayout("points")||[],v=e.getLayout("points")||[],_=0;_<l.length;_++){var b=l[_],x=!0,w=void 0,M=void 0;switch(b.cmd){case"=":w=2*b.idx,M=2*b.idx1;var S=y[w],L=y[w+1],k=v[M],T=v[M+1];(isNaN(S)||isNaN(L))&&(S=k,L=T),u.push(S,L),c.push(k,T),d.push(n[w],n[w+1]),h.push(i[M],i[M+1]),m.push(e.getRawIndex(b.idx1));break;case"+":var D=b.idx,C=g.dataDimsForPoint,I=r.dataToPoint([e.get(C[0],D),e.get(C[1],D)]);M=2*D,u.push(I[0],I[1]),c.push(v[M],v[M+1]);var A=ik(g,r,e,D);d.push(A[0],A[1]),h.push(i[M],i[M+1]),m.push(e.getRawIndex(D));break;case"-":x=!1}x&&(f.push(b),p.push(p.length))}p.sort((function(t,e){return m[t]-m[e]}));var Y=u.length,O=ak(Y),P=ak(Y),E=ak(Y),N=ak(Y),R=[];for(_=0;_<p.length;_++){var H=p[_],j=2*_,z=2*H;O[j]=u[z],O[j+1]=u[z+1],P[j]=c[z],P[j+1]=c[z+1],E[j]=d[z],E[j+1]=d[z+1],N[j]=h[z],N[j+1]=h[z+1],R[_]=f[H]}return{current:O,next:P,stackedOnCurrent:E,stackedOnNext:N,status:R}}var uk=Math.min,ck=Math.max;function dk(t,e){return isNaN(t)||isNaN(e)}function hk(t,e,n,i,r,o,a,s,l){for(var u,c,d,h,f,p,m=n,g=0;g<i;g++){var y=e[2*m],v=e[2*m+1];if(m>=r||m<0)break;if(dk(y,v)){if(l){m+=o;continue}break}if(m===n)t[o>0?"moveTo":"lineTo"](y,v),d=y,h=v;else{var _=y-u,b=v-c;if(_*_+b*b<.5){m+=o;continue}if(a>0){var x=m+o,w=e[2*x],M=e[2*x+1],S=g+1;if(l)while(dk(w,M)&&S<i)S++,x+=o,w=e[2*x],M=e[2*x+1];var L=.5,k=0,T=0,D=void 0,C=void 0;if(S>=i||dk(w,M))f=y,p=v;else{k=w-u,T=M-c;var I=y-u,A=w-y,Y=v-c,O=M-v,P=void 0,E=void 0;"x"===s?(P=Math.abs(I),E=Math.abs(A),f=y-P*a,p=v,D=y+P*a,C=v):"y"===s?(P=Math.abs(Y),E=Math.abs(O),f=y,p=v-P*a,D=y,C=v+P*a):(P=Math.sqrt(I*I+Y*Y),E=Math.sqrt(A*A+O*O),L=E/(E+P),f=y-k*a*(1-L),p=v-T*a*(1-L),D=y+k*a*L,C=v+T*a*L,D=uk(D,ck(w,y)),C=uk(C,ck(M,v)),D=ck(D,uk(w,y)),C=ck(C,uk(M,v)),k=D-y,T=C-v,f=y-k*P/E,p=v-T*P/E,f=uk(f,ck(u,y)),p=uk(p,ck(c,v)),f=ck(f,uk(u,y)),p=ck(p,uk(c,v)),k=y-f,T=v-p,D=y+k*E/P,C=v+T*E/P)}t.bezierCurveTo(d,h,f,p,y,v),d=D,h=C}else t.lineTo(y,v)}u=y,c=v,m+=o}return g}var fk=function(){function t(){this.smooth=0,this.smoothConstraint=!0}return t}(),pk=function(t){function e(e){var n=t.call(this,e)||this;return n.type="ec-polyline",n}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new fk},e.prototype.buildPath=function(t,e){var n=e.points,i=0,r=n.length/2;if(e.connectNulls){for(;r>0;r--)if(!dk(n[2*r-2],n[2*r-1]))break;for(;i<r;i++)if(!dk(n[2*i],n[2*i+1]))break}while(i<r)i+=hk(t,n,i,r,r,1,e.smooth,e.smoothMonotone,e.connectNulls)+1},e.prototype.getPointOn=function(t,e){this.path||(this.createPathProxy(),this.buildPath(this.path,this.shape));for(var n,i,r=this.path,o=r.data,a=Bs.CMD,s="x"===e,l=[],u=0;u<o.length;){var c=o[u++],d=void 0,h=void 0,f=void 0,p=void 0,m=void 0,g=void 0,y=void 0;switch(c){case a.M:n=o[u++],i=o[u++];break;case a.L:if(d=o[u++],h=o[u++],y=s?(t-n)/(d-n):(t-i)/(h-i),y<=1&&y>=0){var v=s?(h-i)*y+i:(d-n)*y+n;return s?[t,v]:[v,t]}n=d,i=h;break;case a.C:d=o[u++],h=o[u++],f=o[u++],p=o[u++],m=o[u++],g=o[u++];var _=s?Xa(n,d,f,m,t,l):Xa(i,h,p,g,t,l);if(_>0)for(var b=0;b<_;b++){var x=l[b];if(x<=1&&x>=0){v=s?Ua(i,h,p,g,x):Ua(n,d,f,m,x);return s?[t,v]:[v,t]}}n=m,i=g;break}}},e}(pl),mk=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e}(fk),gk=function(t){function e(e){var n=t.call(this,e)||this;return n.type="ec-polygon",n}return o(e,t),e.prototype.getDefaultShape=function(){return new mk},e.prototype.buildPath=function(t,e){var n=e.points,i=e.stackedOnPoints,r=0,o=n.length/2,a=e.smoothMonotone;if(e.connectNulls){for(;o>0;o--)if(!dk(n[2*o-2],n[2*o-1]))break;for(;r<o;r++)if(!dk(n[2*r],n[2*r+1]))break}while(r<o){var s=hk(t,n,r,o,o,1,e.smooth,a,e.connectNulls);hk(t,i,r+s-1,s,o,-1,e.stackedOnSmooth,a,e.connectNulls),r+=s+1,t.closePath()}},e}(pl);function yk(t,e,n,i,r){var o=t.getArea(),a=o.x,s=o.y,l=o.width,u=o.height,c=n.get(["lineStyle","width"])||2;a-=c/2,s-=c/2,l+=c,u+=c,a=Math.floor(a),l=Math.round(l);var d=new Al({shape:{x:a,y:s,width:l,height:u}});if(e){var h=t.getBaseAxis(),f=h.isHorizontal(),p=h.inverse;f?(p&&(d.shape.x+=l),d.shape.width=0):(p||(d.shape.y+=u),d.shape.height=0);var m="function"===typeof r?function(t){r(t,d)}:null;eh(d,{shape:{width:l,height:u,x:a,y:s}},n,null,i,m)}return d}function vk(t,e,n){var i=t.getArea(),r=Rr(i.r0,1),o=Rr(i.r,1),a=new qc({shape:{cx:Rr(t.cx,1),cy:Rr(t.cy,1),r0:r,r:o,startAngle:i.startAngle,endAngle:i.endAngle,clockwise:i.clockwise}});if(e){var s="angle"===t.getBaseAxis().dim;s?a.shape.endAngle=i.startAngle:a.shape.r=r,eh(a,{shape:{endAngle:i.endAngle,r:o}},n)}return a}function _k(t,e,n,i,r){return t?"polar"===t.type?vk(t,e,n):"cartesian2d"===t.type?yk(t,e,n,i,r):null:null}function bk(t,e){return t.type===e}function xk(t,e){if(t.length===e.length){for(var n=0;n<t.length;n++)if(t[n]!==e[n])return;return!0}}function wk(t){for(var e=1/0,n=1/0,i=-1/0,r=-1/0,o=0;o<t.length;){var a=t[o++],s=t[o++];isNaN(a)||(e=Math.min(a,e),i=Math.max(a,i)),isNaN(s)||(n=Math.min(s,n),r=Math.max(s,r))}return[[e,n],[i,r]]}function Mk(t,e){var n=wk(t),i=n[0],r=n[1],o=wk(e),a=o[0],s=o[1];return Math.max(Math.abs(i[0]-a[0]),Math.abs(i[1]-a[1]),Math.abs(r[0]-s[0]),Math.abs(r[1]-s[1]))}function Sk(t){return"number"===typeof t?t:t?.5:0}function Lk(t,e,n){if(!n.valueDim)return[];for(var i=e.count(),r=ak(2*i),o=0;o<i;o++){var a=ik(n,t,e,o);r[2*o]=a[0],r[2*o+1]=a[1]}return r}function kk(t,e,n){for(var i=e.getBaseAxis(),r="x"===i.dim||"radius"===i.dim?0:1,o=[],a=0,s=[],l=[],u=[];a<t.length-2;a+=2)switch(u[0]=t[a+2],u[1]=t[a+3],l[0]=t[a],l[1]=t[a+1],o.push(l[0],l[1]),n){case"end":s[r]=u[r],s[1-r]=l[1-r],o.push(s[0],s[1]);break;case"middle":var c=(l[r]+u[r])/2,d=[];s[r]=d[r]=c,s[1-r]=l[1-r],d[1-r]=u[1-r],o.push(s[0],s[1]),o.push(d[0],d[1]);break;default:s[r]=l[r],s[1-r]=u[1-r],o.push(s[0],s[1])}return o.push(t[a++],t[a++]),o}function Tk(t,e){var n=t.getVisual("visualMeta");if(n&&n.length&&t.count()&&"cartesian2d"===e.type){for(var i,r,o=n.length-1;o>=0;o--){var a=n[o].dimension,s=t.dimensions[a],l=t.getDimensionInfo(s);if(i=l&&l.coordDim,"x"===i||"y"===i){r=n[o];break}}if(r){var u=e.getAxis(i),c=H(r.stops,(function(t){return{offset:0,coord:u.toGlobalCoord(u.dataToCoord(t.value,!0)),color:t.color}})),d=c.length,h=r.outerColors.slice();d&&c[0].coord>c[d-1].coord&&(c.reverse(),h.reverse());var f=10,p=c[0].coord-f,m=c[d-1].coord+f,g=m-p;if(g<.001)return"transparent";R(c,(function(t){t.offset=(t.coord-p)/g})),c.push({offset:d?c[d-1].offset:.5,color:h[1]||"transparent"}),c.unshift({offset:d?c[0].offset:.5,color:h[0]||"transparent"});var y=new Ld(0,0,0,0,c,!0);return y[i]=p,y[i+"2"]=m,y}}}function Dk(t,e,n){var i=t.get("showAllSymbol"),r="auto"===i;if(!i||r){var o=n.getAxesByScale("ordinal")[0];if(o&&(!r||!Ck(o,e))){var a=e.mapDimension(o.dim),s={};return R(o.getViewLabels(),(function(t){var e=o.scale.getRawOrdinalNumber(t.tickValue);s[e]=1})),function(t){return!s.hasOwnProperty(e.get(a,t))}}}}function Ck(t,e){var n=t.getExtent(),i=Math.abs(n[1]-n[0])/t.scale.count();isNaN(i)&&(i=0);for(var r=e.count(),o=Math.max(1,Math.round(r/5)),a=0;a<r;a+=o)if(1.5*XL.getSymbolSize(e,a)[t.isHorizontal()?1:0]>i)return!1;return!0}function Ik(t,e){return isNaN(t)||isNaN(e)}function Ak(t){for(var e=t.length/2;e>0;e--)if(!Ik(t[2*e-2],t[2*e-1]))break;return e-1}function Yk(t,e){return[t[2*e],t[2*e+1]]}function Ok(t,e,n){for(var i,r,o=t.length/2,a="x"===n?0:1,s=0,l=-1,u=0;u<o;u++)if(r=t[2*u+a],!isNaN(r)&&!isNaN(t[2*u+1-a]))if(0!==u){if(i<=e&&r>=e||i>=e&&r<=e){l=u;break}s=u,i=r}else i=r;return{range:[s,l],t:(e-i)/(r-i)}}function Pk(t,e,n,i){if(bk(e,"cartesian2d")){var r=i.getModel("endLabel"),o=r.get("show"),a=r.get("valueAnimation"),s=i.getData(),l={lastFrameIndex:0},u=o?function(n,i){t._endLabelOnDuring(n,i,s,l,a,r,e)}:null,c=e.getBaseAxis().isHorizontal(),d=yk(e,n,i,(function(){var e=t._endLabel;e&&n&&null!=l.originalX&&e.attr({x:l.originalX,y:l.originalY})}),u);if(!i.get("clip",!0)){var h=d.shape,f=Math.max(h.width,h.height);c?(h.y-=f,h.height+=2*f):(h.x-=f,h.width+=2*f)}return u&&u(1,d),d}return vk(e,n,i)}function Ek(t,e){var n=e.getBaseAxis(),i=n.isHorizontal(),r=n.inverse,o=i?r?"right":"left":"center",a=i?"middle":r?"top":"bottom";return{normal:{align:t.get("align")||o,verticalAlign:t.get("verticalAlign")||a}}}var Nk=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.init=function(){var t=new Mr,e=new tk;this.group.add(e.group),this._symbolDraw=e,this._lineGroup=t},e.prototype.render=function(t,e,n){var i=this,r=t.coordinateSystem,o=this.group,a=t.getData(),s=t.getModel("lineStyle"),l=t.getModel("areaStyle"),u=a.getLayout("points")||[],c="polar"===r.type,d=this._coordSys,h=this._symbolDraw,f=this._polyline,p=this._polygon,m=this._lineGroup,g=t.get("animation"),y=!l.isEmpty(),v=l.get("origin"),_=ek(r,a,v),b=y&&Lk(r,a,_),x=t.get("showSymbol"),w=x&&!c&&Dk(t,a,r),M=this._data;M&&M.eachItemGraphicEl((function(t,e){t.__temp&&(o.remove(t),M.setItemGraphicEl(e,null))})),x||h.remove(),o.add(m);var S,L=!c&&t.get("step");r&&r.getArea&&t.get("clip",!0)&&(S=r.getArea(),null!=S.width?(S.x-=.1,S.y-=.1,S.width+=.2,S.height+=.2):S.r0&&(S.r0-=.5,S.r+=.5)),this._clipShapeForSymbol=S;var k=Tk(a,r)||a.getVisual("style")[a.getVisual("drawType")];f&&d.type===r.type&&L===this._step?(y&&!p?p=this._newPolygon(u,b):p&&!y&&(m.remove(p),p=this._polygon=null),c||this._initOrUpdateEndLabel(t,r,ip(k)),m.setClipPath(Pk(this,r,!1,t)),x&&h.updateData(a,{isIgnore:w,clipShape:S,disableAnimation:!0,getSymbolPoint:function(t){return[u[2*t],u[2*t+1]]}}),xk(this._stackedOnPoints,b)&&xk(this._points,u)||(g?this._doUpdateAnimation(a,b,r,n,L,v):(L&&(u=kk(u,r,L),b&&(b=kk(b,r,L))),f.setShape({points:u}),p&&p.setShape({points:u,stackedOnPoints:b})))):(x&&h.updateData(a,{isIgnore:w,clipShape:S,disableAnimation:!0,getSymbolPoint:function(t){return[u[2*t],u[2*t+1]]}}),g&&this._initSymbolLabelAnimation(a,r,S),L&&(u=kk(u,r,L),b&&(b=kk(b,r,L))),f=this._newPolyline(u),y&&(p=this._newPolygon(u,b)),c||this._initOrUpdateEndLabel(t,r,ip(k)),m.setClipPath(Pk(this,r,!0,t)));var T=t.get(["emphasis","focus"]),D=t.get(["emphasis","blurScope"]);if(f.useStyle(A(s.getLineStyle(),{fill:"none",stroke:k,lineJoin:"bevel"})),Zu(f,t,"lineStyle"),f.style.lineWidth>0&&"bolder"===t.get(["emphasis","lineStyle","width"])){var C=f.getState("emphasis").style;C.lineWidth=+f.style.lineWidth+1}$l(f).seriesIndex=t.seriesIndex,$u(f,T,D);var I=Sk(t.get("smooth")),Y=t.get("smoothMonotone"),O=t.get("connectNulls");if(f.setShape({smooth:I,smoothMonotone:Y,connectNulls:O}),p){var P=a.getCalculationInfo("stackedOnSeries"),E=0;p.useStyle(A(l.getAreaStyle(),{fill:k,opacity:.7,lineJoin:"bevel",decal:a.getVisual("style").decal})),P&&(E=Sk(P.get("smooth"))),p.setShape({smooth:I,stackedOnSmooth:E,smoothMonotone:Y,connectNulls:O}),Zu(p,t,"areaStyle"),$l(p).seriesIndex=t.seriesIndex,$u(p,T,D)}var N=function(t){i._changePolyState(t)};a.eachItemGraphicEl((function(t){t&&(t.onHoverStateChange=N)})),this._polyline.onHoverStateChange=N,this._data=a,this._coordSys=r,this._stackedOnPoints=b,this._points=u,this._step=L,this._valueOrigin=v},e.prototype.dispose=function(){},e.prototype.highlight=function(t,e,n,i){var r=t.getData(),o=Co(r,i);if(this._changePolyState("emphasis"),!(o instanceof Array)&&null!=o&&o>=0){var a=r.getLayout("points"),s=r.getItemGraphicEl(o);if(!s){var l=a[2*o],u=a[2*o+1];if(isNaN(l)||isNaN(u))return;if(this._clipShapeForSymbol&&!this._clipShapeForSymbol.contain(l,u))return;var c=t.get("zlevel"),d=t.get("z");s=new XL(r,o),s.x=l,s.y=u,s.setZ(c,d);var h=s.getSymbolPath().getTextContent();h&&(h.zlevel=c,h.z=d,h.z2=this._polyline.z2+1),s.__temp=!0,r.setItemGraphicEl(o,s),s.stopSymbolAnimation(!0),this.group.add(s)}s.highlight()}else Vy.prototype.highlight.call(this,t,e,n,i)},e.prototype.downplay=function(t,e,n,i){var r=t.getData(),o=Co(r,i);if(this._changePolyState("normal"),null!=o&&o>=0){var a=r.getItemGraphicEl(o);a&&(a.__temp?(r.setItemGraphicEl(o,null),this.group.remove(a)):a.downplay())}else Vy.prototype.downplay.call(this,t,e,n,i)},e.prototype._changePolyState=function(t){var e=this._polygon;bu(this._polyline,t),e&&bu(e,t)},e.prototype._newPolyline=function(t){var e=this._polyline;return e&&this._lineGroup.remove(e),e=new pk({shape:{points:t},segmentIgnoreThreshold:2,z2:10}),this._lineGroup.add(e),this._polyline=e,e},e.prototype._newPolygon=function(t,e){var n=this._polygon;return n&&this._lineGroup.remove(n),n=new gk({shape:{points:t,stackedOnPoints:e},segmentIgnoreThreshold:2}),this._lineGroup.add(n),this._polygon=n,n},e.prototype._initSymbolLabelAnimation=function(t,e,n){var i,r,o=e.getBaseAxis(),a=o.inverse;"cartesian2d"===e.type?(i=o.isHorizontal(),r=!1):"polar"===e.type&&(i="angle"===o.dim,r=!0);var s=t.hostModel,l=s.get("animationDuration");"function"===typeof l&&(l=l(null));var u=s.get("animationDelay")||0,c="function"===typeof u?u(null):u;t.eachItemGraphicEl((function(t,o){var s=t;if(s){var d=[t.x,t.y],h=void 0,f=void 0,p=void 0;if(n)if(r){var m=n,g=e.pointToCoord(d);i?(h=m.startAngle,f=m.endAngle,p=-g[1]/180*Math.PI):(h=m.r0,f=m.r,p=g[0])}else{var y=n;i?(h=y.x,f=y.x+y.width,p=t.x):(h=y.y+y.height,f=y.y,p=t.y)}var v=f===h?0:(p-h)/(f-h);a&&(v=1-v);var _="function"===typeof u?u(o):l*v+c,b=s.getSymbolPath(),x=b.getTextContent();s.attr({scaleX:0,scaleY:0}),s.animateTo({scaleX:1,scaleY:1},{duration:200,delay:_}),x&&x.animateFrom({style:{opacity:0}},{duration:300,delay:_}),b.disableLabelAnimation=!0}}))},e.prototype._initOrUpdateEndLabel=function(t,e,n){var i=t.getModel("endLabel");if(i.get("show")){var r=t.getData(),o=this._polyline,a=this._endLabel;a||(a=this._endLabel=new Gl({z2:200}),a.ignoreClip=!0,o.setTextContent(this._endLabel),o.disableLabelAnimation=!0);var s=Ak(r.getLayout("points"));s>=0&&(Mh(o,Sh(t,"endLabel"),{inheritColor:n,labelFetcher:t,labelDataIndex:s,defaultText:function(t,e,n){return null!=n?$L(r,n):GL(r,t)},enableTextSetter:!0},Ek(i,e)),o.textConfig.position=null)}else this._endLabel&&(this._polyline.removeTextContent(),this._endLabel=null)},e.prototype._endLabelOnDuring=function(t,e,n,i,r,o,a){var s=this._endLabel,l=this._polyline;if(s){t<1&&null==i.originalX&&(i.originalX=s.x,i.originalY=s.y);var u=n.getLayout("points"),c=n.hostModel,d=c.get("connectNulls"),h=o.get("precision"),f=o.get("distance")||0,p=a.getBaseAxis(),m=p.isHorizontal(),g=p.inverse,y=e.shape,v=g?m?y.x:y.y+y.height:m?y.x+y.width:y.y,_=(m?f:0)*(g?-1:1),b=(m?0:-f)*(g?-1:1),x=m?"x":"y",w=Ok(u,v,x),M=w.range,S=M[1]-M[0],L=void 0;if(S>=1){if(S>1&&!d){var k=Yk(u,M[0]);s.attr({x:k[0]+_,y:k[1]+b}),r&&(L=c.getRawValue(M[0]))}else{k=l.getPointOn(v,x);k&&s.attr({x:k[0]+_,y:k[1]+b});var T=c.getRawValue(M[0]),D=c.getRawValue(M[1]);r&&(L=Wo(n,h,T,D,w.t))}i.lastFrameIndex=M[0]}else{var C=1===t||i.lastFrameIndex>0?M[0]:0;k=Yk(u,C);r&&(L=c.getRawValue(C)),s.attr({x:k[0]+_,y:k[1]+b})}r&&Ph(s).setLabelText(L)}},e.prototype._doUpdateAnimation=function(t,e,n,i,r,o){var a=this._polyline,s=this._polygon,l=t.hostModel,u=lk(this._data,t,this._stackedOnPoints,e,this._coordSys,n,this._valueOrigin,o),c=u.current,d=u.stackedOnCurrent,h=u.next,f=u.stackedOnNext;if(r&&(c=kk(u.current,n,r),d=kk(u.stackedOnCurrent,n,r),h=kk(u.next,n,r),f=kk(u.stackedOnNext,n,r)),Mk(c,h)>3e3||s&&Mk(d,f)>3e3)return a.setShape({points:h}),void(s&&s.setShape({points:h,stackedOnPoints:f}));a.shape.__points=u.current,a.shape.points=c;var p={shape:{points:h}};u.current!==c&&(p.shape.__points=u.next),a.stopAnimation(),th(a,p,l),s&&(s.setShape({points:c,stackedOnPoints:d}),s.stopAnimation(),th(s,{shape:{stackedOnPoints:f}},l),a.shape.points!==s.shape.points&&(s.shape.points=a.shape.points));for(var m=[],g=u.status,y=0;y<g.length;y++){var v=g[y].cmd;if("="===v){var _=t.getItemGraphicEl(g[y].idx1);_&&m.push({el:_,ptIdx:y})}}a.animators&&a.animators.length&&a.animators[0].during((function(){s&&s.dirtyShape();for(var t=a.shape.__points,e=0;e<m.length;e++){var n=m[e].el,i=2*m[e].ptIdx;n.x=t[i],n.y=t[i+1],n.markRedraw()}}))},e.prototype.remove=function(t){var e=this.group,n=this._data;this._lineGroup.removeAll(),this._symbolDraw.remove(!0),n&&n.eachItemGraphicEl((function(t,i){t.__temp&&(e.remove(t),n.setItemGraphicEl(i,null))})),this._polyline=this._polygon=this._coordSys=this._points=this._stackedOnPoints=this._endLabel=this._data=null},e.type="line",e}(Vy),Rk=Nk;function Hk(t,e){return{seriesType:t,plan:Ey(),reset:function(t){var n=t.getData(),i=t.coordinateSystem,r=t.pipelineContext,o=e||r.large;if(i){var a=H(i.dimensions,(function(t){return n.mapDimension(t)})).slice(0,2),s=a.length,l=n.getCalculationInfo("stackResultDimension");RL(n,a[0])&&(a[0]=l),RL(n,a[1])&&(a[1]=l);var u=n.getDimensionInfo(a[0]),c=n.getDimensionInfo(a[1]),d=u&&u.index,h=c&&c.index;return s&&{progress:function(t,e){for(var n=t.end-t.start,r=o&&ak(n*s),a=[],l=[],u=t.start,c=0;u<t.end;u++){var f=void 0;if(1===s){var p=e.getByDimIdx(d,u);f=i.dataToPoint(p,null,l)}else a[0]=e.getByDimIdx(d,u),a[1]=e.getByDimIdx(h,u),f=i.dataToPoint(a,null,l);o?(r[c++]=f[0],r[c++]=f[1]):e.setItemLayout(u,f.slice())}o&&e.setLayout("points",r)}}}}}}var jk={average:function(t){for(var e=0,n=0,i=0;i<t.length;i++)isNaN(t[i])||(e+=t[i],n++);return 0===n?NaN:e/n},sum:function(t){for(var e=0,n=0;n<t.length;n++)e+=t[n]||0;return e},max:function(t){for(var e=-1/0,n=0;n<t.length;n++)t[n]>e&&(e=t[n]);return isFinite(e)?e:NaN},min:function(t){for(var e=1/0,n=0;n<t.length;n++)t[n]<e&&(e=t[n]);return isFinite(e)?e:NaN},nearest:function(t){return t[0]}},zk=function(t){return Math.round(t.length/2)};function Wk(t){return{seriesType:t,reset:function(t,e,n){var i=t.getData(),r=t.get("sampling"),o=t.coordinateSystem,a=i.count();if(a>10&&"cartesian2d"===o.type&&r){var s=o.getBaseAxis(),l=o.getOtherAxis(s),u=s.getExtent(),c=n.getDevicePixelRatio(),d=Math.abs(u[1]-u[0])*(c||1),h=Math.round(a/d);if(h>1){"lttb"===r&&t.setData(i.lttbDownSample(i.mapDimension(l.dim),1/h));var f=void 0;"string"===typeof r?f=jk[r]:"function"===typeof r&&(f=r),f&&t.setData(i.downSample(i.mapDimension(l.dim),1/h,f,zk))}}}}}function Bk(t){t.registerChartView(Rk),t.registerSeriesModel(VL),t.registerLayout(Hk("line",!0)),t.registerVisual({seriesType:"line",reset:function(t){var e=t.getData(),n=t.getModel("lineStyle").getLineStyle();n&&!n.stroke&&(n.stroke=e.getVisual("style").fill),e.setVisual("legendLineStyle",n)}}),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,Wk("line"))}var Fk="__ec_stack_",Vk=.5,Gk="undefined"!==typeof Float32Array?Float32Array:Array;function $k(t){return t.get("stack")||Fk+t.seriesIndex}function Uk(t){return t.dim+t.index}function qk(t){var e=[],n=t.axis,i="axis0";if("category"===n.type){for(var r=n.getBandWidth(),o=0;o<t.count;o++)e.push(A({bandWidth:r,axisKey:i,stackId:Fk+o},t));var a=Kk(e),s=[];for(o=0;o<t.count;o++){var l=a[i][Fk+o];l.offsetCenter=l.offset+l.width/2,s.push(l)}return s}}function Xk(t,e){var n=[];return e.eachSeriesByType(t,(function(t){nT(t)&&!iT(t)&&n.push(t)})),n}function Zk(t){var e={};R(t,(function(t){var n=t.coordinateSystem,i=n.getBaseAxis();if("time"===i.type||"value"===i.type)for(var r=t.getData(),o=i.dim+"_"+i.index,a=r.mapDimension(i.dim),s=0,l=r.count();s<l;++s){var u=r.get(a,s);e[o]?e[o].push(u):e[o]=[u]}}));var n={};for(var i in e)if(e.hasOwnProperty(i)){var r=e[i];if(r){r.sort((function(t,e){return t-e}));for(var o=null,a=1;a<r.length;++a){var s=r[a]-r[a-1];s>0&&(o=null===o?s:Math.min(o,s))}n[i]=o}}return n}function Jk(t){var e=Zk(t),n=[];return R(t,(function(t){var i,r=t.coordinateSystem,o=r.getBaseAxis(),a=o.getExtent();if("category"===o.type)i=o.getBandWidth();else if("value"===o.type||"time"===o.type){var s=o.dim+"_"+o.index,l=e[s],u=Math.abs(a[1]-a[0]),c=o.scale.getExtent(),d=Math.abs(c[1]-c[0]);i=l?u/d*l:u}else{var h=t.getData();i=Math.abs(a[1]-a[0])/h.count()}var f=Nr(t.get("barWidth"),i),p=Nr(t.get("barMaxWidth"),i),m=Nr(t.get("barMinWidth")||1,i),g=t.get("barGap"),y=t.get("barCategoryGap");n.push({bandWidth:i,barWidth:f,barMaxWidth:p,barMinWidth:m,barGap:g,barCategoryGap:y,axisKey:Uk(o),stackId:$k(t)})})),Kk(n)}function Kk(t){var e={};R(t,(function(t,n){var i=t.axisKey,r=t.bandWidth,o=e[i]||{bandWidth:r,remainedWidth:r,autoWidthCount:0,categoryGap:null,gap:"20%",stacks:{}},a=o.stacks;e[i]=o;var s=t.stackId;a[s]||o.autoWidthCount++,a[s]=a[s]||{width:0,maxWidth:0};var l=t.barWidth;l&&!a[s].width&&(a[s].width=l,l=Math.min(o.remainedWidth,l),o.remainedWidth-=l);var u=t.barMaxWidth;u&&(a[s].maxWidth=u);var c=t.barMinWidth;c&&(a[s].minWidth=c);var d=t.barGap;null!=d&&(o.gap=d);var h=t.barCategoryGap;null!=h&&(o.categoryGap=h)}));var n={};return R(e,(function(t,e){n[e]={};var i=t.stacks,r=t.bandWidth,o=t.categoryGap;if(null==o){var a=B(i).length;o=Math.max(35-4*a,15)+"%"}var s=Nr(o,r),l=Nr(t.gap,1),u=t.remainedWidth,c=t.autoWidthCount,d=(u-s)/(c+(c-1)*l);d=Math.max(d,0),R(i,(function(t){var e=t.maxWidth,n=t.minWidth;if(t.width){i=t.width;e&&(i=Math.min(i,e)),n&&(i=Math.max(i,n)),t.width=i,u-=i+l*i,c--}else{var i=d;e&&e<i&&(i=Math.min(e,u)),n&&n>i&&(i=n),i!==d&&(t.width=i,u-=i+l*i,c--)}})),d=(u-s)/(c+(c-1)*l),d=Math.max(d,0);var h,f=0;R(i,(function(t,e){t.width||(t.width=d),h=t,f+=t.width*(1+l)})),h&&(f-=h.width*l);var p=-f/2;R(i,(function(t,i){n[e][i]=n[e][i]||{bandWidth:r,offset:p,width:t.width},p+=t.width*(1+l)}))})),n}function Qk(t,e,n){if(t&&e){var i=t[Uk(e)];return null!=i&&null!=n?i[$k(n)]:i}}function tT(t,e){var n=Xk(t,e),i=Jk(n),r={};R(n,(function(t){var e=t.getData(),n=t.coordinateSystem,o=n.getBaseAxis(),a=$k(t),s=i[Uk(o)][a],l=s.offset,u=s.width,c=n.getOtherAxis(o),d=t.get("barMinHeight")||0;r[a]=r[a]||[],e.setLayout({bandWidth:s.bandWidth,offset:l,size:u});for(var h=e.mapDimension(c.dim),f=e.mapDimension(o.dim),p=RL(e,h),m=c.isHorizontal(),g=rT(o,c,p),y=0,v=e.count();y<v;y++){var _=e.get(h,y),b=e.get(f,y),x=_>=0?"p":"n",w=g;p&&(r[a][b]||(r[a][b]={p:g,n:g}),w=r[a][b][x]);var M=void 0,S=void 0,L=void 0,k=void 0;if(m){var T=n.dataToPoint([_,b]);M=w,S=T[1]+l,L=T[0]-g,k=u,Math.abs(L)<d&&(L=(L<0?-1:1)*d),isNaN(L)||p&&(r[a][b][x]+=L)}else{T=n.dataToPoint([b,_]);M=T[0]+l,S=w,L=u,k=T[1]-g,Math.abs(k)<d&&(k=(k<=0?-1:1)*d),isNaN(k)||p&&(r[a][b][x]+=k)}e.setItemLayout(y,{x:M,y:S,width:L,height:k})}}))}var eT={seriesType:"bar",plan:Ey(),reset:function(t){if(nT(t)&&iT(t)){var e=t.getData(),n=t.coordinateSystem,i=n.master.getRect(),r=n.getBaseAxis(),o=n.getOtherAxis(r),a=e.mapDimension(o.dim),s=e.mapDimension(r.dim),l=o.isHorizontal(),u=l?0:1,c=Qk(Jk([t]),r,t).width;return c>Vk||(c=Vk),{progress:function(t,e){var d,h=t.count,f=new Gk(2*h),p=new Gk(2*h),m=new Gk(h),g=[],y=[],v=0,_=0;while(null!=(d=t.next()))y[u]=e.get(a,d),y[1-u]=e.get(s,d),g=n.dataToPoint(y,null),p[v]=l?i.x+i.width:g[0],f[v++]=g[0],p[v]=l?g[1]:i.y+i.height,f[v++]=g[1],m[_++]=d;e.setLayout({largePoints:f,largeDataIndices:m,largeBackgroundPoints:p,barWidth:c,valueAxisStart:rT(r,o,!1),backgroundStart:l?i.x:i.y,valueAxisHorizontal:l})}}}}};function nT(t){return t.coordinateSystem&&"cartesian2d"===t.coordinateSystem.type}function iT(t){return t.pipelineContext&&t.pipelineContext.large}function rT(t,e,n){return e.toGlobalCoord(e.dataToCoord("log"===e.type?1:0))}var oT=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(t,e){return BL(this.getSource(),this,{useEncodeDefaulter:!0})},e.prototype.getMarkerPosition=function(t){var e=this.coordinateSystem;if(e){var n=e.dataToPoint(e.clampData(t)),i=this.getData(),r=i.getLayout("offset"),o=i.getLayout("size"),a=e.getBaseAxis().isHorizontal()?0:1;return n[a]+=r+o/2,n}return[NaN,NaN]},e.type="series.__base_bar__",e.defaultOption={zlevel:0,z:2,coordinateSystem:"cartesian2d",legendHoverLink:!0,barMinHeight:0,barMinAngle:0,large:!1,largeThreshold:400,progressive:3e3,progressiveChunkMode:"mod"},e}(Yy);Yy.registerClass(oT);var aT=oT,sT=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(){return BL(this.getSource(),this,{useEncodeDefaulter:!0,createInvertedIndices:!!this.get("realtimeSort",!0)||null})},e.prototype.getProgressive=function(){return!!this.get("large")&&this.get("progressive")},e.prototype.getProgressiveThreshold=function(){var t=this.get("progressiveThreshold"),e=this.get("largeThreshold");return e>t&&(t=e),t},e.prototype.brushSelector=function(t,e,n){return n.rect(e.getItemLayout(t))},e.type="series.bar",e.dependencies=["grid","polar"],e.defaultOption=Qh(aT.defaultOption,{clip:!0,roundCap:!1,showBackground:!1,backgroundStyle:{color:"rgba(180, 180, 180, 0.2)",borderColor:null,borderWidth:0,borderType:"solid",borderRadius:0,shadowBlur:0,shadowColor:null,shadowOffsetX:0,shadowOffsetY:0,opacity:1},select:{itemStyle:{borderColor:"#212121"}},realtimeSort:!1}),e}(aT),lT=sT,uT=function(){function t(){this.cx=0,this.cy=0,this.r0=0,this.r=0,this.startAngle=0,this.endAngle=2*Math.PI,this.clockwise=!0}return t}(),cT=function(t){function e(e){var n=t.call(this,e)||this;return n.type="sausage",n}return o(e,t),e.prototype.getDefaultShape=function(){return new uT},e.prototype.buildPath=function(t,e){var n=e.cx,i=e.cy,r=Math.max(e.r0||0,0),o=Math.max(e.r,0),a=.5*(o-r),s=r+a,l=e.startAngle,u=e.endAngle,c=e.clockwise,d=Math.cos(l),h=Math.sin(l),f=Math.cos(u),p=Math.sin(u),m=c?u-l<2*Math.PI:l-u<2*Math.PI;m&&(t.moveTo(d*r+n,h*r+i),t.arc(d*s+n,h*s+i,a,-Math.PI+l,l,!c)),t.arc(n,i,o,l,u,!c),t.moveTo(f*o+n,p*o+i),t.arc(f*s+n,p*s+i,a,u-2*Math.PI,u-Math.PI,!c),0!==r&&(t.arc(n,i,r,u,l,c),t.moveTo(d*r+n,p*r+i)),t.closePath()},e}(pl),dT=cT,hT=[0,0],fT=Math.max,pT=Math.min;function mT(t,e){var n=t.getArea&&t.getArea();if(bk(t,"cartesian2d")){var i=t.getBaseAxis();if("category"!==i.type||!i.onBand){var r=e.getLayout("bandWidth");i.isHorizontal()?(n.x-=r,n.width+=2*r):(n.y-=r,n.height+=2*r)}}return n}var gT=function(t){function e(){var n=t.call(this)||this;return n.type=e.type,n._isFirstFrame=!0,n}return o(e,t),e.prototype.render=function(t,e,n,i){this._model=t,this._removeOnRenderedListener(n),this._updateDrawMode(t);var r=t.get("coordinateSystem");("cartesian2d"===r||"polar"===r)&&(this._isLargeDraw?this._renderLarge(t,e,n):this._renderNormal(t,e,n,i))},e.prototype.incrementalPrepareRender=function(t){this._clear(),this._updateDrawMode(t),this._updateLargeClip(t)},e.prototype.incrementalRender=function(t,e){this._incrementalRenderLarge(t,e)},e.prototype._updateDrawMode=function(t){var e=t.pipelineContext.large;null!=this._isLargeDraw&&e===this._isLargeDraw||(this._isLargeDraw=e,this._clear())},e.prototype._renderNormal=function(t,e,n,i){var r,o=this.group,a=t.getData(),s=this._data,l=t.coordinateSystem,u=l.getBaseAxis();"cartesian2d"===l.type?r=u.isHorizontal():"polar"===l.type&&(r="angle"===u.dim);var c=t.isAnimationEnabled()?t:null,d=_T(t,l);d&&this._enableRealtimeSort(d,a,n);var h=t.get("clip",!0)||d,f=mT(l,a);o.removeClipPath();var p=t.get("roundCap",!0),m=t.get("showBackground",!0),g=t.getModel("backgroundStyle"),y=g.get("borderRadius")||0,v=[],_=this._backgroundEls,b=i&&i.isInitSort,x=i&&"changeAxisOrder"===i.type;function w(t){var e=xT[l.type](a,t),n=OT(l,r,e);return n.useStyle(g.getItemStyle()),"cartesian2d"===l.type&&n.setShape("r",y),v[t]=n,n}a.diff(s).add((function(e){var n=a.getItemModel(e),i=xT[l.type](a,e,n);if(m&&w(e),a.hasValue(e)){var s=!1;h&&(s=yT[l.type](f,i));var g=vT[l.type](t,a,e,i,r,c,u.model,!1,p);MT(g,a,e,n,i,t,r,"polar"===l.type),b?g.attr({shape:i}):d?bT(d,c,g,i,e,r,!1,!1):eh(g,{shape:i},t,e),a.setItemGraphicEl(e,g),o.add(g),g.ignore=s}})).update((function(e,n){var i=a.getItemModel(e),M=xT[l.type](a,e,i);if(m){var S=void 0;0===_.length?S=w(n):(S=_[n],S.useStyle(g.getItemStyle()),"cartesian2d"===l.type&&S.setShape("r",y),v[e]=S);var L=xT[l.type](a,e),k=YT(r,L,l);th(S,{shape:k},c,e)}var T=s.getItemGraphicEl(n);if(!a.hasValue(e))return o.remove(T),void(T=null);var D=!1;h&&(D=yT[l.type](f,M),D&&o.remove(T)),T||(T=vT[l.type](t,a,e,M,r,c,u.model,!!T,p)),x||MT(T,a,e,i,M,t,r,"polar"===l.type),b?T.attr({shape:M}):d?bT(d,c,T,M,e,r,!0,x):th(T,{shape:M},t,e,null),a.setItemGraphicEl(e,T),T.ignore=D,o.add(T)})).remove((function(e){var n=s.getItemGraphicEl(e);n&&rh(n,t,e)})).execute();var M=this._backgroundGroup||(this._backgroundGroup=new Mr);M.removeAll();for(var S=0;S<v.length;++S)M.add(v[S]);o.add(M),this._backgroundEls=v,this._data=a},e.prototype._renderLarge=function(t,e,n){this._clear(),TT(t,this.group),this._updateLargeClip(t)},e.prototype._incrementalRenderLarge=function(t,e){this._removeBackground(),TT(e,this.group,!0)},e.prototype._updateLargeClip=function(t){var e=t.get("clip",!0)?_k(t.coordinateSystem,!1,t):null;e?this.group.setClipPath(e):this.group.removeClipPath()},e.prototype._enableRealtimeSort=function(t,e,n){var i=this;if(e.count()){var r=t.baseAxis;if(this._isFirstFrame)this._dispatchInitSort(e,t,n),this._isFirstFrame=!1;else{var o=function(t){var n=e.getItemGraphicEl(t);if(n){var i=n.shape;return(r.isHorizontal()?Math.abs(i.height):Math.abs(i.width))||0}return 0};this._onRendered=function(){i._updateSortWithinSameData(e,o,r,n)},n.getZr().on("rendered",this._onRendered)}}},e.prototype._dataSort=function(t,e,n){var i=[];return t.each(t.mapDimension(e.dim),(function(t,e){var r=n(e);r=null==r?NaN:r,i.push({dataIndex:e,mappedValue:r,ordinalNumber:t})})),i.sort((function(t,e){return e.mappedValue-t.mappedValue})),{ordinalNumbers:H(i,(function(t){return t.ordinalNumber}))}},e.prototype._isOrderChangedWithinSameData=function(t,e,n){for(var i=n.scale,r=t.mapDimension(n.dim),o=Number.MAX_VALUE,a=0,s=i.getOrdinalMeta().categories.length;a<s;++a){var l=t.rawIndexOf(r,i.getRawOrdinalNumber(a)),u=l<0?Number.MIN_VALUE:e(t.indexOfRawIndex(l));if(u>o)return!0;o=u}return!1},e.prototype._isOrderDifferentInView=function(t,e){for(var n=e.scale,i=n.getExtent(),r=Math.max(0,i[0]),o=Math.min(i[1],n.getOrdinalMeta().categories.length-1);r<=o;++r)if(t.ordinalNumbers[r]!==n.getRawOrdinalNumber(r))return!0},e.prototype._updateSortWithinSameData=function(t,e,n,i){if(this._isOrderChangedWithinSameData(t,e,n)){var r=this._dataSort(t,n,e);this._isOrderDifferentInView(r,n)&&(this._removeOnRenderedListener(i),i.dispatchAction({type:"changeAxisOrder",componentType:n.dim+"Axis",axisId:n.index,sortInfo:r}))}},e.prototype._dispatchInitSort=function(t,e,n){var i=e.baseAxis,r=this._dataSort(t,i,(function(n){return t.get(t.mapDimension(e.otherAxis.dim),n)}));n.dispatchAction({type:"changeAxisOrder",componentType:i.dim+"Axis",isInitSort:!0,axisId:i.index,sortInfo:r,animation:{duration:0}})},e.prototype.remove=function(t,e){this._clear(this._model),this._removeOnRenderedListener(e)},e.prototype.dispose=function(t,e){this._removeOnRenderedListener(e)},e.prototype._removeOnRenderedListener=function(t){this._onRendered&&(t.getZr().off("rendered",this._onRendered),this._onRendered=null)},e.prototype._clear=function(t){var e=this.group,n=this._data;t&&t.isAnimationEnabled()&&n&&!this._isLargeDraw?(this._removeBackground(),this._backgroundEls=[],n.eachItemGraphicEl((function(e){rh(e,t,$l(e).dataIndex)}))):e.removeAll(),this._data=null,this._isFirstFrame=!0},e.prototype._removeBackground=function(){this.group.remove(this._backgroundGroup),this._backgroundGroup=null},e.type="bar",e}(Vy),yT={cartesian2d:function(t,e){var n=e.width<0?-1:1,i=e.height<0?-1:1;n<0&&(e.x+=e.width,e.width=-e.width),i<0&&(e.y+=e.height,e.height=-e.height);var r=t.x+t.width,o=t.y+t.height,a=fT(e.x,t.x),s=pT(e.x+e.width,r),l=fT(e.y,t.y),u=pT(e.y+e.height,o),c=s<a,d=u<l;return e.x=c&&a>r?s:a,e.y=d&&l>o?u:l,e.width=c?0:s-a,e.height=d?0:u-l,n<0&&(e.x+=e.width,e.width=-e.width),i<0&&(e.y+=e.height,e.height=-e.height),c||d},polar:function(t,e){var n=e.r0<=e.r?1:-1;if(n<0){var i=e.r;e.r=e.r0,e.r0=i}var r=pT(e.r,t.r),o=fT(e.r0,t.r0);e.r=r,e.r0=o;var a=r-o<0;if(n<0){i=e.r;e.r=e.r0,e.r0=i}return a}},vT={cartesian2d:function(t,e,n,i,r,o,a,s,l){var u=new Al({shape:I({},i),z2:1});if(u.__dataIndex=n,u.name="item",o){var c=u.shape,d=r?"height":"width";c[d]=0}return u},polar:function(t,e,n,i,r,o,a,s,l){var u=i.startAngle<i.endAngle,c=!r&&l?dT:qc,d=new c({shape:A({clockwise:u},i),z2:1});if(d.name="item",o){var h=d.shape,f=r?"r":"endAngle",p={};h[f]=r?0:i.startAngle,p[f]=i[f],(s?th:eh)(d,{shape:p},o)}return d}};function _T(t,e){var n=t.get("realtimeSort",!0),i=e.getBaseAxis();if(n&&"category"===i.type&&"cartesian2d"===e.type)return{baseAxis:i,otherAxis:e.getOtherAxis(i)}}function bT(t,e,n,i,r,o,a,s){var l,u;o?(u={x:i.x,width:i.width},l={y:i.y,height:i.height}):(u={y:i.y,height:i.height},l={x:i.x,width:i.width}),s||(a?th:eh)(n,{shape:l},e,r,null);var c=e?t.baseAxis.model:null;(a?th:eh)(n,{shape:u},c,r)}var xT={cartesian2d:function(t,e,n){var i=t.getItemLayout(e),r=n?ST(n,i):0,o=i.width>0?1:-1,a=i.height>0?1:-1;return{x:i.x+o*r/2,y:i.y+a*r/2,width:i.width-o*r,height:i.height-a*r}},polar:function(t,e,n){var i=t.getItemLayout(e);return{cx:i.cx,cy:i.cy,r0:i.r0,r:i.r,startAngle:i.startAngle,endAngle:i.endAngle}}};function wT(t){return null!=t.startAngle&&null!=t.endAngle&&t.startAngle===t.endAngle}function MT(t,e,n,i,r,o,a,s){var l=e.getItemVisual(n,"style");s||t.setShape("r",i.get(["itemStyle","borderRadius"])||0),t.useStyle(l);var u=i.getShallow("cursor");if(u&&t.attr("cursor",u),!s){var c=a?r.height>0?"bottom":"top":r.width>0?"left":"right",d=Sh(i);Mh(t,d,{labelFetcher:o,labelDataIndex:n,defaultText:GL(o.getData(),n),inheritColor:l.fill,defaultOpacity:l.opacity,defaultOutsidePosition:c});var h=t.getTextContent();Eh(h,d,o.getRawValue(n),(function(t){return $L(e,t)}))}var f=i.getModel(["emphasis"]);$u(t,f.get("focus"),f.get("blurScope")),Zu(t,i),wT(r)&&(t.style.fill="none",t.style.stroke="none",R(t.states,(function(t){t.style&&(t.style.fill=t.style.stroke="none")})))}function ST(t,e){var n=t.get(["itemStyle","borderColor"]);if(!n||"none"===n)return 0;var i=t.get(["itemStyle","borderWidth"])||0,r=isNaN(e.width)?Number.MAX_VALUE:Math.abs(e.width),o=isNaN(e.height)?Number.MAX_VALUE:Math.abs(e.height);return Math.min(i,r,o)}var LT=function(){function t(){}return t}(),kT=function(t){function e(e){var n=t.call(this,e)||this;return n.type="largeBar",n}return o(e,t),e.prototype.getDefaultShape=function(){return new LT},e.prototype.buildPath=function(t,e){for(var n=e.points,i=this.__startPoint,r=this.__baseDimIdx,o=0;o<n.length;o+=2)i[r]=n[o+r],t.moveTo(i[0],i[1]),t.lineTo(n[o],n[o+1])},e}(pl);function TT(t,e,n){var i=t.getData(),r=[],o=i.getLayout("valueAxisHorizontal")?1:0;r[1-o]=i.getLayout("valueAxisStart");var a=i.getLayout("largeDataIndices"),s=i.getLayout("barWidth"),l=t.getModel("backgroundStyle"),u=t.get("showBackground",!0);if(u){var c=i.getLayout("largeBackgroundPoints"),d=[];d[1-o]=i.getLayout("backgroundStart");var h=new kT({shape:{points:c},incremental:!!n,silent:!0,z2:0});h.__startPoint=d,h.__baseDimIdx=o,h.__largeDataIndices=a,h.__barWidth=s,AT(h,l,i),e.add(h)}var f=new kT({shape:{points:i.getLayout("largePoints")},incremental:!!n});f.__startPoint=r,f.__baseDimIdx=o,f.__largeDataIndices=a,f.__barWidth=s,e.add(f),IT(f,t,i),$l(f).seriesIndex=t.seriesIndex,t.get("silent")||(f.on("mousedown",DT),f.on("mousemove",DT))}var DT=qy((function(t){var e=this,n=CT(e,t.offsetX,t.offsetY);$l(e).dataIndex=n>=0?n:null}),30,!1);function CT(t,e,n){var i=t.__baseDimIdx,r=1-i,o=t.shape.points,a=t.__largeDataIndices,s=Math.abs(t.__barWidth/2),l=t.__startPoint[r];hT[0]=e,hT[1]=n;for(var u=hT[i],c=hT[1-i],d=u-s,h=u+s,f=0,p=o.length/2;f<p;f++){var m=2*f,g=o[m+i],y=o[m+r];if(g>=d&&g<=h&&(l<=y?c>=l&&c<=y:c>=y&&c<=l))return a[f]}return-1}function IT(t,e,n){var i=n.getVisual("style");t.useStyle(I({},i)),t.style.fill=null,t.style.stroke=i.fill,t.style.lineWidth=n.getLayout("barWidth")}function AT(t,e,n){var i=e.get("borderColor")||e.get("color"),r=e.getItemStyle();t.useStyle(r),t.style.fill=null,t.style.stroke=i,t.style.lineWidth=n.getLayout("barWidth")}function YT(t,e,n){if(bk(n,"cartesian2d")){var i=e,r=n.getArea();return{x:t?i.x:r.x,y:t?r.y:i.y,width:t?i.width:r.width,height:t?r.height:i.height}}r=n.getArea();var o=e;return{cx:r.cx,cy:r.cy,r0:t?r.r0:o.r0,r:t?r.r:o.r,startAngle:t?o.startAngle:0,endAngle:t?o.endAngle:2*Math.PI}}function OT(t,e,n){var i="polar"===t.type?qc:Al;return new i({shape:YT(e,n,t),silent:!0,z2:0})}var PT=gT;function ET(t){t.registerChartView(PT),t.registerSeriesModel(lT),t.registerLayout(t.PRIORITY.VISUAL.LAYOUT,G(tT,"bar")),t.registerLayout(t.PRIORITY.VISUAL.PROGRESSIVE_LAYOUT,eT),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,Wk("bar")),t.registerAction({type:"changeAxisOrder",event:"changeAxisOrder",update:"update"},(function(t,e){var n=t.componentType||"series";e.eachComponent({mainType:n,query:t},(function(e){t.sortInfo&&e.axis.setCategorySortInfo(t.sortInfo)}))}))}var NT=2*Math.PI,RT=Math.PI/180;function HT(t,e){return dp(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function jT(t,e,n){e.eachSeriesByType(t,(function(t){var e=t.getData(),i=e.mapDimension("value"),r=HT(t,n),o=t.get("center"),a=t.get("radius");$(a)||(a=[0,a]),$(o)||(o=[o,o]);var s=Nr(r.width,n.getWidth()),l=Nr(r.height,n.getHeight()),u=Math.min(s,l),c=Nr(o[0],s)+r.x,d=Nr(o[1],l)+r.y,h=Nr(a[0],u/2),f=Nr(a[1],u/2),p=-t.get("startAngle")*RT,m=t.get("minAngle")*RT,g=0;e.each(i,(function(t){!isNaN(t)&&g++}));var y=e.getSum(i),v=Math.PI/(y||g)*2,_=t.get("clockwise"),b=t.get("roseType"),x=t.get("stillShowZeroSum"),w=e.getDataExtent(i);w[0]=0;var M=NT,S=0,L=p,k=_?1:-1;if(e.setLayout({viewRect:r,r:f}),e.each(i,(function(t,n){var i;if(isNaN(t))e.setItemLayout(n,{angle:NaN,startAngle:NaN,endAngle:NaN,clockwise:_,cx:c,cy:d,r0:h,r:b?NaN:f});else{i="area"!==b?0===y&&x?v:t*v:NT/g,i<m?(i=m,M-=m):S+=t;var r=L+k*i;e.setItemLayout(n,{angle:i,startAngle:L,endAngle:r,clockwise:_,cx:c,cy:d,r0:h,r:b?Er(t,w,[h,f]):f}),L=r}})),M<NT&&g)if(M<=.001){var T=NT/g;e.each(i,(function(t,n){if(!isNaN(t)){var i=e.getItemLayout(n);i.angle=T,i.startAngle=p+k*n*T,i.endAngle=p+k*(n+1)*T}}))}else v=M/S,L=p,e.each(i,(function(t,n){if(!isNaN(t)){var i=e.getItemLayout(n),r=i.angle===m?m:t*v;i.startAngle=L,i.endAngle=L+k*r,L+=k*r}}))}))}function zT(t){return{seriesType:t,reset:function(t,e){var n=e.findComponents({mainType:"legend"});if(n&&n.length){var i=t.getData();i.filterSelf((function(t){for(var e=i.getName(t),r=0;r<n.length;r++)if(!n[r].isSelected(e))return!1;return!0}))}}}}var WT=Math.PI/180;function BT(t,e,n,i,r,o,a,s,l,u){if(!(t.length<2)){for(var c=t.length,d=0;d<c;d++)if("outer"===t[d].position&&"labelLine"===t[d].labelAlignTo){var h=t[d].label.x-u;t[d].linePoints[1][0]+=h,t[d].label.x=u}d_(t,l,l+a)&&p(t)}function f(t){for(var o=t.rB,a=o*o,s=0;s<t.list.length;s++){var l=t.list[s],u=Math.abs(l.label.y-n),c=i+l.len,d=c*c,h=Math.sqrt((1-Math.abs(u*u/a))*d);l.label.x=e+(h+l.len2)*r}}function p(t){for(var o={list:[],maxY:0},a={list:[],maxY:0},s=0;s<t.length;s++)if("none"===t[s].labelAlignTo){var l=t[s],u=l.label.y>n?a:o,c=Math.abs(l.label.y-n);if(c>u.maxY){var d=l.label.x-e-l.len2*r,h=i+l.len,p=Math.abs(d)<h?Math.sqrt(c*c/(1-d*d/h/h)):h;u.rB=p,u.maxY=c}u.list.push(l)}f(o),f(a)}}function FT(t,e,n,i,r,o,a,s){for(var l=[],u=[],c=Number.MAX_VALUE,d=-Number.MAX_VALUE,h=0;h<t.length;h++){var f=t[h].label;VT(t[h])||(f.x<e?(c=Math.min(c,f.x),l.push(t[h])):(d=Math.max(d,f.x),u.push(t[h])))}BT(u,e,n,i,1,r,o,a,s,d),BT(l,e,n,i,-1,r,o,a,s,c);for(h=0;h<t.length;h++){var p=t[h];f=p.label;if(!VT(p)){var m=p.linePoints;if(m){var g="edge"===p.labelAlignTo,y=p.rect.width,v=void 0;v=g?f.x<e?m[2][0]-p.labelDistance-a-p.edgeDistance:a+r-p.edgeDistance-m[2][0]-p.labelDistance:f.x<e?f.x-a-p.bleedMargin:a+r-f.x-p.bleedMargin,v<p.rect.width&&(p.label.style.width=v,"edge"===p.labelAlignTo&&(y=v));var _=m[1][0]-m[2][0];g?f.x<e?m[2][0]=a+p.edgeDistance+y+p.labelDistance:m[2][0]=a+r-p.edgeDistance-y-p.labelDistance:(f.x<e?m[2][0]=f.x+p.labelDistance:m[2][0]=f.x-p.labelDistance,m[1][0]=m[2][0]+_),m[1][1]=m[2][1]=f.y}}}}function VT(t){return"center"===t.position}function GT(t){var e,n,i=t.getData(),r=[],o=!1,a=(t.get("minShowLabelAngle")||0)*WT,s=i.getLayout("viewRect"),l=i.getLayout("r"),u=s.width,c=s.x,d=s.y,h=s.height;function f(t){t.ignore=!0}function p(t){if(!t.ignore)return!0;for(var e in t.states)if(!1===t.states[e].ignore)return!0;return!1}i.each((function(t){var s=i.getItemGraphicEl(t),d=s.shape,h=s.getTextContent(),m=s.getTextGuideLine(),g=i.getItemModel(t),y=g.getModel("label"),v=y.get("position")||g.get(["emphasis","label","position"]),_=y.get("distanceToLabelLine"),b=y.get("alignTo"),x=Nr(y.get("edgeDistance"),u),w=y.get("bleedMargin"),M=g.getModel("labelLine"),S=M.get("length");S=Nr(S,u);var L=M.get("length2");if(L=Nr(L,u),Math.abs(d.endAngle-d.startAngle)<a)return R(h.states,f),void(h.ignore=!0);if(p(h)){var k,T,D,C,I=(d.startAngle+d.endAngle)/2,A=Math.cos(I),Y=Math.sin(I);e=d.cx,n=d.cy;var O,P="inside"===v||"inner"===v;if("center"===v)k=d.cx,T=d.cy,C="center";else{var E=(P?(d.r+d.r0)/2*A:d.r*A)+e,N=(P?(d.r+d.r0)/2*Y:d.r*Y)+n;if(k=E+3*A,T=N+3*Y,!P){var H=E+A*(S+l-d.r),j=N+Y*(S+l-d.r),z=H+(A<0?-1:1)*L,W=j;k="edge"===b?A<0?c+x:c+u-x:z+(A<0?-_:_),T=W,D=[[E,N],[H,j],[z,W]]}C=P?"center":"edge"===b?A>0?"right":"left":A>0?"left":"right"}var B=y.get("rotate");if(O="number"===typeof B?B*(Math.PI/180):B?A<0?-I+Math.PI:-I:0,o=!!O,h.x=k,h.y=T,h.rotation=O,h.setStyle({verticalAlign:"middle"}),P){h.setStyle({align:C});var F=h.states.select;F&&(F.x+=h.x,F.y+=h.y)}else{var V=h.getBoundingRect().clone();V.applyTransform(h.getComputedTransform());var G=(h.style.margin||0)+2.1;V.y-=G/2,V.height+=G,r.push({label:h,labelLine:m,position:v,len:S,len2:L,minTurnAngle:M.get("minTurnAngle"),maxSurfaceAngle:M.get("maxSurfaceAngle"),surfaceNormal:new Wi(A,Y),linePoints:D,textAlign:C,labelDistance:_,labelAlignTo:b,edgeDistance:x,bleedMargin:w,rect:V})}s.setTextConfig({inside:P})}})),!o&&t.get("avoidLabelOverlap")&&FT(r,e,n,l,u,h,c,d);for(var m=0;m<r.length;m++){var g=r[m],y=g.label,v=g.labelLine,_=isNaN(y.x)||isNaN(y.y);if(y){y.setStyle({align:g.textAlign}),_&&(R(y.states,f),y.ignore=!0);var b=y.states.select;b&&(b.x+=y.x,b.y+=y.y)}if(v){var x=g.linePoints;_||!x?(R(v.states,f),v.ignore=!0):(n_(x,g.minTurnAngle),i_(x,g.surfaceNormal,g.maxSurfaceAngle),v.setShape({points:x}),y.__hostTarget.textGuideLineConfig={anchor:new Wi(x[0][0],x[0][1])})}}}function $T(t,e){var n=t.get("borderRadius");return null==n?null:($(n)||(n=[n,n]),{innerCornerRadius:lr(n[0],e.r0),cornerRadius:lr(n[1],e.r)})}var UT=function(t){function e(e,n,i){var r=t.call(this)||this;r.z2=2;var o=new Gl;return r.setTextContent(o),r.updateData(e,n,i,!0),r}return o(e,t),e.prototype.updateData=function(t,e,n,i){var r=this,o=t.hostModel,s=t.getItemModel(e),l=s.getModel("emphasis"),u=t.getItemLayout(e),c=I($T(s.getModel("itemStyle"),u)||{},u);if(isNaN(c.startAngle))r.setShape(c);else{if(i){r.setShape(c);var d=o.getShallow("animationType");"scale"===d?(r.shape.r=u.r0,eh(r,{shape:{r:u.r}},o,e)):null!=n?(r.setShape({startAngle:n,endAngle:n}),eh(r,{shape:{startAngle:u.startAngle,endAngle:u.endAngle}},o,e)):(r.shape.endAngle=u.startAngle,th(r,{shape:{endAngle:u.endAngle}},o,e))}else th(r,{shape:c},o,e);r.useStyle(t.getItemVisual(e,"style")),Zu(r,s);var h=(u.startAngle+u.endAngle)/2,f=o.get("selectedOffset"),p=Math.cos(h)*f,m=Math.sin(h)*f,g=s.getShallow("cursor");g&&r.attr("cursor",g),this._updateLabel(o,t,e),r.ensureState("emphasis").shape=a({r:u.r+(l.get("scale")&&l.get("scaleSize")||0)},$T(l.getModel("itemStyle"),u)),I(r.ensureState("select"),{x:p,y:m,shape:$T(s.getModel(["select","itemStyle"]),u)}),I(r.ensureState("blur"),{shape:$T(s.getModel(["blur","itemStyle"]),u)});var y=r.getTextGuideLine(),v=r.getTextContent();y&&I(y.ensureState("select"),{x:p,y:m}),I(v.ensureState("select"),{x:p,y:m}),$u(this,l.get("focus"),l.get("blurScope"))}},e.prototype._updateLabel=function(t,e,n){var i=this,r=e.getItemModel(n),o=r.getModel("labelLine"),a=e.getItemVisual(n,"style"),s=a&&a.fill,l=a&&a.opacity;Mh(i,Sh(r),{labelFetcher:e.hostModel,labelDataIndex:n,inheritColor:s,defaultOpacity:l,defaultText:t.getFormattedLabel(n,"normal")||e.getName(n)});var u=i.getTextContent();i.setTextConfig({position:null,rotation:null}),u.attr({z2:10});var c=t.get(["label","position"]);if("outside"!==c&&"outer"!==c)i.removeTextGuideLine();else{var d=this.getTextGuideLine();d||(d=new sd,this.setTextGuideLine(d)),a_(this,s_(r),{stroke:s,opacity:st(o.get(["lineStyle","opacity"]),l,1)})}},e}(qc),qT=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.ignoreLabelLineUpdate=!0,e}return o(e,t),e.prototype.init=function(){var t=new Mr;this._sectorGroup=t},e.prototype.render=function(t,e,n,i){var r,o=t.getData(),a=this._data,s=this.group;if(!a&&o.count()>0){for(var l=o.getItemLayout(0),u=1;isNaN(l&&l.startAngle)&&u<o.count();++u)l=o.getItemLayout(u);l&&(r=l.startAngle)}o.diff(a).add((function(t){var e=new UT(o,t,r);o.setItemGraphicEl(t,e),s.add(e)})).update((function(t,e){var n=a.getItemGraphicEl(e);n.updateData(o,t,r),n.off("click"),s.add(n),o.setItemGraphicEl(t,n)})).remove((function(e){var n=a.getItemGraphicEl(e);rh(n,t,e)})).execute(),GT(t),"expansion"!==t.get("animationTypeUpdate")&&(this._data=o)},e.prototype.dispose=function(){},e.prototype.containPoint=function(t,e){var n=e.getData(),i=n.getItemLayout(0);if(i){var r=t[0]-i.cx,o=t[1]-i.cy,a=Math.sqrt(r*r+o*o);return a<=i.r&&a>=i.r0}},e.type="pie",e}(Vy),XT=qT;function ZT(t,e,n){e=$(e)&&{coordDimensions:e}||I({},e);var i=t.getSource(),r=AL(i,e),o=new kL(r,t);return o.initData(i,n),o}var JT=function(){function t(t,e){this._getDataWithEncodedVisual=t,this._getRawData=e}return t.prototype.getAllNames=function(){var t=this._getRawData();return t.mapArray(t.getName)},t.prototype.containName=function(t){var e=this._getRawData();return e.indexOfName(t)>=0},t.prototype.indexOfName=function(t){var e=this._getDataWithEncodedVisual();return e.indexOfName(t)},t.prototype.getItemVisual=function(t,e){var n=this._getDataWithEncodedVisual();return n.getItemVisual(t,e)},t}(),KT=JT,QT=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.useColorPaletteOnData=!0,e}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new KT(V(this.getData,this),V(this.getRawData,this)),this._defaultLabelLine(e)},e.prototype.mergeOption=function(){t.prototype.mergeOption.apply(this,arguments)},e.prototype.getInitialData=function(){return ZT(this,{coordDimensions:["value"],encodeDefaulter:G(Hp,this)})},e.prototype.getDataParams=function(e){var n=this.getData(),i=t.prototype.getDataParams.call(this,e),r=[];return n.each(n.mapDimension("value"),(function(t){r.push(t)})),i.percent=Br(r,e,n.hostModel.get("percentPrecision")),i.$vars.push("percent"),i},e.prototype._defaultLabelLine=function(t){lo(t,"labelLine",["show"]);var e=t.labelLine,n=t.emphasis.labelLine;e.show=e.show&&t.label.show,n.show=n.show&&t.emphasis.label.show},e.type="series.pie",e.defaultOption={zlevel:0,z:2,legendHoverLink:!0,center:["50%","50%"],radius:[0,"75%"],clockwise:!0,startAngle:90,minAngle:0,minShowLabelAngle:0,selectedOffset:10,percentPrecision:2,stillShowZeroSum:!0,left:0,top:0,right:0,bottom:0,width:null,height:null,label:{rotate:0,show:!0,overflow:"truncate",position:"outer",alignTo:"none",edgeDistance:"25%",bleedMargin:10,distanceToLabelLine:5},labelLine:{show:!0,length:15,length2:15,smooth:!1,minTurnAngle:90,maxSurfaceAngle:90,lineStyle:{width:1,type:"solid"}},itemStyle:{borderWidth:1},labelLayout:{hideOverlap:!0},emphasis:{scale:!0,scaleSize:5},avoidLabelOverlap:!0,animationType:"expansion",animationDuration:1e3,animationTypeUpdate:"transition",animationEasingUpdate:"cubicInOut",animationDurationUpdate:500,animationEasing:"cubicInOut"},e}(Yy),tD=QT;function eD(t){t.registerChartView(XT),t.registerSeriesModel(tD),M_("pie",t.registerAction),t.registerLayout(G(jT,"pie")),t.registerProcessor(zT("pie"))}var nD=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){return BL(this.getSource(),this,{useEncodeDefaulter:!0})},e.prototype.getProgressive=function(){var t=this.option.progressive;return null==t?this.option.large?5e3:this.get("progressive"):t},e.prototype.getProgressiveThreshold=function(){var t=this.option.progressiveThreshold;return null==t?this.option.large?1e4:this.get("progressiveThreshold"):t},e.prototype.brushSelector=function(t,e,n){return n.point(e.getItemLayout(t))},e.type="series.scatter",e.dependencies=["grid","polar","geo","singleAxis","calendar"],e.defaultOption={coordinateSystem:"cartesian2d",zlevel:0,z:2,legendHoverLink:!0,symbolSize:10,large:!1,largeThreshold:2e3,itemStyle:{opacity:.8},emphasis:{scale:!0},clip:!0,select:{itemStyle:{borderColor:"#212121"}}},e}(Yy),iD=nD,rD=4,oD=function(){function t(){}return t}(),aD=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new oD},e.prototype.buildPath=function(t,e){var n=e.points,i=e.size,r=this.symbolProxy,o=r.shape,a=t.getContext?t.getContext():t,s=a&&i[0]<rD;if(s)this._ctx=a;else{this._ctx=null;for(var l=0;l<n.length;){var u=n[l++],c=n[l++];isNaN(u)||isNaN(c)||(this.softClipShape&&!this.softClipShape.contain(u,c)||(o.x=u-i[0]/2,o.y=c-i[1]/2,o.width=i[0],o.height=i[1],r.buildPath(t,o,!0)))}}},e.prototype.afterBrush=function(){var t=this.shape,e=t.points,n=t.size,i=this._ctx;if(i)for(var r=0;r<e.length;){var o=e[r++],a=e[r++];isNaN(o)||isNaN(a)||(this.softClipShape&&!this.softClipShape.contain(o,a)||i.fillRect(o-n[0]/2,a-n[1]/2,n[0],n[1]))}},e.prototype.findDataIndex=function(t,e){for(var n=this.shape,i=n.points,r=n.size,o=Math.max(r[0],4),a=Math.max(r[1],4),s=i.length/2-1;s>=0;s--){var l=2*s,u=i[l]-o/2,c=i[l+1]-a/2;if(t>=u&&e>=c&&t<=u+o&&e<=c+a)return s}return-1},e}(pl),sD=function(){function t(){this.group=new Mr}return t.prototype.isPersistent=function(){return!this._incremental},t.prototype.updateData=function(t,e){this.group.removeAll();var n=new aD({rectHover:!0,cursor:"default"});n.setShape({points:t.getLayout("points")}),this._setCommon(n,t,!1,e),this.group.add(n),this._incremental=null},t.prototype.updateLayout=function(t){if(!this._incremental){var e=t.getLayout("points");this.group.eachChild((function(t){if(null!=t.startIndex){var n=2*(t.endIndex-t.startIndex),i=4*t.startIndex*2;e=new Float32Array(e.buffer,i,n)}t.setShape("points",e)}))}},t.prototype.incrementalPrepareUpdate=function(t){this.group.removeAll(),this._clearIncremental(),t.count()>2e6?(this._incremental||(this._incremental=new Nd({silent:!0})),this.group.add(this._incremental)):this._incremental=null},t.prototype.incrementalUpdate=function(t,e,n){var i;this._incremental?(i=new aD,this._incremental.addDisplayable(i,!0)):(i=new aD({rectHover:!0,cursor:"default",startIndex:t.start,endIndex:t.end}),i.incremental=!0,this.group.add(i)),i.setShape({points:e.getLayout("points")}),this._setCommon(i,e,!!this._incremental,n)},t.prototype._setCommon=function(t,e,n,i){var r=e.hostModel;i=i||{};var o=e.getVisual("symbolSize");t.setShape("size",o instanceof Array?o:[o,o]),t.softClipShape=i.clipShape||null,t.symbolProxy=j_(e.getVisual("symbol"),0,0,0,0),t.setColor=t.symbolProxy.setColor;var a=t.shape.size[0]<rD;t.useStyle(r.getModel("itemStyle").getItemStyle(a?["color","shadowBlur","shadowColor"]:["color"]));var s=e.getVisual("style"),l=s&&s.fill;if(l&&t.setColor(l),!n){var u=$l(t);u.seriesIndex=r.seriesIndex,t.on("mousemove",(function(e){u.dataIndex=null;var n=t.findDataIndex(e.offsetX,e.offsetY);n>=0&&(u.dataIndex=n+(t.startIndex||0))}))}},t.prototype.remove=function(){this._clearIncremental(),this._incremental=null,this.group.removeAll()},t.prototype._clearIncremental=function(){var t=this._incremental;t&&t.clearDisplaybles()},t}(),lD=sD,uD=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=t.getData(),r=this._updateSymbolDraw(i,t);r.updateData(i,{clipShape:this._getClipShape(t)}),this._finished=!0},e.prototype.incrementalPrepareRender=function(t,e,n){var i=t.getData(),r=this._updateSymbolDraw(i,t);r.incrementalPrepareUpdate(i),this._finished=!1},e.prototype.incrementalRender=function(t,e,n){this._symbolDraw.incrementalUpdate(t,e.getData(),{clipShape:this._getClipShape(e)}),this._finished=t.end===e.getData().count()},e.prototype.updateTransform=function(t,e,n){var i=t.getData();if(this.group.dirty(),!this._finished||i.count()>1e4||!this._symbolDraw.isPersistent())return{update:!0};var r=Hk("").reset(t,e,n);r.progress&&r.progress({start:0,end:i.count(),count:i.count()},i),this._symbolDraw.updateLayout(i)},e.prototype._getClipShape=function(t){var e=t.coordinateSystem,n=e&&e.getArea&&e.getArea();return t.get("clip",!0)?n:null},e.prototype._updateSymbolDraw=function(t,e){var n=this._symbolDraw,i=e.pipelineContext,r=i.large;return n&&r===this._isLargeDraw||(n&&n.remove(),n=this._symbolDraw=r?new lD:new tk,this._isLargeDraw=r,this.group.removeAll()),this.group.add(n.group),n},e.prototype.remove=function(t,e){this._symbolDraw&&this._symbolDraw.remove(!0),this._symbolDraw=null},e.prototype.dispose=function(){},e.type="scatter",e}(Vy),cD=uD,dD=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.type="grid",e.dependencies=["xAxis","yAxis"],e.layoutMode="box",e.defaultOption={show:!1,zlevel:0,z:0,left:"10%",top:60,right:"10%",bottom:70,containLabel:!1,backgroundColor:"rgba(0,0,0,0)",borderWidth:1,borderColor:"#ccc"},e}(xp),hD=dD,fD=function(){function t(){}return t.prototype.getNeedCrossZero=function(){var t=this.option;return!t.scale},t.prototype.getCoordSysModel=function(){},t}(),pD=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.getCoordSysModel=function(){return this.getReferringComponents("grid",Po).models[0]},e.type="cartesian2dAxis",e}(xp);E(pD,fD);var mD={show:!0,zlevel:0,z:0,inverse:!1,name:"",nameLocation:"end",nameRotate:null,nameTruncate:{maxWidth:null,ellipsis:"...",placeholder:"."},nameTextStyle:{},nameGap:15,silent:!1,triggerEvent:!1,tooltip:{show:!1},axisPointer:{},axisLine:{show:!0,onZero:!0,onZeroAxisIndex:null,lineStyle:{color:"#6E7079",width:1,type:"solid"},symbol:["none","none"],symbolSize:[10,15]},axisTick:{show:!0,inside:!1,length:5,lineStyle:{width:1}},axisLabel:{show:!0,inside:!1,rotate:0,showMinLabel:null,showMaxLabel:null,margin:8,fontSize:12},splitLine:{show:!0,lineStyle:{color:["#E0E6F1"],width:1,type:"solid"}},splitArea:{show:!1,areaStyle:{color:["rgba(250,250,250,0.2)","rgba(210,219,238,0.2)"]}}},gD=D({boundaryGap:!0,deduplication:null,splitLine:{show:!1},axisTick:{alignWithLabel:!1,interval:"auto"},axisLabel:{interval:"auto"}},mD),yD=D({boundaryGap:[0,0],axisLine:{show:"auto"},axisTick:{show:"auto"},splitNumber:5,minorTick:{show:!1,splitNumber:5,length:3,lineStyle:{}},minorSplitLine:{show:!1,lineStyle:{color:"#F4F7FD",width:1}}},mD),vD=D({scale:!0,splitNumber:6,axisLabel:{showMinLabel:!1,showMaxLabel:!1,rich:{primary:{fontWeight:"bold"}}},splitLine:{show:!1}},yD),_D=A({scale:!0,logBase:10},yD),bD={category:gD,value:yD,time:vD,log:_D},xD=function(){function t(t){this.categories=t.categories||[],this._needCollect=t.needCollect,this._deduplication=t.deduplication}return t.createByAxisModel=function(e){var n=e.option,i=n.data,r=i&&H(i,wD);return new t({categories:r,needCollect:!r,deduplication:!1!==n.dedplication})},t.prototype.getOrdinal=function(t){return this._getOrCreateMap().get(t)},t.prototype.parseAndCollect=function(t){var e,n=this._needCollect;if("string"!==typeof t&&!n)return t;if(n&&!this._deduplication)return e=this.categories.length,this.categories[e]=t,e;var i=this._getOrCreateMap();return e=i.get(t),null==e&&(n?(e=this.categories.length,this.categories[e]=t,i.set(t,e)):e=NaN),e},t.prototype._getOrCreateMap=function(){return this._map||(this._map=gt(this.categories))},t}();function wD(t){return J(t)&&null!=t.value?t.value:t+""}var MD=xD,SD={value:1,category:1,time:1,log:1};function LD(t,e,n,i){R(SD,(function(r,a){var s=D(D({},bD[a],!0),i,!0),l=function(t){function n(){for(var n=[],i=0;i<arguments.length;i++)n[i]=arguments[i];var r=t.apply(this,n)||this;return r.type=e+"Axis."+a,r}return o(n,t),n.prototype.mergeDefaultAndTheme=function(t,e){var n=pp(this),i=n?gp(t):{},r=e.getTheme();D(t,r.get(a+"Axis")),D(t,this.getDefaultOption()),t.type=kD(t),n&&mp(t,i,n)},n.prototype.optionUpdated=function(){var t=this.option;"category"===t.type&&(this.__ordinalMeta=MD.createByAxisModel(this))},n.prototype.getCategories=function(t){var e=this.option;if("category"===e.type)return t?e.data:this.__ordinalMeta.categories},n.prototype.getOrdinalMeta=function(){return this.__ordinalMeta},n.type=e+"Axis."+a,n.defaultOption=s,n}(n);t.registerComponentModel(l)})),t.registerSubTypeDefaulter(e+"Axis",kD)}function kD(t){return t.type||(t.data?"category":"value")}var TD=function(){function t(t){this._setting=t||{},this._extent=[1/0,-1/0]}return t.prototype.getSetting=function(t){return this._setting[t]},t.prototype.unionExtent=function(t){var e=this._extent;t[0]<e[0]&&(e[0]=t[0]),t[1]>e[1]&&(e[1]=t[1])},t.prototype.unionExtentFromData=function(t,e){this.unionExtent(t.getApproximateExtent(e))},t.prototype.getExtent=function(){return this._extent.slice()},t.prototype.setExtent=function(t,e){var n=this._extent;isNaN(t)||(n[0]=t),isNaN(e)||(n[1]=e)},t.prototype.isInExtentRange=function(t){return this._extent[0]<=t&&this._extent[1]>=t},t.prototype.isBlank=function(){return this._isBlank},t.prototype.setBlank=function(t){this._isBlank=t},t}();ea(TD);var DD=TD,CD=Rr;function ID(t,e,n,i){var r={},o=t[1]-t[0],a=r.interval=Jr(o/e,!0);null!=n&&a<n&&(a=r.interval=n),null!=i&&a>i&&(a=r.interval=i);var s=r.intervalPrecision=AD(a),l=r.niceTickExtent=[CD(Math.ceil(t[0]/a)*a,s),CD(Math.floor(t[1]/a)*a,s)];return OD(l,t),r}function AD(t){return jr(t)+2}function YD(t,e,n){t[e]=Math.max(Math.min(t[e],n[1]),n[0])}function OD(t,e){!isFinite(t[0])&&(t[0]=e[0]),!isFinite(t[1])&&(t[1]=e[1]),YD(t,0,e),YD(t,1,e),t[0]>t[1]&&(t[0]=t[1])}function PD(t,e){return t>=e[0]&&t<=e[1]}function ED(t,e){return e[1]===e[0]?.5:(t-e[0])/(e[1]-e[0])}function ND(t,e){return t*(e[1]-e[0])+e[0]}var RD=function(t){function e(e){var n=t.call(this,e)||this;n.type="ordinal";var i=n.getSetting("ordinalMeta");return i||(i=new MD({})),$(i)&&(i=new MD({categories:H(i,(function(t){return J(t)?t.value:t}))})),n._ordinalMeta=i,n._extent=n.getSetting("extent")||[0,i.categories.length-1],n}return o(e,t),e.prototype.parse=function(t){return"string"===typeof t?this._ordinalMeta.getOrdinal(t):Math.round(t)},e.prototype.contain=function(t){return t=this.parse(t),PD(t,this._extent)&&null!=this._ordinalMeta.categories[t]},e.prototype.normalize=function(t){return t=this._getTickNumber(this.parse(t)),ED(t,this._extent)},e.prototype.scale=function(t){return t=Math.round(ND(t,this._extent)),this.getRawOrdinalNumber(t)},e.prototype.getTicks=function(){var t=[],e=this._extent,n=e[0];while(n<=e[1])t.push({value:n}),n++;return t},e.prototype.getMinorTicks=function(t){},e.prototype.setSortInfo=function(t){if(null!=t){for(var e=t.ordinalNumbers,n=this._ordinalNumbersByTick=[],i=this._ticksByOrdinalNumber=[],r=0,o=this._ordinalMeta.categories.length,a=Math.min(o,e.length);r<a;++r){var s=e[r];n[r]=s,i[s]=r}for(var l=0;r<o;++r){while(null!=i[l])l++;n.push(l),i[l]=r}}else this._ordinalNumbersByTick=this._ticksByOrdinalNumber=null},e.prototype._getTickNumber=function(t){var e=this._ticksByOrdinalNumber;return e&&t>=0&&t<e.length?e[t]:t},e.prototype.getRawOrdinalNumber=function(t){var e=this._ordinalNumbersByTick;return e&&t>=0&&t<e.length?e[t]:t},e.prototype.getLabel=function(t){if(!this.isBlank()){var e=this.getRawOrdinalNumber(t.value),n=this._ordinalMeta.categories[e];return null==n?"":n+""}},e.prototype.count=function(){return this._extent[1]-this._extent[0]+1},e.prototype.unionExtentFromData=function(t,e){this.unionExtent(t.getApproximateExtent(e))},e.prototype.isInExtentRange=function(t){return t=this._getTickNumber(t),this._extent[0]<=t&&this._extent[1]>=t},e.prototype.getOrdinalMeta=function(){return this._ordinalMeta},e.prototype.niceTicks=function(){},e.prototype.niceExtent=function(){},e.type="ordinal",e}(DD);DD.registerClass(RD);var HD=RD,jD=Rr,zD=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="interval",e._interval=0,e._intervalPrecision=2,e}return o(e,t),e.prototype.parse=function(t){return t},e.prototype.contain=function(t){return PD(t,this._extent)},e.prototype.normalize=function(t){return ED(t,this._extent)},e.prototype.scale=function(t){return ND(t,this._extent)},e.prototype.setExtent=function(t,e){var n=this._extent;isNaN(t)||(n[0]=parseFloat(t)),isNaN(e)||(n[1]=parseFloat(e))},e.prototype.unionExtent=function(t){var e=this._extent;t[0]<e[0]&&(e[0]=t[0]),t[1]>e[1]&&(e[1]=t[1]),this.setExtent(e[0],e[1])},e.prototype.getInterval=function(){return this._interval},e.prototype.setInterval=function(t){this._interval=t,this._niceExtent=this._extent.slice(),this._intervalPrecision=AD(t)},e.prototype.getTicks=function(t){var e=this._interval,n=this._extent,i=this._niceExtent,r=this._intervalPrecision,o=[];if(!e)return o;var a=1e4;n[0]<i[0]&&(t?o.push({value:jD(i[0]-e,r)}):o.push({value:n[0]}));var s=i[0];while(s<=i[1]){if(o.push({value:s}),s=jD(s+e,r),s===o[o.length-1].value)break;if(o.length>a)return[]}var l=o.length?o[o.length-1].value:i[1];return n[1]>l&&(t?o.push({value:jD(l+e,r)}):o.push({value:n[1]})),o},e.prototype.getMinorTicks=function(t){for(var e=this.getTicks(!0),n=[],i=this.getExtent(),r=1;r<e.length;r++){var o=e[r],a=e[r-1],s=0,l=[],u=o.value-a.value,c=u/t;while(s<t-1){var d=jD(a.value+(s+1)*c);d>i[0]&&d<i[1]&&l.push(d),s++}n.push(l)}return n},e.prototype.getLabel=function(t,e){if(null==t)return"";var n=e&&e.precision;null==n?n=jr(t.value)||0:"auto"===n&&(n=this._intervalPrecision);var i=jD(t.value,n,!0);return Gf(i)},e.prototype.niceTicks=function(t,e,n){t=t||5;var i=this._extent,r=i[1]-i[0];if(isFinite(r)){r<0&&(r=-r,i.reverse());var o=ID(i,t,e,n);this._intervalPrecision=o.intervalPrecision,this._interval=o.interval,this._niceExtent=o.niceTickExtent}},e.prototype.niceExtent=function(t){var e=this._extent;if(e[0]===e[1])if(0!==e[0]){var n=e[0];t.fixMax||(e[1]+=n/2),e[0]-=n/2}else e[1]=1;var i=e[1]-e[0];isFinite(i)||(e[0]=0,e[1]=1),this.niceTicks(t.splitNumber,t.minInterval,t.maxInterval);var r=this._interval;t.fixMin||(e[0]=jD(Math.floor(e[0]/r)*r)),t.fixMax||(e[1]=jD(Math.ceil(e[1]/r)*r))},e.type="interval",e}(DD);DD.registerClass(zD);var WD=zD,BD=function(t,e,n,i){while(n<i){var r=n+i>>>1;t[r][1]<e?n=r+1:i=r}return n},FD=function(t){function e(e){var n=t.call(this,e)||this;return n.type="time",n}return o(e,t),e.prototype.getLabel=function(t){var e=this.getSetting("useUTC");return Tf(t.value,bf[kf(Sf(this._minLevelUnit))]||bf.second,e,this.getSetting("locale"))},e.prototype.getFormattedLabel=function(t,e,n){var i=this.getSetting("useUTC"),r=this.getSetting("locale");return Df(t,e,n,r,i)},e.prototype.getTicks=function(t){var e=this._interval,n=this._extent,i=[];if(!e)return i;i.push({value:n[0],level:0});var r=this.getSetting("useUTC"),o=KD(this._minLevelUnit,this._approxInterval,r,n);return i=i.concat(o),i.push({value:n[1],level:0}),i},e.prototype.niceExtent=function(t){var e=this._extent;if(e[0]===e[1]&&(e[0]-=gf,e[1]+=gf),e[1]===-1/0&&e[0]===1/0){var n=new Date;e[1]=+new Date(n.getFullYear(),n.getMonth(),n.getDate()),e[0]=e[1]-gf}this.niceTicks(t.splitNumber,t.minInterval,t.maxInterval)},e.prototype.niceTicks=function(t,e,n){t=t||10;var i=this._extent,r=i[1]-i[0];this._approxInterval=r/t,null!=e&&this._approxInterval<e&&(this._approxInterval=e),null!=n&&this._approxInterval>n&&(this._approxInterval=n);var o=VD.length,a=Math.min(BD(VD,this._approxInterval,0,o),o-1);this._interval=VD[a][1],this._minLevelUnit=VD[Math.max(a-1,0)][0]},e.prototype.parse=function(t){return"number"===typeof t?t:+qr(t)},e.prototype.contain=function(t){return PD(this.parse(t),this._extent)},e.prototype.normalize=function(t){return ED(this.parse(t),this._extent)},e.prototype.scale=function(t){return ND(t,this._extent)},e.type="time",e}(WD),VD=[["second",ff],["minute",pf],["hour",mf],["quarter-day",6*mf],["half-day",12*mf],["day",1.2*gf],["half-week",3.5*gf],["week",7*gf],["month",31*gf],["quarter",95*gf],["half-year",yf/2],["year",yf]];function GD(t,e,n,i){var r=qr(e),o=qr(n),a=function(t){return If(r,t,i)===If(o,t,i)},s=function(){return a("year")},l=function(){return s()&&a("month")},u=function(){return l()&&a("day")},c=function(){return u()&&a("hour")},d=function(){return c()&&a("minute")},h=function(){return d()&&a("second")},f=function(){return h()&&a("millisecond")};switch(t){case"year":return s();case"month":return l();case"day":return u();case"hour":return c();case"minute":return d();case"second":return h();case"millisecond":return f()}}function $D(t,e){return t/=gf,t>16?16:t>7.5?7:t>3.5?4:t>1.5?2:1}function UD(t){var e=30*gf;return t/=e,t>6?6:t>3?3:t>2?2:1}function qD(t){return t/=mf,t>12?12:t>6?6:t>3.5?4:t>2?2:1}function XD(t,e){return t/=e?pf:ff,t>30?30:t>20?20:t>15?15:t>10?10:t>5?5:t>2?2:1}function ZD(t){return Jr(t,!0)}function JD(t,e,n){var i=new Date(t);switch(Sf(e)){case"year":case"month":i[jf(n)](0);case"day":i[zf(n)](1);case"hour":i[Wf(n)](0);case"minute":i[Bf(n)](0);case"second":i[Ff(n)](0),i[Vf(n)](0)}return i.getTime()}function KD(t,e,n,i){var r=1e4,o=wf,a=0;function s(t,e,n,r,o,a,s){var l=new Date(e),u=e,c=l[r]();while(u<n&&u<=i[1])s.push({value:u}),c+=t,l[o](c),u=l.getTime();s.push({value:u,notAdd:!0})}function l(t,r,o){var a=[],l=!r.length;if(!GD(Sf(t),i[0],i[1],n)){l&&(r=[{value:JD(new Date(i[0]),t,n)},{value:i[1]}]);for(var u=0;u<r.length-1;u++){var c=r[u].value,d=r[u+1].value;if(c!==d){var h=void 0,f=void 0,p=void 0,m=!1;switch(t){case"year":h=Math.max(1,Math.round(e/gf/365)),f=Af(n),p=Hf(n);break;case"half-year":case"quarter":case"month":h=UD(e),f=Yf(n),p=jf(n);break;case"week":case"half-week":case"day":h=$D(e,31),f=Of(n),p=zf(n),m=!0;break;case"half-day":case"quarter-day":case"hour":h=qD(e),f=Pf(n),p=Wf(n);break;case"minute":h=XD(e,!0),f=Ef(n),p=Bf(n);break;case"second":h=XD(e,!1),f=Nf(n),p=Ff(n);break;case"millisecond":h=ZD(e),f=Rf(n),p=Vf(n);break}s(h,c,d,f,p,m,a),"year"===t&&o.length>1&&0===u&&o.unshift({value:o[0].value-h})}}for(u=0;u<a.length;u++)o.push(a[u]);return a}}for(var u=[],c=[],d=0,h=0,f=0;f<o.length&&a++<r;++f){var p=Sf(o[f]);if(Lf(o[f])){l(o[f],u[u.length-1]||[],c);var m=o[f+1]?Sf(o[f+1]):null;if(p!==m){if(c.length){h=d,c.sort((function(t,e){return t.value-e.value}));for(var g=[],y=0;y<c.length;++y){var v=c[y].value;0!==y&&c[y-1].value===v||(g.push(c[y]),v>=i[0]&&v<=i[1]&&d++)}var _=(i[1]-i[0])/e;if(d>1.5*_&&h>_/1.5)break;if(u.push(g),d>_||t===o[f])break}c=[]}}}var b=z(H(u,(function(t){return z(t,(function(t){return t.value>=i[0]&&t.value<=i[1]&&!t.notAdd}))})),(function(t){return t.length>0})),x=[],w=b.length-1;for(f=0;f<b.length;++f)for(var M=b[f],S=0;S<M.length;++S)x.push({value:M[S].value,level:w-f});x.sort((function(t,e){return t.value-e.value}));var L=[];for(f=0;f<x.length;++f)0!==f&&x[f].value===x[f-1].value||L.push(x[f]);return L}DD.registerClass(FD);var QD=FD,tC=DD.prototype,eC=WD.prototype,nC=Rr,iC=Math.floor,rC=Math.ceil,oC=Math.pow,aC=Math.log,sC=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="log",e.base=10,e._originalScale=new WD,e._interval=0,e}return o(e,t),e.prototype.getTicks=function(t){var e=this._originalScale,n=this._extent,i=e.getExtent(),r=eC.getTicks.call(this,t);return H(r,(function(t){var e=t.value,r=Rr(oC(this.base,e));return r=e===n[0]&&this._fixMin?uC(r,i[0]):r,r=e===n[1]&&this._fixMax?uC(r,i[1]):r,{value:r}}),this)},e.prototype.setExtent=function(t,e){var n=this.base;t=aC(t)/aC(n),e=aC(e)/aC(n),eC.setExtent.call(this,t,e)},e.prototype.getExtent=function(){var t=this.base,e=tC.getExtent.call(this);e[0]=oC(t,e[0]),e[1]=oC(t,e[1]);var n=this._originalScale,i=n.getExtent();return this._fixMin&&(e[0]=uC(e[0],i[0])),this._fixMax&&(e[1]=uC(e[1],i[1])),e},e.prototype.unionExtent=function(t){this._originalScale.unionExtent(t);var e=this.base;t[0]=aC(t[0])/aC(e),t[1]=aC(t[1])/aC(e),tC.unionExtent.call(this,t)},e.prototype.unionExtentFromData=function(t,e){this.unionExtent(t.getApproximateExtent(e))},e.prototype.niceTicks=function(t){t=t||10;var e=this._extent,n=e[1]-e[0];if(!(n===1/0||n<=0)){var i=Xr(n),r=t/n*i;r<=.5&&(i*=10);while(!isNaN(i)&&Math.abs(i)<1&&Math.abs(i)>0)i*=10;var o=[Rr(rC(e[0]/i)*i),Rr(iC(e[1]/i)*i)];this._interval=i,this._niceExtent=o}},e.prototype.niceExtent=function(t){eC.niceExtent.call(this,t),this._fixMin=t.fixMin,this._fixMax=t.fixMax},e.prototype.parse=function(t){return t},e.prototype.contain=function(t){return t=aC(t)/aC(this.base),PD(t,this._extent)},e.prototype.normalize=function(t){return t=aC(t)/aC(this.base),ED(t,this._extent)},e.prototype.scale=function(t){return t=ND(t,this._extent),oC(this.base,t)},e.type="log",e}(DD),lC=sC.prototype;function uC(t,e){return nC(t,jr(e))}lC.getMinorTicks=eC.getMinorTicks,lC.getLabel=eC.getLabel,DD.registerClass(sC);var cC=sC,dC=function(){function t(t,e,n){this._prepareParams(t,e,n)}return t.prototype._prepareParams=function(t,e,n){n[1]<n[0]&&(n=[NaN,NaN]),this._dataMin=n[0],this._dataMax=n[1];var i=this._isOrdinal="ordinal"===t.type;this._needCrossZero=e.getNeedCrossZero&&e.getNeedCrossZero();var r=this._modelMinRaw=e.get("min",!0);U(r)?this._modelMinNum=mC(t,r({min:n[0],max:n[1]})):"dataMin"!==r&&(this._modelMinNum=mC(t,r));var o=this._modelMaxRaw=e.get("max",!0);if(U(o)?this._modelMaxNum=mC(t,o({min:n[0],max:n[1]})):"dataMax"!==o&&(this._modelMaxNum=mC(t,o)),i)this._axisDataLen=e.getCategories().length;else{var a=e.get("boundaryGap"),s=$(a)?a:[a||0,a||0];"boolean"===typeof s[0]||"boolean"===typeof s[1]?this._boundaryGapInner=[0,0]:this._boundaryGapInner=[lr(s[0],1),lr(s[1],1)]}},t.prototype.calculate=function(){var t=this._isOrdinal,e=this._dataMin,n=this._dataMax,i=this._axisDataLen,r=this._boundaryGapInner,o=t?null:n-e||Math.abs(e),a="dataMin"===this._modelMinRaw?e:this._modelMinNum,s="dataMax"===this._modelMaxRaw?n:this._modelMaxNum,l=null!=a,u=null!=s;null==a&&(a=t?i?0:NaN:e-r[0]*o),null==s&&(s=t?i?i-1:NaN:n+r[1]*o),(null==a||!isFinite(a))&&(a=NaN),(null==s||!isFinite(s))&&(s=NaN),a>s&&(a=NaN,s=NaN);var c=rt(a)||rt(s)||t&&!i;this._needCrossZero&&(a>0&&s>0&&!l&&(a=0),a<0&&s<0&&!u&&(s=0));var d=this._determinedMin,h=this._determinedMax;return null!=d&&(a=d,l=!0),null!=h&&(s=h,u=!0),{min:a,max:s,minFixed:l,maxFixed:u,isBlank:c}},t.prototype.modifyDataMinMax=function(t,e){this[fC[t]]=e},t.prototype.setDeterminedMinMax=function(t,e){var n=hC[t];this[n]=e},t.prototype.freeze=function(){this.frozen=!0},t}(),hC={min:"_determinedMin",max:"_determinedMax"},fC={min:"_dataMin",max:"_dataMax"};function pC(t,e,n){var i=t.rawExtentInfo;return i||(i=new dC(t,e,n),t.rawExtentInfo=i,i)}function mC(t,e){return null==e?null:rt(e)?NaN:t.parse(e)}function gC(t,e){var n=t.type,i=pC(t,e,t.getExtent()).calculate();t.setBlank(i.isBlank);var r=i.min,o=i.max,a=e.ecModel;if(a&&"time"===n){var s=Xk("bar",a),l=!1;if(R(s,(function(t){l=l||t.getBaseAxis()===e.axis})),l){var u=Jk(s),c=yC(r,o,e,u);r=c.min,o=c.max}}return{extent:[r,o],fixMin:i.minFixed,fixMax:i.maxFixed}}function yC(t,e,n,i){var r=n.axis.getExtent(),o=r[1]-r[0],a=Qk(i,n.axis);if(void 0===a)return{min:t,max:e};var s=1/0;R(a,(function(t){s=Math.min(t.offset,s)}));var l=-1/0;R(a,(function(t){l=Math.max(t.offset+t.width,l)})),s=Math.abs(s),l=Math.abs(l);var u=s+l,c=e-t,d=1-(s+l)/o,h=c/d-c;return e+=h*(l/u),t-=h*(s/u),{min:t,max:e}}function vC(t,e){var n=gC(t,e),i=n.extent,r=e.get("splitNumber");t instanceof cC&&(t.base=e.get("logBase"));var o=t.type;t.setExtent(i[0],i[1]),t.niceExtent({splitNumber:r,fixMin:n.fixMin,fixMax:n.fixMax,minInterval:"interval"===o||"time"===o?e.get("minInterval"):null,maxInterval:"interval"===o||"time"===o?e.get("maxInterval"):null});var a=e.get("interval");null!=a&&t.setInterval&&t.setInterval(a)}function _C(t,e){if(e=e||t.get("type"),e)switch(e){case"category":return new HD({ordinalMeta:t.getOrdinalMeta?t.getOrdinalMeta():t.getCategories(),extent:[1/0,-1/0]});case"time":return new QD({locale:t.ecModel.getLocaleModel(),useUTC:t.ecModel.get("useUTC")});default:return new(DD.getClass(e)||WD)}}function bC(t){var e=t.scale.getExtent(),n=e[0],i=e[1];return!(n>0&&i>0||n<0&&i<0)}function xC(t){var e=t.getLabelModel().get("formatter"),n="category"===t.type?t.scale.getExtent()[0]:null;return"time"===t.scale.type?function(e){return function(n,i){return t.scale.getFormattedLabel(n,i,e)}}(e):"string"===typeof e?function(e){return function(n){var i=t.scale.getLabel(n),r=e.replace("{value}",null!=i?i:"");return r}}(e):"function"===typeof e?function(e){return function(i,r){return null!=n&&(r=i.value-n),e(wC(t,i),r,null!=i.level?{level:i.level}:null)}}(e):function(e){return t.scale.getLabel(e)}}function wC(t,e){return"category"===t.type?t.scale.getLabel(e):e.value}function MC(t){var e=t.model,n=t.scale;if(e.get(["axisLabel","show"])&&!n.isBlank()){var i,r,o=n.getExtent();n instanceof HD?r=n.count():(i=n.getTicks(),r=i.length);var a,s=t.getLabelModel(),l=xC(t),u=1;r>40&&(u=Math.ceil(r/40));for(var c=0;c<r;c+=u){var d=i?i[c]:{value:o[0]+c},h=l(d,c),f=s.getTextRect(h),p=SC(f,s.get("rotate")||0);a?a.union(p):a=p}return a}}function SC(t,e){var n=e*Math.PI/180,i=t.width,r=t.height,o=i*Math.abs(Math.cos(n))+Math.abs(r*Math.sin(n)),a=i*Math.abs(Math.sin(n))+Math.abs(r*Math.cos(n)),s=new Ji(t.x,t.y,o,a);return s}function LC(t){var e=t.get("interval");return null==e?"auto":e}function kC(t){return"category"===t.type&&0===LC(t.getLabelModel())}function TC(t,e){var n={};return R(t.mapDimensionsAll(e),(function(e){n[HL(t,e)]=!0})),B(n)}function DC(t,e,n){e&&R(TC(e,n),(function(n){var i=e.getApproximateExtent(n);i[0]<t[0]&&(t[0]=i[0]),i[1]>t[1]&&(t[1]=i[1])}))}var CC=function(){function t(t){this.type="cartesian",this._dimList=[],this._axes={},this.name=t||""}return t.prototype.getAxis=function(t){return this._axes[t]},t.prototype.getAxes=function(){return H(this._dimList,(function(t){return this._axes[t]}),this)},t.prototype.getAxesByScale=function(t){return t=t.toLowerCase(),z(this.getAxes(),(function(e){return e.scale.type===t}))},t.prototype.addAxis=function(t){var e=t.dim;this._axes[e]=t,this._dimList.push(e)},t}(),IC=CC,AC=["x","y"];function YC(t){return"interval"===t.type||"time"===t.type}var OC=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="cartesian2d",e.dimensions=AC,e}return o(e,t),e.prototype.calcAffineTransform=function(){this._transform=this._invTransform=null;var t=this.getAxis("x").scale,e=this.getAxis("y").scale;if(YC(t)&&YC(e)){var n=t.getExtent(),i=e.getExtent(),r=this.dataToPoint([n[0],i[0]]),o=this.dataToPoint([n[1],i[1]]),a=n[1]-n[0],s=i[1]-i[0];if(a&&s){var l=(o[0]-r[0])/a,u=(o[1]-r[1])/s,c=r[0]-n[0]*l,d=r[1]-i[0]*u,h=this._transform=[l,0,0,u,c,d];this._invTransform=Ti([],h)}}},e.prototype.getBaseAxis=function(){return this.getAxesByScale("ordinal")[0]||this.getAxesByScale("time")[0]||this.getAxis("x")},e.prototype.containPoint=function(t){var e=this.getAxis("x"),n=this.getAxis("y");return e.contain(e.toLocalCoord(t[0]))&&n.contain(n.toLocalCoord(t[1]))},e.prototype.containData=function(t){return this.getAxis("x").containData(t[0])&&this.getAxis("y").containData(t[1])},e.prototype.dataToPoint=function(t,e,n){n=n||[];var i=t[0],r=t[1];if(this._transform&&null!=i&&isFinite(i)&&null!=r&&isFinite(r))return jt(n,t,this._transform);var o=this.getAxis("x"),a=this.getAxis("y");return n[0]=o.toGlobalCoord(o.dataToCoord(i,e)),n[1]=a.toGlobalCoord(a.dataToCoord(r,e)),n},e.prototype.clampData=function(t,e){var n=this.getAxis("x").scale,i=this.getAxis("y").scale,r=n.getExtent(),o=i.getExtent(),a=n.parse(t[0]),s=i.parse(t[1]);return e=e||[],e[0]=Math.min(Math.max(Math.min(r[0],r[1]),a),Math.max(r[0],r[1])),e[1]=Math.min(Math.max(Math.min(o[0],o[1]),s),Math.max(o[0],o[1])),e},e.prototype.pointToData=function(t,e){var n=[];if(this._invTransform)return jt(n,t,this._invTransform);var i=this.getAxis("x"),r=this.getAxis("y");return n[0]=i.coordToData(i.toLocalCoord(t[0]),e),n[1]=r.coordToData(r.toLocalCoord(t[1]),e),n},e.prototype.getOtherAxis=function(t){return this.getAxis("x"===t.dim?"y":"x")},e.prototype.getArea=function(){var t=this.getAxis("x").getGlobalExtent(),e=this.getAxis("y").getGlobalExtent(),n=Math.min(t[0],t[1]),i=Math.min(e[0],e[1]),r=Math.max(t[0],t[1])-n,o=Math.max(e[0],e[1])-i;return new Ji(n,i,r,o)},e}(IC),PC=OC,EC=Io();function NC(t){return"category"===t.type?HC(t):WC(t)}function RC(t,e){return"category"===t.type?zC(t,e):{ticks:H(t.scale.getTicks(),(function(t){return t.value}))}}function HC(t){var e=t.getLabelModel(),n=jC(t,e);return!e.get("show")||t.scale.isBlank()?{labels:[],labelCategoryInterval:n.labelCategoryInterval}:n}function jC(t,e){var n,i,r=BC(t,"labels"),o=LC(e),a=FC(r,o);return a||(U(o)?n=XC(t,o):(i="auto"===o?GC(t):o,n=qC(t,i)),VC(r,o,{labels:n,labelCategoryInterval:i}))}function zC(t,e){var n,i,r=BC(t,"ticks"),o=LC(e),a=FC(r,o);if(a)return a;if(e.get("show")&&!t.scale.isBlank()||(n=[]),U(o))n=XC(t,o,!0);else if("auto"===o){var s=jC(t,t.getLabelModel());i=s.labelCategoryInterval,n=H(s.labels,(function(t){return t.tickValue}))}else i=o,n=qC(t,i,!0);return VC(r,o,{ticks:n,tickCategoryInterval:i})}function WC(t){var e=t.scale.getTicks(),n=xC(t);return{labels:H(e,(function(e,i){return{formattedLabel:n(e,i),rawLabel:t.scale.getLabel(e),tickValue:e.value}}))}}function BC(t,e){return EC(t)[e]||(EC(t)[e]=[])}function FC(t,e){for(var n=0;n<t.length;n++)if(t[n].key===e)return t[n].value}function VC(t,e,n){return t.push({key:e,value:n}),n}function GC(t){var e=EC(t).autoInterval;return null!=e?e:EC(t).autoInterval=t.calculateCategoryInterval()}function $C(t){var e=UC(t),n=xC(t),i=(e.axisRotate-e.labelRotate)/180*Math.PI,r=t.scale,o=r.getExtent(),a=r.count();if(o[1]-o[0]<1)return 0;var s=1;a>40&&(s=Math.max(1,Math.floor(a/40)));for(var l=o[0],u=t.dataToCoord(l+1)-t.dataToCoord(l),c=Math.abs(u*Math.cos(i)),d=Math.abs(u*Math.sin(i)),h=0,f=0;l<=o[1];l+=s){var p=0,m=0,g=rr(n({value:l}),e.font,"center","top");p=1.3*g.width,m=1.3*g.height,h=Math.max(h,p,7),f=Math.max(f,m,7)}var y=h/c,v=f/d;isNaN(y)&&(y=1/0),isNaN(v)&&(v=1/0);var _=Math.max(0,Math.floor(Math.min(y,v))),b=EC(t.model),x=t.getExtent(),w=b.lastAutoInterval,M=b.lastTickCount;return null!=w&&null!=M&&Math.abs(w-_)<=1&&Math.abs(M-a)<=1&&w>_&&b.axisExtent0===x[0]&&b.axisExtent1===x[1]?_=w:(b.lastTickCount=a,b.lastAutoInterval=_,b.axisExtent0=x[0],b.axisExtent1=x[1]),_}function UC(t){var e=t.getLabelModel();return{axisRotate:t.getRotate?t.getRotate():t.isHorizontal&&!t.isHorizontal()?90:0,labelRotate:e.get("rotate")||0,font:e.getFont()}}function qC(t,e,n){var i=xC(t),r=t.scale,o=r.getExtent(),a=t.getLabelModel(),s=[],l=Math.max((e||0)+1,1),u=o[0],c=r.count();0!==u&&l>1&&c/l>2&&(u=Math.round(Math.ceil(u/l)*l));var d=kC(t),h=a.get("showMinLabel")||d,f=a.get("showMaxLabel")||d;h&&u!==o[0]&&m(o[0]);for(var p=u;p<=o[1];p+=l)m(p);function m(t){var e={value:t};s.push(n?t:{formattedLabel:i(e),rawLabel:r.getLabel(e),tickValue:t})}return f&&p-l!==o[1]&&m(o[1]),s}function XC(t,e,n){var i=t.scale,r=xC(t),o=[];return R(i.getTicks(),(function(t){var a=i.getLabel(t),s=t.value;e(t.value,a)&&o.push(n?s:{formattedLabel:r(t),rawLabel:a,tickValue:s})})),o}var ZC=[0,1],JC=function(){function t(t,e,n){this.onBand=!1,this.inverse=!1,this.dim=t,this.scale=e,this._extent=n||[0,0]}return t.prototype.contain=function(t){var e=this._extent,n=Math.min(e[0],e[1]),i=Math.max(e[0],e[1]);return t>=n&&t<=i},t.prototype.containData=function(t){return this.scale.contain(t)},t.prototype.getExtent=function(){return this._extent.slice()},t.prototype.getPixelPrecision=function(t){return Wr(t||this.scale.getExtent(),this._extent)},t.prototype.setExtent=function(t,e){var n=this._extent;n[0]=t,n[1]=e},t.prototype.dataToCoord=function(t,e){var n=this._extent,i=this.scale;return t=i.normalize(t),this.onBand&&"ordinal"===i.type&&(n=n.slice(),KC(n,i.count())),Er(t,ZC,n,e)},t.prototype.coordToData=function(t,e){var n=this._extent,i=this.scale;this.onBand&&"ordinal"===i.type&&(n=n.slice(),KC(n,i.count()));var r=Er(t,n,ZC,e);return this.scale.scale(r)},t.prototype.pointToData=function(t,e){},t.prototype.getTicksCoords=function(t){t=t||{};var e=t.tickModel||this.getTickModel(),n=RC(this,e),i=n.ticks,r=H(i,(function(t){return{coord:this.dataToCoord("ordinal"===this.scale.type?this.scale.getRawOrdinalNumber(t):t),tickValue:t}}),this),o=e.get("alignWithLabel");return QC(this,r,o,t.clamp),r},t.prototype.getMinorTicksCoords=function(){if("ordinal"===this.scale.type)return[];var t=this.model.getModel("minorTick"),e=t.get("splitNumber");e>0&&e<100||(e=5);var n=this.scale.getMinorTicks(e),i=H(n,(function(t){return H(t,(function(t){return{coord:this.dataToCoord(t),tickValue:t}}),this)}),this);return i},t.prototype.getViewLabels=function(){return NC(this).labels},t.prototype.getLabelModel=function(){return this.model.getModel("axisLabel")},t.prototype.getTickModel=function(){return this.model.getModel("axisTick")},t.prototype.getBandWidth=function(){var t=this._extent,e=this.scale.getExtent(),n=e[1]-e[0]+(this.onBand?1:0);0===n&&(n=1);var i=Math.abs(t[1]-t[0]);return Math.abs(i)/n},t.prototype.calculateCategoryInterval=function(){return $C(this)},t}();function KC(t,e){var n=t[1]-t[0],i=e,r=n/i/2;t[0]+=r,t[1]-=r}function QC(t,e,n,i){var r=e.length;if(t.onBand&&!n&&r){var o,a,s=t.getExtent();if(1===r)e[0].coord=s[0],o=e[1]={coord:s[0]};else{var l=e[r-1].tickValue-e[0].tickValue,u=(e[r-1].coord-e[0].coord)/l;R(e,(function(t){t.coord-=u/2}));var c=t.scale.getExtent();a=1+c[1]-e[r-1].tickValue,o={coord:e[r-1].coord+u*a},e.push(o)}var d=s[0]>s[1];h(e[0].coord,s[0])&&(i?e[0].coord=s[0]:e.shift()),i&&h(s[0],e[0].coord)&&e.unshift({coord:s[0]}),h(s[1],o.coord)&&(i?o.coord=s[1]:e.pop()),i&&h(o.coord,s[1])&&e.push({coord:s[1]})}function h(t,e){return t=Rr(t),e=Rr(e),d?t>e:t<e}}var tI=JC,eI=function(t){function e(e,n,i,r,o){var a=t.call(this,e,n,i)||this;return a.index=0,a.type=r||"value",a.position=o||"bottom",a}return o(e,t),e.prototype.isHorizontal=function(){var t=this.position;return"top"===t||"bottom"===t},e.prototype.getGlobalExtent=function(t){var e=this.getExtent();return e[0]=this.toGlobalCoord(e[0]),e[1]=this.toGlobalCoord(e[1]),t&&e[0]>e[1]&&e.reverse(),e},e.prototype.pointToData=function(t,e){return this.coordToData(this.toLocalCoord(t["x"===this.dim?0:1]),e)},e.prototype.setCategorySortInfo=function(t){if("category"!==this.type)return!1;this.model.option.categorySortInfo=t,this.scale.setSortInfo(t)},e}(tI),nI=eI;function iI(t,e,n){n=n||{};var i=t.coordinateSystem,r=e.axis,o={},a=r.getAxesOnZeroOf()[0],s=r.position,l=a?"onZero":s,u=r.dim,c=i.getRect(),d=[c.x,c.x+c.width,c.y,c.y+c.height],h={left:0,right:1,top:0,bottom:1,onZero:2},f=e.get("offset")||0,p="x"===u?[d[2]-f,d[3]+f]:[d[0]-f,d[1]+f];if(a){var m=a.toGlobalCoord(a.dataToCoord(0));p[h.onZero]=Math.max(Math.min(m,p[1]),p[0])}o.position=["y"===u?p[h[l]]:d[0],"x"===u?p[h[l]]:d[3]],o.rotation=Math.PI/2*("x"===u?0:1);var g={top:-1,bottom:1,left:-1,right:1};o.labelDirection=o.tickDirection=o.nameDirection=g[s],o.labelOffset=a?p[h[s]]-p[h.onZero]:0,e.get(["axisTick","inside"])&&(o.tickDirection=-o.tickDirection),ot(n.labelInside,e.get(["axisLabel","inside"]))&&(o.labelDirection=-o.labelDirection);var y=e.get(["axisLabel","rotate"]);return o.labelRotate="top"===l?-y:y,o.z2=1,o}function rI(t){return"cartesian2d"===t.get("coordinateSystem")}function oI(t){var e={xAxisModel:null,yAxisModel:null};return R(e,(function(n,i){var r=i.replace(/Model$/,""),o=t.getReferringComponents(r,Po).models[0];e[i]=o})),e}var aI=function(){function t(t,e,n){this.type="grid",this._coordsMap={},this._coordsList=[],this._axesMap={},this._axesList=[],this.axisPointerEnabled=!0,this.dimensions=AC,this._initCartesian(t,e,n),this.model=t}return t.prototype.getRect=function(){return this._rect},t.prototype.update=function(t,e){var n=this._axesMap;this._updateScale(t,this.model),R(n.x,(function(t){vC(t.scale,t.model)})),R(n.y,(function(t){vC(t.scale,t.model)}));var i={};R(n.x,(function(t){lI(n,"y",t,i)})),R(n.y,(function(t){lI(n,"x",t,i)})),this.resize(this.model,e)},t.prototype.resize=function(t,e,n){var i=t.getBoxLayoutParams(),r=!n&&t.get("containLabel"),o=dp(i,{width:e.getWidth(),height:e.getHeight()});this._rect=o;var a=this._axesList;function s(){R(a,(function(t){var e=t.isHorizontal(),n=e?[0,o.width]:[0,o.height],i=t.inverse?1:0;t.setExtent(n[i],n[1-i]),cI(t,e?o.x:o.y)}))}s(),r&&(R(a,(function(t){if(!t.model.get(["axisLabel","inside"])){var e=MC(t);if(e){var n=t.isHorizontal()?"height":"width",i=t.model.get(["axisLabel","margin"]);o[n]-=e[n]+i,"top"===t.position?o.y+=e.height+i:"left"===t.position&&(o.x+=e.width+i)}}})),s()),R(this._coordsList,(function(t){t.calcAffineTransform()}))},t.prototype.getAxis=function(t,e){var n=this._axesMap[t];if(null!=n)return n[e||0]},t.prototype.getAxes=function(){return this._axesList.slice()},t.prototype.getCartesian=function(t,e){if(null!=t&&null!=e){var n="x"+t+"y"+e;return this._coordsMap[n]}J(t)&&(e=t.yAxisIndex,t=t.xAxisIndex);for(var i=0,r=this._coordsList;i<r.length;i++)if(r[i].getAxis("x").index===t||r[i].getAxis("y").index===e)return r[i]},t.prototype.getCartesians=function(){return this._coordsList.slice()},t.prototype.convertToPixel=function(t,e,n){var i=this._findConvertTarget(e);return i.cartesian?i.cartesian.dataToPoint(n):i.axis?i.axis.toGlobalCoord(i.axis.dataToCoord(n)):null},t.prototype.convertFromPixel=function(t,e,n){var i=this._findConvertTarget(e);return i.cartesian?i.cartesian.pointToData(n):i.axis?i.axis.coordToData(i.axis.toLocalCoord(n)):null},t.prototype._findConvertTarget=function(t){var e,n,i=t.seriesModel,r=t.xAxisModel||i&&i.getReferringComponents("xAxis",Po).models[0],o=t.yAxisModel||i&&i.getReferringComponents("yAxis",Po).models[0],a=t.gridModel,s=this._coordsList;if(i)e=i.coordinateSystem,O(s,e)<0&&(e=null);else if(r&&o)e=this.getCartesian(r.componentIndex,o.componentIndex);else if(r)n=this.getAxis("x",r.componentIndex);else if(o)n=this.getAxis("y",o.componentIndex);else if(a){var l=a.coordinateSystem;l===this&&(e=this._coordsList[0])}return{cartesian:e,axis:n}},t.prototype.containPoint=function(t){var e=this._coordsList[0];if(e)return e.containPoint(t)},t.prototype._initCartesian=function(t,e,n){var i=this,r=this,o={left:!1,right:!1,top:!1,bottom:!1},a={x:{},y:{}},s={x:0,y:0};if(e.eachComponent("xAxis",l("x"),this),e.eachComponent("yAxis",l("y"),this),!s.x||!s.y)return this._axesMap={},void(this._axesList=[]);function l(e){return function(n,i){if(sI(n,t)){var l=n.get("position");"x"===e?"top"!==l&&"bottom"!==l&&(l=o.bottom?"top":"bottom"):"left"!==l&&"right"!==l&&(l=o.left?"right":"left"),o[l]=!0;var u=new nI(e,_C(n),[0,0],n.get("type"),l),c="category"===u.type;u.onBand=c&&n.get("boundaryGap"),u.inverse=n.get("inverse"),n.axis=u,u.model=n,u.grid=r,u.index=i,r._axesList.push(u),a[e][i]=u,s[e]++}}}this._axesMap=a,R(a.x,(function(e,n){R(a.y,(function(r,o){var a="x"+n+"y"+o,s=new PC(a);s.master=i,s.model=t,i._coordsMap[a]=s,i._coordsList.push(s),s.addAxis(e),s.addAxis(r)}))}))},t.prototype._updateScale=function(t,e){function n(t,e){R(TC(t,e.dim),(function(n){e.scale.unionExtentFromData(t,n)}))}R(this._axesList,(function(t){if(t.scale.setExtent(1/0,-1/0),"category"===t.type){var e=t.model.get("categorySortInfo");t.scale.setSortInfo(e)}})),t.eachSeries((function(t){if(rI(t)){var i=oI(t),r=i.xAxisModel,o=i.yAxisModel;if(!sI(r,e)||!sI(o,e))return;var a=this.getCartesian(r.componentIndex,o.componentIndex),s=t.getData(),l=a.getAxis("x"),u=a.getAxis("y");"list"===s.type&&(n(s,l),n(s,u))}}),this)},t.prototype.getTooltipAxes=function(t){var e=[],n=[];return R(this.getCartesians(),(function(i){var r=null!=t&&"auto"!==t?i.getAxis(t):i.getBaseAxis(),o=i.getOtherAxis(r);O(e,r)<0&&e.push(r),O(n,o)<0&&n.push(o)})),{baseAxes:e,otherAxes:n}},t.create=function(e,n){var i=[];return e.eachComponent("grid",(function(r,o){var a=new t(r,e,n);a.name="grid_"+o,a.resize(r,n,!0),r.coordinateSystem=a,i.push(a)})),e.eachSeries((function(t){if(rI(t)){var e=oI(t),n=e.xAxisModel,i=e.yAxisModel,r=n.getCoordSysModel();0;var o=r.coordinateSystem;t.coordinateSystem=o.getCartesian(n.componentIndex,i.componentIndex)}})),i},t.dimensions=AC,t}();function sI(t,e){return t.getCoordSysModel()===e}function lI(t,e,n,i){n.getAxesOnZeroOf=function(){return r?[r]:[]};var r,o=t[e],a=n.model,s=a.get(["axisLine","onZero"]),l=a.get(["axisLine","onZeroAxisIndex"]);if(s){if(null!=l)uI(o[l])&&(r=o[l]);else for(var u in o)if(o.hasOwnProperty(u)&&uI(o[u])&&!i[c(o[u])]){r=o[u];break}r&&(i[c(r)]=!0)}function c(t){return t.dim+"_"+t.index}}function uI(t){return t&&"category"!==t.type&&"time"!==t.type&&bC(t)}function cI(t,e){var n=t.getExtent(),i=n[0]+n[1];t.toGlobalCoord="x"===t.dim?function(t){return t+e}:function(t){return i-t+e},t.toLocalCoord="x"===t.dim?function(t){return t-e}:function(t){return i-t+e}}var dI=aI,hI=Math.PI,fI=function(){function t(t,e){this.group=new Mr,this.opt=e,this.axisModel=t,A(e,{labelOffset:0,nameDirection:1,tickDirection:1,labelDirection:1,silent:!0,handleAutoShown:function(){return!0}});var n=new Mr({x:e.position[0],y:e.position[1],rotation:e.rotation});n.updateTransform(),this._transformGroup=n}return t.prototype.hasBuilder=function(t){return!!pI[t]},t.prototype.add=function(t){pI[t](this.opt,this.axisModel,this.group,this._transformGroup)},t.prototype.getGroup=function(){return this.group},t.innerTextLayout=function(t,e,n){var i,r,o=Gr(e-t);return $r(o)?(r=n>0?"top":"bottom",i="center"):$r(o-hI)?(r=n>0?"bottom":"top",i="center"):(r="middle",i=o>0&&o<hI?n>0?"right":"left":n>0?"left":"right"),{rotation:o,textAlign:i,textVerticalAlign:r}},t.makeAxisEventDataBase=function(t){var e={componentType:t.mainType,componentIndex:t.componentIndex};return e[t.mainType+"Index"]=t.componentIndex,e},t.isLabelSilent=function(t){var e=t.get("tooltip");return t.get("silent")||!(t.get("triggerEvent")||e&&e.show)},t}(),pI={axisLine:function(t,e,n,i){var r=e.get(["axisLine","show"]);if("auto"===r&&t.handleAutoShown&&(r=t.handleAutoShown("axisLine")),r){var o=e.axis.getExtent(),a=i.transform,s=[o[0],0],l=[o[1],0];a&&(jt(s,s,a),jt(l,l,a));var u=I({lineCap:"round"},e.getModel(["axisLine","lineStyle"]).getLineStyle()),c=new dd({subPixelOptimize:!0,shape:{x1:s[0],y1:s[1],x2:l[0],y2:l[1]},style:u,strokeContainThreshold:t.strokeContainThreshold||5,silent:!0,z2:1});c.anid="line",n.add(c);var d=e.get(["axisLine","symbol"]),h=e.get(["axisLine","symbolSize"]),f=e.get(["axisLine","symbolOffset"])||0;if("number"===typeof f&&(f=[f,f]),null!=d){"string"===typeof d&&(d=[d,d]),"string"!==typeof h&&"number"!==typeof h||(h=[h,h]);var p=h[0],m=h[1];R([{rotate:t.rotation+Math.PI/2,offset:f[0],r:0},{rotate:t.rotation-Math.PI/2,offset:f[1],r:Math.sqrt((s[0]-l[0])*(s[0]-l[0])+(s[1]-l[1])*(s[1]-l[1]))}],(function(e,i){if("none"!==d[i]&&null!=d[i]){var r=j_(d[i],-p/2,-m/2,p,m,u.stroke,!0),o=e.r+e.offset;r.attr({rotation:e.rotate,x:s[0]+o*Math.cos(t.rotation),y:s[1]-o*Math.sin(t.rotation),silent:!0,z2:11}),n.add(r)}}))}}},axisTickLabel:function(t,e,n,i){var r=xI(n,i,e,t),o=MI(n,i,e,t);gI(e,o,r),wI(n,i,e,t.tickDirection)},axisName:function(t,e,n,i){var r=ot(t.axisName,e.get("name"));if(r){var o,a,s=e.get("nameLocation"),l=t.nameDirection,u=e.getModel("nameTextStyle"),c=e.get("nameGap")||0,d=e.axis.getExtent(),h=d[0]>d[1]?-1:1,f=["start"===s?d[0]-h*c:"end"===s?d[1]+h*c:(d[0]+d[1])/2,_I(s)?t.labelOffset+l*c:0],p=e.get("nameRotate");null!=p&&(p=p*hI/180),_I(s)?o=fI.innerTextLayout(t.rotation,null!=p?p:t.rotation,l):(o=mI(t.rotation,s,p||0,d),a=t.axisNameAvailableWidth,null!=a&&(a=Math.abs(a/Math.sin(o.rotation)),!isFinite(a)&&(a=null)));var m=u.getFont(),g=e.get("nameTruncate",!0)||{},y=g.ellipsis,v=ot(t.nameTruncateMaxWidth,g.maxWidth,a),_=new Gl({x:f[0],y:f[1],rotation:o.rotation,silent:fI.isLabelSilent(e),style:Lh(u,{text:r,font:m,overflow:"truncate",width:v,ellipsis:y,fill:u.getTextColor()||e.get(["axisLine","lineStyle","color"]),align:u.get("align")||o.textAlign,verticalAlign:u.get("verticalAlign")||o.textVerticalAlign}),z2:1});if(_h({el:_,componentModel:e,itemName:r}),_.__fullText=r,_.anid="name",e.get("triggerEvent")){var b=fI.makeAxisEventDataBase(e);b.targetType="axisName",b.name=r,$l(_).eventData=b}i.add(_),_.updateTransform(),n.add(_),_.decomposeTransform()}}};function mI(t,e,n,i){var r,o,a=Gr(n-t),s=i[0]>i[1],l="start"===e&&!s||"start"!==e&&s;return $r(a-hI/2)?(o=l?"bottom":"top",r="center"):$r(a-1.5*hI)?(o=l?"top":"bottom",r="center"):(o="middle",r=a<1.5*hI&&a>hI/2?l?"left":"right":l?"right":"left"),{rotation:a,textAlign:r,textVerticalAlign:o}}function gI(t,e,n){if(!kC(t.axis)){var i=t.get(["axisLabel","showMinLabel"]),r=t.get(["axisLabel","showMaxLabel"]);e=e||[],n=n||[];var o=e[0],a=e[1],s=e[e.length-1],l=e[e.length-2],u=n[0],c=n[1],d=n[n.length-1],h=n[n.length-2];!1===i?(yI(o),yI(u)):vI(o,a)&&(i?(yI(a),yI(c)):(yI(o),yI(u))),!1===r?(yI(s),yI(d)):vI(l,s)&&(r?(yI(l),yI(h)):(yI(s),yI(d)))}}function yI(t){t&&(t.ignore=!0)}function vI(t,e){var n=t&&t.getBoundingRect().clone(),i=e&&e.getBoundingRect().clone();if(n&&i){var r=xi([]);return Li(r,r,-t.rotation),n.applyTransform(Mi([],r,t.getLocalTransform())),i.applyTransform(Mi([],r,e.getLocalTransform())),n.intersect(i)}}function _I(t){return"middle"===t||"center"===t}function bI(t,e,n,i,r){for(var o=[],a=[],s=[],l=0;l<t.length;l++){var u=t[l].coord;a[0]=u,a[1]=0,s[0]=u,s[1]=n,e&&(jt(a,a,e),jt(s,s,e));var c=new dd({subPixelOptimize:!0,shape:{x1:a[0],y1:a[1],x2:s[0],y2:s[1]},style:i,z2:2,autoBatch:!0,silent:!0});c.anid=r+"_"+t[l].tickValue,o.push(c)}return o}function xI(t,e,n,i){var r=n.axis,o=n.getModel("axisTick"),a=o.get("show");if("auto"===a&&i.handleAutoShown&&(a=i.handleAutoShown("axisTick")),a&&!r.scale.isBlank()){for(var s=o.getModel("lineStyle"),l=i.tickDirection*o.get("length"),u=r.getTicksCoords(),c=bI(u,e.transform,l,A(s.getLineStyle(),{stroke:n.get(["axisLine","lineStyle","color"])}),"ticks"),d=0;d<c.length;d++)t.add(c[d]);return c}}function wI(t,e,n,i){var r=n.axis,o=n.getModel("minorTick");if(o.get("show")&&!r.scale.isBlank()){var a=r.getMinorTicksCoords();if(a.length)for(var s=o.getModel("lineStyle"),l=i*o.get("length"),u=A(s.getLineStyle(),A(n.getModel("axisTick").getLineStyle(),{stroke:n.get(["axisLine","lineStyle","color"])})),c=0;c<a.length;c++)for(var d=bI(a[c],e.transform,l,u,"minorticks_"+c),h=0;h<d.length;h++)t.add(d[h])}}function MI(t,e,n,i){var r=n.axis,o=ot(i.axisLabelShow,n.get(["axisLabel","show"]));if(o&&!r.scale.isBlank()){var a=n.getModel("axisLabel"),s=a.get("margin"),l=r.getViewLabels(),u=(ot(i.labelRotate,a.get("rotate"))||0)*hI/180,c=fI.innerTextLayout(i.rotation,u,i.labelDirection),d=n.getCategories&&n.getCategories(!0),h=[],f=fI.isLabelSilent(n),p=n.get("triggerEvent");return R(l,(function(o,l){var u="ordinal"===r.scale.type?r.scale.getRawOrdinalNumber(o.tickValue):o.tickValue,m=o.formattedLabel,g=o.rawLabel,y=a;if(d&&d[u]){var v=d[u];J(v)&&v.textStyle&&(y=new qh(v.textStyle,a,n.ecModel))}var _=y.getTextColor()||n.get(["axisLine","lineStyle","color"]),b=r.dataToCoord(u),x=new Gl({x:b,y:i.labelOffset+i.labelDirection*s,rotation:c.rotation,silent:f,z2:10,style:Lh(y,{text:m,align:y.getShallow("align",!0)||c.textAlign,verticalAlign:y.getShallow("verticalAlign",!0)||y.getShallow("baseline",!0)||c.textVerticalAlign,fill:"function"===typeof _?_("category"===r.type?g:"value"===r.type?u+"":u,l):_})});if(x.anid="label_"+u,p){var w=fI.makeAxisEventDataBase(n);w.targetType="axisLabel",w.value=g,$l(x).eventData=w}e.add(x),x.updateTransform(),h.push(x),t.add(x),x.decomposeTransform()})),h}}var SI=fI;function LI(t,e){var n={axesInfo:{},seriesInvolved:!1,coordSysAxesInfo:{},coordSysMap:{}};return kI(n,t,e),n.seriesInvolved&&DI(n,t),n}function kI(t,e,n){var i=e.getComponent("tooltip"),r=e.getComponent("axisPointer"),o=r.get("link",!0)||[],a=[];R(n.getCoordinateSystems(),(function(n){if(n.axisPointerEnabled){var s=EI(n.model),l=t.coordSysAxesInfo[s]={};t.coordSysMap[s]=n;var u=n.model,c=u.getModel("tooltip",i);if(R(n.getAxes(),G(p,!1,null)),n.getTooltipAxes&&i&&c.get("show")){var d="axis"===c.get("trigger"),h="cross"===c.get(["axisPointer","type"]),f=n.getTooltipAxes(c.get(["axisPointer","axis"]));(d||h)&&R(f.baseAxes,G(p,!h||"cross",d)),h&&R(f.otherAxes,G(p,"cross",!1))}}function p(i,s,u){var d=u.model.getModel("axisPointer",r),h=d.get("show");if(h&&("auto"!==h||i||PI(d))){null==s&&(s=d.get("triggerTooltip")),d=i?TI(u,c,r,e,i,s):d;var f=d.get("snap"),p=EI(u.model),m=s||f||"category"===u.type,g=t.axesInfo[p]={key:p,axis:u,coordSys:n,axisPointerModel:d,triggerTooltip:s,involveSeries:m,snap:f,useHandle:PI(d),seriesModels:[],linkGroup:null};l[p]=g,t.seriesInvolved=t.seriesInvolved||m;var y=CI(o,u);if(null!=y){var v=a[y]||(a[y]={axesInfo:{}});v.axesInfo[p]=g,v.mapper=o[y].mapper,g.linkGroup=v}}}}))}function TI(t,e,n,i,r,o){var a=e.getModel("axisPointer"),s=["type","snap","lineStyle","shadowStyle","label","animation","animationDurationUpdate","animationEasingUpdate","z"],l={};R(s,(function(t){l[t]=T(a.get(t))})),l.snap="category"!==t.type&&!!o,"cross"===a.get("type")&&(l.type="line");var u=l.label||(l.label={});if(null==u.show&&(u.show=!1),"cross"===r){var c=a.get(["label","show"]);if(u.show=null==c||c,!o){var d=l.lineStyle=a.get("crossStyle");d&&A(u,d.textStyle)}}return t.model.getModel("axisPointer",new qh(l,n,i))}function DI(t,e){e.eachSeries((function(e){var n=e.coordinateSystem,i=e.get(["tooltip","trigger"],!0),r=e.get(["tooltip","show"],!0);n&&"none"!==i&&!1!==i&&"item"!==i&&!1!==r&&!1!==e.get(["axisPointer","show"],!0)&&R(t.coordSysAxesInfo[EI(n.model)],(function(t){var i=t.axis;n.getAxis(i.dim)===i&&(t.seriesModels.push(e),null==t.seriesDataCount&&(t.seriesDataCount=0),t.seriesDataCount+=e.getData().count())}))}))}function CI(t,e){for(var n=e.model,i=e.dim,r=0;r<t.length;r++){var o=t[r]||{};if(II(o[i+"AxisId"],n.id)||II(o[i+"AxisIndex"],n.componentIndex)||II(o[i+"AxisName"],n.name))return r}}function II(t,e){return"all"===t||$(t)&&O(t,e)>=0||t===e}function AI(t){var e=YI(t);if(e){var n=e.axisPointerModel,i=e.axis.scale,r=n.option,o=n.get("status"),a=n.get("value");null!=a&&(a=i.parse(a));var s=PI(n);null==o&&(r.status=s?"show":"hide");var l=i.getExtent().slice();l[0]>l[1]&&l.reverse(),(null==a||a>l[1])&&(a=l[1]),a<l[0]&&(a=l[0]),r.value=a,s&&(r.status=e.axis.scale.isBlank()?"hide":"show")}}function YI(t){var e=(t.ecModel.getComponent("axisPointer")||{}).coordSysAxesInfo;return e&&e.axesInfo[EI(t)]}function OI(t){var e=YI(t);return e&&e.axisPointerModel}function PI(t){return!!t.get(["handle","show"])}function EI(t){return t.type+"||"+t.id}var NI={},RI=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(e,n,i,r){this.axisPointerClass&&AI(e),t.prototype.render.apply(this,arguments),this._doUpdateAxisPointerClass(e,i,!0)},e.prototype.updateAxisPointer=function(t,e,n,i){this._doUpdateAxisPointerClass(t,n,!1)},e.prototype.remove=function(t,e){var n=this._axisPointer;n&&n.remove(e)},e.prototype.dispose=function(e,n){this._disposeAxisPointer(n),t.prototype.dispose.apply(this,arguments)},e.prototype._doUpdateAxisPointerClass=function(t,n,i){var r=e.getAxisPointerClass(this.axisPointerClass);if(r){var o=OI(t);o?(this._axisPointer||(this._axisPointer=new r)).render(t,o,n,i):this._disposeAxisPointer(n)}},e.prototype._disposeAxisPointer=function(t){this._axisPointer&&this._axisPointer.dispose(t),this._axisPointer=null},e.registerAxisPointerClass=function(t,e){NI[t]=e},e.getAxisPointerClass=function(t){return t&&NI[t]},e.type="axis",e}(Py),HI=RI,jI=Io();function zI(t,e,n,i){var r=n.axis;if(!r.scale.isBlank()){var o=n.getModel("splitArea"),a=o.getModel("areaStyle"),s=a.get("color"),l=i.coordinateSystem.getRect(),u=r.getTicksCoords({tickModel:o,clamp:!0});if(u.length){var c=s.length,d=jI(t).splitAreaColors,h=gt(),f=0;if(d)for(var p=0;p<u.length;p++){var m=d.get(u[p].tickValue);if(null!=m){f=(m+(c-1)*p)%c;break}}var g=r.toGlobalCoord(u[0].coord),y=a.getAreaStyle();s=$(s)?s:[s];for(p=1;p<u.length;p++){var v=r.toGlobalCoord(u[p].coord),_=void 0,b=void 0,x=void 0,w=void 0;r.isHorizontal()?(_=g,b=l.y,x=v-_,w=l.height,g=_+x):(_=l.x,b=g,x=l.width,w=v-b,g=b+w);var M=u[p-1].tickValue;null!=M&&h.set(M,f),e.add(new Al({anid:null!=M?"area_"+M:null,shape:{x:_,y:b,width:x,height:w},style:A({fill:s[f]},y),autoBatch:!0,silent:!0})),f=(f+1)%c}jI(t).splitAreaColors=h}}}function WI(t){jI(t).splitAreaColors=null}var BI=["axisLine","axisTickLabel","axisName"],FI=["splitArea","splitLine","minorSplitLine"],VI=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="CartesianAxisPointer",n}return o(e,t),e.prototype.render=function(e,n,i,r){this.group.removeAll();var o=this._axisGroup;if(this._axisGroup=new Mr,this.group.add(this._axisGroup),e.get("show")){var a=e.getCoordSysModel(),s=iI(a,e),l=new SI(e,I({handleAutoShown:function(t){for(var n=a.coordinateSystem.getCartesians(),i=0;i<n.length;i++){var r=n[i].getOtherAxis(e.axis).type;if("value"===r||"log"===r)return!0}return!1}},s));R(BI,l.add,l),this._axisGroup.add(l.getGroup()),R(FI,(function(t){e.get([t,"show"])&&GI[t](this,this._axisGroup,e,a)}),this),dh(o,this._axisGroup,e),t.prototype.render.call(this,e,n,i,r)}},e.prototype.remove=function(){WI(this)},e.type="cartesianAxis",e}(HI),GI={splitLine:function(t,e,n,i){var r=n.axis;if(!r.scale.isBlank()){var o=n.getModel("splitLine"),a=o.getModel("lineStyle"),s=a.get("color");s=$(s)?s:[s];for(var l=i.coordinateSystem.getRect(),u=r.isHorizontal(),c=0,d=r.getTicksCoords({tickModel:o}),h=[],f=[],p=a.getLineStyle(),m=0;m<d.length;m++){var g=r.toGlobalCoord(d[m].coord);u?(h[0]=g,h[1]=l.y,f[0]=g,f[1]=l.y+l.height):(h[0]=l.x,h[1]=g,f[0]=l.x+l.width,f[1]=g);var y=c++%s.length,v=d[m].tickValue;e.add(new dd({anid:null!=v?"line_"+d[m].tickValue:null,subPixelOptimize:!0,autoBatch:!0,shape:{x1:h[0],y1:h[1],x2:f[0],y2:f[1]},style:A({stroke:s[y]},p),silent:!0}))}}},minorSplitLine:function(t,e,n,i){var r=n.axis,o=n.getModel("minorSplitLine"),a=o.getModel("lineStyle"),s=i.coordinateSystem.getRect(),l=r.isHorizontal(),u=r.getMinorTicksCoords();if(u.length)for(var c=[],d=[],h=a.getLineStyle(),f=0;f<u.length;f++)for(var p=0;p<u[f].length;p++){var m=r.toGlobalCoord(u[f][p].coord);l?(c[0]=m,c[1]=s.y,d[0]=m,d[1]=s.y+s.height):(c[0]=s.x,c[1]=m,d[0]=s.x+s.width,d[1]=m),e.add(new dd({anid:"minor_line_"+u[f][p].tickValue,subPixelOptimize:!0,autoBatch:!0,shape:{x1:c[0],y1:c[1],x2:d[0],y2:d[1]},style:h,silent:!0}))}},splitArea:function(t,e,n,i){zI(t,e,n,i)}},$I=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="xAxis",e}(VI),UI=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type=$I.type,e}return o(e,t),e.type="yAxis",e}(VI),qI=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="grid",e}return o(e,t),e.prototype.render=function(t,e){this.group.removeAll(),t.get("show")&&this.group.add(new Al({shape:t.coordinateSystem.getRect(),style:A({fill:t.get("backgroundColor")},t.getItemStyle()),silent:!0,z2:-1}))},e.type="grid",e}(Py),XI={offset:0};function ZI(t){t.registerComponentView(qI),t.registerComponentModel(hD),t.registerCoordinateSystem("cartesian2d",dI),LD(t,"x",pD,XI),LD(t,"y",pD,XI),t.registerComponentView($I),t.registerComponentView(UI),t.registerPreprocessor((function(t){t.xAxis&&t.yAxis&&!t.grid&&(t.grid={})}))}function JI(t){vM(ZI),t.registerSeriesModel(iD),t.registerChartView(cD),t.registerLayout(Hk("scatter"))}function KI(t){t.eachSeriesByType("radar",(function(t){var e=t.getData(),n=[],i=t.coordinateSystem;if(i){var r=i.getIndicatorAxes();R(r,(function(t,o){e.each(e.mapDimension(r[o].dim),(function(t,e){n[e]=n[e]||[];var r=i.dataToPoint(t,o);n[e][o]=QI(r)?r:tA(i)}))})),e.each((function(t){var r=W(n[t],(function(t){return QI(t)}))||tA(i);n[t].push(r.slice()),e.setItemLayout(t,n[t])}))}}))}function QI(t){return!isNaN(t[0])&&!isNaN(t[1])}function tA(t){return[t.cx,t.cy]}function eA(t){var e=t.polar;if(e){$(e)||(e=[e]);var n=[];R(e,(function(e,i){e.indicator?(e.type&&!e.shape&&(e.shape=e.type),t.radar=t.radar||[],$(t.radar)||(t.radar=[t.radar]),t.radar.push(e)):n.push(e)})),t.polar=n}R(t.series,(function(t){t&&"radar"===t.type&&t.polarIndex&&(t.radarIndex=t.polarIndex)}))}function nA(t){return $(t)||(t=[+t,+t]),t}var iA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var r=t.coordinateSystem,o=this.group,a=t.getData(),s=this._data;function l(t,e){var n=t.getItemVisual(e,"symbol")||"circle";if("none"!==n){var i=nA(t.getItemVisual(e,"symbolSize")),r=j_(n,-1,-1,2,2),o=t.getItemVisual(e,"symbolRotate")||0;return r.attr({style:{strokeNoScale:!0},z2:100,scaleX:i[0]/2,scaleY:i[1]/2,rotation:o*Math.PI/180||0}),r}}function u(e,n,r,o,a,s){r.removeAll();for(var u=0;u<n.length-1;u++){var c=l(o,a);c&&(c.__dimIdx=u,e[u]?(c.setPosition(e[u]),i[s?"initProps":"updateProps"](c,{x:n[u][0],y:n[u][1]},t,a)):c.setPosition(n[u]),r.add(c))}}function c(t){return H(t,(function(t){return[r.cx,r.cy]}))}a.diff(s).add((function(e){var n=a.getItemLayout(e);if(n){var i=new rd,r=new sd,o={shape:{points:n}};i.shape.points=c(n),r.shape.points=c(n),eh(i,o,t,e),eh(r,o,t,e);var s=new Mr,l=new Mr;s.add(r),s.add(i),s.add(l),u(r.shape.points,n,l,a,e,!0),a.setItemGraphicEl(e,s)}})).update((function(e,n){var i=s.getItemGraphicEl(n),r=i.childAt(0),o=i.childAt(1),l=i.childAt(2),c={shape:{points:a.getItemLayout(e)}};c.shape.points&&(u(r.shape.points,c.shape.points,l,a,e,!1),th(r,c,t),th(o,c,t),a.setItemGraphicEl(e,i))})).remove((function(t){o.remove(s.getItemGraphicEl(t))})).execute(),a.eachItemGraphicEl((function(t,e){var n=a.getItemModel(e),i=t.childAt(0),r=t.childAt(1),s=t.childAt(2),l=a.getItemVisual(e,"style"),u=l.fill;o.add(t),i.useStyle(A(n.getModel("lineStyle").getLineStyle(),{fill:"none",stroke:u})),Zu(i,n,"lineStyle"),Zu(r,n,"areaStyle");var c=n.getModel("areaStyle"),d=c.isEmpty()&&c.parentModel.isEmpty();r.ignore=d,R(["emphasis","select","blur"],(function(t){var e=n.getModel([t,"areaStyle"]),i=e.isEmpty()&&e.parentModel.isEmpty();r.ensureState(t).ignore=i&&d})),r.useStyle(A(c.getAreaStyle(),{fill:u,opacity:.7,decal:l.decal}));var h=n.getModel("emphasis"),f=h.getModel("itemStyle").getItemStyle();s.eachChild((function(t){if(t instanceof wl){var i=t.style;t.useStyle(I({image:i.image,x:i.x,y:i.y,width:i.width,height:i.height},l))}else t.useStyle(l),t.setColor(u);var r=t.ensureState("emphasis");r.style=T(f);var o=a.get(a.dimensions[t.__dimIdx],e);(null==o||isNaN(o))&&(o=""),Mh(t,Sh(n),{labelFetcher:a.hostModel,labelDataIndex:e,labelDimIndex:t.__dimIdx,defaultText:o,inheritColor:u,defaultOpacity:l.opacity})})),$u(t,h.get("focus"),h.get("blurScope"))})),this._data=a},e.prototype.remove=function(){this.group.removeAll(),this._data=null},e.type="radar",e}(Vy),rA=iA,oA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.useColorPaletteOnData=!0,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new KT(V(this.getData,this),V(this.getRawData,this))},e.prototype.getInitialData=function(t,e){return ZT(this,{generateCoord:"indicator_",generateCoordCount:1/0})},e.prototype.formatTooltip=function(t,e,n){var i=this.getData(),r=this.coordinateSystem,o=r.getIndicatorAxes(),a=this.getData().getName(t),s=""===a?this.name:a,l=gy(this,t);return oy("section",{header:s,sortBlocks:!0,blocks:H(o,(function(e){var n=i.get(i.mapDimension(e.dim),t);return oy("nameValue",{markerType:"subItem",markerColor:l,name:e.name,value:n,sortParam:n})}))})},e.prototype.getTooltipPosition=function(t){if(null!=t)for(var e=this.getData(),n=this.coordinateSystem,i=e.getValues(H(n.dimensions,(function(t){return e.mapDimension(t)})),t),r=0,o=i.length;r<o;r++)if(!isNaN(i[r])){var a=n.getIndicatorAxes();return n.coordToPoint(a[r].dataToCoord(i[r]),r)}},e.type="series.radar",e.dependencies=["radar"],e.defaultOption={zlevel:0,z:2,coordinateSystem:"radar",legendHoverLink:!0,radarIndex:0,lineStyle:{width:2,type:"solid"},label:{position:"top"},symbolSize:8},e}(Yy),aA=oA,sA=bD.value;function lA(t,e){return A({show:e},t)}var uA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(){var t=this.get("boundaryGap"),e=this.get("splitNumber"),n=this.get("scale"),i=this.get("axisLine"),r=this.get("axisTick"),o=this.get("axisLabel"),a=this.get("axisName"),s=this.get(["axisName","show"]),l=this.get(["axisName","formatter"]),u=this.get("axisNameGap"),c=this.get("triggerEvent"),d=H(this.get("indicator")||[],(function(d){null!=d.max&&d.max>0&&!d.min?d.min=0:null!=d.min&&d.min<0&&!d.max&&(d.max=0);var h=a;null!=d.color&&(h=A({color:d.color},a));var f=D(T(d),{boundaryGap:t,splitNumber:e,scale:n,axisLine:i,axisTick:r,axisLabel:o,name:d.text,nameLocation:"end",nameGap:u,nameTextStyle:h,triggerEvent:c},!1);if(s||(f.name=""),"string"===typeof l){var p=f.name;f.name=l.replace("{value}",null!=p?p:"")}else"function"===typeof l&&(f.name=l(f.name,f));var m=new qh(f,null,this.ecModel);return E(m,fD.prototype),m.mainType="radar",m.componentIndex=this.componentIndex,m}),this);this._indicatorModels=d},e.prototype.getIndicatorModels=function(){return this._indicatorModels},e.type="radar",e.defaultOption={zlevel:0,z:0,center:["50%","50%"],radius:"75%",startAngle:90,axisName:{show:!0},boundaryGap:[0,0],splitNumber:5,axisNameGap:15,scale:!1,shape:"polygon",axisLine:D({lineStyle:{color:"#bbb"}},sA.axisLine),axisLabel:lA(sA.axisLabel,!1),axisTick:lA(sA.axisTick,!1),splitLine:lA(sA.splitLine,!0),splitArea:lA(sA.splitArea,!0),indicator:[]},e}(xp),cA=uA,dA=["axisLine","axisTickLabel","axisName"],hA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=this.group;i.removeAll(),this._buildAxes(t),this._buildSplitLineAndArea(t)},e.prototype._buildAxes=function(t){var e=t.coordinateSystem,n=e.getIndicatorAxes(),i=H(n,(function(t){var n=new SI(t.model,{position:[e.cx,e.cy],rotation:t.angle,labelDirection:-1,tickDirection:-1,nameDirection:1});return n}));R(i,(function(t){R(dA,t.add,t),this.group.add(t.getGroup())}),this)},e.prototype._buildSplitLineAndArea=function(t){var e=t.coordinateSystem,n=e.getIndicatorAxes();if(n.length){var i=t.get("shape"),r=t.getModel("splitLine"),o=t.getModel("splitArea"),a=r.getModel("lineStyle"),s=o.getModel("areaStyle"),l=r.get("show"),u=o.get("show"),c=a.get("color"),d=s.get("color"),h=$(c)?c:[c],f=$(d)?d:[d],p=[],m=[];if("circle"===i)for(var g=n[0].getTicksCoords(),y=e.cx,v=e.cy,_=0;_<g.length;_++){if(l){var b=D(p,h,_);p[b].push(new Dc({shape:{cx:y,cy:v,r:g[_].coord}}))}if(u&&_<g.length-1){b=D(m,f,_);m[b].push(new Jc({shape:{cx:y,cy:v,r0:g[_].coord,r:g[_+1].coord}}))}}else{var x,w=H(n,(function(t,n){var i=t.getTicksCoords();return x=null==x?i.length-1:Math.min(i.length-1,x),H(i,(function(t){return e.coordToPoint(t.coord,n)}))})),M=[];for(_=0;_<=x;_++){for(var S=[],L=0;L<n.length;L++)S.push(w[L][_]);if(S[0]&&S.push(S[0].slice()),l){b=D(p,h,_);p[b].push(new sd({shape:{points:S}}))}if(u&&M){b=D(m,f,_-1);m[b].push(new rd({shape:{points:S.concat(M)}}))}M=S.slice().reverse()}}var k=a.getLineStyle(),T=s.getAreaStyle();R(m,(function(t,e){this.group.add(qd(t,{style:A({stroke:"none",fill:f[e%f.length]},T),silent:!0}))}),this),R(p,(function(t,e){this.group.add(qd(t,{style:A({fill:"none",stroke:h[e%h.length]},k),silent:!0}))}),this)}function D(t,e,n){var i=n%e.length;return t[i]=t[i]||[],i}},e.type="radar",e}(Py),fA=hA,pA=function(t){function e(e,n,i){var r=t.call(this,e,n,i)||this;return r.type="value",r.angle=0,r.name="",r}return o(e,t),e}(tI),mA=pA,gA=function(){function t(t,e,n){this.dimensions=[],this._model=t,this._indicatorAxes=H(t.getIndicatorModels(),(function(t,e){var n="indicator_"+e,i=new mA(n,new WD);return i.name=t.get("name"),i.model=t,t.axis=i,this.dimensions.push(n),i}),this),this.resize(t,n)}return t.prototype.getIndicatorAxes=function(){return this._indicatorAxes},t.prototype.dataToPoint=function(t,e){var n=this._indicatorAxes[e];return this.coordToPoint(n.dataToCoord(t),e)},t.prototype.coordToPoint=function(t,e){var n=this._indicatorAxes[e],i=n.angle,r=this.cx+t*Math.cos(i),o=this.cy-t*Math.sin(i);return[r,o]},t.prototype.pointToData=function(t){var e=t[0]-this.cx,n=t[1]-this.cy,i=Math.sqrt(e*e+n*n);e/=i,n/=i;for(var r,o=Math.atan2(-n,e),a=1/0,s=-1,l=0;l<this._indicatorAxes.length;l++){var u=this._indicatorAxes[l],c=Math.abs(o-u.angle);c<a&&(r=u,s=l,a=c)}return[s,+(r&&r.coordToData(i))]},t.prototype.resize=function(t,e){var n=t.get("center"),i=e.getWidth(),r=e.getHeight(),o=Math.min(i,r)/2;this.cx=Nr(n[0],i),this.cy=Nr(n[1],r),this.startAngle=t.get("startAngle")*Math.PI/180;var a=t.get("radius");"string"!==typeof a&&"number"!==typeof a||(a=[0,a]),this.r0=Nr(a[0],o),this.r=Nr(a[1],o),R(this._indicatorAxes,(function(t,e){t.setExtent(this.r0,this.r);var n=this.startAngle+e*Math.PI*2/this._indicatorAxes.length;n=Math.atan2(Math.sin(n),Math.cos(n)),t.angle=n}),this)},t.prototype.update=function(t,e){var n=this._indicatorAxes,i=this._model;R(n,(function(t){t.scale.setExtent(1/0,-1/0)})),t.eachSeriesByType("radar",(function(e,r){if("radar"===e.get("coordinateSystem")&&t.getComponent("radar",e.get("radarIndex"))===i){var o=e.getData();R(n,(function(t){t.scale.unionExtentFromData(o,o.mapDimension(t.dim))}))}}),this);var r=i.get("splitNumber");function o(t){var e=Math.pow(10,Math.floor(Math.log(t)/Math.LN10)),n=t/e;return 2===n?n=5:n*=2,n*e}R(n,(function(t,e){var n=gC(t.scale,t.model).extent;vC(t.scale,t.model);var i=t.model,a=t.scale,s=mC(a,i.get("min",!0)),l=mC(a,i.get("max",!0)),u=a.getInterval();if(null!=s&&null!=l)a.setExtent(+s,+l),a.setInterval((l-s)/r);else if(null!=s){var c=void 0;do{c=s+u*r,a.setExtent(+s,c),a.setInterval(u),u=o(u)}while(c<n[1]&&isFinite(c)&&isFinite(n[1]))}else if(null!=l){var d=void 0;do{d=l-u*r,a.setExtent(d,+l),a.setInterval(u),u=o(u)}while(d>n[0]&&isFinite(d)&&isFinite(n[0]))}else{var h=a.getTicks().length-1;h>r&&(u=o(u));c=Math.ceil(n[1]/u)*u,d=Rr(c-u*r);a.setExtent(d,c),a.setInterval(u)}}))},t.prototype.convertToPixel=function(t,e,n){return null},t.prototype.convertFromPixel=function(t,e,n){return null},t.prototype.containPoint=function(t){return!1},t.create=function(e,n){var i=[];return e.eachComponent("radar",(function(r){var o=new t(r,e,n);i.push(o),r.coordinateSystem=o})),e.eachSeriesByType("radar",(function(t){"radar"===t.get("coordinateSystem")&&(t.coordinateSystem=i[t.get("radarIndex")||0])})),i},t.dimensions=[],t}(),yA=gA;function vA(t){t.registerCoordinateSystem("radar",yA),t.registerComponentModel(cA),t.registerComponentView(fA),t.registerVisual({seriesType:"radar",reset:function(t){var e=t.getData();e.each((function(t){e.setItemVisual(t,"legendIcon","roundRect")})),e.setVisual("legendIcon","roundRect")}})}function _A(t){vM(vA),t.registerChartView(rA),t.registerSeriesModel(aA),t.registerLayout(KI),t.registerProcessor(zT("radar")),t.registerPreprocessor(eA)}var bA="\0_ec_interaction_mutex";function xA(t,e,n){var i=SA(t);i[e]=n}function wA(t,e,n){var i=SA(t),r=i[e];r===n&&(i[e]=null)}function MA(t,e){return!!SA(t)[e]}function SA(t){return t[bA]||(t[bA]={})}sM({type:"takeGlobalCursor",event:"globalCursorTaken",update:"update"},(function(){}));var LA=function(t){function e(e){var n=t.call(this)||this;n._zr=e;var i=V(n._mousedownHandler,n),r=V(n._mousemoveHandler,n),o=V(n._mouseupHandler,n),a=V(n._mousewheelHandler,n),s=V(n._pinchHandler,n);return n.enable=function(t,n){this.disable(),this._opt=A(T(n)||{},{zoomOnMouseWheel:!0,moveOnMouseMove:!0,moveOnMouseWheel:!1,preventDefaultMouseMove:!0}),null==t&&(t=!0),!0!==t&&"move"!==t&&"pan"!==t||(e.on("mousedown",i),e.on("mousemove",r),e.on("mouseup",o)),!0!==t&&"scale"!==t&&"zoom"!==t||(e.on("mousewheel",a),e.on("pinch",s))},n.disable=function(){e.off("mousedown",i),e.off("mousemove",r),e.off("mouseup",o),e.off("mousewheel",a),e.off("pinch",s)},n}return o(e,t),e.prototype.isDragging=function(){return this._dragging},e.prototype.isPinching=function(){return this._pinching},e.prototype.setPointerChecker=function(t){this.pointerChecker=t},e.prototype.dispose=function(){this.disable()},e.prototype._mousedownHandler=function(t){if(!(pe(t)||t.target&&t.target.draggable)){var e=t.offsetX,n=t.offsetY;this.pointerChecker&&this.pointerChecker(t,e,n)&&(this._x=e,this._y=n,this._dragging=!0)}},e.prototype._mousemoveHandler=function(t){if(this._dragging&&DA("moveOnMouseMove",t,this._opt)&&"pinch"!==t.gestureEvent&&!MA(this._zr,"globalPan")){var e=t.offsetX,n=t.offsetY,i=this._x,r=this._y,o=e-i,a=n-r;this._x=e,this._y=n,this._opt.preventDefaultMouseMove&&fe(t.event),TA(this,"pan","moveOnMouseMove",t,{dx:o,dy:a,oldX:i,oldY:r,newX:e,newY:n,isAvailableBehavior:null})}},e.prototype._mouseupHandler=function(t){pe(t)||(this._dragging=!1)},e.prototype._mousewheelHandler=function(t){var e=DA("zoomOnMouseWheel",t,this._opt),n=DA("moveOnMouseWheel",t,this._opt),i=t.wheelDelta,r=Math.abs(i),o=t.offsetX,a=t.offsetY;if(0!==i&&(e||n)){if(e){var s=r>3?1.4:r>1?1.2:1.1,l=i>0?s:1/s;kA(this,"zoom","zoomOnMouseWheel",t,{scale:l,originX:o,originY:a,isAvailableBehavior:null})}if(n){var u=Math.abs(i),c=(i>0?1:-1)*(u>3?.4:u>1?.15:.05);kA(this,"scrollMove","moveOnMouseWheel",t,{scrollDelta:c,originX:o,originY:a,isAvailableBehavior:null})}}},e.prototype._pinchHandler=function(t){if(!MA(this._zr,"globalPan")){var e=t.pinchScale>1?1.1:1/1.1;kA(this,"zoom",null,t,{scale:e,originX:t.pinchX,originY:t.pinchY,isAvailableBehavior:null})}},e}($t);function kA(t,e,n,i,r){t.pointerChecker&&t.pointerChecker(i,r.originX,r.originY)&&(fe(i.event),TA(t,e,n,i,r))}function TA(t,e,n,i,r){r.isAvailableBehavior=V(DA,null,n,i),t.trigger(e,r)}function DA(t,e,n){var i=n[t];return!t||i&&(!q(i)||e.event[i+"Key"])}var CA=LA;function IA(t,e,n){var i=t.target;i.x+=e,i.y+=n,i.dirty()}function AA(t,e,n,i){var r=t.target,o=t.zoomLimit,a=t.zoom=t.zoom||1;if(a*=e,o){var s=o.min||0,l=o.max||1/0;a=Math.max(Math.min(l,a),s)}var u=a/t.zoom;t.zoom=a,r.x-=(n-r.x)*(u-1),r.y-=(i-r.y)*(u-1),r.scaleX*=u,r.scaleY*=u,r.dirty()}var YA={axisPointer:1,tooltip:1,brush:1};function OA(t,e,n){var i=e.getComponentByElement(t.topTarget),r=i&&i.coordinateSystem;return i&&i!==n&&!YA.hasOwnProperty(i.mainType)&&r&&r.model!==n}var PA=["rect","circle","line","ellipse","polygon","polyline","path"],EA=gt(PA),NA=gt(PA.concat(["g"])),RA=gt(PA.concat(["g"])),HA=Io();function jA(t){var e=t.getItemStyle(),n=t.get("areaColor");return null!=n&&(e.fill=n),e}var zA=function(){function t(t){var e=new Mr;this.uid=Zh("ec_map_draw"),this._controller=new CA(t.getZr()),this._controllerHost={target:e},this.group=e,e.add(this._regionsGroup=new Mr),e.add(this._svgGroup=new Mr)}return t.prototype.draw=function(t,e,n,i,r){var o="geo"===t.mainType,a=t.getData&&t.getData();o&&e.eachComponent({mainType:"series",subType:"map"},(function(e){a||e.getHostGeoModel()!==t||(a=e.getData())}));var s=t.coordinateSystem,l=this._regionsGroup,u=this.group,c=s.getTransformInfo(),d=c.raw,h=c.roam,f=!l.childAt(0)||r;f?(u.x=h.x,u.y=h.y,u.scaleX=h.scaleX,u.scaleY=h.scaleY,u.dirty()):th(u,h,t);var p=a&&a.getVisual("visualMeta")&&a.getVisual("visualMeta").length>0,m={api:n,geo:s,mapOrGeoModel:t,data:a,isVisualEncodedByVisualMap:p,isGeo:o,transformInfoRaw:d};"geoJSON"===s.resourceType?this._buildGeoJSON(m):"geoSVG"===s.resourceType&&this._buildSVG(m),this._updateController(t,e,n),this._updateMapSelectHandler(t,l,n,i)},t.prototype._buildGeoJSON=function(t){var e=this._regionsGroupByName=gt(),n=gt(),i=this._regionsGroup,r=t.transformInfoRaw,o=t.mapOrGeoModel,a=t.data,s=function(t){return[t[0]*r.scaleX+r.x,t[1]*r.scaleY+r.y]};i.removeAll(),R(t.geo.regions,(function(r){var l=r.name,u=e.get(l),c=n.get(l)||{},d=c.dataIdx,h=c.regionModel;u||(u=e.set(l,new Mr),i.add(u),d=a?a.indexOfName(l):null,h=t.isGeo?o.getRegionModel(l):a?a.getItemModel(d):null,n.set(l,{dataIdx:d,regionModel:h}));var f=new xd({segmentIgnoreThreshold:1,shape:{paths:[]}});u.add(f),R(r.geometries,(function(t){if("polygon"===t.type){for(var e=[],n=0;n<t.exterior.length;++n)e.push(s(t.exterior[n]));f.shape.paths.push(new rd({segmentIgnoreThreshold:1,shape:{points:e}}));for(n=0;n<(t.interiors?t.interiors.length:0);++n){for(var i=t.interiors[n],r=[],o=0;o<i.length;++o)r.push(s(i[o]));f.shape.paths.push(new rd({segmentIgnoreThreshold:1,shape:{points:r}}))}}})),WA(t,f,d,h),f instanceof Ea&&(f.culling=!0);var p=s(r.getCenter());BA(t,f,l,h,o,d,p)})),e.each((function(e,i){var r=n.get(i),a=r.dataIdx,s=r.regionModel;FA(t,e,i,s,o,a),VA(t,e,i,s,o),GA(t,e,i,s,o)}),this)},t.prototype._buildSVG=function(t){var e=t.geo.map,n=t.transformInfoRaw;this._svgGroup.x=n.x,this._svgGroup.y=n.y,this._svgGroup.scaleX=n.scaleX,this._svgGroup.scaleY=n.scaleY,this._svgResourceChanged(e)&&(this._freeSVG(),this._useSVG(e));var i=this._svgDispatcherMap=gt(),r=!1;R(this._svgGraphicRecord.named,(function(e){var n=e.name,o=t.mapOrGeoModel,a=t.data,s=e.svgNodeTagLower,l=e.el,u=a?a.indexOfName(n):null,c=o.getRegionModel(n);if(null!=EA.get(s)&&l instanceof Ea&&WA(t,l,u,c),l instanceof Ea&&(l.culling=!0),l.z2EmphasisLift=0,!e.namedFrom&&(null!=RA.get(s)&&BA(t,l,n,c,o,u,null),FA(t,l,n,c,o,u),VA(t,l,n,c,o),null!=NA.get(s))){var d=GA(t,l,n,c,o);"self"===d&&(r=!0);var h=i.get(n)||i.set(n,[]);h.push(l)}}),this),this._enableBlurEntireSVG(r,t)},t.prototype._enableBlurEntireSVG=function(t,e){if(t&&e.isGeo){var n=e.mapOrGeoModel.getModel(["blur","itemStyle"]).getItemStyle(),i=n.opacity;this._svgGraphicRecord.root.traverse((function(t){if(!t.isGroup){ku(t);var e=t.ensureState("blur").style||{};null==e.opacity&&null!=i&&(e.opacity=i),t.ensureState("emphasis")}}))}},t.prototype.remove=function(){this._regionsGroup.removeAll(),this._regionsGroupByName=null,this._svgGroup.removeAll(),this._freeSVG(),this._controller.dispose(),this._controllerHost=null},t.prototype.findHighDownDispatchers=function(t,e){if(null==t)return[];var n=e.coordinateSystem;if("geoJSON"===n.resourceType){var i=this._regionsGroupByName;if(i){var r=i.get(t);return r?[r]:[]}}else if("geoSVG"===n.resourceType)return this._svgDispatcherMap&&this._svgDispatcherMap.get(t)||[]},t.prototype._svgResourceChanged=function(t){return this._svgMapName!==t},t.prototype._useSVG=function(t){var e=Ox.getGeoResource(t);if(e&&"geoSVG"===e.type){var n=e.useGraphic(this.uid);this._svgGroup.add(n.root),this._svgGraphicRecord=n,this._svgMapName=t}},t.prototype._freeSVG=function(){var t=this._svgMapName;if(null!=t){var e=Ox.getGeoResource(t);e&&"geoSVG"===e.type&&e.freeGraphic(this.uid),this._svgGraphicRecord=null,this._svgDispatcherMap=null,this._svgGroup.removeAll(),this._svgMapName=null}},t.prototype._updateController=function(t,e,n){var i=t.coordinateSystem,r=this._controller,o=this._controllerHost;o.zoomLimit=t.get("scaleLimit"),o.zoom=i.getZoom(),r.enable(t.get("roam")||!1);var a=t.mainType;function s(){var e={type:"geoRoam",componentType:a};return e[a+"Id"]=t.id,e}r.off("pan").on("pan",(function(t){this._mouseDownFlag=!1,IA(o,t.dx,t.dy),n.dispatchAction(I(s(),{dx:t.dx,dy:t.dy}))}),this),r.off("zoom").on("zoom",(function(t){this._mouseDownFlag=!1,AA(o,t.scale,t.originX,t.originY),n.dispatchAction(I(s(),{zoom:t.scale,originX:t.originX,originY:t.originY}))}),this),r.setPointerChecker((function(e,r,o){return i.containPoint([r,o])&&!OA(e,n,t)}))},t.prototype.resetForLabelLayout=function(){this.group.traverse((function(t){var e=t.getTextContent();e&&(e.ignore=HA(e).ignore)}))},t.prototype._updateMapSelectHandler=function(t,e,n,i){var r=this;e.off("mousedown"),e.off("click"),t.get("selectedMode")&&(e.on("mousedown",(function(){r._mouseDownFlag=!0})),e.on("click",(function(t){r._mouseDownFlag&&(r._mouseDownFlag=!1)})))},t}();function WA(t,e,n,i){var r=i.getModel("itemStyle"),o=i.getModel(["emphasis","itemStyle"]),a=i.getModel(["blur","itemStyle"]),s=i.getModel(["select","itemStyle"]),l=jA(r),u=jA(o),c=jA(s),d=jA(a),h=t.data;if(h){var f=h.getItemVisual(n,"style"),p=h.getItemVisual(n,"decal");t.isVisualEncodedByVisualMap&&f.fill&&(l.fill=f.fill),p&&(l.decal=wb(p,t.api))}e.setStyle(l),e.style.strokeNoScale=!0,e.ensureState("emphasis").style=u,e.ensureState("select").style=c,e.ensureState("blur").style=d,ku(e)}function BA(t,e,n,i,r,o,a){var s=t.data,l=t.isGeo,u=s&&isNaN(s.get(s.mapDimension("value"),o)),c=s&&s.getItemLayout(o);if(l||u||c&&c.showLabel){var d=l?n:o,h=void 0;(!s||o>=0)&&(h=r);var f=a?{normal:{align:"center",verticalAlign:"middle"}}:null;Mh(e,Sh(i),{labelFetcher:h,labelDataIndex:d,defaultText:n},f);var p=e.getTextContent();if(p&&(HA(p).ignore=p.ignore,e.textConfig&&a)){var m=e.getBoundingRect().clone();e.textConfig.layoutRect=m,e.textConfig.position=[(a[0]-m.x)/m.width*100+"%",(a[1]-m.y)/m.height*100+"%"]}e.disableLabelAnimation=!0}else e.removeTextContent(),e.removeTextConfig(),e.disableLabelAnimation=null}function FA(t,e,n,i,r,o){t.data?t.data.setItemGraphicEl(o,e):$l(e).eventData={componentType:"geo",componentIndex:r.componentIndex,geoIndex:r.componentIndex,name:n,region:i&&i.option||{}}}function VA(t,e,n,i,r){t.data||_h({el:e,componentModel:r,itemName:n,itemTooltipOption:i.get("tooltip")})}function GA(t,e,n,i,r){e.highDownSilentOnTouch=!!r.get("selectedMode");var o=i.getModel("emphasis"),a=o.get("focus");return $u(e,a,o.get("blurScope")),t.isGeo&&Qu(e,r,n),a}var $A=zA,UA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,i){if(!i||"mapToggleSelect"!==i.type||i.from!==this.uid){var r=this.group;if(r.removeAll(),!t.getHostGeoModel()){if(this._mapDraw&&i&&"geoRoam"===i.type&&this._mapDraw.resetForLabelLayout(),i&&"geoRoam"===i.type&&"series"===i.componentType&&i.seriesId===t.id){o=this._mapDraw;o&&r.add(o.group)}else if(t.needsDrawMap){var o=this._mapDraw||new $A(n);r.add(o.group),o.draw(t,e,n,this,i),this._mapDraw=o}else this._mapDraw&&this._mapDraw.remove(),this._mapDraw=null;t.get("showLegendSymbol")&&e.getComponent("legend")&&this._renderSymbols(t,e,n)}}},e.prototype.remove=function(){this._mapDraw&&this._mapDraw.remove(),this._mapDraw=null,this.group.removeAll()},e.prototype.dispose=function(){this._mapDraw&&this._mapDraw.remove(),this._mapDraw=null},e.prototype._renderSymbols=function(t,e,n){var i=t.originalData,r=this.group;i.each(i.mapDimension("value"),(function(e,n){if(!isNaN(e)){var o=i.getItemLayout(n);if(o&&o.point){var a=o.point,s=o.offset,l=new Dc({style:{fill:t.getData().getVisual("style").fill},shape:{cx:a[0]+9*s,cy:a[1],r:3},silent:!0,z2:8+(s?0:eu+1)});if(!s){var u=t.mainSeries.getData(),c=i.getName(n),d=u.indexOfName(c),h=i.getItemModel(n),f=h.getModel("label"),p=u.getItemGraphicEl(d);Mh(l,Sh(h),{labelFetcher:{getFormattedLabel:function(e,n){return t.getFormattedLabel(d,n)}}}),l.disableLabelAnimation=!0,f.get("position")||l.setTextConfig({position:"bottom"}),p.onHoverStateChange=function(t){bu(l,t)}}r.add(l)}}}))},e.type="map",e}(Vy),qA=UA,XA=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.needsDrawMap=!1,n.seriesGroup=[],n.getTooltipPosition=function(t){if(null!=t){var e=this.getData().getName(t),n=this.coordinateSystem,i=n.getRegion(e);return i&&n.dataToPoint(i.getCenter())}},n}return o(e,t),e.prototype.getInitialData=function(t){for(var e=ZT(this,{coordDimensions:["value"],encodeDefaulter:G(Hp,this)}),n=gt(),i=[],r=0,o=e.count();r<o;r++){var a=e.getName(r);n.set(a,!0)}var s=Ox.load(this.getMapType(),this.option.nameMap,this.option.nameProperty);return R(s.regions,(function(t){var e=t.name;n.get(e)||i.push(e)})),e.appendValues([],i),e},e.prototype.getHostGeoModel=function(){var t=this.option.geoIndex;return null!=t?this.ecModel.getComponent("geo",t):null},e.prototype.getMapType=function(){return(this.getHostGeoModel()||this).option.map},e.prototype.getRawValue=function(t){var e=this.getData();return e.get(e.mapDimension("value"),t)},e.prototype.getRegionModel=function(t){var e=this.getData();return e.getItemModel(e.indexOfName(t))},e.prototype.formatTooltip=function(t,e,n){for(var i=this.getData(),r=this.getRawValue(t),o=i.getName(t),a=this.seriesGroup,s=[],l=0;l<a.length;l++){var u=a[l].originalData.indexOfName(o),c=i.mapDimension("value");isNaN(a[l].originalData.get(c,u))||s.push(a[l].name)}return oy("section",{header:s.join(", "),noHeader:!s.length,blocks:[oy("nameValue",{name:o,value:r})]})},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.getLegendIcon=function(t){var e=t.icon||"roundRect",n=j_(e,0,0,t.itemWidth,t.itemHeight,t.itemStyle.fill);return n.setStyle(t.itemStyle),n.style.stroke="none",e.indexOf("empty")>-1&&(n.style.stroke=n.style.fill,n.style.fill="#fff",n.style.lineWidth=2),n},e.type="series.map",e.dependencies=["geo"],e.layoutMode="box",e.defaultOption={zlevel:0,z:2,coordinateSystem:"geo",map:"",left:"center",top:"center",aspectScale:null,showLegendSymbol:!0,boundingCoords:null,center:null,zoom:1,scaleLimit:null,selectedMode:!0,label:{show:!1,color:"#000"},itemStyle:{borderWidth:.5,borderColor:"#444",areaColor:"#eee"},emphasis:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{areaColor:"rgba(255,215,0,0.8)"}},select:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{color:"rgba(255,215,0,0.8)"}},nameProperty:"name"},e}(Yy),ZA=XA;function JA(t,e){var n={};return R(t,(function(t){t.each(t.mapDimension("value"),(function(e,i){var r="ec-"+t.getName(i);n[r]=n[r]||[],isNaN(e)||n[r].push(e)}))})),t[0].map(t[0].mapDimension("value"),(function(i,r){for(var o,a="ec-"+t[0].getName(r),s=0,l=1/0,u=-1/0,c=n[a].length,d=0;d<c;d++)l=Math.min(l,n[a][d]),u=Math.max(u,n[a][d]),s+=n[a][d];return o="min"===e?l:"max"===e?u:"average"===e?s/c:s,0===c?NaN:o}))}function KA(t){var e={};t.eachSeriesByType("map",(function(t){var n=t.getHostGeoModel(),i=n?"o"+n.id:"i"+t.getMapType();(e[i]=e[i]||[]).push(t)})),R(e,(function(t,e){for(var n=JA(H(t,(function(t){return t.getData()})),t[0].get("mapValueCalculation")),i=0;i<t.length;i++)t[i].originalData=t[i].getData();for(i=0;i<t.length;i++)t[i].seriesGroup=t,t[i].needsDrawMap=0===i&&!t[i].getHostGeoModel(),t[i].setData(n.cloneShallow()),t[i].mainSeries=t[0]}))}function QA(t){var e={};t.eachSeriesByType("map",(function(n){var i=n.getMapType();if(!n.getHostGeoModel()&&!e[i]){var r={};R(n.seriesGroup,(function(e){var n=e.coordinateSystem,i=e.originalData;e.get("showLegendSymbol")&&t.getComponent("legend")&&i.each(i.mapDimension("value"),(function(t,e){var o=i.getName(e),a=n.getRegion(o);if(a&&!isNaN(t)){var s=r[o]||0,l=n.dataToPoint(a.getCenter());r[o]=s+1,i.setItemLayout(e,{point:l,offset:s})}}))}));var o=n.getData();o.each((function(t){var e=o.getName(t),n=o.getItemLayout(t)||{};n.showLabel=!r[e],o.setItemLayout(t,n)})),e[i]=!0}}))}var tY=jt,eY=function(t){function e(e){var n=t.call(this)||this;return n.type="view",n.dimensions=["x","y"],n._roamTransformable=new ji,n._rawTransformable=new ji,n.name=e,n}return o(e,t),e.prototype.setBoundingRect=function(t,e,n,i){return this._rect=new Ji(t,e,n,i),this._rect},e.prototype.getBoundingRect=function(){return this._rect},e.prototype.setViewRect=function(t,e,n,i){this._transformTo(t,e,n,i),this._viewRect=new Ji(t,e,n,i)},e.prototype._transformTo=function(t,e,n,i){var r=this.getBoundingRect(),o=this._rawTransformable;o.transform=r.calculateTransform(new Ji(t,e,n,i));var a=o.parent;o.parent=null,o.decomposeTransform(),o.parent=a,this._updateTransform()},e.prototype.setCenter=function(t){t&&(this._center=t,this._updateCenterAndZoom())},e.prototype.setZoom=function(t){t=t||1;var e=this.zoomLimit;e&&(null!=e.max&&(t=Math.min(e.max,t)),null!=e.min&&(t=Math.max(e.min,t))),this._zoom=t,this._updateCenterAndZoom()},e.prototype.getDefaultCenter=function(){var t=this.getBoundingRect(),e=t.x+t.width/2,n=t.y+t.height/2;return[e,n]},e.prototype.getCenter=function(){return this._center||this.getDefaultCenter()},e.prototype.getZoom=function(){return this._zoom||1},e.prototype.getRoamTransform=function(){return this._roamTransformable.getLocalTransform()},e.prototype._updateCenterAndZoom=function(){var t=this._rawTransformable.getLocalTransform(),e=this._roamTransformable,n=this.getDefaultCenter(),i=this.getCenter(),r=this.getZoom();i=jt([],i,t),n=jt([],n,t),e.originX=i[0],e.originY=i[1],e.x=n[0]-i[0],e.y=n[1]-i[1],e.scaleX=e.scaleY=r,this._updateTransform()},e.prototype._updateTransform=function(){var t=this._roamTransformable,e=this._rawTransformable;e.parent=t,t.updateTransform(),e.updateTransform(),wi(this.transform||(this.transform=[]),e.transform||bi()),this._rawTransform=e.getLocalTransform(),this.invTransform=this.invTransform||[],Ti(this.invTransform,this.transform),this.decomposeTransform()},e.prototype.getTransformInfo=function(){var t=this._rawTransformable,e=this._roamTransformable,n=new ji;return n.transform=e.transform,n.decomposeTransform(),{roam:{x:n.x,y:n.y,scaleX:n.scaleX,scaleY:n.scaleY},raw:{x:t.x,y:t.y,scaleX:t.scaleX,scaleY:t.scaleY}}},e.prototype.getViewRect=function(){return this._viewRect},e.prototype.getViewRectAfterRoam=function(){var t=this.getBoundingRect().clone();return t.applyTransform(this.transform),t},e.prototype.dataToPoint=function(t,e,n){var i=e?this._rawTransform:this.transform;return n=n||[],i?tY(n,t,i):St(n,t)},e.prototype.pointToData=function(t){var e=this.invTransform;return e?tY([],t,e):[t[0],t[1]]},e.prototype.convertToPixel=function(t,e,n){var i=nY(e);return i===this?i.dataToPoint(n):null},e.prototype.convertFromPixel=function(t,e,n){var i=nY(e);return i===this?i.pointToData(n):null},e.prototype.containPoint=function(t){return this.getViewRectAfterRoam().contain(t[0],t[1])},e.dimensions=["x","y"],e}(ji);function nY(t){var e=t.seriesModel;return e?e.coordinateSystem:null}var iY=eY,rY={geoJSON:{aspectScale:.75,invertLongitute:!0},geoSVG:{aspectScale:1,invertLongitute:!1}},oY=function(t){function e(e,n,i){var r=t.call(this,e)||this;r.dimensions=["lng","lat"],r.type="geo",r._nameCoordMap=gt(),r.map=n;var o=Ox.load(n,i.nameMap,i.nameProperty),a=Ox.getGeoResource(n);r.resourceType=a?a.type:null;var s=rY[a.type];r._regionsMap=o.regionsMap,r._invertLongitute=s.invertLongitute,r.regions=o.regions,r.aspectScale=at(i.aspectScale,s.aspectScale);var l=o.boundingRect;return r.setBoundingRect(l.x,l.y,l.width,l.height),r}return o(e,t),e.prototype._transformTo=function(t,e,n,i){var r=this.getBoundingRect(),o=this._invertLongitute;r=r.clone(),o&&(r.y=-r.y-r.height);var a=this._rawTransformable;a.transform=r.calculateTransform(new Ji(t,e,n,i));var s=a.parent;a.parent=null,a.decomposeTransform(),a.parent=s,o&&(a.scaleY=-a.scaleY),this._updateTransform()},e.prototype.getRegion=function(t){return this._regionsMap.get(t)},e.prototype.getRegionByCoord=function(t){for(var e=this.regions,n=0;n<e.length;n++){var i=e[n];if("geoJSON"===i.type&&i.contain(t))return e[n]}},e.prototype.addGeoCoord=function(t,e){this._nameCoordMap.set(t,e)},e.prototype.getGeoCoord=function(t){var e=this._regionsMap.get(t);return this._nameCoordMap.get(t)||e&&e.getCenter()},e.prototype.dataToPoint=function(t,e,n){if("string"===typeof t&&(t=this.getGeoCoord(t)),t)return iY.prototype.dataToPoint.call(this,t,e,n)},e.prototype.convertToPixel=function(t,e,n){var i=aY(e);return i===this?i.dataToPoint(n):null},e.prototype.convertFromPixel=function(t,e,n){var i=aY(e);return i===this?i.pointToData(n):null},e}(iY);function aY(t){var e=t.geoModel,n=t.seriesModel;return e?e.coordinateSystem:n?n.coordinateSystem||(n.getReferringComponents("geo",Po).models[0]||{}).coordinateSystem:null}E(oY,iY);var sY=oY;function lY(t,e){var n=t.get("boundingCoords");if(null!=n){var i=n[0],r=n[1];isNaN(i[0])||isNaN(i[1])||isNaN(r[0])||isNaN(r[1])||this.setBoundingRect(i[0],i[1],r[0]-i[0],r[1]-i[1])}var o,a,s,l=this.getBoundingRect(),u=t.get("layoutCenter"),c=t.get("layoutSize"),d=e.getWidth(),h=e.getHeight(),f=l.width/l.height*this.aspectScale,p=!1;if(u&&c&&(o=[Nr(u[0],d),Nr(u[1],h)],a=Nr(c,Math.min(d,h)),isNaN(o[0])||isNaN(o[1])||isNaN(a)||(p=!0)),p)s={},f>1?(s.width=a,s.height=a/f):(s.height=a,s.width=a*f),s.y=o[1]-s.height/2,s.x=o[0]-s.width/2;else{var m=t.getBoxLayoutParams();m.aspect=f,s=dp(m,{width:d,height:h})}this.setViewRect(s.x,s.y,s.width,s.height),this.setCenter(t.get("center")),this.setZoom(t.get("zoom"))}function uY(t,e){R(e.get("geoCoord"),(function(e,n){t.addGeoCoord(n,e)}))}var cY=function(){function t(){this.dimensions=sY.prototype.dimensions}return t.prototype.create=function(t,e){var n=[];t.eachComponent("geo",(function(t,i){var r=t.get("map"),o=new sY(r+i,r,{nameMap:t.get("nameMap"),nameProperty:t.get("nameProperty"),aspectScale:t.get("aspectScale")});o.zoomLimit=t.get("scaleLimit"),n.push(o),t.coordinateSystem=o,o.model=t,o.resize=lY,o.resize(t,e)})),t.eachSeries((function(t){var e=t.get("coordinateSystem");if("geo"===e){var i=t.get("geoIndex")||0;t.coordinateSystem=n[i]}}));var i={};return t.eachSeriesByType("map",(function(t){if(!t.getHostGeoModel()){var e=t.getMapType();i[e]=i[e]||[],i[e].push(t)}})),R(i,(function(t,i){var r=H(t,(function(t){return t.get("nameMap")})),o=new sY(i,i,{nameMap:C(r),nameProperty:t[0].get("nameProperty"),aspectScale:t[0].get("aspectScale")});o.zoomLimit=ot.apply(null,H(t,(function(t){return t.get("scaleLimit")}))),n.push(o),o.resize=lY,o.resize(t[0],e),R(t,(function(t){t.coordinateSystem=o,uY(o,t)}))})),n},t.prototype.getFilledRegions=function(t,e,n,i){for(var r=(t||[]).slice(),o=gt(),a=0;a<r.length;a++)o.set(r[a].name,r[a]);var s=Ox.load(e,n,i);return R(s.regions,(function(t){var e=t.name;!o.get(e)&&r.push({name:e})})),r},t}(),dY=new cY,hY=dY,fY=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e,n){var i=Ox.getGeoResource(t.map);if(i&&"geoJSON"===i.type){var r=t.itemStyle=t.itemStyle||{};"color"in r||(r.color="#eee")}this.mergeDefaultAndTheme(t,n),lo(t,"label",["show"])},e.prototype.optionUpdated=function(){var t=this,e=this.option;e.regions=hY.getFilledRegions(e.regions,e.map,e.nameMap,e.nameProperty);var n={};this._optionModelMap=j(e.regions||[],(function(e,i){var r=i.name;return r&&(e.set(r,new qh(i,t,t.ecModel)),i.selected&&(n[r]=!0)),e}),gt()),e.selectedMap||(e.selectedMap=n)},e.prototype.getRegionModel=function(t){return this._optionModelMap.get(t)||new qh(null,this,this.ecModel)},e.prototype.getFormattedLabel=function(t,e){var n=this.getRegionModel(t),i="normal"===e?n.get(["label","formatter"]):n.get(["emphasis","label","formatter"]),r={name:t};return"function"===typeof i?(r.status=e,i(r)):"string"===typeof i?i.replace("{a}",null!=t?t:""):void 0},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.select=function(t){var e=this.option,n=e.selectedMode;if(n){"multiple"!==n&&(e.selectedMap=null);var i=e.selectedMap||(e.selectedMap={});i[t]=!0}},e.prototype.unSelect=function(t){var e=this.option.selectedMap;e&&(e[t]=!1)},e.prototype.toggleSelected=function(t){this[this.isSelected(t)?"unSelect":"select"](t)},e.prototype.isSelected=function(t){var e=this.option.selectedMap;return!(!e||!e[t])},e.type="geo",e.layoutMode="box",e.defaultOption={zlevel:0,z:0,show:!0,left:"center",top:"center",aspectScale:null,silent:!1,map:"",boundingCoords:null,center:null,zoom:1,scaleLimit:null,label:{show:!1,color:"#000"},itemStyle:{borderWidth:.5,borderColor:"#444"},emphasis:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{color:"rgba(255,215,0,0.8)"}},select:{label:{show:!0,color:"rgb(100,0,0)"},itemStyle:{color:"rgba(255,215,0,0.8)"}},regions:[]},e}(xp),pY=fY;function mY(t,e,n){var i=t.getZoom(),r=t.getCenter(),o=e.zoom,a=t.dataToPoint(r);if(null!=e.dx&&null!=e.dy&&(a[0]-=e.dx,a[1]-=e.dy,t.setCenter(t.pointToData(a))),null!=o){if(n){var s=n.min||0,l=n.max||1/0;o=Math.max(Math.min(i*o,l),s)/i}t.scaleX*=o,t.scaleY*=o;var u=(e.originX-t.x)*(o-1),c=(e.originY-t.y)*(o-1);t.x-=u,t.y-=c,t.updateTransform(),t.setCenter(t.pointToData(a)),t.setZoom(o*i)}return{center:t.getCenter(),zoom:t.getZoom()}}var gY=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.focusBlurEnabled=!0,n}return o(e,t),e.prototype.init=function(t,e){var n=new $A(e);this._mapDraw=n,this.group.add(n.group),this._api=e},e.prototype.render=function(t,e,n,i){var r=this._mapDraw;t.get("show")?r.draw(t,e,n,this,i):this._mapDraw.group.removeAll(),r.group.on("click",this._handleRegionClick,this),r.group.silent=t.get("silent"),this._model=t,this.updateSelectStatus(t,e,n)},e.prototype._handleRegionClick=function(t){var e;k_(t.target,(function(t){return null!=(e=$l(t).eventData)}),!0),e&&this._api.dispatchAction({type:"geoToggleSelect",geoId:this._model.id,name:e.name})},e.prototype.updateSelectStatus=function(t,e,n){var i=this;this._mapDraw.group.traverse((function(t){var e=$l(t).eventData;if(e)return i._model.isSelected(e.name)?n.enterSelect(t):n.leaveSelect(t),!0}))},e.prototype.findHighDownDispatchers=function(t){return this._mapDraw&&this._mapDraw.findHighDownDispatchers(t,this._model)},e.prototype.dispose=function(){this._mapDraw&&this._mapDraw.remove()},e.type="geo",e}(Py),yY=gY;function vY(t){function e(e,n){n.update="geo:updateSelectStatus",t.registerAction(n,(function(t,n){var i={},r=[];return n.eachComponent({mainType:"geo",query:t},(function(n){n[e](t.name);var o=n.coordinateSystem;R(o.regions,(function(t){i[t.name]=n.isSelected(t.name)||!1}));var a=[];R(i,(function(t,e){i[e]&&a.push(e)})),r.push({geoIndex:n.componentIndex,name:a})})),{selected:i,allSelected:r,name:t.name}}))}t.registerCoordinateSystem("geo",hY),t.registerComponentModel(pY),t.registerComponentView(yY),e("toggleSelected",{type:"geoToggleSelect",event:"geoselectchanged"}),e("select",{type:"geoSelect",event:"geoselected"}),e("unSelect",{type:"geoUnSelect",event:"geounselected"}),t.registerAction({type:"geoRoam",event:"geoRoam",update:"updateTransform"},(function(t,e){var n=t.componentType||"series";e.eachComponent({mainType:n,query:t},(function(e){var i=e.coordinateSystem;if("geo"===i.type){var r=mY(i,t,e.get("scaleLimit"));e.setCenter&&e.setCenter(r.center),e.setZoom&&e.setZoom(r.zoom),"series"===n&&R(e.seriesGroup,(function(t){t.setCenter(r.center),t.setZoom(r.zoom)}))}}))}))}function _Y(t){vM(vY),t.registerChartView(qA),t.registerSeriesModel(ZA),t.registerLayout(QA),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,KA),M_("map",t.registerAction)}function bY(t){var e=t;e.hierNode={defaultAncestor:null,ancestor:e,prelim:0,modifier:0,change:0,shift:0,i:0,thread:null};var n,i,r=[e];while(n=r.pop())if(i=n.children,n.isExpand&&i.length)for(var o=i.length,a=o-1;a>=0;a--){var s=i[a];s.hierNode={defaultAncestor:null,ancestor:s,prelim:0,modifier:0,change:0,shift:0,i:a,thread:null},r.push(s)}}function xY(t,e){var n=t.isExpand?t.children:[],i=t.parentNode.children,r=t.hierNode.i?i[t.hierNode.i-1]:null;if(n.length){kY(t);var o=(n[0].hierNode.prelim+n[n.length-1].hierNode.prelim)/2;r?(t.hierNode.prelim=r.hierNode.prelim+e(t,r),t.hierNode.modifier=t.hierNode.prelim-o):t.hierNode.prelim=o}else r&&(t.hierNode.prelim=r.hierNode.prelim+e(t,r));t.parentNode.hierNode.defaultAncestor=TY(t,r,t.parentNode.hierNode.defaultAncestor||i[0],e)}function wY(t){var e=t.hierNode.prelim+t.parentNode.hierNode.modifier;t.setLayout({x:e},!0),t.hierNode.modifier+=t.parentNode.hierNode.modifier}function MY(t){return arguments.length?t:YY}function SY(t,e){return t-=Math.PI/2,{x:e*Math.cos(t),y:e*Math.sin(t)}}function LY(t,e){return dp(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function kY(t){var e=t.children,n=e.length,i=0,r=0;while(--n>=0){var o=e[n];o.hierNode.prelim+=i,o.hierNode.modifier+=i,r+=o.hierNode.change,i+=o.hierNode.shift+r}}function TY(t,e,n,i){if(e){var r=t,o=t,a=o.parentNode.children[0],s=e,l=r.hierNode.modifier,u=o.hierNode.modifier,c=a.hierNode.modifier,d=s.hierNode.modifier;while(s=DY(s),o=CY(o),s&&o){r=DY(r),a=CY(a),r.hierNode.ancestor=t;var h=s.hierNode.prelim+d-o.hierNode.prelim-u+i(s,o);h>0&&(AY(IY(s,t,n),t,h),u+=h,l+=h),d+=s.hierNode.modifier,u+=o.hierNode.modifier,l+=r.hierNode.modifier,c+=a.hierNode.modifier}s&&!DY(r)&&(r.hierNode.thread=s,r.hierNode.modifier+=d-l),o&&!CY(a)&&(a.hierNode.thread=o,a.hierNode.modifier+=u-c,n=t)}return n}function DY(t){var e=t.children;return e.length&&t.isExpand?e[e.length-1]:t.hierNode.thread}function CY(t){var e=t.children;return e.length&&t.isExpand?e[0]:t.hierNode.thread}function IY(t,e,n){return t.hierNode.ancestor.parentNode===e.parentNode?t.hierNode.ancestor:n}function AY(t,e,n){var i=n/(e.hierNode.i-t.hierNode.i);e.hierNode.change-=i,e.hierNode.shift+=n,e.hierNode.modifier+=n,e.hierNode.prelim+=n,t.hierNode.change+=i}function YY(t,e){return t.parentNode===e.parentNode?1:2}var OY=function(){function t(){this.parentPoint=[],this.childPoints=[]}return t}(),PY=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new OY},e.prototype.buildPath=function(t,e){var n=e.childPoints,i=n.length,r=e.parentPoint,o=n[0],a=n[i-1];if(1===i)return t.moveTo(r[0],r[1]),void t.lineTo(o[0],o[1]);var s=e.orient,l="TB"===s||"BT"===s?0:1,u=1-l,c=Nr(e.forkPosition,1),d=[];d[l]=r[l],d[u]=r[u]+(a[u]-r[u])*c,t.moveTo(r[0],r[1]),t.lineTo(d[0],d[1]),t.moveTo(o[0],o[1]),d[l]=o[l],t.lineTo(d[0],d[1]),d[l]=a[l],t.lineTo(d[0],d[1]),t.lineTo(a[0],a[1]);for(var h=1;h<i-1;h++){var f=n[h];t.moveTo(f[0],f[1]),d[l]=f[l],t.lineTo(d[0],d[1])}},e}(pl),EY=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._mainGroup=new Mr,n}return o(e,t),e.prototype.init=function(t,e){this._controller=new CA(e.getZr()),this._controllerHost={target:this.group},this.group.add(this._mainGroup)},e.prototype.render=function(t,e,n){var i=t.getData(),r=t.layoutInfo,o=this._mainGroup,a=t.get("layout");"radial"===a?(o.x=r.x+r.width/2,o.y=r.y+r.height/2):(o.x=r.x,o.y=r.y),this._updateViewCoordSys(t),this._updateController(t,e,n);var s=this._data;i.diff(s).add((function(e){NY(i,e)&&RY(i,e,null,o,t)})).update((function(e,n){var r=s.getItemGraphicEl(n);NY(i,e)?RY(i,e,r,o,t):r&&jY(s,n,r,o,t)})).remove((function(e){var n=s.getItemGraphicEl(e);n&&jY(s,e,n,o,t)})).execute(),this._nodeScaleRatio=t.get("nodeScaleRatio"),this._updateNodeAndLinkScale(t),!0===t.get("expandAndCollapse")&&i.eachItemGraphicEl((function(e,i){e.off("click").on("click",(function(){n.dispatchAction({type:"treeExpandAndCollapse",seriesId:t.id,dataIndex:i})}))})),this._data=i},e.prototype._updateViewCoordSys=function(t){var e=t.getData(),n=[];e.each((function(t){var i=e.getItemLayout(t);!i||isNaN(i.x)||isNaN(i.y)||n.push([+i.x,+i.y])}));var i=[],r=[];ms(n,i,r);var o=this._min,a=this._max;r[0]-i[0]===0&&(i[0]=o?o[0]:i[0]-1,r[0]=a?a[0]:r[0]+1),r[1]-i[1]===0&&(i[1]=o?o[1]:i[1]-1,r[1]=a?a[1]:r[1]+1);var s=t.coordinateSystem=new iY;s.zoomLimit=t.get("scaleLimit"),s.setBoundingRect(i[0],i[1],r[0]-i[0],r[1]-i[1]),s.setCenter(t.get("center")),s.setZoom(t.get("zoom")),this.group.attr({x:s.x,y:s.y,scaleX:s.scaleX,scaleY:s.scaleY}),this._min=i,this._max=r},e.prototype._updateController=function(t,e,n){var i=this,r=this._controller,o=this._controllerHost,a=this.group;r.setPointerChecker((function(e,i,r){var o=a.getBoundingRect();return o.applyTransform(a.transform),o.contain(i,r)&&!OA(e,n,t)})),r.enable(t.get("roam")),o.zoomLimit=t.get("scaleLimit"),o.zoom=t.coordinateSystem.getZoom(),r.off("pan").off("zoom").on("pan",(function(e){IA(o,e.dx,e.dy),n.dispatchAction({seriesId:t.id,type:"treeRoam",dx:e.dx,dy:e.dy})})).on("zoom",(function(e){AA(o,e.scale,e.originX,e.originY),n.dispatchAction({seriesId:t.id,type:"treeRoam",zoom:e.scale,originX:e.originX,originY:e.originY}),i._updateNodeAndLinkScale(t),n.updateLabelLayout()}))},e.prototype._updateNodeAndLinkScale=function(t){var e=t.getData(),n=this._getNodeGlobalScale(t);e.eachItemGraphicEl((function(t,e){t.setSymbolScale(n)}))},e.prototype._getNodeGlobalScale=function(t){var e=t.coordinateSystem;if("view"!==e.type)return 1;var n=this._nodeScaleRatio,i=e.scaleX||1,r=e.getZoom(),o=(r-1)*n+1;return o/i},e.prototype.dispose=function(){this._controller&&this._controller.dispose(),this._controllerHost=null},e.prototype.remove=function(){this._mainGroup.removeAll(),this._data=null},e.type="tree",e}(Vy);function NY(t,e){var n=t.getItemLayout(e);return n&&!isNaN(n.x)&&!isNaN(n.y)}function RY(t,e,n,i,r){var o=!n,a=t.tree.getNodeByDataIndex(e),s=a.getModel(),l=a.getVisual("style").fill,u=!1===a.isExpand&&0!==a.children.length?l:"#fff",c=t.tree.root,d=a.parentNode===c?a:a.parentNode||a,h=t.getItemGraphicEl(d.dataIndex),f=d.getLayout(),p=h?{x:h.__oldX,y:h.__oldY,rawX:h.__radialOldRawX,rawY:h.__radialOldRawY}:f,m=a.getLayout();o?(n=new XL(t,e,null,{symbolInnerColor:u,useNameLabel:!0}),n.x=p.x,n.y=p.y):n.updateData(t,e,null,{symbolInnerColor:u,useNameLabel:!0}),n.__radialOldRawX=n.__radialRawX,n.__radialOldRawY=n.__radialRawY,n.__radialRawX=m.rawX,n.__radialRawY=m.rawY,i.add(n),t.setItemGraphicEl(e,n),n.__oldX=n.x,n.__oldY=n.y,th(n,{x:m.x,y:m.y},r);var g=n.getSymbolPath();if("radial"===r.get("layout")){var y=c.children[0],v=y.getLayout(),_=y.children.length,b=void 0,x=void 0;if(m.x===v.x&&!0===a.isExpand){var w={x:(y.children[0].getLayout().x+y.children[_-1].getLayout().x)/2,y:(y.children[0].getLayout().y+y.children[_-1].getLayout().y)/2};b=Math.atan2(w.y-v.y,w.x-v.x),b<0&&(b=2*Math.PI+b),x=w.x<v.x,x&&(b-=Math.PI)}else b=Math.atan2(m.y-v.y,m.x-v.x),b<0&&(b=2*Math.PI+b),0===a.children.length||0!==a.children.length&&!1===a.isExpand?(x=m.x<v.x,x&&(b-=Math.PI)):(x=m.x>v.x,x||(b-=Math.PI));var M=x?"left":"right",S=s.getModel("label"),L=S.get("rotate"),k=L*(Math.PI/180),T=g.getTextContent();T&&(g.setTextConfig({position:S.get("position")||M,rotation:null==L?-b:k,origin:"center"}),T.setStyle("verticalAlign","middle"))}var D=s.get(["emphasis","focus"]),C="ancestor"===D?a.getAncestorsIndices():"descendant"===D?a.getDescendantIndices():null;C&&($l(n).focus=C),HY(r,a,c,n,p,f,m,i),n.__edge&&(n.onHoverStateChange=function(e){if("blur"!==e){var i=a.parentNode&&t.getItemGraphicEl(a.parentNode.dataIndex);i&&i.hoverState===Jl||bu(n.__edge,e)}})}function HY(t,e,n,i,r,o,a,s){var l=e.getModel(),u=t.get("edgeShape"),c=t.get("layout"),d=t.getOrient(),h=t.get(["lineStyle","curveness"]),f=t.get("edgeForkPosition"),p=l.getModel("lineStyle").getLineStyle(),m=i.__edge;if("curve"===u)e.parentNode&&e.parentNode!==n&&(m||(m=i.__edge=new gd({shape:zY(c,d,h,r,r)})),th(m,{shape:zY(c,d,h,o,a)},t));else if("polyline"===u)if("orthogonal"===c){if(e!==n&&e.children&&0!==e.children.length&&!0===e.isExpand){for(var g=e.children,y=[],v=0;v<g.length;v++){var _=g[v].getLayout();y.push([_.x,_.y])}m||(m=i.__edge=new PY({shape:{parentPoint:[a.x,a.y],childPoints:[[a.x,a.y]],orient:d,forkPosition:f}})),th(m,{shape:{parentPoint:[a.x,a.y],childPoints:y}},t)}}else 0;m&&(m.useStyle(A({strokeNoScale:!0,fill:null},p)),Zu(m,l,"lineStyle"),ku(m),s.add(m))}function jY(t,e,n,i,r){var o,a=t.tree.getNodeByDataIndex(e),s=t.tree.root,l=a.parentNode===s?a:a.parentNode||a;while(o=l.getLayout(),null==o)l=l.parentNode===s?l:l.parentNode||l;var u={duration:r.get("animationDurationUpdate"),easing:r.get("animationEasingUpdate")};nh(n,{x:o.x+1,y:o.y+1},r,{cb:function(){i.remove(n),t.setItemGraphicEl(e,null)},removeOpt:u}),n.fadeOut(null,{fadeLabel:!0,animation:u});var c=t.getItemGraphicEl(l.dataIndex),d=c.__edge,h=n.__edge||(!1===l.isExpand||1===l.children.length?d:void 0),f=r.get("edgeShape"),p=r.get("layout"),m=r.get("orient"),g=r.get(["lineStyle","curveness"]);h&&("curve"===f?nh(h,{shape:zY(p,m,g,o,o),style:{opacity:0}},r,{cb:function(){i.remove(h)},removeOpt:u}):"polyline"===f&&"orthogonal"===r.get("layout")&&nh(h,{shape:{parentPoint:[o.x,o.y],childPoints:[[o.x,o.y]]},style:{opacity:0}},r,{cb:function(){i.remove(h)},removeOpt:u}))}function zY(t,e,n,i,r){var o,a,s,l,u,c,d,h;if("radial"===t){u=i.rawX,d=i.rawY,c=r.rawX,h=r.rawY;var f=SY(u,d),p=SY(u,d+(h-d)*n),m=SY(c,h+(d-h)*n),g=SY(c,h);return{x1:f.x||0,y1:f.y||0,x2:g.x||0,y2:g.y||0,cpx1:p.x||0,cpy1:p.y||0,cpx2:m.x||0,cpy2:m.y||0}}return u=i.x,d=i.y,c=r.x,h=r.y,"LR"!==e&&"RL"!==e||(o=u+(c-u)*n,a=d,s=c+(u-c)*n,l=h),"TB"!==e&&"BT"!==e||(o=u,a=d+(h-d)*n,s=c,l=h+(d-h)*n),{x1:u,y1:d,x2:c,y2:h,cpx1:o,cpy1:a,cpx2:s,cpy2:l}}var WY=EY,BY=Io();function FY(t){var e=t.mainData,n=t.datas;n||(n={main:e},t.datasAttr={main:"data"}),t.datas=t.mainData=null,ZY(e,n,t),R(n,(function(n){R(e.TRANSFERABLE_METHODS,(function(e){n.wrapMethod(e,G(VY,t))}))})),e.wrapMethod("cloneShallow",G($Y,t)),R(e.CHANGABLE_METHODS,(function(n){e.wrapMethod(n,G(GY,t))})),ct(n[e.dataType]===e)}function VY(t,e){if(XY(this)){var n=I({},BY(this).datas);n[this.dataType]=e,ZY(e,n,t)}else JY(e,this.dataType,BY(this).mainData,t);return e}function GY(t,e){return t.struct&&t.struct.update(),e}function $Y(t,e){return R(BY(e).datas,(function(n,i){n!==e&&JY(n.cloneShallow(),i,e,t)})),e}function UY(t){var e=BY(this).mainData;return null==t||null==e?e:BY(e).datas[t]}function qY(){var t=BY(this).mainData;return null==t?[{data:t}]:H(B(BY(t).datas),(function(e){return{type:e,data:BY(t).datas[e]}}))}function XY(t){return BY(t).mainData===t}function ZY(t,e,n){BY(t).datas={},R(e,(function(e,i){JY(e,i,t,n)}))}function JY(t,e,n,i){BY(n).datas[e]=t,BY(t).mainData=n,t.dataType=e,i.struct&&(t[i.structAttr]=i.struct,i.struct[i.datasAttr[e]]=t),t.getLinkedData=UY,t.getLinkedDataAll=qY}var KY=FY,QY=function(){function t(t,e){this.depth=0,this.height=0,this.dataIndex=-1,this.children=[],this.viewChildren=[],this.isExpand=!1,this.name=t||"",this.hostTree=e}return t.prototype.isRemoved=function(){return this.dataIndex<0},t.prototype.eachNode=function(t,e,n){"function"===typeof t&&(n=e,e=t,t=null),t=t||{},q(t)&&(t={order:t});var i,r=t.order||"preorder",o=this[t.attr||"children"];"preorder"===r&&(i=e.call(n,this));for(var a=0;!i&&a<o.length;a++)o[a].eachNode(t,e,n);"postorder"===r&&e.call(n,this)},t.prototype.updateDepthAndHeight=function(t){var e=0;this.depth=t;for(var n=0;n<this.children.length;n++){var i=this.children[n];i.updateDepthAndHeight(t+1),i.height>e&&(e=i.height)}this.height=e+1},t.prototype.getNodeById=function(t){if(this.getId()===t)return this;for(var e=0,n=this.children,i=n.length;e<i;e++){var r=n[e].getNodeById(t);if(r)return r}},t.prototype.contains=function(t){if(t===this)return!0;for(var e=0,n=this.children,i=n.length;e<i;e++){var r=n[e].contains(t);if(r)return r}},t.prototype.getAncestors=function(t){var e=[],n=t?this:this.parentNode;while(n)e.push(n),n=n.parentNode;return e.reverse(),e},t.prototype.getAncestorsIndices=function(){var t=[],e=this;while(e)t.push(e.dataIndex),e=e.parentNode;return t.reverse(),t},t.prototype.getDescendantIndices=function(){var t=[];return this.eachNode((function(e){t.push(e.dataIndex)})),t},t.prototype.getValue=function(t){var e=this.hostTree.data;return e.get(e.getDimension(t||"value"),this.dataIndex)},t.prototype.setLayout=function(t,e){this.dataIndex>=0&&this.hostTree.data.setItemLayout(this.dataIndex,t,e)},t.prototype.getLayout=function(){return this.hostTree.data.getItemLayout(this.dataIndex)},t.prototype.getModel=function(t){if(!(this.dataIndex<0)){var e=this.hostTree,n=e.data.getItemModel(this.dataIndex);return n.getModel(t)}},t.prototype.getLevelModel=function(){return(this.hostTree.levelModels||[])[this.depth]},t.prototype.setVisual=function(t,e){this.dataIndex>=0&&this.hostTree.data.setItemVisual(this.dataIndex,t,e)},t.prototype.getVisual=function(t){return this.hostTree.data.getItemVisual(this.dataIndex,t)},t.prototype.getRawIndex=function(){return this.hostTree.data.getRawIndex(this.dataIndex)},t.prototype.getId=function(){return this.hostTree.data.getId(this.dataIndex)},t.prototype.isAncestorOf=function(t){var e=t.parentNode;while(e){if(e===this)return!0;e=e.parentNode}return!1},t.prototype.isDescendantOf=function(t){return t!==this&&t.isAncestorOf(this)},t}(),tO=function(){function t(t){this.type="tree",this._nodes=[],this.hostModel=t}return t.prototype.eachNode=function(t,e,n){this.root.eachNode(t,e,n)},t.prototype.getNodeByDataIndex=function(t){var e=this.data.getRawIndex(t);return this._nodes[e]},t.prototype.getNodeById=function(t){return this.root.getNodeById(t)},t.prototype.update=function(){for(var t=this.data,e=this._nodes,n=0,i=e.length;n<i;n++)e[n].dataIndex=-1;for(n=0,i=t.count();n<i;n++)e[t.getRawIndex(n)].dataIndex=n},t.prototype.clearLayouts=function(){this.data.clearItemLayouts()},t.createTree=function(e,n,i){var r=new t(n),o=[],a=1;function s(t,e){var n=t.value;a=Math.max(a,$(n)?n.length:1),o.push(t);var i=new QY(wo(t.name,""),r);e?eO(i,e):r.root=i,r._nodes.push(i);var l=t.children;if(l)for(var u=0;u<l.length;u++)s(l[u],i)}s(e),r.root.updateDepthAndHeight(0);var l=AL(o,{coordDimensions:["value"],dimensionsCount:a}),u=new kL(l,n);return u.initData(o),i&&i(u),KY({mainData:u,struct:r,structAttr:"tree"}),r.update(),r},t}();function eO(t,e){var n=e.children;t.parentNode!==e&&(n.push(t),t.parentNode=e)}var nO=tO;function iO(t,e,n){if(t&&O(e,t.type)>=0){var i=n.getData().tree.root,r=t.targetNode;if("string"===typeof r&&(r=i.getNodeById(r)),r&&i.contains(r))return{node:r};var o=t.targetNodeId;if(null!=o&&(r=i.getNodeById(o)))return{node:r}}}function rO(t){var e=[];while(t)t=t.parentNode,t&&e.push(t);return e.reverse()}function oO(t,e){var n=rO(t);return O(n,e)>=0}function aO(t,e){var n=[];while(t){var i=t.dataIndex;n.push({name:t.name,dataIndex:i,value:e.getRawValue(i)}),t=t.parentNode}return n.reverse(),n}var sO=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.hasSymbolVisual=!0,e.ignoreStyleOnData=!0,e}return o(e,t),e.prototype.getInitialData=function(t){var e={name:t.name,children:t.data},n=t.leaves||{},i=new qh(n,this,this.ecModel),r=nO.createTree(e,this,o);function o(t){t.wrapMethod("getItemModel",(function(t,e){var n=r.getNodeByDataIndex(e);return n&&n.children.length&&n.isExpand||(t.parentModel=i),t}))}var a=0;r.eachNode("preorder",(function(t){t.depth>a&&(a=t.depth)}));var s=t.expandAndCollapse,l=s&&t.initialTreeDepth>=0?t.initialTreeDepth:a;return r.root.eachNode("preorder",(function(t){var e=t.hostTree.data.getRawDataItem(t.dataIndex);t.isExpand=e&&null!=e.collapsed?!e.collapsed:t.depth<=l})),r.data},e.prototype.getOrient=function(){var t=this.get("orient");return"horizontal"===t?t="LR":"vertical"===t&&(t="TB"),t},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.formatTooltip=function(t,e,n){var i=this.getData().tree,r=i.root.children[0],o=i.getNodeByDataIndex(t),a=o.getValue(),s=o.name;while(o&&o!==r)s=o.parentNode.name+"."+s,o=o.parentNode;return oy("nameValue",{name:s,value:a,noValue:isNaN(a)||null==a})},e.prototype.getDataParams=function(e){var n=t.prototype.getDataParams.apply(this,arguments),i=this.getData().tree.getNodeByDataIndex(e);return n.treeAncestors=aO(i,this),n},e.type="series.tree",e.layoutMode="box",e.defaultOption={zlevel:0,z:2,coordinateSystem:"view",left:"12%",top:"12%",right:"12%",bottom:"12%",layout:"orthogonal",edgeShape:"curve",edgeForkPosition:"50%",roam:!1,nodeScaleRatio:.4,center:null,zoom:1,orient:"LR",symbol:"emptyCircle",symbolSize:7,expandAndCollapse:!0,initialTreeDepth:2,lineStyle:{color:"#ccc",width:1.5,curveness:.5},itemStyle:{color:"lightsteelblue",borderWidth:1.5},label:{show:!0},animationEasing:"linear",animationDuration:700,animationDurationUpdate:500},e}(Yy),lO=sO;function uO(t,e,n){var i,r=[t],o=[];while(i=r.pop())if(o.push(i),i.isExpand){var a=i.children;if(a.length)for(var s=0;s<a.length;s++)r.push(a[s])}while(i=o.pop())e(i,n)}function cO(t,e){var n,i=[t];while(n=i.pop())if(e(n),n.isExpand){var r=n.children;if(r.length)for(var o=r.length-1;o>=0;o--)i.push(r[o])}}function dO(t,e){t.eachSeriesByType("tree",(function(t){hO(t,e)}))}function hO(t,e){var n=LY(t,e);t.layoutInfo=n;var i=t.get("layout"),r=0,o=0,a=null;"radial"===i?(r=2*Math.PI,o=Math.min(n.height,n.width)/2,a=MY((function(t,e){return(t.parentNode===e.parentNode?1:2)/t.depth}))):(r=n.width,o=n.height,a=MY());var s=t.getData().tree.root,l=s.children[0];if(l){bY(s),uO(l,xY,a),s.hierNode.modifier=-l.hierNode.prelim,cO(l,wY);var u=l,c=l,d=l;cO(l,(function(t){var e=t.getLayout().x;e<u.getLayout().x&&(u=t),e>c.getLayout().x&&(c=t),t.depth>d.depth&&(d=t)}));var h=u===c?1:a(u,c)/2,f=h-u.getLayout().x,p=0,m=0,g=0,y=0;if("radial"===i)p=r/(c.getLayout().x+h+f),m=o/(d.depth-1||1),cO(l,(function(t){g=(t.getLayout().x+f)*p,y=(t.depth-1)*m;var e=SY(g,y);t.setLayout({x:e.x,y:e.y,rawX:g,rawY:y},!0)}));else{var v=t.getOrient();"RL"===v||"LR"===v?(m=o/(c.getLayout().x+h+f),p=r/(d.depth-1||1),cO(l,(function(t){y=(t.getLayout().x+f)*m,g="LR"===v?(t.depth-1)*p:r-(t.depth-1)*p,t.setLayout({x:g,y:y},!0)}))):"TB"!==v&&"BT"!==v||(p=r/(c.getLayout().x+h+f),m=o/(d.depth-1||1),cO(l,(function(t){g=(t.getLayout().x+f)*p,y="TB"===v?(t.depth-1)*m:o-(t.depth-1)*m,t.setLayout({x:g,y:y},!0)})))}}}function fO(t){t.eachSeriesByType("tree",(function(t){var e=t.getData(),n=e.tree;n.eachNode((function(t){var n=t.getModel(),i=n.getModel("itemStyle").getItemStyle(),r=e.ensureUniqueItemVisual(t.dataIndex,"style");I(r,i)}))}))}function pO(t){t.registerAction({type:"treeExpandAndCollapse",event:"treeExpandAndCollapse",update:"update"},(function(t,e){e.eachComponent({mainType:"series",subType:"tree",query:t},(function(e){var n=t.dataIndex,i=e.getData().tree,r=i.getNodeByDataIndex(n);r.isExpand=!r.isExpand}))})),t.registerAction({type:"treeRoam",event:"treeRoam",update:"none"},(function(t,e){e.eachComponent({mainType:"series",subType:"tree",query:t},(function(e){var n=e.coordinateSystem,i=mY(n,t);e.setCenter&&e.setCenter(i.center),e.setZoom&&e.setZoom(i.zoom)}))}))}function mO(t){t.registerChartView(WY),t.registerSeriesModel(lO),t.registerLayout(dO),t.registerVisual(fO),pO(t)}var gO=function(){},yO=["treemapZoomToNode","treemapRender","treemapMove"];function vO(t){for(var e=0;e<yO.length;e++)t.registerAction({type:yO[e],update:"updateView"},gO);t.registerAction({type:"treemapRootToNode",update:"updateView"},(function(t,e){function n(e,n){var i=["treemapZoomToNode","treemapRootToNode"],r=iO(t,i,e);if(r){var o=e.getViewRoot();o&&(t.direction=oO(o,r.node)?"rollUp":"drillDown"),e.resetViewRoot(r.node)}}e.eachComponent({mainType:"series",subType:"treemap",query:t},n)}))}function _O(t){var e=t.getData(),n=e.tree,i={};n.eachNode((function(e){var n=e;while(n&&n.depth>1)n=n.parentNode;var r=Xp(t.ecModel,n.name||n.dataIndex+"",i);e.setVisual("decal",r)}))}var bO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.preventUsingHoverLayer=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){var n={name:t.name,children:t.data};xO(n);var i=t.levels||[],r=this.designatedVisualItemStyle={},o=new qh({itemStyle:r},this,e);i=t.levels=wO(i,e);var a=H(i||[],(function(t){return new qh(t,o,e)}),this),s=nO.createTree(n,this,l);function l(t){t.wrapMethod("getItemModel",(function(t,e){var n=s.getNodeByDataIndex(e),i=n?a[n.depth]:null;return t.parentModel=i||o,t}))}return s.data},e.prototype.optionUpdated=function(){this.resetViewRoot()},e.prototype.formatTooltip=function(t,e,n){var i=this.getData(),r=this.getRawValue(t),o=i.getName(t);return oy("nameValue",{name:o,value:r})},e.prototype.getDataParams=function(e){var n=t.prototype.getDataParams.apply(this,arguments),i=this.getData().tree.getNodeByDataIndex(e);return n.treePathInfo=aO(i,this),n},e.prototype.setLayoutInfo=function(t){this.layoutInfo=this.layoutInfo||{},I(this.layoutInfo,t)},e.prototype.mapIdToIndex=function(t){var e=this._idIndexMap;e||(e=this._idIndexMap=gt(),this._idIndexMapCount=0);var n=e.get(t);return null==n&&e.set(t,n=this._idIndexMapCount++),n},e.prototype.getViewRoot=function(){return this._viewRoot},e.prototype.resetViewRoot=function(t){t?this._viewRoot=t:t=this._viewRoot;var e=this.getRawData().tree.root;t&&(t===e||e.contains(t))||(this._viewRoot=e)},e.prototype.enableAriaDecal=function(){_O(this)},e.type="series.treemap",e.layoutMode="box",e.defaultOption={progressive:0,left:"center",top:"middle",width:"80%",height:"80%",sort:!0,clipWindow:"origin",squareRatio:.5*(1+Math.sqrt(5)),leafDepth:null,drillDownIcon:"▶",zoomToNodeRatio:.1024,roam:!0,nodeClick:"zoomToNode",animation:!0,animationDurationUpdate:900,animationEasing:"quinticInOut",breadcrumb:{show:!0,height:22,left:"center",top:"bottom",emptyItemWidth:25,itemStyle:{color:"rgba(0,0,0,0.7)",textStyle:{color:"#fff"}}},label:{show:!0,distance:0,padding:5,position:"inside",color:"#fff",overflow:"truncate"},upperLabel:{show:!1,position:[0,"50%"],height:20,overflow:"truncate",verticalAlign:"middle"},itemStyle:{color:null,colorAlpha:null,colorSaturation:null,borderWidth:0,gapWidth:0,borderColor:"#fff",borderColorSaturation:null},emphasis:{upperLabel:{show:!0,position:[0,"50%"],ellipsis:!0,verticalAlign:"middle"}},visualDimension:0,visualMin:null,visualMax:null,color:[],colorAlpha:null,colorSaturation:null,colorMappingBy:"index",visibleMin:10,childrenVisibleMin:null,levels:[]},e}(Yy);function xO(t){var e=0;R(t.children,(function(t){xO(t);var n=t.value;$(n)&&(n=n[0]),e+=n}));var n=t.value;$(n)&&(n=n[0]),(null==n||isNaN(n))&&(n=e),n<0&&(n=0),$(t.value)?t.value[0]=n:t.value=n}function wO(t,e){var n=so(e.get("color")),i=so(e.get(["aria","decal","decals"]));if(n){var r,o;t=t||[],R(t,(function(t){var e=new qh(t),n=e.get("color"),i=e.get("decal");(e.get(["itemStyle","color"])||n&&"none"!==n)&&(r=!0),(e.get(["itemStyle","decal"])||i&&"none"!==i)&&(o=!0)}));var a=t[0]||(t[0]={});return r||(a.color=n.slice()),!o&&i&&(a.decal=i.slice()),t}}var MO=bO,SO=8,LO=8,kO=5,TO=function(){function t(t){this.group=new Mr,t.add(this.group)}return t.prototype.render=function(t,e,n,i){var r=t.getModel("breadcrumb"),o=this.group;if(o.removeAll(),r.get("show")&&n){var a=r.getModel("itemStyle"),s=a.getModel("textStyle"),l={pos:{left:r.get("left"),right:r.get("right"),top:r.get("top"),bottom:r.get("bottom")},box:{width:e.getWidth(),height:e.getHeight()},emptyItemWidth:r.get("emptyItemWidth"),totalWidth:0,renderList:[]};this._prepare(n,l,s),this._renderContent(t,l,a,s,i),hp(o,l.pos,l.box)}},t.prototype._prepare=function(t,e,n){for(var i=t;i;i=i.parentNode){var r=wo(i.getModel().get("name"),""),o=n.getTextRect(r),a=Math.max(o.width+2*SO,e.emptyItemWidth);e.totalWidth+=a+LO,e.renderList.push({node:i,text:r,width:a})}},t.prototype._renderContent=function(t,e,n,i,r){for(var o=0,a=e.emptyItemWidth,s=t.get(["breadcrumb","height"]),l=cp(e.pos,e.box),u=e.totalWidth,c=e.renderList,d=c.length-1;d>=0;d--){var h=c[d],f=h.node,p=h.width,m=h.text;u>l.width&&(u-=p-a,p=a,m=null);var g=new rd({shape:{points:DO(o,0,p,s,d===c.length-1,0===d)},style:A(n.getItemStyle(),{lineJoin:"bevel"}),textContent:new Gl({style:{text:m,fill:i.getTextColor(),font:i.getFont()}}),textConfig:{position:"inside"},z2:1e4*eu,onclick:G(r,f)});g.disableLabelAnimation=!0,this.group.add(g),CO(g,t,f),o+=p+LO}},t.prototype.remove=function(){this.group.removeAll()},t}();function DO(t,e,n,i,r,o){var a=[[r?t:t-kO,e],[t+n,e],[t+n,e+i],[r?t:t-kO,e+i]];return!o&&a.splice(2,0,[t+n+kO,e+i/2]),!r&&a.push([t,e+i/2]),a}function CO(t,e,n){$l(t).eventData={componentType:"series",componentSubType:"treemap",componentIndex:e.componentIndex,seriesIndex:e.componentIndex,seriesName:e.name,seriesType:"treemap",selfType:"breadcrumb",nodeData:{dataIndex:n&&n.dataIndex,name:n&&n.name},treePathInfo:n&&aO(n,e)}}var IO=TO,AO=function(){function t(){this._storage=[],this._elExistsMap={}}return t.prototype.add=function(t,e,n,i,r){return!this._elExistsMap[t.id]&&(this._elExistsMap[t.id]=!0,this._storage.push({el:t,target:e,duration:n,delay:i,easing:r}),!0)},t.prototype.finished=function(t){return this._finishedCallback=t,this},t.prototype.start=function(){for(var t=this,e=this._storage.length,n=function(){e--,e<=0&&(t._storage.length=0,t._elExistsMap={},t._finishedCallback&&t._finishedCallback())},i=0,r=this._storage.length;i<r;i++){var o=this._storage[i];o.el.animateTo(o.target,{duration:o.duration,delay:o.delay,easing:o.easing,setToFinal:!0,done:n,aborted:n})}return this},t}();function YO(){return new AO}var OO=Mr,PO=Al,EO=3,NO="label",RO="upperLabel",HO=10*eu,jO=2*eu,zO=3*eu,WO=na([["fill","color"],["stroke","strokeColor"],["lineWidth","strokeWidth"],["shadowBlur"],["shadowOffsetX"],["shadowOffsetY"],["shadowColor"]]),BO=function(t){var e=WO(t);return e.stroke=e.fill=e.lineWidth=null,e},FO=Io(),VO=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._state="ready",n._storage=GO(),n}return o(e,t),e.prototype.render=function(t,e,n,i){var r=e.findComponents({mainType:"series",subType:"treemap",query:i});if(!(O(r,t)<0)){this.seriesModel=t,this.api=n,this.ecModel=e;var o=["treemapZoomToNode","treemapRootToNode"],a=iO(i,o,t),s=i&&i.type,l=t.layoutInfo,u=!this._oldTree,c=this._storage,d="treemapRootToNode"===s&&a&&c?{rootNodeGroup:c.nodeGroup[a.node.getRawIndex()],direction:i.direction}:null,h=this._giveContainerGroup(l),f=this._doRender(h,t,d);u||s&&"treemapZoomToNode"!==s&&"treemapRootToNode"!==s?f.renderFinally():this._doAnimation(h,f,t,d),this._resetController(n),this._renderBreadcrumb(t,n,a)}},e.prototype._giveContainerGroup=function(t){var e=this._containerGroup;return e||(e=this._containerGroup=new OO,this._initEvents(e),this.group.add(e)),e.x=t.x,e.y=t.y,e},e.prototype._doRender=function(t,e,n){var i=e.getData().tree,r=this._oldTree,o=GO(),a=GO(),s=this._storage,l=[];function u(t,i,r,u){return $O(e,a,s,n,o,l,t,i,r,u)}d(i.root?[i.root]:[],r&&r.root?[r.root]:[],t,i===r||!r,0);var c=h(s);return this._oldTree=i,this._storage=a,{lastsForAnimation:o,willDeleteEls:c,renderFinally:f};function d(t,e,n,i,r){function o(t){return t.getId()}function a(o,a){var s=null!=o?t[o]:null,l=null!=a?e[a]:null,c=u(s,l,n,r);c&&d(s&&s.viewChildren||[],l&&l.viewChildren||[],c,i,r+1)}i?(e=t,R(t,(function(t,e){!t.isRemoved()&&a(e,e)}))):new $S(e,t,o,o).add(a).update(a).remove(G(a,null)).execute()}function h(t){var e=GO();return t&&R(t,(function(t,n){var i=e[n];R(t,(function(t){t&&(i.push(t),FO(t).willDelete=!0)}))})),e}function f(){R(c,(function(t){R(t,(function(t){t.parent&&t.parent.remove(t)}))})),R(l,(function(t){t.invisible=!0,t.dirty()}))}},e.prototype._doAnimation=function(t,e,n,i){if(n.get("animation")){var r=n.get("animationDurationUpdate"),o=n.get("animationEasing"),a=(U(r)?0:r)||0,s=(U(o)?null:o)||"cubicOut",l=YO();R(e.willDeleteEls,(function(t,e){R(t,(function(t,n){if(!t.invisible){var r,o=t.parent,u=FO(o);if(i&&"drillDown"===i.direction)r=o===i.rootNodeGroup?{shape:{x:0,y:0,width:u.nodeWidth,height:u.nodeHeight},style:{opacity:0}}:{style:{opacity:0}};else{var c=0,d=0;u.willDelete||(c=u.nodeWidth/2,d=u.nodeHeight/2),r="nodeGroup"===e?{x:c,y:d,style:{opacity:0}}:{shape:{x:c,y:d,width:0,height:0},style:{opacity:0}}}r&&l.add(t,r,a,0,s)}}))})),R(this._storage,(function(t,n){R(t,(function(t,i){var r=e.lastsForAnimation[n][i],o={};r&&(t instanceof Mr?null!=r.oldX&&(o.x=t.x,o.y=t.y,t.x=r.oldX,t.y=r.oldY):(r.oldShape&&(o.shape=I({},t.shape),t.setShape(r.oldShape)),r.fadein?(t.setStyle("opacity",0),o.style={opacity:1}):1!==t.style.opacity&&(o.style={opacity:1})),l.add(t,o,a,0,s))}))}),this),this._state="animating",l.finished(V((function(){this._state="ready",e.renderFinally()}),this)).start()}},e.prototype._resetController=function(t){var e=this._controller;e||(e=this._controller=new CA(t.getZr()),e.enable(this.seriesModel.get("roam")),e.on("pan",V(this._onPan,this)),e.on("zoom",V(this._onZoom,this)));var n=new Ji(0,0,t.getWidth(),t.getHeight());e.setPointerChecker((function(t,e,i){return n.contain(e,i)}))},e.prototype._clearController=function(){var t=this._controller;t&&(t.dispose(),t=null)},e.prototype._onPan=function(t){if("animating"!==this._state&&(Math.abs(t.dx)>EO||Math.abs(t.dy)>EO)){var e=this.seriesModel.getData().tree.root;if(!e)return;var n=e.getLayout();if(!n)return;this.api.dispatchAction({type:"treemapMove",from:this.uid,seriesId:this.seriesModel.id,rootRect:{x:n.x+t.dx,y:n.y+t.dy,width:n.width,height:n.height}})}},e.prototype._onZoom=function(t){var e=t.originX,n=t.originY;if("animating"!==this._state){var i=this.seriesModel.getData().tree.root;if(!i)return;var r=i.getLayout();if(!r)return;var o=new Ji(r.x,r.y,r.width,r.height),a=this.seriesModel.layoutInfo;e-=a.x,n-=a.y;var s=bi();Si(s,s,[-e,-n]),ki(s,s,[t.scale,t.scale]),Si(s,s,[e,n]),o.applyTransform(s),this.api.dispatchAction({type:"treemapRender",from:this.uid,seriesId:this.seriesModel.id,rootRect:{x:o.x,y:o.y,width:o.width,height:o.height}})}},e.prototype._initEvents=function(t){var e=this;t.on("click",(function(t){if("ready"===e._state){var n=e.seriesModel.get("nodeClick",!0);if(n){var i=e.findTarget(t.offsetX,t.offsetY);if(i){var r=i.node;if(r.getLayout().isLeafRoot)e._rootToNode(i);else if("zoomToNode"===n)e._zoomToNode(i);else if("link"===n){var o=r.hostTree.data.getItemModel(r.dataIndex),a=o.get("link",!0),s=o.get("target",!0)||"blank";a&&rp(a,s)}}}}}),this)},e.prototype._renderBreadcrumb=function(t,e,n){var i=this;n||(n=null!=t.get("leafDepth",!0)?{node:t.getViewRoot()}:this.findTarget(e.getWidth()/2,e.getHeight()/2),n||(n={node:t.getData().tree.root})),(this._breadcrumb||(this._breadcrumb=new IO(this.group))).render(t,e,n.node,(function(e){"animating"!==i._state&&(oO(t.getViewRoot(),e)?i._rootToNode({node:e}):i._zoomToNode({node:e}))}))},e.prototype.remove=function(){this._clearController(),this._containerGroup&&this._containerGroup.removeAll(),this._storage=GO(),this._state="ready",this._breadcrumb&&this._breadcrumb.remove()},e.prototype.dispose=function(){this._clearController()},e.prototype._zoomToNode=function(t){this.api.dispatchAction({type:"treemapZoomToNode",from:this.uid,seriesId:this.seriesModel.id,targetNode:t.node})},e.prototype._rootToNode=function(t){this.api.dispatchAction({type:"treemapRootToNode",from:this.uid,seriesId:this.seriesModel.id,targetNode:t.node})},e.prototype.findTarget=function(t,e){var n,i=this.seriesModel.getViewRoot();return i.eachNode({attr:"viewChildren",order:"preorder"},(function(i){var r=this._storage.background[i.getRawIndex()];if(r){var o=r.transformCoordToLocal(t,e),a=r.shape;if(!(a.x<=o[0]&&o[0]<=a.x+a.width&&a.y<=o[1]&&o[1]<=a.y+a.height))return!1;n={node:i,offsetX:o[0],offsetY:o[1]}}}),this),n},e.type="treemap",e}(Vy);function GO(){return{nodeGroup:[],background:[],content:[]}}function $O(t,e,n,i,r,o,a,s,l,u){if(a){var c=a.getLayout(),d=t.getData(),h=a.getModel();if(d.setItemGraphicEl(a.dataIndex,null),c&&c.isInView){var f=c.width,p=c.height,m=c.borderWidth,g=c.invisible,y=a.getRawIndex(),v=s&&s.getRawIndex(),_=a.viewChildren,b=c.upperHeight,x=_&&_.length,w=h.getModel("itemStyle"),M=h.getModel(["emphasis","itemStyle"]),S=h.getModel(["blur","itemStyle"]),L=h.getModel(["select","itemStyle"]),k=w.get("borderRadius")||0,T=j("nodeGroup",OO);if(T){if(l.add(T),T.x=c.x||0,T.y=c.y||0,T.markRedraw(),FO(T).nodeWidth=f,FO(T).nodeHeight=p,c.isAboveViewRoot)return T;var D=j("background",PO,u,jO);D&&P(T,D,x&&c.upperLabelHeight);var C=h.get(["emphasis","focus"]),A=h.get(["emphasis","blurScope"]),Y="ancestor"===C?a.getAncestorsIndices():"descendant"===C?a.getDescendantIndices():C;if(x)Ku(T)&&Ju(T,!1),D&&(Ju(D,!0),d.setItemGraphicEl(a.dataIndex,D),Uu(D,Y,A));else{var O=j("content",PO,u,zO);O&&E(T,O),D&&Ku(D)&&Ju(D,!1),Ju(T,!0),d.setItemGraphicEl(a.dataIndex,T),Uu(T,Y,A)}return T}}}function P(e,n,i){var r=$l(n);if(r.dataIndex=a.dataIndex,r.seriesIndex=t.seriesIndex,n.setShape({x:0,y:0,width:f,height:p,r:k}),g)N(n);else{n.invisible=!1;var o=a.getVisual("style"),s=o.stroke,l=BO(w);l.fill=s;var u=WO(M);u.fill=M.get("borderColor");var c=WO(S);c.fill=S.get("borderColor");var d=WO(L);if(d.fill=L.get("borderColor"),i){var h=f-2*m;R(n,s,o.opacity,{x:m,y:0,width:h,height:b})}else n.removeTextContent();n.setStyle(l),n.ensureState("emphasis").style=u,n.ensureState("blur").style=c,n.ensureState("select").style=d,ku(n)}e.add(n)}function E(e,n){var i=$l(n);i.dataIndex=a.dataIndex,i.seriesIndex=t.seriesIndex;var r=Math.max(f-2*m,0),o=Math.max(p-2*m,0);if(n.culling=!0,n.setShape({x:m,y:m,width:r,height:o,r:k}),g)N(n);else{n.invisible=!1;var s=a.getVisual("style"),l=s.fill,u=BO(w);u.fill=l,u.decal=s.decal;var c=WO(M),d=WO(S),h=WO(L);R(n,l,s.opacity,null),n.setStyle(u),n.ensureState("emphasis").style=c,n.ensureState("blur").style=d,n.ensureState("select").style=h,ku(n)}e.add(n)}function N(t){!t.invisible&&o.push(t)}function R(e,n,i,r){var o=h.getModel(r?RO:NO),s=wo(h.get("name"),null),l=o.getShallow("show");Mh(e,Sh(h,r?RO:NO),{defaultText:l?s:null,inheritColor:n,defaultOpacity:i,labelFetcher:t,labelDataIndex:a.dataIndex});var u=e.getTextContent(),d=u.style,f=ut(d.padding||0);r&&(e.setTextConfig({layoutRect:r}),u.disableLabelLayout=!0),u.beforeUpdate=function(){var t=Math.max((r?r.width:e.shape.width)-f[1]-f[3],0),n=Math.max((r?r.height:e.shape.height)-f[0]-f[2],0);d.width===t&&d.height===n||u.setStyle({width:t,height:n})},d.truncateMinChar=2,d.lineOverflow="truncate",H(d,r,c);var p=u.getState("emphasis");H(p?p.style:null,r,c)}function H(e,n,i){var r=e?e.text:null;if(!n&&i.isLeafRoot&&null!=r){var o=t.get("drillDownIcon",!0);e.text=o?o+" "+r:r}}function j(t,i,o,a){var s=null!=v&&n[t][v],l=r[t];return s?(n[t][v]=null,z(l,s)):g||(s=new i,s instanceof Ea&&(s.z2=UO(o,a)),W(l,s)),e[t][y]=s}function z(t,e){var n=t[y]={};e instanceof OO?(n.oldX=e.x,n.oldY=e.y):n.oldShape=I({},e.shape)}function W(t,e){var n=t[y]={},o=a.parentNode,s=e instanceof Mr;if(o&&(!i||"drillDown"===i.direction)){var l=0,u=0,c=r.background[o.getRawIndex()];!i&&c&&c.oldShape&&(l=c.oldShape.width,u=c.oldShape.height),s?(n.oldX=0,n.oldY=u):n.oldShape={x:l,y:u,width:0,height:0}}n.fadein=!s}}function UO(t,e){return t*HO+e}var qO=VO,XO=R,ZO=J,JO=-1,KO=function(){function t(e){var n=e.mappingMethod,i=e.type,r=this.option=T(e);this.type=i,this.mappingMethod=n,this._normalizeData=cP[n];var o=t.visualHandlers[i];this.applyVisual=o.applyVisual,this.getColorMapper=o.getColorMapper,this._normalizedToVisual=o._normalizedToVisual[n],"piecewise"===n?(eP(r),QO(r)):"category"===n?r.categories?tP(r):eP(r,!0):(ct("linear"!==n||r.dataExtent),eP(r))}return t.prototype.mapValueToVisual=function(t){var e=this._normalizeData(t);return this._normalizedToVisual(e,t)},t.prototype.getNormalizer=function(){return V(this._normalizeData,this)},t.listVisualTypes=function(){return B(t.visualHandlers)},t.isValidType=function(e){return t.visualHandlers.hasOwnProperty(e)},t.eachVisual=function(t,e,n){J(t)?R(t,e,n):e.call(n,t)},t.mapVisual=function(e,n,i){var r,o=$(e)?[]:J(e)?{}:(r=!0,null);return t.eachVisual(e,(function(t,e){var a=n.call(i,t,e);r?o=a:o[e]=a})),o},t.retrieveVisuals=function(e){var n,i={};return e&&XO(t.visualHandlers,(function(t,r){e.hasOwnProperty(r)&&(i[r]=e[r],n=!0)})),n?i:null},t.prepareVisualTypes=function(t){if($(t))t=t.slice();else{if(!ZO(t))return[];var e=[];XO(t,(function(t,n){e.push(n)})),t=e}return t.sort((function(t,e){return"color"===e&&"color"!==t&&0===t.indexOf("color")?1:-1})),t},t.dependsOn=function(t,e){return"color"===e?!(!t||0!==t.indexOf(e)):t===e},t.findPieceIndex=function(t,e,n){for(var i,r=1/0,o=0,a=e.length;o<a;o++){var s=e[o].value;if(null!=s){if(s===t||"string"===typeof s&&s===t+"")return o;n&&d(s,o)}}for(o=0,a=e.length;o<a;o++){var l=e[o],u=l.interval,c=l.close;if(u){if(u[0]===-1/0){if(dP(c[1],t,u[1]))return o}else if(u[1]===1/0){if(dP(c[0],u[0],t))return o}else if(dP(c[0],u[0],t)&&dP(c[1],t,u[1]))return o;n&&d(u[0],o),n&&d(u[1],o)}}if(n)return t===1/0?e.length-1:t===-1/0?0:i;function d(e,n){var o=Math.abs(e-t);o<r&&(r=o,i=n)}},t.visualHandlers={color:{applyVisual:rP("color"),getColorMapper:function(){var t=this.option;return V("category"===t.mappingMethod?function(t,e){return!e&&(t=this._normalizeData(t)),oP.call(this,t)}:function(e,n,i){var r=!!i;return!n&&(e=this._normalizeData(e)),i=wn(e,t.parsedVisual,i),r?i:Ln(i,"rgba")},this)},_normalizedToVisual:{linear:function(t){return Ln(wn(t,this.option.parsedVisual),"rgba")},category:oP,piecewise:function(t,e){var n=lP.call(this,e);return null==n&&(n=Ln(wn(t,this.option.parsedVisual),"rgba")),n},fixed:aP}},colorHue:nP((function(t,e){return Mn(t,e)})),colorSaturation:nP((function(t,e){return Mn(t,null,e)})),colorLightness:nP((function(t,e){return Mn(t,null,null,e)})),colorAlpha:nP((function(t,e){return Sn(t,e)})),decal:{applyVisual:rP("decal"),_normalizedToVisual:{linear:null,category:oP,piecewise:null,fixed:null}},opacity:{applyVisual:rP("opacity"),_normalizedToVisual:sP([0,1])},liftZ:{applyVisual:rP("liftZ"),_normalizedToVisual:{linear:aP,category:aP,piecewise:aP,fixed:aP}},symbol:{applyVisual:function(t,e,n){var i=this.mapValueToVisual(t);n("symbol",i)},_normalizedToVisual:{linear:iP,category:oP,piecewise:function(t,e){var n=lP.call(this,e);return null==n&&(n=iP.call(this,t)),n},fixed:aP}},symbolSize:{applyVisual:rP("symbolSize"),_normalizedToVisual:sP([0,1])}},t}();function QO(t){var e=t.pieceList;t.hasSpecialVisual=!1,R(e,(function(e,n){e.originIndex=n,null!=e.visual&&(t.hasSpecialVisual=!0)}))}function tP(t){var e=t.categories,n=t.categoryMap={},i=t.visual;if(XO(e,(function(t,e){n[t]=e})),!$(i)){var r=[];J(i)?XO(i,(function(t,e){var i=n[e];r[null!=i?i:JO]=t})):r[JO]=i,i=uP(t,r)}for(var o=e.length-1;o>=0;o--)null==i[o]&&(delete n[e[o]],e.pop())}function eP(t,e){var n=t.visual,i=[];J(n)?XO(n,(function(t){i.push(t)})):null!=n&&i.push(n);var r={color:1,symbol:1};e||1!==i.length||r.hasOwnProperty(t.type)||(i[1]=i[0]),uP(t,i)}function nP(t){return{applyVisual:function(e,n,i){var r=this.mapValueToVisual(e);i("color",t(n("color"),r))},_normalizedToVisual:sP([0,1])}}function iP(t){var e=this.option.visual;return e[Math.round(Er(t,[0,1],[0,e.length-1],!0))]||{}}function rP(t){return function(e,n,i){i(t,this.mapValueToVisual(e))}}function oP(t){var e=this.option.visual;return e[this.option.loop&&t!==JO?t%e.length:t]}function aP(){return this.option.visual[0]}function sP(t){return{linear:function(e){return Er(e,t,this.option.visual,!0)},category:oP,piecewise:function(e,n){var i=lP.call(this,n);return null==i&&(i=Er(e,t,this.option.visual,!0)),i},fixed:aP}}function lP(t){var e=this.option,n=e.pieceList;if(e.hasSpecialVisual){var i=KO.findPieceIndex(t,n),r=n[i];if(r&&r.visual)return r.visual[this.type]}}function uP(t,e){return t.visual=e,"color"===t.type&&(t.parsedVisual=H(e,(function(t){return yn(t)}))),e}var cP={linear:function(t){return Er(t,this.option.dataExtent,[0,1],!0)},piecewise:function(t){var e=this.option.pieceList,n=KO.findPieceIndex(t,e,!0);if(null!=n)return Er(n,[0,e.length-1],[0,1],!0)},category:function(t){var e=this.option.categories?this.option.categoryMap[t]:t;return null==e?JO:e},fixed:bt};function dP(t,e,n){return t?e<=n:e<n}var hP=KO,fP="itemStyle",pP=Io(),mP={seriesType:"treemap",reset:function(t){var e=t.getData().tree,n=e.root;n.isRemoved()||gP(n,{},t.getViewRoot().getAncestors(),t)}};function gP(t,e,n,i){var r=t.getModel(),o=t.getLayout(),a=t.hostTree.data;if(o&&!o.invisible&&o.isInView){var s,l=r.getModel(fP),u=yP(l,e,i),c=a.ensureUniqueItemVisual(t.dataIndex,"style"),d=l.get("borderColor"),h=l.get("borderColorSaturation");null!=h&&(s=vP(u),d=_P(h,s)),c.stroke=d;var f=t.viewChildren;if(f&&f.length){var p=xP(t,r,o,l,u,f);R(f,(function(t,e){if(t.depth>=n.length||t===n[t.depth]){var o=MP(r,u,t,e,p,i);gP(t,o,n,i)}}))}else s=vP(u),c.fill=s}}function yP(t,e,n){var i=I({},e),r=n.designatedVisualItemStyle;return R(["color","colorAlpha","colorSaturation"],(function(n){r[n]=e[n];var o=t.get(n);r[n]=null,null!=o&&(i[n]=o)})),i}function vP(t){var e=bP(t,"color");if(e){var n=bP(t,"colorAlpha"),i=bP(t,"colorSaturation");return i&&(e=Mn(e,null,null,i)),n&&(e=Sn(e,n)),e}}function _P(t,e){return null!=e?Mn(e,null,null,t):null}function bP(t,e){var n=t[e];if(null!=n&&"none"!==n)return n}function xP(t,e,n,i,r,o){if(o&&o.length){var a=wP(e,"color")||null!=r.color&&"none"!==r.color&&(wP(e,"colorAlpha")||wP(e,"colorSaturation"));if(a){var s=e.get("visualMin"),l=e.get("visualMax"),u=n.dataExtent.slice();null!=s&&s<u[0]&&(u[0]=s),null!=l&&l>u[1]&&(u[1]=l);var c=e.get("colorMappingBy"),d={type:a.name,dataExtent:u,visual:a.range};"color"!==d.type||"index"!==c&&"id"!==c?d.mappingMethod="linear":(d.mappingMethod="category",d.loop=!0);var h=new hP(d);return pP(h).drColorMappingBy=c,h}}}function wP(t,e){var n=t.get(e);return $(n)&&n.length?{name:e,range:n}:null}function MP(t,e,n,i,r,o){var a=I({},e);if(r){var s=r.type,l="color"===s&&pP(r).drColorMappingBy,u="index"===l?i:"id"===l?o.mapIdToIndex(n.getId()):n.getValue(t.get("visualDimension"));a[s]=r.mapValueToVisual(u)}return a}var SP=Math.max,LP=Math.min,kP=ot,TP=R,DP=["itemStyle","borderWidth"],CP=["itemStyle","gapWidth"],IP=["upperLabel","show"],AP=["upperLabel","height"],YP={seriesType:"treemap",reset:function(t,e,n,i){var r=n.getWidth(),o=n.getHeight(),a=t.option,s=dp(t.getBoxLayoutParams(),{width:n.getWidth(),height:n.getHeight()}),l=a.size||[],u=Nr(kP(s.width,l[0]),r),c=Nr(kP(s.height,l[1]),o),d=i&&i.type,h=["treemapZoomToNode","treemapRootToNode"],f=iO(i,h,t),p="treemapRender"===d||"treemapMove"===d?i.rootRect:null,m=t.getViewRoot(),g=rO(m);if("treemapMove"!==d){var y="treemapZoomToNode"===d?zP(t,f,m,u,c):p?[p.width,p.height]:[u,c],v=a.sort;v&&"asc"!==v&&"desc"!==v&&(v="desc");var _={squareRatio:a.squareRatio,sort:v,leafDepth:a.leafDepth};m.hostTree.clearLayouts();var b={x:0,y:0,width:y[0],height:y[1],area:y[0]*y[1]};m.setLayout(b),OP(m,_,!1,0),b=m.getLayout(),TP(g,(function(t,e){var n=(g[e+1]||m).getValue();t.setLayout(I({dataExtent:[n,n],borderWidth:0,upperHeight:0},b))}))}var x=t.getData().tree.root;x.setLayout(WP(s,p,f),!0),t.setLayoutInfo(s),BP(x,new Ji(-s.x,-s.y,r,o),g,m,0)}};function OP(t,e,n,i){var r,o;if(!t.isRemoved()){var a=t.getLayout();r=a.width,o=a.height;var s=t.getModel(),l=s.get(DP),u=s.get(CP)/2,c=FP(s),d=Math.max(l,c),h=l-u,f=d-u;t.setLayout({borderWidth:l,upperHeight:d,upperLabelHeight:c},!0),r=SP(r-2*h,0),o=SP(o-h-f,0);var p=r*o,m=PP(t,s,p,e,n,i);if(m.length){var g={x:h,y:f,width:r,height:o},y=LP(r,o),v=1/0,_=[];_.area=0;for(var b=0,x=m.length;b<x;){var w=m[b];_.push(w),_.area+=w.getLayout().area;var M=HP(_,y,e.squareRatio);M<=v?(b++,v=M):(_.area-=_.pop().getLayout().area,jP(_,y,g,u,!1),y=LP(g.width,g.height),_.length=_.area=0,v=1/0)}if(_.length&&jP(_,y,g,u,!0),!n){var S=s.get("childrenVisibleMin");null!=S&&p<S&&(n=!0)}for(b=0,x=m.length;b<x;b++)OP(m[b],e,n,i+1)}}}function PP(t,e,n,i,r,o){var a=t.children||[],s=i.sort;"asc"!==s&&"desc"!==s&&(s=null);var l=null!=i.leafDepth&&i.leafDepth<=o;if(r&&!l)return t.viewChildren=[];a=z(a,(function(t){return!t.isRemoved()})),NP(a,s);var u=RP(e,a,s);if(0===u.sum)return t.viewChildren=[];if(u.sum=EP(e,n,u.sum,s,a),0===u.sum)return t.viewChildren=[];for(var c=0,d=a.length;c<d;c++){var h=a[c].getValue()/u.sum*n;a[c].setLayout({area:h})}return l&&(a.length&&t.setLayout({isLeafRoot:!0},!0),a.length=0),t.viewChildren=a,t.setLayout({dataExtent:u.dataExtent},!0),a}function EP(t,e,n,i,r){if(!i)return n;for(var o=t.get("visibleMin"),a=r.length,s=a,l=a-1;l>=0;l--){var u=r["asc"===i?a-l-1:l].getValue();u/n*e<o&&(s=l,n-=u)}return"asc"===i?r.splice(0,a-s):r.splice(s,a-s),n}function NP(t,e){return e&&t.sort((function(t,n){var i="asc"===e?t.getValue()-n.getValue():n.getValue()-t.getValue();return 0===i?"asc"===e?t.dataIndex-n.dataIndex:n.dataIndex-t.dataIndex:i})),t}function RP(t,e,n){for(var i=0,r=0,o=e.length;r<o;r++)i+=e[r].getValue();var a,s=t.get("visualDimension");return e&&e.length?"value"===s&&n?(a=[e[e.length-1].getValue(),e[0].getValue()],"asc"===n&&a.reverse()):(a=[1/0,-1/0],TP(e,(function(t){var e=t.getValue(s);e<a[0]&&(a[0]=e),e>a[1]&&(a[1]=e)}))):a=[NaN,NaN],{sum:i,dataExtent:a}}function HP(t,e,n){for(var i=0,r=1/0,o=0,a=void 0,s=t.length;o<s;o++)a=t[o].getLayout().area,a&&(a<r&&(r=a),a>i&&(i=a));var l=t.area*t.area,u=e*e*n;return l?SP(u*i/l,l/(u*r)):1/0}function jP(t,e,n,i,r){var o=e===n.width?0:1,a=1-o,s=["x","y"],l=["width","height"],u=n[s[o]],c=e?t.area/e:0;(r||c>n[l[a]])&&(c=n[l[a]]);for(var d=0,h=t.length;d<h;d++){var f=t[d],p={},m=c?f.getLayout().area/c:0,g=p[l[a]]=SP(c-2*i,0),y=n[s[o]]+n[l[o]]-u,v=d===h-1||y<m?y:m,_=p[l[o]]=SP(v-2*i,0);p[s[a]]=n[s[a]]+LP(i,g/2),p[s[o]]=u+LP(i,_/2),u+=v,f.setLayout(p,!0)}n[s[a]]+=c,n[l[a]]-=c}function zP(t,e,n,i,r){var o,a=(e||{}).node,s=[i,r];if(!a||a===n)return s;var l=i*r,u=l*t.option.zoomToNodeRatio;while(o=a.parentNode){for(var c=0,d=o.children,h=0,f=d.length;h<f;h++)c+=d[h].getValue();var p=a.getValue();if(0===p)return s;u*=c/p;var m=o.getModel(),g=m.get(DP),y=Math.max(g,FP(m));u+=4*g*g+(3*g+y)*Math.pow(u,.5),u>Vr&&(u=Vr),a=o}u<l&&(u=l);var v=Math.pow(u/l,.5);return[i*v,r*v]}function WP(t,e,n){if(e)return{x:e.x,y:e.y};var i={x:0,y:0};if(!n)return i;var r=n.node,o=r.getLayout();if(!o)return i;var a=[o.width/2,o.height/2],s=r;while(s){var l=s.getLayout();a[0]+=l.x,a[1]+=l.y,s=s.parentNode}return{x:t.width/2-a[0],y:t.height/2-a[1]}}function BP(t,e,n,i,r){var o=t.getLayout(),a=n[r],s=a&&a===t;if(!(a&&!s||r===n.length&&t!==i)){t.setLayout({isInView:!0,invisible:!s&&!e.intersect(o),isAboveViewRoot:s},!0);var l=new Ji(e.x-o.x,e.y-o.y,e.width,e.height);TP(t.viewChildren||[],(function(t){BP(t,l,n,i,r+1)}))}}function FP(t){return t.get(IP)?t.get(AP):0}function VP(t){t.registerSeriesModel(MO),t.registerChartView(qO),t.registerVisual(mP),t.registerLayout(YP),vO(t)}function GP(t){var e=t.findComponents({mainType:"legend"});e&&e.length&&t.eachSeriesByType("graph",(function(t){var n=t.getCategoriesData(),i=t.getGraph(),r=i.data,o=n.mapArray(n.getName);r.filterSelf((function(t){var n=r.getItemModel(t),i=n.getShallow("category");if(null!=i){"number"===typeof i&&(i=o[i]);for(var a=0;a<e.length;a++)if(!e[a].isSelected(i))return!1}return!0}))}))}function $P(t){var e={};t.eachSeriesByType("graph",(function(t){var n=t.getCategoriesData(),i=t.getData(),r={};n.each((function(i){var o=n.getName(i);r["ec-"+o]=i;var a=n.getItemModel(i),s=a.getModel("itemStyle").getItemStyle();s.fill||(s.fill=t.getColorFromPalette(o,e)),n.setItemVisual(i,"style",s);for(var l=["symbol","symbolSize","symbolKeepAspect"],u=0;u<l.length;u++){var c=a.getShallow(l[u],!0);null!=c&&n.setItemVisual(i,l[u],c)}})),n.count()&&i.each((function(t){var e=i.getItemModel(t),o=e.getShallow("category");if(null!=o){"string"===typeof o&&(o=r["ec-"+o]);var a=n.getItemVisual(o,"style"),s=i.ensureUniqueItemVisual(t,"style");I(s,a);for(var l=["symbol","symbolSize","symbolKeepAspect"],u=0;u<l.length;u++)i.setItemVisual(t,l[u],n.getItemVisual(o,l[u]))}}))}))}function UP(t){return t instanceof Array||(t=[t,t]),t}function qP(t){t.eachSeriesByType("graph",(function(t){var e=t.getGraph(),n=t.getEdgeData(),i=UP(t.get("edgeSymbol")),r=UP(t.get("edgeSymbolSize"));n.setVisual("fromSymbol",i&&i[0]),n.setVisual("toSymbol",i&&i[1]),n.setVisual("fromSymbolSize",r&&r[0]),n.setVisual("toSymbolSize",r&&r[1]),n.setVisual("style",t.getModel("lineStyle").getLineStyle()),n.each((function(t){var i=n.getItemModel(t),r=e.getEdgeByIndex(t),o=UP(i.getShallow("symbol",!0)),a=UP(i.getShallow("symbolSize",!0)),s=i.getModel("lineStyle").getLineStyle(),l=n.ensureUniqueItemVisual(t,"style");switch(I(l,s),l.stroke){case"source":var u=r.node1.getVisual("style");l.stroke=u&&u.fill;break;case"target":u=r.node2.getVisual("style");l.stroke=u&&u.fill;break}o[0]&&r.setVisual("fromSymbol",o[0]),o[1]&&r.setVisual("toSymbol",o[1]),a[0]&&r.setVisual("fromSymbolSize",a[0]),a[1]&&r.setVisual("toSymbolSize",a[1])}))}))}var XP="--\x3e",ZP=function(t){return t.get("autoCurveness")||null},JP=function(t,e){var n=ZP(t),i=20,r=[];if("number"===typeof n)i=n;else if($(n))return void(t.__curvenessList=n);e>i&&(i=e);var o=i%2?i+2:i+3;r=[];for(var a=0;a<o;a++)r.push((a%2?a+1:a)/10*(a%2?-1:1));t.__curvenessList=r},KP=function(t,e,n){var i=[t.id,t.dataIndex].join("."),r=[e.id,e.dataIndex].join(".");return[n.uid,i,r].join(XP)},QP=function(t){var e=t.split(XP);return[e[0],e[2],e[1]].join(XP)},tE=function(t,e){var n=KP(t.node1,t.node2,e);return e.__edgeMap[n]},eE=function(t,e){var n=nE(KP(t.node1,t.node2,e),e),i=nE(KP(t.node2,t.node1,e),e);return n+i},nE=function(t,e){var n=e.__edgeMap;return n[t]?n[t].length:0};function iE(t){ZP(t)&&(t.__curvenessList=[],t.__edgeMap={},JP(t))}function rE(t,e,n,i){if(ZP(n)){var r=KP(t,e,n),o=n.__edgeMap,a=o[QP(r)];o[r]&&!a?o[r].isForward=!0:a&&o[r]&&(a.isForward=!0,o[r].isForward=!1),o[r]=o[r]||[],o[r].push(i)}}function oE(t,e,n,i){var r=ZP(e),o=$(r);if(!r)return null;var a=tE(t,e);if(!a)return null;for(var s=-1,l=0;l<a.length;l++)if(a[l]===n){s=l;break}var u=eE(t,e);JP(e,u),t.lineStyle=t.lineStyle||{};var c=KP(t.node1,t.node2,e),d=e.__curvenessList,h=o||u%2?0:1;if(a.isForward)return d[h+s];var f=QP(c),p=nE(f,e),m=d[s+p+h];return i?o?r&&0===r[0]?(p+h)%2?m:-m:((p%2?0:1)+h)%2?m:-m:(p+h)%2?m:-m:d[s+p+h]}function aE(t){var e=t.coordinateSystem;if(!e||"view"===e.type){var n=t.getGraph();n.eachNode((function(t){var e=t.getModel();t.setLayout([+e.get("x"),+e.get("y")])})),sE(n,t)}}function sE(t,e){t.eachEdge((function(t,n){var i=st(t.getModel().get(["lineStyle","curveness"]),-oE(t,e,n,!0),0),r=Lt(t.node1.getLayout()),o=Lt(t.node2.getLayout()),a=[r,o];+i&&a.push([(r[0]+o[0])/2-(r[1]-o[1])*i,(r[1]+o[1])/2-(o[0]-r[0])*i]),t.setLayout(a)}))}function lE(t,e){t.eachSeriesByType("graph",(function(t){var e=t.get("layout"),n=t.coordinateSystem;if(n&&"view"!==n.type){var i=t.getData(),r=[];R(n.dimensions,(function(t){r=r.concat(i.mapDimensionsAll(t))}));for(var o=0;o<i.count();o++){for(var a=[],s=!1,l=0;l<r.length;l++){var u=i.get(r[l],o);isNaN(u)||(s=!0),a.push(u)}s?i.setItemLayout(o,n.dataToPoint(a)):i.setItemLayout(o,[NaN,NaN])}sE(i.graph,t)}else e&&"none"!==e||aE(t)}))}function uE(t){var e=t.coordinateSystem;if("view"!==e.type)return 1;var n=t.option.nodeScaleRatio,i=e.scaleX,r=e.getZoom(),o=(r-1)*n+1;return o/i}function cE(t){var e=t.getVisual("symbolSize");return e instanceof Array&&(e=(e[0]+e[1])/2),+e}var dE=Math.PI,hE=[];function fE(t,e){var n=t.coordinateSystem;if(!n||"view"===n.type){var i=n.getBoundingRect(),r=t.getData(),o=r.graph,a=i.width/2+i.x,s=i.height/2+i.y,l=Math.min(i.width,i.height)/2,u=r.count();r.setLayout({cx:a,cy:s}),u&&(pE[e](t,o,r,l,a,s,u),o.eachEdge((function(e,n){var i,r=st(e.getModel().get(["lineStyle","curveness"]),oE(e,t,n),0),o=Lt(e.node1.getLayout()),l=Lt(e.node2.getLayout()),u=(o[0]+l[0])/2,c=(o[1]+l[1])/2;+r&&(r*=3,i=[a*r+u*(1-r),s*r+c*(1-r)]),e.setLayout([o,l,i])})))}}var pE={value:function(t,e,n,i,r,o,a){var s=0,l=n.getSum("value"),u=2*Math.PI/(l||a);e.eachNode((function(t){var e=t.getValue("value"),n=u*(l?e:1)/2;s+=n,t.setLayout([i*Math.cos(s)+r,i*Math.sin(s)+o]),s+=n}))},symbolSize:function(t,e,n,i,r,o,a){var s=0;hE.length=a;var l=uE(t);e.eachNode((function(t){var e=cE(t);isNaN(e)&&(e=2),e<0&&(e=0),e*=l;var n=Math.asin(e/2/i);isNaN(n)&&(n=dE/2),hE[t.dataIndex]=n,s+=2*n}));var u=(2*dE-s)/a/2,c=0;e.eachNode((function(t){var e=u+hE[t.dataIndex];c+=e,t.setLayout([i*Math.cos(c)+r,i*Math.sin(c)+o]),c+=e}))}};function mE(t){t.eachSeriesByType("graph",(function(t){"circular"===t.get("layout")&&fE(t,"symbolSize")}))}var gE=Dt;function yE(t,e,n){for(var i=t,r=e,o=n.rect,a=o.width,s=o.height,l=[o.x+a/2,o.y+s/2],u=null==n.gravity?.1:n.gravity,c=0;c<i.length;c++){var d=i[c];d.p||(d.p=Mt(a*(Math.random()-.5)+l[0],s*(Math.random()-.5)+l[1])),d.pp=Lt(d.p),d.edges=null}var h,f,p=null==n.friction?.6:n.friction,m=p;return{warmUp:function(){m=.8*p},setFixed:function(t){i[t].fixed=!0},setUnfixed:function(t){i[t].fixed=!1},beforeStep:function(t){h=t},afterStep:function(t){f=t},step:function(t){h&&h(i,r);for(var e=[],n=i.length,o=0;o<r.length;o++){var a=r[o];if(!a.ignoreForceLayout){var s=a.n1,c=a.n2;Ct(e,c.p,s.p);var d=It(e)-a.d,p=c.w/(s.w+c.w);isNaN(p)&&(p=0),Ot(e,e),!s.fixed&&gE(s.p,s.p,e,p*d*m),!c.fixed&&gE(c.p,c.p,e,-(1-p)*d*m)}}for(o=0;o<n;o++){var g=i[o];g.fixed||(Ct(e,l,g.p),gE(g.p,g.p,e,u*m))}for(o=0;o<n;o++){s=i[o];for(var y=o+1;y<n;y++){c=i[y];Ct(e,c.p,s.p);d=It(e);0===d&&(kt(e,Math.random()-.5,Math.random()-.5),d=1);var v=(s.rep+c.rep)/d/d;!s.fixed&&gE(s.pp,s.pp,e,v),!c.fixed&&gE(c.pp,c.pp,e,-v)}}var _=[];for(o=0;o<n;o++){g=i[o];g.fixed||(Ct(_,g.p,g.pp),gE(g.p,g.p,_,m),St(g.pp,g.p))}m*=.992;var b=m<.01;f&&f(i,r,b),t&&t(b)}}}function vE(t){t.eachSeriesByType("graph",(function(t){var e=t.coordinateSystem;if(!e||"view"===e.type)if("force"===t.get("layout")){var n=t.preservedPoints||{},i=t.getGraph(),r=i.data,o=i.edgeData,a=t.getModel("force"),s=a.get("initLayout");t.preservedPoints?r.each((function(t){var e=r.getId(t);r.setItemLayout(t,n[e]||[NaN,NaN])})):s&&"none"!==s?"circular"===s&&fE(t,"value"):aE(t);var l=r.getDataExtent("value"),u=o.getDataExtent("value"),c=a.get("repulsion"),d=a.get("edgeLength"),h=$(c)?c:[c,c],f=$(d)?d:[d,d];f=[f[1],f[0]];var p=r.mapArray("value",(function(t,e){var n=r.getItemLayout(e),i=Er(t,l,h);return isNaN(i)&&(i=(h[0]+h[1])/2),{w:i,rep:i,fixed:r.getItemModel(e).get("fixed"),p:!n||isNaN(n[0])||isNaN(n[1])?null:n}})),m=o.mapArray("value",(function(e,n){var r=i.getEdgeByIndex(n),o=Er(e,u,f);isNaN(o)&&(o=(f[0]+f[1])/2);var a=r.getModel(),s=st(r.getModel().get(["lineStyle","curveness"]),-oE(r,t,n,!0),0);return{n1:p[r.node1.dataIndex],n2:p[r.node2.dataIndex],d:o,curveness:s,ignoreForceLayout:a.get("ignoreForceLayout")}})),g=e.getBoundingRect(),y=yE(p,m,{rect:g,gravity:a.get("gravity"),friction:a.get("friction")});y.beforeStep((function(t,e){for(var n=0,r=t.length;n<r;n++)t[n].fixed&&St(t[n].p,i.getNodeByIndex(n).getLayout())})),y.afterStep((function(t,e,o){for(var a=0,s=t.length;a<s;a++)t[a].fixed||i.getNodeByIndex(a).setLayout(t[a].p),n[r.getId(a)]=t[a].p;for(a=0,s=e.length;a<s;a++){var l=e[a],u=i.getEdgeByIndex(a),c=l.n1.p,d=l.n2.p,h=u.getLayout();h=h?h.slice():[],h[0]=h[0]||[],h[1]=h[1]||[],St(h[0],c),St(h[1],d),+l.curveness&&(h[2]=[(c[0]+d[0])/2-(c[1]-d[1])*l.curveness,(c[1]+d[1])/2-(d[0]-c[0])*l.curveness]),u.setLayout(h)}})),t.forceLayout=y,t.preservedPoints=n,y.step()}else t.forceLayout=null}))}function _E(t,e,n){var i=I(t.getBoxLayoutParams(),{aspect:n});return dp(i,{width:e.getWidth(),height:e.getHeight()})}function bE(t,e){var n=[];return t.eachSeriesByType("graph",(function(t){var i=t.get("coordinateSystem");if(!i||"view"===i){var r=t.getData(),o=r.mapArray((function(t){var e=r.getItemModel(t);return[+e.get("x"),+e.get("y")]})),a=[],s=[];ms(o,a,s),s[0]-a[0]===0&&(s[0]+=1,a[0]-=1),s[1]-a[1]===0&&(s[1]+=1,a[1]-=1);var l=(s[0]-a[0])/(s[1]-a[1]),u=_E(t,e,l);isNaN(l)&&(a=[u.x,u.y],s=[u.x+u.width,u.y+u.height]);var c=s[0]-a[0],d=s[1]-a[1],h=u.width,f=u.height,p=t.coordinateSystem=new iY;p.zoomLimit=t.get("scaleLimit"),p.setBoundingRect(a[0],a[1],c,d),p.setViewRect(u.x,u.y,h,f),p.setCenter(t.get("center")),p.setZoom(t.get("zoom")),n.push(p)}})),n}var xE=dd.prototype,wE=gd.prototype,ME=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.percent=1}return t}();(function(t){function e(){return null!==t&&t.apply(this,arguments)||this}o(e,t)})(ME);function SE(t){return isNaN(+t.cpx1)||isNaN(+t.cpy1)}var LE=function(t){function e(e){var n=t.call(this,e)||this;return n.type="ec-line",n}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new ME},e.prototype.buildPath=function(t,e){SE(e)?xE.buildPath.call(this,t,e):wE.buildPath.call(this,t,e)},e.prototype.pointAt=function(t){return SE(this.shape)?xE.pointAt.call(this,t):wE.pointAt.call(this,t)},e.prototype.tangentAt=function(t){var e=this.shape,n=SE(e)?[e.x2-e.x1,e.y2-e.y1]:wE.tangentAt.call(this,t);return Ot(n,n)},e}(pl),kE=LE,TE=["fromSymbol","toSymbol"];function DE(t){return"_"+t+"Type"}function CE(t,e,n){var i=e.getItemVisual(n,t);if(i&&"none"!==i){var r=e.getItemVisual(n,t+"Size"),o=e.getItemVisual(n,t+"Rotate"),a=e.getItemVisual(n,t+"Offset")||0,s=e.getItemVisual(n,t+"KeepAspect"),l=$(r)?r:[r,r],u=$(a)?a:[a,a];u[0]=Nr(u[0],l[0]),u[1]=Nr(at(u[1],u[0]),l[1]);var c=j_(i,-l[0]/2+u[0],-l[1]/2+u[1],l[0],l[1],null,s);return c.__specifiedRotation=null==o||isNaN(o)?void 0:+o*Math.PI/180||0,c.name=t,c}}function IE(t){var e=new kE({name:"line",subPixelOptimize:!0});return AE(e.shape,t),e}function AE(t,e){t.x1=e[0][0],t.y1=e[0][1],t.x2=e[1][0],t.y2=e[1][1],t.percent=1;var n=e[2];n?(t.cpx1=n[0],t.cpy1=n[1]):(t.cpx1=NaN,t.cpy1=NaN)}var YE=function(t){function e(e,n,i){var r=t.call(this)||this;return r._createLine(e,n,i),r}return o(e,t),e.prototype._createLine=function(t,e,n){var i=t.hostModel,r=t.getItemLayout(e),o=IE(r);o.shape.percent=0,eh(o,{shape:{percent:1}},i,e),this.add(o),R(TE,(function(n){var i=CE(n,t,e);this.add(i),this[DE(n)]=t.getItemVisual(e,n)}),this),this._updateCommonStl(t,e,n)},e.prototype.updateData=function(t,e,n){var i=t.hostModel,r=this.childOfName("line"),o=t.getItemLayout(e),a={shape:{}};AE(a.shape,o),th(r,a,i,e),R(TE,(function(n){var i=t.getItemVisual(e,n),r=DE(n);if(this[r]!==i){this.remove(this.childOfName(n));var o=CE(n,t,e);this.add(o)}this[r]=i}),this),this._updateCommonStl(t,e,n)},e.prototype.getLinePath=function(){return this.childAt(0)},e.prototype._updateCommonStl=function(t,e,n){var i=t.hostModel,r=this.childOfName("line"),o=n&&n.emphasisLineStyle,a=n&&n.blurLineStyle,s=n&&n.selectLineStyle,l=n&&n.labelStatesModels;if(!n||t.hasItemOption){var u=t.getItemModel(e);o=u.getModel(["emphasis","lineStyle"]).getLineStyle(),a=u.getModel(["blur","lineStyle"]).getLineStyle(),s=u.getModel(["select","lineStyle"]).getLineStyle(),l=Sh(u)}var c=t.getItemVisual(e,"style"),d=c.stroke;r.useStyle(c),r.style.fill=null,r.style.strokeNoScale=!0,r.ensureState("emphasis").style=o,r.ensureState("blur").style=a,r.ensureState("select").style=s,R(TE,(function(t){var e=this.childOfName(t);if(e){e.setColor(d),e.style.opacity=c.opacity;for(var n=0;n<Ql.length;n++){var i=Ql[n],o=r.getState(i);if(o){var a=o.style||{},s=e.ensureState(i),l=s.style||(s.style={});null!=a.stroke&&(l[e.__isEmptyBrush?"stroke":"fill"]=a.stroke),null!=a.opacity&&(l.opacity=a.opacity)}}e.markRedraw()}}),this);var h=i.getRawValue(e);Mh(this,l,{labelDataIndex:e,labelFetcher:{getFormattedLabel:function(e,n){return i.getFormattedLabel(e,n,t.dataType)}},inheritColor:d||"#000",defaultOpacity:c.opacity,defaultText:(null==h?t.getName(e):isFinite(h)?Rr(h):h)+""});var f=this.getTextContent();if(f){var p=l.normal;f.__align=f.style.align,f.__verticalAlign=f.style.verticalAlign,f.__position=p.get("position")||"middle";var m=p.get("distance");$(m)||(m=[m,m]),f.__labelDistance=m}this.setTextConfig({position:null,local:!0,inside:!1}),$u(this)},e.prototype.highlight=function(){Cu(this)},e.prototype.downplay=function(){Iu(this)},e.prototype.updateLayout=function(t,e){this.setLinePoints(t.getItemLayout(e))},e.prototype.setLinePoints=function(t){var e=this.childOfName("line");AE(e.shape,t),e.dirty()},e.prototype.beforeUpdate=function(){var t=this,e=t.childOfName("fromSymbol"),n=t.childOfName("toSymbol"),i=t.getTextContent();if(e||n||i&&!i.ignore){var r=1,o=this.parent;while(o)o.scaleX&&(r/=o.scaleX),o=o.parent;var a=t.childOfName("line");if(this.__dirty||a.__dirty){var s=a.shape.percent,l=a.pointAt(0),u=a.pointAt(s),c=Ct([],u,l);if(Ot(c,c),e&&(e.setPosition(l),M(e,0),e.scaleX=e.scaleY=r*s,e.markRedraw()),n&&(n.setPosition(u),M(n,1),n.scaleX=n.scaleY=r*s,n.markRedraw()),i&&!i.ignore){i.x=i.y=0,i.originX=i.originY=0;var d=void 0,h=void 0,f=i.__labelDistance,p=f[0]*r,m=f[1]*r,g=s/2,y=a.tangentAt(g),v=[y[1],-y[0]],_=a.pointAt(g);v[1]>0&&(v[0]=-v[0],v[1]=-v[1]);var b=y[0]<0?-1:1;if("start"!==i.__position&&"end"!==i.__position){var x=-Math.atan2(y[1],y[0]);u[0]<l[0]&&(x=Math.PI+x),i.rotation=x}var w=void 0;switch(i.__position){case"insideStartTop":case"insideMiddleTop":case"insideEndTop":case"middle":w=-m,h="bottom";break;case"insideStartBottom":case"insideMiddleBottom":case"insideEndBottom":w=m,h="top";break;default:w=0,h="middle"}switch(i.__position){case"end":i.x=c[0]*p+u[0],i.y=c[1]*m+u[1],d=c[0]>.8?"left":c[0]<-.8?"right":"center",h=c[1]>.8?"top":c[1]<-.8?"bottom":"middle";break;case"start":i.x=-c[0]*p+l[0],i.y=-c[1]*m+l[1],d=c[0]>.8?"right":c[0]<-.8?"left":"center",h=c[1]>.8?"bottom":c[1]<-.8?"top":"middle";break;case"insideStartTop":case"insideStart":case"insideStartBottom":i.x=p*b+l[0],i.y=l[1]+w,d=y[0]<0?"right":"left",i.originX=-p*b,i.originY=-w;break;case"insideMiddleTop":case"insideMiddle":case"insideMiddleBottom":case"middle":i.x=_[0],i.y=_[1]+w,d="center",i.originY=-w;break;case"insideEndTop":case"insideEnd":case"insideEndBottom":i.x=-p*b+u[0],i.y=u[1]+w,d=y[0]>=0?"right":"left",i.originX=p*b,i.originY=-w;break}i.scaleX=i.scaleY=r,i.setStyle({verticalAlign:i.__verticalAlign||h,align:i.__align||d})}}}function M(t,e){var n=t.__specifiedRotation;if(null==n){var i=a.tangentAt(e);t.attr("rotation",(1===e?-1:1)*Math.PI/2-Math.atan2(i[1],i[0]))}else t.attr("rotation",n)}},e}(Mr),OE=YE,PE=function(){function t(t){this.group=new Mr,this._LineCtor=t||OE}return t.prototype.isPersistent=function(){return!0},t.prototype.updateData=function(t){var e=this,n=this,i=n.group,r=n._lineData;n._lineData=t,r||i.removeAll();var o=NE(t);t.diff(r).add((function(n){e._doAdd(t,n,o)})).update((function(n,i){e._doUpdate(r,t,i,n,o)})).remove((function(t){i.remove(r.getItemGraphicEl(t))})).execute()},t.prototype.updateLayout=function(){var t=this._lineData;t&&t.eachItemGraphicEl((function(e,n){e.updateLayout(t,n)}),this)},t.prototype.incrementalPrepareUpdate=function(t){this._seriesScope=NE(t),this._lineData=null,this.group.removeAll()},t.prototype.incrementalUpdate=function(t,e){function n(t){t.isGroup||EE(t)||(t.incremental=!0,t.ensureState("emphasis").hoverLayer=!0)}for(var i=t.start;i<t.end;i++){var r=e.getItemLayout(i);if(HE(r)){var o=new this._LineCtor(e,i,this._seriesScope);o.traverse(n),this.group.add(o),e.setItemGraphicEl(i,o)}}},t.prototype.remove=function(){this.group.removeAll()},t.prototype._doAdd=function(t,e,n){var i=t.getItemLayout(e);if(HE(i)){var r=new this._LineCtor(t,e,n);t.setItemGraphicEl(e,r),this.group.add(r)}},t.prototype._doUpdate=function(t,e,n,i,r){var o=t.getItemGraphicEl(n);HE(e.getItemLayout(i))?(o?o.updateData(e,i,r):o=new this._LineCtor(e,i,r),e.setItemGraphicEl(i,o),this.group.add(o)):this.group.remove(o)},t}();function EE(t){return t.animators&&t.animators.length>0}function NE(t){var e=t.hostModel;return{lineStyle:e.getModel("lineStyle").getLineStyle(),emphasisLineStyle:e.getModel(["emphasis","lineStyle"]).getLineStyle(),blurLineStyle:e.getModel(["blur","lineStyle"]).getLineStyle(),selectLineStyle:e.getModel(["select","lineStyle"]).getLineStyle(),labelStatesModels:Sh(e)}}function RE(t){return isNaN(t[0])||isNaN(t[1])}function HE(t){return!RE(t[0])&&!RE(t[1])}var jE=PE,zE=[],WE=[],BE=[],FE=ts,VE=Rt,GE=Math.abs;function $E(t,e,n){for(var i,r=t[0],o=t[1],a=t[2],s=1/0,l=n*n,u=.1,c=.1;c<=.9;c+=.1){zE[0]=FE(r[0],o[0],a[0],c),zE[1]=FE(r[1],o[1],a[1],c);var d=GE(VE(zE,e)-l);d<s&&(s=d,i=c)}for(var h=0;h<32;h++){var f=i+u;WE[0]=FE(r[0],o[0],a[0],i),WE[1]=FE(r[1],o[1],a[1],i),BE[0]=FE(r[0],o[0],a[0],f),BE[1]=FE(r[1],o[1],a[1],f);d=VE(WE,e)-l;if(GE(d)<.01)break;var p=VE(BE,e)-l;u/=2,d<0?p>=0?i+=u:i-=u:p>=0?i-=u:i+=u}return i}function UE(t,e){var n=[],i=rs,r=[[],[],[]],o=[[],[]],a=[];e/=2,t.eachEdge((function(t,s){var l=t.getLayout(),u=t.getVisual("fromSymbol"),c=t.getVisual("toSymbol");l.__original||(l.__original=[Lt(l[0]),Lt(l[1])],l[2]&&l.__original.push(Lt(l[2])));var d=l.__original;if(null!=l[2]){if(St(r[0],d[0]),St(r[1],d[2]),St(r[2],d[1]),u&&"none"!==u){var h=cE(t.node1),f=$E(r,d[0],h*e);i(r[0][0],r[1][0],r[2][0],f,n),r[0][0]=n[3],r[1][0]=n[4],i(r[0][1],r[1][1],r[2][1],f,n),r[0][1]=n[3],r[1][1]=n[4]}if(c&&"none"!==c){h=cE(t.node2),f=$E(r,d[1],h*e);i(r[0][0],r[1][0],r[2][0],f,n),r[1][0]=n[1],r[2][0]=n[2],i(r[0][1],r[1][1],r[2][1],f,n),r[1][1]=n[1],r[2][1]=n[2]}St(l[0],r[0]),St(l[1],r[2]),St(l[2],r[1])}else{if(St(o[0],d[0]),St(o[1],d[1]),Ct(a,o[1],o[0]),Ot(a,a),u&&"none"!==u){h=cE(t.node1);Dt(o[0],o[0],a,h*e)}if(c&&"none"!==c){h=cE(t.node2);Dt(o[1],o[1],a,-h*e)}St(l[0],o[0]),St(l[1],o[1])}}))}function qE(t){return"view"===t.type}var XE=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){var n=new tk,i=new jE,r=this.group;this._controller=new CA(e.getZr()),this._controllerHost={target:r},r.add(n.group),r.add(i.group),this._symbolDraw=n,this._lineDraw=i,this._firstRender=!0},e.prototype.render=function(t,e,n){var i=this,r=t.coordinateSystem;this._model=t;var o=this._symbolDraw,a=this._lineDraw,s=this.group;if(qE(r)){var l={x:r.x,y:r.y,scaleX:r.scaleX,scaleY:r.scaleY};this._firstRender?s.attr(l):th(s,l,t)}UE(t.getGraph(),uE(t));var u=t.getData();o.updateData(u);var c=t.getEdgeData();a.updateData(c),this._updateNodeAndLinkScale(),this._updateController(t,e,n),clearTimeout(this._layoutTimeout);var d=t.forceLayout,h=t.get(["force","layoutAnimation"]);d&&this._startForceLayoutIteration(d,h),u.graph.eachNode((function(t){var e=t.dataIndex,n=t.getGraphicEl(),r=t.getModel();n.off("drag").off("dragend");var o=r.get("draggable");o&&n.on("drag",(function(){d&&(d.warmUp(),!i._layouting&&i._startForceLayoutIteration(d,h),d.setFixed(e),u.setItemLayout(e,[n.x,n.y]))})).on("dragend",(function(){d&&d.setUnfixed(e)})),n.setDraggable(o&&!!d);var a=r.get(["emphasis","focus"]);"adjacency"===a&&($l(n).focus=t.getAdjacentDataIndices())})),u.graph.eachEdge((function(t){var e=t.getGraphicEl(),n=t.getModel().get(["emphasis","focus"]);"adjacency"===n&&($l(e).focus={edge:[t.dataIndex],node:[t.node1.dataIndex,t.node2.dataIndex]})}));var f="circular"===t.get("layout")&&t.get(["circular","rotateLabel"]),p=u.getLayout("cx"),m=u.getLayout("cy");u.eachItemGraphicEl((function(t,e){var n=u.getItemModel(e),i=n.get(["label","rotate"])||0,r=t.getSymbolPath();if(f){var o=u.getItemLayout(e),a=Math.atan2(o[1]-m,o[0]-p);a<0&&(a=2*Math.PI+a);var s=o[0]<p;s&&(a-=Math.PI);var l=s?"left":"right";r.setTextConfig({rotation:-a,position:l,origin:"center"});var c=r.ensureState("emphasis");I(c.textConfig||(c.textConfig={}),{position:l})}else r.setTextConfig({rotation:i*=Math.PI/180})})),this._firstRender=!1},e.prototype.dispose=function(){this._controller&&this._controller.dispose(),this._controllerHost=null},e.prototype._startForceLayoutIteration=function(t,e){var n=this;(function i(){t.step((function(t){n.updateLayout(n._model),(n._layouting=!t)&&(e?n._layoutTimeout=setTimeout(i,16):i())}))})()},e.prototype._updateController=function(t,e,n){var i=this,r=this._controller,o=this._controllerHost,a=this.group;r.setPointerChecker((function(e,i,r){var o=a.getBoundingRect();return o.applyTransform(a.transform),o.contain(i,r)&&!OA(e,n,t)})),qE(t.coordinateSystem)?(r.enable(t.get("roam")),o.zoomLimit=t.get("scaleLimit"),o.zoom=t.coordinateSystem.getZoom(),r.off("pan").off("zoom").on("pan",(function(e){IA(o,e.dx,e.dy),n.dispatchAction({seriesId:t.id,type:"graphRoam",dx:e.dx,dy:e.dy})})).on("zoom",(function(e){AA(o,e.scale,e.originX,e.originY),n.dispatchAction({seriesId:t.id,type:"graphRoam",zoom:e.scale,originX:e.originX,originY:e.originY}),i._updateNodeAndLinkScale(),UE(t.getGraph(),uE(t)),i._lineDraw.updateLayout(),n.updateLabelLayout()}))):r.disable()},e.prototype._updateNodeAndLinkScale=function(){var t=this._model,e=t.getData(),n=uE(t);e.eachItemGraphicEl((function(t,e){t.setSymbolScale(n)}))},e.prototype.updateLayout=function(t){UE(t.getGraph(),uE(t)),this._symbolDraw.updateLayout(),this._lineDraw.updateLayout()},e.prototype.remove=function(t,e){this._symbolDraw&&this._symbolDraw.remove(),this._lineDraw&&this._lineDraw.remove()},e.type="graph",e}(Vy),ZE=XE;function JE(t){return"_EC_"+t}var KE=function(){function t(t){this.type="graph",this.nodes=[],this.edges=[],this._nodesMap={},this._edgesMap={},this._directed=t||!1}return t.prototype.isDirected=function(){return this._directed},t.prototype.addNode=function(t,e){t=null==t?""+e:""+t;var n=this._nodesMap;if(!n[JE(t)]){var i=new QE(t,e);return i.hostGraph=this,this.nodes.push(i),n[JE(t)]=i,i}},t.prototype.getNodeByIndex=function(t){var e=this.data.getRawIndex(t);return this.nodes[e]},t.prototype.getNodeById=function(t){return this._nodesMap[JE(t)]},t.prototype.addEdge=function(t,e,n){var i=this._nodesMap,r=this._edgesMap;if("number"===typeof t&&(t=this.nodes[t]),"number"===typeof e&&(e=this.nodes[e]),t instanceof QE||(t=i[JE(t)]),e instanceof QE||(e=i[JE(e)]),t&&e){var o=t.id+"-"+e.id,a=new tN(t,e,n);return a.hostGraph=this,this._directed&&(t.outEdges.push(a),e.inEdges.push(a)),t.edges.push(a),t!==e&&e.edges.push(a),this.edges.push(a),r[o]=a,a}},t.prototype.getEdgeByIndex=function(t){var e=this.edgeData.getRawIndex(t);return this.edges[e]},t.prototype.getEdge=function(t,e){t instanceof QE&&(t=t.id),e instanceof QE&&(e=e.id);var n=this._edgesMap;return this._directed?n[t+"-"+e]:n[t+"-"+e]||n[e+"-"+t]},t.prototype.eachNode=function(t,e){for(var n=this.nodes,i=n.length,r=0;r<i;r++)n[r].dataIndex>=0&&t.call(e,n[r],r)},t.prototype.eachEdge=function(t,e){for(var n=this.edges,i=n.length,r=0;r<i;r++)n[r].dataIndex>=0&&n[r].node1.dataIndex>=0&&n[r].node2.dataIndex>=0&&t.call(e,n[r],r)},t.prototype.breadthFirstTraverse=function(t,e,n,i){if(e instanceof QE||(e=this._nodesMap[JE(e)]),e){for(var r="out"===n?"outEdges":"in"===n?"inEdges":"edges",o=0;o<this.nodes.length;o++)this.nodes[o].__visited=!1;if(!t.call(i,e,null)){var a=[e];while(a.length){var s=a.shift(),l=s[r];for(o=0;o<l.length;o++){var u=l[o],c=u.node1===s?u.node2:u.node1;if(!c.__visited){if(t.call(i,c,s))return;a.push(c),c.__visited=!0}}}}}},t.prototype.update=function(){for(var t=this.data,e=this.edgeData,n=this.nodes,i=this.edges,r=0,o=n.length;r<o;r++)n[r].dataIndex=-1;for(r=0,o=t.count();r<o;r++)n[t.getRawIndex(r)].dataIndex=r;e.filterSelf((function(t){var n=i[e.getRawIndex(t)];return n.node1.dataIndex>=0&&n.node2.dataIndex>=0}));for(r=0,o=i.length;r<o;r++)i[r].dataIndex=-1;for(r=0,o=e.count();r<o;r++)i[e.getRawIndex(r)].dataIndex=r},t.prototype.clone=function(){for(var e=new t(this._directed),n=this.nodes,i=this.edges,r=0;r<n.length;r++)e.addNode(n[r].id,n[r].dataIndex);for(r=0;r<i.length;r++){var o=i[r];e.addEdge(o.node1.id,o.node2.id,o.dataIndex)}return e},t}(),QE=function(){function t(t,e){this.inEdges=[],this.outEdges=[],this.edges=[],this.dataIndex=-1,this.id=null==t?"":t,this.dataIndex=null==e?-1:e}return t.prototype.degree=function(){return this.edges.length},t.prototype.inDegree=function(){return this.inEdges.length},t.prototype.outDegree=function(){return this.outEdges.length},t.prototype.getModel=function(t){if(!(this.dataIndex<0)){var e=this.hostGraph,n=e.data.getItemModel(this.dataIndex);return n.getModel(t)}},t.prototype.getAdjacentDataIndices=function(){for(var t={edge:[],node:[]},e=0;e<this.edges.length;e++){var n=this.edges[e];n.dataIndex<0||(t.edge.push(n.dataIndex),t.node.push(n.node1.dataIndex,n.node2.dataIndex))}return t},t}(),tN=function(){function t(t,e,n){this.dataIndex=-1,this.node1=t,this.node2=e,this.dataIndex=null==n?-1:n}return t.prototype.getModel=function(t){if(!(this.dataIndex<0)){var e=this.hostGraph,n=e.edgeData.getItemModel(this.dataIndex);return n.getModel(t)}},t.prototype.getAdjacentDataIndices=function(){return{edge:[this.dataIndex],node:[this.node1.dataIndex,this.node2.dataIndex]}},t}();function eN(t,e){return{getValue:function(n){var i=this[t][e];return i.get(i.getDimension(n||"value"),this.dataIndex)},setVisual:function(n,i){this.dataIndex>=0&&this[t][e].setItemVisual(this.dataIndex,n,i)},getVisual:function(n){return this[t][e].getItemVisual(this.dataIndex,n)},setLayout:function(n,i){this.dataIndex>=0&&this[t][e].setItemLayout(this.dataIndex,n,i)},getLayout:function(){return this[t][e].getItemLayout(this.dataIndex)},getGraphicEl:function(){return this[t][e].getItemGraphicEl(this.dataIndex)},getRawIndex:function(){return this[t][e].getRawIndex(this.dataIndex)}}}E(QE,eN("hostGraph","data")),E(tN,eN("hostGraph","edgeData"));var nN=KE;function iN(t,e,n,i,r){for(var o=new nN(i),a=0;a<t.length;a++)o.addNode(ot(t[a].id,t[a].name,a),a);var s=[],l=[],u=0;for(a=0;a<e.length;a++){var c=e[a],d=c.source,h=c.target;o.addEdge(d,h,u)&&(l.push(c),s.push(ot(wo(c.id,null),d+" > "+h)),u++)}var f,p=n.get("coordinateSystem");if("cartesian2d"===p||"polar"===p)f=BL(t,n);else{var m=vm.get(p),g=m&&m.dimensions||[];O(g,"value")<0&&g.concat(["value"]);var y=AL(t,{coordDimensions:g});f=new kL(y,n),f.initData(t)}var v=new kL(["value"],n);return v.initData(l,s),r&&r(f,v),KY({mainData:f,struct:o,structAttr:"graph",datas:{node:f,edge:v},datasAttr:{node:"data",edge:"edgeData"}}),o.update(),o}var rN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments);var n=this;function i(){return n._categoriesData}this.legendVisualProvider=new KT(i,i),this.fillDataTextStyle(e.edges||e.links),this._updateCategoriesData()},e.prototype.mergeOption=function(e){t.prototype.mergeOption.apply(this,arguments),this.fillDataTextStyle(e.edges||e.links),this._updateCategoriesData()},e.prototype.mergeDefaultAndTheme=function(e){t.prototype.mergeDefaultAndTheme.apply(this,arguments),lo(e,"edgeLabel",["show"])},e.prototype.getInitialData=function(t,e){var n=t.edges||t.links||[],i=t.data||t.nodes||[],r=this;if(i&&n){iE(this);var o=iN(i,n,this,!0,a);return R(o.edges,(function(t){rE(t.node1,t.node2,this,t.dataIndex)}),this),o.data}function a(t,e){t.wrapMethod("getItemModel",(function(t){var e=r._categoriesModels,n=t.getShallow("category"),i=e[n];return i&&(i.parentModel=t.parentModel,t.parentModel=i),t}));var n=qh.prototype.getModel;function i(t,e){var i=n.call(this,t,e);return i.resolveParentPath=o,i}function o(t){if(t&&("label"===t[0]||"label"===t[1])){var e=t.slice();return"label"===t[0]?e[0]="edgeLabel":"label"===t[1]&&(e[1]="edgeLabel"),e}return t}e.wrapMethod("getItemModel",(function(t){return t.resolveParentPath=o,t.getModel=i,t}))}},e.prototype.getGraph=function(){return this.getData().graph},e.prototype.getEdgeData=function(){return this.getGraph().edgeData},e.prototype.getCategoriesData=function(){return this._categoriesData},e.prototype.formatTooltip=function(t,e,n){if("edge"===n){var i=this.getData(),r=this.getDataParams(t,n),o=i.graph.getEdgeByIndex(t),a=i.getName(o.node1.dataIndex),s=i.getName(o.node2.dataIndex),l=[];return null!=a&&l.push(a),null!=s&&l.push(s),oy("nameValue",{name:l.join(" > "),value:r.value,noValue:null==r.value})}var u=_y({series:this,dataIndex:t,multipleSeries:e});return u},e.prototype._updateCategoriesData=function(){var t=H(this.option.categories||[],(function(t){return null!=t.value?t:I({value:0},t)})),e=new kL(["value"],this);e.initData(t),this._categoriesData=e,this._categoriesModels=e.mapArray((function(t){return e.getItemModel(t)}))},e.prototype.setZoom=function(t){this.option.zoom=t},e.prototype.setCenter=function(t){this.option.center=t},e.prototype.isAnimationEnabled=function(){return t.prototype.isAnimationEnabled.call(this)&&!("force"===this.get("layout")&&this.get(["force","layoutAnimation"]))},e.type="series.graph",e.dependencies=["grid","polar","geo","singleAxis","calendar"],e.defaultOption={zlevel:0,z:2,coordinateSystem:"view",legendHoverLink:!0,layout:null,circular:{rotateLabel:!1},force:{initLayout:null,repulsion:[0,50],gravity:.1,friction:.6,edgeLength:30,layoutAnimation:!0},left:"center",top:"center",symbol:"circle",symbolSize:10,edgeSymbol:["none","none"],edgeSymbolSize:10,edgeLabel:{position:"middle",distance:5},draggable:!1,roam:!1,center:null,zoom:1,nodeScaleRatio:.6,label:{show:!1,formatter:"{b}"},itemStyle:{},lineStyle:{color:"#aaa",width:1,opacity:.5},emphasis:{scale:!0,label:{show:!0}},select:{itemStyle:{borderColor:"#212121"}}},e}(Yy),oN=rN,aN={type:"graphRoam",event:"graphRoam",update:"none"};function sN(t){t.registerChartView(ZE),t.registerSeriesModel(oN),t.registerProcessor(GP),t.registerVisual($P),t.registerVisual(qP),t.registerLayout(lE),t.registerLayout(t.PRIORITY.VISUAL.POST_CHART_LAYOUT,mE),t.registerLayout(vE),t.registerCoordinateSystem("graphView",{dimensions:iY.dimensions,create:bE}),t.registerAction({type:"focusNodeAdjacency",event:"focusNodeAdjacency",update:"series:focusNodeAdjacency"},(function(){})),t.registerAction({type:"unfocusNodeAdjacency",event:"unfocusNodeAdjacency",update:"series:unfocusNodeAdjacency"},(function(){})),t.registerAction(aN,(function(t,e){e.eachComponent({mainType:"series",query:t},(function(e){var n=e.coordinateSystem,i=mY(n,t);e.setCenter&&e.setCenter(i.center),e.setZoom&&e.setZoom(i.zoom)}))}))}var lN=function(){function t(){this.angle=0,this.width=10,this.r=10,this.x=0,this.y=0}return t}(),uN=function(t){function e(e){var n=t.call(this,e)||this;return n.type="pointer",n}return o(e,t),e.prototype.getDefaultShape=function(){return new lN},e.prototype.buildPath=function(t,e){var n=Math.cos,i=Math.sin,r=e.r,o=e.width,a=e.angle,s=e.x-n(a)*o*(o>=r/3?1:2),l=e.y-i(a)*o*(o>=r/3?1:2);a=e.angle-Math.PI/2,t.moveTo(s,l),t.lineTo(e.x+n(a)*o,e.y+i(a)*o),t.lineTo(e.x+n(e.angle)*r,e.y+i(e.angle)*r),t.lineTo(e.x-n(a)*o,e.y-i(a)*o),t.lineTo(s,l)},e}(pl),cN=uN;function dN(t,e){var n=t.get("center"),i=e.getWidth(),r=e.getHeight(),o=Math.min(i,r),a=Nr(n[0],e.getWidth()),s=Nr(n[1],e.getHeight()),l=Nr(t.get("radius"),o/2);return{cx:a,cy:s,r:l}}function hN(t,e){var n=null==t?"":t+"";return e&&("string"===typeof e?n=e.replace("{value}",n):"function"===typeof e&&(n=e(t))),n}var fN=2*Math.PI,pN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){this.group.removeAll();var i=t.get(["axisLine","lineStyle","color"]),r=dN(t,n);this._renderMain(t,e,n,i,r),this._data=t.getData()},e.prototype.dispose=function(){},e.prototype._renderMain=function(t,e,n,i,r){for(var o=this.group,a=t.get("clockwise"),s=-t.get("startAngle")/180*Math.PI,l=-t.get("endAngle")/180*Math.PI,u=t.getModel("axisLine"),c=u.get("roundCap"),d=c?dT:qc,h=u.get("show"),f=u.getModel("lineStyle"),p=f.get("width"),m=(l-s)%fN||l===s?(l-s)%fN:fN,g=s,y=0;h&&y<i.length;y++){var v=Math.min(Math.max(i[y][0],0),1);l=s+m*v;var _=new d({shape:{startAngle:g,endAngle:l,cx:r.cx,cy:r.cy,clockwise:a,r0:r.r-p,r:r.r},silent:!0});_.setStyle({fill:i[y][1]}),_.setStyle(f.getLineStyle(["color","width"])),o.add(_),g=l}var b=function(t){if(t<=0)return i[0][1];var e;for(e=0;e<i.length;e++)if(i[e][0]>=t&&(0===e?0:i[e-1][0])<t)return i[e][1];return i[e-1][1]};if(!a){var x=s;s=l,l=x}this._renderTicks(t,e,n,b,r,s,l,a,p),this._renderTitleAndDetail(t,e,n,b,r),this._renderAnchor(t,r),this._renderPointer(t,e,n,b,r,s,l,a,p)},e.prototype._renderTicks=function(t,e,n,i,r,o,a,s,l){for(var u,c,d=this.group,h=r.cx,f=r.cy,p=r.r,m=+t.get("min"),g=+t.get("max"),y=t.getModel("splitLine"),v=t.getModel("axisTick"),_=t.getModel("axisLabel"),b=t.get("splitNumber"),x=v.get("splitNumber"),w=Nr(y.get("length"),p),M=Nr(v.get("length"),p),S=o,L=(a-o)/b,k=L/x,T=y.getModel("lineStyle").getLineStyle(),D=v.getModel("lineStyle").getLineStyle(),C=y.get("distance"),I=0;I<=b;I++){if(u=Math.cos(S),c=Math.sin(S),y.get("show")){var A=C?C+l:l,Y=new dd({shape:{x1:u*(p-A)+h,y1:c*(p-A)+f,x2:u*(p-w-A)+h,y2:c*(p-w-A)+f},style:T,silent:!0});"auto"===T.stroke&&Y.setStyle({stroke:i(I/b)}),d.add(Y)}if(_.get("show")){A=_.get("distance")+C;var O=hN(Rr(I/b*(g-m)+m),_.get("formatter")),P=i(I/b);d.add(new Gl({style:Lh(_,{text:O,x:u*(p-w-A)+h,y:c*(p-w-A)+f,verticalAlign:c<-.8?"top":c>.8?"bottom":"middle",align:u<-.4?"left":u>.4?"right":"center"},{inheritColor:P}),silent:!0}))}if(v.get("show")&&I!==b){A=v.get("distance");A=A?A+l:l;for(var E=0;E<=x;E++){u=Math.cos(S),c=Math.sin(S);var N=new dd({shape:{x1:u*(p-A)+h,y1:c*(p-A)+f,x2:u*(p-M-A)+h,y2:c*(p-M-A)+f},silent:!0,style:D});"auto"===D.stroke&&N.setStyle({stroke:i((I+E/x)/b)}),d.add(N),S+=k}S-=k}else S+=L}},e.prototype._renderPointer=function(t,e,n,i,r,o,a,s,l){var u=this.group,c=this._data,d=this._progressEls,h=[],f=t.get(["pointer","show"]),p=t.getModel("progress"),m=p.get("show"),g=t.getData(),y=g.mapDimension("value"),v=+t.get("min"),_=+t.get("max"),b=[v,_],x=[o,a];function w(e,n){var i,o=g.getItemModel(e),a=o.getModel("pointer"),s=Nr(a.get("width"),r.r),l=Nr(a.get("length"),r.r),u=t.get(["pointer","icon"]),c=a.get("offsetCenter"),d=Nr(c[0],r.r),h=Nr(c[1],r.r),f=a.get("keepAspect");return i=u?j_(u,d-s/2,h-l,s,l,null,f):new cN({shape:{angle:-Math.PI/2,width:s,r:l,x:d,y:h}}),i.rotation=-(n+Math.PI/2),i.x=r.cx,i.y=r.cy,i}function M(t,e){var n=p.get("roundCap"),i=n?dT:qc,a=p.get("overlap"),u=a?p.get("width"):l/g.count(),c=a?r.r-u:r.r-(t+1)*u,d=a?r.r:r.r-t*u,h=new i({shape:{startAngle:o,endAngle:e,cx:r.cx,cy:r.cy,clockwise:s,r0:c,r:d}});return a&&(h.z2=_-g.get(y,t)%_),h}(m||f)&&(g.diff(c).add((function(e){if(f){var n=w(e,o);eh(n,{rotation:-(Er(g.get(y,e),b,x,!0)+Math.PI/2)},t),u.add(n),g.setItemGraphicEl(e,n)}if(m){var i=M(e,o),r=p.get("clip");eh(i,{shape:{endAngle:Er(g.get(y,e),b,x,r)}},t),u.add(i),h[e]=i}})).update((function(e,n){if(f){var i=c.getItemGraphicEl(n),r=i?i.rotation:o,a=w(e,r);a.rotation=r,th(a,{rotation:-(Er(g.get(y,e),b,x,!0)+Math.PI/2)},t),u.add(a),g.setItemGraphicEl(e,a)}if(m){var s=d[n],l=s?s.shape.endAngle:o,v=M(e,l),_=p.get("clip");th(v,{shape:{endAngle:Er(g.get(y,e),b,x,_)}},t),u.add(v),h[e]=v}})).execute(),g.each((function(t){var e=g.getItemModel(t),n=e.getModel("emphasis");if(f){var r=g.getItemGraphicEl(t),o=g.getItemVisual(t,"style"),a=o.fill;if(r instanceof wl){var s=r.style;r.useStyle(I({image:s.image,x:s.x,y:s.y,width:s.width,height:s.height},o))}else r.useStyle(o),"pointer"!==r.type&&r.setColor(a);r.setStyle(e.getModel(["pointer","itemStyle"]).getItemStyle()),"auto"===r.style.fill&&r.setStyle("fill",i(Er(g.get(y,t),b,[0,1],!0))),r.z2EmphasisLift=0,Zu(r,e),$u(r,n.get("focus"),n.get("blurScope"))}if(m){var l=h[t];l.useStyle(g.getItemVisual(t,"style")),l.setStyle(e.getModel(["progress","itemStyle"]).getItemStyle()),l.z2EmphasisLift=0,Zu(l,e),$u(l,n.get("focus"),n.get("blurScope"))}})),this._progressEls=h)},e.prototype._renderAnchor=function(t,e){var n=t.getModel("anchor"),i=n.get("show");if(i){var r=n.get("size"),o=n.get("icon"),a=n.get("offsetCenter"),s=n.get("keepAspect"),l=j_(o,e.cx-r/2+Nr(a[0],e.r),e.cy-r/2+Nr(a[1],e.r),r,r,null,s);l.z2=n.get("showAbove")?1:0,l.setStyle(n.getModel("itemStyle").getItemStyle()),this.group.add(l)}},e.prototype._renderTitleAndDetail=function(t,e,n,i,r){var o=this,a=t.getData(),s=a.mapDimension("value"),l=+t.get("min"),u=+t.get("max"),c=new Mr,d=[],h=[],f=t.isAnimationEnabled();a.diff(this._data).add((function(t){d[t]=new Gl({silent:!0}),h[t]=new Gl({silent:!0})})).update((function(t,e){d[t]=o._titleEls[e],h[t]=o._detailEls[e]})).execute(),a.each((function(e){var n=a.getItemModel(e),o=a.get(s,e),p=new Mr,m=i(Er(o,[l,u],[0,1],!0)),g=n.getModel("title");if(g.get("show")){var y=g.get("offsetCenter"),v=r.cx+Nr(y[0],r.r),_=r.cy+Nr(y[1],r.r),b=d[e];b.attr({style:Lh(g,{x:v,y:_,text:a.getName(e),align:"center",verticalAlign:"middle"},{inheritColor:m})}),p.add(b)}var x=n.getModel("detail");if(x.get("show")){var w=x.get("offsetCenter"),M=r.cx+Nr(w[0],r.r),S=r.cy+Nr(w[1],r.r),L=Nr(x.get("width"),r.r),k=Nr(x.get("height"),r.r),T=t.get(["progress","show"])?a.getItemVisual(e,"style").fill:m,D=(b=h[e],x.get("formatter"));b.attr({style:Lh(x,{x:M,y:S,text:hN(o,D),width:isNaN(L)?null:L,height:isNaN(k)?null:k,align:"center",verticalAlign:"middle"},{inheritColor:T})}),Eh(b,{normal:x},o,(function(t){return hN(t,D)})),f&&Nh(b,e,a,t,{getFormattedLabel:function(t,e,n,i,r,a){return hN(a?a.interpolatedValue:o,D)}}),p.add(b)}c.add(p)})),this.group.add(c),this._titleEls=d,this._detailEls=h},e.type="gauge",e}(Vy),mN=pN,gN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.visualStyleAccessPath="itemStyle",n.useColorPaletteOnData=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){return ZT(this,["value"])},e.type="series.gauge",e.defaultOption={zlevel:0,z:2,center:["50%","50%"],legendHoverLink:!0,radius:"75%",startAngle:225,endAngle:-45,clockwise:!0,min:0,max:100,splitNumber:10,axisLine:{show:!0,roundCap:!1,lineStyle:{color:[[1,"#E6EBF8"]],width:10}},progress:{show:!1,overlap:!0,width:10,roundCap:!1,clip:!0},splitLine:{show:!0,length:10,distance:10,lineStyle:{color:"#63677A",width:3,type:"solid"}},axisTick:{show:!0,splitNumber:5,length:6,distance:10,lineStyle:{color:"#63677A",width:1,type:"solid"}},axisLabel:{show:!0,distance:15,color:"#464646",fontSize:12},pointer:{icon:null,offsetCenter:[0,0],show:!0,length:"60%",width:6,keepAspect:!1},anchor:{show:!1,showAbove:!1,size:6,icon:"circle",offsetCenter:[0,0],keepAspect:!1,itemStyle:{color:"#fff",borderWidth:0,borderColor:"#5470c6"}},title:{show:!0,offsetCenter:[0,"20%"],color:"#464646",fontSize:16,valueAnimation:!1},detail:{show:!0,backgroundColor:"rgba(0,0,0,0)",borderWidth:0,borderColor:"#ccc",width:100,height:null,padding:[5,10],offsetCenter:[0,"40%"],color:"#464646",fontSize:30,fontWeight:"bold",lineHeight:30,valueAnimation:!1}},e}(Yy),yN=gN;function vN(t){t.registerChartView(mN),t.registerSeriesModel(yN)}var _N=["itemStyle","opacity"],bN=function(t){function e(e,n){var i=t.call(this)||this,r=i,o=new sd,a=new Gl;return r.setTextContent(a),i.setTextGuideLine(o),i.updateData(e,n,!0),i}return o(e,t),e.prototype.updateData=function(t,e,n){var i=this,r=t.hostModel,o=t.getItemModel(e),a=t.getItemLayout(e),s=o.getModel("emphasis"),l=o.get(_N);l=null==l?1:l,i.useStyle(t.getItemVisual(e,"style")),i.style.lineJoin="round",n?(i.setShape({points:a.points}),i.style.opacity=0,eh(i,{style:{opacity:l}},r,e)):th(i,{style:{opacity:l},shape:{points:a.points}},r,e),Zu(i,o),this._updateLabel(t,e),$u(this,s.get("focus"),s.get("blurScope"))},e.prototype._updateLabel=function(t,e){var n=this,i=this.getTextGuideLine(),r=n.getTextContent(),o=t.hostModel,a=t.getItemModel(e),s=t.getItemLayout(e),l=s.label,u=t.getItemVisual(e,"style"),c=u.fill;Mh(r,Sh(a),{labelFetcher:t.hostModel,labelDataIndex:e,defaultOpacity:u.opacity,defaultText:t.getName(e)},{normal:{align:l.textAlign,verticalAlign:l.verticalAlign}}),n.setTextConfig({local:!0,inside:!!l.inside,insideStroke:c,outsideFill:c});var d=l.linePoints;i.setShape({points:d}),n.textGuideLineConfig={anchor:d?new Wi(d[0][0],d[0][1]):null},th(r,{style:{x:l.x,y:l.y}},o,e),r.attr({rotation:l.rotation,originX:l.x,originY:l.y,z2:10}),a_(n,s_(a),{stroke:c})},e}(rd),xN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.ignoreLabelLineUpdate=!0,n}return o(e,t),e.prototype.render=function(t,e,n){var i=t.getData(),r=this._data,o=this.group;i.diff(r).add((function(t){var e=new bN(i,t);i.setItemGraphicEl(t,e),o.add(e)})).update((function(t,e){var n=r.getItemGraphicEl(e);n.updateData(i,t),o.add(n),i.setItemGraphicEl(t,n)})).remove((function(e){var n=r.getItemGraphicEl(e);rh(n,t,e)})).execute(),this._data=i},e.prototype.remove=function(){this.group.removeAll(),this._data=null},e.prototype.dispose=function(){},e.type="funnel",e}(Vy),wN=xN,MN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.useColorPaletteOnData=!0,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new KT(V(this.getData,this),V(this.getRawData,this)),this._defaultLabelLine(e)},e.prototype.getInitialData=function(t,e){return ZT(this,{coordDimensions:["value"],encodeDefaulter:G(Hp,this)})},e.prototype._defaultLabelLine=function(t){lo(t,"labelLine",["show"]);var e=t.labelLine,n=t.emphasis.labelLine;e.show=e.show&&t.label.show,n.show=n.show&&t.emphasis.label.show},e.prototype.getDataParams=function(e){var n=this.getData(),i=t.prototype.getDataParams.call(this,e),r=n.mapDimension("value"),o=n.getSum(r);return i.percent=o?+(n.get(r,e)/o*100).toFixed(2):0,i.$vars.push("percent"),i},e.type="series.funnel",e.defaultOption={zlevel:0,z:2,legendHoverLink:!0,left:80,top:60,right:80,bottom:60,minSize:"0%",maxSize:"100%",sort:"descending",orient:"vertical",gap:0,funnelAlign:"center",label:{show:!0,position:"outer"},labelLine:{show:!0,length:20,lineStyle:{width:1}},itemStyle:{borderColor:"#fff",borderWidth:1},emphasis:{label:{show:!0}},select:{itemStyle:{borderColor:"#212121"}}},e}(Yy),SN=MN;function LN(t,e){return dp(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function kN(t,e){for(var n=t.mapDimension("value"),i=t.mapArray(n,(function(t){return t})),r=[],o="ascending"===e,a=0,s=t.count();a<s;a++)r[a]=a;return"function"===typeof e?r.sort(e):"none"!==e&&r.sort((function(t,e){return o?i[t]-i[e]:i[e]-i[t]})),r}function TN(t){var e=t.hostModel,n=e.get("orient");t.each((function(e){var i,r,o,a,s=t.getItemModel(e),l=s.getModel("label"),u=l.get("position"),c=s.getModel("labelLine"),d=t.getItemLayout(e),h=d.points,f="inner"===u||"inside"===u||"center"===u||"insideLeft"===u||"insideRight"===u;if(f)"insideLeft"===u?(r=(h[0][0]+h[3][0])/2+5,o=(h[0][1]+h[3][1])/2,i="left"):"insideRight"===u?(r=(h[1][0]+h[2][0])/2-5,o=(h[1][1]+h[2][1])/2,i="right"):(r=(h[0][0]+h[1][0]+h[2][0]+h[3][0])/4,o=(h[0][1]+h[1][1]+h[2][1]+h[3][1])/4,i="center"),a=[[r,o],[r,o]];else{var p=void 0,m=void 0,g=void 0,y=void 0,v=c.get("length");0,"left"===u?(p=(h[3][0]+h[0][0])/2,m=(h[3][1]+h[0][1])/2,g=p-v,r=g-5,i="right"):"right"===u?(p=(h[1][0]+h[2][0])/2,m=(h[1][1]+h[2][1])/2,g=p+v,r=g+5,i="left"):"top"===u?(p=(h[3][0]+h[0][0])/2,m=(h[3][1]+h[0][1])/2,y=m-v,o=y-5,i="center"):"bottom"===u?(p=(h[1][0]+h[2][0])/2,m=(h[1][1]+h[2][1])/2,y=m+v,o=y+5,i="center"):"rightTop"===u?(p="horizontal"===n?h[3][0]:h[1][0],m="horizontal"===n?h[3][1]:h[1][1],"horizontal"===n?(y=m-v,o=y-5,i="center"):(g=p+v,r=g+5,i="top")):"rightBottom"===u?(p=h[2][0],m=h[2][1],"horizontal"===n?(y=m+v,o=y+5,i="center"):(g=p+v,r=g+5,i="bottom")):"leftTop"===u?(p=h[0][0],m="horizontal"===n?h[0][1]:h[1][1],"horizontal"===n?(y=m-v,o=y-5,i="center"):(g=p-v,r=g-5,i="right")):"leftBottom"===u?(p="horizontal"===n?h[1][0]:h[3][0],m="horizontal"===n?h[1][1]:h[2][1],"horizontal"===n?(y=m+v,o=y+5,i="center"):(g=p-v,r=g-5,i="right")):(p=(h[1][0]+h[2][0])/2,m=(h[1][1]+h[2][1])/2,"horizontal"===n?(y=m+v,o=y+5,i="center"):(g=p+v,r=g+5,i="left")),"horizontal"===n?(g=p,r=g):(y=m,o=y),a=[[p,m],[g,y]]}d.label={linePoints:a,x:r,y:o,verticalAlign:"middle",textAlign:i,inside:f}}))}function DN(t,e){t.eachSeriesByType("funnel",(function(t){var n=t.getData(),i=n.mapDimension("value"),r=t.get("sort"),o=LN(t,e),a=t.get("orient"),s=o.width,l=o.height,u=kN(n,r),c=o.x,d=o.y,h="horizontal"===a?[Nr(t.get("minSize"),l),Nr(t.get("maxSize"),l)]:[Nr(t.get("minSize"),s),Nr(t.get("maxSize"),s)],f=n.getDataExtent(i),p=t.get("min"),m=t.get("max");null==p&&(p=Math.min(f[0],0)),null==m&&(m=f[1]);var g=t.get("funnelAlign"),y=t.get("gap"),v="horizontal"===a?s:l,_=(v-y*(n.count()-1))/n.count(),b=function(t,e){if("horizontal"===a){var r=n.get(i,t)||0,o=Er(r,[p,m],h,!0),u=void 0;switch(g){case"top":u=d;break;case"center":u=d+(l-o)/2;break;case"bottom":u=d+(l-o);break}return[[e,u],[e,u+o]]}var f,y=n.get(i,t)||0,v=Er(y,[p,m],h,!0);switch(g){case"left":f=c;break;case"center":f=c+(s-v)/2;break;case"right":f=c+s-v;break}return[[f,e],[f+v,e]]};"ascending"===r&&(_=-_,y=-y,"horizontal"===a?c+=s:d+=l,u=u.reverse());for(var x=0;x<u.length;x++){var w=u[x],M=u[x+1],S=n.getItemModel(w);if("horizontal"===a){var L=S.get(["itemStyle","width"]);null==L?L=_:(L=Nr(L,s),"ascending"===r&&(L=-L));var k=b(w,c),T=b(M,c+L);c+=L+y,n.setItemLayout(w,{points:k.concat(T.slice().reverse())})}else{var D=S.get(["itemStyle","height"]);null==D?D=_:(D=Nr(D,l),"ascending"===r&&(D=-D));k=b(w,d),T=b(M,d+D);d+=D+y,n.setItemLayout(w,{points:k.concat(T.slice().reverse())})}}TN(n)}))}function CN(t){t.registerChartView(wN),t.registerSeriesModel(SN),t.registerLayout(DN),t.registerProcessor(zT("funnel"))}var IN=.3,AN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._dataGroup=new Mr,n._initialized=!1,n}return o(e,t),e.prototype.init=function(){this.group.add(this._dataGroup)},e.prototype.render=function(t,e,n,i){var r=this._dataGroup,o=t.getData(),a=this._data,s=t.coordinateSystem,l=s.dimensions,u=EN(t);function c(t){var e=PN(o,r,t,l,s);NN(e,o,t,u)}function d(e,n){var i=a.getItemGraphicEl(n),r=ON(o,e,l,s);o.setItemGraphicEl(e,i),th(i,{shape:{points:r}},t,e),NN(i,o,e,u)}function h(t){var e=a.getItemGraphicEl(t);r.remove(e)}if(o.diff(a).add(c).update(d).remove(h).execute(),!this._initialized){this._initialized=!0;var f=YN(s,t,(function(){setTimeout((function(){r.removeClipPath()}))}));r.setClipPath(f)}this._data=o},e.prototype.incrementalPrepareRender=function(t,e,n){this._initialized=!0,this._data=null,this._dataGroup.removeAll()},e.prototype.incrementalRender=function(t,e,n){for(var i=e.getData(),r=e.coordinateSystem,o=r.dimensions,a=EN(e),s=t.start;s<t.end;s++){var l=PN(i,this._dataGroup,s,o,r);l.incremental=!0,NN(l,i,s,a)}},e.prototype.remove=function(){this._dataGroup&&this._dataGroup.removeAll(),this._data=null},e.type="parallel",e}(Vy);function YN(t,e,n){var i=t.model,r=t.getRect(),o=new Al({shape:{x:r.x,y:r.y,width:r.width,height:r.height}}),a="horizontal"===i.get("layout")?"width":"height";return o.setShape(a,0),eh(o,{shape:{width:r.width,height:r.height}},e,n),o}function ON(t,e,n,i){for(var r=[],o=0;o<n.length;o++){var a=n[o],s=t.get(t.mapDimension(a),e);RN(s,i.getAxis(a).type)||r.push(i.dataToPoint(s,a))}return r}function PN(t,e,n,i,r){var o=ON(t,n,i,r),a=new sd({shape:{points:o},z2:10});return e.add(a),t.setItemGraphicEl(n,a),a}function EN(t){var e=t.get("smooth",!0);return!0===e&&(e=IN),e=to(e),rt(e)&&(e=0),{smooth:e}}function NN(t,e,n,i){t.useStyle(e.getItemVisual(n,"style")),t.style.fill=null,t.setShape("smooth",i.smooth);var r=e.getItemModel(n),o=r.getModel("emphasis");Zu(t,r,"lineStyle"),$u(t,o.get("focus"),o.get("blurScope"))}function RN(t,e){return"category"===e?null==t:null==t||isNaN(t)}var HN=AN,jN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.visualStyleAccessPath="lineStyle",n.visualDrawType="stroke",n}return o(e,t),e.prototype.getInitialData=function(t,e){return BL(this.getSource(),this,{useEncodeDefaulter:V(zN,null,this)})},e.prototype.getRawIndicesByActiveState=function(t){var e=this.coordinateSystem,n=this.getData(),i=[];return e.eachActiveState(n,(function(e,r){t===e&&i.push(n.getRawIndex(r))})),i},e.type="series.parallel",e.dependencies=["parallel"],e.defaultOption={zlevel:0,z:2,coordinateSystem:"parallel",parallelIndex:0,label:{show:!1},inactiveOpacity:.05,activeOpacity:1,lineStyle:{width:1,opacity:.45,type:"solid"},emphasis:{label:{show:!1}},progressive:500,smooth:!1,animationEasing:"linear"},e}(Yy);function zN(t){var e=t.ecModel.getComponent("parallel",t.get("parallelIndex"));if(e){var n={};return R(e.dimensions,(function(t){var e=WN(t);n[t]=e})),n}}function WN(t){return+t.replace("dim","")}var BN=jN,FN=["lineStyle","opacity"],VN={seriesType:"parallel",reset:function(t,e){var n=t.coordinateSystem,i={normal:t.get(["lineStyle","opacity"]),active:t.get("activeOpacity"),inactive:t.get("inactiveOpacity")};return{progress:function(t,e){n.eachActiveState(e,(function(t,n){var r=i[t];if("normal"===t&&e.hasItemOption){var o=e.getItemModel(n).get(FN,!0);null!=o&&(r=o)}var a=e.ensureUniqueItemVisual(n,"style");a.opacity=r}),t.start,t.end)}}}},GN=VN;function $N(t){UN(t),qN(t)}function UN(t){if(!t.parallel){var e=!1;R(t.series,(function(t){t&&"parallel"===t.type&&(e=!0)})),e&&(t.parallel=[{}])}}function qN(t){var e=so(t.parallelAxis);R(e,(function(e){if(J(e)){var n=e.parallelIndex||0,i=so(t.parallel)[n];i&&i.parallelAxisDefault&&D(e,i.parallelAxisDefault,!1)}}))}var XN=5,ZN=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){this._model=t,this._api=n,this._handlers||(this._handlers={},R(JN,(function(t,e){n.getZr().on(e,this._handlers[e]=V(t,this))}),this)),Xy(this,"_throttledDispatchExpand",t.get("axisExpandRate"),"fixRate")},e.prototype.dispose=function(t,e){R(this._handlers,(function(t,n){e.getZr().off(n,t)})),this._handlers=null},e.prototype._throttledDispatchExpand=function(t){this._dispatchExpand(t)},e.prototype._dispatchExpand=function(t){t&&this._api.dispatchAction(I({type:"parallelAxisExpand"},t))},e.type="parallel",e}(Py),JN={mousedown:function(t){KN(this,"click")&&(this._mouseDownPoint=[t.offsetX,t.offsetY])},mouseup:function(t){var e=this._mouseDownPoint;if(KN(this,"click")&&e){var n=[t.offsetX,t.offsetY],i=Math.pow(e[0]-n[0],2)+Math.pow(e[1]-n[1],2);if(i>XN)return;var r=this._model.coordinateSystem.getSlidedAxisExpandWindow([t.offsetX,t.offsetY]);"none"!==r.behavior&&this._dispatchExpand({axisExpandWindow:r.axisExpandWindow})}this._mouseDownPoint=null},mousemove:function(t){if(!this._mouseDownPoint&&KN(this,"mousemove")){var e=this._model,n=e.coordinateSystem.getSlidedAxisExpandWindow([t.offsetX,t.offsetY]),i=n.behavior;"jump"===i&&this._throttledDispatchExpand.debounceNextCall(e.get("axisExpandDebounce")),this._throttledDispatchExpand("none"===i?null:{axisExpandWindow:n.axisExpandWindow,animation:"jump"===i?null:{duration:0}})}}};function KN(t,e){var n=t._model;return n.get("axisExpandable")&&n.get("axisExpandTriggerOn")===e}var QN=ZN,tR=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){t.prototype.init.apply(this,arguments),this.mergeOption({})},e.prototype.mergeOption=function(t){var e=this.option;t&&D(e,t,!0),this._initDimensions()},e.prototype.contains=function(t,e){var n=t.get("parallelIndex");return null!=n&&e.getComponent("parallel",n)===this},e.prototype.setAxisExpand=function(t){R(["axisExpandable","axisExpandCenter","axisExpandCount","axisExpandWidth","axisExpandWindow"],(function(e){t.hasOwnProperty(e)&&(this.option[e]=t[e])}),this)},e.prototype._initDimensions=function(){var t=this.dimensions=[],e=this.parallelAxisIndex=[],n=z(this.ecModel.queryComponents({mainType:"parallelAxis"}),(function(t){return(t.get("parallelIndex")||0)===this.componentIndex}),this);R(n,(function(n){t.push("dim"+n.get("dim")),e.push(n.componentIndex)}))},e.type="parallel",e.dependencies=["parallelAxis"],e.layoutMode="box",e.defaultOption={zlevel:0,z:0,left:80,top:60,right:80,bottom:60,layout:"horizontal",axisExpandable:!1,axisExpandCenter:null,axisExpandCount:0,axisExpandWidth:50,axisExpandRate:17,axisExpandDebounce:50,axisExpandSlideTriggerArea:[-.15,.05,.4],axisExpandTriggerOn:"click",parallelAxisDefault:null},e}(xp),eR=tR,nR=function(t){function e(e,n,i,r,o){var a=t.call(this,e,n,i)||this;return a.type=r||"value",a.axisIndex=o,a}return o(e,t),e.prototype.isHorizontal=function(){return"horizontal"!==this.coordinateSystem.getModel().get("layout")},e}(tI),iR=nR;function rR(t,e,n,i,r,o){t=t||0;var a=n[1]-n[0];if(null!=r&&(r=aR(r,[0,a])),null!=o&&(o=Math.max(o,null!=r?r:0)),"all"===i){var s=Math.abs(e[1]-e[0]);s=aR(s,[0,a]),r=o=aR(s,[r,o]),i=0}e[0]=aR(e[0],n),e[1]=aR(e[1],n);var l=oR(e,i);e[i]+=t;var u,c=r||0,d=n.slice();return l.sign<0?d[0]+=c:d[1]-=c,e[i]=aR(e[i],d),u=oR(e,i),null!=r&&(u.sign!==l.sign||u.span<r)&&(e[1-i]=e[i]+l.sign*r),u=oR(e,i),null!=o&&u.span>o&&(e[1-i]=e[i]+u.sign*o),e}function oR(t,e){var n=t[e]-t[1-e];return{span:Math.abs(n),sign:n>0?-1:n<0?1:e?-1:1}}function aR(t,e){return Math.min(null!=e[1]?e[1]:1/0,Math.max(null!=e[0]?e[0]:-1/0,t))}var sR=R,lR=Math.min,uR=Math.max,cR=Math.floor,dR=Math.ceil,hR=Rr,fR=Math.PI,pR=function(){function t(t,e,n){this.type="parallel",this._axesMap=gt(),this._axesLayout={},this.dimensions=t.dimensions,this._model=t,this._init(t,e,n)}return t.prototype._init=function(t,e,n){var i=t.dimensions,r=t.parallelAxisIndex;sR(i,(function(t,n){var i=r[n],o=e.getComponent("parallelAxis",i),a=this._axesMap.set(t,new iR(t,_C(o),[0,0],o.get("type"),i)),s="category"===a.type;a.onBand=s&&o.get("boundaryGap"),a.inverse=o.get("inverse"),o.axis=a,a.model=o,a.coordinateSystem=o.coordinateSystem=this}),this)},t.prototype.update=function(t,e){this._updateAxesFromSeries(this._model,t)},t.prototype.containPoint=function(t){var e=this._makeLayoutInfo(),n=e.axisBase,i=e.layoutBase,r=e.pixelDimIndex,o=t[1-r],a=t[r];return o>=n&&o<=n+e.axisLength&&a>=i&&a<=i+e.layoutLength},t.prototype.getModel=function(){return this._model},t.prototype._updateAxesFromSeries=function(t,e){e.eachSeries((function(n){if(t.contains(n,e)){var i=n.getData();sR(this.dimensions,(function(t){var e=this._axesMap.get(t);e.scale.unionExtentFromData(i,i.mapDimension(t)),vC(e.scale,e.model)}),this)}}),this)},t.prototype.resize=function(t,e){this._rect=dp(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()}),this._layoutAxes()},t.prototype.getRect=function(){return this._rect},t.prototype._makeLayoutInfo=function(){var t,e=this._model,n=this._rect,i=["x","y"],r=["width","height"],o=e.get("layout"),a="horizontal"===o?0:1,s=n[r[a]],l=[0,s],u=this.dimensions.length,c=mR(e.get("axisExpandWidth"),l),d=mR(e.get("axisExpandCount")||0,[0,u]),h=e.get("axisExpandable")&&u>3&&u>d&&d>1&&c>0&&s>0,f=e.get("axisExpandWindow");if(f)t=mR(f[1]-f[0],l),f[1]=f[0]+t;else{t=mR(c*(d-1),l);var p=e.get("axisExpandCenter")||cR(u/2);f=[c*p-t/2],f[1]=f[0]+t}var m=(s-t)/(u-d);m<3&&(m=0);var g=[cR(hR(f[0]/c,1))+1,dR(hR(f[1]/c,1))-1],y=m/c*f[0];return{layout:o,pixelDimIndex:a,layoutBase:n[i[a]],layoutLength:s,axisBase:n[i[1-a]],axisLength:n[r[1-a]],axisExpandable:h,axisExpandWidth:c,axisCollapseWidth:m,axisExpandWindow:f,axisCount:u,winInnerIndices:g,axisExpandWindow0Pos:y}},t.prototype._layoutAxes=function(){var t=this._rect,e=this._axesMap,n=this.dimensions,i=this._makeLayoutInfo(),r=i.layout;e.each((function(t){var e=[0,i.axisLength],n=t.inverse?1:0;t.setExtent(e[n],e[1-n])})),sR(n,(function(e,n){var o=(i.axisExpandable?yR:gR)(n,i),a={horizontal:{x:o.position,y:i.axisLength},vertical:{x:0,y:o.position}},s={horizontal:fR/2,vertical:0},l=[a[r].x+t.x,a[r].y+t.y],u=s[r],c=bi();Li(c,c,u),Si(c,c,l),this._axesLayout[e]={position:l,rotation:u,transform:c,axisNameAvailableWidth:o.axisNameAvailableWidth,axisLabelShow:o.axisLabelShow,nameTruncateMaxWidth:o.nameTruncateMaxWidth,tickDirection:1,labelDirection:1}}),this)},t.prototype.getAxis=function(t){return this._axesMap.get(t)},t.prototype.dataToPoint=function(t,e){return this.axisCoordToPoint(this._axesMap.get(e).dataToCoord(t),e)},t.prototype.eachActiveState=function(t,e,n,i){null==n&&(n=0),null==i&&(i=t.count());var r=this._axesMap,o=this.dimensions,a=[],s=[];R(o,(function(e){a.push(t.mapDimension(e)),s.push(r.get(e).model)}));for(var l=this.hasAxisBrushed(),u=n;u<i;u++){var c=void 0;if(l){c="active";for(var d=t.getValues(a,u),h=0,f=o.length;h<f;h++){var p=s[h].getActiveState(d[h]);if("inactive"===p){c="inactive";break}}}else c="normal";e(c,u)}},t.prototype.hasAxisBrushed=function(){for(var t=this.dimensions,e=this._axesMap,n=!1,i=0,r=t.length;i<r;i++)"normal"!==e.get(t[i]).model.getActiveState()&&(n=!0);return n},t.prototype.axisCoordToPoint=function(t,e){var n=this._axesLayout[e];return sh([t,0],n.transform)},t.prototype.getAxisLayout=function(t){return T(this._axesLayout[t])},t.prototype.getSlidedAxisExpandWindow=function(t){var e=this._makeLayoutInfo(),n=e.pixelDimIndex,i=e.axisExpandWindow.slice(),r=i[1]-i[0],o=[0,e.axisExpandWidth*(e.axisCount-1)];if(!this.containPoint(t))return{behavior:"none",axisExpandWindow:i};var a,s=t[n]-e.layoutBase-e.axisExpandWindow0Pos,l="slide",u=e.axisCollapseWidth,c=this._model.get("axisExpandSlideTriggerArea"),d=null!=c[0];if(u)d&&u&&s<r*c[0]?(l="jump",a=s-r*c[2]):d&&u&&s>r*(1-c[0])?(l="jump",a=s-r*(1-c[2])):(a=s-r*c[1])>=0&&(a=s-r*(1-c[1]))<=0&&(a=0),a*=e.axisExpandWidth/u,a?rR(a,i,o,"all"):l="none";else{var h=i[1]-i[0],f=o[1]*s/h;i=[uR(0,f-h/2)],i[1]=lR(o[1],i[0]+h),i[0]=i[1]-h}return{axisExpandWindow:i,behavior:l}},t}();function mR(t,e){return lR(uR(t,e[0]),e[1])}function gR(t,e){var n=e.layoutLength/(e.axisCount-1);return{position:n*t,axisNameAvailableWidth:n,axisLabelShow:!0}}function yR(t,e){var n,i,r=e.layoutLength,o=e.axisExpandWidth,a=e.axisCount,s=e.axisCollapseWidth,l=e.winInnerIndices,u=s,c=!1;return t<l[0]?(n=t*s,i=s):t<=l[1]?(n=e.axisExpandWindow0Pos+t*o-e.axisExpandWindow[0],u=o,c=!0):(n=r-(a-1-t)*s,i=s),{position:n,axisNameAvailableWidth:u,axisLabelShow:c,nameTruncateMaxWidth:i}}var vR=pR;function _R(t,e){var n=[];return t.eachComponent("parallel",(function(i,r){var o=new vR(i,t,e);o.name="parallel_"+r,o.resize(i,e),i.coordinateSystem=o,o.model=i,n.push(o)})),t.eachSeries((function(t){if("parallel"===t.get("coordinateSystem")){var e=t.getReferringComponents("parallel",Po).models[0];t.coordinateSystem=e.coordinateSystem}})),n}var bR={create:_R},xR=bR,wR=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.activeIntervals=[],n}return o(e,t),e.prototype.getAreaSelectStyle=function(){return na([["fill","color"],["lineWidth","borderWidth"],["stroke","borderColor"],["width","width"],["opacity","opacity"]])(this.getModel("areaSelectStyle"))},e.prototype.setActiveIntervals=function(t){var e=this.activeIntervals=T(t);if(e)for(var n=e.length-1;n>=0;n--)Hr(e[n])},e.prototype.getActiveState=function(t){var e=this.activeIntervals;if(!e.length)return"normal";if(null==t||isNaN(+t))return"inactive";if(1===e.length){var n=e[0];if(n[0]<=t&&t<=n[1])return"active"}else for(var i=0,r=e.length;i<r;i++)if(e[i][0]<=t&&t<=e[i][1])return"active";return"inactive"},e}(xp);E(wR,fD);var MR=wR,SR=!0,LR=Math.min,kR=Math.max,TR=Math.pow,DR=1e4,CR=6,IR=6,AR="globalPan",YR={w:[0,0],e:[0,1],n:[1,0],s:[1,1]},OR={w:"ew",e:"ew",n:"ns",s:"ns",ne:"nesw",sw:"nesw",nw:"nwse",se:"nwse"},PR={brushStyle:{lineWidth:2,stroke:"rgba(210,219,238,0.3)",fill:"#D2DBEE"},transformable:!0,brushMode:"single",removeOnClick:!1},ER=0,NR=function(t){function e(e){var n=t.call(this)||this;return n._track=[],n._covers=[],n._handlers={},n._zr=e,n.group=new Mr,n._uid="brushController_"+ER++,R(pH,(function(t,e){this._handlers[e]=V(t,this)}),n),n}return o(e,t),e.prototype.enableBrush=function(t){return this._brushType&&this._doDisableBrush(),t.brushType&&this._doEnableBrush(t),this},e.prototype._doEnableBrush=function(t){var e=this._zr;this._enableGlobalPan||xA(e,AR,this._uid),R(this._handlers,(function(t,n){e.on(n,t)})),this._brushType=t.brushType,this._brushOption=D(T(PR),t,!0)},e.prototype._doDisableBrush=function(){var t=this._zr;wA(t,AR,this._uid),R(this._handlers,(function(e,n){t.off(n,e)})),this._brushType=this._brushOption=null},e.prototype.setPanels=function(t){if(t&&t.length){var e=this._panels={};R(t,(function(t){e[t.panelId]=T(t)}))}else this._panels=null;return this},e.prototype.mount=function(t){t=t||{},this._enableGlobalPan=t.enableGlobalPan;var e=this.group;return this._zr.add(e),e.attr({x:t.x||0,y:t.y||0,rotation:t.rotation||0,scaleX:t.scaleX||1,scaleY:t.scaleY||1}),this._transform=e.getLocalTransform(),this},e.prototype.updateCovers=function(t){t=H(t,(function(t){return D(T(PR),t,!0)}));var e="\0-brush-index-",n=this._covers,i=this._covers=[],r=this,o=this._creatingCover;return new $S(n,t,s,a).add(l).update(l).remove(u).execute(),this;function a(t,n){return(null!=t.id?t.id:e+n)+"-"+t.brushType}function s(t,e){return a(t.__brushOption,e)}function l(e,a){var s=t[e];if(null!=a&&n[a]===o)i[e]=n[a];else{var l=i[e]=null!=a?(n[a].__brushOption=s,n[a]):HR(r,RR(r,s));WR(r,l)}}function u(t){n[t]!==o&&r.group.remove(n[t])}},e.prototype.unmount=function(){return this.enableBrush(!1),GR(this),this._zr.remove(this.group),this},e.prototype.dispose=function(){this.unmount(),this.off()},e}($t);function RR(t,e){var n=yH[e.brushType].createCover(t,e);return n.__brushOption=e,zR(n,e),t.group.add(n),n}function HR(t,e){var n=BR(e);return n.endCreating&&(n.endCreating(t,e),zR(e,e.__brushOption)),e}function jR(t,e){var n=e.__brushOption;BR(e).updateCoverShape(t,e,n.range,n)}function zR(t,e){var n=e.z;null==n&&(n=DR),t.traverse((function(t){t.z=n,t.z2=n}))}function WR(t,e){BR(e).updateCommon(t,e),jR(t,e)}function BR(t){return yH[t.__brushOption.brushType]}function FR(t,e,n){var i,r=t._panels;if(!r)return SR;var o=t._transform;return R(r,(function(t){t.isTargetByCursor(e,n,o)&&(i=t)})),i}function VR(t,e){var n=t._panels;if(!n)return SR;var i=e.__brushOption.panelId;return null!=i?n[i]:SR}function GR(t){var e=t._covers,n=e.length;return R(e,(function(e){t.group.remove(e)}),t),e.length=0,!!n}function $R(t,e){var n=H(t._covers,(function(t){var e=t.__brushOption,n=T(e.range);return{brushType:e.brushType,panelId:e.panelId,range:n}}));t.trigger("brush",{areas:n,isEnd:!!e.isEnd,removeOnClick:!!e.removeOnClick})}function UR(t){var e=t._track;if(!e.length)return!1;var n=e[e.length-1],i=e[0],r=n[0]-i[0],o=n[1]-i[1],a=TR(r*r+o*o,.5);return a>CR}function qR(t){var e=t.length-1;return e<0&&(e=0),[t[0],t[e]]}function XR(t,e,n,i){var r=new Mr;return r.add(new Al({name:"main",style:QR(n),silent:!0,draggable:!0,cursor:"move",drift:G(rH,t,e,r,["n","s","w","e"]),ondragend:G($R,e,{isEnd:!0})})),R(i,(function(n){r.add(new Al({name:n.join(""),style:{opacity:0},draggable:!0,silent:!0,invisible:!0,drift:G(rH,t,e,r,n),ondragend:G($R,e,{isEnd:!0})}))})),r}function ZR(t,e,n,i){var r=i.brushStyle.lineWidth||0,o=kR(r,IR),a=n[0][0],s=n[1][0],l=a-r/2,u=s-r/2,c=n[0][1],d=n[1][1],h=c-o+r/2,f=d-o+r/2,p=c-a,m=d-s,g=p+r,y=m+r;KR(t,e,"main",a,s,p,m),i.transformable&&(KR(t,e,"w",l,u,o,y),KR(t,e,"e",h,u,o,y),KR(t,e,"n",l,u,g,o),KR(t,e,"s",l,f,g,o),KR(t,e,"nw",l,u,o,o),KR(t,e,"ne",h,u,o,o),KR(t,e,"sw",l,f,o,o),KR(t,e,"se",h,f,o,o))}function JR(t,e){var n=e.__brushOption,i=n.transformable,r=e.childAt(0);r.useStyle(QR(n)),r.attr({silent:!i,cursor:i?"move":"default"}),R([["w"],["e"],["n"],["s"],["s","e"],["s","w"],["n","e"],["n","w"]],(function(n){var r=e.childOfName(n.join("")),o=1===n.length?nH(t,n[0]):iH(t,n);r&&r.attr({silent:!i,invisible:!i,cursor:i?OR[o]+"-resize":null})}))}function KR(t,e,n,i,r,o,a){var s=e.childOfName(n);s&&s.setShape(lH(sH(t,e,[[i,r],[i+o,r+a]])))}function QR(t){return A({strokeNoScale:!0},t.brushStyle)}function tH(t,e,n,i){var r=[LR(t,n),LR(e,i)],o=[kR(t,n),kR(e,i)];return[[r[0],o[0]],[r[1],o[1]]]}function eH(t){return ah(t.group)}function nH(t,e){var n={w:"left",e:"right",n:"top",s:"bottom"},i={left:"w",right:"e",top:"n",bottom:"s"},r=lh(n[e],eH(t));return i[r]}function iH(t,e){var n=[nH(t,e[0]),nH(t,e[1])];return("e"===n[0]||"w"===n[0])&&n.reverse(),n.join("")}function rH(t,e,n,i,r,o){var a=n.__brushOption,s=t.toRectRange(a.range),l=aH(e,r,o);R(i,(function(t){var e=YR[t];s[e[0]][e[1]]+=l[e[0]]})),a.range=t.fromRectRange(tH(s[0][0],s[1][0],s[0][1],s[1][1])),WR(e,n),$R(e,{isEnd:!1})}function oH(t,e,n,i){var r=e.__brushOption.range,o=aH(t,n,i);R(r,(function(t){t[0]+=o[0],t[1]+=o[1]})),WR(t,e),$R(t,{isEnd:!1})}function aH(t,e,n){var i=t.group,r=i.transformCoordToLocal(e,n),o=i.transformCoordToLocal(0,0);return[r[0]-o[0],r[1]-o[1]]}function sH(t,e,n){var i=VR(t,e);return i&&i!==SR?i.clipPath(n,t._transform):T(n)}function lH(t){var e=LR(t[0][0],t[1][0]),n=LR(t[0][1],t[1][1]),i=kR(t[0][0],t[1][0]),r=kR(t[0][1],t[1][1]);return{x:e,y:n,width:i-e,height:r-n}}function uH(t,e,n){if(t._brushType&&!gH(t,e.offsetX,e.offsetY)){var i=t._zr,r=t._covers,o=FR(t,e,n);if(!t._dragging)for(var a=0;a<r.length;a++){var s=r[a].__brushOption;if(o&&(o===SR||s.panelId===o.panelId)&&yH[s.brushType].contain(r[a],n[0],n[1]))return}o&&i.setCursorStyle("crosshair")}}function cH(t){var e=t.event;e.preventDefault&&e.preventDefault()}function dH(t,e,n){return t.childOfName("main").contain(e,n)}function hH(t,e,n,i){var r,o=t._creatingCover,a=t._creatingPanel,s=t._brushOption;if(t._track.push(n.slice()),UR(t)||o){if(a&&!o){"single"===s.brushMode&&GR(t);var l=T(s);l.brushType=fH(l.brushType,a),l.panelId=a===SR?null:a.panelId,o=t._creatingCover=RR(t,l),t._covers.push(o)}if(o){var u=yH[fH(t._brushType,a)],c=o.__brushOption;c.range=u.getCreatingRange(sH(t,o,t._track)),i&&(HR(t,o),u.updateCommon(t,o)),jR(t,o),r={isEnd:i}}}else i&&"single"===s.brushMode&&s.removeOnClick&&FR(t,e,n)&&GR(t)&&(r={isEnd:i,removeOnClick:!0});return r}function fH(t,e){return"auto"===t?e.defaultBrushType:t}var pH={mousedown:function(t){if(this._dragging)mH(this,t);else if(!t.target||!t.target.draggable){cH(t);var e=this.group.transformCoordToLocal(t.offsetX,t.offsetY);this._creatingCover=null;var n=this._creatingPanel=FR(this,t,e);n&&(this._dragging=!0,this._track=[e.slice()])}},mousemove:function(t){var e=t.offsetX,n=t.offsetY,i=this.group.transformCoordToLocal(e,n);if(uH(this,t,i),this._dragging){cH(t);var r=hH(this,t,i,!1);r&&$R(this,r)}},mouseup:function(t){mH(this,t)}};function mH(t,e){if(t._dragging){cH(e);var n=e.offsetX,i=e.offsetY,r=t.group.transformCoordToLocal(n,i),o=hH(t,e,r,!0);t._dragging=!1,t._track=[],t._creatingCover=null,o&&$R(t,o)}}function gH(t,e,n){var i=t._zr;return e<0||e>i.getWidth()||n<0||n>i.getHeight()}var yH={lineX:vH(0),lineY:vH(1),rect:{createCover:function(t,e){function n(t){return t}return XR({toRectRange:n,fromRectRange:n},t,e,[["w"],["e"],["n"],["s"],["s","e"],["s","w"],["n","e"],["n","w"]])},getCreatingRange:function(t){var e=qR(t);return tH(e[1][0],e[1][1],e[0][0],e[0][1])},updateCoverShape:function(t,e,n,i){ZR(t,e,n,i)},updateCommon:JR,contain:dH},polygon:{createCover:function(t,e){var n=new Mr;return n.add(new sd({name:"main",style:QR(e),silent:!0})),n},getCreatingRange:function(t){return t},endCreating:function(t,e){e.remove(e.childAt(0)),e.add(new rd({name:"main",draggable:!0,drift:G(oH,t,e),ondragend:G($R,t,{isEnd:!0})}))},updateCoverShape:function(t,e,n,i){e.childAt(0).setShape({points:sH(t,e,n)})},updateCommon:JR,contain:dH}};function vH(t){return{createCover:function(e,n){return XR({toRectRange:function(e){var n=[e,[0,100]];return t&&n.reverse(),n},fromRectRange:function(e){return e[t]}},e,n,[[["w"],["e"]],[["n"],["s"]]][t])},getCreatingRange:function(e){var n=qR(e),i=LR(n[0][t],n[1][t]),r=kR(n[0][t],n[1][t]);return[i,r]},updateCoverShape:function(e,n,i,r){var o,a=VR(e,n);if(a!==SR&&a.getLinearBrushOtherExtent)o=a.getLinearBrushOtherExtent(t);else{var s=e._zr;o=[0,[s.getWidth(),s.getHeight()][1-t]]}var l=[i,o];t&&l.reverse(),ZR(e,n,l,r)},updateCommon:JR,contain:dH}}var _H=NR;function bH(t){return t=MH(t),function(e){return hh(e,t)}}function xH(t,e){return t=MH(t),function(n){var i=null!=e?e:n,r=i?t.width:t.height,o=i?t.x:t.y;return[o,o+(r||0)]}}function wH(t,e,n){var i=MH(t);return function(t,r){return i.contain(r[0],r[1])&&!OA(t,e,n)}}function MH(t){return Ji.create(t)}var SH=["axisLine","axisTickLabel","axisName"],LH=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(e,n){t.prototype.init.apply(this,arguments),(this._brushController=new _H(n.getZr())).on("brush",V(this._onBrush,this))},e.prototype.render=function(t,e,n,i){if(!kH(t,e,i)){this.axisModel=t,this.api=n,this.group.removeAll();var r=this._axisGroup;if(this._axisGroup=new Mr,this.group.add(this._axisGroup),t.get("show")){var o=DH(t,e),a=o.coordinateSystem,s=t.getAreaSelectStyle(),l=s.width,u=t.axis.dim,c=a.getAxisLayout(u),d=I({strokeContainThreshold:l},c),h=new SI(t,d);R(SH,h.add,h),this._axisGroup.add(h.getGroup()),this._refreshBrushController(d,s,t,o,l,n),dh(r,this._axisGroup,t)}}},e.prototype._refreshBrushController=function(t,e,n,i,r,o){var a=n.axis.getExtent(),s=a[1]-a[0],l=Math.min(30,.1*Math.abs(s)),u=Ji.create({x:a[0],y:-r/2,width:s,height:r});u.x-=l,u.width+=2*l,this._brushController.mount({enableGlobalPan:!0,rotation:t.rotation,x:t.position[0],y:t.position[1]}).setPanels([{panelId:"pl",clipPath:bH(u),isTargetByCursor:wH(u,o,i),getLinearBrushOtherExtent:xH(u,0)}]).enableBrush({brushType:"lineX",brushStyle:e,removeOnClick:!0}).updateCovers(TH(n))},e.prototype._onBrush=function(t){var e=t.areas,n=this.axisModel,i=n.axis,r=H(e,(function(t){return[i.coordToData(t.range[0],!0),i.coordToData(t.range[1],!0)]}));(!n.option.realtime===t.isEnd||t.removeOnClick)&&this.api.dispatchAction({type:"axisAreaSelect",parallelAxisId:n.id,intervals:r})},e.prototype.dispose=function(){this._brushController.dispose()},e.type="parallelAxis",e}(Py);function kH(t,e,n){return n&&"axisAreaSelect"===n.type&&e.findComponents({mainType:"parallelAxis",query:n})[0]===t}function TH(t){var e=t.axis;return H(t.activeIntervals,(function(t){return{brushType:"lineX",panelId:"pl",range:[e.dataToCoord(t[0],!0),e.dataToCoord(t[1],!0)]}}))}function DH(t,e){return e.getComponent("parallel",t.get("parallelIndex"))}var CH=LH,IH={type:"axisAreaSelect",event:"axisAreaSelected"};function AH(t){t.registerAction(IH,(function(t,e){e.eachComponent({mainType:"parallelAxis",query:t},(function(e){e.axis.model.setActiveIntervals(t.intervals)}))})),t.registerAction("parallelAxisExpand",(function(t,e){e.eachComponent({mainType:"parallel",query:t},(function(e){e.setAxisExpand(t)}))}))}var YH={type:"value",areaSelectStyle:{width:20,borderWidth:1,borderColor:"rgba(160,197,232)",color:"rgba(160,197,232)",opacity:.3},realtime:!0,z:10};function OH(t){t.registerComponentView(QN),t.registerComponentModel(eR),t.registerCoordinateSystem("parallel",xR),t.registerPreprocessor($N),t.registerComponentModel(MR),t.registerComponentView(CH),LD(t,"parallel",MR,YH),AH(t)}function PH(t){vM(OH),t.registerChartView(HN),t.registerSeriesModel(BN),t.registerVisual(t.PRIORITY.VISUAL.BRUSH,GN)}var EH=function(){function t(){this.x1=0,this.y1=0,this.x2=0,this.y2=0,this.cpx1=0,this.cpy1=0,this.cpx2=0,this.cpy2=0,this.extent=0}return t}(),NH=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultShape=function(){return new EH},e.prototype.buildPath=function(t,e){var n=e.extent;t.moveTo(e.x1,e.y1),t.bezierCurveTo(e.cpx1,e.cpy1,e.cpx2,e.cpy2,e.x2,e.y2),"vertical"===e.orient?(t.lineTo(e.x2+n,e.y2),t.bezierCurveTo(e.cpx2+n,e.cpy2,e.cpx1+n,e.cpy1,e.x1+n,e.y1)):(t.lineTo(e.x2,e.y2+n),t.bezierCurveTo(e.cpx2,e.cpy2+n,e.cpx1,e.cpy1+n,e.x1,e.y1+n)),t.closePath()},e.prototype.highlight=function(){Cu(this)},e.prototype.downplay=function(){Iu(this)},e}(pl),RH=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._focusAdjacencyDisabled=!1,n}return o(e,t),e.prototype.render=function(t,e,n){var i=this,r=t.getGraph(),o=this.group,a=t.layoutInfo,s=a.width,l=a.height,u=t.getData(),c=t.getData("edge"),d=t.get("orient");this._model=t,o.removeAll(),o.x=a.x,o.y=a.y,r.eachEdge((function(e){var n=new NH,i=$l(n);i.dataIndex=e.dataIndex,i.seriesIndex=t.seriesIndex,i.dataType="edge";var r,a,u,h,f,p,m,g,y=e.getModel(),v=y.getModel("lineStyle"),_=v.get("curveness"),b=e.node1.getLayout(),x=e.node1.getModel(),w=x.get("localX"),M=x.get("localY"),S=e.node2.getLayout(),L=e.node2.getModel(),k=L.get("localX"),T=L.get("localY"),D=e.getLayout();switch(n.shape.extent=Math.max(1,D.dy),n.shape.orient=d,"vertical"===d?(r=(null!=w?w*s:b.x)+D.sy,a=(null!=M?M*l:b.y)+b.dy,u=(null!=k?k*s:S.x)+D.ty,h=null!=T?T*l:S.y,f=r,p=a*(1-_)+h*_,m=u,g=a*_+h*(1-_)):(r=(null!=w?w*s:b.x)+b.dx,a=(null!=M?M*l:b.y)+D.sy,u=null!=k?k*s:S.x,h=(null!=T?T*l:S.y)+D.ty,f=r*(1-_)+u*_,p=a,m=r*_+u*(1-_),g=h),n.setShape({x1:r,y1:a,x2:u,y2:h,cpx1:f,cpy1:p,cpx2:m,cpy2:g}),n.useStyle(v.getItemStyle()),n.style.fill){case"source":n.style.fill=e.node1.getVisual("color"),n.style.decal=e.node1.getVisual("style").decal;break;case"target":n.style.fill=e.node2.getVisual("color"),n.style.decal=e.node2.getVisual("style").decal;break;case"gradient":var C=e.node1.getVisual("color"),I=e.node2.getVisual("color");"string"===typeof C&&"string"===typeof I&&(n.style.fill=new Ld(0,0,1,0,[{color:C,offset:0},{color:I,offset:1}]))}var A=y.getModel("emphasis");Zu(n,y,"lineStyle",(function(t){return t.getItemStyle()})),o.add(n),c.setItemGraphicEl(e.dataIndex,n);var Y=A.get("focus");$u(n,"adjacency"===Y?e.getAdjacentDataIndices():Y,A.get("blurScope")),$l(n).dataType="edge"})),r.eachNode((function(e){var n=e.getLayout(),i=e.getModel(),r=i.get("localX"),a=i.get("localY"),c=i.getModel("emphasis"),d=new Al({shape:{x:null!=r?r*s:n.x,y:null!=a?a*l:n.y,width:n.dx,height:n.dy},style:i.getModel("itemStyle").getItemStyle(),z2:10});Mh(d,Sh(i),{labelFetcher:t,labelDataIndex:e.dataIndex,defaultText:e.id}),d.disableLabelAnimation=!0,d.setStyle("fill",e.getVisual("color")),d.setStyle("decal",e.getVisual("style").decal),Zu(d,i),o.add(d),u.setItemGraphicEl(e.dataIndex,d),$l(d).dataType="node";var h=c.get("focus");$u(d,"adjacency"===h?e.getAdjacentDataIndices():h,c.get("blurScope"))})),u.eachItemGraphicEl((function(e,r){var o=u.getItemModel(r);o.get("draggable")&&(e.drift=function(e,o){i._focusAdjacencyDisabled=!0,this.shape.x+=e,this.shape.y+=o,this.dirty(),n.dispatchAction({type:"dragNode",seriesId:t.id,dataIndex:u.getRawIndex(r),localX:this.shape.x/s,localY:this.shape.y/l})},e.ondragend=function(){i._focusAdjacencyDisabled=!1},e.draggable=!0,e.cursor="move")})),!this._data&&t.isAnimationEnabled()&&o.setClipPath(HH(o.getBoundingRect(),t,(function(){o.removeClipPath()}))),this._data=t.getData()},e.prototype.dispose=function(){},e.type="sankey",e}(Vy);function HH(t,e,n){var i=new Al({shape:{x:t.x-10,y:t.y-10,width:0,height:t.height+20}});return eh(i,{shape:{width:t.width+20}},e,n),i}var jH=RH,zH=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(t,e){var n=t.edges||t.links,i=t.data||t.nodes,r=t.levels;this.levelModels=[];for(var o=this.levelModels,a=0;a<r.length;a++)null!=r[a].depth&&r[a].depth>=0&&(o[r[a].depth]=new qh(r[a],this,e));if(i&&n){var s=iN(i,n,this,!0,l);return s.data}function l(t,e){t.wrapMethod("getItemModel",(function(t,e){var n=t.parentModel,i=n.getData().getItemLayout(e);if(i){var r=i.depth,o=n.levelModels[r];o&&(t.parentModel=o)}return t})),e.wrapMethod("getItemModel",(function(t,e){var n=t.parentModel,i=n.getGraph().getEdgeByIndex(e),r=i.node1.getLayout();if(r){var o=r.depth,a=n.levelModels[o];a&&(t.parentModel=a)}return t}))}},e.prototype.setNodePosition=function(t,e){var n=this.option.data[t];n.localX=e[0],n.localY=e[1]},e.prototype.getGraph=function(){return this.getData().graph},e.prototype.getEdgeData=function(){return this.getGraph().edgeData},e.prototype.formatTooltip=function(t,e,n){function i(t){return isNaN(t)||null==t}if("edge"===n){var r=this.getDataParams(t,n),o=r.data,a=r.value,s=o.source+" -- "+o.target;return oy("nameValue",{name:s,value:a,noValue:i(a)})}var l=this.getGraph().getNodeByIndex(t),u=l.getLayout().value,c=this.getDataParams(t,n).data.name;return oy("nameValue",{name:null!=c?c+"":null,value:u,noValue:i(u)})},e.prototype.optionUpdated=function(){},e.prototype.getDataParams=function(e,n){var i=t.prototype.getDataParams.call(this,e,n);if(null==i.value&&"node"===n){var r=this.getGraph().getNodeByIndex(e),o=r.getLayout().value;i.value=o}return i},e.type="series.sankey",e.defaultOption={zlevel:0,z:2,coordinateSystem:"view",left:"5%",top:"5%",right:"20%",bottom:"5%",orient:"horizontal",nodeWidth:20,nodeGap:8,draggable:!0,layoutIterations:32,label:{show:!0,position:"right",fontSize:12},levels:[],nodeAlign:"justify",lineStyle:{color:"#314656",opacity:.2,curveness:.5},emphasis:{label:{show:!0},lineStyle:{opacity:.5}},select:{itemStyle:{borderColor:"#212121"}},animationEasing:"linear",animationDuration:1e3},e}(Yy),WH=zH;function BH(t,e){t.eachSeriesByType("sankey",(function(t){var n=t.get("nodeWidth"),i=t.get("nodeGap"),r=FH(t,e);t.layoutInfo=r;var o=r.width,a=r.height,s=t.getGraph(),l=s.nodes,u=s.edges;GH(l);var c=z(l,(function(t){return 0===t.getLayout().value})),d=0!==c.length?0:t.get("layoutIterations"),h=t.get("orient"),f=t.get("nodeAlign");VH(l,u,n,i,o,a,d,h,f)}))}function FH(t,e){return dp(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()})}function VH(t,e,n,i,r,o,a,s,l){$H(t,e,n,r,o,s,l),JH(t,e,o,r,i,a,s),cj(t,s)}function GH(t){R(t,(function(t){var e=lj(t.outEdges,sj),n=lj(t.inEdges,sj),i=t.getValue()||0,r=Math.max(e,n,i);t.setLayout({value:r},!0)}))}function $H(t,e,n,i,r,o,a){for(var s=[],l=[],u=[],c=[],d=0,h=0;h<e.length;h++)s[h]=1;for(h=0;h<t.length;h++)l[h]=t[h].inEdges.length,0===l[h]&&u.push(t[h]);var f=-1;while(u.length){for(var p=0;p<u.length;p++){var m=u[p],g=m.hostGraph.data.getRawDataItem(m.dataIndex),y=null!=g.depth&&g.depth>=0;y&&g.depth>f&&(f=g.depth),m.setLayout({depth:y?g.depth:d},!0),"vertical"===o?m.setLayout({dy:n},!0):m.setLayout({dx:n},!0);for(var v=0;v<m.outEdges.length;v++){var _=m.outEdges[v],b=e.indexOf(_);s[b]=0;var x=_.node2,w=t.indexOf(x);0===--l[w]&&c.indexOf(x)<0&&c.push(x)}}++d,u=c,c=[]}for(h=0;h<s.length;h++)if(1===s[h])throw new Error("Sankey is a DAG, the original data has cycle!");var M=f>d-1?f:d-1;a&&"left"!==a&&qH(t,a,o,M);var S="vertical"===o?(r-n)/M:(i-n)/M;ZH(t,S,o)}function UH(t){var e=t.hostGraph.data.getRawDataItem(t.dataIndex);return null!=e.depth&&e.depth>=0}function qH(t,e,n,i){if("right"===e){var r=[],o=t,a=0;while(o.length){for(var s=0;s<o.length;s++){var l=o[s];l.setLayout({skNodeHeight:a},!0);for(var u=0;u<l.inEdges.length;u++){var c=l.inEdges[u];r.indexOf(c.node1)<0&&r.push(c.node1)}}o=r,r=[],++a}R(t,(function(t){UH(t)||t.setLayout({depth:Math.max(0,i-t.getLayout().skNodeHeight)},!0)}))}else"justify"===e&&XH(t,i)}function XH(t,e){R(t,(function(t){UH(t)||t.outEdges.length||t.setLayout({depth:e},!0)}))}function ZH(t,e,n){R(t,(function(t){var i=t.getLayout().depth*e;"vertical"===n?t.setLayout({y:i},!0):t.setLayout({x:i},!0)}))}function JH(t,e,n,i,r,o,a){var s=KH(t,a);QH(s,e,n,i,r,a),tj(s,r,n,i,a);for(var l=1;o>0;o--)l*=.99,ej(s,l,a),tj(s,r,n,i,a),uj(s,l,a),tj(s,r,n,i,a)}function KH(t,e){var n=[],i="vertical"===e?"y":"x",r=zo(t,(function(t){return t.getLayout()[i]}));return r.keys.sort((function(t,e){return t-e})),R(r.keys,(function(t){n.push(r.buckets.get(t))})),n}function QH(t,e,n,i,r,o){var a=1/0;R(t,(function(t){var e=t.length,s=0;R(t,(function(t){s+=t.getLayout().value}));var l="vertical"===o?(i-(e-1)*r)/s:(n-(e-1)*r)/s;l<a&&(a=l)})),R(t,(function(t){R(t,(function(t,e){var n=t.getLayout().value*a;"vertical"===o?(t.setLayout({x:e},!0),t.setLayout({dx:n},!0)):(t.setLayout({y:e},!0),t.setLayout({dy:n},!0))}))})),R(e,(function(t){var e=+t.getValue()*a;t.setLayout({dy:e},!0)}))}function tj(t,e,n,i,r){var o="vertical"===r?"x":"y";R(t,(function(t){var a,s,l;t.sort((function(t,e){return t.getLayout()[o]-e.getLayout()[o]}));for(var u=0,c=t.length,d="vertical"===r?"dx":"dy",h=0;h<c;h++)s=t[h],l=u-s.getLayout()[o],l>0&&(a=s.getLayout()[o]+l,"vertical"===r?s.setLayout({x:a},!0):s.setLayout({y:a},!0)),u=s.getLayout()[o]+s.getLayout()[d]+e;var f="vertical"===r?i:n;if(l=u-e-f,l>0){a=s.getLayout()[o]-l,"vertical"===r?s.setLayout({x:a},!0):s.setLayout({y:a},!0),u=a;for(h=c-2;h>=0;--h)s=t[h],l=s.getLayout()[o]+s.getLayout()[d]+e-u,l>0&&(a=s.getLayout()[o]-l,"vertical"===r?s.setLayout({x:a},!0):s.setLayout({y:a},!0)),u=s.getLayout()[o]}}))}function ej(t,e,n){R(t.slice().reverse(),(function(t){R(t,(function(t){if(t.outEdges.length){var i=lj(t.outEdges,nj,n)/lj(t.outEdges,sj);if(isNaN(i)){var r=t.outEdges.length;i=r?lj(t.outEdges,ij,n)/r:0}if("vertical"===n){var o=t.getLayout().x+(i-aj(t,n))*e;t.setLayout({x:o},!0)}else{var a=t.getLayout().y+(i-aj(t,n))*e;t.setLayout({y:a},!0)}}}))}))}function nj(t,e){return aj(t.node2,e)*t.getValue()}function ij(t,e){return aj(t.node2,e)}function rj(t,e){return aj(t.node1,e)*t.getValue()}function oj(t,e){return aj(t.node1,e)}function aj(t,e){return"vertical"===e?t.getLayout().x+t.getLayout().dx/2:t.getLayout().y+t.getLayout().dy/2}function sj(t){return t.getValue()}function lj(t,e,n){var i=0,r=t.length,o=-1;while(++o<r){var a=+e(t[o],n);isNaN(a)||(i+=a)}return i}function uj(t,e,n){R(t,(function(t){R(t,(function(t){if(t.inEdges.length){var i=lj(t.inEdges,rj,n)/lj(t.inEdges,sj);if(isNaN(i)){var r=t.inEdges.length;i=r?lj(t.inEdges,oj,n)/r:0}if("vertical"===n){var o=t.getLayout().x+(i-aj(t,n))*e;t.setLayout({x:o},!0)}else{var a=t.getLayout().y+(i-aj(t,n))*e;t.setLayout({y:a},!0)}}}))}))}function cj(t,e){var n="vertical"===e?"x":"y";R(t,(function(t){t.outEdges.sort((function(t,e){return t.node2.getLayout()[n]-e.node2.getLayout()[n]})),t.inEdges.sort((function(t,e){return t.node1.getLayout()[n]-e.node1.getLayout()[n]}))})),R(t,(function(t){var e=0,n=0;R(t.outEdges,(function(t){t.setLayout({sy:e},!0),e+=t.getLayout().dy})),R(t.inEdges,(function(t){t.setLayout({ty:n},!0),n+=t.getLayout().dy}))}))}function dj(t){t.eachSeriesByType("sankey",(function(t){var e=t.getGraph(),n=e.nodes;if(n.length){var i=1/0,r=-1/0;R(n,(function(t){var e=t.getLayout().value;e<i&&(i=e),e>r&&(r=e)})),R(n,(function(e){var n=new hP({type:"color",mappingMethod:"linear",dataExtent:[i,r],visual:t.get("color")}),o=n.mapValueToVisual(e.getLayout().value),a=e.getModel().get(["itemStyle","color"]);null!=a?(e.setVisual("color",a),e.setVisual("style",{fill:a})):(e.setVisual("color",o),e.setVisual("style",{fill:o}))}))}}))}function hj(t){t.registerChartView(jH),t.registerSeriesModel(WH),t.registerLayout(BH),t.registerVisual(dj),t.registerAction({type:"dragNode",event:"dragnode",update:"update"},(function(t,e){e.eachComponent({mainType:"series",subType:"sankey",query:t},(function(e){e.setNodePosition(t.dataIndex,[t.localX,t.localY])}))}))}var fj=function(){function t(){}return t.prototype.getInitialData=function(t,e){var n,i,r=e.getComponent("xAxis",this.get("xAxisIndex")),o=e.getComponent("yAxis",this.get("yAxisIndex")),a=r.get("type"),s=o.get("type");"category"===a?(t.layout="horizontal",n=r.getOrdinalMeta(),i=!0):"category"===s?(t.layout="vertical",n=o.getOrdinalMeta(),i=!0):t.layout=t.layout||"horizontal";var l=["x","y"],u="horizontal"===t.layout?0:1,c=this._baseAxisDim=l[u],d=l[1-u],h=[r,o],f=h[u].get("type"),p=h[1-u].get("type"),m=t.data;if(m&&i){var g=[];R(m,(function(t,e){var n;$(t)?(n=t.slice(),t.unshift(e)):$(t.value)?(n=t.value.slice(),t.value.unshift(e)):n=t,g.push(n)})),t.data=g}var y=this.defaultValueDimensions,v=[{name:c,type:XS(f),ordinalMeta:n,otherDims:{tooltip:!1,itemName:0},dimsDef:["base"]},{name:d,type:XS(p),dimsDef:y.slice()}];return ZT(this,{coordDimensions:v,dimensionsCount:y.length+1,encodeDefaulter:G(Rp,v,this)})},t.prototype.getBaseAxis=function(){var t=this._baseAxisDim;return this.ecModel.getComponent(t+"Axis",this.get(t+"AxisIndex")).axis},t}(),pj=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.defaultValueDimensions=[{name:"min",defaultTooltip:!0},{name:"Q1",defaultTooltip:!0},{name:"median",defaultTooltip:!0},{name:"Q3",defaultTooltip:!0},{name:"max",defaultTooltip:!0}],n.visualDrawType="stroke",n}return o(e,t),e.type="series.boxplot",e.dependencies=["xAxis","yAxis","grid"],e.defaultOption={zlevel:0,z:2,coordinateSystem:"cartesian2d",legendHoverLink:!0,layout:null,boxWidth:[7,50],itemStyle:{color:"#fff",borderWidth:1},emphasis:{scale:!0,itemStyle:{borderWidth:2,shadowBlur:5,shadowOffsetX:1,shadowOffsetY:1,shadowColor:"rgba(0,0,0,0.2)"}},animationDuration:800},e}(Yy);E(pj,fj,!0);var mj=pj,gj=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=t.getData(),r=this.group,o=this._data;this._data||r.removeAll();var a="horizontal"===t.get("layout")?1:0;i.diff(o).add((function(t){if(i.hasValue(t)){var e=i.getItemLayout(t),n=_j(e,i,t,a,!0);i.setItemGraphicEl(t,n),r.add(n)}})).update((function(t,e){var n=o.getItemGraphicEl(e);if(i.hasValue(t)){var s=i.getItemLayout(t);n?bj(s,n,i,t):n=_j(s,i,t,a),r.add(n),i.setItemGraphicEl(t,n)}else r.remove(n)})).remove((function(t){var e=o.getItemGraphicEl(t);e&&r.remove(e)})).execute(),this._data=i},e.prototype.remove=function(t){var e=this.group,n=this._data;this._data=null,n&&n.eachItemGraphicEl((function(t){t&&e.remove(t)}))},e.type="boxplot",e}(Vy),yj=function(){function t(){}return t}(),vj=function(t){function e(e){var n=t.call(this,e)||this;return n.type="boxplotBoxPath",n}return o(e,t),e.prototype.getDefaultShape=function(){return new yj},e.prototype.buildPath=function(t,e){var n=e.points,i=0;for(t.moveTo(n[i][0],n[i][1]),i++;i<4;i++)t.lineTo(n[i][0],n[i][1]);for(t.closePath();i<n.length;i++)t.moveTo(n[i][0],n[i][1]),i++,t.lineTo(n[i][0],n[i][1])},e}(pl);function _j(t,e,n,i,r){var o=t.ends,a=new vj({shape:{points:r?xj(o,i,t):o}});return bj(t,a,e,n,r),a}function bj(t,e,n,r,o){var a=n.hostModel,s=i[o?"initProps":"updateProps"];s(e,{shape:{points:t.ends}},a,r),e.useStyle(n.getItemVisual(r,"style")),e.style.strokeNoScale=!0,e.z2=100;var l=n.getItemModel(r);Zu(e,l),$u(e,l.get(["emphasis","focus"]),l.get(["emphasis","blurScope"]))}function xj(t,e,n){return H(t,(function(t){return t=t.slice(),t[e]=n.initBaseline,t}))}var wj=gj;function Mj(t,e){}var Sj=R;function Lj(t){var e=kj(t);Sj(e,(function(t){var e=t.seriesModels;e.length&&(Tj(t),Sj(e,(function(e,n){Dj(e,t.boxOffsetList[n],t.boxWidthList[n])})))}))}function kj(t){var e=[],n=[];return t.eachSeriesByType("boxplot",(function(t){var i=t.getBaseAxis(),r=O(n,i);r<0&&(r=n.length,n[r]=i,e[r]={axis:i,seriesModels:[]}),e[r].seriesModels.push(t)})),e}function Tj(t){var e,n,i=t.axis,r=t.seriesModels,o=r.length,a=t.boxWidthList=[],s=t.boxOffsetList=[],l=[];if("category"===i.type)n=i.getBandWidth();else{var u=0;Sj(r,(function(t){u=Math.max(u,t.getData().count())})),e=i.getExtent(),Math.abs(e[1]-e[0])}Sj(r,(function(t){var e=t.get("boxWidth");$(e)||(e=[e,e]),l.push([Nr(e[0],n)||0,Nr(e[1],n)||0])}));var c=.8*n-2,d=c/o*.3,h=(c-d*(o-1))/o,f=h/2-c/2;Sj(r,(function(t,e){s.push(f),f+=d+h,a.push(Math.min(Math.max(h,l[e][0]),l[e][1]))}))}function Dj(t,e,n){var i=t.coordinateSystem,r=t.getData(),o=n/2,a="horizontal"===t.get("layout")?0:1,s=1-a,l=["x","y"],u=r.mapDimension(l[a]),c=r.mapDimensionsAll(l[s]);if(!(null==u||c.length<5))for(var d=0;d<r.count();d++){var h=r.get(u,d),f=_(h,c[2],d),p=_(h,c[0],d),m=_(h,c[1],d),g=_(h,c[3],d),y=_(h,c[4],d),v=[];b(v,m,!1),b(v,g,!0),v.push(p,m,y,g),x(v,p),x(v,y),x(v,f),r.setItemLayout(d,{initBaseline:f[s],ends:v})}function _(t,n,o){var l,u=r.get(n,o),c=[];return c[a]=t,c[s]=u,isNaN(t)||isNaN(u)?l=[NaN,NaN]:(l=i.dataToPoint(c),l[a]+=e),l}function b(t,e,n){var i=e.slice(),r=e.slice();i[a]+=o,r[a]-=o,n?t.push(i,r):t.push(r,i)}function x(t,e){var n=e.slice(),i=e.slice();n[a]-=o,i[a]+=o,t.push(n,i)}}function Cj(t,e){e=e||{};for(var n=[],i=[],r=e.boundIQR,o="none"===r||0===r,a=0;a<t.length;a++){var s=Hr(t[a].slice()),l=Kr(s,.25),u=Kr(s,.5),c=Kr(s,.75),d=s[0],h=s[s.length-1],f=(null==r?1.5:r)*(c-l),p=o?d:Math.max(d,l-f),m=o?h:Math.min(h,c+f),g=e.itemNameFormatter,y=U(g)?g({value:a}):q(g)?g.replace("{value}",a+""):a+"";n.push([y,p,l,u,c,m]);for(var v=0;v<s.length;v++){var _=s[v];if(_<p||_>m){var b=[y,_];i.push(b)}}}return{boxData:n,outliers:i}}var Ij={type:"echarts:boxplot",transform:function(t){var e=t.upstream;if(e.sourceFormat!==Tp){var n="";0,im(n)}var i=Cj(e.getRawData(),t.config);return[{dimensions:["ItemName","Low","Q1","Q2","Q3","High"],data:i.boxData},{data:i.outliers}]}};function Aj(t){t.registerSeriesModel(mj),t.registerChartView(wj),t.registerVisual(Mj),t.registerLayout(Lj),t.registerTransform(Ij)}var Yj=["color","borderColor"],Oj=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){this.group.removeClipPath(),this._updateDrawMode(t),this._isLargeDraw?this._renderLarge(t):this._renderNormal(t)},e.prototype.incrementalPrepareRender=function(t,e,n){this._clear(),this._updateDrawMode(t)},e.prototype.incrementalRender=function(t,e,n,i){this._isLargeDraw?this._incrementalRenderLarge(t,e):this._incrementalRenderNormal(t,e)},e.prototype._updateDrawMode=function(t){var e=t.pipelineContext.large;null!=this._isLargeDraw&&e===this._isLargeDraw||(this._isLargeDraw=e,this._clear())},e.prototype._renderNormal=function(t){var e=t.getData(),n=this._data,i=this.group,r=e.getLayout("isSimpleBox"),o=t.get("clip",!0),a=t.coordinateSystem,s=a.getArea&&a.getArea();this._data||i.removeAll(),e.diff(n).add((function(n){if(e.hasValue(n)){var a=e.getItemLayout(n);if(o&&Rj(s,a))return;var l=Nj(a,n,!0);eh(l,{shape:{points:a.ends}},t,n),Hj(l,e,n,r),i.add(l),e.setItemGraphicEl(n,l)}})).update((function(a,l){var u=n.getItemGraphicEl(l);if(e.hasValue(a)){var c=e.getItemLayout(a);o&&Rj(s,c)?i.remove(u):(u?th(u,{shape:{points:c.ends}},t,a):u=Nj(c,a),Hj(u,e,a,r),i.add(u),e.setItemGraphicEl(a,u))}else i.remove(u)})).remove((function(t){var e=n.getItemGraphicEl(t);e&&i.remove(e)})).execute(),this._data=e},e.prototype._renderLarge=function(t){this._clear(),Bj(t,this.group);var e=t.get("clip",!0)?_k(t.coordinateSystem,!1,t):null;e?this.group.setClipPath(e):this.group.removeClipPath()},e.prototype._incrementalRenderNormal=function(t,e){var n,i=e.getData(),r=i.getLayout("isSimpleBox");while(null!=(n=t.next())){var o=i.getItemLayout(n),a=Nj(o,n);Hj(a,i,n,r),a.incremental=!0,this.group.add(a)}},e.prototype._incrementalRenderLarge=function(t,e){Bj(e,this.group,!0)},e.prototype.remove=function(t){this._clear()},e.prototype._clear=function(){this.group.removeAll(),this._data=null},e.type="candlestick",e}(Vy),Pj=function(){function t(){}return t}(),Ej=function(t){function e(e){var n=t.call(this,e)||this;return n.type="normalCandlestickBox",n}return o(e,t),e.prototype.getDefaultShape=function(){return new Pj},e.prototype.buildPath=function(t,e){var n=e.points;this.__simpleBox?(t.moveTo(n[4][0],n[4][1]),t.lineTo(n[6][0],n[6][1])):(t.moveTo(n[0][0],n[0][1]),t.lineTo(n[1][0],n[1][1]),t.lineTo(n[2][0],n[2][1]),t.lineTo(n[3][0],n[3][1]),t.closePath(),t.moveTo(n[4][0],n[4][1]),t.lineTo(n[5][0],n[5][1]),t.moveTo(n[6][0],n[6][1]),t.lineTo(n[7][0],n[7][1]))},e}(pl);function Nj(t,e,n){var i=t.ends;return new Ej({shape:{points:n?jj(i,t):i},z2:100})}function Rj(t,e){for(var n=!0,i=0;i<e.ends.length;i++)if(t.contain(e.ends[i][0],e.ends[i][1])){n=!1;break}return n}function Hj(t,e,n,i){var r=e.getItemModel(n);t.useStyle(e.getItemVisual(n,"style")),t.style.strokeNoScale=!0,t.__simpleBox=i,Zu(t,r)}function jj(t,e){return H(t,(function(t){return t=t.slice(),t[1]=e.initBaseline,t}))}var zj=function(){function t(){}return t}(),Wj=function(t){function e(e){var n=t.call(this,e)||this;return n.type="largeCandlestickBox",n}return o(e,t),e.prototype.getDefaultShape=function(){return new zj},e.prototype.buildPath=function(t,e){for(var n=e.points,i=0;i<n.length;)if(this.__sign===n[i++]){var r=n[i++];t.moveTo(r,n[i++]),t.lineTo(r,n[i++])}else i+=3},e}(pl);function Bj(t,e,n){var i=t.getData(),r=i.getLayout("largePoints"),o=new Wj({shape:{points:r},__sign:1});e.add(o);var a=new Wj({shape:{points:r},__sign:-1});e.add(a),Fj(1,o,t,i),Fj(-1,a,t,i),n&&(o.incremental=!0,a.incremental=!0)}function Fj(t,e,n,i){var r=n.get(["itemStyle",t>0?"borderColor":"borderColor0"])||n.get(["itemStyle",t>0?"color":"color0"]),o=n.getModel("itemStyle").getItemStyle(Yj);e.useStyle(o),e.style.fill=null,e.style.stroke=r}var Vj=Oj,Gj=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.defaultValueDimensions=[{name:"open",defaultTooltip:!0},{name:"close",defaultTooltip:!0},{name:"lowest",defaultTooltip:!0},{name:"highest",defaultTooltip:!0}],n}return o(e,t),e.prototype.getShadowDim=function(){return"open"},e.prototype.brushSelector=function(t,e,n){var i=e.getItemLayout(t);return i&&n.rect(i.brushRect)},e.type="series.candlestick",e.dependencies=["xAxis","yAxis","grid"],e.defaultOption={zlevel:0,z:2,coordinateSystem:"cartesian2d",legendHoverLink:!0,layout:null,clip:!0,itemStyle:{color:"#eb5454",color0:"#47b262",borderColor:"#eb5454",borderColor0:"#47b262",borderWidth:1},emphasis:{scale:!0,itemStyle:{borderWidth:2}},barMaxWidth:null,barMinWidth:null,barWidth:null,large:!0,largeThreshold:600,progressive:3e3,progressiveThreshold:1e4,progressiveChunkMode:"mod",animationEasing:"linear",animationDuration:300},e}(Yy);E(Gj,fj,!0);var $j=Gj;function Uj(t){t&&$(t.series)&&R(t.series,(function(t){J(t)&&"k"===t.type&&(t.type="candlestick")}))}var qj=["itemStyle","borderColor"],Xj=["itemStyle","borderColor0"],Zj=["itemStyle","color"],Jj=["itemStyle","color0"],Kj={seriesType:"candlestick",plan:Ey(),performRawSeries:!0,reset:function(t,e){function n(t,e){return e.get(t>0?Zj:Jj)}function i(t,e){return e.get(t>0?qj:Xj)}if(!e.isSeriesFiltered(t)){var r=t.pipelineContext.large;return!r&&{progress:function(t,e){var r;while(null!=(r=t.next())){var o=e.getItemModel(r),a=e.getItemLayout(r).sign,s=o.getItemStyle();s.fill=n(a,o),s.stroke=i(a,o)||s.fill;var l=e.ensureUniqueItemVisual(r,"style");I(l,s)}}}}}},Qj=Kj,tz="undefined"!==typeof Float32Array?Float32Array:Array,ez={seriesType:"candlestick",plan:Ey(),reset:function(t){var e=t.coordinateSystem,n=t.getData(),i=iz(t,n),r=0,o=1,a=["x","y"],s=n.mapDimension(a[r]),l=n.mapDimensionsAll(a[o]),u=l[0],c=l[1],d=l[2],h=l[3];if(n.setLayout({candleWidth:i,isSimpleBox:i<=1.3}),!(null==s||l.length<4))return{progress:t.pipelineContext.large?p:f};function f(t,n){var a;while(null!=(a=t.next())){var l=n.get(s,a),f=n.get(u,a),p=n.get(c,a),m=n.get(d,a),g=n.get(h,a),y=Math.min(f,p),v=Math.max(f,p),_=S(y,l),b=S(v,l),x=S(m,l),w=S(g,l),M=[];L(M,b,0),L(M,_,1),M.push(T(w),T(b),T(x),T(_)),n.setItemLayout(a,{sign:nz(n,a,f,p,c),initBaseline:f>p?b[o]:_[o],ends:M,brushRect:k(m,g,l)})}function S(t,n){var i=[];return i[r]=n,i[o]=t,isNaN(n)||isNaN(t)?[NaN,NaN]:e.dataToPoint(i)}function L(t,e,n){var o=e.slice(),a=e.slice();o[r]=Kd(o[r]+i/2,1,!1),a[r]=Kd(a[r]-i/2,1,!0),n?t.push(o,a):t.push(a,o)}function k(t,e,n){var a=S(t,n),s=S(e,n);return a[r]-=i/2,s[r]-=i/2,{x:a[0],y:a[1],width:o?i:s[0]-a[0],height:o?s[1]-a[1]:i}}function T(t){return t[r]=Kd(t[r],1),t}}function p(t,n){var i,a,l=new tz(4*t.count),f=0,p=[],m=[];while(null!=(a=t.next())){var g=n.get(s,a),y=n.get(u,a),v=n.get(c,a),_=n.get(d,a),b=n.get(h,a);isNaN(g)||isNaN(_)||isNaN(b)?(l[f++]=NaN,f+=3):(l[f++]=nz(n,a,y,v,c),p[r]=g,p[o]=_,i=e.dataToPoint(p,null,m),l[f++]=i?i[0]:NaN,l[f++]=i?i[1]:NaN,p[o]=b,i=e.dataToPoint(p,null,m),l[f++]=i?i[1]:NaN)}n.setLayout("largePoints",l)}}};function nz(t,e,n,i,r){var o;return o=n>i?-1:n<i?1:e>0?t.get(r,e-1)<=i?1:-1:1,o}function iz(t,e){var n,i=t.getBaseAxis(),r="category"===i.type?i.getBandWidth():(n=i.getExtent(),Math.abs(n[1]-n[0])/e.count()),o=Nr(at(t.get("barMaxWidth"),r),r),a=Nr(at(t.get("barMinWidth"),1),r),s=t.get("barWidth");return null!=s?Nr(s,r):Math.max(Math.min(r/2,o),a)}var rz=ez;function oz(t){t.registerChartView(Vj),t.registerSeriesModel($j),t.registerPreprocessor(Uj),t.registerVisual(Qj),t.registerLayout(rz)}var az=3;function sz(t){return $(t)||(t=[+t,+t]),t}function lz(t,e){var n=e.rippleEffectColor||e.color;t.eachChild((function(t){t.attr({z:e.z,zlevel:e.zlevel,style:{stroke:"stroke"===e.brushType?n:null,fill:"fill"===e.brushType?n:null}})}))}var uz=function(t){function e(e,n){var i=t.call(this)||this,r=new XL(e,n),o=new Mr;return i.add(r),i.add(o),i.updateData(e,n),i}return o(e,t),e.prototype.stopEffectAnimation=function(){this.childAt(1).removeAll()},e.prototype.startEffectAnimation=function(t){for(var e=t.symbolType,n=t.color,i=this.childAt(1),r=0;r<az;r++){var o=j_(e,-1,-1,2,2,n);o.attr({style:{strokeNoScale:!0},z2:99,silent:!0,scaleX:.5,scaleY:.5});var a=-r/az*t.period+t.effectOffset;o.animate("",!0).when(t.period,{scaleX:t.rippleScale/2,scaleY:t.rippleScale/2}).delay(a).start(),o.animateStyle(!0).when(t.period,{opacity:0}).delay(a).start(),i.add(o)}lz(i,t)},e.prototype.updateEffectAnimation=function(t){for(var e=this._effectCfg,n=this.childAt(1),i=["symbolType","period","rippleScale"],r=0;r<i.length;r++){var o=i[r];if(e[o]!==t[o])return this.stopEffectAnimation(),void this.startEffectAnimation(t)}lz(n,t)},e.prototype.highlight=function(){Cu(this)},e.prototype.downplay=function(){Iu(this)},e.prototype.updateData=function(t,e){var n=this,i=t.hostModel;this.childAt(0).updateData(t,e);var r=this.childAt(1),o=t.getItemModel(e),a=t.getItemVisual(e,"symbol"),s=sz(t.getItemVisual(e,"symbolSize")),l=t.getItemVisual(e,"style"),u=l&&l.fill;r.setScale(s),r.traverse((function(t){t.setStyle("fill",u)}));var c=t.getItemVisual(e,"symbolOffset");c&&($(c)||(c=[c,c]),r.x=Nr(c[0],s[0]),r.y=Nr(at(c[1],c[0])||0,s[1]));var d=t.getItemVisual(e,"symbolRotate");r.rotation=(d||0)*Math.PI/180||0;var h={};h.showEffectOn=i.get("showEffectOn"),h.rippleScale=o.get(["rippleEffect","scale"]),h.brushType=o.get(["rippleEffect","brushType"]),h.period=1e3*o.get(["rippleEffect","period"]),h.effectOffset=e/t.count(),h.z=i.getShallow("z")||0,h.zlevel=i.getShallow("zlevel")||0,h.symbolType=a,h.color=u,h.rippleEffectColor=o.get(["rippleEffect","color"]),this.off("mouseover").off("mouseout").off("emphasis").off("normal"),"render"===h.showEffectOn?(this._effectCfg?this.updateEffectAnimation(h):this.startEffectAnimation(h),this._effectCfg=h):(this._effectCfg=null,this.stopEffectAnimation(),this.onHoverStateChange=function(t){"emphasis"===t?"render"!==h.showEffectOn&&n.startEffectAnimation(h):"normal"===t&&"render"!==h.showEffectOn&&n.stopEffectAnimation()}),this._effectCfg=h,$u(this)},e.prototype.fadeOut=function(t){this.off("mouseover").off("mouseout"),t&&t()},e}(Mr),cz=uz,dz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){this._symbolDraw=new tk(cz)},e.prototype.render=function(t,e,n){var i=t.getData(),r=this._symbolDraw;r.updateData(i,{clipShape:this._getClipShape(t)}),this.group.add(r.group)},e.prototype._getClipShape=function(t){var e=t.coordinateSystem,n=e&&e.getArea&&e.getArea();return t.get("clip",!0)?n:null},e.prototype.updateTransform=function(t,e,n){var i=t.getData();this.group.dirty();var r=Hk("").reset(t,e,n);r.progress&&r.progress({start:0,end:i.count(),count:i.count()},i),this._symbolDraw.updateLayout()},e.prototype._updateGroupTransform=function(t){var e=t.coordinateSystem;e&&e.getRoamTransform&&(this.group.transform=Di(e.getRoamTransform()),this.group.decomposeTransform())},e.prototype.remove=function(t,e){this._symbolDraw&&this._symbolDraw.remove(!0)},e.type="effectScatter",e}(Vy),hz=dz,fz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){return BL(this.getSource(),this,{useEncodeDefaulter:!0})},e.prototype.brushSelector=function(t,e,n){return n.point(e.getItemLayout(t))},e.type="series.effectScatter",e.dependencies=["grid","polar"],e.defaultOption={coordinateSystem:"cartesian2d",zlevel:0,z:2,legendHoverLink:!0,effectType:"ripple",progressive:0,showEffectOn:"render",clip:!0,rippleEffect:{period:4,scale:2.5,brushType:"fill"},symbolSize:10},e}(Yy),pz=fz;function mz(t){t.registerChartView(hz),t.registerSeriesModel(pz),t.registerLayout(Hk("effectScatter"))}var gz=function(t){function e(e,n,i){var r=t.call(this)||this;return r.add(r.createLine(e,n,i)),r._updateEffectSymbol(e,n),r}return o(e,t),e.prototype.createLine=function(t,e,n){return new OE(t,e,n)},e.prototype._updateEffectSymbol=function(t,e){var n=t.getItemModel(e),i=n.getModel("effect"),r=i.get("symbolSize"),o=i.get("symbol");$(r)||(r=[r,r]);var a=t.getItemVisual(e,"style"),s=i.get("color")||a&&a.stroke,l=this.childAt(1);this._symbolType!==o&&(this.remove(l),l=j_(o,-.5,-.5,1,1,s),l.z2=100,l.culling=!0,this.add(l)),l&&(l.setStyle("shadowColor",s),l.setStyle(i.getItemStyle(["color"])),l.scaleX=r[0],l.scaleY=r[1],l.setColor(s),this._symbolType=o,this._symbolScale=r,this._updateEffectAnimation(t,i,e))},e.prototype._updateEffectAnimation=function(t,e,n){var i=this.childAt(1);if(i){var r=this,o=t.getItemLayout(n),a=1e3*e.get("period"),s=e.get("loop"),l=e.get("constantSpeed"),u=ot(e.get("delay"),(function(e){return e/t.count()*a/3}));if(i.ignore=!0,this._updateAnimationPoints(i,o),l>0&&(a=this._getLineLength(i)/l*1e3),(a!==this._period||s!==this._loop)&&(i.stopAnimation(),a>0)){var c=void 0;c="function"===typeof u?u(n):u,i.__t>0&&(c=-a*i.__t),i.__t=0;var d=i.animate("",s).when(a,{__t:1}).delay(c).during((function(){r._updateSymbolPosition(i)}));s||d.done((function(){r.remove(i)})),d.start()}this._period=a,this._loop=s}},e.prototype._getLineLength=function(t){return Et(t.__p1,t.__cp1)+Et(t.__cp1,t.__p2)},e.prototype._updateAnimationPoints=function(t,e){t.__p1=e[0],t.__p2=e[1],t.__cp1=e[2]||[(e[0][0]+e[1][0])/2,(e[0][1]+e[1][1])/2]},e.prototype.updateData=function(t,e,n){this.childAt(0).updateData(t,e,n),this._updateEffectSymbol(t,e)},e.prototype._updateSymbolPosition=function(t){var e=t.__p1,n=t.__p2,i=t.__cp1,r=t.__t,o=[t.x,t.y],a=o.slice(),s=ts,l=es;o[0]=s(e[0],i[0],n[0],r),o[1]=s(e[1],i[1],n[1],r);var u=l(e[0],i[0],n[0],r),c=l(e[1],i[1],n[1],r);t.rotation=-Math.atan2(c,u)-Math.PI/2,"line"!==this._symbolType&&"rect"!==this._symbolType&&"roundRect"!==this._symbolType||(void 0!==t.__lastT&&t.__lastT<t.__t?(t.scaleY=1.05*Et(a,o),1===r&&(o[0]=a[0]+(o[0]-a[0])/2,o[1]=a[1]+(o[1]-a[1])/2)):1===t.__lastT?t.scaleY=2*Et(e,o):t.scaleY=this._symbolScale[1]),t.__lastT=t.__t,t.ignore=!1,t.x=o[0],t.y=o[1]},e.prototype.updateLayout=function(t,e){this.childAt(0).updateLayout(t,e);var n=t.getItemModel(e).getModel("effect");this._updateEffectAnimation(t,n,e)},e}(Mr),yz=gz,vz=function(t){function e(e,n,i){var r=t.call(this)||this;return r._createPolyline(e,n,i),r}return o(e,t),e.prototype._createPolyline=function(t,e,n){var i=t.getItemLayout(e),r=new sd({shape:{points:i}});this.add(r),this._updateCommonStl(t,e,n)},e.prototype.updateData=function(t,e,n){var i=t.hostModel,r=this.childAt(0),o={shape:{points:t.getItemLayout(e)}};th(r,o,i,e),this._updateCommonStl(t,e,n)},e.prototype._updateCommonStl=function(t,e,n){var i=this.childAt(0),r=t.getItemModel(e),o=n&&n.emphasisLineStyle;n&&!t.hasItemOption||(o=r.getModel(["emphasis","lineStyle"]).getLineStyle()),i.useStyle(t.getItemVisual(e,"style")),i.style.fill=null,i.style.strokeNoScale=!0;var a=i.ensureState("emphasis");a.style=o,$u(this)},e.prototype.updateLayout=function(t,e){var n=this.childAt(0);n.setShape("points",t.getItemLayout(e))},e}(Mr),_z=vz,bz=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e._lastFrame=0,e._lastFramePercent=0,e}return o(e,t),e.prototype.createLine=function(t,e,n){return new _z(t,e,n)},e.prototype._updateAnimationPoints=function(t,e){this._points=e;for(var n=[0],i=0,r=1;r<e.length;r++){var o=e[r-1],a=e[r];i+=Et(o,a),n.push(i)}if(0!==i){for(r=0;r<n.length;r++)n[r]/=i;this._offsets=n,this._length=i}else this._length=0},e.prototype._getLineLength=function(){return this._length},e.prototype._updateSymbolPosition=function(t){var e=t.__t,n=this._points,i=this._offsets,r=n.length;if(i){var o,a=this._lastFrame;if(e<this._lastFramePercent){var s=Math.min(a+1,r-1);for(o=s;o>=0;o--)if(i[o]<=e)break;o=Math.min(o,r-2)}else{for(o=a;o<r;o++)if(i[o]>e)break;o=Math.min(o-1,r-2)}var l=(e-i[o])/(i[o+1]-i[o]),u=n[o],c=n[o+1];t.x=u[0]*(1-l)+l*c[0],t.y=u[1]*(1-l)+l*c[1];var d=c[0]-u[0],h=c[1]-u[1];t.rotation=-Math.atan2(h,d)-Math.PI/2,this._lastFrame=o,this._lastFramePercent=e,t.ignore=!1}},e}(yz),xz=bz,wz=function(){function t(){this.polyline=!1,this.curveness=0,this.segs=[]}return t}(),Mz=function(t){function e(e){return t.call(this,e)||this}return o(e,t),e.prototype.getDefaultStyle=function(){return{stroke:"#000",fill:null}},e.prototype.getDefaultShape=function(){return new wz},e.prototype.buildPath=function(t,e){var n=e.segs,i=e.curveness;if(e.polyline)for(var r=0;r<n.length;){var o=n[r++];if(o>0){t.moveTo(n[r++],n[r++]);for(var a=1;a<o;a++)t.lineTo(n[r++],n[r++])}}else for(r=0;r<n.length;){var s=n[r++],l=n[r++],u=n[r++],c=n[r++];if(t.moveTo(s,l),i>0){var d=(s+u)/2-(l-c)*i,h=(l+c)/2-(u-s)*i;t.quadraticCurveTo(d,h,u,c)}else t.lineTo(u,c)}},e.prototype.findDataIndex=function(t,e){var n=this.shape,i=n.segs,r=n.curveness,o=this.style.lineWidth;if(n.polyline)for(var a=0,s=0;s<i.length;){var l=i[s++];if(l>0)for(var u=i[s++],c=i[s++],d=1;d<l;d++){var h=i[s++],f=i[s++];if(Fs(u,c,h,f,o,t,e))return a}a++}else for(a=0,s=0;s<i.length;){u=i[s++],c=i[s++],h=i[s++],f=i[s++];if(r>0){var p=(u+h)/2-(c-f)*r,m=(c+f)/2-(h-u)*r;if(Gs(u,c,p,m,h,f,o,t,e))return a}else if(Fs(u,c,h,f,o,t,e))return a;a++}return-1},e}(pl),Sz=function(){function t(){this.group=new Mr}return t.prototype.isPersistent=function(){return!this._incremental},t.prototype.updateData=function(t){this.group.removeAll();var e=new Mz({rectHover:!0,cursor:"default"});e.setShape({segs:t.getLayout("linesPoints")}),this._setCommon(e,t),this.group.add(e),this._incremental=null},t.prototype.incrementalPrepareUpdate=function(t){this.group.removeAll(),this._clearIncremental(),t.count()>5e5?(this._incremental||(this._incremental=new Nd({silent:!0})),this.group.add(this._incremental)):this._incremental=null},t.prototype.incrementalUpdate=function(t,e){var n=new Mz;n.setShape({segs:e.getLayout("linesPoints")}),this._setCommon(n,e,!!this._incremental),this._incremental?this._incremental.addDisplayable(n,!0):(n.rectHover=!0,n.cursor="default",n.__startIndex=t.start,this.group.add(n))},t.prototype.remove=function(){this._clearIncremental(),this._incremental=null,this.group.removeAll()},t.prototype._setCommon=function(t,e,n){var i=e.hostModel;t.setShape({polyline:i.get("polyline"),curveness:i.get(["lineStyle","curveness"])}),t.useStyle(i.getModel("lineStyle").getLineStyle()),t.style.strokeNoScale=!0;var r=e.getVisual("style");if(r&&r.stroke&&t.setStyle("stroke",r.stroke),t.setStyle("fill",null),!n){var o=$l(t);o.seriesIndex=i.seriesIndex,t.on("mousemove",(function(e){o.dataIndex=null;var n=t.findDataIndex(e.offsetX,e.offsetY);n>0&&(o.dataIndex=n+t.__startIndex)}))}},t.prototype._clearIncremental=function(){var t=this._incremental;t&&t.clearDisplaybles()},t}(),Lz=Sz,kz={seriesType:"lines",plan:Ey(),reset:function(t){var e=t.coordinateSystem,n=t.get("polyline"),i=t.pipelineContext.large;return{progress:function(r,o){var a=[];if(i){var s=void 0,l=r.end-r.start;if(n){for(var u=0,c=r.start;c<r.end;c++)u+=t.getLineCoordsCount(c);s=new Float32Array(l+2*u)}else s=new Float32Array(4*l);var d=0,h=[];for(c=r.start;c<r.end;c++){var f=t.getLineCoords(c,a);n&&(s[d++]=f);for(var p=0;p<f;p++)h=e.dataToPoint(a[p],!1,h),s[d++]=h[0],s[d++]=h[1]}o.setLayout("linesPoints",s)}else for(c=r.start;c<r.end;c++){var m=o.getItemModel(c),g=(f=t.getLineCoords(c,a),[]);if(n)for(var y=0;y<f;y++)g.push(e.dataToPoint(a[y]));else{g[0]=e.dataToPoint(a[0]),g[1]=e.dataToPoint(a[1]);var v=m.get(["lineStyle","curveness"]);+v&&(g[2]=[(g[0][0]+g[1][0])/2-(g[0][1]-g[1][1])*v,(g[0][1]+g[1][1])/2-(g[1][0]-g[0][0])*v])}o.setItemLayout(c,g)}}}}},Tz=kz,Dz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=t.getData(),r=this._updateLineDraw(i,t),o=t.get("zlevel"),a=t.get(["effect","trailLength"]),s=n.getZr(),l="svg"===s.painter.getType();(l||s.painter.getLayer(o).clear(!0),null==this._lastZlevel||l||s.configLayer(this._lastZlevel,{motionBlur:!1}),this._showEffect(t)&&a)&&(l||s.configLayer(o,{motionBlur:!0,lastFrameAlpha:Math.max(Math.min(a/10+.9,1),0)}));r.updateData(i);var u=t.get("clip",!0)&&_k(t.coordinateSystem,!1,t);u?this.group.setClipPath(u):this.group.removeClipPath(),this._lastZlevel=o,this._finished=!0},e.prototype.incrementalPrepareRender=function(t,e,n){var i=t.getData(),r=this._updateLineDraw(i,t);r.incrementalPrepareUpdate(i),this._clearLayer(n),this._finished=!1},e.prototype.incrementalRender=function(t,e,n){this._lineDraw.incrementalUpdate(t,e.getData()),this._finished=t.end===e.getData().count()},e.prototype.updateTransform=function(t,e,n){var i=t.getData(),r=t.pipelineContext;if(!this._finished||r.large||r.progressiveRender)return{update:!0};var o=Tz.reset(t,e,n);o.progress&&o.progress({start:0,end:i.count(),count:i.count()},i),this._lineDraw.updateLayout(),this._clearLayer(n)},e.prototype._updateLineDraw=function(t,e){var n=this._lineDraw,i=this._showEffect(e),r=!!e.get("polyline"),o=e.pipelineContext,a=o.large;return n&&i===this._hasEffet&&r===this._isPolyline&&a===this._isLargeDraw||(n&&n.remove(),n=this._lineDraw=a?new Lz:new jE(r?i?xz:_z:i?yz:OE),this._hasEffet=i,this._isPolyline=r,this._isLargeDraw=a,this.group.removeAll()),this.group.add(n.group),n},e.prototype._showEffect=function(t){return!!t.get(["effect","show"])},e.prototype._clearLayer=function(t){var e=t.getZr(),n="svg"===e.painter.getType();n||null==this._lastZlevel||e.painter.getLayer(this._lastZlevel).clear(!0)},e.prototype.remove=function(t,e){this._lineDraw&&this._lineDraw.remove(),this._lineDraw=null,this._clearLayer(e)},e.type="lines",e}(Vy),Cz=Dz,Iz="undefined"===typeof Uint32Array?Array:Uint32Array,Az="undefined"===typeof Float64Array?Array:Float64Array;function Yz(t){var e=t.data;e&&e[0]&&e[0][0]&&e[0][0].coord&&(t.data=H(e,(function(t){var e=[t[0].coord,t[1].coord],n={coords:e};return t[0].name&&(n.fromName=t[0].name),t[1].name&&(n.toName=t[1].name),C([n,t[0],t[1]])})))}var Oz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.visualStyleAccessPath="lineStyle",n.visualDrawType="stroke",n}return o(e,t),e.prototype.init=function(e){e.data=e.data||[],Yz(e);var n=this._processFlatCoordsArray(e.data);this._flatCoords=n.flatCoords,this._flatCoordsOffset=n.flatCoordsOffset,n.flatCoords&&(e.data=new Float32Array(n.count)),t.prototype.init.apply(this,arguments)},e.prototype.mergeOption=function(e){if(Yz(e),e.data){var n=this._processFlatCoordsArray(e.data);this._flatCoords=n.flatCoords,this._flatCoordsOffset=n.flatCoordsOffset,n.flatCoords&&(e.data=new Float32Array(n.count))}t.prototype.mergeOption.apply(this,arguments)},e.prototype.appendData=function(t){var e=this._processFlatCoordsArray(t.data);e.flatCoords&&(this._flatCoords?(this._flatCoords=yt(this._flatCoords,e.flatCoords),this._flatCoordsOffset=yt(this._flatCoordsOffset,e.flatCoordsOffset)):(this._flatCoords=e.flatCoords,this._flatCoordsOffset=e.flatCoordsOffset),t.data=new Float32Array(e.count)),this.getRawData().appendData(t.data)},e.prototype._getCoordsFromItemModel=function(t){var e=this.getData().getItemModel(t),n=e.option instanceof Array?e.option:e.getShallow("coords");return n},e.prototype.getLineCoordsCount=function(t){return this._flatCoordsOffset?this._flatCoordsOffset[2*t+1]:this._getCoordsFromItemModel(t).length},e.prototype.getLineCoords=function(t,e){if(this._flatCoordsOffset){for(var n=this._flatCoordsOffset[2*t],i=this._flatCoordsOffset[2*t+1],r=0;r<i;r++)e[r]=e[r]||[],e[r][0]=this._flatCoords[n+2*r],e[r][1]=this._flatCoords[n+2*r+1];return i}var o=this._getCoordsFromItemModel(t);for(r=0;r<o.length;r++)e[r]=e[r]||[],e[r][0]=o[r][0],e[r][1]=o[r][1];return o.length},e.prototype._processFlatCoordsArray=function(t){var e=0;if(this._flatCoords&&(e=this._flatCoords.length),"number"===typeof t[0]){for(var n=t.length,i=new Iz(n),r=new Az(n),o=0,a=0,s=0,l=0;l<n;){s++;var u=t[l++];i[a++]=o+e,i[a++]=u;for(var c=0;c<u;c++){var d=t[l++],h=t[l++];r[o++]=d,r[o++]=h}}return{flatCoordsOffset:new Uint32Array(i.buffer,0,a),flatCoords:r,count:s}}return{flatCoordsOffset:null,flatCoords:null,count:t.length}},e.prototype.getInitialData=function(t,e){var n=new kL(["value"],this);return n.hasItemOption=!1,n.initData(t.data,[],(function(t,e,i,r){if(t instanceof Array)return NaN;n.hasItemOption=!0;var o=t.value;return null!=o?o instanceof Array?o[r]:o:void 0})),n},e.prototype.formatTooltip=function(t,e,n){var i=this.getData(),r=i.getItemModel(t),o=r.get("name");if(o)return o;var a=r.get("fromName"),s=r.get("toName"),l=[];return null!=a&&l.push(a),null!=s&&l.push(s),oy("nameValue",{name:l.join(" > ")})},e.prototype.preventIncremental=function(){return!!this.get(["effect","show"])},e.prototype.getProgressive=function(){var t=this.option.progressive;return null==t?this.option.large?1e4:this.get("progressive"):t},e.prototype.getProgressiveThreshold=function(){var t=this.option.progressiveThreshold;return null==t?this.option.large?2e4:this.get("progressiveThreshold"):t},e.type="series.lines",e.dependencies=["grid","polar","geo","calendar"],e.defaultOption={coordinateSystem:"geo",zlevel:0,z:2,legendHoverLink:!0,xAxisIndex:0,yAxisIndex:0,symbol:["none","none"],symbolSize:[10,10],geoIndex:0,effect:{show:!1,period:4,constantSpeed:0,symbol:"circle",symbolSize:3,loop:!0,trailLength:.2},large:!1,largeThreshold:2e3,polyline:!1,clip:!0,label:{show:!1,position:"end"},lineStyle:{opacity:.5}},e}(Yy),Pz=Oz;function Ez(t){return t instanceof Array||(t=[t,t]),t}var Nz={seriesType:"lines",reset:function(t){var e=Ez(t.get("symbol")),n=Ez(t.get("symbolSize")),i=t.getData();function r(t,e){var n=t.getItemModel(e),i=Ez(n.getShallow("symbol",!0)),r=Ez(n.getShallow("symbolSize",!0));i[0]&&t.setItemVisual(e,"fromSymbol",i[0]),i[1]&&t.setItemVisual(e,"toSymbol",i[1]),r[0]&&t.setItemVisual(e,"fromSymbolSize",r[0]),r[1]&&t.setItemVisual(e,"toSymbolSize",r[1])}return i.setVisual("fromSymbol",e&&e[0]),i.setVisual("toSymbol",e&&e[1]),i.setVisual("fromSymbolSize",n&&n[0]),i.setVisual("toSymbolSize",n&&n[1]),{dataEach:i.hasItemOption?r:null}}},Rz=Nz;function Hz(t){t.registerChartView(Cz),t.registerSeriesModel(Pz),t.registerLayout(Tz),t.registerVisual(Rz)}var jz=256,zz=function(){function t(){this.blurSize=30,this.pointSize=20,this.maxOpacity=1,this.minOpacity=0,this._gradientPixels={inRange:null,outOfRange:null};var t=Y();this.canvas=t}return t.prototype.update=function(t,e,n,i,r,o){var a=this._getBrush(),s=this._getGradient(r,"inRange"),l=this._getGradient(r,"outOfRange"),u=this.pointSize+this.blurSize,c=this.canvas,d=c.getContext("2d"),h=t.length;c.width=e,c.height=n;for(var f=0;f<h;++f){var p=t[f],m=p[0],g=p[1],y=p[2],v=i(y);d.globalAlpha=v,d.drawImage(a,m-u,g-u)}if(!c.width||!c.height)return c;var _=d.getImageData(0,0,c.width,c.height),b=_.data,x=0,w=b.length,M=this.minOpacity,S=this.maxOpacity,L=S-M;while(x<w){v=b[x+3]/256;var k=4*Math.floor(v*(jz-1));if(v>0){var T=o(v)?s:l;v>0&&(v=v*L+M),b[x++]=T[k],b[x++]=T[k+1],b[x++]=T[k+2],b[x++]=T[k+3]*v*256}else x+=4}return d.putImageData(_,0,0),c},t.prototype._getBrush=function(){var t=this._brushCanvas||(this._brushCanvas=Y()),e=this.pointSize+this.blurSize,n=2*e;t.width=n,t.height=n;var i=t.getContext("2d");return i.clearRect(0,0,n,n),i.shadowOffsetX=n,i.shadowBlur=this.blurSize,i.shadowColor="#000",i.beginPath(),i.arc(-e,e,this.pointSize,0,2*Math.PI,!0),i.closePath(),i.fill(),t},t.prototype._getGradient=function(t,e){for(var n=this._gradientPixels,i=n[e]||(n[e]=new Uint8ClampedArray(1024)),r=[0,0,0,0],o=0,a=0;a<256;a++)t[e](a/255,!0,r),i[o++]=r[0],i[o++]=r[1],i[o++]=r[2],i[o++]=r[3];return i},t}(),Wz=zz;function Bz(t,e,n){var i=t[1]-t[0];e=H(e,(function(e){return{interval:[(e.interval[0]-t[0])/i,(e.interval[1]-t[0])/i]}}));var r=e.length,o=0;return function(t){var i;for(i=o;i<r;i++){var a=e[i].interval;if(a[0]<=t&&t<=a[1]){o=i;break}}if(i===r)for(i=o-1;i>=0;i--){a=e[i].interval;if(a[0]<=t&&t<=a[1]){o=i;break}}return i>=0&&i<r&&n[i]}}function Fz(t,e){var n=t[1]-t[0];return e=[(e[0]-t[0])/n,(e[1]-t[0])/n],function(t){return t>=e[0]&&t<=e[1]}}function Vz(t){var e=t.dimensions;return"lng"===e[0]&&"lat"===e[1]}var Gz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i;e.eachComponent("visualMap",(function(e){e.eachTargetSeries((function(n){n===t&&(i=e)}))})),this.group.removeAll(),this._incrementalDisplayable=null;var r=t.coordinateSystem;"cartesian2d"===r.type||"calendar"===r.type?this._renderOnCartesianAndCalendar(t,n,0,t.getData().count()):Vz(r)&&this._renderOnGeo(r,t,i,n)},e.prototype.incrementalPrepareRender=function(t,e,n){this.group.removeAll()},e.prototype.incrementalRender=function(t,e,n,i){var r=e.coordinateSystem;r&&(Vz(r)?this.render(e,n,i):this._renderOnCartesianAndCalendar(e,i,t.start,t.end,!0))},e.prototype._renderOnCartesianAndCalendar=function(t,e,n,i,r){var o,a,s,l,u=t.coordinateSystem;if(bk(u,"cartesian2d")){var c=u.getAxis("x"),d=u.getAxis("y");0,o=c.getBandWidth(),a=d.getBandWidth(),s=c.scale.getExtent(),l=d.scale.getExtent()}for(var h=this.group,f=t.getData(),p=t.getModel(["emphasis","itemStyle"]).getItemStyle(),m=t.getModel(["blur","itemStyle"]).getItemStyle(),g=t.getModel(["select","itemStyle"]).getItemStyle(),y=Sh(t),v=t.get(["emphasis","focus"]),_=t.get(["emphasis","blurScope"]),b=bk(u,"cartesian2d")?[f.mapDimension("x"),f.mapDimension("y"),f.mapDimension("value")]:[f.mapDimension("time"),f.mapDimension("value")],x=n;x<i;x++){var w=void 0,M=f.getItemVisual(x,"style");if(bk(u,"cartesian2d")){var S=f.get(b[0],x),L=f.get(b[1],x);if(isNaN(f.get(b[2],x))||S<s[0]||S>s[1]||L<l[0]||L>l[1])continue;var k=u.dataToPoint([S,L]);w=new Al({shape:{x:Math.floor(Math.round(k[0])-o/2),y:Math.floor(Math.round(k[1])-a/2),width:Math.ceil(o),height:Math.ceil(a)},style:M})}else{if(isNaN(f.get(b[1],x)))continue;w=new Al({z2:1,shape:u.dataToRect([f.get(b[0],x)]).contentShape,style:M})}var T=f.getItemModel(x);if(f.hasItemOption){var D=T.getModel("emphasis");p=D.getModel("itemStyle").getItemStyle(),m=T.getModel(["blur","itemStyle"]).getItemStyle(),g=T.getModel(["select","itemStyle"]).getItemStyle(),v=D.get("focus"),_=D.get("blurScope"),y=Sh(T)}var C=t.getRawValue(x),I="-";C&&null!=C[2]&&(I=C[2]+""),Mh(w,y,{labelFetcher:t,labelDataIndex:x,defaultOpacity:M.opacity,defaultText:I}),w.ensureState("emphasis").style=p,w.ensureState("blur").style=m,w.ensureState("select").style=g,$u(w,v,_),w.incremental=r,r&&(w.states.emphasis.hoverLayer=!0),h.add(w),f.setItemGraphicEl(x,w)}},e.prototype._renderOnGeo=function(t,e,n,i){var r=n.targetVisuals.inRange,o=n.targetVisuals.outOfRange,a=e.getData(),s=this._hmLayer||this._hmLayer||new Wz;s.blurSize=e.get("blurSize"),s.pointSize=e.get("pointSize"),s.minOpacity=e.get("minOpacity"),s.maxOpacity=e.get("maxOpacity");var l=t.getViewRect().clone(),u=t.getRoamTransform();l.applyTransform(u);var c=Math.max(l.x,0),d=Math.max(l.y,0),h=Math.min(l.width+l.x,i.getWidth()),f=Math.min(l.height+l.y,i.getHeight()),p=h-c,m=f-d,g=[a.mapDimension("lng"),a.mapDimension("lat"),a.mapDimension("value")],y=a.mapArray(g,(function(e,n,i){var r=t.dataToPoint([e,n]);return r[0]-=c,r[1]-=d,r.push(i),r})),v=n.getExtent(),_="visualMap.continuous"===n.type?Fz(v,n.option.range):Bz(v,n.getPieceList(),n.option.selected);s.update(y,p,m,r.color.getNormalizer(),{inRange:r.color.getColorMapper(),outOfRange:o.color.getColorMapper()},_);var b=new wl({style:{width:p,height:m,x:c,y:d,image:s.canvas},silent:!0});this.group.add(b)},e.type="heatmap",e}(Vy),$z=Gz,Uz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getInitialData=function(t,e){return BL(this.getSource(),this,{generateCoord:"value"})},e.prototype.preventIncremental=function(){var t=vm.get(this.get("coordinateSystem"));if(t&&t.dimensions)return"lng"===t.dimensions[0]&&"lat"===t.dimensions[1]},e.type="series.heatmap",e.dependencies=["grid","geo","calendar"],e.defaultOption={coordinateSystem:"cartesian2d",zlevel:0,z:2,geoIndex:0,blurSize:30,pointSize:20,maxOpacity:1,minOpacity:0,select:{itemStyle:{borderColor:"#212121"}}},e}(Yy),qz=Uz;function Xz(t){t.registerChartView($z),t.registerSeriesModel(qz)}var Zz=["itemStyle","borderWidth"],Jz=[{xy:"x",wh:"width",index:0,posDesc:["left","right"]},{xy:"y",wh:"height",index:1,posDesc:["top","bottom"]}],Kz=new Dc,Qz=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=this.group,r=t.getData(),o=this._data,a=t.coordinateSystem,s=a.getBaseAxis(),l=s.isHorizontal(),u=a.master.getRect(),c={ecSize:{width:n.getWidth(),height:n.getHeight()},seriesModel:t,coordSys:a,coordSysExtent:[[u.x,u.x+u.width],[u.y,u.y+u.height]],isHorizontal:l,valueDim:Jz[+l],categoryDim:Jz[1-+l]};return r.diff(o).add((function(t){if(r.hasValue(t)){var e=dW(r,t),n=tW(r,t,e,c),o=pW(r,c,n);r.setItemGraphicEl(t,o),i.add(o),bW(o,c,n)}})).update((function(t,e){var n=o.getItemGraphicEl(e);if(r.hasValue(t)){var a=dW(r,t),s=tW(r,t,a,c),l=yW(r,s);n&&l!==n.__pictorialShapeStr&&(i.remove(n),r.setItemGraphicEl(t,null),n=null),n?mW(n,c,s):n=pW(r,c,s,!0),r.setItemGraphicEl(t,n),n.__pictorialSymbolMeta=s,i.add(n),bW(n,c,s)}else i.remove(n)})).remove((function(t){var e=o.getItemGraphicEl(t);e&&gW(o,t,e.__pictorialSymbolMeta.animationModel,e)})).execute(),this._data=r,this.group},e.prototype.remove=function(t,e){var n=this.group,i=this._data;t.get("animation")?i&&i.eachItemGraphicEl((function(e){gW(i,$l(e).dataIndex,t,e)})):n.removeAll()},e.type="pictorialBar",e}(Vy);function tW(t,e,n,i){var r=t.getItemLayout(e),o=n.get("symbolRepeat"),a=n.get("symbolClip"),s=n.get("symbolPosition")||"start",l=n.get("symbolRotate"),u=(l||0)*Math.PI/180||0,c=n.get("symbolPatternSize")||2,d=n.isAnimationEnabled(),h={dataIndex:e,layout:r,itemModel:n,symbolType:t.getItemVisual(e,"symbol")||"circle",style:t.getItemVisual(e,"style"),symbolClip:a,symbolRepeat:o,symbolRepeatDirection:n.get("symbolRepeatDirection"),symbolPatternSize:c,rotation:u,animationModel:d?n:null,hoverScale:d&&n.get(["emphasis","scale"]),z2:n.getShallow("z",!0)||0};eW(n,o,r,i,h),iW(t,e,r,o,a,h.boundingLength,h.pxSign,c,i,h),rW(n,h.symbolScale,u,i,h);var f=h.symbolSize,p=n.get("symbolOffset");return $(p)&&(p=[Nr(p[0],f[0]),Nr(p[1],f[1])]),oW(n,f,r,o,a,p,s,h.valueLineWidth,h.boundingLength,h.repeatCutLength,i,h),h}function eW(t,e,n,i,r){var o,a=i.valueDim,s=t.get("symbolBoundingData"),l=i.coordSys.getOtherAxis(i.coordSys.getBaseAxis()),u=l.toGlobalCoord(l.dataToCoord(0)),c=1-+(n[a.wh]<=0);if($(s)){var d=[nW(l,s[0])-u,nW(l,s[1])-u];d[1]<d[0]&&d.reverse(),o=d[c]}else o=null!=s?nW(l,s)-u:e?i.coordSysExtent[a.index][c]-u:n[a.wh];r.boundingLength=o,e&&(r.repeatCutLength=n[a.wh]),r.pxSign=o>0?1:o<0?-1:0}function nW(t,e){return t.toGlobalCoord(t.dataToCoord(t.scale.parse(e)))}function iW(t,e,n,i,r,o,a,s,l,u){var c,d=l.valueDim,h=l.categoryDim,f=Math.abs(n[h.wh]),p=t.getItemVisual(e,"symbolSize");c=$(p)?p.slice():null==p?["100%","100%"]:[p,p],c[h.index]=Nr(c[h.index],f),c[d.index]=Nr(c[d.index],i?f:Math.abs(o)),u.symbolSize=c;var m=u.symbolScale=[c[0]/s,c[1]/s];m[d.index]*=(l.isHorizontal?-1:1)*a}function rW(t,e,n,i,r){var o=t.get(Zz)||0;o&&(Kz.attr({scaleX:e[0],scaleY:e[1],rotation:n}),Kz.updateTransform(),o/=Kz.getLineScale(),o*=e[i.valueDim.index]),r.valueLineWidth=o}function oW(t,e,n,i,r,o,a,s,l,u,c,d){var h=c.categoryDim,f=c.valueDim,p=d.pxSign,m=Math.max(e[f.index]+s,0),g=m;if(i){var y=Math.abs(l),v=ot(t.get("symbolMargin"),"15%")+"",_=!1;v.lastIndexOf("!")===v.length-1&&(_=!0,v=v.slice(0,v.length-1));var b=Nr(v,e[f.index]),x=Math.max(m+2*b,0),w=_?0:2*b,M=eo(i),S=M?i:xW((y+w)/x),L=y-S*m;b=L/2/(_?S:S-1),x=m+2*b,w=_?0:2*b,M||"fixed"===i||(S=u?xW((Math.abs(u)+w)/x):0),g=S*x-w,d.repeatTimes=S,d.symbolMargin=b}var k=p*(g/2),T=d.pathPosition=[];T[h.index]=n[h.wh]/2,T[f.index]="start"===a?k:"end"===a?l-k:l/2,o&&(T[0]+=o[0],T[1]+=o[1]);var D=d.bundlePosition=[];D[h.index]=n[h.xy],D[f.index]=n[f.xy];var C=d.barRectShape=I({},n);C[f.wh]=p*Math.max(Math.abs(n[f.wh]),Math.abs(T[f.index]+k)),C[h.wh]=n[h.wh];var A=d.clipShape={};A[h.xy]=-n[h.xy],A[h.wh]=c.ecSize[h.wh],A[f.xy]=0,A[f.wh]=n[f.wh]}function aW(t){var e=t.symbolPatternSize,n=j_(t.symbolType,-e/2,-e/2,e,e);return n.attr({culling:!0}),"image"!==n.type&&n.setStyle({strokeNoScale:!0}),n}function sW(t,e,n,i){var r=t.__pictorialBundle,o=n.symbolSize,a=n.valueLineWidth,s=n.pathPosition,l=e.valueDim,u=n.repeatTimes||0,c=0,d=o[e.valueDim.index]+a+2*n.symbolMargin;for(vW(t,(function(t){t.__pictorialAnimationIndex=c,t.__pictorialRepeatTimes=u,c<u?_W(t,null,p(c),n,i):_W(t,null,{scaleX:0,scaleY:0},n,i,(function(){r.remove(t)})),c++}));c<u;c++){var h=aW(n);h.__pictorialAnimationIndex=c,h.__pictorialRepeatTimes=u,r.add(h);var f=p(c);_W(h,{x:f.x,y:f.y,scaleX:0,scaleY:0},{scaleX:f.scaleX,scaleY:f.scaleY,rotation:f.rotation},n,i)}function p(t){var e=s.slice(),i=n.pxSign,r=t;return("start"===n.symbolRepeatDirection?i>0:i<0)&&(r=u-1-t),e[l.index]=d*(r-u/2+.5)+s[l.index],{x:e[0],y:e[1],scaleX:n.symbolScale[0],scaleY:n.symbolScale[1],rotation:n.rotation}}}function lW(t,e,n,i){var r=t.__pictorialBundle,o=t.__pictorialMainPath;o?_W(o,null,{x:n.pathPosition[0],y:n.pathPosition[1],scaleX:n.symbolScale[0],scaleY:n.symbolScale[1],rotation:n.rotation},n,i):(o=t.__pictorialMainPath=aW(n),r.add(o),_W(o,{x:n.pathPosition[0],y:n.pathPosition[1],scaleX:0,scaleY:0,rotation:n.rotation},{scaleX:n.symbolScale[0],scaleY:n.symbolScale[1]},n,i))}function uW(t,e,n){var i=I({},e.barRectShape),r=t.__pictorialBarRect;r?_W(r,null,{shape:i},e,n):(r=t.__pictorialBarRect=new Al({z2:2,shape:i,silent:!0,style:{stroke:"transparent",fill:"transparent",lineWidth:0}}),t.add(r))}function cW(t,e,n,r){if(n.symbolClip){var o=t.__pictorialClipPath,a=I({},n.clipShape),s=e.valueDim,l=n.animationModel,u=n.dataIndex;if(o)th(o,{shape:a},l,u);else{a[s.wh]=0,o=new Al({shape:a}),t.__pictorialBundle.setClipPath(o),t.__pictorialClipPath=o;var c={};c[s.wh]=n.clipShape[s.wh],i[r?"updateProps":"initProps"](o,{shape:c},l,u)}}}function dW(t,e){var n=t.getItemModel(e);return n.getAnimationDelayParams=hW,n.isAnimationEnabled=fW,n}function hW(t){return{index:t.__pictorialAnimationIndex,count:t.__pictorialRepeatTimes}}function fW(){return this.parentModel.isAnimationEnabled()&&!!this.getShallow("animation")}function pW(t,e,n,i){var r=new Mr,o=new Mr;return r.add(o),r.__pictorialBundle=o,o.x=n.bundlePosition[0],o.y=n.bundlePosition[1],n.symbolRepeat?sW(r,e,n):lW(r,e,n),uW(r,n,i),cW(r,e,n,i),r.__pictorialShapeStr=yW(t,n),r.__pictorialSymbolMeta=n,r}function mW(t,e,n){var i=n.animationModel,r=n.dataIndex,o=t.__pictorialBundle;th(o,{x:n.bundlePosition[0],y:n.bundlePosition[1]},i,r),n.symbolRepeat?sW(t,e,n,!0):lW(t,e,n,!0),uW(t,n,!0),cW(t,e,n,!0)}function gW(t,e,n,i){var r=i.__pictorialBarRect;r&&r.removeTextContent();var o=[];vW(i,(function(t){o.push(t)})),i.__pictorialMainPath&&o.push(i.__pictorialMainPath),i.__pictorialClipPath&&(n=null),R(o,(function(t){nh(t,{scaleX:0,scaleY:0},n,e,(function(){i.parent&&i.parent.remove(i)}))})),t.setItemGraphicEl(e,null)}function yW(t,e){return[t.getItemVisual(e.dataIndex,"symbol")||"none",!!e.symbolRepeat,!!e.symbolClip].join(":")}function vW(t,e,n){R(t.__pictorialBundle.children(),(function(i){i!==t.__pictorialBarRect&&e.call(n,i)}))}function _W(t,e,n,r,o,a){e&&t.attr(e),r.symbolClip&&!o?n&&t.attr(n):n&&i[o?"updateProps":"initProps"](t,n,r.animationModel,r.dataIndex,a)}function bW(t,e,n){var i=n.dataIndex,r=n.itemModel,o=r.getModel("emphasis"),a=o.getModel("itemStyle").getItemStyle(),s=r.getModel(["blur","itemStyle"]).getItemStyle(),l=r.getModel(["select","itemStyle"]).getItemStyle(),u=r.getShallow("cursor"),c=o.get("focus"),d=o.get("blurScope"),h=o.get("scale");vW(t,(function(t){if(t instanceof wl){var e=t.style;t.useStyle(I({image:e.image,x:e.x,y:e.y,width:e.width,height:e.height},n.style))}else t.useStyle(n.style);var i=t.ensureState("emphasis");i.style=a,h&&(i.scaleX=1.1*t.scaleX,i.scaleY=1.1*t.scaleY),t.ensureState("blur").style=s,t.ensureState("select").style=l,u&&(t.cursor=u),t.z2=n.z2}));var f=e.valueDim.posDesc[+(n.boundingLength>0)],p=t.__pictorialBarRect;Mh(p,Sh(r),{labelFetcher:e.seriesModel,labelDataIndex:i,defaultText:GL(e.seriesModel.getData(),i),inheritColor:n.style.fill,defaultOpacity:n.style.opacity,defaultOutsidePosition:f}),$u(t,c,d)}function xW(t){var e=Math.round(t);return Math.abs(t-e)<1e-4?e:Math.ceil(t)}var wW=Qz,MW=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.hasSymbolVisual=!0,n.defaultSymbol="roundRect",n}return o(e,t),e.prototype.getInitialData=function(e){return e.stack=null,t.prototype.getInitialData.apply(this,arguments)},e.type="series.pictorialBar",e.dependencies=["grid"],e.defaultOption=Qh(aT.defaultOption,{symbol:"circle",symbolSize:null,symbolRotate:null,symbolPosition:null,symbolOffset:null,symbolMargin:null,symbolRepeat:!1,symbolRepeatDirection:"end",symbolClip:!1,symbolBoundingData:null,symbolPatternSize:400,barGap:"-100%",progressive:0,emphasis:{scale:!1},select:{itemStyle:{borderColor:"#212121"}}}),e}(aT),SW=MW;function LW(t){t.registerChartView(wW),t.registerSeriesModel(SW),t.registerLayout(G(tT,"pictorialBar"))}var kW=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._layers=[],n}return o(e,t),e.prototype.render=function(t,e,n){var i=t.getData(),r=this,o=this.group,a=t.getLayerSeries(),s=i.getLayout("layoutInfo"),l=s.rect,u=s.boundaryGap;function c(t){return t.name}o.x=0,o.y=l.y+u[0];var d=new $S(this._layersSeries||[],a,c,c),h=[];function f(e,n,s){var l=r._layers;if("remove"!==e){for(var u,c,d=[],f=[],p=a[n].indices,m=0;m<p.length;m++){var g=i.getItemLayout(p[m]),y=g.x,v=g.y0,_=g.y;d.push(y,v),f.push(y,v+_),u=i.getItemVisual(p[m],"style")}var b=i.getItemLayout(p[0]),x=t.getModel("label"),w=x.get("margin"),M=t.getModel("emphasis");if("add"===e){var S=h[n]=new Mr;c=new gk({shape:{points:d,stackedOnPoints:f,smooth:.4,stackedOnSmooth:.4,smoothConstraint:!1},z2:0}),S.add(c),o.add(S),t.isAnimationEnabled()&&c.setClipPath(TW(c.getBoundingRect(),t,(function(){c.removeClipPath()})))}else{S=l[s];c=S.childAt(0),o.add(S),h[n]=S,th(c,{shape:{points:d,stackedOnPoints:f}},t)}Mh(c,Sh(t),{labelDataIndex:p[m-1],defaultText:i.getName(p[m-1]),inheritColor:u.fill},{normal:{verticalAlign:"middle"}}),c.setTextConfig({position:null,local:!0});var L=c.getTextContent();L&&(L.x=b.x-w,L.y=b.y0+b.y/2),c.useStyle(u),i.setItemGraphicEl(n,c),Zu(c,t),$u(c,M.get("focus"),M.get("blurScope"))}else o.remove(l[n])}d.add(V(f,this,"add")).update(V(f,this,"update")).remove(V(f,this,"remove")).execute(),this._layersSeries=a,this._layers=h},e.type="themeRiver",e}(Vy);function TW(t,e,n){var i=new Al({shape:{x:t.x-10,y:t.y-10,width:0,height:t.height+20}});return eh(i,{shape:{x:t.x-50,width:t.width+100,height:t.height+20}},e,n),i}var DW=kW,CW=2,IW=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.useColorPaletteOnData=!0,n}return o(e,t),e.prototype.init=function(e){t.prototype.init.apply(this,arguments),this.legendVisualProvider=new KT(V(this.getData,this),V(this.getRawData,this))},e.prototype.fixData=function(t){var e=t.length,n={},i=zo(t,(function(t){return n.hasOwnProperty(t[0]+"")||(n[t[0]+""]=-1),t[2]})),r=[];i.buckets.each((function(t,e){r.push({name:e,dataList:t})}));for(var o=r.length,a=0;a<o;++a){for(var s=r[a].name,l=0;l<r[a].dataList.length;++l){var u=r[a].dataList[l][0]+"";n[u]=a}for(var u in n)n.hasOwnProperty(u)&&n[u]!==a&&(n[u]=a,t[e]=[u,0,s],e++)}return t},e.prototype.getInitialData=function(t,e){for(var n=this.getReferringComponents("singleAxis",Po).models[0],i=n.get("type"),r=z(t.data,(function(t){return void 0!==t[2]})),o=this.fixData(r||[]),a=[],s=this.nameMap=gt(),l=0,u=0;u<o.length;++u)a.push(o[u][CW]),s.get(o[u][CW])||(s.set(o[u][CW],l),l++);var c=AL(o,{coordDimensions:["single"],dimensionsDefine:[{name:"time",type:XS(i)},{name:"value",type:"float"},{name:"name",type:"ordinal"}],encodeDefine:{single:0,value:1,itemName:2}}),d=new kL(c,this);return d.initData(o),d},e.prototype.getLayerSeries=function(){for(var t=this.getData(),e=t.count(),n=[],i=0;i<e;++i)n[i]=i;var r=t.mapDimension("single"),o=zo(n,(function(e){return t.get("name",e)})),a=[];return o.buckets.each((function(e,n){e.sort((function(e,n){return t.get(r,e)-t.get(r,n)})),a.push({name:n,indices:e})})),a},e.prototype.getAxisTooltipData=function(t,e,n){$(t)||(t=t?[t]:[]);for(var i,r=this.getData(),o=this.getLayerSeries(),a=[],s=o.length,l=0;l<s;++l){for(var u=Number.MAX_VALUE,c=-1,d=o[l].indices.length,h=0;h<d;++h){var f=r.get(t[0],o[l].indices[h]),p=Math.abs(f-e);p<=u&&(i=f,u=p,c=o[l].indices[h])}a.push(c)}return{dataIndices:a,nestestValue:i}},e.prototype.formatTooltip=function(t,e,n){var i=this.getData(),r=i.getName(t),o=i.get(i.mapDimension("value"),t);return oy("nameValue",{name:r,value:o})},e.type="series.themeRiver",e.dependencies=["singleAxis"],e.defaultOption={zlevel:0,z:2,coordinateSystem:"singleAxis",boundaryGap:["10%","10%"],singleAxisIndex:0,animationEasing:"linear",label:{margin:4,show:!0,position:"left",fontSize:11},emphasis:{label:{show:!0}}},e}(Yy),AW=IW;function YW(t,e){t.eachSeriesByType("themeRiver",(function(t){var e=t.getData(),n=t.coordinateSystem,i={},r=n.getRect();i.rect=r;var o=t.get("boundaryGap"),a=n.getAxis();if(i.boundaryGap=o,"horizontal"===a.orient){o[0]=Nr(o[0],r.height),o[1]=Nr(o[1],r.height);var s=r.height-o[0]-o[1];OW(e,t,s)}else{o[0]=Nr(o[0],r.width),o[1]=Nr(o[1],r.width);var l=r.width-o[0]-o[1];OW(e,t,l)}e.setLayout("layoutInfo",i)}))}function OW(t,e,n){if(t.count())for(var i,r=e.coordinateSystem,o=e.getLayerSeries(),a=t.mapDimension("single"),s=t.mapDimension("value"),l=H(o,(function(e){return H(e.indices,(function(e){var n=r.dataToPoint(t.get(a,e));return n[1]=t.get(s,e),n}))})),u=PW(l),c=u.y0,d=n/u.max,h=o.length,f=o[0].indices.length,p=0;p<f;++p){i=c[p]*d,t.setItemLayout(o[0].indices[p],{layerIndex:0,x:l[0][p][0],y0:i,y:l[0][p][1]*d});for(var m=1;m<h;++m)i+=l[m-1][p][1]*d,t.setItemLayout(o[m].indices[p],{layerIndex:m,x:l[m][p][0],y0:i,y:l[m][p][1]*d})}}function PW(t){for(var e=t.length,n=t[0].length,i=[],r=[],o=0,a=0;a<n;++a){for(var s=0,l=0;l<e;++l)s+=t[l][a][1];s>o&&(o=s),i.push(s)}for(var u=0;u<n;++u)r[u]=(o-i[u])/2;o=0;for(var c=0;c<n;++c){var d=i[c]+r[c];d>o&&(o=d)}return{y0:r,max:o}}function EW(t){t.registerChartView(DW),t.registerSeriesModel(AW),t.registerLayout(YW),t.registerProcessor(zT("themeRiver"))}var NW=2,RW=4,HW=function(t){function e(e,n,i,r){var o=t.call(this)||this;o.z2=NW,o.textConfig={inside:!0},$l(o).seriesIndex=n.seriesIndex;var a=new Gl({z2:RW,silent:e.getModel().get(["label","silent"])});return o.setTextContent(a),o.updateData(!0,e,n,i,r),o}return o(e,t),e.prototype.updateData=function(t,e,n,i,r){this.node=e,e.piece=this,n=n||this._seriesModel,i=i||this._ecModel;var o=this;$l(o).dataIndex=e.dataIndex;var a=e.getModel(),s=a.getModel("emphasis"),l=e.getLayout(),u=I({},l);u.label=null;var c=e.getVisual("style");c.lineJoin="bevel";var d=e.getVisual("decal");d&&(c.decal=wb(d,r));var h=$T(a.getModel("itemStyle"),u);I(u,h),R(Ql,(function(t){var e=o.ensureState(t),n=a.getModel([t,"itemStyle"]);e.style=n.getItemStyle();var i=$T(n,u);i&&(e.shape=i)})),t?(o.setShape(u),o.shape.r=l.r0,th(o,{shape:{r:l.r}},n,e.dataIndex)):th(o,{shape:u},n),o.useStyle(c),this._updateLabel(n);var f=a.getShallow("cursor");f&&o.attr("cursor",f),this._seriesModel=n||this._seriesModel,this._ecModel=i||this._ecModel;var p=s.get("focus"),m="ancestor"===p?e.getAncestorsIndices():"descendant"===p?e.getDescendantIndices():p;$u(this,m,s.get("blurScope"))},e.prototype._updateLabel=function(t){var e=this,n=this.node.getModel(),i=n.getModel("label"),r=this.node.getLayout(),o=r.endAngle-r.startAngle,a=(r.startAngle+r.endAngle)/2,s=Math.cos(a),l=Math.sin(a),u=this,c=u.getTextContent(),d=this.node.dataIndex,h=i.get("minAngle")/180*Math.PI,f=i.get("show")&&!(null!=h&&Math.abs(o)<h);function p(t,e){var n=t.get(e);return null==n?i.get(e):n}c.ignore=!f,R(tu,(function(i){var o="normal"===i?n.getModel("label"):n.getModel([i,"label"]),h="normal"===i,f=h?c:c.ensureState(i),m=t.getFormattedLabel(d,i);h&&(m=m||e.node.name),f.style=Lh(o,{},null,"normal"!==i,!0),m&&(f.style.text=m);var g=o.get("show");null==g||h||(f.ignore=!g);var y,v=p(o,"position"),_=h?u:u.states[i],b=_.style.fill;_.textConfig={outsideFill:"inherit"===o.get("color")?b:null,inside:"outside"!==v};var x=p(o,"distance")||0,w=p(o,"align");"outside"===v?(y=r.r+x,w=a>Math.PI/2?"right":"left"):w&&"center"!==w?"left"===w?(y=r.r0+x,a>Math.PI/2&&(w="right")):"right"===w&&(y=r.r-x,a>Math.PI/2&&(w="left")):(y=(r.r+r.r0)/2,w="center"),f.style.align=w,f.style.verticalAlign=p(o,"verticalAlign")||"middle",f.x=y*s+r.cx,f.y=y*l+r.cy;var M=p(o,"rotate"),S=0;"radial"===M?(S=-a,S<-Math.PI/2&&(S+=Math.PI)):"tangential"===M?(S=Math.PI/2-a,S>Math.PI/2?S-=Math.PI:S<-Math.PI/2&&(S+=Math.PI)):"number"===typeof M&&(S=M*Math.PI/180),f.rotation=S})),c.dirtyStyle()},e}(qc),jW=HW,zW="sunburstRootToNode",WW="sunburstHighlight",BW="sunburstUnhighlight";function FW(t){t.registerAction({type:zW,update:"updateView"},(function(t,e){function n(e,n){var i=iO(t,[zW],e);if(i){var r=e.getViewRoot();r&&(t.direction=oO(r,i.node)?"rollUp":"drillDown"),e.resetViewRoot(i.node)}}e.eachComponent({mainType:"series",subType:"sunburst",query:t},n)})),t.registerAction({type:WW,update:"none"},(function(t,e,n){function i(e){var n=iO(t,[WW],e);n&&(t.dataIndex=n.node.dataIndex)}t=I({},t),e.eachComponent({mainType:"series",subType:"sunburst",query:t},i),n.dispatchAction(I(t,{type:"highlight"}))})),t.registerAction({type:BW,update:"updateView"},(function(t,e,n){t=I({},t),n.dispatchAction(I(t,{type:"downplay"}))}))}var VW=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,i){var r=this;this.seriesModel=t,this.api=n,this.ecModel=e;var o=t.getData(),a=o.tree.root,s=t.getViewRoot(),l=this.group,u=t.get("renderLabelForZeroData"),c=[];s.eachNode((function(t){c.push(t)}));var d=this._oldChildren||[];function h(t,e){function n(t){return t.getId()}function i(n,i){var r=null==n?null:t[n],o=null==i?null:e[i];f(r,o)}0===t.length&&0===e.length||new $S(e,t,n,n).add(i).update(i).remove(G(i,null)).execute()}function f(i,r){if(u||!i||i.getValue()||(i=null),i!==a&&r!==a)if(r&&r.piece)i?(r.piece.updateData(!1,i,t,e,n),o.setItemGraphicEl(i.dataIndex,r.piece)):p(r);else if(i){var s=new jW(i,t,e,n);l.add(s),o.setItemGraphicEl(i.dataIndex,s)}}function p(t){t&&t.piece&&(l.remove(t.piece),t.piece=null)}function m(i,o){o.depth>0?(r.virtualPiece?r.virtualPiece.updateData(!1,i,t,e,n):(r.virtualPiece=new jW(i,t,e,n),l.add(r.virtualPiece)),o.piece.off("click"),r.virtualPiece.on("click",(function(t){r._rootToNode(o.parentNode)}))):r.virtualPiece&&(l.remove(r.virtualPiece),r.virtualPiece=null)}h(c,d),m(a,s),this._initEvents(),this._oldChildren=c},e.prototype._initEvents=function(){var t=this;this.group.off("click"),this.group.on("click",(function(e){var n=!1,i=t.seriesModel.getViewRoot();i.eachNode((function(i){if(!n&&i.piece&&i.piece===e.target){var r=i.getModel().get("nodeClick");if("rootToNode"===r)t._rootToNode(i);else if("link"===r){var o=i.getModel(),a=o.get("link");if(a){var s=o.get("target",!0)||"_blank";rp(a,s)}}n=!0}}))}))},e.prototype._rootToNode=function(t){t!==this.seriesModel.getViewRoot()&&this.api.dispatchAction({type:zW,from:this.uid,seriesId:this.seriesModel.id,targetNode:t})},e.prototype.containPoint=function(t,e){var n=e.getData(),i=n.getItemLayout(0);if(i){var r=t[0]-i.cx,o=t[1]-i.cy,a=Math.sqrt(r*r+o*o);return a<=i.r&&a>=i.r0}},e.type="sunburst",e}(Vy),GW=VW,$W=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.ignoreStyleOnData=!0,n}return o(e,t),e.prototype.getInitialData=function(t,e){var n={name:t.name,children:t.data};UW(n);var i=H(t.levels||[],(function(t){return new qh(t,this,e)}),this),r=nO.createTree(n,this,o);function o(t){t.wrapMethod("getItemModel",(function(t,e){var n=r.getNodeByDataIndex(e),o=i[n.depth];return o&&(t.parentModel=o),t}))}return r.data},e.prototype.optionUpdated=function(){this.resetViewRoot()},e.prototype.getDataParams=function(e){var n=t.prototype.getDataParams.apply(this,arguments),i=this.getData().tree.getNodeByDataIndex(e);return n.treePathInfo=aO(i,this),n},e.prototype.getViewRoot=function(){return this._viewRoot},e.prototype.resetViewRoot=function(t){t?this._viewRoot=t:t=this._viewRoot;var e=this.getRawData().tree.root;t&&(t===e||e.contains(t))||(this._viewRoot=e)},e.prototype.enableAriaDecal=function(){_O(this)},e.type="series.sunburst",e.defaultOption={zlevel:0,z:2,center:["50%","50%"],radius:[0,"75%"],clockwise:!0,startAngle:90,minAngle:0,stillShowZeroSum:!0,nodeClick:"rootToNode",renderLabelForZeroData:!1,label:{rotate:"radial",show:!0,opacity:1,align:"center",position:"inside",distance:5,silent:!0},itemStyle:{borderWidth:1,borderColor:"white",borderType:"solid",shadowBlur:0,shadowColor:"rgba(0, 0, 0, 0.2)",shadowOffsetX:0,shadowOffsetY:0,opacity:1},emphasis:{focus:"descendant"},blur:{itemStyle:{opacity:.2},label:{opacity:.1}},animationType:"expansion",animationDuration:1e3,animationDurationUpdate:500,data:[],levels:[],sort:"desc"},e}(Yy);function UW(t){var e=0;R(t.children,(function(t){UW(t);var n=t.value;$(n)&&(n=n[0]),e+=n}));var n=t.value;$(n)&&(n=n[0]),(null==n||isNaN(n))&&(n=e),n<0&&(n=0),$(t.value)?t.value[0]=n:t.value=n}var qW=$W,XW=Math.PI/180;function ZW(t,e,n){e.eachSeriesByType(t,(function(t){var e=t.get("center"),i=t.get("radius");$(i)||(i=[0,i]),$(e)||(e=[e,e]);var r=n.getWidth(),o=n.getHeight(),a=Math.min(r,o),s=Nr(e[0],r),l=Nr(e[1],o),u=Nr(i[0],a/2),c=Nr(i[1],a/2),d=-t.get("startAngle")*XW,h=t.get("minAngle")*XW,f=t.getData().tree.root,p=t.getViewRoot(),m=p.depth,g=t.get("sort");null!=g&&JW(p,g);var y=0;R(p.children,(function(t){!isNaN(t.getValue())&&y++}));var v=p.getValue(),_=Math.PI/(v||y)*2,b=p.depth>0,x=p.height-(b?-1:1),w=(c-u)/(x||1),M=t.get("clockwise"),S=t.get("stillShowZeroSum"),L=M?1:-1,k=function(t,e){if(t){var n=e;if(t!==f){var i=t.getValue(),r=0===v&&S?_:i*_;r<h&&(r=h),n=e+L*r;var o=t.depth-m-(b?-1:1),c=u+w*o,d=u+w*(o+1),p=t.getModel();null!=p.get("r0")&&(c=Nr(p.get("r0"),a/2)),null!=p.get("r")&&(d=Nr(p.get("r"),a/2)),t.setLayout({angle:r,startAngle:e,endAngle:n,clockwise:M,cx:s,cy:l,r0:c,r:d})}if(t.children&&t.children.length){var g=0;R(t.children,(function(t){g+=k(t,e+g)}))}return n-e}};if(b){var T=u,D=u+w,C=2*Math.PI;f.setLayout({angle:C,startAngle:d,endAngle:d+C,clockwise:M,cx:s,cy:l,r0:T,r:D})}k(p,d)}))}function JW(t,e){var n=t.children||[];t.children=KW(n,e),n.length&&R(t.children,(function(t){JW(t,e)}))}function KW(t,e){if("function"===typeof e){var n=H(t,(function(t,e){var n=t.getValue();return{params:{depth:t.depth,height:t.height,dataIndex:t.dataIndex,getValue:function(){return n}},index:e}}));return n.sort((function(t,n){return e(t.params,n.params)})),H(n,(function(e){return t[e.index]}))}var i="asc"===e;return t.sort((function(t,e){var n=(t.getValue()-e.getValue())*(i?1:-1);return 0===n?(t.dataIndex-e.dataIndex)*(i?-1:1):n}))}function QW(t){var e={};function n(t,n,i){var r=t;while(r&&r.depth>1)r=r.parentNode;var o=n.getColorFromPalette(r.name||r.dataIndex+"",e);return t.depth>1&&"string"===typeof o&&(o=bn(o,(t.depth-1)/(i-1)*.5)),o}t.eachSeriesByType("sunburst",(function(t){var e=t.getData(),i=e.tree;i.eachNode((function(r){var o=r.getModel(),a=o.getModel("itemStyle").getItemStyle();a.fill||(a.fill=n(r,t,i.root.height));var s=e.ensureUniqueItemVisual(r.dataIndex,"style");I(s,a)}))}))}function tB(t){t.registerChartView(GW),t.registerSeriesModel(qW),t.registerLayout(G(ZW,"sunburst")),t.registerProcessor(G(zT,"sunburst")),t.registerVisual(QW),FW(t)}function eB(t,e){return e=e||[0,0],H(["x","y"],(function(n,i){var r=this.getAxis(n),o=e[i],a=t[i]/2;return"category"===r.type?r.getBandWidth():Math.abs(r.dataToCoord(o-a)-r.dataToCoord(o+a))}),this)}function nB(t){var e=t.master.getRect();return{coordSys:{type:"cartesian2d",x:e.x,y:e.y,width:e.width,height:e.height},api:{coord:function(e){return t.dataToPoint(e)},size:V(eB,t)}}}function iB(t,e){return e=e||[0,0],H([0,1],(function(n){var i=e[n],r=t[n]/2,o=[],a=[];return o[n]=i-r,a[n]=i+r,o[1-n]=a[1-n]=e[1-n],Math.abs(this.dataToPoint(o)[n]-this.dataToPoint(a)[n])}),this)}function rB(t){var e=t.getBoundingRect();return{coordSys:{type:"geo",x:e.x,y:e.y,width:e.width,height:e.height,zoom:t.getZoom()},api:{coord:function(e){return t.dataToPoint(e)},size:V(iB,t)}}}function oB(t,e){var n=this.getAxis(),i=e instanceof Array?e[0]:e,r=(t instanceof Array?t[0]:t)/2;return"category"===n.type?n.getBandWidth():Math.abs(n.dataToCoord(i-r)-n.dataToCoord(i+r))}function aB(t){var e=t.getRect();return{coordSys:{type:"singleAxis",x:e.x,y:e.y,width:e.width,height:e.height},api:{coord:function(e){return t.dataToPoint(e)},size:V(oB,t)}}}function sB(t,e){return e=e||[0,0],H(["Radius","Angle"],(function(n,i){var r="get"+n+"Axis",o=this[r](),a=e[i],s=t[i]/2,l="category"===o.type?o.getBandWidth():Math.abs(o.dataToCoord(a-s)-o.dataToCoord(a+s));return"Angle"===n&&(l=l*Math.PI/180),l}),this)}function lB(t){var e=t.getRadiusAxis(),n=t.getAngleAxis(),i=e.getExtent();return i[0]>i[1]&&i.reverse(),{coordSys:{type:"polar",cx:t.cx,cy:t.cy,r:i[1],r0:i[0]},api:{coord:function(i){var r=e.dataToRadius(i[0]),o=n.dataToAngle(i[1]),a=t.coordToPoint([r,o]);return a.push(r,o*Math.PI/180),a},size:V(sB,t)}}}function uB(t){var e=t.getRect(),n=t.getRangeInfo();return{coordSys:{type:"calendar",x:e.x,y:e.y,width:e.width,height:e.height,cellWidth:t.getCellWidth(),cellHeight:t.getCellHeight(),rangeInfo:{start:n.start,end:n.end,weeks:n.weeks,dayCount:n.allDay}},api:{coord:function(e,n){return t.dataToPoint(e,n)}}}}function cB(t,e,n,i){return t&&(t.legacy||!1!==t.legacy&&!n&&!i&&"tspan"!==e&&("text"===e||_t(t,"text")))}function dB(t,e,n){var i,r,o,a=t;if("text"===e)o=a;else{o={},_t(a,"text")&&(o.text=a.text),_t(a,"rich")&&(o.rich=a.rich),_t(a,"textFill")&&(o.fill=a.textFill),_t(a,"textStroke")&&(o.stroke=a.textStroke),r={type:"text",style:o,silent:!0},i={};var s=_t(a,"textPosition");n?i.position=s?a.textPosition:"inside":s&&(i.position=a.textPosition),_t(a,"textPosition")&&(i.position=a.textPosition),_t(a,"textOffset")&&(i.offset=a.textOffset),_t(a,"textRotation")&&(i.rotation=a.textRotation),_t(a,"textDistance")&&(i.distance=a.textDistance)}return hB(o,t),R(o.rich,(function(t){hB(t,t)})),{textConfig:i,textContent:r}}function hB(t,e){e&&(e.font=e.textFont||e.font,_t(e,"textStrokeWidth")&&(t.lineWidth=e.textStrokeWidth),_t(e,"textAlign")&&(t.align=e.textAlign),_t(e,"textVerticalAlign")&&(t.verticalAlign=e.textVerticalAlign),_t(e,"textLineHeight")&&(t.lineHeight=e.textLineHeight),_t(e,"textWidth")&&(t.width=e.textWidth),_t(e,"textHeight")&&(t.height=e.textHeight),_t(e,"textBackgroundColor")&&(t.backgroundColor=e.textBackgroundColor),_t(e,"textPadding")&&(t.padding=e.textPadding),_t(e,"textBorderColor")&&(t.borderColor=e.textBorderColor),_t(e,"textBorderWidth")&&(t.borderWidth=e.textBorderWidth),_t(e,"textBorderRadius")&&(t.borderRadius=e.textBorderRadius),_t(e,"textBoxShadowColor")&&(t.shadowColor=e.textBoxShadowColor),_t(e,"textBoxShadowBlur")&&(t.shadowBlur=e.textBoxShadowBlur),_t(e,"textBoxShadowOffsetX")&&(t.shadowOffsetX=e.textBoxShadowOffsetX),_t(e,"textBoxShadowOffsetY")&&(t.shadowOffsetY=e.textBoxShadowOffsetY))}function fB(t,e,n){var i=t;i.textPosition=i.textPosition||n.position||"inside",null!=n.offset&&(i.textOffset=n.offset),null!=n.rotation&&(i.textRotation=n.rotation),null!=n.distance&&(i.textDistance=n.distance);var r=i.textPosition.indexOf("inside")>=0,o=t.fill||"#000";pB(i,e);var a=null==i.textFill;return r?a&&(i.textFill=n.insideFill||"#fff",!i.textStroke&&n.insideStroke&&(i.textStroke=n.insideStroke),!i.textStroke&&(i.textStroke=o),null==i.textStrokeWidth&&(i.textStrokeWidth=2)):(a&&(i.textFill=t.fill||n.outsideFill||"#000"),!i.textStroke&&n.outsideStroke&&(i.textStroke=n.outsideStroke)),i.text=e.text,i.rich=e.rich,R(e.rich,(function(t){pB(t,t)})),i}function pB(t,e){e&&(_t(e,"fill")&&(t.textFill=e.fill),_t(e,"stroke")&&(t.textStroke=e.fill),_t(e,"lineWidth")&&(t.textStrokeWidth=e.lineWidth),_t(e,"font")&&(t.font=e.font),_t(e,"fontStyle")&&(t.fontStyle=e.fontStyle),_t(e,"fontWeight")&&(t.fontWeight=e.fontWeight),_t(e,"fontSize")&&(t.fontSize=e.fontSize),_t(e,"fontFamily")&&(t.fontFamily=e.fontFamily),_t(e,"align")&&(t.textAlign=e.align),_t(e,"verticalAlign")&&(t.textVerticalAlign=e.verticalAlign),_t(e,"lineHeight")&&(t.textLineHeight=e.lineHeight),_t(e,"width")&&(t.textWidth=e.width),_t(e,"height")&&(t.textHeight=e.height),_t(e,"backgroundColor")&&(t.textBackgroundColor=e.backgroundColor),_t(e,"padding")&&(t.textPadding=e.padding),_t(e,"borderColor")&&(t.textBorderColor=e.borderColor),_t(e,"borderWidth")&&(t.textBorderWidth=e.borderWidth),_t(e,"borderRadius")&&(t.textBorderRadius=e.borderRadius),_t(e,"shadowColor")&&(t.textBoxShadowColor=e.shadowColor),_t(e,"shadowBlur")&&(t.textBoxShadowBlur=e.shadowBlur),_t(e,"shadowOffsetX")&&(t.textBoxShadowOffsetX=e.shadowOffsetX),_t(e,"shadowOffsetY")&&(t.textBoxShadowOffsetY=e.shadowOffsetY),_t(e,"textShadowColor")&&(t.textShadowColor=e.textShadowColor),_t(e,"textShadowBlur")&&(t.textShadowBlur=e.textShadowBlur),_t(e,"textShadowOffsetX")&&(t.textShadowOffsetX=e.textShadowOffsetX),_t(e,"textShadowOffsetY")&&(t.textShadowOffsetY=e.textShadowOffsetY))}var mB=Bs.CMD,gB=2*Math.PI,yB=["x","y"],vB=["width","height"],_B=[];function bB(t,e){return Math.abs(t-e)<1e-5}function xB(t){var e,n,i,r,o,a=t.data,s=t.len(),l=[],u=0,c=0,d=0,h=0;function f(t,n){e&&e.length>2&&l.push(e),e=[t,n]}function p(t,n,i,r){bB(t,i)&&bB(n,r)||e.push(t,n,i,r,i,r)}function m(t,n,i,r,o,a){var s=Math.abs(n-t),l=4*Math.tan(s/4)/3,u=n<t?-1:1,c=Math.cos(t),d=Math.sin(t),h=Math.cos(n),f=Math.sin(n),p=c*o+i,m=d*a+r,g=h*o+i,y=f*a+r,v=o*l*u,_=a*l*u;e.push(p-v*d,m+_*c,g+v*f,y-_*h,g,y)}for(var g=0;g<s;){var y=a[g++],v=1===g;switch(v&&(u=a[g],c=a[g+1],d=u,h=c,y!==mB.L&&y!==mB.C&&y!==mB.Q||(e=[d,h])),y){case mB.M:u=d=a[g++],c=h=a[g++],f(d,h);break;case mB.L:n=a[g++],i=a[g++],p(u,c,n,i),u=n,c=i;break;case mB.C:e.push(a[g++],a[g++],a[g++],a[g++],u=a[g++],c=a[g++]);break;case mB.Q:n=a[g++],i=a[g++],r=a[g++],o=a[g++],e.push(u+2/3*(n-u),c+2/3*(i-c),r+2/3*(n-r),o+2/3*(i-o),r,o),u=r,c=o;break;case mB.A:var _=a[g++],b=a[g++],x=a[g++],w=a[g++],M=a[g++],S=a[g++]+M;g+=1;var L=!a[g++];n=Math.cos(M)*x+_,i=Math.sin(M)*w+b,v?(d=n,h=i,f(d,h)):p(u,c,n,i),u=Math.cos(S)*x+_,c=Math.sin(S)*w+b;for(var k=(L?-1:1)*Math.PI/2,T=M;L?T>S:T<S;T+=k){var D=L?Math.max(T+k,S):Math.min(T+k,S);m(T,D,_,b,x,w)}break;case mB.R:d=u=a[g++],h=c=a[g++],n=d+a[g++],i=h+a[g++],f(n,h),p(n,h,n,i),p(n,i,d,i),p(d,i,d,h),p(d,h,n,h);break;case mB.Z:e&&p(u,c,d,h),u=d,c=h;break}}return e&&e.length>2&&l.push(e),l}function wB(t,e){var n=t.length,i=e.length;if(n===i)return[t,e];for(var r=n<i?t:e,o=Math.min(n,i),a=Math.abs(i-n)/6,s=(o-2)/6,l=Math.ceil(a/s)+1,u=[r[0],r[1]],c=a,d=[],h=[],f=2;f<o;){var p=r[f-2],m=r[f-1],g=r[f++],y=r[f++],v=r[f++],_=r[f++],b=r[f++],x=r[f++];if(c<=0)u.push(g,y,v,_,b,x);else{for(var w=Math.min(c,l-1)+1,M=1;M<=w;M++){var S=M/w;Ja(p,g,v,b,S,d),Ja(m,y,_,x,S,h),p=d[3],m=h[3],u.push(d[1],h[1],d[2],h[2],p,m),g=d[5],y=h[5],v=d[6],_=h[6]}c-=w-1}}return r===t?[u,e]:[t,u]}function MB(t,e){for(var n=t.length,i=t[n-2],r=t[n-1],o=[],a=0;a<e.length;)o[a++]=i,o[a++]=r;return o}function SB(t,e){for(var n,i,r,o=[],a=[],s=0;s<Math.max(t.length,e.length);s++){var l=t[s],u=e[s],c=void 0,d=void 0;l?u?(n=wB(l,u),c=n[0],d=n[1],i=c,r=d):(d=MB(r||l,l),c=l):(c=MB(i||u,u),d=u),o.push(c),a.push(d)}return[o,a]}function LB(t){for(var e=0,n=0,i=0,r=t.length,o=0,a=r-2;o<r;a=o,o+=2){var s=t[a],l=t[a+1],u=t[o],c=t[o+1],d=s*c-u*l;e+=d,n+=(s+u)*d,i+=(l+c)*d}return 0===e?[t[0]||0,t[1]||0]:[n/e/3,i/e/3,e]}function kB(t,e,n,i){for(var r=(t.length-2)/6,o=1/0,a=0,s=t.length,l=s-2,u=0;u<r;u++){for(var c=6*u,d=0,h=0;h<s;h+=2){var f=0===h?c:(c+h-2)%l+2,p=t[f]-n[0],m=t[f+1]-n[1],g=e[h]-i[0],y=e[h+1]-i[1],v=g-p,_=y-m;d+=v*v+_*_}d<o&&(o=d,a=u)}return a}function TB(t){for(var e=[],n=t.length,i=0;i<n;i+=2)e[i]=t[n-i-2],e[i+1]=t[n-i-1];return e}function DB(t,e,n,i){for(var r,o=[],a=0;a<t.length;a++){var s=t[a],l=e[a],u=LB(s),c=LB(l);null==r&&(r=u[2]<0!==c[2]<0);var d=[],h=[],f=0,p=1/0,m=[],g=s.length;r&&(s=TB(s));for(var y=6*kB(s,l,u,c),v=g-2,_=0;_<v;_+=2){var b=(y+_)%v+2;d[_+2]=s[b]-u[0],d[_+3]=s[b+1]-u[1]}if(d[0]=s[y]-u[0],d[1]=s[y+1]-u[1],n>0)for(var x=i/n,w=-i/2;w<=i/2;w+=x){var M=Math.sin(w),S=Math.cos(w),L=0;for(_=0;_<s.length;_+=2){var k=d[_],T=d[_+1],D=l[_]-c[0],C=l[_+1]-c[1],I=D*S-C*M,A=D*M+C*S;m[_]=I,m[_+1]=A;var Y=I-k,O=A-T;L+=Y*Y+O*O}if(L<p){p=L,f=w;for(var P=0;P<m.length;P++)h[P]=m[P]}}else for(var E=0;E<g;E+=2)h[E]=l[E]-c[0],h[E+1]=l[E+1]-c[1];o.push({from:d,to:h,fromCp:u,toCp:c,rotation:-f})}return o}function CB(t,e,n){var i,r;if(!t||!e)return e;!t.path&&t.createPathProxy(),i=t.path,i.beginPath(),t.buildPath(i,t.shape),!e.path&&e.createPathProxy(),r=e.path,r===i&&(r=new Bs(!1)),r.beginPath(),PB(e)?e.__oldBuildPath(r,e.shape):e.buildPath(r,e.shape);var o=SB(xB(i),xB(r)),a=o[0],s=o[1],l=DB(a,s,10,Math.PI);AB(e,l,0);var u=n&&n.done,c=n&&n.aborted,d=n&&n.during;return e.animateTo({__morphT:1},A({during:function(t){e.dirtyShape(),d&&d(t)},done:function(){OB(e),e.createPathProxy(),e.dirtyShape(),u&&u()},aborted:function(){c&&c()}},n)),e}function IB(t){for(var e=this.__morphingData,n=this.__morphT,i=1-n,r=[],o=0;o<e.length;o++){var a=e[o],s=a.from,l=a.to,u=a.rotation*n,c=a.fromCp,d=a.toCp,h=Math.sin(u),f=Math.cos(u);Ht(r,c,d,n);for(var p=0;p<s.length;p+=2){var m=s[p],g=s[p+1],y=l[p],v=l[p+1],_=m*i+y*n,b=g*i+v*n;_B[p]=_*f-b*h+r[0],_B[p+1]=_*h+b*f+r[1]}for(p=0;p<s.length;)0===p&&t.moveTo(_B[p++],_B[p++]),t.bezierCurveTo(_B[p++],_B[p++],_B[p++],_B[p++],_B[p++],_B[p++])}}function AB(t,e,n){if(PB(t))YB(t,e,n);else{var i=t;i.__oldBuildPath=i.buildPath,i.buildPath=IB,YB(i,e,n)}}function YB(t,e,n){t.__morphingData=e,t.__morphT=n}function OB(t){PB(t)&&(t.buildPath=t.__oldBuildPath,t.__oldBuildPath=t.__morphingData=null)}function PB(t){return null!=t.__oldBuildPath}function EB(t){return!!t.__combiningSubList}function NB(t){return PB(t)||EB(t)}function RB(t,e,n,i){for(var r=[],o=0,a=0;a<t.length;a++){var s=t[a];if(EB(s)){for(var l=s.__combiningSubList,u=0;u<l.length;u++)r.push(l[u]);o+=l.length}else r.push(s),o++}if(o){var c=n?n.dividingMethod:null,d=$B(e,o,c);ct(d.length===o);var h=n&&n.done,f=n&&n.aborted,p=n&&n.during,m=0,g=!1,y=A({during:function(t){p&&p(t)},done:function(){m++,m===d.length&&(jB(e),h&&h())},aborted:function(){g||(g=!0,f&&f())}},n);for(a=0;a<o;a++){var v=r[a],_=d[a];i&&i(e,_,!0),CB(v,_,y)}return HB(e,d),{fromIndividuals:r,toIndividuals:d,count:o}}}function HB(t,e){if(EB(t))zB(t,e);else{var n=t;zB(n,e),n.__oldAddSelfToZr=t.addSelfToZr,n.__oldRemoveSelfFromZr=t.removeSelfFromZr,n.addSelfToZr=WB,n.removeSelfFromZr=FB,n.__oldBuildPath=n.buildPath,n.buildPath=bt,n.childrenRef=VB}}function jB(t){if(EB(t)){var e=t;zB(e,null),e.addSelfToZr=e.__oldAddSelfToZr,e.removeSelfFromZr=e.__oldRemoveSelfFromZr,e.buildPath=e.__oldBuildPath,e.childrenRef=e.__combiningSubList=e.__oldAddSelfToZr=e.__oldRemoveSelfFromZr=e.__oldBuildPath=null}}function zB(t,e){if(t.__combiningSubList!==e){if(BB(t,"removeSelfFromZr"),t.__combiningSubList=e,e)for(var n=0;n<e.length;n++)e[n].parent=t;BB(t,"addSelfToZr")}}function WB(t){this.__oldAddSelfToZr(t),BB(this,"addSelfToZr")}function BB(t,e){var n=t.__combiningSubList,i=t.__zr;if(n&&i)for(var r=0;r<n.length;r++){var o=n[r];o[e](i)}}function FB(t){this.__oldRemoveSelfFromZr(t);for(var e=this.__combiningSubList,n=0;n<e.length;n++){var i=e[n];i.removeSelfFromZr(t)}}function VB(){return this.__combiningSubList}function GB(t,e,n,i){var r,o=e.length,a=n?n.dividingMethod:null,s=!1;if(EB(t)){var l=t.__combiningSubList;l.length===o?r=l:(r=$B(t,o,a),s=!0)}else r=$B(t,o,a),s=!0;ct(r.length===o);for(var u=0;u<o;u++)s&&i&&i(t,r[u],!1),CB(r[u],e[u],n);return{fromIndividuals:r,toIndividuals:e,count:o}}function $B(t,e,n){return"duplicate"===n?qB(t,e):UB(t,e)}function UB(t,e){var n=[];if(e<=0)return n;if(1===e)return qB(t,e);if(t instanceof Al)for(var i=t.shape,r=i.height>i.width?1:0,o=vB[r],a=yB[r],s=i[o]/e,l=i[a],u=0;u<e;u++,l+=s){var c={x:i.x,y:i.y,width:i.width,height:i.height};c[a]=l,c[o]=u<e-1?s:i[a]+i[o]-l;var d=new Al({shape:c});n.push(d)}else{if(!(t instanceof qc))return qB(t,e);i=t.shape;var h=i.clockwise,f=i.startAngle,p=i.endAngle,m=XB(f,i.endAngle,h),g=(m-f)/e,y=f;for(u=0;u<e;u++,y+=g){d=new qc({shape:{cx:i.cx,cy:i.cy,r:i.r,r0:i.r0,clockwise:h,startAngle:y,endAngle:u===e-1?p:y+g}});n.push(d)}}return n}function qB(t,e){var n=[];if(e<=0)return n;for(var i=t.constructor,r=0;r<e;r++){var o=new i({shape:T(t.shape)});n.push(o)}return n}function XB(t,e,n){return e+gB*Math[n?"ceil":"floor"]((t-e)/gB)}var ZB=Io(),JB={x:1,y:1,scaleX:1,scaleY:1,originX:1,originY:1,rotation:1},KB=(B(JB).join(", "),{color:"fill",borderColor:"stroke"}),QB={symbol:1,symbolSize:1,symbolKeepAspect:1,legendIcon:1,visualMeta:1,liftZ:1,decal:1},tF="emphasis",eF="normal",nF="blur",iF="select",rF=[eF,tF,nF,iF],oF={normal:["itemStyle"],emphasis:[tF,"itemStyle"],blur:[nF,"itemStyle"],select:[iF,"itemStyle"]},aF={normal:["label"],emphasis:[tF,"label"],blur:[nF,"label"],select:[iF,"label"]},sF="e\0\0",lF={normal:{},emphasis:{},blur:{},select:{}},uF={position:["x","y"],scale:["scaleX","scaleY"],origin:["originX","originY"]},cF=new ji,dF={cartesian2d:nB,geo:rB,singleAxis:aB,polar:lB,calendar:uB},hF=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(){this.currentZLevel=this.get("zlevel",!0),this.currentZ=this.get("z",!0)},e.prototype.getInitialData=function(t,e){return BL(this.getSource(),this)},e.prototype.getDataParams=function(e,n,i){var r=t.prototype.getDataParams.call(this,e,n);return i&&(r.info=ZB(i).info),r},e.type="series.custom",e.dependencies=["grid","polar","geo","singleAxis","calendar"],e.defaultOption={coordinateSystem:"cartesian2d",zlevel:0,z:2,legendHoverLink:!0,clip:!1},e}(Yy),fF=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,i){var r=this._data,o=t.getData(),a=this.group,s=NF(t,o,e,n);r||a.removeAll();var l=t.__transientTransitionOpt;if(!l||null!=l.from&&null!=l.to){var u=new nV(t,l),c=l?"multiple":"oneToOne";new $S(r?r.getIndices():[],o.getIndices(),pF(r,c,l&&l.from),pF(o,c,l&&l.to),null,c).add((function(e){HF(n,null,e,s(e,i),t,a,o,null)})).remove((function(e){JF(r.getItemGraphicEl(e),t,a)})).update((function(e,l){u.reset("oneToOne");var c=r.getItemGraphicEl(l);u.findAndAddFrom(c),u.hasFrom()&&(eV(c,a),c=null),HF(n,c,e,s(e,i),t,a,o,u),u.applyMorphing()})).updateManyToOne((function(e,l){u.reset("manyToOne");for(var c=0;c<l.length;c++){var d=r.getItemGraphicEl(l[c]);u.findAndAddFrom(d),eV(d,a)}HF(n,null,e,s(e,i),t,a,o,u),u.applyMorphing()})).updateOneToMany((function(e,l){u.reset("oneToMany");var c=e.length,d=r.getItemGraphicEl(l);u.findAndAddFrom(d),eV(d,a);for(var h=0;h<c;h++)HF(n,null,e[h],s(e[h],i),t,a,o,u);u.applyMorphing()})).execute()}else r&&r.each((function(e){JF(r.getItemGraphicEl(e),t,a)})),o.each((function(e){HF(n,null,e,s(e,i),t,a,o,null)}));var d=t.get("clip",!0)?_k(t.coordinateSystem,!1,t):null;d?a.setClipPath(d):a.removeClipPath(),this._data=o},e.prototype.incrementalPrepareRender=function(t,e,n){this.group.removeAll(),this._data=null},e.prototype.incrementalRender=function(t,e,n,i,r){var o=e.getData(),a=NF(e,o,n,i);function s(t){t.isGroup||(t.incremental=!0,t.ensureState("emphasis").hoverLayer=!0)}for(var l=t.start;l<t.end;l++){var u=HF(null,null,l,a(l,r),e,this.group,o,null);u&&u.traverse(s)}},e.prototype.filterForExposedEvent=function(t,e,n,i){var r=e.element;if(null==r||n.name===r)return!0;while((n=n.__hostTarget||n.parent)&&n!==this.group)if(n.name===r)return!0;return!1},e.type="custom",e}(Vy);function pF(t,e,n){if(t){if("oneToOne"===e)return function(e,n){return t.getId(n)};var i=t.getDimension(n),r=t.getDimensionInfo(i);if(!r){var o="";0,im(o)}var a=r.ordinalMeta;return function(e,n){var r=t.get(i,n);return a&&(r=a.categories[r]),null==r||rt(r)?e+"":"_ec_"+r}}}function mF(t){var e,n=t.type;if("path"===n){var i=t.shape,r=null!=i.width&&null!=i.height?{x:i.x||0,y:i.y||0,width:i.width,height:i.height}:null,o=KF(i);e=Gd(o,null,r,i.layout||"center"),ZB(e).customPathData=o}else if("image"===n)e=new wl({}),ZB(e).customImagePath=t.style.image;else if("text"===n)e=new Gl({});else if("group"===n)e=new Mr;else{if("compoundPath"===n)throw new Error('"compoundPath" is not supported yet.');var a=Vd(n);if(!a){var s="";0,im(s)}e=new a}return ZB(e).customGraphicType=n,e.name=t.name,e.z2EmphasisLift=1,e.z2SelectLift=1,e}function gF(t,e,n,i,r,o,a,s,l,u){var c={},d={},h=e.isGroup?null:e;!n&&_F("shape",e,null,r,c,l),bF("shape",r,d),!n&&_F("extra",e,null,r,c,l),bF("extra",r,d),!n&&xF(e,null,r,c,l),wF(r,d);var f=a&&a.normal.cfg;if(f&&e.setTextConfig(f),"text"===e.type&&o){var p=o;_t(p,"textFill")&&(p.fill=p.textFill),_t(p,"textStroke")&&(p.stroke=p.textStroke)}if(o){var m=void 0,g=tV(e)?o.decal:null;t&&g&&(g.dirty=!0,m=wb(g,t)),o.__decalPattern=m}return!n&&MF(e,null,r,o,c,l),h&&_t(r,"invisible")&&(h.invisible=r.invisible),n||(yF(e,d,o),vF(e,i,r,s,c,l)),_t(r,"silent")&&(e.silent=r.silent),_t(r,"ignore")&&(e.ignore=r.ignore),u||_t(r,"info")&&(ZB(e).info=r.info),o?e.dirty():e.markRedraw(),n?d:null}function yF(t,e,n){var i=t.isGroup?null:t;if(i&&n){var r=n.__decalPattern,o=void 0;r&&(o=n.decal,n.decal=r),i.useStyle(n),r&&(n.decal=o);for(var a=i.animators,s=0;s<a.length;s++){var l=a[s];"style"===l.targetName&&l.changeTarget(i.style)}}e&&t.attr(e)}function vF(t,e,n,i,r,o){if(r){var a=n.during;ZB(t).userDuring=a;var s=a?V(CF,{el:t,userDuring:a}):null,l={dataIndex:e,isFrom:!0,during:s};o?eh(t,r,i,l):th(t,r,i,l)}}function _F(t,e,n,i,r,o){var a=i[t];if(a){var s,l=e[t],u=a.enterFrom;if(o&&u){!s&&(s=r[t]={});for(var c=B(u),d=0;d<c.length;d++){var h=c[d];s[h]=u[h]}}if(!o&&l&&(null==n||"shape"!==t))if(a.transition){!s&&(s=r[t]={});var f=so(a.transition);for(d=0;d<f.length;d++){h=f[d];var p=l[h];0,s[h]=p}}else if(O(i.transition,t)>=0){!s&&(s=r[t]={});var m=B(l);for(d=0;d<m.length;d++){h=m[d],p=l[h];LF(a[h],p)&&(s[h]=p)}}var g=a.leaveTo;if(g){var y=kF(e),v=y[t]||(y[t]={}),_=B(g);for(d=0;d<_.length;d++){h=_[d];v[h]=g[h]}}}}function bF(t,e,n){var i=e[t];if(i)for(var r=n[t]={},o=B(i),a=0;a<o.length;a++){var s=o[a];r[s]=jn(i[s])}}function xF(t,e,n,i,r){var o=n.enterFrom;if(r&&o)for(var a=B(o),s=0;s<a.length;s++){var l=a[s];0,i[l]=o[l]}if(!r)if(e){var u=SF(e,t);EF(i,"x",u),EF(i,"y",u),EF(i,"scaleX",u),EF(i,"scaleY",u),EF(i,"originX",u),EF(i,"originY",u),EF(i,"rotation",u)}else if(n.transition){var c=so(n.transition);for(s=0;s<c.length;s++){l=c[s];if("style"!==l&&"shape"!==l&&"extra"!==l){var d=t[l];0,i[l]=d}}}else EF(i,"x",t),EF(i,"y",t);var h=n.leaveTo;if(h){var f=kF(t),p=B(h);for(s=0;s<p.length;s++){l=p[s];0,f[l]=h[l]}}}function wF(t,e){OF(t,e,"position"),OF(t,e,"scale"),OF(t,e,"origin"),PF(t,e,"x"),PF(t,e,"y"),PF(t,e,"scaleX"),PF(t,e,"scaleY"),PF(t,e,"originX"),PF(t,e,"originY"),PF(t,e,"rotation")}function MF(t,e,n,i,r,o){if(i){var a,s=e||t,l=s.style,u=i.enterFrom;if(o&&u){var c=B(u);!a&&(a=r.style={});for(var d=0;d<c.length;d++){var h=c[d];a[h]=u[h]}}if(!o&&l)if(i.transition){var f=so(i.transition);!a&&(a=r.style={});for(d=0;d<f.length;d++){h=f[d];var p=l[h];a[h]=p}}else if(t.getAnimationStyleProps&&O(n.transition,"style")>=0){var m=t.getAnimationStyleProps(),g=m?m.style:null;if(g){!a&&(a=r.style={});var y=B(i);for(d=0;d<y.length;d++){h=y[d];if(g[h]){p=l[h];a[h]=p}}}}var v=i.leaveTo;if(v){var _=B(v),b=kF(t),x=b.style||(b.style={});for(d=0;d<_.length;d++){h=_[d];x[h]=v[h]}}}}function SF(t,e){if(!t||t===e||t.parent===e.parent)return t;var n=cF.transform||(cF.transform=xi([])),i=t.getComputedTransform();i?wi(n,i):xi(n);var r=e.parent;return r&&r.getComputedTransform(),cF.originX=t.originX,cF.originY=t.originY,cF.parent=r,cF.decomposeTransform(),cF}function LF(t,e){return N(t)?t!==e:null!=t&&isFinite(t)}function kF(t){var e=ZB(t);return e.leaveToProps||(e.leaveToProps={})}var TF={},DF={setTransform:function(t,e){return TF.el[t]=e,this},getTransform:function(t){return TF.el[t]},setShape:function(t,e){var n=TF.el.shape||(TF.el.shape={});return n[t]=e,TF.isShapeDirty=!0,this},getShape:function(t){var e=TF.el.shape;if(e)return e[t]},setStyle:function(t,e){var n=TF.el.style;return n&&(n[t]=e,TF.isStyleDirty=!0),this},getStyle:function(t){var e=TF.el.style;if(e)return e[t]},setExtra:function(t,e){var n=TF.el.extra||(TF.el.extra={});return n[t]=e,this},getExtra:function(t){var e=TF.el.extra;if(e)return e[t]}};function CF(){var t=this,e=t.el;if(e){var n=ZB(e).userDuring,i=t.userDuring;n===i?(TF.el=e,TF.isShapeDirty=!1,TF.isStyleDirty=!1,i(DF),TF.isShapeDirty&&e.dirtyShape&&e.dirtyShape(),TF.isStyleDirty&&e.dirtyStyle&&e.dirtyStyle()):t.el=t.userDuring=null}}function IF(t,e,n,i,r,o,a){var s=e.isGroup?null:e,l=r&&r[t].cfg;if(s){var u=s.ensureState(t);if(!1===i){var c=s.getState(t);c&&(c.style=null)}else u.style=i||null;l&&(u.textConfig=l),ku(s)}}function AF(t,e,n,i){if(!t.isGroup){var r=t,o=n.currentZ,a=n.currentZLevel;r.z=o,r.zlevel=a;var s=e.z2;null!=s&&(r.z2=s||0);for(var l=0;l<rF.length;l++)YF(r,e,rF[l])}}function YF(t,e,n){var i,r=n===eF,o=r?e:VF(e,n),a=o?o.z2:null;null!=a&&(i=r?t:t.ensureState(n),i.z2=a||0)}function OF(t,e,n,i){var r=t[n],o=uF[n];r&&(i?(e[o[0]]=i[o[0]],e[o[1]]=i[o[1]]):(e[o[0]]=r[0],e[o[1]]=r[1]))}function PF(t,e,n,i){null!=t[n]&&(e[n]=i?i[n]:t[n])}function EF(t,e,n){n&&(t[e]=n[e])}function NF(t,e,n,i){var r=t.get("renderItem"),o=t.coordinateSystem,a={};o&&(a=o.prepareCustoms?o.prepareCustoms(o):dF[o.type](o));for(var s,l,u=A({getWidth:i.getWidth,getHeight:i.getHeight,getZr:i.getZr,getDevicePixelRatio:i.getDevicePixelRatio,value:b,style:w,ordinalRawValue:x,styleEmphasis:M,visual:k,barLayout:T,currentSeriesIndices:D,font:C},a.api||{}),c={context:{},seriesId:t.id,seriesName:t.name,seriesIndex:t.seriesIndex,coordSys:a.coordSys,dataInsideLength:e.count(),encode:RF(t.getData())},d={},h={},f={},p={},m=0;m<rF.length;m++){var g=rF[m];f[g]=t.getModel(oF[g]),p[g]=t.getModel(aF[g])}function y(t){return t===s?l||(l=e.getItemModel(t)):e.getItemModel(t)}function v(t,n){return e.hasItemOption?t===s?d[n]||(d[n]=y(t).getModel(oF[n])):y(t).getModel(oF[n]):f[n]}function _(t,n){return e.hasItemOption?t===s?h[n]||(h[n]=y(t).getModel(aF[n])):y(t).getModel(aF[n]):p[n]}return function(t,n){return s=t,l=null,d={},h={},r&&r(A({dataIndexInside:t,dataIndex:e.getRawIndex(t),actionType:n?n.type:null},c),u)};function b(t,n){return null==n&&(n=s),e.get(e.getDimension(t||0),n)}function x(t,n){null==n&&(n=s);var i=e.getDimensionInfo(t||0);if(i){var r=e.get(i.name,n),o=i&&i.ordinalMeta;return o?o.categories[r]:r}}function w(n,i){null==i&&(i=s);var r=e.getItemVisual(i,"style"),o=r&&r.fill,a=r&&r.opacity,l=v(i,eF).getItemStyle();null!=o&&(l.fill=o),null!=a&&(l.opacity=a);var u={inheritColor:q(o)?o:"#000"},c=_(i,eF),d=Lh(c,null,u,!1,!0);d.text=c.getShallow("show")?at(t.getFormattedLabel(i,eF),GL(e,i)):null;var h=kh(c,u,!1);return L(n,l),l=fB(l,d,h),n&&S(l,n),l.legacy=!0,l}function M(n,i){null==i&&(i=s);var r=v(i,tF).getItemStyle(),o=_(i,tF),a=Lh(o,null,null,!0,!0);a.text=o.getShallow("show")?st(t.getFormattedLabel(i,tF),t.getFormattedLabel(i,eF),GL(e,i)):null;var l=kh(o,null,!0);return L(n,r),r=fB(r,a,l),n&&S(r,n),r.legacy=!0,r}function S(t,e){for(var n in e)_t(e,n)&&(t[n]=e[n])}function L(t,e){t&&(t.textFill&&(e.textFill=t.textFill),t.textPosition&&(e.textPosition=t.textPosition))}function k(t,n){if(null==n&&(n=s),_t(KB,t)){var i=e.getItemVisual(n,"style");return i?i[KB[t]]:null}if(_t(QB,t))return e.getItemVisual(n,t)}function T(t){if("cartesian2d"===o.type){var e=o.getBaseAxis();return qk(A({axis:e},t))}}function D(){return n.getCurrentSeriesIndices()}function C(t){return Oh(t,n)}}function RF(t){var e={};return R(t.dimensions,(function(n,i){var r=t.getDimensionInfo(n);if(!r.isExtraCoord){var o=r.coordDim,a=e[o]=e[o]||[];a[r.coordDimIndex]=i}})),e}function HF(t,e,n,i,r,o,a,s){if(i)return e=jF(t,e,n,i,r,o,!0,s),e&&a.setItemGraphicEl(n,e),e&&$u(e,i.focus,i.blurScope),e;eV(e,o)}function jF(t,e,n,i,r,o,a,s){var l=-1;e&&zF(e,i)&&(l=o.childrenRef().indexOf(e),e=null);var u=!e;e?e.clearStates():e=mF(i);var c=ZB(e).canMorph=i.morph&&tV(e),d=c&&s&&s.hasFrom(),h=u&&!d;lF.normal.cfg=lF.normal.conOpt=lF.emphasis.cfg=lF.emphasis.conOpt=lF.blur.cfg=lF.blur.conOpt=lF.select.cfg=lF.select.conOpt=null,lF.isLegacy=!1,BF(e,n,i,r,h,lF),WF(e,n,i,r,h);var f=gF(t,e,d,n,i,i.style,lF,r,h,!1);d&&s.addTo(e,i,n,f);for(var p=0;p<rF.length;p++){var m=rF[p];if(m!==eF){var g=VF(i,m),y=GF(i,g,m);IF(m,e,g,y,lF,a,!1)}}return AF(e,i,r,lF),"group"===i.type&&$F(t,e,n,i,r,s),l>=0?o.replaceAt(e,l):o.add(e),e}function zF(t,e){var n=ZB(t),i=e.type,r=e.shape,o=e.style;return null!=i&&i!==n.customGraphicType||"path"===i&&QF(r)&&KF(r)!==n.customPathData||"image"===i&&_t(o,"image")&&o.image!==n.customImagePath}function WF(t,e,n,i,r){var o=n.clipPath;if(!1===o)t&&t.getClipPath()&&t.removeClipPath();else if(o){var a=t.getClipPath();a&&zF(a,o)&&(a=null),a||(a=mF(o),t.setClipPath(a)),gF(null,a,null,e,o,null,null,i,r,!1)}}function BF(t,e,n,i,r,o){if(!t.isGroup){FF(n,null,o),FF(n,tF,o);var a=o.normal.conOpt,s=o.emphasis.conOpt,l=o.blur.conOpt,u=o.select.conOpt;if(null!=a||null!=s||null!=u||null!=l){var c=t.getTextContent();if(!1===a)c&&t.removeTextContent();else{a=o.normal.conOpt=a||{type:"text"},c?c.clearStates():(c=mF(a),t.setTextContent(c));var d=a&&a.style;gF(null,c,null,e,a,d,null,i,r,!0);for(var h=0;h<rF.length;h++){var f=rF[h];if(f!==eF){var p=o[f].conOpt;IF(f,c,p,GF(a,p,f),null,!1,!0)}}d?c.dirty():c.markRedraw()}}}}function FF(t,e,n){var i=e?VF(t,e):t,r=e?GF(t,i,tF):t.style,o=t.type,a=i?i.textConfig:null,s=t.textContent,l=s?e?VF(s,e):s:null;if(r&&(n.isLegacy||cB(r,o,!!a,!!l))){n.isLegacy=!0;var u=dB(r,o,!e);!a&&u.textConfig&&(a=u.textConfig),!l&&u.textContent&&(l=u.textContent)}if(!e&&l){var c=l;!c.type&&(c.type="text")}var d=e?n[e]:n.normal;d.cfg=a,d.conOpt=l}function VF(t,e){return e?t?t[e]:null:t}function GF(t,e,n){var i=e&&e.style;return null==i&&n===tF&&t&&(i=t.styleEmphasis),i}function $F(t,e,n,i,r,o){var a=i.children,s=a?a.length:0,l=i.$mergeChildren,u="byName"===l||i.diffChildrenByName,c=!1===l;if(s||u||c)if(u)UF({api:t,oldChildren:e.children()||[],newChildren:a||[],dataIndex:n,seriesModel:r,group:e,morphPreparation:o});else{c&&e.removeAll();for(var d=0;d<s;d++)a[d]&&jF(t,e.childAt(d),n,a[d],r,e,!1,o);for(var h=e.childCount()-1;h>=d;h--)JF(e.childAt(h),r,e)}}function UF(t){new $S(t.oldChildren,t.newChildren,qF,qF,t).add(XF).update(XF).remove(ZF).execute()}function qF(t,e){var n=t&&t.name;return null!=n?n:sF+e}function XF(t,e){var n=this.context,i=null!=t?n.newChildren[t]:null,r=null!=e?n.oldChildren[e]:null;jF(n.api,r,n.dataIndex,i,n.seriesModel,n.group,!1,n.morphPreparation)}function ZF(t){var e=this.context,n=e.oldChildren[t];JF(n,e.seriesModel,e.group)}function JF(t,e,n){if(t){var i=ZB(t).leaveToProps;i?th(t,i,e,{cb:function(){n.remove(t)}}):n.remove(t)}}function KF(t){return t&&(t.pathData||t.d)}function QF(t){return t&&(_t(t,"pathData")||_t(t,"d"))}function tV(t){return t&&t instanceof pl}function eV(t,e){t&&e.remove(t)}var nV=function(){function t(t,e){this._fromList=[],this._toList=[],this._toElOptionList=[],this._allPropsFinalList=[],this._toDataIndices=[],this._morphConfigList=[],this._seriesModel=t,this._transOpt=e}return t.prototype.hasFrom=function(){return!!this._fromList.length},t.prototype.findAndAddFrom=function(t){if(t&&(ZB(t).canMorph&&this._fromList.push(t),t.isGroup))for(var e=t.childrenRef(),n=0;n<e.length;n++)this.findAndAddFrom(e[n])},t.prototype.addTo=function(t,e,n,i){t&&(this._toList.push(t),this._toElOptionList.push(e),this._toDataIndices.push(n),this._allPropsFinalList.push(i))},t.prototype.applyMorphing=function(){var t=this._type,e=this._fromList,n=this._toList,i=n.length,r=e.length;if(r&&i)if("oneToOne"===t)for(var o=0;o<i;o++)this._oneToOneForSingleTo(o,o);else if("manyToOne"===t)for(var a=Math.max(1,Math.floor(r/i)),s=(o=0,0);o<i;o++,s+=a){var l=o+1>=i?r-s:a;this._manyToOneForSingleTo(o,s>=r?null:s,l)}else if("oneToMany"===t)for(var u=Math.max(1,Math.floor(i/r)),c=0,d=0;c<i;c+=u,d++){var h=c+u>=i?i-c:u;this._oneToManyForSingleFrom(c,h,d>=r?null:d)}},t.prototype._oneToOneForSingleTo=function(t,e){var n=this._toList[t],i=this._toElOptionList[t],r=this._toDataIndices[t],o=this._allPropsFinalList[t],a=this._fromList[e],s=this._getOrCreateMorphConfig(r),l=s.duration;if(a&&EB(a)){if(yF(n,o,i.style),l){var u=RB([a],n,s,iV);this._processResultIndividuals(u,t,null)}}else{var c=l&&a&&(a!==n||NB(a))?a:null,d={};_F("shape",n,c,i,d,!1),_F("extra",n,c,i,d,!1),xF(n,c,i,d,!1),MF(n,c,i,i.style,d,!1),yF(n,o,i.style),c&&CB(c,n,s),vF(n,r,i,this._seriesModel,d,!1)}},t.prototype._manyToOneForSingleTo=function(t,e,n){var i=this._toList[t],r=this._toElOptionList[t],o=this._allPropsFinalList[t];yF(i,o,r.style);var a=this._getOrCreateMorphConfig(this._toDataIndices[t]);if(a.duration&&null!=e){for(var s=[],l=e;l<n;l++)s.push(this._fromList[l]);var u=RB(s,i,a,iV);this._processResultIndividuals(u,t,null)}},t.prototype._oneToManyForSingleFrom=function(t,e,n){for(var i=null==n?null:this._fromList[n],r=this._toList,o=[],a=t;a<e;a++){var s=r[a];yF(s,this._allPropsFinalList[a],this._toElOptionList[a].style),o.push(s)}var l=this._getOrCreateMorphConfig(this._toDataIndices[t]);if(l.duration&&i){var u=GB(i,o,l,iV);this._processResultIndividuals(u,t,e)}},t.prototype._processResultIndividuals=function(t,e,n){for(var i=null!=n,r=0;r<t.count;r++){var o=t.fromIndividuals[r],a=t.toIndividuals[r],s=e+(i?r:0),l=this._toElOptionList[s],u=this._toDataIndices[s],c={};xF(a,o,l,c,!1),MF(a,o,l,l.style,c,!1),vF(a,u,l,this._seriesModel,c,!1)}},t.prototype._getOrCreateMorphConfig=function(t){var e,n,i,r=this._morphConfigList,o=r[t];if(o)return o;var a=this._seriesModel,s=this._transOpt;if(a.isAnimationEnabled()){var l=void 0;if(a&&a.ecModel){var u=a.ecModel.getUpdatePayload();l=u&&u.animation}if(l)e=l.duration||0,n=l.easing||"cubicOut",i=l.delay||0;else{n=a.get("animationEasingUpdate");var c=a.get("animationDelayUpdate");i=U(c)?c(t):c;var d=a.get("animationDurationUpdate");e=U(d)?d(t):d}}return o={duration:e||0,delay:i,easing:n,dividingMethod:s?s.dividingMethod:null},r[t]=o,o},t.prototype.reset=function(t){this._type=t,this._fromList.length=this._toList.length=this._toElOptionList.length=this._allPropsFinalList.length=this._toDataIndices.length=0},t}();function iV(t,e,n){e.style=n?T(t.style):t.style,e.zlevel=t.zlevel,e.z=t.z,e.z2=t.z2}function rV(t){t.registerChartView(fF),t.registerSeriesModel(hF)}var oV=Io(),aV=T,sV=V,lV=function(){function t(){this._dragging=!1,this.animationThreshold=15}return t.prototype.render=function(t,e,n,i){var r=e.get("value"),o=e.get("status");if(this._axisModel=t,this._axisPointerModel=e,this._api=n,i||this._lastValue!==r||this._lastStatus!==o){this._lastValue=r,this._lastStatus=o;var a=this._group,s=this._handle;if(!o||"hide"===o)return a&&a.hide(),void(s&&s.hide());a&&a.show(),s&&s.show();var l={};this.makeElOption(l,r,t,e,n);var u=l.graphicKey;u!==this._lastGraphicKey&&this.clear(n),this._lastGraphicKey=u;var c=this._moveAnimation=this.determineAnimation(t,e);if(a){var d=G(uV,e,c);this.updatePointerEl(a,l,d),this.updateLabelEl(a,l,d,e)}else a=this._group=new Mr,this.createPointerEl(a,l,t,e),this.createLabelEl(a,l,t,e),n.getZr().add(a);fV(a,e,!0),this._renderHandle(r)}},t.prototype.remove=function(t){this.clear(t)},t.prototype.dispose=function(t){this.clear(t)},t.prototype.determineAnimation=function(t,e){var n=e.get("animation"),i=t.axis,r="category"===i.type,o=e.get("snap");if(!o&&!r)return!1;if("auto"===n||null==n){var a=this.animationThreshold;if(r&&i.getBandWidth()>a)return!0;if(o){var s=YI(t).seriesDataCount,l=i.getExtent();return Math.abs(l[0]-l[1])/s>a}return!1}return!0===n},t.prototype.makeElOption=function(t,e,n,i,r){},t.prototype.createPointerEl=function(t,e,n,r){var o=e.pointer;if(o){var a=oV(t).pointerEl=new i[o.type](aV(e.pointer));t.add(a)}},t.prototype.createLabelEl=function(t,e,n,i){if(e.label){var r=oV(t).labelEl=new Gl(aV(e.label));t.add(r),dV(r,i)}},t.prototype.updatePointerEl=function(t,e,n){var i=oV(t).pointerEl;i&&e.pointer&&(i.setStyle(e.pointer.style),n(i,{shape:e.pointer.shape}))},t.prototype.updateLabelEl=function(t,e,n,i){var r=oV(t).labelEl;r&&(r.setStyle(e.label.style),n(r,{x:e.label.x,y:e.label.y}),dV(r,i))},t.prototype._renderHandle=function(t){if(!this._dragging&&this.updateHandleTransform){var e,n=this._axisPointerModel,i=this._api.getZr(),r=this._handle,o=n.getModel("handle"),a=n.get("status");if(!o.get("show")||!a||"hide"===a)return r&&i.remove(r),void(this._handle=null);this._handle||(e=!0,r=this._handle=ph(o.get("icon"),{cursor:"move",draggable:!0,onmousemove:function(t){fe(t.event)},onmousedown:sV(this._onHandleDragMove,this,0,0),drift:sV(this._onHandleDragMove,this),ondragend:sV(this._onHandleDragEnd,this)}),i.add(r)),fV(r,n,!1),r.setStyle(o.getItemStyle(null,["color","borderColor","borderWidth","opacity","shadowColor","shadowBlur","shadowOffsetX","shadowOffsetY"]));var s=o.get("size");$(s)||(s=[s,s]),r.scaleX=s[0]/2,r.scaleY=s[1]/2,Xy(this,"_doDispatchAxisPointer",o.get("throttle")||0,"fixRate"),this._moveHandleToValue(t,e)}},t.prototype._moveHandleToValue=function(t,e){uV(this._axisPointerModel,!e&&this._moveAnimation,this._handle,hV(this.getHandleTransform(t,this._axisModel,this._axisPointerModel)))},t.prototype._onHandleDragMove=function(t,e){var n=this._handle;if(n){this._dragging=!0;var i=this.updateHandleTransform(hV(n),[t,e],this._axisModel,this._axisPointerModel);this._payloadInfo=i,n.stopAnimation(),n.attr(hV(i)),oV(n).lastProp=null,this._doDispatchAxisPointer()}},t.prototype._doDispatchAxisPointer=function(){var t=this._handle;if(t){var e=this._payloadInfo,n=this._axisModel;this._api.dispatchAction({type:"updateAxisPointer",x:e.cursorPoint[0],y:e.cursorPoint[1],tooltipOption:e.tooltipOption,axesInfo:[{axisDim:n.axis.dim,axisIndex:n.componentIndex}]})}},t.prototype._onHandleDragEnd=function(){this._dragging=!1;var t=this._handle;if(t){var e=this._axisPointerModel.get("value");this._moveHandleToValue(e),this._api.dispatchAction({type:"hideTip"})}},t.prototype.clear=function(t){this._lastValue=null,this._lastStatus=null;var e=t.getZr(),n=this._group,i=this._handle;e&&n&&(this._lastGraphicKey=null,n&&e.remove(n),i&&e.remove(i),this._group=null,this._handle=null,this._payloadInfo=null)},t.prototype.doClear=function(){},t.prototype.buildLabel=function(t,e,n){return n=n||0,{x:t[n],y:t[1-n],width:e[n],height:e[1-n]}},t}();function uV(t,e,n,i){cV(oV(n).lastProp,i)||(oV(n).lastProp=i,e?th(n,i,t):(n.stopAnimation(),n.attr(i)))}function cV(t,e){if(J(t)&&J(e)){var n=!0;return R(e,(function(e,i){n=n&&cV(t[i],e)})),!!n}return t===e}function dV(t,e){t[e.get(["label","show"])?"show":"hide"]()}function hV(t){return{x:t.x||0,y:t.y||0,rotation:t.rotation||0}}function fV(t,e,n){var i=e.get("z"),r=e.get("zlevel");t&&t.traverse((function(t){"group"!==t.type&&(null!=i&&(t.z=i),null!=r&&(t.zlevel=r),t.silent=n)}))}var pV=lV;function mV(t){var e,n=t.get("type"),i=t.getModel(n+"Style");return"line"===n?(e=i.getLineStyle(),e.fill=null):"shadow"===n&&(e=i.getAreaStyle(),e.stroke=null),e}function gV(t,e,n,i,r){var o=n.get("value"),a=vV(o,e.axis,e.ecModel,n.get("seriesDataIndices"),{precision:n.get(["label","precision"]),formatter:n.get(["label","formatter"])}),s=n.getModel("label"),l=Uf(s.get("padding")||0),u=s.getFont(),c=rr(a,u),d=r.position,h=c.width+l[1]+l[3],f=c.height+l[0]+l[2],p=r.align;"right"===p&&(d[0]-=h),"center"===p&&(d[0]-=h/2);var m=r.verticalAlign;"bottom"===m&&(d[1]-=f),"middle"===m&&(d[1]-=f/2),yV(d,h,f,i);var g=s.get("backgroundColor");g&&"auto"!==g||(g=e.get(["axisLine","lineStyle","color"])),t.label={x:d[0],y:d[1],style:Lh(s,{text:a,font:u,fill:s.getTextColor(),padding:l,backgroundColor:g}),z2:10}}function yV(t,e,n,i){var r=i.getWidth(),o=i.getHeight();t[0]=Math.min(t[0]+e,r)-e,t[1]=Math.min(t[1]+n,o)-n,t[0]=Math.max(t[0],0),t[1]=Math.max(t[1],0)}function vV(t,e,n,i,r){t=e.scale.parse(t);var o=e.scale.getLabel({value:t},{precision:r.precision}),a=r.formatter;if(a){var s={value:wC(e,{value:t}),axisDimension:e.dim,axisIndex:e.index,seriesData:[]};R(i,(function(t){var e=n.getSeriesByIndex(t.seriesIndex),i=t.dataIndexInside,r=e&&e.getDataParams(i);r&&s.seriesData.push(r)})),q(a)?o=a.replace("{value}",o):U(a)&&(o=a(s))}return o}function _V(t,e,n){var i=bi();return Li(i,i,n.rotation),Si(i,i,n.position),sh([t.dataToCoord(e),(n.labelOffset||0)+(n.labelDirection||1)*(n.labelMargin||0)],i)}function bV(t,e,n,i,r,o){var a=SI.innerTextLayout(n.rotation,0,n.labelDirection);n.labelMargin=r.get(["label","margin"]),gV(e,i,r,o,{position:_V(i.axis,t,n),align:a.textAlign,verticalAlign:a.textVerticalAlign})}function xV(t,e,n){return n=n||0,{x1:t[n],y1:t[1-n],x2:e[n],y2:e[1-n]}}function wV(t,e,n){return n=n||0,{x:t[n],y:t[1-n],width:e[n],height:e[1-n]}}function MV(t,e,n,i,r,o){return{cx:t,cy:e,r0:n,r:i,startAngle:r,endAngle:o,clockwise:!0}}var SV=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.makeElOption=function(t,e,n,i,r){var o=n.axis,a=o.grid,s=i.get("type"),l=LV(a,o).getOtherAxis(o).getGlobalExtent(),u=o.toGlobalCoord(o.dataToCoord(e,!0));if(s&&"none"!==s){var c=mV(i),d=kV[s](o,u,l);d.style=c,t.graphicKey=d.type,t.pointer=d}var h=iI(a.model,n);bV(e,t,h,n,i,r)},e.prototype.getHandleTransform=function(t,e,n){var i=iI(e.axis.grid.model,e,{labelInside:!1});i.labelMargin=n.get(["handle","margin"]);var r=_V(e.axis,t,i);return{x:r[0],y:r[1],rotation:i.rotation+(i.labelDirection<0?Math.PI:0)}},e.prototype.updateHandleTransform=function(t,e,n,i){var r=n.axis,o=r.grid,a=r.getGlobalExtent(!0),s=LV(o,r).getOtherAxis(r).getGlobalExtent(),l="x"===r.dim?0:1,u=[t.x,t.y];u[l]+=e[l],u[l]=Math.min(a[1],u[l]),u[l]=Math.max(a[0],u[l]);var c=(s[1]+s[0])/2,d=[c,c];d[l]=u[l];var h=[{verticalAlign:"middle"},{align:"center"}];return{x:u[0],y:u[1],rotation:t.rotation,cursorPoint:d,tooltipOption:h[l]}},e}(pV);function LV(t,e){var n={};return n[e.dim+"AxisIndex"]=e.index,t.getCartesian(n)}var kV={line:function(t,e,n){var i=xV([e,n[0]],[e,n[1]],TV(t));return{type:"Line",subPixelOptimize:!0,shape:i}},shadow:function(t,e,n){var i=Math.max(1,t.getBandWidth()),r=n[1]-n[0];return{type:"Rect",shape:wV([e-i/2,n[0]],[i,r],TV(t))}}};function TV(t){return"x"===t.dim?0:1}var DV=SV,CV=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="axisPointer",e.defaultOption={show:"auto",zlevel:0,z:50,type:"line",snap:!1,triggerTooltip:!0,value:null,status:null,link:[],animation:null,animationDurationUpdate:200,lineStyle:{color:"#B9BEC9",width:1,type:"dashed"},shadowStyle:{color:"rgba(210,219,238,0.2)"},label:{show:!0,formatter:null,precision:"auto",margin:3,color:"#fff",padding:[5,7,5,7],backgroundColor:"auto",borderColor:null,borderWidth:0,borderRadius:3},handle:{show:!1,icon:"M10.7,11.9v-1.3H9.3v1.3c-4.9,0.3-8.8,4.4-8.8,9.4c0,5,3.9,9.1,8.8,9.4h1.3c4.9-0.3,8.8-4.4,8.8-9.4C19.5,16.3,15.6,12.2,10.7,11.9z M13.3,24.4H6.7v-1.2h6.6z M13.3,22H6.7v-1.2h6.6z M13.3,19.6H6.7v-1.2h6.6z",size:45,margin:50,color:"#333",shadowBlur:3,shadowColor:"#aaa",shadowOffsetX:0,shadowOffsetY:2,throttle:40}},e}(xp),IV=CV,AV=Io(),YV=R;function OV(t,e,n){if(!h.node){var i=e.getZr();AV(i).records||(AV(i).records={}),PV(i,e);var r=AV(i).records[t]||(AV(i).records[t]={});r.handler=n}}function PV(t,e){function n(n,i){t.on(n,(function(n){var r=HV(e);YV(AV(t).records,(function(t){t&&i(t,n,r.dispatchAction)})),EV(r.pendings,e)}))}AV(t).initialized||(AV(t).initialized=!0,n("click",G(RV,"click")),n("mousemove",G(RV,"mousemove")),n("globalout",NV))}function EV(t,e){var n,i=t.showTip.length,r=t.hideTip.length;i?n=t.showTip[i-1]:r&&(n=t.hideTip[r-1]),n&&(n.dispatchAction=null,e.dispatchAction(n))}function NV(t,e,n){t.handler("leave",null,n)}function RV(t,e,n,i){e.handler(t,n,i)}function HV(t){var e={showTip:[],hideTip:[]},n=function(i){var r=e[i.type];r?r.push(i):(i.dispatchAction=n,t.dispatchAction(i))};return{dispatchAction:n,pendings:e}}function jV(t,e){if(!h.node){var n=e.getZr(),i=(AV(n).records||{})[t];i&&(AV(n).records[t]=null)}}var zV=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=e.getComponent("tooltip"),r=t.get("triggerOn")||i&&i.get("triggerOn")||"mousemove|click";OV("axisPointer",n,(function(t,e,n){"none"!==r&&("leave"===t||r.indexOf(t)>=0)&&n({type:"updateAxisPointer",currTrigger:t,x:e&&e.offsetX,y:e&&e.offsetY})}))},e.prototype.remove=function(t,e){jV("axisPointer",e)},e.prototype.dispose=function(t,e){jV("axisPointer",e)},e.type="axisPointer",e}(Py),WV=zV;function BV(t,e){var n,i=[],r=t.seriesIndex;if(null==r||!(n=e.getSeriesByIndex(r)))return{point:[]};var o=n.getData(),a=Co(o,t);if(null==a||a<0||$(a))return{point:[]};var s=o.getItemGraphicEl(a),l=n.coordinateSystem;if(n.getTooltipPosition)i=n.getTooltipPosition(a)||[];else if(l&&l.dataToPoint)if(t.isStacked){var u=l.getBaseAxis(),c=l.getOtherAxis(u),d=c.dim,h=u.dim,f="x"===d||"radius"===d?1:0,p=o.mapDimension(h),m=[];m[f]=o.get(p,a),m[1-f]=o.get(o.getCalculationInfo("stackResultDimension"),a),i=l.dataToPoint(m)||[]}else i=l.dataToPoint(o.getValues(H(l.dimensions,(function(t){return o.mapDimension(t)})),a))||[];else if(s){var g=s.getBoundingRect().clone();g.applyTransform(s.transform),i=[g.x+g.width/2,g.y+g.height/2]}return{point:i,el:s}}var FV=Io();function VV(t,e,n){var i=t.currTrigger,r=[t.x,t.y],o=t,a=t.dispatchAction||V(n.dispatchAction,n),s=e.getComponent("axisPointer").coordSysAxesInfo;if(s){tG(r)&&(r=BV({seriesIndex:o.seriesIndex,dataIndex:o.dataIndex},e).point);var l=tG(r),u=o.axesInfo,c=s.axesInfo,d="leave"===i||tG(r),h={},f={},p={list:[],map:{}},m={showPointer:G(UV,f),showTooltip:G(qV,p)};R(s.coordSysMap,(function(t,e){var n=l||t.containPoint(r);R(s.coordSysAxesInfo[e],(function(t,e){var i=t.axis,o=KV(u,t);if(!d&&n&&(!u||o)){var a=o&&o.value;null!=a||l||(a=i.pointToData(r)),null!=a&&GV(t,a,m,!1,h)}}))}));var g={};return R(c,(function(t,e){var n=t.linkGroup;n&&!f[e]&&R(n.axesInfo,(function(e,i){var r=f[i];if(e!==t&&r){var o=r.value;n.mapper&&(o=t.axis.scale.parse(n.mapper(o,QV(e),QV(t)))),g[t.key]=o}}))})),R(g,(function(t,e){GV(c[e],t,m,!0,h)})),XV(f,c,h),ZV(p,r,t,a),JV(c,a,n),h}}function GV(t,e,n,i,r){var o=t.axis;if(!o.scale.isBlank()&&o.containData(e))if(t.involveSeries){var a=$V(e,t),s=a.payloadBatch,l=a.snapToValue;s[0]&&null==r.seriesIndex&&I(r,s[0]),!i&&t.snap&&o.containData(l)&&null!=l&&(e=l),n.showPointer(t,e,s),n.showTooltip(t,a,l)}else n.showPointer(t,e)}function $V(t,e){var n=e.axis,i=n.dim,r=t,o=[],a=Number.MAX_VALUE,s=-1;return R(e.seriesModels,(function(e,l){var u,c,d=e.getData().mapDimensionsAll(i);if(e.getAxisTooltipData){var h=e.getAxisTooltipData(d,t,n);c=h.dataIndices,u=h.nestestValue}else{if(c=e.getData().indicesOfNearest(d[0],t,"category"===n.type?.5:null),!c.length)return;u=e.getData().get(d[0],c[0])}if(null!=u&&isFinite(u)){var f=t-u,p=Math.abs(f);p<=a&&((p<a||f>=0&&s<0)&&(a=p,s=f,r=u,o.length=0),R(c,(function(t){o.push({seriesIndex:e.seriesIndex,dataIndexInside:t,dataIndex:e.getData().getRawIndex(t)})})))}})),{payloadBatch:o,snapToValue:r}}function UV(t,e,n,i){t[e.key]={value:n,payloadBatch:i}}function qV(t,e,n,i){var r=n.payloadBatch,o=e.axis,a=o.model,s=e.axisPointerModel;if(e.triggerTooltip&&r.length){var l=e.coordSys.model,u=EI(l),c=t.map[u];c||(c=t.map[u]={coordSysId:l.id,coordSysIndex:l.componentIndex,coordSysType:l.type,coordSysMainType:l.mainType,dataByAxis:[]},t.list.push(c)),c.dataByAxis.push({axisDim:o.dim,axisIndex:a.componentIndex,axisType:a.type,axisId:a.id,value:i,valueLabelOpt:{precision:s.get(["label","precision"]),formatter:s.get(["label","formatter"])},seriesDataIndices:r.slice()})}}function XV(t,e,n){var i=n.axesInfo=[];R(e,(function(e,n){var r=e.axisPointerModel.option,o=t[n];o?(!e.useHandle&&(r.status="show"),r.value=o.value,r.seriesDataIndices=(o.payloadBatch||[]).slice()):!e.useHandle&&(r.status="hide"),"show"===r.status&&i.push({axisDim:e.axis.dim,axisIndex:e.axis.model.componentIndex,value:r.value})}))}function ZV(t,e,n,i){if(!tG(e)&&t.list.length){var r=((t.list[0].dataByAxis[0]||{}).seriesDataIndices||[])[0]||{};i({type:"showTip",escapeConnect:!0,x:e[0],y:e[1],tooltipOption:n.tooltipOption,position:n.position,dataIndexInside:r.dataIndexInside,dataIndex:r.dataIndex,seriesIndex:r.seriesIndex,dataByCoordSys:t.list})}else i({type:"hideTip"})}function JV(t,e,n){var i=n.getZr(),r="axisPointerLastHighlights",o=FV(i)[r]||{},a=FV(i)[r]={};R(t,(function(t,e){var n=t.axisPointerModel.option;"show"===n.status&&R(n.seriesDataIndices,(function(t){var e=t.seriesIndex+" | "+t.dataIndex;a[e]=t}))}));var s=[],l=[];R(o,(function(t,e){!a[e]&&l.push(t)})),R(a,(function(t,e){!o[e]&&s.push(t)})),l.length&&n.dispatchAction({type:"downplay",escapeConnect:!0,notBlur:!0,batch:l}),s.length&&n.dispatchAction({type:"highlight",escapeConnect:!0,notBlur:!0,batch:s})}function KV(t,e){for(var n=0;n<(t||[]).length;n++){var i=t[n];if(e.axis.dim===i.axisDim&&e.axis.model.componentIndex===i.axisIndex)return i}}function QV(t){var e=t.axis.model,n={},i=n.axisDim=t.axis.dim;return n.axisIndex=n[i+"AxisIndex"]=e.componentIndex,n.axisName=n[i+"AxisName"]=e.name,n.axisId=n[i+"AxisId"]=e.id,n}function tG(t){return!t||null==t[0]||isNaN(t[0])||null==t[1]||isNaN(t[1])}function eG(t){HI.registerAxisPointerClass("CartesianAxisPointer",DV),t.registerComponentModel(IV),t.registerComponentView(WV),t.registerPreprocessor((function(t){if(t){(!t.axisPointer||0===t.axisPointer.length)&&(t.axisPointer={});var e=t.axisPointer.link;e&&!$(e)&&(t.axisPointer.link=[e])}})),t.registerProcessor(t.PRIORITY.PROCESSOR.STATISTIC,(function(t,e){t.getComponent("axisPointer").coordSysAxesInfo=LI(t,e)})),t.registerAction({type:"updateAxisPointer",event:"updateAxisPointer",update:":updateAxisPointer"},VV)}function nG(t){vM(ZI),vM(eG)}var iG=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.makeElOption=function(t,e,n,i,r){var o=n.axis;"angle"===o.dim&&(this.animationThreshold=Math.PI/18);var a=o.polar,s=a.getOtherAxis(o),l=s.getExtent(),u=o.dataToCoord(e),c=i.get("type");if(c&&"none"!==c){var d=mV(i),h=oG[c](o,a,u,l);h.style=d,t.graphicKey=h.type,t.pointer=h}var f=i.get(["label","margin"]),p=rG(e,n,i,a,f);gV(t,n,i,r,p)},e}(pV);function rG(t,e,n,i,r){var o=e.axis,a=o.dataToCoord(t),s=i.getAngleAxis().getExtent()[0];s=s/180*Math.PI;var l,u,c,d=i.getRadiusAxis().getExtent();if("radius"===o.dim){var h=bi();Li(h,h,s),Si(h,h,[i.cx,i.cy]),l=sh([a,-r],h);var f=e.getModel("axisLabel").get("rotate")||0,p=SI.innerTextLayout(s,f*Math.PI/180,-1);u=p.textAlign,c=p.textVerticalAlign}else{var m=d[1];l=i.coordToPoint([m+r,a]);var g=i.cx,y=i.cy;u=Math.abs(l[0]-g)/m<.3?"center":l[0]>g?"left":"right",c=Math.abs(l[1]-y)/m<.3?"middle":l[1]>y?"top":"bottom"}return{position:l,align:u,verticalAlign:c}}var oG={line:function(t,e,n,i){return"angle"===t.dim?{type:"Line",shape:xV(e.coordToPoint([i[0],n]),e.coordToPoint([i[1],n]))}:{type:"Circle",shape:{cx:e.cx,cy:e.cy,r:n}}},shadow:function(t,e,n,i){var r=Math.max(1,t.getBandWidth()),o=Math.PI/180;return"angle"===t.dim?{type:"Sector",shape:MV(e.cx,e.cy,i[0],i[1],(-n-r/2)*o,(r/2-n)*o)}:{type:"Sector",shape:MV(e.cx,e.cy,n-r/2,n+r/2,0,2*Math.PI)}}},aG=iG,sG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.findAxisModel=function(t){var e,n=this.ecModel;return n.eachComponent(t,(function(t){t.getCoordSysModel()===this&&(e=t)}),this),e},e.type="polar",e.dependencies=["radiusAxis","angleAxis"],e.defaultOption={zlevel:0,z:0,center:["50%","50%"],radius:"80%"},e}(xp),lG=sG,uG=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.getCoordSysModel=function(){return this.getReferringComponents("polar",Po).models[0]},e.type="polarAxis",e}(xp);E(uG,fD);var cG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="angleAxis",e}(uG),dG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="radiusAxis",e}(uG),hG=function(t){function e(e,n){return t.call(this,"radius",e,n)||this}return o(e,t),e.prototype.pointToData=function(t,e){return this.polar.pointToData(t,e)["radius"===this.dim?0:1]},e}(tI);hG.prototype.dataToRadius=tI.prototype.dataToCoord,hG.prototype.radiusToData=tI.prototype.coordToData;var fG=hG,pG=Io(),mG=function(t){function e(e,n){return t.call(this,"angle",e,n||[0,360])||this}return o(e,t),e.prototype.pointToData=function(t,e){return this.polar.pointToData(t,e)["radius"===this.dim?0:1]},e.prototype.calculateCategoryInterval=function(){var t=this,e=t.getLabelModel(),n=t.scale,i=n.getExtent(),r=n.count();if(i[1]-i[0]<1)return 0;var o=i[0],a=t.dataToCoord(o+1)-t.dataToCoord(o),s=Math.abs(a),l=rr(null==o?"":o+"",e.getFont(),"center","top"),u=Math.max(l.height,7),c=u/s;isNaN(c)&&(c=1/0);var d=Math.max(0,Math.floor(c)),h=pG(t.model),f=h.lastAutoInterval,p=h.lastTickCount;return null!=f&&null!=p&&Math.abs(f-d)<=1&&Math.abs(p-r)<=1&&f>d?d=f:(h.lastTickCount=r,h.lastAutoInterval=d),d},e}(tI);mG.prototype.dataToAngle=tI.prototype.dataToCoord,mG.prototype.angleToData=tI.prototype.coordToData;var gG=mG,yG=function(){function t(t){this.dimensions=["radius","angle"],this.type="polar",this.cx=0,this.cy=0,this._radiusAxis=new fG,this._angleAxis=new gG,this.axisPointerEnabled=!0,this.name=t||"",this._radiusAxis.polar=this._angleAxis.polar=this}return t.prototype.containPoint=function(t){var e=this.pointToCoord(t);return this._radiusAxis.contain(e[0])&&this._angleAxis.contain(e[1])},t.prototype.containData=function(t){return this._radiusAxis.containData(t[0])&&this._angleAxis.containData(t[1])},t.prototype.getAxis=function(t){var e="_"+t+"Axis";return this[e]},t.prototype.getAxes=function(){return[this._radiusAxis,this._angleAxis]},t.prototype.getAxesByScale=function(t){var e=[],n=this._angleAxis,i=this._radiusAxis;return n.scale.type===t&&e.push(n),i.scale.type===t&&e.push(i),e},t.prototype.getAngleAxis=function(){return this._angleAxis},t.prototype.getRadiusAxis=function(){return this._radiusAxis},t.prototype.getOtherAxis=function(t){var e=this._angleAxis;return t===e?this._radiusAxis:e},t.prototype.getBaseAxis=function(){return this.getAxesByScale("ordinal")[0]||this.getAxesByScale("time")[0]||this.getAngleAxis()},t.prototype.getTooltipAxes=function(t){var e=null!=t&&"auto"!==t?this.getAxis(t):this.getBaseAxis();return{baseAxes:[e],otherAxes:[this.getOtherAxis(e)]}},t.prototype.dataToPoint=function(t,e){return this.coordToPoint([this._radiusAxis.dataToRadius(t[0],e),this._angleAxis.dataToAngle(t[1],e)])},t.prototype.pointToData=function(t,e){var n=this.pointToCoord(t);return[this._radiusAxis.radiusToData(n[0],e),this._angleAxis.angleToData(n[1],e)]},t.prototype.pointToCoord=function(t){var e=t[0]-this.cx,n=t[1]-this.cy,i=this.getAngleAxis(),r=i.getExtent(),o=Math.min(r[0],r[1]),a=Math.max(r[0],r[1]);i.inverse?o=a-360:a=o+360;var s=Math.sqrt(e*e+n*n);e/=s,n/=s;var l=Math.atan2(-n,e)/Math.PI*180,u=l<o?1:-1;while(l<o||l>a)l+=360*u;return[s,l]},t.prototype.coordToPoint=function(t){var e=t[0],n=t[1]/180*Math.PI,i=Math.cos(n)*e+this.cx,r=-Math.sin(n)*e+this.cy;return[i,r]},t.prototype.getArea=function(){var t=this.getAngleAxis(),e=this.getRadiusAxis(),n=e.getExtent().slice();n[0]>n[1]&&n.reverse();var i=t.getExtent(),r=Math.PI/180;return{cx:this.cx,cy:this.cy,r0:n[0],r:n[1],startAngle:-i[0]*r,endAngle:-i[1]*r,clockwise:t.inverse,contain:function(t,e){var n=t-this.cx,i=e-this.cy,r=n*n+i*i,o=this.r,a=this.r0;return r<=o*o&&r>=a*a}}},t.prototype.convertToPixel=function(t,e,n){var i=vG(e);return i===this?this.dataToPoint(n):null},t.prototype.convertFromPixel=function(t,e,n){var i=vG(e);return i===this?this.pointToData(n):null},t}();function vG(t){var e=t.seriesModel,n=t.polarModel;return n&&n.coordinateSystem||e&&e.coordinateSystem}var _G=yG;function bG(t,e,n){var i=e.get("center"),r=n.getWidth(),o=n.getHeight();t.cx=Nr(i[0],r),t.cy=Nr(i[1],o);var a=t.getRadiusAxis(),s=Math.min(r,o)/2,l=e.get("radius");null==l?l=[0,"100%"]:$(l)||(l=[0,l]);var u=[Nr(l[0],s),Nr(l[1],s)];a.inverse?a.setExtent(u[1],u[0]):a.setExtent(u[0],u[1])}function xG(t,e){var n=this,i=n.getAngleAxis(),r=n.getRadiusAxis();if(i.scale.setExtent(1/0,-1/0),r.scale.setExtent(1/0,-1/0),t.eachSeries((function(t){if(t.coordinateSystem===n){var e=t.getData();R(TC(e,"radius"),(function(t){r.scale.unionExtentFromData(e,t)})),R(TC(e,"angle"),(function(t){i.scale.unionExtentFromData(e,t)}))}})),vC(i.scale,i.model),vC(r.scale,r.model),"category"===i.type&&!i.onBand){var o=i.getExtent(),a=360/i.scale.count();i.inverse?o[1]+=a:o[1]-=a,i.setExtent(o[0],o[1])}}function wG(t){return"angleAxis"===t.mainType}function MG(t,e){if(t.type=e.get("type"),t.scale=_C(e),t.onBand=e.get("boundaryGap")&&"category"===t.type,t.inverse=e.get("inverse"),wG(e)){t.inverse=t.inverse!==e.get("clockwise");var n=e.get("startAngle");t.setExtent(n,n+(t.inverse?-360:360))}e.axis=t,t.model=e}var SG={dimensions:_G.prototype.dimensions,create:function(t,e){var n=[];return t.eachComponent("polar",(function(t,i){var r=new _G(i+"");r.update=xG;var o=r.getRadiusAxis(),a=r.getAngleAxis(),s=t.findAxisModel("radiusAxis"),l=t.findAxisModel("angleAxis");MG(o,s),MG(a,l),bG(r,t,e),n.push(r),t.coordinateSystem=r,r.model=t})),t.eachSeries((function(t){if("polar"===t.get("coordinateSystem")){var e=t.getReferringComponents("polar",Po).models[0];0,t.coordinateSystem=e.coordinateSystem}})),n}},LG=SG,kG=["axisLine","axisLabel","axisTick","minorTick","splitLine","minorSplitLine","splitArea"];function TG(t,e,n){e[1]>e[0]&&(e=e.slice().reverse());var i=t.coordToPoint([e[0],n]),r=t.coordToPoint([e[1],n]);return{x1:i[0],y1:i[1],x2:r[0],y2:r[1]}}function DG(t){var e=t.getRadiusAxis();return e.inverse?0:1}function CG(t){var e=t[0],n=t[t.length-1];e&&n&&Math.abs(Math.abs(e.coord-n.coord)-360)<1e-4&&t.pop()}var IG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="PolarAxisPointer",n}return o(e,t),e.prototype.render=function(t,e){if(this.group.removeAll(),t.get("show")){var n=t.axis,i=n.polar,r=i.getRadiusAxis().getExtent(),o=n.getTicksCoords(),a=n.getMinorTicksCoords(),s=H(n.getViewLabels(),(function(t){t=T(t);var e=n.scale,i="ordinal"===e.type?e.getRawOrdinalNumber(t.tickValue):t.tickValue;return t.coord=n.dataToCoord(i),t}));CG(s),CG(o),R(kG,(function(e){!t.get([e,"show"])||n.scale.isBlank()&&"axisLine"!==e||AG[e](this.group,t,i,o,a,r,s)}),this)}},e.type="angleAxis",e}(HI),AG={axisLine:function(t,e,n,i,r,o){var a,s=e.getModel(["axisLine","lineStyle"]),l=DG(n),u=l?0:1;a=0===o[u]?new Dc({shape:{cx:n.cx,cy:n.cy,r:o[l]},style:s.getLineStyle(),z2:1,silent:!0}):new Jc({shape:{cx:n.cx,cy:n.cy,r:o[l],r0:o[u]},style:s.getLineStyle(),z2:1,silent:!0}),a.style.fill=null,t.add(a)},axisTick:function(t,e,n,i,r,o){var a=e.getModel("axisTick"),s=(a.get("inside")?-1:1)*a.get("length"),l=o[DG(n)],u=H(i,(function(t){return new dd({shape:TG(n,[l,l+s],t.coord)})}));t.add(qd(u,{style:A(a.getModel("lineStyle").getLineStyle(),{stroke:e.get(["axisLine","lineStyle","color"])})}))},minorTick:function(t,e,n,i,r,o){if(r.length){for(var a=e.getModel("axisTick"),s=e.getModel("minorTick"),l=(a.get("inside")?-1:1)*s.get("length"),u=o[DG(n)],c=[],d=0;d<r.length;d++)for(var h=0;h<r[d].length;h++)c.push(new dd({shape:TG(n,[u,u+l],r[d][h].coord)}));t.add(qd(c,{style:A(s.getModel("lineStyle").getLineStyle(),A(a.getLineStyle(),{stroke:e.get(["axisLine","lineStyle","color"])}))}))}},axisLabel:function(t,e,n,i,r,o,a){var s=e.getCategories(!0),l=e.getModel("axisLabel"),u=l.get("margin"),c=e.get("triggerEvent");R(a,(function(i,r){var a=l,d=i.tickValue,h=o[DG(n)],f=n.coordToPoint([h+u,i.coord]),p=n.cx,m=n.cy,g=Math.abs(f[0]-p)/h<.3?"center":f[0]>p?"left":"right",y=Math.abs(f[1]-m)/h<.3?"middle":f[1]>m?"top":"bottom";if(s&&s[d]){var v=s[d];J(v)&&v.textStyle&&(a=new qh(v.textStyle,l,l.ecModel))}var _=new Gl({silent:SI.isLabelSilent(e),style:Lh(a,{x:f[0],y:f[1],fill:a.getTextColor()||e.get(["axisLine","lineStyle","color"]),text:i.formattedLabel,align:g,verticalAlign:y})});if(t.add(_),c){var b=SI.makeAxisEventDataBase(e);b.targetType="axisLabel",b.value=i.rawLabel,$l(_).eventData=b}}),this)},splitLine:function(t,e,n,i,r,o){var a=e.getModel("splitLine"),s=a.getModel("lineStyle"),l=s.get("color"),u=0;l=l instanceof Array?l:[l];for(var c=[],d=0;d<i.length;d++){var h=u++%l.length;c[h]=c[h]||[],c[h].push(new dd({shape:TG(n,o,i[d].coord)}))}for(d=0;d<c.length;d++)t.add(qd(c[d],{style:A({stroke:l[d%l.length]},s.getLineStyle()),silent:!0,z:e.get("z")}))},minorSplitLine:function(t,e,n,i,r,o){if(r.length){for(var a=e.getModel("minorSplitLine"),s=a.getModel("lineStyle"),l=[],u=0;u<r.length;u++)for(var c=0;c<r[u].length;c++)l.push(new dd({shape:TG(n,o,r[u][c].coord)}));t.add(qd(l,{style:s.getLineStyle(),silent:!0,z:e.get("z")}))}},splitArea:function(t,e,n,i,r,o){if(i.length){var a=e.getModel("splitArea"),s=a.getModel("areaStyle"),l=s.get("color"),u=0;l=l instanceof Array?l:[l];for(var c=[],d=Math.PI/180,h=-i[0].coord*d,f=Math.min(o[0],o[1]),p=Math.max(o[0],o[1]),m=e.get("clockwise"),g=1,y=i.length;g<=y;g++){var v=g===y?i[0].coord:i[g].coord,_=u++%l.length;c[_]=c[_]||[],c[_].push(new qc({shape:{cx:n.cx,cy:n.cy,r0:f,r:p,startAngle:h,endAngle:-v*d,clockwise:m},silent:!0})),h=-v*d}for(g=0;g<c.length;g++)t.add(qd(c[g],{style:A({fill:l[g%l.length]},s.getAreaStyle()),silent:!0}))}}},YG=IG,OG=["axisLine","axisTickLabel","axisName"],PG=["splitLine","splitArea","minorSplitLine"],EG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="PolarAxisPointer",n}return o(e,t),e.prototype.render=function(t,e){if(this.group.removeAll(),t.get("show")){var n=this._axisGroup,i=this._axisGroup=new Mr;this.group.add(i);var r=t.axis,o=r.polar,a=o.getAngleAxis(),s=r.getTicksCoords(),l=r.getMinorTicksCoords(),u=a.getExtent()[0],c=r.getExtent(),d=RG(o,t,u),h=new SI(t,d);R(OG,h.add,h),i.add(h.getGroup()),dh(n,i,t),R(PG,(function(e){t.get([e,"show"])&&!r.scale.isBlank()&&NG[e](this.group,t,o,u,c,s,l)}),this)}},e.type="radiusAxis",e}(HI),NG={splitLine:function(t,e,n,i,r,o){var a=e.getModel("splitLine"),s=a.getModel("lineStyle"),l=s.get("color"),u=0;l=l instanceof Array?l:[l];for(var c=[],d=0;d<o.length;d++){var h=u++%l.length;c[h]=c[h]||[],c[h].push(new Dc({shape:{cx:n.cx,cy:n.cy,r:o[d].coord}}))}for(d=0;d<c.length;d++)t.add(qd(c[d],{style:A({stroke:l[d%l.length],fill:null},s.getLineStyle()),silent:!0}))},minorSplitLine:function(t,e,n,i,r,o,a){if(a.length){for(var s=e.getModel("minorSplitLine"),l=s.getModel("lineStyle"),u=[],c=0;c<a.length;c++)for(var d=0;d<a[c].length;d++)u.push(new Dc({shape:{cx:n.cx,cy:n.cy,r:a[c][d].coord}}));t.add(qd(u,{style:A({fill:null},l.getLineStyle()),silent:!0}))}},splitArea:function(t,e,n,i,r,o){if(o.length){var a=e.getModel("splitArea"),s=a.getModel("areaStyle"),l=s.get("color"),u=0;l=l instanceof Array?l:[l];for(var c=[],d=o[0].coord,h=1;h<o.length;h++){var f=u++%l.length;c[f]=c[f]||[],c[f].push(new qc({shape:{cx:n.cx,cy:n.cy,r0:d,r:o[h].coord,startAngle:0,endAngle:2*Math.PI},silent:!0})),d=o[h].coord}for(h=0;h<c.length;h++)t.add(qd(c[h],{style:A({fill:l[h%l.length]},s.getAreaStyle()),silent:!0}))}}};function RG(t,e,n){return{position:[t.cx,t.cy],rotation:n/180*Math.PI,labelDirection:-1,tickDirection:-1,nameDirection:1,labelRotate:e.getModel("axisLabel").get("rotate"),z2:1}}var HG=EG;function jG(t){return t.get("stack")||"__ec_stack_"+t.seriesIndex}function zG(t,e){return e.dim+t.model.componentIndex}function WG(t,e,n){var i={},r=BG(z(e.getSeriesByType(t),(function(t){return!e.isSeriesFiltered(t)&&t.coordinateSystem&&"polar"===t.coordinateSystem.type})));e.eachSeriesByType(t,(function(t){if("polar"===t.coordinateSystem.type){var e=t.getData(),n=t.coordinateSystem,o=n.getBaseAxis(),a=zG(n,o),s=jG(t),l=r[a][s],u=l.offset,c=l.width,d=n.getOtherAxis(o),h=t.coordinateSystem.cx,f=t.coordinateSystem.cy,p=t.get("barMinHeight")||0,m=t.get("barMinAngle")||0;i[s]=i[s]||[];for(var g=e.mapDimension(d.dim),y=e.mapDimension(o.dim),v=RL(e,g),_="radius"!==o.dim||!t.get("roundCap",!0),b=d.dataToCoord(0),x=0,w=e.count();x<w;x++){var M=e.get(g,x),S=e.get(y,x),L=M>=0?"p":"n",k=b;v&&(i[s][S]||(i[s][S]={p:b,n:b}),k=i[s][S][L]);var T=void 0,D=void 0,C=void 0,I=void 0;if("radius"===d.dim){var A=d.dataToCoord(M)-b,Y=o.dataToCoord(S);Math.abs(A)<p&&(A=(A<0?-1:1)*p),T=k,D=k+A,C=Y-u,I=C-c,v&&(i[s][S][L]=D)}else{var O=d.dataToCoord(M,_)-b,P=o.dataToCoord(S);Math.abs(O)<m&&(O=(O<0?-1:1)*m),T=P+u,D=T+c,C=k,I=k+O,v&&(i[s][S][L]=I)}e.setItemLayout(x,{cx:h,cy:f,r0:T,r:D,startAngle:-C*Math.PI/180,endAngle:-I*Math.PI/180})}}}))}function BG(t){var e={};R(t,(function(t,n){var i=t.getData(),r=t.coordinateSystem,o=r.getBaseAxis(),a=zG(r,o),s=o.getExtent(),l="category"===o.type?o.getBandWidth():Math.abs(s[1]-s[0])/i.count(),u=e[a]||{bandWidth:l,remainedWidth:l,autoWidthCount:0,categoryGap:"20%",gap:"30%",stacks:{}},c=u.stacks;e[a]=u;var d=jG(t);c[d]||u.autoWidthCount++,c[d]=c[d]||{width:0,maxWidth:0};var h=Nr(t.get("barWidth"),l),f=Nr(t.get("barMaxWidth"),l),p=t.get("barGap"),m=t.get("barCategoryGap");h&&!c[d].width&&(h=Math.min(u.remainedWidth,h),c[d].width=h,u.remainedWidth-=h),f&&(c[d].maxWidth=f),null!=p&&(u.gap=p),null!=m&&(u.categoryGap=m)}));var n={};return R(e,(function(t,e){n[e]={};var i=t.stacks,r=t.bandWidth,o=Nr(t.categoryGap,r),a=Nr(t.gap,1),s=t.remainedWidth,l=t.autoWidthCount,u=(s-o)/(l+(l-1)*a);u=Math.max(u,0),R(i,(function(t,e){var n=t.maxWidth;n&&n<u&&(n=Math.min(n,s),t.width&&(n=Math.min(n,t.width)),s-=n,t.width=n,l--)})),u=(s-o)/(l+(l-1)*a),u=Math.max(u,0);var c,d=0;R(i,(function(t,e){t.width||(t.width=u),c=t,d+=t.width*(1+a)})),c&&(d-=c.width*a);var h=-d/2;R(i,(function(t,i){n[e][i]=n[e][i]||{offset:h,width:t.width},h+=t.width*(1+a)}))})),n}var FG=WG,VG={startAngle:90,clockwise:!0,splitNumber:12,axisLabel:{rotate:0}},GG={splitNumber:5},$G=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="polar",e}(Py);function UG(t){vM(eG),HI.registerAxisPointerClass("PolarAxisPointer",aG),t.registerCoordinateSystem("polar",LG),t.registerComponentModel(lG),t.registerComponentView($G),LD(t,"angle",cG,VG),LD(t,"radius",dG,GG),t.registerComponentView(YG),t.registerComponentView(HG),t.registerLayout(G(FG,"bar"))}function qG(t,e){e=e||{};var n=t.coordinateSystem,i=t.axis,r={},o=i.position,a=i.orient,s=n.getRect(),l=[s.x,s.x+s.width,s.y,s.y+s.height],u={horizontal:{top:l[2],bottom:l[3]},vertical:{left:l[0],right:l[1]}};r.position=["vertical"===a?u.vertical[o]:l[0],"horizontal"===a?u.horizontal[o]:l[3]];var c={horizontal:0,vertical:1};r.rotation=Math.PI/2*c[a];var d={top:-1,bottom:1,right:1,left:-1};r.labelDirection=r.tickDirection=r.nameDirection=d[o],t.get(["axisTick","inside"])&&(r.tickDirection=-r.tickDirection),ot(e.labelInside,t.get(["axisLabel","inside"]))&&(r.labelDirection=-r.labelDirection);var h=e.rotate;return null==h&&(h=t.get(["axisLabel","rotate"])),r.labelRotation="top"===o?-h:h,r.z2=1,r}var XG=["axisLine","axisTickLabel","axisName"],ZG=["splitArea","splitLine"],JG=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.axisPointerClass="SingleAxisPointer",n}return o(e,t),e.prototype.render=function(e,n,i,r){var o=this.group;o.removeAll();var a=this._axisGroup;this._axisGroup=new Mr;var s=qG(e),l=new SI(e,s);R(XG,l.add,l),o.add(this._axisGroup),o.add(l.getGroup()),R(ZG,(function(t){e.get([t,"show"])&&KG[t](this,this.group,this._axisGroup,e)}),this),dh(a,this._axisGroup,e),t.prototype.render.call(this,e,n,i,r)},e.prototype.remove=function(){WI(this)},e.type="singleAxis",e}(HI),KG={splitLine:function(t,e,n,i){var r=i.axis;if(!r.scale.isBlank()){var o=i.getModel("splitLine"),a=o.getModel("lineStyle"),s=a.get("color");s=s instanceof Array?s:[s];for(var l=i.coordinateSystem.getRect(),u=r.isHorizontal(),c=[],d=0,h=r.getTicksCoords({tickModel:o}),f=[],p=[],m=0;m<h.length;++m){var g=r.toGlobalCoord(h[m].coord);u?(f[0]=g,f[1]=l.y,p[0]=g,p[1]=l.y+l.height):(f[0]=l.x,f[1]=g,p[0]=l.x+l.width,p[1]=g);var y=d++%s.length;c[y]=c[y]||[],c[y].push(new dd({subPixelOptimize:!0,shape:{x1:f[0],y1:f[1],x2:p[0],y2:p[1]},silent:!0}))}var v=a.getLineStyle(["color"]);for(m=0;m<c.length;++m)e.add(qd(c[m],{style:A({stroke:s[m%s.length]},v),silent:!0}))}},splitArea:function(t,e,n,i){zI(t,n,i,i)}},QG=JG,t$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.getCoordSysModel=function(){return this},e.type="singleAxis",e.layoutMode="box",e.defaultOption={left:"5%",top:"5%",right:"5%",bottom:"5%",type:"value",position:"bottom",orient:"horizontal",axisLine:{show:!0,lineStyle:{width:1,type:"solid"}},tooltip:{show:!0},axisTick:{show:!0,length:6,lineStyle:{width:1}},axisLabel:{show:!0,interval:"auto"},splitLine:{show:!0,lineStyle:{type:"dashed",opacity:.2}}},e}(xp);E(t$,fD.prototype);var e$=t$,n$=function(t){function e(e,n,i,r,o){var a=t.call(this,e,n,i)||this;return a.type=r||"value",a.position=o||"bottom",a}return o(e,t),e.prototype.isHorizontal=function(){var t=this.position;return"top"===t||"bottom"===t},e.prototype.pointToData=function(t,e){return this.coordinateSystem.pointToData(t)[0]},e}(tI),i$=n$,r$=function(){function t(t,e,n){this.type="single",this.dimension="single",this.dimensions=["single"],this.axisPointerEnabled=!0,this.model=t,this._init(t,e,n)}return t.prototype._init=function(t,e,n){var i=this.dimension,r=new i$(i,_C(t),[0,0],t.get("type"),t.get("position")),o="category"===r.type;r.onBand=o&&t.get("boundaryGap"),r.inverse=t.get("inverse"),r.orient=t.get("orient"),t.axis=r,r.model=t,r.coordinateSystem=this,this._axis=r},t.prototype.update=function(t,e){t.eachSeries((function(t){if(t.coordinateSystem===this){var e=t.getData();R(e.mapDimensionsAll(this.dimension),(function(t){this._axis.scale.unionExtentFromData(e,t)}),this),vC(this._axis.scale,this._axis.model)}}),this)},t.prototype.resize=function(t,e){this._rect=dp({left:t.get("left"),top:t.get("top"),right:t.get("right"),bottom:t.get("bottom"),width:t.get("width"),height:t.get("height")},{width:e.getWidth(),height:e.getHeight()}),this._adjustAxis()},t.prototype.getRect=function(){return this._rect},t.prototype._adjustAxis=function(){var t=this._rect,e=this._axis,n=e.isHorizontal(),i=n?[0,t.width]:[0,t.height],r=e.reverse?1:0;e.setExtent(i[r],i[1-r]),this._updateAxisTransform(e,n?t.x:t.y)},t.prototype._updateAxisTransform=function(t,e){var n=t.getExtent(),i=n[0]+n[1],r=t.isHorizontal();t.toGlobalCoord=r?function(t){return t+e}:function(t){return i-t+e},t.toLocalCoord=r?function(t){return t-e}:function(t){return i-t+e}},t.prototype.getAxis=function(){return this._axis},t.prototype.getBaseAxis=function(){return this._axis},t.prototype.getAxes=function(){return[this._axis]},t.prototype.getTooltipAxes=function(){return{baseAxes:[this.getAxis()],otherAxes:[]}},t.prototype.containPoint=function(t){var e=this.getRect(),n=this.getAxis(),i=n.orient;return"horizontal"===i?n.contain(n.toLocalCoord(t[0]))&&t[1]>=e.y&&t[1]<=e.y+e.height:n.contain(n.toLocalCoord(t[1]))&&t[0]>=e.y&&t[0]<=e.y+e.height},t.prototype.pointToData=function(t){var e=this.getAxis();return[e.coordToData(e.toLocalCoord(t["horizontal"===e.orient?0:1]))]},t.prototype.dataToPoint=function(t){var e=this.getAxis(),n=this.getRect(),i=[],r="horizontal"===e.orient?0:1;return t instanceof Array&&(t=t[0]),i[r]=e.toGlobalCoord(e.dataToCoord(+t)),i[1-r]=0===r?n.y+n.height/2:n.x+n.width/2,i},t.prototype.convertToPixel=function(t,e,n){var i=o$(e);return i===this?this.dataToPoint(n):null},t.prototype.convertFromPixel=function(t,e,n){var i=o$(e);return i===this?this.pointToData(n):null},t}();function o$(t){var e=t.seriesModel,n=t.singleAxisModel;return n&&n.coordinateSystem||e&&e.coordinateSystem}var a$=r$;function s$(t,e){var n=[];return t.eachComponent("singleAxis",(function(i,r){var o=new a$(i,t,e);o.name="single_"+r,o.resize(i,e),i.coordinateSystem=o,n.push(o)})),t.eachSeries((function(t){if("singleAxis"===t.get("coordinateSystem")){var e=t.getReferringComponents("singleAxis",Po).models[0];t.coordinateSystem=e&&e.coordinateSystem}})),n}var l$={create:s$,dimensions:a$.prototype.dimensions},u$=l$,c$=["x","y"],d$=["width","height"],h$=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.makeElOption=function(t,e,n,i,r){var o=n.axis,a=o.coordinateSystem,s=m$(a,1-p$(o)),l=a.dataToPoint(e)[0],u=i.get("type");if(u&&"none"!==u){var c=mV(i),d=f$[u](o,l,s);d.style=c,t.graphicKey=d.type,t.pointer=d}var h=qG(n);bV(e,t,h,n,i,r)},e.prototype.getHandleTransform=function(t,e,n){var i=qG(e,{labelInside:!1});i.labelMargin=n.get(["handle","margin"]);var r=_V(e.axis,t,i);return{x:r[0],y:r[1],rotation:i.rotation+(i.labelDirection<0?Math.PI:0)}},e.prototype.updateHandleTransform=function(t,e,n,i){var r=n.axis,o=r.coordinateSystem,a=p$(r),s=m$(o,a),l=[t.x,t.y];l[a]+=e[a],l[a]=Math.min(s[1],l[a]),l[a]=Math.max(s[0],l[a]);var u=m$(o,1-a),c=(u[1]+u[0])/2,d=[c,c];return d[a]=l[a],{x:l[0],y:l[1],rotation:t.rotation,cursorPoint:d,tooltipOption:{verticalAlign:"middle"}}},e}(pV),f$={line:function(t,e,n){var i=xV([e,n[0]],[e,n[1]],p$(t));return{type:"Line",subPixelOptimize:!0,shape:i}},shadow:function(t,e,n){var i=t.getBandWidth(),r=n[1]-n[0];return{type:"Rect",shape:wV([e-i/2,n[0]],[i,r],p$(t))}}};function p$(t){return t.isHorizontal()?0:1}function m$(t,e){var n=t.getRect();return[n[c$[e]],n[c$[e]]+n[d$[e]]]}var g$=h$,y$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="single",e}(Py);function v$(t){vM(eG),HI.registerAxisPointerClass("SingleAxisPointer",g$),t.registerComponentView(y$),t.registerComponentView(QG),t.registerComponentModel(e$),LD(t,"single",e$,e$.defaultOption),t.registerCoordinateSystem("single",u$)}var _$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(e,n,i){var r=gp(e);t.prototype.init.apply(this,arguments),b$(e,r)},e.prototype.mergeOption=function(e){t.prototype.mergeOption.apply(this,arguments),b$(this.option,e)},e.prototype.getCellSize=function(){return this.option.cellSize},e.type="calendar",e.defaultOption={zlevel:0,z:2,left:80,top:60,cellSize:20,orient:"horizontal",splitLine:{show:!0,lineStyle:{color:"#000",width:1,type:"solid"}},itemStyle:{color:"#fff",borderWidth:1,borderColor:"#ccc"},dayLabel:{show:!0,firstDay:0,position:"start",margin:"50%",nameMap:"en",color:"#000"},monthLabel:{show:!0,position:"start",margin:5,align:"center",nameMap:"en",formatter:null,color:"#000"},yearLabel:{show:!0,position:null,margin:30,formatter:null,color:"#ccc",fontFamily:"sans-serif",fontWeight:"bolder",fontSize:20}},e}(xp);function b$(t,e){var n,i=t.cellSize;n=$(i)?i:t.cellSize=[i,i],1===n.length&&(n[1]=n[0]);var r=H([0,1],(function(t){return fp(e,t)&&(n[t]="auto"),null!=n[t]&&"auto"!==n[t]}));mp(t,e,{type:"box",ignoreSize:r})}var x$=_$,w$={EN:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],CN:["一月","二月","三月","四月","五月","六月","七月","八月","九月","十月","十一月","十二月"]},M$={EN:["S","M","T","W","T","F","S"],CN:["日","一","二","三","四","五","六"]},S$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){var i=this.group;i.removeAll();var r=t.coordinateSystem,o=r.getRangeInfo(),a=r.getOrient();this._renderDayRect(t,o,i),this._renderLines(t,o,a,i),this._renderYearText(t,o,a,i),this._renderMonthText(t,a,i),this._renderWeekText(t,o,a,i)},e.prototype._renderDayRect=function(t,e,n){for(var i=t.coordinateSystem,r=t.getModel("itemStyle").getItemStyle(),o=i.getCellWidth(),a=i.getCellHeight(),s=e.start.time;s<=e.end.time;s=i.getNextNDay(s,1).time){var l=i.dataToRect([s],!1).tl,u=new Al({shape:{x:l[0],y:l[1],width:o,height:a},cursor:"default",style:r});n.add(u)}},e.prototype._renderLines=function(t,e,n,i){var r=this,o=t.coordinateSystem,a=t.getModel(["splitLine","lineStyle"]).getLineStyle(),s=t.get(["splitLine","show"]),l=a.lineWidth;this._tlpoints=[],this._blpoints=[],this._firstDayOfMonth=[],this._firstDayPoints=[];for(var u=e.start,c=0;u.time<=e.end.time;c++){h(u.formatedDate),0===c&&(u=o.getDateInfo(e.start.y+"-"+e.start.m));var d=u.date;d.setMonth(d.getMonth()+1),u=o.getDateInfo(d)}function h(e){r._firstDayOfMonth.push(o.getDateInfo(e)),r._firstDayPoints.push(o.dataToRect([e],!1).tl);var l=r._getLinePointsOfOneWeek(t,e,n);r._tlpoints.push(l[0]),r._blpoints.push(l[l.length-1]),s&&r._drawSplitline(l,a,i)}h(o.getNextNDay(e.end.time,1).formatedDate),s&&this._drawSplitline(r._getEdgesPoints(r._tlpoints,l,n),a,i),s&&this._drawSplitline(r._getEdgesPoints(r._blpoints,l,n),a,i)},e.prototype._getEdgesPoints=function(t,e,n){var i=[t[0].slice(),t[t.length-1].slice()],r="horizontal"===n?0:1;return i[0][r]=i[0][r]-e/2,i[1][r]=i[1][r]+e/2,i},e.prototype._drawSplitline=function(t,e,n){var i=new sd({z2:20,shape:{points:t},style:e});n.add(i)},e.prototype._getLinePointsOfOneWeek=function(t,e,n){for(var i=t.coordinateSystem,r=i.getDateInfo(e),o=[],a=0;a<7;a++){var s=i.getNextNDay(r.time,a),l=i.dataToRect([s.time],!1);o[2*s.day]=l.tl,o[2*s.day+1]=l["horizontal"===n?"bl":"tr"]}return o},e.prototype._formatterLabel=function(t,e){return"string"===typeof t&&t?ep(t,e):"function"===typeof t?t(e):e.nameMap},e.prototype._yearTextPositionControl=function(t,e,n,i,r){var o=e[0],a=e[1],s=["center","bottom"];"bottom"===i?(a+=r,s=["center","top"]):"left"===i?o-=r:"right"===i?(o+=r,s=["center","top"]):a-=r;var l=0;return"left"!==i&&"right"!==i||(l=Math.PI/2),{rotation:l,x:o,y:a,style:{align:s[0],verticalAlign:s[1]}}},e.prototype._renderYearText=function(t,e,n,i){var r=t.getModel("yearLabel");if(r.get("show")){var o=r.get("margin"),a=r.get("position");a||(a="horizontal"!==n?"top":"left");var s=[this._tlpoints[this._tlpoints.length-1],this._blpoints[0]],l=(s[0][0]+s[1][0])/2,u=(s[0][1]+s[1][1])/2,c="horizontal"===n?0:1,d={top:[l,s[c][1]],bottom:[l,s[1-c][1]],left:[s[1-c][0],u],right:[s[c][0],u]},h=e.start.y;+e.end.y>+e.start.y&&(h=h+"-"+e.end.y);var f=r.get("formatter"),p={start:e.start.y,end:e.end.y,nameMap:h},m=this._formatterLabel(f,p),g=new Gl({z2:30,style:Lh(r,{text:m})});g.attr(this._yearTextPositionControl(g,d[a],n,a,o)),i.add(g)}},e.prototype._monthTextPositionControl=function(t,e,n,i,r){var o="left",a="top",s=t[0],l=t[1];return"horizontal"===n?(l+=r,e&&(o="center"),"start"===i&&(a="bottom")):(s+=r,e&&(a="middle"),"start"===i&&(o="right")),{x:s,y:l,align:o,verticalAlign:a}},e.prototype._renderMonthText=function(t,e,n){var i=t.getModel("monthLabel");if(i.get("show")){var r=i.get("nameMap"),o=i.get("margin"),a=i.get("position"),s=i.get("align"),l=[this._tlpoints,this._blpoints];q(r)&&(r=w$[r.toUpperCase()]||[]);var u="start"===a?0:1,c="horizontal"===e?0:1;o="start"===a?-o:o;for(var d="center"===s,h=0;h<l[u].length-1;h++){var f=l[u][h].slice(),p=this._firstDayOfMonth[h];if(d){var m=this._firstDayPoints[h];f[c]=(m[c]+l[0][h+1][c])/2}var g=i.get("formatter"),y=r[+p.m-1],v={yyyy:p.y,yy:(p.y+"").slice(2),MM:p.m,M:+p.m,nameMap:y},_=this._formatterLabel(g,v),b=new Gl({z2:30,style:I(Lh(i,{text:_}),this._monthTextPositionControl(f,d,e,a,o))});n.add(b)}}},e.prototype._weekTextPositionControl=function(t,e,n,i,r){var o="center",a="middle",s=t[0],l=t[1],u="start"===n;return"horizontal"===e?(s=s+i+(u?1:-1)*r[0]/2,o=u?"right":"left"):(l=l+i+(u?1:-1)*r[1]/2,a=u?"bottom":"top"),{x:s,y:l,align:o,verticalAlign:a}},e.prototype._renderWeekText=function(t,e,n,i){var r=t.getModel("dayLabel");if(r.get("show")){var o=t.coordinateSystem,a=r.get("position"),s=r.get("nameMap"),l=r.get("margin"),u=o.getFirstDayOfWeek();q(s)&&(s=M$[s.toUpperCase()]||[]);var c=o.getNextNDay(e.end.time,7-e.lweek).time,d=[o.getCellWidth(),o.getCellHeight()];l=Nr(l,Math.min(d[1],d[0])),"start"===a&&(c=o.getNextNDay(e.start.time,-(7+e.fweek)).time,l=-l);for(var h=0;h<7;h++){var f=o.getNextNDay(c,h),p=o.dataToRect([f.time],!1).center,m=h;m=Math.abs((h+u)%7);var g=new Gl({z2:30,style:I(Lh(r,{text:s[m]}),this._weekTextPositionControl(p,n,a,l,d))});i.add(g)}}},e.type="calendar",e}(Py),L$=S$,k$=864e5,T$=function(){function t(e,n,i){this.type="calendar",this.dimensions=t.dimensions,this.getDimensionsInfo=t.getDimensionsInfo,this._model=e}return t.getDimensionsInfo=function(){return[{name:"time",type:"time"},"value"]},t.prototype.getRangeInfo=function(){return this._rangeInfo},t.prototype.getModel=function(){return this._model},t.prototype.getRect=function(){return this._rect},t.prototype.getCellWidth=function(){return this._sw},t.prototype.getCellHeight=function(){return this._sh},t.prototype.getOrient=function(){return this._orient},t.prototype.getFirstDayOfWeek=function(){return this._firstDayOfWeek},t.prototype.getDateInfo=function(t){t=qr(t);var e=t.getFullYear(),n=t.getMonth()+1,i=n<10?"0"+n:""+n,r=t.getDate(),o=r<10?"0"+r:""+r,a=t.getDay();return a=Math.abs((a+7-this.getFirstDayOfWeek())%7),{y:e+"",m:i,d:o,day:a,time:t.getTime(),formatedDate:e+"-"+i+"-"+o,date:t}},t.prototype.getNextNDay=function(t,e){return e=e||0,0===e||(t=new Date(this.getDateInfo(t).time),t.setDate(t.getDate()+e)),this.getDateInfo(t)},t.prototype.update=function(t,e){this._firstDayOfWeek=+this._model.getModel("dayLabel").get("firstDay"),this._orient=this._model.get("orient"),this._lineWidth=this._model.getModel("itemStyle").getItemStyle().lineWidth||0,this._rangeInfo=this._getRangeInfo(this._initRangeOption());var n=this._rangeInfo.weeks||1,i=["width","height"],r=this._model.getCellSize().slice(),o=this._model.getBoxLayoutParams(),a="horizontal"===this._orient?[n,7]:[7,n];R([0,1],(function(t){u(r,t)&&(o[i[t]]=r[t]*a[t])}));var s={width:e.getWidth(),height:e.getHeight()},l=this._rect=dp(o,s);function u(t,e){return null!=t[e]&&"auto"!==t[e]}R([0,1],(function(t){u(r,t)||(r[t]=l[i[t]]/a[t])})),this._sw=r[0],this._sh=r[1]},t.prototype.dataToPoint=function(t,e){$(t)&&(t=t[0]),null==e&&(e=!0);var n=this.getDateInfo(t),i=this._rangeInfo,r=n.formatedDate;if(e&&!(n.time>=i.start.time&&n.time<i.end.time+k$))return[NaN,NaN];var o=n.day,a=this._getRangeInfo([i.start.time,r]).nthWeek;return"vertical"===this._orient?[this._rect.x+o*this._sw+this._sw/2,this._rect.y+a*this._sh+this._sh/2]:[this._rect.x+a*this._sw+this._sw/2,this._rect.y+o*this._sh+this._sh/2]},t.prototype.pointToData=function(t){var e=this.pointToDate(t);return e&&e.time},t.prototype.dataToRect=function(t,e){var n=this.dataToPoint(t,e);return{contentShape:{x:n[0]-(this._sw-this._lineWidth)/2,y:n[1]-(this._sh-this._lineWidth)/2,width:this._sw-this._lineWidth,height:this._sh-this._lineWidth},center:n,tl:[n[0]-this._sw/2,n[1]-this._sh/2],tr:[n[0]+this._sw/2,n[1]-this._sh/2],br:[n[0]+this._sw/2,n[1]+this._sh/2],bl:[n[0]-this._sw/2,n[1]+this._sh/2]}},t.prototype.pointToDate=function(t){var e=Math.floor((t[0]-this._rect.x)/this._sw)+1,n=Math.floor((t[1]-this._rect.y)/this._sh)+1,i=this._rangeInfo.range;return"vertical"===this._orient?this._getDateByWeeksAndDay(n,e-1,i):this._getDateByWeeksAndDay(e,n-1,i)},t.prototype.convertToPixel=function(t,e,n){var i=D$(e);return i===this?i.dataToPoint(n):null},t.prototype.convertFromPixel=function(t,e,n){var i=D$(e);return i===this?i.pointToData(n):null},t.prototype.containPoint=function(t){return!1},t.prototype._initRangeOption=function(){var t,e=this._model.get("range");if($(e)&&1===e.length&&(e=e[0]),$(e))t=e;else{var n=e.toString();if(/^\d{4}$/.test(n)&&(t=[n+"-01-01",n+"-12-31"]),/^\d{4}[\/|-]\d{1,2}$/.test(n)){var i=this.getDateInfo(n),r=i.date;r.setMonth(r.getMonth()+1);var o=this.getNextNDay(r,-1);t=[i.formatedDate,o.formatedDate]}/^\d{4}[\/|-]\d{1,2}[\/|-]\d{1,2}$/.test(n)&&(t=[n,n])}if(!t)return e;var a=this._getRangeInfo(t);return a.start.time>a.end.time&&t.reverse(),t},t.prototype._getRangeInfo=function(t){var e,n=[this.getDateInfo(t[0]),this.getDateInfo(t[1])];n[0].time>n[1].time&&(e=!0,n.reverse());var i=Math.floor(n[1].time/k$)-Math.floor(n[0].time/k$)+1,r=new Date(n[0].time),o=r.getDate(),a=n[1].date.getDate();r.setDate(o+i-1);var s=r.getDate();if(s!==a){var l=r.getTime()-n[1].time>0?1:-1;while((s=r.getDate())!==a&&(r.getTime()-n[1].time)*l>0)i-=l,r.setDate(s-l)}var u=Math.floor((i+n[0].day+6)/7),c=e?1-u:u-1;return e&&n.reverse(),{range:[n[0].formatedDate,n[1].formatedDate],start:n[0],end:n[1],allDay:i,weeks:u,nthWeek:c,fweek:n[0].day,lweek:n[1].day}},t.prototype._getDateByWeeksAndDay=function(t,e,n){var i=this._getRangeInfo(n);if(t>i.weeks||0===t&&e<i.fweek||t===i.weeks&&e>i.lweek)return null;var r=7*(t-1)-i.fweek+e,o=new Date(i.start.time);return o.setDate(+i.start.d+r),this.getDateInfo(o)},t.create=function(e,n){var i=[];return e.eachComponent("calendar",(function(r){var o=new t(r,e,n);i.push(o),r.coordinateSystem=o})),e.eachSeries((function(t){"calendar"===t.get("coordinateSystem")&&(t.coordinateSystem=i[t.get("calendarIndex")||0])})),i},t.dimensions=["time","value"],t}();function D$(t){var e=t.calendarModel,n=t.seriesModel,i=e?e.coordinateSystem:n?n.coordinateSystem:null;return i}var C$=T$;function I$(t){t.registerComponentModel(x$),t.registerComponentView(L$),t.registerCoordinateSystem("calendar",C$)}var A$=Io(),Y$={path:null,compoundPath:null,group:Mr,image:wl,text:Gl},O$=function(t){var e=t.graphic;$(e)?e[0]&&e[0].elements?t.graphic=[t.graphic[0]]:t.graphic=[{elements:e}]:e&&!e.elements&&(t.graphic=[{elements:[e]}])},P$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.preventAutoZ=!0,n}return o(e,t),e.prototype.mergeOption=function(e,n){var i=this.option.elements;this.option.elements=null,t.prototype.mergeOption.call(this,e,n),this.option.elements=i},e.prototype.optionUpdated=function(t,e){var n=this.option,i=(e?n:t).elements,r=n.elements=e?[]:n.elements,o=[];this._flatten(i,o,null);var a=fo(r,o,"normalMerge"),s=this._elOptionsToUpdate=[];R(a,(function(t,e){var n=t.newOption;n&&(s.push(n),z$(t,n),W$(r,e,n),B$(r[e],n))}),this);for(var l=r.length-1;l>=0;l--)null==r[l]?r.splice(l,1):delete r[l].$action},e.prototype._flatten=function(t,e,n){R(t,(function(t){if(t){n&&(t.parentOption=n),e.push(t);var i=t.children;"group"===t.type&&i&&this._flatten(i,e,t),delete t.children}}),this)},e.prototype.useElOptionsToUpdate=function(){var t=this._elOptionsToUpdate;return this._elOptionsToUpdate=null,t},e.type="graphic",e.defaultOption={elements:[]},e}(xp),E$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){this._elMap=gt()},e.prototype.render=function(t,e,n){t!==this._lastGraphicModel&&this._clear(),this._lastGraphicModel=t,this._updateElements(t),this._relocate(t,n)},e.prototype._updateElements=function(t){var e=t.useElOptionsToUpdate();if(e){var n=this._elMap,i=this.group;R(e,(function(e){var r=wo(e.id,null),o=null!=r?n.get(r):null,a=wo(e.parentId,null),s=null!=a?n.get(a):i,l=e.type,u=e.style;"text"===l&&u&&e.hv&&e.hv[1]&&(u.textVerticalAlign=u.textBaseline=u.verticalAlign=u.align=null);var c=e.textContent,d=e.textConfig;if(u&&cB(u,l,!!d,!!c)){var h=dB(u,l,!0);!d&&h.textConfig&&(d=e.textConfig=h.textConfig),!c&&h.textContent&&(c=h.textContent)}var f=H$(e);var p=e.$action||"merge";"merge"===p?o?o.attr(f):N$(r,s,f,n):"replace"===p?(R$(o,n),N$(r,s,f,n)):"remove"===p&&R$(o,n);var m=n.get(r);if(m&&c)if("merge"===p){var g=m.getTextContent();g?g.attr(c):m.setTextContent(new Gl(c))}else"replace"===p&&m.setTextContent(new Gl(c));if(m){var y=A$(m);y.__ecGraphicWidthOption=e.width,y.__ecGraphicHeightOption=e.height,F$(m,t,e),_h({el:m,componentModel:t,itemName:m.name,itemTooltipOption:e.tooltip})}}))}},e.prototype._relocate=function(t,e){for(var n=t.option.elements,i=this.group,r=this._elMap,o=e.getWidth(),a=e.getHeight(),s=0;s<n.length;s++){var l=n[s],u=wo(l.id,null),c=null!=u?r.get(u):null;if(c&&c.isGroup){var d=c.parent,h=d===i,f=A$(c),p=A$(d);f.__ecGraphicWidth=Nr(f.__ecGraphicWidthOption,h?o:p.__ecGraphicWidth)||0,f.__ecGraphicHeight=Nr(f.__ecGraphicHeightOption,h?a:p.__ecGraphicHeight)||0}}for(s=n.length-1;s>=0;s--){l=n[s],u=wo(l.id,null),c=null!=u?r.get(u):null;if(c){d=c.parent,p=A$(d);var m=d===i?{width:o,height:a}:{width:p.__ecGraphicWidth,height:p.__ecGraphicHeight};hp(c,l,m,null,{hv:l.hv,boundingMode:l.bounding})}}},e.prototype._clear=function(){var t=this._elMap;t.each((function(e){R$(e,t)})),this._elMap=gt()},e.prototype.dispose=function(){this._clear()},e.type="graphic",e}(Py);function N$(t,e,n,i){var r=n.type;var o=_t(Y$,r)?Y$[r]:Vd(r);var a=new o(n);e.add(a),i.set(t,a),A$(a).__ecGraphicId=t}function R$(t,e){var n=t&&t.parent;n&&("group"===t.type&&t.traverse((function(t){R$(t,e)})),e.removeKey(A$(t).__ecGraphicId),n.remove(t))}function H$(t){return t=I({},t),R(["id","parentId","$action","hv","bounding","textContent"].concat(ap),(function(e){delete t[e]})),t}function j$(t,e){var n;return R(e,(function(e){null!=t[e]&&"auto"!==t[e]&&(n=!0)})),n}function z$(t,e){var n=t.existing;if(e.id=t.keyInfo.id,!e.type&&n&&(e.type=n.type),null==e.parentId){var i=e.parentOption;i?e.parentId=i.id:n&&(e.parentId=n.parentId)}e.parentOption=null}function W$(t,e,n){var i=I({},n),r=t[e],o=n.$action||"merge";"merge"===o?r?(D(r,i,!0),mp(r,i,{ignoreSize:!0}),yp(n,r)):t[e]=i:"replace"===o?t[e]=i:"remove"===o&&r&&(t[e]=null)}function B$(t,e){if(t&&(t.hv=e.hv=[j$(e,["left","right"]),j$(e,["top","bottom"])],"group"===t.type)){var n=t,i=e;null==n.width&&(n.width=i.width=0),null==n.height&&(n.height=i.height=0)}}function F$(t,e,n){var i=$l(t).eventData;t.silent||t.ignore||i||(i=$l(t).eventData={componentType:"graphic",componentIndex:e.componentIndex,name:t.name}),i&&(i.info=n.info)}function V$(t){t.registerComponentModel(P$),t.registerComponentView(E$),t.registerPreprocessor(O$)}var G$=["x","y","radius","angle","single"],$$=["cartesian2d","polar","singleAxis"];function U$(t){var e=t.get("coordinateSystem");return O($$,e)>=0}function q$(t){return t+"Axis"}function X$(t,e){var n,i=gt(),r=[],o=gt();t.eachComponent({mainType:"dataZoom",query:e},(function(t){o.get(t.uid)||s(t)}));do{n=!1,t.eachComponent("dataZoom",a)}while(n);function a(t){!o.get(t.uid)&&l(t)&&(s(t),n=!0)}function s(t){o.set(t.uid,!0),r.push(t),u(t)}function l(t){var e=!1;return t.eachTargetAxis((function(t,n){var r=i.get(t);r&&r[n]&&(e=!0)})),e}function u(t){t.eachTargetAxis((function(t,e){(i.get(t)||i.set(t,[]))[e]=!0}))}return r}function Z$(t){var e=t.ecModel,n={infoList:[],infoMap:gt()};return t.eachTargetAxis((function(t,i){var r=e.getComponent(q$(t),i);if(r){var o=r.getCoordSysModel();if(o){var a=o.uid,s=n.infoMap.get(a);s||(s={model:o,axisModels:[]},n.infoList.push(s),n.infoMap.set(a,s)),s.axisModels.push(r)}}})),n}var J$=function(){function t(){this.indexList=[],this.indexMap=[]}return t.prototype.add=function(t){this.indexMap[t]||(this.indexList.push(t),this.indexMap[t]=!0)},t}(),K$=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._autoThrottle=!0,n._noTarget=!0,n._rangePropMode=["percent","percent"],n}return o(e,t),e.prototype.init=function(t,e,n){var i=Q$(t);this.settledOption=i,this.mergeDefaultAndTheme(t,n),this._doInit(i)},e.prototype.mergeOption=function(t){var e=Q$(t);D(this.option,t,!0),D(this.settledOption,e,!0),this._doInit(e)},e.prototype._doInit=function(t){var e=this.option;this._setDefaultThrottle(t),this._updateRangeUse(t);var n=this.settledOption;R([["start","startValue"],["end","endValue"]],(function(t,i){"value"===this._rangePropMode[i]&&(e[t[0]]=n[t[0]]=null)}),this),this._resetTarget()},e.prototype._resetTarget=function(){var t=this.get("orient",!0),e=this._targetAxisInfoMap=gt(),n=this._fillSpecifiedTargetAxis(e);n?this._orient=t||this._makeAutoOrientByTargetAxis():(this._orient=t||"horizontal",this._fillAutoTargetAxisByOrient(e,this._orient)),this._noTarget=!0,e.each((function(t){t.indexList.length&&(this._noTarget=!1)}),this)},e.prototype._fillSpecifiedTargetAxis=function(t){var e=!1;return R(G$,(function(n){var i=this.getReferringComponents(q$(n),Eo);if(i.specified){e=!0;var r=new J$;R(i.models,(function(t){r.add(t.componentIndex)})),t.set(n,r)}}),this),e},e.prototype._fillAutoTargetAxisByOrient=function(t,e){var n=this.ecModel,i=!0;if(i){var r="vertical"===e?"y":"x",o=n.findComponents({mainType:r+"Axis"});a(o,r)}if(i){o=n.findComponents({mainType:"singleAxis",filter:function(t){return t.get("orient",!0)===e}});a(o,"single")}function a(e,n){var r=e[0];if(r){var o=new J$;if(o.add(r.componentIndex),t.set(n,o),i=!1,"x"===n||"y"===n){var a=r.getReferringComponents("grid",Po).models[0];a&&R(e,(function(t){r.componentIndex!==t.componentIndex&&a===t.getReferringComponents("grid",Po).models[0]&&o.add(t.componentIndex)}))}}}i&&R(G$,(function(e){if(i){var r=n.findComponents({mainType:q$(e),filter:function(t){return"category"===t.get("type",!0)}});if(r[0]){var o=new J$;o.add(r[0].componentIndex),t.set(e,o),i=!1}}}),this)},e.prototype._makeAutoOrientByTargetAxis=function(){var t;return this.eachTargetAxis((function(e){!t&&(t=e)}),this),"y"===t?"vertical":"horizontal"},e.prototype._setDefaultThrottle=function(t){if(t.hasOwnProperty("throttle")&&(this._autoThrottle=!1),this._autoThrottle){var e=this.ecModel.option;this.option.throttle=e.animation&&e.animationDurationUpdate>0?100:20}},e.prototype._updateRangeUse=function(t){var e=this._rangePropMode,n=this.get("rangeMode");R([["start","startValue"],["end","endValue"]],(function(i,r){var o=null!=t[i[0]],a=null!=t[i[1]];o&&!a?e[r]="percent":!o&&a?e[r]="value":n?e[r]=n[r]:o&&(e[r]="percent")}))},e.prototype.noTarget=function(){return this._noTarget},e.prototype.getFirstTargetAxisModel=function(){var t;return this.eachTargetAxis((function(e,n){null==t&&(t=this.ecModel.getComponent(q$(e),n))}),this),t},e.prototype.eachTargetAxis=function(t,e){this._targetAxisInfoMap.each((function(n,i){R(n.indexList,(function(n){t.call(e,i,n)}))}))},e.prototype.getAxisProxy=function(t,e){var n=this.getAxisModel(t,e);if(n)return n.__dzAxisProxy},e.prototype.getAxisModel=function(t,e){var n=this._targetAxisInfoMap.get(t);if(n&&n.indexMap[e])return this.ecModel.getComponent(q$(t),e)},e.prototype.setRawRange=function(t){var e=this.option,n=this.settledOption;R([["start","startValue"],["end","endValue"]],(function(i){null==t[i[0]]&&null==t[i[1]]||(e[i[0]]=n[i[0]]=t[i[0]],e[i[1]]=n[i[1]]=t[i[1]])}),this),this._updateRangeUse(t)},e.prototype.setCalculatedRange=function(t){var e=this.option;R(["start","startValue","end","endValue"],(function(n){e[n]=t[n]}))},e.prototype.getPercentRange=function(){var t=this.findRepresentativeAxisProxy();if(t)return t.getDataPercentWindow()},e.prototype.getValueRange=function(t,e){if(null!=t||null!=e)return this.getAxisProxy(t,e).getDataValueWindow();var n=this.findRepresentativeAxisProxy();return n?n.getDataValueWindow():void 0},e.prototype.findRepresentativeAxisProxy=function(t){if(t)return t.__dzAxisProxy;for(var e,n=this._targetAxisInfoMap.keys(),i=0;i<n.length;i++)for(var r=n[i],o=this._targetAxisInfoMap.get(r),a=0;a<o.indexList.length;a++){var s=this.getAxisProxy(r,o.indexList[a]);if(s.hostedBy(this))return s;e||(e=s)}return e},e.prototype.getRangePropMode=function(){return this._rangePropMode.slice()},e.prototype.getOrient=function(){return this._orient},e.type="dataZoom",e.dependencies=["xAxis","yAxis","radiusAxis","angleAxis","singleAxis","series","toolbox"],e.defaultOption={zlevel:0,z:4,filterMode:"filter",start:0,end:100},e}(xp);function Q$(t){var e={};return R(["start","end","startValue","endValue","throttle"],(function(n){t.hasOwnProperty(n)&&(e[n]=t[n])})),e}var tU=K$,eU=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.select",e}(tU),nU=eU,iU=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n,i){this.dataZoomModel=t,this.ecModel=e,this.api=n},e.type="dataZoom",e}(Py),rU=iU,oU=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.select",e}(rU),aU=oU,sU=R,lU=Hr,uU=function(){function t(t,e,n,i){this._dimName=t,this._axisIndex=e,this.ecModel=i,this._dataZoomModel=n}return t.prototype.hostedBy=function(t){return this._dataZoomModel===t},t.prototype.getDataValueWindow=function(){return this._valueWindow.slice()},t.prototype.getDataPercentWindow=function(){return this._percentWindow.slice()},t.prototype.getTargetSeriesModels=function(){var t=[];return this.ecModel.eachSeries((function(e){if(U$(e)){var n=q$(this._dimName),i=e.getReferringComponents(n,Po).models[0];i&&this._axisIndex===i.componentIndex&&t.push(e)}}),this),t},t.prototype.getAxisModel=function(){return this.ecModel.getComponent(this._dimName+"Axis",this._axisIndex)},t.prototype.getMinMaxSpan=function(){return T(this._minMaxSpan)},t.prototype.calculateDataWindow=function(t){var e,n=this._dataExtent,i=this.getAxisModel(),r=i.axis.scale,o=this._dataZoomModel.getRangePropMode(),a=[0,100],s=[],l=[];sU(["start","end"],(function(i,u){var c=t[i],d=t[i+"Value"];"percent"===o[u]?(null==c&&(c=a[u]),d=r.parse(Er(c,a,n))):(e=!0,d=null==d?n[u]:r.parse(d),c=Er(d,n,a)),l[u]=d,s[u]=c})),lU(l),lU(s);var u=this._minMaxSpan;function c(t,e,n,i,o){var a=o?"Span":"ValueSpan";rR(0,t,n,"all",u["min"+a],u["max"+a]);for(var s=0;s<2;s++)e[s]=Er(t[s],n,i,!0),o&&(e[s]=r.parse(e[s]))}return e?c(l,s,n,a,!1):c(s,l,a,n,!0),{valueWindow:l,percentWindow:s}},t.prototype.reset=function(t){if(t===this._dataZoomModel){var e=this.getTargetSeriesModels();this._dataExtent=cU(this,this._dimName,e),this._updateMinMaxSpan();var n=this.calculateDataWindow(t.settledOption);this._valueWindow=n.valueWindow,this._percentWindow=n.percentWindow,this._setAxisModel()}},t.prototype.filterData=function(t,e){if(t===this._dataZoomModel){var n=this._dimName,i=this.getTargetSeriesModels(),r=t.get("filterMode"),o=this._valueWindow;"none"!==r&&sU(i,(function(t){var e=t.getData(),i=e.mapDimensionsAll(n);i.length&&("weakFilter"===r?e.filterSelf((function(t){for(var n,r,a,s=0;s<i.length;s++){var l=e.get(i[s],t),u=!isNaN(l),c=l<o[0],d=l>o[1];if(u&&!c&&!d)return!0;u&&(a=!0),c&&(n=!0),d&&(r=!0)}return a&&n&&r})):sU(i,(function(n){if("empty"===r)t.setData(e=e.map(n,(function(t){return a(t)?t:NaN})));else{var i={};i[n]=o,e.selectRange(i)}})),sU(i,(function(t){e.setApproximateExtent(o,t)})))}))}function a(t){return t>=o[0]&&t<=o[1]}},t.prototype._updateMinMaxSpan=function(){var t=this._minMaxSpan={},e=this._dataZoomModel,n=this._dataExtent;sU(["min","max"],(function(i){var r=e.get(i+"Span"),o=e.get(i+"ValueSpan");null!=o&&(o=this.getAxisModel().axis.scale.parse(o)),null!=o?r=Er(n[0]+o,n,[0,100],!0):null!=r&&(o=Er(r,[0,100],n,!0)-n[0]),t[i+"Span"]=r,t[i+"ValueSpan"]=o}),this)},t.prototype._setAxisModel=function(){var t=this.getAxisModel(),e=this._percentWindow,n=this._valueWindow;if(e){var i=Wr(n,[0,500]);i=Math.min(i,20);var r=t.axis.scale.rawExtentInfo;0!==e[0]&&r.setDeterminedMinMax("min",+n[0].toFixed(i)),100!==e[1]&&r.setDeterminedMinMax("max",+n[1].toFixed(i)),r.freeze()}},t}();function cU(t,e,n){var i=[1/0,-1/0];sU(n,(function(t){DC(i,t.getData(),e)}));var r=t.getAxisModel(),o=pC(r.axis.scale,r,i).calculate();return[o.min,o.max]}var dU=uU,hU={getTargetSeries:function(t){function e(e){t.eachComponent("dataZoom",(function(n){n.eachTargetAxis((function(i,r){var o=t.getComponent(q$(i),r);e(i,r,o,n)}))}))}e((function(t,e,n,i){n.__dzAxisProxy=null}));var n=[];e((function(e,i,r,o){r.__dzAxisProxy||(r.__dzAxisProxy=new dU(e,i,o,t),n.push(r.__dzAxisProxy))}));var i=gt();return R(n,(function(t){R(t.getTargetSeriesModels(),(function(t){i.set(t.uid,t)}))})),i},overallReset:function(t,e){t.eachComponent("dataZoom",(function(t){t.eachTargetAxis((function(e,n){t.getAxisProxy(e,n).reset(t)})),t.eachTargetAxis((function(n,i){t.getAxisProxy(n,i).filterData(t,e)}))})),t.eachComponent("dataZoom",(function(t){var e=t.findRepresentativeAxisProxy();if(e){var n=e.getDataPercentWindow(),i=e.getDataValueWindow();t.setCalculatedRange({start:n[0],end:n[1],startValue:i[0],endValue:i[1]})}}))}},fU=hU;function pU(t){t.registerAction("dataZoom",(function(t,e){var n=X$(e,t);R(n,(function(e){e.setRawRange({start:t.start,end:t.end,startValue:t.startValue,endValue:t.endValue})}))}))}var mU=!1;function gU(t){mU||(mU=!0,t.registerProcessor(t.PRIORITY.PROCESSOR.FILTER,fU),pU(t),t.registerSubTypeDefaulter("dataZoom",(function(){return"slider"})))}function yU(t){t.registerComponentModel(nU),t.registerComponentView(aU),gU(t)}var vU=function(){function t(){}return t}(),_U={};function bU(t,e){_U[t]=e}function xU(t){return _U[t]}var wU=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(){t.prototype.optionUpdated.apply(this,arguments);var e=this.ecModel;R(this.option.feature,(function(t,n){var i=xU(n);i&&(i.getDefaultOption&&(i.defaultOption=i.getDefaultOption(e)),D(t,i.defaultOption))}))},e.type="toolbox",e.layoutMode={type:"box",ignoreSize:!0},e.defaultOption={show:!0,z:6,zlevel:0,orient:"horizontal",left:"right",top:"top",backgroundColor:"transparent",borderColor:"#ccc",borderRadius:0,borderWidth:0,padding:5,itemSize:15,itemGap:8,showTitle:!0,iconStyle:{borderColor:"#666",color:"none"},emphasis:{iconStyle:{borderColor:"#3E98C5"}},tooltip:{show:!1,position:"bottom"}},e}(xp),MU=wU;function SU(t,e,n){var i=e.getBoxLayoutParams(),r=e.get("padding"),o={width:n.getWidth(),height:n.getHeight()},a=dp(i,o,r);up(e.get("orient"),t,e.get("itemGap"),a.width,a.height),hp(t,i,o,r)}function LU(t,e){var n=Uf(e.get("padding")),i=e.getItemStyle(["color","opacity"]);return i.fill=e.get("backgroundColor"),t=new Al({shape:{x:t.x-n[3],y:t.y-n[0],width:t.width+n[1]+n[3],height:t.height+n[0]+n[2],r:e.get("borderRadius")},style:i,silent:!0,z2:-1}),t}var kU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.render=function(t,e,n,i){var r=this.group;if(r.removeAll(),t.get("show")){var o=+t.get("itemSize"),a=t.get("feature")||{},s=this._features||(this._features={}),l=[];R(a,(function(t,e){l.push(e)})),new $S(this._featureNames||[],l).add(u).update(u).remove(G(u,null)).execute(),this._featureNames=l,SU(r,t,n),r.add(LU(r.getBoundingRect(),t)),r.eachChild((function(t){var e=t.__title,i=t.ensureState("emphasis"),a=i.textConfig||(i.textConfig={}),s=t.getTextContent(),l=s&&s.states.emphasis;if(l&&!U(l)&&e){var u=l.style||(l.style={}),c=rr(e,Gl.makeFont(u)),d=t.x+r.x,h=t.y+r.y+o,f=!1;h+c.height>n.getHeight()&&(a.position="top",f=!0);var p=f?-5-c.height:o+8;d+c.width/2>n.getWidth()?(a.position=["100%",p],u.align="right"):d-c.width/2<0&&(a.position=[0,p],u.align="left")}}))}function u(r,o){var u,d=l[r],h=l[o],f=a[d],p=new qh(f,t,t.ecModel);if(i&&null!=i.newTitle&&i.featureName===d&&(f.title=i.newTitle),d&&!h){if(TU(d))u={onclick:p.option.onclick,featureName:d};else{var m=xU(d);if(!m)return;u=new m}s[d]=u}else if(u=s[h],!u)return;u.uid=Zh("toolbox-feature"),u.model=p,u.ecModel=e,u.api=n;var g=u instanceof vU;d||!h?!p.get("show")||g&&u.unusable?g&&u.remove&&u.remove(e,n):(c(p,u,d),p.setIconStatus=function(t,e){var n=this.option,i=this.iconPaths;n.iconStatus=n.iconStatus||{},n.iconStatus[t]=e,i[t]&&("emphasis"===e?Cu:Iu)(i[t])},u instanceof vU&&u.render&&u.render(p,e,n,i)):g&&u.dispose&&u.dispose(e,n)}function c(i,a,s){var l,u,c=i.getModel("iconStyle"),d=i.getModel(["emphasis","iconStyle"]),h=a instanceof vU&&a.getIcons?a.getIcons():i.get("icon"),f=i.get("title")||{};"string"===typeof h?(l={},l[s]=h):l=h,"string"===typeof f?(u={},u[s]=f):u=f;var p=i.iconPaths={};R(l,(function(s,l){var h=ph(s,{},{x:-o/2,y:-o/2,width:o,height:o});h.setStyle(c.getItemStyle());var f=h.ensureState("emphasis");f.style=d.getItemStyle();var m=new Gl({style:{text:u[l],align:d.get("textAlign"),borderRadius:d.get("textBorderRadius"),padding:d.get("textPadding"),fill:null},ignore:!0});h.setTextContent(m),_h({el:h,componentModel:t,itemName:l,formatterParamsExtra:{title:u[l]}}),h.__title=u[l],h.on("mouseover",(function(){var e=d.getItemStyle(),n="vertical"===t.get("orient")?null==t.get("right")?"right":"left":null==t.get("bottom")?"bottom":"top";m.setStyle({fill:d.get("textFill")||e.fill||e.stroke||"#000",backgroundColor:d.get("textBackgroundColor")}),h.setTextConfig({position:d.get("textPosition")||n}),m.ignore=!t.get("showTitle"),Cu(this)})).on("mouseout",(function(){"emphasis"!==i.get(["iconStatus",l])&&Iu(this),m.hide()})),("emphasis"===i.get(["iconStatus",l])?Cu:Iu)(h),r.add(h),h.on("click",V(a.onclick,a,e,n,l)),p[l]=h}))}},e.prototype.updateView=function(t,e,n,i){R(this._features,(function(t){t instanceof vU&&t.updateView&&t.updateView(t.model,e,n,i)}))},e.prototype.remove=function(t,e){R(this._features,(function(n){n instanceof vU&&n.remove&&n.remove(t,e)})),this.group.removeAll()},e.prototype.dispose=function(t,e){R(this._features,(function(n){n instanceof vU&&n.dispose&&n.dispose(t,e)}))},e.type="toolbox",e}(Py);function TU(t){return 0===t.indexOf("my")}var DU=kU,CU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.onclick=function(t,e){var n=this.model,i=n.get("name")||t.get("title.0.text")||"echarts",r="svg"===e.getZr().painter.getType(),o=r?"svg":n.get("type",!0)||"png",a=e.getConnectedDataURL({type:o,backgroundColor:n.get("backgroundColor",!0)||t.get("backgroundColor")||"#fff",connectedBackgroundColor:n.get("connectedBackgroundColor"),excludeComponents:n.get("excludeComponents"),pixelRatio:n.get("pixelRatio")});if("function"!==typeof MouseEvent||!h.browser.newEdge&&(h.browser.ie||h.browser.edge))if(window.navigator.msSaveOrOpenBlob||r){var s=a.split(","),l=s[0].indexOf("base64")>-1,u=r?decodeURIComponent(s[1]):s[1];l&&(u=window.atob(u));var c=i+"."+o;if(window.navigator.msSaveOrOpenBlob){var d=u.length,f=new Uint8Array(d);while(d--)f[d]=u.charCodeAt(d);var p=new Blob([f]);window.navigator.msSaveOrOpenBlob(p,c)}else{var m=document.createElement("iframe");document.body.appendChild(m);var g=m.contentWindow,y=g.document;y.open("image/svg+xml","replace"),y.write(u),y.close(),g.focus(),y.execCommand("SaveAs",!0,c),document.body.removeChild(m)}}else{var v=n.get("lang"),_='<body style="margin:0;"><img src="'+a+'" style="max-width:100%;" title="'+(v&&v[0]||"")+'" /></body>',b=window.open();b.document.write(_),b.document.title=i}else{var x=document.createElement("a");x.download=i+"."+o,x.target="_blank",x.href=a;var w=new MouseEvent("click",{view:document.defaultView,bubbles:!0,cancelable:!1});x.dispatchEvent(w)}},e.getDefaultOption=function(t){var e={show:!0,icon:"M4.7,22.9L29.3,45.5L54.7,23.4M4.6,43.6L4.6,58L53.8,58L53.8,43.6M29.2,45.1L29.2,0",title:t.getLocale(["toolbox","saveAsImage","title"]),type:"png",connectedBackgroundColor:"#fff",name:"",excludeComponents:["toolbox"],lang:t.getLocale(["toolbox","saveAsImage","lang"])};return e},e}(vU);CU.prototype.unusable=!h.canvasSupported;var IU=CU,AU="__ec_magicType_stack__",YU=[["line","bar"],["stack"]],OU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.getIcons=function(){var t=this.model,e=t.get("icon"),n={};return R(t.get("type"),(function(t){e[t]&&(n[t]=e[t])})),n},e.getDefaultOption=function(t){var e={show:!0,type:[],icon:{line:"M4.1,28.9h7.1l9.3-22l7.4,38l9.7-19.7l3,12.8h14.9M4.1,58h51.4",bar:"M6.7,22.9h10V48h-10V22.9zM24.9,13h10v35h-10V13zM43.2,2h10v46h-10V2zM3.1,58h53.7",stack:"M8.2,38.4l-8.4,4.1l30.6,15.3L60,42.5l-8.1-4.1l-21.5,11L8.2,38.4z M51.9,30l-8.1,4.2l-13.4,6.9l-13.9-6.9L8.2,30l-8.4,4.2l8.4,4.2l22.2,11l21.5-11l8.1-4.2L51.9,30z M51.9,21.7l-8.1,4.2L35.7,30l-5.3,2.8L24.9,30l-8.4-4.1l-8.3-4.2l-8.4,4.2L8.2,30l8.3,4.2l13.9,6.9l13.4-6.9l8.1-4.2l8.1-4.1L51.9,21.7zM30.4,2.2L-0.2,17.5l8.4,4.1l8.3,4.2l8.4,4.2l5.5,2.7l5.3-2.7l8.1-4.2l8.1-4.2l8.1-4.1L30.4,2.2z"},title:t.getLocale(["toolbox","magicType","title"]),option:{},seriesIndex:{}};return e},e.prototype.onclick=function(t,e,n){var i=this.model,r=i.get(["seriesIndex",n]);if(PU[n]){var o,a={series:[]},s=function(t){var e=t.subType,r=t.id,o=PU[n](e,r,t,i);o&&(A(o,t.option),a.series.push(o));var s=t.coordinateSystem;if(s&&"cartesian2d"===s.type&&("line"===n||"bar"===n)){var l=s.getAxesByScale("ordinal")[0];if(l){var u=l.dim,c=u+"Axis",d=t.getReferringComponents(c,Po).models[0],h=d.componentIndex;a[c]=a[c]||[];for(var f=0;f<=h;f++)a[c][h]=a[c][h]||{};a[c][h].boundaryGap="bar"===n}}};R(YU,(function(t){O(t,n)>=0&&R(t,(function(t){i.setIconStatus(t,"normal")}))})),i.setIconStatus(n,"emphasis"),t.eachComponent({mainType:"series",query:null==r?null:{seriesIndex:r}},s);var l=n;"stack"===n&&(o=D({stack:i.option.title.tiled,tiled:i.option.title.stack},i.option.title),"emphasis"!==i.get(["iconStatus",n])&&(l="tiled")),e.dispatchAction({type:"changeMagicType",currentType:l,newOption:a,newTitle:o,featureName:"magicType"})}},e}(vU),PU={line:function(t,e,n,i){if("bar"===t)return D({id:e,type:"line",data:n.get("data"),stack:n.get("stack"),markPoint:n.get("markPoint"),markLine:n.get("markLine")},i.get(["option","line"])||{},!0)},bar:function(t,e,n,i){if("line"===t)return D({id:e,type:"bar",data:n.get("data"),stack:n.get("stack"),markPoint:n.get("markPoint"),markLine:n.get("markLine")},i.get(["option","bar"])||{},!0)},stack:function(t,e,n,i){var r=n.get("stack")===AU;if("line"===t||"bar"===t)return i.setIconStatus("stack",r?"normal":"emphasis"),D({id:e,stack:r?"":AU},i.get(["option","stack"])||{},!0)}};sM({type:"changeMagicType",event:"magicTypeChanged",update:"prepareAndUpdate"},(function(t,e){e.mergeOption(t.newOption)}));var EU=OU,NU=new Array(60).join("-"),RU="\t";function HU(t){var e={},n=[],i=[];return t.eachRawSeries((function(t){var r=t.coordinateSystem;if(!r||"cartesian2d"!==r.type&&"polar"!==r.type)n.push(t);else{var o=r.getBaseAxis();if("category"===o.type){var a=o.dim+"_"+o.index;e[a]||(e[a]={categoryAxis:o,valueAxis:r.getOtherAxis(o),series:[]},i.push({axisDim:o.dim,axisIndex:o.index})),e[a].series.push(t)}else n.push(t)}})),{seriesGroupByCategoryAxis:e,other:n,meta:i}}function jU(t){var e=[];return R(t,(function(t,n){var i=t.categoryAxis,r=t.valueAxis,o=r.dim,a=[" "].concat(H(t.series,(function(t){return t.name}))),s=[i.model.getCategories()];R(t.series,(function(t){var e=t.getRawData();s.push(t.getRawData().mapArray(e.mapDimension(o),(function(t){return t})))}));for(var l=[a.join(RU)],u=0;u<s[0].length;u++){for(var c=[],d=0;d<s.length;d++)c.push(s[d][u]);l.push(c.join(RU))}e.push(l.join("\n"))})),e.join("\n\n"+NU+"\n\n")}function zU(t){return H(t,(function(t){var e=t.getRawData(),n=[t.name],i=[];return e.each(e.dimensions,(function(){for(var t=arguments.length,r=arguments[t-1],o=e.getName(r),a=0;a<t-1;a++)i[a]=arguments[a];n.push((o?o+RU:"")+i.join(RU))})),n.join("\n")})).join("\n\n"+NU+"\n\n")}function WU(t){var e=HU(t);return{value:z([jU(e.seriesGroupByCategoryAxis),zU(e.other)],(function(t){return!!t.replace(/[\n\t\s]/g,"")})).join("\n\n"+NU+"\n\n"),meta:e.meta}}function BU(t){return t.replace(/^\s\s*/,"").replace(/\s\s*$/,"")}function FU(t){var e=t.slice(0,t.indexOf("\n"));if(e.indexOf(RU)>=0)return!0}var VU=new RegExp("["+RU+"]+","g");function GU(t){for(var e=t.split(/\n+/g),n=BU(e.shift()).split(VU),i=[],r=H(n,(function(t){return{name:t,data:[]}})),o=0;o<e.length;o++){var a=BU(e[o]).split(VU);i.push(a.shift());for(var s=0;s<a.length;s++)r[s]&&(r[s].data[o]=a[s])}return{series:r,categories:i}}function $U(t){for(var e=t.split(/\n+/g),n=BU(e.shift()),i=[],r=0;r<e.length;r++){var o=BU(e[r]);if(o){var a=o.split(VU),s="",l=void 0,u=!1;isNaN(a[0])?(u=!0,s=a[0],a=a.slice(1),i[r]={name:s,value:[]},l=i[r].value):l=i[r]=[];for(var c=0;c<a.length;c++)l.push(+a[c]);1===l.length&&(u?i[r].value=l[0]:i[r]=l[0])}}return{name:n,data:i}}function UU(t,e){var n=t.split(new RegExp("\n*"+NU+"\n*","g")),i={series:[]};return R(n,(function(t,n){if(FU(t)){var r=GU(t),o=e[n],a=o.axisDim+"Axis";o&&(i[a]=i[a]||[],i[a][o.axisIndex]={data:r.categories},i.series=i.series.concat(r.series))}else{r=$U(t);i.series.push(r)}})),i}var qU=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.onclick=function(t,e){var n=e.getDom(),i=this.model;this._dom&&n.removeChild(this._dom);var r=document.createElement("div");r.style.cssText="position:absolute;left:5px;top:5px;bottom:5px;right:5px;",r.style.backgroundColor=i.get("backgroundColor")||"#fff";var o=document.createElement("h4"),a=i.get("lang")||[];o.innerHTML=a[0]||i.get("title"),o.style.cssText="margin: 10px 20px;",o.style.color=i.get("textColor");var s=document.createElement("div"),l=document.createElement("textarea");s.style.cssText="display:block;width:100%;overflow:auto;";var u=i.get("optionToContent"),c=i.get("contentToOption"),d=WU(t);if("function"===typeof u){var h=u(e.getOption());"string"===typeof h?s.innerHTML=h:tt(h)&&s.appendChild(h)}else s.appendChild(l),l.readOnly=i.get("readOnly"),l.style.cssText="width:100%;height:100%;font-family:monospace;font-size:14px;line-height:1.6rem;",l.style.color=i.get("textColor"),l.style.borderColor=i.get("textareaBorderColor"),l.style.backgroundColor=i.get("textareaColor"),l.value=d.value;var f=d.meta,p=document.createElement("div");p.style.cssText="position:absolute;bottom:0;left:0;right:0;";var m="float:right;margin-right:20px;border:none;cursor:pointer;padding:2px 5px;font-size:12px;border-radius:3px",g=document.createElement("div"),y=document.createElement("div");m+=";background-color:"+i.get("buttonColor"),m+=";color:"+i.get("buttonTextColor");var v=this;function _(){n.removeChild(r),v._dom=null}de(g,"click",_),de(y,"click",(function(){if(null==c&&null!=u||null!=c&&null==u)_();else{var t;try{t="function"===typeof c?c(s,e.getOption()):UU(l.value,f)}catch(n){throw _(),new Error("Data view format error "+n)}t&&e.dispatchAction({type:"changeDataView",newOption:t}),_()}})),g.innerHTML=a[1],y.innerHTML=a[2],y.style.cssText=m,g.style.cssText=m,!i.get("readOnly")&&p.appendChild(y),p.appendChild(g),r.appendChild(o),r.appendChild(s),r.appendChild(p),s.style.height=n.clientHeight-80+"px",n.appendChild(r),this._dom=r},e.prototype.remove=function(t,e){this._dom&&e.getDom().removeChild(this._dom)},e.prototype.dispose=function(t,e){this.remove(t,e)},e.getDefaultOption=function(t){var e={show:!0,readOnly:!1,optionToContent:null,contentToOption:null,icon:"M17.5,17.3H33 M17.5,17.3H33 M45.4,29.5h-28 M11.5,2v56H51V14.8L38.4,2H11.5z M38.4,2.2v12.7H51 M45.4,41.7h-28",title:t.getLocale(["toolbox","dataView","title"]),lang:t.getLocale(["toolbox","dataView","lang"]),backgroundColor:"#fff",textColor:"#000",textareaColor:"#fff",textareaBorderColor:"#333",buttonColor:"#c23531",buttonTextColor:"#fff"};return e},e}(vU);function XU(t,e){return H(t,(function(t,n){var i=e&&e[n];if(J(i)&&!$(i)){var r=J(t)&&!$(t);r||(t={value:t});var o=null!=i.name&&null==t.name;return t=A(t,i),o&&delete t.name,t}return t}))}sM({type:"changeDataView",event:"dataViewChanged",update:"prepareAndUpdate"},(function(t,e){var n=[];R(t.newOption.series,(function(t){var i=e.getSeriesByName(t.name)[0];if(i){var r=i.get("data");n.push({name:t.name,data:XU(t.data,r)})}else n.push(I({type:"scatter"},t))})),e.mergeOption(A({series:n},t.newOption))}));var ZU=qU,JU=R,KU=Io();function QU(t,e){var n=iq(t);JU(e,(function(e,i){for(var r=n.length-1;r>=0;r--){var o=n[r];if(o[i])break}if(r<0){var a=t.queryComponents({mainType:"dataZoom",subType:"select",id:i})[0];if(a){var s=a.getPercentRange();n[0][i]={dataZoomId:i,start:s[0],end:s[1]}}}})),n.push(e)}function tq(t){var e=iq(t),n=e[e.length-1];e.length>1&&e.pop();var i={};return JU(n,(function(t,n){for(var r=e.length-1;r>=0;r--)if(t=e[r][n],t){i[n]=t;break}})),i}function eq(t){KU(t).snapshots=null}function nq(t){return iq(t).length}function iq(t){var e=KU(t);return e.snapshots||(e.snapshots=[{}]),e.snapshots}var rq=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.onclick=function(t,e){eq(t),e.dispatchAction({type:"restore",from:this.uid})},e.getDefaultOption=function(t){var e={show:!0,icon:"M3.8,33.4 M47,18.9h9.8V8.7 M56.3,20.1 C52.1,9,40.5,0.6,26.8,2.1C12.6,3.7,1.6,16.2,2.1,30.6 M13,41.1H3.1v10.2 M3.7,39.9c4.2,11.1,15.8,19.5,29.5,18 c14.2-1.6,25.2-14.1,24.7-28.5",title:t.getLocale(["toolbox","restore","title"])};return e},e}(vU);sM({type:"restore",event:"restore",update:"prepareAndUpdate"},(function(t,e){e.resetOption("recreate")}));var oq=rq,aq=["grid","xAxis","yAxis","geo","graph","polar","radiusAxis","angleAxis","bmap"],sq=function(){function t(t,e,n){var i=this;this._targetInfoList=[];var r=uq(e,t);R(cq,(function(t,e){(!n||!n.include||O(n.include,e)>=0)&&t(r,i._targetInfoList)}))}return t.prototype.setOutputRanges=function(t,e){return this.matchOutputRanges(t,e,(function(t,e,n){if((t.coordRanges||(t.coordRanges=[])).push(e),!t.coordRange){t.coordRange=e;var i=fq[t.brushType](0,n,e);t.__rangeOffset={offset:mq[t.brushType](i.values,t.range,[1,1]),xyMinMax:i.xyMinMax}}})),t},t.prototype.matchOutputRanges=function(t,e,n){R(t,(function(t){var i=this.findTargetInfo(t,e);i&&!0!==i&&R(i.coordSyses,(function(i){var r=fq[t.brushType](1,i,t.range,!0);n(t,r.values,i,e)}))}),this)},t.prototype.setInputRanges=function(t,e){R(t,(function(t){var n=this.findTargetInfo(t,e);if(t.range=t.range||[],n&&!0!==n){t.panelId=n.panelId;var i=fq[t.brushType](0,n.coordSys,t.coordRange),r=t.__rangeOffset;t.range=r?mq[t.brushType](i.values,r.offset,yq(i.xyMinMax,r.xyMinMax)):i.values}}),this)},t.prototype.makePanelOpts=function(t,e){return H(this._targetInfoList,(function(n){var i=n.getPanelRect();return{panelId:n.panelId,defaultBrushType:e?e(n):null,clipPath:bH(i),isTargetByCursor:wH(i,t,n.coordSysModel),getLinearBrushOtherExtent:xH(i)}}))},t.prototype.controlSeries=function(t,e,n){var i=this.findTargetInfo(t,n);return!0===i||i&&O(i.coordSyses,e.coordinateSystem)>=0},t.prototype.findTargetInfo=function(t,e){for(var n=this._targetInfoList,i=uq(e,t),r=0;r<n.length;r++){var o=n[r],a=t.panelId;if(a){if(o.panelId===a)return o}else for(var s=0;s<dq.length;s++)if(dq[s](i,o))return o}return!0},t}();function lq(t){return t[0]>t[1]&&t.reverse(),t}function uq(t,e){return Yo(t,e,{includeMainTypes:aq})}var cq={grid:function(t,e){var n=t.xAxisModels,i=t.yAxisModels,r=t.gridModels,o=gt(),a={},s={};(n||i||r)&&(R(n,(function(t){var e=t.axis.grid.model;o.set(e.id,e),a[e.id]=!0})),R(i,(function(t){var e=t.axis.grid.model;o.set(e.id,e),s[e.id]=!0})),R(r,(function(t){o.set(t.id,t),a[t.id]=!0,s[t.id]=!0})),o.each((function(t){var r=t.coordinateSystem,o=[];R(r.getCartesians(),(function(t,e){(O(n,t.getAxis("x").model)>=0||O(i,t.getAxis("y").model)>=0)&&o.push(t)})),e.push({panelId:"grid--"+t.id,gridModel:t,coordSysModel:t,coordSys:o[0],coordSyses:o,getPanelRect:hq.grid,xAxisDeclared:a[t.id],yAxisDeclared:s[t.id]})})))},geo:function(t,e){R(t.geoModels,(function(t){var n=t.coordinateSystem;e.push({panelId:"geo--"+t.id,geoModel:t,coordSysModel:t,coordSys:n,coordSyses:[n],getPanelRect:hq.geo})}))}},dq=[function(t,e){var n=t.xAxisModel,i=t.yAxisModel,r=t.gridModel;return!r&&n&&(r=n.axis.grid.model),!r&&i&&(r=i.axis.grid.model),r&&r===e.gridModel},function(t,e){var n=t.geoModel;return n&&n===e.geoModel}],hq={grid:function(){return this.coordSys.master.getRect().clone()},geo:function(){var t=this.coordSys,e=t.getBoundingRect().clone();return e.applyTransform(ah(t)),e}},fq={lineX:G(pq,0),lineY:G(pq,1),rect:function(t,e,n,i){var r=t?e.pointToData([n[0][0],n[1][0]],i):e.dataToPoint([n[0][0],n[1][0]],i),o=t?e.pointToData([n[0][1],n[1][1]],i):e.dataToPoint([n[0][1],n[1][1]],i),a=[lq([r[0],o[0]]),lq([r[1],o[1]])];return{values:a,xyMinMax:a}},polygon:function(t,e,n,i){var r=[[1/0,-1/0],[1/0,-1/0]],o=H(n,(function(n){var o=t?e.pointToData(n,i):e.dataToPoint(n,i);return r[0][0]=Math.min(r[0][0],o[0]),r[1][0]=Math.min(r[1][0],o[1]),r[0][1]=Math.max(r[0][1],o[0]),r[1][1]=Math.max(r[1][1],o[1]),o}));return{values:o,xyMinMax:r}}};function pq(t,e,n,i){var r=n.getAxis(["x","y"][t]),o=lq(H([0,1],(function(t){return e?r.coordToData(r.toLocalCoord(i[t]),!0):r.toGlobalCoord(r.dataToCoord(i[t]))}))),a=[];return a[t]=o,a[1-t]=[NaN,NaN],{values:o,xyMinMax:a}}var mq={lineX:G(gq,0),lineY:G(gq,1),rect:function(t,e,n){return[[t[0][0]-n[0]*e[0][0],t[0][1]-n[0]*e[0][1]],[t[1][0]-n[1]*e[1][0],t[1][1]-n[1]*e[1][1]]]},polygon:function(t,e,n){return H(t,(function(t,i){return[t[0]-n[0]*e[i][0],t[1]-n[1]*e[i][1]]}))}};function gq(t,e,n,i){return[e[0]-i[t]*n[0],e[1]-i[t]*n[1]]}function yq(t,e){var n=vq(t),i=vq(e),r=[n[0]/i[0],n[1]/i[1]];return isNaN(r[0])&&(r[0]=1),isNaN(r[1])&&(r[1]=1),r}function vq(t){return t?[t[0][1]-t[0][0],t[1][1]-t[1][0]]:[NaN,NaN]}var _q=sq,bq=R,xq=Lo("toolbox-dataZoom_"),wq=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.render=function(t,e,n,i){this._brushController||(this._brushController=new _H(n.getZr()),this._brushController.on("brush",V(this._onBrush,this)).mount()),kq(t,e,this,i,n),Lq(t,e)},e.prototype.onclick=function(t,e,n){Mq[n].call(this)},e.prototype.remove=function(t,e){this._brushController&&this._brushController.unmount()},e.prototype.dispose=function(t,e){this._brushController&&this._brushController.dispose()},e.prototype._onBrush=function(t){var e=t.areas;if(t.isEnd&&e.length){var n={},i=this.ecModel;this._brushController.updateCovers([]);var r=new _q(Sq(this.model),i,{include:["grid"]});r.matchOutputRanges(e,i,(function(t,e,n){if("cartesian2d"===n.type){var i=t.brushType;"rect"===i?(o("x",n,e[0]),o("y",n,e[1])):o({lineX:"x",lineY:"y"}[i],n,e)}})),QU(i,n),this._dispatchZoomAction(n)}function o(t,e,r){var o=e.getAxis(t),s=o.model,l=a(t,s,i),u=l.findRepresentativeAxisProxy(s).getMinMaxSpan();null==u.minValueSpan&&null==u.maxValueSpan||(r=rR(0,r.slice(),o.scale.getExtent(),0,u.minValueSpan,u.maxValueSpan)),l&&(n[l.id]={dataZoomId:l.id,startValue:r[0],endValue:r[1]})}function a(t,e,n){var i;return n.eachComponent({mainType:"dataZoom",subType:"select"},(function(n){var r=n.getAxisModel(t,e.componentIndex);r&&(i=n)})),i}},e.prototype._dispatchZoomAction=function(t){var e=[];bq(t,(function(t,n){e.push(T(t))})),e.length&&this.api.dispatchAction({type:"dataZoom",from:this.uid,batch:e})},e.getDefaultOption=function(t){var e={show:!0,filterMode:"filter",icon:{zoom:"M0,13.5h26.9 M13.5,26.9V0 M32.1,13.5H58V58H13.5 V32.1",back:"M22,1.4L9.9,13.5l12.3,12.3 M10.3,13.5H54.9v44.6 H10.3v-26"},title:t.getLocale(["toolbox","dataZoom","title"]),brushStyle:{borderWidth:0,color:"rgba(210,219,238,0.2)"}};return e},e}(vU),Mq={zoom:function(){var t=!this._isZoomActive;this.api.dispatchAction({type:"takeGlobalCursor",key:"dataZoomSelect",dataZoomSelectActive:t})},back:function(){this._dispatchZoomAction(tq(this.ecModel))}};function Sq(t){var e={xAxisIndex:t.get("xAxisIndex",!0),yAxisIndex:t.get("yAxisIndex",!0),xAxisId:t.get("xAxisId",!0),yAxisId:t.get("yAxisId",!0)};return null==e.xAxisIndex&&null==e.xAxisId&&(e.xAxisIndex="all"),null==e.yAxisIndex&&null==e.yAxisId&&(e.yAxisIndex="all"),e}function Lq(t,e){t.setIconStatus("back",nq(e)>1?"emphasis":"normal")}function kq(t,e,n,i,r){var o=n._isZoomActive;i&&"takeGlobalCursor"===i.type&&(o="dataZoomSelect"===i.key&&i.dataZoomSelectActive),n._isZoomActive=o,t.setIconStatus("zoom",o?"emphasis":"normal");var a=new _q(Sq(t),e,{include:["grid"]}),s=a.makePanelOpts(r,(function(t){return t.xAxisDeclared&&!t.yAxisDeclared?"lineX":!t.xAxisDeclared&&t.yAxisDeclared?"lineY":"rect"}));n._brushController.setPanels(s).enableBrush(!(!o||!s.length)&&{brushType:"auto",brushStyle:t.getModel("brushStyle").getItemStyle()})}Vp("dataZoom",(function(t){var e=t.getComponent("toolbox",0),n=["feature","dataZoom"];if(e&&null!=e.get(n)){var i=e.getModel(n),r=[],o=Sq(i),a=Yo(t,o);return bq(a.xAxisModels,(function(t){return s(t,"xAxis","xAxisIndex")})),bq(a.yAxisModels,(function(t){return s(t,"yAxis","yAxisIndex")})),r}function s(t,e,n){var o=t.componentIndex,a={type:"select",$fromToolbox:!0,filterMode:i.get("filterMode",!0)||"filter",id:xq+e+o};a[n]=o,r.push(a)}}));var Tq=wq;function Dq(t){t.registerComponentModel(MU),t.registerComponentView(DU),bU("saveAsImage",IU),bU("magicType",EU),bU("dataView",ZU),bU("dataZoom",Tq),bU("restore",oq),vM(yU)}var Cq=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="tooltip",e.dependencies=["axisPointer"],e.defaultOption={zlevel:0,z:60,show:!0,showContent:!0,trigger:"item",triggerOn:"mousemove|click",alwaysShowContent:!1,displayMode:"single",renderMode:"auto",confine:null,showDelay:0,hideDelay:100,transitionDuration:.4,enterable:!1,backgroundColor:"#fff",shadowBlur:10,shadowColor:"rgba(0, 0, 0, .2)",shadowOffsetX:1,shadowOffsetY:2,borderRadius:4,borderWidth:1,padding:null,extraCssText:"",axisPointer:{type:"line",axis:"auto",animation:"auto",animationDurationUpdate:200,animationEasingUpdate:"exponentialOut",crossStyle:{color:"#999",width:1,type:"dashed",textStyle:{}}},textStyle:{color:"#666",fontSize:14}},e}(xp),Iq=Cq;function Aq(t){var e=t.get("confine");return null!=e?!!e:"richText"===t.get("renderMode")}function Yq(t){if(h.domSupported)for(var e=document.documentElement.style,n=0,i=t.length;n<i;n++)if(t[n]in e)return t[n]}var Oq=Yq(["transform","webkitTransform","OTransform","MozTransform","msTransform"]),Pq=Yq(["webkitTransition","transition","OTransition","MozTransition","msTransition"]);function Eq(t,e){if(!t)return e;e=$f(e,!0);var n=t.indexOf(e);return t=-1===n?e:"-"+t.slice(0,n)+"-"+e,t.toLowerCase()}function Nq(t,e){var n=t.currentStyle||document.defaultView&&document.defaultView.getComputedStyle(t);return n?e?n[e]:n:null}var Rq=Eq(Pq,"transition"),Hq=Eq(Oq,"transform"),jq="position:absolute;display:block;border-style:solid;white-space:nowrap;z-index:9999999;"+(h.transform3dSupported?"will-change:transform;":"");function zq(t){return t="left"===t?"right":"right"===t?"left":"top"===t?"bottom":"top",t}function Wq(t,e,n){if(!q(n)||"inside"===n)return"";e=ip(e);var i=zq(n),r=i+":-6px;",o=Hq+":";O(["left","right"],i)>-1?(r+="top:50%",o+="translateY(-50%) rotate("+("left"===i?-225:-45)+"deg)"):(r+="left:50%",o+="translateX(-50%) rotate("+("top"===i?225:45)+"deg)");var a=e+" solid 1px;",s=["position:absolute;width:10px;height:10px;",r+";"+o+";","border-bottom:"+a,"border-right:"+a,"background-color:"+t+";","box-shadow:8px 8px 16px -3px #000;"];return'<div style="'+s.join("")+'"></div>'}function Bq(t,e){var n="cubic-bezier(0.23,1,0.32,1)",i=" "+t/2+"s "+n,r="opacity"+i+",visibility"+i;return e||(i=" "+t+"s "+n,r+=h.transformSupported?","+Hq+i:",left"+i+",top"+i),Rq+":"+r}function Fq(t,e,n){var i=t.toFixed(0)+"px",r=e.toFixed(0)+"px";if(!h.transformSupported)return n?"top:"+r+";left:"+i+";":[["top",r],["left",i]];var o=h.transform3dSupported,a="translate"+(o?"3d":"")+"("+i+","+r+(o?",0":"")+")";return n?"top:0;left:0;"+Hq+":"+a+";":[["top",0],["left",0],[Oq,a]]}function Vq(t){var e=[],n=t.get("fontSize"),i=t.getTextColor();i&&e.push("color:"+i),e.push("font:"+t.getFont()),n&&e.push("line-height:"+Math.round(3*n/2)+"px");var r=t.get("textShadowColor"),o=t.get("textShadowBlur")||0,a=t.get("textShadowOffsetX")||0,s=t.get("textShadowOffsetY")||0;return r&&o&&e.push("text-shadow:"+a+"px "+s+"px "+o+"px "+r),R(["decoration","align"],(function(n){var i=t.get(n);i&&e.push("text-"+n+":"+i)})),e.join(";")}function Gq(t,e,n){var i=[],r=t.get("transitionDuration"),o=t.get("backgroundColor"),a=t.get("shadowBlur"),s=t.get("shadowColor"),l=t.get("shadowOffsetX"),u=t.get("shadowOffsetY"),c=t.getModel("textStyle"),d=yy(t,"html"),f=l+"px "+u+"px "+a+"px "+s;return i.push("box-shadow:"+f),e&&r&&i.push(Bq(r,n)),o&&(h.canvasSupported?i.push("background-color:"+o):(i.push("background-color:#"+xn(o)),i.push("filter:alpha(opacity=70)"))),R(["width","color","radius"],(function(e){var n="border-"+e,r=$f(n),o=t.get(r);null!=o&&i.push(n+":"+o+("color"===e?"":"px"))})),i.push(Vq(c)),null!=d&&i.push("padding:"+Uf(d).join("px ")+"px"),i.join(";")+";"}function $q(t,e,n,i,r){var o=e&&e.painter;if(n){var a=o&&o.getViewportRoot();a&&Kt(t,a,document.body,i,r)}else{t[0]=i,t[1]=r;var s=o&&o.getViewportRootOffset();s&&(t[0]+=s.offsetLeft,t[1]+=s.offsetTop)}t[2]=t[0]/e.getWidth(),t[3]=t[1]/e.getHeight()}var Uq=function(){function t(t,e,n){if(this._show=!1,this._styleCoord=[0,0,0,0],this._enterable=!0,this._firstShow=!0,this._longHide=!0,h.wxa)return null;var i=document.createElement("div");i.domBelongToZr=!0,this.el=i;var r=this._zr=e.getZr(),o=this._appendToBody=n&&n.appendToBody;$q(this._styleCoord,r,o,e.getWidth()/2,e.getHeight()/2),o?document.body.appendChild(i):t.appendChild(i),this._container=t;var a=this;i.onmouseenter=function(){a._enterable&&(clearTimeout(a._hideTimeout),a._show=!0),a._inContent=!0},i.onmousemove=function(t){if(t=t||window.event,!a._enterable){var e=r.handler,n=r.painter.getViewportRoot();ue(n,t,!0),e.dispatch("mousemove",t)}},i.onmouseleave=function(){a._inContent=!1,a._enterable&&a._show&&a.hideLater(a._hideDelay)}}return t.prototype.update=function(t){var e=this._container,n=Nq(e,"position"),i=e.style;"absolute"!==i.position&&"absolute"!==n&&(i.position="relative");var r=t.get("alwaysShowContent");r&&this._moveIfResized(),this.el.className=t.get("className")||""},t.prototype.show=function(t,e){clearTimeout(this._hideTimeout),clearTimeout(this._longHideTimeout);var n=this.el,i=n.style,r=this._styleCoord;n.innerHTML?i.cssText=jq+Gq(t,!this._firstShow,this._longHide)+Fq(r[0],r[1],!0)+"border-color:"+ip(e)+";"+(t.get("extraCssText")||"")+";pointer-event:"+(this._enterable?"auto":"none"):i.display="none",this._show=!0,this._firstShow=!1,this._longHide=!1},t.prototype.setContent=function(t,e,n,i,r){if(null!=t){var o=this.el;if(q(r)&&"item"===n.get("trigger")&&!Aq(n)&&(t+=Wq(n.get("backgroundColor"),i,r)),q(t))o.innerHTML=t;else if(t){o.innerHTML="",$(t)||(t=[t]);for(var a=0;a<t.length;a++)tt(t[a])&&t[a].parentNode!==o&&o.appendChild(t[a])}}},t.prototype.setEnterable=function(t){this._enterable=t},t.prototype.getSize=function(){var t=this.el;return[t.clientWidth,t.clientHeight]},t.prototype.moveTo=function(t,e){var n=this._styleCoord;if($q(n,this._zr,this._appendToBody,t,e),null!=n[0]&&null!=n[1]){var i=this.el.style,r=Fq(n[0],n[1]);R(r,(function(t){i[t[0]]=t[1]}))}},t.prototype._moveIfResized=function(){var t=this._styleCoord[2],e=this._styleCoord[3];this.moveTo(t*this._zr.getWidth(),e*this._zr.getHeight())},t.prototype.hide=function(){var t=this,e=this.el.style;e.visibility="hidden",e.opacity="0",h.transform3dSupported&&(e.willChange=""),this._show=!1,this._longHideTimeout=setTimeout((function(){return t._longHide=!0}),500)},t.prototype.hideLater=function(t){!this._show||this._inContent&&this._enterable||(t?(this._hideDelay=t,this._show=!1,this._hideTimeout=setTimeout(V(this.hide,this),t)):this.hide())},t.prototype.isShow=function(){return this._show},t.prototype.dispose=function(){this.el.parentNode.removeChild(this.el)},t.prototype.getOuterSize=function(){var t=this.el.clientWidth,e=this.el.clientHeight,n=Nq(this.el);return n&&(t+=parseInt(n.borderLeftWidth,10)+parseInt(n.borderRightWidth,10),e+=parseInt(n.borderTopWidth,10)+parseInt(n.borderBottomWidth,10)),{width:t,height:e}},t}(),qq=Uq,Xq=function(){function t(t){this._show=!1,this._styleCoord=[0,0,0,0],this._enterable=!0,this._zr=t.getZr(),Kq(this._styleCoord,this._zr,t.getWidth()/2,t.getHeight()/2)}return t.prototype.update=function(t){var e=t.get("alwaysShowContent");e&&this._moveIfResized()},t.prototype.show=function(){this._hideTimeout&&clearTimeout(this._hideTimeout),this.el.show(),this._show=!0},t.prototype.setContent=function(t,e,n,i,r){J(t)&&im(""),this.el&&this._zr.remove(this.el);var o=n.getModel("textStyle");this.el=new Gl({style:{rich:e.richTextStyles,text:t,lineHeight:22,backgroundColor:n.get("backgroundColor"),borderRadius:n.get("borderRadius"),borderWidth:1,borderColor:i,shadowColor:n.get("shadowColor"),shadowBlur:n.get("shadowBlur"),shadowOffsetX:n.get("shadowOffsetX"),shadowOffsetY:n.get("shadowOffsetY"),textShadowColor:o.get("textShadowColor"),textShadowBlur:o.get("textShadowBlur")||0,textShadowOffsetX:o.get("textShadowOffsetX")||0,textShadowOffsetY:o.get("textShadowOffsetY")||0,fill:n.get(["textStyle","color"]),padding:yy(n,"richText"),verticalAlign:"top",align:"left"},z:n.get("z")}),this._zr.add(this.el);var a=this;this.el.on("mouseover",(function(){a._enterable&&(clearTimeout(a._hideTimeout),a._show=!0),a._inContent=!0})),this.el.on("mouseout",(function(){a._enterable&&a._show&&a.hideLater(a._hideDelay),a._inContent=!1}))},t.prototype.setEnterable=function(t){this._enterable=t},t.prototype.getSize=function(){var t=this.el,e=this.el.getBoundingRect(),n=Jq(t.style);return[e.width+n.left+n.right,e.height+n.top+n.bottom]},t.prototype.moveTo=function(t,e){var n=this.el;if(n){var i=this._styleCoord;Kq(i,this._zr,t,e),t=i[0],e=i[1];var r=n.style,o=Zq(r.borderWidth||0),a=Jq(r);n.x=t+o+a.left,n.y=e+o+a.top,n.markRedraw()}},t.prototype._moveIfResized=function(){var t=this._styleCoord[2],e=this._styleCoord[3];this.moveTo(t*this._zr.getWidth(),e*this._zr.getHeight())},t.prototype.hide=function(){this.el&&this.el.hide(),this._show=!1},t.prototype.hideLater=function(t){!this._show||this._inContent&&this._enterable||(t?(this._hideDelay=t,this._show=!1,this._hideTimeout=setTimeout(V(this.hide,this),t)):this.hide())},t.prototype.isShow=function(){return this._show},t.prototype.getOuterSize=function(){var t=this.getSize();return{width:t[0],height:t[1]}},t.prototype.dispose=function(){this._zr.remove(this.el)},t}();function Zq(t){return Math.max(0,t)}function Jq(t){var e=Zq(t.shadowBlur||0),n=Zq(t.shadowOffsetX||0),i=Zq(t.shadowOffsetY||0);return{left:Zq(e-n),right:Zq(e+n),top:Zq(e-i),bottom:Zq(e+i)}}function Kq(t,e,n,i){t[0]=n,t[1]=i,t[2]=t[0]/e.getWidth(),t[3]=t[1]/e.getHeight()}var Qq=Xq,tX=V,eX=R,nX=Nr,iX=new Al({shape:{x:-1,y:-1,width:2,height:2}}),rX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){if(!h.node){var n=t.getComponent("tooltip"),i=n.get("renderMode");this._renderMode=jo(i),this._tooltipContent="richText"===this._renderMode?new Qq(e):new qq(e.getDom(),e,{appendToBody:n.get("appendToBody",!0)})}},e.prototype.render=function(t,e,n){if(!h.node){this.group.removeAll(),this._tooltipModel=t,this._ecModel=e,this._api=n,this._alwaysShowContent=t.get("alwaysShowContent");var i=this._tooltipContent;i.update(t),i.setEnterable(t.get("enterable")),this._initGlobalListener(),this._keepShow()}},e.prototype._initGlobalListener=function(){var t=this._tooltipModel,e=t.get("triggerOn");OV("itemTooltip",this._api,tX((function(t,n,i){"none"!==e&&(e.indexOf(t)>=0?this._tryShow(n,i):"leave"===t&&this._hide(i))}),this))},e.prototype._keepShow=function(){var t=this._tooltipModel,e=this._ecModel,n=this._api;if(null!=this._lastX&&null!=this._lastY&&"none"!==t.get("triggerOn")){var i=this;clearTimeout(this._refreshUpdateTimeout),this._refreshUpdateTimeout=setTimeout((function(){!n.isDisposed()&&i.manuallyShowTip(t,e,n,{x:i._lastX,y:i._lastY,dataByCoordSys:i._lastDataByCoordSys})}))}},e.prototype.manuallyShowTip=function(t,e,n,i){if(i.from!==this.uid&&!h.node){var r=aX(i,n);this._ticket="";var o=i.dataByCoordSys,a=dX(i,e,n);if(a){var s=a.el.getBoundingRect().clone();s.applyTransform(a.el.transform),this._tryShow({offsetX:s.x+s.width/2,offsetY:s.y+s.height/2,target:a.el,position:i.position,positionDefault:"bottom"},r)}else if(i.tooltip&&null!=i.x&&null!=i.y){var l=iX;l.x=i.x,l.y=i.y,l.update(),$l(l).tooltipConfig={name:null,option:i.tooltip},this._tryShow({offsetX:i.x,offsetY:i.y,target:l},r)}else if(o)this._tryShow({offsetX:i.x,offsetY:i.y,position:i.position,dataByCoordSys:o,tooltipOption:i.tooltipOption},r);else if(null!=i.seriesIndex){if(this._manuallyAxisShowTip(t,e,n,i))return;var u=BV(i,e),c=u.point[0],d=u.point[1];null!=c&&null!=d&&this._tryShow({offsetX:c,offsetY:d,target:u.el,position:i.position,positionDefault:"bottom"},r)}else null!=i.x&&null!=i.y&&(n.dispatchAction({type:"updateAxisPointer",x:i.x,y:i.y}),this._tryShow({offsetX:i.x,offsetY:i.y,position:i.position,target:n.getZr().findHover(i.x,i.y).target},r))}},e.prototype.manuallyHideTip=function(t,e,n,i){var r=this._tooltipContent;!this._alwaysShowContent&&this._tooltipModel&&r.hideLater(this._tooltipModel.get("hideDelay")),this._lastX=this._lastY=this._lastDataByCoordSys=null,i.from!==this.uid&&this._hide(aX(i,n))},e.prototype._manuallyAxisShowTip=function(t,e,n,i){var r=i.seriesIndex,o=i.dataIndex,a=e.getComponent("axisPointer").coordSysAxesInfo;if(null!=r&&null!=o&&null!=a){var s=e.getSeriesByIndex(r);if(s){var l=s.getData(),u=oX([l.getItemModel(o),s,(s.coordinateSystem||{}).model],this._tooltipModel);if("axis"===u.get("trigger"))return n.dispatchAction({type:"updateAxisPointer",seriesIndex:r,dataIndex:o,position:i.position}),!0}}},e.prototype._tryShow=function(t,e){var n=t.target,i=this._tooltipModel;if(i){this._lastX=t.offsetX,this._lastY=t.offsetY;var r=t.dataByCoordSys;if(r&&r.length)this._showAxisTooltip(r,t);else if(n){var o,a;this._lastDataByCoordSys=null,k_(n,(function(t){return null!=$l(t).dataIndex?(o=t,!0):null!=$l(t).tooltipConfig?(a=t,!0):void 0}),!0),o?this._showSeriesItemTooltip(t,o,e):a?this._showComponentItemTooltip(t,a,e):this._hide(e)}else this._lastDataByCoordSys=null,this._hide(e)}},e.prototype._showOrMove=function(t,e){var n=t.get("showDelay");e=V(e,this),clearTimeout(this._showTimout),n>0?this._showTimout=setTimeout(e,n):e()},e.prototype._showAxisTooltip=function(t,e){var n=this._ecModel,i=this._tooltipModel,r=[e.offsetX,e.offsetY],o=oX([e.tooltipOption],i),a=this._renderMode,s=[],l=oy("section",{blocks:[],noHeader:!0}),u=[],c=new vy;eX(t,(function(t){eX(t.dataByAxis,(function(t){var e=n.getComponent(t.axisDim+"Axis",t.axisIndex),i=t.value;if(e&&null!=i){var r=vV(i,e.axis,n,t.seriesDataIndices,t.valueLabelOpt),o=oy("section",{header:r,noHeader:!dt(r),sortBlocks:!0,blocks:[]});l.blocks.push(o),R(t.seriesDataIndices,(function(l){var d=n.getSeriesByIndex(l.seriesIndex),h=l.dataIndexInside,f=d.getDataParams(h);f.axisDim=t.axisDim,f.axisIndex=t.axisIndex,f.axisType=t.axisType,f.axisId=t.axisId,f.axisValue=wC(e.axis,{value:i}),f.axisValueLabel=r,f.marker=c.makeTooltipMarker("item",ip(f.color),a);var p=Dg(d.formatTooltip(h,!0,null));p.markupFragment&&o.blocks.push(p.markupFragment),p.markupText&&u.push(p.markupText),s.push(f)}))}}))})),l.blocks.reverse(),u.reverse();var d=e.position,h=o.get("order"),f=uy(l,c,a,h,n.get("useUTC"),o.get("textStyle"));f&&u.unshift(f);var p="richText"===a?"\n\n":"<br/>",m=u.join(p);this._showOrMove(o,(function(){this._updateContentNotChangedOnAxis(t)?this._updatePosition(o,d,r[0],r[1],this._tooltipContent,s):this._showTooltipContent(o,m,s,Math.random()+"",r[0],r[1],d,null,c)}))},e.prototype._showSeriesItemTooltip=function(t,e,n){var i=this._ecModel,r=$l(e),o=r.seriesIndex,a=i.getSeriesByIndex(o),s=r.dataModel||a,l=r.dataIndex,u=r.dataType,c=s.getData(u),d=this._renderMode,h=t.positionDefault,f=oX([c.getItemModel(l),s,a&&(a.coordinateSystem||{}).model],this._tooltipModel,h?{position:h}:null),p=f.get("trigger");if(null==p||"item"===p){var m=s.getDataParams(l,u),g=new vy;m.marker=g.makeTooltipMarker("item",ip(m.color),d);var y=Dg(s.formatTooltip(l,!1,u)),v=f.get("order"),_=y.markupFragment?uy(y.markupFragment,g,d,v,i.get("useUTC"),f.get("textStyle")):y.markupText,b="item_"+s.name+"_"+l;this._showOrMove(f,(function(){this._showTooltipContent(f,_,m,b,t.offsetX,t.offsetY,t.position,t.target,g)})),n({type:"showTip",dataIndexInside:l,dataIndex:c.getRawIndex(l),seriesIndex:o,from:this.uid})}},e.prototype._showComponentItemTooltip=function(t,e,n){var i=$l(e),r=i.tooltipConfig,o=r.option||{};if(q(o)){var a=o;o={content:a,formatter:a}}var s=[o],l=this._ecModel.getComponent(i.componentMainType,i.componentIndex);l&&s.push(l),s.push({formatter:o.content});var u=t.positionDefault,c=oX(s,this._tooltipModel,u?{position:u}:null),d=c.get("content"),h=Math.random()+"",f=new vy;this._showOrMove(c,(function(){var n=T(c.get("formatterParams")||{});this._showTooltipContent(c,d,n,h,t.offsetX,t.offsetY,t.position,e,f)})),n({type:"showTip",from:this.uid})},e.prototype._showTooltipContent=function(t,e,n,i,r,o,a,s,l){if(this._ticket="",t.get("showContent")&&t.get("show")){var u=this._tooltipContent,c=t.get("formatter");a=a||t.get("position");var d=e,h=this._getNearestPoint([r,o],n,t.get("trigger"),t.get("borderColor")),f=h.color;if(c&&q(c)){var p=t.ecModel.get("useUTC"),m=$(n)?n[0]:n,g=m&&m.axisType&&m.axisType.indexOf("time")>=0;d=c,g&&(d=Tf(m.axisValue,d,p)),d=tp(d,n,!0)}else if(U(c)){var y=tX((function(e,i){e===this._ticket&&(u.setContent(i,l,t,f,a),this._updatePosition(t,a,r,o,u,n,s))}),this);this._ticket=i,d=c(n,i,y)}u.setContent(d,l,t,f,a),u.show(t,f),this._updatePosition(t,a,r,o,u,n,s)}},e.prototype._getNearestPoint=function(t,e,n,i){return"axis"===n||$(e)?{color:i||("html"===this._renderMode?"#fff":"none")}:$(e)?void 0:{color:i||e.color||e.borderColor}},e.prototype._updatePosition=function(t,e,n,i,r,o,a){var s=this._api.getWidth(),l=this._api.getHeight();e=e||t.get("position");var u=r.getSize(),c=t.get("align"),d=t.get("verticalAlign"),h=a&&a.getBoundingRect().clone();if(a&&h.applyTransform(a.transform),U(e)&&(e=e([n,i],o,r.el,h,{viewSize:[s,l],contentSize:u.slice()})),$(e))n=nX(e[0],s),i=nX(e[1],l);else if(J(e)){var f=e;f.width=u[0],f.height=u[1];var p=dp(f,{width:s,height:l});n=p.x,i=p.y,c=null,d=null}else if(q(e)&&a){var m=uX(e,h,u);n=m[0],i=m[1]}else{m=sX(n,i,r,s,l,c?null:20,d?null:20);n=m[0],i=m[1]}if(c&&(n-=cX(c)?u[0]/2:"right"===c?u[0]:0),d&&(i-=cX(d)?u[1]/2:"bottom"===d?u[1]:0),Aq(t)){m=lX(n,i,r,s,l);n=m[0],i=m[1]}r.moveTo(n,i)},e.prototype._updateContentNotChangedOnAxis=function(t){var e=this._lastDataByCoordSys,n=!!e&&e.length===t.length;return n&&eX(e,(function(e,i){var r=e.dataByAxis||[],o=t[i]||{},a=o.dataByAxis||[];n=n&&r.length===a.length,n&&eX(r,(function(t,e){var i=a[e]||{},r=t.seriesDataIndices||[],o=i.seriesDataIndices||[];n=n&&t.value===i.value&&t.axisType===i.axisType&&t.axisId===i.axisId&&r.length===o.length,n&&eX(r,(function(t,e){var i=o[e];n=n&&t.seriesIndex===i.seriesIndex&&t.dataIndex===i.dataIndex}))}))})),this._lastDataByCoordSys=t,!!n},e.prototype._hide=function(t){this._lastDataByCoordSys=null,t({type:"hideTip",from:this.uid})},e.prototype.dispose=function(t,e){h.node||(this._tooltipContent.dispose(),jV("itemTooltip",e))},e.type="tooltip",e}(Py);function oX(t,e,n){var i,r=e.ecModel;n?(i=new qh(n,r,r),i=new qh(e.option,i,r)):i=e;for(var o=t.length-1;o>=0;o--){var a=t[o];a&&(a instanceof qh&&(a=a.get("tooltip",!0)),q(a)&&(a={formatter:a}),a&&(i=new qh(a,i,r)))}return i}function aX(t,e){return t.dispatchAction||V(e.dispatchAction,e)}function sX(t,e,n,i,r,o,a){var s=n.getOuterSize(),l=s.width,u=s.height;return null!=o&&(t+l+o+2>i?t-=l+o:t+=o),null!=a&&(e+u+a>r?e-=u+a:e+=a),[t,e]}function lX(t,e,n,i,r){var o=n.getOuterSize(),a=o.width,s=o.height;return t=Math.min(t+a,i)-a,e=Math.min(e+s,r)-s,t=Math.max(t,0),e=Math.max(e,0),[t,e]}function uX(t,e,n){var i=n[0],r=n[1],o=10,a=5,s=0,l=0,u=e.width,c=e.height;switch(t){case"inside":s=e.x+u/2-i/2,l=e.y+c/2-r/2;break;case"top":s=e.x+u/2-i/2,l=e.y-r-o;break;case"bottom":s=e.x+u/2-i/2,l=e.y+c+o;break;case"left":s=e.x-i-o-a,l=e.y+c/2-r/2;break;case"right":s=e.x+u+o+a,l=e.y+c/2-r/2}return[s,l]}function cX(t){return"center"===t||"middle"===t}function dX(t,e,n){var i=Oo(t).queryOptionMap,r=i.keys()[0];if(r&&"series"!==r){var o=No(e,r,i.get(r),{useDefault:!1,enableAll:!1,enableNone:!1}),a=o.models[0];if(a){var s,l=n.getViewOfComponentModel(a);return l.group.traverse((function(e){var n=$l(e).tooltipConfig;if(n&&n.name===t.name)return s=e,!0})),s?{componentMainType:r,componentIndex:a.componentIndex,el:s}:void 0}}}var hX=rX;function fX(t){vM(eG),t.registerComponentModel(Iq),t.registerComponentView(hX),t.registerAction({type:"showTip",event:"showTip",update:"tooltip:manuallyShowTip"},(function(){})),t.registerAction({type:"hideTip",event:"hideTip",update:"tooltip:manuallyHideTip"},(function(){}))}var pX=["rect","polygon","keep","clear"];function mX(t,e){var n=so(t?t.brush:[]);if(n.length){var i=[];R(n,(function(t){var e=t.hasOwnProperty("toolbox")?t.toolbox:[];e instanceof Array&&(i=i.concat(e))}));var r=t&&t.toolbox;$(r)&&(r=r[0]),r||(r={feature:{}},t.toolbox=[r]);var o=r.feature||(r.feature={}),a=o.brush||(o.brush={}),s=a.type||(a.type=[]);s.push.apply(s,i),gX(s),e&&!s.length&&s.push.apply(s,pX)}}function gX(t){var e={};R(t,(function(t){e[t]=1})),t.length=0,R(e,(function(e,n){t.push(n)}))}var yX=R;function vX(t){if(t)for(var e in t)if(t.hasOwnProperty(e))return!0}function _X(t,e,n){var i={};return yX(e,(function(e){var o=i[e]=r();yX(t[e],(function(t,i){if(hP.isValidType(i)){var r={type:i,visual:t};n&&n(r,e),o[i]=new hP(r),"opacity"===i&&(r=T(r),r.type="colorAlpha",o.__hidden.__alphaForOpacity=new hP(r))}}))})),i;function r(){var t=function(){};t.prototype.__hidden=t.prototype;var e=new t;return e}}function bX(t,e,n){var i;R(n,(function(t){e.hasOwnProperty(t)&&vX(e[t])&&(i=!0)})),i&&R(n,(function(n){e.hasOwnProperty(n)&&vX(e[n])?t[n]=T(e[n]):delete t[n]}))}function xX(t,e,n,i,r,o){var a,s={};function l(t){return Ev(n,a,t)}function u(t,e){Rv(n,a,t,e)}function c(t,c){a=null==o?t:c;var d=n.getRawDataItem(a);if(!d||!1!==d.visualMap)for(var h=i.call(r,t),f=e[h],p=s[h],m=0,g=p.length;m<g;m++){var y=p[m];f[y]&&f[y].applyVisual(t,l,u)}}R(t,(function(t){var n=hP.prepareVisualTypes(e[t]);s[t]=n})),null==o?n.each(c):n.each([o],c)}function wX(t,e,n,i){var r={};return R(t,(function(t){var n=hP.prepareVisualTypes(e[t]);r[t]=n})),{progress:function(t,o){var a,s;function l(t){return Ev(o,s,t)}function u(t,e){Rv(o,s,t,e)}null!=i&&(a=o.getDimension(i));while(null!=(s=t.next())){var c=o.getRawDataItem(s);if(!c||!1!==c.visualMap)for(var d=null!=i?o.get(a,s):s,h=n(d),f=e[h],p=r[h],m=0,g=p.length;m<g;m++){var y=p[m];f[y]&&f[y].applyVisual(d,l,u)}}}}}function MX(t){var e=t.brushType,n={point:function(i){return SX[e].point(i,n,t)},rect:function(i){return SX[e].rect(i,n,t)}};return n}var SX={lineX:LX(0),lineY:LX(1),rect:{point:function(t,e,n){return t&&n.boundingRect.contain(t[0],t[1])},rect:function(t,e,n){return t&&n.boundingRect.intersect(t)}},polygon:{point:function(t,e,n){return t&&n.boundingRect.contain(t[0],t[1])&&ix(n.range,t[0],t[1])},rect:function(t,e,n){var i=n.range;if(!t||i.length<=1)return!1;var r=t.x,o=t.y,a=t.width,s=t.height,l=i[0];return!!(ix(i,r,o)||ix(i,r+a,o)||ix(i,r,o+s)||ix(i,r+a,o+s)||Ji.create(t).contain(l[0],l[1])||mh(r,o,r+a,o,i)||mh(r,o,r,o+s,i)||mh(r+a,o,r+a,o+s,i)||mh(r,o+s,r+a,o+s,i))||void 0}}};function LX(t){var e=["x","y"],n=["width","height"];return{point:function(e,n,i){if(e){var r=i.range,o=e[t];return kX(o,r)}},rect:function(i,r,o){if(i){var a=o.range,s=[i[e[t]],i[e[t]]+i[n[t]]];return s[1]<s[0]&&s.reverse(),kX(s[0],a)||kX(s[1],a)||kX(a[0],s)||kX(a[1],s)}}}}function kX(t,e){return e[0]<=t&&t<=e[1]}var TX=["inBrush","outOfBrush"],DX="__ecBrushSelect",CX="__ecInBrushSelectEvent";function IX(t){t.eachComponent({mainType:"brush"},(function(e){var n=e.brushTargetManager=new _q(e.option,t);n.setInputRanges(e.areas,t)}))}function AX(t,e,n){var i,r,o=[];t.eachComponent({mainType:"brush"},(function(t){n&&"takeGlobalCursor"===n.type&&t.setBrushOption("brush"===n.key?n.brushOption:{brushType:!1})})),IX(t),t.eachComponent({mainType:"brush"},(function(e,n){var a={brushId:e.id,brushIndex:n,brushName:e.name,areas:T(e.areas),selected:[]};o.push(a);var s=e.option,l=s.brushLink,u=[],c=[],d=[],h=!1;n||(i=s.throttleType,r=s.throttleDelay);var f=H(e.areas,(function(t){var e=NX[t.brushType],n=A({boundingRect:e?e(t):void 0},t);return n.selectors=MX(n),n})),p=_X(e.option,TX,(function(t){t.mappingMethod="fixed"}));function m(t){return"all"===l||!!u[t]}function g(t){return!!t.length}function y(t,e){var n=t.coordinateSystem;h=h||n.hasAxisBrushed(),m(e)&&n.eachActiveState(t.getData(),(function(t,e){"active"===t&&(c[e]=1)}))}function v(n,i,r){if(n.brushSelector&&!EX(e,i)&&(R(f,(function(i){e.brushTargetManager.controlSeries(i,n,t)&&r.push(i),h=h||g(r)})),m(i)&&g(r))){var o=n.getData();o.each((function(t){PX(n,r,o,t)&&(c[t]=1)}))}}$(l)&&R(l,(function(t){u[t]=1})),t.eachSeries((function(t,e){var n=d[e]=[];"parallel"===t.subType?y(t,e):v(t,e,n)})),t.eachSeries((function(t,e){var n={seriesId:t.id,seriesIndex:e,seriesName:t.name,dataIndex:[]};a.selected.push(n);var i=d[e],r=t.getData(),o=m(e)?function(t){return c[t]?(n.dataIndex.push(r.getRawIndex(t)),"inBrush"):"outOfBrush"}:function(e){return PX(t,i,r,e)?(n.dataIndex.push(r.getRawIndex(e)),"inBrush"):"outOfBrush"};(m(e)?h:g(i))&&xX(TX,p,r,o)}))})),YX(e,i,r,o,n)}function YX(t,e,n,i,r){if(r){var o=t.getZr();if(!o[CX]){o[DX]||(o[DX]=OX);var a=Xy(o,DX,n,e);a(t,i)}}}function OX(t,e){if(!t.isDisposed()){var n=t.getZr();n[CX]=!0,t.dispatchAction({type:"brushSelect",batch:e}),n[CX]=!1}}function PX(t,e,n,i){for(var r=0,o=e.length;r<o;r++){var a=e[r];if(t.brushSelector(i,n,a.selectors,a))return!0}}function EX(t,e){var n=t.option.seriesIndex;return null!=n&&"all"!==n&&($(n)?O(n,e)<0:e!==n)}var NX={rect:function(t){return RX(t.range)},polygon:function(t){for(var e,n=t.range,i=0,r=n.length;i<r;i++){e=e||[[1/0,-1/0],[1/0,-1/0]];var o=n[i];o[0]<e[0][0]&&(e[0][0]=o[0]),o[0]>e[0][1]&&(e[0][1]=o[0]),o[1]<e[1][0]&&(e[1][0]=o[1]),o[1]>e[1][1]&&(e[1][1]=o[1])}return e&&RX(e)}};function RX(t){return new Ji(t[0][0],t[1][0],t[0][1]-t[0][0],t[1][1]-t[1][0])}var HX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){this.ecModel=t,this.api=e,this.model,(this._brushController=new _H(e.getZr())).on("brush",V(this._onBrush,this)).mount()},e.prototype.render=function(t,e,n,i){this.model=t,this._updateController(t,e,n,i)},e.prototype.updateTransform=function(t,e,n,i){IX(e),this._updateController(t,e,n,i)},e.prototype.updateVisual=function(t,e,n,i){this.updateTransform(t,e,n,i)},e.prototype.updateView=function(t,e,n,i){this._updateController(t,e,n,i)},e.prototype._updateController=function(t,e,n,i){(!i||i.$from!==t.id)&&this._brushController.setPanels(t.brushTargetManager.makePanelOpts(n)).enableBrush(t.brushOption).updateCovers(t.areas.slice())},e.prototype.dispose=function(){this._brushController.dispose()},e.prototype._onBrush=function(t){var e=this.model.id,n=this.model.brushTargetManager.setOutputRanges(t.areas,this.ecModel);(!t.isEnd||t.removeOnClick)&&this.api.dispatchAction({type:"brush",brushId:e,areas:T(n),$from:e}),t.isEnd&&this.api.dispatchAction({type:"brushEnd",brushId:e,areas:T(n),$from:e})},e.type="brush",e}(Py),jX=HX,zX="#ddd",WX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.areas=[],n.brushOption={},n}return o(e,t),e.prototype.optionUpdated=function(t,e){var n=this.option;!e&&bX(n,t,["inBrush","outOfBrush"]);var i=n.inBrush=n.inBrush||{};n.outOfBrush=n.outOfBrush||{color:zX},i.hasOwnProperty("liftZ")||(i.liftZ=5)},e.prototype.setAreas=function(t){t&&(this.areas=H(t,(function(t){return BX(this.option,t)}),this))},e.prototype.setBrushOption=function(t){this.brushOption=BX(this.option,t),this.brushType=this.brushOption.brushType},e.type="brush",e.dependencies=["geo","grid","xAxis","yAxis","parallel","series"],e.defaultOption={seriesIndex:"all",brushType:"rect",brushMode:"single",transformable:!0,brushStyle:{borderWidth:1,color:"rgba(210,219,238,0.3)",borderColor:"#D2DBEE"},throttleType:"fixRate",throttleDelay:0,removeOnClick:!0,z:1e4},e}(xp);function BX(t,e){return D({brushType:t.brushType,brushMode:t.brushMode,transformable:t.transformable,brushStyle:new qh(t.brushStyle).getItemStyle(),removeOnClick:t.removeOnClick,z:t.z},e,!0)}var FX=WX,VX=["rect","polygon","lineX","lineY","keep","clear"],GX=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return o(e,t),e.prototype.render=function(t,e,n){var i,r,o;e.eachComponent({mainType:"brush"},(function(t){i=t.brushType,r=t.brushOption.brushMode||"single",o=o||!!t.areas.length})),this._brushType=i,this._brushMode=r,R(t.get("type",!0),(function(e){t.setIconStatus(e,("keep"===e?"multiple"===r:"clear"===e?o:e===i)?"emphasis":"normal")}))},e.prototype.updateView=function(t,e,n){this.render(t,e,n)},e.prototype.getIcons=function(){var t=this.model,e=t.get("icon",!0),n={};return R(t.get("type",!0),(function(t){e[t]&&(n[t]=e[t])})),n},e.prototype.onclick=function(t,e,n){var i=this._brushType,r=this._brushMode;"clear"===n?(e.dispatchAction({type:"axisAreaSelect",intervals:[]}),e.dispatchAction({type:"brush",command:"clear",areas:[]})):e.dispatchAction({type:"takeGlobalCursor",key:"brush",brushOption:{brushType:"keep"===n?i:i!==n&&n,brushMode:"keep"===n?"multiple"===r?"single":"multiple":r}})},e.getDefaultOption=function(t){var e={show:!0,type:VX.slice(),icon:{rect:"M7.3,34.7 M0.4,10V-0.2h9.8 M89.6,10V-0.2h-9.8 M0.4,60v10.2h9.8 M89.6,60v10.2h-9.8 M12.3,22.4V10.5h13.1 M33.6,10.5h7.8 M49.1,10.5h7.8 M77.5,22.4V10.5h-13 M12.3,31.1v8.2 M77.7,31.1v8.2 M12.3,47.6v11.9h13.1 M33.6,59.5h7.6 M49.1,59.5 h7.7 M77.5,47.6v11.9h-13",polygon:"M55.2,34.9c1.7,0,3.1,1.4,3.1,3.1s-1.4,3.1-3.1,3.1 s-3.1-1.4-3.1-3.1S53.5,34.9,55.2,34.9z M50.4,51c1.7,0,3.1,1.4,3.1,3.1c0,1.7-1.4,3.1-3.1,3.1c-1.7,0-3.1-1.4-3.1-3.1 C47.3,52.4,48.7,51,50.4,51z M55.6,37.1l1.5-7.8 M60.1,13.5l1.6-8.7l-7.8,4 M59,19l-1,5.3 M24,16.1l6.4,4.9l6.4-3.3 M48.5,11.6 l-5.9,3.1 M19.1,12.8L9.7,5.1l1.1,7.7 M13.4,29.8l1,7.3l6.6,1.6 M11.6,18.4l1,6.1 M32.8,41.9 M26.6,40.4 M27.3,40.2l6.1,1.6 M49.9,52.1l-5.6-7.6l-4.9-1.2",lineX:"M15.2,30 M19.7,15.6V1.9H29 M34.8,1.9H40.4 M55.3,15.6V1.9H45.9 M19.7,44.4V58.1H29 M34.8,58.1H40.4 M55.3,44.4 V58.1H45.9 M12.5,20.3l-9.4,9.6l9.6,9.8 M3.1,29.9h16.5 M62.5,20.3l9.4,9.6L62.3,39.7 M71.9,29.9H55.4",lineY:"M38.8,7.7 M52.7,12h13.2v9 M65.9,26.6V32 M52.7,46.3h13.2v-9 M24.9,12H11.8v9 M11.8,26.6V32 M24.9,46.3H11.8v-9 M48.2,5.1l-9.3-9l-9.4,9.2 M38.9-3.9V12 M48.2,53.3l-9.3,9l-9.4-9.2 M38.9,62.3V46.4",keep:"M4,10.5V1h10.3 M20.7,1h6.1 M33,1h6.1 M55.4,10.5V1H45.2 M4,17.3v6.6 M55.6,17.3v6.6 M4,30.5V40h10.3 M20.7,40 h6.1 M33,40h6.1 M55.4,30.5V40H45.2 M21,18.9h62.9v48.6H21V18.9z",clear:"M22,14.7l30.9,31 M52.9,14.7L22,45.7 M4.7,16.8V4.2h13.1 M26,4.2h7.8 M41.6,4.2h7.8 M70.3,16.8V4.2H57.2 M4.7,25.9v8.6 M70.3,25.9v8.6 M4.7,43.2v12.6h13.1 M26,55.8h7.8 M41.6,55.8h7.8 M70.3,43.2v12.6H57.2"},title:t.getLocale(["toolbox","brush","title"])};return e},e}(vU),$X=GX;function UX(t){t.registerComponentView(jX),t.registerComponentModel(FX),t.registerPreprocessor(mX),t.registerVisual(t.PRIORITY.VISUAL.BRUSH,AX),t.registerAction({type:"brush",event:"brush",update:"updateVisual"},(function(t,e){e.eachComponent({mainType:"brush",query:t},(function(e){e.setAreas(t.areas)}))})),t.registerAction({type:"brushSelect",event:"brushSelected",update:"none"},(function(){})),t.registerAction({type:"brushEnd",event:"brushEnd",update:"none"},(function(){})),bU("brush",$X)}var qX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.layoutMode={type:"box",ignoreSize:!0},n}return o(e,t),e.type="title",e.defaultOption={zlevel:0,z:6,show:!0,text:"",target:"blank",subtext:"",subtarget:"blank",left:0,top:0,backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",borderWidth:0,padding:5,itemGap:10,textStyle:{fontSize:18,fontWeight:"bold",color:"#464646"},subtextStyle:{fontSize:12,color:"#6E7079"}},e}(xp),XX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.render=function(t,e,n){if(this.group.removeAll(),t.get("show")){var i=this.group,r=t.getModel("textStyle"),o=t.getModel("subtextStyle"),a=t.get("textAlign"),s=at(t.get("textBaseline"),t.get("textVerticalAlign")),l=new Gl({style:Lh(r,{text:t.get("text"),fill:r.getTextColor()},{disableBox:!0}),z2:10}),u=l.getBoundingRect(),c=t.get("subtext"),d=new Gl({style:Lh(o,{text:c,fill:o.getTextColor(),y:u.height+t.get("itemGap"),verticalAlign:"top"},{disableBox:!0}),z2:10}),h=t.get("link"),f=t.get("sublink"),p=t.get("triggerEvent",!0);l.silent=!h&&!p,d.silent=!f&&!p,h&&l.on("click",(function(){rp(h,"_"+t.get("target"))})),f&&d.on("click",(function(){rp(f,"_"+t.get("subtarget"))})),$l(l).eventData=$l(d).eventData=p?{componentType:"title",componentIndex:t.componentIndex}:null,i.add(l),c&&i.add(d);var m=i.getBoundingRect(),g=t.getBoxLayoutParams();g.width=m.width,g.height=m.height;var y=dp(g,{width:n.getWidth(),height:n.getHeight()},t.get("padding"));a||(a=t.get("left")||t.get("right"),"middle"===a&&(a="center"),"right"===a?y.x+=y.width:"center"===a&&(y.x+=y.width/2)),s||(s=t.get("top")||t.get("bottom"),"center"===s&&(s="middle"),"bottom"===s?y.y+=y.height:"middle"===s&&(y.y+=y.height/2),s=s||"top"),i.x=y.x,i.y=y.y,i.markRedraw();var v={align:a,verticalAlign:s};l.setStyle(v),d.setStyle(v),m=i.getBoundingRect();var _=y.margin,b=t.getItemStyle(["color","opacity"]);b.fill=t.get("backgroundColor");var x=new Al({shape:{x:m.x-_[3],y:m.y-_[0],width:m.width+_[1]+_[3],height:m.height+_[0]+_[2],r:t.get("borderRadius")},style:b,subPixelOptimize:!0,silent:!0});i.add(x)}},e.type="title",e}(Py);function ZX(t){t.registerComponentModel(qX),t.registerComponentView(XX)}var JX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.layoutMode="box",n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n),this._initData()},e.prototype.mergeOption=function(e){t.prototype.mergeOption.apply(this,arguments),this._initData()},e.prototype.setCurrentIndex=function(t){null==t&&(t=this.option.currentIndex);var e=this._data.count();this.option.loop?t=(t%e+e)%e:(t>=e&&(t=e-1),t<0&&(t=0)),this.option.currentIndex=t},e.prototype.getCurrentIndex=function(){return this.option.currentIndex},e.prototype.isIndexMax=function(){return this.getCurrentIndex()>=this._data.count()-1},e.prototype.setPlayState=function(t){this.option.autoPlay=!!t},e.prototype.getPlayState=function(){return!!this.option.autoPlay},e.prototype._initData=function(){var t,e=this.option,n=e.data||[],i=e.axisType,r=this._names=[];"category"===i?(t=[],R(n,(function(e,n){var i,o=wo(co(e),"");J(e)?(i=T(e),i.value=n):i=n,t.push(i),r.push(o)}))):t=n;var o={category:"ordinal",time:"time",value:"number"}[i]||"number",a=this._data=new kL([{name:"value",type:o}],this);a.initData(t,r)},e.prototype.getData=function(){return this._data},e.prototype.getCategories=function(){if("category"===this.get("axisType"))return this._names.slice()},e.type="timeline",e.defaultOption={zlevel:0,z:4,show:!0,axisType:"time",realtime:!0,left:"20%",top:null,right:"20%",bottom:0,width:null,height:40,padding:5,controlPosition:"left",autoPlay:!1,rewind:!1,loop:!0,playInterval:2e3,currentIndex:0,itemStyle:{},label:{color:"#000"},data:[]},e}(xp),KX=JX,QX=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="timeline.slider",e.defaultOption=Qh(KX.defaultOption,{backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",borderWidth:0,orient:"horizontal",inverse:!1,tooltip:{trigger:"item"},symbol:"circle",symbolSize:12,lineStyle:{show:!0,width:2,color:"#DAE1F5"},label:{position:"auto",show:!0,interval:"auto",rotate:0,color:"#A4B1D7"},itemStyle:{color:"#A4B1D7",borderWidth:1},checkpointStyle:{symbol:"circle",symbolSize:15,color:"#316bf3",borderColor:"#fff",borderWidth:2,shadowBlur:2,shadowOffsetX:1,shadowOffsetY:1,shadowColor:"rgba(0, 0, 0, 0.3)",animation:!0,animationDuration:300,animationEasing:"quinticInOut"},controlStyle:{show:!0,showPlayBtn:!0,showPrevBtn:!0,showNextBtn:!0,itemSize:24,itemGap:12,position:"left",playIcon:"path://M31.6,53C17.5,53,6,41.5,6,27.4S17.5,1.8,31.6,1.8C45.7,1.8,57.2,13.3,57.2,27.4S45.7,53,31.6,53z M31.6,3.3 C18.4,3.3,7.5,14.1,7.5,27.4c0,13.3,10.8,24.1,24.1,24.1C44.9,51.5,55.7,40.7,55.7,27.4C55.7,14.1,44.9,3.3,31.6,3.3z M24.9,21.3 c0-2.2,1.6-3.1,3.5-2l10.5,6.1c1.899,1.1,1.899,2.9,0,4l-10.5,6.1c-1.9,1.1-3.5,0.2-3.5-2V21.3z",stopIcon:"path://M30.9,53.2C16.8,53.2,5.3,41.7,5.3,27.6S16.8,2,30.9,2C45,2,56.4,13.5,56.4,27.6S45,53.2,30.9,53.2z M30.9,3.5C17.6,3.5,6.8,14.4,6.8,27.6c0,13.3,10.8,24.1,24.101,24.1C44.2,51.7,55,40.9,55,27.6C54.9,14.4,44.1,3.5,30.9,3.5z M36.9,35.8c0,0.601-0.4,1-0.9,1h-1.3c-0.5,0-0.9-0.399-0.9-1V19.5c0-0.6,0.4-1,0.9-1H36c0.5,0,0.9,0.4,0.9,1V35.8z M27.8,35.8 c0,0.601-0.4,1-0.9,1h-1.3c-0.5,0-0.9-0.399-0.9-1V19.5c0-0.6,0.4-1,0.9-1H27c0.5,0,0.9,0.4,0.9,1L27.8,35.8L27.8,35.8z",nextIcon:"M2,18.5A1.52,1.52,0,0,1,.92,18a1.49,1.49,0,0,1,0-2.12L7.81,9.36,1,3.11A1.5,1.5,0,1,1,3,.89l8,7.34a1.48,1.48,0,0,1,.49,1.09,1.51,1.51,0,0,1-.46,1.1L3,18.08A1.5,1.5,0,0,1,2,18.5Z",prevIcon:"M10,.5A1.52,1.52,0,0,1,11.08,1a1.49,1.49,0,0,1,0,2.12L4.19,9.64,11,15.89a1.5,1.5,0,1,1-2,2.22L1,10.77A1.48,1.48,0,0,1,.5,9.68,1.51,1.51,0,0,1,1,8.58L9,.92A1.5,1.5,0,0,1,10,.5Z",prevBtnSize:18,nextBtnSize:18,color:"#A4B1D7",borderColor:"#A4B1D7",borderWidth:1},emphasis:{label:{show:!0,color:"#6f778d"},itemStyle:{color:"#316BF3"},controlStyle:{color:"#316BF3",borderColor:"#316BF3",borderWidth:2}},progress:{lineStyle:{color:"#316BF3"},itemStyle:{color:"#316BF3"},label:{color:"#6f778d"}},data:[]}),e}(KX);E(QX,Tg.prototype);var tZ=QX,eZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="timeline",e}(Py),nZ=eZ,iZ=function(t){function e(e,n,i,r){var o=t.call(this,e,n,i)||this;return o.type=r||"value",o}return o(e,t),e.prototype.getLabelModel=function(){return this.model.getModel("label")},e.prototype.isHorizontal=function(){return"horizontal"===this.model.get("orient")},e}(tI),rZ=iZ,oZ=Math.PI,aZ=Io(),sZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(t,e){this.api=e},e.prototype.render=function(t,e,n){if(this.model=t,this.api=n,this.ecModel=e,this.group.removeAll(),t.get("show",!0)){var i=this._layout(t,n),r=this._createGroup("_mainGroup"),o=this._createGroup("_labelGroup"),a=this._axis=this._createAxis(i,t);t.formatTooltip=function(t){var e=a.scale.getLabel({value:t});return oy("nameValue",{noName:!0,value:e})},R(["AxisLine","AxisTick","Control","CurrentPointer"],(function(e){this["_render"+e](i,r,a,t)}),this),this._renderAxisLabel(i,o,a,t),this._position(i,t)}this._doPlayStop(),this._updateTicksStatus()},e.prototype.remove=function(){this._clearTimer(),this.group.removeAll()},e.prototype.dispose=function(){this._clearTimer()},e.prototype._layout=function(t,e){var n,i=t.get(["label","position"]),r=t.get("orient"),o=uZ(t,e);n=null==i||"auto"===i?"horizontal"===r?o.y+o.height/2<e.getHeight()/2?"-":"+":o.x+o.width/2<e.getWidth()/2?"+":"-":q(i)?{horizontal:{top:"-",bottom:"+"},vertical:{left:"-",right:"+"}}[r][i]:i;var a,s,l,u={horizontal:"center",vertical:n>=0||"+"===n?"left":"right"},c={horizontal:n>=0||"+"===n?"top":"bottom",vertical:"middle"},d={horizontal:0,vertical:oZ/2},h="vertical"===r?o.height:o.width,f=t.getModel("controlStyle"),p=f.get("show",!0),m=p?f.get("itemSize"):0,g=p?f.get("itemGap"):0,y=m+g,v=t.get(["label","rotate"])||0;v=v*oZ/180;var _=f.get("position",!0),b=p&&f.get("showPlayBtn",!0),x=p&&f.get("showPrevBtn",!0),w=p&&f.get("showNextBtn",!0),M=0,S=h;"left"===_||"bottom"===_?(b&&(a=[0,0],M+=y),x&&(s=[M,0],M+=y),w&&(l=[S-m,0],S-=y)):(b&&(a=[S-m,0],S-=y),x&&(s=[0,0],M+=y),w&&(l=[S-m,0],S-=y));var L=[M,S];return t.get("inverse")&&L.reverse(),{viewRect:o,mainLength:h,orient:r,rotation:d[r],labelRotation:v,labelPosOpt:n,labelAlign:t.get(["label","align"])||u[r],labelBaseline:t.get(["label","verticalAlign"])||t.get(["label","baseline"])||c[r],playPosition:a,prevBtnPosition:s,nextBtnPosition:l,axisExtent:L,controlSize:m,controlGap:g}},e.prototype._position=function(t,e){var n=this._mainGroup,i=this._labelGroup,r=t.viewRect;if("vertical"===t.orient){var o=bi(),a=r.x,s=r.y+r.height;Si(o,o,[-a,-s]),Li(o,o,-oZ/2),Si(o,o,[a,s]),r=r.clone(),r.applyTransform(o)}var l=g(r),u=g(n.getBoundingRect()),c=g(i.getBoundingRect()),d=[n.x,n.y],h=[i.x,i.y];h[0]=d[0]=l[0][0];var f=t.labelPosOpt;if(null==f||q(f)){var p="+"===f?0:1;y(d,u,l,1,p),y(h,c,l,1,1-p)}else{p=f>=0?0:1;y(d,u,l,1,p),h[1]=d[1]+f}function m(t){t.originX=l[0][0]-t.x,t.originY=l[1][0]-t.y}function g(t){return[[t.x,t.x+t.width],[t.y,t.y+t.height]]}function y(t,e,n,i,r){t[i]+=n[i][r]-e[i][r]}n.setPosition(d),i.setPosition(h),n.rotation=i.rotation=t.rotation,m(n),m(i)},e.prototype._createAxis=function(t,e){var n=e.getData(),i=e.get("axisType"),r=lZ(e,i);r.getTicks=function(){return n.mapArray(["value"],(function(t){return{value:t}}))};var o=n.getDataExtent("value");r.setExtent(o[0],o[1]),r.niceTicks();var a=new rZ("value",r,t.axisExtent,i);return a.model=e,a},e.prototype._createGroup=function(t){var e=this[t]=new Mr;return this.group.add(e),e},e.prototype._renderAxisLine=function(t,e,n,i){var r=n.getExtent();if(i.get(["lineStyle","show"])){var o=new dd({shape:{x1:r[0],y1:0,x2:r[1],y2:0},style:I({lineCap:"round"},i.getModel("lineStyle").getLineStyle()),silent:!0,z2:1});e.add(o);var a=this._progressLine=new dd({shape:{x1:r[0],x2:this._currentPointer?this._currentPointer.x:r[0],y1:0,y2:0},style:A({lineCap:"round",lineWidth:o.style.lineWidth},i.getModel(["progress","lineStyle"]).getLineStyle()),silent:!0,z2:1});e.add(a)}},e.prototype._renderAxisTick=function(t,e,n,i){var r=this,o=i.getData(),a=n.scale.getTicks();this._tickSymbols=[],R(a,(function(t){var a=n.dataToCoord(t.value),s=o.getItemModel(t.value),l=s.getModel("itemStyle"),u=s.getModel(["emphasis","itemStyle"]),c=s.getModel(["progress","itemStyle"]),d={x:a,y:0,onclick:V(r._changeTimeline,r,t.value)},h=dZ(s,l,e,d);h.ensureState("emphasis").style=u.getItemStyle(),h.ensureState("progress").style=c.getItemStyle(),$u(h);var f=$l(h);s.get("tooltip")?(f.dataIndex=t.value,f.dataModel=i):f.dataIndex=f.dataModel=null,r._tickSymbols.push(h)}))},e.prototype._renderAxisLabel=function(t,e,n,i){var r=this,o=n.getLabelModel();if(o.get("show")){var a=i.getData(),s=n.getViewLabels();this._tickLabels=[],R(s,(function(i){var o=i.tickValue,s=a.getItemModel(o),l=s.getModel("label"),u=s.getModel(["emphasis","label"]),c=s.getModel(["progress","label"]),d=n.dataToCoord(i.tickValue),h=new Gl({x:d,y:0,rotation:t.labelRotation-t.rotation,onclick:V(r._changeTimeline,r,o),silent:!1,style:Lh(l,{text:i.formattedLabel,align:t.labelAlign,verticalAlign:t.labelBaseline})});h.ensureState("emphasis").style=Lh(u),h.ensureState("progress").style=Lh(c),e.add(h),$u(h),aZ(h).dataIndex=o,r._tickLabels.push(h)}))}},e.prototype._renderControl=function(t,e,n,i){var r=t.controlSize,o=t.rotation,a=i.getModel("controlStyle").getItemStyle(),s=i.getModel(["emphasis","controlStyle"]).getItemStyle(),l=i.getPlayState(),u=i.get("inverse",!0);function c(t,n,l,u){if(t){var c=lr(at(i.get(["controlStyle",n+"BtnSize"]),r),r),d=[0,-c/2,c,c],h=cZ(i,n+"Icon",d,{x:t[0],y:t[1],originX:r/2,originY:0,rotation:u?-o:0,rectHover:!0,style:a,onclick:l});h.ensureState("emphasis").style=s,e.add(h),$u(h)}}c(t.nextBtnPosition,"next",V(this._changeTimeline,this,u?"-":"+")),c(t.prevBtnPosition,"prev",V(this._changeTimeline,this,u?"+":"-")),c(t.playPosition,l?"stop":"play",V(this._handlePlayClick,this,!l),!0)},e.prototype._renderCurrentPointer=function(t,e,n,i){var r=i.getData(),o=i.getCurrentIndex(),a=r.getItemModel(o).getModel("checkpointStyle"),s=this,l={onCreate:function(t){t.draggable=!0,t.drift=V(s._handlePointerDrag,s),t.ondragend=V(s._handlePointerDragend,s),hZ(t,s._progressLine,o,n,i,!0)},onUpdate:function(t){hZ(t,s._progressLine,o,n,i)}};this._currentPointer=dZ(a,a,this._mainGroup,{},this._currentPointer,l)},e.prototype._handlePlayClick=function(t){this._clearTimer(),this.api.dispatchAction({type:"timelinePlayChange",playState:t,from:this.uid})},e.prototype._handlePointerDrag=function(t,e,n){this._clearTimer(),this._pointerChangeTimeline([n.offsetX,n.offsetY])},e.prototype._handlePointerDragend=function(t){this._pointerChangeTimeline([t.offsetX,t.offsetY],!0)},e.prototype._pointerChangeTimeline=function(t,e){var n=this._toAxisCoord(t)[0],i=this._axis,r=Hr(i.getExtent().slice());n>r[1]&&(n=r[1]),n<r[0]&&(n=r[0]),this._currentPointer.x=n,this._currentPointer.markRedraw(),this._progressLine.shape.x2=n,this._progressLine.dirty();var o=this._findNearestTick(n),a=this.model;(e||o!==a.getCurrentIndex()&&a.get("realtime"))&&this._changeTimeline(o)},e.prototype._doPlayStop=function(){var t=this;this._clearTimer(),this.model.getPlayState()&&(this._timer=setTimeout((function(){var e=t.model;t._changeTimeline(e.getCurrentIndex()+(e.get("rewind",!0)?-1:1))}),this.model.get("playInterval")))},e.prototype._toAxisCoord=function(t){var e=this._mainGroup.getLocalTransform();return sh(t,e,!0)},e.prototype._findNearestTick=function(t){var e,n=this.model.getData(),i=1/0,r=this._axis;return n.each(["value"],(function(n,o){var a=r.dataToCoord(n),s=Math.abs(a-t);s<i&&(i=s,e=o)})),e},e.prototype._clearTimer=function(){this._timer&&(clearTimeout(this._timer),this._timer=null)},e.prototype._changeTimeline=function(t){var e=this.model.getCurrentIndex();"+"===t?t=e+1:"-"===t&&(t=e-1),this.api.dispatchAction({type:"timelineChange",currentIndex:t,from:this.uid})},e.prototype._updateTicksStatus=function(){var t=this.model.getCurrentIndex(),e=this._tickSymbols,n=this._tickLabels;if(e)for(var i=0;i<e.length;i++)e&&e[i]&&e[i].toggleState("progress",i<t);if(n)for(i=0;i<n.length;i++)n&&n[i]&&n[i].toggleState("progress",aZ(n[i]).dataIndex<=t)},e.type="timeline.slider",e}(nZ);function lZ(t,e){if(e=e||t.get("type"),e)switch(e){case"category":return new HD({ordinalMeta:t.getCategories(),extent:[1/0,-1/0]});case"time":return new QD({locale:t.ecModel.getLocaleModel(),useUTC:t.ecModel.get("useUTC")});default:return new WD}}function uZ(t,e){return dp(t.getBoxLayoutParams(),{width:e.getWidth(),height:e.getHeight()},t.get("padding"))}function cZ(t,e,n,i){var r=i.style,o=ph(t.get(["controlStyle",e]),i||{},new Ji(n[0],n[1],n[2],n[3]));return r&&o.setStyle(r),o}function dZ(t,e,n,i,r,o){var a=e.get("color");if(r)r.setColor(a),n.add(r),o&&o.onUpdate(r);else{var s=t.get("symbol");r=j_(s,-1,-1,2,2,a),r.setStyle("strokeNoScale",!0),n.add(r),o&&o.onCreate(r)}var l=e.getItemStyle(["color"]);r.setStyle(l),i=D({rectHover:!0,z2:100},i,!0);var u=t.get("symbolSize");u=u instanceof Array?u.slice():[+u,+u],i.scaleX=u[0]/2,i.scaleY=u[1]/2;var c=t.get("symbolOffset");c&&(i.x=i.x||0,i.y=i.y||0,i.x+=Nr(c[0],u[0]),i.y+=Nr(c[1],u[1]));var d=t.get("symbolRotate");return i.rotation=(d||0)*Math.PI/180||0,r.attr(i),r.updateTransform(),r}function hZ(t,e,n,i,r,o){if(!t.dragging){var a=r.getModel("checkpointStyle"),s=i.dataToCoord(r.getData().get("value",n));if(o||!a.get("animation",!0))t.attr({x:s,y:0}),e&&e.attr({shape:{x2:s}});else{var l={duration:a.get("animationDuration",!0),easing:a.get("animationEasing",!0)};t.stopAnimation(null,!0),t.animateTo({x:s,y:0},l),e&&e.animateTo({shape:{x2:s}},l)}}}var fZ=sZ;function pZ(t){t.registerAction({type:"timelineChange",event:"timelineChanged",update:"prepareAndUpdate"},(function(t,e){var n=e.getComponent("timeline");return n&&null!=t.currentIndex&&(n.setCurrentIndex(t.currentIndex),!n.get("loop",!0)&&n.isIndexMax()&&n.setPlayState(!1)),e.resetOption("timeline",{replaceMerge:n.get("replaceMerge",!0)}),A({currentIndex:n.option.currentIndex},t)})),t.registerAction({type:"timelinePlayChange",event:"timelinePlayChanged",update:"update"},(function(t,e){var n=e.getComponent("timeline");n&&null!=t.playState&&n.setPlayState(t.playState)}))}function mZ(t){var e=t&&t.timeline;$(e)||(e=e?[e]:[]),R(e,(function(t){t&&gZ(t)}))}function gZ(t){var e=t.type,n={number:"value",time:"time"};if(n[e]&&(t.axisType=n[e],delete t.type),yZ(t),vZ(t,"controlPosition")){var i=t.controlStyle||(t.controlStyle={});vZ(i,"position")||(i.position=t.controlPosition),"none"!==i.position||vZ(i,"show")||(i.show=!1,delete i.position),delete t.controlPosition}R(t.data||[],(function(t){J(t)&&!$(t)&&(!vZ(t,"value")&&vZ(t,"name")&&(t.value=t.name),yZ(t))}))}function yZ(t){var e=t.itemStyle||(t.itemStyle={}),n=e.emphasis||(e.emphasis={}),i=t.label||t.label||{},r=i.normal||(i.normal={}),o={normal:1,emphasis:1};R(i,(function(t,e){o[e]||vZ(r,e)||(r[e]=t)})),n.label&&!vZ(i,"emphasis")&&(i.emphasis=n.label,delete n.label)}function vZ(t,e){return t.hasOwnProperty(e)}function _Z(t){t.registerComponentModel(tZ),t.registerComponentView(fZ),t.registerSubTypeDefaulter("timeline",(function(){return"slider"})),pZ(t),t.registerPreprocessor(mZ)}function bZ(t,e){if(!t)return!1;for(var n=$(t)?t:[t],i=0;i<n.length;i++)if(n[i]&&n[i][e])return!0;return!1}function xZ(t){lo(t,"label",["show"])}var wZ=Io(),MZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.createdBySelf=!1,n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n),this._mergeOption(t,n,!1,!0)},e.prototype.isAnimationEnabled=function(){if(h.node)return!1;var t=this.__hostSeries;return this.getShallow("animation")&&t&&t.isAnimationEnabled()},e.prototype.mergeOption=function(t,e){this._mergeOption(t,e,!1,!1)},e.prototype._mergeOption=function(t,e,n,i){var r=this.mainType;n||e.eachSeries((function(t){var n=t.get(this.mainType,!0),o=wZ(t)[r];n&&n.data?(o?o._mergeOption(n,e,!0):(i&&xZ(n),R(n.data,(function(t){t instanceof Array?(xZ(t[0]),xZ(t[1])):xZ(t)})),o=this.createMarkerModelFromSeries(n,this,e),I(o,{mainType:this.mainType,seriesIndex:t.seriesIndex,name:t.name,createdBySelf:!0}),o.__hostSeries=t),wZ(t)[r]=o):wZ(t)[r]=null}),this)},e.prototype.formatTooltip=function(t,e,n){var i=this.getData(),r=this.getRawValue(t),o=i.getName(t);return oy("section",{header:this.name,blocks:[oy("nameValue",{name:o,value:r,noName:!o,noValue:null==r})]})},e.prototype.getData=function(){return this._data},e.prototype.setData=function(t){this._data=t},e.getMarkerModelFromSeries=function(t,e){return wZ(t)[e]},e.type="marker",e.dependencies=["series","grid","polar","geo"],e}(xp);E(MZ,Tg.prototype);var SZ=MZ,LZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.createMarkerModelFromSeries=function(t,n,i){return new e(t,n,i)},e.type="markPoint",e.defaultOption={zlevel:0,z:5,symbol:"pin",symbolSize:50,tooltip:{trigger:"item"},label:{show:!0,position:"inside"},itemStyle:{borderWidth:2},emphasis:{label:{show:!0}}},e}(SZ),kZ=LZ;function TZ(t){return!(isNaN(parseFloat(t.x))&&isNaN(parseFloat(t.y)))}function DZ(t){return!isNaN(parseFloat(t.x))&&!isNaN(parseFloat(t.y))}function CZ(t,e,n,i,r,o){var a=[],s=RL(e,i),l=s?e.getCalculationInfo("stackResultDimension"):i,u=NZ(e,l,t),c=e.indicesOfNearest(l,u)[0];a[r]=e.get(n,c),a[o]=e.get(l,c);var d=e.get(i,c),h=jr(e.get(i,c));return h=Math.min(h,20),h>=0&&(a[o]=+a[o].toFixed(h)),[a,d]}var IZ={min:G(CZ,"min"),max:G(CZ,"max"),average:G(CZ,"average"),median:G(CZ,"median")};function AZ(t,e){var n=t.getData(),i=t.coordinateSystem;if(e&&!DZ(e)&&!$(e.coord)&&i){var r=i.dimensions,o=YZ(e,n,i,t);if(e=T(e),e.type&&IZ[e.type]&&o.baseAxis&&o.valueAxis){var a=O(r,o.baseAxis.dim),s=O(r,o.valueAxis.dim),l=IZ[e.type](n,o.baseDataDim,o.valueDataDim,a,s);e.coord=l[0],e.value=l[1]}else{for(var u=[null!=e.xAxis?e.xAxis:e.radiusAxis,null!=e.yAxis?e.yAxis:e.angleAxis],c=0;c<2;c++)IZ[u[c]]&&(u[c]=NZ(n,n.mapDimension(r[c]),u[c]));e.coord=u}}return e}function YZ(t,e,n,i){var r={};return null!=t.valueIndex||null!=t.valueDim?(r.valueDataDim=null!=t.valueIndex?e.getDimension(t.valueIndex):t.valueDim,r.valueAxis=n.getAxis(OZ(i,r.valueDataDim)),r.baseAxis=n.getOtherAxis(r.valueAxis),r.baseDataDim=e.mapDimension(r.baseAxis.dim)):(r.baseAxis=i.getBaseAxis(),r.valueAxis=n.getOtherAxis(r.baseAxis),r.baseDataDim=e.mapDimension(r.baseAxis.dim),r.valueDataDim=e.mapDimension(r.valueAxis.dim)),r}function OZ(t,e){var n=t.getData(),i=n.dimensions;e=n.getDimension(e);for(var r=0;r<i.length;r++){var o=n.getDimensionInfo(i[r]);if(o.name===e)return o.coordDim}}function PZ(t,e){return!(t&&t.containData&&e.coord&&!TZ(e))||t.containData(e.coord)}function EZ(t,e,n,i){return i<2?t.coord&&t.coord[i]:t.value}function NZ(t,e,n){if("average"===n){var i=0,r=0;return t.each(e,(function(t,e){isNaN(t)||(i+=t,r++)})),i/r}return"median"===n?t.getMedian(e):t.getDataExtent(e)["max"===n?1:0]}var RZ=Io(),HZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.init=function(){this.markerGroupMap=gt()},e.prototype.render=function(t,e,n){var i=this,r=this.markerGroupMap;r.each((function(t){RZ(t).keep=!1})),e.eachSeries((function(t){var r=SZ.getMarkerModelFromSeries(t,i.type);r&&i.renderSeries(t,r,e,n)})),r.each((function(t){!RZ(t).keep&&i.group.remove(t.group)}))},e.prototype.markKeep=function(t){RZ(t).keep=!0},e.prototype.blurSeries=function(t){var e=this;R(t,(function(t){var n=SZ.getMarkerModelFromSeries(t,e.type);if(n){var i=n.getData();i.eachItemGraphicEl((function(t){t&&Au(t)}))}}))},e.type="marker",e}(Py),jZ=HZ;function zZ(t,e,n){var i=e.coordinateSystem;t.each((function(r){var o,a=t.getItemModel(r),s=Nr(a.get("x"),n.getWidth()),l=Nr(a.get("y"),n.getHeight());if(isNaN(s)||isNaN(l)){if(e.getMarkerPosition)o=e.getMarkerPosition(t.getValues(t.dimensions,r));else if(i){var u=t.get(i.dimensions[0],r),c=t.get(i.dimensions[1],r);o=i.dataToPoint([u,c])}}else o=[s,l];isNaN(s)||(o[0]=s),isNaN(l)||(o[1]=l),t.setItemLayout(r,o)}))}var WZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.updateTransform=function(t,e,n){e.eachSeries((function(t){var e=SZ.getMarkerModelFromSeries(t,"markPoint");e&&(zZ(e.getData(),t,n),this.markerGroupMap.get(t.id).updateLayout())}),this)},e.prototype.renderSeries=function(t,e,n,i){var r=t.coordinateSystem,o=t.id,a=t.getData(),s=this.markerGroupMap,l=s.get(o)||s.set(o,new tk),u=BZ(r,t,e);e.setData(u),zZ(e.getData(),t,i),u.each((function(t){var n=u.getItemModel(t),i=n.getShallow("symbol"),r=n.getShallow("symbolSize"),o=n.getShallow("symbolRotate");if(U(i)||U(r)||U(o)){var s=e.getRawValue(t),l=e.getDataParams(t);U(i)&&(i=i(s,l)),U(r)&&(r=r(s,l)),U(o)&&(o=o(s,l))}var c=n.getModel("itemStyle").getItemStyle(),d=Nv(a,"color");c.fill||(c.fill=d),u.setItemVisual(t,{symbol:i,symbolSize:r,symbolRotate:o,style:c})})),l.updateData(u),this.group.add(l.group),u.eachItemGraphicEl((function(t){t.traverse((function(t){$l(t).dataModel=e}))})),this.markKeep(l),l.group.silent=e.get("silent")||t.get("silent")},e.type="markPoint",e}(jZ);function BZ(t,e,n){var i;i=t?H(t&&t.dimensions,(function(t){var n=e.getData().getDimensionInfo(e.getData().mapDimension(t))||{};return A({name:t},n)})):[{name:"value",type:"float"}];var r=new kL(i,n),o=H(n.get("data"),G(AZ,e));return t&&(o=z(o,G(PZ,t))),r.initData(o,null,t?EZ:function(t){return t.value}),r}var FZ=WZ;function VZ(t){t.registerComponentModel(kZ),t.registerComponentView(FZ),t.registerPreprocessor((function(t){bZ(t.series,"markPoint")&&(t.markPoint=t.markPoint||{})}))}var GZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.createMarkerModelFromSeries=function(t,n,i){return new e(t,n,i)},e.type="markLine",e.defaultOption={zlevel:0,z:5,symbol:["circle","arrow"],symbolSize:[8,16],symbolOffset:0,precision:2,tooltip:{trigger:"item"},label:{show:!0,position:"end",distance:5},lineStyle:{type:"dashed"},emphasis:{label:{show:!0},lineStyle:{width:3}},animationEasing:"linear"},e}(SZ),$Z=GZ,UZ=Io(),qZ=function(t,e,n,i){var r,o=t.getData();if($(i))r=i;else{var a=i.type;if("min"===a||"max"===a||"average"===a||"median"===a||null!=i.xAxis||null!=i.yAxis){var s=void 0,l=void 0;if(null!=i.yAxis||null!=i.xAxis)s=e.getAxis(null!=i.yAxis?"y":"x"),l=ot(i.yAxis,i.xAxis);else{var u=YZ(i,o,e,t);s=u.valueAxis;var c=HL(o,u.valueDataDim);l=NZ(o,c,a)}var d="x"===s.dim?0:1,h=1-d,f=T(i),p={coord:[]};f.type=null,f.coord=[],f.coord[h]=-1/0,p.coord[h]=1/0;var m=n.get("precision");m>=0&&"number"===typeof l&&(l=+l.toFixed(Math.min(m,20))),f.coord[d]=p.coord[d]=l,r=[f,p,{type:a,valueIndex:i.valueIndex,value:l}]}else r=[]}var g=[AZ(t,r[0]),AZ(t,r[1]),I({},r[2])];return g[2].type=g[2].type||null,D(g[2],g[0]),D(g[2],g[1]),g};function XZ(t){return!isNaN(t)&&!isFinite(t)}function ZZ(t,e,n,i){var r=1-t,o=i.dimensions[t];return XZ(e[r])&&XZ(n[r])&&e[t]===n[t]&&i.getAxis(o).containData(e[t])}function JZ(t,e){if("cartesian2d"===t.type){var n=e[0].coord,i=e[1].coord;if(n&&i&&(ZZ(1,n,i,t)||ZZ(0,n,i,t)))return!0}return PZ(t,e[0])&&PZ(t,e[1])}function KZ(t,e,n,i,r){var o,a=i.coordinateSystem,s=t.getItemModel(e),l=Nr(s.get("x"),r.getWidth()),u=Nr(s.get("y"),r.getHeight());if(isNaN(l)||isNaN(u)){if(i.getMarkerPosition)o=i.getMarkerPosition(t.getValues(t.dimensions,e));else{var c=a.dimensions,d=t.get(c[0],e),h=t.get(c[1],e);o=a.dataToPoint([d,h])}if(bk(a,"cartesian2d")){var f=a.getAxis("x"),p=a.getAxis("y");c=a.dimensions;XZ(t.get(c[0],e))?o[0]=f.toGlobalCoord(f.getExtent()[n?0:1]):XZ(t.get(c[1],e))&&(o[1]=p.toGlobalCoord(p.getExtent()[n?0:1]))}isNaN(l)||(o[0]=l),isNaN(u)||(o[1]=u)}else o=[l,u];t.setItemLayout(e,o)}var QZ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.updateTransform=function(t,e,n){e.eachSeries((function(t){var e=SZ.getMarkerModelFromSeries(t,"markLine");if(e){var i=e.getData(),r=UZ(e).from,o=UZ(e).to;r.each((function(e){KZ(r,e,!0,t,n),KZ(o,e,!1,t,n)})),i.each((function(t){i.setItemLayout(t,[r.getItemLayout(t),o.getItemLayout(t)])})),this.markerGroupMap.get(t.id).updateLayout()}}),this)},e.prototype.renderSeries=function(t,e,n,i){var r=t.coordinateSystem,o=t.id,a=t.getData(),s=this.markerGroupMap,l=s.get(o)||s.set(o,new jE);this.group.add(l.group);var u=tJ(r,t,e),c=u.from,d=u.to,h=u.line;UZ(e).from=c,UZ(e).to=d,e.setData(h);var f=e.get("symbol"),p=e.get("symbolSize"),m=e.get("symbolRotate"),g=e.get("symbolOffset");function y(e,n,r){var o=e.getItemModel(n);KZ(e,n,r,t,i);var s=o.getModel("itemStyle").getItemStyle();null==s.fill&&(s.fill=Nv(a,"color")),e.setItemVisual(n,{symbolKeepAspect:o.get("symbolKeepAspect"),symbolOffset:at(o.get("symbolOffset"),g[r?0:1]),symbolRotate:at(o.get("symbolRotate",!0),m[r?0:1]),symbolSize:at(o.get("symbolSize"),p[r?0:1]),symbol:at(o.get("symbol",!0),f[r?0:1]),style:s})}$(f)||(f=[f,f]),$(p)||(p=[p,p]),$(m)||(m=[m,m]),$(g)||(g=[g,g]),u.from.each((function(t){y(c,t,!0),y(d,t,!1)})),h.each((function(t){var e=h.getItemModel(t).getModel("lineStyle").getLineStyle();h.setItemLayout(t,[c.getItemLayout(t),d.getItemLayout(t)]),null==e.stroke&&(e.stroke=c.getItemVisual(t,"style").fill),h.setItemVisual(t,{fromSymbolKeepAspect:c.getItemVisual(t,"symbolKeepAspect"),fromSymbolOffset:c.getItemVisual(t,"symbolOffset"),fromSymbolRotate:c.getItemVisual(t,"symbolRotate"),fromSymbolSize:c.getItemVisual(t,"symbolSize"),fromSymbol:c.getItemVisual(t,"symbol"),toSymbolKeepAspect:d.getItemVisual(t,"symbolKeepAspect"),toSymbolOffset:d.getItemVisual(t,"symbolOffset"),toSymbolRotate:d.getItemVisual(t,"symbolRotate"),toSymbolSize:d.getItemVisual(t,"symbolSize"),toSymbol:d.getItemVisual(t,"symbol"),style:e})})),l.updateData(h),u.line.eachItemGraphicEl((function(t,n){t.traverse((function(t){$l(t).dataModel=e}))})),this.markKeep(l),l.group.silent=e.get("silent")||t.get("silent")},e.type="markLine",e}(jZ);function tJ(t,e,n){var i;i=t?H(t&&t.dimensions,(function(t){var n=e.getData().getDimensionInfo(e.getData().mapDimension(t))||{};return A({name:t},n)})):[{name:"value",type:"float"}];var r=new kL(i,n),o=new kL(i,n),a=new kL([],n),s=H(n.get("data"),G(qZ,e,t,n));t&&(s=z(s,G(JZ,t)));var l=t?EZ:function(t){return t.value};return r.initData(H(s,(function(t){return t[0]})),null,l),o.initData(H(s,(function(t){return t[1]})),null,l),a.initData(H(s,(function(t){return t[2]}))),a.hasItemOption=!0,{from:r,to:o,line:a}}var eJ=QZ;function nJ(t){t.registerComponentModel($Z),t.registerComponentView(eJ),t.registerPreprocessor((function(t){bZ(t.series,"markLine")&&(t.markLine=t.markLine||{})}))}var iJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.createMarkerModelFromSeries=function(t,n,i){return new e(t,n,i)},e.type="markArea",e.defaultOption={zlevel:0,z:1,tooltip:{trigger:"item"},animation:!1,label:{show:!0,position:"top"},itemStyle:{borderWidth:0},emphasis:{label:{show:!0,position:"top"}}},e}(SZ),rJ=iJ,oJ=Io(),aJ=function(t,e,n,i){var r=AZ(t,i[0]),o=AZ(t,i[1]),a=r.coord,s=o.coord;a[0]=ot(a[0],-1/0),a[1]=ot(a[1],-1/0),s[0]=ot(s[0],1/0),s[1]=ot(s[1],1/0);var l=C([{},r,o]);return l.coord=[r.coord,o.coord],l.x0=r.x,l.y0=r.y,l.x1=o.x,l.y1=o.y,l};function sJ(t){return!isNaN(t)&&!isFinite(t)}function lJ(t,e,n,i){var r=1-t;return sJ(e[r])&&sJ(n[r])}function uJ(t,e){var n=e.coord[0],i=e.coord[1];return!!(bk(t,"cartesian2d")&&n&&i&&(lJ(1,n,i,t)||lJ(0,n,i,t)))||(PZ(t,{coord:n,x:e.x0,y:e.y0})||PZ(t,{coord:i,x:e.x1,y:e.y1}))}function cJ(t,e,n,i,r){var o,a=i.coordinateSystem,s=t.getItemModel(e),l=Nr(s.get(n[0]),r.getWidth()),u=Nr(s.get(n[1]),r.getHeight());if(isNaN(l)||isNaN(u)){if(i.getMarkerPosition)o=i.getMarkerPosition(t.getValues(n,e));else{var c=t.get(n[0],e),d=t.get(n[1],e),h=[c,d];a.clampData&&a.clampData(h,h),o=a.dataToPoint(h,!0)}if(bk(a,"cartesian2d")){var f=a.getAxis("x"),p=a.getAxis("y");c=t.get(n[0],e),d=t.get(n[1],e);sJ(c)?o[0]=f.toGlobalCoord(f.getExtent()["x0"===n[0]?0:1]):sJ(d)&&(o[1]=p.toGlobalCoord(p.getExtent()["y0"===n[1]?0:1]))}isNaN(l)||(o[0]=l),isNaN(u)||(o[1]=u)}else o=[l,u];return o}var dJ=[["x0","y0"],["x1","y0"],["x1","y1"],["x0","y1"]],hJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.updateTransform=function(t,e,n){e.eachSeries((function(t){var e=SZ.getMarkerModelFromSeries(t,"markArea");if(e){var i=e.getData();i.each((function(e){var r=H(dJ,(function(r){return cJ(i,e,r,t,n)}));i.setItemLayout(e,r);var o=i.getItemGraphicEl(e);o.setShape("points",r)}))}}),this)},e.prototype.renderSeries=function(t,e,n,i){var r=t.coordinateSystem,o=t.id,a=t.getData(),s=this.markerGroupMap,l=s.get(o)||s.set(o,{group:new Mr});this.group.add(l.group),this.markKeep(l);var u=fJ(r,t,e);e.setData(u),u.each((function(e){var n=H(dJ,(function(n){return cJ(u,e,n,t,i)})),o=r.getAxis("x").scale,s=r.getAxis("y").scale,l=o.getExtent(),c=s.getExtent(),d=[o.parse(u.get("x0",e)),o.parse(u.get("x1",e))],h=[s.parse(u.get("y0",e)),s.parse(u.get("y1",e))];Hr(d),Hr(h);var f=!(l[0]>d[1]||l[1]<d[0]||c[0]>h[1]||c[1]<h[0]),p=!f;u.setItemLayout(e,{points:n,allClipped:p});var m=u.getItemModel(e).getModel("itemStyle").getItemStyle(),g=Nv(a,"color");m.fill||(m.fill=g,"string"===typeof m.fill&&(m.fill=Sn(m.fill,.4))),m.stroke||(m.stroke=g),u.setItemVisual(e,"style",m)})),u.diff(oJ(l).data).add((function(t){var e=u.getItemLayout(t);if(!e.allClipped){var n=new rd({shape:{points:e.points}});u.setItemGraphicEl(t,n),l.group.add(n)}})).update((function(t,n){var i=oJ(l).data.getItemGraphicEl(n),r=u.getItemLayout(t);r.allClipped?i&&l.group.remove(i):(i?th(i,{shape:{points:r.points}},e,t):i=new rd({shape:{points:r.points}}),u.setItemGraphicEl(t,i),l.group.add(i))})).remove((function(t){var e=oJ(l).data.getItemGraphicEl(t);l.group.remove(e)})).execute(),u.eachItemGraphicEl((function(t,n){var i=u.getItemModel(n),r=u.getItemVisual(n,"style");t.useStyle(u.getItemVisual(n,"style")),Mh(t,Sh(i),{labelFetcher:e,labelDataIndex:n,defaultText:u.getName(n)||"",inheritColor:"string"===typeof r.fill?Sn(r.fill,1):"#000"}),Zu(t,i),$u(t),$l(t).dataModel=e})),oJ(l).data=u,l.group.silent=e.get("silent")||t.get("silent")},e.type="markArea",e}(jZ);function fJ(t,e,n){var i,r,o=["x0","y0","x1","y1"];t?(i=H(t&&t.dimensions,(function(t){var n=e.getData(),i=n.getDimensionInfo(n.mapDimension(t))||{};return A({name:t},i)})),r=new kL(H(o,(function(t,e){return{name:t,type:i[e%2].type}})),n)):(i=[{name:"value",type:"float"}],r=new kL(i,n));var a=H(n.get("data"),G(aJ,e,t,n));t&&(a=z(a,G(uJ,t)));var s=t?function(t,e,n,i){return t.coord[Math.floor(i/2)][i%2]}:function(t){return t.value};return r.initData(a,null,s),r.hasItemOption=!0,r}var pJ=hJ;function mJ(t){t.registerComponentModel(rJ),t.registerComponentView(pJ),t.registerPreprocessor((function(t){bZ(t.series,"markArea")&&(t.markArea=t.markArea||{})}))}var gJ=function(t,e){return"all"===e?{type:"all",title:t.getLocale(["legend","selector","all"])}:"inverse"===e?{type:"inverse",title:t.getLocale(["legend","selector","inverse"])}:void 0},yJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.layoutMode={type:"box",ignoreSize:!0},n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n),t.selected=t.selected||{},this._updateSelector(t)},e.prototype.mergeOption=function(e,n){t.prototype.mergeOption.call(this,e,n),this._updateSelector(e)},e.prototype._updateSelector=function(t){var e=t.selector,n=this.ecModel;!0===e&&(e=t.selector=["all","inverse"]),$(e)&&R(e,(function(t,i){q(t)&&(t={type:t}),e[i]=D(t,gJ(n,t.type))}))},e.prototype.optionUpdated=function(){this._updateData(this.ecModel);var t=this._data;if(t[0]&&"single"===this.get("selectedMode")){for(var e=!1,n=0;n<t.length;n++){var i=t[n].get("name");if(this.isSelected(i)){this.select(i),e=!0;break}}!e&&this.select(t[0].get("name"))}},e.prototype._updateData=function(t){var e=[],n=[];t.eachRawSeries((function(i){var r,o=i.name;if(n.push(o),i.legendVisualProvider){var a=i.legendVisualProvider,s=a.getAllNames();t.isSeriesFiltered(i)||(n=n.concat(s)),s.length?e=e.concat(s):r=!0}else r=!0;r&&Mo(i)&&e.push(i.name)})),this._availableNames=n;var i=this.get("data")||e,r=H(i,(function(t){return"string"!==typeof t&&"number"!==typeof t||(t={name:t}),new qh(t,this,this.ecModel)}),this);this._data=r},e.prototype.getData=function(){return this._data},e.prototype.select=function(t){var e=this.option.selected,n=this.get("selectedMode");if("single"===n){var i=this._data;R(i,(function(t){e[t.get("name")]=!1}))}e[t]=!0},e.prototype.unSelect=function(t){"single"!==this.get("selectedMode")&&(this.option.selected[t]=!1)},e.prototype.toggleSelected=function(t){var e=this.option.selected;e.hasOwnProperty(t)||(e[t]=!0),this[e[t]?"unSelect":"select"](t)},e.prototype.allSelect=function(){var t=this._data,e=this.option.selected;R(t,(function(t){e[t.get("name",!0)]=!0}))},e.prototype.inverseSelect=function(){var t=this._data,e=this.option.selected;R(t,(function(t){var n=t.get("name",!0);e.hasOwnProperty(n)||(e[n]=!0),e[n]=!e[n]}))},e.prototype.isSelected=function(t){var e=this.option.selected;return!(e.hasOwnProperty(t)&&!e[t])&&O(this._availableNames,t)>=0},e.prototype.getOrient=function(){return"vertical"===this.get("orient")?{index:1,name:"vertical"}:{index:0,name:"horizontal"}},e.type="legend.plain",e.dependencies=["series"],e.defaultOption={zlevel:0,z:4,show:!0,orient:"horizontal",left:"center",top:0,align:"auto",backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",borderRadius:0,borderWidth:0,padding:5,itemGap:10,itemWidth:25,itemHeight:14,symbolRotate:"inherit",inactiveColor:"#ccc",inactiveBorderColor:"#ccc",inactiveBorderWidth:"auto",itemStyle:{color:"inherit",opacity:"inherit",decal:"inherit",shadowBlur:0,shadowColor:null,shadowOffsetX:0,shadowOffsetY:0,borderColor:"inherit",borderWidth:"auto",borderCap:"inherit",borderJoin:"inherit",borderDashOffset:"inherit",borderMiterLimit:"inherit"},lineStyle:{width:"auto",color:"inherit",inactiveColor:"#ccc",inactiveWidth:2,opacity:"inherit",type:"inherit",cap:"inherit",join:"inherit",dashOffset:"inherit",miterLimit:"inherit",shadowBlur:0,shadowColor:null,shadowOffsetX:0,shadowOffsetY:0},textStyle:{color:"#333"},selectedMode:!0,selector:!1,selectorLabel:{show:!0,borderRadius:10,padding:[3,5,3,5],fontSize:12,fontFamily:" sans-serif",color:"#666",borderWidth:1,borderColor:"#666"},emphasis:{selectorLabel:{show:!0,color:"#eee",backgroundColor:"#666"}},selectorPosition:"auto",selectorItemGap:7,selectorButtonGap:10,tooltip:{show:!1}},e}(xp),vJ=yJ,_J=G,bJ=R,xJ=Mr,wJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.newlineDisabled=!1,n}return o(e,t),e.prototype.init=function(){this.group.add(this._contentGroup=new xJ),this.group.add(this._selectorGroup=new xJ),this._isFirstRender=!0},e.prototype.getContentGroup=function(){return this._contentGroup},e.prototype.getSelectorGroup=function(){return this._selectorGroup},e.prototype.render=function(t,e,n){var i=this._isFirstRender;if(this._isFirstRender=!1,this.resetInner(),t.get("show",!0)){var r=t.get("align"),o=t.get("orient");r&&"auto"!==r||(r="right"===t.get("left")&&"vertical"===o?"right":"left");var a=t.get("selector",!0),s=t.get("selectorPosition",!0);!a||s&&"auto"!==s||(s="horizontal"===o?"end":"start"),this.renderInner(r,t,e,n,a,o,s);var l=t.getBoxLayoutParams(),u={width:n.getWidth(),height:n.getHeight()},c=t.get("padding"),d=dp(l,u,c),h=this.layoutInner(t,r,d,i,a,s),f=dp(A({width:h.width,height:h.height},l),u,c);this.group.x=f.x-h.x,this.group.y=f.y-h.y,this.group.markRedraw(),this.group.add(this._backgroundEl=LU(h,t))}},e.prototype.resetInner=function(){this.getContentGroup().removeAll(),this._backgroundEl&&this.group.remove(this._backgroundEl),this.getSelectorGroup().removeAll()},e.prototype.renderInner=function(t,e,n,i,r,o,a){var s=this.getContentGroup(),l=gt(),u=e.get("selectedMode"),c=[];n.eachRawSeries((function(t){!t.get("legendHoverLink")&&c.push(t.id)})),bJ(e.getData(),(function(r,o){var a=r.get("name");if(!this.newlineDisabled&&(""===a||"\n"===a)){var d=new xJ;return d.newline=!0,void s.add(d)}var h=n.getSeriesByName(a)[0];if(!l.get(a)){if(h){var f=h.getData(),p=f.getVisual("legendLineStyle")||{},m=f.getVisual("legendIcon"),g=f.getVisual("style"),y=this._createItem(h,a,o,r,e,t,p,g,m,u);y.on("click",_J(LJ,a,null,i,c)).on("mouseover",_J(TJ,h.name,null,i,c)).on("mouseout",_J(DJ,h.name,null,i,c)),l.set(a,!0)}else n.eachRawSeries((function(n){if(!l.get(a)&&n.legendVisualProvider){var s=n.legendVisualProvider;if(!s.containName(a))return;var d=s.indexOfName(a),h=s.getItemVisual(d,"style"),f=s.getItemVisual(d,"legendIcon"),p=yn(h.fill);p&&0===p[3]&&(p[3]=.2,h.fill=Ln(p,"rgba"));var m=this._createItem(n,a,o,r,e,t,{},h,f,u);m.on("click",_J(LJ,null,a,i,c)).on("mouseover",_J(TJ,null,a,i,c)).on("mouseout",_J(DJ,null,a,i,c)),l.set(a,!0)}}),this);0}}),this),r&&this._createSelector(r,e,i,o,a)},e.prototype._createSelector=function(t,e,n,i,r){var o=this.getSelectorGroup();bJ(t,(function(t){var i=t.type,r=new Gl({style:{x:0,y:0,align:"center",verticalAlign:"middle"},onclick:function(){n.dispatchAction({type:"all"===i?"legendAllSelect":"legendInverseSelect"})}});o.add(r);var a=e.getModel("selectorLabel"),s=e.getModel(["emphasis","selectorLabel"]);Mh(r,{normal:a,emphasis:s},{defaultText:t.title}),$u(r)}))},e.prototype._createItem=function(t,e,n,i,r,o,a,s,l,u){var c=t.visualDrawType,d=r.get("itemWidth"),h=r.get("itemHeight"),f=r.isSelected(e),p=i.get("symbolRotate"),m=i.get("icon");l=m||l||"roundRect";var g=r.getModel("lineStyle"),y=MJ(l,i,g,a,s,c,f),v=new xJ,_=i.getModel("textStyle");if("function"!==typeof t.getLegendIcon||m&&"inherit"!==m){var b="inherit"===m&&t.getData().getVisual("symbol")?"inherit"===p?t.getData().getVisual("symbolRotate"):p:0;v.add(SJ({itemWidth:d,itemHeight:h,icon:l,iconRotate:b,itemStyle:y.itemStyle,lineStyle:y.lineStyle}))}else v.add(t.getLegendIcon({itemWidth:d,itemHeight:h,icon:l,iconRotate:p,itemStyle:y.itemStyle,lineStyle:y.lineStyle}));var x="left"===o?d+5:-5,w=o,M=r.get("formatter"),S=e;"string"===typeof M&&M?S=M.replace("{name}",null!=e?e:""):"function"===typeof M&&(S=M(e));var L=i.get("inactiveColor");v.add(new Gl({style:Lh(_,{text:S,x:x,y:h/2,fill:f?_.getTextColor():L,align:w,verticalAlign:"middle"})}));var k=new Al({shape:v.getBoundingRect(),invisible:!0}),T=i.getModel("tooltip");return T.get("show")&&_h({el:k,componentModel:r,itemName:e,itemTooltipOption:T.option}),v.add(k),v.eachChild((function(t){t.silent=!0})),k.silent=!u,this.getContentGroup().add(v),$u(v),v.__legendDataIndex=n,v},e.prototype.layoutInner=function(t,e,n,i,r,o){var a=this.getContentGroup(),s=this.getSelectorGroup();up(t.get("orient"),a,t.get("itemGap"),n.width,n.height);var l=a.getBoundingRect(),u=[-l.x,-l.y];if(s.markRedraw(),a.markRedraw(),r){up("horizontal",s,t.get("selectorItemGap",!0));var c=s.getBoundingRect(),d=[-c.x,-c.y],h=t.get("selectorButtonGap",!0),f=t.getOrient().index,p=0===f?"width":"height",m=0===f?"height":"width",g=0===f?"y":"x";"end"===o?d[f]+=l[p]+h:u[f]+=c[p]+h,d[1-f]+=l[m]/2-c[m]/2,s.x=d[0],s.y=d[1],a.x=u[0],a.y=u[1];var y={x:0,y:0};return y[p]=l[p]+h+c[p],y[m]=Math.max(l[m],c[m]),y[g]=Math.min(0,c[g]+d[1-f]),y}return a.x=u[0],a.y=u[1],this.group.getBoundingRect()},e.prototype.remove=function(){this.getContentGroup().removeAll(),this._isFirstRender=!0},e.type="legend.plain",e}(Py);function MJ(t,e,n,i,r,o,a){for(var s=e.getModel("itemStyle"),l=Vh.concat([["decal"]]),u={},c=0;c<l.length;++c){var d=l[c][l[c].length-1],h=l[c][0],f=s.getShallow(d);if("inherit"===f)switch(h){case"fill":u.fill=r[o];break;case"stroke":u.stroke=r[0===t.lastIndexOf("empty",0)?"fill":"stroke"];break;case"opacity":u.opacity=("fill"===o?r:i).opacity;break;default:u[h]=r[h]}else"auto"===f&&"lineWidth"===h?u.lineWidth=r.lineWidth>0?2:0:u[h]=f}var p=e.getModel("lineStyle"),m=Wh.concat([["inactiveColor"],["inactiveWidth"]]),g={};for(c=0;c<m.length;++c){d=m[c][1],h=m[c][0],f=p.getShallow(d);"inherit"===f?g[h]=i[h]:"auto"===f&&"lineWidth"===h?g.lineWidth=i.lineWidth>0?2:0:g[h]=f}if("auto"===u.fill&&(u.fill=r.fill),"auto"===u.stroke&&(u.stroke=r.fill),"auto"===g.stroke&&(g.stroke=r.fill),!a){var y=e.get("inactiveBorderWidth"),v=u[t.indexOf("empty")>-1?"fill":"stroke"];u.lineWidth="auto"===y?r.lineWidth>0&&v?2:0:u.lineWidth,u.fill=e.get("inactiveColor"),u.stroke=e.get("inactiveBorderColor"),g.stroke=n.get("inactiveColor"),g.lineWidth=n.get("inactiveWidth")}return{itemStyle:u,lineStyle:g}}function SJ(t){var e=t.icon||"roundRect",n=j_(e,0,0,t.itemWidth,t.itemHeight,t.itemStyle.fill);return n.setStyle(t.itemStyle),n.rotation=(t.iconRotate||0)*Math.PI/180,n.setOrigin([t.itemWidth/2,t.itemHeight/2]),e.indexOf("empty")>-1&&(n.style.stroke=n.style.fill,n.style.fill="#fff",n.style.lineWidth=2),n}function LJ(t,e,n,i){DJ(t,e,n,i),n.dispatchAction({type:"legendToggleSelect",name:null!=t?t:e}),TJ(t,e,n,i)}function kJ(t){var e,n=t.getZr().storage.getDisplayList(),i=0,r=n.length;while(i<r&&!(e=n[i].states.emphasis))i++;return e&&e.hoverLayer}function TJ(t,e,n,i){kJ(n)||n.dispatchAction({type:"highlight",seriesName:t,name:e,excludeSeriesId:i})}function DJ(t,e,n,i){kJ(n)||n.dispatchAction({type:"downplay",seriesName:t,name:e,excludeSeriesId:i})}var CJ=wJ;function IJ(t){var e=t.findComponents({mainType:"legend"});e&&e.length&&t.filterSeries((function(t){for(var n=0;n<e.length;n++)if(!e[n].isSelected(t.name))return!1;return!0}))}function AJ(t,e,n){var i,r={},o="toggleSelected"===t;return n.eachComponent("legend",(function(n){o&&null!=i?n[i?"select":"unSelect"](e.name):"allSelect"===t||"inverseSelect"===t?n[t]():(n[t](e.name),i=n.isSelected(e.name));var a=n.getData();R(a,(function(t){var e=t.get("name");if("\n"!==e&&""!==e){var i=n.isSelected(e);r.hasOwnProperty(e)?r[e]=r[e]&&i:r[e]=i}}))})),"allSelect"===t||"inverseSelect"===t?{selected:r}:{name:e.name,selected:r}}function YJ(t){t.registerAction("legendToggleSelect","legendselectchanged",G(AJ,"toggleSelected")),t.registerAction("legendAllSelect","legendselectall",G(AJ,"allSelect")),t.registerAction("legendInverseSelect","legendinverseselect",G(AJ,"inverseSelect")),t.registerAction("legendSelect","legendselected",G(AJ,"select")),t.registerAction("legendUnSelect","legendunselected",G(AJ,"unSelect"))}function OJ(t){t.registerComponentModel(vJ),t.registerComponentView(CJ),t.registerProcessor(t.PRIORITY.PROCESSOR.SERIES_FILTER,IJ),t.registerSubTypeDefaulter("legend",(function(){return"plain"})),YJ(t)}var PJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.setScrollDataIndex=function(t){this.option.scrollDataIndex=t},e.prototype.init=function(e,n,i){var r=gp(e);t.prototype.init.call(this,e,n,i),EJ(this,e,r)},e.prototype.mergeOption=function(e,n){t.prototype.mergeOption.call(this,e,n),EJ(this,this.option,e)},e.type="legend.scroll",e.defaultOption=Qh(vJ.defaultOption,{scrollDataIndex:0,pageButtonItemGap:5,pageButtonGap:null,pageButtonPosition:"end",pageFormatter:"{current}/{total}",pageIcons:{horizontal:["M0,0L12,-10L12,10z","M0,0L-12,-10L-12,10z"],vertical:["M0,0L20,0L10,-20z","M0,0L20,0L10,20z"]},pageIconColor:"#2f4554",pageIconInactiveColor:"#aaa",pageIconSize:15,pageTextStyle:{color:"#333"},animationDurationUpdate:800}),e}(vJ);function EJ(t,e,n){var i=t.getOrient(),r=[1,1];r[i.index]=0,mp(e,n,{type:"box",ignoreSize:!!r})}var NJ=PJ,RJ=Mr,HJ=["width","height"],jJ=["x","y"],zJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.newlineDisabled=!0,n._currentIndex=0,n}return o(e,t),e.prototype.init=function(){t.prototype.init.call(this),this.group.add(this._containerGroup=new RJ),this._containerGroup.add(this.getContentGroup()),this.group.add(this._controllerGroup=new RJ)},e.prototype.resetInner=function(){t.prototype.resetInner.call(this),this._controllerGroup.removeAll(),this._containerGroup.removeClipPath(),this._containerGroup.__rectSize=null},e.prototype.renderInner=function(e,n,i,r,o,a,s){var l=this;t.prototype.renderInner.call(this,e,n,i,r,o,a,s);var u=this._controllerGroup,c=n.get("pageIconSize",!0),d=$(c)?c:[c,c];f("pagePrev",0);var h=n.getModel("pageTextStyle");function f(t,e){var i=t+"DataIndex",o=ph(n.get("pageIcons",!0)[n.getOrient().name][e],{onclick:V(l._pageGo,l,i,n,r)},{x:-d[0]/2,y:-d[1]/2,width:d[0],height:d[1]});o.name=t,u.add(o)}u.add(new Gl({name:"pageText",style:{text:"xx/xx",fill:h.getTextColor(),font:h.getFont(),verticalAlign:"middle",align:"center"},silent:!0})),f("pageNext",1)},e.prototype.layoutInner=function(t,e,n,i,r,o){var a=this.getSelectorGroup(),s=t.getOrient().index,l=HJ[s],u=jJ[s],c=HJ[1-s],d=jJ[1-s];r&&up("horizontal",a,t.get("selectorItemGap",!0));var h=t.get("selectorButtonGap",!0),f=a.getBoundingRect(),p=[-f.x,-f.y],m=T(n);r&&(m[l]=n[l]-f[l]-h);var g=this._layoutContentAndController(t,i,m,s,l,c,d,u);if(r){if("end"===o)p[s]+=g[l]+h;else{var y=f[l]+h;p[s]-=y,g[u]-=y}g[l]+=f[l]+h,p[1-s]+=g[d]+g[c]/2-f[c]/2,g[c]=Math.max(g[c],f[c]),g[d]=Math.min(g[d],f[d]+p[1-s]),a.x=p[0],a.y=p[1],a.markRedraw()}return g},e.prototype._layoutContentAndController=function(t,e,n,i,r,o,a,s){var l=this.getContentGroup(),u=this._containerGroup,c=this._controllerGroup;up(t.get("orient"),l,t.get("itemGap"),i?n.width:null,i?null:n.height),up("horizontal",c,t.get("pageButtonItemGap",!0));var d=l.getBoundingRect(),h=c.getBoundingRect(),f=this._showController=d[r]>n[r],p=[-d.x,-d.y];e||(p[i]=l[s]);var m=[0,0],g=[-h.x,-h.y],y=at(t.get("pageButtonGap",!0),t.get("itemGap",!0));if(f){var v=t.get("pageButtonPosition",!0);"end"===v?g[i]+=n[r]-h[r]:m[i]+=h[r]+y}g[1-i]+=d[o]/2-h[o]/2,l.setPosition(p),u.setPosition(m),c.setPosition(g);var _={x:0,y:0};if(_[r]=f?n[r]:d[r],_[o]=Math.max(d[o],h[o]),_[a]=Math.min(0,h[a]+g[1-i]),u.__rectSize=n[r],f){var b={x:0,y:0};b[r]=Math.max(n[r]-h[r]-y,0),b[o]=_[o],u.setClipPath(new Al({shape:b})),u.__rectSize=b[r]}else c.eachChild((function(t){t.attr({invisible:!0,silent:!0})}));var x=this._getPageInfo(t);return null!=x.pageIndex&&th(l,{x:x.contentPosition[0],y:x.contentPosition[1]},f?t:null),this._updatePageInfoView(t,x),_},e.prototype._pageGo=function(t,e,n){var i=this._getPageInfo(e)[t];null!=i&&n.dispatchAction({type:"legendScroll",scrollDataIndex:i,legendId:e.id})},e.prototype._updatePageInfoView=function(t,e){var n=this._controllerGroup;R(["pagePrev","pageNext"],(function(i){var r=i+"DataIndex",o=null!=e[r],a=n.childOfName(i);a&&(a.setStyle("fill",o?t.get("pageIconColor",!0):t.get("pageIconInactiveColor",!0)),a.cursor=o?"pointer":"default")}));var i=n.childOfName("pageText"),r=t.get("pageFormatter"),o=e.pageIndex,a=null!=o?o+1:0,s=e.pageCount;i&&r&&i.setStyle("text",q(r)?r.replace("{current}",null==a?"":a+"").replace("{total}",null==s?"":s+""):r({current:a,total:s}))},e.prototype._getPageInfo=function(t){var e=t.get("scrollDataIndex",!0),n=this.getContentGroup(),i=this._containerGroup.__rectSize,r=t.getOrient().index,o=HJ[r],a=jJ[r],s=this._findTargetItemIndex(e),l=n.children(),u=l[s],c=l.length,d=c?1:0,h={contentPosition:[n.x,n.y],pageCount:d,pageIndex:d-1,pagePrevDataIndex:null,pageNextDataIndex:null};if(!u)return h;var f=v(u);h.contentPosition[r]=-f.s;for(var p=s+1,m=f,g=f,y=null;p<=c;++p)y=v(l[p]),(!y&&g.e>m.s+i||y&&!_(y,m.s))&&(m=g.i>m.i?g:y,m&&(null==h.pageNextDataIndex&&(h.pageNextDataIndex=m.i),++h.pageCount)),g=y;for(p=s-1,m=f,g=f,y=null;p>=-1;--p)y=v(l[p]),y&&_(g,y.s)||!(m.i<g.i)||(g=m,null==h.pagePrevDataIndex&&(h.pagePrevDataIndex=m.i),++h.pageCount,++h.pageIndex),m=y;return h;function v(t){if(t){var e=t.getBoundingRect(),n=e[a]+t[a];return{s:n,e:n+e[o],i:t.__legendDataIndex}}}function _(t,e){return t.e>=e&&t.s<=e+i}},e.prototype._findTargetItemIndex=function(t){if(!this._showController)return 0;var e,n,i=this.getContentGroup();return i.eachChild((function(i,r){var o=i.__legendDataIndex;null==n&&null!=o&&(n=r),o===t&&(e=r)})),null!=e?e:n},e.type="legend.scroll",e}(CJ),WJ=zJ;function BJ(t){t.registerAction("legendScroll","legendscroll",(function(t,e){var n=t.scrollDataIndex;null!=n&&e.eachComponent({mainType:"legend",subType:"scroll",query:t},(function(t){t.setScrollDataIndex(n)}))}))}function FJ(t){vM(OJ),t.registerComponentModel(NJ),t.registerComponentView(WJ),BJ(t)}function VJ(t){vM(OJ),vM(FJ)}var GJ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.inside",e.defaultOption=Qh(tU.defaultOption,{disabled:!1,zoomLock:!1,zoomOnMouseWheel:!0,moveOnMouseMove:!0,moveOnMouseWheel:!1,preventDefaultMouseMove:!0}),e}(tU),$J=GJ,UJ=Io();function qJ(t,e,n){UJ(t).coordSysRecordMap.each((function(t){var i=t.dataZoomInfoMap.get(e.uid);i&&(i.getRange=n)}))}function XJ(t,e){for(var n=UJ(t).coordSysRecordMap,i=n.keys(),r=0;r<i.length;r++){var o=i[r],a=n.get(o),s=a.dataZoomInfoMap;if(s){var l=e.uid,u=s.get(l);u&&(s.removeKey(l),s.keys().length||ZJ(n,a))}}}function ZJ(t,e){if(e){t.removeKey(e.model.uid);var n=e.controller;n&&n.dispose()}}function JJ(t,e){var n={model:e,containsPoint:G(QJ,e),dispatchAction:G(KJ,t),dataZoomInfoMap:null,controller:null},i=n.controller=new CA(t.getZr());return R(["pan","zoom","scrollMove"],(function(t){i.on(t,(function(e){var i=[];n.dataZoomInfoMap.each((function(r){if(e.isAvailableBehavior(r.model.option)){var o=(r.getRange||{})[t],a=o&&o(r.dzReferCoordSysInfo,n.model.mainType,n.controller,e);!r.model.get("disabled",!0)&&a&&i.push({dataZoomId:r.model.id,start:a[0],end:a[1]})}})),i.length&&n.dispatchAction(i)}))})),n}function KJ(t,e){t.dispatchAction({type:"dataZoom",animation:{easing:"cubicOut",duration:100},batch:e})}function QJ(t,e,n,i){return t.coordinateSystem.containPoint([n,i])}function tK(t){var e,n="type_",i={type_true:2,type_move:1,type_false:0,type_undefined:-1},r=!0;return t.each((function(t){var o=t.model,a=!o.get("disabled",!0)&&(!o.get("zoomLock",!0)||"move");i[n+a]>i[n+e]&&(e=a),r=r&&o.get("preventDefaultMouseMove",!0)})),{controlType:e,opt:{zoomOnMouseWheel:!0,moveOnMouseMove:!0,moveOnMouseWheel:!0,preventDefaultMouseMove:!!r}}}function eK(t){t.registerProcessor(t.PRIORITY.PROCESSOR.FILTER,(function(t,e){var n=UJ(e),i=n.coordSysRecordMap||(n.coordSysRecordMap=gt());i.each((function(t){t.dataZoomInfoMap=null})),t.eachComponent({mainType:"dataZoom",subType:"inside"},(function(t){var n=Z$(t);R(n.infoList,(function(n){var r=n.model.uid,o=i.get(r)||i.set(r,JJ(e,n.model)),a=o.dataZoomInfoMap||(o.dataZoomInfoMap=gt());a.set(t.uid,{dzReferCoordSysInfo:n,model:t,getRange:null})}))})),i.each((function(t){var e,n=t.controller,r=t.dataZoomInfoMap;if(r){var o=r.keys()[0];null!=o&&(e=r.get(o))}if(e){var a=tK(r);n.enable(a.controlType,a.opt),n.setPointerChecker(t.containsPoint),Xy(t,"dispatchAction",e.model.get("throttle",!0),"fixRate")}else ZJ(i,t)}))}))}var nK=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="dataZoom.inside",e}return o(e,t),e.prototype.render=function(e,n,i){t.prototype.render.apply(this,arguments),e.noTarget()?this._clear():(this.range=e.getPercentRange(),qJ(i,e,{pan:V(iK.pan,this),zoom:V(iK.zoom,this),scrollMove:V(iK.scrollMove,this)}))},e.prototype.dispose=function(){this._clear(),t.prototype.dispose.apply(this,arguments)},e.prototype._clear=function(){XJ(this.api,this.dataZoomModel),this.range=null},e.type="dataZoom.inside",e}(rU),iK={zoom:function(t,e,n,i){var r=this.range,o=r.slice(),a=t.axisModels[0];if(a){var s=oK[e](null,[i.originX,i.originY],a,n,t),l=(s.signal>0?s.pixelStart+s.pixelLength-s.pixel:s.pixel-s.pixelStart)/s.pixelLength*(o[1]-o[0])+o[0],u=Math.max(1/i.scale,0);o[0]=(o[0]-l)*u+l,o[1]=(o[1]-l)*u+l;var c=this.dataZoomModel.findRepresentativeAxisProxy().getMinMaxSpan();return rR(0,o,[0,100],0,c.minSpan,c.maxSpan),this.range=o,r[0]!==o[0]||r[1]!==o[1]?o:void 0}},pan:rK((function(t,e,n,i,r,o){var a=oK[i]([o.oldX,o.oldY],[o.newX,o.newY],e,r,n);return a.signal*(t[1]-t[0])*a.pixel/a.pixelLength})),scrollMove:rK((function(t,e,n,i,r,o){var a=oK[i]([0,0],[o.scrollDelta,o.scrollDelta],e,r,n);return a.signal*(t[1]-t[0])*o.scrollDelta}))};function rK(t){return function(e,n,i,r){var o=this.range,a=o.slice(),s=e.axisModels[0];if(s){var l=t(a,s,e,n,i,r);return rR(l,a,[0,100],"all"),this.range=a,o[0]!==a[0]||o[1]!==a[1]?a:void 0}}}var oK={grid:function(t,e,n,i,r){var o=n.axis,a={},s=r.model.coordinateSystem.getRect();return t=t||[0,0],"x"===o.dim?(a.pixel=e[0]-t[0],a.pixelLength=s.width,a.pixelStart=s.x,a.signal=o.inverse?1:-1):(a.pixel=e[1]-t[1],a.pixelLength=s.height,a.pixelStart=s.y,a.signal=o.inverse?-1:1),a},polar:function(t,e,n,i,r){var o=n.axis,a={},s=r.model.coordinateSystem,l=s.getRadiusAxis().getExtent(),u=s.getAngleAxis().getExtent();return t=t?s.pointToCoord(t):[0,0],e=s.pointToCoord(e),"radiusAxis"===n.mainType?(a.pixel=e[0]-t[0],a.pixelLength=l[1]-l[0],a.pixelStart=l[0],a.signal=o.inverse?1:-1):(a.pixel=e[1]-t[1],a.pixelLength=u[1]-u[0],a.pixelStart=u[0],a.signal=o.inverse?-1:1),a},singleAxis:function(t,e,n,i,r){var o=n.axis,a=r.model.coordinateSystem.getRect(),s={};return t=t||[0,0],"horizontal"===o.orient?(s.pixel=e[0]-t[0],s.pixelLength=a.width,s.pixelStart=a.x,s.signal=o.inverse?1:-1):(s.pixel=e[1]-t[1],s.pixelLength=a.height,s.pixelStart=a.y,s.signal=o.inverse?-1:1),s}},aK=nK;function sK(t){gU(t),t.registerComponentModel($J),t.registerComponentView(aK),eK(t)}var lK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.type="dataZoom.slider",e.layoutMode="box",e.defaultOption=Qh(tU.defaultOption,{show:!0,right:"ph",top:"ph",width:"ph",height:"ph",left:null,bottom:null,borderColor:"#d2dbee",borderRadius:3,backgroundColor:"rgba(47,69,84,0)",dataBackground:{lineStyle:{color:"#d2dbee",width:.5},areaStyle:{color:"#d2dbee",opacity:.2}},selectedDataBackground:{lineStyle:{color:"#8fb0f7",width:.5},areaStyle:{color:"#8fb0f7",opacity:.2}},fillerColor:"rgba(135,175,274,0.2)",handleIcon:"path://M-9.35,34.56V42m0-40V9.5m-2,0h4a2,2,0,0,1,2,2v21a2,2,0,0,1-2,2h-4a2,2,0,0,1-2-2v-21A2,2,0,0,1-11.35,9.5Z",handleSize:"100%",handleStyle:{color:"#fff",borderColor:"#ACB8D1"},moveHandleSize:7,moveHandleIcon:"path://M-320.9-50L-320.9-50c18.1,0,27.1,9,27.1,27.1V85.7c0,18.1-9,27.1-27.1,27.1l0,0c-18.1,0-27.1-9-27.1-27.1V-22.9C-348-41-339-50-320.9-50z M-212.3-50L-212.3-50c18.1,0,27.1,9,27.1,27.1V85.7c0,18.1-9,27.1-27.1,27.1l0,0c-18.1,0-27.1-9-27.1-27.1V-22.9C-239.4-41-230.4-50-212.3-50z M-103.7-50L-103.7-50c18.1,0,27.1,9,27.1,27.1V85.7c0,18.1-9,27.1-27.1,27.1l0,0c-18.1,0-27.1-9-27.1-27.1V-22.9C-130.9-41-121.8-50-103.7-50z",moveHandleStyle:{color:"#D2DBEE",opacity:.7},showDetail:!0,showDataShadow:"auto",realtime:!0,zoomLock:!1,textStyle:{color:"#6E7079"},brushSelect:!0,brushStyle:{color:"rgba(135,175,274,0.15)"},emphasis:{handleStyle:{borderColor:"#8FB0F7"},moveHandleStyle:{color:"#8FB0F7"}}}),e}(tU),uK=lK,cK=Al,dK=7,hK=1,fK=30,pK=7,mK="horizontal",gK="vertical",yK=5,vK=["line","bar","candlestick","scatter"],_K={easing:"cubicOut",duration:100},bK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._displayables={},n}return o(e,t),e.prototype.init=function(t,e){this.api=e,this._onBrush=V(this._onBrush,this),this._onBrushEnd=V(this._onBrushEnd,this)},e.prototype.render=function(e,n,i,r){if(t.prototype.render.apply(this,arguments),Xy(this,"_dispatchZoomAction",e.get("throttle"),"fixRate"),this._orient=e.getOrient(),!1!==e.get("show")){if(e.noTarget())return this._clear(),void this.group.removeAll();r&&"dataZoom"===r.type&&r.from===this.uid||this._buildView(),this._updateView()}else this.group.removeAll()},e.prototype.dispose=function(){this._clear(),t.prototype.dispose.apply(this,arguments)},e.prototype._clear=function(){Zy(this,"_dispatchZoomAction");var t=this.api.getZr();t.off("mousemove",this._onBrush),t.off("mouseup",this._onBrushEnd)},e.prototype._buildView=function(){var t=this.group;t.removeAll(),this._brushing=!1,this._displayables.brushRect=null,this._resetLocation(),this._resetInterval();var e=this._displayables.sliderGroup=new Mr;this._renderBackground(),this._renderHandle(),this._renderDataShadow(),t.add(e),this._positionGroup()},e.prototype._resetLocation=function(){var t=this.dataZoomModel,e=this.api,n=t.get("brushSelect"),i=n?pK:0,r=this._findCoordRect(),o={width:e.getWidth(),height:e.getHeight()},a=this._orient===mK?{right:o.width-r.x-r.width,top:o.height-fK-dK-i,width:r.width,height:fK}:{right:dK,top:r.y,width:fK,height:r.height},s=gp(t.option);R(["right","top","width","height"],(function(t){"ph"===s[t]&&(s[t]=a[t])}));var l=dp(s,o);this._location={x:l.x,y:l.y},this._size=[l.width,l.height],this._orient===gK&&this._size.reverse()},e.prototype._positionGroup=function(){var t=this.group,e=this._location,n=this._orient,i=this.dataZoomModel.getFirstTargetAxisModel(),r=i&&i.get("inverse"),o=this._displayables.sliderGroup,a=(this._dataShadowInfo||{}).otherAxisInverse;o.attr(n!==mK||r?n===mK&&r?{scaleY:a?1:-1,scaleX:-1}:n!==gK||r?{scaleY:a?-1:1,scaleX:-1,rotation:Math.PI/2}:{scaleY:a?-1:1,scaleX:1,rotation:Math.PI/2}:{scaleY:a?1:-1,scaleX:1});var s=t.getBoundingRect([o]);t.x=e.x-s.x,t.y=e.y-s.y,t.markRedraw()},e.prototype._getViewExtent=function(){return[0,this._size[0]]},e.prototype._renderBackground=function(){var t=this.dataZoomModel,e=this._size,n=this._displayables.sliderGroup,i=t.get("brushSelect");n.add(new cK({silent:!0,shape:{x:0,y:0,width:e[0],height:e[1]},style:{fill:t.get("backgroundColor")},z2:-40}));var r=new cK({shape:{x:0,y:0,width:e[0],height:e[1]},style:{fill:"transparent"},z2:0,onclick:V(this._onClickPanel,this)}),o=this.api.getZr();i?(r.on("mousedown",this._onBrushStart,this),r.cursor="crosshair",o.on("mousemove",this._onBrush),o.on("mouseup",this._onBrushEnd)):(o.off("mousemove",this._onBrush),o.off("mouseup",this._onBrushEnd)),n.add(r)},e.prototype._renderDataShadow=function(){var t=this._dataShadowInfo=this._prepareDataShadowInfo();if(this._displayables.dataShadowSegs=[],t){var e=this._size,n=t.series,i=n.getRawData(),r=n.getShadowDim?n.getShadowDim():t.otherDim;if(null!=r){var o=i.getDataExtent(r),a=.3*(o[1]-o[0]);o=[o[0]-a,o[1]+a];var s,l=[0,e[1]],u=[0,e[0]],c=[[e[0],0],[0,0]],d=[],h=u[1]/(i.count()-1),f=0,p=Math.round(i.count()/e[0]);i.each([r],(function(t,e){if(p>0&&e%p)f+=h;else{var n=null==t||isNaN(t)||""===t,i=n?0:Er(t,o,l,!0);n&&!s&&e?(c.push([c[c.length-1][0],0]),d.push([d[d.length-1][0],0])):!n&&s&&(c.push([f,0]),d.push([f,0])),c.push([f,i]),d.push([f,i]),f+=h,s=n}}));for(var m=this.dataZoomModel,g=0;g<3;g++){var y=v(1===g);this._displayables.sliderGroup.add(y),this._displayables.dataShadowSegs.push(y)}}}function v(t){var e=m.getModel(t?"selectedDataBackground":"dataBackground"),n=new Mr,i=new rd({shape:{points:c},segmentIgnoreThreshold:1,style:e.getModel("areaStyle").getAreaStyle(),silent:!0,z2:-20}),r=new sd({shape:{points:d},segmentIgnoreThreshold:1,style:e.getModel("lineStyle").getLineStyle(),silent:!0,z2:-19});return n.add(i),n.add(r),n}},e.prototype._prepareDataShadowInfo=function(){var t=this.dataZoomModel,e=t.get("showDataShadow");if(!1!==e){var n,i=this.ecModel;return t.eachTargetAxis((function(r,o){var a=t.getAxisProxy(r,o).getTargetSeriesModels();R(a,(function(t){if(!n&&!(!0!==e&&O(vK,t.get("type"))<0)){var a,s=i.getComponent(q$(r),o).axis,l=xK(r),u=t.coordinateSystem;null!=l&&u.getOtherAxis&&(a=u.getOtherAxis(s).inverse),l=t.getData().mapDimension(l),n={thisAxis:s,series:t,thisDim:r,otherDim:l,otherAxisInverse:a}}}),this)}),this),n}},e.prototype._renderHandle=function(){var t=this.group,e=this._displayables,n=e.handles=[null,null],i=e.handleLabels=[null,null],r=this._displayables.sliderGroup,o=this._size,a=this.dataZoomModel,s=this.api,l=a.get("borderRadius")||0,u=a.get("brushSelect"),c=e.filler=new cK({silent:u,style:{fill:a.get("fillerColor")},textConfig:{position:"inside"}});r.add(c),r.add(new cK({silent:!0,subPixelOptimize:!0,shape:{x:0,y:0,width:o[0],height:o[1],r:l},style:{stroke:a.get("dataBackgroundColor")||a.get("borderColor"),lineWidth:hK,fill:"rgba(0,0,0,0)"}})),R([0,1],(function(e){var o=a.get("handleIcon");!N_[o]&&o.indexOf("path://")<0&&o.indexOf("image://")<0&&(o="path://"+o);var s=j_(o,-1,0,2,2,null,!0);s.attr({cursor:wK(this._orient),draggable:!0,drift:V(this._onDragMove,this,e),ondragend:V(this._onDragEnd,this),onmouseover:V(this._showDataInfo,this,!0),onmouseout:V(this._showDataInfo,this,!1),z2:5});var l=s.getBoundingRect(),u=a.get("handleSize");this._handleHeight=Nr(u,this._size[1]),this._handleWidth=l.width/l.height*this._handleHeight,s.setStyle(a.getModel("handleStyle").getItemStyle()),s.style.strokeNoScale=!0,s.rectHover=!0,s.ensureState("emphasis").style=a.getModel(["emphasis","handleStyle"]).getItemStyle(),$u(s);var c=a.get("handleColor");null!=c&&(s.style.fill=c),r.add(n[e]=s);var d=a.getModel("textStyle");t.add(i[e]=new Gl({silent:!0,invisible:!0,style:Lh(d,{x:0,y:0,text:"",verticalAlign:"middle",align:"center",fill:d.getTextColor(),font:d.getFont()}),z2:10}))}),this);var d=c;if(u){var h=Nr(a.get("moveHandleSize"),o[1]),f=e.moveHandle=new Al({style:a.getModel("moveHandleStyle").getItemStyle(),silent:!0,shape:{r:[0,0,2,2],y:o[1]-.5,height:h}}),p=.8*h,m=e.moveHandleIcon=j_(a.get("moveHandleIcon"),-p/2,-p/2,p,p,"#fff",!0);m.silent=!0,m.y=o[1]+h/2-.5,f.ensureState("emphasis").style=a.getModel(["emphasis","moveHandleStyle"]).getItemStyle();var g=Math.min(o[1]/2,Math.max(h,10));d=e.moveZone=new Al({invisible:!0,shape:{y:o[1]-g,height:h+g}}),d.on("mouseover",(function(){s.enterEmphasis(f)})).on("mouseout",(function(){s.leaveEmphasis(f)})),r.add(f),r.add(m),r.add(d)}d.attr({draggable:!0,cursor:wK(this._orient),drift:V(this._onDragMove,this,"all"),ondragstart:V(this._showDataInfo,this,!0),ondragend:V(this._onDragEnd,this),onmouseover:V(this._showDataInfo,this,!0),onmouseout:V(this._showDataInfo,this,!1)})},e.prototype._resetInterval=function(){var t=this._range=this.dataZoomModel.getPercentRange(),e=this._getViewExtent();this._handleEnds=[Er(t[0],[0,100],e,!0),Er(t[1],[0,100],e,!0)]},e.prototype._updateInterval=function(t,e){var n=this.dataZoomModel,i=this._handleEnds,r=this._getViewExtent(),o=n.findRepresentativeAxisProxy().getMinMaxSpan(),a=[0,100];rR(e,i,r,n.get("zoomLock")?"all":t,null!=o.minSpan?Er(o.minSpan,a,r,!0):null,null!=o.maxSpan?Er(o.maxSpan,a,r,!0):null);var s=this._range,l=this._range=Hr([Er(i[0],r,a,!0),Er(i[1],r,a,!0)]);return!s||s[0]!==l[0]||s[1]!==l[1]},e.prototype._updateView=function(t){var e=this._displayables,n=this._handleEnds,i=Hr(n.slice()),r=this._size;R([0,1],(function(t){var i=e.handles[t],o=this._handleHeight;i.attr({scaleX:o/2,scaleY:o/2,x:n[t]+(t?-1:1),y:r[1]/2-o/2})}),this),e.filler.setShape({x:i[0],y:0,width:i[1]-i[0],height:r[1]});var o={x:i[0],width:i[1]-i[0]};e.moveHandle&&(e.moveHandle.setShape(o),e.moveZone.setShape(o),e.moveZone.getBoundingRect(),e.moveHandleIcon&&e.moveHandleIcon.attr("x",o.x+o.width/2));for(var a=e.dataShadowSegs,s=[0,i[0],i[1],r[0]],l=0;l<a.length;l++){var u=a[l],c=u.getClipPath();c||(c=new Al,u.setClipPath(c)),c.setShape({x:s[l],y:0,width:s[l+1]-s[l],height:r[1]})}this._updateDataInfo(t)},e.prototype._updateDataInfo=function(t){var e=this.dataZoomModel,n=this._displayables,i=n.handleLabels,r=this._orient,o=["",""];if(e.get("showDetail")){var a=e.findRepresentativeAxisProxy();if(a){var s=a.getAxisModel().axis,l=this._range,u=t?a.calculateDataWindow({start:l[0],end:l[1]}).valueWindow:a.getDataValueWindow();o=[this._formatLabel(u[0],s),this._formatLabel(u[1],s)]}}var c=Hr(this._handleEnds.slice());function d(t){var e=ah(n.handles[t].parent,this.group),a=lh(0===t?"right":"left",e),s=this._handleWidth/2+yK,l=sh([c[t]+(0===t?-s:s),this._size[1]/2],e);i[t].setStyle({x:l[0],y:l[1],verticalAlign:r===mK?"middle":a,align:r===mK?a:"center",text:o[t]})}d.call(this,0),d.call(this,1)},e.prototype._formatLabel=function(t,e){var n=this.dataZoomModel,i=n.get("labelFormatter"),r=n.get("labelPrecision");null!=r&&"auto"!==r||(r=e.getPixelPrecision());var o=null==t||isNaN(t)?"":"category"===e.type||"time"===e.type?e.scale.getLabel({value:Math.round(t)}):t.toFixed(Math.min(r,20));return U(i)?i(t,o):q(i)?i.replace("{value}",o):o},e.prototype._showDataInfo=function(t){t=this._dragging||t;var e=this._displayables,n=e.handleLabels;n[0].attr("invisible",!t),n[1].attr("invisible",!t),e.moveHandle&&this.api[t?"enterEmphasis":"leaveEmphasis"](e.moveHandle,1)},e.prototype._onDragMove=function(t,e,n,i){this._dragging=!0,fe(i.event);var r=this._displayables.sliderGroup.getLocalTransform(),o=sh([e,n],r,!0),a=this._updateInterval(t,o[0]),s=this.dataZoomModel.get("realtime");this._updateView(!s),a&&s&&this._dispatchZoomAction(!0)},e.prototype._onDragEnd=function(){this._dragging=!1,this._showDataInfo(!1);var t=this.dataZoomModel.get("realtime");!t&&this._dispatchZoomAction(!1)},e.prototype._onClickPanel=function(t){var e=this._size,n=this._displayables.sliderGroup.transformCoordToLocal(t.offsetX,t.offsetY);if(!(n[0]<0||n[0]>e[0]||n[1]<0||n[1]>e[1])){var i=this._handleEnds,r=(i[0]+i[1])/2,o=this._updateInterval("all",n[0]-r);this._updateView(),o&&this._dispatchZoomAction(!1)}},e.prototype._onBrushStart=function(t){var e=t.offsetX,n=t.offsetY;this._brushStart=new Wi(e,n),this._brushing=!0,this._brushStartTime=+new Date},e.prototype._onBrushEnd=function(t){if(this._brushing){var e=this._displayables.brushRect;if(this._brushing=!1,e){e.attr("ignore",!0);var n=e.shape,i=+new Date;if(!(i-this._brushStartTime<200&&Math.abs(n.width)<5)){var r=this._getViewExtent(),o=[0,100];this._range=Hr([Er(n.x,r,o,!0),Er(n.x+n.width,r,o,!0)]),this._handleEnds=[n.x,n.x+n.width],this._updateView(),this._dispatchZoomAction(!1)}}}},e.prototype._onBrush=function(t){this._brushing&&(fe(t.event),this._updateBrushRect(t.offsetX,t.offsetY))},e.prototype._updateBrushRect=function(t,e){var n=this._displayables,i=this.dataZoomModel,r=n.brushRect;r||(r=n.brushRect=new cK({silent:!0,style:i.getModel("brushStyle").getItemStyle()}),n.sliderGroup.add(r)),r.attr("ignore",!1);var o=this._brushStart,a=this._displayables.sliderGroup,s=a.transformCoordToLocal(t,e),l=a.transformCoordToLocal(o.x,o.y),u=this._size;s[0]=Math.max(Math.min(u[0],s[0]),0),r.setShape({x:l[0],y:0,width:s[0]-l[0],height:u[1]})},e.prototype._dispatchZoomAction=function(t){var e=this._range;this.api.dispatchAction({type:"dataZoom",from:this.uid,dataZoomId:this.dataZoomModel.id,animation:t?_K:null,start:e[0],end:e[1]})},e.prototype._findCoordRect=function(){var t,e=Z$(this.dataZoomModel).infoList;if(!t&&e.length){var n=e[0].model.coordinateSystem;t=n.getRect&&n.getRect()}if(!t){var i=this.api.getWidth(),r=this.api.getHeight();t={x:.2*i,y:.2*r,width:.6*i,height:.6*r}}return t},e.type="dataZoom.slider",e}(rU);function xK(t){var e={x:"y",y:"x",radius:"angle",angle:"radius"};return e[t]}function wK(t){return"vertical"===t?"ns-resize":"ew-resize"}var MK=bK;function SK(t){t.registerComponentModel(uK),t.registerComponentView(MK),gU(t)}function LK(t){vM(sK),vM(SK)}var kK={get:function(t,e,n){var i=T((TK[t]||{})[e]);return n&&$(i)?i[i.length-1]:i}},TK={color:{active:["#006edd","#e0ffff"],inactive:["rgba(0,0,0,0)"]},colorHue:{active:[0,360],inactive:[0,0]},colorSaturation:{active:[.3,1],inactive:[0,0]},colorLightness:{active:[.9,.5],inactive:[0,0]},colorAlpha:{active:[.3,1],inactive:[0,0]},opacity:{active:[.3,1],inactive:[0,0]},symbol:{active:["circle","roundRect","diamond"],inactive:["none"]},symbolSize:{active:[10,50],inactive:[0,0]}},DK=kK,CK=hP.mapVisual,IK=hP.eachVisual,AK=$,YK=R,OK=Hr,PK=Er,EK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.stateList=["inRange","outOfRange"],n.replacableOptionKeys=["inRange","outOfRange","target","controller","color"],n.layoutMode={type:"box",ignoreSize:!0},n.dataBound=[-1/0,1/0],n.targetVisuals={},n.controllerVisuals={},n}return o(e,t),e.prototype.init=function(t,e,n){this.mergeDefaultAndTheme(t,n)},e.prototype.optionUpdated=function(t,e){var n=this.option;h.canvasSupported||(n.realtime=!1),!e&&bX(n,t,this.replacableOptionKeys),this.textStyleModel=this.getModel("textStyle"),this.resetItemSize(),this.completeVisualOption()},e.prototype.resetVisual=function(t){var e=this.stateList;t=V(t,this),this.controllerVisuals=_X(this.option.controller,e,t),this.targetVisuals=_X(this.option.target,e,t)},e.prototype.getItemSymbol=function(){return null},e.prototype.getTargetSeriesIndices=function(){var t=this.option.seriesIndex,e=[];return null==t||"all"===t?this.ecModel.eachSeries((function(t,n){e.push(n)})):e=so(t),e},e.prototype.eachTargetSeries=function(t,e){R(this.getTargetSeriesIndices(),(function(n){var i=this.ecModel.getSeriesByIndex(n);i&&t.call(e,i)}),this)},e.prototype.isTargetSeries=function(t){var e=!1;return this.eachTargetSeries((function(n){n===t&&(e=!0)})),e},e.prototype.formatValueText=function(t,e,n){var i,r=this.option,o=r.precision,a=this.dataBound,s=r.formatter;n=n||["<",">"],$(t)&&(t=t.slice(),i=!0);var l=e?t:i?[u(t[0]),u(t[1])]:u(t);return q(s)?s.replace("{value}",i?l[0]:l).replace("{value2}",i?l[1]:l):U(s)?i?s(t[0],t[1]):s(t):i?t[0]===a[0]?n[0]+" "+l[1]:t[1]===a[1]?n[1]+" "+l[0]:l[0]+" - "+l[1]:l;function u(t){return t===a[0]?"min":t===a[1]?"max":(+t).toFixed(Math.min(o,20))}},e.prototype.resetExtent=function(){var t=this.option,e=OK([t.min,t.max]);this._dataExtent=e},e.prototype.getDataDimension=function(t){var e=this.option.dimension,n=t.dimensions;if(null!=e||n.length){if(null!=e)return t.getDimension(e);for(var i=t.dimensions,r=i.length-1;r>=0;r--){var o=i[r],a=t.getDimensionInfo(o);if(!a.isCalculationCoord)return o}}},e.prototype.getExtent=function(){return this._dataExtent.slice()},e.prototype.completeVisualOption=function(){var t=this.ecModel,e=this.option,n={inRange:e.inRange,outOfRange:e.outOfRange},i=e.target||(e.target={}),r=e.controller||(e.controller={});D(i,n),D(r,n);var o=this.isCategory();function a(n){AK(e.color)&&!n.inRange&&(n.inRange={color:e.color.slice().reverse()}),n.inRange=n.inRange||{color:t.get("gradientColor")}}function s(t,e,n){var i=t[e],r=t[n];i&&!r&&(r=t[n]={},YK(i,(function(t,e){if(hP.isValidType(e)){var n=DK.get(e,"inactive",o);null!=n&&(r[e]=n,"color"!==e||r.hasOwnProperty("opacity")||r.hasOwnProperty("colorAlpha")||(r.opacity=[0,0]))}})))}function l(t){var e=(t.inRange||{}).symbol||(t.outOfRange||{}).symbol,n=(t.inRange||{}).symbolSize||(t.outOfRange||{}).symbolSize,i=this.get("inactiveColor"),r=this.getItemSymbol(),a=r||"roundRect";YK(this.stateList,(function(r){var s=this.itemSize,l=t[r];l||(l=t[r]={color:o?i:[i]}),null==l.symbol&&(l.symbol=e&&T(e)||(o?a:[a])),null==l.symbolSize&&(l.symbolSize=n&&T(n)||(o?s[0]:[s[0],s[0]])),l.symbol=CK(l.symbol,(function(t){return"none"===t?a:t}));var u=l.symbolSize;if(null!=u){var c=-1/0;IK(u,(function(t){t>c&&(c=t)})),l.symbolSize=CK(u,(function(t){return PK(t,[0,c],[0,s[0]],!0)}))}}),this)}a.call(this,i),a.call(this,r),s.call(this,i,"inRange","outOfRange"),l.call(this,r)},e.prototype.resetItemSize=function(){this.itemSize=[parseFloat(this.get("itemWidth")),parseFloat(this.get("itemHeight"))]},e.prototype.isCategory=function(){return!!this.option.categories},e.prototype.setSelected=function(t){},e.prototype.getSelected=function(){return null},e.prototype.getValueState=function(t){return null},e.prototype.getVisualMeta=function(t){return null},e.type="visualMap",e.dependencies=["series"],e.defaultOption={show:!0,zlevel:0,z:4,seriesIndex:"all",min:0,max:200,left:0,right:null,top:null,bottom:0,itemWidth:null,itemHeight:null,inverse:!1,orient:"vertical",backgroundColor:"rgba(0,0,0,0)",borderColor:"#ccc",contentColor:"#5793f3",inactiveColor:"#aaa",borderWidth:0,padding:5,textGap:10,precision:0,textStyle:{color:"#333"}},e}(xp),NK=EK,RK=[20,140],HK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.optionUpdated=function(e,n){t.prototype.optionUpdated.apply(this,arguments),this.resetExtent(),this.resetVisual((function(t){t.mappingMethod="linear",t.dataExtent=this.getExtent()})),this._resetRange()},e.prototype.resetItemSize=function(){t.prototype.resetItemSize.apply(this,arguments);var e=this.itemSize;(null==e[0]||isNaN(e[0]))&&(e[0]=RK[0]),(null==e[1]||isNaN(e[1]))&&(e[1]=RK[1])},e.prototype._resetRange=function(){var t=this.getExtent(),e=this.option.range;!e||e.auto?(t.auto=1,this.option.range=t):$(e)&&(e[0]>e[1]&&e.reverse(),e[0]=Math.max(e[0],t[0]),e[1]=Math.min(e[1],t[1]))},e.prototype.completeVisualOption=function(){t.prototype.completeVisualOption.apply(this,arguments),R(this.stateList,(function(t){var e=this.option.controller[t].symbolSize;e&&e[0]!==e[1]&&(e[0]=e[1]/3)}),this)},e.prototype.setSelected=function(t){this.option.range=t.slice(),this._resetRange()},e.prototype.getSelected=function(){var t=this.getExtent(),e=Hr((this.get("range")||[]).slice());return e[0]>t[1]&&(e[0]=t[1]),e[1]>t[1]&&(e[1]=t[1]),e[0]<t[0]&&(e[0]=t[0]),e[1]<t[0]&&(e[1]=t[0]),e},e.prototype.getValueState=function(t){var e=this.option.range,n=this.getExtent();return(e[0]<=n[0]||e[0]<=t)&&(e[1]>=n[1]||t<=e[1])?"inRange":"outOfRange"},e.prototype.findTargetDataIndices=function(t){var e=[];return this.eachTargetSeries((function(n){var i=[],r=n.getData();r.each(this.getDataDimension(r),(function(e,n){t[0]<=e&&e<=t[1]&&i.push(n)}),this),e.push({seriesId:n.id,dataIndex:i})}),this),e},e.prototype.getVisualMeta=function(t){var e=jK(this,"outOfRange",this.getExtent()),n=jK(this,"inRange",this.option.range.slice()),i=[];function r(e,n){i.push({value:e,color:t(e,n)})}for(var o=0,a=0,s=n.length,l=e.length;a<l&&(!n.length||e[a]<=n[0]);a++)e[a]<n[o]&&r(e[a],"outOfRange");for(var u=1;o<s;o++,u=0)u&&i.length&&r(n[o],"outOfRange"),r(n[o],"inRange");for(u=1;a<l;a++)(!n.length||n[n.length-1]<e[a])&&(u&&(i.length&&r(i[i.length-1].value,"outOfRange"),u=0),r(e[a],"outOfRange"));var c=i.length;return{stops:i,outerColors:[c?i[0].color:"transparent",c?i[c-1].color:"transparent"]}},e.type="visualMap.continuous",e.defaultOption=Qh(NK.defaultOption,{align:"auto",calculable:!1,hoverLink:!0,realtime:!0,handleIcon:"path://M-11.39,9.77h0a3.5,3.5,0,0,1-3.5,3.5h-22a3.5,3.5,0,0,1-3.5-3.5h0a3.5,3.5,0,0,1,3.5-3.5h22A3.5,3.5,0,0,1-11.39,9.77Z",handleSize:"120%",handleStyle:{borderColor:"#fff",borderWidth:1},indicatorIcon:"circle",indicatorSize:"50%",indicatorStyle:{borderColor:"#fff",borderWidth:2,shadowBlur:2,shadowOffsetX:1,shadowOffsetY:1,shadowColor:"rgba(0,0,0,0.2)"}}),e}(NK);function jK(t,e,n){if(n[0]===n[1])return n.slice();for(var i=200,r=(n[1]-n[0])/i,o=n[0],a=[],s=0;s<=i&&o<n[1];s++)a.push(o),o+=r;return a.push(n[1]),a}var zK=HK,WK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n.autoPositionValues={left:1,right:1,top:1,bottom:1},n}return o(e,t),e.prototype.init=function(t,e){this.ecModel=t,this.api=e},e.prototype.render=function(t,e,n,i){this.visualMapModel=t,!1!==t.get("show")?this.doRender(t,e,n,i):this.group.removeAll()},e.prototype.renderBackground=function(t){var e=this.visualMapModel,n=Uf(e.get("padding")||0),i=t.getBoundingRect();t.add(new Al({z2:-1,silent:!0,shape:{x:i.x-n[3],y:i.y-n[0],width:i.width+n[3]+n[1],height:i.height+n[0]+n[2]},style:{fill:e.get("backgroundColor"),stroke:e.get("borderColor"),lineWidth:e.get("borderWidth")}}))},e.prototype.getControllerVisual=function(t,e,n){n=n||{};var i=n.forceState,r=this.visualMapModel,o={};if("color"===e){var a=r.get("contentColor");o.color=a}function s(t){return o[t]}function l(t,e){o[t]=e}var u=r.controllerVisuals[i||r.getValueState(t)],c=hP.prepareVisualTypes(u);return R(c,(function(i){var r=u[i];n.convertOpacityToAlpha&&"opacity"===i&&(i="colorAlpha",r=u.__alphaForOpacity),hP.dependsOn(i,e)&&r&&r.applyVisual(t,s,l)})),o[e]},e.prototype.positionGroup=function(t){var e=this.visualMapModel,n=this.api;hp(t,e.getBoxLayoutParams(),{width:n.getWidth(),height:n.getHeight()})},e.prototype.doRender=function(t,e,n,i){},e.type="visualMap",e}(Py),BK=WK,FK=[["left","right","width"],["top","bottom","height"]];function VK(t,e,n){var i=t.option,r=i.align;if(null!=r&&"auto"!==r)return r;for(var o={width:e.getWidth(),height:e.getHeight()},a="horizontal"===i.orient?1:0,s=FK[a],l=[0,null,10],u={},c=0;c<3;c++)u[FK[1-a][c]]=l[c],u[s[c]]=2===c?n[0]:i[s[c]];var d=[["x","width",3],["y","height",0]][a],h=dp(u,o,i.padding);return s[(h.margin[d[2]]||0)+h[d[0]]+.5*h[d[1]]<.5*o[d[1]]?0:1]}function GK(t,e){return R(t||[],(function(t){null!=t.dataIndex&&(t.dataIndexInside=t.dataIndex,t.dataIndex=null),t.highlightKey="visualMap"+(e?e.componentIndex:"")})),t}var $K=Er,UK=R,qK=Math.min,XK=Math.max,ZK=12,JK=6,KK=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._shapes={},n._dataInterval=[],n._handleEnds=[],n._hoverLinkDataIndices=[],n}return o(e,t),e.prototype.doRender=function(t,e,n,i){this._api=n,i&&"selectDataRange"===i.type&&i.from===this.uid||this._buildView()},e.prototype._buildView=function(){this.group.removeAll();var t=this.visualMapModel,e=this.group;this._orient=t.get("orient"),this._useHandle=t.get("calculable"),this._resetInterval(),this._renderBar(e);var n=t.get("text");this._renderEndsText(e,n,0),this._renderEndsText(e,n,1),this._updateView(!0),this.renderBackground(e),this._updateView(),this._enableHoverLinkToSeries(),this._enableHoverLinkFromSeries(),this.positionGroup(e)},e.prototype._renderEndsText=function(t,e,n){if(e){var i=e[1-n];i=null!=i?i+"":"";var r=this.visualMapModel,o=r.get("textGap"),a=r.itemSize,s=this._shapes.mainGroup,l=this._applyTransform([a[0]/2,0===n?-o:a[1]+o],s),u=this._applyTransform(0===n?"bottom":"top",s),c=this._orient,d=this.visualMapModel.textStyleModel;this.group.add(new Gl({style:{x:l[0],y:l[1],verticalAlign:"horizontal"===c?"middle":u,align:"horizontal"===c?u:"center",text:i,font:d.getFont(),fill:d.getTextColor()}}))}},e.prototype._renderBar=function(t){var e=this.visualMapModel,n=this._shapes,i=e.itemSize,r=this._orient,o=this._useHandle,a=VK(e,this.api,i),s=n.mainGroup=this._createBarGroup(a),l=new Mr;s.add(l),l.add(n.outOfRange=QK()),l.add(n.inRange=QK(null,o?nQ(this._orient):null,V(this._dragHandle,this,"all",!1),V(this._dragHandle,this,"all",!0))),l.setClipPath(new Al({shape:{x:0,y:0,width:i[0],height:i[1],r:3}}));var u=e.textStyleModel.getTextRect("国"),c=XK(u.width,u.height);o&&(n.handleThumbs=[],n.handleLabels=[],n.handleLabelPoints=[],this._createHandle(e,s,0,i,c,r),this._createHandle(e,s,1,i,c,r)),this._createIndicator(e,s,i,c,r),t.add(s)},e.prototype._createHandle=function(t,e,n,i,r,o){var a=V(this._dragHandle,this,n,!1),s=V(this._dragHandle,this,n,!0),l=lr(t.get("handleSize"),i[0]),u=j_(t.get("handleIcon"),-l/2,-l/2,l,l,null,!0),c=nQ(this._orient);u.attr({cursor:c,draggable:!0,drift:a,ondragend:s,onmousemove:function(t){fe(t.event)}}),u.x=i[0]/2,u.useStyle(t.getModel("handleStyle").getItemStyle()),u.setStyle({strokeNoScale:!0,strokeFirst:!0}),u.style.lineWidth*=2,u.ensureState("emphasis").style=t.getModel(["emphasis","handleStyle"]).getItemStyle(),Ju(u,!0),e.add(u);var d=this.visualMapModel.textStyleModel,h=new Gl({cursor:c,draggable:!0,drift:a,onmousemove:function(t){fe(t.event)},ondragend:s,style:{x:0,y:0,text:"",font:d.getFont(),fill:d.getTextColor()}});h.ensureState("blur").style={opacity:.1},h.stateTransition={duration:200},this.group.add(h);var f=[l,0],p=this._shapes;p.handleThumbs[n]=u,p.handleLabelPoints[n]=f,p.handleLabels[n]=h},e.prototype._createIndicator=function(t,e,n,i,r){var o=lr(t.get("indicatorSize"),n[0]),a=j_(t.get("indicatorIcon"),-o/2,-o/2,o,o,null,!0);a.attr({cursor:"move",invisible:!0,silent:!0,x:n[0]/2});var s=t.getModel("indicatorStyle").getItemStyle();if(a instanceof wl){var l=a.style;a.useStyle(I({image:l.image,x:l.x,y:l.y,width:l.width,height:l.height},s))}else a.useStyle(s);e.add(a);var u=this.visualMapModel.textStyleModel,c=new Gl({silent:!0,invisible:!0,style:{x:0,y:0,text:"",font:u.getFont(),fill:u.getTextColor()}});this.group.add(c);var d=[("horizontal"===r?i/2:JK)+n[0]/2,0],h=this._shapes;h.indicator=a,h.indicatorLabel=c,h.indicatorLabelPoint=d,this._firstShowIndicator=!0},e.prototype._dragHandle=function(t,e,n,i){if(this._useHandle){if(this._dragging=!e,!e){var r=this._applyTransform([n,i],this._shapes.mainGroup,!0);this._updateInterval(t,r[1]),this._hideIndicator(),this._updateView()}e===!this.visualMapModel.get("realtime")&&this.api.dispatchAction({type:"selectDataRange",from:this.uid,visualMapId:this.visualMapModel.id,selected:this._dataInterval.slice()}),e?!this._hovering&&this._clearHoverLinkToSeries():eQ(this.visualMapModel)&&this._doHoverLinkToSeries(this._handleEnds[t],!1)}},e.prototype._resetInterval=function(){var t=this.visualMapModel,e=this._dataInterval=t.getSelected(),n=t.getExtent(),i=[0,t.itemSize[1]];this._handleEnds=[$K(e[0],n,i,!0),$K(e[1],n,i,!0)]},e.prototype._updateInterval=function(t,e){e=e||0;var n=this.visualMapModel,i=this._handleEnds,r=[0,n.itemSize[1]];rR(e,i,r,t,0);var o=n.getExtent();this._dataInterval=[$K(i[0],r,o,!0),$K(i[1],r,o,!0)]},e.prototype._updateView=function(t){var e=this.visualMapModel,n=e.getExtent(),i=this._shapes,r=[0,e.itemSize[1]],o=t?r:this._handleEnds,a=this._createBarVisual(this._dataInterval,n,o,"inRange"),s=this._createBarVisual(n,n,r,"outOfRange");i.inRange.setStyle({fill:a.barColor}).setShape("points",a.barPoints),i.outOfRange.setStyle({fill:s.barColor}).setShape("points",s.barPoints),this._updateHandle(o,a)},e.prototype._createBarVisual=function(t,e,n,i){var r={forceState:i,convertOpacityToAlpha:!0},o=this._makeColorGradient(t,r),a=[this.getControllerVisual(t[0],"symbolSize",r),this.getControllerVisual(t[1],"symbolSize",r)],s=this._createBarPoints(n,a);return{barColor:new Ld(0,0,0,1,o),barPoints:s,handlesColor:[o[0].color,o[o.length-1].color]}},e.prototype._makeColorGradient=function(t,e){var n=100,i=[],r=(t[1]-t[0])/n;i.push({color:this.getControllerVisual(t[0],"color",e),offset:0});for(var o=1;o<n;o++){var a=t[0]+r*o;if(a>t[1])break;i.push({color:this.getControllerVisual(a,"color",e),offset:o/n})}return i.push({color:this.getControllerVisual(t[1],"color",e),offset:1}),i},e.prototype._createBarPoints=function(t,e){var n=this.visualMapModel.itemSize;return[[n[0]-e[0],t[0]],[n[0],t[0]],[n[0],t[1]],[n[0]-e[1],t[1]]]},e.prototype._createBarGroup=function(t){var e=this._orient,n=this.visualMapModel.get("inverse");return new Mr("horizontal"!==e||n?"horizontal"===e&&n?{scaleX:"bottom"===t?-1:1,rotation:-Math.PI/2}:"vertical"!==e||n?{scaleX:"left"===t?1:-1}:{scaleX:"left"===t?1:-1,scaleY:-1}:{scaleX:"bottom"===t?1:-1,rotation:Math.PI/2})},e.prototype._updateHandle=function(t,e){if(this._useHandle){var n=this._shapes,i=this.visualMapModel,r=n.handleThumbs,o=n.handleLabels,a=i.itemSize,s=i.getExtent();UK([0,1],(function(l){var u=r[l];u.setStyle("fill",e.handlesColor[l]),u.y=t[l];var c=$K(t[l],[0,a[1]],s,!0),d=this.getControllerVisual(c,"symbolSize");u.scaleX=u.scaleY=d/a[0],u.x=a[0]-d/2;var h=sh(n.handleLabelPoints[l],ah(u,this.group));o[l].setStyle({x:h[0],y:h[1],text:i.formatValueText(this._dataInterval[l]),verticalAlign:"middle",align:"vertical"===this._orient?this._applyTransform("left",n.mainGroup):"center"})}),this)}},e.prototype._showIndicator=function(t,e,n,i){var r=this.visualMapModel,o=r.getExtent(),a=r.itemSize,s=[0,a[1]],l=this._shapes,u=l.indicator;if(u){u.attr("invisible",!1);var c={convertOpacityToAlpha:!0},d=this.getControllerVisual(t,"color",c),h=this.getControllerVisual(t,"symbolSize"),f=$K(t,o,s,!0),p=a[0]-h/2,m={x:u.x,y:u.y};u.y=f,u.x=p;var g=sh(l.indicatorLabelPoint,ah(u,this.group)),y=l.indicatorLabel;y.attr("invisible",!1);var v=this._applyTransform("left",l.mainGroup),_=this._orient,b="horizontal"===_;y.setStyle({text:(n||"")+r.formatValueText(e),verticalAlign:b?v:"middle",align:b?"center":v});var x={x:p,y:f,style:{fill:d}},w={style:{x:g[0],y:g[1]}};if(r.ecModel.isAnimationEnabled()&&!this._firstShowIndicator){var M={duration:100,easing:"cubicInOut",additive:!0};u.x=m.x,u.y=m.y,u.animateTo(x,M),y.animateTo(w,M)}else u.attr(x),y.attr(w);this._firstShowIndicator=!1;var S=this._shapes.handleLabels;if(S)for(var L=0;L<S.length;L++)this._api.enterBlur(S[L])}},e.prototype._enableHoverLinkToSeries=function(){var t=this;this._shapes.mainGroup.on("mousemove",(function(e){if(t._hovering=!0,!t._dragging){var n=t.visualMapModel.itemSize,i=t._applyTransform([e.offsetX,e.offsetY],t._shapes.mainGroup,!0,!0);i[1]=qK(XK(0,i[1]),n[1]),t._doHoverLinkToSeries(i[1],0<=i[0]&&i[0]<=n[0])}})).on("mouseout",(function(){t._hovering=!1,!t._dragging&&t._clearHoverLinkToSeries()}))},e.prototype._enableHoverLinkFromSeries=function(){var t=this.api.getZr();this.visualMapModel.option.hoverLink?(t.on("mouseover",this._hoverLinkFromSeriesMouseOver,this),t.on("mouseout",this._hideIndicator,this)):this._clearHoverLinkFromSeries()},e.prototype._doHoverLinkToSeries=function(t,e){var n=this.visualMapModel,i=n.itemSize;if(n.option.hoverLink){var r=[0,i[1]],o=n.getExtent();t=qK(XK(r[0],t),r[1]);var a=tQ(n,o,r),s=[t-a,t+a],l=$K(t,r,o,!0),u=[$K(s[0],r,o,!0),$K(s[1],r,o,!0)];s[0]<r[0]&&(u[0]=-1/0),s[1]>r[1]&&(u[1]=1/0),e&&(u[0]===-1/0?this._showIndicator(l,u[1],"< ",a):u[1]===1/0?this._showIndicator(l,u[0],"> ",a):this._showIndicator(l,l,"≈ ",a));var c=this._hoverLinkDataIndices,d=[];(e||eQ(n))&&(d=this._hoverLinkDataIndices=n.findTargetDataIndices(u));var h=Do(c,d);this._dispatchHighDown("downplay",GK(h[0],n)),this._dispatchHighDown("highlight",GK(h[1],n))}},e.prototype._hoverLinkFromSeriesMouseOver=function(t){var e=t.target,n=this.visualMapModel;if(e&&null!=$l(e).dataIndex){var i=$l(e),r=this.ecModel.getSeriesByIndex(i.seriesIndex);if(n.isTargetSeries(r)){var o=r.getData(i.dataType),a=o.get(n.getDataDimension(o),i.dataIndex);isNaN(a)||this._showIndicator(a,a)}}},e.prototype._hideIndicator=function(){var t=this._shapes;t.indicator&&t.indicator.attr("invisible",!0),t.indicatorLabel&&t.indicatorLabel.attr("invisible",!0);var e=this._shapes.handleLabels;if(e)for(var n=0;n<e.length;n++)this._api.leaveBlur(e[n])},e.prototype._clearHoverLinkToSeries=function(){this._hideIndicator();var t=this._hoverLinkDataIndices;this._dispatchHighDown("downplay",GK(t,this.visualMapModel)),t.length=0},e.prototype._clearHoverLinkFromSeries=function(){this._hideIndicator();var t=this.api.getZr();t.off("mouseover",this._hoverLinkFromSeriesMouseOver),t.off("mouseout",this._hideIndicator)},e.prototype._applyTransform=function(t,e,n,i){var r=ah(e,i?null:this.group);return $(t)?sh(t,r,n):lh(t,r,n)},e.prototype._dispatchHighDown=function(t,e){e&&e.length&&this.api.dispatchAction({type:t,batch:e})},e.prototype.dispose=function(){this._clearHoverLinkFromSeries(),this._clearHoverLinkToSeries()},e.prototype.remove=function(){this._clearHoverLinkFromSeries(),this._clearHoverLinkToSeries()},e.type="visualMap.continuous",e}(BK);function QK(t,e,n,i){return new rd({shape:{points:t},draggable:!!n,cursor:e,drift:n,onmousemove:function(t){fe(t.event)},ondragend:i})}function tQ(t,e,n){var i=ZK/2,r=t.get("hoverLinkDataSize");return r&&(i=$K(r,e,n,!0)/2),i}function eQ(t){var e=t.get("hoverLinkOnHandle");return!!(null==e?t.get("realtime"):e)}function nQ(t){return"vertical"===t?"ns-resize":"ew-resize"}var iQ=KK,rQ={type:"selectDataRange",event:"dataRangeSelected",update:"update"},oQ=function(t,e){e.eachComponent({mainType:"visualMap",query:t},(function(e){e.setSelected(t.selected)}))},aQ=[{createOnAllSeries:!0,reset:function(t,e){var n=[];return e.eachComponent("visualMap",(function(e){var i=t.pipelineContext;!e.isTargetSeries(t)||i&&i.large||n.push(wX(e.stateList,e.targetVisuals,V(e.getValueState,e),e.getDataDimension(t.getData())))})),n}},{createOnAllSeries:!0,reset:function(t,e){var n=t.getData(),i=[];e.eachComponent("visualMap",(function(e){if(e.isTargetSeries(t)){var r=e.getVisualMeta(V(sQ,null,t,e))||{stops:[],outerColors:[]},o=e.getDataDimension(n),a=n.getDimensionInfo(o);null!=a&&(r.dimension=a.index,i.push(r))}})),t.getData().setVisual("visualMeta",i)}}];function sQ(t,e,n,i){for(var r=e.targetVisuals[i],o=hP.prepareVisualTypes(r),a={color:Nv(t.getData(),"color")},s=0,l=o.length;s<l;s++){var u=o[s],c=r["opacity"===u?"__alphaForOpacity":u];c&&c.applyVisual(n,d,h)}return a.color;function d(t){return a[t]}function h(t,e){a[t]=e}}var lQ=R;function uQ(t){var e=t&&t.visualMap;$(e)||(e=e?[e]:[]),lQ(e,(function(t){if(t){cQ(t,"splitList")&&!cQ(t,"pieces")&&(t.pieces=t.splitList,delete t.splitList);var e=t.pieces;e&&$(e)&&lQ(e,(function(t){J(t)&&(cQ(t,"start")&&!cQ(t,"min")&&(t.min=t.start),cQ(t,"end")&&!cQ(t,"max")&&(t.max=t.end))}))}}))}function cQ(t,e){return t&&t.hasOwnProperty&&t.hasOwnProperty(e)}var dQ=!1;function hQ(t){dQ||(dQ=!0,t.registerSubTypeDefaulter("visualMap",(function(t){return t.categories||(t.pieces?t.pieces.length>0:t.splitNumber>0)&&!t.calculable?"piecewise":"continuous"})),t.registerAction(rQ,oQ),R(aQ,(function(e){t.registerVisual(t.PRIORITY.VISUAL.COMPONENT,e)})),t.registerPreprocessor(uQ))}function fQ(t){t.registerComponentModel(zK),t.registerComponentView(iQ),hQ(t)}var pQ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n._pieceList=[],n}return o(e,t),e.prototype.optionUpdated=function(e,n){t.prototype.optionUpdated.apply(this,arguments),this.resetExtent();var i=this._mode=this._determineMode();this._pieceList=[],mQ[this._mode].call(this,this._pieceList),this._resetSelected(e,n);var r=this.option.categories;this.resetVisual((function(t,e){"categories"===i?(t.mappingMethod="category",t.categories=T(r)):(t.dataExtent=this.getExtent(),t.mappingMethod="piecewise",t.pieceList=H(this._pieceList,(function(t){return t=T(t),"inRange"!==e&&(t.visual=null),t})))}))},e.prototype.completeVisualOption=function(){var e=this.option,n={},i=hP.listVisualTypes(),r=this.isCategory();function o(t,e,n){return t&&t[e]&&t[e].hasOwnProperty(n)}R(e.pieces,(function(t){R(i,(function(e){t.hasOwnProperty(e)&&(n[e]=1)}))})),R(n,(function(t,n){var i=!1;R(this.stateList,(function(t){i=i||o(e,t,n)||o(e.target,t,n)}),this),!i&&R(this.stateList,(function(t){(e[t]||(e[t]={}))[n]=DK.get(n,"inRange"===t?"active":"inactive",r)}))}),this),t.prototype.completeVisualOption.apply(this,arguments)},e.prototype._resetSelected=function(t,e){var n=this.option,i=this._pieceList,r=(e?n:t).selected||{};if(n.selected=r,R(i,(function(t,e){var n=this.getSelectedMapKey(t);r.hasOwnProperty(n)||(r[n]=!0)}),this),"single"===n.selectedMode){var o=!1;R(i,(function(t,e){var n=this.getSelectedMapKey(t);r[n]&&(o?r[n]=!1:o=!0)}),this)}},e.prototype.getItemSymbol=function(){return this.get("itemSymbol")},e.prototype.getSelectedMapKey=function(t){return"categories"===this._mode?t.value+"":t.index+""},e.prototype.getPieceList=function(){return this._pieceList},e.prototype._determineMode=function(){var t=this.option;return t.pieces&&t.pieces.length>0?"pieces":this.option.categories?"categories":"splitNumber"},e.prototype.setSelected=function(t){this.option.selected=T(t)},e.prototype.getValueState=function(t){var e=hP.findPieceIndex(t,this._pieceList);return null!=e&&this.option.selected[this.getSelectedMapKey(this._pieceList[e])]?"inRange":"outOfRange"},e.prototype.findTargetDataIndices=function(t){var e=[],n=this._pieceList;return this.eachTargetSeries((function(i){var r=[],o=i.getData();o.each(this.getDataDimension(o),(function(e,i){var o=hP.findPieceIndex(e,n);o===t&&r.push(i)}),this),e.push({seriesId:i.id,dataIndex:r})}),this),e},e.prototype.getRepresentValue=function(t){var e;if(this.isCategory())e=t.value;else if(null!=t.value)e=t.value;else{var n=t.interval||[];e=n[0]===-1/0&&n[1]===1/0?0:(n[0]+n[1])/2}return e},e.prototype.getVisualMeta=function(t){if(!this.isCategory()){var e=[],n=["",""],i=this,r=this._pieceList.slice();if(r.length){var o=r[0].interval[0];o!==-1/0&&r.unshift({interval:[-1/0,o]}),o=r[r.length-1].interval[1],o!==1/0&&r.push({interval:[o,1/0]})}else r.push({interval:[-1/0,1/0]});var a=-1/0;return R(r,(function(t){var e=t.interval;e&&(e[0]>a&&s([a,e[0]],"outOfRange"),s(e.slice()),a=e[1])}),this),{stops:e,outerColors:n}}function s(r,o){var a=i.getRepresentValue({interval:r});o||(o=i.getValueState(a));var s=t(a,o);r[0]===-1/0?n[0]=s:r[1]===1/0?n[1]=s:e.push({value:r[0],color:s},{value:r[1],color:s})}},e.type="visualMap.piecewise",e.defaultOption=Qh(NK.defaultOption,{selected:null,minOpen:!1,maxOpen:!1,align:"auto",itemWidth:20,itemHeight:14,itemSymbol:"roundRect",pieces:null,categories:null,splitNumber:5,selectedMode:"multiple",itemGap:10,hoverLink:!0}),e}(NK),mQ={splitNumber:function(t){var e=this.option,n=Math.min(e.precision,20),i=this.getExtent(),r=e.splitNumber;r=Math.max(parseInt(r,10),1),e.splitNumber=r;var o=(i[1]-i[0])/r;while(+o.toFixed(n)!==o&&n<5)n++;e.precision=n,o=+o.toFixed(n),e.minOpen&&t.push({interval:[-1/0,i[0]],close:[0,0]});for(var a=0,s=i[0];a<r;s+=o,a++){var l=a===r-1?i[1]:s+o;t.push({interval:[s,l],close:[1,1]})}e.maxOpen&&t.push({interval:[i[1],1/0],close:[0,0]}),Qr(t),R(t,(function(t,e){t.index=e,t.text=this.formatValueText(t.interval)}),this)},categories:function(t){var e=this.option;R(e.categories,(function(e){t.push({text:this.formatValueText(e,!0),value:e})}),this),gQ(e,t)},pieces:function(t){var e=this.option;R(e.pieces,(function(e,n){J(e)||(e={value:e});var i={text:"",index:n};if(null!=e.label&&(i.text=e.label),e.hasOwnProperty("value")){var r=i.value=e.value;i.interval=[r,r],i.close=[1,1]}else{for(var o=i.interval=[],a=i.close=[0,0],s=[1,0,1],l=[-1/0,1/0],u=[],c=0;c<2;c++){for(var d=[["gte","gt","min"],["lte","lt","max"]][c],h=0;h<3&&null==o[c];h++)o[c]=e[d[h]],a[c]=s[h],u[c]=2===h;null==o[c]&&(o[c]=l[c])}u[0]&&o[1]===1/0&&(a[0]=0),u[1]&&o[0]===-1/0&&(a[1]=0),o[0]===o[1]&&a[0]&&a[1]&&(i.value=o[0])}i.visual=hP.retrieveVisuals(e),t.push(i)}),this),gQ(e,t),Qr(t),R(t,(function(t){var e=t.close,n=[["<","≤"][e[1]],[">","≥"][e[0]]];t.text=t.text||this.formatValueText(null!=t.value?t.value:t.interval,!1,n)}),this)}};function gQ(t,e){var n=t.inverse;("vertical"===t.orient?!n:n)&&e.reverse()}var yQ=pQ,vQ=function(t){function e(){var n=null!==t&&t.apply(this,arguments)||this;return n.type=e.type,n}return o(e,t),e.prototype.doRender=function(){var t=this.group;t.removeAll();var e=this.visualMapModel,n=e.get("textGap"),i=e.textStyleModel,r=i.getFont(),o=i.getTextColor(),a=this._getItemAlign(),s=e.itemSize,l=this._getViewData(),u=l.endsText,c=ot(e.get("showLabel",!0),!u);u&&this._renderEndsText(t,u[0],s,c,a),R(l.viewPieceList,(function(i){var l=i.piece,u=new Mr;u.onclick=V(this._onItemClick,this,l),this._enableHoverLink(u,i.indexInModelPieceList);var d=e.getRepresentValue(l);if(this._createItemSymbol(u,d,[0,0,s[0],s[1]]),c){var h=this.visualMapModel.getValueState(d);u.add(new Gl({style:{x:"right"===a?-n:s[0]+n,y:s[1]/2,text:l.text,verticalAlign:"middle",align:a,font:r,fill:o,opacity:"outOfRange"===h?.5:1}}))}t.add(u)}),this),u&&this._renderEndsText(t,u[1],s,c,a),up(e.get("orient"),t,e.get("itemGap")),this.renderBackground(t),this.positionGroup(t)},e.prototype._enableHoverLink=function(t,e){var n=this;t.on("mouseover",(function(){return i("highlight")})).on("mouseout",(function(){return i("downplay")}));var i=function(t){var i=n.visualMapModel;i.option.hoverLink&&n.api.dispatchAction({type:t,batch:GK(i.findTargetDataIndices(e),i)})}},e.prototype._getItemAlign=function(){var t=this.visualMapModel,e=t.option;if("vertical"===e.orient)return VK(t,this.api,t.itemSize);var n=e.align;return n&&"auto"!==n||(n="left"),n},e.prototype._renderEndsText=function(t,e,n,i,r){if(e){var o=new Mr,a=this.visualMapModel.textStyleModel;o.add(new Gl({style:{x:i?"right"===r?n[0]:0:n[0]/2,y:n[1]/2,verticalAlign:"middle",align:i?r:"center",text:e,font:a.getFont(),fill:a.getTextColor()}})),t.add(o)}},e.prototype._getViewData=function(){var t=this.visualMapModel,e=H(t.getPieceList(),(function(t,e){return{piece:t,indexInModelPieceList:e}})),n=t.get("text"),i=t.get("orient"),r=t.get("inverse");return("horizontal"===i?r:!r)?e.reverse():n&&(n=n.slice().reverse()),{viewPieceList:e,endsText:n}},e.prototype._createItemSymbol=function(t,e,n){t.add(j_(this.getControllerVisual(e,"symbol"),n[0],n[1],n[2],n[3],this.getControllerVisual(e,"color")))},e.prototype._onItemClick=function(t){var e=this.visualMapModel,n=e.option,i=T(n.selected),r=e.getSelectedMapKey(t);"single"===n.selectedMode?(i[r]=!0,R(i,(function(t,e){i[e]=e===r}))):i[r]=!i[r],this.api.dispatchAction({type:"selectDataRange",from:this.uid,visualMapId:this.visualMapModel.id,selected:i})},e.type="visualMap.piecewise",e}(BK),_Q=vQ;function bQ(t){t.registerComponentModel(yQ),t.registerComponentView(_Q),hQ(t)}function xQ(t){vM(fQ),vM(bQ)}var wQ={label:{enabled:!0},decal:{show:!1}},MQ=Io(),SQ={};function LQ(t,e){var n=t.getModel("aria");if(n.get("enabled")){var i=T(wQ);D(i.label,t.getLocaleModel().get("aria"),!1),D(n.option,i,!1),r(),o()}function r(){var e=n.getModel("decal"),i=e.get("show");if(i){var r=gt();t.eachSeries((function(t){if(t.useColorPaletteOnData){var e=r.get(t.type);e||(e={},r.set(t.type,e)),MQ(t).scope=e}})),t.eachRawSeries((function(e){if(!t.isSeriesFiltered(e))if("function"!==typeof e.enableAriaDecal){var n=e.getData();if(e.useColorPaletteOnData){var i=e.getRawData(),r={},o=MQ(e).scope;n.each((function(t){var e=n.getRawIndex(t);r[e]=t}));var a=i.count();i.each((function(t){var s=r[t],l=i.getName(t)||t+"",c=Xp(e.ecModel,l,o,a),d=n.getItemVisual(s,"decal");n.setItemVisual(s,"decal",u(d,c))}))}else{var s=Xp(e.ecModel,e.name,SQ,t.getSeriesCount()),l=n.getVisual("decal");n.setVisual("decal",u(l,s))}}else e.enableAriaDecal();function u(t,e){var n=t?I(I({},e),t):e;return n.dirty=!0,n}}))}}function o(){var i=t.getLocaleModel().get("aria"),r=n.getModel("label");if(r.option=A(r.option,i),r.get("enabled")){var o=e.getZr().dom;if(r.get("description"))o.setAttribute("aria-label",r.get("description"));else{var u,c=t.getSeriesCount(),d=r.get(["data","maxCount"])||10,h=r.get(["series","maxCount"])||10,f=Math.min(c,h);if(!(c<1)){var p=s();if(p){var m=r.get(["general","withTitle"]);u=a(m,{title:p})}else u=r.get(["general","withoutTitle"]);var g=[],y=c>1?r.get(["series","multiple","prefix"]):r.get(["series","single","prefix"]);u+=a(y,{seriesCount:c}),t.eachSeries((function(t,e){if(e<f){var n=void 0,i=t.get("name"),o=i?"withName":"withoutName";n=c>1?r.get(["series","multiple",o]):r.get(["series","single",o]),n=a(n,{seriesId:t.seriesIndex,seriesName:t.get("name"),seriesType:l(t.subType)});var s=t.getData();if(s.count()>d){var u=r.get(["data","partialData"]);n+=a(u,{displayCnt:d})}else n+=r.get(["data","allData"]);for(var h=[],p=0;p<s.count();p++)if(p<d){var m=s.getName(p),y=Lg(s,p),v=r.get(["data",m?"withName":"withoutName"]);h.push(a(v,{name:m,value:y}))}var _=r.get(["data","separator","middle"]),b=r.get(["data","separator","end"]);n+=h.join(_)+b,g.push(n)}}));var v=r.getModel(["series","multiple","separator"]),_=v.get("middle"),b=v.get("end");u+=g.join(_)+b,o.setAttribute("aria-label",u)}}}}function a(t,e){if("string"!==typeof t)return t;var n=t;return R(e,(function(t,e){n=n.replace(new RegExp("\\{\\s*"+e+"\\s*\\}","g"),t)})),n}function s(){var e=t.get("title");return e&&e.length&&(e=e[0]),e&&e.text}function l(e){return t.getLocaleModel().get(["series","typeNames"])[e]||"自定义图"}}function kQ(t){if(t&&t.aria){var e=t.aria;null!=e.show&&(e.enabled=e.show),e.label=e.label||{},R(["description","general","series","data"],(function(t){null!=e[t]&&(e.label[t]=e[t])}))}}function TQ(t){t.registerPreprocessor(kQ),t.registerVisual(t.PRIORITY.VISUAL.ARIA,LQ)}var DQ={value:"eq","<":"lt","<=":"lte",">":"gt",">=":"gte","=":"eq","!=":"ne","<>":"ne"},CQ=function(){function t(t){var e=this._condVal=q(t)?new RegExp(t):it(t)?t:null;if(null==e){var n="";0,im(n)}}return t.prototype.evaluate=function(t){var e=typeof t;return"string"===e?this._condVal.test(t):"number"===e&&this._condVal.test(t+"")},t}(),IQ=function(){function t(){}return t.prototype.evaluate=function(){return this.value},t}(),AQ=function(){function t(){}return t.prototype.evaluate=function(){for(var t=this.children,e=0;e<t.length;e++)if(!t[e].evaluate())return!1;return!0},t}(),YQ=function(){function t(){}return t.prototype.evaluate=function(){for(var t=this.children,e=0;e<t.length;e++)if(t[e].evaluate())return!0;return!1},t}(),OQ=function(){function t(){}return t.prototype.evaluate=function(){return!this.child.evaluate()},t}(),PQ=function(){function t(){}return t.prototype.evaluate=function(){for(var t=!!this.valueParser,e=this.getValue,n=e(this.valueGetterParam),i=t?this.valueParser(n):null,r=0;r<this.subCondList.length;r++)if(!this.subCondList[r].evaluate(t?i:n))return!1;return!0},t}();function EQ(t,e){if(!0===t||!1===t){var n=new IQ;return n.value=t,n}var i="";return jQ(t)||im(i),t.and?NQ("and",t,e):t.or?NQ("or",t,e):t.not?RQ(t,e):HQ(t,e)}function NQ(t,e,n){var i=e[t],r="";$(i)||im(r),i.length||im(r);var o="and"===t?new AQ:new YQ;return o.children=H(i,(function(t){return EQ(t,n)})),o.children.length||im(r),o}function RQ(t,e){var n=t.not,i="";jQ(n)||im(i);var r=new OQ;return r.child=EQ(n,e),r.child||im(i),r}function HQ(t,e){for(var n="",i=e.prepareGetValue(t),r=[],o=B(t),a=t.parser,s=a?Pg(a):null,l=0;l<o.length;l++){var u=o[l];if("parser"!==u&&!e.valueGetterAttrMap.get(u)){var c=_t(DQ,u)?DQ[u]:u,d=t[u],h=s?s(d):d,f=jg(c,h)||"reg"===c&&new CQ(h);f||im(n),r.push(f)}}r.length||im(n);var p=new PQ;return p.valueGetterParam=i,p.valueParser=s,p.getValue=e.getValue,p.subCondList=r,p}function jQ(t){return J(t)&&!N(t)}var zQ=function(){function t(t,e){this._cond=EQ(t,e)}return t.prototype.evaluate=function(){return this._cond.evaluate()},t}();function WQ(t,e){return new zQ(t,e)}var BQ={type:"echarts:filter",transform:function(t){for(var e,n=t.upstream,i=WQ(t.config,{valueGetterAttrMap:gt({dimension:!0}),prepareGetValue:function(t){var e="",i=t.dimension;_t(t,"dimension")||im(e);var r=n.getDimensionInfo(i);return r||im(e),{dimIdx:r.index}},getValue:function(t){return n.retrieveValueFromItem(e,t.dimIdx)}}),r=[],o=0,a=n.count();o<a;o++)e=n.getRawDataItem(o),i.evaluate()&&r.push(e);return{data:r}}};var FQ={type:"echarts:sort",transform:function(t){var e=t.upstream,n=t.config,i="",r=so(n);r.length||im(i);var o=[];R(r,(function(t){var n=t.dimension,r=t.order,a=t.parser,s=t.incomparable;if(null==n&&im(i),"asc"!==r&&"desc"!==r&&im(i),s&&"min"!==s&&"max"!==s){var l="";0,im(l)}if("asc"!==r&&"desc"!==r){var u="";0,im(u)}var c=e.getDimensionInfo(n);c||im(i);var d=a?Pg(a):null;a&&!d&&im(i),o.push({dimIdx:c.index,parser:d,comparator:new Rg(r,s)})}));var a=e.sourceFormat;a!==Tp&&a!==Dp&&im(i);for(var s=[],l=0,u=e.count();l<u;l++)s.push(e.getRawDataItem(l));return s.sort((function(t,n){for(var i=0;i<o.length;i++){var r=o[i],a=e.retrieveValueFromItem(t,r.dimIdx),s=e.retrieveValueFromItem(n,r.dimIdx);r.parser&&(a=r.parser(a),s=r.parser(s));var l=r.comparator.evaluate(a,s);if(0!==l)return l}return 0})),{data:s}}};function VQ(t){t.registerTransform(BQ),t.registerTransform(FQ)}var GQ=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="dataset",e}return o(e,t),e.prototype.init=function(e,n,i){t.prototype.init.call(this,e,n,i),this._sourceManager=new Jg(this),Kg(this)},e.prototype.mergeOption=function(e,n){t.prototype.mergeOption.call(this,e,n),Kg(this)},e.prototype.optionUpdated=function(){this._sourceManager.dirty()},e.prototype.getSourceManager=function(){return this._sourceManager},e.type="dataset",e.defaultOption={seriesLayoutBy:Yp},e}(xp),$Q=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.type="dataset",e}return o(e,t),e.type="dataset",e}(Py);function UQ(t){t.registerComponentModel(GQ),t.registerComponentView($Q)}vM([YM]),vM([BS]),vM([Bk,ET,eD,JI,_A,_Y,mO,VP,sN,vN,CN,PH,hj,Aj,oz,mz,Hz,Xz,LW,EW,tB,rV]),vM(nG),vM(UG),vM(vY),vM(v$),vM(OH),vM(I$),vM(V$),vM(Dq),vM(fX),vM(eG),vM(UX),vM(ZX),vM(_Z),vM(VZ),vM(nJ),vM(mJ),vM(VJ),vM(LK),vM(sK),vM(SK),vM(xQ),vM(fQ),vM(bQ),vM(TQ),vM(VQ),vM(UQ)},"31de":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("ul",{staticClass:"ivu-dropdown-menu"},[t._t("default")],2)},r=[],o={name:"DropdownMenu"},a=o,s=n("2877"),l=Object(s["a"])(a,i,r,!1,null,null,null);e["a"]=l.exports},"32b3":function(t,e,n){var i=n("872a"),r=n("9638"),o=Object.prototype,a=o.hasOwnProperty;function s(t,e,n){var o=t[e];a.call(t,e)&&r(o,n)&&(void 0!==n||e in t)||i(t,e,n)}t.exports=s},"32f4":function(t,e,n){var i=n("2d7c"),r=n("d327"),o=Object.prototype,a=o.propertyIsEnumerable,s=Object.getOwnPropertySymbols,l=s?function(t){return null==t?[]:(t=Object(t),i(s(t),(function(e){return a.call(t,e)})))}:r;t.exports=l},"342f":function(t,e,n){var i=n("d066");t.exports=i("navigator","userAgent")||""},"34ac":function(t,e,n){var i=n("9520"),r=n("1368"),o=n("1a8c"),a=n("dc57"),s=/[\\^$.*+?()[\]{}|]/g,l=/^\[object .+?Constructor\]$/,u=Function.prototype,c=Object.prototype,d=u.toString,h=c.hasOwnProperty,f=RegExp("^"+d.call(h).replace(s,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$");function p(t){if(!o(t)||r(t))return!1;var e=i(t)?f:l;return e.test(a(t))}t.exports=p},"35a1":function(t,e,n){var i=n("f5df"),r=n("3f8c"),o=n("b622"),a=o("iterator");t.exports=function(t){if(void 0!=t)return t[a]||t["@@iterator"]||r[i(t)]}},3698:function(t,e){function n(t,e){return null==t?void 0:t[e]}t.exports=n},3729:function(t,e,n){var i=n("9e69"),r=n("00fd"),o=n("29f3"),a="[object Null]",s="[object Undefined]",l=i?i.toStringTag:void 0;function u(t){return null==t?void 0===t?s:a:l&&l in Object(t)?r(t):o(t)}t.exports=u},"37e8":function(t,e,n){var i=n("83ab"),r=n("9bf2"),o=n("825a"),a=n("df75");t.exports=i?Object.defineProperties:function(t,e){o(t);var n,i=a(e),s=i.length,l=0;while(s>l)r.f(t,n=i[l++],e[n]);return t}},3818:function(t,e,n){var i=n("7e64"),r=n("8057"),o=n("32b3"),a=n("5b01"),s=n("0f0f"),l=n("e5383"),u=n("4359"),c=n("54eb"),d=n("1041"),h=n("a994"),f=n("1bac"),p=n("42a2"),m=n("c87c"),g=n("c2b6"),y=n("fa21"),v=n("6747"),_=n("0d24"),b=n("cc45"),x=n("1a8c"),w=n("d7ee"),M=n("ec69"),S=n("9934"),L=1,k=2,T=4,D="[object Arguments]",C="[object Array]",I="[object Boolean]",A="[object Date]",Y="[object Error]",O="[object Function]",P="[object GeneratorFunction]",E="[object Map]",N="[object Number]",R="[object Object]",H="[object RegExp]",j="[object Set]",z="[object String]",W="[object Symbol]",B="[object WeakMap]",F="[object ArrayBuffer]",V="[object DataView]",G="[object Float32Array]",$="[object Float64Array]",U="[object Int8Array]",q="[object Int16Array]",X="[object Int32Array]",Z="[object Uint8Array]",J="[object Uint8ClampedArray]",K="[object Uint16Array]",Q="[object Uint32Array]",tt={};function et(t,e,n,C,I,A){var Y,E=e&L,N=e&k,H=e&T;if(n&&(Y=I?n(t,C,I,A):n(t)),void 0!==Y)return Y;if(!x(t))return t;var j=v(t);if(j){if(Y=m(t),!E)return u(t,Y)}else{var z=p(t),W=z==O||z==P;if(_(t))return l(t,E);if(z==R||z==D||W&&!I){if(Y=N||W?{}:y(t),!E)return N?d(t,s(Y,t)):c(t,a(Y,t))}else{if(!tt[z])return I?t:{};Y=g(t,z,E)}}A||(A=new i);var B=A.get(t);if(B)return B;A.set(t,Y),w(t)?t.forEach((function(i){Y.add(et(i,e,n,i,t,A))})):b(t)&&t.forEach((function(i,r){Y.set(r,et(i,e,n,r,t,A))}));var F=H?N?f:h:N?S:M,V=j?void 0:F(t);return r(V||t,(function(i,r){V&&(r=i,i=t[r]),o(Y,r,et(i,e,n,r,t,A))})),Y}tt[D]=tt[C]=tt[F]=tt[V]=tt[I]=tt[A]=tt[G]=tt[$]=tt[U]=tt[q]=tt[X]=tt[E]=tt[N]=tt[R]=tt[H]=tt[j]=tt[z]=tt[W]=tt[Z]=tt[J]=tt[K]=tt[Q]=!0,tt[Y]=tt[O]=tt[B]=!1,t.exports=et},3852:function(t,e,n){var i=n("96f3"),r=n("e2c0");function o(t,e){return null!=t&&r(t,e,i)}t.exports=o},"387f":function(t,e,n){"use strict";t.exports=function(t,e,n,i,r){return t.config=e,n&&(t.code=n),t.request=i,t.response=r,t.isAxiosError=!0,t.toJSON=function(){return{message:this.message,name:this.name,description:this.description,number:this.number,fileName:this.fileName,lineNumber:this.lineNumber,columnNumber:this.columnNumber,stack:this.stack,config:this.config,code:this.code}},t}},3886:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-ca",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"YYYY-MM-DD",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY h:mm A",LLLL:"dddd, MMMM D, YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n}});return e}))},3934:function(t,e,n){"use strict";var i=n("c532");t.exports=i.isStandardBrowserEnv()?function(){var t,e=/(msie|trident)/i.test(navigator.userAgent),n=document.createElement("a");function r(t){var i=t;return e&&(n.setAttribute("href",i),i=n.href),n.setAttribute("href",i),{href:n.href,protocol:n.protocol?n.protocol.replace(/:$/,""):"",host:n.host,search:n.search?n.search.replace(/^\?/,""):"",hash:n.hash?n.hash.replace(/^#/,""):"",hostname:n.hostname,port:n.port,pathname:"/"===n.pathname.charAt(0)?n.pathname:"/"+n.pathname}}return t=r(window.location.href),function(e){var n=i.isString(e)?r(e):e;return n.protocol===t.protocol&&n.host===t.host}}():function(){return function(){return!0}}()},"39a6":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-gb",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},"39bd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"१",2:"२",3:"३",4:"४",5:"५",6:"६",7:"७",8:"८",9:"९",0:"०"},n={"१":"1","२":"2","३":"3","४":"4","५":"5","६":"6","७":"7","८":"8","९":"9","०":"0"};function i(t,e,n,i){var r="";if(e)switch(n){case"s":r="काही सेकंद";break;case"ss":r="%d सेकंद";break;case"m":r="एक मिनिट";break;case"mm":r="%d मिनिटे";break;case"h":r="एक तास";break;case"hh":r="%d तास";break;case"d":r="एक दिवस";break;case"dd":r="%d दिवस";break;case"M":r="एक महिना";break;case"MM":r="%d महिने";break;case"y":r="एक वर्ष";break;case"yy":r="%d वर्षे";break}else switch(n){case"s":r="काही सेकंदां";break;case"ss":r="%d सेकंदां";break;case"m":r="एका मिनिटा";break;case"mm":r="%d मिनिटां";break;case"h":r="एका तासा";break;case"hh":r="%d तासां";break;case"d":r="एका दिवसा";break;case"dd":r="%d दिवसां";break;case"M":r="एका महिन्या";break;case"MM":r="%d महिन्यां";break;case"y":r="एका वर्षा";break;case"yy":r="%d वर्षां";break}return r.replace(/%d/i,t)}var r=t.defineLocale("mr",{months:"जानेवारी_फेब्रुवारी_मार्च_एप्रिल_मे_जून_जुलै_ऑगस्ट_सप्टेंबर_ऑक्टोबर_नोव्हेंबर_डिसेंबर".split("_"),monthsShort:"जाने._फेब्रु._मार्च._एप्रि._मे._जून._जुलै._ऑग._सप्टें._ऑक्टो._नोव्हें._डिसें.".split("_"),monthsParseExact:!0,weekdays:"रविवार_सोमवार_मंगळवार_बुधवार_गुरूवार_शुक्रवार_शनिवार".split("_"),weekdaysShort:"रवि_सोम_मंगळ_बुध_गुरू_शुक्र_शनि".split("_"),weekdaysMin:"र_सो_मं_बु_गु_शु_श".split("_"),longDateFormat:{LT:"A h:mm वाजता",LTS:"A h:mm:ss वाजता",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm वाजता",LLLL:"dddd, D MMMM YYYY, A h:mm वाजता"},calendar:{sameDay:"[आज] LT",nextDay:"[उद्या] LT",nextWeek:"dddd, LT",lastDay:"[काल] LT",lastWeek:"[मागील] dddd, LT",sameElse:"L"},relativeTime:{future:"%sमध्ये",past:"%sपूर्वी",s:i,ss:i,m:i,mm:i,h:i,hh:i,d:i,dd:i,M:i,MM:i,y:i,yy:i},preparse:function(t){return t.replace(/[१२३४५६७८९०]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/पहाटे|सकाळी|दुपारी|सायंकाळी|रात्री/,meridiemHour:function(t,e){return 12===t&&(t=0),"पहाटे"===e||"सकाळी"===e?t:"दुपारी"===e||"सायंकाळी"===e||"रात्री"===e?t>=12?t:t+12:void 0},meridiem:function(t,e,n){return t>=0&&t<6?"पहाटे":t<12?"सकाळी":t<17?"दुपारी":t<20?"सायंकाळी":"रात्री"},week:{dow:0,doy:6}});return r}))},"39ff":function(t,e,n){var i=n("0b07"),r=n("2b3e"),o=i(r,"WeakMap");t.exports=o},"3a0e":function(t,e,n){var i=n("86e1"),r=n("100e"),o=r(i);t.exports=o},"3a27":function(t,e,n){"use strict";n.d(e,"i",(function(){return o})),n.d(e,"g",(function(){return s})),n.d(e,"a",(function(){return l})),n.d(e,"h",(function(){return h})),n.d(e,"e",(function(){return f})),n.d(e,"d",(function(){return p})),n.d(e,"f",(function(){return m})),n.d(e,"c",(function(){return g})),n.d(e,"b",(function(){return y})),n.d(e,"j",(function(){return v}));var i=n("2b0e");const r=i["a"].prototype.$isServer;function o(t,e){for(let n=0;n<e.length;n++)if(t===e[n])return!0;return!1}let a;function s(t){if(r)return 0;if(t||void 0===a){const t=document.createElement("div");t.style.width="100%",t.style.height="200px";const e=document.createElement("div"),n=e.style;n.position="absolute",n.top=0,n.left=0,n.pointerEvents="none",n.visibility="hidden",n.width="200px",n.height="150px",n.overflow="hidden",e.appendChild(t),document.body.appendChild(e);const i=t.offsetWidth;e.style.overflow="scroll";let r=t.offsetWidth;i===r&&(r=e.clientWidth),document.body.removeChild(e),a=i-r}return a}const l=!r&&(window.MutationObserver||window.WebKitMutationObserver||window.MozMutationObserver||!1),u=/([\:\-\_]+(.))/g,c=/^moz([A-Z])/;function d(t){return t.replace(u,(function(t,e,n,i){return i?n.toUpperCase():n})).replace(c,"Moz$1")}function h(t,e){if(!t||!e)return null;e=d(e),"float"===e&&(e="cssFloat");try{const n=document.defaultView.getComputedStyle(t,"");return t.style[e]||n?n[e]:null}catch(n){return t.style[e]}}function f(t,e,n){n="string"===typeof e?[e]:e;let i=t.$parent,r=i.$options.name;while(i&&(!r||n.indexOf(r)<0))i=i.$parent,i&&(r=i.$options.name);return i}function p(t,e){const n=t.$children;let i=null;if(n.length)for(const r of n){const t=r.$options.name;if(t===e){i=r;break}if(i=p(r,e),i)break}return i}function m(t,e){return t.$children.reduce((t,n)=>{n.$options.name===e&&t.push(n);const i=m(n,e);return t.concat(i)},[])}function g(t,e,n=!0){let i=t.$parent.$children.filter(t=>t.$options.name===e),r=i.findIndex(e=>e._uid===t._uid);return n&&i.splice(r,1),i}const y={xs:"480px",sm:"576px",md:"768px",lg:"992px",xl:"1200px",xxl:"1600px"};function v(){if("undefined"!==typeof window){const t=t=>({media:t,matches:!1,on(){},off(){}});window.matchMedia=window.matchMedia||t}}},"3a39":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"१",2:"२",3:"३",4:"४",5:"५",6:"६",7:"७",8:"८",9:"९",0:"०"},n={"१":"1","२":"2","३":"3","४":"4","५":"5","६":"6","७":"7","८":"8","९":"9","०":"0"},i=t.defineLocale("ne",{months:"जनवरी_फेब्रुवरी_मार्च_अप्रिल_मई_जुन_जुलाई_अगष्ट_सेप्टेम्बर_अक्टोबर_नोभेम्बर_डिसेम्बर".split("_"),monthsShort:"जन._फेब्रु._मार्च_अप्रि._मई_जुन_जुलाई._अग._सेप्ट._अक्टो._नोभे._डिसे.".split("_"),monthsParseExact:!0,weekdays:"आइतबार_सोमबार_मङ्गलबार_बुधबार_बिहिबार_शुक्रबार_शनिबार".split("_"),weekdaysShort:"आइत._सोम._मङ्गल._बुध._बिहि._शुक्र._शनि.".split("_"),weekdaysMin:"आ._सो._मं._बु._बि._शु._श.".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"Aको h:mm बजे",LTS:"Aको h:mm:ss बजे",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, Aको h:mm बजे",LLLL:"dddd, D MMMM YYYY, Aको h:mm बजे"},preparse:function(t){return t.replace(/[१२३४५६७८९०]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/राति|बिहान|दिउँसो|साँझ/,meridiemHour:function(t,e){return 12===t&&(t=0),"राति"===e?t<4?t:t+12:"बिहान"===e?t:"दिउँसो"===e?t>=10?t:t+12:"साँझ"===e?t+12:void 0},meridiem:function(t,e,n){return t<3?"राति":t<12?"बिहान":t<16?"दिउँसो":t<20?"साँझ":"राति"},calendar:{sameDay:"[आज] LT",nextDay:"[भोलि] LT",nextWeek:"[आउँदो] dddd[,] LT",lastDay:"[हिजो] LT",lastWeek:"[गएको] dddd[,] LT",sameElse:"L"},relativeTime:{future:"%sमा",past:"%s अगाडि",s:"केही क्षण",ss:"%d सेकेण्ड",m:"एक मिनेट",mm:"%d मिनेट",h:"एक घण्टा",hh:"%d घण्टा",d:"एक दिन",dd:"%d दिन",M:"एक महिना",MM:"%d महिना",y:"एक बर्ष",yy:"%d बर्ष"},week:{dow:0,doy:6}});return i}))},"3a6c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-mo",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"週日_週一_週二_週三_週四_週五_週六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日dddd HH:mm",l:"D/M/YYYY",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"中午"===e?t>=11?t:t+12:"下午"===e||"晚上"===e?t+12:void 0},meridiem:function(t,e,n){var i=100*t+e;return i<600?"凌晨":i<900?"早上":i<1130?"上午":i<1230?"中午":i<1800?"下午":"晚上"},calendar:{sameDay:"[今天] LT",nextDay:"[明天] LT",nextWeek:"[下]dddd LT",lastDay:"[昨天] LT",lastWeek:"[上]dddd LT",sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|週)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"週";default:return t}},relativeTime:{future:"%s內",past:"%s前",s:"幾秒",ss:"%d 秒",m:"1 分鐘",mm:"%d 分鐘",h:"1 小時",hh:"%d 小時",d:"1 天",dd:"%d 天",M:"1 個月",MM:"%d 個月",y:"1 年",yy:"%d 年"}});return e}))},"3b1b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={0:"-ум",1:"-ум",2:"-юм",3:"-юм",4:"-ум",5:"-ум",6:"-ум",7:"-ум",8:"-ум",9:"-ум",10:"-ум",12:"-ум",13:"-ум",20:"-ум",30:"-юм",40:"-ум",50:"-ум",60:"-ум",70:"-ум",80:"-ум",90:"-ум",100:"-ум"},n=t.defineLocale("tg",{months:{format:"январи_феврали_марти_апрели_майи_июни_июли_августи_сентябри_октябри_ноябри_декабри".split("_"),standalone:"январ_феврал_март_апрел_май_июн_июл_август_сентябр_октябр_ноябр_декабр".split("_")},monthsShort:"янв_фев_мар_апр_май_июн_июл_авг_сен_окт_ноя_дек".split("_"),weekdays:"якшанбе_душанбе_сешанбе_чоршанбе_панҷшанбе_ҷумъа_шанбе".split("_"),weekdaysShort:"яшб_дшб_сшб_чшб_пшб_ҷум_шнб".split("_"),weekdaysMin:"яш_дш_сш_чш_пш_ҷм_шб".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Имрӯз соати] LT",nextDay:"[Фардо соати] LT",lastDay:"[Дирӯз соати] LT",nextWeek:"dddd[и] [ҳафтаи оянда соати] LT",lastWeek:"dddd[и] [ҳафтаи гузашта соати] LT",sameElse:"L"},relativeTime:{future:"баъди %s",past:"%s пеш",s:"якчанд сония",m:"як дақиқа",mm:"%d дақиқа",h:"як соат",hh:"%d соат",d:"як рӯз",dd:"%d рӯз",M:"як моҳ",MM:"%d моҳ",y:"як сол",yy:"%d сол"},meridiemParse:/шаб|субҳ|рӯз|бегоҳ/,meridiemHour:function(t,e){return 12===t&&(t=0),"шаб"===e?t<4?t:t+12:"субҳ"===e?t:"рӯз"===e?t>=11?t:t+12:"бегоҳ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"шаб":t<11?"субҳ":t<16?"рӯз":t<19?"бегоҳ":"шаб"},dayOfMonthOrdinalParse:/\d{1,2}-(ум|юм)/,ordinal:function(t){var n=t%10,i=t>=100?100:null;return t+(e[t]||e[n]||e[i])},week:{dow:1,doy:7}});return n}))},"3b4a":function(t,e,n){var i=n("0b07"),r=function(){try{var t=i(Object,"defineProperty");return t({},"",{}),t}catch(e){}}();t.exports=r},"3bb4":function(t,e,n){var i=n("08cc"),r=n("ec69");function o(t){var e=r(t),n=e.length;while(n--){var o=e[n],a=t[o];e[n]=[o,a,i(a)]}return e}t.exports=o},"3bbe":function(t,e,n){var i=n("861d");t.exports=function(t){if(!i(t)&&null!==t)throw TypeError("Can't set "+String(t)+" as a prototype");return t}},"3c0d":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="leden_únor_březen_duben_květen_červen_červenec_srpen_září_říjen_listopad_prosinec".split("_"),n="led_úno_bře_dub_kvě_čvn_čvc_srp_zář_říj_lis_pro".split("_"),i=[/^led/i,/^úno/i,/^bře/i,/^dub/i,/^kvě/i,/^(čvn|červen$|června)/i,/^(čvc|červenec|července)/i,/^srp/i,/^zář/i,/^říj/i,/^lis/i,/^pro/i],r=/^(leden|únor|březen|duben|květen|červenec|července|červen|června|srpen|září|říjen|listopad|prosinec|led|úno|bře|dub|kvě|čvn|čvc|srp|zář|říj|lis|pro)/i;function o(t){return t>1&&t<5&&1!==~~(t/10)}function a(t,e,n,i){var r=t+" ";switch(n){case"s":return e||i?"pár sekund":"pár sekundami";case"ss":return e||i?r+(o(t)?"sekundy":"sekund"):r+"sekundami";case"m":return e?"minuta":i?"minutu":"minutou";case"mm":return e||i?r+(o(t)?"minuty":"minut"):r+"minutami";case"h":return e?"hodina":i?"hodinu":"hodinou";case"hh":return e||i?r+(o(t)?"hodiny":"hodin"):r+"hodinami";case"d":return e||i?"den":"dnem";case"dd":return e||i?r+(o(t)?"dny":"dní"):r+"dny";case"M":return e||i?"měsíc":"měsícem";case"MM":return e||i?r+(o(t)?"měsíce":"měsíců"):r+"měsíci";case"y":return e||i?"rok":"rokem";case"yy":return e||i?r+(o(t)?"roky":"let"):r+"lety"}}var s=t.defineLocale("cs",{months:e,monthsShort:n,monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:/^(leden|ledna|února|únor|březen|března|duben|dubna|květen|května|červenec|července|červen|června|srpen|srpna|září|říjen|října|listopadu|listopad|prosinec|prosince)/i,monthsShortStrictRegex:/^(led|úno|bře|dub|kvě|čvn|čvc|srp|zář|říj|lis|pro)/i,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"neděle_pondělí_úterý_středa_čtvrtek_pátek_sobota".split("_"),weekdaysShort:"ne_po_út_st_čt_pá_so".split("_"),weekdaysMin:"ne_po_út_st_čt_pá_so".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd D. MMMM YYYY H:mm",l:"D. M. YYYY"},calendar:{sameDay:"[dnes v] LT",nextDay:"[zítra v] LT",nextWeek:function(){switch(this.day()){case 0:return"[v neděli v] LT";case 1:case 2:return"[v] dddd [v] LT";case 3:return"[ve středu v] LT";case 4:return"[ve čtvrtek v] LT";case 5:return"[v pátek v] LT";case 6:return"[v sobotu v] LT"}},lastDay:"[včera v] LT",lastWeek:function(){switch(this.day()){case 0:return"[minulou neděli v] LT";case 1:case 2:return"[minulé] dddd [v] LT";case 3:return"[minulou středu v] LT";case 4:case 5:return"[minulý] dddd [v] LT";case 6:return"[minulou sobotu v] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"před %s",s:a,ss:a,m:a,mm:a,h:a,hh:a,d:a,dd:a,M:a,MM:a,y:a,yy:a},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return s}))},"3ca3":function(t,e,n){"use strict";var i=n("6547").charAt,r=n("69f3"),o=n("7dd0"),a="String Iterator",s=r.set,l=r.getterFor(a);o(String,"String",(function(t){s(this,{type:a,string:String(t),index:0})}),(function(){var t,e=l(this),n=e.string,r=e.index;return r>=n.length?{value:void 0,done:!0}:(t=i(n,r),e.index+=t.length,{value:t,done:!1})}))},"3cfe":function(t,e,n){var i=n("4bb5");function r(t,e){return null==t||i(t,e)}t.exports=r},"3de5":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"௧",2:"௨",3:"௩",4:"௪",5:"௫",6:"௬",7:"௭",8:"௮",9:"௯",0:"௦"},n={"௧":"1","௨":"2","௩":"3","௪":"4","௫":"5","௬":"6","௭":"7","௮":"8","௯":"9","௦":"0"},i=t.defineLocale("ta",{months:"ஜனவரி_பிப்ரவரி_மார்ச்_ஏப்ரல்_மே_ஜூன்_ஜூலை_ஆகஸ்ட்_செப்டெம்பர்_அக்டோபர்_நவம்பர்_டிசம்பர்".split("_"),monthsShort:"ஜனவரி_பிப்ரவரி_மார்ச்_ஏப்ரல்_மே_ஜூன்_ஜூலை_ஆகஸ்ட்_செப்டெம்பர்_அக்டோபர்_நவம்பர்_டிசம்பர்".split("_"),weekdays:"ஞாயிற்றுக்கிழமை_திங்கட்கிழமை_செவ்வாய்கிழமை_புதன்கிழமை_வியாழக்கிழமை_வெள்ளிக்கிழமை_சனிக்கிழமை".split("_"),weekdaysShort:"ஞாயிறு_திங்கள்_செவ்வாய்_புதன்_வியாழன்_வெள்ளி_சனி".split("_"),weekdaysMin:"ஞா_தி_செ_பு_வி_வெ_ச".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, HH:mm",LLLL:"dddd, D MMMM YYYY, HH:mm"},calendar:{sameDay:"[இன்று] LT",nextDay:"[நாளை] LT",nextWeek:"dddd, LT",lastDay:"[நேற்று] LT",lastWeek:"[கடந்த வாரம்] dddd, LT",sameElse:"L"},relativeTime:{future:"%s இல்",past:"%s முன்",s:"ஒரு சில விநாடிகள்",ss:"%d விநாடிகள்",m:"ஒரு நிமிடம்",mm:"%d நிமிடங்கள்",h:"ஒரு மணி நேரம்",hh:"%d மணி நேரம்",d:"ஒரு நாள்",dd:"%d நாட்கள்",M:"ஒரு மாதம்",MM:"%d மாதங்கள்",y:"ஒரு வருடம்",yy:"%d ஆண்டுகள்"},dayOfMonthOrdinalParse:/\d{1,2}வது/,ordinal:function(t){return t+"வது"},preparse:function(t){return t.replace(/[௧௨௩௪௫௬௭௮௯௦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/யாமம்|வைகறை|காலை|நண்பகல்|எற்பாடு|மாலை/,meridiem:function(t,e,n){return t<2?" யாமம்":t<6?" வைகறை":t<10?" காலை":t<14?" நண்பகல்":t<18?" எற்பாடு":t<22?" மாலை":" யாமம்"},meridiemHour:function(t,e){return 12===t&&(t=0),"யாமம்"===e?t<2?t:t+12:"வைகறை"===e||"காலை"===e||"நண்பகல்"===e&&t>=10?t:t+12},week:{dow:0,doy:6}});return i}))},"3e92":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"೧",2:"೨",3:"೩",4:"೪",5:"೫",6:"೬",7:"೭",8:"೮",9:"೯",0:"೦"},n={"೧":"1","೨":"2","೩":"3","೪":"4","೫":"5","೬":"6","೭":"7","೮":"8","೯":"9","೦":"0"},i=t.defineLocale("kn",{months:"ಜನವರಿ_ಫೆಬ್ರವರಿ_ಮಾರ್ಚ್_ಏಪ್ರಿಲ್_ಮೇ_ಜೂನ್_ಜುಲೈ_ಆಗಸ್ಟ್_ಸೆಪ್ಟೆಂಬರ್_ಅಕ್ಟೋಬರ್_ನವೆಂಬರ್_ಡಿಸೆಂಬರ್".split("_"),monthsShort:"ಜನ_ಫೆಬ್ರ_ಮಾರ್ಚ್_ಏಪ್ರಿಲ್_ಮೇ_ಜೂನ್_ಜುಲೈ_ಆಗಸ್ಟ್_ಸೆಪ್ಟೆಂ_ಅಕ್ಟೋ_ನವೆಂ_ಡಿಸೆಂ".split("_"),monthsParseExact:!0,weekdays:"ಭಾನುವಾರ_ಸೋಮವಾರ_ಮಂಗಳವಾರ_ಬುಧವಾರ_ಗುರುವಾರ_ಶುಕ್ರವಾರ_ಶನಿವಾರ".split("_"),weekdaysShort:"ಭಾನು_ಸೋಮ_ಮಂಗಳ_ಬುಧ_ಗುರು_ಶುಕ್ರ_ಶನಿ".split("_"),weekdaysMin:"ಭಾ_ಸೋ_ಮಂ_ಬು_ಗು_ಶು_ಶ".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm",LLLL:"dddd, D MMMM YYYY, A h:mm"},calendar:{sameDay:"[ಇಂದು] LT",nextDay:"[ನಾಳೆ] LT",nextWeek:"dddd, LT",lastDay:"[ನಿನ್ನೆ] LT",lastWeek:"[ಕೊನೆಯ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s ನಂತರ",past:"%s ಹಿಂದೆ",s:"ಕೆಲವು ಕ್ಷಣಗಳು",ss:"%d ಸೆಕೆಂಡುಗಳು",m:"ಒಂದು ನಿಮಿಷ",mm:"%d ನಿಮಿಷ",h:"ಒಂದು ಗಂಟೆ",hh:"%d ಗಂಟೆ",d:"ಒಂದು ದಿನ",dd:"%d ದಿನ",M:"ಒಂದು ತಿಂಗಳು",MM:"%d ತಿಂಗಳು",y:"ಒಂದು ವರ್ಷ",yy:"%d ವರ್ಷ"},preparse:function(t){return t.replace(/[೧೨೩೪೫೬೭೮೯೦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/ರಾತ್ರಿ|ಬೆಳಿಗ್ಗೆ|ಮಧ್ಯಾಹ್ನ|ಸಂಜೆ/,meridiemHour:function(t,e){return 12===t&&(t=0),"ರಾತ್ರಿ"===e?t<4?t:t+12:"ಬೆಳಿಗ್ಗೆ"===e?t:"ಮಧ್ಯಾಹ್ನ"===e?t>=10?t:t+12:"ಸಂಜೆ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"ರಾತ್ರಿ":t<10?"ಬೆಳಿಗ್ಗೆ":t<17?"ಮಧ್ಯಾಹ್ನ":t<20?"ಸಂಜೆ":"ರಾತ್ರಿ"},dayOfMonthOrdinalParse:/\d{1,2}(ನೇ)/,ordinal:function(t){return t+"ನೇ"},week:{dow:0,doy:6}});return i}))},"3eea":function(t,e,n){var i=n("7948"),r=n("3818"),o=n("4bb5"),a=n("e2e4"),s=n("8eeb"),l=n("e0e7"),u=n("c6cf"),c=n("1bac"),d=1,h=2,f=4,p=u((function(t,e){var n={};if(null==t)return n;var u=!1;e=i(e,(function(e){return e=a(e,t),u||(u=e.length>1),e})),s(t,c(t),n),u&&(n=r(n,d|h|f,l));var p=e.length;while(p--)o(n,e[p]);return n}));t.exports=p},"3f84":function(t,e,n){var i=n("85e3"),r=n("100e"),o=n("e031"),a=n("2411"),s=r((function(t){return t.push(void 0,o),i(a,void 0,t)}));t.exports=s},"3f8c":function(t,e){t.exports={}},"3ff1":function(t,e,n){var i=n("266a"),r=n("ec69");function o(t){return null==t?[]:i(t,r(t))}t.exports=o},"408a":function(t,e,n){var i=n("c6b6");t.exports=function(t){if("number"!=typeof t&&"Number"!=i(t))throw TypeError("Incorrect invocation");return+t}},4160:function(t,e,n){"use strict";var i=n("23e7"),r=n("17c2");i({target:"Array",proto:!0,forced:[].forEach!=r},{forEach:r})},"41c3":function(t,e,n){var i=n("1a8c"),r=n("eac5"),o=n("ec8c"),a=Object.prototype,s=a.hasOwnProperty;function l(t){if(!i(t))return o(t);var e=r(t),n=[];for(var a in t)("constructor"!=a||!e&&s.call(t,a))&&n.push(a);return n}t.exports=l},"423e":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ar-kw",{months:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),monthsShort:"يناير_فبراير_مارس_أبريل_ماي_يونيو_يوليوز_غشت_شتنبر_أكتوبر_نونبر_دجنبر".split("_"),weekdays:"الأحد_الإتنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"احد_اتنين_ثلاثاء_اربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},week:{dow:0,doy:12}});return e}))},4245:function(t,e,n){var i=n("1290");function r(t,e){var n=t.__data__;return i(e)?n["string"==typeof e?"string":"hash"]:n.map}t.exports=r},42454:function(t,e,n){var i=n("f909"),r=n("2ec1"),o=r((function(t,e,n){i(t,e,n)}));t.exports=o},4284:function(t,e){function n(t,e){var n=-1,i=null==t?0:t.length;while(++n<i)if(e(t[n],n,t))return!0;return!1}t.exports=n},"428f":function(t,e,n){var i=n("da84");t.exports=i},"42a2":function(t,e,n){var i=n("b5a7"),r=n("79bc"),o=n("1cec"),a=n("c869"),s=n("39ff"),l=n("3729"),u=n("dc57"),c="[object Map]",d="[object Object]",h="[object Promise]",f="[object Set]",p="[object WeakMap]",m="[object DataView]",g=u(i),y=u(r),v=u(o),_=u(a),b=u(s),x=l;(i&&x(new i(new ArrayBuffer(1)))!=m||r&&x(new r)!=c||o&&x(o.resolve())!=h||a&&x(new a)!=f||s&&x(new s)!=p)&&(x=function(t){var e=l(t),n=e==d?t.constructor:void 0,i=n?u(n):"";if(i)switch(i){case g:return m;case y:return c;case v:return h;case _:return f;case b:return p}return e}),t.exports=x},"432f":function(t,e,n){"use strict";
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */const i="undefined"!==typeof window&&null!=window.customElements&&void 0!==window.customElements.polyfillWrapFlushCallback,r=(t,e,n=null)=>{while(e!==n){const n=e.nextSibling;t.removeChild(e),e=n}},o=`{{lit-${String(Math.random()).slice(2)}}}`,a=`\x3c!--${o}--\x3e`,s=new RegExp(`${o}|${a}`),l="$lit$";class u{constructor(t,e){this.parts=[],this.element=e;const n=[],i=[],r=document.createTreeWalker(e.content,133,null,!1);let a=0,u=-1,d=0;const{strings:p,values:{length:m}}=t;while(d<m){const t=r.nextNode();if(null!==t){if(u++,1===t.nodeType){if(t.hasAttributes()){const e=t.attributes,{length:n}=e;let i=0;for(let t=0;t<n;t++)c(e[t].name,l)&&i++;while(i-- >0){const e=p[d],n=f.exec(e)[2],i=n.toLowerCase()+l,r=t.getAttribute(i);t.removeAttribute(i);const o=r.split(s);this.parts.push({type:"attribute",index:u,name:n,strings:o}),d+=o.length-1}}"TEMPLATE"===t.tagName&&(i.push(t),r.currentNode=t.content)}else if(3===t.nodeType){const e=t.data;if(e.indexOf(o)>=0){const i=t.parentNode,r=e.split(s),o=r.length-1;for(let e=0;e<o;e++){let n,o=r[e];if(""===o)n=h();else{const t=f.exec(o);null!==t&&c(t[2],l)&&(o=o.slice(0,t.index)+t[1]+t[2].slice(0,-l.length)+t[3]),n=document.createTextNode(o)}i.insertBefore(n,t),this.parts.push({type:"node",index:++u})}""===r[o]?(i.insertBefore(h(),t),n.push(t)):t.data=r[o],d+=o}}else if(8===t.nodeType)if(t.data===o){const e=t.parentNode;null!==t.previousSibling&&u!==a||(u++,e.insertBefore(h(),t)),a=u,this.parts.push({type:"node",index:u}),null===t.nextSibling?t.data="":(n.push(t),u--),d++}else{let e=-1;while(-1!==(e=t.data.indexOf(o,e+1)))this.parts.push({type:"node",index:-1}),d++}}else r.currentNode=i.pop()}for(const o of n)o.parentNode.removeChild(o)}}const c=(t,e)=>{const n=t.length-e.length;return n>=0&&t.slice(n)===e},d=t=>-1!==t.index,h=()=>document.createComment(""),f=/([ \x09\x0a\x0c\x0d])([^\0-\x1F\x7F-\x9F "'>=/]+)([ \x09\x0a\x0c\x0d]*=[ \x09\x0a\x0c\x0d]*(?:[^ \x09\x0a\x0c\x0d"'`<>=]*|"[^"]*|'[^']*))$/,p=133;function m(t,e){const{element:{content:n},parts:i}=t,r=document.createTreeWalker(n,p,null,!1);let o=y(i),a=i[o],s=-1,l=0;const u=[];let c=null;while(r.nextNode()){s++;const t=r.currentNode;t.previousSibling===c&&(c=null),e.has(t)&&(u.push(t),null===c&&(c=t)),null!==c&&l++;while(void 0!==a&&a.index===s)a.index=null!==c?-1:a.index-l,o=y(i,o),a=i[o]}u.forEach(t=>t.parentNode.removeChild(t))}const g=t=>{let e=11===t.nodeType?0:1;const n=document.createTreeWalker(t,p,null,!1);while(n.nextNode())e++;return e},y=(t,e=-1)=>{for(let n=e+1;n<t.length;n++){const e=t[n];if(d(e))return n}return-1};function v(t,e,n=null){const{element:{content:i},parts:r}=t;if(null===n||void 0===n)return void i.appendChild(e);const o=document.createTreeWalker(i,p,null,!1);let a=y(r),s=0,l=-1;while(o.nextNode()){l++;const t=o.currentNode;t===n&&(s=g(e),n.parentNode.insertBefore(e,n));while(-1!==a&&r[a].index===l){if(s>0){while(-1!==a)r[a].index+=s,a=y(r,a);return}a=y(r,a)}}}
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */const _=new WeakMap,b=t=>"function"===typeof t&&_.has(t),x={},w={};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
class M{constructor(t,e,n){this.__parts=[],this.template=t,this.processor=e,this.options=n}update(t){let e=0;for(const n of this.__parts)void 0!==n&&n.setValue(t[e]),e++;for(const n of this.__parts)void 0!==n&&n.commit()}_clone(){const t=i?this.template.element.content.cloneNode(!0):document.importNode(this.template.element.content,!0),e=[],n=this.template.parts,r=document.createTreeWalker(t,133,null,!1);let o,a=0,s=0,l=r.nextNode();while(a<n.length)if(o=n[a],d(o)){while(s<o.index)s++,"TEMPLATE"===l.nodeName&&(e.push(l),r.currentNode=l.content),null===(l=r.nextNode())&&(r.currentNode=e.pop(),l=r.nextNode());if("node"===o.type){const t=this.processor.handleTextExpression(this.options);t.insertAfterNode(l.previousSibling),this.__parts.push(t)}else this.__parts.push(...this.processor.handleAttributeExpressions(l,o.name,o.strings,this.options));a++}else this.__parts.push(void 0),a++;return i&&(document.adoptNode(t),customElements.upgrade(t)),t}}
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */const S=window.trustedTypes&&trustedTypes.createPolicy("lit-html",{createHTML:t=>t}),L=` ${o} `;class k{constructor(t,e,n,i){this.strings=t,this.values=e,this.type=n,this.processor=i}getHTML(){const t=this.strings.length-1;let e="",n=!1;for(let i=0;i<t;i++){const t=this.strings[i],r=t.lastIndexOf("\x3c!--");n=(r>-1||n)&&-1===t.indexOf("--\x3e",r+1);const s=f.exec(t);e+=null===s?t+(n?L:a):t.substr(0,s.index)+s[1]+s[2]+l+s[3]+o}return e+=this.strings[t],e}getTemplateElement(){const t=document.createElement("template");let e=this.getHTML();return void 0!==S&&(e=S.createHTML(e)),t.innerHTML=e,t}}
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
const T=t=>null===t||!("object"===typeof t||"function"===typeof t),D=t=>Array.isArray(t)||!(!t||!t[Symbol.iterator]);class C{constructor(t,e,n){this.dirty=!0,this.element=t,this.name=e,this.strings=n,this.parts=[];for(let i=0;i<n.length-1;i++)this.parts[i]=this._createPart()}_createPart(){return new I(this)}_getValue(){const t=this.strings,e=t.length-1,n=this.parts;if(1===e&&""===t[0]&&""===t[1]){const t=n[0].value;if("symbol"===typeof t)return String(t);if("string"===typeof t||!D(t))return t}let i="";for(let r=0;r<e;r++){i+=t[r];const e=n[r];if(void 0!==e){const t=e.value;if(T(t)||!D(t))i+="string"===typeof t?t:String(t);else for(const e of t)i+="string"===typeof e?e:String(e)}}return i+=t[e],i}commit(){this.dirty&&(this.dirty=!1,this.element.setAttribute(this.name,this._getValue()))}}class I{constructor(t){this.value=void 0,this.committer=t}setValue(t){t===x||T(t)&&t===this.value||(this.value=t,b(t)||(this.committer.dirty=!0))}commit(){while(b(this.value)){const t=this.value;this.value=x,t(this)}this.value!==x&&this.committer.commit()}}class A{constructor(t){this.value=void 0,this.__pendingValue=void 0,this.options=t}appendInto(t){this.startNode=t.appendChild(h()),this.endNode=t.appendChild(h())}insertAfterNode(t){this.startNode=t,this.endNode=t.nextSibling}appendIntoPart(t){t.__insert(this.startNode=h()),t.__insert(this.endNode=h())}insertAfterPart(t){t.__insert(this.startNode=h()),this.endNode=t.endNode,t.endNode=this.startNode}setValue(t){this.__pendingValue=t}commit(){if(null===this.startNode.parentNode)return;while(b(this.__pendingValue)){const t=this.__pendingValue;this.__pendingValue=x,t(this)}const t=this.__pendingValue;t!==x&&(T(t)?t!==this.value&&this.__commitText(t):t instanceof k?this.__commitTemplateResult(t):t instanceof Node?this.__commitNode(t):D(t)?this.__commitIterable(t):t===w?(this.value=w,this.clear()):this.__commitText(t))}__insert(t){this.endNode.parentNode.insertBefore(t,this.endNode)}__commitNode(t){this.value!==t&&(this.clear(),this.__insert(t),this.value=t)}__commitText(t){const e=this.startNode.nextSibling;t=null==t?"":t;const n="string"===typeof t?t:String(t);e===this.endNode.previousSibling&&3===e.nodeType?e.data=n:this.__commitNode(document.createTextNode(n)),this.value=t}__commitTemplateResult(t){const e=this.options.templateFactory(t);if(this.value instanceof M&&this.value.template===e)this.value.update(t.values);else{const n=new M(e,t.processor,this.options),i=n._clone();n.update(t.values),this.__commitNode(i),this.value=n}}__commitIterable(t){Array.isArray(this.value)||(this.value=[],this.clear());const e=this.value;let n,i=0;for(const r of t)n=e[i],void 0===n&&(n=new A(this.options),e.push(n),0===i?n.appendIntoPart(this):n.insertAfterPart(e[i-1])),n.setValue(r),n.commit(),i++;i<e.length&&(e.length=i,this.clear(n&&n.endNode))}clear(t=this.startNode){r(this.startNode.parentNode,t.nextSibling,this.endNode)}}class Y{constructor(t,e,n){if(this.value=void 0,this.__pendingValue=void 0,2!==n.length||""!==n[0]||""!==n[1])throw new Error("Boolean attributes can only contain a single expression");this.element=t,this.name=e,this.strings=n}setValue(t){this.__pendingValue=t}commit(){while(b(this.__pendingValue)){const t=this.__pendingValue;this.__pendingValue=x,t(this)}if(this.__pendingValue===x)return;const t=!!this.__pendingValue;this.value!==t&&(t?this.element.setAttribute(this.name,""):this.element.removeAttribute(this.name),this.value=t),this.__pendingValue=x}}class O extends C{constructor(t,e,n){super(t,e,n),this.single=2===n.length&&""===n[0]&&""===n[1]}_createPart(){return new P(this)}_getValue(){return this.single?this.parts[0].value:super._getValue()}commit(){this.dirty&&(this.dirty=!1,this.element[this.name]=this._getValue())}}class P extends I{}let E=!1;(()=>{try{const t={get capture(){return E=!0,!1}};window.addEventListener("test",t,t),window.removeEventListener("test",t,t)}catch(t){}})();class N{constructor(t,e,n){this.value=void 0,this.__pendingValue=void 0,this.element=t,this.eventName=e,this.eventContext=n,this.__boundHandleEvent=t=>this.handleEvent(t)}setValue(t){this.__pendingValue=t}commit(){while(b(this.__pendingValue)){const t=this.__pendingValue;this.__pendingValue=x,t(this)}if(this.__pendingValue===x)return;const t=this.__pendingValue,e=this.value,n=null==t||null!=e&&(t.capture!==e.capture||t.once!==e.once||t.passive!==e.passive),i=null!=t&&(null==e||n);n&&this.element.removeEventListener(this.eventName,this.__boundHandleEvent,this.__options),i&&(this.__options=R(t),this.element.addEventListener(this.eventName,this.__boundHandleEvent,this.__options)),this.value=t,this.__pendingValue=x}handleEvent(t){"function"===typeof this.value?this.value.call(this.eventContext||this.element,t):this.value.handleEvent(t)}}const R=t=>t&&(E?{capture:t.capture,passive:t.passive,once:t.once}:t.capture)
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */;function H(t){let e=j.get(t.type);void 0===e&&(e={stringsArray:new WeakMap,keyString:new Map},j.set(t.type,e));let n=e.stringsArray.get(t.strings);if(void 0!==n)return n;const i=t.strings.join(o);return n=e.keyString.get(i),void 0===n&&(n=new u(t,t.getTemplateElement()),e.keyString.set(i,n)),e.stringsArray.set(t.strings,n),n}const j=new Map,z=new WeakMap,W=(t,e,n)=>{let i=z.get(e);void 0===i&&(r(e,e.firstChild),z.set(e,i=new A(Object.assign({templateFactory:H},n))),i.appendInto(e)),i.setValue(t),i.commit()};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
class B{handleAttributeExpressions(t,e,n,i){const r=e[0];if("."===r){const i=new O(t,e.slice(1),n);return i.parts}if("@"===r)return[new N(t,e.slice(1),i.eventContext)];if("?"===r)return[new Y(t,e.slice(1),n)];const o=new C(t,e,n);return o.parts}handleTextExpression(t){return new A(t)}}const F=new B;
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */"undefined"!==typeof window&&(window["litHtmlVersions"]||(window["litHtmlVersions"]=[])).push("1.4.1");const V=(t,...e)=>new k(t,e,"html",F),G=(t,e)=>`${t}--${e}`;let $=!0;("undefined"===typeof window.ShadyCSS||"undefined"===typeof window.ShadyCSS.prepareTemplateDom)&&($=!1);const U=t=>e=>{const n=G(e.type,t);let i=j.get(n);void 0===i&&(i={stringsArray:new WeakMap,keyString:new Map},j.set(n,i));let r=i.stringsArray.get(e.strings);if(void 0!==r)return r;const a=e.strings.join(o);if(r=i.keyString.get(a),void 0===r){const n=e.getTemplateElement();$&&window.ShadyCSS.prepareTemplateDom(n,t),r=new u(e,n),i.keyString.set(a,r)}return i.stringsArray.set(e.strings,r),r},q=["html","svg"],X=t=>{q.forEach(e=>{const n=j.get(G(e,t));void 0!==n&&n.keyString.forEach(t=>{const{element:{content:e}}=t,n=new Set;Array.from(e.querySelectorAll("style")).forEach(t=>{n.add(t)}),m(t,n)})})},Z=new Set,J=(t,e,n)=>{Z.add(t);const i=n?n.element:document.createElement("template"),r=e.querySelectorAll("style"),{length:o}=r;if(0===o)return void window.ShadyCSS.prepareTemplateStyles(i,t);const a=document.createElement("style");for(let u=0;u<o;u++){const t=r[u];t.parentNode.removeChild(t),a.textContent+=t.textContent}X(t);const s=i.content;n?v(n,a,s.firstChild):s.insertBefore(a,s.firstChild),window.ShadyCSS.prepareTemplateStyles(i,t);const l=s.querySelector("style");if(window.ShadyCSS.nativeShadow&&null!==l)e.insertBefore(l.cloneNode(!0),e.firstChild);else if(n){s.insertBefore(a,s.firstChild);const t=new Set;t.add(a),m(n,t)}},K=(t,e,n)=>{if(!n||"object"!==typeof n||!n.scopeName)throw new Error("The `scopeName` option is required.");const i=n.scopeName,o=z.has(e),a=$&&11===e.nodeType&&!!e.host,s=a&&!Z.has(i),l=s?document.createDocumentFragment():e;if(W(t,l,Object.assign({templateFactory:U(i)},n)),s){const t=z.get(l);z.delete(l);const n=t.value instanceof M?t.value.template:void 0;J(i,l,n),r(e,e.firstChild),e.appendChild(l),z.set(e,t)}!o&&a&&window.ShadyCSS.styleElement(e.host)};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
var Q;window.JSCompiler_renameProperty=(t,e)=>t;const tt={toAttribute(t,e){switch(e){case Boolean:return t?"":null;case Object:case Array:return null==t?t:JSON.stringify(t)}return t},fromAttribute(t,e){switch(e){case Boolean:return null!==t;case Number:return null===t?null:Number(t);case Object:case Array:return JSON.parse(t)}return t}},et=(t,e)=>e!==t&&(e===e||t===t),nt={attribute:!0,type:String,converter:tt,reflect:!1,hasChanged:et},it=1,rt=4,ot=8,at=16,st="finalized";class lt extends HTMLElement{constructor(){super(),this.initialize()}static get observedAttributes(){this.finalize();const t=[];return this._classProperties.forEach((e,n)=>{const i=this._attributeNameForProperty(n,e);void 0!==i&&(this._attributeToPropertyMap.set(i,n),t.push(i))}),t}static _ensureClassProperties(){if(!this.hasOwnProperty(JSCompiler_renameProperty("_classProperties",this))){this._classProperties=new Map;const t=Object.getPrototypeOf(this)._classProperties;void 0!==t&&t.forEach((t,e)=>this._classProperties.set(e,t))}}static createProperty(t,e=nt){if(this._ensureClassProperties(),this._classProperties.set(t,e),e.noAccessor||this.prototype.hasOwnProperty(t))return;const n="symbol"===typeof t?Symbol():"__"+t,i=this.getPropertyDescriptor(t,n,e);void 0!==i&&Object.defineProperty(this.prototype,t,i)}static getPropertyDescriptor(t,e,n){return{get(){return this[e]},set(i){const r=this[t];this[e]=i,this.requestUpdateInternal(t,r,n)},configurable:!0,enumerable:!0}}static getPropertyOptions(t){return this._classProperties&&this._classProperties.get(t)||nt}static finalize(){const t=Object.getPrototypeOf(this);if(t.hasOwnProperty(st)||t.finalize(),this[st]=!0,this._ensureClassProperties(),this._attributeToPropertyMap=new Map,this.hasOwnProperty(JSCompiler_renameProperty("properties",this))){const t=this.properties,e=[...Object.getOwnPropertyNames(t),..."function"===typeof Object.getOwnPropertySymbols?Object.getOwnPropertySymbols(t):[]];for(const n of e)this.createProperty(n,t[n])}}static _attributeNameForProperty(t,e){const n=e.attribute;return!1===n?void 0:"string"===typeof n?n:"string"===typeof t?t.toLowerCase():void 0}static _valueHasChanged(t,e,n=et){return n(t,e)}static _propertyValueFromAttribute(t,e){const n=e.type,i=e.converter||tt,r="function"===typeof i?i:i.fromAttribute;return r?r(t,n):t}static _propertyValueToAttribute(t,e){if(void 0===e.reflect)return;const n=e.type,i=e.converter,r=i&&i.toAttribute||tt.toAttribute;return r(t,n)}initialize(){this._updateState=0,this._updatePromise=new Promise(t=>this._enableUpdatingResolver=t),this._changedProperties=new Map,this._saveInstanceProperties(),this.requestUpdateInternal()}_saveInstanceProperties(){this.constructor._classProperties.forEach((t,e)=>{if(this.hasOwnProperty(e)){const t=this[e];delete this[e],this._instanceProperties||(this._instanceProperties=new Map),this._instanceProperties.set(e,t)}})}_applyInstanceProperties(){this._instanceProperties.forEach((t,e)=>this[e]=t),this._instanceProperties=void 0}connectedCallback(){this.enableUpdating()}enableUpdating(){void 0!==this._enableUpdatingResolver&&(this._enableUpdatingResolver(),this._enableUpdatingResolver=void 0)}disconnectedCallback(){}attributeChangedCallback(t,e,n){e!==n&&this._attributeToProperty(t,n)}_propertyToAttribute(t,e,n=nt){const i=this.constructor,r=i._attributeNameForProperty(t,n);if(void 0!==r){const t=i._propertyValueToAttribute(e,n);if(void 0===t)return;this._updateState=this._updateState|ot,null==t?this.removeAttribute(r):this.setAttribute(r,t),this._updateState=this._updateState&~ot}}_attributeToProperty(t,e){if(this._updateState&ot)return;const n=this.constructor,i=n._attributeToPropertyMap.get(t);if(void 0!==i){const t=n.getPropertyOptions(i);this._updateState=this._updateState|at,this[i]=n._propertyValueFromAttribute(e,t),this._updateState=this._updateState&~at}}requestUpdateInternal(t,e,n){let i=!0;if(void 0!==t){const r=this.constructor;n=n||r.getPropertyOptions(t),r._valueHasChanged(this[t],e,n.hasChanged)?(this._changedProperties.has(t)||this._changedProperties.set(t,e),!0!==n.reflect||this._updateState&at||(void 0===this._reflectingProperties&&(this._reflectingProperties=new Map),this._reflectingProperties.set(t,n))):i=!1}!this._hasRequestedUpdate&&i&&(this._updatePromise=this._enqueueUpdate())}requestUpdate(t,e){return this.requestUpdateInternal(t,e),this.updateComplete}async _enqueueUpdate(){this._updateState=this._updateState|rt;try{await this._updatePromise}catch(e){}const t=this.performUpdate();return null!=t&&await t,!this._hasRequestedUpdate}get _hasRequestedUpdate(){return this._updateState&rt}get hasUpdated(){return this._updateState&it}performUpdate(){if(!this._hasRequestedUpdate)return;this._instanceProperties&&this._applyInstanceProperties();let t=!1;const e=this._changedProperties;try{t=this.shouldUpdate(e),t?this.update(e):this._markUpdated()}catch(n){throw t=!1,this._markUpdated(),n}t&&(this._updateState&it||(this._updateState=this._updateState|it,this.firstUpdated(e)),this.updated(e))}_markUpdated(){this._changedProperties=new Map,this._updateState=this._updateState&~rt}get updateComplete(){return this._getUpdateComplete()}_getUpdateComplete(){return this.getUpdateComplete()}getUpdateComplete(){return this._updatePromise}shouldUpdate(t){return!0}update(t){void 0!==this._reflectingProperties&&this._reflectingProperties.size>0&&(this._reflectingProperties.forEach((t,e)=>this._propertyToAttribute(e,this[e],t)),this._reflectingProperties=void 0),this._markUpdated()}updated(t){}firstUpdated(t){}}Q=st,lt[Q]=!0;const ut=Element.prototype;ut.msMatchesSelector||ut.webkitMatchesSelector;
/**
@license
Copyright (c) 2019 The Polymer Project Authors. All rights reserved.
This code may only be used under the BSD style license found at
http://polymer.github.io/LICENSE.txt The complete set of authors may be found at
http://polymer.github.io/AUTHORS.txt The complete set of contributors may be
found at http://polymer.github.io/CONTRIBUTORS.txt Code distributed by Google as
part of the polymer project is also subject to an additional IP rights grant
found at http://polymer.github.io/PATENTS.txt
*/
const ct=window.ShadowRoot&&(void 0===window.ShadyCSS||window.ShadyCSS.nativeShadow)&&"adoptedStyleSheets"in Document.prototype&&"replace"in CSSStyleSheet.prototype,dt=Symbol();class ht{constructor(t,e){if(e!==dt)throw new Error("CSSResult is not constructable. Use `unsafeCSS` or `css` instead.");this.cssText=t}get styleSheet(){return void 0===this._styleSheet&&(ct?(this._styleSheet=new CSSStyleSheet,this._styleSheet.replaceSync(this.cssText)):this._styleSheet=null),this._styleSheet}toString(){return this.cssText}}const ft=t=>new ht(String(t),dt),pt=t=>{if(t instanceof ht)return t.cssText;if("number"===typeof t)return t;throw new Error(`Value passed to 'css' function must be a 'css' function result: ${t}. Use 'unsafeCSS' to pass non-literal values, but\n            take care to ensure page security.`)},mt=(t,...e)=>{const n=e.reduce((e,n,i)=>e+pt(n)+t[i+1],t[0]);return new ht(n,dt)};
/**
 * @license
 * Copyright (c) 2017 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at
 * http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at
 * http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at
 * http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at
 * http://polymer.github.io/PATENTS.txt
 */
(window["litElementVersions"]||(window["litElementVersions"]=[])).push("2.5.1");const gt={};class yt extends lt{static getStyles(){return this.styles}static _getUniqueStyles(){if(this.hasOwnProperty(JSCompiler_renameProperty("_styles",this)))return;const t=this.getStyles();if(Array.isArray(t)){const e=(t,n)=>t.reduceRight((t,n)=>Array.isArray(n)?e(n,t):(t.add(n),t),n),n=e(t,new Set),i=[];n.forEach(t=>i.unshift(t)),this._styles=i}else this._styles=void 0===t?[]:[t];this._styles=this._styles.map(t=>{if(t instanceof CSSStyleSheet&&!ct){const e=Array.prototype.slice.call(t.cssRules).reduce((t,e)=>t+e.cssText,"");return ft(e)}return t})}initialize(){super.initialize(),this.constructor._getUniqueStyles(),this.renderRoot=this.createRenderRoot(),window.ShadowRoot&&this.renderRoot instanceof window.ShadowRoot&&this.adoptStyles()}createRenderRoot(){return this.attachShadow(this.constructor.shadowRootOptions)}adoptStyles(){const t=this.constructor._styles;0!==t.length&&(void 0===window.ShadyCSS||window.ShadyCSS.nativeShadow?ct?this.renderRoot.adoptedStyleSheets=t.map(t=>t instanceof CSSStyleSheet?t:t.styleSheet):this._needsShimAdoptedStyleSheets=!0:window.ShadyCSS.ScopingShim.prepareAdoptedCssText(t.map(t=>t.cssText),this.localName))}connectedCallback(){super.connectedCallback(),this.hasUpdated&&void 0!==window.ShadyCSS&&window.ShadyCSS.styleElement(this)}update(t){const e=this.render();super.update(t),e!==gt&&this.constructor.render(e,this.renderRoot,{scopeName:this.localName,eventContext:this}),this._needsShimAdoptedStyleSheets&&(this._needsShimAdoptedStyleSheets=!1,this.constructor._styles.forEach(t=>{const e=document.createElement("style");e.textContent=t.cssText,this.renderRoot.appendChild(e)}))}render(){return gt}}yt["finalized"]=!0,yt.render=K,yt.shadowRootOptions={mode:"open"};const vt=[{name:"初始资金",key:"start_balance",fix_decs:2},{name:"结束资金",key:"end_balance",fix_decs:2},{name:"开始交易日",key:"start_date"},{name:"结束交易日",key:"end_date"},{name:"累计盈利天数",key:"cum_profit_days",fix_decs:0},{name:"累计亏损天数",key:"cum_loss_days",fix_decs:0},{name:"年化收益率",key:"annual_yield",is_percent:!0,fix_decs:2},{name:"最大回撤",key:"max_drawdown",is_percent:!0,fix_decs:2},{name:"总手续费",key:"commission",fix_decs:2},{name:"开仓次数",key:"open_times"},{name:"平仓次数",key:"close_times"},{name:"日均风险度",key:"daily_risk_ratio",is_percent:!0,fix_decs:2},{name:"最大连续盈利天数",key:"max_cont_profit_days"},{name:"最大连续亏损天数",key:"max_cont_loss_days"},{name:"年化夏普率",key:"sharpe_ratio",fix_decs:2},{name:"年化卡玛比率",key:"calmar_ratio",fix_decs:2},{name:"年化索提诺比率",key:"sortino_ratio",fix_decs:2}];class _t extends yt{static get properties(){return{metries:{type:Object}}}static get styles(){return mt`
    table.metries {
      text-indent: 0px;
      border: 2px solid black;
      border-collapse: collapse;
    }
    table.metries>tr {
        border: 1px solid rgba(0, 0, 0, .2);
    }
    table.metries>tr>td {
      padding: 0.5rem;
    }
    table.metries>tr>td.text-right {
      text-align: right;
    }
  `}constructor(){super(),this.metries={},this.columns=vt}formatter(t){return t.fix_decs?t.is_percent?Number(100*this.metries[t.key]).toFixed(t.fix_decs)+"%":Number(this.metries[t.key]).toFixed(t.fix_decs):this.metries[t.key]}render(){return V`
      <table class="metries">
        ${this.columns.map((t,e,n)=>V`
            <tr>
              <td>${t.name}</td>
              <td class="text-right">${this.formatter(t)}</td>
            </tr>`)}
      </table>
      `}}customElements.define("tq-report-metrics",_t)},4359:function(t,e){function n(t,e){var n=-1,i=t.length;e||(e=Array(i));while(++n<i)e[n]=t[n];return e}t.exports=n},4362:function(t,e,n){e.nextTick=function(t){var e=Array.prototype.slice.call(arguments);e.shift(),setTimeout((function(){t.apply(null,e)}),0)},e.platform=e.arch=e.execPath=e.title="browser",e.pid=1,e.browser=!0,e.env={},e.argv=[],e.binding=function(t){throw new Error("No such module. (Possibly not yet loaded)")},function(){var t,i="/";e.cwd=function(){return i},e.chdir=function(e){t||(t=n("df7c")),i=t.resolve(e,i)}}(),e.exit=e.kill=e.umask=e.dlopen=e.uptime=e.memoryUsage=e.uvCounters=function(){},e.features={}},"43ad":function(t,e,n){var i=n("23e2");function r(t,e){return function(n,r){return i(n,t,e(r),{})}}t.exports=r},"440c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r={m:["eng Minutt","enger Minutt"],h:["eng Stonn","enger Stonn"],d:["een Dag","engem Dag"],M:["ee Mount","engem Mount"],y:["ee Joer","engem Joer"]};return e?r[n][0]:r[n][1]}function n(t){var e=t.substr(0,t.indexOf(" "));return r(e)?"a "+t:"an "+t}function i(t){var e=t.substr(0,t.indexOf(" "));return r(e)?"viru "+t:"virun "+t}function r(t){if(t=parseInt(t,10),isNaN(t))return!1;if(t<0)return!0;if(t<10)return 4<=t&&t<=7;if(t<100){var e=t%10,n=t/10;return r(0===e?n:e)}if(t<1e4){while(t>=10)t/=10;return r(t)}return t/=1e3,r(t)}var o=t.defineLocale("lb",{months:"Januar_Februar_Mäerz_Abrëll_Mee_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jan._Febr._Mrz._Abr._Mee_Jun._Jul._Aug._Sept._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonndeg_Méindeg_Dënschdeg_Mëttwoch_Donneschdeg_Freideg_Samschdeg".split("_"),weekdaysShort:"So._Mé._Dë._Më._Do._Fr._Sa.".split("_"),weekdaysMin:"So_Mé_Dë_Më_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm [Auer]",LTS:"H:mm:ss [Auer]",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm [Auer]",LLLL:"dddd, D. MMMM YYYY H:mm [Auer]"},calendar:{sameDay:"[Haut um] LT",sameElse:"L",nextDay:"[Muer um] LT",nextWeek:"dddd [um] LT",lastDay:"[Gëschter um] LT",lastWeek:function(){switch(this.day()){case 2:case 4:return"[Leschten] dddd [um] LT";default:return"[Leschte] dddd [um] LT"}}},relativeTime:{future:n,past:i,s:"e puer Sekonnen",ss:"%d Sekonnen",m:e,mm:"%d Minutten",h:e,hh:"%d Stonnen",d:e,dd:"%d Deeg",M:e,MM:"%d Méint",y:e,yy:"%d Joer"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return o}))},4416:function(t,e){function n(t){var e=null==t?0:t.length;return e?t[e-1]:void 0}t.exports=n},4467:function(t,e,n){var i=n("e2e4"),r=n("9520"),o=n("f4d6");function a(t,e,n){e=i(e,t);var a=-1,s=e.length;s||(s=1,t=void 0);while(++a<s){var l=null==t?void 0:t[o(e[a])];void 0===l&&(a=s,l=n),t=r(l)?l.call(t):l}return t}t.exports=a},4472:function(t,e,n){var i=n("872a"),r=n("242e"),o=n("badf");function a(t,e){var n={};return e=o(e,3),r(t,(function(t,r,o){i(n,e(t,r,o),t)})),n}t.exports=a},"44ad":function(t,e,n){var i=n("d039"),r=n("c6b6"),o="".split;t.exports=i((function(){return!Object("z").propertyIsEnumerable(0)}))?function(t){return"String"==r(t)?o.call(t,""):Object(t)}:Object},"44d2":function(t,e,n){var i=n("b622"),r=n("7c73"),o=n("9bf2"),a=i("unscopables"),s=Array.prototype;void 0==s[a]&&o.f(s,a,{configurable:!0,value:r(null)}),t.exports=function(t){s[a][t]=!0}},"44de":function(t,e,n){var i=n("da84");t.exports=function(t,e){var n=i.console;n&&n.error&&(1===arguments.length?n.error(t):n.error(t,e))}},"44e7":function(t,e,n){var i=n("861d"),r=n("c6b6"),o=n("b622"),a=o("match");t.exports=function(t){var e;return i(t)&&(void 0!==(e=t[a])?!!e:"RegExp"==r(t))}},"45fc":function(t,e,n){"use strict";var i=n("23e7"),r=n("b727").some,o=n("a640"),a=n("ae40"),s=o("some"),l=a("some");i({target:"Array",proto:!0,forced:!s||!l},{some:function(t){return r(this,t,arguments.length>1?arguments[1]:void 0)}})},"466d":function(t,e,n){"use strict";var i=n("d784"),r=n("825a"),o=n("50c4"),a=n("1d80"),s=n("8aa5"),l=n("14c3");i("match",1,(function(t,e,n){return[function(e){var n=a(this),i=void 0==e?void 0:e[t];return void 0!==i?i.call(e,n):new RegExp(e)[t](String(n))},function(t){var i=n(e,t,this);if(i.done)return i.value;var a=r(t),u=String(this);if(!a.global)return l(a,u);var c=a.unicode;a.lastIndex=0;var d,h=[],f=0;while(null!==(d=l(a,u))){var p=String(d[0]);h[f]=p,""===p&&(a.lastIndex=s(u,o(a.lastIndex),c)),f++}return 0===f?null:h}]}))},"467f":function(t,e,n){"use strict";var i=n("2d83");t.exports=function(t,e,n){var r=n.config.validateStatus;n.status&&r&&!r(n.status)?e(i("Request failed with status code "+n.status,n.config,null,n.request,n)):t(n)}},4840:function(t,e,n){var i=n("825a"),r=n("1c0b"),o=n("b622"),a=o("species");t.exports=function(t,e){var n,o=i(t).constructor;return void 0===o||void 0==(n=i(o)[a])?e:r(n)}},"485c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"-inci",5:"-inci",8:"-inci",70:"-inci",80:"-inci",2:"-nci",7:"-nci",20:"-nci",50:"-nci",3:"-üncü",4:"-üncü",100:"-üncü",6:"-ncı",9:"-uncu",10:"-uncu",30:"-uncu",60:"-ıncı",90:"-ıncı"},n=t.defineLocale("az",{months:"yanvar_fevral_mart_aprel_may_iyun_iyul_avqust_sentyabr_oktyabr_noyabr_dekabr".split("_"),monthsShort:"yan_fev_mar_apr_may_iyn_iyl_avq_sen_okt_noy_dek".split("_"),weekdays:"Bazar_Bazar ertəsi_Çərşənbə axşamı_Çərşənbə_Cümə axşamı_Cümə_Şənbə".split("_"),weekdaysShort:"Baz_BzE_ÇAx_Çər_CAx_Cüm_Şən".split("_"),weekdaysMin:"Bz_BE_ÇA_Çə_CA_Cü_Şə".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[bugün saat] LT",nextDay:"[sabah saat] LT",nextWeek:"[gələn həftə] dddd [saat] LT",lastDay:"[dünən] LT",lastWeek:"[keçən həftə] dddd [saat] LT",sameElse:"L"},relativeTime:{future:"%s sonra",past:"%s əvvəl",s:"bir neçə saniyə",ss:"%d saniyə",m:"bir dəqiqə",mm:"%d dəqiqə",h:"bir saat",hh:"%d saat",d:"bir gün",dd:"%d gün",M:"bir ay",MM:"%d ay",y:"bir il",yy:"%d il"},meridiemParse:/gecə|səhər|gündüz|axşam/,isPM:function(t){return/^(gündüz|axşam)$/.test(t)},meridiem:function(t,e,n){return t<4?"gecə":t<12?"səhər":t<17?"gündüz":"axşam"},dayOfMonthOrdinalParse:/\d{1,2}-(ıncı|inci|nci|üncü|ncı|uncu)/,ordinal:function(t){if(0===t)return t+"-ıncı";var n=t%10,i=t%100-n,r=t>=100?100:null;return t+(e[n]||e[i]||e[r])},week:{dow:1,doy:7}});return n}))},4930:function(t,e,n){var i=n("d039");t.exports=!!Object.getOwnPropertySymbols&&!i((function(){return!String(Symbol())}))},"49ab":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-hk",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"週日_週一_週二_週三_週四_週五_週六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日dddd HH:mm",l:"YYYY/M/D",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"中午"===e?t>=11?t:t+12:"下午"===e||"晚上"===e?t+12:void 0},meridiem:function(t,e,n){var i=100*t+e;return i<600?"凌晨":i<900?"早上":i<1200?"上午":1200===i?"中午":i<1800?"下午":"晚上"},calendar:{sameDay:"[今天]LT",nextDay:"[明天]LT",nextWeek:"[下]ddddLT",lastDay:"[昨天]LT",lastWeek:"[上]ddddLT",sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|週)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"週";default:return t}},relativeTime:{future:"%s後",past:"%s前",s:"幾秒",ss:"%d 秒",m:"1 分鐘",mm:"%d 分鐘",h:"1 小時",hh:"%d 小時",d:"1 天",dd:"%d 天",M:"1 個月",MM:"%d 個月",y:"1 年",yy:"%d 年"}});return e}))},"49ad":function(t,e,n){"use strict";t.exports=function(t){var e={};function n(n){var i=t.get(n);return void 0===i?[]:e[i]||[]}function i(n,i){var r=t.get(n);e[r]||(e[r]=[]),e[r].push(i)}function r(t,e){for(var i=n(t),r=0,o=i.length;r<o;++r)if(i[r]===e){i.splice(r,1);break}}function o(t){var e=n(t);e&&(e.length=0)}return{get:n,add:i,removeListener:r,removeAllListeners:o}}},"49f4":function(t,e,n){var i=n("6044");function r(){this.__data__=i?i(null):{},this.size=0}t.exports=r},"4a7b":function(t,e,n){"use strict";var i=n("c532");t.exports=function(t,e){e=e||{};var n={},r=["url","method","data"],o=["headers","auth","proxy","params"],a=["baseURL","transformRequest","transformResponse","paramsSerializer","timeout","timeoutMessage","withCredentials","adapter","responseType","xsrfCookieName","xsrfHeaderName","onUploadProgress","onDownloadProgress","decompress","maxContentLength","maxBodyLength","maxRedirects","transport","httpAgent","httpsAgent","cancelToken","socketPath","responseEncoding"],s=["validateStatus"];function l(t,e){return i.isPlainObject(t)&&i.isPlainObject(e)?i.merge(t,e):i.isPlainObject(e)?i.merge({},e):i.isArray(e)?e.slice():e}function u(r){i.isUndefined(e[r])?i.isUndefined(t[r])||(n[r]=l(void 0,t[r])):n[r]=l(t[r],e[r])}i.forEach(r,(function(t){i.isUndefined(e[t])||(n[t]=l(void 0,e[t]))})),i.forEach(o,u),i.forEach(a,(function(r){i.isUndefined(e[r])?i.isUndefined(t[r])||(n[r]=l(void 0,t[r])):n[r]=l(void 0,e[r])})),i.forEach(s,(function(i){i in e?n[i]=l(t[i],e[i]):i in t&&(n[i]=l(void 0,t[i]))}));var c=r.concat(o).concat(a).concat(s),d=Object.keys(t).concat(Object.keys(e)).filter((function(t){return-1===c.indexOf(t)}));return i.forEach(d,u),n}},"4ba9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n){var i=t+" ";switch(n){case"ss":return i+=1===t?"sekunda":2===t||3===t||4===t?"sekunde":"sekundi",i;case"m":return e?"jedna minuta":"jedne minute";case"mm":return i+=1===t?"minuta":2===t||3===t||4===t?"minute":"minuta",i;case"h":return e?"jedan sat":"jednog sata";case"hh":return i+=1===t?"sat":2===t||3===t||4===t?"sata":"sati",i;case"dd":return i+=1===t?"dan":"dana",i;case"MM":return i+=1===t?"mjesec":2===t||3===t||4===t?"mjeseca":"mjeseci",i;case"yy":return i+=1===t?"godina":2===t||3===t||4===t?"godine":"godina",i}}var n=t.defineLocale("hr",{months:{format:"siječnja_veljače_ožujka_travnja_svibnja_lipnja_srpnja_kolovoza_rujna_listopada_studenoga_prosinca".split("_"),standalone:"siječanj_veljača_ožujak_travanj_svibanj_lipanj_srpanj_kolovoz_rujan_listopad_studeni_prosinac".split("_")},monthsShort:"sij._velj._ožu._tra._svi._lip._srp._kol._ruj._lis._stu._pro.".split("_"),monthsParseExact:!0,weekdays:"nedjelja_ponedjeljak_utorak_srijeda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sri._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"Do MMMM YYYY",LLL:"Do MMMM YYYY H:mm",LLLL:"dddd, Do MMMM YYYY H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedjelju] [u] LT";case 3:return"[u] [srijedu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[jučer u] LT",lastWeek:function(){switch(this.day()){case 0:return"[prošlu] [nedjelju] [u] LT";case 3:return"[prošlu] [srijedu] [u] LT";case 6:return"[prošle] [subote] [u] LT";case 1:case 2:case 4:case 5:return"[prošli] dddd [u] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"prije %s",s:"par sekundi",ss:e,m:e,mm:e,h:e,hh:e,d:"dan",dd:e,M:"mjesec",MM:e,y:"godinu",yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},"4bb5":function(t,e,n){var i=n("e2e4"),r=n("4416"),o=n("8296"),a=n("f4d6");function s(t,e){return e=i(e,t),t=o(t,e),null==t||delete t[a(r(e))]}t.exports=s},"4be8":function(t,e,n){var i=n("30e0"),r=n("ec69");function o(t,e){return t&&i(t,e,r)}t.exports=o},"4d64":function(t,e,n){var i=n("fc6a"),r=n("50c4"),o=n("23cb"),a=function(t){return function(e,n,a){var s,l=i(e),u=r(l.length),c=o(a,u);if(t&&n!=n){while(u>c)if(s=l[c++],s!=s)return!0}else for(;u>c;c++)if((t||c in l)&&l[c]===n)return t||c||0;return!t&&-1}};t.exports={includes:a(!0),indexOf:a(!1)}},"4d8b":function(t,e,n){var i=n("9b02");function r(t,e){var n=-1,r=e.length,o=Array(r),a=null==t;while(++n<r)o[n]=a?void 0:i(t,e[n]);return o}t.exports=r},"4d8c":function(t,e,n){var i=n("5c69");function r(t){var e=null==t?0:t.length;return e?i(t,1):[]}t.exports=r},"4de4":function(t,e,n){"use strict";var i=n("23e7"),r=n("b727").filter,o=n("1dde"),a=n("ae40"),s=o("filter"),l=a("filter");i({target:"Array",proto:!0,forced:!s||!l},{filter:function(t){return r(this,t,arguments.length>1?arguments[1]:void 0)}})},"4df4":function(t,e,n){"use strict";var i=n("0366"),r=n("7b0b"),o=n("9bdd"),a=n("e95a"),s=n("50c4"),l=n("8418"),u=n("35a1");t.exports=function(t){var e,n,c,d,h,f,p=r(t),m="function"==typeof this?this:Array,g=arguments.length,y=g>1?arguments[1]:void 0,v=void 0!==y,_=u(p),b=0;if(v&&(y=i(y,g>2?arguments[2]:void 0,2)),void 0==_||m==Array&&a(_))for(e=s(p.length),n=new m(e);e>b;b++)f=v?y(p[b],b):p[b],l(n,b,f);else for(d=_.call(p),h=d.next,n=new m;!(c=h.call(d)).done;b++)f=v?o(d,y,[c.value,b],!0):c.value,l(n,b,f);return n.length=b,n}},"4f50":function(t,e,n){var i=n("b760"),r=n("e5383"),o=n("c8fe"),a=n("4359"),s=n("fa21"),l=n("d370"),u=n("6747"),c=n("dcbe"),d=n("0d24"),h=n("9520"),f=n("1a8c"),p=n("60ed"),m=n("73ac"),g=n("8adb"),y=n("8de2");function v(t,e,n,v,_,b,x){var w=g(t,n),M=g(e,n),S=x.get(M);if(S)i(t,n,S);else{var L=b?b(w,M,n+"",t,e,x):void 0,k=void 0===L;if(k){var T=u(M),D=!T&&d(M),C=!T&&!D&&m(M);L=M,T||D||C?u(w)?L=w:c(w)?L=a(w):D?(k=!1,L=r(M,!0)):C?(k=!1,L=o(M,!0)):L=[]:p(M)||l(M)?(L=w,l(w)?L=y(w):f(w)&&!h(w)||(L=s(M))):k=!1}k&&(x.set(M,L),_(L,M,v,b,x),x["delete"](M)),i(t,n,L)}}t.exports=v},5038:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("id",{months:"Januari_Februari_Maret_April_Mei_Juni_Juli_Agustus_September_Oktober_November_Desember".split("_"),monthsShort:"Jan_Feb_Mar_Apr_Mei_Jun_Jul_Agt_Sep_Okt_Nov_Des".split("_"),weekdays:"Minggu_Senin_Selasa_Rabu_Kamis_Jumat_Sabtu".split("_"),weekdaysShort:"Min_Sen_Sel_Rab_Kam_Jum_Sab".split("_"),weekdaysMin:"Mg_Sn_Sl_Rb_Km_Jm_Sb".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/pagi|siang|sore|malam/,meridiemHour:function(t,e){return 12===t&&(t=0),"pagi"===e?t:"siang"===e?t>=11?t:t+12:"sore"===e||"malam"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"pagi":t<15?"siang":t<19?"sore":"malam"},calendar:{sameDay:"[Hari ini pukul] LT",nextDay:"[Besok pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kemarin pukul] LT",lastWeek:"dddd [lalu pukul] LT",sameElse:"L"},relativeTime:{future:"dalam %s",past:"%s yang lalu",s:"beberapa detik",ss:"%d detik",m:"semenit",mm:"%d menit",h:"sejam",hh:"%d jam",d:"sehari",dd:"%d hari",M:"sebulan",MM:"%d bulan",y:"setahun",yy:"%d tahun"},week:{dow:0,doy:6}});return e}))},5058:function(t,e,n){"use strict";t.exports=function(t){var e=t.idGenerator,n=t.stateHandler.getState;function i(t){var e=n(t);return e&&void 0!==e.id?e.id:null}function r(t){var i=n(t);if(!i)throw new Error("setId required the element to have a resize detection state.");var r=e.generate();return i.id=r,r}return{get:i,set:r}}},"50bf":function(t,e,n){"use strict";var i=t.exports={};function r(t,e,n){var i=t[e];return void 0!==i&&null!==i||void 0===n?i:n}i.getOption=r},"50c4":function(t,e,n){var i=n("a691"),r=Math.min;t.exports=function(t){return t>0?r(i(t),9007199254740991):0}},"50ca":function(t,e,n){var i=n("8057"),r=n("7530"),o=n("242e"),a=n("badf"),s=n("2dcb"),l=n("6747"),u=n("0d24"),c=n("9520"),d=n("1a8c"),h=n("73ac");function f(t,e,n){var f=l(t),p=f||u(t)||h(t);if(e=a(e,4),null==n){var m=t&&t.constructor;n=p?f?new m:[]:d(t)&&c(m)?r(s(t)):{}}return(p?i:o)(t,(function(t,i,r){return e(n,t,i,r)})),n}t.exports=f},"50d8":function(t,e){function n(t,e){var n=-1,i=Array(t);while(++n<t)i[n]=e(n);return i}t.exports=n},5120:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["Eanáir","Feabhra","Márta","Aibreán","Bealtaine","Meitheamh","Iúil","Lúnasa","Meán Fómhair","Deireadh Fómhair","Samhain","Nollaig"],n=["Ean","Feabh","Márt","Aib","Beal","Meith","Iúil","Lún","M.F.","D.F.","Samh","Noll"],i=["Dé Domhnaigh","Dé Luain","Dé Máirt","Dé Céadaoin","Déardaoin","Dé hAoine","Dé Sathairn"],r=["Domh","Luan","Máirt","Céad","Déar","Aoine","Sath"],o=["Do","Lu","Má","Cé","Dé","A","Sa"],a=t.defineLocale("ga",{months:e,monthsShort:n,monthsParseExact:!0,weekdays:i,weekdaysShort:r,weekdaysMin:o,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Inniu ag] LT",nextDay:"[Amárach ag] LT",nextWeek:"dddd [ag] LT",lastDay:"[Inné ag] LT",lastWeek:"dddd [seo caite] [ag] LT",sameElse:"L"},relativeTime:{future:"i %s",past:"%s ó shin",s:"cúpla soicind",ss:"%d soicind",m:"nóiméad",mm:"%d nóiméad",h:"uair an chloig",hh:"%d uair an chloig",d:"lá",dd:"%d lá",M:"mí",MM:"%d míonna",y:"bliain",yy:"%d bliain"},dayOfMonthOrdinalParse:/\d{1,2}(d|na|mh)/,ordinal:function(t){var e=1===t?"d":t%10===2?"na":"mh";return t+e},week:{dow:1,doy:4}});return a}))},5135:function(t,e){var n={}.hasOwnProperty;t.exports=function(t,e){return n.call(t,e)}},5250:function(t,e,n){"use strict";var i=n("099a");e["a"]=i["a"]},5270:function(t,e,n){"use strict";var i=n("c532"),r=n("c401"),o=n("2e67"),a=n("2444");function s(t){t.cancelToken&&t.cancelToken.throwIfRequested()}t.exports=function(t){s(t),t.headers=t.headers||{},t.data=r(t.data,t.headers,t.transformRequest),t.headers=i.merge(t.headers.common||{},t.headers[t.method]||{},t.headers),i.forEach(["delete","get","head","post","put","patch","common"],(function(e){delete t.headers[e]}));var e=t.adapter||a.adapter;return e(t).then((function(e){return s(t),e.data=r(e.data,e.headers,t.transformResponse),e}),(function(e){return o(e)||(s(t),e&&e.response&&(e.response.data=r(e.response.data,e.response.headers,t.transformResponse))),Promise.reject(e)}))}},5294:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["جنوری","فروری","مارچ","اپریل","مئی","جون","جولائی","اگست","ستمبر","اکتوبر","نومبر","دسمبر"],n=["اتوار","پیر","منگل","بدھ","جمعرات","جمعہ","ہفتہ"],i=t.defineLocale("ur",{months:e,monthsShort:e,weekdays:n,weekdaysShort:n,weekdaysMin:n,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd، D MMMM YYYY HH:mm"},meridiemParse:/صبح|شام/,isPM:function(t){return"شام"===t},meridiem:function(t,e,n){return t<12?"صبح":"شام"},calendar:{sameDay:"[آج بوقت] LT",nextDay:"[کل بوقت] LT",nextWeek:"dddd [بوقت] LT",lastDay:"[گذشتہ روز بوقت] LT",lastWeek:"[گذشتہ] dddd [بوقت] LT",sameElse:"L"},relativeTime:{future:"%s بعد",past:"%s قبل",s:"چند سیکنڈ",ss:"%d سیکنڈ",m:"ایک منٹ",mm:"%d منٹ",h:"ایک گھنٹہ",hh:"%d گھنٹے",d:"ایک دن",dd:"%d دن",M:"ایک ماہ",MM:"%d ماہ",y:"ایک سال",yy:"%d سال"},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:1,doy:4}});return i}))},"52bd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ss",{months:"Bhimbidvwane_Indlovana_Indlov'lenkhulu_Mabasa_Inkhwekhweti_Inhlaba_Kholwane_Ingci_Inyoni_Imphala_Lweti_Ingongoni".split("_"),monthsShort:"Bhi_Ina_Inu_Mab_Ink_Inh_Kho_Igc_Iny_Imp_Lwe_Igo".split("_"),weekdays:"Lisontfo_Umsombuluko_Lesibili_Lesitsatfu_Lesine_Lesihlanu_Umgcibelo".split("_"),weekdaysShort:"Lis_Umb_Lsb_Les_Lsi_Lsh_Umg".split("_"),weekdaysMin:"Li_Us_Lb_Lt_Ls_Lh_Ug".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Namuhla nga] LT",nextDay:"[Kusasa nga] LT",nextWeek:"dddd [nga] LT",lastDay:"[Itolo nga] LT",lastWeek:"dddd [leliphelile] [nga] LT",sameElse:"L"},relativeTime:{future:"nga %s",past:"wenteka nga %s",s:"emizuzwana lomcane",ss:"%d mzuzwana",m:"umzuzu",mm:"%d emizuzu",h:"lihora",hh:"%d emahora",d:"lilanga",dd:"%d emalanga",M:"inyanga",MM:"%d tinyanga",y:"umnyaka",yy:"%d iminyaka"},meridiemParse:/ekuseni|emini|entsambama|ebusuku/,meridiem:function(t,e,n){return t<11?"ekuseni":t<15?"emini":t<19?"entsambama":"ebusuku"},meridiemHour:function(t,e){return 12===t&&(t=0),"ekuseni"===e?t:"emini"===e?t>=11?t:t+12:"entsambama"===e||"ebusuku"===e?0===t?0:t+12:void 0},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:"%d",week:{dow:1,doy:4}});return e}))},5334:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{directives:[{name:"show",rawName:"v-show",value:t.show,expression:"show"}],class:t.prefixCls,style:t.contentStyle},[t._t("default")],2)},r=[],o=(n("b0c0"),n("a9e3"),"ivu-tabs-tabpane"),a={name:"TabPane",inject:["TabsInstance"],props:{name:{type:String},label:{type:[String,Function],default:""},icon:{type:String},disabled:{type:Boolean,default:!1},closable:{type:Boolean,default:null},tab:{type:String},index:{type:Number},contextMenu:{type:Boolean,default:!1}},data:function(){return{prefixCls:o,show:!0,currentName:this.name}},computed:{contentStyle:function(){return{visibility:this.TabsInstance.activeKey!==this.currentName?"hidden":"visible"}}},methods:{updateNav:function(){this.TabsInstance.updateNav()}},watch:{name:function(t){this.currentName=t,this.updateNav()},label:function(){this.updateNav()},icon:function(){this.updateNav()},disabled:function(){this.updateNav()}},mounted:function(){this.updateNav()},destroyed:function(){this.updateNav()}},s=a,l=n("2877"),u=Object(l["a"])(s,i,r,!1,null,null,null);e["a"]=u.exports},"53ca":function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));n("a4d3"),n("e01a"),n("d28b"),n("d3b7"),n("3ca3"),n("ddb0");function i(t){return i="function"===typeof Symbol&&"symbol"===typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"===typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},i(t)}},"54eb":function(t,e,n){var i=n("8eeb"),r=n("32f4");function o(t,e){return i(t,r(t),e)}t.exports=o},"55a3":function(t,e){function n(t){return this.__data__.has(t)}t.exports=n},"55c9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),i=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],r=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es-us",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,i){return t?/-MMM-/.test(i)?n[t.month()]:e[t.month()]:e},monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"MM/DD/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY h:mm A",LLLL:"dddd, D [de] MMMM [de] YYYY h:mm A"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:0,doy:6}});return o}))},5692:function(t,e,n){var i=n("c430"),r=n("c6cd");(t.exports=function(t,e){return r[t]||(r[t]=void 0!==e?e:{})})("versions",[]).push({version:"3.8.3",mode:i?"pure":"global",copyright:"© 2021 Denis Pushkarev (zloirock.ru)"})},"56ef":function(t,e,n){var i=n("d066"),r=n("241c"),o=n("7418"),a=n("825a");t.exports=i("Reflect","ownKeys")||function(t){var e=r.f(a(t)),n=o.f;return n?e.concat(n(t)):e}},"576c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tet",{months:"Janeiru_Fevereiru_Marsu_Abril_Maiu_Juñu_Jullu_Agustu_Setembru_Outubru_Novembru_Dezembru".split("_"),monthsShort:"Jan_Fev_Mar_Abr_Mai_Jun_Jul_Ago_Set_Out_Nov_Dez".split("_"),weekdays:"Domingu_Segunda_Tersa_Kuarta_Kinta_Sesta_Sabadu".split("_"),weekdaysShort:"Dom_Seg_Ters_Kua_Kint_Sest_Sab".split("_"),weekdaysMin:"Do_Seg_Te_Ku_Ki_Ses_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Ohin iha] LT",nextDay:"[Aban iha] LT",nextWeek:"dddd [iha] LT",lastDay:"[Horiseik iha] LT",lastWeek:"dddd [semana kotuk] [iha] LT",sameElse:"L"},relativeTime:{future:"iha %s",past:"%s liuba",s:"segundu balun",ss:"segundu %d",m:"minutu ida",mm:"minutu %d",h:"oras ida",hh:"oras %d",d:"loron ida",dd:"loron %d",M:"fulan ida",MM:"fulan %d",y:"tinan ida",yy:"tinan %d"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},"57a5":function(t,e,n){var i=n("91e9"),r=i(Object.keys,Object);t.exports=r},"585a":function(t,e,n){(function(e){var n="object"==typeof e&&e&&e.Object===Object&&e;t.exports=n}).call(this,n("c8ba"))},5899:function(t,e){t.exports="\t\n\v\f\r                　\u2028\u2029\ufeff"},"58a8":function(t,e,n){var i=n("1d80"),r=n("5899"),o="["+r+"]",a=RegExp("^"+o+o+"*"),s=RegExp(o+o+"*$"),l=function(t){return function(e){var n=String(i(e));return 1&t&&(n=n.replace(a,"")),2&t&&(n=n.replace(s,"")),n}};t.exports={start:l(1),end:l(2),trim:l(3)}},"598a":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["ޖެނުއަރީ","ފެބްރުއަރީ","މާރިޗު","އޭޕްރީލު","މޭ","ޖޫން","ޖުލައި","އޯގަސްޓު","ސެޕްޓެމްބަރު","އޮކްޓޯބަރު","ނޮވެމްބަރު","ޑިސެމްބަރު"],n=["އާދިއްތަ","ހޯމަ","އަންގާރަ","ބުދަ","ބުރާސްފަތި","ހުކުރު","ހޮނިހިރު"],i=t.defineLocale("dv",{months:e,monthsShort:e,weekdays:n,weekdaysShort:n,weekdaysMin:"އާދި_ހޯމަ_އަން_ބުދަ_ބުރާ_ހުކު_ހޮނި".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/M/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/މކ|މފ/,isPM:function(t){return"މފ"===t},meridiem:function(t,e,n){return t<12?"މކ":"މފ"},calendar:{sameDay:"[މިއަދު] LT",nextDay:"[މާދަމާ] LT",nextWeek:"dddd LT",lastDay:"[އިއްޔެ] LT",lastWeek:"[ފާއިތުވި] dddd LT",sameElse:"L"},relativeTime:{future:"ތެރޭގައި %s",past:"ކުރިން %s",s:"ސިކުންތުކޮޅެއް",ss:"d% ސިކުންތު",m:"މިނިޓެއް",mm:"މިނިޓު %d",h:"ގަޑިއިރެއް",hh:"ގަޑިއިރު %d",d:"ދުވަހެއް",dd:"ދުވަސް %d",M:"މަހެއް",MM:"މަސް %d",y:"އަހަރެއް",yy:"އަހަރު %d"},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:7,doy:12}});return i}))},"5a34":function(t,e,n){var i=n("44e7");t.exports=function(t){if(i(t))throw TypeError("The method doesn't accept regular expressions");return t}},"5aff":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"'inji",5:"'inji",8:"'inji",70:"'inji",80:"'inji",2:"'nji",7:"'nji",20:"'nji",50:"'nji",3:"'ünji",4:"'ünji",100:"'ünji",6:"'njy",9:"'unjy",10:"'unjy",30:"'unjy",60:"'ynjy",90:"'ynjy"},n=t.defineLocale("tk",{months:"Ýanwar_Fewral_Mart_Aprel_Maý_Iýun_Iýul_Awgust_Sentýabr_Oktýabr_Noýabr_Dekabr".split("_"),monthsShort:"Ýan_Few_Mar_Apr_Maý_Iýn_Iýl_Awg_Sen_Okt_Noý_Dek".split("_"),weekdays:"Ýekşenbe_Duşenbe_Sişenbe_Çarşenbe_Penşenbe_Anna_Şenbe".split("_"),weekdaysShort:"Ýek_Duş_Siş_Çar_Pen_Ann_Şen".split("_"),weekdaysMin:"Ýk_Dş_Sş_Çr_Pn_An_Şn".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[bugün sagat] LT",nextDay:"[ertir sagat] LT",nextWeek:"[indiki] dddd [sagat] LT",lastDay:"[düýn] LT",lastWeek:"[geçen] dddd [sagat] LT",sameElse:"L"},relativeTime:{future:"%s soň",past:"%s öň",s:"birnäçe sekunt",m:"bir minut",mm:"%d minut",h:"bir sagat",hh:"%d sagat",d:"bir gün",dd:"%d gün",M:"bir aý",MM:"%d aý",y:"bir ýyl",yy:"%d ýyl"},ordinal:function(t,n){switch(n){case"d":case"D":case"Do":case"DD":return t;default:if(0===t)return t+"'unjy";var i=t%10,r=t%100-i,o=t>=100?100:null;return t+(e[i]||e[r]||e[o])}},week:{dow:1,doy:7}});return n}))},"5b01":function(t,e,n){var i=n("8eeb"),r=n("ec69");function o(t,e){return t&&i(e,r(e),t)}t.exports=o},"5b14":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="vasárnap hétfőn kedden szerdán csütörtökön pénteken szombaton".split(" ");function n(t,e,n,i){var r=t;switch(n){case"s":return i||e?"néhány másodperc":"néhány másodperce";case"ss":return r+(i||e)?" másodperc":" másodperce";case"m":return"egy"+(i||e?" perc":" perce");case"mm":return r+(i||e?" perc":" perce");case"h":return"egy"+(i||e?" óra":" órája");case"hh":return r+(i||e?" óra":" órája");case"d":return"egy"+(i||e?" nap":" napja");case"dd":return r+(i||e?" nap":" napja");case"M":return"egy"+(i||e?" hónap":" hónapja");case"MM":return r+(i||e?" hónap":" hónapja");case"y":return"egy"+(i||e?" év":" éve");case"yy":return r+(i||e?" év":" éve")}return""}function i(t){return(t?"":"[múlt] ")+"["+e[this.day()]+"] LT[-kor]"}var r=t.defineLocale("hu",{months:"január_február_március_április_május_június_július_augusztus_szeptember_október_november_december".split("_"),monthsShort:"jan._feb._márc._ápr._máj._jún._júl._aug._szept._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"vasárnap_hétfő_kedd_szerda_csütörtök_péntek_szombat".split("_"),weekdaysShort:"vas_hét_kedd_sze_csüt_pén_szo".split("_"),weekdaysMin:"v_h_k_sze_cs_p_szo".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"YYYY.MM.DD.",LL:"YYYY. MMMM D.",LLL:"YYYY. MMMM D. H:mm",LLLL:"YYYY. MMMM D., dddd H:mm"},meridiemParse:/de|du/i,isPM:function(t){return"u"===t.charAt(1).toLowerCase()},meridiem:function(t,e,n){return t<12?!0===n?"de":"DE":!0===n?"du":"DU"},calendar:{sameDay:"[ma] LT[-kor]",nextDay:"[holnap] LT[-kor]",nextWeek:function(){return i.call(this,!0)},lastDay:"[tegnap] LT[-kor]",lastWeek:function(){return i.call(this,!1)},sameElse:"L"},relativeTime:{future:"%s múlva",past:"%s",s:n,ss:n,m:n,mm:n,h:n,hh:n,d:n,dd:n,M:n,MM:n,y:n,yy:n},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return r}))},"5be5":function(t,e,n){"use strict";t.exports=function(t){var e=t.stateHandler.getState;function n(t){var n=e(t);return n&&!!n.isDetectable}function i(t){e(t).isDetectable=!0}function r(t){return!!e(t).busy}function o(t,n){e(t).busy=!!n}return{isDetectable:n,markAsDetectable:i,isBusy:r,markBusy:o}}},"5c3a":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-cn",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"周日_周一_周二_周三_周四_周五_周六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日Ah点mm分",LLLL:"YYYY年M月D日ddddAh点mm分",l:"YYYY/M/D",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"下午"===e||"晚上"===e?t+12:t>=11?t:t+12},meridiem:function(t,e,n){var i=100*t+e;return i<600?"凌晨":i<900?"早上":i<1130?"上午":i<1230?"中午":i<1800?"下午":"晚上"},calendar:{sameDay:"[今天]LT",nextDay:"[明天]LT",nextWeek:function(t){return t.week()!==this.week()?"[下]dddLT":"[本]dddLT"},lastDay:"[昨天]LT",lastWeek:function(t){return this.week()!==t.week()?"[上]dddLT":"[本]dddLT"},sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|周)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"周";default:return t}},relativeTime:{future:"%s后",past:"%s前",s:"几秒",ss:"%d 秒",m:"1 分钟",mm:"%d 分钟",h:"1 小时",hh:"%d 小时",d:"1 天",dd:"%d 天",w:"1 周",ww:"%d 周",M:"1 个月",MM:"%d 个月",y:"1 年",yy:"%d 年"},week:{dow:1,doy:4}});return e}))},"5c69":function(t,e,n){var i=n("087d"),r=n("0621");function o(t,e,n,a,s){var l=-1,u=t.length;n||(n=r),s||(s=[]);while(++l<u){var c=t[l];e>0&&n(c)?e>1?o(c,e-1,n,a,s):i(s,c):a||(s[s.length]=c)}return s}t.exports=o},"5c6c":function(t,e){t.exports=function(t,e){return{enumerable:!(1&t),configurable:!(2&t),writable:!(4&t),value:e}}},"5cbb":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("te",{months:"జనవరి_ఫిబ్రవరి_మార్చి_ఏప్రిల్_మే_జూన్_జులై_ఆగస్టు_సెప్టెంబర్_అక్టోబర్_నవంబర్_డిసెంబర్".split("_"),monthsShort:"జన._ఫిబ్ర._మార్చి_ఏప్రి._మే_జూన్_జులై_ఆగ._సెప్._అక్టో._నవ._డిసె.".split("_"),monthsParseExact:!0,weekdays:"ఆదివారం_సోమవారం_మంగళవారం_బుధవారం_గురువారం_శుక్రవారం_శనివారం".split("_"),weekdaysShort:"ఆది_సోమ_మంగళ_బుధ_గురు_శుక్ర_శని".split("_"),weekdaysMin:"ఆ_సో_మం_బు_గు_శు_శ".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm",LLLL:"dddd, D MMMM YYYY, A h:mm"},calendar:{sameDay:"[నేడు] LT",nextDay:"[రేపు] LT",nextWeek:"dddd, LT",lastDay:"[నిన్న] LT",lastWeek:"[గత] dddd, LT",sameElse:"L"},relativeTime:{future:"%s లో",past:"%s క్రితం",s:"కొన్ని క్షణాలు",ss:"%d సెకన్లు",m:"ఒక నిమిషం",mm:"%d నిమిషాలు",h:"ఒక గంట",hh:"%d గంటలు",d:"ఒక రోజు",dd:"%d రోజులు",M:"ఒక నెల",MM:"%d నెలలు",y:"ఒక సంవత్సరం",yy:"%d సంవత్సరాలు"},dayOfMonthOrdinalParse:/\d{1,2}వ/,ordinal:"%dవ",meridiemParse:/రాత్రి|ఉదయం|మధ్యాహ్నం|సాయంత్రం/,meridiemHour:function(t,e){return 12===t&&(t=0),"రాత్రి"===e?t<4?t:t+12:"ఉదయం"===e?t:"మధ్యాహ్నం"===e?t>=10?t:t+12:"సాయంత్రం"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"రాత్రి":t<10?"ఉదయం":t<17?"మధ్యాహ్నం":t<20?"సాయంత్రం":"రాత్రి"},week:{dow:0,doy:6}});return e}))},"5d89":function(t,e,n){var i=n("f8af");function r(t,e){var n=e?i(t.buffer):t.buffer;return new t.constructor(n,t.byteOffset,t.byteLength)}t.exports=r},"5e2e":function(t,e,n){var i=n("28c9"),r=n("69d5"),o=n("b4c0"),a=n("fba5"),s=n("67ca");function l(t){var e=-1,n=null==t?0:t.length;this.clear();while(++e<n){var i=t[e];this.set(i[0],i[1])}}l.prototype.clear=i,l.prototype["delete"]=r,l.prototype.get=o,l.prototype.has=a,l.prototype.set=s,t.exports=l},"5e89":function(t,e,n){var i=n("861d"),r=Math.floor;t.exports=function(t){return!i(t)&&isFinite(t)&&r(t)===t}},"5f02":function(t,e,n){"use strict";t.exports=function(t){return"object"===typeof t&&!0===t.isAxiosError}},"5fbd":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("sv",{months:"januari_februari_mars_april_maj_juni_juli_augusti_september_oktober_november_december".split("_"),monthsShort:"jan_feb_mar_apr_maj_jun_jul_aug_sep_okt_nov_dec".split("_"),weekdays:"söndag_måndag_tisdag_onsdag_torsdag_fredag_lördag".split("_"),weekdaysShort:"sön_mån_tis_ons_tor_fre_lör".split("_"),weekdaysMin:"sö_må_ti_on_to_fr_lö".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [kl.] HH:mm",LLLL:"dddd D MMMM YYYY [kl.] HH:mm",lll:"D MMM YYYY HH:mm",llll:"ddd D MMM YYYY HH:mm"},calendar:{sameDay:"[Idag] LT",nextDay:"[Imorgon] LT",lastDay:"[Igår] LT",nextWeek:"[På] dddd LT",lastWeek:"[I] dddd[s] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"för %s sedan",s:"några sekunder",ss:"%d sekunder",m:"en minut",mm:"%d minuter",h:"en timme",hh:"%d timmar",d:"en dag",dd:"%d dagar",M:"en månad",MM:"%d månader",y:"ett år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}(\:e|\:a)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?":e":1===e||2===e?":a":":e";return t+n},week:{dow:1,doy:4}});return e}))},6029:function(t,e,n){"use strict";var i=Array.prototype.slice,r=function(t){return t},o=1,a=2,s=3,l=4,u=1e-6;function c(t){return"translate("+t+",0)"}function d(t){return"translate(0,"+t+")"}function h(t){return e=>+t(e)}function f(t,e){return e=Math.max(0,t.bandwidth()-2*e)/2,t.round()&&(e=Math.round(e)),n=>+t(n)+e}function p(){return!this.__axis}function m(t,e){var n=[],m=null,g=null,y=6,v=6,_=3,b="undefined"!==typeof window&&window.devicePixelRatio>1?0:.5,x=t===o||t===l?-1:1,w=t===l||t===a?"x":"y",M=t===o||t===s?c:d;function S(i){var c=null==m?e.ticks?e.ticks.apply(e,n):e.domain():m,d=null==g?e.tickFormat?e.tickFormat.apply(e,n):r:g,S=Math.max(y,0)+_,L=e.range(),k=+L[0]+b,T=+L[L.length-1]+b,D=(e.bandwidth?f:h)(e.copy(),b),C=i.selection?i.selection():i,I=C.selectAll(".domain").data([null]),A=C.selectAll(".tick").data(c,e).order(),Y=A.exit(),O=A.enter().append("g").attr("class","tick"),P=A.select("line"),E=A.select("text");I=I.merge(I.enter().insert("path",".tick").attr("class","domain").attr("stroke","currentColor")),A=A.merge(O),P=P.merge(O.append("line").attr("stroke","currentColor").attr(w+"2",x*y)),E=E.merge(O.append("text").attr("fill","currentColor").attr(w,x*S).attr("dy",t===o?"0em":t===s?"0.71em":"0.32em")),i!==C&&(I=I.transition(i),A=A.transition(i),P=P.transition(i),E=E.transition(i),Y=Y.transition(i).attr("opacity",u).attr("transform",(function(t){return isFinite(t=D(t))?M(t+b):this.getAttribute("transform")})),O.attr("opacity",u).attr("transform",(function(t){var e=this.parentNode.__axis;return M((e&&isFinite(e=e(t))?e:D(t))+b)}))),Y.remove(),I.attr("d",t===l||t===a?v?"M"+x*v+","+k+"H"+b+"V"+T+"H"+x*v:"M"+b+","+k+"V"+T:v?"M"+k+","+x*v+"V"+b+"H"+T+"V"+x*v:"M"+k+","+b+"H"+T),A.attr("opacity",1).attr("transform",(function(t){return M(D(t)+b)})),P.attr(w+"2",x*y),E.attr(w,x*S).text(d),C.filter(p).attr("fill","none").attr("font-size",10).attr("font-family","sans-serif").attr("text-anchor",t===a?"start":t===l?"end":"middle"),C.each((function(){this.__axis=D}))}return S.scale=function(t){return arguments.length?(e=t,S):e},S.ticks=function(){return n=i.call(arguments),S},S.tickArguments=function(t){return arguments.length?(n=null==t?[]:i.call(t),S):n.slice()},S.tickValues=function(t){return arguments.length?(m=null==t?null:i.call(t),S):m&&m.slice()},S.tickFormat=function(t){return arguments.length?(g=t,S):g},S.tickSize=function(t){return arguments.length?(y=v=+t,S):y},S.tickSizeInner=function(t){return arguments.length?(y=+t,S):y},S.tickSizeOuter=function(t){return arguments.length?(v=+t,S):v},S.tickPadding=function(t){return arguments.length?(_=+t,S):_},S.offset=function(t){return arguments.length?(b=+t,S):b},S}function g(t){return m(a,t)}function y(t){return m(s,t)}function v(t){return m(l,t)}var _=Math.sqrt(50),b=Math.sqrt(10),x=Math.sqrt(2),w=function(t,e,n){var i,r,o,a,s=-1;if(e=+e,t=+t,n=+n,t===e&&n>0)return[t];if((i=e<t)&&(r=t,t=e,e=r),0===(a=M(t,e,n))||!isFinite(a))return[];if(a>0){let n=Math.round(t/a),i=Math.round(e/a);n*a<t&&++n,i*a>e&&--i,o=new Array(r=i-n+1);while(++s<r)o[s]=(n+s)*a}else{a=-a;let n=Math.round(t*a),i=Math.round(e*a);n/a<t&&++n,i/a>e&&--i,o=new Array(r=i-n+1);while(++s<r)o[s]=(n+s)/a}return i&&o.reverse(),o};function M(t,e,n){var i=(e-t)/Math.max(0,n),r=Math.floor(Math.log(i)/Math.LN10),o=i/Math.pow(10,r);return r>=0?(o>=_?10:o>=b?5:o>=x?2:1)*Math.pow(10,r):-Math.pow(10,-r)/(o>=_?10:o>=b?5:o>=x?2:1)}function S(t,e,n){var i=Math.abs(e-t)/Math.max(0,n),r=Math.pow(10,Math.floor(Math.log(i)/Math.LN10)),o=i/r;return o>=_?r*=10:o>=b?r*=5:o>=x&&(r*=2),e<t?-r:r}var L=function(t,e){return t<e?-1:t>e?1:t>=e?0:NaN},k=function(t){let e=t,n=t;function i(t,e,i,r){null==i&&(i=0),null==r&&(r=t.length);while(i<r){const o=i+r>>>1;n(t[o],e)<0?i=o+1:r=o}return i}function r(t,e,i,r){null==i&&(i=0),null==r&&(r=t.length);while(i<r){const o=i+r>>>1;n(t[o],e)>0?r=o:i=o+1}return i}function o(t,n,r,o){null==r&&(r=0),null==o&&(o=t.length);const a=i(t,n,r,o-1);return a>r&&e(t[a-1],n)>-e(t[a],n)?a-1:a}return 1===t.length&&(e=(e,n)=>t(e)-n,n=T(t)),{left:i,center:o,right:r}};function T(t){return(e,n)=>L(t(e),n)}var D=function(t){return null===t?NaN:+t};const C=k(L),I=C.right;C.left,k(D).center;var A=I,Y=function(t,e,n){t.prototype=e.prototype=n,n.constructor=t};function O(t,e){var n=Object.create(t.prototype);for(var i in e)n[i]=e[i];return n}function P(){}var E=.7,N=1/E,R="\\s*([+-]?\\d+)\\s*",H="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)\\s*",j="\\s*([+-]?\\d*\\.?\\d+(?:[eE][+-]?\\d+)?)%\\s*",z=/^#([0-9a-f]{3,8})$/,W=new RegExp("^rgb\\("+[R,R,R]+"\\)$"),B=new RegExp("^rgb\\("+[j,j,j]+"\\)$"),F=new RegExp("^rgba\\("+[R,R,R,H]+"\\)$"),V=new RegExp("^rgba\\("+[j,j,j,H]+"\\)$"),G=new RegExp("^hsl\\("+[H,j,j]+"\\)$"),$=new RegExp("^hsla\\("+[H,j,j,H]+"\\)$"),U={aliceblue:15792383,antiquewhite:16444375,aqua:65535,aquamarine:8388564,azure:15794175,beige:16119260,bisque:16770244,black:0,blanchedalmond:16772045,blue:255,blueviolet:9055202,brown:10824234,burlywood:14596231,cadetblue:6266528,chartreuse:8388352,chocolate:13789470,coral:16744272,cornflowerblue:6591981,cornsilk:16775388,crimson:14423100,cyan:65535,darkblue:139,darkcyan:35723,darkgoldenrod:12092939,darkgray:11119017,darkgreen:25600,darkgrey:11119017,darkkhaki:12433259,darkmagenta:9109643,darkolivegreen:5597999,darkorange:16747520,darkorchid:10040012,darkred:9109504,darksalmon:15308410,darkseagreen:9419919,darkslateblue:4734347,darkslategray:3100495,darkslategrey:3100495,darkturquoise:52945,darkviolet:9699539,deeppink:16716947,deepskyblue:49151,dimgray:6908265,dimgrey:6908265,dodgerblue:2003199,firebrick:11674146,floralwhite:16775920,forestgreen:2263842,fuchsia:16711935,gainsboro:14474460,ghostwhite:16316671,gold:16766720,goldenrod:14329120,gray:8421504,green:32768,greenyellow:11403055,grey:8421504,honeydew:15794160,hotpink:16738740,indianred:13458524,indigo:4915330,ivory:16777200,khaki:15787660,lavender:15132410,lavenderblush:16773365,lawngreen:8190976,lemonchiffon:16775885,lightblue:11393254,lightcoral:15761536,lightcyan:14745599,lightgoldenrodyellow:16448210,lightgray:13882323,lightgreen:9498256,lightgrey:13882323,lightpink:16758465,lightsalmon:16752762,lightseagreen:2142890,lightskyblue:8900346,lightslategray:7833753,lightslategrey:7833753,lightsteelblue:11584734,lightyellow:16777184,lime:65280,limegreen:3329330,linen:16445670,magenta:16711935,maroon:8388608,mediumaquamarine:6737322,mediumblue:205,mediumorchid:12211667,mediumpurple:9662683,mediumseagreen:3978097,mediumslateblue:8087790,mediumspringgreen:64154,mediumturquoise:4772300,mediumvioletred:13047173,midnightblue:1644912,mintcream:16121850,mistyrose:16770273,moccasin:16770229,navajowhite:16768685,navy:128,oldlace:16643558,olive:8421376,olivedrab:7048739,orange:16753920,orangered:16729344,orchid:14315734,palegoldenrod:15657130,palegreen:10025880,paleturquoise:11529966,palevioletred:14381203,papayawhip:16773077,peachpuff:16767673,peru:13468991,pink:16761035,plum:14524637,powderblue:11591910,purple:8388736,rebeccapurple:6697881,red:16711680,rosybrown:12357519,royalblue:4286945,saddlebrown:9127187,salmon:16416882,sandybrown:16032864,seagreen:3050327,seashell:16774638,sienna:10506797,silver:12632256,skyblue:8900331,slateblue:6970061,slategray:7372944,slategrey:7372944,snow:16775930,springgreen:65407,steelblue:4620980,tan:13808780,teal:32896,thistle:14204888,tomato:16737095,turquoise:4251856,violet:15631086,wheat:16113331,white:16777215,whitesmoke:16119285,yellow:16776960,yellowgreen:10145074};function q(){return this.rgb().formatHex()}function X(){return st(this).formatHsl()}function Z(){return this.rgb().formatRgb()}function J(t){var e,n;return t=(t+"").trim().toLowerCase(),(e=z.exec(t))?(n=e[1].length,e=parseInt(e[1],16),6===n?K(e):3===n?new nt(e>>8&15|e>>4&240,e>>4&15|240&e,(15&e)<<4|15&e,1):8===n?Q(e>>24&255,e>>16&255,e>>8&255,(255&e)/255):4===n?Q(e>>12&15|e>>8&240,e>>8&15|e>>4&240,e>>4&15|240&e,((15&e)<<4|15&e)/255):null):(e=W.exec(t))?new nt(e[1],e[2],e[3],1):(e=B.exec(t))?new nt(255*e[1]/100,255*e[2]/100,255*e[3]/100,1):(e=F.exec(t))?Q(e[1],e[2],e[3],e[4]):(e=V.exec(t))?Q(255*e[1]/100,255*e[2]/100,255*e[3]/100,e[4]):(e=G.exec(t))?at(e[1],e[2]/100,e[3]/100,1):(e=$.exec(t))?at(e[1],e[2]/100,e[3]/100,e[4]):U.hasOwnProperty(t)?K(U[t]):"transparent"===t?new nt(NaN,NaN,NaN,0):null}function K(t){return new nt(t>>16&255,t>>8&255,255&t,1)}function Q(t,e,n,i){return i<=0&&(t=e=n=NaN),new nt(t,e,n,i)}function tt(t){return t instanceof P||(t=J(t)),t?(t=t.rgb(),new nt(t.r,t.g,t.b,t.opacity)):new nt}function et(t,e,n,i){return 1===arguments.length?tt(t):new nt(t,e,n,null==i?1:i)}function nt(t,e,n,i){this.r=+t,this.g=+e,this.b=+n,this.opacity=+i}function it(){return"#"+ot(this.r)+ot(this.g)+ot(this.b)}function rt(){var t=this.opacity;return t=isNaN(t)?1:Math.max(0,Math.min(1,t)),(1===t?"rgb(":"rgba(")+Math.max(0,Math.min(255,Math.round(this.r)||0))+", "+Math.max(0,Math.min(255,Math.round(this.g)||0))+", "+Math.max(0,Math.min(255,Math.round(this.b)||0))+(1===t?")":", "+t+")")}function ot(t){return t=Math.max(0,Math.min(255,Math.round(t)||0)),(t<16?"0":"")+t.toString(16)}function at(t,e,n,i){return i<=0?t=e=n=NaN:n<=0||n>=1?t=e=NaN:e<=0&&(t=NaN),new ut(t,e,n,i)}function st(t){if(t instanceof ut)return new ut(t.h,t.s,t.l,t.opacity);if(t instanceof P||(t=J(t)),!t)return new ut;if(t instanceof ut)return t;t=t.rgb();var e=t.r/255,n=t.g/255,i=t.b/255,r=Math.min(e,n,i),o=Math.max(e,n,i),a=NaN,s=o-r,l=(o+r)/2;return s?(a=e===o?(n-i)/s+6*(n<i):n===o?(i-e)/s+2:(e-n)/s+4,s/=l<.5?o+r:2-o-r,a*=60):s=l>0&&l<1?0:a,new ut(a,s,l,t.opacity)}function lt(t,e,n,i){return 1===arguments.length?st(t):new ut(t,e,n,null==i?1:i)}function ut(t,e,n,i){this.h=+t,this.s=+e,this.l=+n,this.opacity=+i}function ct(t,e,n){return 255*(t<60?e+(n-e)*t/60:t<180?n:t<240?e+(n-e)*(240-t)/60:e)}function dt(t,e,n,i,r){var o=t*t,a=o*t;return((1-3*t+3*o-a)*e+(4-6*o+3*a)*n+(1+3*t+3*o-3*a)*i+a*r)/6}Y(P,J,{copy:function(t){return Object.assign(new this.constructor,this,t)},displayable:function(){return this.rgb().displayable()},hex:q,formatHex:q,formatHsl:X,formatRgb:Z,toString:Z}),Y(nt,et,O(P,{brighter:function(t){return t=null==t?N:Math.pow(N,t),new nt(this.r*t,this.g*t,this.b*t,this.opacity)},darker:function(t){return t=null==t?E:Math.pow(E,t),new nt(this.r*t,this.g*t,this.b*t,this.opacity)},rgb:function(){return this},displayable:function(){return-.5<=this.r&&this.r<255.5&&-.5<=this.g&&this.g<255.5&&-.5<=this.b&&this.b<255.5&&0<=this.opacity&&this.opacity<=1},hex:it,formatHex:it,formatRgb:rt,toString:rt})),Y(ut,lt,O(P,{brighter:function(t){return t=null==t?N:Math.pow(N,t),new ut(this.h,this.s,this.l*t,this.opacity)},darker:function(t){return t=null==t?E:Math.pow(E,t),new ut(this.h,this.s,this.l*t,this.opacity)},rgb:function(){var t=this.h%360+360*(this.h<0),e=isNaN(t)||isNaN(this.s)?0:this.s,n=this.l,i=n+(n<.5?n:1-n)*e,r=2*n-i;return new nt(ct(t>=240?t-240:t+120,r,i),ct(t,r,i),ct(t<120?t+240:t-120,r,i),this.opacity)},displayable:function(){return(0<=this.s&&this.s<=1||isNaN(this.s))&&0<=this.l&&this.l<=1&&0<=this.opacity&&this.opacity<=1},formatHsl:function(){var t=this.opacity;return t=isNaN(t)?1:Math.max(0,Math.min(1,t)),(1===t?"hsl(":"hsla(")+(this.h||0)+", "+100*(this.s||0)+"%, "+100*(this.l||0)+"%"+(1===t?")":", "+t+")")}}));var ht=function(t){var e=t.length-1;return function(n){var i=n<=0?n=0:n>=1?(n=1,e-1):Math.floor(n*e),r=t[i],o=t[i+1],a=i>0?t[i-1]:2*r-o,s=i<e-1?t[i+2]:2*o-r;return dt((n-i/e)*e,a,r,o,s)}},ft=function(t){var e=t.length;return function(n){var i=Math.floor(((n%=1)<0?++n:n)*e),r=t[(i+e-1)%e],o=t[i%e],a=t[(i+1)%e],s=t[(i+2)%e];return dt((n-i/e)*e,r,o,a,s)}},pt=t=>()=>t;function mt(t,e){return function(n){return t+n*e}}function gt(t,e,n){return t=Math.pow(t,n),e=Math.pow(e,n)-t,n=1/n,function(i){return Math.pow(t+i*e,n)}}function yt(t){return 1===(t=+t)?vt:function(e,n){return n-e?gt(e,n,t):pt(isNaN(e)?n:e)}}function vt(t,e){var n=e-t;return n?mt(t,n):pt(isNaN(t)?e:t)}var _t=function t(e){var n=yt(e);function i(t,e){var i=n((t=et(t)).r,(e=et(e)).r),r=n(t.g,e.g),o=n(t.b,e.b),a=vt(t.opacity,e.opacity);return function(e){return t.r=i(e),t.g=r(e),t.b=o(e),t.opacity=a(e),t+""}}return i.gamma=t,i}(1);function bt(t){return function(e){var n,i,r=e.length,o=new Array(r),a=new Array(r),s=new Array(r);for(n=0;n<r;++n)i=et(e[n]),o[n]=i.r||0,a[n]=i.g||0,s[n]=i.b||0;return o=t(o),a=t(a),s=t(s),i.opacity=1,function(t){return i.r=o(t),i.g=a(t),i.b=s(t),i+""}}}bt(ht),bt(ft);var xt=function(t,e){e||(e=[]);var n,i=t?Math.min(e.length,t.length):0,r=e.slice();return function(o){for(n=0;n<i;++n)r[n]=t[n]*(1-o)+e[n]*o;return r}};function wt(t){return ArrayBuffer.isView(t)&&!(t instanceof DataView)}function Mt(t,e){var n,i=e?e.length:0,r=t?Math.min(i,t.length):0,o=new Array(r),a=new Array(i);for(n=0;n<r;++n)o[n]=Yt(t[n],e[n]);for(;n<i;++n)a[n]=e[n];return function(t){for(n=0;n<r;++n)a[n]=o[n](t);return a}}var St=function(t,e){var n=new Date;return t=+t,e=+e,function(i){return n.setTime(t*(1-i)+e*i),n}},Lt=function(t,e){return t=+t,e=+e,function(n){return t*(1-n)+e*n}},kt=function(t,e){var n,i={},r={};for(n in null!==t&&"object"===typeof t||(t={}),null!==e&&"object"===typeof e||(e={}),e)n in t?i[n]=Yt(t[n],e[n]):r[n]=e[n];return function(t){for(n in i)r[n]=i[n](t);return r}},Tt=/[-+]?(?:\d+\.?\d*|\.?\d+)(?:[eE][-+]?\d+)?/g,Dt=new RegExp(Tt.source,"g");function Ct(t){return function(){return t}}function It(t){return function(e){return t(e)+""}}var At=function(t,e){var n,i,r,o=Tt.lastIndex=Dt.lastIndex=0,a=-1,s=[],l=[];t+="",e+="";while((n=Tt.exec(t))&&(i=Dt.exec(e)))(r=i.index)>o&&(r=e.slice(o,r),s[a]?s[a]+=r:s[++a]=r),(n=n[0])===(i=i[0])?s[a]?s[a]+=i:s[++a]=i:(s[++a]=null,l.push({i:a,x:Lt(n,i)})),o=Dt.lastIndex;return o<e.length&&(r=e.slice(o),s[a]?s[a]+=r:s[++a]=r),s.length<2?l[0]?It(l[0].x):Ct(e):(e=l.length,function(t){for(var n,i=0;i<e;++i)s[(n=l[i]).i]=n.x(t);return s.join("")})},Yt=function(t,e){var n,i=typeof e;return null==e||"boolean"===i?pt(e):("number"===i?Lt:"string"===i?(n=J(e))?(e=n,_t):At:e instanceof J?_t:e instanceof Date?St:wt(e)?xt:Array.isArray(e)?Mt:"function"!==typeof e.valueOf&&"function"!==typeof e.toString||isNaN(e)?kt:Lt)(t,e)},Ot=function(t,e){return t=+t,e=+e,function(n){return Math.round(t*(1-n)+e*n)}};function Pt(t){return function(){return t}}function Et(t){return+t}var Nt=[0,1];function Rt(t){return t}function Ht(t,e){return(e-=t=+t)?function(n){return(n-t)/e}:Pt(isNaN(e)?NaN:.5)}function jt(t,e){var n;return t>e&&(n=t,t=e,e=n),function(n){return Math.max(t,Math.min(e,n))}}function zt(t,e,n){var i=t[0],r=t[1],o=e[0],a=e[1];return r<i?(i=Ht(r,i),o=n(a,o)):(i=Ht(i,r),o=n(o,a)),function(t){return o(i(t))}}function Wt(t,e,n){var i=Math.min(t.length,e.length)-1,r=new Array(i),o=new Array(i),a=-1;t[i]<t[0]&&(t=t.slice().reverse(),e=e.slice().reverse());while(++a<i)r[a]=Ht(t[a],t[a+1]),o[a]=n(e[a],e[a+1]);return function(e){var n=A(t,e,1,i)-1;return o[n](r[n](e))}}function Bt(t,e){return e.domain(t.domain()).range(t.range()).interpolate(t.interpolate()).clamp(t.clamp()).unknown(t.unknown())}function Ft(){var t,e,n,i,r,o,a=Nt,s=Nt,l=Yt,u=Rt;function c(){var t=Math.min(a.length,s.length);return u!==Rt&&(u=jt(a[0],a[t-1])),i=t>2?Wt:zt,r=o=null,d}function d(e){return null==e||isNaN(e=+e)?n:(r||(r=i(a.map(t),s,l)))(t(u(e)))}return d.invert=function(n){return u(e((o||(o=i(s,a.map(t),Lt)))(n)))},d.domain=function(t){return arguments.length?(a=Array.from(t,Et),c()):a.slice()},d.range=function(t){return arguments.length?(s=Array.from(t),c()):s.slice()},d.rangeRound=function(t){return s=Array.from(t),l=Ot,c()},d.clamp=function(t){return arguments.length?(u=!!t||Rt,c()):u!==Rt},d.interpolate=function(t){return arguments.length?(l=t,c()):l},d.unknown=function(t){return arguments.length?(n=t,d):n},function(n,i){return t=n,e=i,c()}}function Vt(){return Ft()(Rt,Rt)}function Gt(t,e){switch(arguments.length){case 0:break;case 1:this.range(t);break;default:this.range(e).domain(t);break}return this}var $t=/^(?:(.)?([<>=^]))?([+\-( ])?([$#])?(0)?(\d+)?(,)?(\.\d+)?(~)?([a-z%])?$/i;function Ut(t){if(!(e=$t.exec(t)))throw new Error("invalid format: "+t);var e;return new qt({fill:e[1],align:e[2],sign:e[3],symbol:e[4],zero:e[5],width:e[6],comma:e[7],precision:e[8]&&e[8].slice(1),trim:e[9],type:e[10]})}function qt(t){this.fill=void 0===t.fill?" ":t.fill+"",this.align=void 0===t.align?">":t.align+"",this.sign=void 0===t.sign?"-":t.sign+"",this.symbol=void 0===t.symbol?"":t.symbol+"",this.zero=!!t.zero,this.width=void 0===t.width?void 0:+t.width,this.comma=!!t.comma,this.precision=void 0===t.precision?void 0:+t.precision,this.trim=!!t.trim,this.type=void 0===t.type?"":t.type+""}Ut.prototype=qt.prototype,qt.prototype.toString=function(){return this.fill+this.align+this.sign+this.symbol+(this.zero?"0":"")+(void 0===this.width?"":Math.max(1,0|this.width))+(this.comma?",":"")+(void 0===this.precision?"":"."+Math.max(0,0|this.precision))+(this.trim?"~":"")+this.type};var Xt=function(t){return Math.abs(t=Math.round(t))>=1e21?t.toLocaleString("en").replace(/,/g,""):t.toString(10)};function Zt(t,e){if((n=(t=e?t.toExponential(e-1):t.toExponential()).indexOf("e"))<0)return null;var n,i=t.slice(0,n);return[i.length>1?i[0]+i.slice(2):i,+t.slice(n+1)]}var Jt,Kt,Qt,te,ee=function(t){return t=Zt(Math.abs(t)),t?t[1]:NaN},ne=function(t,e){return Math.max(0,3*Math.max(-8,Math.min(8,Math.floor(ee(e)/3)))-ee(Math.abs(t)))},ie=function(t,e){return function(n,i){var r=n.length,o=[],a=0,s=t[0],l=0;while(r>0&&s>0){if(l+s+1>i&&(s=Math.max(1,i-l)),o.push(n.substring(r-=s,r+s)),(l+=s+1)>i)break;s=t[a=(a+1)%t.length]}return o.reverse().join(e)}},re=function(t){return function(e){return e.replace(/[0-9]/g,(function(e){return t[+e]}))}},oe=function(t){t:for(var e,n=t.length,i=1,r=-1;i<n;++i)switch(t[i]){case".":r=e=i;break;case"0":0===r&&(r=i),e=i;break;default:if(!+t[i])break t;r>0&&(r=0);break}return r>0?t.slice(0,r)+t.slice(e+1):t},ae=function(t,e){var n=Zt(t,e);if(!n)return t+"";var i=n[0],r=n[1],o=r-(Jt=3*Math.max(-8,Math.min(8,Math.floor(r/3))))+1,a=i.length;return o===a?i:o>a?i+new Array(o-a+1).join("0"):o>0?i.slice(0,o)+"."+i.slice(o):"0."+new Array(1-o).join("0")+Zt(t,Math.max(0,e+o-1))[0]},se=function(t,e){var n=Zt(t,e);if(!n)return t+"";var i=n[0],r=n[1];return r<0?"0."+new Array(-r).join("0")+i:i.length>r+1?i.slice(0,r+1)+"."+i.slice(r+1):i+new Array(r-i.length+2).join("0")},le={"%":(t,e)=>(100*t).toFixed(e),b:t=>Math.round(t).toString(2),c:t=>t+"",d:Xt,e:(t,e)=>t.toExponential(e),f:(t,e)=>t.toFixed(e),g:(t,e)=>t.toPrecision(e),o:t=>Math.round(t).toString(8),p:(t,e)=>se(100*t,e),r:se,s:ae,X:t=>Math.round(t).toString(16).toUpperCase(),x:t=>Math.round(t).toString(16)},ue=function(t){return t},ce=Array.prototype.map,de=["y","z","a","f","p","n","µ","m","","k","M","G","T","P","E","Z","Y"],he=function(t){var e=void 0===t.grouping||void 0===t.thousands?ue:ie(ce.call(t.grouping,Number),t.thousands+""),n=void 0===t.currency?"":t.currency[0]+"",i=void 0===t.currency?"":t.currency[1]+"",r=void 0===t.decimal?".":t.decimal+"",o=void 0===t.numerals?ue:re(ce.call(t.numerals,String)),a=void 0===t.percent?"%":t.percent+"",s=void 0===t.minus?"−":t.minus+"",l=void 0===t.nan?"NaN":t.nan+"";function u(t){t=Ut(t);var u=t.fill,c=t.align,d=t.sign,h=t.symbol,f=t.zero,p=t.width,m=t.comma,g=t.precision,y=t.trim,v=t.type;"n"===v?(m=!0,v="g"):le[v]||(void 0===g&&(g=12),y=!0,v="g"),(f||"0"===u&&"="===c)&&(f=!0,u="0",c="=");var _="$"===h?n:"#"===h&&/[boxX]/.test(v)?"0"+v.toLowerCase():"",b="$"===h?i:/[%p]/.test(v)?a:"",x=le[v],w=/[defgprs%]/.test(v);function M(t){var n,i,a,h=_,M=b;if("c"===v)M=x(t)+M,t="";else{t=+t;var S=t<0||1/t<0;if(t=isNaN(t)?l:x(Math.abs(t),g),y&&(t=oe(t)),S&&0===+t&&"+"!==d&&(S=!1),h=(S?"("===d?d:s:"-"===d||"("===d?"":d)+h,M=("s"===v?de[8+Jt/3]:"")+M+(S&&"("===d?")":""),w){n=-1,i=t.length;while(++n<i)if(a=t.charCodeAt(n),48>a||a>57){M=(46===a?r+t.slice(n+1):t.slice(n))+M,t=t.slice(0,n);break}}}m&&!f&&(t=e(t,1/0));var L=h.length+t.length+M.length,k=L<p?new Array(p-L+1).join(u):"";switch(m&&f&&(t=e(k+t,k.length?p-M.length:1/0),k=""),c){case"<":t=h+t+M+k;break;case"=":t=h+k+t+M;break;case"^":t=k.slice(0,L=k.length>>1)+h+t+M+k.slice(L);break;default:t=k+h+t+M;break}return o(t)}return g=void 0===g?6:/[gprs]/.test(v)?Math.max(1,Math.min(21,g)):Math.max(0,Math.min(20,g)),M.toString=function(){return t+""},M}function c(t,e){var n=u((t=Ut(t),t.type="f",t)),i=3*Math.max(-8,Math.min(8,Math.floor(ee(e)/3))),r=Math.pow(10,-i),o=de[8+i/3];return function(t){return n(r*t)+o}}return{format:u,formatPrefix:c}};function fe(t){return Kt=he(t),Qt=Kt.format,te=Kt.formatPrefix,Kt}fe({thousands:",",grouping:[3],currency:["$",""]});var pe=function(t,e){return t=Math.abs(t),e=Math.abs(e)-t,Math.max(0,ee(e)-ee(t))+1},me=function(t){return Math.max(0,-ee(Math.abs(t)))};function ge(t,e,n,i){var r,o=S(t,e,n);switch(i=Ut(null==i?",f":i),i.type){case"s":var a=Math.max(Math.abs(t),Math.abs(e));return null!=i.precision||isNaN(r=ne(o,a))||(i.precision=r),te(i,a);case"":case"e":case"g":case"p":case"r":null!=i.precision||isNaN(r=pe(o,Math.max(Math.abs(t),Math.abs(e))))||(i.precision=r-("e"===i.type));break;case"f":case"%":null!=i.precision||isNaN(r=me(o))||(i.precision=r-2*("%"===i.type));break}return Qt(i)}function ye(t){var e=t.domain;return t.ticks=function(t){var n=e();return w(n[0],n[n.length-1],null==t?10:t)},t.tickFormat=function(t,n){var i=e();return ge(i[0],i[i.length-1],null==t?10:t,n)},t.nice=function(n){null==n&&(n=10);var i,r,o=e(),a=0,s=o.length-1,l=o[a],u=o[s],c=10;u<l&&(r=l,l=u,u=r,r=a,a=s,s=r);while(c-- >0){if(r=M(l,u,n),r===i)return o[a]=l,o[s]=u,e(o);if(r>0)l=Math.floor(l/r)*r,u=Math.ceil(u/r)*r;else{if(!(r<0))break;l=Math.ceil(l*r)/r,u=Math.floor(u*r)/r}i=r}return t},t}function ve(){var t=Vt();return t.copy=function(){return Bt(t,ve())},Gt.apply(t,arguments),ye(t)}function _e(){}var be=function(t){return null==t?_e:function(){return this.querySelector(t)}},xe=function(t){"function"!==typeof t&&(t=be(t));for(var e=this._groups,n=e.length,i=new Array(n),r=0;r<n;++r)for(var o,a,s=e[r],l=s.length,u=i[r]=new Array(l),c=0;c<l;++c)(o=s[c])&&(a=t.call(o,o.__data__,c,s))&&("__data__"in o&&(a.__data__=o.__data__),u[c]=a);return new mi(i,this._parents)},we=function(t){return"object"===typeof t&&"length"in t?t:Array.from(t)};function Me(){return[]}var Se=function(t){return null==t?Me:function(){return this.querySelectorAll(t)}};function Le(t){return function(){var e=t.apply(this,arguments);return null==e?[]:we(e)}}var ke=function(t){t="function"===typeof t?Le(t):Se(t);for(var e=this._groups,n=e.length,i=[],r=[],o=0;o<n;++o)for(var a,s=e[o],l=s.length,u=0;u<l;++u)(a=s[u])&&(i.push(t.call(a,a.__data__,u,s)),r.push(a));return new mi(i,r)},Te=function(t){return function(){return this.matches(t)}};function De(t){return function(e){return e.matches(t)}}var Ce=Array.prototype.find;function Ie(t){return function(){return Ce.call(this.children,t)}}function Ae(){return this.firstElementChild}var Ye=function(t){return this.select(null==t?Ae:Ie("function"===typeof t?t:De(t)))},Oe=Array.prototype.filter;function Pe(){return this.children}function Ee(t){return function(){return Oe.call(this.children,t)}}var Ne=function(t){return this.selectAll(null==t?Pe:Ee("function"===typeof t?t:De(t)))},Re=function(t){"function"!==typeof t&&(t=Te(t));for(var e=this._groups,n=e.length,i=new Array(n),r=0;r<n;++r)for(var o,a=e[r],s=a.length,l=i[r]=[],u=0;u<s;++u)(o=a[u])&&t.call(o,o.__data__,u,a)&&l.push(o);return new mi(i,this._parents)},He=function(t){return new Array(t.length)},je=function(){return new mi(this._enter||this._groups.map(He),this._parents)};function ze(t,e){this.ownerDocument=t.ownerDocument,this.namespaceURI=t.namespaceURI,this._next=null,this._parent=t,this.__data__=e}ze.prototype={constructor:ze,appendChild:function(t){return this._parent.insertBefore(t,this._next)},insertBefore:function(t,e){return this._parent.insertBefore(t,e)},querySelector:function(t){return this._parent.querySelector(t)},querySelectorAll:function(t){return this._parent.querySelectorAll(t)}};var We=function(t){return function(){return t}};function Be(t,e,n,i,r,o){for(var a,s=0,l=e.length,u=o.length;s<u;++s)(a=e[s])?(a.__data__=o[s],i[s]=a):n[s]=new ze(t,o[s]);for(;s<l;++s)(a=e[s])&&(r[s]=a)}function Fe(t,e,n,i,r,o,a){var s,l,u,c=new Map,d=e.length,h=o.length,f=new Array(d);for(s=0;s<d;++s)(l=e[s])&&(f[s]=u=a.call(l,l.__data__,s,e)+"",c.has(u)?r[s]=l:c.set(u,l));for(s=0;s<h;++s)u=a.call(t,o[s],s,o)+"",(l=c.get(u))?(i[s]=l,l.__data__=o[s],c.delete(u)):n[s]=new ze(t,o[s]);for(s=0;s<d;++s)(l=e[s])&&c.get(f[s])===l&&(r[s]=l)}function Ve(t){return t.__data__}var Ge=function(t,e){if(!arguments.length)return Array.from(this,Ve);var n=e?Fe:Be,i=this._parents,r=this._groups;"function"!==typeof t&&(t=We(t));for(var o=r.length,a=new Array(o),s=new Array(o),l=new Array(o),u=0;u<o;++u){var c=i[u],d=r[u],h=d.length,f=we(t.call(c,c&&c.__data__,u,i)),p=f.length,m=s[u]=new Array(p),g=a[u]=new Array(p),y=l[u]=new Array(h);n(c,d,m,g,y,f,e);for(var v,_,b=0,x=0;b<p;++b)if(v=m[b]){b>=x&&(x=b+1);while(!(_=g[x])&&++x<p);v._next=_||null}}return a=new mi(a,i),a._enter=s,a._exit=l,a},$e=function(){return new mi(this._exit||this._groups.map(He),this._parents)},Ue=function(t,e,n){var i=this.enter(),r=this,o=this.exit();return i="function"===typeof t?t(i):i.append(t+""),null!=e&&(r=e(r)),null==n?o.remove():n(o),i&&r?i.merge(r).order():r},qe=function(t){if(!(t instanceof mi))throw new Error("invalid merge");for(var e=this._groups,n=t._groups,i=e.length,r=n.length,o=Math.min(i,r),a=new Array(i),s=0;s<o;++s)for(var l,u=e[s],c=n[s],d=u.length,h=a[s]=new Array(d),f=0;f<d;++f)(l=u[f]||c[f])&&(h[f]=l);for(;s<i;++s)a[s]=e[s];return new mi(a,this._parents)},Xe=function(){for(var t=this._groups,e=-1,n=t.length;++e<n;)for(var i,r=t[e],o=r.length-1,a=r[o];--o>=0;)(i=r[o])&&(a&&4^i.compareDocumentPosition(a)&&a.parentNode.insertBefore(i,a),a=i);return this},Ze=function(t){function e(e,n){return e&&n?t(e.__data__,n.__data__):!e-!n}t||(t=Je);for(var n=this._groups,i=n.length,r=new Array(i),o=0;o<i;++o){for(var a,s=n[o],l=s.length,u=r[o]=new Array(l),c=0;c<l;++c)(a=s[c])&&(u[c]=a);u.sort(e)}return new mi(r,this._parents).order()};function Je(t,e){return t<e?-1:t>e?1:t>=e?0:NaN}var Ke=function(){var t=arguments[0];return arguments[0]=this,t.apply(null,arguments),this},Qe=function(){return Array.from(this)},tn=function(){for(var t=this._groups,e=0,n=t.length;e<n;++e)for(var i=t[e],r=0,o=i.length;r<o;++r){var a=i[r];if(a)return a}return null},en=function(){let t=0;for(const e of this)++t;return t},nn=function(){return!this.node()},rn=function(t){for(var e=this._groups,n=0,i=e.length;n<i;++n)for(var r,o=e[n],a=0,s=o.length;a<s;++a)(r=o[a])&&t.call(r,r.__data__,a,o);return this},on="http://www.w3.org/1999/xhtml",an={svg:"http://www.w3.org/2000/svg",xhtml:on,xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/"},sn=function(t){var e=t+="",n=e.indexOf(":");return n>=0&&"xmlns"!==(e=t.slice(0,n))&&(t=t.slice(n+1)),an.hasOwnProperty(e)?{space:an[e],local:t}:t};function ln(t){return function(){this.removeAttribute(t)}}function un(t){return function(){this.removeAttributeNS(t.space,t.local)}}function cn(t,e){return function(){this.setAttribute(t,e)}}function dn(t,e){return function(){this.setAttributeNS(t.space,t.local,e)}}function hn(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttribute(t):this.setAttribute(t,n)}}function fn(t,e){return function(){var n=e.apply(this,arguments);null==n?this.removeAttributeNS(t.space,t.local):this.setAttributeNS(t.space,t.local,n)}}var pn=function(t,e){var n=sn(t);if(arguments.length<2){var i=this.node();return n.local?i.getAttributeNS(n.space,n.local):i.getAttribute(n)}return this.each((null==e?n.local?un:ln:"function"===typeof e?n.local?fn:hn:n.local?dn:cn)(n,e))},mn=function(t){return t.ownerDocument&&t.ownerDocument.defaultView||t.document&&t||t.defaultView};function gn(t){return function(){this.style.removeProperty(t)}}function yn(t,e,n){return function(){this.style.setProperty(t,e,n)}}function vn(t,e,n){return function(){var i=e.apply(this,arguments);null==i?this.style.removeProperty(t):this.style.setProperty(t,i,n)}}var _n=function(t,e,n){return arguments.length>1?this.each((null==e?gn:"function"===typeof e?vn:yn)(t,e,null==n?"":n)):bn(this.node(),t)};function bn(t,e){return t.style.getPropertyValue(e)||mn(t).getComputedStyle(t,null).getPropertyValue(e)}function xn(t){return function(){delete this[t]}}function wn(t,e){return function(){this[t]=e}}function Mn(t,e){return function(){var n=e.apply(this,arguments);null==n?delete this[t]:this[t]=n}}var Sn=function(t,e){return arguments.length>1?this.each((null==e?xn:"function"===typeof e?Mn:wn)(t,e)):this.node()[t]};function Ln(t){return t.trim().split(/^|\s+/)}function kn(t){return t.classList||new Tn(t)}function Tn(t){this._node=t,this._names=Ln(t.getAttribute("class")||"")}function Dn(t,e){var n=kn(t),i=-1,r=e.length;while(++i<r)n.add(e[i])}function Cn(t,e){var n=kn(t),i=-1,r=e.length;while(++i<r)n.remove(e[i])}function In(t){return function(){Dn(this,t)}}function An(t){return function(){Cn(this,t)}}function Yn(t,e){return function(){(e.apply(this,arguments)?Dn:Cn)(this,t)}}Tn.prototype={add:function(t){var e=this._names.indexOf(t);e<0&&(this._names.push(t),this._node.setAttribute("class",this._names.join(" ")))},remove:function(t){var e=this._names.indexOf(t);e>=0&&(this._names.splice(e,1),this._node.setAttribute("class",this._names.join(" ")))},contains:function(t){return this._names.indexOf(t)>=0}};var On=function(t,e){var n=Ln(t+"");if(arguments.length<2){var i=kn(this.node()),r=-1,o=n.length;while(++r<o)if(!i.contains(n[r]))return!1;return!0}return this.each(("function"===typeof e?Yn:e?In:An)(n,e))};function Pn(){this.textContent=""}function En(t){return function(){this.textContent=t}}function Nn(t){return function(){var e=t.apply(this,arguments);this.textContent=null==e?"":e}}var Rn=function(t){return arguments.length?this.each(null==t?Pn:("function"===typeof t?Nn:En)(t)):this.node().textContent};function Hn(){this.innerHTML=""}function jn(t){return function(){this.innerHTML=t}}function zn(t){return function(){var e=t.apply(this,arguments);this.innerHTML=null==e?"":e}}var Wn=function(t){return arguments.length?this.each(null==t?Hn:("function"===typeof t?zn:jn)(t)):this.node().innerHTML};function Bn(){this.nextSibling&&this.parentNode.appendChild(this)}var Fn=function(){return this.each(Bn)};function Vn(){this.previousSibling&&this.parentNode.insertBefore(this,this.parentNode.firstChild)}var Gn=function(){return this.each(Vn)};function $n(t){return function(){var e=this.ownerDocument,n=this.namespaceURI;return n===on&&e.documentElement.namespaceURI===on?e.createElement(t):e.createElementNS(n,t)}}function Un(t){return function(){return this.ownerDocument.createElementNS(t.space,t.local)}}var qn=function(t){var e=sn(t);return(e.local?Un:$n)(e)},Xn=function(t){var e="function"===typeof t?t:qn(t);return this.select((function(){return this.appendChild(e.apply(this,arguments))}))};function Zn(){return null}var Jn=function(t,e){var n="function"===typeof t?t:qn(t),i=null==e?Zn:"function"===typeof e?e:be(e);return this.select((function(){return this.insertBefore(n.apply(this,arguments),i.apply(this,arguments)||null)}))};function Kn(){var t=this.parentNode;t&&t.removeChild(this)}var Qn=function(){return this.each(Kn)};function ti(){var t=this.cloneNode(!1),e=this.parentNode;return e?e.insertBefore(t,this.nextSibling):t}function ei(){var t=this.cloneNode(!0),e=this.parentNode;return e?e.insertBefore(t,this.nextSibling):t}var ni=function(t){return this.select(t?ei:ti)},ii=function(t){return arguments.length?this.property("__data__",t):this.node().__data__};function ri(t){return function(e){t.call(this,e,this.__data__)}}function oi(t){return t.trim().split(/^|\s+/).map((function(t){var e="",n=t.indexOf(".");return n>=0&&(e=t.slice(n+1),t=t.slice(0,n)),{type:t,name:e}}))}function ai(t){return function(){var e=this.__on;if(e){for(var n,i=0,r=-1,o=e.length;i<o;++i)n=e[i],t.type&&n.type!==t.type||n.name!==t.name?e[++r]=n:this.removeEventListener(n.type,n.listener,n.options);++r?e.length=r:delete this.__on}}}function si(t,e,n){return function(){var i,r=this.__on,o=ri(e);if(r)for(var a=0,s=r.length;a<s;++a)if((i=r[a]).type===t.type&&i.name===t.name)return this.removeEventListener(i.type,i.listener,i.options),this.addEventListener(i.type,i.listener=o,i.options=n),void(i.value=e);this.addEventListener(t.type,o,n),i={type:t.type,name:t.name,value:e,listener:o,options:n},r?r.push(i):this.__on=[i]}}var li=function(t,e,n){var i,r,o=oi(t+""),a=o.length;if(!(arguments.length<2)){for(s=e?si:ai,i=0;i<a;++i)this.each(s(o[i],e,n));return this}var s=this.node().__on;if(s)for(var l,u=0,c=s.length;u<c;++u)for(i=0,l=s[u];i<a;++i)if((r=o[i]).type===l.type&&r.name===l.name)return l.value};function ui(t,e,n){var i=mn(t),r=i.CustomEvent;"function"===typeof r?r=new r(e,n):(r=i.document.createEvent("Event"),n?(r.initEvent(e,n.bubbles,n.cancelable),r.detail=n.detail):r.initEvent(e,!1,!1)),t.dispatchEvent(r)}function ci(t,e){return function(){return ui(this,t,e)}}function di(t,e){return function(){return ui(this,t,e.apply(this,arguments))}}var hi=function(t,e){return this.each(("function"===typeof e?di:ci)(t,e))},fi=function*(){for(var t=this._groups,e=0,n=t.length;e<n;++e)for(var i,r=t[e],o=0,a=r.length;o<a;++o)(i=r[o])&&(yield i)},pi=[null];function mi(t,e){this._groups=t,this._parents=e}function gi(){return new mi([[document.documentElement]],pi)}function yi(){return this}mi.prototype=gi.prototype={constructor:mi,select:xe,selectAll:ke,selectChild:Ye,selectChildren:Ne,filter:Re,data:Ge,enter:je,exit:$e,join:Ue,merge:qe,selection:yi,order:Xe,sort:Ze,call:Ke,nodes:Qe,node:tn,size:en,empty:nn,each:rn,attr:pn,style:_n,property:Sn,classed:On,text:Rn,html:Wn,raise:Fn,lower:Gn,append:Xn,insert:Jn,remove:Qn,clone:ni,datum:ii,on:li,dispatch:hi,[Symbol.iterator]:fi};var vi=function(t){return"string"===typeof t?new mi([[document.querySelector(t)]],[document.documentElement]):new mi([[t]],pi)};function _i(){return window.devicePixelRatio||1}function bi(t=12,e="normal",n="Helvetica Neue"){return`${e} ${t*_i()}px ${n}`}function xi(t,e,n,{color:i,lineWidth:r}){const o=_i();wi(t,`M ${e.x*o} ${e.y*o} L ${n.x*o} ${n.y*o}`,{lineWidth:r||1,strokeStyle:i||"#FF0000"})}function wi(t,e,{lineWidth:n,lineDash:i,strokeStyle:r,fillStyle:o}){t.beginPath(),i&&t.setLineDash(i),t.lineWidth=n||_i(),t.strokeStyle=r||"none",t.fillStyle=o||"none",t[r?"stroke":"fill"](new Path2D(e))}function Mi(t){let e="string"==typeof t?t:"";if(n=t,Number.isInteger(Number(n))){const n=t.toString(16).padStart(8,"F"),i=parseInt(n.slice(0,2),16)/255;e=`rgba(${parseInt(n.slice(2,4),16)}, ${parseInt(n.slice(4,6),16)}, ${parseInt(n.slice(6),16)}, ${i})`}var n;return e}class Si{constructor({chart:t,boardId:e,width:n,height:i,align:r,container:o,priceDecs:a}){this.chart=t,this.width=n,this.height=i,this.boardId=e,this.container=o.dom,this.align=r||"right",this.priceDecs=a||0,this.scale=ve().rangeRound([this.height,0]),this.axis="left"===this.align?v(this.scale):g(this.scale),this.rootElement=vi(this.container).append("svg").attr("class","axis").attr("width",this.width).attr("height",this.height),this.rootG=this.rootElement.append("g").attr("class","y axis "+this.align),this.annotationG=this.rootElement.append("g"),this.annotationRect=this.annotationG.append("rect").attr("fill","#fbd3a0"),this.annotationText=this.annotationG.append("text"),"left"===this.align&&this.rootG.attr("transform","translate("+this.width+",0)")}update({width:t,height:e}){t&&t!=this.width&&(this.width=t),e&&e!=this.height&&(this.height=e),this.scale.rangeRound([this.height,0]),this.rootElement.attr("width",this.width).attr("height",this.height)}getTicks(){const t=[],e=this;return this.rootG.selectAll(".tick").each((function(n){t.push(Math.round(e.scale(n)))})),t}draw(t){this.scale.domain(t),this.axis.ticks("main"===this.boardId?6:3,`.${this.priceDecs}f`),this.rootG.call(this.axis)}showYAnnotation(t){const e=this.scale.invert(t).toFixed(this.priceDecs);isNaN(e)?this.annotationG.attr("visibility","hidden"):(this.annotationG.attr("visibility","visible"),this.annotationRect.attr("width",this.width-2).attr("height",12).attr("transform","translate(1,"+(t-6)+")"),this.annotationText.text(e).attr("transform","translate(3,"+(t+5)+")"))}hideYAnnotation(){this.annotationG.attr("visibility","hidden")}}class Li{constructor(t,e,n){this.fn=t,this.context=e,this.once=n||!1}}class ki{constructor(){this._events={}}addListener(t,e,n,i){if("function"!=typeof e)throw new TypeError("The listener must be a function");const r=new Li(e,n||this,i||!1);this._events[t]?this._events[t].push(r):this._events[t]=[r]}once(t,e,n){this.addListener(t,e,n||this,!0)}removeListener(t,e){if("function"!=typeof e)throw new TypeError("The listener must be a function");if(!this._events[t])return;let n=!1;for(const i in this._events[t])e==this._events[t][i].fn&&(n=!0,this._events[t][i]=void 0);n&&(this._events[t]=this._events[t].filter(t=>t)),0==this._events[t].length&&delete this._events[t]}removeAllListeners(t){this._events[t]&&delete this._events[t]}emit(t,...e){if(!this._events[t])return;let n=!1;for(const i in this._events[t]){const r=this._events[t][i];r.once&&(n=!0,this._events[t][i]=void 0),r.fn.call(r.context,...e)}n&&(this._events[t]=this._events[t].filter(t=>t),0==this._events[t].length&&delete this._events[t])}getEventNames(){return Object.keys(this._events)}getListeners(t){if(!this._events[t])return[];const e=[];for(const n of this._events[t])e.push(n);return e}}ki.prototype.off=ki.prototype.removeListener,ki.prototype.on=ki.prototype.addListener;class Ti{constructor({id:t,chart:e,board:n,yAlign:i}){this.id=t||(new Date).getTime(),this.chart=e,this.board=n,this.boardId=this.board.boardId,this.yAlign=i||"right",this._show=!0}update(t){this.xPos1=void 0===t.xPos1?this.xPos1:t.xPos1,this.x1=void 0===t.x1?this.x1:t.x1,this.dt1=void 0===t.dt1?this.dt1:t.dt1,this.yPos1=void 0===t.yPos1?this.yPos1:t.yPos1,this.y1=void 0===t.y1?this.y1:t.y1,this.xPos2=void 0===t.xPos2?this.xPos2:t.xPos2,this.x2=void 0===t.x2?this.x2:t.x2,this.dt2=void 0===t.dt2?this.dt2:t.dt2,this.yPos2=void 0===t.yPos2?this.yPos2:t.yPos2,this.y2=void 0===t.y2?this.y2:t.y2}show(){this._show||(this._show=!0)}hide(){this._show&&(this._show=!1)}revertSomeToPos(t){const e={x:null,y:null};return Number.isFinite(this["xPos"+t])?e.x=this["xPos"+t]:Number.isFinite(this["x"+t])?e.x=this.revertIdToPos(this["x"+t]):Number.isFinite(this["dt"+t])&&(e.x=this.revertIdToPos(this.revertDtToId(this["dt"+t]))),Number.isFinite(this["yPos"+t])?e.y=this["yPos"+t]:Number.isFinite(this["y"+t])&&(e.y=this.revertValueToYPos(this["y"+t])),e}revertDtToId(t){const e=this.chart.mainSeries,[n,i]=[this.chart.range.leftId,this.chart.range.rightId];if(!e||!e.data||!e.data[n]||t<e.data[n].datetime)return null;const r=e.data[i]?i:e.last_id;if(t<=e.data[r].datetime)for(let o=n;o<r+1;o++)if(t-e.data[o].datetime<=0)return o;return null}revertIdToPos(t){return Math.round(this.chart.timeAxis.scale(t))+Math.floor(this.chart.bar.barWidth/2)}revertValueToYPos(t){return t&&this.board[this.yAlign+"YAxis"]?Math.round(this.board[this.yAlign+"YAxis"].scale(t)):null}isPointInChart(t){const{x:e,y:n}=t;return e>=0&&e<=this.board.width&&n>=0&&n<=this.board.height}}class Di extends Ti{constructor(t){super(t),this.type="text",this.color=t.color||"#000000",this.xPos1=t.xPos1,this.x1=t.x1,this.dt1=t.dt1,this.yPos1=t.yPos1,this.y1=t.y1,this.text=t.text,this.size=t.size||12}update(t){super.update(t),this.text=void 0===t.text?this.text:t.text}draw(t){if(!this._show)return;const e=this.chart.bar,n=this.x1*e.step,i=this.revertValueToYPos(this.y1),r=_i();t.fillStyle=this.color,t.font=bi(this.size),t.fillText(this.text,n*r,(i+this.size)*r)}}class Ci extends Ti{constructor(t){super(t),this.type="line",this.color=t.color,this.lineWidth=t.lineWidth,this.xPos1=t.xPos1,this.x1=t.x1,this.dt1=t.dt1,this.yPos1=t.yPos1,this.y1=t.y1,this.xPos2=t.xPos2,this.x2=t.x2,this.dt2=t.dt2,this.yPos2=t.yPos2,this.y2=t.y2,this.text=t.text,this.clamp=t.clamp}update(t){super.update(t),this.x1=t.x1||this.x1,this.y1=t.y1||this.y1,this.text=void 0===t.text?this.text:t.text,this.clamp=void 0===t.clamp?this.clamp:t.clamp}draw(t){if(!this._show)return;const e=this.chart.bar;xi(t,{x:this.x1*e.step,y:this.revertValueToYPos(this.y1)},{x:this.x2*e.step,y:this.revertValueToYPos(this.y2)},{color:this.color,lineWidth:this.lineWidth})}}class Ii extends Ti{constructor(t){super(t),this.type="h-line",this.color=t.stroke,this.lineWidth=t.lineWidth,this.y1=t.y1,this.text=t.text||""}update(t){super.update(t),this.y1=t.y1||this.y1,this.text=t.text||this.text}draw(t){if(!this._show)return;const e=this.revertValueToYPos(this.y1),n=this.chart.chartCtrl.pixelLeft,i=this.chart.chartCtrl.pixelLeft+this.board.width,r=_i();t.font=bi(12),xi(t,{x:n,y:e},{x:i,y:e},{color:this.color,lineWidth:this.lineWidth}),t.fillStyle=this.color,t.fillText(this.text,n*r,(e-1)*r)}}class Ai extends Ti{constructor(t){super(t),this.type="rect",this.fill=t.fill||"none",this.color=t.color||"#bbbbbb",this.lineWidth=t.lineWidth||1,this.xPos1=t.xPos1,this.x1=t.x1,this.dt1=t.dt1,this.yPos1=t.yPos1,this.y1=t.y1,this.xPos2=t.xPos2,this.x2=t.x2,this.dt2=t.dt2,this.yPos2=t.yPos2,this.y2=t.y2,this.width=t.width,this.height=t.height}update(t){super.update(t),this.width=void 0===t.width?this.width:t.width,this.height=void 0===t.height?this.height:t.height}draw(t){if(!this._show)return;const e=this.chart.bar,n=this.x1*e.step,i=this.revertValueToYPos(this.y1),r=this.x2*e.step+this.chart.bar.barWidth-n,o=this.revertValueToYPos(this.y2)-i;t.lineWidth=this.lineWidth,t.fillStyle=this.fill||"blue",t.strokeStyle=this.color||"red";const a=_i();t.rect(n*a,i*a,r*a,o*a),t.fill(),t.stroke()}}class Yi extends Ti{constructor(t){super(t),this.type="hightlight-bar",this.fillStyle=t.fill||"#d0edfd",this.x1=t.x1}update(t){super.update(t),this.x1=t.x1||this.x1}draw(t){if(!this._show)return;const e=this.chart.bar,n=this.x1*e.step,i=e.barWidth,r=this.board.height;t.beginPath(),t.fillStyle=this.fillStyle;const o=_i();t.rect(n*o,0,i*o,r*o),t.fill()}}class Oi extends Ti{constructor(t){super(t),this.type="arrow",this.x1=t.x1,this.width=t.width,this.height=t.height,this.direction=t.direction}update(t){super.update(t),this.width=void 0===t.width?this.width:t.width,this.height=void 0===t.height?this.height:t.height,this.direction=void 0===t.direction?this.direction:t.direction}draw(){const t=this.chart.bar,e={x:this.x1*t.step,y:this.revertValueToYPos(this.y1)},n=this.width||t.barWidth,i=this.height||t.barWidth,r="up"===this.direction?1:-1,o="none"===this.stroke?0:.5;if(Number.isFinite(e.x)&&Number.isFinite(e.y)&&this.isPointInChart(e)){const t=Math.round(n/3),a=`M ${e.x+n/2} ${e.y} l ${-n/2} ${r*i/2} l ${t-o} 0 l 0 ${r*i/2} l ${t} 0 l 0 ${-1*r*i/2} l ${t} 0 z`;this.arrowG.attr("d",a),this.show()}else this.hide()}}class Pi{constructor(t,e,n){this.chart=t,this.board=e,this.canvas=n,this._showAll=!0,this.marks={}}addMark(t){return t.id=t.id||(new Date).getTime(),this.marks[t.id]?this.marks[t.id].update(t):(t.chart=t.chart||this.chart,t.boardId=t.boardId||"main",t.board=t.board||this.chart.boards[t.boardId],t.yAlign=t.yAlign||"right","line"===t.type?this.marks[t.id]=new Ci(t):"h-line"===t.type?this.marks[t.id]=new Ii(t):"rect"===t.type?this.marks[t.id]=new Ai(t):"arrow"===t.type?this.marks[t.id]=new Oi(t):"text"===t.type?this.marks[t.id]=new Di(t):"hightlight-bar"===t.type&&(this.marks[t.id]=new Yi(t))),this.marks[t.id]}removeMarkAll(){this.marks={}}removeMark(t){this.marks[t]&&delete this.marks[t]}showMark(t){t&&this.marks[t]&&this.marks[t].show()}hideMark(t){t&&this.marks[t]&&this.marks[t].hide()}showAllMark(){this._showAll||(this._showAll=!0)}hideAllMark(){this._showAll&&(this._showAll=!1)}draw(){if(this._showAll)for(const t in this.marks)this.marks[t].draw(this.canvas.ctx)}}class Ei{constructor({width:t,height:e}){this.dom=document.createElement("table"),this.dom.setAttribute("cellpading","0"),this.dom.setAttribute("cellspacing","0"),this.dom.setAttribute("class","tqchart-table"),this.width=t||500,this.height=e||300}resize({width:t,height:e}){this.width=t||this.width,this.height=e||this.height,this.dom.style.width=this.width+"px",this.dom.style.height=this.height+"px"}}class Ni{constructor(t){this.dom=document.createElement("tr");for(const e in t)this.dom.setAttribute(e,t[e])}}class Ri{constructor({width:t,height:e,minWidth:n},i){this.dom=document.createElement("td"),this.dom.style.width=t+"px",this.dom.style.height=e+"px",this.width=t||400,this.height=e||150,n&&(this.dom.style.minWidth=n+"px");for(const r in i)this.dom.setAttribute(r,i[r])}resize({width:t,height:e}){this.width=t||this.width,this.height=e||this.height,this.dom.style.width=this.width+"px",this.dom.style.height=this.height+"px",this.dom.style.minWidth=this.width+"px"}}class Hi{constructor({width:t,height:e},n){this.dom=document.createElement("canvas"),this.ctx=this.dom.getContext("2d"),this.width=t||300,this.height=e||150,this.resize({width:t,height:e});for(const i in n)this.dom.setAttribute(i,n[i])}resize({width:t,height:e}){this.width=t||this.width,this.height=e||this.height;const n=_i();this.dom.width=this.width*n,this.dom.height=this.height*n,this.dom.style.width=this.width+"px",this.dom.style.height=this.height+"px"}clear(){this.clearRect({width:this.width,height:this.height})}clearRect({width:t,height:e}){const n=_i();t=t||this.width,e=e||this.height,this.ctx.clearRect(0,0,t*n,e*n)}}class ji extends ki{constructor({boardId:t,chart:e,container:n,width:i,height:r,leftWidth:o,rightWidth:a,rightPriceTick:s,rightPriceDecs:l,leftPriceTick:u,leftPriceDecs:c}){super(),this.boardId=t,this.chart=e,this.container=n,this.width=i||0,this.height=r||0,this.leftWidth=o||0,this.rightWidth=a||0,this._initElement(),this._initYAxis(l,c),this._initManagers(),this.show=!0}_initElement(){this._rootElement=new Ni({class:"board-container"}),this._leftTd=new Ri({width:this.leftWidth,height:this.height,minWidth:this.leftWidth},{class:"left-y-container"}),this._centerTd=new Ri({width:this.width-this.leftWidth-this.rightWidth,height:this.height},{class:"canvas-container"}),this._centerCanvas=new Hi({width:this.width-this.leftWidth-this.rightWidth,height:this.height},{class:"plot-canvas"}),this._centerTopCanvas=new Hi({width:this.width-this.leftWidth-this.rightWidth,height:this.height},{class:"plot-top-canvas","data-board-id":this.boardId}),this._centerTd.dom.appendChild(this._centerCanvas.dom),this._centerTd.dom.appendChild(this._centerTopCanvas.dom),this._rightTd=new Ri({width:this.rightWidth,height:this.height,minWidth:this.rightWidth},{class:"right-y-container"}),this._rootElement.dom.appendChild(this._leftTd.dom),this._rootElement.dom.appendChild(this._centerTd.dom),this._rootElement.dom.appendChild(this._rightTd.dom)}_initYAxis(t,e){this._rightTd&&(this.rightYAxis=new Si({chart:this.chart,width:this.rightWidth,height:this.height,boardId:this.boardId,container:this._rightTd,align:"right",priceDecs:t})),this._leftTd&&(this.leftYAxis=new Si({chart:this.chart,width:this.leftWidth,height:this.height,boardId:this.boardId,container:this._leftTd,align:"left",priceDecs:e}))}_initManagers(){this.marksManager=new Pi(this.chart,this,this._centerCanvas),this.hightlightManager=new Pi(this.chart,this,this._centerCanvas)}setDomain(t){this.domain=t}setPriceDecs(t){this.rightYAxis.priceDecs=t}drawYAnnotation(t,e){e?(this.leftYAxis.showYAnnotation(t),this.rightYAxis.showYAnnotation(t)):(this.leftYAxis.hideYAnnotation(),this.rightYAxis.hideYAnnotation())}addMark(t){"back"===t.layer?this.hightlightManager.addMark(t):this.marksManager.addMark(t)}removeMark(t){this.hightlightManager.removeMark(t),this.marksManager.removeMark(t)}removeMarkAll(){this.hightlightManager.removeMarkAll(),this.marksManager.removeMarkAll()}update({width:t,height:e}){if((t<=-1||e<=-1)&&this.show?(this.show=!1,this._rootElement.dom.previousElementSibling.getAttribute("class").indexOf("separator-container")>-1&&this._rootElement.dom.previousElementSibling.setAttribute("class","separator-container hide"),this._rootElement.dom.setAttribute("class","board-container hide"),this._centerCanvas.clear(),this._centerTopCanvas.clear()):!this.show&&t>0&&e>0&&(this.show=!0,this._rootElement.dom.previousElementSibling.getAttribute("class").indexOf("separator-container")>-1&&this._rootElement.dom.previousElementSibling.setAttribute("class","separator-container"),this._rootElement.dom.setAttribute("class","board-container")),t>0||e>0){t>0&&this.width!=t&&(this.width=t),e>0&&this.height!=e&&(this.height=e,this._leftTd.resize({width:this.leftWidth,height:this.height}),this._rightTd.resize({width:this.rightWidth,height:this.height}),this.leftYAxis&&this.leftYAxis.update({width:this.leftWidth,height:this.height}),this.rightYAxis&&this.rightYAxis.update({width:this.rightWidth,height:this.height}));const n=this.width-this.leftWidth-this.rightWidth;this._centerTd.resize({width:n,height:this.height}),this._centerCanvas.resize({width:n,height:this.height}),this._centerTopCanvas.resize({width:n,height:this.height})}}drawClearRect(){this.show&&this._centerCanvas.clear()}drawYAxis(){this.show}drawBackground(){this.show&&this.hightlightManager.draw()}drawFrontMark(){this.show&&this.marksManager.draw()}hideCrosshair(){this._centerTopCanvas.clear()}drawTitleText(t,e){const n=this._centerTopCanvas.ctx,i=_i();n.font=bi(14),n.fillStyle="#000000";let r="";const o=this.chart.bar.getKlineId(t);if(this.chart.mainSeries&&this.chart.mainSeries.data[o])for(const a in this.chart.plots){const t=this.chart.plots[a];t.boardId===this.boardId&&(r+=t.getText(o))}n.fillText(r,1*i,15*i)}drawCrosshair(t,e){if(!this.show)return;this._centerTopCanvas.clear();const n=this._centerTopCanvas.ctx,i=_i();n.save(),n.lineWidth=1*i,n.setLineDash([5,5]),n.strokeStyle="#666666",n.beginPath(),n.moveTo(t*i,0*i),n.lineTo(t*i,this.height*i),n.moveTo(0*i,e*i),n.lineTo(this.width*i,e*i),n.stroke(),n.restore()}drawCrosshairVertical(t){this._centerTopCanvas.clear();const e=this._centerTopCanvas.ctx,n=_i();e.save(),e.lineWidth=1*n,e.setLineDash([5,5]),e.strokeStyle="#666666",e.beginPath(),e.moveTo(t*n,0*n),e.lineTo(t*n,this.height*n),e.stroke(),e.restore()}}class zi extends ji{_initElement(){this._rootElement=new Ni({class:"board-container"}),this._leftTd=new Ri({width:this.leftWidth,height:this.height}),this._centerTd=new Ri({width:this.width-this.leftWidth-this.rightWidth,height:this.height}),this._rightTd=new Ri({width:this.rightWidth,height:this.height}),this._rootElement.dom.appendChild(this._leftTd.dom),this._rootElement.dom.appendChild(this._centerTd.dom),this._rootElement.dom.appendChild(this._rightTd.dom)}_initYAxis(){}_initManagers(){}}class Wi{constructor(){this._rootElement=document.createElement("tr"),this._rootElement.setAttribute("class","separator-container"),this._td=document.createElement("td"),this._td.setAttribute("class","separator"),this._td.setAttribute("colspan","3"),this._separatorDiv=document.createElement("div"),this._separatorDiv.setAttribute("class","handle"),this._td.appendChild(this._separatorDiv),this._rootElement.appendChild(this._td)}}class Bi extends ki{init({chart:t,width:e,height:n,margin:i}){this.chart=t,this.width=e,this.height=n,this.margin=i,this.table=new Ei({width:this.width,height:this.height}),this.root=this.table.dom,this.boards=[],this.initX=null,this.isClick=!0,this.isDraging=!1,this.dragTimeout=null,this._initEventListeners(),this.timeAxisBoard=new zi({chart:this.chart,width:this.width,height:this.margin.bottom,leftWidth:this.margin.left,rightWidth:this.margin.right}),this.table.dom.appendChild(this.timeAxisBoard._rootElement.dom)}getBoardContainerByBoardId(t){for(const e of this.boards)if(t===e.boardId)return e;return null}update(){const t=function(t,e,n){const i=e-1*(t-1),r=[];if(n){r[0]={top:0,height:i};for(let e=1;e<t;e++)r[e]={top:-1,height:-1}}else{const e=1/t,n=Math.ceil(i*e*.7),o=i-n*(t-1);r[0]={top:0,height:o};for(let i=1;i<t;i++)r[i]={height:n,top:r[i-1].top+r[i-1].height+1}}return r}(this.boards.length,this.height-this.margin.bottom,this.chart.onlyMainBoard||!1);for(let e=0;e<t.length;e++)this.boards[e].update({width:this.width,height:t[e].height})}resize({width:t,height:e}){this.width=t,this.height=e,this.table.resize({width:this.width,height:this.height}),this.update()}addBoard(t){for(const e of this.boards)if(e.boardId===t.boardId)return;if(this.boards.push(t),this.boards.length>1){const t=new Wi;this.table.dom.insertBefore(t._rootElement,this.timeAxisBoard._rootElement.dom)}return this.table.dom.insertBefore(t._rootElement.dom,this.timeAxisBoard._rootElement.dom),this.update(),t}_getBoardIdByEvent(t){return"CANVAS"==t.target.nodeName&&"plot-top-canvas"==t.target.getAttribute("class")&&t.target.dataset.boardId}_initEventListeners(){const t=this;this.root.addEventListener("mouseup",(function(e){e.preventDefault(),t.dragTimeout&&(clearTimeout(t.dragTimeout),t.dragTimeout=null)}),!1),this.root.addEventListener("mousemove",(function(e){e.preventDefault();const n=t._getBoardIdByEvent(e);n&&(t.isDraging?t.emit("dragmove",n,e):t.emit("move",n,e))}),!1);const e=function(n){n.preventDefault(),n.stopPropagation(),t.isDraging&&(t.isClick=!1,t.isDraging=!1,t.emit("dragend",t.boardId,n),t.root.removeEventListener("mouseup",e))};this.root.addEventListener("mousedown",(function(n){n.preventDefault(),n.stopPropagation();const i=t._getBoardIdByEvent(n);i&&(t.dragTimeout=setTimeout((function(){t.isClick=!0,t.isDraging=!0,t.emit("dragstart",i,n),t.root.addEventListener("mouseup",e,!1)}),100))}),!1),this.root.addEventListener("mouseenter",(function(t){t.preventDefault(),t.stopPropagation()}),!1),this.root.addEventListener("mouseleave",(function(e){e.preventDefault(),e.stopPropagation(),t.isDraging&&(t.isDraging=!1,t.emit("dragend",t.boardId,e))}),!1),this.root.addEventListener("wheel",(function(e){e.preventDefault(),e.stopPropagation(),t._getBoardIdByEvent(e)&&t.emit("zoom",t.boardId,e)}))}boardId(t,e,n){throw new Error("Method not implemented.")}deleteBoard(t){if("main"===t)return;let e=-1;for(let r=0;r<this.boards.length;r++)this.boards[r].boardId===t&&(e=r);if(-1===e)return;const n=this.boards[e];this.boards.splice(e,1);const i=n._rootElement.dom.previousSibling;return this.table.dom.removeChild(i),this.table.dom.removeChild(n._rootElement.dom),this.update(),n}hideRow(){}showRow(){}}const Fi=/^Y{1,4}|M{1,4}|D{1,2}|d{1,4}|H{1,2}|m{1,2}|s{1,2}|SSS/g;function Vi(t,e="YYYY-MM-DD HH:mm:ss"){const n=(t=t instanceof Date?t:new Date(t)).getFullYear(),i=t.getMonth(),r=t.getDate(),o=t.getHours(),a=t.getMinutes(),s=t.getSeconds(),l=t.getMilliseconds(),u={YY:String(n).slice(-2),YYYY:String(n),M:String(i+1),MM:String(i+1).padStart(2,"0"),D:String(r),DD:String(r).padStart(2,"0"),H:String(o),HH:String(o).padStart(2,"0"),m:String(a),mm:String(a).padStart(2,"0"),s:String(s),ss:String(s).padStart(2,"0"),SSS:String(l).padStart(3,"0")};return e.replace(Fi,t=>u[t])}class Gi{constructor({chart:t,width:e,height:n,container:i}){this.chart=t,this.width=e,this.height=n,this.container=i.dom,this.scale=ve().rangeRound([0,this.width]),this.axis=y(this.scale),this.rootElement=vi(this.container).append("svg").attr("class","axis").attr("width",this.width).attr("height",this.height),this.rootG=this.rootElement.append("g").attr("class","y axis "+this.align),this.annotationG=this.rootElement.append("g"),this.annotationRect=this.annotationG.append("rect").attr("fill","#fbd3a0"),this.annotationText=this.annotationG.append("text")}update({width:t,height:e}){t&&t!=this.width&&(this.width=t),e&&e!=this.height&&(this.height=e),this.rootElement.attr("width",this.width).attr("height",this.height)}getTicks(){return this.axis.getTicks()}draw(t,e,n){const i=t+Math.ceil(this.width/this.chart.bar.step);this.scale.rangeRound([0,(i-t)*this.chart.bar.step]),this.scale.domain([t,i]);const r=t*this.chart.bar.step-this.chart.chartCtrl.pixelLeft+Math.round(this.chart.bar.barWidth/2);this.rootG.attr("transform","translate("+r+",0)"),this.axis.tickFormat((t,e)=>{return n[t]&&n[t].datetime?(i=n[t].datetime/1e6,r=this.chart.duration,Vi(i,r/1e9/86400==0?"YYYY-MM-DD":r/1e3%3600==0?"MM-DD HH:mm":r/1e3%60==0?"HH:mm":"HH:mm:ss")):"";var i,r}),this.rootG.call(this.axis)}drawAnnotation(){}showXAnnotation(t){const e=Math.floor(t/this.chart.bar.step),n=this.scale.domain()[0]+e;if(isNaN(n)||!this.chart.mainSeries||!this.chart.mainSeries.data[n])return void this.annotationG.attr("visibility","hidden");const i=Vi(this.chart.mainSeries.data[n].datetime/1e6,"YYYY-MM-DD HH:mm:ss");this.annotationG.attr("visibility","visible"),this.annotationRect.attr("width",118).attr("height",16).attr("transform","translate("+(t-42)+",1)"),this.annotationText.text(i).attr("transform","translate("+(t-40)+",14)")}hideXAnnotation(){this.annotationG.attr("visibility","hidden")}}var $i={min:function(t,e){let n;if(void 0===e)for(const i of t)null!=i&&(n>i||void 0===n&&i>=i)&&(n=i);else{let i=-1;for(let r of t)null!=(r=e(r,++i,t))&&(n>r||void 0===n&&r>=r)&&(n=r)}return n},max:function(t,e){let n;if(void 0===e)for(const i of t)null!=i&&(n<i||void 0===n&&i>=i)&&(n=i);else{let i=-1;for(let r of t)null!=(r=e(r,++i,t))&&(n<r||void 0===n&&r>=r)&&(n=r)}return n}};const Ui={candle:{getYDomain(t,e,n){const i=n.slice(t,e+1),r=$i.min(i,t=>t&&t.low),o=$i.max(i,t=>t&&t.high),a=.5*(o-r)||4*this.chart.price_ticks;return[r-a,o+a]},draw(t,e,n,i){const r=_i(),o={upBody:"",upLine:"",downBody:"",downLine:"",equalLine:""};for(let a=e;a<=n;a++){const t=i[a];if(!t)continue;if(t.datetime>this.chart.maxDatetime)break;const e=this.yScale(t.open),n=this.yScale(t.close),s=this.yScale(t.high),l=this.yScale(t.low),u=a*this.xScale.step,c=Math.round(u+this.xScale.barWidth/2)-1;if(t.close===t.open)o.equalLine+=`M${u*r} ${(e+.5)*r} l${this.xScale.barWidth*r} 0 M${(c+.5)*r} ${s*r} l0 ${(l-s)*r}`;else{const i=t.close>t.open?"up":"down",a=n>e?Math.max(n-e,1):Math.min(n-e,-1);o[i+"Body"]+=`M${u*r} ${e*r} l${this.xScale.barWidth*r} 0 l0 ${a*r} l${-this.xScale.barWidth*r} 0 l0 ${-a*r}`,o[i+"Line"]+=`M${(c+.5)*r} ${s*r} l0 ${(l-s)*r}`}}wi(t,o.upBody,{fillStyle:"red"}),wi(t,o.upLine,{lineWidth:r,strokeStyle:"red"}),wi(t,o.downBody,{fillStyle:"green"}),wi(t,o.downLine,{lineWidth:r,strokeStyle:"green"}),wi(t,o.equalLine,{lineWidth:r,strokeStyle:"black"})}},close:{getYDomain(t,e,n){const i=n.slice(t,e+1);let[r,o]=[$i.min(i,t=>t&&t.close),$i.max(i,t=>t&&t.close)];return r-=.2*(o-r),o+=.2*(o-r),[r,o]},draw(t,e,n,i){const r=_i();let o="";for(let a=e;a<=n;a++){const t=i[a];if(t){if(t.datetime>this.chart.maxDatetime)break;o+=`${a==e?"M":"L"} ${(a*this.xScale.step+Math.round(this.xScale.barWidth/2)-1)*r} ${this.yScale(t.close)*r}`}}wi(t,o,{lineWidth:r,strokeStyle:"#C62828"})}},volume:{getYDomain(t,e,n){const i=n.slice(t,e+1);return[0,1.2*$i.max(i,t=>t&&t.volume)]},draw(t,e,n,i){const r=_i();let o={up:"",down:""};for(let a=e;a<=n;a++){const t=i[a];if(!t)continue;if(!t.volume)continue;if(t.datetime>this.chart.maxDatetime)break;const e=a*this.xScale.step*r,n=this.xScale.barWidth*r,s=this.yScale.range()[0]*r,l=this.yScale(t.volume)*r-s;o[t.close>=t.open?"up":"down"]+=`M${e} ${s} l${n} 0 l0 ${l} l${-n} 0 l0 ${-l}`}wi(t,o.up,{fillStyle:"red"}),wi(t,o.down,{fillStyle:"green"})}},oi:{getYDomain(t,e,n){const i=n.slice(t,e+1);let[r,o]=[$i.min(i,t=>t&&t.close_oi),$i.max(i,t=>t&&t.close_oi)];return r-=.2*(o-r),o+=.2*(o-r),[r,o]},draw(t,e,n,i){const r=_i();let o=null,a="";for(let s=e;s<=n;s++){const t=i[s];if(t&&t.close_oi){if(t.datetime>this.chart.maxDatetime)break;a+=`${o==s-1?"L":"M"} ${(s*this.xScale.step+Math.round(this.xScale.barWidth/2)-1)*r} ${this.yScale(t.close_oi)*r}`,o=s}}wi(t,a,{lineWidth:2*r,strokeStyle:"#C62828"})}},line:{getYDomain(t,e,n){if(Number.isFinite(this.min)&&Number.isFinite(this.max))return[this.min,this.max];{const i=n.slice(t,e+1),[r,o]=[$i.min(i,t=>t&&t[this.propName]),$i.max(i,t=>t&&t[this.propName])],a=.2*(o-r);return[r-a,o+a]}},draw(t,e,n,i){if(!this.yScale||!this.xScale)return;const r=_i();let o=null,a="";for(let l=e;l<=n;l++){const t=i[l],e=l*this.xScale.step+Math.round(this.xScale.barWidth/2)-1;t&&t[this.propName]&&(a+=`${o&&o==l-1?"L":"M"} ${e*r} ${this.yScale(t[this.propName])*r}`,o=l)}const s={lineWidth:this.width*r,strokeStyle:this.color};this.dash&&(s.lineDash=[5,5]),wi(t,a,s)}},bar:{getYDomain(t,e,n){if(Number.isFinite(this.min)&&Number.isFinite(this.max))return[this.min,this.max];{const i=n.slice(t,e+1);return[0,$i.max(i,t=>t&&t[this.propName])]}},draw(t,e,n,i){if(!this.yScale||!this.xScale)return;const r=_i();let o="";for(let a=e;a<=n;a++){const t=i[a];if(!t)continue;if(!t[this.propName])continue;const e=a*this.xScale.step*r,n=this.yScale(t[this.propName])*r,s=this.xScale.barWidth*r,l=(n-this.yScale.range[0])*r;o+=`M${e} ${n} l${s} 0 l0 ${l} l${-s} 0 l0 ${-l}`}wi(t,o,{fillStyle:this.color||"red"})}},dot:{getYDomain(t,e,n){if(Number.isFinite(this.min)&&Number.isFinite(this.max))return[this.min,this.max];{const i=n.slice(t,e+1);return[$i.min(i,t=>t&&t[this.propName]),$i.max(i,t=>t&&t[this.propName])]}},draw(t,e,n,i){if(!this.yScale||!this.xScale)return;const r=_i();let o="";for(let a=e;a<=n;a++){const t=i[a];if(!t)continue;if(!t[this.propName])continue;let e=a*this.xScale.step+this.bar.step/2,n=this.yScale(t[this.propName]);if(Number.isNaN(n)||Number.isNaN(e))continue;const s=3*r,l=3*r;e*=r,n*=r,o+="M"+(e-s)+" "+n+"a"+s+" "+l+" 0 1 0 "+2*s+" 0a"+s+" "+l+" 0 1 0 "+-2*s+" 0z"}wi(t,o,{fillStyle:this.color||"red"})}}};class qi{constructor(t,e,n,i,r,o={}){this.chart=t,this.boardId=e,this.board=this.chart.boards[this.boardId],this.plotId=n,this.type=i,this.yAlign=r||"right",this.xScale=this.chart.bar,this.yScale=this.board[this.yAlign+"YAxis"].scale,this.data=o.data,this.color=o.color||"red",this.width=Number.isFinite(o.width)?Math.round(o.width):1,this.dash=!!o.dash,this.min=o.min,this.max=o.max,this.isHollow=o.isHollow,this._initType(o)}_initType(t){switch(this.propName=t.propName||"value",this.type){case"candle":break;case"volume":this.propName="volume";break;case"oi":this.propName="close_oi";break;case"close":this.propName="close"}}getYDomain(t,e,n){let i=this.type;return"dash"==this.type&&(i="line"),Ui[i].getYDomain.bind(this)(t,e,n)}getText(t){if(this.data&&this.data[t]){const e=this.data[t];if("candle"===this.type&&e)return` O(${e.open}) H(${e.high}) L(${e.low}) C(${e.close})`;if(e&&e[this.propName]){let t=String(e[this.propName]);return t.length>12?(t=function(t,e=2){const n=Number(t),i=Number.isInteger(e)?e:2;return Number.isFinite(n)?n.toFixed(i):String(t)}(e[this.propName],"main"===this.boardId&&"right"===this.yAlign&&this.chart.price_decs?this.chart.price_decs:2),` ${this.plotId} (${t})`):` ${this.plotId} (${e[this.propName]})`}}return""}draw(t,e,n,i){this.data=i||this.data;let r=this.type;"dash"==this.type&&(r="line"),Ui[r].draw.bind(this)(t,e,n,this.data)}}class Xi{constructor(t,e){this._reset(t,e)}_reset(t,e){this.leftId=Number.isInteger(t)?t:-1,this.rightId=Number.isInteger(e)?e:-1}reset(t,e){return!(!this.isInvalid()&&t===this.leftId&&e===this.rightId)&&(this._reset(t,e),!0)}isInvalid(){return-1===this.leftId||-1===this.rightId}isEqual(t){return t.leftId===this.leftId&&t.rightId===this.rightId}copy(t){t instanceof Xi&&(this.leftId=t.leftId,this.rightId=t.rightId)}static copy(t){return new Xi(t.leftId,t.rightId)}}class Zi{constructor(t){this.chart=t,this.range=new Xi(-1,-1),this.pixelLeft=-1}get fullPixel(){return this.lastId&&this.lastId>-1?Math.max(this.lastId*this.chart.bar.step,this.canvasWidth):this.canvasWidth}get lastId(){return this.chart.mainSeries&&this.chart.mainSeries.last_id}get canvasWidth(){return this.chart.width-this.chart.margin.left-this.chart.margin.right}init(){this._initRange()}_initRange(){let[t,e]=[-1,-1];this.chart.mainSeries&&-1!==this.chart.mainSeries.last_id&&(e=this.chart.mainSeries.last_id,t=e-this.chart.bar.barNumbers+1,t+=5,t<0&&(t=0),this.pixelLeft=t*this.chart.bar.step),this.range.reset(t,e)}getRange(){return this.range.isInvalid()&&this._initRange(),this.range}setRange(t,e){t>=0&&e>=t?this.range.reset(t,e)&&(this.pixelLeft=this.range.leftId*this.chart.bar.step):this.range.reset(-1,-1)}move(t){if(this.getRange().isInvalid())return!1;let e=this.pixelLeft-t;if(e<0?e=0:e>=this.fullPixel-this.canvasWidth/2&&(e=Math.round(this.fullPixel-this.canvasWidth/2)),this.pixelLeft===e)return!1;this.pixelLeft=e;const n=Math.floor(e/this.chart.bar.step);let i=n+Math.ceil(this.canvasWidth/this.chart.bar.step);return i>this.lastId&&(i=this.lastId),this.range.reset(n,i),!0}zoom(t){const e=this.chart.bar.step-t,n=this.range.rightId*e-this.pixelLeft;this.pixelLeft=this.range.rightId*this.chart.bar.step-n;const i=Math.ceil(n/this.chart.bar.step);let r=this.range.rightId-i+1;r<0&&(r=0);let o=r+Math.ceil(this.canvasWidth/this.chart.bar.step);return o>this.lastId&&(o=this.lastId),this.range.reset(r,o),!0}resetByKlineId(t){const e=(t=t||this.range.rightId)*this.chart.bar.step-this.pixelLeft;this.pixelLeft=this.range.rightId*this.chart.bar.step-e;const n=Math.ceil(e/this.chart.bar.step);let i=this.range.rightId-n+1;i<0&&(i=0);let r=i+Math.ceil(this.canvasWidth/this.chart.bar.step);return r>this.lastId&&(r=this.lastId),this.range.reset(i,r),!0}revertIdToPx(t){return t*this.chart.bar.step}revertDtToId(t,e,n){const i=this.chart.mainSeries;if(!i||!i.data||this.range.isInvalid())return null;let r=!1;e>=0&&n>e?r=!0:(e=this.range.leftId,n=this.range.rightId);let[o,a]=[e,n];if(a=a<=i.last_id?a:i.last_id,!i.data[o]||!i.data[a])return null;if(t>=i.data[a].datetime&&t<i.data[a].datetime+this.chart.duration)return a;if(t>=i.data[o].datetime&&t<i.data[a].datetime+this.chart.duration){for(let e=o;e<a+1&&i.data[e];e++)if(t>=i.data[e].datetime&&t<i.data[e].datetime+this.chart.duration)return e;return null}if(r)return null;if(t>i.data[a].datetime){if(a===i.last_id)return null;for(let e=a;e<i.last_id&&i.data[e]&&i.data[e];e++)if(t-i.data[e].datetime<=0)return e>=1?e-1:e;return null}if(t<i.data[o].datetime){if(0===o)return null;for(let e=o;e>=0&&i.data[e];e--)if(t-i.data[e].datetime>=0)return e;return null}}}class Ji{constructor(t){this.chart=t,this.board=this.chart.boards.main,this.trades={}}add(t,e){this.trades[t]||(this.trades[t]=e)}removeAll(){this.trades={}}revertValueToYPos(t){return t&&this.board.rightYAxis?Math.round(this.board.rightYAxis.scale(t)):null}draw(){const t=[],e={},n={},i=this.chart.mainSeries;let[r,o]=[this.chart.range.leftId,this.chart.range.rightId];if(o=i.data[o]?o:i.last_id,!i.data[r]||!i.data[o])return;const[a,s]=[i.data[r].datetime,i.data[o].datetime];for(const l in this.trades){const u=this.trades[l];if(u.trade_date_time>=a&&u.trade_date_time<=s+this.chart.duration){const a=this.chart.chartCtrl.revertDtToId(u.trade_date_time,r,o);if(!a)continue;if("BUY"===u.direction)if(e[a])e[a].volume+=u.volume;else{const n={x:this.chart.chartCtrl.revertIdToPx(a),y:this.revertValueToYPos(i.data[a].low),dir:"up",diff:"OPEN"===u.offset?0:.5,width:this.chart.bar.barWidth,volume:u.volume};e[a]=n,t.push(n)}else if("SELL"===u.direction)if(n[a])n[a].volume+=u.volume;else{const e={x:this.chart.chartCtrl.revertIdToPx(a),y:this.revertValueToYPos(i.data[a].high),dir:"down",diff:"OPEN"===u.offset?0:.5,width:this.chart.bar.barWidth,volume:u.volume};n[a]=e,t.push(e)}}}this.drawArrows(t)}drawArrows(t){const e=this.board._centerCanvas.ctx,n=_i(),i={up:"",down:""};for(let r=0;r<t.length;r++){const e=t[r],o="up"===e.dir?1:-1,a=e.width,s=e.width,l=Math.round(s/3);i[e.dir]+=`M ${(e.x+s/2)*n} ${e.y*n} l ${-s/2*n} ${o*a/2*n} l ${(l-e.diff)*n} 0 l 0 ${o*a/2*n} l ${l*n} 0 l 0 ${-1*o*a/2*n} l ${l*n} 0 z`}e.beginPath(),e.fillStyle="red",e.fill(new Path2D(i.up)),e.fillStyle="green",e.fill(new Path2D(i.down))}}class Ki{constructor(t){this.fullWidth=t,this.barWidth=7,this.barNumbers=this._getBarNumbers(),this._domain=null,this.rangeToDomain={},this.domainToRange={},this.domainRange={}}get padding(){return this.barWidth<=3?0:this.barWidth<=17?1:2}get step(){return this.barWidth+(this.barWidth<=3?0:this.barWidth<=17?1:2)}_getBarNumbers(){return Math.round(this.fullWidth/this.step)}set domain(t){this._domain=t,this._updateDomainRange()}get domain(){return this._domain}_updateDomainRange(){this.domainRange={};for(let t=0;t<this.barNumbers;t++)this.domainToRange[this.domain[0]+t]=t*this.step}zoom(t){if(0===(t=Math.round(t)))return;const e=this.barWidth+t+t%2;e<1||e>35||(this.barWidth=e,this.barNumbers=this._getBarNumbers(),this._updateDomainRange())}setFullWidth(t){this.fullWidth=t,this.barNumbers=this._getBarNumbers()}setBarWidth(t=7){this.barWidth=Math.max(Math.min(Math.round(t),35),1),this.barNumbers=this._getBarNumbers()}setBarNumbers(t){this.barNumbers=Math.round(t);let e=Math.floor(this.fullWidth/this.barNumbers);e=Math.max(Math.min(e,35),1),this.barWidth!==e&&this.setBarWidth(e),this._updateDomainRange()}setDomain(t){this.domain=t,this._updateDomainRange()}getPosX(t){return this.domainToRange[t]}getKlineId(t){if(this.domain&&this.domain[0])return this.domain[0]+Math.floor(t/this.step)}getKlineIdTicks(){const t=Math.round((this.domain[1]-this.domain[0]-4)/7),e=[this.domain[0]+2];for(let n=1;n<7;n++)e.push(e[n-1]+t);return e}}e["a"]=class extends ki{constructor(t){if(super(),"string"==typeof t)this.rootContainer=document.querySelector("#"+t);else{if(!(t instanceof HTMLElement&&t.tagName))return;this.rootContainer=t}}init({width:t,height:e,margin:n,symbol:i,duration:r,price_decs:o,price_ticks:a,mainSeries:s}){this.width=t,this.height=e,this.rootContainer.style.width=this.width+"px",this.rootContainer.style.height=this.height+"px",this.margin=n?Object.assign(n,{top:0,right:0,bottom:0,left:0}):{top:1,right:50,bottom:20,left:50},this.symbol=i||"",this.duration=r||6e10,this.price_decs=o||2,this.price_ticks=a||1,this.bar=new Ki(t-this.margin.left-this.margin.right),this.range=new Xi,this.mainSeries=s,this.tableContainer=new Bi,this.chartCtrl=new Zi(this),this.tableContainer.init({chart:this,width:this.width,height:this.height,margin:this.margin}),this._initEventListenners(),this.timeAxis=new Gi({chart:this,width:this.width-this.margin.left-this.margin.right,height:this.margin.bottom,container:this.tableContainer.timeAxisBoard._centerTd}),this.mainBoard=new ji({boardId:"main",chart:this,width:this.width,leftWidth:this.margin.left,rightWidth:this.margin.right,rightPriceDecs:this.price_decs}),this.boards={main:this.mainBoard,sub:new ji({boardId:"sub",chart:this,width:this.width,leftWidth:this.margin.left,rightWidth:this.margin.right,rightPriceDecs:0,leftPriceDecs:0})},this.tableContainer.addBoard(this.mainBoard),this.tableContainer.addBoard(this.boards.sub),this.tableContainer.update(),this.plots={mainkline:new qi(this,"main","candle","candle","right"),volume:new qi(this,"sub","volume","volume","right"),oi:new qi(this,"sub","oi","oi","left")},this.chartCtrl.init(),this.rootContainer.appendChild(this.tableContainer.root),this.tradeArrowManager=new Ji(this)}setRange(t,e){"close"===this.mainType&&this.bar.setBarNumbers(e-t+1),this.chartCtrl.setRange(t,e),this.draw()}setPriceDecs(t){this.price_decs=t||this.price_decs,this.boards.main.setPriceDecs(this.price_decs)}setMainSeries(t,e,n,i="candle",{price_decs:r,price_ticks:o}){t&&e&&n&&(this.symbol=t,this.duration=e,this.mainSeries=n,this.mainType=i,this.chartCtrl.init(),this.price_decs=r||this.price_decs,this.price_ticks=o||this.price_ticks,this.boards.main.setPriceDecs(this.price_decs),this._resetBoards(),this.draw())}_resetBoards(){let t=Object.keys(this.plots);for(let n in t)-1===["mainkline","volume","oi"].indexOf(t[n])&&delete this.plots[t[n]];let e=Object.keys(this.boards);for(let n in e)-1===["main","sub"].indexOf(e[n])&&(delete this.boards[e[n]],this.tableContainer.deleteBoard(e[n]))}draw(t){if("candle"===this.mainType&&"close"===this.plots.mainkline.type&&(this.bar.setBarWidth(),this.plots.mainkline.type="candle",this.chartCtrl.setRange()),"close"===this.mainType&&"candle"===this.plots.mainkline.type&&(this.bar.setBarWidth(),this.plots.mainkline.type="close"),!this.mainSeries)return;if(!this.mainSeries.last_id||-1===this.mainSeries.last_id)return;if(this.mainSeries.last_id-this.range.rightId==1){const{leftId:t,rightId:e}=this.chartCtrl.getRange();this.bar.barNumbers-(e-t)>2?this.chartCtrl.setRange(t,this.mainSeries.last_id):this.chartCtrl.setRange(this.mainSeries.last_id-(e-t),this.mainSeries.last_id)}const{leftId:e,rightId:n}=this.chartCtrl.getRange();this.bar.setDomain([e,n]),this.timeAxis.draw(e,n,this.mainSeries.data);for(const i in this.boards){const t=this.boards[i],r={left:[1/0,-1/0],right:[1/0,-1/0]};for(const s in this.plots){const t=this.plots[s];if(i!=t.boardId)continue;let o=null;o="mainkline"===s||"volume"===s||"oi"===s?t.getYDomain(e,n,this.mainSeries.data):t.getYDomain(e,n,t.data||this.mainSeries.data),o[0]=Math.min(r[t.yAlign][0],o[0]),o[1]=Math.max(r[t.yAlign][1],o[1]),Number.isFinite(o[0])&&Number.isFinite(o[1])&&(r[t.yAlign]=o)}t.leftYAxis.draw(r.left),t.rightYAxis.draw(r.right),t.drawClearRect(e,n,this.mainSeries.data);const o=t._centerCanvas.ctx,a=_i();o.save(),o.translate(-this.chartCtrl.pixelLeft*a,0),t.drawBackground(e,n,this.mainSeries.data);for(const s in this.plots){const t=this.plots[s];i==t.boardId&&("mainkline"===s||"volume"===s||"oi"===s?t.draw(o,e,n,this.mainSeries.data):t.draw(o,e,n,this.plots[s].data||this.mainSeries.data))}t.drawFrontMark(e,n,this.mainSeries.data),"main"==i&&this.tradeArrowManager.draw(),o.restore()}if(!this.chartCtrl.range.isEqual(this.range)){this.range.copy(this.chartCtrl.range);const t=this.range.leftId-100,e=this.range.rightId+100;this.mainSeries.data[t]&&this.mainSeries.data[e]||this.emit("showRangeChanged",this.range)}}resize({width:t,height:e}){t===this.width&&e===this.height||(this.width=t,this.height=e,this.bar.setFullWidth(this.width-this.margin.left-this.margin.right),this.timeAxis.update({width:this.width-this.margin.left-this.margin.right,height:this.margin.bottom}),this.tableContainer.resize({width:this.width,height:this.height}),this.chartCtrl.resetByKlineId(),this.draw())}removeMark(t){for(const e in this.boards)this.boards[e].removeMark(t)}removeMarkAll(){for(const t in this.boards)this.boards[t].removeMarkAll()}addHighlightBar(t,e){for(const n in this.boards)this.boards[n].addMark({id:t,boardId:"main",type:"hightlight-bar",x1:e,layer:"back"});this.draw()}_getKlineIdByDatetime(t){let e=null,n=1/0;for(let i=this.range.leftId;i<=this.range.rightId;i++)if(e||(e=this.mainSeries[i]),e&&e<=t){const r=t-e;if(!(r>0))return 0==r?i:n+r>0?i-1:i;n=r}else if(e&&e>t)return null;return null}addTradeArrow(t,e){this.tradeArrowManager.add(t,e)}addPositionRecord(t,e){}_debounce(t,e){let n=null;return function(){const i=this,r=arguments;clearTimeout(n),n=setTimeout((function(){t.apply(i,r)}),e)}}_initEventListenners(){const t=this;this.tableContainer.addListener("dragstart",(function(e,n){if("close"!==t.mainType)for(const i in t.boards)t.boards[i].hideCrosshair()})),this.tableContainer.addListener("dragmove",(function(e,n){"close"!==t.mainType&&0!=n.movementX&&t.chartCtrl.move(n.movementX)&&t.draw("dragmove")})),this.tableContainer.addListener("dragend",(function(e,n){t.mainType})),this.tableContainer.addListener("zoom",(function(e,n){if("close"===t.mainType)return;const i=n.deltaY;Math.abs(i)>0&&t._debounce(t._zoom,60).bind(t)(i>0?-2:2)})),this.tableContainer.addListener("move",(function(e,n){for(const i in t.boards){const r=t.boards[i];e===i?(r.drawCrosshair(n.offsetX,n.offsetY),r.drawYAnnotation(n.offsetY,!0)):(r.drawCrosshairVertical(n.offsetX),r.drawYAnnotation(n.offsetY,!1)),r.drawTitleText(n.offsetX,n.offsetY)}t.timeAxis.showXAnnotation(n.offsetX)}))}_zoom(t){if("fixed"===this.mode)return;const e=this.bar.step;this.bar.zoom(t),e!==this.bar.step&&(this.chartCtrl.zoom(this.bar.step-e),this.draw("zoom"))}addSeries(t,e){const n=e.board?e.board.toLowerCase():"main";if(!this.boards[n]){const t=new ji({boardId:n,chart:this,width:this.width,leftWidth:this.margin.left,rightWidth:this.margin.right,rightPriceTick:1,rightPriceDecs:0,leftPriceTick:1,leftPriceDecs:0});this.boards[n]=t,this.tableContainer.addBoard(t)}"KSERIAL"!==e.type&&"SERIAL"!==e.type||this.plots[t]?this.plots[t]&&e.color&&(this.plots[t].color=Mi(e.color)):"KSERIAL"===e.type?this.plots[t]=new qi(this,n,t,"candle",e.yAlign,{data:e.data}):this.plots[t]=new qi(this,n,t,e.style.toLowerCase(),e.yAlign,{data:e.data,color:Mi(e.color),width:e.width,dash:"dash"===e.style.toLowerCase()}),"TEXT"===e.type?this.boards[n].addMark({id:t,boardId:n,type:"text",x1:e.x1,y1:e.y1,color:Mi(e.color),size:e.size,text:e.text}):["LINE","SEG","RAY"].includes(e.type)?this.boards[n].addMark({id:t,boardId:n,type:"line",x1:e.x1,x2:e.x2,y1:e.y1,y2:e.y2,color:Mi(e.color),lineWidth:e.width,text:"",clamp:!1}):"BOX"===e.type&&this.boards[n].addMark({id:t,boardId:n,type:"rect",x1:e.x1,x2:e.x2,y1:e.y1,y2:e.y2,fill:Mi(e.bg_color),color:Mi(e.color),lineWidth:e.width||1}),this.draw()}addMark(t){this.boards[t.boardId]&&this.boards[t.boardId].addMark(t)}setOnlyMainBoard(t){this.onlyMainBoard!==t&&(this.onlyMainBoard=t,this.tableContainer.update(),this.draw())}setMaxDatetime(t){this.maxDatetime=t}}},6044:function(t,e,n){var i=n("0b07"),r=i(Object,"create");t.exports=r},"605d":function(t,e,n){var i=n("c6b6"),r=n("da84");t.exports="process"==i(r.process)},"60da":function(t,e,n){"use strict";var i=n("83ab"),r=n("d039"),o=n("df75"),a=n("7418"),s=n("d1e7"),l=n("7b0b"),u=n("44ad"),c=Object.assign,d=Object.defineProperty;t.exports=!c||r((function(){if(i&&1!==c({b:1},c(d({},"a",{enumerable:!0,get:function(){d(this,"b",{value:3,enumerable:!1})}}),{b:2})).b)return!0;var t={},e={},n=Symbol(),r="abcdefghijklmnopqrst";return t[n]=7,r.split("").forEach((function(t){e[t]=t})),7!=c({},t)[n]||o(c({},e)).join("")!=r}))?function(t,e){var n=l(t),r=arguments.length,c=1,d=a.f,h=s.f;while(r>c){var f,p=u(arguments[c++]),m=d?o(p).concat(d(p)):o(p),g=m.length,y=0;while(g>y)f=m[y++],i&&!h.call(p,f)||(n[f]=p[f])}return n}:c},"60ed":function(t,e,n){var i=n("3729"),r=n("2dcb"),o=n("1310"),a="[object Object]",s=Function.prototype,l=Object.prototype,u=s.toString,c=l.hasOwnProperty,d=u.call(Object);function h(t){if(!o(t)||i(t)!=a)return!1;var e=r(t);if(null===e)return!0;var n=c.call(e,"constructor")&&e.constructor;return"function"==typeof n&&n instanceof n&&u.call(n)==d}t.exports=h},6117:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ug-cn",{months:"يانۋار_فېۋرال_مارت_ئاپرېل_ماي_ئىيۇن_ئىيۇل_ئاۋغۇست_سېنتەبىر_ئۆكتەبىر_نويابىر_دېكابىر".split("_"),monthsShort:"يانۋار_فېۋرال_مارت_ئاپرېل_ماي_ئىيۇن_ئىيۇل_ئاۋغۇست_سېنتەبىر_ئۆكتەبىر_نويابىر_دېكابىر".split("_"),weekdays:"يەكشەنبە_دۈشەنبە_سەيشەنبە_چارشەنبە_پەيشەنبە_جۈمە_شەنبە".split("_"),weekdaysShort:"يە_دۈ_سە_چا_پە_جۈ_شە".split("_"),weekdaysMin:"يە_دۈ_سە_چا_پە_جۈ_شە".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY-يىلىM-ئاينىڭD-كۈنى",LLL:"YYYY-يىلىM-ئاينىڭD-كۈنى، HH:mm",LLLL:"dddd، YYYY-يىلىM-ئاينىڭD-كۈنى، HH:mm"},meridiemParse:/يېرىم كېچە|سەھەر|چۈشتىن بۇرۇن|چۈش|چۈشتىن كېيىن|كەچ/,meridiemHour:function(t,e){return 12===t&&(t=0),"يېرىم كېچە"===e||"سەھەر"===e||"چۈشتىن بۇرۇن"===e?t:"چۈشتىن كېيىن"===e||"كەچ"===e?t+12:t>=11?t:t+12},meridiem:function(t,e,n){var i=100*t+e;return i<600?"يېرىم كېچە":i<900?"سەھەر":i<1130?"چۈشتىن بۇرۇن":i<1230?"چۈش":i<1800?"چۈشتىن كېيىن":"كەچ"},calendar:{sameDay:"[بۈگۈن سائەت] LT",nextDay:"[ئەتە سائەت] LT",nextWeek:"[كېلەركى] dddd [سائەت] LT",lastDay:"[تۆنۈگۈن] LT",lastWeek:"[ئالدىنقى] dddd [سائەت] LT",sameElse:"L"},relativeTime:{future:"%s كېيىن",past:"%s بۇرۇن",s:"نەچچە سېكونت",ss:"%d سېكونت",m:"بىر مىنۇت",mm:"%d مىنۇت",h:"بىر سائەت",hh:"%d سائەت",d:"بىر كۈن",dd:"%d كۈن",M:"بىر ئاي",MM:"%d ئاي",y:"بىر يىل",yy:"%d يىل"},dayOfMonthOrdinalParse:/\d{1,2}(-كۈنى|-ئاي|-ھەپتە)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"-كۈنى";case"w":case"W":return t+"-ھەپتە";default:return t}},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:1,doy:7}});return e}))},6252:function(t,e,n){"use strict";var i=n("a77e");const r="ivu-notice",o="ivu-icon",a="ivu_notice_key_";let s,l=24,u=4.5,c=1;const d={info:"ios-information-circle",success:"ios-checkmark-circle",warning:"ios-alert",error:"ios-close-circle"};function h(){return s=s||i["a"].newInstance({prefixCls:r,styles:{top:l+"px",right:0}}),s}function f(t,e){const n=e.title||"",i=e.desc||"",s=e.name||`${a}${c}`,l=e.onClose||function(){},f=e.render,p=0===e.duration?0:e.duration||u;c++;let m,g,y=h();const v=e.render&&!n?"":i||e.render?` ${r}-with-desc`:"";if("normal"==t)g=!1,m=`\n            <div class="${r}-custom-content ${r}-with-normal ${v}">\n                <div class="${r}-title">${n}</div>\n                <div class="${r}-desc">${i}</div>\n            </div>\n        `;else{const e=d[t],a=""===v?"":"-outline";g=!0,m=`\n            <div class="${r}-custom-content ${r}-with-icon ${r}-with-${t} ${v}">\n                <span class="${r}-icon ${r}-icon-${t}">\n                    <i class="${o} ${o}-${e}${a}"></i>\n                </span>\n                <div class="${r}-title">${n}</div>\n                <div class="${r}-desc">${i}</div>\n            </div>\n        `}y.notice({name:s.toString(),duration:p,styles:{},transitionName:"move-notice",content:m,withIcon:g,render:f,hasTitle:!!n,onClose:l,closable:!0,type:"notice"})}e["a"]={open(t){return f("normal",t)},info(t){return f("info",t)},success(t){return f("success",t)},warning(t){return f("warning",t)},error(t){return f("error",t)},config(t){t.top&&(l=t.top),(t.duration||0===t.duration)&&(u=t.duration)},close(t){if(!t)return!1;t=t.toString(),s&&s.remove(t)},destroy(){let t=h();s=null,t.destroy("ivu-notice")}}},"62e4":function(t,e){t.exports=function(t){return t.webpackPolyfill||(t.deprecate=function(){},t.paths=[],t.children||(t.children=[]),Object.defineProperty(t,"loaded",{enumerable:!0,get:function(){return t.l}}),Object.defineProperty(t,"id",{enumerable:!0,get:function(){return t.i}}),t.webpackPolyfill=1),t}},6403:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ms-my",{months:"Januari_Februari_Mac_April_Mei_Jun_Julai_Ogos_September_Oktober_November_Disember".split("_"),monthsShort:"Jan_Feb_Mac_Apr_Mei_Jun_Jul_Ogs_Sep_Okt_Nov_Dis".split("_"),weekdays:"Ahad_Isnin_Selasa_Rabu_Khamis_Jumaat_Sabtu".split("_"),weekdaysShort:"Ahd_Isn_Sel_Rab_Kha_Jum_Sab".split("_"),weekdaysMin:"Ah_Is_Sl_Rb_Km_Jm_Sb".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/pagi|tengahari|petang|malam/,meridiemHour:function(t,e){return 12===t&&(t=0),"pagi"===e?t:"tengahari"===e?t>=11?t:t+12:"petang"===e||"malam"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"pagi":t<15?"tengahari":t<19?"petang":"malam"},calendar:{sameDay:"[Hari ini pukul] LT",nextDay:"[Esok pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kelmarin pukul] LT",lastWeek:"dddd [lepas pukul] LT",sameElse:"L"},relativeTime:{future:"dalam %s",past:"%s yang lepas",s:"beberapa saat",ss:"%d saat",m:"seminit",mm:"%d minit",h:"sejam",hh:"%d jam",d:"sehari",dd:"%d hari",M:"sebulan",MM:"%d bulan",y:"setahun",yy:"%d tahun"},week:{dow:1,doy:7}});return e}))},"642a":function(t,e,n){var i=n("966f"),r=n("3bb4"),o=n("20ec");function a(t){var e=r(t);return 1==e.length&&e[0][2]?o(e[0][0],e[0][1]):function(n){return n===t||i(n,t,e)}}t.exports=a},6547:function(t,e,n){var i=n("a691"),r=n("1d80"),o=function(t){return function(e,n){var o,a,s=String(r(e)),l=i(n),u=s.length;return l<0||l>=u?t?"":void 0:(o=s.charCodeAt(l),o<55296||o>56319||l+1===u||(a=s.charCodeAt(l+1))<56320||a>57343?t?s.charAt(l):o:t?s.slice(l,l+2):a-56320+(o-55296<<10)+65536)}};t.exports={codeAt:o(!1),charAt:o(!0)}},"656b":function(t,e,n){var i=n("e2e4"),r=n("f4d6");function o(t,e){e=i(e,t);var n=0,o=e.length;while(null!=t&&n<o)t=t[r(e[n++])];return n&&n==o?t:void 0}t.exports=o},"65db":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("eo",{months:"januaro_februaro_marto_aprilo_majo_junio_julio_aŭgusto_septembro_oktobro_novembro_decembro".split("_"),monthsShort:"jan_feb_mart_apr_maj_jun_jul_aŭg_sept_okt_nov_dec".split("_"),weekdays:"dimanĉo_lundo_mardo_merkredo_ĵaŭdo_vendredo_sabato".split("_"),weekdaysShort:"dim_lun_mard_merk_ĵaŭ_ven_sab".split("_"),weekdaysMin:"di_lu_ma_me_ĵa_ve_sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"[la] D[-an de] MMMM, YYYY",LLL:"[la] D[-an de] MMMM, YYYY HH:mm",LLLL:"dddd[n], [la] D[-an de] MMMM, YYYY HH:mm",llll:"ddd, [la] D[-an de] MMM, YYYY HH:mm"},meridiemParse:/[ap]\.t\.m/i,isPM:function(t){return"p"===t.charAt(0).toLowerCase()},meridiem:function(t,e,n){return t>11?n?"p.t.m.":"P.T.M.":n?"a.t.m.":"A.T.M."},calendar:{sameDay:"[Hodiaŭ je] LT",nextDay:"[Morgaŭ je] LT",nextWeek:"dddd[n je] LT",lastDay:"[Hieraŭ je] LT",lastWeek:"[pasintan] dddd[n je] LT",sameElse:"L"},relativeTime:{future:"post %s",past:"antaŭ %s",s:"kelkaj sekundoj",ss:"%d sekundoj",m:"unu minuto",mm:"%d minutoj",h:"unu horo",hh:"%d horoj",d:"unu tago",dd:"%d tagoj",M:"unu monato",MM:"%d monatoj",y:"unu jaro",yy:"%d jaroj"},dayOfMonthOrdinalParse:/\d{1,2}a/,ordinal:"%da",week:{dow:1,doy:7}});return e}))},"65f0":function(t,e,n){var i=n("861d"),r=n("e8b5"),o=n("b622"),a=o("species");t.exports=function(t,e){var n;return r(t)&&(n=t.constructor,"function"!=typeof n||n!==Array&&!r(n.prototype)?i(n)&&(n=n[a],null===n&&(n=void 0)):n=void 0),new(void 0===n?Array:n)(0===e?0:e)}},6747:function(t,e){var n=Array.isArray;t.exports=n},6784:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["جنوري","فيبروري","مارچ","اپريل","مئي","جون","جولاءِ","آگسٽ","سيپٽمبر","آڪٽوبر","نومبر","ڊسمبر"],n=["آچر","سومر","اڱارو","اربع","خميس","جمع","ڇنڇر"],i=t.defineLocale("sd",{months:e,monthsShort:e,weekdays:n,weekdaysShort:n,weekdaysMin:n,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd، D MMMM YYYY HH:mm"},meridiemParse:/صبح|شام/,isPM:function(t){return"شام"===t},meridiem:function(t,e,n){return t<12?"صبح":"شام"},calendar:{sameDay:"[اڄ] LT",nextDay:"[سڀاڻي] LT",nextWeek:"dddd [اڳين هفتي تي] LT",lastDay:"[ڪالهه] LT",lastWeek:"[گزريل هفتي] dddd [تي] LT",sameElse:"L"},relativeTime:{future:"%s پوء",past:"%s اڳ",s:"چند سيڪنڊ",ss:"%d سيڪنڊ",m:"هڪ منٽ",mm:"%d منٽ",h:"هڪ ڪلاڪ",hh:"%d ڪلاڪ",d:"هڪ ڏينهن",dd:"%d ڏينهن",M:"هڪ مهينو",MM:"%d مهينا",y:"هڪ سال",yy:"%d سال"},preparse:function(t){return t.replace(/،/g,",")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:1,doy:4}});return i}))},"67ca":function(t,e,n){var i=n("cb5a");function r(t,e){var n=this.__data__,r=i(n,t);return r<0?(++this.size,n.push([t,e])):n[r][1]=e,this}t.exports=r},6887:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n){var i={mm:"munutenn",MM:"miz",dd:"devezh"};return t+" "+r(i[n],t)}function n(t){switch(i(t)){case 1:case 3:case 4:case 5:case 9:return t+" bloaz";default:return t+" vloaz"}}function i(t){return t>9?i(t%10):t}function r(t,e){return 2===e?o(t):t}function o(t){var e={m:"v",b:"v",d:"z"};return void 0===e[t.charAt(0)]?t:e[t.charAt(0)]+t.substring(1)}var a=[/^gen/i,/^c[ʼ\']hwe/i,/^meu/i,/^ebr/i,/^mae/i,/^(mez|eve)/i,/^gou/i,/^eos/i,/^gwe/i,/^her/i,/^du/i,/^ker/i],s=/^(genver|c[ʼ\']hwevrer|meurzh|ebrel|mae|mezheven|gouere|eost|gwengolo|here|du|kerzu|gen|c[ʼ\']hwe|meu|ebr|mae|eve|gou|eos|gwe|her|du|ker)/i,l=/^(genver|c[ʼ\']hwevrer|meurzh|ebrel|mae|mezheven|gouere|eost|gwengolo|here|du|kerzu)/i,u=/^(gen|c[ʼ\']hwe|meu|ebr|mae|eve|gou|eos|gwe|her|du|ker)/i,c=[/^sul/i,/^lun/i,/^meurzh/i,/^merc[ʼ\']her/i,/^yaou/i,/^gwener/i,/^sadorn/i],d=[/^Sul/i,/^Lun/i,/^Meu/i,/^Mer/i,/^Yao/i,/^Gwe/i,/^Sad/i],h=[/^Su/i,/^Lu/i,/^Me([^r]|$)/i,/^Mer/i,/^Ya/i,/^Gw/i,/^Sa/i],f=t.defineLocale("br",{months:"Genver_Cʼhwevrer_Meurzh_Ebrel_Mae_Mezheven_Gouere_Eost_Gwengolo_Here_Du_Kerzu".split("_"),monthsShort:"Gen_Cʼhwe_Meu_Ebr_Mae_Eve_Gou_Eos_Gwe_Her_Du_Ker".split("_"),weekdays:"Sul_Lun_Meurzh_Mercʼher_Yaou_Gwener_Sadorn".split("_"),weekdaysShort:"Sul_Lun_Meu_Mer_Yao_Gwe_Sad".split("_"),weekdaysMin:"Su_Lu_Me_Mer_Ya_Gw_Sa".split("_"),weekdaysParse:h,fullWeekdaysParse:c,shortWeekdaysParse:d,minWeekdaysParse:h,monthsRegex:s,monthsShortRegex:s,monthsStrictRegex:l,monthsShortStrictRegex:u,monthsParse:a,longMonthsParse:a,shortMonthsParse:a,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [a viz] MMMM YYYY",LLL:"D [a viz] MMMM YYYY HH:mm",LLLL:"dddd, D [a viz] MMMM YYYY HH:mm"},calendar:{sameDay:"[Hiziv da] LT",nextDay:"[Warcʼhoazh da] LT",nextWeek:"dddd [da] LT",lastDay:"[Decʼh da] LT",lastWeek:"dddd [paset da] LT",sameElse:"L"},relativeTime:{future:"a-benn %s",past:"%s ʼzo",s:"un nebeud segondennoù",ss:"%d eilenn",m:"ur vunutenn",mm:e,h:"un eur",hh:"%d eur",d:"un devezh",dd:e,M:"ur miz",MM:e,y:"ur bloaz",yy:n},dayOfMonthOrdinalParse:/\d{1,2}(añ|vet)/,ordinal:function(t){var e=1===t?"añ":"vet";return t+e},week:{dow:1,doy:4},meridiemParse:/a.m.|g.m./,isPM:function(t){return"g.m."===t},meridiem:function(t,e,n){return t<12?"a.m.":"g.m."}});return f}))},"688b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("mi",{months:"Kohi-tāte_Hui-tanguru_Poutū-te-rangi_Paenga-whāwhā_Haratua_Pipiri_Hōngoingoi_Here-turi-kōkā_Mahuru_Whiringa-ā-nuku_Whiringa-ā-rangi_Hakihea".split("_"),monthsShort:"Kohi_Hui_Pou_Pae_Hara_Pipi_Hōngoi_Here_Mahu_Whi-nu_Whi-ra_Haki".split("_"),monthsRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,3}/i,monthsStrictRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,3}/i,monthsShortRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,3}/i,monthsShortStrictRegex:/(?:['a-z\u0101\u014D\u016B]+\-?){1,2}/i,weekdays:"Rātapu_Mane_Tūrei_Wenerei_Tāite_Paraire_Hātarei".split("_"),weekdaysShort:"Ta_Ma_Tū_We_Tāi_Pa_Hā".split("_"),weekdaysMin:"Ta_Ma_Tū_We_Tāi_Pa_Hā".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [i] HH:mm",LLLL:"dddd, D MMMM YYYY [i] HH:mm"},calendar:{sameDay:"[i teie mahana, i] LT",nextDay:"[apopo i] LT",nextWeek:"dddd [i] LT",lastDay:"[inanahi i] LT",lastWeek:"dddd [whakamutunga i] LT",sameElse:"L"},relativeTime:{future:"i roto i %s",past:"%s i mua",s:"te hēkona ruarua",ss:"%d hēkona",m:"he meneti",mm:"%d meneti",h:"te haora",hh:"%d haora",d:"he ra",dd:"%d ra",M:"he marama",MM:"%d marama",y:"he tau",yy:"%d tau"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},6909:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("mk",{months:"јануари_февруари_март_април_мај_јуни_јули_август_септември_октомври_ноември_декември".split("_"),monthsShort:"јан_фев_мар_апр_мај_јун_јул_авг_сеп_окт_ное_дек".split("_"),weekdays:"недела_понеделник_вторник_среда_четврток_петок_сабота".split("_"),weekdaysShort:"нед_пон_вто_сре_чет_пет_саб".split("_"),weekdaysMin:"нe_пo_вт_ср_че_пе_сa".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY H:mm",LLLL:"dddd, D MMMM YYYY H:mm"},calendar:{sameDay:"[Денес во] LT",nextDay:"[Утре во] LT",nextWeek:"[Во] dddd [во] LT",lastDay:"[Вчера во] LT",lastWeek:function(){switch(this.day()){case 0:case 3:case 6:return"[Изминатата] dddd [во] LT";case 1:case 2:case 4:case 5:return"[Изминатиот] dddd [во] LT"}},sameElse:"L"},relativeTime:{future:"за %s",past:"пред %s",s:"неколку секунди",ss:"%d секунди",m:"една минута",mm:"%d минути",h:"еден час",hh:"%d часа",d:"еден ден",dd:"%d дена",M:"еден месец",MM:"%d месеци",y:"една година",yy:"%d години"},dayOfMonthOrdinalParse:/\d{1,2}-(ев|ен|ти|ви|ри|ми)/,ordinal:function(t){var e=t%10,n=t%100;return 0===t?t+"-ев":0===n?t+"-ен":n>10&&n<20?t+"-ти":1===e?t+"-ви":2===e?t+"-ри":7===e||8===e?t+"-ми":t+"-ти"},week:{dow:1,doy:7}});return e}))},"69d5":function(t,e,n){var i=n("cb5a"),r=Array.prototype,o=r.splice;function a(t){var e=this.__data__,n=i(e,t);if(n<0)return!1;var r=e.length-1;return n==r?e.pop():o.call(e,n,1),--this.size,!0}t.exports=a},"69f3":function(t,e,n){var i,r,o,a=n("7f9a"),s=n("da84"),l=n("861d"),u=n("9112"),c=n("5135"),d=n("c6cd"),h=n("f772"),f=n("d012"),p=s.WeakMap,m=function(t){return o(t)?r(t):i(t,{})},g=function(t){return function(e){var n;if(!l(e)||(n=r(e)).type!==t)throw TypeError("Incompatible receiver, "+t+" required");return n}};if(a){var y=d.state||(d.state=new p),v=y.get,_=y.has,b=y.set;i=function(t,e){return e.facade=t,b.call(y,t,e),e},r=function(t){return v.call(y,t)||{}},o=function(t){return _.call(y,t)}}else{var x=h("state");f[x]=!0,i=function(t,e){return e.facade=t,u(t,x,e),e},r=function(t){return c(t,x)?t[x]:{}},o=function(t){return c(t,x)}}t.exports={set:i,get:r,has:o,enforce:m,getterFor:g}},"6b75":function(t,e,n){"use strict";function i(t,e){(null==e||e>t.length)&&(e=t.length);for(var n=0,i=new Array(e);n<e;n++)i[n]=t[n];return i}n.d(e,"a",(function(){return i}))},"6ce3":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("nb",{months:"januar_februar_mars_april_mai_juni_juli_august_september_oktober_november_desember".split("_"),monthsShort:"jan._feb._mars_apr._mai_juni_juli_aug._sep._okt._nov._des.".split("_"),monthsParseExact:!0,weekdays:"søndag_mandag_tirsdag_onsdag_torsdag_fredag_lørdag".split("_"),weekdaysShort:"sø._ma._ti._on._to._fr._lø.".split("_"),weekdaysMin:"sø_ma_ti_on_to_fr_lø".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY [kl.] HH:mm",LLLL:"dddd D. MMMM YYYY [kl.] HH:mm"},calendar:{sameDay:"[i dag kl.] LT",nextDay:"[i morgen kl.] LT",nextWeek:"dddd [kl.] LT",lastDay:"[i går kl.] LT",lastWeek:"[forrige] dddd [kl.] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"%s siden",s:"noen sekunder",ss:"%d sekunder",m:"ett minutt",mm:"%d minutter",h:"en time",hh:"%d timer",d:"en dag",dd:"%d dager",w:"en uke",ww:"%d uker",M:"en måned",MM:"%d måneder",y:"ett år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},"6d79":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={0:"-ші",1:"-ші",2:"-ші",3:"-ші",4:"-ші",5:"-ші",6:"-шы",7:"-ші",8:"-ші",9:"-шы",10:"-шы",20:"-шы",30:"-шы",40:"-шы",50:"-ші",60:"-шы",70:"-ші",80:"-ші",90:"-шы",100:"-ші"},n=t.defineLocale("kk",{months:"қаңтар_ақпан_наурыз_сәуір_мамыр_маусым_шілде_тамыз_қыркүйек_қазан_қараша_желтоқсан".split("_"),monthsShort:"қаң_ақп_нау_сәу_мам_мау_шіл_там_қыр_қаз_қар_жел".split("_"),weekdays:"жексенбі_дүйсенбі_сейсенбі_сәрсенбі_бейсенбі_жұма_сенбі".split("_"),weekdaysShort:"жек_дүй_сей_сәр_бей_жұм_сен".split("_"),weekdaysMin:"жк_дй_сй_ср_бй_жм_сн".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Бүгін сағат] LT",nextDay:"[Ертең сағат] LT",nextWeek:"dddd [сағат] LT",lastDay:"[Кеше сағат] LT",lastWeek:"[Өткен аптаның] dddd [сағат] LT",sameElse:"L"},relativeTime:{future:"%s ішінде",past:"%s бұрын",s:"бірнеше секунд",ss:"%d секунд",m:"бір минут",mm:"%d минут",h:"бір сағат",hh:"%d сағат",d:"бір күн",dd:"%d күн",M:"бір ай",MM:"%d ай",y:"бір жыл",yy:"%d жыл"},dayOfMonthOrdinalParse:/\d{1,2}-(ші|шы)/,ordinal:function(t){var n=t%10,i=t>=100?100:null;return t+(e[t]||e[n]||e[i])},week:{dow:1,doy:7}});return n}))},"6d83":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ar-tn",{months:"جانفي_فيفري_مارس_أفريل_ماي_جوان_جويلية_أوت_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),monthsShort:"جانفي_فيفري_مارس_أفريل_ماي_جوان_جويلية_أوت_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},week:{dow:1,doy:4}});return e}))},"6def":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return t.dot?n("span",{ref:"badge",class:t.classes},[t._t("default"),n("sup",{directives:[{name:"show",rawName:"v-show",value:t.badge,expression:"badge"}],class:t.dotClasses,style:t.styles})],2):t.status||t.color?n("span",{ref:"badge",staticClass:"ivu-badge-status",class:t.classes},[n("span",{class:t.statusClasses,style:t.statusStyles}),n("span",{staticClass:"ivu-badge-status-text"},[t._t("text",[t._v(t._s(t.text))])],2)]):n("span",{ref:"badge",class:t.classes},[t._t("default"),t.$slots.count?n("sup",{class:t.customCountClasses,style:t.styles},[t._t("count")],2):t.hasCount?n("sup",{directives:[{name:"show",rawName:"v-show",value:t.badge,expression:"badge"}],class:t.countClasses,style:t.styles},[t._t("text",[t._v(t._s(t.finalCount))])],2):t._e()],2)},r=[],o=(n("a9e3"),n("ade3")),a=n("3a27"),s=["blue","green","red","yellow","pink","magenta","volcano","orange","gold","lime","cyan","geekblue","purple"],l="ivu-badge",u={name:"Badge",props:{count:Number,dot:{type:Boolean,default:!1},overflowCount:{type:[Number,String],default:99},className:String,showZero:{type:Boolean,default:!1},text:{type:String,default:""},status:{validator:function(t){return Object(a["i"])(t,["success","processing","default","error","warning"])}},type:{validator:function(t){return Object(a["i"])(t,["success","primary","normal","error","warning","info"])}},offset:{type:Array},color:{type:String}},computed:{classes:function(){return"".concat(l)},dotClasses:function(){return"".concat(l,"-dot")},countClasses:function(){var t;return["".concat(l,"-count"),(t={},Object(o["a"])(t,"".concat(this.className),!!this.className),Object(o["a"])(t,"".concat(l,"-count-alone"),this.alone),Object(o["a"])(t,"".concat(l,"-count-").concat(this.type),!!this.type),t)]},customCountClasses:function(){return["".concat(l,"-count"),"".concat(l,"-count-custom"),Object(o["a"])({},"".concat(this.className),!!this.className)]},statusClasses:function(){var t;return["".concat(l,"-status-dot"),(t={},Object(o["a"])(t,"".concat(l,"-status-").concat(this.status),!!this.status),Object(o["a"])(t,"".concat(l,"-status-").concat(this.color),!!this.color&&Object(a["i"])(this.color,s)),t)]},statusStyles:function(){return Object(a["i"])(this.color,s)?{}:{backgroundColor:this.color}},styles:function(){var t={};return this.offset&&2===this.offset.length&&(t["margin-top"]="".concat(this.offset[0],"px"),t["margin-right"]="".concat(this.offset[1],"px")),t},finalCount:function(){return""!==this.text?this.text:parseInt(this.count)>=parseInt(this.overflowCount)?"".concat(this.overflowCount,"+"):this.count},badge:function(){var t=!1;return this.count&&(t=!(0===parseInt(this.count))),this.dot&&(t=!0,null!==this.count&&0===parseInt(this.count)&&(t=!1)),""!==this.text&&(t=!0),t||this.showZero},hasCount:function(){return!(!this.count&&""===this.text)||!(!this.showZero||0!==parseInt(this.count))},alone:function(){return void 0===this.$slots.default}}},c=u,d=n("2877"),h=Object(d["a"])(c,i,r,!1,null,null,null),f=h.exports;e["a"]=f},"6e29":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{directives:[{name:"click-outside",rawName:"v-click-outside",value:t.onClickoutside,expression:"onClickoutside"}],class:[t.prefixCls],on:{mouseenter:t.handleMouseenter,mouseleave:t.handleMouseleave}},[n("div",{ref:"reference",class:t.relClasses,on:{click:t.handleClick,contextmenu:function(e){return e.preventDefault(),t.handleRightClick(e)}}},[t._t("default")],2),n("transition",{attrs:{name:"transition-drop"}},[n("Drop",{directives:[{name:"show",rawName:"v-show",value:t.currentVisible,expression:"currentVisible"},{name:"transfer-dom",rawName:"v-transfer-dom"}],ref:"drop",class:t.dropdownCls,attrs:{placement:t.placement,"data-transfer":t.transfer,transfer:t.transfer},nativeOn:{mouseenter:function(e){return t.handleMouseenter(e)},mouseleave:function(e){return t.handleMouseleave(e)}}},[t._t("list")],2)],1)],1)},r=[],o=(n("c975"),n("ade3")),a=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{staticClass:"ivu-select-dropdown",class:t.className,style:t.styles},[t._t("default")],2)},s=[],l=(n("b0c0"),n("ac1f"),n("1276"),n("2b0e")),u=n("3a27"),c=n("c177"),d=l["a"].prototype.$isServer,h=d?function(){}:n("d2e4"),f={name:"Drop",props:{placement:{type:String,default:"bottom-start"},className:{type:String},transfer:{type:Boolean}},data:function(){return{popper:null,width:"",popperStatus:!1,tIndex:this.handleGetIndex()}},computed:{styles:function(){var t={};return this.width&&(t.minWidth="".concat(this.width,"px")),this.transfer&&(t["z-index"]=1060+this.tIndex),t}},methods:{update:function(){var t=this;d||(this.popper?this.$nextTick((function(){t.popper.update(),t.popperStatus=!0})):this.$nextTick((function(){t.popper=new h(t.$parent.$refs.reference,t.$el,{placement:t.placement,modifiers:{computeStyle:{gpuAcceleration:!1},preventOverflow:{boundariesElement:"window"}},onCreate:function(){t.resetTransformOrigin(),t.$nextTick(t.popper.update())},onUpdate:function(){t.resetTransformOrigin()}})})),"iSelect"===this.$parent.$options.name&&(this.width=parseInt(Object(u["h"])(this.$parent.$el,"width"))),this.tIndex=this.handleGetIndex())},destroy:function(){var t=this;this.popper&&setTimeout((function(){t.popper&&!t.popperStatus&&(t.popper.destroy(),t.popper=null),t.popperStatus=!1}),300)},resetTransformOrigin:function(){if(this.popper){var t=this.popper.popper.getAttribute("x-placement"),e=t.split("-")[0],n=t.split("-")[1],i="left"===t||"right"===t;i||(this.popper.popper.style.transformOrigin="bottom"===e||"top"!==e&&"start"===n?"center top":"center bottom")}},handleGetIndex:function(){return Object(c["a"])(),c["b"]}},created:function(){this.$on("on-update-popper",this.update),this.$on("on-destroy-popper",this.destroy)},beforeDestroy:function(){this.popper&&this.popper.destroy()}},p=f,m=n("2877"),g=Object(m["a"])(p,a,s,!1,null,null,null),y=g.exports,v={bind(t,e,n){function i(n){if(t.contains(n.target))return!1;e.expression&&e.value(n)}t.__vueClickOutside__=i,document.addEventListener("click",i)},update(){},unbind(t,e){document.removeEventListener("click",t.__vueClickOutside__),delete t.__vueClickOutside__}},_=n("c780"),b="ivu-dropdown",x={name:"Dropdown",directives:{clickOutside:v,TransferDom:_["a"]},components:{Drop:y},props:{trigger:{validator:function(t){return Object(u["i"])(t,["click","hover","custom","contextMenu"])},default:"hover"},placement:{validator:function(t){return Object(u["i"])(t,["top","top-start","top-end","bottom","bottom-start","bottom-end","left","left-start","left-end","right","right-start","right-end"])},default:"bottom"},visible:{type:Boolean,default:!1},transfer:{type:Boolean,default:function(){return!(!this.$IVIEW||""===this.$IVIEW.transfer)&&this.$IVIEW.transfer}},transferClassName:{type:String},stopPropagation:{type:Boolean,default:!1},capture:{type:Boolean,default:function(){return!this.$IVIEW||this.$IVIEW.capture}}},computed:{transition:function(){return["bottom-start","bottom","bottom-end"].indexOf(this.placement)>-1?"slide-up":"fade"},dropdownCls:function(){var t;return t={},Object(o["a"])(t,b+"-transfer",this.transfer),Object(o["a"])(t,this.transferClassName,this.transferClassName),t},relClasses:function(){return["".concat(b,"-rel"),Object(o["a"])({},"".concat(b,"-rel-user-select-none"),"contextMenu"===this.trigger)]}},data:function(){return{prefixCls:b,currentVisible:this.visible}},watch:{visible:function(t){this.currentVisible=t},currentVisible:function(t){t?this.$refs.drop.update():this.$refs.drop.destroy(),this.$emit("on-visible-change",t)}},methods:{handleClick:function(){if("custom"===this.trigger)return!1;if("click"!==this.trigger)return!1;var t=this.hasParent();t||(this.currentVisible=!this.currentVisible)},handleRightClick:function(){return"custom"!==this.trigger&&("contextMenu"===this.trigger&&void(this.currentVisible=!this.currentVisible))},handleMouseenter:function(){var t=this;return"custom"!==this.trigger&&("hover"===this.trigger&&(this.timeout&&clearTimeout(this.timeout),void(this.timeout=setTimeout((function(){t.currentVisible=!0}),250))))},handleMouseleave:function(){var t=this;return"custom"!==this.trigger&&("hover"===this.trigger&&void(this.timeout&&(clearTimeout(this.timeout),this.timeout=setTimeout((function(){t.currentVisible=!1}),150))))},onClickoutside:function(t){this.handleClose(),this.handleRightClose(),this.currentVisible&&this.$emit("on-clickoutside",t)},handleClose:function(){return"custom"!==this.trigger&&("click"===this.trigger&&void(this.currentVisible=!1))},handleRightClose:function(){return"custom"!==this.trigger&&("contextMenu"===this.trigger&&void(this.currentVisible=!1))},hasParent:function(){var t=Object(u["e"])(this,"Dropdown");return t||!1}},mounted:function(){var t=this;this.$on("on-click",(function(e){if(!t.stopPropagation){var n=t.hasParent();n&&n.$emit("on-click",e)}})),this.$on("on-hover-click",(function(){var e=t.hasParent();e?(t.$nextTick((function(){if("custom"===t.trigger)return!1;t.currentVisible=!1})),e.$emit("on-hover-click")):t.$nextTick((function(){if("custom"===t.trigger)return!1;t.currentVisible=!1}))})),this.$on("on-haschild-click",(function(){t.$nextTick((function(){if("custom"===t.trigger)return!1;t.currentVisible=!0}));var e=t.hasParent();e&&e.$emit("on-haschild-click")}))}},w=x,M=Object(m["a"])(w,i,r,!1,null,null,null);e["a"]=M.exports},"6e98":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("it",{months:"gennaio_febbraio_marzo_aprile_maggio_giugno_luglio_agosto_settembre_ottobre_novembre_dicembre".split("_"),monthsShort:"gen_feb_mar_apr_mag_giu_lug_ago_set_ott_nov_dic".split("_"),weekdays:"domenica_lunedì_martedì_mercoledì_giovedì_venerdì_sabato".split("_"),weekdaysShort:"dom_lun_mar_mer_gio_ven_sab".split("_"),weekdaysMin:"do_lu_ma_me_gi_ve_sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:function(){return"[Oggi a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},nextDay:function(){return"[Domani a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},nextWeek:function(){return"dddd [a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},lastDay:function(){return"[Ieri a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"},lastWeek:function(){switch(this.day()){case 0:return"[La scorsa] dddd [a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT";default:return"[Lo scorso] dddd [a"+(this.hours()>1?"lle ":0===this.hours()?" ":"ll'")+"]LT"}},sameElse:"L"},relativeTime:{future:"tra %s",past:"%s fa",s:"alcuni secondi",ss:"%d secondi",m:"un minuto",mm:"%d minuti",h:"un'ora",hh:"%d ore",d:"un giorno",dd:"%d giorni",w:"una settimana",ww:"%d settimane",M:"un mese",MM:"%d mesi",y:"un anno",yy:"%d anni"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},"6eeb":function(t,e,n){var i=n("da84"),r=n("9112"),o=n("5135"),a=n("ce4e"),s=n("8925"),l=n("69f3"),u=l.get,c=l.enforce,d=String(String).split("String");(t.exports=function(t,e,n,s){var l,u=!!s&&!!s.unsafe,h=!!s&&!!s.enumerable,f=!!s&&!!s.noTargetGet;"function"==typeof n&&("string"!=typeof e||o(n,"name")||r(n,"name",e),l=c(n),l.source||(l.source=d.join("string"==typeof e?e:""))),t!==i?(u?!f&&t[e]&&(h=!0):delete t[e],h?t[e]=n:r(t,e,n)):h?t[e]=n:a(e,n)})(Function.prototype,"toString",(function(){return"function"==typeof this&&u(this).source||s(this)}))},"6f12":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("it-ch",{months:"gennaio_febbraio_marzo_aprile_maggio_giugno_luglio_agosto_settembre_ottobre_novembre_dicembre".split("_"),monthsShort:"gen_feb_mar_apr_mag_giu_lug_ago_set_ott_nov_dic".split("_"),weekdays:"domenica_lunedì_martedì_mercoledì_giovedì_venerdì_sabato".split("_"),weekdaysShort:"dom_lun_mar_mer_gio_ven_sab".split("_"),weekdaysMin:"do_lu_ma_me_gi_ve_sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Oggi alle] LT",nextDay:"[Domani alle] LT",nextWeek:"dddd [alle] LT",lastDay:"[Ieri alle] LT",lastWeek:function(){switch(this.day()){case 0:return"[la scorsa] dddd [alle] LT";default:return"[lo scorso] dddd [alle] LT"}},sameElse:"L"},relativeTime:{future:function(t){return(/^[0-9].+$/.test(t)?"tra":"in")+" "+t},past:"%s fa",s:"alcuni secondi",ss:"%d secondi",m:"un minuto",mm:"%d minuti",h:"un'ora",hh:"%d ore",d:"un giorno",dd:"%d giorni",M:"un mese",MM:"%d mesi",y:"un anno",yy:"%d anni"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},"6f50":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-nz",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},"6f53":function(t,e,n){var i=n("83ab"),r=n("df75"),o=n("fc6a"),a=n("d1e7").f,s=function(t){return function(e){var n,s=o(e),l=r(s),u=l.length,c=0,d=[];while(u>c)n=l[c++],i&&!a.call(s,n)||d.push(t?[n,s[n]]:s[n]);return d}};t.exports={entries:s(!0),values:s(!1)}},"6f6c":function(t,e){var n=/\w*$/;function i(t){var e=new t.constructor(t.source,n.exec(t));return e.lastIndex=t.lastIndex,e}t.exports=i},"6fcd":function(t,e,n){var i=n("50d8"),r=n("d370"),o=n("6747"),a=n("0d24"),s=n("c098"),l=n("73ac"),u=Object.prototype,c=u.hasOwnProperty;function d(t,e){var n=o(t),u=!n&&r(t),d=!n&&!u&&a(t),h=!n&&!u&&!d&&l(t),f=n||u||d||h,p=f?i(t.length,String):[],m=p.length;for(var g in t)!e&&!c.call(t,g)||f&&("length"==g||d&&("offset"==g||"parent"==g)||h&&("buffer"==g||"byteLength"==g||"byteOffset"==g)||s(g,m))||p.push(g);return p}t.exports=d},7118:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="jan._feb._mrt._apr._mai_jun._jul._aug._sep._okt._nov._des.".split("_"),n="jan_feb_mrt_apr_mai_jun_jul_aug_sep_okt_nov_des".split("_"),i=t.defineLocale("fy",{months:"jannewaris_febrewaris_maart_april_maaie_juny_july_augustus_septimber_oktober_novimber_desimber".split("_"),monthsShort:function(t,i){return t?/-MMM-/.test(i)?n[t.month()]:e[t.month()]:e},monthsParseExact:!0,weekdays:"snein_moandei_tiisdei_woansdei_tongersdei_freed_sneon".split("_"),weekdaysShort:"si._mo._ti._wo._to._fr._so.".split("_"),weekdaysMin:"Si_Mo_Ti_Wo_To_Fr_So".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[hjoed om] LT",nextDay:"[moarn om] LT",nextWeek:"dddd [om] LT",lastDay:"[juster om] LT",lastWeek:"[ôfrûne] dddd [om] LT",sameElse:"L"},relativeTime:{future:"oer %s",past:"%s lyn",s:"in pear sekonden",ss:"%d sekonden",m:"ien minút",mm:"%d minuten",h:"ien oere",hh:"%d oeren",d:"ien dei",dd:"%d dagen",M:"ien moanne",MM:"%d moannen",y:"ien jier",yy:"%d jierren"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return i}))},7156:function(t,e,n){var i=n("861d"),r=n("d2bb");t.exports=function(t,e,n){var o,a;return r&&"function"==typeof(o=e.constructor)&&o!==n&&i(a=o.prototype)&&a!==n.prototype&&r(t,a),t}},7159:function(t,e,n){var i=n("266a"),r=n("9934");function o(t){return null==t?[]:i(t,r(t))}t.exports=o},"72af":function(t,e,n){var i=n("99cd"),r=i();t.exports=r},"72f0":function(t,e){function n(t){return function(){return t}}t.exports=n},7333:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-il",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n}});return e}))},"73ac":function(t,e,n){var i=n("743f"),r=n("b047"),o=n("99d3"),a=o&&o.isTypedArray,s=a?r(a):i;t.exports=s},7418:function(t,e){e.f=Object.getOwnPropertySymbols},"743f":function(t,e,n){var i=n("3729"),r=n("b218"),o=n("1310"),a="[object Arguments]",s="[object Array]",l="[object Boolean]",u="[object Date]",c="[object Error]",d="[object Function]",h="[object Map]",f="[object Number]",p="[object Object]",m="[object RegExp]",g="[object Set]",y="[object String]",v="[object WeakMap]",_="[object ArrayBuffer]",b="[object DataView]",x="[object Float32Array]",w="[object Float64Array]",M="[object Int8Array]",S="[object Int16Array]",L="[object Int32Array]",k="[object Uint8Array]",T="[object Uint8ClampedArray]",D="[object Uint16Array]",C="[object Uint32Array]",I={};function A(t){return o(t)&&r(t.length)&&!!I[i(t)]}I[x]=I[w]=I[M]=I[S]=I[L]=I[k]=I[T]=I[D]=I[C]=!0,I[a]=I[s]=I[_]=I[l]=I[b]=I[u]=I[c]=I[d]=I[h]=I[f]=I[p]=I[m]=I[g]=I[y]=I[v]=!1,t.exports=A},"746f":function(t,e,n){var i=n("428f"),r=n("5135"),o=n("e538"),a=n("9bf2").f;t.exports=function(t){var e=i.Symbol||(i.Symbol={});r(e,t)||a(e,t,{value:o.f(t)})}},"74c8":function(t,e,n){var i=n("972ca"),r=n("242e"),o=n("badf");function a(t,e){return i(t,o(e,3),r)}t.exports=a},"74dc":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("sw",{months:"Januari_Februari_Machi_Aprili_Mei_Juni_Julai_Agosti_Septemba_Oktoba_Novemba_Desemba".split("_"),monthsShort:"Jan_Feb_Mac_Apr_Mei_Jun_Jul_Ago_Sep_Okt_Nov_Des".split("_"),weekdays:"Jumapili_Jumatatu_Jumanne_Jumatano_Alhamisi_Ijumaa_Jumamosi".split("_"),weekdaysShort:"Jpl_Jtat_Jnne_Jtan_Alh_Ijm_Jmos".split("_"),weekdaysMin:"J2_J3_J4_J5_Al_Ij_J1".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"hh:mm A",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[leo saa] LT",nextDay:"[kesho saa] LT",nextWeek:"[wiki ijayo] dddd [saat] LT",lastDay:"[jana] LT",lastWeek:"[wiki iliyopita] dddd [saat] LT",sameElse:"L"},relativeTime:{future:"%s baadaye",past:"tokea %s",s:"hivi punde",ss:"sekunde %d",m:"dakika moja",mm:"dakika %d",h:"saa limoja",hh:"masaa %d",d:"siku moja",dd:"siku %d",M:"mwezi mmoja",MM:"miezi %d",y:"mwaka mmoja",yy:"miaka %d"},week:{dow:1,doy:7}});return e}))},7530:function(t,e,n){var i=n("1a8c"),r=Object.create,o=function(){function t(){}return function(e){if(!i(e))return{};if(r)return r(e);t.prototype=e;var n=new t;return t.prototype=void 0,n}}();t.exports=o},"76dd":function(t,e,n){var i=n("ce86");function r(t){return null==t?"":i(t)}t.exports=r},"76ea":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes,style:t.styles},[t._t("default")],2)},r=[],o=(n("99af"),n("4160"),n("a9e3"),n("b64b"),n("159b"),n("53ca")),a=n("ade3"),s=n("3a27"),l="ivu-col",u={name:"iCol",props:{span:[Number,String],order:[Number,String],offset:[Number,String],push:[Number,String],pull:[Number,String],className:String,xs:[Number,Object],sm:[Number,Object],md:[Number,Object],lg:[Number,Object],xl:[Number,Object],xxl:[Number,Object]},data:function(){return{gutter:0}},computed:{classes:function(){var t,e=this,n=["".concat(l),(t={},Object(a["a"])(t,"".concat(l,"-span-").concat(this.span),this.span),Object(a["a"])(t,"".concat(l,"-order-").concat(this.order),this.order),Object(a["a"])(t,"".concat(l,"-offset-").concat(this.offset),this.offset),Object(a["a"])(t,"".concat(l,"-push-").concat(this.push),this.push),Object(a["a"])(t,"".concat(l,"-pull-").concat(this.pull),this.pull),Object(a["a"])(t,"".concat(this.className),!!this.className),t)];return["xs","sm","md","lg","xl","xxl"].forEach((function(t){if("number"===typeof e[t])n.push("".concat(l,"-span-").concat(t,"-").concat(e[t]));else if("object"===Object(o["a"])(e[t])){var i=e[t];Object.keys(i).forEach((function(e){n.push("span"!==e?"".concat(l,"-").concat(t,"-").concat(e,"-").concat(i[e]):"".concat(l,"-span-").concat(t,"-").concat(i[e]))}))}})),n},styles:function(){var t={};return 0!==this.gutter&&(t={paddingLeft:this.gutter/2+"px",paddingRight:this.gutter/2+"px"}),t}},methods:{updateGutter:function(){var t=Object(s["e"])(this,"Row");t&&t.updateGutter(t.gutter)}},mounted:function(){this.updateGutter()},beforeDestroy:function(){this.updateGutter()}},c=u,d=n("2877"),h=Object(d["a"])(c,i,r,!1,null,null,null),f=h.exports;e["a"]=f},"77c1":function(t,e,n){var i=n("7948"),r=n("badf"),o=n("89d9"),a=n("1bac");function s(t,e){if(null==t)return{};var n=i(a(t),(function(t){return[t]}));return e=r(e),o(t,n,(function(t,n){return e(t,n[0])}))}t.exports=s},7839:function(t,e){t.exports=["constructor","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","toLocaleString","toString","valueOf"]},7948:function(t,e){function n(t,e){var n=-1,i=null==t?0:t.length,r=Array(i);while(++n<i)r[n]=e(t[n],n,t);return r}t.exports=n},"79bc":function(t,e,n){var i=n("0b07"),r=n("2b3e"),o=i(r,"Map");t.exports=o},"7a12":function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n(t.tagName,t._b({tag:"component",class:t.classes,attrs:{disabled:t.itemDisabled},on:{click:t.handleClickLink}},"component",t.tagProps,!1),[t.loading?n("Icon",{staticClass:"ivu-load-loop",attrs:{type:"ios-loading"}}):t._e(),!t.icon&&!t.customIcon||t.loading?t._e():n("Icon",{attrs:{type:t.icon,custom:t.customIcon}}),t.showSlot?n("span",{ref:"slot"},[t._t("default")],2):t._e()],1)},r=[],o=n("ade3"),a=n("5250"),s=n("3a27"),l={props:{to:{type:[Object,String]},replace:{type:Boolean,default:!1},target:{type:String,validator(t){return Object(s["i"])(t,["_blank","_self","_parent","_top"])},default:"_self"},append:{type:Boolean,required:!1,default:!1}},computed:{linkUrl(){const t=typeof this.to;if("string"!==t)return null;if(this.to.includes("//"))return this.to;const e=this.$router;if(e){const t=this.$route,n=e.resolve(this.to,t,this.append);return n?n.href:this.to}return this.to}},methods:{handleClick(t=!1){const e=this.$router;if(t){let t=this.to;if(e){const n=this.$route,i=e.resolve(this.to,n,this.append);t=i?i.href:this.to}window.open(t)}else e?"string"===typeof this.to&&this.to.includes("//")?window.location.href=this.to:this.replace?this.$router.replace(this.to,()=>{}):this.$router.push(this.to,()=>{}):window.location.href=this.to},handleCheckClick(t,e=!1){if(this.to){if("_blank"===this.target)return!1;t.preventDefault(),this.handleClick(e)}}}},u=n("9f81"),c="ivu-btn",d={name:"Button",mixins:[l,u["a"]],components:{Icon:a["a"]},props:{type:{validator:function(t){return Object(s["i"])(t,["default","primary","dashed","text","info","success","warning","error"])},default:"default"},shape:{validator:function(t){return Object(s["i"])(t,["circle","circle-outline"])}},size:{validator:function(t){return Object(s["i"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},loading:Boolean,disabled:Boolean,htmlType:{default:"button",validator:function(t){return Object(s["i"])(t,["button","submit","reset"])}},icon:{type:String,default:""},customIcon:{type:String,default:""},long:{type:Boolean,default:!1},ghost:{type:Boolean,default:!1}},computed:{showSlot:function(){return!!this.$slots.default},classes:function(){var t;return["".concat(c),"".concat(c,"-").concat(this.type),(t={},Object(o["a"])(t,"".concat(c,"-long"),this.long),Object(o["a"])(t,"".concat(c,"-").concat(this.shape),!!this.shape),Object(o["a"])(t,"".concat(c,"-").concat(this.size),"default"!==this.size),Object(o["a"])(t,"".concat(c,"-loading"),null!=this.loading&&this.loading),Object(o["a"])(t,"".concat(c,"-icon-only"),!this.showSlot&&(!!this.icon||!!this.customIcon||this.loading)),Object(o["a"])(t,"".concat(c,"-ghost"),this.ghost),t)]},isHrefPattern:function(){var t=this.to;return!!t},tagName:function(){var t=this.isHrefPattern;return t?"a":"button"},tagProps:function(){var t=this.isHrefPattern;if(t){var e=this.linkUrl,n=this.target;return{href:e,target:n}}var i=this.htmlType;return{type:i}}},methods:{handleClickLink:function(t){this.$emit("click",t);var e=t.ctrlKey||t.metaKey;this.handleCheckClick(t,e)}}},h=d,f=n("2877"),p=Object(f["a"])(h,i,r,!1,null,null,null);e["a"]=p.exports},"7a48":function(t,e,n){var i=n("6044"),r=Object.prototype,o=r.hasOwnProperty;function a(t){var e=this.__data__;return i?void 0!==e[t]:o.call(e,t)}t.exports=a},"7a77":function(t,e,n){"use strict";function i(t){this.message=t}i.prototype.toString=function(){return"Cancel"+(this.message?": "+this.message:"")},i.prototype.__CANCEL__=!0,t.exports=i},"7aa2":function(t,e,n){var i=n("ec47"),r=n("9934"),o=i(r);t.exports=o},"7aac":function(t,e,n){"use strict";var i=n("c532");t.exports=i.isStandardBrowserEnv()?function(){return{write:function(t,e,n,r,o,a){var s=[];s.push(t+"="+encodeURIComponent(e)),i.isNumber(n)&&s.push("expires="+new Date(n).toGMTString()),i.isString(r)&&s.push("path="+r),i.isString(o)&&s.push("domain="+o),!0===a&&s.push("secure"),document.cookie=s.join("; ")},read:function(t){var e=document.cookie.match(new RegExp("(^|;\\s*)("+t+")=([^;]*)"));return e?decodeURIComponent(e[3]):null},remove:function(t){this.write(t,"",Date.now()-864e5)}}}():function(){return{write:function(){},read:function(){return null},remove:function(){}}}()},"7b0b":function(t,e,n){var i=n("1d80");t.exports=function(t){return Object(i(t))}},"7b83":function(t,e,n){var i=n("7c64"),r=n("93ed"),o=n("2478"),a=n("a524"),s=n("1fc8");function l(t){var e=-1,n=null==t?0:t.length;this.clear();while(++e<n){var i=t[e];this.set(i[0],i[1])}}l.prototype.clear=i,l.prototype["delete"]=r,l.prototype.get=o,l.prototype.has=a,l.prototype.set=s,t.exports=l},"7b97":function(t,e,n){var i=n("7e64"),r=n("a2be"),o=n("1c3c"),a=n("b1e5"),s=n("42a2"),l=n("6747"),u=n("0d24"),c=n("73ac"),d=1,h="[object Arguments]",f="[object Array]",p="[object Object]",m=Object.prototype,g=m.hasOwnProperty;function y(t,e,n,m,y,v){var _=l(t),b=l(e),x=_?f:s(t),w=b?f:s(e);x=x==h?p:x,w=w==h?p:w;var M=x==p,S=w==p,L=x==w;if(L&&u(t)){if(!u(e))return!1;_=!0,M=!1}if(L&&!M)return v||(v=new i),_||c(t)?r(t,e,n,m,y,v):o(t,e,x,n,m,y,v);if(!(n&d)){var k=M&&g.call(t,"__wrapped__"),T=S&&g.call(e,"__wrapped__");if(k||T){var D=k?t.value():t,C=T?e.value():e;return v||(v=new i),y(D,C,n,m,v)}}return!!L&&(v||(v=new i),a(t,e,n,m,y,v))}t.exports=y},"7be6":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="január_február_marec_apríl_máj_jún_júl_august_september_október_november_december".split("_"),n="jan_feb_mar_apr_máj_jún_júl_aug_sep_okt_nov_dec".split("_");function i(t){return t>1&&t<5}function r(t,e,n,r){var o=t+" ";switch(n){case"s":return e||r?"pár sekúnd":"pár sekundami";case"ss":return e||r?o+(i(t)?"sekundy":"sekúnd"):o+"sekundami";case"m":return e?"minúta":r?"minútu":"minútou";case"mm":return e||r?o+(i(t)?"minúty":"minút"):o+"minútami";case"h":return e?"hodina":r?"hodinu":"hodinou";case"hh":return e||r?o+(i(t)?"hodiny":"hodín"):o+"hodinami";case"d":return e||r?"deň":"dňom";case"dd":return e||r?o+(i(t)?"dni":"dní"):o+"dňami";case"M":return e||r?"mesiac":"mesiacom";case"MM":return e||r?o+(i(t)?"mesiace":"mesiacov"):o+"mesiacmi";case"y":return e||r?"rok":"rokom";case"yy":return e||r?o+(i(t)?"roky":"rokov"):o+"rokmi"}}var o=t.defineLocale("sk",{months:e,monthsShort:n,weekdays:"nedeľa_pondelok_utorok_streda_štvrtok_piatok_sobota".split("_"),weekdaysShort:"ne_po_ut_st_št_pi_so".split("_"),weekdaysMin:"ne_po_ut_st_št_pi_so".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd D. MMMM YYYY H:mm"},calendar:{sameDay:"[dnes o] LT",nextDay:"[zajtra o] LT",nextWeek:function(){switch(this.day()){case 0:return"[v nedeľu o] LT";case 1:case 2:return"[v] dddd [o] LT";case 3:return"[v stredu o] LT";case 4:return"[vo štvrtok o] LT";case 5:return"[v piatok o] LT";case 6:return"[v sobotu o] LT"}},lastDay:"[včera o] LT",lastWeek:function(){switch(this.day()){case 0:return"[minulú nedeľu o] LT";case 1:case 2:return"[minulý] dddd [o] LT";case 3:return"[minulú stredu o] LT";case 4:case 5:return"[minulý] dddd [o] LT";case 6:return"[minulú sobotu o] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"pred %s",s:r,ss:r,m:r,mm:r,h:r,hh:r,d:r,dd:r,M:r,MM:r,y:r,yy:r},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return o}))},"7c64":function(t,e,n){var i=n("e24b"),r=n("5e2e"),o=n("79bc");function a(){this.size=0,this.__data__={hash:new i,map:new(o||r),string:new i}}t.exports=a},"7c73":function(t,e,n){var i,r=n("825a"),o=n("37e8"),a=n("7839"),s=n("d012"),l=n("1be4"),u=n("cc12"),c=n("f772"),d=">",h="<",f="prototype",p="script",m=c("IE_PROTO"),g=function(){},y=function(t){return h+p+d+t+h+"/"+p+d},v=function(t){t.write(y("")),t.close();var e=t.parentWindow.Object;return t=null,e},_=function(){var t,e=u("iframe"),n="java"+p+":";return e.style.display="none",l.appendChild(e),e.src=String(n),t=e.contentWindow.document,t.open(),t.write(y("document.F=Object")),t.close(),t.F},b=function(){try{i=document.domain&&new ActiveXObject("htmlfile")}catch(e){}b=i?v(i):_();var t=a.length;while(t--)delete b[f][a[t]];return b()};s[m]=!0,t.exports=Object.create||function(t,e){var n;return null!==t?(g[f]=r(t),n=new g,g[f]=null,n[m]=t):n=b(),void 0===e?n:o(n,e)}},"7d1f":function(t,e,n){var i=n("087d"),r=n("6747");function o(t,e,n){var o=e(t);return r(t)?o:i(o,n(t))}t.exports=o},"7dd0":function(t,e,n){"use strict";var i=n("23e7"),r=n("9ed3"),o=n("e163"),a=n("d2bb"),s=n("d44e"),l=n("9112"),u=n("6eeb"),c=n("b622"),d=n("c430"),h=n("3f8c"),f=n("ae93"),p=f.IteratorPrototype,m=f.BUGGY_SAFARI_ITERATORS,g=c("iterator"),y="keys",v="values",_="entries",b=function(){return this};t.exports=function(t,e,n,c,f,x,w){r(n,e,c);var M,S,L,k=function(t){if(t===f&&A)return A;if(!m&&t in C)return C[t];switch(t){case y:return function(){return new n(this,t)};case v:return function(){return new n(this,t)};case _:return function(){return new n(this,t)}}return function(){return new n(this)}},T=e+" Iterator",D=!1,C=t.prototype,I=C[g]||C["@@iterator"]||f&&C[f],A=!m&&I||k(f),Y="Array"==e&&C.entries||I;if(Y&&(M=o(Y.call(new t)),p!==Object.prototype&&M.next&&(d||o(M)===p||(a?a(M,p):"function"!=typeof M[g]&&l(M,g,b)),s(M,T,!0,!0),d&&(h[T]=b))),f==v&&I&&I.name!==v&&(D=!0,A=function(){return I.call(this)}),d&&!w||C[g]===A||l(C,g,A),h[e]=A,f)if(S={values:k(v),keys:x?A:k(y),entries:k(_)},w)for(L in S)(m||D||!(L in C))&&u(C,L,S[L]);else i({target:e,proto:!0,forced:m||D},S);return S}},"7e64":function(t,e,n){var i=n("5e2e"),r=n("efb6"),o=n("2fcc"),a=n("802a"),s=n("55a3"),l=n("d02c");function u(t){var e=this.__data__=new i(t);this.size=e.size}u.prototype.clear=r,u.prototype["delete"]=o,u.prototype.get=a,u.prototype.has=s,u.prototype.set=l,t.exports=u},"7ed2":function(t,e){var n="__lodash_hash_undefined__";function i(t){return this.__data__.set(t,n),this}t.exports=i},"7f33":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("yo",{months:"Sẹ́rẹ́_Èrèlè_Ẹrẹ̀nà_Ìgbé_Èbibi_Òkùdu_Agẹmo_Ògún_Owewe_Ọ̀wàrà_Bélú_Ọ̀pẹ̀̀".split("_"),monthsShort:"Sẹ́r_Èrl_Ẹrn_Ìgb_Èbi_Òkù_Agẹ_Ògú_Owe_Ọ̀wà_Bél_Ọ̀pẹ̀̀".split("_"),weekdays:"Àìkú_Ajé_Ìsẹ́gun_Ọjọ́rú_Ọjọ́bọ_Ẹtì_Àbámẹ́ta".split("_"),weekdaysShort:"Àìk_Ajé_Ìsẹ́_Ọjr_Ọjb_Ẹtì_Àbá".split("_"),weekdaysMin:"Àì_Aj_Ìs_Ọr_Ọb_Ẹt_Àb".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Ònì ni] LT",nextDay:"[Ọ̀la ni] LT",nextWeek:"dddd [Ọsẹ̀ tón'bọ] [ni] LT",lastDay:"[Àna ni] LT",lastWeek:"dddd [Ọsẹ̀ tólọ́] [ni] LT",sameElse:"L"},relativeTime:{future:"ní %s",past:"%s kọjá",s:"ìsẹjú aayá die",ss:"aayá %d",m:"ìsẹjú kan",mm:"ìsẹjú %d",h:"wákati kan",hh:"wákati %d",d:"ọjọ́ kan",dd:"ọjọ́ %d",M:"osù kan",MM:"osù %d",y:"ọdún kan",yy:"ọdún %d"},dayOfMonthOrdinalParse:/ọjọ́\s\d{1,2}/,ordinal:"ọjọ́ %d",week:{dow:1,doy:4}});return e}))},"7f9a":function(t,e,n){var i=n("da84"),r=n("8925"),o=i.WeakMap;t.exports="function"===typeof o&&/native code/.test(r(o))},"802a":function(t,e){function n(t){return this.__data__.get(t)}t.exports=n},8057:function(t,e){function n(t,e){var n=-1,i=null==t?0:t.length;while(++n<i)if(!1===e(t[n],n,t))break;return t}t.exports=n},8144:function(t,e,n){"use strict";var i=n("2430");e["a"]=i["a"]},8155:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r=t+" ";switch(n){case"s":return e||i?"nekaj sekund":"nekaj sekundami";case"ss":return r+=1===t?e?"sekundo":"sekundi":2===t?e||i?"sekundi":"sekundah":t<5?e||i?"sekunde":"sekundah":"sekund",r;case"m":return e?"ena minuta":"eno minuto";case"mm":return r+=1===t?e?"minuta":"minuto":2===t?e||i?"minuti":"minutama":t<5?e||i?"minute":"minutami":e||i?"minut":"minutami",r;case"h":return e?"ena ura":"eno uro";case"hh":return r+=1===t?e?"ura":"uro":2===t?e||i?"uri":"urama":t<5?e||i?"ure":"urami":e||i?"ur":"urami",r;case"d":return e||i?"en dan":"enim dnem";case"dd":return r+=1===t?e||i?"dan":"dnem":2===t?e||i?"dni":"dnevoma":e||i?"dni":"dnevi",r;case"M":return e||i?"en mesec":"enim mesecem";case"MM":return r+=1===t?e||i?"mesec":"mesecem":2===t?e||i?"meseca":"mesecema":t<5?e||i?"mesece":"meseci":e||i?"mesecev":"meseci",r;case"y":return e||i?"eno leto":"enim letom";case"yy":return r+=1===t?e||i?"leto":"letom":2===t?e||i?"leti":"letoma":t<5?e||i?"leta":"leti":e||i?"let":"leti",r}}var n=t.defineLocale("sl",{months:"januar_februar_marec_april_maj_junij_julij_avgust_september_oktober_november_december".split("_"),monthsShort:"jan._feb._mar._apr._maj._jun._jul._avg._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedelja_ponedeljek_torek_sreda_četrtek_petek_sobota".split("_"),weekdaysShort:"ned._pon._tor._sre._čet._pet._sob.".split("_"),weekdaysMin:"ne_po_to_sr_če_pe_so".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD. MM. YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[danes ob] LT",nextDay:"[jutri ob] LT",nextWeek:function(){switch(this.day()){case 0:return"[v] [nedeljo] [ob] LT";case 3:return"[v] [sredo] [ob] LT";case 6:return"[v] [soboto] [ob] LT";case 1:case 2:case 4:case 5:return"[v] dddd [ob] LT"}},lastDay:"[včeraj ob] LT",lastWeek:function(){switch(this.day()){case 0:return"[prejšnjo] [nedeljo] [ob] LT";case 3:return"[prejšnjo] [sredo] [ob] LT";case 6:return"[prejšnjo] [soboto] [ob] LT";case 1:case 2:case 4:case 5:return"[prejšnji] dddd [ob] LT"}},sameElse:"L"},relativeTime:{future:"čez %s",past:"pred %s",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},"81e9":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="nolla yksi kaksi kolme neljä viisi kuusi seitsemän kahdeksan yhdeksän".split(" "),n=["nolla","yhden","kahden","kolmen","neljän","viiden","kuuden",e[7],e[8],e[9]];function i(t,e,n,i){var o="";switch(n){case"s":return i?"muutaman sekunnin":"muutama sekunti";case"ss":o=i?"sekunnin":"sekuntia";break;case"m":return i?"minuutin":"minuutti";case"mm":o=i?"minuutin":"minuuttia";break;case"h":return i?"tunnin":"tunti";case"hh":o=i?"tunnin":"tuntia";break;case"d":return i?"päivän":"päivä";case"dd":o=i?"päivän":"päivää";break;case"M":return i?"kuukauden":"kuukausi";case"MM":o=i?"kuukauden":"kuukautta";break;case"y":return i?"vuoden":"vuosi";case"yy":o=i?"vuoden":"vuotta";break}return o=r(t,i)+" "+o,o}function r(t,i){return t<10?i?n[t]:e[t]:t}var o=t.defineLocale("fi",{months:"tammikuu_helmikuu_maaliskuu_huhtikuu_toukokuu_kesäkuu_heinäkuu_elokuu_syyskuu_lokakuu_marraskuu_joulukuu".split("_"),monthsShort:"tammi_helmi_maalis_huhti_touko_kesä_heinä_elo_syys_loka_marras_joulu".split("_"),weekdays:"sunnuntai_maanantai_tiistai_keskiviikko_torstai_perjantai_lauantai".split("_"),weekdaysShort:"su_ma_ti_ke_to_pe_la".split("_"),weekdaysMin:"su_ma_ti_ke_to_pe_la".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD.MM.YYYY",LL:"Do MMMM[ta] YYYY",LLL:"Do MMMM[ta] YYYY, [klo] HH.mm",LLLL:"dddd, Do MMMM[ta] YYYY, [klo] HH.mm",l:"D.M.YYYY",ll:"Do MMM YYYY",lll:"Do MMM YYYY, [klo] HH.mm",llll:"ddd, Do MMM YYYY, [klo] HH.mm"},calendar:{sameDay:"[tänään] [klo] LT",nextDay:"[huomenna] [klo] LT",nextWeek:"dddd [klo] LT",lastDay:"[eilen] [klo] LT",lastWeek:"[viime] dddd[na] [klo] LT",sameElse:"L"},relativeTime:{future:"%s päästä",past:"%s sitten",s:i,ss:i,m:i,mm:i,h:i,hh:i,d:i,dd:i,M:i,MM:i,y:i,yy:i},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return o}))},8230:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"١",2:"٢",3:"٣",4:"٤",5:"٥",6:"٦",7:"٧",8:"٨",9:"٩",0:"٠"},n={"١":"1","٢":"2","٣":"3","٤":"4","٥":"5","٦":"6","٧":"7","٨":"8","٩":"9","٠":"0"},i=t.defineLocale("ar-sa",{months:"يناير_فبراير_مارس_أبريل_مايو_يونيو_يوليو_أغسطس_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),monthsShort:"يناير_فبراير_مارس_أبريل_مايو_يونيو_يوليو_أغسطس_سبتمبر_أكتوبر_نوفمبر_ديسمبر".split("_"),weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم على الساعة] LT",nextDay:"[غدا على الساعة] LT",nextWeek:"dddd [على الساعة] LT",lastDay:"[أمس على الساعة] LT",lastWeek:"dddd [على الساعة] LT",sameElse:"L"},relativeTime:{future:"في %s",past:"منذ %s",s:"ثوان",ss:"%d ثانية",m:"دقيقة",mm:"%d دقائق",h:"ساعة",hh:"%d ساعات",d:"يوم",dd:"%d أيام",M:"شهر",MM:"%d أشهر",y:"سنة",yy:"%d سنوات"},preparse:function(t){return t.replace(/[١٢٣٤٥٦٧٨٩٠]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:0,doy:6}});return i}))},"825a":function(t,e,n){var i=n("861d");t.exports=function(t){if(!i(t))throw TypeError(String(t)+" is not an object");return t}},8296:function(t,e,n){var i=n("656b"),r=n("2b10");function o(t,e){return e.length<2?t:i(t,r(e,0,-1))}t.exports=o},8332:function(t,e,n){t.exports=n("7aa2")},"83ab":function(t,e,n){var i=n("d039");t.exports=!i((function(){return 7!=Object.defineProperty({},1,{get:function(){return 7}})[1]}))},"83b9":function(t,e,n){"use strict";var i=n("d925"),r=n("e683");t.exports=function(t,e){return t&&!i(e)?r(t,e):e}},8418:function(t,e,n){"use strict";var i=n("c04e"),r=n("9bf2"),o=n("5c6c");t.exports=function(t,e,n){var a=i(e);a in t?r.f(t,a,o(0,n)):t[a]=n}},"84aa":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("bg",{months:"януари_февруари_март_април_май_юни_юли_август_септември_октомври_ноември_декември".split("_"),monthsShort:"яну_фев_мар_апр_май_юни_юли_авг_сеп_окт_ное_дек".split("_"),weekdays:"неделя_понеделник_вторник_сряда_четвъртък_петък_събота".split("_"),weekdaysShort:"нед_пон_вто_сря_чет_пет_съб".split("_"),weekdaysMin:"нд_пн_вт_ср_чт_пт_сб".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY H:mm",LLLL:"dddd, D MMMM YYYY H:mm"},calendar:{sameDay:"[Днес в] LT",nextDay:"[Утре в] LT",nextWeek:"dddd [в] LT",lastDay:"[Вчера в] LT",lastWeek:function(){switch(this.day()){case 0:case 3:case 6:return"[Миналата] dddd [в] LT";case 1:case 2:case 4:case 5:return"[Миналия] dddd [в] LT"}},sameElse:"L"},relativeTime:{future:"след %s",past:"преди %s",s:"няколко секунди",ss:"%d секунди",m:"минута",mm:"%d минути",h:"час",hh:"%d часа",d:"ден",dd:"%d дена",w:"седмица",ww:"%d седмици",M:"месец",MM:"%d месеца",y:"година",yy:"%d години"},dayOfMonthOrdinalParse:/\d{1,2}-(ев|ен|ти|ви|ри|ми)/,ordinal:function(t){var e=t%10,n=t%100;return 0===t?t+"-ев":0===n?t+"-ен":n>10&&n<20?t+"-ти":1===e?t+"-ви":2===e?t+"-ри":7===e||8===e?t+"-ми":t+"-ти"},week:{dow:1,doy:7}});return e}))},"85e3":function(t,e){function n(t,e,n){switch(n.length){case 0:return t.call(e);case 1:return t.call(e,n[0]);case 2:return t.call(e,n[0],n[1]);case 3:return t.call(e,n[0],n[1],n[2])}return t.apply(e,n)}t.exports=n},8604:function(t,e,n){var i=n("26e8"),r=n("e2c0");function o(t,e){return null!=t&&r(t,e,i)}t.exports=o},"861d":function(t,e){t.exports=function(t){return"object"===typeof t?null!==t:"function"===typeof t}},8629:function(t,e,n){"use strict";var i=n("5334");e["a"]=i["a"]},8679:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.wrapClasses},[t._t("default")],2)},r=[],o=(n("45fc"),n("b0c0"),n("ade3")),a="ivu-layout",s={name:"Layout",data:function(){return{hasSider:!1}},computed:{wrapClasses:function(){return["".concat(a),Object(o["a"])({},"".concat(a,"-has-sider"),this.hasSider)]}},methods:{findSider:function(){return this.$children.some((function(t){return"Sider"===t.$options.name}))}},mounted:function(){this.hasSider=this.findSider()}},l=s,u=n("2877"),c=Object(u["a"])(l,i,r,!1,null,null,null),d=c.exports,h=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.wrapClasses},[t._t("default")],2)},f=[],p="ivu-layout",m={name:"Header",computed:{wrapClasses:function(){return"".concat(p,"-header")}}},g=m,y=Object(u["a"])(g,h,f,!1,null,null,null),v=y.exports,_=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.wrapClasses,style:t.wrapStyles},[n("span",{directives:[{name:"show",rawName:"v-show",value:t.showZeroTrigger,expression:"showZeroTrigger"}],class:t.zeroWidthTriggerClasses,on:{click:t.toggleCollapse}},[n("i",{staticClass:"ivu-icon ivu-icon-ios-menu"})]),n("div",{class:t.childClasses},[t._t("default")],2),t._t("trigger",[n("div",{directives:[{name:"show",rawName:"v-show",value:t.showBottomTrigger,expression:"showBottomTrigger"}],class:t.triggerClasses,style:{width:t.siderWidth+"px"},on:{click:t.toggleCollapse}},[n("i",{class:t.triggerIconClasses})])])],2)},b=[],x=(n("a9e3"),n("de48")),w=n("3a27"),M="ivu-layout-sider";Object(w["j"])();var S={name:"Sider",props:{value:{type:Boolean,default:!1},width:{type:[Number,String],default:200},collapsedWidth:{type:[Number,String],default:64},hideTrigger:{type:Boolean,default:!1},breakpoint:{type:String,validator:function(t){return Object(w["i"])(t,["xs","sm","md","lg","xl","xxl"])}},collapsible:{type:Boolean,default:!1},defaultCollapsed:{type:Boolean,default:!1},reverseArrow:{type:Boolean,default:!1}},data:function(){return{prefixCls:M,mediaMatched:!1}},computed:{wrapClasses:function(){return["".concat(M),this.siderWidth?"":"".concat(M,"-zero-width"),this.value?"".concat(M,"-collapsed"):""]},wrapStyles:function(){return{width:"".concat(this.siderWidth,"px"),minWidth:"".concat(this.siderWidth,"px"),maxWidth:"".concat(this.siderWidth,"px"),flex:"0 0 ".concat(this.siderWidth,"px")}},triggerClasses:function(){return["".concat(M,"-trigger"),this.value?"".concat(M,"-trigger-collapsed"):""]},childClasses:function(){return"".concat(this.prefixCls,"-children")},zeroWidthTriggerClasses:function(){return["".concat(M,"-zero-width-trigger"),this.reverseArrow?"".concat(M,"-zero-width-trigger-left"):""]},triggerIconClasses:function(){return["ivu-icon","ivu-icon-ios-arrow-".concat(this.reverseArrow?"forward":"back"),"".concat(M,"-trigger-icon")]},siderWidth:function(){return this.collapsible?this.value?this.mediaMatched?0:parseInt(this.collapsedWidth):parseInt(this.width):this.width},showZeroTrigger:function(){return!!this.collapsible&&(this.mediaMatched&&!this.hideTrigger||0===parseInt(this.collapsedWidth)&&this.value&&!this.hideTrigger)},showBottomTrigger:function(){return!!this.collapsible&&(!this.mediaMatched&&!this.hideTrigger)}},methods:{toggleCollapse:function(){var t=!!this.collapsible&&!this.value;this.$emit("input",t)},matchMedia:function(){var t;window.matchMedia&&(t=window.matchMedia);var e=this.mediaMatched;this.mediaMatched=t("(max-width: ".concat(w["b"][this.breakpoint],")")).matches,this.mediaMatched!==e&&this.$emit("input",this.mediaMatched)},onWindowResize:function(){this.matchMedia()}},watch:{value:function(t){this.$emit("on-collapse",t)}},mounted:function(){this.defaultCollapsed&&this.$emit("input",this.defaultCollapsed),void 0!==this.breakpoint&&(Object(x["b"])(window,"resize",this.onWindowResize),this.matchMedia())},beforeDestroy:function(){void 0!==this.breakpoint&&Object(x["a"])(window,"resize",this.onWindowResize)}},L=S,k=Object(u["a"])(L,_,b,!1,null,null,null),T=k.exports,D=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.wrapClasses},[t._t("default")],2)},C=[],I="ivu-layout",A={name:"Content",computed:{wrapClasses:function(){return"".concat(I,"-content")}}},Y=A,O=Object(u["a"])(Y,D,C,!1,null,null,null),P=O.exports,E=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.wrapClasses},[t._t("default")],2)},N=[],R="ivu-layout",H={name:"Footer",computed:{wrapClasses:function(){return"".concat(R,"-footer")}}},j=H,z=Object(u["a"])(j,E,N,!1,null,null,null),W=z.exports;d.Header=v,d.Sider=T,d.Content=P,d.Footer=W;e["a"]=d},8689:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"၁",2:"၂",3:"၃",4:"၄",5:"၅",6:"၆",7:"၇",8:"၈",9:"၉",0:"၀"},n={"၁":"1","၂":"2","၃":"3","၄":"4","၅":"5","၆":"6","၇":"7","၈":"8","၉":"9","၀":"0"},i=t.defineLocale("my",{months:"ဇန်နဝါရီ_ဖေဖော်ဝါရီ_မတ်_ဧပြီ_မေ_ဇွန်_ဇူလိုင်_သြဂုတ်_စက်တင်ဘာ_အောက်တိုဘာ_နိုဝင်ဘာ_ဒီဇင်ဘာ".split("_"),monthsShort:"ဇန်_ဖေ_မတ်_ပြီ_မေ_ဇွန်_လိုင်_သြ_စက်_အောက်_နို_ဒီ".split("_"),weekdays:"တနင်္ဂနွေ_တနင်္လာ_အင်္ဂါ_ဗုဒ္ဓဟူး_ကြာသပတေး_သောကြာ_စနေ".split("_"),weekdaysShort:"နွေ_လာ_ဂါ_ဟူး_ကြာ_သော_နေ".split("_"),weekdaysMin:"နွေ_လာ_ဂါ_ဟူး_ကြာ_သော_နေ".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[ယနေ.] LT [မှာ]",nextDay:"[မနက်ဖြန်] LT [မှာ]",nextWeek:"dddd LT [မှာ]",lastDay:"[မနေ.က] LT [မှာ]",lastWeek:"[ပြီးခဲ့သော] dddd LT [မှာ]",sameElse:"L"},relativeTime:{future:"လာမည့် %s မှာ",past:"လွန်ခဲ့သော %s က",s:"စက္ကန်.အနည်းငယ်",ss:"%d စက္ကန့်",m:"တစ်မိနစ်",mm:"%d မိနစ်",h:"တစ်နာရီ",hh:"%d နာရီ",d:"တစ်ရက်",dd:"%d ရက်",M:"တစ်လ",MM:"%d လ",y:"တစ်နှစ်",yy:"%d နှစ်"},preparse:function(t){return t.replace(/[၁၂၃၄၅၆၇၈၉၀]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},week:{dow:1,doy:4}});return i}))},"86e1":function(t,e,n){var i=n("85e3"),r=n("e2e4"),o=n("4416"),a=n("8296"),s=n("f4d6");function l(t,e,n){e=r(e,t),t=a(t,e);var l=null==t?t:t[s(o(e))];return null==l?void 0:i(l,t,n)}t.exports=l},"872a":function(t,e,n){var i=n("3b4a");function r(t,e,n){"__proto__"==e&&i?i(t,e,{configurable:!0,enumerable:!0,value:n,writable:!0}):t[e]=n}t.exports=r},8840:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("gl",{months:"xaneiro_febreiro_marzo_abril_maio_xuño_xullo_agosto_setembro_outubro_novembro_decembro".split("_"),monthsShort:"xan._feb._mar._abr._mai._xuñ._xul._ago._set._out._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"domingo_luns_martes_mércores_xoves_venres_sábado".split("_"),weekdaysShort:"dom._lun._mar._mér._xov._ven._sáb.".split("_"),weekdaysMin:"do_lu_ma_mé_xo_ve_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY H:mm",LLLL:"dddd, D [de] MMMM [de] YYYY H:mm"},calendar:{sameDay:function(){return"[hoxe "+(1!==this.hours()?"ás":"á")+"] LT"},nextDay:function(){return"[mañá "+(1!==this.hours()?"ás":"á")+"] LT"},nextWeek:function(){return"dddd ["+(1!==this.hours()?"ás":"a")+"] LT"},lastDay:function(){return"[onte "+(1!==this.hours()?"á":"a")+"] LT"},lastWeek:function(){return"[o] dddd [pasado "+(1!==this.hours()?"ás":"a")+"] LT"},sameElse:"L"},relativeTime:{future:function(t){return 0===t.indexOf("un")?"n"+t:"en "+t},past:"hai %s",s:"uns segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"unha hora",hh:"%d horas",d:"un día",dd:"%d días",M:"un mes",MM:"%d meses",y:"un ano",yy:"%d anos"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},8925:function(t,e,n){var i=n("c6cd"),r=Function.toString;"function"!=typeof i.inspectSource&&(i.inspectSource=function(t){return r.call(t)}),t.exports=i.inspectSource},"898b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),i=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],r=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,i){return t?/-MMM-/.test(i)?n[t.month()]:e[t.month()]:e},monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY H:mm",LLLL:"dddd, D [de] MMMM [de] YYYY H:mm"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4},invalidDate:"Fecha inválida"});return o}))},"89d9":function(t,e,n){var i=n("656b"),r=n("159a"),o=n("e2e4");function a(t,e,n){var a=-1,s=e.length,l={};while(++a<s){var u=e[a],c=i(t,u);n(c,u)&&r(l,o(u,t),c)}return l}t.exports=a},"8aa5":function(t,e,n){"use strict";var i=n("6547").charAt;t.exports=function(t,e,n){return e+(n?i(t,e).length:1)}},"8adb":function(t,e){function n(t,e){if(("constructor"!==e||"function"!==typeof t[e])&&"__proto__"!=e)return t[e]}t.exports=n},"8ba4":function(t,e,n){var i=n("23e7"),r=n("5e89");i({target:"Number",stat:!0},{isInteger:r})},"8c4f":function(t,e,n){"use strict";
/*!
  * vue-router v3.5.0
  * (c) 2021 Evan You
  * @license MIT
  */function i(t,e){0}function r(t,e){for(var n in e)t[n]=e[n];return t}var o=/[!'()*]/g,a=function(t){return"%"+t.charCodeAt(0).toString(16)},s=/%2C/g,l=function(t){return encodeURIComponent(t).replace(o,a).replace(s,",")};function u(t){try{return decodeURIComponent(t)}catch(e){0}return t}function c(t,e,n){void 0===e&&(e={});var i,r=n||h;try{i=r(t||"")}catch(s){i={}}for(var o in e){var a=e[o];i[o]=Array.isArray(a)?a.map(d):d(a)}return i}var d=function(t){return null==t||"object"===typeof t?t:String(t)};function h(t){var e={};return t=t.trim().replace(/^(\?|#|&)/,""),t?(t.split("&").forEach((function(t){var n=t.replace(/\+/g," ").split("="),i=u(n.shift()),r=n.length>0?u(n.join("=")):null;void 0===e[i]?e[i]=r:Array.isArray(e[i])?e[i].push(r):e[i]=[e[i],r]})),e):e}function f(t){var e=t?Object.keys(t).map((function(e){var n=t[e];if(void 0===n)return"";if(null===n)return l(e);if(Array.isArray(n)){var i=[];return n.forEach((function(t){void 0!==t&&(null===t?i.push(l(e)):i.push(l(e)+"="+l(t)))})),i.join("&")}return l(e)+"="+l(n)})).filter((function(t){return t.length>0})).join("&"):null;return e?"?"+e:""}var p=/\/?$/;function m(t,e,n,i){var r=i&&i.options.stringifyQuery,o=e.query||{};try{o=g(o)}catch(s){}var a={name:e.name||t&&t.name,meta:t&&t.meta||{},path:e.path||"/",hash:e.hash||"",query:o,params:e.params||{},fullPath:_(e,r),matched:t?v(t):[]};return n&&(a.redirectedFrom=_(n,r)),Object.freeze(a)}function g(t){if(Array.isArray(t))return t.map(g);if(t&&"object"===typeof t){var e={};for(var n in t)e[n]=g(t[n]);return e}return t}var y=m(null,{path:"/"});function v(t){var e=[];while(t)e.unshift(t),t=t.parent;return e}function _(t,e){var n=t.path,i=t.query;void 0===i&&(i={});var r=t.hash;void 0===r&&(r="");var o=e||f;return(n||"/")+o(i)+r}function b(t,e,n){return e===y?t===e:!!e&&(t.path&&e.path?t.path.replace(p,"")===e.path.replace(p,"")&&(n||t.hash===e.hash&&x(t.query,e.query)):!(!t.name||!e.name)&&(t.name===e.name&&(n||t.hash===e.hash&&x(t.query,e.query)&&x(t.params,e.params))))}function x(t,e){if(void 0===t&&(t={}),void 0===e&&(e={}),!t||!e)return t===e;var n=Object.keys(t).sort(),i=Object.keys(e).sort();return n.length===i.length&&n.every((function(n,r){var o=t[n],a=i[r];if(a!==n)return!1;var s=e[n];return null==o||null==s?o===s:"object"===typeof o&&"object"===typeof s?x(o,s):String(o)===String(s)}))}function w(t,e){return 0===t.path.replace(p,"/").indexOf(e.path.replace(p,"/"))&&(!e.hash||t.hash===e.hash)&&M(t.query,e.query)}function M(t,e){for(var n in e)if(!(n in t))return!1;return!0}function S(t){for(var e=0;e<t.matched.length;e++){var n=t.matched[e];for(var i in n.instances){var r=n.instances[i],o=n.enteredCbs[i];if(r&&o){delete n.enteredCbs[i];for(var a=0;a<o.length;a++)r._isBeingDestroyed||o[a](r)}}}}var L={name:"RouterView",functional:!0,props:{name:{type:String,default:"default"}},render:function(t,e){var n=e.props,i=e.children,o=e.parent,a=e.data;a.routerView=!0;var s=o.$createElement,l=n.name,u=o.$route,c=o._routerViewCache||(o._routerViewCache={}),d=0,h=!1;while(o&&o._routerRoot!==o){var f=o.$vnode?o.$vnode.data:{};f.routerView&&d++,f.keepAlive&&o._directInactive&&o._inactive&&(h=!0),o=o.$parent}if(a.routerViewDepth=d,h){var p=c[l],m=p&&p.component;return m?(p.configProps&&k(m,a,p.route,p.configProps),s(m,a,i)):s()}var g=u.matched[d],y=g&&g.components[l];if(!g||!y)return c[l]=null,s();c[l]={component:y},a.registerRouteInstance=function(t,e){var n=g.instances[l];(e&&n!==t||!e&&n===t)&&(g.instances[l]=e)},(a.hook||(a.hook={})).prepatch=function(t,e){g.instances[l]=e.componentInstance},a.hook.init=function(t){t.data.keepAlive&&t.componentInstance&&t.componentInstance!==g.instances[l]&&(g.instances[l]=t.componentInstance),S(u)};var v=g.props&&g.props[l];return v&&(r(c[l],{route:u,configProps:v}),k(y,a,u,v)),s(y,a,i)}};function k(t,e,n,i){var o=e.props=T(n,i);if(o){o=e.props=r({},o);var a=e.attrs=e.attrs||{};for(var s in o)t.props&&s in t.props||(a[s]=o[s],delete o[s])}}function T(t,e){switch(typeof e){case"undefined":return;case"object":return e;case"function":return e(t);case"boolean":return e?t.params:void 0;default:0}}function D(t,e,n){var i=t.charAt(0);if("/"===i)return t;if("?"===i||"#"===i)return e+t;var r=e.split("/");n&&r[r.length-1]||r.pop();for(var o=t.replace(/^\//,"").split("/"),a=0;a<o.length;a++){var s=o[a];".."===s?r.pop():"."!==s&&r.push(s)}return""!==r[0]&&r.unshift(""),r.join("/")}function C(t){var e="",n="",i=t.indexOf("#");i>=0&&(e=t.slice(i),t=t.slice(0,i));var r=t.indexOf("?");return r>=0&&(n=t.slice(r+1),t=t.slice(0,r)),{path:t,query:n,hash:e}}function I(t){return t.replace(/\/\//g,"/")}var A=Array.isArray||function(t){return"[object Array]"==Object.prototype.toString.call(t)},Y=J,O=H,P=j,E=B,N=Z,R=new RegExp(["(\\\\.)","([\\/.])?(?:(?:\\:(\\w+)(?:\\(((?:\\\\.|[^\\\\()])+)\\))?|\\(((?:\\\\.|[^\\\\()])+)\\))([+*?])?|(\\*))"].join("|"),"g");function H(t,e){var n,i=[],r=0,o=0,a="",s=e&&e.delimiter||"/";while(null!=(n=R.exec(t))){var l=n[0],u=n[1],c=n.index;if(a+=t.slice(o,c),o=c+l.length,u)a+=u[1];else{var d=t[o],h=n[2],f=n[3],p=n[4],m=n[5],g=n[6],y=n[7];a&&(i.push(a),a="");var v=null!=h&&null!=d&&d!==h,_="+"===g||"*"===g,b="?"===g||"*"===g,x=n[2]||s,w=p||m;i.push({name:f||r++,prefix:h||"",delimiter:x,optional:b,repeat:_,partial:v,asterisk:!!y,pattern:w?V(w):y?".*":"[^"+F(x)+"]+?"})}}return o<t.length&&(a+=t.substr(o)),a&&i.push(a),i}function j(t,e){return B(H(t,e),e)}function z(t){return encodeURI(t).replace(/[\/?#]/g,(function(t){return"%"+t.charCodeAt(0).toString(16).toUpperCase()}))}function W(t){return encodeURI(t).replace(/[?#]/g,(function(t){return"%"+t.charCodeAt(0).toString(16).toUpperCase()}))}function B(t,e){for(var n=new Array(t.length),i=0;i<t.length;i++)"object"===typeof t[i]&&(n[i]=new RegExp("^(?:"+t[i].pattern+")$",$(e)));return function(e,i){for(var r="",o=e||{},a=i||{},s=a.pretty?z:encodeURIComponent,l=0;l<t.length;l++){var u=t[l];if("string"!==typeof u){var c,d=o[u.name];if(null==d){if(u.optional){u.partial&&(r+=u.prefix);continue}throw new TypeError('Expected "'+u.name+'" to be defined')}if(A(d)){if(!u.repeat)throw new TypeError('Expected "'+u.name+'" to not repeat, but received `'+JSON.stringify(d)+"`");if(0===d.length){if(u.optional)continue;throw new TypeError('Expected "'+u.name+'" to not be empty')}for(var h=0;h<d.length;h++){if(c=s(d[h]),!n[l].test(c))throw new TypeError('Expected all "'+u.name+'" to match "'+u.pattern+'", but received `'+JSON.stringify(c)+"`");r+=(0===h?u.prefix:u.delimiter)+c}}else{if(c=u.asterisk?W(d):s(d),!n[l].test(c))throw new TypeError('Expected "'+u.name+'" to match "'+u.pattern+'", but received "'+c+'"');r+=u.prefix+c}}else r+=u}return r}}function F(t){return t.replace(/([.+*?=^!:${}()[\]|\/\\])/g,"\\$1")}function V(t){return t.replace(/([=!:$\/()])/g,"\\$1")}function G(t,e){return t.keys=e,t}function $(t){return t&&t.sensitive?"":"i"}function U(t,e){var n=t.source.match(/\((?!\?)/g);if(n)for(var i=0;i<n.length;i++)e.push({name:i,prefix:null,delimiter:null,optional:!1,repeat:!1,partial:!1,asterisk:!1,pattern:null});return G(t,e)}function q(t,e,n){for(var i=[],r=0;r<t.length;r++)i.push(J(t[r],e,n).source);var o=new RegExp("(?:"+i.join("|")+")",$(n));return G(o,e)}function X(t,e,n){return Z(H(t,n),e,n)}function Z(t,e,n){A(e)||(n=e||n,e=[]),n=n||{};for(var i=n.strict,r=!1!==n.end,o="",a=0;a<t.length;a++){var s=t[a];if("string"===typeof s)o+=F(s);else{var l=F(s.prefix),u="(?:"+s.pattern+")";e.push(s),s.repeat&&(u+="(?:"+l+u+")*"),u=s.optional?s.partial?l+"("+u+")?":"(?:"+l+"("+u+"))?":l+"("+u+")",o+=u}}var c=F(n.delimiter||"/"),d=o.slice(-c.length)===c;return i||(o=(d?o.slice(0,-c.length):o)+"(?:"+c+"(?=$))?"),o+=r?"$":i&&d?"":"(?="+c+"|$)",G(new RegExp("^"+o,$(n)),e)}function J(t,e,n){return A(e)||(n=e||n,e=[]),n=n||{},t instanceof RegExp?U(t,e):A(t)?q(t,e,n):X(t,e,n)}Y.parse=O,Y.compile=P,Y.tokensToFunction=E,Y.tokensToRegExp=N;var K=Object.create(null);function Q(t,e,n){e=e||{};try{var i=K[t]||(K[t]=Y.compile(t));return"string"===typeof e.pathMatch&&(e[0]=e.pathMatch),i(e,{pretty:!0})}catch(r){return""}finally{delete e[0]}}function tt(t,e,n,i){var o="string"===typeof t?{path:t}:t;if(o._normalized)return o;if(o.name){o=r({},t);var a=o.params;return a&&"object"===typeof a&&(o.params=r({},a)),o}if(!o.path&&o.params&&e){o=r({},o),o._normalized=!0;var s=r(r({},e.params),o.params);if(e.name)o.name=e.name,o.params=s;else if(e.matched.length){var l=e.matched[e.matched.length-1].path;o.path=Q(l,s,"path "+e.path)}else 0;return o}var u=C(o.path||""),d=e&&e.path||"/",h=u.path?D(u.path,d,n||o.append):d,f=c(u.query,o.query,i&&i.options.parseQuery),p=o.hash||u.hash;return p&&"#"!==p.charAt(0)&&(p="#"+p),{_normalized:!0,path:h,query:f,hash:p}}var et,nt=[String,Object],it=[String,Array],rt=function(){},ot={name:"RouterLink",props:{to:{type:nt,required:!0},tag:{type:String,default:"a"},custom:Boolean,exact:Boolean,exactPath:Boolean,append:Boolean,replace:Boolean,activeClass:String,exactActiveClass:String,ariaCurrentValue:{type:String,default:"page"},event:{type:it,default:"click"}},render:function(t){var e=this,n=this.$router,i=this.$route,o=n.resolve(this.to,i,this.append),a=o.location,s=o.route,l=o.href,u={},c=n.options.linkActiveClass,d=n.options.linkExactActiveClass,h=null==c?"router-link-active":c,f=null==d?"router-link-exact-active":d,p=null==this.activeClass?h:this.activeClass,g=null==this.exactActiveClass?f:this.exactActiveClass,y=s.redirectedFrom?m(null,tt(s.redirectedFrom),null,n):s;u[g]=b(i,y,this.exactPath),u[p]=this.exact||this.exactPath?u[g]:w(i,y);var v=u[g]?this.ariaCurrentValue:null,_=function(t){at(t)&&(e.replace?n.replace(a,rt):n.push(a,rt))},x={click:at};Array.isArray(this.event)?this.event.forEach((function(t){x[t]=_})):x[this.event]=_;var M={class:u},S=!this.$scopedSlots.$hasNormal&&this.$scopedSlots.default&&this.$scopedSlots.default({href:l,route:s,navigate:_,isActive:u[p],isExactActive:u[g]});if(S){if(1===S.length)return S[0];if(S.length>1||!S.length)return 0===S.length?t():t("span",{},S)}if("a"===this.tag)M.on=x,M.attrs={href:l,"aria-current":v};else{var L=st(this.$slots.default);if(L){L.isStatic=!1;var k=L.data=r({},L.data);for(var T in k.on=k.on||{},k.on){var D=k.on[T];T in x&&(k.on[T]=Array.isArray(D)?D:[D])}for(var C in x)C in k.on?k.on[C].push(x[C]):k.on[C]=_;var I=L.data.attrs=r({},L.data.attrs);I.href=l,I["aria-current"]=v}else M.on=x}return t(this.tag,M,this.$slots.default)}};function at(t){if(!(t.metaKey||t.altKey||t.ctrlKey||t.shiftKey)&&!t.defaultPrevented&&(void 0===t.button||0===t.button)){if(t.currentTarget&&t.currentTarget.getAttribute){var e=t.currentTarget.getAttribute("target");if(/\b_blank\b/i.test(e))return}return t.preventDefault&&t.preventDefault(),!0}}function st(t){if(t)for(var e,n=0;n<t.length;n++){if(e=t[n],"a"===e.tag)return e;if(e.children&&(e=st(e.children)))return e}}function lt(t){if(!lt.installed||et!==t){lt.installed=!0,et=t;var e=function(t){return void 0!==t},n=function(t,n){var i=t.$options._parentVnode;e(i)&&e(i=i.data)&&e(i=i.registerRouteInstance)&&i(t,n)};t.mixin({beforeCreate:function(){e(this.$options.router)?(this._routerRoot=this,this._router=this.$options.router,this._router.init(this),t.util.defineReactive(this,"_route",this._router.history.current)):this._routerRoot=this.$parent&&this.$parent._routerRoot||this,n(this,this)},destroyed:function(){n(this)}}),Object.defineProperty(t.prototype,"$router",{get:function(){return this._routerRoot._router}}),Object.defineProperty(t.prototype,"$route",{get:function(){return this._routerRoot._route}}),t.component("RouterView",L),t.component("RouterLink",ot);var i=t.config.optionMergeStrategies;i.beforeRouteEnter=i.beforeRouteLeave=i.beforeRouteUpdate=i.created}}var ut="undefined"!==typeof window;function ct(t,e,n,i,r){var o=e||[],a=n||Object.create(null),s=i||Object.create(null);t.forEach((function(t){dt(o,a,s,t,r)}));for(var l=0,u=o.length;l<u;l++)"*"===o[l]&&(o.push(o.splice(l,1)[0]),u--,l--);return{pathList:o,pathMap:a,nameMap:s}}function dt(t,e,n,i,r,o){var a=i.path,s=i.name;var l=i.pathToRegexpOptions||{},u=ft(a,r,l.strict);"boolean"===typeof i.caseSensitive&&(l.sensitive=i.caseSensitive);var c={path:u,regex:ht(u,l),components:i.components||{default:i.component},alias:i.alias?"string"===typeof i.alias?[i.alias]:i.alias:[],instances:{},enteredCbs:{},name:s,parent:r,matchAs:o,redirect:i.redirect,beforeEnter:i.beforeEnter,meta:i.meta||{},props:null==i.props?{}:i.components?i.props:{default:i.props}};if(i.children&&i.children.forEach((function(i){var r=o?I(o+"/"+i.path):void 0;dt(t,e,n,i,c,r)})),e[c.path]||(t.push(c.path),e[c.path]=c),void 0!==i.alias)for(var d=Array.isArray(i.alias)?i.alias:[i.alias],h=0;h<d.length;++h){var f=d[h];0;var p={path:f,children:i.children};dt(t,e,n,p,r,c.path||"/")}s&&(n[s]||(n[s]=c))}function ht(t,e){var n=Y(t,[],e);return n}function ft(t,e,n){return n||(t=t.replace(/\/$/,"")),"/"===t[0]||null==e?t:I(e.path+"/"+t)}function pt(t,e){var n=ct(t),i=n.pathList,r=n.pathMap,o=n.nameMap;function a(t){ct(t,i,r,o)}function s(t,e){var n="object"!==typeof t?o[t]:void 0;ct([e||t],i,r,o,n),n&&ct(n.alias.map((function(t){return{path:t,children:[e]}})),i,r,o,n)}function l(){return i.map((function(t){return r[t]}))}function u(t,n,a){var s=tt(t,n,!1,e),l=s.name;if(l){var u=o[l];if(!u)return h(null,s);var c=u.regex.keys.filter((function(t){return!t.optional})).map((function(t){return t.name}));if("object"!==typeof s.params&&(s.params={}),n&&"object"===typeof n.params)for(var d in n.params)!(d in s.params)&&c.indexOf(d)>-1&&(s.params[d]=n.params[d]);return s.path=Q(u.path,s.params,'named route "'+l+'"'),h(u,s,a)}if(s.path){s.params={};for(var f=0;f<i.length;f++){var p=i[f],m=r[p];if(mt(m.regex,s.path,s.params))return h(m,s,a)}}return h(null,s)}function c(t,n){var i=t.redirect,r="function"===typeof i?i(m(t,n,null,e)):i;if("string"===typeof r&&(r={path:r}),!r||"object"!==typeof r)return h(null,n);var a=r,s=a.name,l=a.path,c=n.query,d=n.hash,f=n.params;if(c=a.hasOwnProperty("query")?a.query:c,d=a.hasOwnProperty("hash")?a.hash:d,f=a.hasOwnProperty("params")?a.params:f,s){o[s];return u({_normalized:!0,name:s,query:c,hash:d,params:f},void 0,n)}if(l){var p=gt(l,t),g=Q(p,f,'redirect route with path "'+p+'"');return u({_normalized:!0,path:g,query:c,hash:d},void 0,n)}return h(null,n)}function d(t,e,n){var i=Q(n,e.params,'aliased route with path "'+n+'"'),r=u({_normalized:!0,path:i});if(r){var o=r.matched,a=o[o.length-1];return e.params=r.params,h(a,e)}return h(null,e)}function h(t,n,i){return t&&t.redirect?c(t,i||n):t&&t.matchAs?d(t,n,t.matchAs):m(t,n,i,e)}return{match:u,addRoute:s,getRoutes:l,addRoutes:a}}function mt(t,e,n){var i=e.match(t);if(!i)return!1;if(!n)return!0;for(var r=1,o=i.length;r<o;++r){var a=t.keys[r-1];a&&(n[a.name||"pathMatch"]="string"===typeof i[r]?u(i[r]):i[r])}return!0}function gt(t,e){return D(t,e.parent?e.parent.path:"/",!0)}var yt=ut&&window.performance&&window.performance.now?window.performance:Date;function vt(){return yt.now().toFixed(3)}var _t=vt();function bt(){return _t}function xt(t){return _t=t}var wt=Object.create(null);function Mt(){"scrollRestoration"in window.history&&(window.history.scrollRestoration="manual");var t=window.location.protocol+"//"+window.location.host,e=window.location.href.replace(t,""),n=r({},window.history.state);return n.key=bt(),window.history.replaceState(n,"",e),window.addEventListener("popstate",kt),function(){window.removeEventListener("popstate",kt)}}function St(t,e,n,i){if(t.app){var r=t.options.scrollBehavior;r&&t.app.$nextTick((function(){var o=Tt(),a=r.call(t,e,n,i?o:null);a&&("function"===typeof a.then?a.then((function(t){Pt(t,o)})).catch((function(t){0})):Pt(a,o))}))}}function Lt(){var t=bt();t&&(wt[t]={x:window.pageXOffset,y:window.pageYOffset})}function kt(t){Lt(),t.state&&t.state.key&&xt(t.state.key)}function Tt(){var t=bt();if(t)return wt[t]}function Dt(t,e){var n=document.documentElement,i=n.getBoundingClientRect(),r=t.getBoundingClientRect();return{x:r.left-i.left-e.x,y:r.top-i.top-e.y}}function Ct(t){return Yt(t.x)||Yt(t.y)}function It(t){return{x:Yt(t.x)?t.x:window.pageXOffset,y:Yt(t.y)?t.y:window.pageYOffset}}function At(t){return{x:Yt(t.x)?t.x:0,y:Yt(t.y)?t.y:0}}function Yt(t){return"number"===typeof t}var Ot=/^#\d/;function Pt(t,e){var n="object"===typeof t;if(n&&"string"===typeof t.selector){var i=Ot.test(t.selector)?document.getElementById(t.selector.slice(1)):document.querySelector(t.selector);if(i){var r=t.offset&&"object"===typeof t.offset?t.offset:{};r=At(r),e=Dt(i,r)}else Ct(t)&&(e=It(t))}else n&&Ct(t)&&(e=It(t));e&&("scrollBehavior"in document.documentElement.style?window.scrollTo({left:e.x,top:e.y,behavior:t.behavior}):window.scrollTo(e.x,e.y))}var Et=ut&&function(){var t=window.navigator.userAgent;return(-1===t.indexOf("Android 2.")&&-1===t.indexOf("Android 4.0")||-1===t.indexOf("Mobile Safari")||-1!==t.indexOf("Chrome")||-1!==t.indexOf("Windows Phone"))&&(window.history&&"function"===typeof window.history.pushState)}();function Nt(t,e){Lt();var n=window.history;try{if(e){var i=r({},n.state);i.key=bt(),n.replaceState(i,"",t)}else n.pushState({key:xt(vt())},"",t)}catch(o){window.location[e?"replace":"assign"](t)}}function Rt(t){Nt(t,!0)}function Ht(t,e,n){var i=function(r){r>=t.length?n():t[r]?e(t[r],(function(){i(r+1)})):i(r+1)};i(0)}var jt={redirected:2,aborted:4,cancelled:8,duplicated:16};function zt(t,e){return Vt(t,e,jt.redirected,'Redirected when going from "'+t.fullPath+'" to "'+$t(e)+'" via a navigation guard.')}function Wt(t,e){var n=Vt(t,e,jt.duplicated,'Avoided redundant navigation to current location: "'+t.fullPath+'".');return n.name="NavigationDuplicated",n}function Bt(t,e){return Vt(t,e,jt.cancelled,'Navigation cancelled from "'+t.fullPath+'" to "'+e.fullPath+'" with a new navigation.')}function Ft(t,e){return Vt(t,e,jt.aborted,'Navigation aborted from "'+t.fullPath+'" to "'+e.fullPath+'" via a navigation guard.')}function Vt(t,e,n,i){var r=new Error(i);return r._isRouter=!0,r.from=t,r.to=e,r.type=n,r}var Gt=["params","query","hash"];function $t(t){if("string"===typeof t)return t;if("path"in t)return t.path;var e={};return Gt.forEach((function(n){n in t&&(e[n]=t[n])})),JSON.stringify(e,null,2)}function Ut(t){return Object.prototype.toString.call(t).indexOf("Error")>-1}function qt(t,e){return Ut(t)&&t._isRouter&&(null==e||t.type===e)}function Xt(t){return function(e,n,i){var r=!1,o=0,a=null;Zt(t,(function(t,e,n,s){if("function"===typeof t&&void 0===t.cid){r=!0,o++;var l,u=te((function(e){Qt(e)&&(e=e.default),t.resolved="function"===typeof e?e:et.extend(e),n.components[s]=e,o--,o<=0&&i()})),c=te((function(t){var e="Failed to resolve async component "+s+": "+t;a||(a=Ut(t)?t:new Error(e),i(a))}));try{l=t(u,c)}catch(h){c(h)}if(l)if("function"===typeof l.then)l.then(u,c);else{var d=l.component;d&&"function"===typeof d.then&&d.then(u,c)}}})),r||i()}}function Zt(t,e){return Jt(t.map((function(t){return Object.keys(t.components).map((function(n){return e(t.components[n],t.instances[n],t,n)}))})))}function Jt(t){return Array.prototype.concat.apply([],t)}var Kt="function"===typeof Symbol&&"symbol"===typeof Symbol.toStringTag;function Qt(t){return t.__esModule||Kt&&"Module"===t[Symbol.toStringTag]}function te(t){var e=!1;return function(){var n=[],i=arguments.length;while(i--)n[i]=arguments[i];if(!e)return e=!0,t.apply(this,n)}}var ee=function(t,e){this.router=t,this.base=ne(e),this.current=y,this.pending=null,this.ready=!1,this.readyCbs=[],this.readyErrorCbs=[],this.errorCbs=[],this.listeners=[]};function ne(t){if(!t)if(ut){var e=document.querySelector("base");t=e&&e.getAttribute("href")||"/",t=t.replace(/^https?:\/\/[^\/]+/,"")}else t="/";return"/"!==t.charAt(0)&&(t="/"+t),t.replace(/\/$/,"")}function ie(t,e){var n,i=Math.max(t.length,e.length);for(n=0;n<i;n++)if(t[n]!==e[n])break;return{updated:e.slice(0,n),activated:e.slice(n),deactivated:t.slice(n)}}function re(t,e,n,i){var r=Zt(t,(function(t,i,r,o){var a=oe(t,e);if(a)return Array.isArray(a)?a.map((function(t){return n(t,i,r,o)})):n(a,i,r,o)}));return Jt(i?r.reverse():r)}function oe(t,e){return"function"!==typeof t&&(t=et.extend(t)),t.options[e]}function ae(t){return re(t,"beforeRouteLeave",le,!0)}function se(t){return re(t,"beforeRouteUpdate",le)}function le(t,e){if(e)return function(){return t.apply(e,arguments)}}function ue(t){return re(t,"beforeRouteEnter",(function(t,e,n,i){return ce(t,n,i)}))}function ce(t,e,n){return function(i,r,o){return t(i,r,(function(t){"function"===typeof t&&(e.enteredCbs[n]||(e.enteredCbs[n]=[]),e.enteredCbs[n].push(t)),o(t)}))}}ee.prototype.listen=function(t){this.cb=t},ee.prototype.onReady=function(t,e){this.ready?t():(this.readyCbs.push(t),e&&this.readyErrorCbs.push(e))},ee.prototype.onError=function(t){this.errorCbs.push(t)},ee.prototype.transitionTo=function(t,e,n){var i,r=this;try{i=this.router.match(t,this.current)}catch(a){throw this.errorCbs.forEach((function(t){t(a)})),a}var o=this.current;this.confirmTransition(i,(function(){r.updateRoute(i),e&&e(i),r.ensureURL(),r.router.afterHooks.forEach((function(t){t&&t(i,o)})),r.ready||(r.ready=!0,r.readyCbs.forEach((function(t){t(i)})))}),(function(t){n&&n(t),t&&!r.ready&&(qt(t,jt.redirected)&&o===y||(r.ready=!0,r.readyErrorCbs.forEach((function(e){e(t)}))))}))},ee.prototype.confirmTransition=function(t,e,n){var r=this,o=this.current;this.pending=t;var a=function(t){!qt(t)&&Ut(t)&&(r.errorCbs.length?r.errorCbs.forEach((function(e){e(t)})):i(!1,"uncaught error during route navigation:")),n&&n(t)},s=t.matched.length-1,l=o.matched.length-1;if(b(t,o)&&s===l&&t.matched[s]===o.matched[l])return this.ensureURL(),a(Wt(o,t));var u=ie(this.current.matched,t.matched),c=u.updated,d=u.deactivated,h=u.activated,f=[].concat(ae(d),this.router.beforeHooks,se(c),h.map((function(t){return t.beforeEnter})),Xt(h)),p=function(e,n){if(r.pending!==t)return a(Bt(o,t));try{e(t,o,(function(e){!1===e?(r.ensureURL(!0),a(Ft(o,t))):Ut(e)?(r.ensureURL(!0),a(e)):"string"===typeof e||"object"===typeof e&&("string"===typeof e.path||"string"===typeof e.name)?(a(zt(o,t)),"object"===typeof e&&e.replace?r.replace(e):r.push(e)):n(e)}))}catch(i){a(i)}};Ht(f,p,(function(){var n=ue(h),i=n.concat(r.router.resolveHooks);Ht(i,p,(function(){if(r.pending!==t)return a(Bt(o,t));r.pending=null,e(t),r.router.app&&r.router.app.$nextTick((function(){S(t)}))}))}))},ee.prototype.updateRoute=function(t){this.current=t,this.cb&&this.cb(t)},ee.prototype.setupListeners=function(){},ee.prototype.teardown=function(){this.listeners.forEach((function(t){t()})),this.listeners=[],this.current=y,this.pending=null};var de=function(t){function e(e,n){t.call(this,e,n),this._startLocation=he(this.base)}return t&&(e.__proto__=t),e.prototype=Object.create(t&&t.prototype),e.prototype.constructor=e,e.prototype.setupListeners=function(){var t=this;if(!(this.listeners.length>0)){var e=this.router,n=e.options.scrollBehavior,i=Et&&n;i&&this.listeners.push(Mt());var r=function(){var n=t.current,r=he(t.base);t.current===y&&r===t._startLocation||t.transitionTo(r,(function(t){i&&St(e,t,n,!0)}))};window.addEventListener("popstate",r),this.listeners.push((function(){window.removeEventListener("popstate",r)}))}},e.prototype.go=function(t){window.history.go(t)},e.prototype.push=function(t,e,n){var i=this,r=this,o=r.current;this.transitionTo(t,(function(t){Nt(I(i.base+t.fullPath)),St(i.router,t,o,!1),e&&e(t)}),n)},e.prototype.replace=function(t,e,n){var i=this,r=this,o=r.current;this.transitionTo(t,(function(t){Rt(I(i.base+t.fullPath)),St(i.router,t,o,!1),e&&e(t)}),n)},e.prototype.ensureURL=function(t){if(he(this.base)!==this.current.fullPath){var e=I(this.base+this.current.fullPath);t?Nt(e):Rt(e)}},e.prototype.getCurrentLocation=function(){return he(this.base)},e}(ee);function he(t){var e=window.location.pathname;return t&&0===e.toLowerCase().indexOf(t.toLowerCase())&&(e=e.slice(t.length)),(e||"/")+window.location.search+window.location.hash}var fe=function(t){function e(e,n,i){t.call(this,e,n),i&&pe(this.base)||me()}return t&&(e.__proto__=t),e.prototype=Object.create(t&&t.prototype),e.prototype.constructor=e,e.prototype.setupListeners=function(){var t=this;if(!(this.listeners.length>0)){var e=this.router,n=e.options.scrollBehavior,i=Et&&n;i&&this.listeners.push(Mt());var r=function(){var e=t.current;me()&&t.transitionTo(ge(),(function(n){i&&St(t.router,n,e,!0),Et||_e(n.fullPath)}))},o=Et?"popstate":"hashchange";window.addEventListener(o,r),this.listeners.push((function(){window.removeEventListener(o,r)}))}},e.prototype.push=function(t,e,n){var i=this,r=this,o=r.current;this.transitionTo(t,(function(t){ve(t.fullPath),St(i.router,t,o,!1),e&&e(t)}),n)},e.prototype.replace=function(t,e,n){var i=this,r=this,o=r.current;this.transitionTo(t,(function(t){_e(t.fullPath),St(i.router,t,o,!1),e&&e(t)}),n)},e.prototype.go=function(t){window.history.go(t)},e.prototype.ensureURL=function(t){var e=this.current.fullPath;ge()!==e&&(t?ve(e):_e(e))},e.prototype.getCurrentLocation=function(){return ge()},e}(ee);function pe(t){var e=he(t);if(!/^\/#/.test(e))return window.location.replace(I(t+"/#"+e)),!0}function me(){var t=ge();return"/"===t.charAt(0)||(_e("/"+t),!1)}function ge(){var t=window.location.href,e=t.indexOf("#");return e<0?"":(t=t.slice(e+1),t)}function ye(t){var e=window.location.href,n=e.indexOf("#"),i=n>=0?e.slice(0,n):e;return i+"#"+t}function ve(t){Et?Nt(ye(t)):window.location.hash=t}function _e(t){Et?Rt(ye(t)):window.location.replace(ye(t))}var be=function(t){function e(e,n){t.call(this,e,n),this.stack=[],this.index=-1}return t&&(e.__proto__=t),e.prototype=Object.create(t&&t.prototype),e.prototype.constructor=e,e.prototype.push=function(t,e,n){var i=this;this.transitionTo(t,(function(t){i.stack=i.stack.slice(0,i.index+1).concat(t),i.index++,e&&e(t)}),n)},e.prototype.replace=function(t,e,n){var i=this;this.transitionTo(t,(function(t){i.stack=i.stack.slice(0,i.index).concat(t),e&&e(t)}),n)},e.prototype.go=function(t){var e=this,n=this.index+t;if(!(n<0||n>=this.stack.length)){var i=this.stack[n];this.confirmTransition(i,(function(){var t=e.current;e.index=n,e.updateRoute(i),e.router.afterHooks.forEach((function(e){e&&e(i,t)}))}),(function(t){qt(t,jt.duplicated)&&(e.index=n)}))}},e.prototype.getCurrentLocation=function(){var t=this.stack[this.stack.length-1];return t?t.fullPath:"/"},e.prototype.ensureURL=function(){},e}(ee),xe=function(t){void 0===t&&(t={}),this.app=null,this.apps=[],this.options=t,this.beforeHooks=[],this.resolveHooks=[],this.afterHooks=[],this.matcher=pt(t.routes||[],this);var e=t.mode||"hash";switch(this.fallback="history"===e&&!Et&&!1!==t.fallback,this.fallback&&(e="hash"),ut||(e="abstract"),this.mode=e,e){case"history":this.history=new de(this,t.base);break;case"hash":this.history=new fe(this,t.base,this.fallback);break;case"abstract":this.history=new be(this,t.base);break;default:0}},we={currentRoute:{configurable:!0}};function Me(t,e){return t.push(e),function(){var n=t.indexOf(e);n>-1&&t.splice(n,1)}}function Se(t,e,n){var i="hash"===n?"#"+e:e;return t?I(t+"/"+i):i}xe.prototype.match=function(t,e,n){return this.matcher.match(t,e,n)},we.currentRoute.get=function(){return this.history&&this.history.current},xe.prototype.init=function(t){var e=this;if(this.apps.push(t),t.$once("hook:destroyed",(function(){var n=e.apps.indexOf(t);n>-1&&e.apps.splice(n,1),e.app===t&&(e.app=e.apps[0]||null),e.app||e.history.teardown()})),!this.app){this.app=t;var n=this.history;if(n instanceof de||n instanceof fe){var i=function(t){var i=n.current,r=e.options.scrollBehavior,o=Et&&r;o&&"fullPath"in t&&St(e,t,i,!1)},r=function(t){n.setupListeners(),i(t)};n.transitionTo(n.getCurrentLocation(),r,r)}n.listen((function(t){e.apps.forEach((function(e){e._route=t}))}))}},xe.prototype.beforeEach=function(t){return Me(this.beforeHooks,t)},xe.prototype.beforeResolve=function(t){return Me(this.resolveHooks,t)},xe.prototype.afterEach=function(t){return Me(this.afterHooks,t)},xe.prototype.onReady=function(t,e){this.history.onReady(t,e)},xe.prototype.onError=function(t){this.history.onError(t)},xe.prototype.push=function(t,e,n){var i=this;if(!e&&!n&&"undefined"!==typeof Promise)return new Promise((function(e,n){i.history.push(t,e,n)}));this.history.push(t,e,n)},xe.prototype.replace=function(t,e,n){var i=this;if(!e&&!n&&"undefined"!==typeof Promise)return new Promise((function(e,n){i.history.replace(t,e,n)}));this.history.replace(t,e,n)},xe.prototype.go=function(t){this.history.go(t)},xe.prototype.back=function(){this.go(-1)},xe.prototype.forward=function(){this.go(1)},xe.prototype.getMatchedComponents=function(t){var e=t?t.matched?t:this.resolve(t).route:this.currentRoute;return e?[].concat.apply([],e.matched.map((function(t){return Object.keys(t.components).map((function(e){return t.components[e]}))}))):[]},xe.prototype.resolve=function(t,e,n){e=e||this.history.current;var i=tt(t,e,n,this),r=this.match(i,e),o=r.redirectedFrom||r.fullPath,a=this.history.base,s=Se(a,o,this.mode);return{location:i,route:r,href:s,normalizedTo:i,resolved:r}},xe.prototype.getRoutes=function(){return this.matcher.getRoutes()},xe.prototype.addRoute=function(t,e){this.matcher.addRoute(t,e),this.history.current!==y&&this.history.transitionTo(this.history.getCurrentLocation())},xe.prototype.addRoutes=function(t){this.matcher.addRoutes(t),this.history.current!==y&&this.history.transitionTo(this.history.getCurrentLocation())},Object.defineProperties(xe.prototype,we),xe.install=lt,xe.version="3.5.0",xe.isNavigationFailure=qt,xe.NavigationFailureType=jt,xe.START_LOCATION=y,ut&&window.Vue&&window.Vue.use(xe),e["a"]=xe},"8d47":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t){return"undefined"!==typeof Function&&t instanceof Function||"[object Function]"===Object.prototype.toString.call(t)}var n=t.defineLocale("el",{monthsNominativeEl:"Ιανουάριος_Φεβρουάριος_Μάρτιος_Απρίλιος_Μάιος_Ιούνιος_Ιούλιος_Αύγουστος_Σεπτέμβριος_Οκτώβριος_Νοέμβριος_Δεκέμβριος".split("_"),monthsGenitiveEl:"Ιανουαρίου_Φεβρουαρίου_Μαρτίου_Απριλίου_Μαΐου_Ιουνίου_Ιουλίου_Αυγούστου_Σεπτεμβρίου_Οκτωβρίου_Νοεμβρίου_Δεκεμβρίου".split("_"),months:function(t,e){return t?"string"===typeof e&&/D/.test(e.substring(0,e.indexOf("MMMM")))?this._monthsGenitiveEl[t.month()]:this._monthsNominativeEl[t.month()]:this._monthsNominativeEl},monthsShort:"Ιαν_Φεβ_Μαρ_Απρ_Μαϊ_Ιουν_Ιουλ_Αυγ_Σεπ_Οκτ_Νοε_Δεκ".split("_"),weekdays:"Κυριακή_Δευτέρα_Τρίτη_Τετάρτη_Πέμπτη_Παρασκευή_Σάββατο".split("_"),weekdaysShort:"Κυρ_Δευ_Τρι_Τετ_Πεμ_Παρ_Σαβ".split("_"),weekdaysMin:"Κυ_Δε_Τρ_Τε_Πε_Πα_Σα".split("_"),meridiem:function(t,e,n){return t>11?n?"μμ":"ΜΜ":n?"πμ":"ΠΜ"},isPM:function(t){return"μ"===(t+"").toLowerCase()[0]},meridiemParse:/[ΠΜ]\.?Μ?\.?/i,longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendarEl:{sameDay:"[Σήμερα {}] LT",nextDay:"[Αύριο {}] LT",nextWeek:"dddd [{}] LT",lastDay:"[Χθες {}] LT",lastWeek:function(){switch(this.day()){case 6:return"[το προηγούμενο] dddd [{}] LT";default:return"[την προηγούμενη] dddd [{}] LT"}},sameElse:"L"},calendar:function(t,n){var i=this._calendarEl[t],r=n&&n.hours();return e(i)&&(i=i.apply(n)),i.replace("{}",r%12===1?"στη":"στις")},relativeTime:{future:"σε %s",past:"%s πριν",s:"λίγα δευτερόλεπτα",ss:"%d δευτερόλεπτα",m:"ένα λεπτό",mm:"%d λεπτά",h:"μία ώρα",hh:"%d ώρες",d:"μία μέρα",dd:"%d μέρες",M:"ένας μήνας",MM:"%d μήνες",y:"ένας χρόνος",yy:"%d χρόνια"},dayOfMonthOrdinalParse:/\d{1,2}η/,ordinal:"%dη",week:{dow:1,doy:4}});return n}))},"8d57":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="styczeń_luty_marzec_kwiecień_maj_czerwiec_lipiec_sierpień_wrzesień_październik_listopad_grudzień".split("_"),n="stycznia_lutego_marca_kwietnia_maja_czerwca_lipca_sierpnia_września_października_listopada_grudnia".split("_"),i=[/^sty/i,/^lut/i,/^mar/i,/^kwi/i,/^maj/i,/^cze/i,/^lip/i,/^sie/i,/^wrz/i,/^paź/i,/^lis/i,/^gru/i];function r(t){return t%10<5&&t%10>1&&~~(t/10)%10!==1}function o(t,e,n){var i=t+" ";switch(n){case"ss":return i+(r(t)?"sekundy":"sekund");case"m":return e?"minuta":"minutę";case"mm":return i+(r(t)?"minuty":"minut");case"h":return e?"godzina":"godzinę";case"hh":return i+(r(t)?"godziny":"godzin");case"ww":return i+(r(t)?"tygodnie":"tygodni");case"MM":return i+(r(t)?"miesiące":"miesięcy");case"yy":return i+(r(t)?"lata":"lat")}}var a=t.defineLocale("pl",{months:function(t,i){return t?/D MMMM/.test(i)?n[t.month()]:e[t.month()]:e},monthsShort:"sty_lut_mar_kwi_maj_cze_lip_sie_wrz_paź_lis_gru".split("_"),monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"niedziela_poniedziałek_wtorek_środa_czwartek_piątek_sobota".split("_"),weekdaysShort:"ndz_pon_wt_śr_czw_pt_sob".split("_"),weekdaysMin:"Nd_Pn_Wt_Śr_Cz_Pt_So".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Dziś o] LT",nextDay:"[Jutro o] LT",nextWeek:function(){switch(this.day()){case 0:return"[W niedzielę o] LT";case 2:return"[We wtorek o] LT";case 3:return"[W środę o] LT";case 6:return"[W sobotę o] LT";default:return"[W] dddd [o] LT"}},lastDay:"[Wczoraj o] LT",lastWeek:function(){switch(this.day()){case 0:return"[W zeszłą niedzielę o] LT";case 3:return"[W zeszłą środę o] LT";case 6:return"[W zeszłą sobotę o] LT";default:return"[W zeszły] dddd [o] LT"}},sameElse:"L"},relativeTime:{future:"za %s",past:"%s temu",s:"kilka sekund",ss:o,m:o,mm:o,h:o,hh:o,d:"1 dzień",dd:"%d dni",w:"tydzień",ww:o,M:"miesiąc",MM:o,y:"rok",yy:o},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return a}))},"8de2":function(t,e,n){var i=n("8eeb"),r=n("9934");function o(t){return i(t,r(t))}t.exports=o},"8df4":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"۱",2:"۲",3:"۳",4:"۴",5:"۵",6:"۶",7:"۷",8:"۸",9:"۹",0:"۰"},n={"۱":"1","۲":"2","۳":"3","۴":"4","۵":"5","۶":"6","۷":"7","۸":"8","۹":"9","۰":"0"},i=t.defineLocale("fa",{months:"ژانویه_فوریه_مارس_آوریل_مه_ژوئن_ژوئیه_اوت_سپتامبر_اکتبر_نوامبر_دسامبر".split("_"),monthsShort:"ژانویه_فوریه_مارس_آوریل_مه_ژوئن_ژوئیه_اوت_سپتامبر_اکتبر_نوامبر_دسامبر".split("_"),weekdays:"یک‌شنبه_دوشنبه_سه‌شنبه_چهارشنبه_پنج‌شنبه_جمعه_شنبه".split("_"),weekdaysShort:"یک‌شنبه_دوشنبه_سه‌شنبه_چهارشنبه_پنج‌شنبه_جمعه_شنبه".split("_"),weekdaysMin:"ی_د_س_چ_پ_ج_ش".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},meridiemParse:/قبل از ظهر|بعد از ظهر/,isPM:function(t){return/بعد از ظهر/.test(t)},meridiem:function(t,e,n){return t<12?"قبل از ظهر":"بعد از ظهر"},calendar:{sameDay:"[امروز ساعت] LT",nextDay:"[فردا ساعت] LT",nextWeek:"dddd [ساعت] LT",lastDay:"[دیروز ساعت] LT",lastWeek:"dddd [پیش] [ساعت] LT",sameElse:"L"},relativeTime:{future:"در %s",past:"%s پیش",s:"چند ثانیه",ss:"%d ثانیه",m:"یک دقیقه",mm:"%d دقیقه",h:"یک ساعت",hh:"%d ساعت",d:"یک روز",dd:"%d روز",M:"یک ماه",MM:"%d ماه",y:"یک سال",yy:"%d سال"},preparse:function(t){return t.replace(/[۰-۹]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},dayOfMonthOrdinalParse:/\d{1,2}م/,ordinal:"%dم",week:{dow:6,doy:12}});return i}))},"8df4b":function(t,e,n){"use strict";var i=n("7a77");function r(t){if("function"!==typeof t)throw new TypeError("executor must be a function.");var e;this.promise=new Promise((function(t){e=t}));var n=this;t((function(t){n.reason||(n.reason=new i(t),e(n.reason))}))}r.prototype.throwIfRequested=function(){if(this.reason)throw this.reason},r.source=function(){var t,e=new r((function(e){t=e}));return{token:e,cancel:t}},t.exports=r},"8e73":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"١",2:"٢",3:"٣",4:"٤",5:"٥",6:"٦",7:"٧",8:"٨",9:"٩",0:"٠"},n={"١":"1","٢":"2","٣":"3","٤":"4","٥":"5","٦":"6","٧":"7","٨":"8","٩":"9","٠":"0"},i=function(t){return 0===t?0:1===t?1:2===t?2:t%100>=3&&t%100<=10?3:t%100>=11?4:5},r={s:["أقل من ثانية","ثانية واحدة",["ثانيتان","ثانيتين"],"%d ثوان","%d ثانية","%d ثانية"],m:["أقل من دقيقة","دقيقة واحدة",["دقيقتان","دقيقتين"],"%d دقائق","%d دقيقة","%d دقيقة"],h:["أقل من ساعة","ساعة واحدة",["ساعتان","ساعتين"],"%d ساعات","%d ساعة","%d ساعة"],d:["أقل من يوم","يوم واحد",["يومان","يومين"],"%d أيام","%d يومًا","%d يوم"],M:["أقل من شهر","شهر واحد",["شهران","شهرين"],"%d أشهر","%d شهرا","%d شهر"],y:["أقل من عام","عام واحد",["عامان","عامين"],"%d أعوام","%d عامًا","%d عام"]},o=function(t){return function(e,n,o,a){var s=i(e),l=r[t][i(e)];return 2===s&&(l=l[n?0:1]),l.replace(/%d/i,e)}},a=["يناير","فبراير","مارس","أبريل","مايو","يونيو","يوليو","أغسطس","سبتمبر","أكتوبر","نوفمبر","ديسمبر"],s=t.defineLocale("ar",{months:a,monthsShort:a,weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/‏M/‏YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم عند الساعة] LT",nextDay:"[غدًا عند الساعة] LT",nextWeek:"dddd [عند الساعة] LT",lastDay:"[أمس عند الساعة] LT",lastWeek:"dddd [عند الساعة] LT",sameElse:"L"},relativeTime:{future:"بعد %s",past:"منذ %s",s:o("s"),ss:o("s"),m:o("m"),mm:o("m"),h:o("h"),hh:o("h"),d:o("d"),dd:o("d"),M:o("M"),MM:o("M"),y:o("y"),yy:o("y")},preparse:function(t){return t.replace(/[١٢٣٤٥٦٧٨٩٠]/g,(function(t){return n[t]})).replace(/،/g,",")},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]})).replace(/,/g,"،")},week:{dow:6,doy:12}});return s}))},"8eeb":function(t,e,n){var i=n("32b3"),r=n("872a");function o(t,e,n,o){var a=!n;n||(n={});var s=-1,l=e.length;while(++s<l){var u=e[s],c=o?o(n[u],t[u],u,n,t):void 0;void 0===c&&(c=t[u]),a?r(n,u,c):i(n,u,c)}return n}t.exports=o},"900d":function(t,e,n){var i=n("8eeb"),r=n("2ec1"),o=n("ec69"),a=r((function(t,e,n,r){i(e,o(e),t,r)}));t.exports=a},9043:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"১",2:"২",3:"৩",4:"৪",5:"৫",6:"৬",7:"৭",8:"৮",9:"৯",0:"০"},n={"১":"1","২":"2","৩":"3","৪":"4","৫":"5","৬":"6","৭":"7","৮":"8","৯":"9","০":"0"},i=t.defineLocale("bn",{months:"জানুয়ারি_ফেব্রুয়ারি_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্টেম্বর_অক্টোবর_নভেম্বর_ডিসেম্বর".split("_"),monthsShort:"জানু_ফেব্রু_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্ট_অক্টো_নভে_ডিসে".split("_"),weekdays:"রবিবার_সোমবার_মঙ্গলবার_বুধবার_বৃহস্পতিবার_শুক্রবার_শনিবার".split("_"),weekdaysShort:"রবি_সোম_মঙ্গল_বুধ_বৃহস্পতি_শুক্র_শনি".split("_"),weekdaysMin:"রবি_সোম_মঙ্গল_বুধ_বৃহ_শুক্র_শনি".split("_"),longDateFormat:{LT:"A h:mm সময়",LTS:"A h:mm:ss সময়",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm সময়",LLLL:"dddd, D MMMM YYYY, A h:mm সময়"},calendar:{sameDay:"[আজ] LT",nextDay:"[আগামীকাল] LT",nextWeek:"dddd, LT",lastDay:"[গতকাল] LT",lastWeek:"[গত] dddd, LT",sameElse:"L"},relativeTime:{future:"%s পরে",past:"%s আগে",s:"কয়েক সেকেন্ড",ss:"%d সেকেন্ড",m:"এক মিনিট",mm:"%d মিনিট",h:"এক ঘন্টা",hh:"%d ঘন্টা",d:"এক দিন",dd:"%d দিন",M:"এক মাস",MM:"%d মাস",y:"এক বছর",yy:"%d বছর"},preparse:function(t){return t.replace(/[১২৩৪৫৬৭৮৯০]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/রাত|সকাল|দুপুর|বিকাল|রাত/,meridiemHour:function(t,e){return 12===t&&(t=0),"রাত"===e&&t>=4||"দুপুর"===e&&t<5||"বিকাল"===e?t+12:t},meridiem:function(t,e,n){return t<4?"রাত":t<10?"সকাল":t<17?"দুপুর":t<20?"বিকাল":"রাত"},week:{dow:0,doy:6}});return i}))},"90b2":function(t,e,n){"use strict";var i=n("31de");e["a"]=i["a"]},"90e3":function(t,e){var n=0,i=Math.random();t.exports=function(t){return"Symbol("+String(void 0===t?"":t)+")_"+(++n+i).toString(36)}},"90ea":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("zh-tw",{months:"一月_二月_三月_四月_五月_六月_七月_八月_九月_十月_十一月_十二月".split("_"),monthsShort:"1月_2月_3月_4月_5月_6月_7月_8月_9月_10月_11月_12月".split("_"),weekdays:"星期日_星期一_星期二_星期三_星期四_星期五_星期六".split("_"),weekdaysShort:"週日_週一_週二_週三_週四_週五_週六".split("_"),weekdaysMin:"日_一_二_三_四_五_六".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY/MM/DD",LL:"YYYY年M月D日",LLL:"YYYY年M月D日 HH:mm",LLLL:"YYYY年M月D日dddd HH:mm",l:"YYYY/M/D",ll:"YYYY年M月D日",lll:"YYYY年M月D日 HH:mm",llll:"YYYY年M月D日dddd HH:mm"},meridiemParse:/凌晨|早上|上午|中午|下午|晚上/,meridiemHour:function(t,e){return 12===t&&(t=0),"凌晨"===e||"早上"===e||"上午"===e?t:"中午"===e?t>=11?t:t+12:"下午"===e||"晚上"===e?t+12:void 0},meridiem:function(t,e,n){var i=100*t+e;return i<600?"凌晨":i<900?"早上":i<1130?"上午":i<1230?"中午":i<1800?"下午":"晚上"},calendar:{sameDay:"[今天] LT",nextDay:"[明天] LT",nextWeek:"[下]dddd LT",lastDay:"[昨天] LT",lastWeek:"[上]dddd LT",sameElse:"L"},dayOfMonthOrdinalParse:/\d{1,2}(日|月|週)/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+"日";case"M":return t+"月";case"w":case"W":return t+"週";default:return t}},relativeTime:{future:"%s後",past:"%s前",s:"幾秒",ss:"%d 秒",m:"1 分鐘",mm:"%d 分鐘",h:"1 小時",hh:"%d 小時",d:"1 天",dd:"%d 天",M:"1 個月",MM:"%d 個月",y:"1 年",yy:"%d 年"}});return e}))},9112:function(t,e,n){var i=n("83ab"),r=n("9bf2"),o=n("5c6c");t.exports=i?function(t,e,n){return r.f(t,e,o(1,n))}:function(t,e,n){return t[e]=n,t}},"91e9":function(t,e){function n(t,e){return function(n){return t(e(n))}}t.exports=n},9259:function(t,e,n){"use strict";var i=n("2b0e"),r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{directives:[{name:"transfer-dom",rawName:"v-transfer-dom"}],attrs:{"data-transfer":t.transfer}},[n("transition",{attrs:{name:t.transitionNames[1]}},[t.showMask?n("div",{directives:[{name:"show",rawName:"v-show",value:t.visible,expression:"visible"}],class:t.maskClasses,style:t.wrapStyles,on:{click:t.handleMask}}):t._e()]),n("div",{class:t.wrapClasses,style:t.wrapStyles,on:{click:t.handleWrapClick}},[n("transition",{attrs:{name:t.transitionNames[0]},on:{"after-leave":t.animationFinish}},[n("div",{directives:[{name:"show",rawName:"v-show",value:t.visible,expression:"visible"}],class:t.classes,style:t.mainStyles,on:{mousedown:t.handleMousedown}},[n("div",{ref:"content",class:t.contentClasses,style:t.contentStyles,on:{click:t.handleClickModal}},[t.closable?n("a",{class:[t.prefixCls+"-close"],on:{click:t.close}},[t._t("close",[n("Icon",{attrs:{type:"ios-close"}})])],2):t._e(),t.showHead?n("div",{class:[t.prefixCls+"-header"],on:{mousedown:t.handleMoveStart}},[t._t("header",[n("div",{class:[t.prefixCls+"-header-inner"]},[t._v(t._s(t.title))])])],2):t._e(),n("div",{class:[t.prefixCls+"-body"]},[t._t("default")],2),t.footerHide?t._e():n("div",{class:[t.prefixCls+"-footer"]},[t._t("footer",[n("i-button",{attrs:{type:"text"},nativeOn:{click:function(e){return t.cancel(e)}}},[t._v(t._s(t.localeCancelText))]),n("i-button",{attrs:{type:"primary",loading:t.buttonLoading},nativeOn:{click:function(e){return t.ok(e)}}},[t._v(t._s(t.localeOkText))])])],2)])])])],1)],1)},o=[],a=(n("4de4"),n("c975"),n("a9e3"),n("ade3")),s=n("5250"),l=n("7a12"),u=n("c780");const c=i["a"].prototype.$isServer;var d=function(t){c||"undefined"!==typeof window.iview&&("langs"in iview||(iview.langs={}),iview.langs[t.i.locale]=t)};const h={i:{locale:"zh-CN",select:{placeholder:"请选择",noMatch:"无匹配数据",loading:"加载中"},table:{noDataText:"暂无数据",noFilteredDataText:"暂无筛选结果",confirmFilter:"筛选",resetFilter:"重置",clearFilter:"全部",sumText:"合计"},datepicker:{selectDate:"选择日期",selectTime:"选择时间",startTime:"开始时间",endTime:"结束时间",clear:"清空",ok:"确定",datePanelLabel:"[yyyy年] [m月]",month:"月",month1:"1 月",month2:"2 月",month3:"3 月",month4:"4 月",month5:"5 月",month6:"6 月",month7:"7 月",month8:"8 月",month9:"9 月",month10:"10 月",month11:"11 月",month12:"12 月",year:"年",weekStartDay:"0",weeks:{sun:"日",mon:"一",tue:"二",wed:"三",thu:"四",fri:"五",sat:"六"},months:{m1:"1月",m2:"2月",m3:"3月",m4:"4月",m5:"5月",m6:"6月",m7:"7月",m8:"8月",m9:"9月",m10:"10月",m11:"11月",m12:"12月"}},transfer:{titles:{source:"源列表",target:"目的列表"},filterPlaceholder:"请输入搜索内容",notFoundText:"列表为空"},modal:{okText:"确定",cancelText:"取消"},poptip:{okText:"确定",cancelText:"取消"},page:{prev:"上一页",next:"下一页",total:"共",item:"条",items:"条",prev5:"向前 5 页",next5:"向后 5 页",page:"条/页",goto:"跳至",p:"页"},rate:{star:"星",stars:"星"},time:{before:"前",after:"后",just:"刚刚",seconds:"秒",minutes:"分钟",hours:"小时",days:"天"},tree:{emptyText:"暂无数据"}}};d(h);var f=h,p=function(t){return m(t)&&!g(t)};function m(t){return!!t&&"object"===typeof t}function g(t){var e=Object.prototype.toString.call(t);return"[object RegExp]"===e||"[object Date]"===e||_(t)}var y="function"===typeof Symbol&&Symbol.for,v=y?Symbol.for("react.element"):60103;function _(t){return t.$$typeof===v}function b(t){return Array.isArray(t)?[]:{}}function x(t,e){return!1!==e.clone&&e.isMergeableObject(t)?S(b(t),t,e):t}function w(t,e,n){return t.concat(e).map((function(t){return x(t,n)}))}function M(t,e,n){var i={};return n.isMergeableObject(t)&&Object.keys(t).forEach((function(e){i[e]=x(t[e],n)})),Object.keys(e).forEach((function(r){n.isMergeableObject(e[r])&&t[r]?i[r]=S(t[r],e[r],n):i[r]=x(e[r],n)})),i}function S(t,e,n){n=n||{},n.arrayMerge=n.arrayMerge||w,n.isMergeableObject=n.isMergeableObject||p;var i=Array.isArray(e),r=Array.isArray(t),o=i===r;return o?i?n.arrayMerge(t,e,n):M(t,e,n):x(e,n)}S.all=function(t,e){if(!Array.isArray(t))throw new Error("first argument should be an array");return t.reduce((function(t,n){return S(t,n,e)}),{})};var L=S,k=L;const T=/(%|)\{([0-9a-zA-Z_]+)\}/g;var D=function(){function t(t,e){return Object.prototype.hasOwnProperty.call(t,e)}function e(e,...n){return 1===n.length&&"object"===typeof n[0]&&(n=n[0]),n&&n.hasOwnProperty||(n={}),e.replace(T,(i,r,o,a)=>{let s;return"{"===e[a-1]&&"}"===e[a+i.length]?o:(s=t(n,o)?n[o]:null,null===s||void 0===s?"":s)})}return e};const C=D(i["a"]);let I=f,A=!1,Y=function(){const t=Object.getPrototypeOf(this||i["a"]).$t;if("function"===typeof t&&i["a"].locale)return A||(A=!0,i["a"].locale(i["a"].config.lang,k(I,i["a"].locale(i["a"].config.lang)||{},{clone:!0}))),t.apply(this,arguments)};const O=function(t,e){let n=Y.apply(this,arguments);if(null!==n&&void 0!==n)return n;const i=t.split(".");let r=I;for(let o=0,a=i.length;o<a;o++){const t=i[o];if(n=r[t],o===a-1)return C(n,e);if(!n)return"";r=n}return""};var P={methods:{t(...t){return O.apply(this,t)}}},E=n("2507"),N=n("b2e6"),R=n("de48"),H=n("3a27"),j=n("c177"),z="ivu-modal",W={name:"Modal",mixins:[P,E["a"],N["a"]],components:{Icon:s["a"],iButton:l["a"]},directives:{TransferDom:u["a"]},props:{value:{type:Boolean,default:!1},closable:{type:Boolean,default:!0},maskClosable:{type:Boolean,default:function(){return!this.$IVIEW||""===this.$IVIEW.modal.maskClosable||this.$IVIEW.modal.maskClosable}},title:{type:String},width:{type:[Number,String],default:520},okText:{type:String},cancelText:{type:String},loading:{type:Boolean,default:!1},styles:{type:Object,default:function(){return{}}},className:{type:String},footerHide:{type:Boolean,default:!1},scrollable:{type:Boolean,default:!1},transitionNames:{type:Array,default:function(){return["ease","fade"]}},transfer:{type:Boolean,default:function(){return!this.$IVIEW||""===this.$IVIEW.transfer||this.$IVIEW.transfer}},fullscreen:{type:Boolean,default:!1},mask:{type:Boolean,default:!0},draggable:{type:Boolean,default:!1},zIndex:{type:Number,default:1e3}},data:function(){return{prefixCls:z,wrapShow:!1,showHead:!0,buttonLoading:!1,visible:this.value,dragData:{x:null,y:null,dragX:null,dragY:null,dragging:!1},modalIndex:this.handleGetModalIndex(),isMouseTriggerIn:!1}},computed:{wrapClasses:function(){var t;return["".concat(z,"-wrap"),(t={},Object(a["a"])(t,"".concat(z,"-hidden"),!this.wrapShow),Object(a["a"])(t,"".concat(this.className),!!this.className),Object(a["a"])(t,"".concat(z,"-no-mask"),!this.showMask),t)]},wrapStyles:function(){return{zIndex:this.modalIndex+this.zIndex}},maskClasses:function(){return"".concat(z,"-mask")},classes:function(){var t;return["".concat(z),(t={},Object(a["a"])(t,"".concat(z,"-fullscreen"),this.fullscreen),Object(a["a"])(t,"".concat(z,"-fullscreen-no-header"),this.fullscreen&&!this.showHead),Object(a["a"])(t,"".concat(z,"-fullscreen-no-footer"),this.fullscreen&&this.footerHide),t)]},contentClasses:function(){var t;return["".concat(z,"-content"),(t={},Object(a["a"])(t,"".concat(z,"-content-no-mask"),!this.showMask),Object(a["a"])(t,"".concat(z,"-content-drag"),this.draggable),Object(a["a"])(t,"".concat(z,"-content-dragging"),this.draggable&&this.dragData.dragging),t)]},mainStyles:function(){var t={},e=parseInt(this.width),n=null!==this.dragData.x?{top:0}:{width:"".concat(e,e<=100?"%":"px")},i=this.styles?this.styles:{};return Object.assign(t,n,i),t},contentStyles:function(){var t={};if(this.draggable){var e=this.styles.top?parseFloat(this.styles.top):0,n=this.styles.left?parseFloat(this.styles.left):0;null!==this.dragData.x&&(t.left="".concat(this.dragData.x-n,"px")),null!==this.dragData.y&&(t.top="".concat(this.dragData.y,"px")),null!==this.dragData.y&&(t.top="".concat(this.dragData.y-e,"px"));var i=parseInt(this.width),r={width:"".concat(i,i<=100?"%":"px")};Object.assign(t,r)}return t},localeOkText:function(){return void 0===this.okText?this.t("i.modal.okText"):this.okText},localeCancelText:function(){return void 0===this.cancelText?this.t("i.modal.cancelText"):this.cancelText},showMask:function(){return!this.draggable&&this.mask}},methods:{close:function(){this.visible=!1,this.$emit("input",!1),this.$emit("on-cancel")},handleMask:function(){this.maskClosable&&this.showMask&&this.close()},handleWrapClick:function(t){if(this.isMouseTriggerIn)this.isMouseTriggerIn=!1;else{var e=t.target.getAttribute("class");e&&e.indexOf("".concat(z,"-wrap"))>-1&&this.handleMask()}},handleMousedown:function(){this.isMouseTriggerIn=!0},cancel:function(){this.close()},ok:function(){this.loading?this.buttonLoading=!0:(this.visible=!1,this.$emit("input",!1)),this.$emit("on-ok")},EscClose:function(t){if(this.visible&&this.closable&&27===t.keyCode){var e=Object(H["f"])(this.$root,"Modal").filter((function(t){return t.$data.visible&&t.$props.closable})),n=e.sort((function(t,e){return t.$data.modalIndex<e.$data.modalIndex?1:-1}))[0];setTimeout((function(){n.close()}),0)}},animationFinish:function(){this.$emit("on-hidden")},handleMoveStart:function(t){if(!this.draggable)return!1;var e=this.$refs.content,n=e.getBoundingClientRect();this.dragData.x=n.x||n.left,this.dragData.y=n.y||n.top;var i={x:t.clientX,y:t.clientY};this.dragData.dragX=i.x,this.dragData.dragY=i.y,this.dragData.dragging=!0,Object(R["b"])(window,"mousemove",this.handleMoveMove),Object(R["b"])(window,"mouseup",this.handleMoveEnd)},handleMoveMove:function(t){if(!this.dragData.dragging)return!1;var e={x:t.clientX,y:t.clientY},n={x:e.x-this.dragData.dragX,y:e.y-this.dragData.dragY};this.dragData.x+=n.x,this.dragData.y+=n.y,this.dragData.dragX=e.x,this.dragData.dragY=e.y},handleMoveEnd:function(){this.dragData.dragging=!1,Object(R["a"])(window,"mousemove",this.handleMoveMove),Object(R["a"])(window,"mouseup",this.handleMoveEnd)},handleGetModalIndex:function(){return Object(j["a"])(),j["b"]},handleClickModal:function(){this.draggable&&(this.modalIndex=this.handleGetModalIndex())}},mounted:function(){this.visible&&(this.wrapShow=!0);var t=!0;void 0!==this.$slots.header||this.title||(t=!1),this.showHead=t,document.addEventListener("keydown",this.EscClose)},beforeDestroy:function(){document.removeEventListener("keydown",this.EscClose),this.removeScrollEffect()},watch:{value:function(t){this.visible=t},visible:function(t){var e=this;!1===t?(this.buttonLoading=!1,this.timer=setTimeout((function(){e.wrapShow=!1,e.removeScrollEffect()}),300)):(this.modalIndex=this.handleGetModalIndex(),this.timer&&clearTimeout(this.timer),this.wrapShow=!0,this.scrollable||this.addScrollEffect()),this.broadcast("Table","on-visible-change",t),this.broadcast("Slider","on-visible-change",t),this.$emit("on-visible-change",t)},loading:function(t){t||(this.buttonLoading=!1)},scrollable:function(t){t?this.removeScrollEffect():this.addScrollEffect()},title:function(t){void 0===this.$slots.header&&(this.showHead=!!t)}}},B=W,F=n("2877"),V=Object(F["a"])(B,r,o,!1,null,null,null),G=V.exports;const $="ivu-modal-confirm";G.newInstance=t=>{const e=t||{},n=new i["a"]({mixins:[P],data:Object.assign({},e,{visible:!1,width:416,title:"",body:"",iconType:"",iconName:"",okText:void 0,cancelText:void 0,showCancel:!1,loading:!1,buttonLoading:!1,scrollable:!1,closable:!1,closing:!1}),render(t){let n,i,r=[];return this.showCancel&&r.push(t(l["a"],{props:{type:"text"},on:{click:this.cancel}},this.localeCancelText)),r.push(t(l["a"],{props:{type:"primary",loading:this.buttonLoading},on:{click:this.ok}},this.localeOkText)),n=this.render?t("div",{attrs:{class:`${$}-body ${$}-body-render`}},[this.render(t)]):t("div",{attrs:{class:$+"-body"}},[t("div",{domProps:{innerHTML:this.body}})]),this.title&&(i=t("div",{attrs:{class:$+"-head"}},[t("div",{class:this.iconTypeCls},[t("i",{class:this.iconNameCls})]),t("div",{attrs:{class:$+"-head-title"},domProps:{innerHTML:this.title}})])),t(G,{props:Object.assign({},e,{width:this.width,scrollable:this.scrollable,closable:this.closable}),domProps:{value:this.visible},on:{input:t=>{this.visible=t},"on-cancel":this.cancel}},[t("div",{attrs:{class:$}},[i,n,t("div",{attrs:{class:$+"-footer"}},r)])])},computed:{iconTypeCls(){return[$+"-head-icon",`${$}-head-icon-${this.iconType}`]},iconNameCls(){return["ivu-icon","ivu-icon-"+this.iconName]},localeOkText(){return this.okText?this.okText:this.t("i.modal.okText")},localeCancelText(){return this.cancelText?this.cancelText:this.t("i.modal.cancelText")}},methods:{cancel(){this.closing||(this.$children[0].visible=!1,this.buttonLoading=!1,this.onCancel(),this.remove())},ok(){this.closing||(this.loading?this.buttonLoading=!0:(this.$children[0].visible=!1,this.remove()),this.onOk())},remove(){this.closing=!0,setTimeout(()=>{this.closing=!1,this.destroy()},300)},destroy(){this.$destroy(),this.$el&&document.body.removeChild(this.$el),this.onRemove()},onOk(){},onCancel(){},onRemove(){}}}),r=n.$mount();document.body.appendChild(r.$el);const o=n.$children[0];return{show(t){switch(o.$parent.showCancel=t.showCancel,o.$parent.iconType=t.icon,t.icon){case"info":o.$parent.iconName="ios-information-circle";break;case"success":o.$parent.iconName="ios-checkmark-circle";break;case"warning":o.$parent.iconName="ios-alert";break;case"error":o.$parent.iconName="ios-close-circle";break;case"confirm":o.$parent.iconName="ios-help-circle";break}"width"in t&&(o.$parent.width=t.width),"closable"in t&&(o.$parent.closable=t.closable),"title"in t&&(o.$parent.title=t.title),"content"in t&&(o.$parent.body=t.content),"okText"in t&&(o.$parent.okText=t.okText),"cancelText"in t&&(o.$parent.cancelText=t.cancelText),"onCancel"in t&&(o.$parent.onCancel=t.onCancel),"onOk"in t&&(o.$parent.onOk=t.onOk),"loading"in t&&(o.$parent.loading=t.loading),"scrollable"in t&&(o.$parent.scrollable=t.scrollable),o.$parent.onRemove=t.onRemove,o.visible=!0},remove(){o.visible=!1,o.$parent.buttonLoading=!1,o.$parent.remove()},component:o}};var U=G;let q;function X(t,e=!0){return q=q||U.newInstance({closable:!1,maskClosable:!1,footerHide:!0,render:t,lockScroll:e}),q}function Z(t){const e="render"in t?t.render:void 0,n=!("lockScroll"in t)||t.lockScroll;let i=X(e,n);t.onRemove=function(){q=null},i.show(t)}U.info=function(t={}){return t.icon="info",t.showCancel=!1,Z(t)},U.success=function(t={}){return t.icon="success",t.showCancel=!1,Z(t)},U.warning=function(t={}){return t.icon="warning",t.showCancel=!1,Z(t)},U.error=function(t={}){return t.icon="error",t.showCancel=!1,Z(t)},U.confirm=function(t={}){return t.icon="confirm",t.showCancel=!0,Z(t)},U.remove=function(){if(!q)return!1;const t=X();t.remove()};e["a"]=U},9263:function(t,e,n){"use strict";var i=n("ad6d"),r=n("9f7f"),o=RegExp.prototype.exec,a=String.prototype.replace,s=o,l=function(){var t=/a/,e=/b*/g;return o.call(t,"a"),o.call(e,"a"),0!==t.lastIndex||0!==e.lastIndex}(),u=r.UNSUPPORTED_Y||r.BROKEN_CARET,c=void 0!==/()??/.exec("")[1],d=l||c||u;d&&(s=function(t){var e,n,r,s,d=this,h=u&&d.sticky,f=i.call(d),p=d.source,m=0,g=t;return h&&(f=f.replace("y",""),-1===f.indexOf("g")&&(f+="g"),g=String(t).slice(d.lastIndex),d.lastIndex>0&&(!d.multiline||d.multiline&&"\n"!==t[d.lastIndex-1])&&(p="(?: "+p+")",g=" "+g,m++),n=new RegExp("^(?:"+p+")",f)),c&&(n=new RegExp("^"+p+"$(?!\\s)",f)),l&&(e=d.lastIndex),r=o.call(h?n:d,g),h?r?(r.input=r.input.slice(m),r[0]=r[0].slice(m),r.index=d.lastIndex,d.lastIndex+=r[0].length):d.lastIndex=0:l&&r&&(d.lastIndex=d.global?r.index+r[0].length:e),c&&r&&r.length>1&&a.call(r[0],n,(function(){for(s=1;s<arguments.length-2;s++)void 0===arguments[s]&&(r[s]=void 0)})),r}),t.exports=s},"93ed":function(t,e,n){var i=n("4245");function r(t){var e=i(this,t)["delete"](t);return this.size-=e?1:0,e}t.exports=r},"94ca":function(t,e,n){var i=n("d039"),r=/#|\.prototype\./,o=function(t,e){var n=s[a(t)];return n==u||n!=l&&("function"==typeof e?i(e):!!e)},a=o.normalize=function(t){return String(t).replace(r,".").toLowerCase()},s=o.data={},l=o.NATIVE="N",u=o.POLYFILL="P";t.exports=o},9520:function(t,e,n){var i=n("3729"),r=n("1a8c"),o="[object AsyncFunction]",a="[object Function]",s="[object GeneratorFunction]",l="[object Proxy]";function u(t){if(!r(t))return!1;var e=i(t);return e==a||e==s||e==o||e==l}t.exports=u},"957c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e){var n=t.split("_");return e%10===1&&e%100!==11?n[0]:e%10>=2&&e%10<=4&&(e%100<10||e%100>=20)?n[1]:n[2]}function n(t,n,i){var r={ss:n?"секунда_секунды_секунд":"секунду_секунды_секунд",mm:n?"минута_минуты_минут":"минуту_минуты_минут",hh:"час_часа_часов",dd:"день_дня_дней",ww:"неделя_недели_недель",MM:"месяц_месяца_месяцев",yy:"год_года_лет"};return"m"===i?n?"минута":"минуту":t+" "+e(r[i],+t)}var i=[/^янв/i,/^фев/i,/^мар/i,/^апр/i,/^ма[йя]/i,/^июн/i,/^июл/i,/^авг/i,/^сен/i,/^окт/i,/^ноя/i,/^дек/i],r=t.defineLocale("ru",{months:{format:"января_февраля_марта_апреля_мая_июня_июля_августа_сентября_октября_ноября_декабря".split("_"),standalone:"январь_февраль_март_апрель_май_июнь_июль_август_сентябрь_октябрь_ноябрь_декабрь".split("_")},monthsShort:{format:"янв._февр._мар._апр._мая_июня_июля_авг._сент._окт._нояб._дек.".split("_"),standalone:"янв._февр._март_апр._май_июнь_июль_авг._сент._окт._нояб._дек.".split("_")},weekdays:{standalone:"воскресенье_понедельник_вторник_среда_четверг_пятница_суббота".split("_"),format:"воскресенье_понедельник_вторник_среду_четверг_пятницу_субботу".split("_"),isFormat:/\[ ?[Вв] ?(?:прошлую|следующую|эту)? ?] ?dddd/},weekdaysShort:"вс_пн_вт_ср_чт_пт_сб".split("_"),weekdaysMin:"вс_пн_вт_ср_чт_пт_сб".split("_"),monthsParse:i,longMonthsParse:i,shortMonthsParse:i,monthsRegex:/^(январ[ья]|янв\.?|феврал[ья]|февр?\.?|марта?|мар\.?|апрел[ья]|апр\.?|ма[йя]|июн[ья]|июн\.?|июл[ья]|июл\.?|августа?|авг\.?|сентябр[ья]|сент?\.?|октябр[ья]|окт\.?|ноябр[ья]|нояб?\.?|декабр[ья]|дек\.?)/i,monthsShortRegex:/^(январ[ья]|янв\.?|феврал[ья]|февр?\.?|марта?|мар\.?|апрел[ья]|апр\.?|ма[йя]|июн[ья]|июн\.?|июл[ья]|июл\.?|августа?|авг\.?|сентябр[ья]|сент?\.?|октябр[ья]|окт\.?|ноябр[ья]|нояб?\.?|декабр[ья]|дек\.?)/i,monthsStrictRegex:/^(январ[яь]|феврал[яь]|марта?|апрел[яь]|ма[яй]|июн[яь]|июл[яь]|августа?|сентябр[яь]|октябр[яь]|ноябр[яь]|декабр[яь])/i,monthsShortStrictRegex:/^(янв\.|февр?\.|мар[т.]|апр\.|ма[яй]|июн[ья.]|июл[ья.]|авг\.|сент?\.|окт\.|нояб?\.|дек\.)/i,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY г.",LLL:"D MMMM YYYY г., H:mm",LLLL:"dddd, D MMMM YYYY г., H:mm"},calendar:{sameDay:"[Сегодня, в] LT",nextDay:"[Завтра, в] LT",lastDay:"[Вчера, в] LT",nextWeek:function(t){if(t.week()===this.week())return 2===this.day()?"[Во] dddd, [в] LT":"[В] dddd, [в] LT";switch(this.day()){case 0:return"[В следующее] dddd, [в] LT";case 1:case 2:case 4:return"[В следующий] dddd, [в] LT";case 3:case 5:case 6:return"[В следующую] dddd, [в] LT"}},lastWeek:function(t){if(t.week()===this.week())return 2===this.day()?"[Во] dddd, [в] LT":"[В] dddd, [в] LT";switch(this.day()){case 0:return"[В прошлое] dddd, [в] LT";case 1:case 2:case 4:return"[В прошлый] dddd, [в] LT";case 3:case 5:case 6:return"[В прошлую] dddd, [в] LT"}},sameElse:"L"},relativeTime:{future:"через %s",past:"%s назад",s:"несколько секунд",ss:n,m:n,mm:n,h:"час",hh:n,d:"день",dd:n,w:"неделя",ww:n,M:"месяц",MM:n,y:"год",yy:n},meridiemParse:/ночи|утра|дня|вечера/i,isPM:function(t){return/^(дня|вечера)$/.test(t)},meridiem:function(t,e,n){return t<4?"ночи":t<12?"утра":t<17?"дня":"вечера"},dayOfMonthOrdinalParse:/\d{1,2}-(й|го|я)/,ordinal:function(t,e){switch(e){case"M":case"d":case"DDD":return t+"-й";case"D":return t+"-го";case"w":case"W":return t+"-я";default:return t}},week:{dow:1,doy:4}});return r}))},"958b":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){switch(n){case"s":return e?"хэдхэн секунд":"хэдхэн секундын";case"ss":return t+(e?" секунд":" секундын");case"m":case"mm":return t+(e?" минут":" минутын");case"h":case"hh":return t+(e?" цаг":" цагийн");case"d":case"dd":return t+(e?" өдөр":" өдрийн");case"M":case"MM":return t+(e?" сар":" сарын");case"y":case"yy":return t+(e?" жил":" жилийн");default:return t}}var n=t.defineLocale("mn",{months:"Нэгдүгээр сар_Хоёрдугаар сар_Гуравдугаар сар_Дөрөвдүгээр сар_Тавдугаар сар_Зургадугаар сар_Долдугаар сар_Наймдугаар сар_Есдүгээр сар_Аравдугаар сар_Арван нэгдүгээр сар_Арван хоёрдугаар сар".split("_"),monthsShort:"1 сар_2 сар_3 сар_4 сар_5 сар_6 сар_7 сар_8 сар_9 сар_10 сар_11 сар_12 сар".split("_"),monthsParseExact:!0,weekdays:"Ням_Даваа_Мягмар_Лхагва_Пүрэв_Баасан_Бямба".split("_"),weekdaysShort:"Ням_Дав_Мяг_Лха_Пүр_Баа_Бям".split("_"),weekdaysMin:"Ня_Да_Мя_Лх_Пү_Ба_Бя".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"YYYY оны MMMMын D",LLL:"YYYY оны MMMMын D HH:mm",LLLL:"dddd, YYYY оны MMMMын D HH:mm"},meridiemParse:/ҮӨ|ҮХ/i,isPM:function(t){return"ҮХ"===t},meridiem:function(t,e,n){return t<12?"ҮӨ":"ҮХ"},calendar:{sameDay:"[Өнөөдөр] LT",nextDay:"[Маргааш] LT",nextWeek:"[Ирэх] dddd LT",lastDay:"[Өчигдөр] LT",lastWeek:"[Өнгөрсөн] dddd LT",sameElse:"L"},relativeTime:{future:"%s дараа",past:"%s өмнө",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2} өдөр/,ordinal:function(t,e){switch(e){case"d":case"D":case"DDD":return t+" өдөр";default:return t}}});return n}))},"95ae":function(t,e,n){var i=n("100e"),r=n("9638"),o=n("9aff"),a=n("9934"),s=Object.prototype,l=s.hasOwnProperty,u=i((function(t,e){t=Object(t);var n=-1,i=e.length,u=i>2?e[2]:void 0;u&&o(e[0],e[1],u)&&(i=1);while(++n<i){var c=e[n],d=a(c),h=-1,f=d.length;while(++h<f){var p=d[h],m=t[p];(void 0===m||r(m,s[p])&&!l.call(t,p))&&(t[p]=c[p])}}return t}));t.exports=u},9609:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={0:"-чү",1:"-чи",2:"-чи",3:"-чү",4:"-чү",5:"-чи",6:"-чы",7:"-чи",8:"-чи",9:"-чу",10:"-чу",20:"-чы",30:"-чу",40:"-чы",50:"-чү",60:"-чы",70:"-чи",80:"-чи",90:"-чу",100:"-чү"},n=t.defineLocale("ky",{months:"январь_февраль_март_апрель_май_июнь_июль_август_сентябрь_октябрь_ноябрь_декабрь".split("_"),monthsShort:"янв_фев_март_апр_май_июнь_июль_авг_сен_окт_ноя_дек".split("_"),weekdays:"Жекшемби_Дүйшөмбү_Шейшемби_Шаршемби_Бейшемби_Жума_Ишемби".split("_"),weekdaysShort:"Жек_Дүй_Шей_Шар_Бей_Жум_Ише".split("_"),weekdaysMin:"Жк_Дй_Шй_Шр_Бй_Жм_Иш".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Бүгүн саат] LT",nextDay:"[Эртең саат] LT",nextWeek:"dddd [саат] LT",lastDay:"[Кечээ саат] LT",lastWeek:"[Өткөн аптанын] dddd [күнү] [саат] LT",sameElse:"L"},relativeTime:{future:"%s ичинде",past:"%s мурун",s:"бирнече секунд",ss:"%d секунд",m:"бир мүнөт",mm:"%d мүнөт",h:"бир саат",hh:"%d саат",d:"бир күн",dd:"%d күн",M:"бир ай",MM:"%d ай",y:"бир жыл",yy:"%d жыл"},dayOfMonthOrdinalParse:/\d{1,2}-(чи|чы|чү|чу)/,ordinal:function(t){var n=t%10,i=t>=100?100:null;return t+(e[t]||e[n]||e[i])},week:{dow:1,doy:7}});return n}))},9638:function(t,e){function n(t,e){return t===e||t!==t&&e!==e}t.exports=n},"966f":function(t,e,n){var i=n("7e64"),r=n("c05f"),o=1,a=2;function s(t,e,n,s){var l=n.length,u=l,c=!s;if(null==t)return!u;t=Object(t);while(l--){var d=n[l];if(c&&d[2]?d[1]!==t[d[0]]:!(d[0]in t))return!1}while(++l<u){d=n[l];var h=d[0],f=t[h],p=d[1];if(c&&d[2]){if(void 0===f&&!(h in t))return!1}else{var m=new i;if(s)var g=s(f,p,h,t,e,m);if(!(void 0===g?r(p,f,o|a,s,m):g))return!1}}return!0}t.exports=s},9686:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"১",2:"২",3:"৩",4:"৪",5:"৫",6:"৬",7:"৭",8:"৮",9:"৯",0:"০"},n={"১":"1","২":"2","৩":"3","৪":"4","৫":"5","৬":"6","৭":"7","৮":"8","৯":"9","০":"0"},i=t.defineLocale("bn-bd",{months:"জানুয়ারি_ফেব্রুয়ারি_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্টেম্বর_অক্টোবর_নভেম্বর_ডিসেম্বর".split("_"),monthsShort:"জানু_ফেব্রু_মার্চ_এপ্রিল_মে_জুন_জুলাই_আগস্ট_সেপ্ট_অক্টো_নভে_ডিসে".split("_"),weekdays:"রবিবার_সোমবার_মঙ্গলবার_বুধবার_বৃহস্পতিবার_শুক্রবার_শনিবার".split("_"),weekdaysShort:"রবি_সোম_মঙ্গল_বুধ_বৃহস্পতি_শুক্র_শনি".split("_"),weekdaysMin:"রবি_সোম_মঙ্গল_বুধ_বৃহ_শুক্র_শনি".split("_"),longDateFormat:{LT:"A h:mm সময়",LTS:"A h:mm:ss সময়",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm সময়",LLLL:"dddd, D MMMM YYYY, A h:mm সময়"},calendar:{sameDay:"[আজ] LT",nextDay:"[আগামীকাল] LT",nextWeek:"dddd, LT",lastDay:"[গতকাল] LT",lastWeek:"[গত] dddd, LT",sameElse:"L"},relativeTime:{future:"%s পরে",past:"%s আগে",s:"কয়েক সেকেন্ড",ss:"%d সেকেন্ড",m:"এক মিনিট",mm:"%d মিনিট",h:"এক ঘন্টা",hh:"%d ঘন্টা",d:"এক দিন",dd:"%d দিন",M:"এক মাস",MM:"%d মাস",y:"এক বছর",yy:"%d বছর"},preparse:function(t){return t.replace(/[১২৩৪৫৬৭৮৯০]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/রাত|ভোর|সকাল|দুপুর|বিকাল|সন্ধ্যা|রাত/,meridiemHour:function(t,e){return 12===t&&(t=0),"রাত"===e?t<4?t:t+12:"ভোর"===e||"সকাল"===e?t:"দুপুর"===e?t>=3?t:t+12:"বিকাল"===e||"সন্ধ্যা"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"রাত":t<6?"ভোর":t<12?"সকাল":t<15?"দুপুর":t<18?"বিকাল":t<20?"সন্ধ্যা":"রাত"},week:{dow:0,doy:6}});return i}))},"96f3":function(t,e){var n=Object.prototype,i=n.hasOwnProperty;function r(t,e){return null!=t&&i.call(t,e)}t.exports=r},"972c":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n){var i={ss:"secunde",mm:"minute",hh:"ore",dd:"zile",ww:"săptămâni",MM:"luni",yy:"ani"},r=" ";return(t%100>=20||t>=100&&t%100===0)&&(r=" de "),t+r+i[n]}var n=t.defineLocale("ro",{months:"ianuarie_februarie_martie_aprilie_mai_iunie_iulie_august_septembrie_octombrie_noiembrie_decembrie".split("_"),monthsShort:"ian._feb._mart._apr._mai_iun._iul._aug._sept._oct._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"duminică_luni_marți_miercuri_joi_vineri_sâmbătă".split("_"),weekdaysShort:"Dum_Lun_Mar_Mie_Joi_Vin_Sâm".split("_"),weekdaysMin:"Du_Lu_Ma_Mi_Jo_Vi_Sâ".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY H:mm",LLLL:"dddd, D MMMM YYYY H:mm"},calendar:{sameDay:"[azi la] LT",nextDay:"[mâine la] LT",nextWeek:"dddd [la] LT",lastDay:"[ieri la] LT",lastWeek:"[fosta] dddd [la] LT",sameElse:"L"},relativeTime:{future:"peste %s",past:"%s în urmă",s:"câteva secunde",ss:e,m:"un minut",mm:e,h:"o oră",hh:e,d:"o zi",dd:e,w:"o săptămână",ww:e,M:"o lună",MM:e,y:"un an",yy:e},week:{dow:1,doy:7}});return n}))},"972ca":function(t,e){function n(t,e,n){var i;return n(t,(function(t,n,r){if(e(t,n,r))return i=n,!1})),i}t.exports=n},9797:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("cy",{months:"Ionawr_Chwefror_Mawrth_Ebrill_Mai_Mehefin_Gorffennaf_Awst_Medi_Hydref_Tachwedd_Rhagfyr".split("_"),monthsShort:"Ion_Chwe_Maw_Ebr_Mai_Meh_Gor_Aws_Med_Hyd_Tach_Rhag".split("_"),weekdays:"Dydd Sul_Dydd Llun_Dydd Mawrth_Dydd Mercher_Dydd Iau_Dydd Gwener_Dydd Sadwrn".split("_"),weekdaysShort:"Sul_Llun_Maw_Mer_Iau_Gwe_Sad".split("_"),weekdaysMin:"Su_Ll_Ma_Me_Ia_Gw_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Heddiw am] LT",nextDay:"[Yfory am] LT",nextWeek:"dddd [am] LT",lastDay:"[Ddoe am] LT",lastWeek:"dddd [diwethaf am] LT",sameElse:"L"},relativeTime:{future:"mewn %s",past:"%s yn ôl",s:"ychydig eiliadau",ss:"%d eiliad",m:"munud",mm:"%d munud",h:"awr",hh:"%d awr",d:"diwrnod",dd:"%d diwrnod",M:"mis",MM:"%d mis",y:"blwyddyn",yy:"%d flynedd"},dayOfMonthOrdinalParse:/\d{1,2}(fed|ain|af|il|ydd|ed|eg)/,ordinal:function(t){var e=t,n="",i=["","af","il","ydd","ydd","ed","ed","ed","fed","fed","fed","eg","fed","eg","eg","fed","eg","eg","fed","eg","fed"];return e>20?n=40===e||50===e||60===e||80===e||100===e?"fed":"ain":e>0&&(n=i[e]),t+n},week:{dow:1,doy:4}});return e}))},9934:function(t,e,n){var i=n("6fcd"),r=n("41c3"),o=n("30c9");function a(t){return o(t)?i(t,!0):r(t)}t.exports=a},9948:function(t,e,n){var i=n("72af"),r=n("1304"),o=n("9934");function a(t,e){return null==t?t:i(t,r(e),o)}t.exports=a},"99af":function(t,e,n){"use strict";var i=n("23e7"),r=n("d039"),o=n("e8b5"),a=n("861d"),s=n("7b0b"),l=n("50c4"),u=n("8418"),c=n("65f0"),d=n("1dde"),h=n("b622"),f=n("2d00"),p=h("isConcatSpreadable"),m=9007199254740991,g="Maximum allowed index exceeded",y=f>=51||!r((function(){var t=[];return t[p]=!1,t.concat()[0]!==t})),v=d("concat"),_=function(t){if(!a(t))return!1;var e=t[p];return void 0!==e?!!e:o(t)},b=!y||!v;i({target:"Array",proto:!0,forced:b},{concat:function(t){var e,n,i,r,o,a=s(this),d=c(a,0),h=0;for(e=-1,i=arguments.length;e<i;e++)if(o=-1===e?a:arguments[e],_(o)){if(r=l(o.length),h+r>m)throw TypeError(g);for(n=0;n<r;n++,h++)n in o&&u(d,h,o[n])}else{if(h>=m)throw TypeError(g);u(d,h++,o)}return d.length=h,d}})},"99cd":function(t,e){function n(t){return function(e,n,i){var r=-1,o=Object(e),a=i(e),s=a.length;while(s--){var l=a[t?s:++r];if(!1===n(o[l],l,o))break}return e}}t.exports=n},"99d3":function(t,e,n){(function(t){var i=n("585a"),r=e&&!e.nodeType&&e,o=r&&"object"==typeof t&&t&&!t.nodeType&&t,a=o&&o.exports===r,s=a&&i.process,l=function(){try{var t=o&&o.require&&o.require("util").types;return t||s&&s.binding&&s.binding("util")}catch(e){}}();t.exports=l}).call(this,n("62e4")(t))},"9aff":function(t,e,n){var i=n("9638"),r=n("30c9"),o=n("c098"),a=n("1a8c");function s(t,e,n){if(!a(n))return!1;var s=typeof e;return!!("number"==s?r(n)&&o(e,n.length):"string"==s&&e in n)&&i(n[e],t)}t.exports=s},"9b02":function(t,e,n){var i=n("656b");function r(t,e,n){var r=null==t?void 0:i(t,e);return void 0===r?n:r}t.exports=r},"9bdd":function(t,e,n){var i=n("825a"),r=n("2a62");t.exports=function(t,e,n,o){try{return o?e(i(n)[0],n[1]):e(n)}catch(a){throw r(t),a}}},"9bf2":function(t,e,n){var i=n("83ab"),r=n("0cfb"),o=n("825a"),a=n("c04e"),s=Object.defineProperty;e.f=i?s:function(t,e,n){if(o(t),e=a(e,!0),o(n),r)try{return s(t,e,n)}catch(i){}if("get"in n||"set"in n)throw TypeError("Accessors not supported");return"value"in n&&(t[e]=n.value),t}},"9e69":function(t,e,n){var i=n("2b3e"),r=i.Symbol;t.exports=r},"9e86":function(t,e,n){var i=n("872a"),r=n("242e"),o=n("badf");function a(t,e){var n={};return e=o(e,3),r(t,(function(t,r,o){i(n,r,e(t,r,o))})),n}t.exports=a},"9ed3":function(t,e,n){"use strict";var i=n("ae93").IteratorPrototype,r=n("7c73"),o=n("5c6c"),a=n("d44e"),s=n("3f8c"),l=function(){return this};t.exports=function(t,e,n){var u=e+" Iterator";return t.prototype=r(i,{next:o(1,n)}),a(t,u,!1,!0),s[u]=l,t}},"9f26":function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=/^(janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre)/i,n=/(janv\.?|févr\.?|mars|avr\.?|mai|juin|juil\.?|août|sept\.?|oct\.?|nov\.?|déc\.?)/i,i=/(janv\.?|févr\.?|mars|avr\.?|mai|juin|juil\.?|août|sept\.?|oct\.?|nov\.?|déc\.?|janvier|février|mars|avril|mai|juin|juillet|août|septembre|octobre|novembre|décembre)/i,r=[/^janv/i,/^févr/i,/^mars/i,/^avr/i,/^mai/i,/^juin/i,/^juil/i,/^août/i,/^sept/i,/^oct/i,/^nov/i,/^déc/i],o=t.defineLocale("fr",{months:"janvier_février_mars_avril_mai_juin_juillet_août_septembre_octobre_novembre_décembre".split("_"),monthsShort:"janv._févr._mars_avr._mai_juin_juil._août_sept._oct._nov._déc.".split("_"),monthsRegex:i,monthsShortRegex:i,monthsStrictRegex:e,monthsShortStrictRegex:n,monthsParse:r,longMonthsParse:r,shortMonthsParse:r,weekdays:"dimanche_lundi_mardi_mercredi_jeudi_vendredi_samedi".split("_"),weekdaysShort:"dim._lun._mar._mer._jeu._ven._sam.".split("_"),weekdaysMin:"di_lu_ma_me_je_ve_sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Aujourd’hui à] LT",nextDay:"[Demain à] LT",nextWeek:"dddd [à] LT",lastDay:"[Hier à] LT",lastWeek:"dddd [dernier à] LT",sameElse:"L"},relativeTime:{future:"dans %s",past:"il y a %s",s:"quelques secondes",ss:"%d secondes",m:"une minute",mm:"%d minutes",h:"une heure",hh:"%d heures",d:"un jour",dd:"%d jours",w:"une semaine",ww:"%d semaines",M:"un mois",MM:"%d mois",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(er|)/,ordinal:function(t,e){switch(e){case"D":return t+(1===t?"er":"");default:case"M":case"Q":case"DDD":case"d":return t+(1===t?"er":"e");case"w":case"W":return t+(1===t?"re":"e")}},week:{dow:1,doy:4}});return o}))},"9f7f":function(t,e,n){"use strict";var i=n("d039");function r(t,e){return RegExp(t,e)}e.UNSUPPORTED_Y=i((function(){var t=r("a","y");return t.lastIndex=2,null!=t.exec("abcd")})),e.BROKEN_CARET=i((function(){var t=r("^r","gy");return t.lastIndex=2,null!=t.exec("str")}))},"9f81":function(t,e,n){"use strict";e["a"]={inject:{FormInstance:{default:""}},computed:{itemDisabled(){let t=this.disabled;return!t&&this.FormInstance&&(t=this.FormInstance.disabled),t}}}},a029:function(t,e,n){var i=n("087d"),r=n("2dcb"),o=n("32f4"),a=n("d327"),s=Object.getOwnPropertySymbols,l=s?function(t){var e=[];while(t)i(e,o(t)),t=r(t);return e}:a;t.exports=l},a0ac:function(t,e){var n="Expected a function";function i(t){if("function"!=typeof t)throw new TypeError(n);return function(){var e=arguments;switch(e.length){case 0:return!t.call(this);case 1:return!t.call(this,e[0]);case 2:return!t.call(this,e[0],e[1]);case 3:return!t.call(this,e[0],e[1],e[2])}return!t.apply(this,e)}}t.exports=i},a2be:function(t,e,n){var i=n("d612"),r=n("4284"),o=n("c584"),a=1,s=2;function l(t,e,n,l,u,c){var d=n&a,h=t.length,f=e.length;if(h!=f&&!(d&&f>h))return!1;var p=c.get(t),m=c.get(e);if(p&&m)return p==e&&m==t;var g=-1,y=!0,v=n&s?new i:void 0;c.set(t,e),c.set(e,t);while(++g<h){var _=t[g],b=e[g];if(l)var x=d?l(b,_,g,e,t,c):l(_,b,g,t,e,c);if(void 0!==x){if(x)continue;y=!1;break}if(v){if(!r(e,(function(t,e){if(!o(v,e)&&(_===t||u(_,t,n,l,c)))return v.push(e)}))){y=!1;break}}else if(_!==b&&!u(_,b,n,l,c)){y=!1;break}}return c["delete"](t),c["delete"](e),y}t.exports=l},a2db:function(t,e,n){var i=n("9e69"),r=i?i.prototype:void 0,o=r?r.valueOf:void 0;function a(t){return o?Object(o.call(t)):{}}t.exports=a},a356:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=function(t){return 0===t?0:1===t?1:2===t?2:t%100>=3&&t%100<=10?3:t%100>=11?4:5},n={s:["أقل من ثانية","ثانية واحدة",["ثانيتان","ثانيتين"],"%d ثوان","%d ثانية","%d ثانية"],m:["أقل من دقيقة","دقيقة واحدة",["دقيقتان","دقيقتين"],"%d دقائق","%d دقيقة","%d دقيقة"],h:["أقل من ساعة","ساعة واحدة",["ساعتان","ساعتين"],"%d ساعات","%d ساعة","%d ساعة"],d:["أقل من يوم","يوم واحد",["يومان","يومين"],"%d أيام","%d يومًا","%d يوم"],M:["أقل من شهر","شهر واحد",["شهران","شهرين"],"%d أشهر","%d شهرا","%d شهر"],y:["أقل من عام","عام واحد",["عامان","عامين"],"%d أعوام","%d عامًا","%d عام"]},i=function(t){return function(i,r,o,a){var s=e(i),l=n[t][e(i)];return 2===s&&(l=l[r?0:1]),l.replace(/%d/i,i)}},r=["جانفي","فيفري","مارس","أفريل","ماي","جوان","جويلية","أوت","سبتمبر","أكتوبر","نوفمبر","ديسمبر"],o=t.defineLocale("ar-dz",{months:r,monthsShort:r,weekdays:"الأحد_الإثنين_الثلاثاء_الأربعاء_الخميس_الجمعة_السبت".split("_"),weekdaysShort:"أحد_إثنين_ثلاثاء_أربعاء_خميس_جمعة_سبت".split("_"),weekdaysMin:"ح_ن_ث_ر_خ_ج_س".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"D/‏M/‏YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},meridiemParse:/ص|م/,isPM:function(t){return"م"===t},meridiem:function(t,e,n){return t<12?"ص":"م"},calendar:{sameDay:"[اليوم عند الساعة] LT",nextDay:"[غدًا عند الساعة] LT",nextWeek:"dddd [عند الساعة] LT",lastDay:"[أمس عند الساعة] LT",lastWeek:"dddd [عند الساعة] LT",sameElse:"L"},relativeTime:{future:"بعد %s",past:"منذ %s",s:i("s"),ss:i("s"),m:i("m"),mm:i("m"),h:i("h"),hh:i("h"),d:i("d"),dd:i("d"),M:i("M"),MM:i("M"),y:i("y"),yy:i("y")},postformat:function(t){return t.replace(/,/g,"،")},week:{dow:0,doy:4}});return o}))},a3fd:function(t,e,n){var i=n("7948");function r(t,e){return i(e,(function(e){return[e,t[e]]}))}t.exports=r},a434:function(t,e,n){"use strict";var i=n("23e7"),r=n("23cb"),o=n("a691"),a=n("50c4"),s=n("7b0b"),l=n("65f0"),u=n("8418"),c=n("1dde"),d=n("ae40"),h=c("splice"),f=d("splice",{ACCESSORS:!0,0:0,1:2}),p=Math.max,m=Math.min,g=9007199254740991,y="Maximum allowed length exceeded";i({target:"Array",proto:!0,forced:!h||!f},{splice:function(t,e){var n,i,c,d,h,f,v=s(this),_=a(v.length),b=r(t,_),x=arguments.length;if(0===x?n=i=0:1===x?(n=0,i=_-b):(n=x-2,i=m(p(o(e),0),_-b)),_+n-i>g)throw TypeError(y);for(c=l(v,i),d=0;d<i;d++)h=b+d,h in v&&u(c,d,v[h]);if(c.length=i,n<i){for(d=b;d<_-i;d++)h=d+i,f=d+n,h in v?v[f]=v[h]:delete v[f];for(d=_;d>_-i+n;d--)delete v[d-1]}else if(n>i)for(d=_-i;d>b;d--)h=d+i-1,f=d+n-1,h in v?v[f]=v[h]:delete v[f];for(d=0;d<n;d++)v[d+b]=arguments[d+2];return v.length=_-i+n,c}})},a454:function(t,e,n){var i=n("72f0"),r=n("3b4a"),o=n("cd9d"),a=r?function(t,e){return r(t,"toString",{configurable:!0,enumerable:!1,value:i(e),writable:!0})}:o;t.exports=a},a4b4:function(t,e,n){var i=n("342f");t.exports=/web0s(?!.*chrome)/i.test(i)},a4d3:function(t,e,n){"use strict";var i=n("23e7"),r=n("da84"),o=n("d066"),a=n("c430"),s=n("83ab"),l=n("4930"),u=n("fdbf"),c=n("d039"),d=n("5135"),h=n("e8b5"),f=n("861d"),p=n("825a"),m=n("7b0b"),g=n("fc6a"),y=n("c04e"),v=n("5c6c"),_=n("7c73"),b=n("df75"),x=n("241c"),w=n("057f"),M=n("7418"),S=n("06cf"),L=n("9bf2"),k=n("d1e7"),T=n("9112"),D=n("6eeb"),C=n("5692"),I=n("f772"),A=n("d012"),Y=n("90e3"),O=n("b622"),P=n("e538"),E=n("746f"),N=n("d44e"),R=n("69f3"),H=n("b727").forEach,j=I("hidden"),z="Symbol",W="prototype",B=O("toPrimitive"),F=R.set,V=R.getterFor(z),G=Object[W],$=r.Symbol,U=o("JSON","stringify"),q=S.f,X=L.f,Z=w.f,J=k.f,K=C("symbols"),Q=C("op-symbols"),tt=C("string-to-symbol-registry"),et=C("symbol-to-string-registry"),nt=C("wks"),it=r.QObject,rt=!it||!it[W]||!it[W].findChild,ot=s&&c((function(){return 7!=_(X({},"a",{get:function(){return X(this,"a",{value:7}).a}})).a}))?function(t,e,n){var i=q(G,e);i&&delete G[e],X(t,e,n),i&&t!==G&&X(G,e,i)}:X,at=function(t,e){var n=K[t]=_($[W]);return F(n,{type:z,tag:t,description:e}),s||(n.description=e),n},st=u?function(t){return"symbol"==typeof t}:function(t){return Object(t)instanceof $},lt=function(t,e,n){t===G&&lt(Q,e,n),p(t);var i=y(e,!0);return p(n),d(K,i)?(n.enumerable?(d(t,j)&&t[j][i]&&(t[j][i]=!1),n=_(n,{enumerable:v(0,!1)})):(d(t,j)||X(t,j,v(1,{})),t[j][i]=!0),ot(t,i,n)):X(t,i,n)},ut=function(t,e){p(t);var n=g(e),i=b(n).concat(pt(n));return H(i,(function(e){s&&!dt.call(n,e)||lt(t,e,n[e])})),t},ct=function(t,e){return void 0===e?_(t):ut(_(t),e)},dt=function(t){var e=y(t,!0),n=J.call(this,e);return!(this===G&&d(K,e)&&!d(Q,e))&&(!(n||!d(this,e)||!d(K,e)||d(this,j)&&this[j][e])||n)},ht=function(t,e){var n=g(t),i=y(e,!0);if(n!==G||!d(K,i)||d(Q,i)){var r=q(n,i);return!r||!d(K,i)||d(n,j)&&n[j][i]||(r.enumerable=!0),r}},ft=function(t){var e=Z(g(t)),n=[];return H(e,(function(t){d(K,t)||d(A,t)||n.push(t)})),n},pt=function(t){var e=t===G,n=Z(e?Q:g(t)),i=[];return H(n,(function(t){!d(K,t)||e&&!d(G,t)||i.push(K[t])})),i};if(l||($=function(){if(this instanceof $)throw TypeError("Symbol is not a constructor");var t=arguments.length&&void 0!==arguments[0]?String(arguments[0]):void 0,e=Y(t),n=function(t){this===G&&n.call(Q,t),d(this,j)&&d(this[j],e)&&(this[j][e]=!1),ot(this,e,v(1,t))};return s&&rt&&ot(G,e,{configurable:!0,set:n}),at(e,t)},D($[W],"toString",(function(){return V(this).tag})),D($,"withoutSetter",(function(t){return at(Y(t),t)})),k.f=dt,L.f=lt,S.f=ht,x.f=w.f=ft,M.f=pt,P.f=function(t){return at(O(t),t)},s&&(X($[W],"description",{configurable:!0,get:function(){return V(this).description}}),a||D(G,"propertyIsEnumerable",dt,{unsafe:!0}))),i({global:!0,wrap:!0,forced:!l,sham:!l},{Symbol:$}),H(b(nt),(function(t){E(t)})),i({target:z,stat:!0,forced:!l},{for:function(t){var e=String(t);if(d(tt,e))return tt[e];var n=$(e);return tt[e]=n,et[n]=e,n},keyFor:function(t){if(!st(t))throw TypeError(t+" is not a symbol");if(d(et,t))return et[t]},useSetter:function(){rt=!0},useSimple:function(){rt=!1}}),i({target:"Object",stat:!0,forced:!l,sham:!s},{create:ct,defineProperty:lt,defineProperties:ut,getOwnPropertyDescriptor:ht}),i({target:"Object",stat:!0,forced:!l},{getOwnPropertyNames:ft,getOwnPropertySymbols:pt}),i({target:"Object",stat:!0,forced:c((function(){M.f(1)}))},{getOwnPropertySymbols:function(t){return M.f(m(t))}}),U){var mt=!l||c((function(){var t=$();return"[null]"!=U([t])||"{}"!=U({a:t})||"{}"!=U(Object(t))}));i({target:"JSON",stat:!0,forced:mt},{stringify:function(t,e,n){var i,r=[t],o=1;while(arguments.length>o)r.push(arguments[o++]);if(i=e,(f(e)||void 0!==t)&&!st(t))return h(e)||(e=function(t,e){if("function"==typeof i&&(e=i.call(this,t,e)),!st(e))return e}),r[1]=e,U.apply(null,r)}})}$[W][B]||T($[W],B,$[W].valueOf),N($,z),A[j]=!0},a524:function(t,e,n){var i=n("4245");function r(t){return i(this,t).has(t)}t.exports=r},a630:function(t,e,n){var i=n("23e7"),r=n("4df4"),o=n("1c7e"),a=!o((function(t){Array.from(t)}));i({target:"Array",stat:!0,forced:a},{from:r})},a640:function(t,e,n){"use strict";var i=n("d039");t.exports=function(t,e){var n=[][t];return!!n&&i((function(){n.call(null,e||function(){throw 1},1)}))}},a691:function(t,e){var n=Math.ceil,i=Math.floor;t.exports=function(t){return isNaN(t=+t)?0:(t>0?i:n)(t)}},a6fb:function(t,e,n){t.exports={assign:n("dce5"),assignIn:n("a9b9"),assignInWith:n("c30c"),assignWith:n("900d"),at:n("b8bb"),create:n("c3f4"),defaults:n("95ae"),defaultsDeep:n("3f84"),entries:n("0b49"),entriesIn:n("8332"),extend:n("cdd8"),extendWith:n("d6dd"),findKey:n("74c8"),findLastKey:n("ab3e"),forIn:n("9948"),forInRight:n("18f6"),forOwn:n("020f"),forOwnRight:n("ca7c"),functions:n("fb25"),functionsIn:n("1693"),get:n("9b02"),has:n("3852"),hasIn:n("8604"),invert:n("27f3"),invertBy:n("e759"),invoke:n("3a0e"),keys:n("ec69"),keysIn:n("9934"),mapKeys:n("4472"),mapValues:n("9e86"),merge:n("42454"),mergeWith:n("2411"),omit:n("3eea"),omitBy:n("dd65"),pick:n("2593"),pickBy:n("77c1"),result:n("4467"),set:n("0f5c"),setWith:n("193b"),toPairs:n("f542"),toPairsIn:n("7aa2"),transform:n("50ca"),unset:n("3cfe"),update:n("ee45"),updateWith:n("e9d1"),values:n("3ff1"),valuesIn:n("7159")}},a77e:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes,style:t.wrapStyles},t._l(t.notices,(function(e){return n("Notice",{key:e.name,attrs:{"prefix-cls":t.prefixCls,styles:e.styles,type:e.type,content:e.content,duration:e.duration,render:e.render,"has-title":e.hasTitle,withIcon:e.withIcon,closable:e.closable,name:e.name,"transition-name":e.transitionName,background:e.background,"msg-type":e.msgType,"on-close":e.onClose}})})),1)},r=[],o=(n("a434"),n("b0c0"),n("ade3")),a=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("transition",{attrs:{name:t.transitionName,appear:""},on:{enter:t.handleEnter,leave:t.handleLeave}},[n("div",{class:t.classes,style:t.styles},["notice"===t.type?[n("div",{ref:"content",class:t.contentClasses,domProps:{innerHTML:t._s(t.content)}}),n("div",{class:t.contentWithIcon},[n("render-cell",{attrs:{render:t.renderFunc}})],1),t.closable?n("a",{class:[t.baseClass+"-close"],on:{click:t.close}},[n("i",{staticClass:"ivu-icon ivu-icon-ios-close"})]):t._e()]:t._e(),"message"===t.type?[n("div",{ref:"content",class:t.messageContentClasses},[n("div",{class:[t.baseClass+"-content-text"],domProps:{innerHTML:t._s(t.content)}}),n("div",{class:[t.baseClass+"-content-text"]},[n("render-cell",{attrs:{render:t.renderFunc}})],1),t.closable?n("a",{class:[t.baseClass+"-close"],on:{click:t.close}},[n("i",{staticClass:"ivu-icon ivu-icon-ios-close"})]):t._e()])]:t._e()],2)])},s=[],l=(n("99af"),n("a9e3"),n("1766")),u={components:{RenderCell:l["a"]},props:{prefixCls:{type:String,default:""},duration:{type:Number,default:1.5},type:{type:String},content:{type:String,default:""},withIcon:Boolean,render:{type:Function},hasTitle:Boolean,styles:{type:Object,default:function(){return{right:"50%"}}},closable:{type:Boolean,default:!1},className:{type:String},name:{type:String,required:!0},onClose:{type:Function},transitionName:{type:String},background:{type:Boolean,default:!1},msgType:{type:String}},data:function(){return{withDesc:!1}},computed:{baseClass:function(){return"".concat(this.prefixCls,"-notice")},renderFunc:function(){return this.render||function(){}},classes:function(){var t;return[this.baseClass,(t={},Object(o["a"])(t,"".concat(this.className),!!this.className),Object(o["a"])(t,"".concat(this.baseClass,"-closable"),this.closable),Object(o["a"])(t,"".concat(this.baseClass,"-with-desc"),this.withDesc),Object(o["a"])(t,"".concat(this.baseClass,"-with-background"),this.background),t)]},contentClasses:function(){return["".concat(this.baseClass,"-content"),void 0!==this.render?"".concat(this.baseClass,"-content-with-render"):""]},messageContentClasses:function(){var t;return["".concat(this.baseClass,"-content"),(t={},Object(o["a"])(t,"".concat(this.baseClass,"-content-").concat(this.msgType),this.msgType),Object(o["a"])(t,"".concat(this.baseClass,"-content-background"),this.background),t)]},contentWithIcon:function(){return[this.withIcon?"".concat(this.prefixCls,"-content-with-icon"):"",!this.hasTitle&&this.withIcon?"".concat(this.prefixCls,"-content-with-render-notitle"):""]},messageClasses:function(){return["".concat(this.baseClass,"-content"),void 0!==this.render?"".concat(this.baseClass,"-content-with-render"):""]}},methods:{clearCloseTimer:function(){this.closeTimer&&(clearTimeout(this.closeTimer),this.closeTimer=null)},close:function(){this.clearCloseTimer(),this.onClose(),this.$parent.close(this.name)},handleEnter:function(t){"message"===this.type&&(t.style.height=t.scrollHeight+"px")},handleLeave:function(t){"message"===this.type&&1!==document.getElementsByClassName("ivu-message-notice").length&&(t.style.height=0,t.style.paddingTop=0,t.style.paddingBottom=0)}},mounted:function(){var t=this;if(this.clearCloseTimer(),0!==this.duration&&(this.closeTimer=setTimeout((function(){t.close()}),1e3*this.duration)),"ivu-notice"===this.prefixCls){var e=this.$refs.content.querySelectorAll(".".concat(this.prefixCls,"-desc"))[0];this.withDesc=!!this.render||!!e&&""!==e.innerHTML}},beforeDestroy:function(){this.clearCloseTimer()}},c=u,d=n("2877"),h=Object(d["a"])(c,a,s,!1,null,null,null),f=h.exports,p=n("c177"),m="ivu-notification",g=0,y=Date.now();function v(){return"ivuNotification_"+y+"_"+g++}var _={components:{Notice:f},props:{prefixCls:{type:String,default:m},styles:{type:Object,default:function(){return{top:"65px",left:"50%"}}},content:{type:String},className:{type:String}},data:function(){return{notices:[],tIndex:this.handleGetIndex()}},computed:{classes:function(){return["".concat(this.prefixCls),Object(o["a"])({},"".concat(this.className),!!this.className)]},wrapStyles:function(){var t=Object.assign({},this.styles);return t["z-index"]=1010+this.tIndex,t}},methods:{add:function(t){var e=t.name||v(),n=Object.assign({styles:{right:"50%"},content:"",duration:1.5,closable:!1,name:e},t);this.notices.push(n),this.tIndex=this.handleGetIndex()},close:function(t){for(var e=this.notices,n=0;n<e.length;n++)if(e[n].name===t){this.notices.splice(n,1);break}},closeAll:function(){this.notices=[]},handleGetIndex:function(){return Object(p["a"])(),p["b"]}}},b=_,x=Object(d["a"])(b,i,r,!1,null,null,null),w=x.exports,M=n("2b0e");w.newInstance=t=>{const e=t||{},n=new M["a"]({render(t){return t(w,{props:e})}}),i=n.$mount();document.body.appendChild(i.$el);const r=n.$children[0];return{notice(t){r.add(t)},remove(t){r.close(t)},component:r,destroy(t){r.closeAll(),setTimeout((function(){document.body.removeChild(document.getElementsByClassName(t)[0])}),500)}}};e["a"]=w},a78b:function(t,e,n){var i=n("656b"),r=n("159a");function o(t,e,n,o){return r(t,e,n(i(t,e)),o)}t.exports=o},a79d:function(t,e,n){"use strict";var i=n("23e7"),r=n("c430"),o=n("fea9"),a=n("d039"),s=n("d066"),l=n("4840"),u=n("cdf9"),c=n("6eeb"),d=!!o&&a((function(){o.prototype["finally"].call({then:function(){}},(function(){}))}));i({target:"Promise",proto:!0,real:!0,forced:d},{finally:function(t){var e=l(this,s("Promise")),n="function"==typeof t;return this.then(n?function(n){return u(e,t()).then((function(){return n}))}:t,n?function(n){return u(e,t()).then((function(){throw n}))}:t)}}),r||"function"!=typeof o||o.prototype["finally"]||c(o.prototype,"finally",s("Promise").prototype["finally"])},a7fa:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("bm",{months:"Zanwuyekalo_Fewuruyekalo_Marisikalo_Awirilikalo_Mɛkalo_Zuwɛnkalo_Zuluyekalo_Utikalo_Sɛtanburukalo_ɔkutɔburukalo_Nowanburukalo_Desanburukalo".split("_"),monthsShort:"Zan_Few_Mar_Awi_Mɛ_Zuw_Zul_Uti_Sɛt_ɔku_Now_Des".split("_"),weekdays:"Kari_Ntɛnɛn_Tarata_Araba_Alamisa_Juma_Sibiri".split("_"),weekdaysShort:"Kar_Ntɛ_Tar_Ara_Ala_Jum_Sib".split("_"),weekdaysMin:"Ka_Nt_Ta_Ar_Al_Ju_Si".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"MMMM [tile] D [san] YYYY",LLL:"MMMM [tile] D [san] YYYY [lɛrɛ] HH:mm",LLLL:"dddd MMMM [tile] D [san] YYYY [lɛrɛ] HH:mm"},calendar:{sameDay:"[Bi lɛrɛ] LT",nextDay:"[Sini lɛrɛ] LT",nextWeek:"dddd [don lɛrɛ] LT",lastDay:"[Kunu lɛrɛ] LT",lastWeek:"dddd [tɛmɛnen lɛrɛ] LT",sameElse:"L"},relativeTime:{future:"%s kɔnɔ",past:"a bɛ %s bɔ",s:"sanga dama dama",ss:"sekondi %d",m:"miniti kelen",mm:"miniti %d",h:"lɛrɛ kelen",hh:"lɛrɛ %d",d:"tile kelen",dd:"tile %d",M:"kalo kelen",MM:"kalo %d",y:"san kelen",yy:"san %d"},week:{dow:1,doy:4}});return e}))},a994:function(t,e,n){var i=n("7d1f"),r=n("32f4"),o=n("ec69");function a(t){return i(t,o,r)}t.exports=a},a9b9:function(t,e,n){var i=n("8eeb"),r=n("2ec1"),o=n("9934"),a=r((function(t,e){i(e,o(e),t)}));t.exports=a},a9e3:function(t,e,n){"use strict";var i=n("83ab"),r=n("da84"),o=n("94ca"),a=n("6eeb"),s=n("5135"),l=n("c6b6"),u=n("7156"),c=n("c04e"),d=n("d039"),h=n("7c73"),f=n("241c").f,p=n("06cf").f,m=n("9bf2").f,g=n("58a8").trim,y="Number",v=r[y],_=v.prototype,b=l(h(_))==y,x=function(t){var e,n,i,r,o,a,s,l,u=c(t,!1);if("string"==typeof u&&u.length>2)if(u=g(u),e=u.charCodeAt(0),43===e||45===e){if(n=u.charCodeAt(2),88===n||120===n)return NaN}else if(48===e){switch(u.charCodeAt(1)){case 66:case 98:i=2,r=49;break;case 79:case 111:i=8,r=55;break;default:return+u}for(o=u.slice(2),a=o.length,s=0;s<a;s++)if(l=o.charCodeAt(s),l<48||l>r)return NaN;return parseInt(o,i)}return+u};if(o(y,!v(" 0o1")||!v("0b1")||v("+0x1"))){for(var w,M=function(t){var e=arguments.length<1?0:t,n=this;return n instanceof M&&(b?d((function(){_.valueOf.call(n)})):l(n)!=y)?u(new v(x(e)),n,M):x(e)},S=i?f(v):"MAX_VALUE,MIN_VALUE,NaN,NEGATIVE_INFINITY,POSITIVE_INFINITY,EPSILON,isFinite,isInteger,isNaN,isSafeInteger,MAX_SAFE_INTEGER,MIN_SAFE_INTEGER,parseFloat,parseInt,isInteger,fromString,range".split(","),L=0;S.length>L;L++)s(v,w=S[L])&&!s(M,w)&&m(M,w,p(v,w));M.prototype=_,_.constructor=M,a(r,y,M)}},aa30:function(t,e,n){"use strict";var i=n("a77e");const r="ivu-message",o="ivu-icon",a="ivu_message_key_",s={top:24,duration:1.5};let l,u=1;const c={info:"ios-information-circle",success:"ios-checkmark-circle",warning:"ios-alert",error:"ios-close-circle",loading:"ios-loading"};function d(){return l=l||i["a"].newInstance({prefixCls:r,styles:{top:s.top+"px"}}),l}function h(t="",e=s.duration,n,i=function(){},l=!1,h=function(){},f=!1){const p=c[n],m="loading"===n?" ivu-load-loop":"";let g=d();return g.notice({name:`${a}${u}`,duration:e,styles:{},transitionName:"move-up",content:`\n            <div class="${r}-custom-content ${r}-${n}">\n                <i class="${o} ${o}-${p} ${m}"></i>\n                <span>${t}</span>\n            </div>\n        `,render:h,onClose:i,closable:l,type:"message",msgType:n,background:f}),function(){let t=u++;return function(){g.remove(`${a}${t}`)}}()}e["a"]={name:"Message",info(t){return this.message("info",t)},success(t){return this.message("success",t)},warning(t){return this.message("warning",t)},error(t){return this.message("error",t)},loading(t){return this.message("loading",t)},message(t,e){return"string"===typeof e&&(e={content:e}),h(e.content,e.duration,t,e.onClose,e.closable,e.render,e.background)},config(t){(t.top||0===t.top)&&(s.top=t.top),(t.duration||0===t.duration)&&(s.duration=t.duration)},destroy(){let t=d();l=null,t.destroy("ivu-message")}}},aaf2:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r={s:["थोडया सॅकंडांनी","थोडे सॅकंड"],ss:[t+" सॅकंडांनी",t+" सॅकंड"],m:["एका मिणटान","एक मिनूट"],mm:[t+" मिणटांनी",t+" मिणटां"],h:["एका वरान","एक वर"],hh:[t+" वरांनी",t+" वरां"],d:["एका दिसान","एक दीस"],dd:[t+" दिसांनी",t+" दीस"],M:["एका म्हयन्यान","एक म्हयनो"],MM:[t+" म्हयन्यानी",t+" म्हयने"],y:["एका वर्सान","एक वर्स"],yy:[t+" वर्सांनी",t+" वर्सां"]};return i?r[n][0]:r[n][1]}var n=t.defineLocale("gom-deva",{months:{standalone:"जानेवारी_फेब्रुवारी_मार्च_एप्रील_मे_जून_जुलय_ऑगस्ट_सप्टेंबर_ऑक्टोबर_नोव्हेंबर_डिसेंबर".split("_"),format:"जानेवारीच्या_फेब्रुवारीच्या_मार्चाच्या_एप्रीलाच्या_मेयाच्या_जूनाच्या_जुलयाच्या_ऑगस्टाच्या_सप्टेंबराच्या_ऑक्टोबराच्या_नोव्हेंबराच्या_डिसेंबराच्या".split("_"),isFormat:/MMMM(\s)+D[oD]?/},monthsShort:"जाने._फेब्रु._मार्च_एप्री._मे_जून_जुल._ऑग._सप्टें._ऑक्टो._नोव्हें._डिसें.".split("_"),monthsParseExact:!0,weekdays:"आयतार_सोमार_मंगळार_बुधवार_बिरेस्तार_सुक्रार_शेनवार".split("_"),weekdaysShort:"आयत._सोम._मंगळ._बुध._ब्रेस्त._सुक्र._शेन.".split("_"),weekdaysMin:"आ_सो_मं_बु_ब्रे_सु_शे".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"A h:mm [वाजतां]",LTS:"A h:mm:ss [वाजतां]",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY A h:mm [वाजतां]",LLLL:"dddd, MMMM Do, YYYY, A h:mm [वाजतां]",llll:"ddd, D MMM YYYY, A h:mm [वाजतां]"},calendar:{sameDay:"[आयज] LT",nextDay:"[फाल्यां] LT",nextWeek:"[फुडलो] dddd[,] LT",lastDay:"[काल] LT",lastWeek:"[फाटलो] dddd[,] LT",sameElse:"L"},relativeTime:{future:"%s",past:"%s आदीं",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:e,M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}(वेर)/,ordinal:function(t,e){switch(e){case"D":return t+"वेर";default:case"M":case"Q":case"DDD":case"d":case"w":case"W":return t}},week:{dow:0,doy:3},meridiemParse:/राती|सकाळीं|दनपारां|सांजे/,meridiemHour:function(t,e){return 12===t&&(t=0),"राती"===e?t<4?t:t+12:"सकाळीं"===e?t:"दनपारां"===e?t>12?t:t+12:"सांजे"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"राती":t<12?"सकाळीं":t<16?"दनपारां":t<20?"सांजे":"राती"}});return n}))},ab13:function(t,e,n){var i=n("b622"),r=i("match");t.exports=function(t){var e=/./;try{"/./"[t](e)}catch(n){try{return e[r]=!1,"/./"[t](e)}catch(i){}}return!1}},ab3e:function(t,e,n){var i=n("972ca"),r=n("4be8"),o=n("badf");function a(t,e){return i(t,o(e,3),r)}t.exports=a},abb4:function(t,e,n){"use strict";t.exports=function(t){function e(){}var n={log:e,warn:e,error:e};if(!t&&window.console){var i=function(t,e){t[e]=function(){var t=console[e];if(t.apply)t.apply(console,arguments);else for(var n=0;n<arguments.length;n++)t(arguments[n])}};i(n,"log"),i(n,"warn"),i(n,"error")}return n}},ac1f:function(t,e,n){"use strict";var i=n("23e7"),r=n("9263");i({target:"RegExp",proto:!0,forced:/./.exec!==r},{exec:r})},ac41:function(t,e){function n(t){var e=-1,n=Array(t.size);return t.forEach((function(t){n[++e]=t})),n}t.exports=n},ad6d:function(t,e,n){"use strict";var i=n("825a");t.exports=function(){var t=i(this),e="";return t.global&&(e+="g"),t.ignoreCase&&(e+="i"),t.multiline&&(e+="m"),t.dotAll&&(e+="s"),t.unicode&&(e+="u"),t.sticky&&(e+="y"),e}},ada2:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e){var n=t.split("_");return e%10===1&&e%100!==11?n[0]:e%10>=2&&e%10<=4&&(e%100<10||e%100>=20)?n[1]:n[2]}function n(t,n,i){var r={ss:n?"секунда_секунди_секунд":"секунду_секунди_секунд",mm:n?"хвилина_хвилини_хвилин":"хвилину_хвилини_хвилин",hh:n?"година_години_годин":"годину_години_годин",dd:"день_дні_днів",MM:"місяць_місяці_місяців",yy:"рік_роки_років"};return"m"===i?n?"хвилина":"хвилину":"h"===i?n?"година":"годину":t+" "+e(r[i],+t)}function i(t,e){var n,i={nominative:"неділя_понеділок_вівторок_середа_четвер_п’ятниця_субота".split("_"),accusative:"неділю_понеділок_вівторок_середу_четвер_п’ятницю_суботу".split("_"),genitive:"неділі_понеділка_вівторка_середи_четверга_п’ятниці_суботи".split("_")};return!0===t?i["nominative"].slice(1,7).concat(i["nominative"].slice(0,1)):t?(n=/(\[[ВвУу]\]) ?dddd/.test(e)?"accusative":/\[?(?:минулої|наступної)? ?\] ?dddd/.test(e)?"genitive":"nominative",i[n][t.day()]):i["nominative"]}function r(t){return function(){return t+"о"+(11===this.hours()?"б":"")+"] LT"}}var o=t.defineLocale("uk",{months:{format:"січня_лютого_березня_квітня_травня_червня_липня_серпня_вересня_жовтня_листопада_грудня".split("_"),standalone:"січень_лютий_березень_квітень_травень_червень_липень_серпень_вересень_жовтень_листопад_грудень".split("_")},monthsShort:"січ_лют_бер_квіт_трав_черв_лип_серп_вер_жовт_лист_груд".split("_"),weekdays:i,weekdaysShort:"нд_пн_вт_ср_чт_пт_сб".split("_"),weekdaysMin:"нд_пн_вт_ср_чт_пт_сб".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY р.",LLL:"D MMMM YYYY р., HH:mm",LLLL:"dddd, D MMMM YYYY р., HH:mm"},calendar:{sameDay:r("[Сьогодні "),nextDay:r("[Завтра "),lastDay:r("[Вчора "),nextWeek:r("[У] dddd ["),lastWeek:function(){switch(this.day()){case 0:case 3:case 5:case 6:return r("[Минулої] dddd [").call(this);case 1:case 2:case 4:return r("[Минулого] dddd [").call(this)}},sameElse:"L"},relativeTime:{future:"за %s",past:"%s тому",s:"декілька секунд",ss:n,m:n,mm:n,h:"годину",hh:n,d:"день",dd:n,M:"місяць",MM:n,y:"рік",yy:n},meridiemParse:/ночі|ранку|дня|вечора/,isPM:function(t){return/^(дня|вечора)$/.test(t)},meridiem:function(t,e,n){return t<4?"ночі":t<12?"ранку":t<17?"дня":"вечора"},dayOfMonthOrdinalParse:/\d{1,2}-(й|го)/,ordinal:function(t,e){switch(e){case"M":case"d":case"DDD":case"w":case"W":return t+"-й";case"D":return t+"-го";default:return t}},week:{dow:1,doy:7}});return o}))},ade3:function(t,e,n){"use strict";function i(t,e,n){return e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t}n.d(e,"a",(function(){return i}))},ae40:function(t,e,n){var i=n("83ab"),r=n("d039"),o=n("5135"),a=Object.defineProperty,s={},l=function(t){throw t};t.exports=function(t,e){if(o(s,t))return s[t];e||(e={});var n=[][t],u=!!o(e,"ACCESSORS")&&e.ACCESSORS,c=o(e,0)?e[0]:l,d=o(e,1)?e[1]:void 0;return s[t]=!!n&&!r((function(){if(u&&!i)return!0;var t={length:-1};u?a(t,1,{enumerable:!0,get:l}):t[1]=1,n.call(t,c,d)}))}},ae93:function(t,e,n){"use strict";var i,r,o,a=n("d039"),s=n("e163"),l=n("9112"),u=n("5135"),c=n("b622"),d=n("c430"),h=c("iterator"),f=!1,p=function(){return this};[].keys&&(o=[].keys(),"next"in o?(r=s(s(o)),r!==Object.prototype&&(i=r)):f=!0);var m=void 0==i||a((function(){var t={};return i[h].call(t)!==t}));m&&(i={}),d&&!m||u(i,h)||l(i,h,p),t.exports={IteratorPrototype:i,BUGGY_SAFARI_ITERATORS:f}},afb9:function(t,e,n){var i=n("2d7c"),r=n("9520");function o(t,e){return i(e,(function(e){return r(t[e])}))}t.exports=o},b041:function(t,e,n){"use strict";var i=n("00ee"),r=n("f5df");t.exports=i?{}.toString:function(){return"[object "+r(this)+"]"}},b047:function(t,e){function n(t){return function(e){return t(e)}}t.exports=n},b0c0:function(t,e,n){var i=n("83ab"),r=n("9bf2").f,o=Function.prototype,a=o.toString,s=/^\s*function ([^ (]*)/,l="name";i&&!(l in o)&&r(o,l,{configurable:!0,get:function(){try{return a.call(this).match(s)[1]}catch(t){return""}}})},b1e5:function(t,e,n){var i=n("a994"),r=1,o=Object.prototype,a=o.hasOwnProperty;function s(t,e,n,o,s,l){var u=n&r,c=i(t),d=c.length,h=i(e),f=h.length;if(d!=f&&!u)return!1;var p=d;while(p--){var m=c[p];if(!(u?m in e:a.call(e,m)))return!1}var g=l.get(t),y=l.get(e);if(g&&y)return g==e&&y==t;var v=!0;l.set(t,e),l.set(e,t);var _=u;while(++p<d){m=c[p];var b=t[m],x=e[m];if(o)var w=u?o(x,b,m,e,t,l):o(b,x,m,t,e,l);if(!(void 0===w?b===x||s(b,x,n,o,l):w)){v=!1;break}_||(_="constructor"==m)}if(v&&!_){var M=t.constructor,S=e.constructor;M==S||!("constructor"in t)||!("constructor"in e)||"function"==typeof M&&M instanceof M&&"function"==typeof S&&S instanceof S||(v=!1)}return l["delete"](t),l["delete"](e),v}t.exports=s},b218:function(t,e){var n=9007199254740991;function i(t){return"number"==typeof t&&t>-1&&t%1==0&&t<=n}t.exports=i},b29d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("lo",{months:"ມັງກອນ_ກຸມພາ_ມີນາ_ເມສາ_ພຶດສະພາ_ມິຖຸນາ_ກໍລະກົດ_ສິງຫາ_ກັນຍາ_ຕຸລາ_ພະຈິກ_ທັນວາ".split("_"),monthsShort:"ມັງກອນ_ກຸມພາ_ມີນາ_ເມສາ_ພຶດສະພາ_ມິຖຸນາ_ກໍລະກົດ_ສິງຫາ_ກັນຍາ_ຕຸລາ_ພະຈິກ_ທັນວາ".split("_"),weekdays:"ອາທິດ_ຈັນ_ອັງຄານ_ພຸດ_ພະຫັດ_ສຸກ_ເສົາ".split("_"),weekdaysShort:"ທິດ_ຈັນ_ອັງຄານ_ພຸດ_ພະຫັດ_ສຸກ_ເສົາ".split("_"),weekdaysMin:"ທ_ຈ_ອຄ_ພ_ພຫ_ສກ_ສ".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"ວັນdddd D MMMM YYYY HH:mm"},meridiemParse:/ຕອນເຊົ້າ|ຕອນແລງ/,isPM:function(t){return"ຕອນແລງ"===t},meridiem:function(t,e,n){return t<12?"ຕອນເຊົ້າ":"ຕອນແລງ"},calendar:{sameDay:"[ມື້ນີ້ເວລາ] LT",nextDay:"[ມື້ອື່ນເວລາ] LT",nextWeek:"[ວັນ]dddd[ໜ້າເວລາ] LT",lastDay:"[ມື້ວານນີ້ເວລາ] LT",lastWeek:"[ວັນ]dddd[ແລ້ວນີ້ເວລາ] LT",sameElse:"L"},relativeTime:{future:"ອີກ %s",past:"%sຜ່ານມາ",s:"ບໍ່ເທົ່າໃດວິນາທີ",ss:"%d ວິນາທີ",m:"1 ນາທີ",mm:"%d ນາທີ",h:"1 ຊົ່ວໂມງ",hh:"%d ຊົ່ວໂມງ",d:"1 ມື້",dd:"%d ມື້",M:"1 ເດືອນ",MM:"%d ເດືອນ",y:"1 ປີ",yy:"%d ປີ"},dayOfMonthOrdinalParse:/(ທີ່)\d{1,2}/,ordinal:function(t){return"ທີ່"+t}});return e}))},b2e6:function(t,e,n){"use strict";var i=n("3a27");e["a"]={props:{lockScroll:{type:Boolean,default:!0}},methods:{checkScrollBar(){let t=window.innerWidth;if(!t){const e=document.documentElement.getBoundingClientRect();t=e.right-Math.abs(e.left)}this.bodyIsOverflowing=document.body.clientWidth<t,this.bodyIsOverflowing&&(this.scrollBarWidth=Object(i["g"])())},checkMaskInVisible(){let t=document.getElementsByClassName("ivu-modal-mask")||[];return Array.from(t).every(t=>"none"===t.style.display||t.classList.contains("fade-leave-to"))},setScrollBar(){this.bodyIsOverflowing&&void 0!==this.scrollBarWidth&&(document.body.style.paddingRight=this.scrollBarWidth+"px")},resetScrollBar(){document.body.style.paddingRight=""},addScrollEffect(){this.lockScroll&&(this.checkScrollBar(),this.setScrollBar(),document.body.style.overflow="hidden")},removeScrollEffect(){this.lockScroll&&this.checkMaskInVisible()&&(document.body.style.overflow="",this.resetScrollBar())}}}},b3e6:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes,style:t.styles},[t._t("default")],2)},r=[],o=(n("99af"),n("4160"),n("a9e3"),n("159b"),n("ade3")),a=n("3a27"),s="ivu-row",l={name:"Row",props:{type:{validator:function(t){return Object(a["i"])(t,["flex"])}},align:{validator:function(t){return Object(a["i"])(t,["top","middle","bottom"])}},justify:{validator:function(t){return Object(a["i"])(t,["start","end","center","space-around","space-between"])}},gutter:{type:Number,default:0},className:String},computed:{classes:function(){var t;return[(t={},Object(o["a"])(t,"".concat(s),!this.type),Object(o["a"])(t,"".concat(s,"-").concat(this.type),!!this.type),Object(o["a"])(t,"".concat(s,"-").concat(this.type,"-").concat(this.align),!!this.align),Object(o["a"])(t,"".concat(s,"-").concat(this.type,"-").concat(this.justify),!!this.justify),Object(o["a"])(t,"".concat(this.className),!!this.className),t)]},styles:function(){var t={};return 0!==this.gutter&&(t={marginLeft:this.gutter/-2+"px",marginRight:this.gutter/-2+"px"}),t}},methods:{updateGutter:function(t){var e=Object(a["d"])(this,"iCol"),n=Object(a["c"])(e,"iCol",!1);n.length&&n.forEach((function(e){0!==t&&(e.gutter=t)}))}},watch:{gutter:function(t){this.updateGutter(t)}}},u=l,c=n("2877"),d=Object(c["a"])(u,i,r,!1,null,null,null),h=d.exports;e["a"]=h},b3eb:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r={m:["eine Minute","einer Minute"],h:["eine Stunde","einer Stunde"],d:["ein Tag","einem Tag"],dd:[t+" Tage",t+" Tagen"],w:["eine Woche","einer Woche"],M:["ein Monat","einem Monat"],MM:[t+" Monate",t+" Monaten"],y:["ein Jahr","einem Jahr"],yy:[t+" Jahre",t+" Jahren"]};return e?r[n][0]:r[n][1]}var n=t.defineLocale("de-at",{months:"Jänner_Februar_März_April_Mai_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jän._Feb._März_Apr._Mai_Juni_Juli_Aug._Sep._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonntag_Montag_Dienstag_Mittwoch_Donnerstag_Freitag_Samstag".split("_"),weekdaysShort:"So._Mo._Di._Mi._Do._Fr._Sa.".split("_"),weekdaysMin:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd, D. MMMM YYYY HH:mm"},calendar:{sameDay:"[heute um] LT [Uhr]",sameElse:"L",nextDay:"[morgen um] LT [Uhr]",nextWeek:"dddd [um] LT [Uhr]",lastDay:"[gestern um] LT [Uhr]",lastWeek:"[letzten] dddd [um] LT [Uhr]"},relativeTime:{future:"in %s",past:"vor %s",s:"ein paar Sekunden",ss:"%d Sekunden",m:e,mm:"%d Minuten",h:e,hh:"%d Stunden",d:e,dd:e,w:e,ww:"%d Wochen",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},b469:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r={m:["eine Minute","einer Minute"],h:["eine Stunde","einer Stunde"],d:["ein Tag","einem Tag"],dd:[t+" Tage",t+" Tagen"],w:["eine Woche","einer Woche"],M:["ein Monat","einem Monat"],MM:[t+" Monate",t+" Monaten"],y:["ein Jahr","einem Jahr"],yy:[t+" Jahre",t+" Jahren"]};return e?r[n][0]:r[n][1]}var n=t.defineLocale("de",{months:"Januar_Februar_März_April_Mai_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jan._Feb._März_Apr._Mai_Juni_Juli_Aug._Sep._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonntag_Montag_Dienstag_Mittwoch_Donnerstag_Freitag_Samstag".split("_"),weekdaysShort:"So._Mo._Di._Mi._Do._Fr._Sa.".split("_"),weekdaysMin:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd, D. MMMM YYYY HH:mm"},calendar:{sameDay:"[heute um] LT [Uhr]",sameElse:"L",nextDay:"[morgen um] LT [Uhr]",nextWeek:"dddd [um] LT [Uhr]",lastDay:"[gestern um] LT [Uhr]",lastWeek:"[letzten] dddd [um] LT [Uhr]"},relativeTime:{future:"in %s",past:"vor %s",s:"ein paar Sekunden",ss:"%d Sekunden",m:e,mm:"%d Minuten",h:e,hh:"%d Stunden",d:e,dd:e,w:e,ww:"%d Wochen",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},b4c0:function(t,e,n){var i=n("cb5a");function r(t){var e=this.__data__,n=i(e,t);return n<0?void 0:e[n][1]}t.exports=r},b50d:function(t,e,n){"use strict";var i=n("c532"),r=n("467f"),o=n("7aac"),a=n("30b5"),s=n("83b9"),l=n("c345"),u=n("3934"),c=n("2d83");t.exports=function(t){return new Promise((function(e,n){var d=t.data,h=t.headers;i.isFormData(d)&&delete h["Content-Type"];var f=new XMLHttpRequest;if(t.auth){var p=t.auth.username||"",m=t.auth.password?unescape(encodeURIComponent(t.auth.password)):"";h.Authorization="Basic "+btoa(p+":"+m)}var g=s(t.baseURL,t.url);if(f.open(t.method.toUpperCase(),a(g,t.params,t.paramsSerializer),!0),f.timeout=t.timeout,f.onreadystatechange=function(){if(f&&4===f.readyState&&(0!==f.status||f.responseURL&&0===f.responseURL.indexOf("file:"))){var i="getAllResponseHeaders"in f?l(f.getAllResponseHeaders()):null,o=t.responseType&&"text"!==t.responseType?f.response:f.responseText,a={data:o,status:f.status,statusText:f.statusText,headers:i,config:t,request:f};r(e,n,a),f=null}},f.onabort=function(){f&&(n(c("Request aborted",t,"ECONNABORTED",f)),f=null)},f.onerror=function(){n(c("Network Error",t,null,f)),f=null},f.ontimeout=function(){var e="timeout of "+t.timeout+"ms exceeded";t.timeoutErrorMessage&&(e=t.timeoutErrorMessage),n(c(e,t,"ECONNABORTED",f)),f=null},i.isStandardBrowserEnv()){var y=(t.withCredentials||u(g))&&t.xsrfCookieName?o.read(t.xsrfCookieName):void 0;y&&(h[t.xsrfHeaderName]=y)}if("setRequestHeader"in f&&i.forEach(h,(function(t,e){"undefined"===typeof d&&"content-type"===e.toLowerCase()?delete h[e]:f.setRequestHeader(e,t)})),i.isUndefined(t.withCredentials)||(f.withCredentials=!!t.withCredentials),t.responseType)try{f.responseType=t.responseType}catch(v){if("json"!==t.responseType)throw v}"function"===typeof t.onDownloadProgress&&f.addEventListener("progress",t.onDownloadProgress),"function"===typeof t.onUploadProgress&&f.upload&&f.upload.addEventListener("progress",t.onUploadProgress),t.cancelToken&&t.cancelToken.promise.then((function(t){f&&(f.abort(),n(t),f=null)})),d||(d=null),f.send(d)}))}},b53d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tzm-latn",{months:"innayr_brˤayrˤ_marˤsˤ_ibrir_mayyw_ywnyw_ywlywz_ɣwšt_šwtanbir_ktˤwbrˤ_nwwanbir_dwjnbir".split("_"),monthsShort:"innayr_brˤayrˤ_marˤsˤ_ibrir_mayyw_ywnyw_ywlywz_ɣwšt_šwtanbir_ktˤwbrˤ_nwwanbir_dwjnbir".split("_"),weekdays:"asamas_aynas_asinas_akras_akwas_asimwas_asiḍyas".split("_"),weekdaysShort:"asamas_aynas_asinas_akras_akwas_asimwas_asiḍyas".split("_"),weekdaysMin:"asamas_aynas_asinas_akras_akwas_asimwas_asiḍyas".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[asdkh g] LT",nextDay:"[aska g] LT",nextWeek:"dddd [g] LT",lastDay:"[assant g] LT",lastWeek:"dddd [g] LT",sameElse:"L"},relativeTime:{future:"dadkh s yan %s",past:"yan %s",s:"imik",ss:"%d imik",m:"minuḍ",mm:"%d minuḍ",h:"saɛa",hh:"%d tassaɛin",d:"ass",dd:"%d ossan",M:"ayowr",MM:"%d iyyirn",y:"asgas",yy:"%d isgasn"},week:{dow:6,doy:12}});return e}))},b540:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("jv",{months:"Januari_Februari_Maret_April_Mei_Juni_Juli_Agustus_September_Oktober_Nopember_Desember".split("_"),monthsShort:"Jan_Feb_Mar_Apr_Mei_Jun_Jul_Ags_Sep_Okt_Nop_Des".split("_"),weekdays:"Minggu_Senen_Seloso_Rebu_Kemis_Jemuwah_Septu".split("_"),weekdaysShort:"Min_Sen_Sel_Reb_Kem_Jem_Sep".split("_"),weekdaysMin:"Mg_Sn_Sl_Rb_Km_Jm_Sp".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/enjing|siyang|sonten|ndalu/,meridiemHour:function(t,e){return 12===t&&(t=0),"enjing"===e?t:"siyang"===e?t>=11?t:t+12:"sonten"===e||"ndalu"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"enjing":t<15?"siyang":t<19?"sonten":"ndalu"},calendar:{sameDay:"[Dinten puniko pukul] LT",nextDay:"[Mbenjang pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kala wingi pukul] LT",lastWeek:"dddd [kepengker pukul] LT",sameElse:"L"},relativeTime:{future:"wonten ing %s",past:"%s ingkang kepengker",s:"sawetawis detik",ss:"%d detik",m:"setunggal menit",mm:"%d menit",h:"setunggal jam",hh:"%d jam",d:"sedinten",dd:"%d dinten",M:"sewulan",MM:"%d wulan",y:"setaun",yy:"%d taun"},week:{dow:1,doy:7}});return e}))},b575:function(t,e,n){var i,r,o,a,s,l,u,c,d=n("da84"),h=n("06cf").f,f=n("2cf4").set,p=n("1cdc"),m=n("a4b4"),g=n("605d"),y=d.MutationObserver||d.WebKitMutationObserver,v=d.document,_=d.process,b=d.Promise,x=h(d,"queueMicrotask"),w=x&&x.value;w||(i=function(){var t,e;g&&(t=_.domain)&&t.exit();while(r){e=r.fn,r=r.next;try{e()}catch(n){throw r?a():o=void 0,n}}o=void 0,t&&t.enter()},p||g||m||!y||!v?b&&b.resolve?(u=b.resolve(void 0),c=u.then,a=function(){c.call(u,i)}):a=g?function(){_.nextTick(i)}:function(){f.call(d,i)}:(s=!0,l=v.createTextNode(""),new y(i).observe(l,{characterData:!0}),a=function(){l.data=s=!s})),t.exports=w||function(t){var e={fn:t,next:void 0};o&&(o.next=e),r||(r=e,a()),o=e}},b5a7:function(t,e,n){var i=n("0b07"),r=n("2b3e"),o=i(r,"DataView");t.exports=o},b5b7:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="ene._feb._mar._abr._may._jun._jul._ago._sep._oct._nov._dic.".split("_"),n="ene_feb_mar_abr_may_jun_jul_ago_sep_oct_nov_dic".split("_"),i=[/^ene/i,/^feb/i,/^mar/i,/^abr/i,/^may/i,/^jun/i,/^jul/i,/^ago/i,/^sep/i,/^oct/i,/^nov/i,/^dic/i],r=/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre|ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,o=t.defineLocale("es-mx",{months:"enero_febrero_marzo_abril_mayo_junio_julio_agosto_septiembre_octubre_noviembre_diciembre".split("_"),monthsShort:function(t,i){return t?/-MMM-/.test(i)?n[t.month()]:e[t.month()]:e},monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:/^(enero|febrero|marzo|abril|mayo|junio|julio|agosto|septiembre|octubre|noviembre|diciembre)/i,monthsShortStrictRegex:/^(ene\.?|feb\.?|mar\.?|abr\.?|may\.?|jun\.?|jul\.?|ago\.?|sep\.?|oct\.?|nov\.?|dic\.?)/i,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"domingo_lunes_martes_miércoles_jueves_viernes_sábado".split("_"),weekdaysShort:"dom._lun._mar._mié._jue._vie._sáb.".split("_"),weekdaysMin:"do_lu_ma_mi_ju_vi_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY H:mm",LLLL:"dddd, D [de] MMMM [de] YYYY H:mm"},calendar:{sameDay:function(){return"[hoy a la"+(1!==this.hours()?"s":"")+"] LT"},nextDay:function(){return"[mañana a la"+(1!==this.hours()?"s":"")+"] LT"},nextWeek:function(){return"dddd [a la"+(1!==this.hours()?"s":"")+"] LT"},lastDay:function(){return"[ayer a la"+(1!==this.hours()?"s":"")+"] LT"},lastWeek:function(){return"[el] dddd [pasado a la"+(1!==this.hours()?"s":"")+"] LT"},sameElse:"L"},relativeTime:{future:"en %s",past:"hace %s",s:"unos segundos",ss:"%d segundos",m:"un minuto",mm:"%d minutos",h:"una hora",hh:"%d horas",d:"un día",dd:"%d días",w:"una semana",ww:"%d semanas",M:"un mes",MM:"%d meses",y:"un año",yy:"%d años"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:0,doy:4},invalidDate:"Fecha inválida"});return o}))},b622:function(t,e,n){var i=n("da84"),r=n("5692"),o=n("5135"),a=n("90e3"),s=n("4930"),l=n("fdbf"),u=r("wks"),c=i.Symbol,d=l?c:c&&c.withoutSetter||a;t.exports=function(t){return o(u,t)||(s&&o(c,t)?u[t]=c[t]:u[t]=d("Symbol."+t)),u[t]}},b64b:function(t,e,n){var i=n("23e7"),r=n("7b0b"),o=n("df75"),a=n("d039"),s=a((function(){o(1)}));i({target:"Object",stat:!0,forced:s},{keys:function(t){return o(r(t))}})},b680:function(t,e,n){"use strict";var i=n("23e7"),r=n("a691"),o=n("408a"),a=n("1148"),s=n("d039"),l=1..toFixed,u=Math.floor,c=function(t,e,n){return 0===e?n:e%2===1?c(t,e-1,n*t):c(t*t,e/2,n)},d=function(t){var e=0,n=t;while(n>=4096)e+=12,n/=4096;while(n>=2)e+=1,n/=2;return e},h=l&&("0.000"!==8e-5.toFixed(3)||"1"!==.9.toFixed(0)||"1.25"!==1.255.toFixed(2)||"1000000000000000128"!==(0xde0b6b3a7640080).toFixed(0))||!s((function(){l.call({})}));i({target:"Number",proto:!0,forced:h},{toFixed:function(t){var e,n,i,s,l=o(this),h=r(t),f=[0,0,0,0,0,0],p="",m="0",g=function(t,e){var n=-1,i=e;while(++n<6)i+=t*f[n],f[n]=i%1e7,i=u(i/1e7)},y=function(t){var e=6,n=0;while(--e>=0)n+=f[e],f[e]=u(n/t),n=n%t*1e7},v=function(){var t=6,e="";while(--t>=0)if(""!==e||0===t||0!==f[t]){var n=String(f[t]);e=""===e?n:e+a.call("0",7-n.length)+n}return e};if(h<0||h>20)throw RangeError("Incorrect fraction digits");if(l!=l)return"NaN";if(l<=-1e21||l>=1e21)return String(l);if(l<0&&(p="-",l=-l),l>1e-21)if(e=d(l*c(2,69,1))-69,n=e<0?l*c(2,-e,1):l/c(2,e,1),n*=4503599627370496,e=52-e,e>0){g(0,n),i=h;while(i>=7)g(1e7,0),i-=7;g(c(10,i,1),0),i=e-1;while(i>=23)y(1<<23),i-=23;y(1<<i),g(1,1),y(2),m=v()}else g(0,n),g(1<<-e,0),m=v()+a.call("0",h);return h>0?(s=m.length,m=p+(s<=h?"0."+a.call("0",h-s)+m:m.slice(0,s-h)+"."+m.slice(s-h))):m=p+m,m}})},b727:function(t,e,n){var i=n("0366"),r=n("44ad"),o=n("7b0b"),a=n("50c4"),s=n("65f0"),l=[].push,u=function(t){var e=1==t,n=2==t,u=3==t,c=4==t,d=6==t,h=7==t,f=5==t||d;return function(p,m,g,y){for(var v,_,b=o(p),x=r(b),w=i(m,g,3),M=a(x.length),S=0,L=y||s,k=e?L(p,M):n||h?L(p,0):void 0;M>S;S++)if((f||S in x)&&(v=x[S],_=w(v,S,b),t))if(e)k[S]=_;else if(_)switch(t){case 3:return!0;case 5:return v;case 6:return S;case 2:l.call(k,v)}else switch(t){case 4:return!1;case 7:l.call(k,v)}return d?-1:u||c?c:k}};t.exports={forEach:u(0),map:u(1),filter:u(2),some:u(3),every:u(4),find:u(5),findIndex:u(6),filterOut:u(7)}},b760:function(t,e,n){var i=n("872a"),r=n("9638");function o(t,e,n){(void 0!==n&&!r(t[e],n)||void 0===n&&!(e in t))&&i(t,e,n)}t.exports=o},b770:function(t,e,n){"use strict";var i=t.exports={};i.forEach=function(t,e){for(var n=0;n<t.length;n++){var i=e(t[n]);if(i)return i}}},b7e9:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-sg",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},b84c:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("nn",{months:"januar_februar_mars_april_mai_juni_juli_august_september_oktober_november_desember".split("_"),monthsShort:"jan._feb._mars_apr._mai_juni_juli_aug._sep._okt._nov._des.".split("_"),monthsParseExact:!0,weekdays:"sundag_måndag_tysdag_onsdag_torsdag_fredag_laurdag".split("_"),weekdaysShort:"su._må._ty._on._to._fr._lau.".split("_"),weekdaysMin:"su_må_ty_on_to_fr_la".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY [kl.] H:mm",LLLL:"dddd D. MMMM YYYY [kl.] HH:mm"},calendar:{sameDay:"[I dag klokka] LT",nextDay:"[I morgon klokka] LT",nextWeek:"dddd [klokka] LT",lastDay:"[I går klokka] LT",lastWeek:"[Føregåande] dddd [klokka] LT",sameElse:"L"},relativeTime:{future:"om %s",past:"%s sidan",s:"nokre sekund",ss:"%d sekund",m:"eit minutt",mm:"%d minutt",h:"ein time",hh:"%d timar",d:"ein dag",dd:"%d dagar",w:"ei veke",ww:"%d veker",M:"ein månad",MM:"%d månader",y:"eit år",yy:"%d år"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},b85c:function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));n("a4d3"),n("e01a"),n("d28b"),n("d3b7"),n("3ca3"),n("ddb0");var i=n("06c5");function r(t,e){var n;if("undefined"===typeof Symbol||null==t[Symbol.iterator]){if(Array.isArray(t)||(n=Object(i["a"])(t))||e&&t&&"number"===typeof t.length){n&&(t=n);var r=0,o=function(){};return{s:o,n:function(){return r>=t.length?{done:!0}:{done:!1,value:t[r++]}},e:function(t){throw t},f:o}}throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}var a,s=!0,l=!1;return{s:function(){n=t[Symbol.iterator]()},n:function(){var t=n.next();return s=t.done,t},e:function(t){l=!0,a=t},f:function(){try{s||null==n["return"]||n["return"]()}finally{if(l)throw a}}}}},b8bb:function(t,e,n){var i=n("4d8b"),r=n("c6cf"),o=r(i);t.exports=o},b97c:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={ss:"sekundes_sekundēm_sekunde_sekundes".split("_"),m:"minūtes_minūtēm_minūte_minūtes".split("_"),mm:"minūtes_minūtēm_minūte_minūtes".split("_"),h:"stundas_stundām_stunda_stundas".split("_"),hh:"stundas_stundām_stunda_stundas".split("_"),d:"dienas_dienām_diena_dienas".split("_"),dd:"dienas_dienām_diena_dienas".split("_"),M:"mēneša_mēnešiem_mēnesis_mēneši".split("_"),MM:"mēneša_mēnešiem_mēnesis_mēneši".split("_"),y:"gada_gadiem_gads_gadi".split("_"),yy:"gada_gadiem_gads_gadi".split("_")};function n(t,e,n){return n?e%10===1&&e%100!==11?t[2]:t[3]:e%10===1&&e%100!==11?t[0]:t[1]}function i(t,i,r){return t+" "+n(e[r],t,i)}function r(t,i,r){return n(e[r],t,i)}function o(t,e){return e?"dažas sekundes":"dažām sekundēm"}var a=t.defineLocale("lv",{months:"janvāris_februāris_marts_aprīlis_maijs_jūnijs_jūlijs_augusts_septembris_oktobris_novembris_decembris".split("_"),monthsShort:"jan_feb_mar_apr_mai_jūn_jūl_aug_sep_okt_nov_dec".split("_"),weekdays:"svētdiena_pirmdiena_otrdiena_trešdiena_ceturtdiena_piektdiena_sestdiena".split("_"),weekdaysShort:"Sv_P_O_T_C_Pk_S".split("_"),weekdaysMin:"Sv_P_O_T_C_Pk_S".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY.",LL:"YYYY. [gada] D. MMMM",LLL:"YYYY. [gada] D. MMMM, HH:mm",LLLL:"YYYY. [gada] D. MMMM, dddd, HH:mm"},calendar:{sameDay:"[Šodien pulksten] LT",nextDay:"[Rīt pulksten] LT",nextWeek:"dddd [pulksten] LT",lastDay:"[Vakar pulksten] LT",lastWeek:"[Pagājušā] dddd [pulksten] LT",sameElse:"L"},relativeTime:{future:"pēc %s",past:"pirms %s",s:o,ss:i,m:r,mm:i,h:r,hh:i,d:r,dd:i,M:r,MM:i,y:r,yy:i},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return a}))},badf:function(t,e,n){var i=n("642a"),r=n("1838"),o=n("cd9d"),a=n("6747"),s=n("f9ce");function l(t){return"function"==typeof t?t:null==t?o:"object"==typeof t?a(t)?r(t[0],t[1]):i(t):s(t)}t.exports=l},bb71:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r={m:["eine Minute","einer Minute"],h:["eine Stunde","einer Stunde"],d:["ein Tag","einem Tag"],dd:[t+" Tage",t+" Tagen"],w:["eine Woche","einer Woche"],M:["ein Monat","einem Monat"],MM:[t+" Monate",t+" Monaten"],y:["ein Jahr","einem Jahr"],yy:[t+" Jahre",t+" Jahren"]};return e?r[n][0]:r[n][1]}var n=t.defineLocale("de-ch",{months:"Januar_Februar_März_April_Mai_Juni_Juli_August_September_Oktober_November_Dezember".split("_"),monthsShort:"Jan._Feb._März_Apr._Mai_Juni_Juli_Aug._Sep._Okt._Nov._Dez.".split("_"),monthsParseExact:!0,weekdays:"Sonntag_Montag_Dienstag_Mittwoch_Donnerstag_Freitag_Samstag".split("_"),weekdaysShort:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysMin:"So_Mo_Di_Mi_Do_Fr_Sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY HH:mm",LLLL:"dddd, D. MMMM YYYY HH:mm"},calendar:{sameDay:"[heute um] LT [Uhr]",sameElse:"L",nextDay:"[morgen um] LT [Uhr]",nextWeek:"dddd [um] LT [Uhr]",lastDay:"[gestern um] LT [Uhr]",lastWeek:"[letzten] dddd [um] LT [Uhr]"},relativeTime:{future:"in %s",past:"vor %s",s:"ein paar Sekunden",ss:"%d Sekunden",m:e,mm:"%d Minuten",h:e,hh:"%d Stunden",d:e,dd:e,w:e,ww:"%d Wochen",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},bbc0:function(t,e,n){var i=n("6044"),r="__lodash_hash_undefined__",o=Object.prototype,a=o.hasOwnProperty;function s(t){var e=this.__data__;if(i){var n=e[t];return n===r?void 0:n}return a.call(e,t)?e[t]:void 0}t.exports=s},bc3a:function(t,e,n){t.exports=n("cee4")},c04e:function(t,e,n){var i=n("861d");t.exports=function(t,e){if(!i(t))return t;var n,r;if(e&&"function"==typeof(n=t.toString)&&!i(r=n.call(t)))return r;if("function"==typeof(n=t.valueOf)&&!i(r=n.call(t)))return r;if(!e&&"function"==typeof(n=t.toString)&&!i(r=n.call(t)))return r;throw TypeError("Can't convert object to primitive value")}},c05f:function(t,e,n){var i=n("7b97"),r=n("1310");function o(t,e,n,a,s){return t===e||(null==t||null==e||!r(t)&&!r(e)?t!==t&&e!==e:i(t,e,n,a,o,s))}t.exports=o},c098:function(t,e){var n=9007199254740991,i=/^(?:0|[1-9]\d*)$/;function r(t,e){var r=typeof t;return e=null==e?n:e,!!e&&("number"==r||"symbol"!=r&&i.test(t))&&t>-1&&t%1==0&&t<e}t.exports=r},c109:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tzm",{months:"ⵉⵏⵏⴰⵢⵔ_ⴱⵕⴰⵢⵕ_ⵎⴰⵕⵚ_ⵉⴱⵔⵉⵔ_ⵎⴰⵢⵢⵓ_ⵢⵓⵏⵢⵓ_ⵢⵓⵍⵢⵓⵣ_ⵖⵓⵛⵜ_ⵛⵓⵜⴰⵏⴱⵉⵔ_ⴽⵟⵓⴱⵕ_ⵏⵓⵡⴰⵏⴱⵉⵔ_ⴷⵓⵊⵏⴱⵉⵔ".split("_"),monthsShort:"ⵉⵏⵏⴰⵢⵔ_ⴱⵕⴰⵢⵕ_ⵎⴰⵕⵚ_ⵉⴱⵔⵉⵔ_ⵎⴰⵢⵢⵓ_ⵢⵓⵏⵢⵓ_ⵢⵓⵍⵢⵓⵣ_ⵖⵓⵛⵜ_ⵛⵓⵜⴰⵏⴱⵉⵔ_ⴽⵟⵓⴱⵕ_ⵏⵓⵡⴰⵏⴱⵉⵔ_ⴷⵓⵊⵏⴱⵉⵔ".split("_"),weekdays:"ⴰⵙⴰⵎⴰⵙ_ⴰⵢⵏⴰⵙ_ⴰⵙⵉⵏⴰⵙ_ⴰⴽⵔⴰⵙ_ⴰⴽⵡⴰⵙ_ⴰⵙⵉⵎⵡⴰⵙ_ⴰⵙⵉⴹⵢⴰⵙ".split("_"),weekdaysShort:"ⴰⵙⴰⵎⴰⵙ_ⴰⵢⵏⴰⵙ_ⴰⵙⵉⵏⴰⵙ_ⴰⴽⵔⴰⵙ_ⴰⴽⵡⴰⵙ_ⴰⵙⵉⵎⵡⴰⵙ_ⴰⵙⵉⴹⵢⴰⵙ".split("_"),weekdaysMin:"ⴰⵙⴰⵎⴰⵙ_ⴰⵢⵏⴰⵙ_ⴰⵙⵉⵏⴰⵙ_ⴰⴽⵔⴰⵙ_ⴰⴽⵡⴰⵙ_ⴰⵙⵉⵎⵡⴰⵙ_ⴰⵙⵉⴹⵢⴰⵙ".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[ⴰⵙⴷⵅ ⴴ] LT",nextDay:"[ⴰⵙⴽⴰ ⴴ] LT",nextWeek:"dddd [ⴴ] LT",lastDay:"[ⴰⵚⴰⵏⵜ ⴴ] LT",lastWeek:"dddd [ⴴ] LT",sameElse:"L"},relativeTime:{future:"ⴷⴰⴷⵅ ⵙ ⵢⴰⵏ %s",past:"ⵢⴰⵏ %s",s:"ⵉⵎⵉⴽ",ss:"%d ⵉⵎⵉⴽ",m:"ⵎⵉⵏⵓⴺ",mm:"%d ⵎⵉⵏⵓⴺ",h:"ⵙⴰⵄⴰ",hh:"%d ⵜⴰⵙⵙⴰⵄⵉⵏ",d:"ⴰⵙⵙ",dd:"%d oⵙⵙⴰⵏ",M:"ⴰⵢoⵓⵔ",MM:"%d ⵉⵢⵢⵉⵔⵏ",y:"ⴰⵙⴳⴰⵙ",yy:"%d ⵉⵙⴳⴰⵙⵏ"},week:{dow:6,doy:12}});return e}))},c177:function(t,e,n){"use strict";n.d(e,"b",(function(){return i})),n.d(e,"a",(function(){return r}));let i=0;function r(){i++}},c1c9:function(t,e,n){var i=n("a454"),r=n("f3c1"),o=r(i);t.exports=o},c1df:function(t,e,n){(function(t){var e;//! moment.js
//! version : 2.29.1
//! authors : Tim Wood, Iskren Chernev, Moment.js contributors
//! license : MIT
//! momentjs.com
(function(e,n){t.exports=n()})(0,(function(){"use strict";var i,r;function o(){return i.apply(null,arguments)}function a(t){i=t}function s(t){return t instanceof Array||"[object Array]"===Object.prototype.toString.call(t)}function l(t){return null!=t&&"[object Object]"===Object.prototype.toString.call(t)}function u(t,e){return Object.prototype.hasOwnProperty.call(t,e)}function c(t){if(Object.getOwnPropertyNames)return 0===Object.getOwnPropertyNames(t).length;var e;for(e in t)if(u(t,e))return!1;return!0}function d(t){return void 0===t}function h(t){return"number"===typeof t||"[object Number]"===Object.prototype.toString.call(t)}function f(t){return t instanceof Date||"[object Date]"===Object.prototype.toString.call(t)}function p(t,e){var n,i=[];for(n=0;n<t.length;++n)i.push(e(t[n],n));return i}function m(t,e){for(var n in e)u(e,n)&&(t[n]=e[n]);return u(e,"toString")&&(t.toString=e.toString),u(e,"valueOf")&&(t.valueOf=e.valueOf),t}function g(t,e,n,i){return Xn(t,e,n,i,!0).utc()}function y(){return{empty:!1,unusedTokens:[],unusedInput:[],overflow:-2,charsLeftOver:0,nullInput:!1,invalidEra:null,invalidMonth:null,invalidFormat:!1,userInvalidated:!1,iso:!1,parsedDateParts:[],era:null,meridiem:null,rfc2822:!1,weekdayMismatch:!1}}function v(t){return null==t._pf&&(t._pf=y()),t._pf}function _(t){if(null==t._isValid){var e=v(t),n=r.call(e.parsedDateParts,(function(t){return null!=t})),i=!isNaN(t._d.getTime())&&e.overflow<0&&!e.empty&&!e.invalidEra&&!e.invalidMonth&&!e.invalidWeekday&&!e.weekdayMismatch&&!e.nullInput&&!e.invalidFormat&&!e.userInvalidated&&(!e.meridiem||e.meridiem&&n);if(t._strict&&(i=i&&0===e.charsLeftOver&&0===e.unusedTokens.length&&void 0===e.bigHour),null!=Object.isFrozen&&Object.isFrozen(t))return i;t._isValid=i}return t._isValid}function b(t){var e=g(NaN);return null!=t?m(v(e),t):v(e).userInvalidated=!0,e}r=Array.prototype.some?Array.prototype.some:function(t){var e,n=Object(this),i=n.length>>>0;for(e=0;e<i;e++)if(e in n&&t.call(this,n[e],e,n))return!0;return!1};var x=o.momentProperties=[],w=!1;function M(t,e){var n,i,r;if(d(e._isAMomentObject)||(t._isAMomentObject=e._isAMomentObject),d(e._i)||(t._i=e._i),d(e._f)||(t._f=e._f),d(e._l)||(t._l=e._l),d(e._strict)||(t._strict=e._strict),d(e._tzm)||(t._tzm=e._tzm),d(e._isUTC)||(t._isUTC=e._isUTC),d(e._offset)||(t._offset=e._offset),d(e._pf)||(t._pf=v(e)),d(e._locale)||(t._locale=e._locale),x.length>0)for(n=0;n<x.length;n++)i=x[n],r=e[i],d(r)||(t[i]=r);return t}function S(t){M(this,t),this._d=new Date(null!=t._d?t._d.getTime():NaN),this.isValid()||(this._d=new Date(NaN)),!1===w&&(w=!0,o.updateOffset(this),w=!1)}function L(t){return t instanceof S||null!=t&&null!=t._isAMomentObject}function k(t){!1===o.suppressDeprecationWarnings&&"undefined"!==typeof console&&console.warn}function T(t,e){var n=!0;return m((function(){if(null!=o.deprecationHandler&&o.deprecationHandler(null,t),n){var i,r,a,s=[];for(r=0;r<arguments.length;r++){if(i="","object"===typeof arguments[r]){for(a in i+="\n["+r+"] ",arguments[0])u(arguments[0],a)&&(i+=a+": "+arguments[0][a]+", ");i=i.slice(0,-2)}else i=arguments[r];s.push(i)}k(t+"\nArguments: "+Array.prototype.slice.call(s).join("")+"\n"+(new Error).stack),n=!1}return e.apply(this,arguments)}),e)}var D,C={};function I(t,e){null!=o.deprecationHandler&&o.deprecationHandler(t,e),C[t]||(k(e),C[t]=!0)}function A(t){return"undefined"!==typeof Function&&t instanceof Function||"[object Function]"===Object.prototype.toString.call(t)}function Y(t){var e,n;for(n in t)u(t,n)&&(e=t[n],A(e)?this[n]=e:this["_"+n]=e);this._config=t,this._dayOfMonthOrdinalParseLenient=new RegExp((this._dayOfMonthOrdinalParse.source||this._ordinalParse.source)+"|"+/\d{1,2}/.source)}function O(t,e){var n,i=m({},t);for(n in e)u(e,n)&&(l(t[n])&&l(e[n])?(i[n]={},m(i[n],t[n]),m(i[n],e[n])):null!=e[n]?i[n]=e[n]:delete i[n]);for(n in t)u(t,n)&&!u(e,n)&&l(t[n])&&(i[n]=m({},i[n]));return i}function P(t){null!=t&&this.set(t)}o.suppressDeprecationWarnings=!1,o.deprecationHandler=null,D=Object.keys?Object.keys:function(t){var e,n=[];for(e in t)u(t,e)&&n.push(e);return n};var E={sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"};function N(t,e,n){var i=this._calendar[t]||this._calendar["sameElse"];return A(i)?i.call(e,n):i}function R(t,e,n){var i=""+Math.abs(t),r=e-i.length,o=t>=0;return(o?n?"+":"":"-")+Math.pow(10,Math.max(0,r)).toString().substr(1)+i}var H=/(\[[^\[]*\])|(\\)?([Hh]mm(ss)?|Mo|MM?M?M?|Do|DDDo|DD?D?D?|ddd?d?|do?|w[o|w]?|W[o|W]?|Qo?|N{1,5}|YYYYYY|YYYYY|YYYY|YY|y{2,4}|yo?|gg(ggg?)?|GG(GGG?)?|e|E|a|A|hh?|HH?|kk?|mm?|ss?|S{1,9}|x|X|zz?|ZZ?|.)/g,j=/(\[[^\[]*\])|(\\)?(LTS|LT|LL?L?L?|l{1,4})/g,z={},W={};function B(t,e,n,i){var r=i;"string"===typeof i&&(r=function(){return this[i]()}),t&&(W[t]=r),e&&(W[e[0]]=function(){return R(r.apply(this,arguments),e[1],e[2])}),n&&(W[n]=function(){return this.localeData().ordinal(r.apply(this,arguments),t)})}function F(t){return t.match(/\[[\s\S]/)?t.replace(/^\[|\]$/g,""):t.replace(/\\/g,"")}function V(t){var e,n,i=t.match(H);for(e=0,n=i.length;e<n;e++)W[i[e]]?i[e]=W[i[e]]:i[e]=F(i[e]);return function(e){var r,o="";for(r=0;r<n;r++)o+=A(i[r])?i[r].call(e,t):i[r];return o}}function G(t,e){return t.isValid()?(e=$(e,t.localeData()),z[e]=z[e]||V(e),z[e](t)):t.localeData().invalidDate()}function $(t,e){var n=5;function i(t){return e.longDateFormat(t)||t}j.lastIndex=0;while(n>=0&&j.test(t))t=t.replace(j,i),j.lastIndex=0,n-=1;return t}var U={LTS:"h:mm:ss A",LT:"h:mm A",L:"MM/DD/YYYY",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY h:mm A",LLLL:"dddd, MMMM D, YYYY h:mm A"};function q(t){var e=this._longDateFormat[t],n=this._longDateFormat[t.toUpperCase()];return e||!n?e:(this._longDateFormat[t]=n.match(H).map((function(t){return"MMMM"===t||"MM"===t||"DD"===t||"dddd"===t?t.slice(1):t})).join(""),this._longDateFormat[t])}var X="Invalid date";function Z(){return this._invalidDate}var J="%d",K=/\d{1,2}/;function Q(t){return this._ordinal.replace("%d",t)}var tt={future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",w:"a week",ww:"%d weeks",M:"a month",MM:"%d months",y:"a year",yy:"%d years"};function et(t,e,n,i){var r=this._relativeTime[n];return A(r)?r(t,e,n,i):r.replace(/%d/i,t)}function nt(t,e){var n=this._relativeTime[t>0?"future":"past"];return A(n)?n(e):n.replace(/%s/i,e)}var it={};function rt(t,e){var n=t.toLowerCase();it[n]=it[n+"s"]=it[e]=t}function ot(t){return"string"===typeof t?it[t]||it[t.toLowerCase()]:void 0}function at(t){var e,n,i={};for(n in t)u(t,n)&&(e=ot(n),e&&(i[e]=t[n]));return i}var st={};function lt(t,e){st[t]=e}function ut(t){var e,n=[];for(e in t)u(t,e)&&n.push({unit:e,priority:st[e]});return n.sort((function(t,e){return t.priority-e.priority})),n}function ct(t){return t%4===0&&t%100!==0||t%400===0}function dt(t){return t<0?Math.ceil(t)||0:Math.floor(t)}function ht(t){var e=+t,n=0;return 0!==e&&isFinite(e)&&(n=dt(e)),n}function ft(t,e){return function(n){return null!=n?(mt(this,t,n),o.updateOffset(this,e),this):pt(this,t)}}function pt(t,e){return t.isValid()?t._d["get"+(t._isUTC?"UTC":"")+e]():NaN}function mt(t,e,n){t.isValid()&&!isNaN(n)&&("FullYear"===e&&ct(t.year())&&1===t.month()&&29===t.date()?(n=ht(n),t._d["set"+(t._isUTC?"UTC":"")+e](n,t.month(),ee(n,t.month()))):t._d["set"+(t._isUTC?"UTC":"")+e](n))}function gt(t){return t=ot(t),A(this[t])?this[t]():this}function yt(t,e){if("object"===typeof t){t=at(t);var n,i=ut(t);for(n=0;n<i.length;n++)this[i[n].unit](t[i[n].unit])}else if(t=ot(t),A(this[t]))return this[t](e);return this}var vt,_t=/\d/,bt=/\d\d/,xt=/\d{3}/,wt=/\d{4}/,Mt=/[+-]?\d{6}/,St=/\d\d?/,Lt=/\d\d\d\d?/,kt=/\d\d\d\d\d\d?/,Tt=/\d{1,3}/,Dt=/\d{1,4}/,Ct=/[+-]?\d{1,6}/,It=/\d+/,At=/[+-]?\d+/,Yt=/Z|[+-]\d\d:?\d\d/gi,Ot=/Z|[+-]\d\d(?::?\d\d)?/gi,Pt=/[+-]?\d+(\.\d{1,3})?/,Et=/[0-9]{0,256}['a-z\u00A0-\u05FF\u0700-\uD7FF\uF900-\uFDCF\uFDF0-\uFF07\uFF10-\uFFEF]{1,256}|[\u0600-\u06FF\/]{1,256}(\s*?[\u0600-\u06FF]{1,256}){1,2}/i;function Nt(t,e,n){vt[t]=A(e)?e:function(t,i){return t&&n?n:e}}function Rt(t,e){return u(vt,t)?vt[t](e._strict,e._locale):new RegExp(Ht(t))}function Ht(t){return jt(t.replace("\\","").replace(/\\(\[)|\\(\])|\[([^\]\[]*)\]|\\(.)/g,(function(t,e,n,i,r){return e||n||i||r})))}function jt(t){return t.replace(/[-\/\\^$*+?.()|[\]{}]/g,"\\$&")}vt={};var zt={};function Wt(t,e){var n,i=e;for("string"===typeof t&&(t=[t]),h(e)&&(i=function(t,n){n[e]=ht(t)}),n=0;n<t.length;n++)zt[t[n]]=i}function Bt(t,e){Wt(t,(function(t,n,i,r){i._w=i._w||{},e(t,i._w,i,r)}))}function Ft(t,e,n){null!=e&&u(zt,t)&&zt[t](e,n._a,n,t)}var Vt,Gt=0,$t=1,Ut=2,qt=3,Xt=4,Zt=5,Jt=6,Kt=7,Qt=8;function te(t,e){return(t%e+e)%e}function ee(t,e){if(isNaN(t)||isNaN(e))return NaN;var n=te(e,12);return t+=(e-n)/12,1===n?ct(t)?29:28:31-n%7%2}Vt=Array.prototype.indexOf?Array.prototype.indexOf:function(t){var e;for(e=0;e<this.length;++e)if(this[e]===t)return e;return-1},B("M",["MM",2],"Mo",(function(){return this.month()+1})),B("MMM",0,0,(function(t){return this.localeData().monthsShort(this,t)})),B("MMMM",0,0,(function(t){return this.localeData().months(this,t)})),rt("month","M"),lt("month",8),Nt("M",St),Nt("MM",St,bt),Nt("MMM",(function(t,e){return e.monthsShortRegex(t)})),Nt("MMMM",(function(t,e){return e.monthsRegex(t)})),Wt(["M","MM"],(function(t,e){e[$t]=ht(t)-1})),Wt(["MMM","MMMM"],(function(t,e,n,i){var r=n._locale.monthsParse(t,i,n._strict);null!=r?e[$t]=r:v(n).invalidMonth=t}));var ne="January_February_March_April_May_June_July_August_September_October_November_December".split("_"),ie="Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),re=/D[oD]?(\[[^\[\]]*\]|\s)+MMMM?/,oe=Et,ae=Et;function se(t,e){return t?s(this._months)?this._months[t.month()]:this._months[(this._months.isFormat||re).test(e)?"format":"standalone"][t.month()]:s(this._months)?this._months:this._months["standalone"]}function le(t,e){return t?s(this._monthsShort)?this._monthsShort[t.month()]:this._monthsShort[re.test(e)?"format":"standalone"][t.month()]:s(this._monthsShort)?this._monthsShort:this._monthsShort["standalone"]}function ue(t,e,n){var i,r,o,a=t.toLocaleLowerCase();if(!this._monthsParse)for(this._monthsParse=[],this._longMonthsParse=[],this._shortMonthsParse=[],i=0;i<12;++i)o=g([2e3,i]),this._shortMonthsParse[i]=this.monthsShort(o,"").toLocaleLowerCase(),this._longMonthsParse[i]=this.months(o,"").toLocaleLowerCase();return n?"MMM"===e?(r=Vt.call(this._shortMonthsParse,a),-1!==r?r:null):(r=Vt.call(this._longMonthsParse,a),-1!==r?r:null):"MMM"===e?(r=Vt.call(this._shortMonthsParse,a),-1!==r?r:(r=Vt.call(this._longMonthsParse,a),-1!==r?r:null)):(r=Vt.call(this._longMonthsParse,a),-1!==r?r:(r=Vt.call(this._shortMonthsParse,a),-1!==r?r:null))}function ce(t,e,n){var i,r,o;if(this._monthsParseExact)return ue.call(this,t,e,n);for(this._monthsParse||(this._monthsParse=[],this._longMonthsParse=[],this._shortMonthsParse=[]),i=0;i<12;i++){if(r=g([2e3,i]),n&&!this._longMonthsParse[i]&&(this._longMonthsParse[i]=new RegExp("^"+this.months(r,"").replace(".","")+"$","i"),this._shortMonthsParse[i]=new RegExp("^"+this.monthsShort(r,"").replace(".","")+"$","i")),n||this._monthsParse[i]||(o="^"+this.months(r,"")+"|^"+this.monthsShort(r,""),this._monthsParse[i]=new RegExp(o.replace(".",""),"i")),n&&"MMMM"===e&&this._longMonthsParse[i].test(t))return i;if(n&&"MMM"===e&&this._shortMonthsParse[i].test(t))return i;if(!n&&this._monthsParse[i].test(t))return i}}function de(t,e){var n;if(!t.isValid())return t;if("string"===typeof e)if(/^\d+$/.test(e))e=ht(e);else if(e=t.localeData().monthsParse(e),!h(e))return t;return n=Math.min(t.date(),ee(t.year(),e)),t._d["set"+(t._isUTC?"UTC":"")+"Month"](e,n),t}function he(t){return null!=t?(de(this,t),o.updateOffset(this,!0),this):pt(this,"Month")}function fe(){return ee(this.year(),this.month())}function pe(t){return this._monthsParseExact?(u(this,"_monthsRegex")||ge.call(this),t?this._monthsShortStrictRegex:this._monthsShortRegex):(u(this,"_monthsShortRegex")||(this._monthsShortRegex=oe),this._monthsShortStrictRegex&&t?this._monthsShortStrictRegex:this._monthsShortRegex)}function me(t){return this._monthsParseExact?(u(this,"_monthsRegex")||ge.call(this),t?this._monthsStrictRegex:this._monthsRegex):(u(this,"_monthsRegex")||(this._monthsRegex=ae),this._monthsStrictRegex&&t?this._monthsStrictRegex:this._monthsRegex)}function ge(){function t(t,e){return e.length-t.length}var e,n,i=[],r=[],o=[];for(e=0;e<12;e++)n=g([2e3,e]),i.push(this.monthsShort(n,"")),r.push(this.months(n,"")),o.push(this.months(n,"")),o.push(this.monthsShort(n,""));for(i.sort(t),r.sort(t),o.sort(t),e=0;e<12;e++)i[e]=jt(i[e]),r[e]=jt(r[e]);for(e=0;e<24;e++)o[e]=jt(o[e]);this._monthsRegex=new RegExp("^("+o.join("|")+")","i"),this._monthsShortRegex=this._monthsRegex,this._monthsStrictRegex=new RegExp("^("+r.join("|")+")","i"),this._monthsShortStrictRegex=new RegExp("^("+i.join("|")+")","i")}function ye(t){return ct(t)?366:365}B("Y",0,0,(function(){var t=this.year();return t<=9999?R(t,4):"+"+t})),B(0,["YY",2],0,(function(){return this.year()%100})),B(0,["YYYY",4],0,"year"),B(0,["YYYYY",5],0,"year"),B(0,["YYYYYY",6,!0],0,"year"),rt("year","y"),lt("year",1),Nt("Y",At),Nt("YY",St,bt),Nt("YYYY",Dt,wt),Nt("YYYYY",Ct,Mt),Nt("YYYYYY",Ct,Mt),Wt(["YYYYY","YYYYYY"],Gt),Wt("YYYY",(function(t,e){e[Gt]=2===t.length?o.parseTwoDigitYear(t):ht(t)})),Wt("YY",(function(t,e){e[Gt]=o.parseTwoDigitYear(t)})),Wt("Y",(function(t,e){e[Gt]=parseInt(t,10)})),o.parseTwoDigitYear=function(t){return ht(t)+(ht(t)>68?1900:2e3)};var ve=ft("FullYear",!0);function _e(){return ct(this.year())}function be(t,e,n,i,r,o,a){var s;return t<100&&t>=0?(s=new Date(t+400,e,n,i,r,o,a),isFinite(s.getFullYear())&&s.setFullYear(t)):s=new Date(t,e,n,i,r,o,a),s}function xe(t){var e,n;return t<100&&t>=0?(n=Array.prototype.slice.call(arguments),n[0]=t+400,e=new Date(Date.UTC.apply(null,n)),isFinite(e.getUTCFullYear())&&e.setUTCFullYear(t)):e=new Date(Date.UTC.apply(null,arguments)),e}function we(t,e,n){var i=7+e-n,r=(7+xe(t,0,i).getUTCDay()-e)%7;return-r+i-1}function Me(t,e,n,i,r){var o,a,s=(7+n-i)%7,l=we(t,i,r),u=1+7*(e-1)+s+l;return u<=0?(o=t-1,a=ye(o)+u):u>ye(t)?(o=t+1,a=u-ye(t)):(o=t,a=u),{year:o,dayOfYear:a}}function Se(t,e,n){var i,r,o=we(t.year(),e,n),a=Math.floor((t.dayOfYear()-o-1)/7)+1;return a<1?(r=t.year()-1,i=a+Le(r,e,n)):a>Le(t.year(),e,n)?(i=a-Le(t.year(),e,n),r=t.year()+1):(r=t.year(),i=a),{week:i,year:r}}function Le(t,e,n){var i=we(t,e,n),r=we(t+1,e,n);return(ye(t)-i+r)/7}function ke(t){return Se(t,this._week.dow,this._week.doy).week}B("w",["ww",2],"wo","week"),B("W",["WW",2],"Wo","isoWeek"),rt("week","w"),rt("isoWeek","W"),lt("week",5),lt("isoWeek",5),Nt("w",St),Nt("ww",St,bt),Nt("W",St),Nt("WW",St,bt),Bt(["w","ww","W","WW"],(function(t,e,n,i){e[i.substr(0,1)]=ht(t)}));var Te={dow:0,doy:6};function De(){return this._week.dow}function Ce(){return this._week.doy}function Ie(t){var e=this.localeData().week(this);return null==t?e:this.add(7*(t-e),"d")}function Ae(t){var e=Se(this,1,4).week;return null==t?e:this.add(7*(t-e),"d")}function Ye(t,e){return"string"!==typeof t?t:isNaN(t)?(t=e.weekdaysParse(t),"number"===typeof t?t:null):parseInt(t,10)}function Oe(t,e){return"string"===typeof t?e.weekdaysParse(t)%7||7:isNaN(t)?null:t}function Pe(t,e){return t.slice(e,7).concat(t.slice(0,e))}B("d",0,"do","day"),B("dd",0,0,(function(t){return this.localeData().weekdaysMin(this,t)})),B("ddd",0,0,(function(t){return this.localeData().weekdaysShort(this,t)})),B("dddd",0,0,(function(t){return this.localeData().weekdays(this,t)})),B("e",0,0,"weekday"),B("E",0,0,"isoWeekday"),rt("day","d"),rt("weekday","e"),rt("isoWeekday","E"),lt("day",11),lt("weekday",11),lt("isoWeekday",11),Nt("d",St),Nt("e",St),Nt("E",St),Nt("dd",(function(t,e){return e.weekdaysMinRegex(t)})),Nt("ddd",(function(t,e){return e.weekdaysShortRegex(t)})),Nt("dddd",(function(t,e){return e.weekdaysRegex(t)})),Bt(["dd","ddd","dddd"],(function(t,e,n,i){var r=n._locale.weekdaysParse(t,i,n._strict);null!=r?e.d=r:v(n).invalidWeekday=t})),Bt(["d","e","E"],(function(t,e,n,i){e[i]=ht(t)}));var Ee="Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),Ne="Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),Re="Su_Mo_Tu_We_Th_Fr_Sa".split("_"),He=Et,je=Et,ze=Et;function We(t,e){var n=s(this._weekdays)?this._weekdays:this._weekdays[t&&!0!==t&&this._weekdays.isFormat.test(e)?"format":"standalone"];return!0===t?Pe(n,this._week.dow):t?n[t.day()]:n}function Be(t){return!0===t?Pe(this._weekdaysShort,this._week.dow):t?this._weekdaysShort[t.day()]:this._weekdaysShort}function Fe(t){return!0===t?Pe(this._weekdaysMin,this._week.dow):t?this._weekdaysMin[t.day()]:this._weekdaysMin}function Ve(t,e,n){var i,r,o,a=t.toLocaleLowerCase();if(!this._weekdaysParse)for(this._weekdaysParse=[],this._shortWeekdaysParse=[],this._minWeekdaysParse=[],i=0;i<7;++i)o=g([2e3,1]).day(i),this._minWeekdaysParse[i]=this.weekdaysMin(o,"").toLocaleLowerCase(),this._shortWeekdaysParse[i]=this.weekdaysShort(o,"").toLocaleLowerCase(),this._weekdaysParse[i]=this.weekdays(o,"").toLocaleLowerCase();return n?"dddd"===e?(r=Vt.call(this._weekdaysParse,a),-1!==r?r:null):"ddd"===e?(r=Vt.call(this._shortWeekdaysParse,a),-1!==r?r:null):(r=Vt.call(this._minWeekdaysParse,a),-1!==r?r:null):"dddd"===e?(r=Vt.call(this._weekdaysParse,a),-1!==r?r:(r=Vt.call(this._shortWeekdaysParse,a),-1!==r?r:(r=Vt.call(this._minWeekdaysParse,a),-1!==r?r:null))):"ddd"===e?(r=Vt.call(this._shortWeekdaysParse,a),-1!==r?r:(r=Vt.call(this._weekdaysParse,a),-1!==r?r:(r=Vt.call(this._minWeekdaysParse,a),-1!==r?r:null))):(r=Vt.call(this._minWeekdaysParse,a),-1!==r?r:(r=Vt.call(this._weekdaysParse,a),-1!==r?r:(r=Vt.call(this._shortWeekdaysParse,a),-1!==r?r:null)))}function Ge(t,e,n){var i,r,o;if(this._weekdaysParseExact)return Ve.call(this,t,e,n);for(this._weekdaysParse||(this._weekdaysParse=[],this._minWeekdaysParse=[],this._shortWeekdaysParse=[],this._fullWeekdaysParse=[]),i=0;i<7;i++){if(r=g([2e3,1]).day(i),n&&!this._fullWeekdaysParse[i]&&(this._fullWeekdaysParse[i]=new RegExp("^"+this.weekdays(r,"").replace(".","\\.?")+"$","i"),this._shortWeekdaysParse[i]=new RegExp("^"+this.weekdaysShort(r,"").replace(".","\\.?")+"$","i"),this._minWeekdaysParse[i]=new RegExp("^"+this.weekdaysMin(r,"").replace(".","\\.?")+"$","i")),this._weekdaysParse[i]||(o="^"+this.weekdays(r,"")+"|^"+this.weekdaysShort(r,"")+"|^"+this.weekdaysMin(r,""),this._weekdaysParse[i]=new RegExp(o.replace(".",""),"i")),n&&"dddd"===e&&this._fullWeekdaysParse[i].test(t))return i;if(n&&"ddd"===e&&this._shortWeekdaysParse[i].test(t))return i;if(n&&"dd"===e&&this._minWeekdaysParse[i].test(t))return i;if(!n&&this._weekdaysParse[i].test(t))return i}}function $e(t){if(!this.isValid())return null!=t?this:NaN;var e=this._isUTC?this._d.getUTCDay():this._d.getDay();return null!=t?(t=Ye(t,this.localeData()),this.add(t-e,"d")):e}function Ue(t){if(!this.isValid())return null!=t?this:NaN;var e=(this.day()+7-this.localeData()._week.dow)%7;return null==t?e:this.add(t-e,"d")}function qe(t){if(!this.isValid())return null!=t?this:NaN;if(null!=t){var e=Oe(t,this.localeData());return this.day(this.day()%7?e:e-7)}return this.day()||7}function Xe(t){return this._weekdaysParseExact?(u(this,"_weekdaysRegex")||Ke.call(this),t?this._weekdaysStrictRegex:this._weekdaysRegex):(u(this,"_weekdaysRegex")||(this._weekdaysRegex=He),this._weekdaysStrictRegex&&t?this._weekdaysStrictRegex:this._weekdaysRegex)}function Ze(t){return this._weekdaysParseExact?(u(this,"_weekdaysRegex")||Ke.call(this),t?this._weekdaysShortStrictRegex:this._weekdaysShortRegex):(u(this,"_weekdaysShortRegex")||(this._weekdaysShortRegex=je),this._weekdaysShortStrictRegex&&t?this._weekdaysShortStrictRegex:this._weekdaysShortRegex)}function Je(t){return this._weekdaysParseExact?(u(this,"_weekdaysRegex")||Ke.call(this),t?this._weekdaysMinStrictRegex:this._weekdaysMinRegex):(u(this,"_weekdaysMinRegex")||(this._weekdaysMinRegex=ze),this._weekdaysMinStrictRegex&&t?this._weekdaysMinStrictRegex:this._weekdaysMinRegex)}function Ke(){function t(t,e){return e.length-t.length}var e,n,i,r,o,a=[],s=[],l=[],u=[];for(e=0;e<7;e++)n=g([2e3,1]).day(e),i=jt(this.weekdaysMin(n,"")),r=jt(this.weekdaysShort(n,"")),o=jt(this.weekdays(n,"")),a.push(i),s.push(r),l.push(o),u.push(i),u.push(r),u.push(o);a.sort(t),s.sort(t),l.sort(t),u.sort(t),this._weekdaysRegex=new RegExp("^("+u.join("|")+")","i"),this._weekdaysShortRegex=this._weekdaysRegex,this._weekdaysMinRegex=this._weekdaysRegex,this._weekdaysStrictRegex=new RegExp("^("+l.join("|")+")","i"),this._weekdaysShortStrictRegex=new RegExp("^("+s.join("|")+")","i"),this._weekdaysMinStrictRegex=new RegExp("^("+a.join("|")+")","i")}function Qe(){return this.hours()%12||12}function tn(){return this.hours()||24}function en(t,e){B(t,0,0,(function(){return this.localeData().meridiem(this.hours(),this.minutes(),e)}))}function nn(t,e){return e._meridiemParse}function rn(t){return"p"===(t+"").toLowerCase().charAt(0)}B("H",["HH",2],0,"hour"),B("h",["hh",2],0,Qe),B("k",["kk",2],0,tn),B("hmm",0,0,(function(){return""+Qe.apply(this)+R(this.minutes(),2)})),B("hmmss",0,0,(function(){return""+Qe.apply(this)+R(this.minutes(),2)+R(this.seconds(),2)})),B("Hmm",0,0,(function(){return""+this.hours()+R(this.minutes(),2)})),B("Hmmss",0,0,(function(){return""+this.hours()+R(this.minutes(),2)+R(this.seconds(),2)})),en("a",!0),en("A",!1),rt("hour","h"),lt("hour",13),Nt("a",nn),Nt("A",nn),Nt("H",St),Nt("h",St),Nt("k",St),Nt("HH",St,bt),Nt("hh",St,bt),Nt("kk",St,bt),Nt("hmm",Lt),Nt("hmmss",kt),Nt("Hmm",Lt),Nt("Hmmss",kt),Wt(["H","HH"],qt),Wt(["k","kk"],(function(t,e,n){var i=ht(t);e[qt]=24===i?0:i})),Wt(["a","A"],(function(t,e,n){n._isPm=n._locale.isPM(t),n._meridiem=t})),Wt(["h","hh"],(function(t,e,n){e[qt]=ht(t),v(n).bigHour=!0})),Wt("hmm",(function(t,e,n){var i=t.length-2;e[qt]=ht(t.substr(0,i)),e[Xt]=ht(t.substr(i)),v(n).bigHour=!0})),Wt("hmmss",(function(t,e,n){var i=t.length-4,r=t.length-2;e[qt]=ht(t.substr(0,i)),e[Xt]=ht(t.substr(i,2)),e[Zt]=ht(t.substr(r)),v(n).bigHour=!0})),Wt("Hmm",(function(t,e,n){var i=t.length-2;e[qt]=ht(t.substr(0,i)),e[Xt]=ht(t.substr(i))})),Wt("Hmmss",(function(t,e,n){var i=t.length-4,r=t.length-2;e[qt]=ht(t.substr(0,i)),e[Xt]=ht(t.substr(i,2)),e[Zt]=ht(t.substr(r))}));var on=/[ap]\.?m?\.?/i,an=ft("Hours",!0);function sn(t,e,n){return t>11?n?"pm":"PM":n?"am":"AM"}var ln,un={calendar:E,longDateFormat:U,invalidDate:X,ordinal:J,dayOfMonthOrdinalParse:K,relativeTime:tt,months:ne,monthsShort:ie,week:Te,weekdays:Ee,weekdaysMin:Re,weekdaysShort:Ne,meridiemParse:on},cn={},dn={};function hn(t,e){var n,i=Math.min(t.length,e.length);for(n=0;n<i;n+=1)if(t[n]!==e[n])return n;return i}function fn(t){return t?t.toLowerCase().replace("_","-"):t}function pn(t){var e,n,i,r,o=0;while(o<t.length){r=fn(t[o]).split("-"),e=r.length,n=fn(t[o+1]),n=n?n.split("-"):null;while(e>0){if(i=mn(r.slice(0,e).join("-")),i)return i;if(n&&n.length>=e&&hn(r,n)>=e-1)break;e--}o++}return ln}function mn(i){var r=null;if(void 0===cn[i]&&"undefined"!==typeof t&&t&&t.exports)try{r=ln._abbr,e,n("4678")("./"+i),gn(r)}catch(o){cn[i]=null}return cn[i]}function gn(t,e){var n;return t&&(n=d(e)?_n(t):yn(t,e),n?ln=n:"undefined"!==typeof console&&console.warn),ln._abbr}function yn(t,e){if(null!==e){var n,i=un;if(e.abbr=t,null!=cn[t])I("defineLocaleOverride","use moment.updateLocale(localeName, config) to change an existing locale. moment.defineLocale(localeName, config) should only be used for creating a new locale See http://momentjs.com/guides/#/warnings/define-locale/ for more info."),i=cn[t]._config;else if(null!=e.parentLocale)if(null!=cn[e.parentLocale])i=cn[e.parentLocale]._config;else{if(n=mn(e.parentLocale),null==n)return dn[e.parentLocale]||(dn[e.parentLocale]=[]),dn[e.parentLocale].push({name:t,config:e}),null;i=n._config}return cn[t]=new P(O(i,e)),dn[t]&&dn[t].forEach((function(t){yn(t.name,t.config)})),gn(t),cn[t]}return delete cn[t],null}function vn(t,e){if(null!=e){var n,i,r=un;null!=cn[t]&&null!=cn[t].parentLocale?cn[t].set(O(cn[t]._config,e)):(i=mn(t),null!=i&&(r=i._config),e=O(r,e),null==i&&(e.abbr=t),n=new P(e),n.parentLocale=cn[t],cn[t]=n),gn(t)}else null!=cn[t]&&(null!=cn[t].parentLocale?(cn[t]=cn[t].parentLocale,t===gn()&&gn(t)):null!=cn[t]&&delete cn[t]);return cn[t]}function _n(t){var e;if(t&&t._locale&&t._locale._abbr&&(t=t._locale._abbr),!t)return ln;if(!s(t)){if(e=mn(t),e)return e;t=[t]}return pn(t)}function bn(){return D(cn)}function xn(t){var e,n=t._a;return n&&-2===v(t).overflow&&(e=n[$t]<0||n[$t]>11?$t:n[Ut]<1||n[Ut]>ee(n[Gt],n[$t])?Ut:n[qt]<0||n[qt]>24||24===n[qt]&&(0!==n[Xt]||0!==n[Zt]||0!==n[Jt])?qt:n[Xt]<0||n[Xt]>59?Xt:n[Zt]<0||n[Zt]>59?Zt:n[Jt]<0||n[Jt]>999?Jt:-1,v(t)._overflowDayOfYear&&(e<Gt||e>Ut)&&(e=Ut),v(t)._overflowWeeks&&-1===e&&(e=Kt),v(t)._overflowWeekday&&-1===e&&(e=Qt),v(t).overflow=e),t}var wn=/^\s*((?:[+-]\d{6}|\d{4})-(?:\d\d-\d\d|W\d\d-\d|W\d\d|\d\d\d|\d\d))(?:(T| )(\d\d(?::\d\d(?::\d\d(?:[.,]\d+)?)?)?)([+-]\d\d(?::?\d\d)?|\s*Z)?)?$/,Mn=/^\s*((?:[+-]\d{6}|\d{4})(?:\d\d\d\d|W\d\d\d|W\d\d|\d\d\d|\d\d|))(?:(T| )(\d\d(?:\d\d(?:\d\d(?:[.,]\d+)?)?)?)([+-]\d\d(?::?\d\d)?|\s*Z)?)?$/,Sn=/Z|[+-]\d\d(?::?\d\d)?/,Ln=[["YYYYYY-MM-DD",/[+-]\d{6}-\d\d-\d\d/],["YYYY-MM-DD",/\d{4}-\d\d-\d\d/],["GGGG-[W]WW-E",/\d{4}-W\d\d-\d/],["GGGG-[W]WW",/\d{4}-W\d\d/,!1],["YYYY-DDD",/\d{4}-\d{3}/],["YYYY-MM",/\d{4}-\d\d/,!1],["YYYYYYMMDD",/[+-]\d{10}/],["YYYYMMDD",/\d{8}/],["GGGG[W]WWE",/\d{4}W\d{3}/],["GGGG[W]WW",/\d{4}W\d{2}/,!1],["YYYYDDD",/\d{7}/],["YYYYMM",/\d{6}/,!1],["YYYY",/\d{4}/,!1]],kn=[["HH:mm:ss.SSSS",/\d\d:\d\d:\d\d\.\d+/],["HH:mm:ss,SSSS",/\d\d:\d\d:\d\d,\d+/],["HH:mm:ss",/\d\d:\d\d:\d\d/],["HH:mm",/\d\d:\d\d/],["HHmmss.SSSS",/\d\d\d\d\d\d\.\d+/],["HHmmss,SSSS",/\d\d\d\d\d\d,\d+/],["HHmmss",/\d\d\d\d\d\d/],["HHmm",/\d\d\d\d/],["HH",/\d\d/]],Tn=/^\/?Date\((-?\d+)/i,Dn=/^(?:(Mon|Tue|Wed|Thu|Fri|Sat|Sun),?\s)?(\d{1,2})\s(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\s(\d{2,4})\s(\d\d):(\d\d)(?::(\d\d))?\s(?:(UT|GMT|[ECMP][SD]T)|([Zz])|([+-]\d{4}))$/,Cn={UT:0,GMT:0,EDT:-240,EST:-300,CDT:-300,CST:-360,MDT:-360,MST:-420,PDT:-420,PST:-480};function In(t){var e,n,i,r,o,a,s=t._i,l=wn.exec(s)||Mn.exec(s);if(l){for(v(t).iso=!0,e=0,n=Ln.length;e<n;e++)if(Ln[e][1].exec(l[1])){r=Ln[e][0],i=!1!==Ln[e][2];break}if(null==r)return void(t._isValid=!1);if(l[3]){for(e=0,n=kn.length;e<n;e++)if(kn[e][1].exec(l[3])){o=(l[2]||" ")+kn[e][0];break}if(null==o)return void(t._isValid=!1)}if(!i&&null!=o)return void(t._isValid=!1);if(l[4]){if(!Sn.exec(l[4]))return void(t._isValid=!1);a="Z"}t._f=r+(o||"")+(a||""),Bn(t)}else t._isValid=!1}function An(t,e,n,i,r,o){var a=[Yn(t),ie.indexOf(e),parseInt(n,10),parseInt(i,10),parseInt(r,10)];return o&&a.push(parseInt(o,10)),a}function Yn(t){var e=parseInt(t,10);return e<=49?2e3+e:e<=999?1900+e:e}function On(t){return t.replace(/\([^)]*\)|[\n\t]/g," ").replace(/(\s\s+)/g," ").replace(/^\s\s*/,"").replace(/\s\s*$/,"")}function Pn(t,e,n){if(t){var i=Ne.indexOf(t),r=new Date(e[0],e[1],e[2]).getDay();if(i!==r)return v(n).weekdayMismatch=!0,n._isValid=!1,!1}return!0}function En(t,e,n){if(t)return Cn[t];if(e)return 0;var i=parseInt(n,10),r=i%100,o=(i-r)/100;return 60*o+r}function Nn(t){var e,n=Dn.exec(On(t._i));if(n){if(e=An(n[4],n[3],n[2],n[5],n[6],n[7]),!Pn(n[1],e,t))return;t._a=e,t._tzm=En(n[8],n[9],n[10]),t._d=xe.apply(null,t._a),t._d.setUTCMinutes(t._d.getUTCMinutes()-t._tzm),v(t).rfc2822=!0}else t._isValid=!1}function Rn(t){var e=Tn.exec(t._i);null===e?(In(t),!1===t._isValid&&(delete t._isValid,Nn(t),!1===t._isValid&&(delete t._isValid,t._strict?t._isValid=!1:o.createFromInputFallback(t)))):t._d=new Date(+e[1])}function Hn(t,e,n){return null!=t?t:null!=e?e:n}function jn(t){var e=new Date(o.now());return t._useUTC?[e.getUTCFullYear(),e.getUTCMonth(),e.getUTCDate()]:[e.getFullYear(),e.getMonth(),e.getDate()]}function zn(t){var e,n,i,r,o,a=[];if(!t._d){for(i=jn(t),t._w&&null==t._a[Ut]&&null==t._a[$t]&&Wn(t),null!=t._dayOfYear&&(o=Hn(t._a[Gt],i[Gt]),(t._dayOfYear>ye(o)||0===t._dayOfYear)&&(v(t)._overflowDayOfYear=!0),n=xe(o,0,t._dayOfYear),t._a[$t]=n.getUTCMonth(),t._a[Ut]=n.getUTCDate()),e=0;e<3&&null==t._a[e];++e)t._a[e]=a[e]=i[e];for(;e<7;e++)t._a[e]=a[e]=null==t._a[e]?2===e?1:0:t._a[e];24===t._a[qt]&&0===t._a[Xt]&&0===t._a[Zt]&&0===t._a[Jt]&&(t._nextDay=!0,t._a[qt]=0),t._d=(t._useUTC?xe:be).apply(null,a),r=t._useUTC?t._d.getUTCDay():t._d.getDay(),null!=t._tzm&&t._d.setUTCMinutes(t._d.getUTCMinutes()-t._tzm),t._nextDay&&(t._a[qt]=24),t._w&&"undefined"!==typeof t._w.d&&t._w.d!==r&&(v(t).weekdayMismatch=!0)}}function Wn(t){var e,n,i,r,o,a,s,l,u;e=t._w,null!=e.GG||null!=e.W||null!=e.E?(o=1,a=4,n=Hn(e.GG,t._a[Gt],Se(Zn(),1,4).year),i=Hn(e.W,1),r=Hn(e.E,1),(r<1||r>7)&&(l=!0)):(o=t._locale._week.dow,a=t._locale._week.doy,u=Se(Zn(),o,a),n=Hn(e.gg,t._a[Gt],u.year),i=Hn(e.w,u.week),null!=e.d?(r=e.d,(r<0||r>6)&&(l=!0)):null!=e.e?(r=e.e+o,(e.e<0||e.e>6)&&(l=!0)):r=o),i<1||i>Le(n,o,a)?v(t)._overflowWeeks=!0:null!=l?v(t)._overflowWeekday=!0:(s=Me(n,i,r,o,a),t._a[Gt]=s.year,t._dayOfYear=s.dayOfYear)}function Bn(t){if(t._f!==o.ISO_8601)if(t._f!==o.RFC_2822){t._a=[],v(t).empty=!0;var e,n,i,r,a,s,l=""+t._i,u=l.length,c=0;for(i=$(t._f,t._locale).match(H)||[],e=0;e<i.length;e++)r=i[e],n=(l.match(Rt(r,t))||[])[0],n&&(a=l.substr(0,l.indexOf(n)),a.length>0&&v(t).unusedInput.push(a),l=l.slice(l.indexOf(n)+n.length),c+=n.length),W[r]?(n?v(t).empty=!1:v(t).unusedTokens.push(r),Ft(r,n,t)):t._strict&&!n&&v(t).unusedTokens.push(r);v(t).charsLeftOver=u-c,l.length>0&&v(t).unusedInput.push(l),t._a[qt]<=12&&!0===v(t).bigHour&&t._a[qt]>0&&(v(t).bigHour=void 0),v(t).parsedDateParts=t._a.slice(0),v(t).meridiem=t._meridiem,t._a[qt]=Fn(t._locale,t._a[qt],t._meridiem),s=v(t).era,null!==s&&(t._a[Gt]=t._locale.erasConvertYear(s,t._a[Gt])),zn(t),xn(t)}else Nn(t);else In(t)}function Fn(t,e,n){var i;return null==n?e:null!=t.meridiemHour?t.meridiemHour(e,n):null!=t.isPM?(i=t.isPM(n),i&&e<12&&(e+=12),i||12!==e||(e=0),e):e}function Vn(t){var e,n,i,r,o,a,s=!1;if(0===t._f.length)return v(t).invalidFormat=!0,void(t._d=new Date(NaN));for(r=0;r<t._f.length;r++)o=0,a=!1,e=M({},t),null!=t._useUTC&&(e._useUTC=t._useUTC),e._f=t._f[r],Bn(e),_(e)&&(a=!0),o+=v(e).charsLeftOver,o+=10*v(e).unusedTokens.length,v(e).score=o,s?o<i&&(i=o,n=e):(null==i||o<i||a)&&(i=o,n=e,a&&(s=!0));m(t,n||e)}function Gn(t){if(!t._d){var e=at(t._i),n=void 0===e.day?e.date:e.day;t._a=p([e.year,e.month,n,e.hour,e.minute,e.second,e.millisecond],(function(t){return t&&parseInt(t,10)})),zn(t)}}function $n(t){var e=new S(xn(Un(t)));return e._nextDay&&(e.add(1,"d"),e._nextDay=void 0),e}function Un(t){var e=t._i,n=t._f;return t._locale=t._locale||_n(t._l),null===e||void 0===n&&""===e?b({nullInput:!0}):("string"===typeof e&&(t._i=e=t._locale.preparse(e)),L(e)?new S(xn(e)):(f(e)?t._d=e:s(n)?Vn(t):n?Bn(t):qn(t),_(t)||(t._d=null),t))}function qn(t){var e=t._i;d(e)?t._d=new Date(o.now()):f(e)?t._d=new Date(e.valueOf()):"string"===typeof e?Rn(t):s(e)?(t._a=p(e.slice(0),(function(t){return parseInt(t,10)})),zn(t)):l(e)?Gn(t):h(e)?t._d=new Date(e):o.createFromInputFallback(t)}function Xn(t,e,n,i,r){var o={};return!0!==e&&!1!==e||(i=e,e=void 0),!0!==n&&!1!==n||(i=n,n=void 0),(l(t)&&c(t)||s(t)&&0===t.length)&&(t=void 0),o._isAMomentObject=!0,o._useUTC=o._isUTC=r,o._l=n,o._i=t,o._f=e,o._strict=i,$n(o)}function Zn(t,e,n,i){return Xn(t,e,n,i,!1)}o.createFromInputFallback=T("value provided is not in a recognized RFC2822 or ISO format. moment construction falls back to js Date(), which is not reliable across all browsers and versions. Non RFC2822/ISO date formats are discouraged. Please refer to http://momentjs.com/guides/#/warnings/js-date/ for more info.",(function(t){t._d=new Date(t._i+(t._useUTC?" UTC":""))})),o.ISO_8601=function(){},o.RFC_2822=function(){};var Jn=T("moment().min is deprecated, use moment.max instead. http://momentjs.com/guides/#/warnings/min-max/",(function(){var t=Zn.apply(null,arguments);return this.isValid()&&t.isValid()?t<this?this:t:b()})),Kn=T("moment().max is deprecated, use moment.min instead. http://momentjs.com/guides/#/warnings/min-max/",(function(){var t=Zn.apply(null,arguments);return this.isValid()&&t.isValid()?t>this?this:t:b()}));function Qn(t,e){var n,i;if(1===e.length&&s(e[0])&&(e=e[0]),!e.length)return Zn();for(n=e[0],i=1;i<e.length;++i)e[i].isValid()&&!e[i][t](n)||(n=e[i]);return n}function ti(){var t=[].slice.call(arguments,0);return Qn("isBefore",t)}function ei(){var t=[].slice.call(arguments,0);return Qn("isAfter",t)}var ni=function(){return Date.now?Date.now():+new Date},ii=["year","quarter","month","week","day","hour","minute","second","millisecond"];function ri(t){var e,n,i=!1;for(e in t)if(u(t,e)&&(-1===Vt.call(ii,e)||null!=t[e]&&isNaN(t[e])))return!1;for(n=0;n<ii.length;++n)if(t[ii[n]]){if(i)return!1;parseFloat(t[ii[n]])!==ht(t[ii[n]])&&(i=!0)}return!0}function oi(){return this._isValid}function ai(){return Ci(NaN)}function si(t){var e=at(t),n=e.year||0,i=e.quarter||0,r=e.month||0,o=e.week||e.isoWeek||0,a=e.day||0,s=e.hour||0,l=e.minute||0,u=e.second||0,c=e.millisecond||0;this._isValid=ri(e),this._milliseconds=+c+1e3*u+6e4*l+1e3*s*60*60,this._days=+a+7*o,this._months=+r+3*i+12*n,this._data={},this._locale=_n(),this._bubble()}function li(t){return t instanceof si}function ui(t){return t<0?-1*Math.round(-1*t):Math.round(t)}function ci(t,e,n){var i,r=Math.min(t.length,e.length),o=Math.abs(t.length-e.length),a=0;for(i=0;i<r;i++)(n&&t[i]!==e[i]||!n&&ht(t[i])!==ht(e[i]))&&a++;return a+o}function di(t,e){B(t,0,0,(function(){var t=this.utcOffset(),n="+";return t<0&&(t=-t,n="-"),n+R(~~(t/60),2)+e+R(~~t%60,2)}))}di("Z",":"),di("ZZ",""),Nt("Z",Ot),Nt("ZZ",Ot),Wt(["Z","ZZ"],(function(t,e,n){n._useUTC=!0,n._tzm=fi(Ot,t)}));var hi=/([\+\-]|\d\d)/gi;function fi(t,e){var n,i,r,o=(e||"").match(t);return null===o?null:(n=o[o.length-1]||[],i=(n+"").match(hi)||["-",0,0],r=60*i[1]+ht(i[2]),0===r?0:"+"===i[0]?r:-r)}function pi(t,e){var n,i;return e._isUTC?(n=e.clone(),i=(L(t)||f(t)?t.valueOf():Zn(t).valueOf())-n.valueOf(),n._d.setTime(n._d.valueOf()+i),o.updateOffset(n,!1),n):Zn(t).local()}function mi(t){return-Math.round(t._d.getTimezoneOffset())}function gi(t,e,n){var i,r=this._offset||0;if(!this.isValid())return null!=t?this:NaN;if(null!=t){if("string"===typeof t){if(t=fi(Ot,t),null===t)return this}else Math.abs(t)<16&&!n&&(t*=60);return!this._isUTC&&e&&(i=mi(this)),this._offset=t,this._isUTC=!0,null!=i&&this.add(i,"m"),r!==t&&(!e||this._changeInProgress?Pi(this,Ci(t-r,"m"),1,!1):this._changeInProgress||(this._changeInProgress=!0,o.updateOffset(this,!0),this._changeInProgress=null)),this}return this._isUTC?r:mi(this)}function yi(t,e){return null!=t?("string"!==typeof t&&(t=-t),this.utcOffset(t,e),this):-this.utcOffset()}function vi(t){return this.utcOffset(0,t)}function _i(t){return this._isUTC&&(this.utcOffset(0,t),this._isUTC=!1,t&&this.subtract(mi(this),"m")),this}function bi(){if(null!=this._tzm)this.utcOffset(this._tzm,!1,!0);else if("string"===typeof this._i){var t=fi(Yt,this._i);null!=t?this.utcOffset(t):this.utcOffset(0,!0)}return this}function xi(t){return!!this.isValid()&&(t=t?Zn(t).utcOffset():0,(this.utcOffset()-t)%60===0)}function wi(){return this.utcOffset()>this.clone().month(0).utcOffset()||this.utcOffset()>this.clone().month(5).utcOffset()}function Mi(){if(!d(this._isDSTShifted))return this._isDSTShifted;var t,e={};return M(e,this),e=Un(e),e._a?(t=e._isUTC?g(e._a):Zn(e._a),this._isDSTShifted=this.isValid()&&ci(e._a,t.toArray())>0):this._isDSTShifted=!1,this._isDSTShifted}function Si(){return!!this.isValid()&&!this._isUTC}function Li(){return!!this.isValid()&&this._isUTC}function ki(){return!!this.isValid()&&(this._isUTC&&0===this._offset)}o.updateOffset=function(){};var Ti=/^(-|\+)?(?:(\d*)[. ])?(\d+):(\d+)(?::(\d+)(\.\d*)?)?$/,Di=/^(-|\+)?P(?:([-+]?[0-9,.]*)Y)?(?:([-+]?[0-9,.]*)M)?(?:([-+]?[0-9,.]*)W)?(?:([-+]?[0-9,.]*)D)?(?:T(?:([-+]?[0-9,.]*)H)?(?:([-+]?[0-9,.]*)M)?(?:([-+]?[0-9,.]*)S)?)?$/;function Ci(t,e){var n,i,r,o=t,a=null;return li(t)?o={ms:t._milliseconds,d:t._days,M:t._months}:h(t)||!isNaN(+t)?(o={},e?o[e]=+t:o.milliseconds=+t):(a=Ti.exec(t))?(n="-"===a[1]?-1:1,o={y:0,d:ht(a[Ut])*n,h:ht(a[qt])*n,m:ht(a[Xt])*n,s:ht(a[Zt])*n,ms:ht(ui(1e3*a[Jt]))*n}):(a=Di.exec(t))?(n="-"===a[1]?-1:1,o={y:Ii(a[2],n),M:Ii(a[3],n),w:Ii(a[4],n),d:Ii(a[5],n),h:Ii(a[6],n),m:Ii(a[7],n),s:Ii(a[8],n)}):null==o?o={}:"object"===typeof o&&("from"in o||"to"in o)&&(r=Yi(Zn(o.from),Zn(o.to)),o={},o.ms=r.milliseconds,o.M=r.months),i=new si(o),li(t)&&u(t,"_locale")&&(i._locale=t._locale),li(t)&&u(t,"_isValid")&&(i._isValid=t._isValid),i}function Ii(t,e){var n=t&&parseFloat(t.replace(",","."));return(isNaN(n)?0:n)*e}function Ai(t,e){var n={};return n.months=e.month()-t.month()+12*(e.year()-t.year()),t.clone().add(n.months,"M").isAfter(e)&&--n.months,n.milliseconds=+e-+t.clone().add(n.months,"M"),n}function Yi(t,e){var n;return t.isValid()&&e.isValid()?(e=pi(e,t),t.isBefore(e)?n=Ai(t,e):(n=Ai(e,t),n.milliseconds=-n.milliseconds,n.months=-n.months),n):{milliseconds:0,months:0}}function Oi(t,e){return function(n,i){var r,o;return null===i||isNaN(+i)||(I(e,"moment()."+e+"(period, number) is deprecated. Please use moment()."+e+"(number, period). See http://momentjs.com/guides/#/warnings/add-inverted-param/ for more info."),o=n,n=i,i=o),r=Ci(n,i),Pi(this,r,t),this}}function Pi(t,e,n,i){var r=e._milliseconds,a=ui(e._days),s=ui(e._months);t.isValid()&&(i=null==i||i,s&&de(t,pt(t,"Month")+s*n),a&&mt(t,"Date",pt(t,"Date")+a*n),r&&t._d.setTime(t._d.valueOf()+r*n),i&&o.updateOffset(t,a||s))}Ci.fn=si.prototype,Ci.invalid=ai;var Ei=Oi(1,"add"),Ni=Oi(-1,"subtract");function Ri(t){return"string"===typeof t||t instanceof String}function Hi(t){return L(t)||f(t)||Ri(t)||h(t)||zi(t)||ji(t)||null===t||void 0===t}function ji(t){var e,n,i=l(t)&&!c(t),r=!1,o=["years","year","y","months","month","M","days","day","d","dates","date","D","hours","hour","h","minutes","minute","m","seconds","second","s","milliseconds","millisecond","ms"];for(e=0;e<o.length;e+=1)n=o[e],r=r||u(t,n);return i&&r}function zi(t){var e=s(t),n=!1;return e&&(n=0===t.filter((function(e){return!h(e)&&Ri(t)})).length),e&&n}function Wi(t){var e,n,i=l(t)&&!c(t),r=!1,o=["sameDay","nextDay","lastDay","nextWeek","lastWeek","sameElse"];for(e=0;e<o.length;e+=1)n=o[e],r=r||u(t,n);return i&&r}function Bi(t,e){var n=t.diff(e,"days",!0);return n<-6?"sameElse":n<-1?"lastWeek":n<0?"lastDay":n<1?"sameDay":n<2?"nextDay":n<7?"nextWeek":"sameElse"}function Fi(t,e){1===arguments.length&&(arguments[0]?Hi(arguments[0])?(t=arguments[0],e=void 0):Wi(arguments[0])&&(e=arguments[0],t=void 0):(t=void 0,e=void 0));var n=t||Zn(),i=pi(n,this).startOf("day"),r=o.calendarFormat(this,i)||"sameElse",a=e&&(A(e[r])?e[r].call(this,n):e[r]);return this.format(a||this.localeData().calendar(r,this,Zn(n)))}function Vi(){return new S(this)}function Gi(t,e){var n=L(t)?t:Zn(t);return!(!this.isValid()||!n.isValid())&&(e=ot(e)||"millisecond","millisecond"===e?this.valueOf()>n.valueOf():n.valueOf()<this.clone().startOf(e).valueOf())}function $i(t,e){var n=L(t)?t:Zn(t);return!(!this.isValid()||!n.isValid())&&(e=ot(e)||"millisecond","millisecond"===e?this.valueOf()<n.valueOf():this.clone().endOf(e).valueOf()<n.valueOf())}function Ui(t,e,n,i){var r=L(t)?t:Zn(t),o=L(e)?e:Zn(e);return!!(this.isValid()&&r.isValid()&&o.isValid())&&(i=i||"()",("("===i[0]?this.isAfter(r,n):!this.isBefore(r,n))&&(")"===i[1]?this.isBefore(o,n):!this.isAfter(o,n)))}function qi(t,e){var n,i=L(t)?t:Zn(t);return!(!this.isValid()||!i.isValid())&&(e=ot(e)||"millisecond","millisecond"===e?this.valueOf()===i.valueOf():(n=i.valueOf(),this.clone().startOf(e).valueOf()<=n&&n<=this.clone().endOf(e).valueOf()))}function Xi(t,e){return this.isSame(t,e)||this.isAfter(t,e)}function Zi(t,e){return this.isSame(t,e)||this.isBefore(t,e)}function Ji(t,e,n){var i,r,o;if(!this.isValid())return NaN;if(i=pi(t,this),!i.isValid())return NaN;switch(r=6e4*(i.utcOffset()-this.utcOffset()),e=ot(e),e){case"year":o=Ki(this,i)/12;break;case"month":o=Ki(this,i);break;case"quarter":o=Ki(this,i)/3;break;case"second":o=(this-i)/1e3;break;case"minute":o=(this-i)/6e4;break;case"hour":o=(this-i)/36e5;break;case"day":o=(this-i-r)/864e5;break;case"week":o=(this-i-r)/6048e5;break;default:o=this-i}return n?o:dt(o)}function Ki(t,e){if(t.date()<e.date())return-Ki(e,t);var n,i,r=12*(e.year()-t.year())+(e.month()-t.month()),o=t.clone().add(r,"months");return e-o<0?(n=t.clone().add(r-1,"months"),i=(e-o)/(o-n)):(n=t.clone().add(r+1,"months"),i=(e-o)/(n-o)),-(r+i)||0}function Qi(){return this.clone().locale("en").format("ddd MMM DD YYYY HH:mm:ss [GMT]ZZ")}function tr(t){if(!this.isValid())return null;var e=!0!==t,n=e?this.clone().utc():this;return n.year()<0||n.year()>9999?G(n,e?"YYYYYY-MM-DD[T]HH:mm:ss.SSS[Z]":"YYYYYY-MM-DD[T]HH:mm:ss.SSSZ"):A(Date.prototype.toISOString)?e?this.toDate().toISOString():new Date(this.valueOf()+60*this.utcOffset()*1e3).toISOString().replace("Z",G(n,"Z")):G(n,e?"YYYY-MM-DD[T]HH:mm:ss.SSS[Z]":"YYYY-MM-DD[T]HH:mm:ss.SSSZ")}function er(){if(!this.isValid())return"moment.invalid(/* "+this._i+" */)";var t,e,n,i,r="moment",o="";return this.isLocal()||(r=0===this.utcOffset()?"moment.utc":"moment.parseZone",o="Z"),t="["+r+'("]',e=0<=this.year()&&this.year()<=9999?"YYYY":"YYYYYY",n="-MM-DD[T]HH:mm:ss.SSS",i=o+'[")]',this.format(t+e+n+i)}function nr(t){t||(t=this.isUtc()?o.defaultFormatUtc:o.defaultFormat);var e=G(this,t);return this.localeData().postformat(e)}function ir(t,e){return this.isValid()&&(L(t)&&t.isValid()||Zn(t).isValid())?Ci({to:this,from:t}).locale(this.locale()).humanize(!e):this.localeData().invalidDate()}function rr(t){return this.from(Zn(),t)}function or(t,e){return this.isValid()&&(L(t)&&t.isValid()||Zn(t).isValid())?Ci({from:this,to:t}).locale(this.locale()).humanize(!e):this.localeData().invalidDate()}function ar(t){return this.to(Zn(),t)}function sr(t){var e;return void 0===t?this._locale._abbr:(e=_n(t),null!=e&&(this._locale=e),this)}o.defaultFormat="YYYY-MM-DDTHH:mm:ssZ",o.defaultFormatUtc="YYYY-MM-DDTHH:mm:ss[Z]";var lr=T("moment().lang() is deprecated. Instead, use moment().localeData() to get the language configuration. Use moment().locale() to change languages.",(function(t){return void 0===t?this.localeData():this.locale(t)}));function ur(){return this._locale}var cr=1e3,dr=60*cr,hr=60*dr,fr=3506328*hr;function pr(t,e){return(t%e+e)%e}function mr(t,e,n){return t<100&&t>=0?new Date(t+400,e,n)-fr:new Date(t,e,n).valueOf()}function gr(t,e,n){return t<100&&t>=0?Date.UTC(t+400,e,n)-fr:Date.UTC(t,e,n)}function yr(t){var e,n;if(t=ot(t),void 0===t||"millisecond"===t||!this.isValid())return this;switch(n=this._isUTC?gr:mr,t){case"year":e=n(this.year(),0,1);break;case"quarter":e=n(this.year(),this.month()-this.month()%3,1);break;case"month":e=n(this.year(),this.month(),1);break;case"week":e=n(this.year(),this.month(),this.date()-this.weekday());break;case"isoWeek":e=n(this.year(),this.month(),this.date()-(this.isoWeekday()-1));break;case"day":case"date":e=n(this.year(),this.month(),this.date());break;case"hour":e=this._d.valueOf(),e-=pr(e+(this._isUTC?0:this.utcOffset()*dr),hr);break;case"minute":e=this._d.valueOf(),e-=pr(e,dr);break;case"second":e=this._d.valueOf(),e-=pr(e,cr);break}return this._d.setTime(e),o.updateOffset(this,!0),this}function vr(t){var e,n;if(t=ot(t),void 0===t||"millisecond"===t||!this.isValid())return this;switch(n=this._isUTC?gr:mr,t){case"year":e=n(this.year()+1,0,1)-1;break;case"quarter":e=n(this.year(),this.month()-this.month()%3+3,1)-1;break;case"month":e=n(this.year(),this.month()+1,1)-1;break;case"week":e=n(this.year(),this.month(),this.date()-this.weekday()+7)-1;break;case"isoWeek":e=n(this.year(),this.month(),this.date()-(this.isoWeekday()-1)+7)-1;break;case"day":case"date":e=n(this.year(),this.month(),this.date()+1)-1;break;case"hour":e=this._d.valueOf(),e+=hr-pr(e+(this._isUTC?0:this.utcOffset()*dr),hr)-1;break;case"minute":e=this._d.valueOf(),e+=dr-pr(e,dr)-1;break;case"second":e=this._d.valueOf(),e+=cr-pr(e,cr)-1;break}return this._d.setTime(e),o.updateOffset(this,!0),this}function _r(){return this._d.valueOf()-6e4*(this._offset||0)}function br(){return Math.floor(this.valueOf()/1e3)}function xr(){return new Date(this.valueOf())}function wr(){var t=this;return[t.year(),t.month(),t.date(),t.hour(),t.minute(),t.second(),t.millisecond()]}function Mr(){var t=this;return{years:t.year(),months:t.month(),date:t.date(),hours:t.hours(),minutes:t.minutes(),seconds:t.seconds(),milliseconds:t.milliseconds()}}function Sr(){return this.isValid()?this.toISOString():null}function Lr(){return _(this)}function kr(){return m({},v(this))}function Tr(){return v(this).overflow}function Dr(){return{input:this._i,format:this._f,locale:this._locale,isUTC:this._isUTC,strict:this._strict}}function Cr(t,e){var n,i,r,a=this._eras||_n("en")._eras;for(n=0,i=a.length;n<i;++n){switch(typeof a[n].since){case"string":r=o(a[n].since).startOf("day"),a[n].since=r.valueOf();break}switch(typeof a[n].until){case"undefined":a[n].until=1/0;break;case"string":r=o(a[n].until).startOf("day").valueOf(),a[n].until=r.valueOf();break}}return a}function Ir(t,e,n){var i,r,o,a,s,l=this.eras();for(t=t.toUpperCase(),i=0,r=l.length;i<r;++i)if(o=l[i].name.toUpperCase(),a=l[i].abbr.toUpperCase(),s=l[i].narrow.toUpperCase(),n)switch(e){case"N":case"NN":case"NNN":if(a===t)return l[i];break;case"NNNN":if(o===t)return l[i];break;case"NNNNN":if(s===t)return l[i];break}else if([o,a,s].indexOf(t)>=0)return l[i]}function Ar(t,e){var n=t.since<=t.until?1:-1;return void 0===e?o(t.since).year():o(t.since).year()+(e-t.offset)*n}function Yr(){var t,e,n,i=this.localeData().eras();for(t=0,e=i.length;t<e;++t){if(n=this.clone().startOf("day").valueOf(),i[t].since<=n&&n<=i[t].until)return i[t].name;if(i[t].until<=n&&n<=i[t].since)return i[t].name}return""}function Or(){var t,e,n,i=this.localeData().eras();for(t=0,e=i.length;t<e;++t){if(n=this.clone().startOf("day").valueOf(),i[t].since<=n&&n<=i[t].until)return i[t].narrow;if(i[t].until<=n&&n<=i[t].since)return i[t].narrow}return""}function Pr(){var t,e,n,i=this.localeData().eras();for(t=0,e=i.length;t<e;++t){if(n=this.clone().startOf("day").valueOf(),i[t].since<=n&&n<=i[t].until)return i[t].abbr;if(i[t].until<=n&&n<=i[t].since)return i[t].abbr}return""}function Er(){var t,e,n,i,r=this.localeData().eras();for(t=0,e=r.length;t<e;++t)if(n=r[t].since<=r[t].until?1:-1,i=this.clone().startOf("day").valueOf(),r[t].since<=i&&i<=r[t].until||r[t].until<=i&&i<=r[t].since)return(this.year()-o(r[t].since).year())*n+r[t].offset;return this.year()}function Nr(t){return u(this,"_erasNameRegex")||Fr.call(this),t?this._erasNameRegex:this._erasRegex}function Rr(t){return u(this,"_erasAbbrRegex")||Fr.call(this),t?this._erasAbbrRegex:this._erasRegex}function Hr(t){return u(this,"_erasNarrowRegex")||Fr.call(this),t?this._erasNarrowRegex:this._erasRegex}function jr(t,e){return e.erasAbbrRegex(t)}function zr(t,e){return e.erasNameRegex(t)}function Wr(t,e){return e.erasNarrowRegex(t)}function Br(t,e){return e._eraYearOrdinalRegex||It}function Fr(){var t,e,n=[],i=[],r=[],o=[],a=this.eras();for(t=0,e=a.length;t<e;++t)i.push(jt(a[t].name)),n.push(jt(a[t].abbr)),r.push(jt(a[t].narrow)),o.push(jt(a[t].name)),o.push(jt(a[t].abbr)),o.push(jt(a[t].narrow));this._erasRegex=new RegExp("^("+o.join("|")+")","i"),this._erasNameRegex=new RegExp("^("+i.join("|")+")","i"),this._erasAbbrRegex=new RegExp("^("+n.join("|")+")","i"),this._erasNarrowRegex=new RegExp("^("+r.join("|")+")","i")}function Vr(t,e){B(0,[t,t.length],0,e)}function Gr(t){return Jr.call(this,t,this.week(),this.weekday(),this.localeData()._week.dow,this.localeData()._week.doy)}function $r(t){return Jr.call(this,t,this.isoWeek(),this.isoWeekday(),1,4)}function Ur(){return Le(this.year(),1,4)}function qr(){return Le(this.isoWeekYear(),1,4)}function Xr(){var t=this.localeData()._week;return Le(this.year(),t.dow,t.doy)}function Zr(){var t=this.localeData()._week;return Le(this.weekYear(),t.dow,t.doy)}function Jr(t,e,n,i,r){var o;return null==t?Se(this,i,r).year:(o=Le(t,i,r),e>o&&(e=o),Kr.call(this,t,e,n,i,r))}function Kr(t,e,n,i,r){var o=Me(t,e,n,i,r),a=xe(o.year,0,o.dayOfYear);return this.year(a.getUTCFullYear()),this.month(a.getUTCMonth()),this.date(a.getUTCDate()),this}function Qr(t){return null==t?Math.ceil((this.month()+1)/3):this.month(3*(t-1)+this.month()%3)}B("N",0,0,"eraAbbr"),B("NN",0,0,"eraAbbr"),B("NNN",0,0,"eraAbbr"),B("NNNN",0,0,"eraName"),B("NNNNN",0,0,"eraNarrow"),B("y",["y",1],"yo","eraYear"),B("y",["yy",2],0,"eraYear"),B("y",["yyy",3],0,"eraYear"),B("y",["yyyy",4],0,"eraYear"),Nt("N",jr),Nt("NN",jr),Nt("NNN",jr),Nt("NNNN",zr),Nt("NNNNN",Wr),Wt(["N","NN","NNN","NNNN","NNNNN"],(function(t,e,n,i){var r=n._locale.erasParse(t,i,n._strict);r?v(n).era=r:v(n).invalidEra=t})),Nt("y",It),Nt("yy",It),Nt("yyy",It),Nt("yyyy",It),Nt("yo",Br),Wt(["y","yy","yyy","yyyy"],Gt),Wt(["yo"],(function(t,e,n,i){var r;n._locale._eraYearOrdinalRegex&&(r=t.match(n._locale._eraYearOrdinalRegex)),n._locale.eraYearOrdinalParse?e[Gt]=n._locale.eraYearOrdinalParse(t,r):e[Gt]=parseInt(t,10)})),B(0,["gg",2],0,(function(){return this.weekYear()%100})),B(0,["GG",2],0,(function(){return this.isoWeekYear()%100})),Vr("gggg","weekYear"),Vr("ggggg","weekYear"),Vr("GGGG","isoWeekYear"),Vr("GGGGG","isoWeekYear"),rt("weekYear","gg"),rt("isoWeekYear","GG"),lt("weekYear",1),lt("isoWeekYear",1),Nt("G",At),Nt("g",At),Nt("GG",St,bt),Nt("gg",St,bt),Nt("GGGG",Dt,wt),Nt("gggg",Dt,wt),Nt("GGGGG",Ct,Mt),Nt("ggggg",Ct,Mt),Bt(["gggg","ggggg","GGGG","GGGGG"],(function(t,e,n,i){e[i.substr(0,2)]=ht(t)})),Bt(["gg","GG"],(function(t,e,n,i){e[i]=o.parseTwoDigitYear(t)})),B("Q",0,"Qo","quarter"),rt("quarter","Q"),lt("quarter",7),Nt("Q",_t),Wt("Q",(function(t,e){e[$t]=3*(ht(t)-1)})),B("D",["DD",2],"Do","date"),rt("date","D"),lt("date",9),Nt("D",St),Nt("DD",St,bt),Nt("Do",(function(t,e){return t?e._dayOfMonthOrdinalParse||e._ordinalParse:e._dayOfMonthOrdinalParseLenient})),Wt(["D","DD"],Ut),Wt("Do",(function(t,e){e[Ut]=ht(t.match(St)[0])}));var to=ft("Date",!0);function eo(t){var e=Math.round((this.clone().startOf("day")-this.clone().startOf("year"))/864e5)+1;return null==t?e:this.add(t-e,"d")}B("DDD",["DDDD",3],"DDDo","dayOfYear"),rt("dayOfYear","DDD"),lt("dayOfYear",4),Nt("DDD",Tt),Nt("DDDD",xt),Wt(["DDD","DDDD"],(function(t,e,n){n._dayOfYear=ht(t)})),B("m",["mm",2],0,"minute"),rt("minute","m"),lt("minute",14),Nt("m",St),Nt("mm",St,bt),Wt(["m","mm"],Xt);var no=ft("Minutes",!1);B("s",["ss",2],0,"second"),rt("second","s"),lt("second",15),Nt("s",St),Nt("ss",St,bt),Wt(["s","ss"],Zt);var io,ro,oo=ft("Seconds",!1);for(B("S",0,0,(function(){return~~(this.millisecond()/100)})),B(0,["SS",2],0,(function(){return~~(this.millisecond()/10)})),B(0,["SSS",3],0,"millisecond"),B(0,["SSSS",4],0,(function(){return 10*this.millisecond()})),B(0,["SSSSS",5],0,(function(){return 100*this.millisecond()})),B(0,["SSSSSS",6],0,(function(){return 1e3*this.millisecond()})),B(0,["SSSSSSS",7],0,(function(){return 1e4*this.millisecond()})),B(0,["SSSSSSSS",8],0,(function(){return 1e5*this.millisecond()})),B(0,["SSSSSSSSS",9],0,(function(){return 1e6*this.millisecond()})),rt("millisecond","ms"),lt("millisecond",16),Nt("S",Tt,_t),Nt("SS",Tt,bt),Nt("SSS",Tt,xt),io="SSSS";io.length<=9;io+="S")Nt(io,It);function ao(t,e){e[Jt]=ht(1e3*("0."+t))}for(io="S";io.length<=9;io+="S")Wt(io,ao);function so(){return this._isUTC?"UTC":""}function lo(){return this._isUTC?"Coordinated Universal Time":""}ro=ft("Milliseconds",!1),B("z",0,0,"zoneAbbr"),B("zz",0,0,"zoneName");var uo=S.prototype;function co(t){return Zn(1e3*t)}function ho(){return Zn.apply(null,arguments).parseZone()}function fo(t){return t}uo.add=Ei,uo.calendar=Fi,uo.clone=Vi,uo.diff=Ji,uo.endOf=vr,uo.format=nr,uo.from=ir,uo.fromNow=rr,uo.to=or,uo.toNow=ar,uo.get=gt,uo.invalidAt=Tr,uo.isAfter=Gi,uo.isBefore=$i,uo.isBetween=Ui,uo.isSame=qi,uo.isSameOrAfter=Xi,uo.isSameOrBefore=Zi,uo.isValid=Lr,uo.lang=lr,uo.locale=sr,uo.localeData=ur,uo.max=Kn,uo.min=Jn,uo.parsingFlags=kr,uo.set=yt,uo.startOf=yr,uo.subtract=Ni,uo.toArray=wr,uo.toObject=Mr,uo.toDate=xr,uo.toISOString=tr,uo.inspect=er,"undefined"!==typeof Symbol&&null!=Symbol.for&&(uo[Symbol.for("nodejs.util.inspect.custom")]=function(){return"Moment<"+this.format()+">"}),uo.toJSON=Sr,uo.toString=Qi,uo.unix=br,uo.valueOf=_r,uo.creationData=Dr,uo.eraName=Yr,uo.eraNarrow=Or,uo.eraAbbr=Pr,uo.eraYear=Er,uo.year=ve,uo.isLeapYear=_e,uo.weekYear=Gr,uo.isoWeekYear=$r,uo.quarter=uo.quarters=Qr,uo.month=he,uo.daysInMonth=fe,uo.week=uo.weeks=Ie,uo.isoWeek=uo.isoWeeks=Ae,uo.weeksInYear=Xr,uo.weeksInWeekYear=Zr,uo.isoWeeksInYear=Ur,uo.isoWeeksInISOWeekYear=qr,uo.date=to,uo.day=uo.days=$e,uo.weekday=Ue,uo.isoWeekday=qe,uo.dayOfYear=eo,uo.hour=uo.hours=an,uo.minute=uo.minutes=no,uo.second=uo.seconds=oo,uo.millisecond=uo.milliseconds=ro,uo.utcOffset=gi,uo.utc=vi,uo.local=_i,uo.parseZone=bi,uo.hasAlignedHourOffset=xi,uo.isDST=wi,uo.isLocal=Si,uo.isUtcOffset=Li,uo.isUtc=ki,uo.isUTC=ki,uo.zoneAbbr=so,uo.zoneName=lo,uo.dates=T("dates accessor is deprecated. Use date instead.",to),uo.months=T("months accessor is deprecated. Use month instead",he),uo.years=T("years accessor is deprecated. Use year instead",ve),uo.zone=T("moment().zone is deprecated, use moment().utcOffset instead. http://momentjs.com/guides/#/warnings/zone/",yi),uo.isDSTShifted=T("isDSTShifted is deprecated. See http://momentjs.com/guides/#/warnings/dst-shifted/ for more information",Mi);var po=P.prototype;function mo(t,e,n,i){var r=_n(),o=g().set(i,e);return r[n](o,t)}function go(t,e,n){if(h(t)&&(e=t,t=void 0),t=t||"",null!=e)return mo(t,e,n,"month");var i,r=[];for(i=0;i<12;i++)r[i]=mo(t,i,n,"month");return r}function yo(t,e,n,i){"boolean"===typeof t?(h(e)&&(n=e,e=void 0),e=e||""):(e=t,n=e,t=!1,h(e)&&(n=e,e=void 0),e=e||"");var r,o=_n(),a=t?o._week.dow:0,s=[];if(null!=n)return mo(e,(n+a)%7,i,"day");for(r=0;r<7;r++)s[r]=mo(e,(r+a)%7,i,"day");return s}function vo(t,e){return go(t,e,"months")}function _o(t,e){return go(t,e,"monthsShort")}function bo(t,e,n){return yo(t,e,n,"weekdays")}function xo(t,e,n){return yo(t,e,n,"weekdaysShort")}function wo(t,e,n){return yo(t,e,n,"weekdaysMin")}po.calendar=N,po.longDateFormat=q,po.invalidDate=Z,po.ordinal=Q,po.preparse=fo,po.postformat=fo,po.relativeTime=et,po.pastFuture=nt,po.set=Y,po.eras=Cr,po.erasParse=Ir,po.erasConvertYear=Ar,po.erasAbbrRegex=Rr,po.erasNameRegex=Nr,po.erasNarrowRegex=Hr,po.months=se,po.monthsShort=le,po.monthsParse=ce,po.monthsRegex=me,po.monthsShortRegex=pe,po.week=ke,po.firstDayOfYear=Ce,po.firstDayOfWeek=De,po.weekdays=We,po.weekdaysMin=Fe,po.weekdaysShort=Be,po.weekdaysParse=Ge,po.weekdaysRegex=Xe,po.weekdaysShortRegex=Ze,po.weekdaysMinRegex=Je,po.isPM=rn,po.meridiem=sn,gn("en",{eras:[{since:"0001-01-01",until:1/0,offset:1,name:"Anno Domini",narrow:"AD",abbr:"AD"},{since:"0000-12-31",until:-1/0,offset:1,name:"Before Christ",narrow:"BC",abbr:"BC"}],dayOfMonthOrdinalParse:/\d{1,2}(th|st|nd|rd)/,ordinal:function(t){var e=t%10,n=1===ht(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n}}),o.lang=T("moment.lang is deprecated. Use moment.locale instead.",gn),o.langData=T("moment.langData is deprecated. Use moment.localeData instead.",_n);var Mo=Math.abs;function So(){var t=this._data;return this._milliseconds=Mo(this._milliseconds),this._days=Mo(this._days),this._months=Mo(this._months),t.milliseconds=Mo(t.milliseconds),t.seconds=Mo(t.seconds),t.minutes=Mo(t.minutes),t.hours=Mo(t.hours),t.months=Mo(t.months),t.years=Mo(t.years),this}function Lo(t,e,n,i){var r=Ci(e,n);return t._milliseconds+=i*r._milliseconds,t._days+=i*r._days,t._months+=i*r._months,t._bubble()}function ko(t,e){return Lo(this,t,e,1)}function To(t,e){return Lo(this,t,e,-1)}function Do(t){return t<0?Math.floor(t):Math.ceil(t)}function Co(){var t,e,n,i,r,o=this._milliseconds,a=this._days,s=this._months,l=this._data;return o>=0&&a>=0&&s>=0||o<=0&&a<=0&&s<=0||(o+=864e5*Do(Ao(s)+a),a=0,s=0),l.milliseconds=o%1e3,t=dt(o/1e3),l.seconds=t%60,e=dt(t/60),l.minutes=e%60,n=dt(e/60),l.hours=n%24,a+=dt(n/24),r=dt(Io(a)),s+=r,a-=Do(Ao(r)),i=dt(s/12),s%=12,l.days=a,l.months=s,l.years=i,this}function Io(t){return 4800*t/146097}function Ao(t){return 146097*t/4800}function Yo(t){if(!this.isValid())return NaN;var e,n,i=this._milliseconds;if(t=ot(t),"month"===t||"quarter"===t||"year"===t)switch(e=this._days+i/864e5,n=this._months+Io(e),t){case"month":return n;case"quarter":return n/3;case"year":return n/12}else switch(e=this._days+Math.round(Ao(this._months)),t){case"week":return e/7+i/6048e5;case"day":return e+i/864e5;case"hour":return 24*e+i/36e5;case"minute":return 1440*e+i/6e4;case"second":return 86400*e+i/1e3;case"millisecond":return Math.floor(864e5*e)+i;default:throw new Error("Unknown unit "+t)}}function Oo(){return this.isValid()?this._milliseconds+864e5*this._days+this._months%12*2592e6+31536e6*ht(this._months/12):NaN}function Po(t){return function(){return this.as(t)}}var Eo=Po("ms"),No=Po("s"),Ro=Po("m"),Ho=Po("h"),jo=Po("d"),zo=Po("w"),Wo=Po("M"),Bo=Po("Q"),Fo=Po("y");function Vo(){return Ci(this)}function Go(t){return t=ot(t),this.isValid()?this[t+"s"]():NaN}function $o(t){return function(){return this.isValid()?this._data[t]:NaN}}var Uo=$o("milliseconds"),qo=$o("seconds"),Xo=$o("minutes"),Zo=$o("hours"),Jo=$o("days"),Ko=$o("months"),Qo=$o("years");function ta(){return dt(this.days()/7)}var ea=Math.round,na={ss:44,s:45,m:45,h:22,d:26,w:null,M:11};function ia(t,e,n,i,r){return r.relativeTime(e||1,!!n,t,i)}function ra(t,e,n,i){var r=Ci(t).abs(),o=ea(r.as("s")),a=ea(r.as("m")),s=ea(r.as("h")),l=ea(r.as("d")),u=ea(r.as("M")),c=ea(r.as("w")),d=ea(r.as("y")),h=o<=n.ss&&["s",o]||o<n.s&&["ss",o]||a<=1&&["m"]||a<n.m&&["mm",a]||s<=1&&["h"]||s<n.h&&["hh",s]||l<=1&&["d"]||l<n.d&&["dd",l];return null!=n.w&&(h=h||c<=1&&["w"]||c<n.w&&["ww",c]),h=h||u<=1&&["M"]||u<n.M&&["MM",u]||d<=1&&["y"]||["yy",d],h[2]=e,h[3]=+t>0,h[4]=i,ia.apply(null,h)}function oa(t){return void 0===t?ea:"function"===typeof t&&(ea=t,!0)}function aa(t,e){return void 0!==na[t]&&(void 0===e?na[t]:(na[t]=e,"s"===t&&(na.ss=e-1),!0))}function sa(t,e){if(!this.isValid())return this.localeData().invalidDate();var n,i,r=!1,o=na;return"object"===typeof t&&(e=t,t=!1),"boolean"===typeof t&&(r=t),"object"===typeof e&&(o=Object.assign({},na,e),null!=e.s&&null==e.ss&&(o.ss=e.s-1)),n=this.localeData(),i=ra(this,!r,o,n),r&&(i=n.pastFuture(+this,i)),n.postformat(i)}var la=Math.abs;function ua(t){return(t>0)-(t<0)||+t}function ca(){if(!this.isValid())return this.localeData().invalidDate();var t,e,n,i,r,o,a,s,l=la(this._milliseconds)/1e3,u=la(this._days),c=la(this._months),d=this.asSeconds();return d?(t=dt(l/60),e=dt(t/60),l%=60,t%=60,n=dt(c/12),c%=12,i=l?l.toFixed(3).replace(/\.?0+$/,""):"",r=d<0?"-":"",o=ua(this._months)!==ua(d)?"-":"",a=ua(this._days)!==ua(d)?"-":"",s=ua(this._milliseconds)!==ua(d)?"-":"",r+"P"+(n?o+n+"Y":"")+(c?o+c+"M":"")+(u?a+u+"D":"")+(e||t||l?"T":"")+(e?s+e+"H":"")+(t?s+t+"M":"")+(l?s+i+"S":"")):"P0D"}var da=si.prototype;return da.isValid=oi,da.abs=So,da.add=ko,da.subtract=To,da.as=Yo,da.asMilliseconds=Eo,da.asSeconds=No,da.asMinutes=Ro,da.asHours=Ho,da.asDays=jo,da.asWeeks=zo,da.asMonths=Wo,da.asQuarters=Bo,da.asYears=Fo,da.valueOf=Oo,da._bubble=Co,da.clone=Vo,da.get=Go,da.milliseconds=Uo,da.seconds=qo,da.minutes=Xo,da.hours=Zo,da.days=Jo,da.weeks=ta,da.months=Ko,da.years=Qo,da.humanize=sa,da.toISOString=ca,da.toString=ca,da.toJSON=ca,da.locale=sr,da.localeData=ur,da.toIsoString=T("toIsoString() is deprecated. Please use toISOString() instead (notice the capitals)",ca),da.lang=lr,B("X",0,0,"unix"),B("x",0,0,"valueOf"),Nt("x",At),Nt("X",Pt),Wt("X",(function(t,e,n){n._d=new Date(1e3*parseFloat(t))})),Wt("x",(function(t,e,n){n._d=new Date(ht(t))})),
//! moment.js
o.version="2.29.1",a(Zn),o.fn=uo,o.min=ti,o.max=ei,o.now=ni,o.utc=g,o.unix=co,o.months=vo,o.isDate=f,o.locale=gn,o.invalid=b,o.duration=Ci,o.isMoment=L,o.weekdays=bo,o.parseZone=ho,o.localeData=_n,o.isDuration=li,o.monthsShort=_o,o.weekdaysMin=wo,o.defineLocale=yn,o.updateLocale=vn,o.locales=bn,o.weekdaysShort=xo,o.normalizeUnits=ot,o.relativeTimeRounding=oa,o.relativeTimeThreshold=aa,o.calendarFormat=Bi,o.prototype=uo,o.HTML5_FMT={DATETIME_LOCAL:"YYYY-MM-DDTHH:mm",DATETIME_LOCAL_SECONDS:"YYYY-MM-DDTHH:mm:ss",DATETIME_LOCAL_MS:"YYYY-MM-DDTHH:mm:ss.SSS",DATE:"YYYY-MM-DD",TIME:"HH:mm",TIME_SECONDS:"HH:mm:ss",TIME_MS:"HH:mm:ss.SSS",WEEK:"GGGG-[W]WW",MONTH:"YYYY-MM"},o}))}).call(this,n("62e4")(t))},c20d:function(t,e,n){var i=n("da84"),r=n("58a8").trim,o=n("5899"),a=i.parseInt,s=/^[+-]?0[Xx]/,l=8!==a(o+"08")||22!==a(o+"0x16");t.exports=l?function(t,e){var n=r(String(t));return a(n,e>>>0||(s.test(n)?16:10))}:a},c274:function(t,e,n){"use strict";var i=n("50bf");function r(){var t={},e=0,n=0,i=0;function r(r,o){o||(o=r,r=0),r>n?n=r:r<i&&(i=r),t[r]||(t[r]=[]),t[r].push(o),e++}function o(){for(var e=i;e<=n;e++)for(var r=t[e],o=0;o<r.length;o++){var a=r[o];a()}}function a(){return e}return{add:r,process:o,size:a}}t.exports=function(t){t=t||{};var e=t.reporter,n=i.getOption(t,"async",!0),o=i.getOption(t,"auto",!0);o&&!n&&(e&&e.warn("Invalid options combination. auto=true and async=false is invalid. Setting async=true."),n=!0);var a,s=r(),l=!1;function u(t,e){!l&&o&&n&&0===s.size()&&h(),s.add(t,e)}function c(){l=!0;while(s.size()){var t=s;s=r(),t.process()}l=!1}function d(t){l||(void 0===t&&(t=n),a&&(f(a),a=null),t?h():c())}function h(){a=p(c)}function f(t){var e=clearTimeout;return e(t)}function p(t){var e=function(t){return setTimeout(t,0)};return e(t)}return{add:u,force:d}}},c2b6:function(t,e,n){var i=n("f8af"),r=n("5d89"),o=n("6f6c"),a=n("a2db"),s=n("c8fe"),l="[object Boolean]",u="[object Date]",c="[object Map]",d="[object Number]",h="[object RegExp]",f="[object Set]",p="[object String]",m="[object Symbol]",g="[object ArrayBuffer]",y="[object DataView]",v="[object Float32Array]",_="[object Float64Array]",b="[object Int8Array]",x="[object Int16Array]",w="[object Int32Array]",M="[object Uint8Array]",S="[object Uint8ClampedArray]",L="[object Uint16Array]",k="[object Uint32Array]";function T(t,e,n){var T=t.constructor;switch(e){case g:return i(t);case l:case u:return new T(+t);case y:return r(t,n);case v:case _:case b:case x:case w:case M:case S:case L:case k:return s(t,n);case c:return new T;case d:case p:return new T(t);case h:return o(t);case f:return new T;case m:return a(t)}}t.exports=T},c30c:function(t,e,n){var i=n("8eeb"),r=n("2ec1"),o=n("9934"),a=r((function(t,e,n,r){i(e,o(e),t,r)}));t.exports=a},c345:function(t,e,n){"use strict";var i=n("c532"),r=["age","authorization","content-length","content-type","etag","expires","from","host","if-modified-since","if-unmodified-since","last-modified","location","max-forwards","proxy-authorization","referer","retry-after","user-agent"];t.exports=function(t){var e,n,o,a={};return t?(i.forEach(t.split("\n"),(function(t){if(o=t.indexOf(":"),e=i.trim(t.substr(0,o)).toLowerCase(),n=i.trim(t.substr(o+1)),e){if(a[e]&&r.indexOf(e)>=0)return;a[e]="set-cookie"===e?(a[e]?a[e]:[]).concat([n]):a[e]?a[e]+", "+n:n}})),a):a}},c3f4:function(t,e,n){var i=n("5b01"),r=n("7530");function o(t,e){var n=r(t);return null==e?n:i(n,e)}t.exports=o},c3fc:function(t,e,n){var i=n("42a2"),r=n("1310"),o="[object Set]";function a(t){return r(t)&&i(t)==o}t.exports=a},c401:function(t,e,n){"use strict";var i=n("c532");t.exports=function(t,e,n){return i.forEach(n,(function(n){t=n(t,e)})),t}},c430:function(t,e){t.exports=!1},c532:function(t,e,n){"use strict";var i=n("1d2b"),r=Object.prototype.toString;function o(t){return"[object Array]"===r.call(t)}function a(t){return"undefined"===typeof t}function s(t){return null!==t&&!a(t)&&null!==t.constructor&&!a(t.constructor)&&"function"===typeof t.constructor.isBuffer&&t.constructor.isBuffer(t)}function l(t){return"[object ArrayBuffer]"===r.call(t)}function u(t){return"undefined"!==typeof FormData&&t instanceof FormData}function c(t){var e;return e="undefined"!==typeof ArrayBuffer&&ArrayBuffer.isView?ArrayBuffer.isView(t):t&&t.buffer&&t.buffer instanceof ArrayBuffer,e}function d(t){return"string"===typeof t}function h(t){return"number"===typeof t}function f(t){return null!==t&&"object"===typeof t}function p(t){if("[object Object]"!==r.call(t))return!1;var e=Object.getPrototypeOf(t);return null===e||e===Object.prototype}function m(t){return"[object Date]"===r.call(t)}function g(t){return"[object File]"===r.call(t)}function y(t){return"[object Blob]"===r.call(t)}function v(t){return"[object Function]"===r.call(t)}function _(t){return f(t)&&v(t.pipe)}function b(t){return"undefined"!==typeof URLSearchParams&&t instanceof URLSearchParams}function x(t){return t.replace(/^\s*/,"").replace(/\s*$/,"")}function w(){return("undefined"===typeof navigator||"ReactNative"!==navigator.product&&"NativeScript"!==navigator.product&&"NS"!==navigator.product)&&("undefined"!==typeof window&&"undefined"!==typeof document)}function M(t,e){if(null!==t&&"undefined"!==typeof t)if("object"!==typeof t&&(t=[t]),o(t))for(var n=0,i=t.length;n<i;n++)e.call(null,t[n],n,t);else for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.call(null,t[r],r,t)}function S(){var t={};function e(e,n){p(t[n])&&p(e)?t[n]=S(t[n],e):p(e)?t[n]=S({},e):o(e)?t[n]=e.slice():t[n]=e}for(var n=0,i=arguments.length;n<i;n++)M(arguments[n],e);return t}function L(t,e,n){return M(e,(function(e,r){t[r]=n&&"function"===typeof e?i(e,n):e})),t}function k(t){return 65279===t.charCodeAt(0)&&(t=t.slice(1)),t}t.exports={isArray:o,isArrayBuffer:l,isBuffer:s,isFormData:u,isArrayBufferView:c,isString:d,isNumber:h,isObject:f,isPlainObject:p,isUndefined:a,isDate:m,isFile:g,isBlob:y,isFunction:v,isStream:_,isURLSearchParams:b,isStandardBrowserEnv:w,forEach:M,merge:S,extend:L,trim:x,stripBOM:k}},c584:function(t,e){function n(t,e){return t.has(e)}t.exports=n},c5c1:function(t,e,n){"use strict";var i=n("6e29"),r=n("31de"),o=n("245a");i["a"].Menu=r["a"],i["a"].Item=o["a"],e["a"]=i["a"]},c6b6:function(t,e){var n={}.toString;t.exports=function(t){return n.call(t).slice(8,-1)}},c6cd:function(t,e,n){var i=n("da84"),r=n("ce4e"),o="__core-js_shared__",a=i[o]||r(o,{});t.exports=a},c6cf:function(t,e,n){var i=n("4d8c"),r=n("2286"),o=n("c1c9");function a(t){return o(r(t,void 0,i),t+"")}t.exports=a},c740:function(t,e,n){"use strict";var i=n("23e7"),r=n("b727").findIndex,o=n("44d2"),a=n("ae40"),s="findIndex",l=!0,u=a(s);s in[]&&Array(1)[s]((function(){l=!1})),i({target:"Array",proto:!0,forced:l||!u},{findIndex:function(t){return r(this,t,arguments.length>1?arguments[1]:void 0)}}),o(s)},c780:function(t,e,n){"use strict";function i(t){return void 0===t&&(t=document.body),!0===t?document.body:t instanceof window.Node?t:document.querySelector(t)}const r={inserted(t,{value:e},n){if(t.dataset&&"true"!==t.dataset.transfer)return!1;t.className=t.className?t.className+" v-transfer-dom":"v-transfer-dom";const r=t.parentNode;if(!r)return;const o=document.createComment("");let a=!1;!1!==e&&(r.replaceChild(o,t),i(e).appendChild(t),a=!0),t.__transferDomData||(t.__transferDomData={parentNode:r,home:o,target:i(e),hasMovedOut:a})},componentUpdated(t,{value:e}){if(t.dataset&&"true"!==t.dataset.transfer)return!1;const n=t.__transferDomData;if(!n)return;const r=n.parentNode,o=n.home,a=n.hasMovedOut;!a&&e?(r.replaceChild(o,t),i(e).appendChild(t),t.__transferDomData=Object.assign({},t.__transferDomData,{hasMovedOut:!0,target:i(e)})):a&&!1===e?(r.replaceChild(t,o),t.__transferDomData=Object.assign({},t.__transferDomData,{hasMovedOut:!1,target:i(e)})):e&&i(e).appendChild(t)},unbind(t){if(t.dataset&&"true"!==t.dataset.transfer)return!1;t.className=t.className.replace("v-transfer-dom","");const e=t.__transferDomData;e&&(!0===t.__transferDomData.hasMovedOut&&t.__transferDomData.parentNode&&t.__transferDomData.parentNode.appendChild(t),t.__transferDomData=null)}};e["a"]=r},c7aa:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("he",{months:"ינואר_פברואר_מרץ_אפריל_מאי_יוני_יולי_אוגוסט_ספטמבר_אוקטובר_נובמבר_דצמבר".split("_"),monthsShort:"ינו׳_פבר׳_מרץ_אפר׳_מאי_יוני_יולי_אוג׳_ספט׳_אוק׳_נוב׳_דצמ׳".split("_"),weekdays:"ראשון_שני_שלישי_רביעי_חמישי_שישי_שבת".split("_"),weekdaysShort:"א׳_ב׳_ג׳_ד׳_ה׳_ו׳_ש׳".split("_"),weekdaysMin:"א_ב_ג_ד_ה_ו_ש".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [ב]MMMM YYYY",LLL:"D [ב]MMMM YYYY HH:mm",LLLL:"dddd, D [ב]MMMM YYYY HH:mm",l:"D/M/YYYY",ll:"D MMM YYYY",lll:"D MMM YYYY HH:mm",llll:"ddd, D MMM YYYY HH:mm"},calendar:{sameDay:"[היום ב־]LT",nextDay:"[מחר ב־]LT",nextWeek:"dddd [בשעה] LT",lastDay:"[אתמול ב־]LT",lastWeek:"[ביום] dddd [האחרון בשעה] LT",sameElse:"L"},relativeTime:{future:"בעוד %s",past:"לפני %s",s:"מספר שניות",ss:"%d שניות",m:"דקה",mm:"%d דקות",h:"שעה",hh:function(t){return 2===t?"שעתיים":t+" שעות"},d:"יום",dd:function(t){return 2===t?"יומיים":t+" ימים"},M:"חודש",MM:function(t){return 2===t?"חודשיים":t+" חודשים"},y:"שנה",yy:function(t){return 2===t?"שנתיים":t%10===0&&10!==t?t+" שנה":t+" שנים"}},meridiemParse:/אחה"צ|לפנה"צ|אחרי הצהריים|לפני הצהריים|לפנות בוקר|בבוקר|בערב/i,isPM:function(t){return/^(אחה"צ|אחרי הצהריים|בערב)$/.test(t)},meridiem:function(t,e,n){return t<5?"לפנות בוקר":t<10?"בבוקר":t<12?n?'לפנה"צ':"לפני הצהריים":t<18?n?'אחה"צ':"אחרי הצהריים":"בערב"}});return e}))},c869:function(t,e,n){var i=n("0b07"),r=n("2b3e"),o=i(r,"Set");t.exports=o},c87c:function(t,e){var n=Object.prototype,i=n.hasOwnProperty;function r(t){var e=t.length,n=new t.constructor(e);return e&&"string"==typeof t[0]&&i.call(t,"index")&&(n.index=t.index,n.input=t.input),n}t.exports=r},c8af:function(t,e,n){"use strict";var i=n("c532");t.exports=function(t,e){i.forEach(t,(function(n,i){i!==e&&i.toUpperCase()===e.toUpperCase()&&(t[e]=n,delete t[i])}))}},c8ba:function(t,e){var n;n=function(){return this}();try{n=n||new Function("return this")()}catch(i){"object"===typeof window&&(n=window)}t.exports=n},c8f3:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("sq",{months:"Janar_Shkurt_Mars_Prill_Maj_Qershor_Korrik_Gusht_Shtator_Tetor_Nëntor_Dhjetor".split("_"),monthsShort:"Jan_Shk_Mar_Pri_Maj_Qer_Kor_Gus_Sht_Tet_Nën_Dhj".split("_"),weekdays:"E Diel_E Hënë_E Martë_E Mërkurë_E Enjte_E Premte_E Shtunë".split("_"),weekdaysShort:"Die_Hën_Mar_Mër_Enj_Pre_Sht".split("_"),weekdaysMin:"D_H_Ma_Më_E_P_Sh".split("_"),weekdaysParseExact:!0,meridiemParse:/PD|MD/,isPM:function(t){return"M"===t.charAt(0)},meridiem:function(t,e,n){return t<12?"PD":"MD"},longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[Sot në] LT",nextDay:"[Nesër në] LT",nextWeek:"dddd [në] LT",lastDay:"[Dje në] LT",lastWeek:"dddd [e kaluar në] LT",sameElse:"L"},relativeTime:{future:"në %s",past:"%s më parë",s:"disa sekonda",ss:"%d sekonda",m:"një minutë",mm:"%d minuta",h:"një orë",hh:"%d orë",d:"një ditë",dd:"%d ditë",M:"një muaj",MM:"%d muaj",y:"një vit",yy:"%d vite"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))},c8fe:function(t,e,n){var i=n("f8af");function r(t,e){var n=e?i(t.buffer):t.buffer;return new t.constructor(n,t.byteOffset,t.length)}t.exports=r},c946:function(t,e,n){"use strict";var i=n("b770").forEach;t.exports=function(t){t=t||{};var e=t.reporter,n=t.batchProcessor,r=t.stateHandler.getState,o=(t.stateHandler.hasState,t.idHandler);if(!n)throw new Error("Missing required dependency: batchProcessor");if(!e)throw new Error("Missing required dependency: reporter.");var a=d(),s="erd_scroll_detection_scrollbar_style",l="erd_scroll_detection_container";function u(t){h(t,s,l)}function c(e){var n=t.important?" !important; ":"; ";return(e.join(n)+n).trim()}function d(){var t=500,e=500,n=document.createElement("div");n.style.cssText=c(["position: absolute","width: "+2*t+"px","height: "+2*e+"px","visibility: hidden","margin: 0","padding: 0"]);var i=document.createElement("div");i.style.cssText=c(["position: absolute","width: "+t+"px","height: "+e+"px","overflow: scroll","visibility: none","top: "+3*-t+"px","left: "+3*-e+"px","visibility: hidden","margin: 0","padding: 0"]),i.appendChild(n),document.body.insertBefore(i,document.body.firstChild);var r=t-i.clientWidth,o=e-i.clientHeight;return document.body.removeChild(i),{width:r,height:o}}function h(t,e,n){function i(n,i){i=i||function(e){t.head.appendChild(e)};var r=t.createElement("style");return r.innerHTML=n,r.id=e,i(r),r}if(!t.getElementById(e)){var r=n+"_animation",o=n+"_animation_active",a="/* Created by the element-resize-detector library. */\n";a+="."+n+" > div::-webkit-scrollbar { "+c(["display: none"])+" }\n\n",a+="."+o+" { "+c(["-webkit-animation-duration: 0.1s","animation-duration: 0.1s","-webkit-animation-name: "+r,"animation-name: "+r])+" }\n",a+="@-webkit-keyframes "+r+" { 0% { opacity: 1; } 50% { opacity: 0; } 100% { opacity: 1; } }\n",a+="@keyframes "+r+" { 0% { opacity: 1; } 50% { opacity: 0; } 100% { opacity: 1; } }",i(a)}}function f(t){t.className+=" "+l+"_animation_active"}function p(t,n,i){if(t.addEventListener)t.addEventListener(n,i);else{if(!t.attachEvent)return e.error("[scroll] Don't know how to add event listeners.");t.attachEvent("on"+n,i)}}function m(t,n,i){if(t.removeEventListener)t.removeEventListener(n,i);else{if(!t.detachEvent)return e.error("[scroll] Don't know how to remove event listeners.");t.detachEvent("on"+n,i)}}function g(t){return r(t).container.childNodes[0].childNodes[0].childNodes[0]}function y(t){return r(t).container.childNodes[0].childNodes[0].childNodes[1]}function v(t,e){var n=r(t).listeners;if(!n.push)throw new Error("Cannot add listener to an element that is not detectable.");r(t).listeners.push(e)}function _(t,s,u){function d(){if(t.debug){var n=Array.prototype.slice.call(arguments);if(n.unshift(o.get(s),"Scroll: "),e.log.apply)e.log.apply(null,n);else for(var i=0;i<n.length;i++)e.log(n[i])}}function h(t){function e(t){return t===t.ownerDocument.body||t.ownerDocument.body.contains(t)}return!e(t)||null===window.getComputedStyle(t)}function m(t){var e=r(t).container.childNodes[0],n=window.getComputedStyle(e);return!n.width||-1===n.width.indexOf("px")}function v(){var t=window.getComputedStyle(s),e={};return e.position=t.position,e.width=s.offsetWidth,e.height=s.offsetHeight,e.top=t.top,e.right=t.right,e.bottom=t.bottom,e.left=t.left,e.widthCSS=t.width,e.heightCSS=t.height,e}function _(){var t=v();r(s).startSize={width:t.width,height:t.height},d("Element start size",r(s).startSize)}function b(){r(s).listeners=[]}function x(){if(d("storeStyle invoked."),r(s)){var t=v();r(s).style=t}else d("Aborting because element has been uninstalled")}function w(t,e,n){r(t).lastWidth=e,r(t).lastHeight=n}function M(t){return g(t).childNodes[0]}function S(){return 2*a.width+1}function L(){return 2*a.height+1}function k(t){return t+10+S()}function T(t){return t+10+L()}function D(t){return 2*t+S()}function C(t){return 2*t+L()}function I(t,e,n){var i=g(t),r=y(t),o=k(e),a=T(n),s=D(e),l=C(n);i.scrollLeft=o,i.scrollTop=a,r.scrollLeft=s,r.scrollTop=l}function A(){var t=r(s).container;if(!t){t=document.createElement("div"),t.className=l,t.style.cssText=c(["visibility: hidden","display: inline","width: 0px","height: 0px","z-index: -1","overflow: hidden","margin: 0","padding: 0"]),r(s).container=t,f(t),s.appendChild(t);var e=function(){r(s).onRendered&&r(s).onRendered()};p(t,"animationstart",e),r(s).onAnimationStart=e}return t}function Y(){function n(){var n=r(s).style;if("static"===n.position){s.style.setProperty("position","relative",t.important?"important":"");var i=function(t,e,n,i){function r(t){return t.replace(/[^-\d\.]/g,"")}var o=n[i];"auto"!==o&&"0"!==r(o)&&(t.warn("An element that is positioned static has style."+i+"="+o+" which is ignored due to the static positioning. The element will need to be positioned relative, so the style."+i+" will be set to 0. Element: ",e),e.style[i]=0)};i(e,s,n,"top"),i(e,s,n,"right"),i(e,s,n,"bottom"),i(e,s,n,"left")}}function i(t,e,n,i){return t=t?t+"px":"0",e=e?e+"px":"0",n=n?n+"px":"0",i=i?i+"px":"0",["left: "+t,"top: "+e,"right: "+i,"bottom: "+n]}if(d("Injecting elements"),r(s)){n();var o=r(s).container;o||(o=A());var u=a.width,h=a.height,f=c(["position: absolute","flex: none","overflow: hidden","z-index: -1","visibility: hidden","width: 100%","height: 100%","left: 0px","top: 0px"]),m=c(["position: absolute","flex: none","overflow: hidden","z-index: -1","visibility: hidden"].concat(i(-(1+u),-(1+h),-h,-u))),g=c(["position: absolute","flex: none","overflow: scroll","z-index: -1","visibility: hidden","width: 100%","height: 100%"]),y=c(["position: absolute","flex: none","overflow: scroll","z-index: -1","visibility: hidden","width: 100%","height: 100%"]),v=c(["position: absolute","left: 0","top: 0"]),_=c(["position: absolute","width: 200%","height: 200%"]),b=document.createElement("div"),x=document.createElement("div"),w=document.createElement("div"),M=document.createElement("div"),S=document.createElement("div"),L=document.createElement("div");b.dir="ltr",b.style.cssText=f,b.className=l,x.className=l,x.style.cssText=m,w.style.cssText=g,M.style.cssText=v,S.style.cssText=y,L.style.cssText=_,w.appendChild(M),S.appendChild(L),x.appendChild(w),x.appendChild(S),b.appendChild(x),o.appendChild(b),p(w,"scroll",k),p(S,"scroll",T),r(s).onExpandScroll=k,r(s).onShrinkScroll=T}else d("Aborting because element has been uninstalled");function k(){r(s).onExpand&&r(s).onExpand()}function T(){r(s).onShrink&&r(s).onShrink()}}function O(){function a(e,n,i){var r=M(e),o=k(n),a=T(i);r.style.setProperty("width",o+"px",t.important?"important":""),r.style.setProperty("height",a+"px",t.important?"important":"")}function l(i){var l=s.offsetWidth,c=s.offsetHeight,h=l!==r(s).lastWidth||c!==r(s).lastHeight;d("Storing current size",l,c),w(s,l,c),n.add(0,(function(){if(h)if(r(s))if(u()){if(t.debug){var n=s.offsetWidth,i=s.offsetHeight;n===l&&i===c||e.warn(o.get(s),"Scroll: Size changed before updating detector elements.")}a(s,l,c)}else d("Aborting because element container has not been initialized");else d("Aborting because element has been uninstalled")})),n.add(1,(function(){r(s)?u()?I(s,l,c):d("Aborting because element container has not been initialized"):d("Aborting because element has been uninstalled")})),h&&i&&n.add(2,(function(){r(s)?u()?i():d("Aborting because element container has not been initialized"):d("Aborting because element has been uninstalled")}))}function u(){return!!r(s).container}function c(){function t(){return void 0===r(s).lastNotifiedWidth}d("notifyListenersIfNeeded invoked");var e=r(s);return t()&&e.lastWidth===e.startSize.width&&e.lastHeight===e.startSize.height?d("Not notifying: Size is the same as the start size, and there has been no notification yet."):e.lastWidth===e.lastNotifiedWidth&&e.lastHeight===e.lastNotifiedHeight?d("Not notifying: Size already notified"):(d("Current size not notified, notifying..."),e.lastNotifiedWidth=e.lastWidth,e.lastNotifiedHeight=e.lastHeight,void i(r(s).listeners,(function(t){t(s)})))}function h(){if(d("startanimation triggered."),m(s))d("Ignoring since element is still unrendered...");else{d("Element rendered.");var t=g(s),e=y(s);0!==t.scrollLeft&&0!==t.scrollTop&&0!==e.scrollLeft&&0!==e.scrollTop||(d("Scrollbars out of sync. Updating detector elements..."),l(c))}}function f(){d("Scroll detected."),m(s)?d("Scroll event fired while unrendered. Ignoring..."):l(c)}if(d("registerListenersAndPositionElements invoked."),r(s)){r(s).onRendered=h,r(s).onExpand=f,r(s).onShrink=f;var p=r(s).style;a(s,p.width,p.height)}else d("Aborting because element has been uninstalled")}function P(){if(d("finalizeDomMutation invoked."),r(s)){var t=r(s).style;w(s,t.width,t.height),I(s,t.width,t.height)}else d("Aborting because element has been uninstalled")}function E(){u(s)}function N(){d("Installing..."),b(),_(),n.add(0,x),n.add(1,Y),n.add(2,O),n.add(3,P),n.add(4,E)}u||(u=s,s=t,t=null),t=t||{},d("Making detectable..."),h(s)?(d("Element is detached"),A(),d("Waiting until element is attached..."),r(s).onRendered=function(){d("Element is now attached"),N()}):N()}function b(t){var e=r(t);e&&(e.onExpandScroll&&m(g(t),"scroll",e.onExpandScroll),e.onShrinkScroll&&m(y(t),"scroll",e.onShrinkScroll),e.onAnimationStart&&m(e.container,"animationstart",e.onAnimationStart),e.container&&t.removeChild(e.container))}return u(window.document),{makeDetectable:_,addListener:v,uninstall:b,initDocument:u}}},c975:function(t,e,n){"use strict";var i=n("23e7"),r=n("4d64").indexOf,o=n("a640"),a=n("ae40"),s=[].indexOf,l=!!s&&1/[1].indexOf(1,-0)<0,u=o("indexOf"),c=a("indexOf",{ACCESSORS:!0,1:0});i({target:"Array",proto:!0,forced:l||!u||!c},{indexOf:function(t){return l?s.apply(this,arguments)||0:r(this,t,arguments.length>1?arguments[1]:void 0)}})},ca7c:function(t,e,n){var i=n("4be8"),r=n("1304");function o(t,e){return t&&i(t,r(e))}t.exports=o},ca84:function(t,e,n){var i=n("5135"),r=n("fc6a"),o=n("4d64").indexOf,a=n("d012");t.exports=function(t,e){var n,s=r(t),l=0,u=[];for(n in s)!i(a,n)&&i(s,n)&&u.push(n);while(e.length>l)i(s,n=e[l++])&&(~o(u,n)||u.push(n));return u}},caad:function(t,e,n){"use strict";var i=n("23e7"),r=n("4d64").includes,o=n("44d2"),a=n("ae40"),s=a("indexOf",{ACCESSORS:!0,1:0});i({target:"Array",proto:!0,forced:!s},{includes:function(t){return r(this,t,arguments.length>1?arguments[1]:void 0)}}),o("includes")},cb5a:function(t,e,n){var i=n("9638");function r(t,e){var n=t.length;while(n--)if(i(t[n][0],e))return n;return-1}t.exports=r},cc12:function(t,e,n){var i=n("da84"),r=n("861d"),o=i.document,a=r(o)&&r(o.createElement);t.exports=function(t){return a?o.createElement(t):{}}},cc45:function(t,e,n){var i=n("1a2d"),r=n("b047"),o=n("99d3"),a=o&&o.isMap,s=a?r(a):i;t.exports=s},cca6:function(t,e,n){var i=n("23e7"),r=n("60da");i({target:"Object",stat:!0,forced:Object.assign!==r},{assign:r})},cd9d:function(t,e){function n(t){return t}t.exports=n},cdd8:function(t,e,n){t.exports=n("a9b9")},cdf9:function(t,e,n){var i=n("825a"),r=n("861d"),o=n("f069");t.exports=function(t,e){if(i(t),r(e)&&e.constructor===t)return e;var n=o.f(t),a=n.resolve;return a(e),n.promise}},ce4e:function(t,e,n){var i=n("da84"),r=n("9112");t.exports=function(t,e){try{r(i,t,e)}catch(n){i[t]=e}return e}},ce86:function(t,e,n){var i=n("9e69"),r=n("7948"),o=n("6747"),a=n("ffd6"),s=1/0,l=i?i.prototype:void 0,u=l?l.toString:void 0;function c(t){if("string"==typeof t)return t;if(o(t))return r(t,c)+"";if(a(t))return u?u.call(t):"";var e=t+"";return"0"==e&&1/t==-s?"-0":e}t.exports=c},cebd:function(t,e){function n(t){var e=-1,n=Array(t.size);return t.forEach((function(t){n[++e]=[t,t]})),n}t.exports=n},cee4:function(t,e,n){"use strict";var i=n("c532"),r=n("1d2b"),o=n("0a06"),a=n("4a7b"),s=n("2444");function l(t){var e=new o(t),n=r(o.prototype.request,e);return i.extend(n,o.prototype,e),i.extend(n,e),n}var u=l(s);u.Axios=o,u.create=function(t){return l(a(u.defaults,t))},u.Cancel=n("7a77"),u.CancelToken=n("8df4b"),u.isCancel=n("2e67"),u.all=function(t){return Promise.all(t)},u.spread=n("0df6"),u.isAxiosError=n("5f02"),t.exports=u,t.exports.default=u},cf1e:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={words:{ss:["sekunda","sekunde","sekundi"],m:["jedan minut","jedne minute"],mm:["minut","minute","minuta"],h:["jedan sat","jednog sata"],hh:["sat","sata","sati"],dd:["dan","dana","dana"],MM:["mesec","meseca","meseci"],yy:["godina","godine","godina"]},correctGrammaticalCase:function(t,e){return 1===t?e[0]:t>=2&&t<=4?e[1]:e[2]},translate:function(t,n,i){var r=e.words[i];return 1===i.length?n?r[0]:r[1]:t+" "+e.correctGrammaticalCase(t,r)}},n=t.defineLocale("sr",{months:"januar_februar_mart_april_maj_jun_jul_avgust_septembar_oktobar_novembar_decembar".split("_"),monthsShort:"jan._feb._mar._apr._maj_jun_jul_avg._sep._okt._nov._dec.".split("_"),monthsParseExact:!0,weekdays:"nedelja_ponedeljak_utorak_sreda_četvrtak_petak_subota".split("_"),weekdaysShort:"ned._pon._uto._sre._čet._pet._sub.".split("_"),weekdaysMin:"ne_po_ut_sr_če_pe_su".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"D. M. YYYY.",LL:"D. MMMM YYYY.",LLL:"D. MMMM YYYY. H:mm",LLLL:"dddd, D. MMMM YYYY. H:mm"},calendar:{sameDay:"[danas u] LT",nextDay:"[sutra u] LT",nextWeek:function(){switch(this.day()){case 0:return"[u] [nedelju] [u] LT";case 3:return"[u] [sredu] [u] LT";case 6:return"[u] [subotu] [u] LT";case 1:case 2:case 4:case 5:return"[u] dddd [u] LT"}},lastDay:"[juče u] LT",lastWeek:function(){var t=["[prošle] [nedelje] [u] LT","[prošlog] [ponedeljka] [u] LT","[prošlog] [utorka] [u] LT","[prošle] [srede] [u] LT","[prošlog] [četvrtka] [u] LT","[prošlog] [petka] [u] LT","[prošle] [subote] [u] LT"];return t[this.day()]},sameElse:"L"},relativeTime:{future:"za %s",past:"pre %s",s:"nekoliko sekundi",ss:e.translate,m:e.translate,mm:e.translate,h:e.translate,hh:e.translate,d:"dan",dd:e.translate,M:"mesec",MM:e.translate,y:"godinu",yy:e.translate},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:7}});return n}))},cf51:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("tzl",{months:"Januar_Fevraglh_Març_Avrïu_Mai_Gün_Julia_Guscht_Setemvar_Listopäts_Noemvar_Zecemvar".split("_"),monthsShort:"Jan_Fev_Mar_Avr_Mai_Gün_Jul_Gus_Set_Lis_Noe_Zec".split("_"),weekdays:"Súladi_Lúneçi_Maitzi_Márcuri_Xhúadi_Viénerçi_Sáturi".split("_"),weekdaysShort:"Súl_Lún_Mai_Már_Xhú_Vié_Sát".split("_"),weekdaysMin:"Sú_Lú_Ma_Má_Xh_Vi_Sá".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD.MM.YYYY",LL:"D. MMMM [dallas] YYYY",LLL:"D. MMMM [dallas] YYYY HH.mm",LLLL:"dddd, [li] D. MMMM [dallas] YYYY HH.mm"},meridiemParse:/d\'o|d\'a/i,isPM:function(t){return"d'o"===t.toLowerCase()},meridiem:function(t,e,n){return t>11?n?"d'o":"D'O":n?"d'a":"D'A"},calendar:{sameDay:"[oxhi à] LT",nextDay:"[demà à] LT",nextWeek:"dddd [à] LT",lastDay:"[ieiri à] LT",lastWeek:"[sür el] dddd [lasteu à] LT",sameElse:"L"},relativeTime:{future:"osprei %s",past:"ja%s",s:n,ss:n,m:n,mm:n,h:n,hh:n,d:n,dd:n,M:n,MM:n,y:n,yy:n},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});function n(t,e,n,i){var r={s:["viensas secunds","'iensas secunds"],ss:[t+" secunds",t+" secunds"],m:["'n míut","'iens míut"],mm:[t+" míuts",t+" míuts"],h:["'n þora","'iensa þora"],hh:[t+" þoras",t+" þoras"],d:["'n ziua","'iensa ziua"],dd:[t+" ziuas",t+" ziuas"],M:["'n mes","'iens mes"],MM:[t+" mesen",t+" mesen"],y:["'n ar","'iens ar"],yy:[t+" ars",t+" ars"]};return i||e?r[n][0]:r[n][1]}return e}))},cf75:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="pagh_wa’_cha’_wej_loS_vagh_jav_Soch_chorgh_Hut".split("_");function n(t){var e=t;return e=-1!==t.indexOf("jaj")?e.slice(0,-3)+"leS":-1!==t.indexOf("jar")?e.slice(0,-3)+"waQ":-1!==t.indexOf("DIS")?e.slice(0,-3)+"nem":e+" pIq",e}function i(t){var e=t;return e=-1!==t.indexOf("jaj")?e.slice(0,-3)+"Hu’":-1!==t.indexOf("jar")?e.slice(0,-3)+"wen":-1!==t.indexOf("DIS")?e.slice(0,-3)+"ben":e+" ret",e}function r(t,e,n,i){var r=o(t);switch(n){case"ss":return r+" lup";case"mm":return r+" tup";case"hh":return r+" rep";case"dd":return r+" jaj";case"MM":return r+" jar";case"yy":return r+" DIS"}}function o(t){var n=Math.floor(t%1e3/100),i=Math.floor(t%100/10),r=t%10,o="";return n>0&&(o+=e[n]+"vatlh"),i>0&&(o+=(""!==o?" ":"")+e[i]+"maH"),r>0&&(o+=(""!==o?" ":"")+e[r]),""===o?"pagh":o}var a=t.defineLocale("tlh",{months:"tera’ jar wa’_tera’ jar cha’_tera’ jar wej_tera’ jar loS_tera’ jar vagh_tera’ jar jav_tera’ jar Soch_tera’ jar chorgh_tera’ jar Hut_tera’ jar wa’maH_tera’ jar wa’maH wa’_tera’ jar wa’maH cha’".split("_"),monthsShort:"jar wa’_jar cha’_jar wej_jar loS_jar vagh_jar jav_jar Soch_jar chorgh_jar Hut_jar wa’maH_jar wa’maH wa’_jar wa’maH cha’".split("_"),monthsParseExact:!0,weekdays:"lojmItjaj_DaSjaj_povjaj_ghItlhjaj_loghjaj_buqjaj_ghInjaj".split("_"),weekdaysShort:"lojmItjaj_DaSjaj_povjaj_ghItlhjaj_loghjaj_buqjaj_ghInjaj".split("_"),weekdaysMin:"lojmItjaj_DaSjaj_povjaj_ghItlhjaj_loghjaj_buqjaj_ghInjaj".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[DaHjaj] LT",nextDay:"[wa’leS] LT",nextWeek:"LLL",lastDay:"[wa’Hu’] LT",lastWeek:"LLL",sameElse:"L"},relativeTime:{future:n,past:i,s:"puS lup",ss:r,m:"wa’ tup",mm:r,h:"wa’ rep",hh:r,d:"wa’ jaj",dd:r,M:"wa’ jar",MM:r,y:"wa’ DIS",yy:r},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return a}))},d012:function(t,e){t.exports={}},d02c:function(t,e,n){var i=n("5e2e"),r=n("79bc"),o=n("7b83"),a=200;function s(t,e){var n=this.__data__;if(n instanceof i){var s=n.__data__;if(!r||s.length<a-1)return s.push([t,e]),this.size=++n.size,this;n=this.__data__=new o(s)}return n.set(t,e),this.size=n.size,this}t.exports=s},d039:function(t,e){t.exports=function(t){try{return!!t()}catch(e){return!0}}},d066:function(t,e,n){var i=n("428f"),r=n("da84"),o=function(t){return"function"==typeof t?t:void 0};t.exports=function(t,e){return arguments.length<2?o(i[t])||o(r[t]):i[t]&&i[t][e]||r[t]&&r[t][e]}},d1e7:function(t,e,n){"use strict";var i={}.propertyIsEnumerable,r=Object.getOwnPropertyDescriptor,o=r&&!i.call({1:2},1);e.f=o?function(t){var e=r(this,t);return!!e&&e.enumerable}:i},d26a:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"༡",2:"༢",3:"༣",4:"༤",5:"༥",6:"༦",7:"༧",8:"༨",9:"༩",0:"༠"},n={"༡":"1","༢":"2","༣":"3","༤":"4","༥":"5","༦":"6","༧":"7","༨":"8","༩":"9","༠":"0"},i=t.defineLocale("bo",{months:"ཟླ་བ་དང་པོ_ཟླ་བ་གཉིས་པ_ཟླ་བ་གསུམ་པ_ཟླ་བ་བཞི་པ_ཟླ་བ་ལྔ་པ_ཟླ་བ་དྲུག་པ_ཟླ་བ་བདུན་པ_ཟླ་བ་བརྒྱད་པ_ཟླ་བ་དགུ་པ_ཟླ་བ་བཅུ་པ_ཟླ་བ་བཅུ་གཅིག་པ_ཟླ་བ་བཅུ་གཉིས་པ".split("_"),monthsShort:"ཟླ་1_ཟླ་2_ཟླ་3_ཟླ་4_ཟླ་5_ཟླ་6_ཟླ་7_ཟླ་8_ཟླ་9_ཟླ་10_ཟླ་11_ཟླ་12".split("_"),monthsShortRegex:/^(ཟླ་\d{1,2})/,monthsParseExact:!0,weekdays:"གཟའ་ཉི་མ་_གཟའ་ཟླ་བ་_གཟའ་མིག་དམར་_གཟའ་ལྷག་པ་_གཟའ་ཕུར་བུ_གཟའ་པ་སངས་_གཟའ་སྤེན་པ་".split("_"),weekdaysShort:"ཉི་མ་_ཟླ་བ་_མིག་དམར་_ལྷག་པ་_ཕུར་བུ_པ་སངས་_སྤེན་པ་".split("_"),weekdaysMin:"ཉི_ཟླ_མིག_ལྷག_ཕུར_སངས_སྤེན".split("_"),longDateFormat:{LT:"A h:mm",LTS:"A h:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm",LLLL:"dddd, D MMMM YYYY, A h:mm"},calendar:{sameDay:"[དི་རིང] LT",nextDay:"[སང་ཉིན] LT",nextWeek:"[བདུན་ཕྲག་རྗེས་མ], LT",lastDay:"[ཁ་སང] LT",lastWeek:"[བདུན་ཕྲག་མཐའ་མ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s ལ་",past:"%s སྔན་ལ",s:"ལམ་སང",ss:"%d སྐར་ཆ།",m:"སྐར་མ་གཅིག",mm:"%d སྐར་མ",h:"ཆུ་ཚོད་གཅིག",hh:"%d ཆུ་ཚོད",d:"ཉིན་གཅིག",dd:"%d ཉིན་",M:"ཟླ་བ་གཅིག",MM:"%d ཟླ་བ",y:"ལོ་གཅིག",yy:"%d ལོ"},preparse:function(t){return t.replace(/[༡༢༣༤༥༦༧༨༩༠]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/མཚན་མོ|ཞོགས་ཀས|ཉིན་གུང|དགོང་དག|མཚན་མོ/,meridiemHour:function(t,e){return 12===t&&(t=0),"མཚན་མོ"===e&&t>=4||"ཉིན་གུང"===e&&t<5||"དགོང་དག"===e?t+12:t},meridiem:function(t,e,n){return t<4?"མཚན་མོ":t<10?"ཞོགས་ཀས":t<17?"ཉིན་གུང":t<20?"དགོང་དག":"མཚན་མོ"},week:{dow:0,doy:6}});return i}))},d28b:function(t,e,n){var i=n("746f");i("iterator")},d2bb:function(t,e,n){var i=n("825a"),r=n("3bbe");t.exports=Object.setPrototypeOf||("__proto__"in{}?function(){var t,e=!1,n={};try{t=Object.getOwnPropertyDescriptor(Object.prototype,"__proto__").set,t.call(n,[]),e=n instanceof Array}catch(o){}return function(n,o){return i(n),r(o),e?t.call(n,o):n.__proto__=o,n}}():void 0)},d2d4:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("pt-br",{months:"janeiro_fevereiro_março_abril_maio_junho_julho_agosto_setembro_outubro_novembro_dezembro".split("_"),monthsShort:"jan_fev_mar_abr_mai_jun_jul_ago_set_out_nov_dez".split("_"),weekdays:"domingo_segunda-feira_terça-feira_quarta-feira_quinta-feira_sexta-feira_sábado".split("_"),weekdaysShort:"dom_seg_ter_qua_qui_sex_sáb".split("_"),weekdaysMin:"do_2ª_3ª_4ª_5ª_6ª_sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY [às] HH:mm",LLLL:"dddd, D [de] MMMM [de] YYYY [às] HH:mm"},calendar:{sameDay:"[Hoje às] LT",nextDay:"[Amanhã às] LT",nextWeek:"dddd [às] LT",lastDay:"[Ontem às] LT",lastWeek:function(){return 0===this.day()||6===this.day()?"[Último] dddd [às] LT":"[Última] dddd [às] LT"},sameElse:"L"},relativeTime:{future:"em %s",past:"há %s",s:"poucos segundos",ss:"%d segundos",m:"um minuto",mm:"%d minutos",h:"uma hora",hh:"%d horas",d:"um dia",dd:"%d dias",M:"um mês",MM:"%d meses",y:"um ano",yy:"%d anos"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",invalidDate:"Data inválida"});return e}))},d2e4:function(t,e,n){(function(e){
/**!
 * @fileOverview Kickass library to create and place poppers near their reference elements.
 * @version 1.16.1
 * @license
 * Copyright (c) 2016 Federico Zivolo and contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
(function(e,n){t.exports=n()})(0,(function(){"use strict";var t="undefined"!==typeof window&&"undefined"!==typeof document&&"undefined"!==typeof navigator,n=function(){for(var e=["Edge","Trident","Firefox"],n=0;n<e.length;n+=1)if(t&&navigator.userAgent.indexOf(e[n])>=0)return 1;return 0}();function i(t){var e=!1;return function(){e||(e=!0,window.Promise.resolve().then((function(){e=!1,t()})))}}function r(t){var e=!1;return function(){e||(e=!0,setTimeout((function(){e=!1,t()}),n))}}var o=t&&window.Promise,a=o?i:r;function s(t){var e={};return t&&"[object Function]"===e.toString.call(t)}function l(t,e){if(1!==t.nodeType)return[];var n=t.ownerDocument.defaultView,i=n.getComputedStyle(t,null);return e?i[e]:i}function u(t){return"HTML"===t.nodeName?t:t.parentNode||t.host}function c(t){if(!t)return document.body;switch(t.nodeName){case"HTML":case"BODY":return t.ownerDocument.body;case"#document":return t.body}var e=l(t),n=e.overflow,i=e.overflowX,r=e.overflowY;return/(auto|scroll|overlay)/.test(n+r+i)?t:c(u(t))}function d(t){return t&&t.referenceNode?t.referenceNode:t}var h=t&&!(!window.MSInputMethodContext||!document.documentMode),f=t&&/MSIE 10/.test(navigator.userAgent);function p(t){return 11===t?h:10===t?f:h||f}function m(t){if(!t)return document.documentElement;var e=p(10)?document.body:null,n=t.offsetParent||null;while(n===e&&t.nextElementSibling)n=(t=t.nextElementSibling).offsetParent;var i=n&&n.nodeName;return i&&"BODY"!==i&&"HTML"!==i?-1!==["TH","TD","TABLE"].indexOf(n.nodeName)&&"static"===l(n,"position")?m(n):n:t?t.ownerDocument.documentElement:document.documentElement}function g(t){var e=t.nodeName;return"BODY"!==e&&("HTML"===e||m(t.firstElementChild)===t)}function y(t){return null!==t.parentNode?y(t.parentNode):t}function v(t,e){if(!t||!t.nodeType||!e||!e.nodeType)return document.documentElement;var n=t.compareDocumentPosition(e)&Node.DOCUMENT_POSITION_FOLLOWING,i=n?t:e,r=n?e:t,o=document.createRange();o.setStart(i,0),o.setEnd(r,0);var a=o.commonAncestorContainer;if(t!==a&&e!==a||i.contains(r))return g(a)?a:m(a);var s=y(t);return s.host?v(s.host,e):v(t,y(e).host)}function _(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:"top",n="top"===e?"scrollTop":"scrollLeft",i=t.nodeName;if("BODY"===i||"HTML"===i){var r=t.ownerDocument.documentElement,o=t.ownerDocument.scrollingElement||r;return o[n]}return t[n]}function b(t,e){var n=arguments.length>2&&void 0!==arguments[2]&&arguments[2],i=_(e,"top"),r=_(e,"left"),o=n?-1:1;return t.top+=i*o,t.bottom+=i*o,t.left+=r*o,t.right+=r*o,t}function x(t,e){var n="x"===e?"Left":"Top",i="Left"===n?"Right":"Bottom";return parseFloat(t["border"+n+"Width"])+parseFloat(t["border"+i+"Width"])}function w(t,e,n,i){return Math.max(e["offset"+t],e["scroll"+t],n["client"+t],n["offset"+t],n["scroll"+t],p(10)?parseInt(n["offset"+t])+parseInt(i["margin"+("Height"===t?"Top":"Left")])+parseInt(i["margin"+("Height"===t?"Bottom":"Right")]):0)}function M(t){var e=t.body,n=t.documentElement,i=p(10)&&getComputedStyle(n);return{height:w("Height",e,n,i),width:w("Width",e,n,i)}}var S=function(t,e){if(!(t instanceof e))throw new TypeError("Cannot call a class as a function")},L=function(){function t(t,e){for(var n=0;n<e.length;n++){var i=e[n];i.enumerable=i.enumerable||!1,i.configurable=!0,"value"in i&&(i.writable=!0),Object.defineProperty(t,i.key,i)}}return function(e,n,i){return n&&t(e.prototype,n),i&&t(e,i),e}}(),k=function(t,e,n){return e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t},T=Object.assign||function(t){for(var e=1;e<arguments.length;e++){var n=arguments[e];for(var i in n)Object.prototype.hasOwnProperty.call(n,i)&&(t[i]=n[i])}return t};function D(t){return T({},t,{right:t.left+t.width,bottom:t.top+t.height})}function C(t){var e={};try{if(p(10)){e=t.getBoundingClientRect();var n=_(t,"top"),i=_(t,"left");e.top+=n,e.left+=i,e.bottom+=n,e.right+=i}else e=t.getBoundingClientRect()}catch(h){}var r={left:e.left,top:e.top,width:e.right-e.left,height:e.bottom-e.top},o="HTML"===t.nodeName?M(t.ownerDocument):{},a=o.width||t.clientWidth||r.width,s=o.height||t.clientHeight||r.height,u=t.offsetWidth-a,c=t.offsetHeight-s;if(u||c){var d=l(t);u-=x(d,"x"),c-=x(d,"y"),r.width-=u,r.height-=c}return D(r)}function I(t,e){var n=arguments.length>2&&void 0!==arguments[2]&&arguments[2],i=p(10),r="HTML"===e.nodeName,o=C(t),a=C(e),s=c(t),u=l(e),d=parseFloat(u.borderTopWidth),h=parseFloat(u.borderLeftWidth);n&&r&&(a.top=Math.max(a.top,0),a.left=Math.max(a.left,0));var f=D({top:o.top-a.top-d,left:o.left-a.left-h,width:o.width,height:o.height});if(f.marginTop=0,f.marginLeft=0,!i&&r){var m=parseFloat(u.marginTop),g=parseFloat(u.marginLeft);f.top-=d-m,f.bottom-=d-m,f.left-=h-g,f.right-=h-g,f.marginTop=m,f.marginLeft=g}return(i&&!n?e.contains(s):e===s&&"BODY"!==s.nodeName)&&(f=b(f,e)),f}function A(t){var e=arguments.length>1&&void 0!==arguments[1]&&arguments[1],n=t.ownerDocument.documentElement,i=I(t,n),r=Math.max(n.clientWidth,window.innerWidth||0),o=Math.max(n.clientHeight,window.innerHeight||0),a=e?0:_(n),s=e?0:_(n,"left"),l={top:a-i.top+i.marginTop,left:s-i.left+i.marginLeft,width:r,height:o};return D(l)}function Y(t){var e=t.nodeName;if("BODY"===e||"HTML"===e)return!1;if("fixed"===l(t,"position"))return!0;var n=u(t);return!!n&&Y(n)}function O(t){if(!t||!t.parentElement||p())return document.documentElement;var e=t.parentElement;while(e&&"none"===l(e,"transform"))e=e.parentElement;return e||document.documentElement}function P(t,e,n,i){var r=arguments.length>4&&void 0!==arguments[4]&&arguments[4],o={top:0,left:0},a=r?O(t):v(t,d(e));if("viewport"===i)o=A(a,r);else{var s=void 0;"scrollParent"===i?(s=c(u(e)),"BODY"===s.nodeName&&(s=t.ownerDocument.documentElement)):s="window"===i?t.ownerDocument.documentElement:i;var l=I(s,a,r);if("HTML"!==s.nodeName||Y(a))o=l;else{var h=M(t.ownerDocument),f=h.height,p=h.width;o.top+=l.top-l.marginTop,o.bottom=f+l.top,o.left+=l.left-l.marginLeft,o.right=p+l.left}}n=n||0;var m="number"===typeof n;return o.left+=m?n:n.left||0,o.top+=m?n:n.top||0,o.right-=m?n:n.right||0,o.bottom-=m?n:n.bottom||0,o}function E(t){var e=t.width,n=t.height;return e*n}function N(t,e,n,i,r){var o=arguments.length>5&&void 0!==arguments[5]?arguments[5]:0;if(-1===t.indexOf("auto"))return t;var a=P(n,i,o,r),s={top:{width:a.width,height:e.top-a.top},right:{width:a.right-e.right,height:a.height},bottom:{width:a.width,height:a.bottom-e.bottom},left:{width:e.left-a.left,height:a.height}},l=Object.keys(s).map((function(t){return T({key:t},s[t],{area:E(s[t])})})).sort((function(t,e){return e.area-t.area})),u=l.filter((function(t){var e=t.width,i=t.height;return e>=n.clientWidth&&i>=n.clientHeight})),c=u.length>0?u[0].key:l[0].key,d=t.split("-")[1];return c+(d?"-"+d:"")}function R(t,e,n){var i=arguments.length>3&&void 0!==arguments[3]?arguments[3]:null,r=i?O(e):v(e,d(n));return I(n,r,i)}function H(t){var e=t.ownerDocument.defaultView,n=e.getComputedStyle(t),i=parseFloat(n.marginTop||0)+parseFloat(n.marginBottom||0),r=parseFloat(n.marginLeft||0)+parseFloat(n.marginRight||0),o={width:t.offsetWidth+r,height:t.offsetHeight+i};return o}function j(t){var e={left:"right",right:"left",bottom:"top",top:"bottom"};return t.replace(/left|right|bottom|top/g,(function(t){return e[t]}))}function z(t,e,n){n=n.split("-")[0];var i=H(t),r={width:i.width,height:i.height},o=-1!==["right","left"].indexOf(n),a=o?"top":"left",s=o?"left":"top",l=o?"height":"width",u=o?"width":"height";return r[a]=e[a]+e[l]/2-i[l]/2,r[s]=n===s?e[s]-i[u]:e[j(s)],r}function W(t,e){return Array.prototype.find?t.find(e):t.filter(e)[0]}function B(t,e,n){if(Array.prototype.findIndex)return t.findIndex((function(t){return t[e]===n}));var i=W(t,(function(t){return t[e]===n}));return t.indexOf(i)}function F(t,e,n){var i=void 0===n?t:t.slice(0,B(t,"name",n));return i.forEach((function(t){t["function"];var n=t["function"]||t.fn;t.enabled&&s(n)&&(e.offsets.popper=D(e.offsets.popper),e.offsets.reference=D(e.offsets.reference),e=n(e,t))})),e}function V(){if(!this.state.isDestroyed){var t={instance:this,styles:{},arrowStyles:{},attributes:{},flipped:!1,offsets:{}};t.offsets.reference=R(this.state,this.popper,this.reference,this.options.positionFixed),t.placement=N(this.options.placement,t.offsets.reference,this.popper,this.reference,this.options.modifiers.flip.boundariesElement,this.options.modifiers.flip.padding),t.originalPlacement=t.placement,t.positionFixed=this.options.positionFixed,t.offsets.popper=z(this.popper,t.offsets.reference,t.placement),t.offsets.popper.position=this.options.positionFixed?"fixed":"absolute",t=F(this.modifiers,t),this.state.isCreated?this.options.onUpdate(t):(this.state.isCreated=!0,this.options.onCreate(t))}}function G(t,e){return t.some((function(t){var n=t.name,i=t.enabled;return i&&n===e}))}function $(t){for(var e=[!1,"ms","Webkit","Moz","O"],n=t.charAt(0).toUpperCase()+t.slice(1),i=0;i<e.length;i++){var r=e[i],o=r?""+r+n:t;if("undefined"!==typeof document.body.style[o])return o}return null}function U(){return this.state.isDestroyed=!0,G(this.modifiers,"applyStyle")&&(this.popper.removeAttribute("x-placement"),this.popper.style.position="",this.popper.style.top="",this.popper.style.left="",this.popper.style.right="",this.popper.style.bottom="",this.popper.style.willChange="",this.popper.style[$("transform")]=""),this.disableEventListeners(),this.options.removeOnDestroy&&this.popper.parentNode.removeChild(this.popper),this}function q(t){var e=t.ownerDocument;return e?e.defaultView:window}function X(t,e,n,i){var r="BODY"===t.nodeName,o=r?t.ownerDocument.defaultView:t;o.addEventListener(e,n,{passive:!0}),r||X(c(o.parentNode),e,n,i),i.push(o)}function Z(t,e,n,i){n.updateBound=i,q(t).addEventListener("resize",n.updateBound,{passive:!0});var r=c(t);return X(r,"scroll",n.updateBound,n.scrollParents),n.scrollElement=r,n.eventsEnabled=!0,n}function J(){this.state.eventsEnabled||(this.state=Z(this.reference,this.options,this.state,this.scheduleUpdate))}function K(t,e){return q(t).removeEventListener("resize",e.updateBound),e.scrollParents.forEach((function(t){t.removeEventListener("scroll",e.updateBound)})),e.updateBound=null,e.scrollParents=[],e.scrollElement=null,e.eventsEnabled=!1,e}function Q(){this.state.eventsEnabled&&(cancelAnimationFrame(this.scheduleUpdate),this.state=K(this.reference,this.state))}function tt(t){return""!==t&&!isNaN(parseFloat(t))&&isFinite(t)}function et(t,e){Object.keys(e).forEach((function(n){var i="";-1!==["width","height","top","right","bottom","left"].indexOf(n)&&tt(e[n])&&(i="px"),t.style[n]=e[n]+i}))}function nt(t,e){Object.keys(e).forEach((function(n){var i=e[n];!1!==i?t.setAttribute(n,e[n]):t.removeAttribute(n)}))}function it(t){return et(t.instance.popper,t.styles),nt(t.instance.popper,t.attributes),t.arrowElement&&Object.keys(t.arrowStyles).length&&et(t.arrowElement,t.arrowStyles),t}function rt(t,e,n,i,r){var o=R(r,e,t,n.positionFixed),a=N(n.placement,o,e,t,n.modifiers.flip.boundariesElement,n.modifiers.flip.padding);return e.setAttribute("x-placement",a),et(e,{position:n.positionFixed?"fixed":"absolute"}),n}function ot(t,e){var n=t.offsets,i=n.popper,r=n.reference,o=Math.round,a=Math.floor,s=function(t){return t},l=o(r.width),u=o(i.width),c=-1!==["left","right"].indexOf(t.placement),d=-1!==t.placement.indexOf("-"),h=l%2===u%2,f=l%2===1&&u%2===1,p=e?c||d||h?o:a:s,m=e?o:s;return{left:p(f&&!d&&e?i.left-1:i.left),top:m(i.top),bottom:m(i.bottom),right:p(i.right)}}var at=t&&/Firefox/i.test(navigator.userAgent);function st(t,e){var n=e.x,i=e.y,r=t.offsets.popper,o=W(t.instance.modifiers,(function(t){return"applyStyle"===t.name})).gpuAcceleration,a=void 0!==o?o:e.gpuAcceleration,s=m(t.instance.popper),l=C(s),u={position:r.position},c=ot(t,window.devicePixelRatio<2||!at),d="bottom"===n?"top":"bottom",h="right"===i?"left":"right",f=$("transform"),p=void 0,g=void 0;if(g="bottom"===d?"HTML"===s.nodeName?-s.clientHeight+c.bottom:-l.height+c.bottom:c.top,p="right"===h?"HTML"===s.nodeName?-s.clientWidth+c.right:-l.width+c.right:c.left,a&&f)u[f]="translate3d("+p+"px, "+g+"px, 0)",u[d]=0,u[h]=0,u.willChange="transform";else{var y="bottom"===d?-1:1,v="right"===h?-1:1;u[d]=g*y,u[h]=p*v,u.willChange=d+", "+h}var _={"x-placement":t.placement};return t.attributes=T({},_,t.attributes),t.styles=T({},u,t.styles),t.arrowStyles=T({},t.offsets.arrow,t.arrowStyles),t}function lt(t,e,n){var i=W(t,(function(t){var n=t.name;return n===e})),r=!!i&&t.some((function(t){return t.name===n&&t.enabled&&t.order<i.order}));if(!r);return r}function ut(t,e){var n;if(!lt(t.instance.modifiers,"arrow","keepTogether"))return t;var i=e.element;if("string"===typeof i){if(i=t.instance.popper.querySelector(i),!i)return t}else if(!t.instance.popper.contains(i))return t;var r=t.placement.split("-")[0],o=t.offsets,a=o.popper,s=o.reference,u=-1!==["left","right"].indexOf(r),c=u?"height":"width",d=u?"Top":"Left",h=d.toLowerCase(),f=u?"left":"top",p=u?"bottom":"right",m=H(i)[c];s[p]-m<a[h]&&(t.offsets.popper[h]-=a[h]-(s[p]-m)),s[h]+m>a[p]&&(t.offsets.popper[h]+=s[h]+m-a[p]),t.offsets.popper=D(t.offsets.popper);var g=s[h]+s[c]/2-m/2,y=l(t.instance.popper),v=parseFloat(y["margin"+d]),_=parseFloat(y["border"+d+"Width"]),b=g-t.offsets.popper[h]-v-_;return b=Math.max(Math.min(a[c]-m,b),0),t.arrowElement=i,t.offsets.arrow=(n={},k(n,h,Math.round(b)),k(n,f,""),n),t}function ct(t){return"end"===t?"start":"start"===t?"end":t}var dt=["auto-start","auto","auto-end","top-start","top","top-end","right-start","right","right-end","bottom-end","bottom","bottom-start","left-end","left","left-start"],ht=dt.slice(3);function ft(t){var e=arguments.length>1&&void 0!==arguments[1]&&arguments[1],n=ht.indexOf(t),i=ht.slice(n+1).concat(ht.slice(0,n));return e?i.reverse():i}var pt={FLIP:"flip",CLOCKWISE:"clockwise",COUNTERCLOCKWISE:"counterclockwise"};function mt(t,e){if(G(t.instance.modifiers,"inner"))return t;if(t.flipped&&t.placement===t.originalPlacement)return t;var n=P(t.instance.popper,t.instance.reference,e.padding,e.boundariesElement,t.positionFixed),i=t.placement.split("-")[0],r=j(i),o=t.placement.split("-")[1]||"",a=[];switch(e.behavior){case pt.FLIP:a=[i,r];break;case pt.CLOCKWISE:a=ft(i);break;case pt.COUNTERCLOCKWISE:a=ft(i,!0);break;default:a=e.behavior}return a.forEach((function(s,l){if(i!==s||a.length===l+1)return t;i=t.placement.split("-")[0],r=j(i);var u=t.offsets.popper,c=t.offsets.reference,d=Math.floor,h="left"===i&&d(u.right)>d(c.left)||"right"===i&&d(u.left)<d(c.right)||"top"===i&&d(u.bottom)>d(c.top)||"bottom"===i&&d(u.top)<d(c.bottom),f=d(u.left)<d(n.left),p=d(u.right)>d(n.right),m=d(u.top)<d(n.top),g=d(u.bottom)>d(n.bottom),y="left"===i&&f||"right"===i&&p||"top"===i&&m||"bottom"===i&&g,v=-1!==["top","bottom"].indexOf(i),_=!!e.flipVariations&&(v&&"start"===o&&f||v&&"end"===o&&p||!v&&"start"===o&&m||!v&&"end"===o&&g),b=!!e.flipVariationsByContent&&(v&&"start"===o&&p||v&&"end"===o&&f||!v&&"start"===o&&g||!v&&"end"===o&&m),x=_||b;(h||y||x)&&(t.flipped=!0,(h||y)&&(i=a[l+1]),x&&(o=ct(o)),t.placement=i+(o?"-"+o:""),t.offsets.popper=T({},t.offsets.popper,z(t.instance.popper,t.offsets.reference,t.placement)),t=F(t.instance.modifiers,t,"flip"))})),t}function gt(t){var e=t.offsets,n=e.popper,i=e.reference,r=t.placement.split("-")[0],o=Math.floor,a=-1!==["top","bottom"].indexOf(r),s=a?"right":"bottom",l=a?"left":"top",u=a?"width":"height";return n[s]<o(i[l])&&(t.offsets.popper[l]=o(i[l])-n[u]),n[l]>o(i[s])&&(t.offsets.popper[l]=o(i[s])),t}function yt(t,e,n,i){var r=t.match(/((?:\-|\+)?\d*\.?\d*)(.*)/),o=+r[1],a=r[2];if(!o)return t;if(0===a.indexOf("%")){var s=void 0;switch(a){case"%p":s=n;break;case"%":case"%r":default:s=i}var l=D(s);return l[e]/100*o}if("vh"===a||"vw"===a){var u=void 0;return u="vh"===a?Math.max(document.documentElement.clientHeight,window.innerHeight||0):Math.max(document.documentElement.clientWidth,window.innerWidth||0),u/100*o}return o}function vt(t,e,n,i){var r=[0,0],o=-1!==["right","left"].indexOf(i),a=t.split(/(\+|\-)/).map((function(t){return t.trim()})),s=a.indexOf(W(a,(function(t){return-1!==t.search(/,|\s/)})));a[s]&&a[s].indexOf(",");var l=/\s*,\s*|\s+/,u=-1!==s?[a.slice(0,s).concat([a[s].split(l)[0]]),[a[s].split(l)[1]].concat(a.slice(s+1))]:[a];return u=u.map((function(t,i){var r=(1===i?!o:o)?"height":"width",a=!1;return t.reduce((function(t,e){return""===t[t.length-1]&&-1!==["+","-"].indexOf(e)?(t[t.length-1]=e,a=!0,t):a?(t[t.length-1]+=e,a=!1,t):t.concat(e)}),[]).map((function(t){return yt(t,r,e,n)}))})),u.forEach((function(t,e){t.forEach((function(n,i){tt(n)&&(r[e]+=n*("-"===t[i-1]?-1:1))}))})),r}function _t(t,e){var n=e.offset,i=t.placement,r=t.offsets,o=r.popper,a=r.reference,s=i.split("-")[0],l=void 0;return l=tt(+n)?[+n,0]:vt(n,o,a,s),"left"===s?(o.top+=l[0],o.left-=l[1]):"right"===s?(o.top+=l[0],o.left+=l[1]):"top"===s?(o.left+=l[0],o.top-=l[1]):"bottom"===s&&(o.left+=l[0],o.top+=l[1]),t.popper=o,t}function bt(t,e){var n=e.boundariesElement||m(t.instance.popper);t.instance.reference===n&&(n=m(n));var i=$("transform"),r=t.instance.popper.style,o=r.top,a=r.left,s=r[i];r.top="",r.left="",r[i]="";var l=P(t.instance.popper,t.instance.reference,e.padding,n,t.positionFixed);r.top=o,r.left=a,r[i]=s,e.boundaries=l;var u=e.priority,c=t.offsets.popper,d={primary:function(t){var n=c[t];return c[t]<l[t]&&!e.escapeWithReference&&(n=Math.max(c[t],l[t])),k({},t,n)},secondary:function(t){var n="right"===t?"left":"top",i=c[n];return c[t]>l[t]&&!e.escapeWithReference&&(i=Math.min(c[n],l[t]-("right"===t?c.width:c.height))),k({},n,i)}};return u.forEach((function(t){var e=-1!==["left","top"].indexOf(t)?"primary":"secondary";c=T({},c,d[e](t))})),t.offsets.popper=c,t}function xt(t){var e=t.placement,n=e.split("-")[0],i=e.split("-")[1];if(i){var r=t.offsets,o=r.reference,a=r.popper,s=-1!==["bottom","top"].indexOf(n),l=s?"left":"top",u=s?"width":"height",c={start:k({},l,o[l]),end:k({},l,o[l]+o[u]-a[u])};t.offsets.popper=T({},a,c[i])}return t}function wt(t){if(!lt(t.instance.modifiers,"hide","preventOverflow"))return t;var e=t.offsets.reference,n=W(t.instance.modifiers,(function(t){return"preventOverflow"===t.name})).boundaries;if(e.bottom<n.top||e.left>n.right||e.top>n.bottom||e.right<n.left){if(!0===t.hide)return t;t.hide=!0,t.attributes["x-out-of-boundaries"]=""}else{if(!1===t.hide)return t;t.hide=!1,t.attributes["x-out-of-boundaries"]=!1}return t}function Mt(t){var e=t.placement,n=e.split("-")[0],i=t.offsets,r=i.popper,o=i.reference,a=-1!==["left","right"].indexOf(n),s=-1===["top","left"].indexOf(n);return r[a?"left":"top"]=o[n]-(s?r[a?"width":"height"]:0),t.placement=j(e),t.offsets.popper=D(r),t}var St={shift:{order:100,enabled:!0,fn:xt},offset:{order:200,enabled:!0,fn:_t,offset:0},preventOverflow:{order:300,enabled:!0,fn:bt,priority:["left","right","top","bottom"],padding:5,boundariesElement:"scrollParent"},keepTogether:{order:400,enabled:!0,fn:gt},arrow:{order:500,enabled:!0,fn:ut,element:"[x-arrow]"},flip:{order:600,enabled:!0,fn:mt,behavior:"flip",padding:5,boundariesElement:"viewport",flipVariations:!1,flipVariationsByContent:!1},inner:{order:700,enabled:!1,fn:Mt},hide:{order:800,enabled:!0,fn:wt},computeStyle:{order:850,enabled:!0,fn:st,gpuAcceleration:!0,x:"bottom",y:"right"},applyStyle:{order:900,enabled:!0,fn:it,onLoad:rt,gpuAcceleration:void 0}},Lt={placement:"bottom",positionFixed:!1,eventsEnabled:!0,removeOnDestroy:!1,onCreate:function(){},onUpdate:function(){},modifiers:St},kt=function(){function t(e,n){var i=this,r=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{};S(this,t),this.scheduleUpdate=function(){return requestAnimationFrame(i.update)},this.update=a(this.update.bind(this)),this.options=T({},t.Defaults,r),this.state={isDestroyed:!1,isCreated:!1,scrollParents:[]},this.reference=e&&e.jquery?e[0]:e,this.popper=n&&n.jquery?n[0]:n,this.options.modifiers={},Object.keys(T({},t.Defaults.modifiers,r.modifiers)).forEach((function(e){i.options.modifiers[e]=T({},t.Defaults.modifiers[e]||{},r.modifiers?r.modifiers[e]:{})})),this.modifiers=Object.keys(this.options.modifiers).map((function(t){return T({name:t},i.options.modifiers[t])})).sort((function(t,e){return t.order-e.order})),this.modifiers.forEach((function(t){t.enabled&&s(t.onLoad)&&t.onLoad(i.reference,i.popper,i.options,t,i.state)})),this.update();var o=this.options.eventsEnabled;o&&this.enableEventListeners(),this.state.eventsEnabled=o}return L(t,[{key:"update",value:function(){return V.call(this)}},{key:"destroy",value:function(){return U.call(this)}},{key:"enableEventListeners",value:function(){return J.call(this)}},{key:"disableEventListeners",value:function(){return Q.call(this)}}]),t}();return kt.Utils=("undefined"!==typeof window?window:e).PopperUtils,kt.placements=dt,kt.Defaults=Lt,kt}))}).call(this,n("c8ba"))},d327:function(t,e){function n(){return[]}t.exports=n},d370:function(t,e,n){var i=n("253c"),r=n("1310"),o=Object.prototype,a=o.hasOwnProperty,s=o.propertyIsEnumerable,l=i(function(){return arguments}())?i:function(t){return r(t)&&a.call(t,"callee")&&!s.call(t,"callee")};t.exports=l},d3b7:function(t,e,n){var i=n("00ee"),r=n("6eeb"),o=n("b041");i||r(Object.prototype,"toString",o,{unsafe:!0})},d44e:function(t,e,n){var i=n("9bf2").f,r=n("5135"),o=n("b622"),a=o("toStringTag");t.exports=function(t,e,n){t&&!r(t=n?t:t.prototype,a)&&i(t,a,{configurable:!0,value:e})}},d612:function(t,e,n){var i=n("7b83"),r=n("7ed2"),o=n("dc0f");function a(t){var e=-1,n=null==t?0:t.length;this.__data__=new i;while(++e<n)this.add(t[e])}a.prototype.add=a.prototype.push=r,a.prototype.has=o,t.exports=a},d69a:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fil",{months:"Enero_Pebrero_Marso_Abril_Mayo_Hunyo_Hulyo_Agosto_Setyembre_Oktubre_Nobyembre_Disyembre".split("_"),monthsShort:"Ene_Peb_Mar_Abr_May_Hun_Hul_Ago_Set_Okt_Nob_Dis".split("_"),weekdays:"Linggo_Lunes_Martes_Miyerkules_Huwebes_Biyernes_Sabado".split("_"),weekdaysShort:"Lin_Lun_Mar_Miy_Huw_Biy_Sab".split("_"),weekdaysMin:"Li_Lu_Ma_Mi_Hu_Bi_Sab".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"MM/D/YYYY",LL:"MMMM D, YYYY",LLL:"MMMM D, YYYY HH:mm",LLLL:"dddd, MMMM DD, YYYY HH:mm"},calendar:{sameDay:"LT [ngayong araw]",nextDay:"[Bukas ng] LT",nextWeek:"LT [sa susunod na] dddd",lastDay:"LT [kahapon]",lastWeek:"LT [noong nakaraang] dddd",sameElse:"L"},relativeTime:{future:"sa loob ng %s",past:"%s ang nakalipas",s:"ilang segundo",ss:"%d segundo",m:"isang minuto",mm:"%d minuto",h:"isang oras",hh:"%d oras",d:"isang araw",dd:"%d araw",M:"isang buwan",MM:"%d buwan",y:"isang taon",yy:"%d taon"},dayOfMonthOrdinalParse:/\d{1,2}/,ordinal:function(t){return t},week:{dow:1,doy:4}});return e}))},d6b6:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("hy-am",{months:{format:"հունվարի_փետրվարի_մարտի_ապրիլի_մայիսի_հունիսի_հուլիսի_օգոստոսի_սեպտեմբերի_հոկտեմբերի_նոյեմբերի_դեկտեմբերի".split("_"),standalone:"հունվար_փետրվար_մարտ_ապրիլ_մայիս_հունիս_հուլիս_օգոստոս_սեպտեմբեր_հոկտեմբեր_նոյեմբեր_դեկտեմբեր".split("_")},monthsShort:"հնվ_փտր_մրտ_ապր_մյս_հնս_հլս_օգս_սպտ_հկտ_նմբ_դկտ".split("_"),weekdays:"կիրակի_երկուշաբթի_երեքշաբթի_չորեքշաբթի_հինգշաբթի_ուրբաթ_շաբաթ".split("_"),weekdaysShort:"կրկ_երկ_երք_չրք_հնգ_ուրբ_շբթ".split("_"),weekdaysMin:"կրկ_երկ_երք_չրք_հնգ_ուրբ_շբթ".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"D MMMM YYYY թ.",LLL:"D MMMM YYYY թ., HH:mm",LLLL:"dddd, D MMMM YYYY թ., HH:mm"},calendar:{sameDay:"[այսօր] LT",nextDay:"[վաղը] LT",lastDay:"[երեկ] LT",nextWeek:function(){return"dddd [օրը ժամը] LT"},lastWeek:function(){return"[անցած] dddd [օրը ժամը] LT"},sameElse:"L"},relativeTime:{future:"%s հետո",past:"%s առաջ",s:"մի քանի վայրկյան",ss:"%d վայրկյան",m:"րոպե",mm:"%d րոպե",h:"ժամ",hh:"%d ժամ",d:"օր",dd:"%d օր",M:"ամիս",MM:"%d ամիս",y:"տարի",yy:"%d տարի"},meridiemParse:/գիշերվա|առավոտվա|ցերեկվա|երեկոյան/,isPM:function(t){return/^(ցերեկվա|երեկոյան)$/.test(t)},meridiem:function(t){return t<4?"գիշերվա":t<12?"առավոտվա":t<17?"ցերեկվա":"երեկոյան"},dayOfMonthOrdinalParse:/\d{1,2}|\d{1,2}-(ին|րդ)/,ordinal:function(t,e){switch(e){case"DDD":case"w":case"W":case"DDDo":return 1===t?t+"-ին":t+"-րդ";default:return t}},week:{dow:1,doy:7}});return e}))},d6dd:function(t,e,n){t.exports=n("c30c")},d6eb:function(t,e,n){"use strict";var i="_erd";function r(t){return t[i]={},o(t)}function o(t){return t[i]}function a(t){delete t[i]}t.exports={initState:r,getState:o,cleanState:a}},d716:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ca",{months:{standalone:"gener_febrer_març_abril_maig_juny_juliol_agost_setembre_octubre_novembre_desembre".split("_"),format:"de gener_de febrer_de març_d'abril_de maig_de juny_de juliol_d'agost_de setembre_d'octubre_de novembre_de desembre".split("_"),isFormat:/D[oD]?(\s)+MMMM/},monthsShort:"gen._febr._març_abr._maig_juny_jul._ag._set._oct._nov._des.".split("_"),monthsParseExact:!0,weekdays:"diumenge_dilluns_dimarts_dimecres_dijous_divendres_dissabte".split("_"),weekdaysShort:"dg._dl._dt._dc._dj._dv._ds.".split("_"),weekdaysMin:"dg_dl_dt_dc_dj_dv_ds".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM [de] YYYY",ll:"D MMM YYYY",LLL:"D MMMM [de] YYYY [a les] H:mm",lll:"D MMM YYYY, H:mm",LLLL:"dddd D MMMM [de] YYYY [a les] H:mm",llll:"ddd D MMM YYYY, H:mm"},calendar:{sameDay:function(){return"[avui a "+(1!==this.hours()?"les":"la")+"] LT"},nextDay:function(){return"[demà a "+(1!==this.hours()?"les":"la")+"] LT"},nextWeek:function(){return"dddd [a "+(1!==this.hours()?"les":"la")+"] LT"},lastDay:function(){return"[ahir a "+(1!==this.hours()?"les":"la")+"] LT"},lastWeek:function(){return"[el] dddd [passat a "+(1!==this.hours()?"les":"la")+"] LT"},sameElse:"L"},relativeTime:{future:"d'aquí %s",past:"fa %s",s:"uns segons",ss:"%d segons",m:"un minut",mm:"%d minuts",h:"una hora",hh:"%d hores",d:"un dia",dd:"%d dies",M:"un mes",MM:"%d mesos",y:"un any",yy:"%d anys"},dayOfMonthOrdinalParse:/\d{1,2}(r|n|t|è|a)/,ordinal:function(t,e){var n=1===t?"r":2===t?"n":3===t?"r":4===t?"t":"è";return"w"!==e&&"W"!==e||(n="a"),t+n},week:{dow:1,doy:4}});return e}))},d784:function(t,e,n){"use strict";n("ac1f");var i=n("6eeb"),r=n("d039"),o=n("b622"),a=n("9263"),s=n("9112"),l=o("species"),u=!r((function(){var t=/./;return t.exec=function(){var t=[];return t.groups={a:"7"},t},"7"!=="".replace(t,"$<a>")})),c=function(){return"$0"==="a".replace(/./,"$0")}(),d=o("replace"),h=function(){return!!/./[d]&&""===/./[d]("a","$0")}(),f=!r((function(){var t=/(?:)/,e=t.exec;t.exec=function(){return e.apply(this,arguments)};var n="ab".split(t);return 2!==n.length||"a"!==n[0]||"b"!==n[1]}));t.exports=function(t,e,n,d){var p=o(t),m=!r((function(){var e={};return e[p]=function(){return 7},7!=""[t](e)})),g=m&&!r((function(){var e=!1,n=/a/;return"split"===t&&(n={},n.constructor={},n.constructor[l]=function(){return n},n.flags="",n[p]=/./[p]),n.exec=function(){return e=!0,null},n[p](""),!e}));if(!m||!g||"replace"===t&&(!u||!c||h)||"split"===t&&!f){var y=/./[p],v=n(p,""[t],(function(t,e,n,i,r){return e.exec===a?m&&!r?{done:!0,value:y.call(e,n,i)}:{done:!0,value:t.call(n,e,i)}:{done:!1}}),{REPLACE_KEEPS_$0:c,REGEXP_REPLACE_SUBSTITUTES_UNDEFINED_CAPTURE:h}),_=v[0],b=v[1];i(String.prototype,t,_),i(RegExp.prototype,p,2==e?function(t,e){return b.call(t,this,e)}:function(t){return b.call(t,this)})}d&&s(RegExp.prototype[p],"sham",!0)}},d7ee:function(t,e,n){var i=n("c3fc"),r=n("b047"),o=n("99d3"),a=o&&o.isSet,s=a?r(a):i;t.exports=s},d81d:function(t,e,n){"use strict";var i=n("23e7"),r=n("b727").map,o=n("1dde"),a=n("ae40"),s=o("map"),l=a("map");i({target:"Array",proto:!0,forced:!s||!l},{map:function(t){return r(this,t,arguments.length>1?arguments[1]:void 0)}})},d925:function(t,e,n){"use strict";t.exports=function(t){return/^([a-z][a-z\d\+\-\.]*:)?\/\//i.test(t)}},d9f8:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("fr-ca",{months:"janvier_février_mars_avril_mai_juin_juillet_août_septembre_octobre_novembre_décembre".split("_"),monthsShort:"janv._févr._mars_avr._mai_juin_juil._août_sept._oct._nov._déc.".split("_"),monthsParseExact:!0,weekdays:"dimanche_lundi_mardi_mercredi_jeudi_vendredi_samedi".split("_"),weekdaysShort:"dim._lun._mar._mer._jeu._ven._sam.".split("_"),weekdaysMin:"di_lu_ma_me_je_ve_sa".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"YYYY-MM-DD",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Aujourd’hui à] LT",nextDay:"[Demain à] LT",nextWeek:"dddd [à] LT",lastDay:"[Hier à] LT",lastWeek:"dddd [dernier à] LT",sameElse:"L"},relativeTime:{future:"dans %s",past:"il y a %s",s:"quelques secondes",ss:"%d secondes",m:"une minute",mm:"%d minutes",h:"une heure",hh:"%d heures",d:"un jour",dd:"%d jours",M:"un mois",MM:"%d mois",y:"un an",yy:"%d ans"},dayOfMonthOrdinalParse:/\d{1,2}(er|e)/,ordinal:function(t,e){switch(e){default:case"M":case"Q":case"D":case"DDD":case"d":return t+(1===t?"er":"e");case"w":case"W":return t+(1===t?"re":"e")}}});return e}))},da03:function(t,e,n){var i=n("2b3e"),r=i["__core-js_shared__"];t.exports=r},da84:function(t,e,n){(function(e){var n=function(t){return t&&t.Math==Math&&t};t.exports=n("object"==typeof globalThis&&globalThis)||n("object"==typeof window&&window)||n("object"==typeof self&&self)||n("object"==typeof e&&e)||function(){return this}()||Function("return this")()}).call(this,n("c8ba"))},db29:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="jan._feb._mrt._apr._mei_jun._jul._aug._sep._okt._nov._dec.".split("_"),n="jan_feb_mrt_apr_mei_jun_jul_aug_sep_okt_nov_dec".split("_"),i=[/^jan/i,/^feb/i,/^maart|mrt.?$/i,/^apr/i,/^mei$/i,/^jun[i.]?$/i,/^jul[i.]?$/i,/^aug/i,/^sep/i,/^okt/i,/^nov/i,/^dec/i],r=/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december|jan\.?|feb\.?|mrt\.?|apr\.?|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,o=t.defineLocale("nl-be",{months:"januari_februari_maart_april_mei_juni_juli_augustus_september_oktober_november_december".split("_"),monthsShort:function(t,i){return t?/-MMM-/.test(i)?n[t.month()]:e[t.month()]:e},monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december)/i,monthsShortStrictRegex:/^(jan\.?|feb\.?|mrt\.?|apr\.?|mei|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"zondag_maandag_dinsdag_woensdag_donderdag_vrijdag_zaterdag".split("_"),weekdaysShort:"zo._ma._di._wo._do._vr._za.".split("_"),weekdaysMin:"zo_ma_di_wo_do_vr_za".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[vandaag om] LT",nextDay:"[morgen om] LT",nextWeek:"dddd [om] LT",lastDay:"[gisteren om] LT",lastWeek:"[afgelopen] dddd [om] LT",sameElse:"L"},relativeTime:{future:"over %s",past:"%s geleden",s:"een paar seconden",ss:"%d seconden",m:"één minuut",mm:"%d minuten",h:"één uur",hh:"%d uur",d:"één dag",dd:"%d dagen",M:"één maand",MM:"%d maanden",y:"één jaar",yy:"%d jaar"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return o}))},dc0f:function(t,e){function n(t){return this.__data__.has(t)}t.exports=n},dc4d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"१",2:"२",3:"३",4:"४",5:"५",6:"६",7:"७",8:"८",9:"९",0:"०"},n={"१":"1","२":"2","३":"3","४":"4","५":"5","६":"6","७":"7","८":"8","९":"9","०":"0"},i=[/^जन/i,/^फ़र|फर/i,/^मार्च/i,/^अप्रै/i,/^मई/i,/^जून/i,/^जुल/i,/^अग/i,/^सितं|सित/i,/^अक्टू/i,/^नव|नवं/i,/^दिसं|दिस/i],r=[/^जन/i,/^फ़र/i,/^मार्च/i,/^अप्रै/i,/^मई/i,/^जून/i,/^जुल/i,/^अग/i,/^सित/i,/^अक्टू/i,/^नव/i,/^दिस/i],o=t.defineLocale("hi",{months:{format:"जनवरी_फ़रवरी_मार्च_अप्रैल_मई_जून_जुलाई_अगस्त_सितम्बर_अक्टूबर_नवम्बर_दिसम्बर".split("_"),standalone:"जनवरी_फरवरी_मार्च_अप्रैल_मई_जून_जुलाई_अगस्त_सितंबर_अक्टूबर_नवंबर_दिसंबर".split("_")},monthsShort:"जन._फ़र._मार्च_अप्रै._मई_जून_जुल._अग._सित._अक्टू._नव._दिस.".split("_"),weekdays:"रविवार_सोमवार_मंगलवार_बुधवार_गुरूवार_शुक्रवार_शनिवार".split("_"),weekdaysShort:"रवि_सोम_मंगल_बुध_गुरू_शुक्र_शनि".split("_"),weekdaysMin:"र_सो_मं_बु_गु_शु_श".split("_"),longDateFormat:{LT:"A h:mm बजे",LTS:"A h:mm:ss बजे",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm बजे",LLLL:"dddd, D MMMM YYYY, A h:mm बजे"},monthsParse:i,longMonthsParse:i,shortMonthsParse:r,monthsRegex:/^(जनवरी|जन\.?|फ़रवरी|फरवरी|फ़र\.?|मार्च?|अप्रैल|अप्रै\.?|मई?|जून?|जुलाई|जुल\.?|अगस्त|अग\.?|सितम्बर|सितंबर|सित\.?|अक्टूबर|अक्टू\.?|नवम्बर|नवंबर|नव\.?|दिसम्बर|दिसंबर|दिस\.?)/i,monthsShortRegex:/^(जनवरी|जन\.?|फ़रवरी|फरवरी|फ़र\.?|मार्च?|अप्रैल|अप्रै\.?|मई?|जून?|जुलाई|जुल\.?|अगस्त|अग\.?|सितम्बर|सितंबर|सित\.?|अक्टूबर|अक्टू\.?|नवम्बर|नवंबर|नव\.?|दिसम्बर|दिसंबर|दिस\.?)/i,monthsStrictRegex:/^(जनवरी?|फ़रवरी|फरवरी?|मार्च?|अप्रैल?|मई?|जून?|जुलाई?|अगस्त?|सितम्बर|सितंबर|सित?\.?|अक्टूबर|अक्टू\.?|नवम्बर|नवंबर?|दिसम्बर|दिसंबर?)/i,monthsShortStrictRegex:/^(जन\.?|फ़र\.?|मार्च?|अप्रै\.?|मई?|जून?|जुल\.?|अग\.?|सित\.?|अक्टू\.?|नव\.?|दिस\.?)/i,calendar:{sameDay:"[आज] LT",nextDay:"[कल] LT",nextWeek:"dddd, LT",lastDay:"[कल] LT",lastWeek:"[पिछले] dddd, LT",sameElse:"L"},relativeTime:{future:"%s में",past:"%s पहले",s:"कुछ ही क्षण",ss:"%d सेकंड",m:"एक मिनट",mm:"%d मिनट",h:"एक घंटा",hh:"%d घंटे",d:"एक दिन",dd:"%d दिन",M:"एक महीने",MM:"%d महीने",y:"एक वर्ष",yy:"%d वर्ष"},preparse:function(t){return t.replace(/[१२३४५६७८९०]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/रात|सुबह|दोपहर|शाम/,meridiemHour:function(t,e){return 12===t&&(t=0),"रात"===e?t<4?t:t+12:"सुबह"===e?t:"दोपहर"===e?t>=10?t:t+12:"शाम"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"रात":t<10?"सुबह":t<17?"दोपहर":t<20?"शाम":"रात"},week:{dow:0,doy:6}});return o}))},dc57:function(t,e){var n=Function.prototype,i=n.toString;function r(t){if(null!=t){try{return i.call(t)}catch(e){}try{return t+""}catch(e){}}return""}t.exports=r},dcbe:function(t,e,n){var i=n("30c9"),r=n("1310");function o(t){return r(t)&&i(t)}t.exports=o},dcbf:function(t,e,n){"use strict";var i=n("f500");e["a"]=i["a"]},dce5:function(t,e,n){var i=n("32b3"),r=n("8eeb"),o=n("2ec1"),a=n("30c9"),s=n("eac5"),l=n("ec69"),u=Object.prototype,c=u.hasOwnProperty,d=o((function(t,e){if(s(e)||a(e))r(e,l(e),t);else for(var n in e)c.call(e,n)&&i(t,n,e[n])}));t.exports=d},dd65:function(t,e,n){var i=n("badf"),r=n("a0ac"),o=n("77c1");function a(t,e){return o(t,r(i(e)))}t.exports=a},ddb0:function(t,e,n){var i=n("da84"),r=n("fdbc"),o=n("e260"),a=n("9112"),s=n("b622"),l=s("iterator"),u=s("toStringTag"),c=o.values;for(var d in r){var h=i[d],f=h&&h.prototype;if(f){if(f[l]!==c)try{a(f,l,c)}catch(m){f[l]=c}if(f[u]||a(f,u,d),r[d])for(var p in o)if(f[p]!==o[p])try{a(f,p,o[p])}catch(m){f[p]=o[p]}}}},de48:function(t,e,n){"use strict";n.d(e,"b",(function(){return o})),n.d(e,"a",(function(){return a}));var i=n("2b0e");const r=i["a"].prototype.$isServer,o=function(){return!r&&document.addEventListener?function(t,e,n,i=!1){t&&e&&n&&t.addEventListener(e,n,i)}:function(t,e,n){t&&e&&n&&t.attachEvent("on"+e,n)}}(),a=function(){return!r&&document.removeEventListener?function(t,e,n,i=!1){t&&e&&t.removeEventListener(e,n,i)}:function(t,e,n){t&&e&&t.detachEvent("on"+e,n)}}()},df75:function(t,e,n){var i=n("ca84"),r=n("7839");t.exports=Object.keys||function(t){return i(t,r)}},df7c:function(t,e,n){(function(t){function n(t,e){for(var n=0,i=t.length-1;i>=0;i--){var r=t[i];"."===r?t.splice(i,1):".."===r?(t.splice(i,1),n++):n&&(t.splice(i,1),n--)}if(e)for(;n--;n)t.unshift("..");return t}function i(t){"string"!==typeof t&&(t+="");var e,n=0,i=-1,r=!0;for(e=t.length-1;e>=0;--e)if(47===t.charCodeAt(e)){if(!r){n=e+1;break}}else-1===i&&(r=!1,i=e+1);return-1===i?"":t.slice(n,i)}function r(t,e){if(t.filter)return t.filter(e);for(var n=[],i=0;i<t.length;i++)e(t[i],i,t)&&n.push(t[i]);return n}e.resolve=function(){for(var e="",i=!1,o=arguments.length-1;o>=-1&&!i;o--){var a=o>=0?arguments[o]:t.cwd();if("string"!==typeof a)throw new TypeError("Arguments to path.resolve must be strings");a&&(e=a+"/"+e,i="/"===a.charAt(0))}return e=n(r(e.split("/"),(function(t){return!!t})),!i).join("/"),(i?"/":"")+e||"."},e.normalize=function(t){var i=e.isAbsolute(t),a="/"===o(t,-1);return t=n(r(t.split("/"),(function(t){return!!t})),!i).join("/"),t||i||(t="."),t&&a&&(t+="/"),(i?"/":"")+t},e.isAbsolute=function(t){return"/"===t.charAt(0)},e.join=function(){var t=Array.prototype.slice.call(arguments,0);return e.normalize(r(t,(function(t,e){if("string"!==typeof t)throw new TypeError("Arguments to path.join must be strings");return t})).join("/"))},e.relative=function(t,n){function i(t){for(var e=0;e<t.length;e++)if(""!==t[e])break;for(var n=t.length-1;n>=0;n--)if(""!==t[n])break;return e>n?[]:t.slice(e,n-e+1)}t=e.resolve(t).substr(1),n=e.resolve(n).substr(1);for(var r=i(t.split("/")),o=i(n.split("/")),a=Math.min(r.length,o.length),s=a,l=0;l<a;l++)if(r[l]!==o[l]){s=l;break}var u=[];for(l=s;l<r.length;l++)u.push("..");return u=u.concat(o.slice(s)),u.join("/")},e.sep="/",e.delimiter=":",e.dirname=function(t){if("string"!==typeof t&&(t+=""),0===t.length)return".";for(var e=t.charCodeAt(0),n=47===e,i=-1,r=!0,o=t.length-1;o>=1;--o)if(e=t.charCodeAt(o),47===e){if(!r){i=o;break}}else r=!1;return-1===i?n?"/":".":n&&1===i?"/":t.slice(0,i)},e.basename=function(t,e){var n=i(t);return e&&n.substr(-1*e.length)===e&&(n=n.substr(0,n.length-e.length)),n},e.extname=function(t){"string"!==typeof t&&(t+="");for(var e=-1,n=0,i=-1,r=!0,o=0,a=t.length-1;a>=0;--a){var s=t.charCodeAt(a);if(47!==s)-1===i&&(r=!1,i=a+1),46===s?-1===e?e=a:1!==o&&(o=1):-1!==e&&(o=-1);else if(!r){n=a+1;break}}return-1===e||-1===i||0===o||1===o&&e===i-1&&e===n+1?"":t.slice(e,i)};var o="b"==="ab".substr(-1)?function(t,e,n){return t.substr(e,n)}:function(t,e,n){return e<0&&(e=t.length+e),t.substr(e,n)}}).call(this,n("4362"))},e01a:function(t,e,n){"use strict";var i=n("23e7"),r=n("83ab"),o=n("da84"),a=n("5135"),s=n("861d"),l=n("9bf2").f,u=n("e893"),c=o.Symbol;if(r&&"function"==typeof c&&(!("description"in c.prototype)||void 0!==c().description)){var d={},h=function(){var t=arguments.length<1||void 0===arguments[0]?void 0:String(arguments[0]),e=this instanceof h?new c(t):void 0===t?c():c(t);return""===t&&(d[e]=!0),e};u(h,c);var f=h.prototype=c.prototype;f.constructor=h;var p=f.toString,m="Symbol(test)"==String(c("test")),g=/^Symbol\((.*)\)[^)]+$/;l(f,"description",{configurable:!0,get:function(){var t=s(this)?this.valueOf():this,e=p.call(t);if(a(d,t))return"";var n=m?e.slice(7,-1):e.replace(g,"$1");return""===n?void 0:n}}),i({global:!0,forced:!0},{Symbol:h})}},e031:function(t,e,n){var i=n("f909"),r=n("1a8c");function o(t,e,n,a,s,l){return r(t)&&r(e)&&(l.set(e,t),i(t,e,void 0,o,l),l["delete"](e)),t}t.exports=o},e0b9:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("transition",{attrs:{name:"fade"}},[n("div",{directives:[{name:"show",rawName:"v-show",value:t.show,expression:"show"}],class:t.classes,style:t.outerStyles},[n("div",{class:t.innerClasses,style:t.styles})])])},r=[],o=(n("a9e3"),n("ade3")),a="ivu-loading-bar",s={name:"LoadingBar",props:{color:{type:String,default:"primary"},failedColor:{type:String,default:"error"},height:{type:Number,default:2}},data:function(){return{percent:0,status:"success",show:!1}},computed:{classes:function(){return"".concat(a)},innerClasses:function(){var t;return["".concat(a,"-inner"),(t={},Object(o["a"])(t,"".concat(a,"-inner-color-primary"),"primary"===this.color&&"success"===this.status),Object(o["a"])(t,"".concat(a,"-inner-failed-color-error"),"error"===this.failedColor&&"error"===this.status),t)]},outerStyles:function(){return{height:"".concat(this.height,"px")}},styles:function(){var t={width:"".concat(this.percent,"%"),height:"".concat(this.height,"px")};return"primary"!==this.color&&"success"===this.status&&(t.backgroundColor=this.color),"error"!==this.failedColor&&"error"===this.status&&(t.backgroundColor=this.failedColor),t}}},l=s,u=n("2877"),c=Object(u["a"])(l,i,r,!1,null,null,null),d=c.exports,h=n("2b0e");d.newInstance=t=>{const e=t||{},n=new h["a"]({data:e,render(t){return t(d,{props:e})}}),i=n.$mount();document.body.appendChild(i.$el);const r=n.$children[0];return{update(t){"percent"in t&&(r.percent=t.percent),t.status&&(r.status=t.status),"show"in t&&(r.show=t.show)},component:r,destroy(){document.body.removeChild(document.getElementsByClassName("ivu-loading-bar")[0])}}};var f=d;let p,m,g="primary",y=800,v="error",_=2;function b(){return p=p||f.newInstance({color:g,failedColor:v,height:_}),p}function x(t){let e=b();e.update(t)}function w(){setTimeout(()=>{x({show:!1}),setTimeout(()=>{x({percent:0})},200)},y)}function M(){m&&(clearInterval(m),m=null)}e["a"]={start(){if(m)return;let t=0;x({percent:t,status:"success",show:!0}),m=setInterval(()=>{t+=Math.floor(3*Math.random()+1),t>95&&M(),x({percent:t,status:"success",show:!0})},200)},update(t){M(),x({percent:t,status:"success",show:!0})},finish(){M(),x({percent:100,status:"success",show:!0}),w()},error(){M(),x({percent:100,status:"error",show:!0}),w()},config(t){t.color&&(g=t.color),t.duration&&(y=t.duration),t.failedColor&&(v=t.failedColor),t.height&&(_=t.height)},destroy(){M();let t=b();p=null,t.destroy()}}},e0c5:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"૧",2:"૨",3:"૩",4:"૪",5:"૫",6:"૬",7:"૭",8:"૮",9:"૯",0:"૦"},n={"૧":"1","૨":"2","૩":"3","૪":"4","૫":"5","૬":"6","૭":"7","૮":"8","૯":"9","૦":"0"},i=t.defineLocale("gu",{months:"જાન્યુઆરી_ફેબ્રુઆરી_માર્ચ_એપ્રિલ_મે_જૂન_જુલાઈ_ઑગસ્ટ_સપ્ટેમ્બર_ઑક્ટ્બર_નવેમ્બર_ડિસેમ્બર".split("_"),monthsShort:"જાન્યુ._ફેબ્રુ._માર્ચ_એપ્રિ._મે_જૂન_જુલા._ઑગ._સપ્ટે._ઑક્ટ્._નવે._ડિસે.".split("_"),monthsParseExact:!0,weekdays:"રવિવાર_સોમવાર_મંગળવાર_બુધ્વાર_ગુરુવાર_શુક્રવાર_શનિવાર".split("_"),weekdaysShort:"રવિ_સોમ_મંગળ_બુધ્_ગુરુ_શુક્ર_શનિ".split("_"),weekdaysMin:"ર_સો_મં_બુ_ગુ_શુ_શ".split("_"),longDateFormat:{LT:"A h:mm વાગ્યે",LTS:"A h:mm:ss વાગ્યે",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm વાગ્યે",LLLL:"dddd, D MMMM YYYY, A h:mm વાગ્યે"},calendar:{sameDay:"[આજ] LT",nextDay:"[કાલે] LT",nextWeek:"dddd, LT",lastDay:"[ગઇકાલે] LT",lastWeek:"[પાછલા] dddd, LT",sameElse:"L"},relativeTime:{future:"%s મા",past:"%s પહેલા",s:"અમુક પળો",ss:"%d સેકંડ",m:"એક મિનિટ",mm:"%d મિનિટ",h:"એક કલાક",hh:"%d કલાક",d:"એક દિવસ",dd:"%d દિવસ",M:"એક મહિનો",MM:"%d મહિનો",y:"એક વર્ષ",yy:"%d વર્ષ"},preparse:function(t){return t.replace(/[૧૨૩૪૫૬૭૮૯૦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/રાત|બપોર|સવાર|સાંજ/,meridiemHour:function(t,e){return 12===t&&(t=0),"રાત"===e?t<4?t:t+12:"સવાર"===e?t:"બપોર"===e?t>=10?t:t+12:"સાંજ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"રાત":t<10?"સવાર":t<17?"બપોર":t<20?"સાંજ":"રાત"},week:{dow:0,doy:6}});return i}))},e0e7:function(t,e,n){var i=n("60ed");function r(t){return i(t)?void 0:t}t.exports=r},e163:function(t,e,n){var i=n("5135"),r=n("7b0b"),o=n("f772"),a=n("e177"),s=o("IE_PROTO"),l=Object.prototype;t.exports=a?Object.getPrototypeOf:function(t){return t=r(t),i(t,s)?t[s]:"function"==typeof t.constructor&&t instanceof t.constructor?t.constructor.prototype:t instanceof Object?l:null}},e177:function(t,e,n){var i=n("d039");t.exports=!i((function(){function t(){}return t.prototype.constructor=null,Object.getPrototypeOf(new t)!==t.prototype}))},e1d3:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-ie",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},e24b:function(t,e,n){var i=n("49f4"),r=n("1efc"),o=n("bbc0"),a=n("7a48"),s=n("2524");function l(t){var e=-1,n=null==t?0:t.length;this.clear();while(++e<n){var i=t[e];this.set(i[0],i[1])}}l.prototype.clear=i,l.prototype["delete"]=r,l.prototype.get=o,l.prototype.has=a,l.prototype.set=s,t.exports=l},e260:function(t,e,n){"use strict";var i=n("fc6a"),r=n("44d2"),o=n("3f8c"),a=n("69f3"),s=n("7dd0"),l="Array Iterator",u=a.set,c=a.getterFor(l);t.exports=s(Array,"Array",(function(t,e){u(this,{type:l,target:i(t),index:0,kind:e})}),(function(){var t=c(this),e=t.target,n=t.kind,i=t.index++;return!e||i>=e.length?(t.target=void 0,{value:void 0,done:!0}):"keys"==n?{value:i,done:!1}:"values"==n?{value:e[i],done:!1}:{value:[i,e[i]],done:!1}}),"values"),o.Arguments=o.Array,r("keys"),r("values"),r("entries")},e285:function(t,e,n){var i=n("da84"),r=i.isFinite;t.exports=Number.isFinite||function(t){return"number"==typeof t&&r(t)}},e2c0:function(t,e,n){var i=n("e2e4"),r=n("d370"),o=n("6747"),a=n("c098"),s=n("b218"),l=n("f4d6");function u(t,e,n){e=i(e,t);var u=-1,c=e.length,d=!1;while(++u<c){var h=l(e[u]);if(!(d=null!=t&&n(t,h)))break;t=t[h]}return d||++u!=c?d:(c=null==t?0:t.length,!!c&&s(c)&&a(h,c)&&(o(t)||r(t)))}t.exports=u},e2cc:function(t,e,n){var i=n("6eeb");t.exports=function(t,e,n){for(var r in e)i(t,r,e[r],n);return t}},e2e4:function(t,e,n){var i=n("6747"),r=n("f608"),o=n("18d8"),a=n("76dd");function s(t,e){return i(t)?t:r(t,e)?[t]:o(a(t))}t.exports=s},e380:function(t,e,n){var i=n("7b83"),r="Expected a function";function o(t,e){if("function"!=typeof t||null!=e&&"function"!=typeof e)throw new TypeError(r);var n=function(){var i=arguments,r=e?e.apply(this,i):i[0],o=n.cache;if(o.has(r))return o.get(r);var a=t.apply(this,i);return n.cache=o.set(r,a)||o,a};return n.cache=new(o.Cache||i),n}o.Cache=i,t.exports=o},e3f5:function(t,e,n){"use strict";var i=n("7a12"),r=n("f500");i["a"].Group=r["a"],e["a"]=i["a"]},e3f8:function(t,e,n){var i=n("656b");function r(t){return function(e){return i(e,t)}}t.exports=r},e538:function(t,e,n){var i=n("b622");e.f=i},e5383:function(t,e,n){(function(t){var i=n("2b3e"),r=e&&!e.nodeType&&e,o=r&&"object"==typeof t&&t&&!t.nodeType&&t,a=o&&o.exports===r,s=a?i.Buffer:void 0,l=s?s.allocUnsafe:void 0;function u(t,e){if(e)return t.slice();var n=t.length,i=l?l(n):new t.constructor(n);return t.copy(i),i}t.exports=u}).call(this,n("62e4")(t))},e667:function(t,e){t.exports=function(t){try{return{error:!1,value:t()}}catch(e){return{error:!0,value:e}}}},e683:function(t,e,n){"use strict";t.exports=function(t,e){return e?t.replace(/\/+$/,"")+"/"+e.replace(/^\/+/,""):t}},e6cf:function(t,e,n){"use strict";var i,r,o,a,s=n("23e7"),l=n("c430"),u=n("da84"),c=n("d066"),d=n("fea9"),h=n("6eeb"),f=n("e2cc"),p=n("d44e"),m=n("2626"),g=n("861d"),y=n("1c0b"),v=n("19aa"),_=n("8925"),b=n("2266"),x=n("1c7e"),w=n("4840"),M=n("2cf4").set,S=n("b575"),L=n("cdf9"),k=n("44de"),T=n("f069"),D=n("e667"),C=n("69f3"),I=n("94ca"),A=n("b622"),Y=n("605d"),O=n("2d00"),P=A("species"),E="Promise",N=C.get,R=C.set,H=C.getterFor(E),j=d,z=u.TypeError,W=u.document,B=u.process,F=c("fetch"),V=T.f,G=V,$=!!(W&&W.createEvent&&u.dispatchEvent),U="function"==typeof PromiseRejectionEvent,q="unhandledrejection",X="rejectionhandled",Z=0,J=1,K=2,Q=1,tt=2,et=I(E,(function(){var t=_(j)!==String(j);if(!t){if(66===O)return!0;if(!Y&&!U)return!0}if(l&&!j.prototype["finally"])return!0;if(O>=51&&/native code/.test(j))return!1;var e=j.resolve(1),n=function(t){t((function(){}),(function(){}))},i=e.constructor={};return i[P]=n,!(e.then((function(){}))instanceof n)})),nt=et||!x((function(t){j.all(t)["catch"]((function(){}))})),it=function(t){var e;return!(!g(t)||"function"!=typeof(e=t.then))&&e},rt=function(t,e){if(!t.notified){t.notified=!0;var n=t.reactions;S((function(){var i=t.value,r=t.state==J,o=0;while(n.length>o){var a,s,l,u=n[o++],c=r?u.ok:u.fail,d=u.resolve,h=u.reject,f=u.domain;try{c?(r||(t.rejection===tt&&lt(t),t.rejection=Q),!0===c?a=i:(f&&f.enter(),a=c(i),f&&(f.exit(),l=!0)),a===u.promise?h(z("Promise-chain cycle")):(s=it(a))?s.call(a,d,h):d(a)):h(i)}catch(p){f&&!l&&f.exit(),h(p)}}t.reactions=[],t.notified=!1,e&&!t.rejection&&at(t)}))}},ot=function(t,e,n){var i,r;$?(i=W.createEvent("Event"),i.promise=e,i.reason=n,i.initEvent(t,!1,!0),u.dispatchEvent(i)):i={promise:e,reason:n},!U&&(r=u["on"+t])?r(i):t===q&&k("Unhandled promise rejection",n)},at=function(t){M.call(u,(function(){var e,n=t.facade,i=t.value,r=st(t);if(r&&(e=D((function(){Y?B.emit("unhandledRejection",i,n):ot(q,n,i)})),t.rejection=Y||st(t)?tt:Q,e.error))throw e.value}))},st=function(t){return t.rejection!==Q&&!t.parent},lt=function(t){M.call(u,(function(){var e=t.facade;Y?B.emit("rejectionHandled",e):ot(X,e,t.value)}))},ut=function(t,e,n){return function(i){t(e,i,n)}},ct=function(t,e,n){t.done||(t.done=!0,n&&(t=n),t.value=e,t.state=K,rt(t,!0))},dt=function(t,e,n){if(!t.done){t.done=!0,n&&(t=n);try{if(t.facade===e)throw z("Promise can't be resolved itself");var i=it(e);i?S((function(){var n={done:!1};try{i.call(e,ut(dt,n,t),ut(ct,n,t))}catch(r){ct(n,r,t)}})):(t.value=e,t.state=J,rt(t,!1))}catch(r){ct({done:!1},r,t)}}};et&&(j=function(t){v(this,j,E),y(t),i.call(this);var e=N(this);try{t(ut(dt,e),ut(ct,e))}catch(n){ct(e,n)}},i=function(t){R(this,{type:E,done:!1,notified:!1,parent:!1,reactions:[],rejection:!1,state:Z,value:void 0})},i.prototype=f(j.prototype,{then:function(t,e){var n=H(this),i=V(w(this,j));return i.ok="function"!=typeof t||t,i.fail="function"==typeof e&&e,i.domain=Y?B.domain:void 0,n.parent=!0,n.reactions.push(i),n.state!=Z&&rt(n,!1),i.promise},catch:function(t){return this.then(void 0,t)}}),r=function(){var t=new i,e=N(t);this.promise=t,this.resolve=ut(dt,e),this.reject=ut(ct,e)},T.f=V=function(t){return t===j||t===o?new r(t):G(t)},l||"function"!=typeof d||(a=d.prototype.then,h(d.prototype,"then",(function(t,e){var n=this;return new j((function(t,e){a.call(n,t,e)})).then(t,e)}),{unsafe:!0}),"function"==typeof F&&s({global:!0,enumerable:!0,forced:!0},{fetch:function(t){return L(j,F.apply(u,arguments))}}))),s({global:!0,wrap:!0,forced:et},{Promise:j}),p(j,E,!1,!0),m(E),o=c(E),s({target:E,stat:!0,forced:et},{reject:function(t){var e=V(this);return e.reject.call(void 0,t),e.promise}}),s({target:E,stat:!0,forced:l||et},{resolve:function(t){return L(l&&this===o?j:this,t)}}),s({target:E,stat:!0,forced:nt},{all:function(t){var e=this,n=V(e),i=n.resolve,r=n.reject,o=D((function(){var n=y(e.resolve),o=[],a=0,s=1;b(t,(function(t){var l=a++,u=!1;o.push(void 0),s++,n.call(e,t).then((function(t){u||(u=!0,o[l]=t,--s||i(o))}),r)})),--s||i(o)}));return o.error&&r(o.value),n.promise},race:function(t){var e=this,n=V(e),i=n.reject,r=D((function(){var r=y(e.resolve);b(t,(function(t){r.call(e,t).then(n.resolve,i)}))}));return r.error&&i(r.value),n.promise}})},e759:function(t,e,n){var i=n("badf"),r=n("43ad"),o=Object.prototype,a=o.hasOwnProperty,s=o.toString,l=r((function(t,e,n){null!=e&&"function"!=typeof e.toString&&(e=s.call(e)),a.call(t,e)?t[e].push(n):t[e]=[n]}),i);t.exports=l},e79d:function(t,e,n){"use strict";var i=n("245a");e["a"]=i["a"]},e81d:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"១",2:"២",3:"៣",4:"៤",5:"៥",6:"៦",7:"៧",8:"៨",9:"៩",0:"០"},n={"១":"1","២":"2","៣":"3","៤":"4","៥":"5","៦":"6","៧":"7","៨":"8","៩":"9","០":"0"},i=t.defineLocale("km",{months:"មករា_កុម្ភៈ_មីនា_មេសា_ឧសភា_មិថុនា_កក្កដា_សីហា_កញ្ញា_តុលា_វិច្ឆិកា_ធ្នូ".split("_"),monthsShort:"មករា_កុម្ភៈ_មីនា_មេសា_ឧសភា_មិថុនា_កក្កដា_សីហា_កញ្ញា_តុលា_វិច្ឆិកា_ធ្នូ".split("_"),weekdays:"អាទិត្យ_ច័ន្ទ_អង្គារ_ពុធ_ព្រហស្បតិ៍_សុក្រ_សៅរ៍".split("_"),weekdaysShort:"អា_ច_អ_ព_ព្រ_សុ_ស".split("_"),weekdaysMin:"អា_ច_អ_ព_ព្រ_សុ_ស".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},meridiemParse:/ព្រឹក|ល្ងាច/,isPM:function(t){return"ល្ងាច"===t},meridiem:function(t,e,n){return t<12?"ព្រឹក":"ល្ងាច"},calendar:{sameDay:"[ថ្ងៃនេះ ម៉ោង] LT",nextDay:"[ស្អែក ម៉ោង] LT",nextWeek:"dddd [ម៉ោង] LT",lastDay:"[ម្សិលមិញ ម៉ោង] LT",lastWeek:"dddd [សប្តាហ៍មុន] [ម៉ោង] LT",sameElse:"L"},relativeTime:{future:"%sទៀត",past:"%sមុន",s:"ប៉ុន្មានវិនាទី",ss:"%d វិនាទី",m:"មួយនាទី",mm:"%d នាទី",h:"មួយម៉ោង",hh:"%d ម៉ោង",d:"មួយថ្ងៃ",dd:"%d ថ្ងៃ",M:"មួយខែ",MM:"%d ខែ",y:"មួយឆ្នាំ",yy:"%d ឆ្នាំ"},dayOfMonthOrdinalParse:/ទី\d{1,2}/,ordinal:"ទី%d",preparse:function(t){return t.replace(/[១២៣៤៥៦៧៨៩០]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},week:{dow:1,doy:4}});return i}))},e893:function(t,e,n){var i=n("5135"),r=n("56ef"),o=n("06cf"),a=n("9bf2");t.exports=function(t,e){for(var n=r(e),s=a.f,l=o.f,u=0;u<n.length;u++){var c=n[u];i(t,c)||s(t,c,l(e,c))}}},e8b5:function(t,e,n){var i=n("c6b6");t.exports=Array.isArray||function(t){return"Array"==i(t)}},e95a:function(t,e,n){var i=n("b622"),r=n("3f8c"),o=i("iterator"),a=Array.prototype;t.exports=function(t){return void 0!==t&&(r.Array===t||a[o]===t)}},e9d1:function(t,e,n){var i=n("a78b"),r=n("1304");function o(t,e,n,o){return o="function"==typeof o?o:void 0,null==t?t:i(t,e,r(n),o)}t.exports=o},eac5:function(t,e){var n=Object.prototype;function i(t){var e=t&&t.constructor,i="function"==typeof e&&e.prototype||n;return t===i}t.exports=i},ebe4:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("ms",{months:"Januari_Februari_Mac_April_Mei_Jun_Julai_Ogos_September_Oktober_November_Disember".split("_"),monthsShort:"Jan_Feb_Mac_Apr_Mei_Jun_Jul_Ogs_Sep_Okt_Nov_Dis".split("_"),weekdays:"Ahad_Isnin_Selasa_Rabu_Khamis_Jumaat_Sabtu".split("_"),weekdaysShort:"Ahd_Isn_Sel_Rab_Kha_Jum_Sab".split("_"),weekdaysMin:"Ah_Is_Sl_Rb_Km_Jm_Sb".split("_"),longDateFormat:{LT:"HH.mm",LTS:"HH.mm.ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY [pukul] HH.mm",LLLL:"dddd, D MMMM YYYY [pukul] HH.mm"},meridiemParse:/pagi|tengahari|petang|malam/,meridiemHour:function(t,e){return 12===t&&(t=0),"pagi"===e?t:"tengahari"===e?t>=11?t:t+12:"petang"===e||"malam"===e?t+12:void 0},meridiem:function(t,e,n){return t<11?"pagi":t<15?"tengahari":t<19?"petang":"malam"},calendar:{sameDay:"[Hari ini pukul] LT",nextDay:"[Esok pukul] LT",nextWeek:"dddd [pukul] LT",lastDay:"[Kelmarin pukul] LT",lastWeek:"dddd [lepas pukul] LT",sameElse:"L"},relativeTime:{future:"dalam %s",past:"%s yang lepas",s:"beberapa saat",ss:"%d saat",m:"seminit",mm:"%d minit",h:"sejam",hh:"%d jam",d:"sehari",dd:"%d hari",M:"sebulan",MM:"%d bulan",y:"setahun",yy:"%d tahun"},week:{dow:1,doy:7}});return e}))},ec18:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
function e(t,e,n,i){var r={s:["mõne sekundi","mõni sekund","paar sekundit"],ss:[t+"sekundi",t+"sekundit"],m:["ühe minuti","üks minut"],mm:[t+" minuti",t+" minutit"],h:["ühe tunni","tund aega","üks tund"],hh:[t+" tunni",t+" tundi"],d:["ühe päeva","üks päev"],M:["kuu aja","kuu aega","üks kuu"],MM:[t+" kuu",t+" kuud"],y:["ühe aasta","aasta","üks aasta"],yy:[t+" aasta",t+" aastat"]};return e?r[n][2]?r[n][2]:r[n][1]:i?r[n][0]:r[n][1]}var n=t.defineLocale("et",{months:"jaanuar_veebruar_märts_aprill_mai_juuni_juuli_august_september_oktoober_november_detsember".split("_"),monthsShort:"jaan_veebr_märts_apr_mai_juuni_juuli_aug_sept_okt_nov_dets".split("_"),weekdays:"pühapäev_esmaspäev_teisipäev_kolmapäev_neljapäev_reede_laupäev".split("_"),weekdaysShort:"P_E_T_K_N_R_L".split("_"),weekdaysMin:"P_E_T_K_N_R_L".split("_"),longDateFormat:{LT:"H:mm",LTS:"H:mm:ss",L:"DD.MM.YYYY",LL:"D. MMMM YYYY",LLL:"D. MMMM YYYY H:mm",LLLL:"dddd, D. MMMM YYYY H:mm"},calendar:{sameDay:"[Täna,] LT",nextDay:"[Homme,] LT",nextWeek:"[Järgmine] dddd LT",lastDay:"[Eile,] LT",lastWeek:"[Eelmine] dddd LT",sameElse:"L"},relativeTime:{future:"%s pärast",past:"%s tagasi",s:e,ss:e,m:e,mm:e,h:e,hh:e,d:e,dd:"%d päeva",M:e,MM:e,y:e,yy:e},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return n}))},ec2e:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("en-in",{months:"January_February_March_April_May_June_July_August_September_October_November_December".split("_"),monthsShort:"Jan_Feb_Mar_Apr_May_Jun_Jul_Aug_Sep_Oct_Nov_Dec".split("_"),weekdays:"Sunday_Monday_Tuesday_Wednesday_Thursday_Friday_Saturday".split("_"),weekdaysShort:"Sun_Mon_Tue_Wed_Thu_Fri_Sat".split("_"),weekdaysMin:"Su_Mo_Tu_We_Th_Fr_Sa".split("_"),longDateFormat:{LT:"h:mm A",LTS:"h:mm:ss A",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY h:mm A",LLLL:"dddd, D MMMM YYYY h:mm A"},calendar:{sameDay:"[Today at] LT",nextDay:"[Tomorrow at] LT",nextWeek:"dddd [at] LT",lastDay:"[Yesterday at] LT",lastWeek:"[Last] dddd [at] LT",sameElse:"L"},relativeTime:{future:"in %s",past:"%s ago",s:"a few seconds",ss:"%d seconds",m:"a minute",mm:"%d minutes",h:"an hour",hh:"%d hours",d:"a day",dd:"%d days",M:"a month",MM:"%d months",y:"a year",yy:"%d years"},dayOfMonthOrdinalParse:/\d{1,2}(st|nd|rd|th)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:0,doy:6}});return e}))},ec47:function(t,e,n){var i=n("a3fd"),r=n("42a2"),o=n("edfa"),a=n("cebd"),s="[object Map]",l="[object Set]";function u(t){return function(e){var n=r(e);return n==s?o(e):n==l?a(e):i(e,t(e))}}t.exports=u},ec69:function(t,e,n){var i=n("6fcd"),r=n("03dd"),o=n("30c9");function a(t){return o(t)?i(t):r(t)}t.exports=a},ec8c:function(t,e){function n(t){var e=[];if(null!=t)for(var n in Object(t))e.push(n);return e}t.exports=n},eda5:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("si",{months:"ජනවාරි_පෙබරවාරි_මාර්තු_අප්‍රේල්_මැයි_ජූනි_ජූලි_අගෝස්තු_සැප්තැම්බර්_ඔක්තෝබර්_නොවැම්බර්_දෙසැම්බර්".split("_"),monthsShort:"ජන_පෙබ_මාර්_අප්_මැයි_ජූනි_ජූලි_අගෝ_සැප්_ඔක්_නොවැ_දෙසැ".split("_"),weekdays:"ඉරිදා_සඳුදා_අඟහරුවාදා_බදාදා_බ්‍රහස්පතින්දා_සිකුරාදා_සෙනසුරාදා".split("_"),weekdaysShort:"ඉරි_සඳු_අඟ_බදා_බ්‍රහ_සිකු_සෙන".split("_"),weekdaysMin:"ඉ_ස_අ_බ_බ්‍ර_සි_සෙ".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"a h:mm",LTS:"a h:mm:ss",L:"YYYY/MM/DD",LL:"YYYY MMMM D",LLL:"YYYY MMMM D, a h:mm",LLLL:"YYYY MMMM D [වැනි] dddd, a h:mm:ss"},calendar:{sameDay:"[අද] LT[ට]",nextDay:"[හෙට] LT[ට]",nextWeek:"dddd LT[ට]",lastDay:"[ඊයේ] LT[ට]",lastWeek:"[පසුගිය] dddd LT[ට]",sameElse:"L"},relativeTime:{future:"%sකින්",past:"%sකට පෙර",s:"තත්පර කිහිපය",ss:"තත්පර %d",m:"මිනිත්තුව",mm:"මිනිත්තු %d",h:"පැය",hh:"පැය %d",d:"දිනය",dd:"දින %d",M:"මාසය",MM:"මාස %d",y:"වසර",yy:"වසර %d"},dayOfMonthOrdinalParse:/\d{1,2} වැනි/,ordinal:function(t){return t+" වැනි"},meridiemParse:/පෙර වරු|පස් වරු|පෙ.ව|ප.ව./,isPM:function(t){return"ප.ව."===t||"පස් වරු"===t},meridiem:function(t,e,n){return t>11?n?"ප.ව.":"පස් වරු":n?"පෙ.ව.":"පෙර වරු"}});return e}))},edfa:function(t,e){function n(t){var e=-1,n=Array(t.size);return t.forEach((function(t,i){n[++e]=[i,t]})),n}t.exports=n},ee45:function(t,e,n){var i=n("a78b"),r=n("1304");function o(t,e,n){return null==t?t:i(t,e,r(n))}t.exports=o},eec4:function(t,e,n){"use strict";var i=n("b770").forEach,r=n("5be5"),o=n("49ad"),a=n("2cef"),s=n("5058"),l=n("abb4"),u=n("18e9"),c=n("c274"),d=n("d6eb"),h=n("18d2"),f=n("c946");function p(t){return Array.isArray(t)||void 0!==t.length}function m(t){if(Array.isArray(t))return t;var e=[];return i(t,(function(t){e.push(t)})),e}function g(t){return t&&1===t.nodeType}function y(t,e,n){var i=t[e];return void 0!==i&&null!==i||void 0===n?i:n}t.exports=function(t){var e;if(t=t||{},t.idHandler)e={get:function(e){return t.idHandler.get(e,!0)},set:t.idHandler.set};else{var n=a(),v=s({idGenerator:n,stateHandler:d});e=v}var _=t.reporter;if(!_){var b=!1===_;_=l(b)}var x=y(t,"batchProcessor",c({reporter:_})),w={};w.callOnAdd=!!y(t,"callOnAdd",!0),w.debug=!!y(t,"debug",!1);var M,S=o(e),L=r({stateHandler:d}),k=y(t,"strategy","object"),T=y(t,"important",!1),D={reporter:_,batchProcessor:x,stateHandler:d,idHandler:e,important:T};if("scroll"===k&&(u.isLegacyOpera()?(_.warn("Scroll strategy is not supported on legacy Opera. Changing to object strategy."),k="object"):u.isIE(9)&&(_.warn("Scroll strategy is not supported on IE9. Changing to object strategy."),k="object")),"scroll"===k)M=f(D);else{if("object"!==k)throw new Error("Invalid strategy name: "+k);M=h(D)}var C={};function I(t,n,r){function o(t){var e=S.get(t);i(e,(function(e){e(t)}))}function a(t,e,n){S.add(e,n),t&&n(e)}if(r||(r=n,n=t,t={}),!n)throw new Error("At least one element required.");if(!r)throw new Error("Listener required.");if(g(n))n=[n];else{if(!p(n))return _.error("Invalid arguments. Must be a DOM element or a collection of DOM elements.");n=m(n)}var s=0,l=y(t,"callOnAdd",w.callOnAdd),u=y(t,"onReady",(function(){})),c=y(t,"debug",w.debug);i(n,(function(t){d.getState(t)||(d.initState(t),e.set(t));var h=e.get(t);if(c&&_.log("Attaching listener to element",h,t),!L.isDetectable(t))return c&&_.log(h,"Not detectable."),L.isBusy(t)?(c&&_.log(h,"System busy making it detectable"),a(l,t,r),C[h]=C[h]||[],void C[h].push((function(){s++,s===n.length&&u()}))):(c&&_.log(h,"Making detectable..."),L.markBusy(t,!0),M.makeDetectable({debug:c,important:T},t,(function(t){if(c&&_.log(h,"onElementDetectable"),d.getState(t)){L.markAsDetectable(t),L.markBusy(t,!1),M.addListener(t,o),a(l,t,r);var e=d.getState(t);if(e&&e.startSize){var f=t.offsetWidth,p=t.offsetHeight;e.startSize.width===f&&e.startSize.height===p||o(t)}C[h]&&i(C[h],(function(t){t()}))}else c&&_.log(h,"Element uninstalled before being detectable.");delete C[h],s++,s===n.length&&u()})));c&&_.log(h,"Already detecable, adding listener."),a(l,t,r),s++})),s===n.length&&u()}function A(t){if(!t)return _.error("At least one element is required.");if(g(t))t=[t];else{if(!p(t))return _.error("Invalid arguments. Must be a DOM element or a collection of DOM elements.");t=m(t)}i(t,(function(t){S.removeAllListeners(t),M.uninstall(t),d.cleanState(t)}))}function Y(t){M.initDocument&&M.initDocument(t)}return{listenTo:I,removeListener:S.removeListener,removeAllListeners:S.removeAllListeners,uninstall:A,initDocument:Y}}},ef5d:function(t,e){function n(t){return function(e){return null==e?void 0:e[t]}}t.exports=n},efb6:function(t,e,n){var i=n("5e2e");function r(){this.__data__=new i,this.size=0}t.exports=r},efc9:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("label",{class:t.wrapClasses},[n("span",{class:t.checkboxClasses},[n("span",{class:t.innerClasses}),t.group?n("input",{directives:[{name:"model",rawName:"v-model",value:t.model,expression:"model"}],class:t.inputClasses,attrs:{type:"checkbox",disabled:t.itemDisabled,name:t.name},domProps:{value:t.label,checked:Array.isArray(t.model)?t._i(t.model,t.label)>-1:t.model},on:{change:[function(e){var n=t.model,i=e.target,r=!!i.checked;if(Array.isArray(n)){var o=t.label,a=t._i(n,o);i.checked?a<0&&(t.model=n.concat([o])):a>-1&&(t.model=n.slice(0,a).concat(n.slice(a+1)))}else t.model=r},t.change],focus:t.onFocus,blur:t.onBlur}}):n("input",{class:t.inputClasses,attrs:{type:"checkbox",disabled:t.itemDisabled,name:t.name},domProps:{checked:t.currentValue},on:{change:t.change,focus:t.onFocus,blur:t.onBlur}})]),t._t("default",[t.showSlot?n("span",[t._v(t._s(t.label))]):t._e()])],2)},r=[],o=(n("a9e3"),n("ade3")),a=n("3a27"),s=n("2507"),l=n("9f81"),u="ivu-checkbox",c={name:"Checkbox",mixins:[s["a"],l["a"]],props:{disabled:{type:Boolean,default:!1},value:{type:[String,Number,Boolean],default:!1},trueValue:{type:[String,Number,Boolean],default:!0},falseValue:{type:[String,Number,Boolean],default:!1},label:{type:[String,Number,Boolean]},indeterminate:{type:Boolean,default:!1},size:{validator:function(t){return Object(a["i"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},name:{type:String},border:{type:Boolean,default:!1}},data:function(){return{model:[],currentValue:this.value,group:!1,showSlot:!0,parent:Object(a["e"])(this,"CheckboxGroup"),focusInner:!1}},computed:{wrapClasses:function(){var t;return["".concat(u,"-wrapper"),(t={},Object(o["a"])(t,"".concat(u,"-group-item"),this.group),Object(o["a"])(t,"".concat(u,"-wrapper-checked"),this.currentValue),Object(o["a"])(t,"".concat(u,"-wrapper-disabled"),this.itemDisabled),Object(o["a"])(t,"".concat(u,"-").concat(this.size),!!this.size),Object(o["a"])(t,"".concat(u,"-border"),this.border),t)]},checkboxClasses:function(){var t;return["".concat(u),(t={},Object(o["a"])(t,"".concat(u,"-checked"),this.currentValue),Object(o["a"])(t,"".concat(u,"-disabled"),this.itemDisabled),Object(o["a"])(t,"".concat(u,"-indeterminate"),this.indeterminate),t)]},innerClasses:function(){return["".concat(u,"-inner"),Object(o["a"])({},"".concat(u,"-focus"),this.focusInner)]},inputClasses:function(){return"".concat(u,"-input")}},mounted:function(){this.parent=Object(a["e"])(this,"CheckboxGroup"),this.parent&&(this.group=!0),this.group?this.parent.updateModel(!0):(this.updateModel(),this.showSlot=void 0!==this.$slots.default)},methods:{change:function(t){if(this.itemDisabled)return!1;var e=t.target.checked;this.currentValue=e;var n=e?this.trueValue:this.falseValue;this.$emit("input",n),this.group?this.parent.change(this.model):(this.$emit("on-change",n),this.dispatch("FormItem","on-form-change",n))},updateModel:function(){this.currentValue=this.value===this.trueValue},onBlur:function(){this.focusInner=!1},onFocus:function(){this.focusInner=!0}},watch:{value:function(t){if(t!==this.trueValue&&t!==this.falseValue)throw"Value should be trueValue or falseValue.";this.updateModel()}}},d=c,h=n("2877"),f=Object(h["a"])(d,i,r,!1,null,null,null),p=f.exports,m=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes},[t._t("default")],2)},g=[],y=(n("4160"),n("c975"),n("159b"),"ivu-checkbox-group"),v={name:"CheckboxGroup",mixins:[s["a"]],props:{value:{type:Array,default:function(){return[]}},size:{validator:function(t){return Object(a["i"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}}},data:function(){return{currentValue:this.value,childrens:[]}},computed:{classes:function(){return["".concat(y),Object(o["a"])({},"ivu-checkbox-".concat(this.size),!!this.size)]}},mounted:function(){this.updateModel(!0)},methods:{updateModel:function(t){if(this.childrens=Object(a["f"])(this,"Checkbox"),this.childrens){var e=this.value;this.childrens.forEach((function(n){n.model=e,t&&(n.currentValue=e.indexOf(n.label)>=0,n.group=!0)}))}},change:function(t){this.currentValue=t,this.$emit("input",t),this.$emit("on-change",t),this.dispatch("FormItem","on-form-change",t)}},watch:{value:function(){this.updateModel(!0)}}},_=v,b=Object(h["a"])(_,m,g,!1,null,null,null),x=b.exports;p.Group=x;e["a"]=p},f00c:function(t,e,n){var i=n("23e7"),r=n("e285");i({target:"Number",stat:!0},{isFinite:r})},f069:function(t,e,n){"use strict";var i=n("1c0b"),r=function(t){var e,n;this.promise=new t((function(t,i){if(void 0!==e||void 0!==n)throw TypeError("Bad Promise constructor");e=t,n=i})),this.resolve=i(e),this.reject=i(n)};t.exports.f=function(t){return new r(t)}},f260:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("pt",{months:"janeiro_fevereiro_março_abril_maio_junho_julho_agosto_setembro_outubro_novembro_dezembro".split("_"),monthsShort:"jan_fev_mar_abr_mai_jun_jul_ago_set_out_nov_dez".split("_"),weekdays:"Domingo_Segunda-feira_Terça-feira_Quarta-feira_Quinta-feira_Sexta-feira_Sábado".split("_"),weekdaysShort:"Dom_Seg_Ter_Qua_Qui_Sex_Sáb".split("_"),weekdaysMin:"Do_2ª_3ª_4ª_5ª_6ª_Sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D [de] MMMM [de] YYYY",LLL:"D [de] MMMM [de] YYYY HH:mm",LLLL:"dddd, D [de] MMMM [de] YYYY HH:mm"},calendar:{sameDay:"[Hoje às] LT",nextDay:"[Amanhã às] LT",nextWeek:"dddd [às] LT",lastDay:"[Ontem às] LT",lastWeek:function(){return 0===this.day()||6===this.day()?"[Último] dddd [às] LT":"[Última] dddd [às] LT"},sameElse:"L"},relativeTime:{future:"em %s",past:"há %s",s:"segundos",ss:"%d segundos",m:"um minuto",mm:"%d minutos",h:"uma hora",hh:"%d horas",d:"um dia",dd:"%d dias",w:"uma semana",ww:"%d semanas",M:"um mês",MM:"%d meses",y:"um ano",yy:"%d anos"},dayOfMonthOrdinalParse:/\d{1,2}º/,ordinal:"%dº",week:{dow:1,doy:4}});return e}))},f3c1:function(t,e){var n=800,i=16,r=Date.now;function o(t){var e=0,o=0;return function(){var a=r(),s=i-(a-o);if(o=a,s>0){if(++e>=n)return arguments[0]}else e=0;return t.apply(void 0,arguments)}}t.exports=o},f3ff:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e={1:"੧",2:"੨",3:"੩",4:"੪",5:"੫",6:"੬",7:"੭",8:"੮",9:"੯",0:"੦"},n={"੧":"1","੨":"2","੩":"3","੪":"4","੫":"5","੬":"6","੭":"7","੮":"8","੯":"9","੦":"0"},i=t.defineLocale("pa-in",{months:"ਜਨਵਰੀ_ਫ਼ਰਵਰੀ_ਮਾਰਚ_ਅਪ੍ਰੈਲ_ਮਈ_ਜੂਨ_ਜੁਲਾਈ_ਅਗਸਤ_ਸਤੰਬਰ_ਅਕਤੂਬਰ_ਨਵੰਬਰ_ਦਸੰਬਰ".split("_"),monthsShort:"ਜਨਵਰੀ_ਫ਼ਰਵਰੀ_ਮਾਰਚ_ਅਪ੍ਰੈਲ_ਮਈ_ਜੂਨ_ਜੁਲਾਈ_ਅਗਸਤ_ਸਤੰਬਰ_ਅਕਤੂਬਰ_ਨਵੰਬਰ_ਦਸੰਬਰ".split("_"),weekdays:"ਐਤਵਾਰ_ਸੋਮਵਾਰ_ਮੰਗਲਵਾਰ_ਬੁਧਵਾਰ_ਵੀਰਵਾਰ_ਸ਼ੁੱਕਰਵਾਰ_ਸ਼ਨੀਚਰਵਾਰ".split("_"),weekdaysShort:"ਐਤ_ਸੋਮ_ਮੰਗਲ_ਬੁਧ_ਵੀਰ_ਸ਼ੁਕਰ_ਸ਼ਨੀ".split("_"),weekdaysMin:"ਐਤ_ਸੋਮ_ਮੰਗਲ_ਬੁਧ_ਵੀਰ_ਸ਼ੁਕਰ_ਸ਼ਨੀ".split("_"),longDateFormat:{LT:"A h:mm ਵਜੇ",LTS:"A h:mm:ss ਵਜੇ",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY, A h:mm ਵਜੇ",LLLL:"dddd, D MMMM YYYY, A h:mm ਵਜੇ"},calendar:{sameDay:"[ਅਜ] LT",nextDay:"[ਕਲ] LT",nextWeek:"[ਅਗਲਾ] dddd, LT",lastDay:"[ਕਲ] LT",lastWeek:"[ਪਿਛਲੇ] dddd, LT",sameElse:"L"},relativeTime:{future:"%s ਵਿੱਚ",past:"%s ਪਿਛਲੇ",s:"ਕੁਝ ਸਕਿੰਟ",ss:"%d ਸਕਿੰਟ",m:"ਇਕ ਮਿੰਟ",mm:"%d ਮਿੰਟ",h:"ਇੱਕ ਘੰਟਾ",hh:"%d ਘੰਟੇ",d:"ਇੱਕ ਦਿਨ",dd:"%d ਦਿਨ",M:"ਇੱਕ ਮਹੀਨਾ",MM:"%d ਮਹੀਨੇ",y:"ਇੱਕ ਸਾਲ",yy:"%d ਸਾਲ"},preparse:function(t){return t.replace(/[੧੨੩੪੫੬੭੮੯੦]/g,(function(t){return n[t]}))},postformat:function(t){return t.replace(/\d/g,(function(t){return e[t]}))},meridiemParse:/ਰਾਤ|ਸਵੇਰ|ਦੁਪਹਿਰ|ਸ਼ਾਮ/,meridiemHour:function(t,e){return 12===t&&(t=0),"ਰਾਤ"===e?t<4?t:t+12:"ਸਵੇਰ"===e?t:"ਦੁਪਹਿਰ"===e?t>=10?t:t+12:"ਸ਼ਾਮ"===e?t+12:void 0},meridiem:function(t,e,n){return t<4?"ਰਾਤ":t<10?"ਸਵੇਰ":t<17?"ਦੁਪਹਿਰ":t<20?"ਸ਼ਾਮ":"ਰਾਤ"},week:{dow:0,doy:6}});return i}))},f4d6:function(t,e,n){var i=n("ffd6"),r=1/0;function o(t){if("string"==typeof t||i(t))return t;var e=t+"";return"0"==e&&1/t==-r?"-0":e}t.exports=o},f500:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes},[t._t("default")],2)},r=[],o=n("ade3"),a=n("3a27"),s="ivu-btn-group",l={name:"ButtonGroup",props:{size:{validator:function(t){return Object(a["i"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},shape:{validator:function(t){return Object(a["i"])(t,["circle","circle-outline"])}},vertical:{type:Boolean,default:!1}},computed:{classes:function(){var t;return["".concat(s),(t={},Object(o["a"])(t,"".concat(s,"-").concat(this.size),!!this.size),Object(o["a"])(t,"".concat(s,"-").concat(this.shape),!!this.shape),Object(o["a"])(t,"".concat(s,"-vertical"),this.vertical),t)]}}},u=l,c=n("2877"),d=Object(c["a"])(u,i,r,!1,null,null,null);e["a"]=d.exports},f542:function(t,e,n){var i=n("ec47"),r=n("ec69"),o=i(r);t.exports=o},f5df:function(t,e,n){var i=n("00ee"),r=n("c6b6"),o=n("b622"),a=o("toStringTag"),s="Arguments"==r(function(){return arguments}()),l=function(t,e){try{return t[e]}catch(n){}};t.exports=i?r:function(t){var e,n,i;return void 0===t?"Undefined":null===t?"Null":"string"==typeof(n=l(e=Object(t),a))?n:s?r(e):"Object"==(i=r(e))&&"function"==typeof e.callee?"Arguments":i}},f608:function(t,e,n){var i=n("6747"),r=n("ffd6"),o=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,a=/^\w*$/;function s(t,e){if(i(t))return!1;var n=typeof t;return!("number"!=n&&"symbol"!=n&&"boolean"!=n&&null!=t&&!r(t))||(a.test(t)||!o.test(t)||null!=e&&t in Object(e))}t.exports=s},f6b4:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=["Am Faoilleach","An Gearran","Am Màrt","An Giblean","An Cèitean","An t-Ògmhios","An t-Iuchar","An Lùnastal","An t-Sultain","An Dàmhair","An t-Samhain","An Dùbhlachd"],n=["Faoi","Gear","Màrt","Gibl","Cèit","Ògmh","Iuch","Lùn","Sult","Dàmh","Samh","Dùbh"],i=["Didòmhnaich","Diluain","Dimàirt","Diciadain","Diardaoin","Dihaoine","Disathairne"],r=["Did","Dil","Dim","Dic","Dia","Dih","Dis"],o=["Dò","Lu","Mà","Ci","Ar","Ha","Sa"],a=t.defineLocale("gd",{months:e,monthsShort:n,monthsParseExact:!0,weekdays:i,weekdaysShort:r,weekdaysMin:o,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[An-diugh aig] LT",nextDay:"[A-màireach aig] LT",nextWeek:"dddd [aig] LT",lastDay:"[An-dè aig] LT",lastWeek:"dddd [seo chaidh] [aig] LT",sameElse:"L"},relativeTime:{future:"ann an %s",past:"bho chionn %s",s:"beagan diogan",ss:"%d diogan",m:"mionaid",mm:"%d mionaidean",h:"uair",hh:"%d uairean",d:"latha",dd:"%d latha",M:"mìos",MM:"%d mìosan",y:"bliadhna",yy:"%d bliadhna"},dayOfMonthOrdinalParse:/\d{1,2}(d|na|mh)/,ordinal:function(t){var e=1===t?"d":t%10===2?"na":"mh";return t+e},week:{dow:1,doy:4}});return a}))},f6b49:function(t,e,n){"use strict";var i=n("c532");function r(){this.handlers=[]}r.prototype.use=function(t,e){return this.handlers.push({fulfilled:t,rejected:e}),this.handlers.length-1},r.prototype.eject=function(t){this.handlers[t]&&(this.handlers[t]=null)},r.prototype.forEach=function(t){i.forEach(this.handlers,(function(e){null!==e&&t(e)}))},t.exports=r},f772:function(t,e,n){var i=n("5692"),r=n("90e3"),o=i("keys");t.exports=function(t){return o[t]||(o[t]=r(t))}},f843:function(t,e,n){"use strict";var i=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{ref:"tabsWrap",class:t.classes},[n("div",{class:[t.prefixCls+"-bar"]},[t.showSlot?n("div",{class:[t.prefixCls+"-nav-right"]},[t._t("extra")],2):t._e(),n("div",{ref:"navContainer",class:[t.prefixCls+"-nav-container"],attrs:{tabindex:"0"},on:{keydown:[t.handleTabKeyNavigation,function(e){return!e.type.indexOf("key")&&t._k(e.keyCode,"space",32,e.key,[" ","Spacebar"])?null:(e.preventDefault(),t.handleTabKeyboardSelect(!1))}]}},[n("div",{ref:"navWrap",class:[t.prefixCls+"-nav-wrap",t.scrollable?t.prefixCls+"-nav-scrollable":""]},[n("span",{class:[t.prefixCls+"-nav-prev",t.scrollable?"":t.prefixCls+"-nav-scroll-disabled"],on:{click:t.scrollPrev}},[n("Icon",{attrs:{type:"ios-arrow-back"}})],1),n("span",{class:[t.prefixCls+"-nav-next",t.scrollable?"":t.prefixCls+"-nav-scroll-disabled"],on:{click:t.scrollNext}},[n("Icon",{attrs:{type:"ios-arrow-forward"}})],1),n("div",{ref:"navScroll",class:[t.prefixCls+"-nav-scroll"],on:{DOMMouseScroll:t.handleScroll,mousewheel:t.handleScroll}},[n("div",{ref:"nav",class:[t.prefixCls+"-nav"],style:t.navStyle},[n("div",{class:t.barClasses,style:t.barStyle}),t._l(t.navList,(function(e,i){return n("div",{class:t.tabCls(e),attrs:{draggable:t.draggable},on:{click:function(e){return t.handleChange(i)},dblclick:function(e){return t.handleDblclick(i)},contextmenu:function(e){return e.stopPropagation(),t.handleContextmenu(i,e)},selectstart:function(e){return e.stopPropagation(),t.handlePreventSelect(i,e)},dragstart:function(e){return t.handleDrag(i,e)},drop:function(e){return t.handleDrop(i,e)},dragover:function(t){t.preventDefault()}}},[""!==e.icon?n("Icon",{attrs:{type:e.icon}}):t._e(),"function"===e.labelType?n("Render",{attrs:{render:e.label}}):[t._v(t._s(e.label))],t.showClose(e)?n("Icon",{class:[t.prefixCls+"-close"],attrs:{type:t.arrowType,custom:t.customArrowType,size:t.arrowSize},nativeOn:{click:function(e){return e.stopPropagation(),t.handleRemove(i)}}}):t._e()],2)}))],2)])])])]),n("div",{ref:"panes",class:t.contentClasses,style:t.contentStyle},[t._t("default")],2),n("div",{staticClass:"ivu-tabs-context-menu",style:t.contextMenuStyles},[n("Dropdown",{attrs:{trigger:"custom",visible:t.contextMenuVisible,transfer:""},on:{"on-clickoutside":t.handleClickContextMenuOutside}},[n("DropdownMenu",{attrs:{slot:"list"},slot:"list"},[t._t("contextMenu")],2)],1)],1)])},r=[],o=(n("99af"),n("4de4"),n("c740"),n("4160"),n("d81d"),n("a434"),n("b0c0"),n("a9e3"),n("ac1f"),n("466d"),n("159b"),n("6b75"));function a(t){if(Array.isArray(t))return Object(o["a"])(t)}n("a4d3"),n("e01a"),n("d28b"),n("a630"),n("d3b7"),n("3ca3"),n("ddb0");function s(t){if("undefined"!==typeof Symbol&&Symbol.iterator in Object(t))return Array.from(t)}var l=n("06c5");function u(){throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}function c(t){return a(t)||s(t)||Object(l["a"])(t)||u()}var d=n("53ca"),h=n("ade3"),f=n("b85c"),p=n("099a"),m=n("1766"),g=n("6e29"),y=n("31de"),v=n("3a27"),_=n("2507"),b=n("eec4"),x=n.n(b),w="ivu-tabs",M=300,S=function t(e,n,i,r){var o=e.findIndex((function(t){return t.name===n})),a=(o+i+e.length)%e.length,s=e[a];return s.disabled?t(e,s.name,i,r):s},L=function t(e,n){try{e.focus()}catch(s){}if(document.activeElement==e&&e!==n)return!0;var i,r=e.children,o=Object(f["a"])(r);try{for(o.s();!(i=o.n()).done;){var a=i.value;if(t(a,n))return!0}}catch(s){o.e(s)}finally{o.f()}return!1},k={name:"Tabs",mixins:[_["a"]],components:{Icon:p["a"],Render:m["a"],Dropdown:g["a"],DropdownMenu:y["a"]},provide:function(){return{TabsInstance:this}},props:{value:{type:[String,Number]},type:{validator:function(t){return Object(v["i"])(t,["line","card"])},default:"line"},size:{validator:function(t){return Object(v["i"])(t,["small","default"])},default:"default"},animated:{type:Boolean,default:!0},captureFocus:{type:Boolean,default:!1},closable:{type:Boolean,default:!1},beforeRemove:Function,name:{type:String},draggable:{type:Boolean,default:!1}},data:function(){return{prefixCls:w,navList:[],barWidth:0,barOffset:0,activeKey:this.value,focusedKey:this.value,showSlot:!1,navStyle:{transform:""},scrollable:!1,transitioning:!1,contextMenuVisible:!1,contextMenuStyles:{top:0,left:0}}},computed:{classes:function(){var t;return["".concat(w),(t={},Object(h["a"])(t,"".concat(w,"-card"),"card"===this.type),Object(h["a"])(t,"".concat(w,"-mini"),"small"===this.size&&"line"===this.type),Object(h["a"])(t,"".concat(w,"-no-animation"),!this.animated),t)]},contentClasses:function(){return["".concat(w,"-content"),Object(h["a"])({},"".concat(w,"-content-animated"),this.animated)]},barClasses:function(){return["".concat(w,"-ink-bar"),Object(h["a"])({},"".concat(w,"-ink-bar-animated"),this.animated)]},contentStyle:function(){var t=this.getTabIndex(this.activeKey),e=0===t?"0%":"-".concat(t,"00%"),n={};return t>-1&&(n={transform:"translateX(".concat(e,") translateZ(0px)")}),n},barStyle:function(){var t={visibility:"hidden",width:"".concat(this.barWidth,"px")};return"line"===this.type&&(t.visibility="visible"),this.animated?t.transform="translate3d(".concat(this.barOffset,"px, 0px, 0px)"):t.left="".concat(this.barOffset,"px"),t},arrowType:function(){var t="ios-close";return this.$IVIEW&&(this.$IVIEW.tabs.customCloseIcon?t="":this.$IVIEW.tabs.closeIcon&&(t=this.$IVIEW.tabs.closeIcon)),t},customArrowType:function(){var t="";return this.$IVIEW&&this.$IVIEW.tabs.customCloseIcon&&(t=this.$IVIEW.tabs.customCloseIcon),t},arrowSize:function(){var t="";return this.$IVIEW&&this.$IVIEW.tabs.closeIconSize&&(t=this.$IVIEW.tabs.closeIconSize),t}},methods:{getTabs:function(){var t=this,e=Object(v["f"])(this,"TabPane"),n=[];return e.forEach((function(e){e.tab&&t.name?e.tab===t.name&&n.push(e):n.push(e)})),n.sort((function(t,e){if(t.index&&e.index)return t.index>e.index?1:-1})),n},updateNav:function(){var t=this;this.navList=[],this.getTabs().forEach((function(e,n){t.navList.push({labelType:Object(d["a"])(e.label),label:e.label,icon:e.icon||"",name:e.currentName||n,disabled:e.disabled,closable:e.closable,contextMenu:e.contextMenu}),e.currentName||(e.currentName=n),0===n&&(t.activeKey||(t.activeKey=e.currentName||n))})),this.updateStatus(),this.updateBar()},updateBar:function(){var t=this;this.$nextTick((function(){var e=t.getTabIndex(t.activeKey);if(t.$refs.nav){var n=t.$refs.nav.querySelectorAll(".".concat(w,"-tab")),i=n[e];if(t.barWidth=i?parseFloat(i.offsetWidth):0,e>0){for(var r=0,o="small"===t.size?0:16,a=0;a<e;a++)r+=parseFloat(n[a].offsetWidth)+o;t.barOffset=r}else t.barOffset=0;t.updateNavScroll()}}))},updateStatus:function(){var t=this,e=this.getTabs();e.forEach((function(e){return e.show=e.currentName===t.activeKey||t.animated}))},tabCls:function(t){var e;return["".concat(w,"-tab"),(e={},Object(h["a"])(e,"".concat(w,"-tab-disabled"),t.disabled),Object(h["a"])(e,"".concat(w,"-tab-active"),t.name===this.activeKey),Object(h["a"])(e,"".concat(w,"-tab-focused"),t.name===this.focusedKey),e)]},handleChange:function(t){var e=this;if(!this.transitioning){this.transitioning=!0,setTimeout((function(){return e.transitioning=!1}),M);var n=this.navList[t];n&&!n.disabled&&(this.activeKey=n.name,this.$emit("input",n.name),this.$emit("on-click",n.name))}},handleDblclick:function(t){var e=this.navList[t];e&&!e.disabled&&this.$emit("on-dblclick",e.name)},handleContextmenu:function(t,e){var n=this;this.contextMenuVisible&&this.handleClickContextMenuOutside(),this.$nextTick((function(){var i=n.navList[t];if(i&&!i.disabled&&i.contextMenu){e.preventDefault();var r=n.$refs.tabsWrap,o=r.getBoundingClientRect(),a={left:"".concat(e.clientX-o.left,"px"),top:"".concat(e.clientY-o.top,"px")};n.contextMenuStyles=a,n.contextMenuVisible=!0,n.$emit("on-contextmenu",i,e,a)}}))},handleClickContextMenuOutside:function(){this.contextMenuVisible=!1},handlePreventSelect:function(t,e){var n=this.navList[t];n&&!n.disabled&&n.contextMenu&&e.preventDefault()},handleTabKeyNavigation:function(t){if(37===t.keyCode||39===t.keyCode){var e=39===t.keyCode?1:-1,n=S(this.navList,this.focusedKey,e);this.focusedKey=n.name}},handleTabKeyboardSelect:function(){var t=arguments.length>0&&void 0!==arguments[0]&&arguments[0];if(!t){var e=this.focusedKey||0,n=this.getTabIndex(e);this.handleChange(n)}},handleRemove:function(t){var e=this;if(!this.beforeRemove)return this.handleRemoveTab(t);var n=this.beforeRemove(t);n&&n.then?n.then((function(){e.handleRemoveTab(t)})):this.handleRemoveTab(t)},handleRemoveTab:function(t){var e=this.getTabs(),n=e[t];if(n.$destroy(),n.currentName===this.activeKey){var i=this.getTabs(),r=-1;if(i.length){var o=e.filter((function(e,n){return!e.disabled&&n<t})),a=e.filter((function(e,n){return!e.disabled&&n>t}));r=a.length?a[0].currentName:o.length?o[o.length-1].currentName:i[0].currentName}this.activeKey=r,this.$emit("input",r)}this.$emit("on-tab-remove",n.currentName),this.updateNav()},showClose:function(t){return"card"===this.type&&(null!==t.closable?t.closable:this.closable)},scrollPrev:function(){var t=this.$refs.navScroll.offsetWidth,e=this.getCurrentScrollOffset();if(e){var n=e>t?e-t:0;this.setOffset(n)}},scrollNext:function(){var t=this.$refs.nav.offsetWidth,e=this.$refs.navScroll.offsetWidth,n=this.getCurrentScrollOffset();if(!(t-n<=e)){var i=t-n>2*e?n+e:t-e;this.setOffset(i)}},getCurrentScrollOffset:function(){var t=this.navStyle;return t.transform?Number(t.transform.match(/translateX\(-(\d+(\.\d+)*)px\)/)[1]):0},getTabIndex:function(t){return this.navList.findIndex((function(e){return e.name===t}))},setOffset:function(t){this.navStyle.transform="translateX(-".concat(t,"px)")},scrollToActiveTab:function(){if(this.scrollable){var t=this.$refs.nav,e=this.$el.querySelector(".".concat(w,"-tab-active"));if(e){var n=this.$refs.navScroll,i=e.getBoundingClientRect(),r=n.getBoundingClientRect(),o=t.getBoundingClientRect(),a=this.getCurrentScrollOffset(),s=a;o.right<r.right&&(s=t.offsetWidth-r.width),i.left<r.left?s=a-(r.left-i.left):i.right>r.right&&(s=a+i.right-r.right),a!==s&&this.setOffset(Math.max(s,0))}}},updateNavScroll:function(){var t=this.$refs.nav.offsetWidth,e=this.$refs.navScroll.offsetWidth,n=this.getCurrentScrollOffset();e<t?(this.scrollable=!0,t-n<e&&this.setOffset(t-e)):(this.scrollable=!1,n>0&&this.setOffset(0))},handleScroll:function(t){t.preventDefault(),t.stopPropagation();var e=t.type,n=0;"DOMMouseScroll"!==e&&"mousewheel"!==e||(n=t.wheelDelta?t.wheelDelta:40*-(t.detail||0)),n>0?this.scrollPrev():this.scrollNext()},handleResize:function(){this.updateNavScroll()},isInsideHiddenElement:function(){var t=this.$el.parentNode;while(t&&t!==document.body){if(t.style&&"none"===t.style.display)return t;t=t.parentNode}return!1},updateVisibility:function(t){var e=this;c(this.$refs.panes.querySelectorAll(".".concat(w,"-tabpane"))).forEach((function(n,i){t===i?(c(n.children).filter((function(t){return t.classList.contains("".concat(w,"-tabpane"))})).forEach((function(t){return t.style.visibility="visible"})),e.captureFocus&&setTimeout((function(){return L(n,n)}),M)):setTimeout((function(){c(n.children).filter((function(t){return t.classList.contains("".concat(w,"-tabpane"))})).forEach((function(t){return t.style.visibility="hidden"}))}),M)}))},handleDrag:function(t,e){var n=this.navList[t];n&&e.dataTransfer.setData("tab-name",n.name)},handleDrop:function(t,e){var n=this.navList[t];if(n){var i=e.dataTransfer.getData("tab-name");e.preventDefault();var r=this.navList.map((function(t){return t.name})),o=parseInt(r.findIndex((function(t){return t===i}))),a=parseInt(r.findIndex((function(t){return t===n.name})));r.splice.apply(r,[a,1].concat(c(r.splice(o,1,r[a])))),this.$emit("on-drag-drop",i,n.name,o,a,r)}}},watch:{value:function(t){this.activeKey=t,this.focusedKey=t},activeKey:function(t){var e=this;this.focusedKey=t,this.updateBar(),this.updateStatus(),this.broadcast("Table","on-visible-change",!0),this.$nextTick((function(){e.scrollToActiveTab()}));var n=Math.max(this.getTabIndex(this.focusedKey),0);this.updateVisibility(n)}},mounted:function(){var t=this;this.showSlot=void 0!==this.$slots.extra,this.observer=x()(),this.observer.listenTo(this.$refs.navWrap,this.handleResize);var e=this.isInsideHiddenElement();e&&(this.mutationObserver=new v["a"]((function(){"none"!==e.style.display&&(t.updateBar(),t.mutationObserver.disconnect())})),this.mutationObserver.observe(e,{attributes:!0,childList:!0,characterData:!0,attributeFilter:["style"]})),this.handleTabKeyboardSelect(!0),this.updateVisibility(this.getTabIndex(this.activeKey))},beforeDestroy:function(){this.observer.removeListener(this.$refs.navWrap,this.handleResize),this.mutationObserver&&this.mutationObserver.disconnect()}},T=k,D=n("2877"),C=Object(D["a"])(T,i,r,!1,null,null,null),I=C.exports,A=n("5334");I.Pane=A["a"];e["a"]=I},f8af:function(t,e,n){var i=n("2474");function r(t){var e=new t.constructor(t.byteLength);return new i(e).set(new i(t)),e}t.exports=r},f909:function(t,e,n){var i=n("7e64"),r=n("b760"),o=n("72af"),a=n("4f50"),s=n("1a8c"),l=n("9934"),u=n("8adb");function c(t,e,n,d,h){t!==e&&o(e,(function(o,l){if(h||(h=new i),s(o))a(t,e,l,n,c,d,h);else{var f=d?d(u(t,l),o,l+"",t,e,h):void 0;void 0===f&&(f=o),r(t,l,f)}}),l)}t.exports=c},f9ce:function(t,e,n){var i=n("ef5d"),r=n("e3f8"),o=n("f608"),a=n("f4d6");function s(t){return o(t)?i(a(t)):r(t)}t.exports=s},fa21:function(t,e,n){var i=n("7530"),r=n("2dcb"),o=n("eac5");function a(t){return"function"!=typeof t.constructor||o(t)?{}:i(r(t))}t.exports=a},facd:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e="jan._feb._mrt._apr._mei_jun._jul._aug._sep._okt._nov._dec.".split("_"),n="jan_feb_mrt_apr_mei_jun_jul_aug_sep_okt_nov_dec".split("_"),i=[/^jan/i,/^feb/i,/^maart|mrt.?$/i,/^apr/i,/^mei$/i,/^jun[i.]?$/i,/^jul[i.]?$/i,/^aug/i,/^sep/i,/^okt/i,/^nov/i,/^dec/i],r=/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december|jan\.?|feb\.?|mrt\.?|apr\.?|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,o=t.defineLocale("nl",{months:"januari_februari_maart_april_mei_juni_juli_augustus_september_oktober_november_december".split("_"),monthsShort:function(t,i){return t?/-MMM-/.test(i)?n[t.month()]:e[t.month()]:e},monthsRegex:r,monthsShortRegex:r,monthsStrictRegex:/^(januari|februari|maart|april|mei|ju[nl]i|augustus|september|oktober|november|december)/i,monthsShortStrictRegex:/^(jan\.?|feb\.?|mrt\.?|apr\.?|mei|ju[nl]\.?|aug\.?|sep\.?|okt\.?|nov\.?|dec\.?)/i,monthsParse:i,longMonthsParse:i,shortMonthsParse:i,weekdays:"zondag_maandag_dinsdag_woensdag_donderdag_vrijdag_zaterdag".split("_"),weekdaysShort:"zo._ma._di._wo._do._vr._za.".split("_"),weekdaysMin:"zo_ma_di_wo_do_vr_za".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD-MM-YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd D MMMM YYYY HH:mm"},calendar:{sameDay:"[vandaag om] LT",nextDay:"[morgen om] LT",nextWeek:"dddd [om] LT",lastDay:"[gisteren om] LT",lastWeek:"[afgelopen] dddd [om] LT",sameElse:"L"},relativeTime:{future:"over %s",past:"%s geleden",s:"een paar seconden",ss:"%d seconden",m:"één minuut",mm:"%d minuten",h:"één uur",hh:"%d uur",d:"één dag",dd:"%d dagen",w:"één week",ww:"%d weken",M:"één maand",MM:"%d maanden",y:"één jaar",yy:"%d jaar"},dayOfMonthOrdinalParse:/\d{1,2}(ste|de)/,ordinal:function(t){return t+(1===t||8===t||t>=20?"ste":"de")},week:{dow:1,doy:4}});return o}))},fb25:function(t,e,n){var i=n("afb9"),r=n("ec69");function o(t){return null==t?[]:i(t,r(t))}t.exports=o},fb6a:function(t,e,n){"use strict";var i=n("23e7"),r=n("861d"),o=n("e8b5"),a=n("23cb"),s=n("50c4"),l=n("fc6a"),u=n("8418"),c=n("b622"),d=n("1dde"),h=n("ae40"),f=d("slice"),p=h("slice",{ACCESSORS:!0,0:0,1:2}),m=c("species"),g=[].slice,y=Math.max;i({target:"Array",proto:!0,forced:!f||!p},{slice:function(t,e){var n,i,c,d=l(this),h=s(d.length),f=a(t,h),p=a(void 0===e?h:e,h);if(o(d)&&(n=d.constructor,"function"!=typeof n||n!==Array&&!o(n.prototype)?r(n)&&(n=n[m],null===n&&(n=void 0)):n=void 0,n===Array||void 0===n))return g.call(d,f,p);for(i=new(void 0===n?Array:n)(y(p-f,0)),c=0;f<p;f++,c++)f in d&&u(i,c,d[f]);return i.length=c,i}})},fba5:function(t,e,n){var i=n("cb5a");function r(t){return i(this.__data__,t)>-1}t.exports=r},fc6a:function(t,e,n){var i=n("44ad"),r=n("1d80");t.exports=function(t){return i(r(t))}},fd7e:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("x-pseudo",{months:"J~áñúá~rý_F~ébrú~árý_~Márc~h_Áp~ríl_~Máý_~Júñé~_Júl~ý_Áú~gúst~_Sép~témb~ér_Ó~ctób~ér_Ñ~óvém~bér_~Décé~mbér".split("_"),monthsShort:"J~áñ_~Féb_~Már_~Ápr_~Máý_~Júñ_~Júl_~Áúg_~Sép_~Óct_~Ñóv_~Déc".split("_"),monthsParseExact:!0,weekdays:"S~úñdá~ý_Mó~ñdáý~_Túé~sdáý~_Wéd~ñésd~áý_T~húrs~dáý_~Fríd~áý_S~átúr~dáý".split("_"),weekdaysShort:"S~úñ_~Móñ_~Túé_~Wéd_~Thú_~Frí_~Sát".split("_"),weekdaysMin:"S~ú_Mó~_Tú_~Wé_T~h_Fr~_Sá".split("_"),weekdaysParseExact:!0,longDateFormat:{LT:"HH:mm",L:"DD/MM/YYYY",LL:"D MMMM YYYY",LLL:"D MMMM YYYY HH:mm",LLLL:"dddd, D MMMM YYYY HH:mm"},calendar:{sameDay:"[T~ódá~ý át] LT",nextDay:"[T~ómó~rró~w át] LT",nextWeek:"dddd [át] LT",lastDay:"[Ý~ést~érdá~ý át] LT",lastWeek:"[L~ást] dddd [át] LT",sameElse:"L"},relativeTime:{future:"í~ñ %s",past:"%s á~gó",s:"á ~féw ~sécó~ñds",ss:"%d s~écóñ~ds",m:"á ~míñ~úté",mm:"%d m~íñú~tés",h:"á~ñ hó~úr",hh:"%d h~óúrs",d:"á ~dáý",dd:"%d d~áýs",M:"á ~móñ~th",MM:"%d m~óñt~hs",y:"á ~ýéár",yy:"%d ý~éárs"},dayOfMonthOrdinalParse:/\d{1,2}(th|st|nd|rd)/,ordinal:function(t){var e=t%10,n=1===~~(t%100/10)?"th":1===e?"st":2===e?"nd":3===e?"rd":"th";return t+n},week:{dow:1,doy:4}});return e}))},fdbc:function(t,e){t.exports={CSSRuleList:0,CSSStyleDeclaration:0,CSSValueList:0,ClientRectList:0,DOMRectList:0,DOMStringList:0,DOMTokenList:1,DataTransferItemList:0,FileList:0,HTMLAllCollection:0,HTMLCollection:0,HTMLFormElement:0,HTMLSelectElement:0,MediaList:0,MimeTypeArray:0,NamedNodeMap:0,NodeList:1,PaintRequestList:0,Plugin:0,PluginArray:0,SVGLengthList:0,SVGNumberList:0,SVGPathSegList:0,SVGPointList:0,SVGStringList:0,SVGTransformList:0,SourceBufferList:0,StyleSheetList:0,TextTrackCueList:0,TextTrackList:0,TouchList:0}},fdbf:function(t,e,n){var i=n("4930");t.exports=i&&!Symbol.sham&&"symbol"==typeof Symbol.iterator},fe66:function(t,e,n){"use strict";var i=n("2b0e"),r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("transition",{attrs:{name:"fade"}},[t.fullscreenVisible?n("div",{class:t.classes},[n("div",{class:t.mainClasses},[n("span",{class:t.dotClasses}),n("div",{class:t.textClasses},[t._t("default")],2)])]):t._e()])},o=[],a=n("ade3"),s=n("3a27"),l=n("b2e6"),u="ivu-spin",c={name:"Spin",mixins:[l["a"]],props:{size:{validator:function(t){return Object(s["i"])(t,["small","large","default"])},default:function(){return this.$IVIEW&&""!==this.$IVIEW.size?this.$IVIEW.size:"default"}},fix:{type:Boolean,default:!1},fullscreen:{type:Boolean,default:!1}},data:function(){return{showText:!1,visible:!1}},computed:{classes:function(){var t;return["".concat(u),(t={},Object(a["a"])(t,"".concat(u,"-").concat(this.size),!!this.size),Object(a["a"])(t,"".concat(u,"-fix"),this.fix),Object(a["a"])(t,"".concat(u,"-show-text"),this.showText),Object(a["a"])(t,"".concat(u,"-fullscreen"),this.fullscreen),t)]},mainClasses:function(){return"".concat(u,"-main")},dotClasses:function(){return"".concat(u,"-dot")},textClasses:function(){return"".concat(u,"-text")},fullscreenVisible:function(){return!this.fullscreen||this.visible}},watch:{visible:function(t){t?this.addScrollEffect():this.removeScrollEffect()}},mounted:function(){this.showText=void 0!==this.$slots.default}},d=c,h=n("2877"),f=Object(h["a"])(d,r,o,!1,null,null,null),p=f.exports,m=n("c177");function g(){return Object(m["a"])(),m["b"]}let y=g();p.newInstance=t=>{const e=t||{},n=new i["a"]({data:Object.assign({},e,{}),render(t){let e="";return e=this.render?t(p,{props:{fix:!0,fullscreen:!0}},[this.render(t)]):t(p,{props:{size:"large",fix:!0,fullscreen:!0}}),t("div",{class:"ivu-spin-fullscreen ivu-spin-fullscreen-wrapper",style:{"z-index":2010+y}},[e])}}),r=n.$mount();document.body.appendChild(r.$el);const o=n.$children[0];return{show(){o.visible=!0,y=g()},remove(t){o.visible=!1,setTimeout((function(){o.$parent.$destroy(),void 0!==document.getElementsByClassName("ivu-spin-fullscreen")[0]&&document.body.removeChild(document.getElementsByClassName("ivu-spin-fullscreen")[0]),t()}),500)},component:o}};var v=p;let _;function b(t){return _=_||v.newInstance({render:t}),_}function x(t){const e="render"in t?t.render:void 0;let n=b(e);n.show(t)}v.show=function(t={}){return x(t)},v.hide=function(){if(!_)return!1;const t=b();t.remove(()=>{_=null})};e["a"]=v},fea9:function(t,e,n){var i=n("da84");t.exports=i.Promise},ffd6:function(t,e,n){var i=n("3729"),r=n("1310"),o="[object Symbol]";function a(t){return"symbol"==typeof t||r(t)&&i(t)==o}t.exports=a},ffff:function(t,e,n){(function(t,e){e(n("c1df"))})(0,(function(t){"use strict";
//! moment.js locale configuration
var e=t.defineLocale("se",{months:"ođđajagemánnu_guovvamánnu_njukčamánnu_cuoŋománnu_miessemánnu_geassemánnu_suoidnemánnu_borgemánnu_čakčamánnu_golggotmánnu_skábmamánnu_juovlamánnu".split("_"),monthsShort:"ođđj_guov_njuk_cuo_mies_geas_suoi_borg_čakč_golg_skáb_juov".split("_"),weekdays:"sotnabeaivi_vuossárga_maŋŋebárga_gaskavahkku_duorastat_bearjadat_lávvardat".split("_"),weekdaysShort:"sotn_vuos_maŋ_gask_duor_bear_láv".split("_"),weekdaysMin:"s_v_m_g_d_b_L".split("_"),longDateFormat:{LT:"HH:mm",LTS:"HH:mm:ss",L:"DD.MM.YYYY",LL:"MMMM D. [b.] YYYY",LLL:"MMMM D. [b.] YYYY [ti.] HH:mm",LLLL:"dddd, MMMM D. [b.] YYYY [ti.] HH:mm"},calendar:{sameDay:"[otne ti] LT",nextDay:"[ihttin ti] LT",nextWeek:"dddd [ti] LT",lastDay:"[ikte ti] LT",lastWeek:"[ovddit] dddd [ti] LT",sameElse:"L"},relativeTime:{future:"%s geažes",past:"maŋit %s",s:"moadde sekunddat",ss:"%d sekunddat",m:"okta minuhta",mm:"%d minuhtat",h:"okta diimmu",hh:"%d diimmut",d:"okta beaivi",dd:"%d beaivvit",M:"okta mánnu",MM:"%d mánut",y:"okta jahki",yy:"%d jagit"},dayOfMonthOrdinalParse:/\d{1,2}\./,ordinal:"%d.",week:{dow:1,doy:4}});return e}))}}]);